'use strict';

var substanceTest = require('substance-test');

/*
  Implements Substance ChangeStore API. This is just a dumb store.
  No integrity checks are made, as this is the task of DocumentEngine
*/
class ChangeStore {
  constructor(seed) {
    this._changes = seed || {};
  }

  /*
    Gets changes for a given document

    @param {String} documentId document id
    @param {Number} sinceVersion since which change (optional)
    @param {Number} toVersion up to and including version (optional)
  */
  getChanges(documentId, sinceVersion, toVersion, cb) {
    if (typeof sinceVersion === 'function') {
      cb = sinceVersion;
      sinceVersion = 0;
    } else if (typeof toVersion === 'function') {
      cb = toVersion;
      toVersion = undefined;
    }
    if (!(documentId && sinceVersion >= 0 && cb)) {
      throw new Error('Invalid arguments')
    }
    let version = this._getVersion(documentId);
    let changes = this._getChanges(documentId);
    changes = changes.slice(sinceVersion, toVersion);
    cb(null, changes, version);
  }

  /*
    Add a change object to the database
  */
  addChange(documentId, change, cb) {
    if (!documentId || !change) {
      throw new Error('Invalid arguments')
    }
    this._addChange(documentId, change);
    let newVersion = this._getVersion(documentId);
    cb(null, newVersion);
  }

  /*
    Delete changes for a given documentId
  */
  deleteChanges(documentId, cb) {
    var deletedChanges = this._deleteChanges(documentId);
    cb(null, deletedChanges.length);
  }

  /*
    Gets the version number for a document
  */
  getVersion(id, cb) {
    cb(null, this._getVersion(id));
  }

  // Handy synchronous helpers
  // -------------------------

  _deleteChanges(documentId) {
    var changes = this._getChanges(documentId);
    delete this._changes[documentId];
    return changes
  }

  _getVersion(documentId) {
    var changes = this._changes[documentId];
    return changes ? changes.length : 0
  }

  _getChanges(documentId) {
    return this._changes[documentId] || []
  }

  _addChange(documentId, change) {
    if (!this._changes[documentId]) {
      this._changes[documentId] = [];
    }
    this._changes[documentId].push(change);
  }
}

/*
  Run the functions in the tasks collection in series.

  NOTE: You can not access results of the executed functions
*/
function series(tasks, cb, i) {
  i = i || 0;
  tasks[i](function(err) {
    // Always stop execution on error
    if (err) return cb(err)
    if (i === tasks.length-1) {
      cb(...arguments); // we are done
    } else {
      series(tasks, cb, i + 1);
    }
  });
}

var async = Object.freeze({
	series: series
});

/*
  An iterator for arrays.

  @class
  @param {Array} arr
 */
class ArrayIterator {

  constructor(arr) {
    this.arr = arr;
    this.pos = -1;
  }

  get _isArrayIterator() {
    return true
  }

  /**
    @returns {Boolean} true if there is another child node left.
   */
  hasNext() {
    return this.pos < this.arr.length - 1
  }

  /**
    Increments the iterator providing the next child node.

    @returns {HTMLElement} The next child node.
   */
  next() {
    this.pos += 1;
    var next = this.arr[this.pos];
    return next
  }

  /**
    Decrements the iterator.
   */
  back() {
    if (this.pos >= 0) {
      this.pos -= 1;
    }
    return this
  }

  peek() {
    return this.arr[this.pos+1]
  }

}

function deleteFromArray(array, value) {
  if (!array) return
  for (var i = 0; i < array.length; i++) {
    if (array[i] === value) {
      array.splice(i, 1);
      i--;
    }
  }
}

// simplified version of TreeIndex for arrays
class ArrayTree {
  add(path, val) {
    if (!this[path]) {
      this[path] = [];
    }
    this[path].push(val);
  }
  remove(path, val) {
    if (this[path]) {
      deleteFromArray(this[path], val);
    }
  }
}

function checkGlobal(value) {
  return (value && value.Object === Object) ? value : null;
}

var freeGlobal = checkGlobal(typeof global == 'object' && global);


var freeSelf = checkGlobal(typeof self == 'object' && self);


var thisGlobal = checkGlobal(typeof undefined == 'object' && undefined);


var root = freeGlobal || freeSelf || thisGlobal || Function('return this')();

var Symbol = root.Symbol;

function isObjectLike(value) {
  return !!value && typeof value == 'object';
}

var symbolTag = '[object Symbol]';


var objectProto = Object.prototype;


var objectToString = objectProto.toString;


function isSymbol(value) {
  return typeof value == 'symbol' ||
    (isObjectLike(value) && objectToString.call(value) == symbolTag);
}

var INFINITY = 1 / 0;


var symbolProto = Symbol ? Symbol.prototype : undefined;
var symbolToString = symbolProto ? symbolProto.toString : undefined;


function baseToString(value) {
  
  if (typeof value == 'string') {
    return value;
  }
  if (isSymbol(value)) {
    return symbolToString ? symbolToString.call(value) : '';
  }
  var result = (value + '');
  return (result == '0' && (1 / value) == -INFINITY) ? '-0' : result;
}

function toString(value) {
  return value == null ? '' : baseToString(value);
}

function baseSlice(array, start, end) {
  var index = -1,
      length = array.length;

  if (start < 0) {
    start = -start > length ? 0 : (length + start);
  }
  end = end > length ? length : end;
  if (end < 0) {
    end += length;
  }
  length = start > end ? 0 : ((end - start) >>> 0);
  start >>>= 0;

  var result = Array(length);
  while (++index < length) {
    result[index] = array[index + start];
  }
  return result;
}

function castSlice(array, start, end) {
  var length = array.length;
  end = end === undefined ? length : end;
  return (!start && end >= length) ? array : baseSlice(array, start, end);
}

var rsAstralRange = '\\ud800-\\udfff';
var rsComboMarksRange = '\\u0300-\\u036f\\ufe20-\\ufe23';
var rsComboSymbolsRange = '\\u20d0-\\u20f0';
var rsVarRange = '\\ufe0e\\ufe0f';


var rsZWJ = '\\u200d';


var reHasComplexSymbol = RegExp('[' + rsZWJ + rsAstralRange  + rsComboMarksRange + rsComboSymbolsRange + rsVarRange + ']');

var rsAstralRange$1 = '\\ud800-\\udfff';
var rsComboMarksRange$1 = '\\u0300-\\u036f\\ufe20-\\ufe23';
var rsComboSymbolsRange$1 = '\\u20d0-\\u20f0';
var rsVarRange$1 = '\\ufe0e\\ufe0f';


var rsAstral = '[' + rsAstralRange$1 + ']';
var rsCombo = '[' + rsComboMarksRange$1 + rsComboSymbolsRange$1 + ']';
var rsFitz = '\\ud83c[\\udffb-\\udfff]';
var rsModifier = '(?:' + rsCombo + '|' + rsFitz + ')';
var rsNonAstral = '[^' + rsAstralRange$1 + ']';
var rsRegional = '(?:\\ud83c[\\udde6-\\uddff]){2}';
var rsSurrPair = '[\\ud800-\\udbff][\\udc00-\\udfff]';
var rsZWJ$1 = '\\u200d';


var reOptMod = rsModifier + '?';
var rsOptVar = '[' + rsVarRange$1 + ']?';
var rsOptJoin = '(?:' + rsZWJ$1 + '(?:' + [rsNonAstral, rsRegional, rsSurrPair].join('|') + ')' + rsOptVar + reOptMod + ')*';
var rsSeq = rsOptVar + reOptMod + rsOptJoin;
var rsSymbol = '(?:' + [rsNonAstral + rsCombo + '?', rsCombo, rsRegional, rsSurrPair, rsAstral].join('|') + ')';


var reComplexSymbol = RegExp(rsFitz + '(?=' + rsFitz + ')|' + rsSymbol + rsSeq, 'g');


function stringToArray(string) {
  return string.match(reComplexSymbol);
}

function createCaseFirst(methodName) {
  return function(string) {
    string = toString(string);

    var strSymbols = reHasComplexSymbol.test(string)
      ? stringToArray(string)
      : undefined;

    var chr = strSymbols
      ? strSymbols[0]
      : string.charAt(0);

    var trailing = strSymbols
      ? castSlice(strSymbols, 1).join('')
      : string.slice(1);

    return chr[methodName]() + trailing;
  };
}

var upperFirst = createCaseFirst('toUpperCase');

function capitalize(string) {
  return upperFirst(toString(string).toLowerCase());
}

function isObject(value) {
  var type = typeof value;
  return !!value && (type == 'object' || type == 'function');
}

var funcTag = '[object Function]';
var genTag = '[object GeneratorFunction]';


var objectProto$1 = Object.prototype;


var objectToString$1 = objectProto$1.toString;


function isFunction(value) {
  
  
  
  var tag = isObject(value) ? objectToString$1.call(value) : '';
  return tag == funcTag || tag == genTag;
}

var NAN = 0 / 0;


var reTrim = /^\s+|\s+$/g;


var reIsBadHex = /^[-+]0x[0-9a-f]+$/i;


var reIsBinary = /^0b[01]+$/i;


var reIsOctal = /^0o[0-7]+$/i;


var freeParseInt = parseInt;


function toNumber(value) {
  if (typeof value == 'number') {
    return value;
  }
  if (isSymbol(value)) {
    return NAN;
  }
  if (isObject(value)) {
    var other = isFunction(value.valueOf) ? value.valueOf() : value;
    value = isObject(other) ? (other + '') : other;
  }
  if (typeof value != 'string') {
    return value === 0 ? value : +value;
  }
  value = value.replace(reTrim, '');
  var isBinary = reIsBinary.test(value);
  return (isBinary || reIsOctal.test(value))
    ? freeParseInt(value.slice(2), isBinary ? 2 : 8)
    : (reIsBadHex.test(value) ? NAN : +value);
}

var isArray = Array.isArray;

function isHostObject(value) {
  
  
  var result = false;
  if (value != null && typeof value.toString != 'function') {
    try {
      result = !!(value + '');
    } catch (e) {}
  }
  return result;
}

var coreJsData = root['__core-js_shared__'];

var maskSrcKey = (function() {
  var uid = /[^.]+$/.exec(coreJsData && coreJsData.keys && coreJsData.keys.IE_PROTO || '');
  return uid ? ('Symbol(src)_1.' + uid) : '';
}());


function isMasked(func) {
  return !!maskSrcKey && (maskSrcKey in func);
}

var funcToString$1 = Function.prototype.toString;


function toSource(func) {
  if (func != null) {
    try {
      return funcToString$1.call(func);
    } catch (e) {}
    try {
      return (func + '');
    } catch (e) {}
  }
  return '';
}

var reRegExpChar = /[\\^$.*+?()[\]{}|]/g;


var reIsHostCtor = /^\[object .+?Constructor\]$/;


var objectProto$2 = Object.prototype;


var funcToString = Function.prototype.toString;


var hasOwnProperty = objectProto$2.hasOwnProperty;


var reIsNative = RegExp('^' +
  funcToString.call(hasOwnProperty).replace(reRegExpChar, '\\$&')
  .replace(/hasOwnProperty|(function).*?(?=\\\()| for .+?(?=\\\])/g, '$1.*?') + '$'
);


function baseIsNative(value) {
  if (!isObject(value) || isMasked(value)) {
    return false;
  }
  var pattern = (isFunction(value) || isHostObject(value)) ? reIsNative : reIsHostCtor;
  return pattern.test(toSource(value));
}

function getValue(object, key) {
  return object == null ? undefined : object[key];
}

function getNative(object, key) {
  var value = getValue(object, key);
  return baseIsNative(value) ? value : undefined;
}

var nativeCreate = getNative(Object, 'create');

function hashClear() {
  this.__data__ = nativeCreate ? nativeCreate(null) : {};
}

function hashDelete(key) {
  return this.has(key) && delete this.__data__[key];
}

var HASH_UNDEFINED = '__lodash_hash_undefined__';


var objectProto$3 = Object.prototype;


var hasOwnProperty$1 = objectProto$3.hasOwnProperty;


function hashGet(key) {
  var data = this.__data__;
  if (nativeCreate) {
    var result = data[key];
    return result === HASH_UNDEFINED ? undefined : result;
  }
  return hasOwnProperty$1.call(data, key) ? data[key] : undefined;
}

var objectProto$4 = Object.prototype;


var hasOwnProperty$2 = objectProto$4.hasOwnProperty;


function hashHas(key) {
  var data = this.__data__;
  return nativeCreate ? data[key] !== undefined : hasOwnProperty$2.call(data, key);
}

var HASH_UNDEFINED$1 = '__lodash_hash_undefined__';


function hashSet(key, value) {
  var data = this.__data__;
  data[key] = (nativeCreate && value === undefined) ? HASH_UNDEFINED$1 : value;
  return this;
}

function Hash(entries) {
  var index = -1,
      length = entries ? entries.length : 0;

  this.clear();
  while (++index < length) {
    var entry = entries[index];
    this.set(entry[0], entry[1]);
  }
}


Hash.prototype.clear = hashClear;
Hash.prototype['delete'] = hashDelete;
Hash.prototype.get = hashGet;
Hash.prototype.has = hashHas;
Hash.prototype.set = hashSet;

function listCacheClear() {
  this.__data__ = [];
}

function eq(value, other) {
  return value === other || (value !== value && other !== other);
}

function assocIndexOf(array, key) {
  var length = array.length;
  while (length--) {
    if (eq(array[length][0], key)) {
      return length;
    }
  }
  return -1;
}

var arrayProto = Array.prototype;


var splice = arrayProto.splice;


function listCacheDelete(key) {
  var data = this.__data__,
      index = assocIndexOf(data, key);

  if (index < 0) {
    return false;
  }
  var lastIndex = data.length - 1;
  if (index == lastIndex) {
    data.pop();
  } else {
    splice.call(data, index, 1);
  }
  return true;
}

function listCacheGet(key) {
  var data = this.__data__,
      index = assocIndexOf(data, key);

  return index < 0 ? undefined : data[index][1];
}

function listCacheHas(key) {
  return assocIndexOf(this.__data__, key) > -1;
}

function listCacheSet(key, value) {
  var data = this.__data__,
      index = assocIndexOf(data, key);

  if (index < 0) {
    data.push([key, value]);
  } else {
    data[index][1] = value;
  }
  return this;
}

function ListCache(entries) {
  var index = -1,
      length = entries ? entries.length : 0;

  this.clear();
  while (++index < length) {
    var entry = entries[index];
    this.set(entry[0], entry[1]);
  }
}


ListCache.prototype.clear = listCacheClear;
ListCache.prototype['delete'] = listCacheDelete;
ListCache.prototype.get = listCacheGet;
ListCache.prototype.has = listCacheHas;
ListCache.prototype.set = listCacheSet;

var Map$1 = getNative(root, 'Map');

function mapCacheClear() {
  this.__data__ = {
    'hash': new Hash,
    'map': new (Map$1 || ListCache),
    'string': new Hash
  };
}

function isKeyable(value) {
  var type = typeof value;
  return (type == 'string' || type == 'number' || type == 'symbol' || type == 'boolean')
    ? (value !== '__proto__')
    : (value === null);
}

function getMapData(map, key) {
  var data = map.__data__;
  return isKeyable(key)
    ? data[typeof key == 'string' ? 'string' : 'hash']
    : data.map;
}

function mapCacheDelete(key) {
  return getMapData(this, key)['delete'](key);
}

function mapCacheGet(key) {
  return getMapData(this, key).get(key);
}

function mapCacheHas(key) {
  return getMapData(this, key).has(key);
}

function mapCacheSet(key, value) {
  getMapData(this, key).set(key, value);
  return this;
}

function MapCache(entries) {
  var index = -1,
      length = entries ? entries.length : 0;

  this.clear();
  while (++index < length) {
    var entry = entries[index];
    this.set(entry[0], entry[1]);
  }
}


MapCache.prototype.clear = mapCacheClear;
MapCache.prototype['delete'] = mapCacheDelete;
MapCache.prototype.get = mapCacheGet;
MapCache.prototype.has = mapCacheHas;
MapCache.prototype.set = mapCacheSet;

var FUNC_ERROR_TEXT$1 = 'Expected a function';


function memoize(func, resolver) {
  if (typeof func != 'function' || (resolver && typeof resolver != 'function')) {
    throw new TypeError(FUNC_ERROR_TEXT$1);
  }
  var memoized = function() {
    var args = arguments,
        key = resolver ? resolver.apply(this, args) : args[0],
        cache = memoized.cache;

    if (cache.has(key)) {
      return cache.get(key);
    }
    var result = func.apply(this, args);
    memoized.cache = cache.set(key, result);
    return result;
  };
  memoized.cache = new (memoize.Cache || MapCache);
  return memoized;
}


memoize.Cache = MapCache;

var rePropName = /[^.[\]]+|\[(?:(-?\d+(?:\.\d+)?)|(["'])((?:(?!\2)[^\\]|\\.)*?)\2)\]|(?=(\.|\[\])(?:\4|$))/g;


var reEscapeChar = /\\(\\)?/g;


var stringToPath = memoize(function(string) {
  var result = [];
  toString(string).replace(rePropName, function(match, number, quote, string) {
    result.push(quote ? string.replace(reEscapeChar, '$1') : (number || match));
  });
  return result;
});

function castPath(value) {
  return isArray(value) ? value : stringToPath(value);
}

var reIsDeepProp = /\.|\[(?:[^[\]]*|(["'])(?:(?!\1)[^\\]|\\.)*?\1)\]/;
var reIsPlainProp = /^\w*$/;


function isKey(value, object) {
  if (isArray(value)) {
    return false;
  }
  var type = typeof value;
  if (type == 'number' || type == 'symbol' || type == 'boolean' ||
      value == null || isSymbol(value)) {
    return true;
  }
  return reIsPlainProp.test(value) || !reIsDeepProp.test(value) ||
    (object != null && value in Object(object));
}

var INFINITY$1 = 1 / 0;


function toKey(value) {
  if (typeof value == 'string' || isSymbol(value)) {
    return value;
  }
  var result = (value + '');
  return (result == '0' && (1 / value) == -INFINITY$1) ? '-0' : result;
}

function baseGet(object, path) {
  path = isKey(path, object) ? [path] : castPath(path);

  var index = 0,
      length = path.length;

  while (object != null && index < length) {
    object = object[toKey(path[index++])];
  }
  return (index && index == length) ? object : undefined;
}

function get(object, path, defaultValue) {
  var result = object == null ? undefined : baseGet(object, path);
  return result === undefined ? defaultValue : result;
}

function stackClear() {
  this.__data__ = new ListCache;
}

function stackDelete(key) {
  return this.__data__['delete'](key);
}

function stackGet(key) {
  return this.__data__.get(key);
}

function stackHas(key) {
  return this.__data__.has(key);
}

var LARGE_ARRAY_SIZE = 200;


function stackSet(key, value) {
  var cache = this.__data__;
  if (cache instanceof ListCache && cache.__data__.length == LARGE_ARRAY_SIZE) {
    cache = this.__data__ = new MapCache(cache.__data__);
  }
  cache.set(key, value);
  return this;
}

function Stack(entries) {
  this.__data__ = new ListCache(entries);
}


Stack.prototype.clear = stackClear;
Stack.prototype['delete'] = stackDelete;
Stack.prototype.get = stackGet;
Stack.prototype.has = stackHas;
Stack.prototype.set = stackSet;

var HASH_UNDEFINED$2 = '__lodash_hash_undefined__';


function setCacheAdd(value) {
  this.__data__.set(value, HASH_UNDEFINED$2);
  return this;
}

function setCacheHas(value) {
  return this.__data__.has(value);
}

function SetCache(values) {
  var index = -1,
      length = values ? values.length : 0;

  this.__data__ = new MapCache;
  while (++index < length) {
    this.add(values[index]);
  }
}


SetCache.prototype.add = SetCache.prototype.push = setCacheAdd;
SetCache.prototype.has = setCacheHas;

function arraySome(array, predicate) {
  var index = -1,
      length = array ? array.length : 0;

  while (++index < length) {
    if (predicate(array[index], index, array)) {
      return true;
    }
  }
  return false;
}

var UNORDERED_COMPARE_FLAG$1 = 1;
var PARTIAL_COMPARE_FLAG$2 = 2;


function equalArrays(array, other, equalFunc, customizer, bitmask, stack) {
  var isPartial = bitmask & PARTIAL_COMPARE_FLAG$2,
      arrLength = array.length,
      othLength = other.length;

  if (arrLength != othLength && !(isPartial && othLength > arrLength)) {
    return false;
  }
  
  var stacked = stack.get(array);
  if (stacked) {
    return stacked == other;
  }
  var index = -1,
      result = true,
      seen = (bitmask & UNORDERED_COMPARE_FLAG$1) ? new SetCache : undefined;

  stack.set(array, other);

  
  while (++index < arrLength) {
    var arrValue = array[index],
        othValue = other[index];

    if (customizer) {
      var compared = isPartial
        ? customizer(othValue, arrValue, index, other, array, stack)
        : customizer(arrValue, othValue, index, array, other, stack);
    }
    if (compared !== undefined) {
      if (compared) {
        continue;
      }
      result = false;
      break;
    }
    
    if (seen) {
      if (!arraySome(other, function(othValue, othIndex) {
            if (!seen.has(othIndex) &&
                (arrValue === othValue || equalFunc(arrValue, othValue, customizer, bitmask, stack))) {
              return seen.add(othIndex);
            }
          })) {
        result = false;
        break;
      }
    } else if (!(
          arrValue === othValue ||
            equalFunc(arrValue, othValue, customizer, bitmask, stack)
        )) {
      result = false;
      break;
    }
  }
  stack['delete'](array);
  return result;
}

var Uint8Array = root.Uint8Array;

function mapToArray(map) {
  var index = -1,
      result = Array(map.size);

  map.forEach(function(value, key) {
    result[++index] = [key, value];
  });
  return result;
}

function setToArray(set) {
  var index = -1,
      result = Array(set.size);

  set.forEach(function(value) {
    result[++index] = value;
  });
  return result;
}

var UNORDERED_COMPARE_FLAG$2 = 1;
var PARTIAL_COMPARE_FLAG$3 = 2;


var boolTag = '[object Boolean]';
var dateTag = '[object Date]';
var errorTag = '[object Error]';
var mapTag = '[object Map]';
var numberTag = '[object Number]';
var regexpTag = '[object RegExp]';
var setTag = '[object Set]';
var stringTag = '[object String]';
var symbolTag$1 = '[object Symbol]';

var arrayBufferTag = '[object ArrayBuffer]';
var dataViewTag = '[object DataView]';


var symbolProto$1 = Symbol ? Symbol.prototype : undefined;
var symbolValueOf = symbolProto$1 ? symbolProto$1.valueOf : undefined;


function equalByTag(object, other, tag, equalFunc, customizer, bitmask, stack) {
  switch (tag) {
    case dataViewTag:
      if ((object.byteLength != other.byteLength) ||
          (object.byteOffset != other.byteOffset)) {
        return false;
      }
      object = object.buffer;
      other = other.buffer;

    case arrayBufferTag:
      if ((object.byteLength != other.byteLength) ||
          !equalFunc(new Uint8Array(object), new Uint8Array(other))) {
        return false;
      }
      return true;

    case boolTag:
    case dateTag:
      
      
      
      return +object == +other;

    case errorTag:
      return object.name == other.name && object.message == other.message;

    case numberTag:
      
      return (object != +object) ? other != +other : object == +other;

    case regexpTag:
    case stringTag:
      
      
      
      return object == (other + '');

    case mapTag:
      var convert = mapToArray;

    case setTag:
      var isPartial = bitmask & PARTIAL_COMPARE_FLAG$3;
      convert || (convert = setToArray);

      if (object.size != other.size && !isPartial) {
        return false;
      }
      
      var stacked = stack.get(object);
      if (stacked) {
        return stacked == other;
      }
      bitmask |= UNORDERED_COMPARE_FLAG$2;
      stack.set(object, other);

      
      return equalArrays(convert(object), convert(other), equalFunc, customizer, bitmask, stack);

    case symbolTag$1:
      if (symbolValueOf) {
        return symbolValueOf.call(object) == symbolValueOf.call(other);
      }
  }
  return false;
}

var nativeGetPrototype = Object.getPrototypeOf;


function getPrototype(value) {
  return nativeGetPrototype(Object(value));
}

var objectProto$6 = Object.prototype;


var hasOwnProperty$4 = objectProto$6.hasOwnProperty;


function baseHas(object, key) {
  
  
  
  return object != null &&
    (hasOwnProperty$4.call(object, key) ||
      (typeof object == 'object' && key in object && getPrototype(object) === null));
}

var nativeKeys = Object.keys;


function baseKeys(object) {
  return nativeKeys(Object(object));
}

function baseTimes(n, iteratee) {
  var index = -1,
      result = Array(n);

  while (++index < n) {
    result[index] = iteratee(index);
  }
  return result;
}

function baseProperty(key) {
  return function(object) {
    return object == null ? undefined : object[key];
  };
}

var getLength = baseProperty('length');

var MAX_SAFE_INTEGER = 9007199254740991;


function isLength(value) {
  return typeof value == 'number' &&
    value > -1 && value % 1 == 0 && value <= MAX_SAFE_INTEGER;
}

function isArrayLike(value) {
  return value != null && isLength(getLength(value)) && !isFunction(value);
}

function isArrayLikeObject(value) {
  return isObjectLike(value) && isArrayLike(value);
}

var argsTag$1 = '[object Arguments]';


var objectProto$7 = Object.prototype;


var hasOwnProperty$5 = objectProto$7.hasOwnProperty;


var objectToString$2 = objectProto$7.toString;


var propertyIsEnumerable = objectProto$7.propertyIsEnumerable;


function isArguments(value) {
  
  return isArrayLikeObject(value) && hasOwnProperty$5.call(value, 'callee') &&
    (!propertyIsEnumerable.call(value, 'callee') || objectToString$2.call(value) == argsTag$1);
}

var stringTag$1 = '[object String]';


var objectProto$8 = Object.prototype;


var objectToString$3 = objectProto$8.toString;


function isString(value) {
  return typeof value == 'string' ||
    (!isArray(value) && isObjectLike(value) && objectToString$3.call(value) == stringTag$1);
}

function indexKeys(object) {
  var length = object ? object.length : undefined;
  if (isLength(length) &&
      (isArray(object) || isString(object) || isArguments(object))) {
    return baseTimes(length, String);
  }
  return null;
}

var MAX_SAFE_INTEGER$1 = 9007199254740991;


var reIsUint = /^(?:0|[1-9]\d*)$/;


function isIndex(value, length) {
  length = length == null ? MAX_SAFE_INTEGER$1 : length;
  return !!length &&
    (typeof value == 'number' || reIsUint.test(value)) &&
    (value > -1 && value % 1 == 0 && value < length);
}

var objectProto$9 = Object.prototype;


function isPrototype(value) {
  var Ctor = value && value.constructor,
      proto = (typeof Ctor == 'function' && Ctor.prototype) || objectProto$9;

  return value === proto;
}

function keys(object) {
  var isProto = isPrototype(object);
  if (!(isProto || isArrayLike(object))) {
    return baseKeys(object);
  }
  var indexes = indexKeys(object),
      skipIndexes = !!indexes,
      result = indexes || [],
      length = result.length;

  for (var key in object) {
    if (baseHas(object, key) &&
        !(skipIndexes && (key == 'length' || isIndex(key, length))) &&
        !(isProto && key == 'constructor')) {
      result.push(key);
    }
  }
  return result;
}

var PARTIAL_COMPARE_FLAG$4 = 2;


function equalObjects(object, other, equalFunc, customizer, bitmask, stack) {
  var isPartial = bitmask & PARTIAL_COMPARE_FLAG$4,
      objProps = keys(object),
      objLength = objProps.length,
      othProps = keys(other),
      othLength = othProps.length;

  if (objLength != othLength && !isPartial) {
    return false;
  }
  var index = objLength;
  while (index--) {
    var key = objProps[index];
    if (!(isPartial ? key in other : baseHas(other, key))) {
      return false;
    }
  }
  
  var stacked = stack.get(object);
  if (stacked) {
    return stacked == other;
  }
  var result = true;
  stack.set(object, other);

  var skipCtor = isPartial;
  while (++index < objLength) {
    key = objProps[index];
    var objValue = object[key],
        othValue = other[key];

    if (customizer) {
      var compared = isPartial
        ? customizer(othValue, objValue, key, other, object, stack)
        : customizer(objValue, othValue, key, object, other, stack);
    }
    
    if (!(compared === undefined
          ? (objValue === othValue || equalFunc(objValue, othValue, customizer, bitmask, stack))
          : compared
        )) {
      result = false;
      break;
    }
    skipCtor || (skipCtor = key == 'constructor');
  }
  if (result && !skipCtor) {
    var objCtor = object.constructor,
        othCtor = other.constructor;

    
    if (objCtor != othCtor &&
        ('constructor' in object && 'constructor' in other) &&
        !(typeof objCtor == 'function' && objCtor instanceof objCtor &&
          typeof othCtor == 'function' && othCtor instanceof othCtor)) {
      result = false;
    }
  }
  stack['delete'](object);
  return result;
}

var DataView = getNative(root, 'DataView');

var Promise$1 = getNative(root, 'Promise');

var Set$1 = getNative(root, 'Set');

var WeakMap$1 = getNative(root, 'WeakMap');

var mapTag$1 = '[object Map]';
var objectTag$1 = '[object Object]';
var promiseTag = '[object Promise]';
var setTag$1 = '[object Set]';
var weakMapTag = '[object WeakMap]';

var dataViewTag$1 = '[object DataView]';


var objectProto$10 = Object.prototype;


var objectToString$4 = objectProto$10.toString;


var dataViewCtorString = toSource(DataView);
var mapCtorString = toSource(Map$1);
var promiseCtorString = toSource(Promise$1);
var setCtorString = toSource(Set$1);
var weakMapCtorString = toSource(WeakMap$1);


function getTag(value) {
  return objectToString$4.call(value);
}



if ((DataView && getTag(new DataView(new ArrayBuffer(1))) != dataViewTag$1) ||
    (Map$1 && getTag(new Map$1) != mapTag$1) ||
    (Promise$1 && getTag(Promise$1.resolve()) != promiseTag) ||
    (Set$1 && getTag(new Set$1) != setTag$1) ||
    (WeakMap$1 && getTag(new WeakMap$1) != weakMapTag)) {
  getTag = function(value) {
    var result = objectToString$4.call(value),
        Ctor = result == objectTag$1 ? value.constructor : undefined,
        ctorString = Ctor ? toSource(Ctor) : undefined;

    if (ctorString) {
      switch (ctorString) {
        case dataViewCtorString: return dataViewTag$1;
        case mapCtorString: return mapTag$1;
        case promiseCtorString: return promiseTag;
        case setCtorString: return setTag$1;
        case weakMapCtorString: return weakMapTag;
      }
    }
    return result;
  };
}

var getTag$1 = getTag;

var argsTag$2 = '[object Arguments]';
var arrayTag$1 = '[object Array]';
var boolTag$1 = '[object Boolean]';
var dateTag$1 = '[object Date]';
var errorTag$1 = '[object Error]';
var funcTag$1 = '[object Function]';
var mapTag$2 = '[object Map]';
var numberTag$1 = '[object Number]';
var objectTag$2 = '[object Object]';
var regexpTag$1 = '[object RegExp]';
var setTag$2 = '[object Set]';
var stringTag$2 = '[object String]';
var weakMapTag$1 = '[object WeakMap]';

var arrayBufferTag$1 = '[object ArrayBuffer]';
var dataViewTag$2 = '[object DataView]';
var float32Tag = '[object Float32Array]';
var float64Tag = '[object Float64Array]';
var int8Tag = '[object Int8Array]';
var int16Tag = '[object Int16Array]';
var int32Tag = '[object Int32Array]';
var uint8Tag = '[object Uint8Array]';
var uint8ClampedTag = '[object Uint8ClampedArray]';
var uint16Tag = '[object Uint16Array]';
var uint32Tag = '[object Uint32Array]';


var typedArrayTags = {};
typedArrayTags[float32Tag] = typedArrayTags[float64Tag] =
typedArrayTags[int8Tag] = typedArrayTags[int16Tag] =
typedArrayTags[int32Tag] = typedArrayTags[uint8Tag] =
typedArrayTags[uint8ClampedTag] = typedArrayTags[uint16Tag] =
typedArrayTags[uint32Tag] = true;
typedArrayTags[argsTag$2] = typedArrayTags[arrayTag$1] =
typedArrayTags[arrayBufferTag$1] = typedArrayTags[boolTag$1] =
typedArrayTags[dataViewTag$2] = typedArrayTags[dateTag$1] =
typedArrayTags[errorTag$1] = typedArrayTags[funcTag$1] =
typedArrayTags[mapTag$2] = typedArrayTags[numberTag$1] =
typedArrayTags[objectTag$2] = typedArrayTags[regexpTag$1] =
typedArrayTags[setTag$2] = typedArrayTags[stringTag$2] =
typedArrayTags[weakMapTag$1] = false;


var objectProto$11 = Object.prototype;


var objectToString$5 = objectProto$11.toString;


function isTypedArray(value) {
  return isObjectLike(value) &&
    isLength(value.length) && !!typedArrayTags[objectToString$5.call(value)];
}

var PARTIAL_COMPARE_FLAG$1 = 2;


var argsTag = '[object Arguments]';
var arrayTag = '[object Array]';
var objectTag = '[object Object]';


var objectProto$5 = Object.prototype;


var hasOwnProperty$3 = objectProto$5.hasOwnProperty;


function baseIsEqualDeep(object, other, equalFunc, customizer, bitmask, stack) {
  var objIsArr = isArray(object),
      othIsArr = isArray(other),
      objTag = arrayTag,
      othTag = arrayTag;

  if (!objIsArr) {
    objTag = getTag$1(object);
    objTag = objTag == argsTag ? objectTag : objTag;
  }
  if (!othIsArr) {
    othTag = getTag$1(other);
    othTag = othTag == argsTag ? objectTag : othTag;
  }
  var objIsObj = objTag == objectTag && !isHostObject(object),
      othIsObj = othTag == objectTag && !isHostObject(other),
      isSameTag = objTag == othTag;

  if (isSameTag && !objIsObj) {
    stack || (stack = new Stack);
    return (objIsArr || isTypedArray(object))
      ? equalArrays(object, other, equalFunc, customizer, bitmask, stack)
      : equalByTag(object, other, objTag, equalFunc, customizer, bitmask, stack);
  }
  if (!(bitmask & PARTIAL_COMPARE_FLAG$1)) {
    var objIsWrapped = objIsObj && hasOwnProperty$3.call(object, '__wrapped__'),
        othIsWrapped = othIsObj && hasOwnProperty$3.call(other, '__wrapped__');

    if (objIsWrapped || othIsWrapped) {
      var objUnwrapped = objIsWrapped ? object.value() : object,
          othUnwrapped = othIsWrapped ? other.value() : other;

      stack || (stack = new Stack);
      return equalFunc(objUnwrapped, othUnwrapped, customizer, bitmask, stack);
    }
  }
  if (!isSameTag) {
    return false;
  }
  stack || (stack = new Stack);
  return equalObjects(object, other, equalFunc, customizer, bitmask, stack);
}

function baseIsEqual(value, other, customizer, bitmask, stack) {
  if (value === other) {
    return true;
  }
  if (value == null || other == null || (!isObject(value) && !isObjectLike(other))) {
    return value !== value && other !== other;
  }
  return baseIsEqualDeep(value, other, baseIsEqual, customizer, bitmask, stack);
}

var UNORDERED_COMPARE_FLAG = 1;
var PARTIAL_COMPARE_FLAG = 2;


function baseIsMatch(object, source, matchData, customizer) {
  var index = matchData.length,
      length = index,
      noCustomizer = !customizer;

  if (object == null) {
    return !length;
  }
  object = Object(object);
  while (index--) {
    var data = matchData[index];
    if ((noCustomizer && data[2])
          ? data[1] !== object[data[0]]
          : !(data[0] in object)
        ) {
      return false;
    }
  }
  while (++index < length) {
    data = matchData[index];
    var key = data[0],
        objValue = object[key],
        srcValue = data[1];

    if (noCustomizer && data[2]) {
      if (objValue === undefined && !(key in object)) {
        return false;
      }
    } else {
      var stack = new Stack;
      if (customizer) {
        var result = customizer(objValue, srcValue, key, object, source, stack);
      }
      if (!(result === undefined
            ? baseIsEqual(srcValue, objValue, customizer, UNORDERED_COMPARE_FLAG | PARTIAL_COMPARE_FLAG, stack)
            : result
          )) {
        return false;
      }
    }
  }
  return true;
}

function isStrictComparable(value) {
  return value === value && !isObject(value);
}

function getMatchData(object) {
  var result = keys(object),
      length = result.length;

  while (length--) {
    var key = result[length],
        value = object[key];

    result[length] = [key, value, isStrictComparable(value)];
  }
  return result;
}

function matchesStrictComparable(key, srcValue) {
  return function(object) {
    if (object == null) {
      return false;
    }
    return object[key] === srcValue &&
      (srcValue !== undefined || (key in Object(object)));
  };
}

function baseMatches(source) {
  var matchData = getMatchData(source);
  if (matchData.length == 1 && matchData[0][2]) {
    return matchesStrictComparable(matchData[0][0], matchData[0][1]);
  }
  return function(object) {
    return object === source || baseIsMatch(object, source, matchData);
  };
}

function baseHasIn(object, key) {
  return object != null && key in Object(object);
}

function hasPath(object, path, hasFunc) {
  path = isKey(path, object) ? [path] : castPath(path);

  var result,
      index = -1,
      length = path.length;

  while (++index < length) {
    var key = toKey(path[index]);
    if (!(result = object != null && hasFunc(object, key))) {
      break;
    }
    object = object[key];
  }
  if (result) {
    return result;
  }
  var length = object ? object.length : 0;
  return !!length && isLength(length) && isIndex(key, length) &&
    (isArray(object) || isString(object) || isArguments(object));
}

function hasIn(object, path) {
  return object != null && hasPath(object, path, baseHasIn);
}

var UNORDERED_COMPARE_FLAG$3 = 1;
var PARTIAL_COMPARE_FLAG$5 = 2;


function baseMatchesProperty(path, srcValue) {
  if (isKey(path) && isStrictComparable(srcValue)) {
    return matchesStrictComparable(toKey(path), srcValue);
  }
  return function(object) {
    var objValue = get(object, path);
    return (objValue === undefined && objValue === srcValue)
      ? hasIn(object, path)
      : baseIsEqual(srcValue, objValue, undefined, UNORDERED_COMPARE_FLAG$3 | PARTIAL_COMPARE_FLAG$5);
  };
}

function identity(value) {
  return value;
}

function basePropertyDeep(path) {
  return function(object) {
    return baseGet(object, path);
  };
}

function property(path) {
  return isKey(path) ? baseProperty(toKey(path)) : basePropertyDeep(path);
}

function baseIteratee(value) {
  
  
  if (typeof value == 'function') {
    return value;
  }
  if (value == null) {
    return identity;
  }
  if (typeof value == 'object') {
    return isArray(value)
      ? baseMatchesProperty(value[0], value[1])
      : baseMatches(value);
  }
  return property(value);
}

function createFind(findIndexFunc) {
  return function(collection, predicate, fromIndex) {
    var iterable = Object(collection);
    predicate = baseIteratee(predicate, 3);
    if (!isArrayLike(collection)) {
      var props = keys(collection);
    }
    var index = findIndexFunc(props || collection, function(value, key) {
      if (props) {
        key = value;
        value = iterable[key];
      }
      return predicate(value, key, iterable);
    }, fromIndex);
    return index > -1 ? collection[props ? props[index] : index] : undefined;
  };
}

function baseFindIndex(array, predicate, fromIndex, fromRight) {
  var length = array.length,
      index = fromIndex + (fromRight ? 1 : -1);

  while ((fromRight ? index-- : ++index < length)) {
    if (predicate(array[index], index, array)) {
      return index;
    }
  }
  return -1;
}

var INFINITY$2 = 1 / 0;
var MAX_INTEGER = 1.7976931348623157e+308;


function toFinite(value) {
  if (!value) {
    return value === 0 ? value : 0;
  }
  value = toNumber(value);
  if (value === INFINITY$2 || value === -INFINITY$2) {
    var sign = (value < 0 ? -1 : 1);
    return sign * MAX_INTEGER;
  }
  return value === value ? value : 0;
}

function toInteger(value) {
  var result = toFinite(value),
      remainder = result % 1;

  return result === result ? (remainder ? result - remainder : result) : 0;
}

var nativeMax$1 = Math.max;


function findIndex(array, predicate, fromIndex) {
  var length = array ? array.length : 0;
  if (!length) {
    return -1;
  }
  var index = fromIndex == null ? 0 : toInteger(fromIndex);
  if (index < 0) {
    index = nativeMax$1(length + index, 0);
  }
  return baseFindIndex(array, baseIteratee(predicate, 3), index);
}

var find = createFind(findIndex);

function isMatch(object, source) {
  return object === source || baseIsMatch(object, source, getMatchData(source));
}

function arrayEach(array, iteratee) {
  var index = -1,
      length = array ? array.length : 0;

  while (++index < length) {
    if (iteratee(array[index], index, array) === false) {
      break;
    }
  }
  return array;
}

function assignMergeValue(object, key, value) {
  if ((value !== undefined && !eq(object[key], value)) ||
      (typeof key == 'number' && value === undefined && !(key in object))) {
    object[key] = value;
  }
}

var objectProto$12 = Object.prototype;


var hasOwnProperty$6 = objectProto$12.hasOwnProperty;


function assignValue(object, key, value) {
  var objValue = object[key];
  if (!(hasOwnProperty$6.call(object, key) && eq(objValue, value)) ||
      (value === undefined && !(key in object))) {
    object[key] = value;
  }
}

function copyObject(source, props, object, customizer) {
  object || (object = {});

  var index = -1,
      length = props.length;

  while (++index < length) {
    var key = props[index];

    var newValue = customizer
      ? customizer(object[key], source[key], key, object, source)
      : source[key];

    assignValue(object, key, newValue);
  }
  return object;
}

function baseAssign(object, source) {
  return object && copyObject(source, keys(source), object);
}

function cloneBuffer(buffer, isDeep) {
  if (isDeep) {
    return buffer.slice();
  }
  var result = new buffer.constructor(buffer.length);
  buffer.copy(result);
  return result;
}

function copyArray(source, array) {
  var index = -1,
      length = source.length;

  array || (array = Array(length));
  while (++index < length) {
    array[index] = source[index];
  }
  return array;
}

function stubArray() {
  return [];
}

var getOwnPropertySymbols = Object.getOwnPropertySymbols;


function getSymbols(object) {
  
  
  return getOwnPropertySymbols(Object(object));
}


if (!getOwnPropertySymbols) {
  getSymbols = stubArray;
}

var getSymbols$1 = getSymbols;

function copySymbols(source, object) {
  return copyObject(source, getSymbols$1(source), object);
}

function arrayPush(array, values) {
  var index = -1,
      length = values.length,
      offset = array.length;

  while (++index < length) {
    array[offset + index] = values[index];
  }
  return array;
}

function baseGetAllKeys(object, keysFunc, symbolsFunc) {
  var result = keysFunc(object);
  return isArray(object) ? result : arrayPush(result, symbolsFunc(object));
}

function getAllKeys(object) {
  return baseGetAllKeys(object, keys, getSymbols$1);
}

var objectProto$13 = Object.prototype;


var hasOwnProperty$7 = objectProto$13.hasOwnProperty;


function initCloneArray(array) {
  var length = array.length,
      result = array.constructor(length);

  
  if (length && typeof array[0] == 'string' && hasOwnProperty$7.call(array, 'index')) {
    result.index = array.index;
    result.input = array.input;
  }
  return result;
}

function cloneArrayBuffer(arrayBuffer) {
  var result = new arrayBuffer.constructor(arrayBuffer.byteLength);
  new Uint8Array(result).set(new Uint8Array(arrayBuffer));
  return result;
}

function cloneDataView(dataView, isDeep) {
  var buffer = isDeep ? cloneArrayBuffer(dataView.buffer) : dataView.buffer;
  return new dataView.constructor(buffer, dataView.byteOffset, dataView.byteLength);
}

function addMapEntry(map, pair) {
  
  map.set(pair[0], pair[1]);
  return map;
}

function arrayReduce(array, iteratee, accumulator, initAccum) {
  var index = -1,
      length = array ? array.length : 0;

  if (initAccum && length) {
    accumulator = array[++index];
  }
  while (++index < length) {
    accumulator = iteratee(accumulator, array[index], index, array);
  }
  return accumulator;
}

function cloneMap(map, isDeep, cloneFunc) {
  var array = isDeep ? cloneFunc(mapToArray(map), true) : mapToArray(map);
  return arrayReduce(array, addMapEntry, new map.constructor);
}

var reFlags = /\w*$/;


function cloneRegExp(regexp) {
  var result = new regexp.constructor(regexp.source, reFlags.exec(regexp));
  result.lastIndex = regexp.lastIndex;
  return result;
}

function addSetEntry(set, value) {
  set.add(value);
  return set;
}

function cloneSet(set, isDeep, cloneFunc) {
  var array = isDeep ? cloneFunc(setToArray(set), true) : setToArray(set);
  return arrayReduce(array, addSetEntry, new set.constructor);
}

var symbolProto$2 = Symbol ? Symbol.prototype : undefined;
var symbolValueOf$1 = symbolProto$2 ? symbolProto$2.valueOf : undefined;


function cloneSymbol(symbol) {
  return symbolValueOf$1 ? Object(symbolValueOf$1.call(symbol)) : {};
}

function cloneTypedArray(typedArray, isDeep) {
  var buffer = isDeep ? cloneArrayBuffer(typedArray.buffer) : typedArray.buffer;
  return new typedArray.constructor(buffer, typedArray.byteOffset, typedArray.length);
}

var boolTag$3 = '[object Boolean]';
var dateTag$3 = '[object Date]';
var mapTag$4 = '[object Map]';
var numberTag$3 = '[object Number]';
var regexpTag$3 = '[object RegExp]';
var setTag$4 = '[object Set]';
var stringTag$4 = '[object String]';
var symbolTag$3 = '[object Symbol]';

var arrayBufferTag$3 = '[object ArrayBuffer]';
var dataViewTag$4 = '[object DataView]';
var float32Tag$2 = '[object Float32Array]';
var float64Tag$2 = '[object Float64Array]';
var int8Tag$2 = '[object Int8Array]';
var int16Tag$2 = '[object Int16Array]';
var int32Tag$2 = '[object Int32Array]';
var uint8Tag$2 = '[object Uint8Array]';
var uint8ClampedTag$2 = '[object Uint8ClampedArray]';
var uint16Tag$2 = '[object Uint16Array]';
var uint32Tag$2 = '[object Uint32Array]';


function initCloneByTag(object, tag, cloneFunc, isDeep) {
  var Ctor = object.constructor;
  switch (tag) {
    case arrayBufferTag$3:
      return cloneArrayBuffer(object);

    case boolTag$3:
    case dateTag$3:
      return new Ctor(+object);

    case dataViewTag$4:
      return cloneDataView(object, isDeep);

    case float32Tag$2: case float64Tag$2:
    case int8Tag$2: case int16Tag$2: case int32Tag$2:
    case uint8Tag$2: case uint8ClampedTag$2: case uint16Tag$2: case uint32Tag$2:
      return cloneTypedArray(object, isDeep);

    case mapTag$4:
      return cloneMap(object, isDeep, cloneFunc);

    case numberTag$3:
    case stringTag$4:
      return new Ctor(object);

    case regexpTag$3:
      return cloneRegExp(object);

    case setTag$4:
      return cloneSet(object, isDeep, cloneFunc);

    case symbolTag$3:
      return cloneSymbol(object);
  }
}

var objectCreate = Object.create;


function baseCreate(proto) {
  return isObject(proto) ? objectCreate(proto) : {};
}

function initCloneObject(object) {
  return (typeof object.constructor == 'function' && !isPrototype(object))
    ? baseCreate(getPrototype(object))
    : {};
}

function stubFalse() {
  return false;
}

var freeExports = typeof exports == 'object' && exports;


var freeModule = freeExports && typeof module == 'object' && module;


var moduleExports = freeModule && freeModule.exports === freeExports;


var Buffer = moduleExports ? root.Buffer : undefined;


var isBuffer = !Buffer ? stubFalse : function(value) {
  return value instanceof Buffer;
};

var argsTag$3 = '[object Arguments]';
var arrayTag$2 = '[object Array]';
var boolTag$2 = '[object Boolean]';
var dateTag$2 = '[object Date]';
var errorTag$2 = '[object Error]';
var funcTag$2 = '[object Function]';
var genTag$1 = '[object GeneratorFunction]';
var mapTag$3 = '[object Map]';
var numberTag$2 = '[object Number]';
var objectTag$3 = '[object Object]';
var regexpTag$2 = '[object RegExp]';
var setTag$3 = '[object Set]';
var stringTag$3 = '[object String]';
var symbolTag$2 = '[object Symbol]';
var weakMapTag$2 = '[object WeakMap]';

var arrayBufferTag$2 = '[object ArrayBuffer]';
var dataViewTag$3 = '[object DataView]';
var float32Tag$1 = '[object Float32Array]';
var float64Tag$1 = '[object Float64Array]';
var int8Tag$1 = '[object Int8Array]';
var int16Tag$1 = '[object Int16Array]';
var int32Tag$1 = '[object Int32Array]';
var uint8Tag$1 = '[object Uint8Array]';
var uint8ClampedTag$1 = '[object Uint8ClampedArray]';
var uint16Tag$1 = '[object Uint16Array]';
var uint32Tag$1 = '[object Uint32Array]';


var cloneableTags = {};
cloneableTags[argsTag$3] = cloneableTags[arrayTag$2] =
cloneableTags[arrayBufferTag$2] = cloneableTags[dataViewTag$3] =
cloneableTags[boolTag$2] = cloneableTags[dateTag$2] =
cloneableTags[float32Tag$1] = cloneableTags[float64Tag$1] =
cloneableTags[int8Tag$1] = cloneableTags[int16Tag$1] =
cloneableTags[int32Tag$1] = cloneableTags[mapTag$3] =
cloneableTags[numberTag$2] = cloneableTags[objectTag$3] =
cloneableTags[regexpTag$2] = cloneableTags[setTag$3] =
cloneableTags[stringTag$3] = cloneableTags[symbolTag$2] =
cloneableTags[uint8Tag$1] = cloneableTags[uint8ClampedTag$1] =
cloneableTags[uint16Tag$1] = cloneableTags[uint32Tag$1] = true;
cloneableTags[errorTag$2] = cloneableTags[funcTag$2] =
cloneableTags[weakMapTag$2] = false;


function baseClone(value, isDeep, isFull, customizer, key, object, stack) {
  var result;
  if (customizer) {
    result = object ? customizer(value, key, object, stack) : customizer(value);
  }
  if (result !== undefined) {
    return result;
  }
  if (!isObject(value)) {
    return value;
  }
  var isArr = isArray(value);
  if (isArr) {
    result = initCloneArray(value);
    if (!isDeep) {
      return copyArray(value, result);
    }
  } else {
    var tag = getTag$1(value),
        isFunc = tag == funcTag$2 || tag == genTag$1;

    if (isBuffer(value)) {
      return cloneBuffer(value, isDeep);
    }
    if (tag == objectTag$3 || tag == argsTag$3 || (isFunc && !object)) {
      if (isHostObject(value)) {
        return object ? value : {};
      }
      result = initCloneObject(isFunc ? {} : value);
      if (!isDeep) {
        return copySymbols(value, baseAssign(result, value));
      }
    } else {
      if (!cloneableTags[tag]) {
        return object ? value : {};
      }
      result = initCloneByTag(value, tag, baseClone, isDeep);
    }
  }
  
  stack || (stack = new Stack);
  var stacked = stack.get(value);
  if (stacked) {
    return stacked;
  }
  stack.set(value, result);

  if (!isArr) {
    var props = isFull ? getAllKeys(value) : keys(value);
  }
  
  arrayEach(props || value, function(subValue, key) {
    if (props) {
      key = subValue;
      subValue = value[key];
    }
    assignValue(result, key, baseClone(subValue, isDeep, isFull, customizer, key, value, stack));
  });
  return result;
}

var objectTag$4 = '[object Object]';


var objectProto$14 = Object.prototype;


var funcToString$2 = Function.prototype.toString;


var hasOwnProperty$8 = objectProto$14.hasOwnProperty;


var objectCtorString = funcToString$2.call(Object);


var objectToString$6 = objectProto$14.toString;


function isPlainObject(value) {
  if (!isObjectLike(value) ||
      objectToString$6.call(value) != objectTag$4 || isHostObject(value)) {
    return false;
  }
  var proto = getPrototype(value);
  if (proto === null) {
    return true;
  }
  var Ctor = hasOwnProperty$8.call(proto, 'constructor') && proto.constructor;
  return (typeof Ctor == 'function' &&
    Ctor instanceof Ctor && funcToString$2.call(Ctor) == objectCtorString);
}

var Reflect = root.Reflect;

function iteratorToArray(iterator) {
  var data,
      result = [];

  while (!(data = iterator.next()).done) {
    result.push(data.value);
  }
  return result;
}

var objectProto$16 = Object.prototype;


var enumerate = Reflect ? Reflect.enumerate : undefined;
var propertyIsEnumerable$1 = objectProto$16.propertyIsEnumerable;


function baseKeysIn(object) {
  object = object == null ? object : Object(object);

  var result = [];
  for (var key in object) {
    result.push(key);
  }
  return result;
}


if (enumerate && !propertyIsEnumerable$1.call({ 'valueOf': 1 }, 'valueOf')) {
  baseKeysIn = function(object) {
    return iteratorToArray(enumerate(object));
  };
}

var baseKeysIn$1 = baseKeysIn;

var objectProto$15 = Object.prototype;


var hasOwnProperty$9 = objectProto$15.hasOwnProperty;


function keysIn(object) {
  var index = -1,
      isProto = isPrototype(object),
      props = baseKeysIn$1(object),
      propsLength = props.length,
      indexes = indexKeys(object),
      skipIndexes = !!indexes,
      result = indexes || [],
      length = result.length;

  while (++index < propsLength) {
    var key = props[index];
    if (!(skipIndexes && (key == 'length' || isIndex(key, length))) &&
        !(key == 'constructor' && (isProto || !hasOwnProperty$9.call(object, key)))) {
      result.push(key);
    }
  }
  return result;
}

function toPlainObject(value) {
  return copyObject(value, keysIn(value));
}

function baseMergeDeep(object, source, key, srcIndex, mergeFunc, customizer, stack) {
  var objValue = object[key],
      srcValue = source[key],
      stacked = stack.get(srcValue);

  if (stacked) {
    assignMergeValue(object, key, stacked);
    return;
  }
  var newValue = customizer
    ? customizer(objValue, srcValue, (key + ''), object, source, stack)
    : undefined;

  var isCommon = newValue === undefined;

  if (isCommon) {
    newValue = srcValue;
    if (isArray(srcValue) || isTypedArray(srcValue)) {
      if (isArray(objValue)) {
        newValue = objValue;
      }
      else if (isArrayLikeObject(objValue)) {
        newValue = copyArray(objValue);
      }
      else {
        isCommon = false;
        newValue = baseClone(srcValue, true);
      }
    }
    else if (isPlainObject(srcValue) || isArguments(srcValue)) {
      if (isArguments(objValue)) {
        newValue = toPlainObject(objValue);
      }
      else if (!isObject(objValue) || (srcIndex && isFunction(objValue))) {
        isCommon = false;
        newValue = baseClone(srcValue, true);
      }
      else {
        newValue = objValue;
      }
    }
    else {
      isCommon = false;
    }
  }
  stack.set(srcValue, newValue);

  if (isCommon) {
    
    mergeFunc(newValue, srcValue, srcIndex, customizer, stack);
  }
  stack['delete'](srcValue);
  assignMergeValue(object, key, newValue);
}

function baseMerge(object, source, srcIndex, customizer, stack) {
  if (object === source) {
    return;
  }
  if (!(isArray(source) || isTypedArray(source))) {
    var props = keysIn(source);
  }
  arrayEach(props || source, function(srcValue, key) {
    if (props) {
      key = srcValue;
      srcValue = source[key];
    }
    if (isObject(srcValue)) {
      stack || (stack = new Stack);
      baseMergeDeep(object, source, key, srcIndex, baseMerge, customizer, stack);
    }
    else {
      var newValue = customizer
        ? customizer(object[key], srcValue, (key + ''), object, source, stack)
        : undefined;

      if (newValue === undefined) {
        newValue = srcValue;
      }
      assignMergeValue(object, key, newValue);
    }
  });
}

function isIterateeCall(value, index, object) {
  if (!isObject(object)) {
    return false;
  }
  var type = typeof index;
  if (type == 'number'
        ? (isArrayLike(object) && isIndex(index, object.length))
        : (type == 'string' && index in object)
      ) {
    return eq(object[index], value);
  }
  return false;
}

function apply(func, thisArg, args) {
  var length = args.length;
  switch (length) {
    case 0: return func.call(thisArg);
    case 1: return func.call(thisArg, args[0]);
    case 2: return func.call(thisArg, args[0], args[1]);
    case 3: return func.call(thisArg, args[0], args[1], args[2]);
  }
  return func.apply(thisArg, args);
}

var FUNC_ERROR_TEXT$2 = 'Expected a function';


var nativeMax$2 = Math.max;


function rest(func, start) {
  if (typeof func != 'function') {
    throw new TypeError(FUNC_ERROR_TEXT$2);
  }
  start = nativeMax$2(start === undefined ? (func.length - 1) : toInteger(start), 0);
  return function() {
    var args = arguments,
        index = -1,
        length = nativeMax$2(args.length - start, 0),
        array = Array(length);

    while (++index < length) {
      array[index] = args[start + index];
    }
    switch (start) {
      case 0: return func.call(this, array);
      case 1: return func.call(this, args[0], array);
      case 2: return func.call(this, args[0], args[1], array);
    }
    var otherArgs = Array(start + 1);
    index = -1;
    while (++index < start) {
      otherArgs[index] = args[index];
    }
    otherArgs[start] = array;
    return apply(func, this, otherArgs);
  };
}

function createAssigner(assigner) {
  return rest(function(object, sources) {
    var index = -1,
        length = sources.length,
        customizer = length > 1 ? sources[length - 1] : undefined,
        guard = length > 2 ? sources[2] : undefined;

    customizer = (assigner.length > 3 && typeof customizer == 'function')
      ? (length--, customizer)
      : undefined;

    if (guard && isIterateeCall(sources[0], sources[1], guard)) {
      customizer = length < 3 ? undefined : customizer;
      length = 1;
    }
    object = Object(object);
    while (++index < length) {
      var source = sources[index];
      if (source) {
        assigner(object, source, index, customizer);
      }
    }
    return object;
  });
}

var merge = createAssigner(function(object, source, srcIndex) {
  baseMerge(object, source, srcIndex);
});

var mergeWith = createAssigner(function(object, source, srcIndex, customizer) {
  baseMerge(object, source, srcIndex, customizer);
});

function baseSet(object, path, value, customizer) {
  path = isKey(path, object) ? [path] : castPath(path);

  var index = -1,
      length = path.length,
      lastIndex = length - 1,
      nested = object;

  while (nested != null && ++index < length) {
    var key = toKey(path[index]);
    if (isObject(nested)) {
      var newValue = value;
      if (index != lastIndex) {
        var objValue = nested[key];
        newValue = customizer ? customizer(objValue, key, nested) : undefined;
        if (newValue === undefined) {
          newValue = objValue == null
            ? (isIndex(path[index + 1]) ? [] : {})
            : objValue;
        }
      }
      assignValue(nested, key, newValue);
    }
    nested = nested[key];
  }
  return object;
}

function setWith(object, path, value, customizer) {
  customizer = typeof customizer == 'function' ? customizer : undefined;
  return object == null ? object : baseSet(object, path, value, customizer);
}

function indexOfNaN(array, fromIndex, fromRight) {
  var length = array.length,
      index = fromIndex + (fromRight ? 1 : -1);

  while ((fromRight ? index-- : ++index < length)) {
    var other = array[index];
    if (other !== other) {
      return index;
    }
  }
  return -1;
}

function baseIndexOf(array, value, fromIndex) {
  if (value !== value) {
    return indexOfNaN(array, fromIndex);
  }
  var index = fromIndex - 1,
      length = array.length;

  while (++index < length) {
    if (array[index] === value) {
      return index;
    }
  }
  return -1;
}

function arrayIncludes(array, value) {
  var length = array ? array.length : 0;
  return !!length && baseIndexOf(array, value, 0) > -1;
}

function arrayIncludesWith(array, value, comparator) {
  var index = -1,
      length = array ? array.length : 0;

  while (++index < length) {
    if (comparator(value, array[index])) {
      return true;
    }
  }
  return false;
}

function cacheHas(cache, key) {
  return cache.has(key);
}

function noop() {
  
}

var INFINITY$3 = 1 / 0;


var createSet = !(Set$1 && (1 / setToArray(new Set$1([,-0]))[1]) == INFINITY$3) ? noop : function(values) {
  return new Set$1(values);
};

var LARGE_ARRAY_SIZE$1 = 200;


function last(array) {
  var length = array ? array.length : 0;
  return length ? array[length - 1] : undefined;
}

function parent(object, path) {
  return path.length == 1 ? object : baseGet(object, baseSlice(path, 0, -1));
}

function baseUnset(object, path) {
  path = isKey(path, object) ? [path] : castPath(path);
  object = parent(object, path);

  var key = toKey(last(path));
  return !(object != null && baseHas(object, key)) || delete object[key];
}

function unset(object, path) {
  return object == null ? true : baseUnset(object, path);
}

function arrayMap(array, iteratee) {
  var index = -1,
      length = array ? array.length : 0,
      result = Array(length);

  while (++index < length) {
    result[index] = iteratee(array[index], index, array);
  }
  return result;
}

function baseUnary(func) {
  return function(value) {
    return func(value);
  };
}

var LARGE_ARRAY_SIZE$2 = 200;


function baseDifference(array, values, iteratee, comparator) {
  var index = -1,
      includes = arrayIncludes,
      isCommon = true,
      length = array.length,
      result = [],
      valuesLength = values.length;

  if (!length) {
    return result;
  }
  if (iteratee) {
    values = arrayMap(values, baseUnary(iteratee));
  }
  if (comparator) {
    includes = arrayIncludesWith;
    isCommon = false;
  }
  else if (values.length >= LARGE_ARRAY_SIZE$2) {
    includes = cacheHas;
    isCommon = false;
    values = new SetCache(values);
  }
  outer:
  while (++index < length) {
    var value = array[index],
        computed = iteratee ? iteratee(value) : value;

    value = (comparator || value !== 0) ? value : 0;
    if (isCommon && computed === computed) {
      var valuesIndex = valuesLength;
      while (valuesIndex--) {
        if (values[valuesIndex] === computed) {
          continue outer;
        }
      }
      result.push(value);
    }
    else if (!includes(values, computed, comparator)) {
      result.push(value);
    }
  }
  return result;
}

var without = rest(function(array, values) {
  return isArrayLikeObject(array)
    ? baseDifference(array, values)
    : [];
});

function isObject$1(val) {
  const type = typeof val;
  return Boolean(val) && (type === 'object' || type === 'function')
}

function isArray$1(a) {
  return Array.isArray(a)
}

function clone(val) {
  if (isArray$1(val)) {
    return val.slice(0)
  }
  if (isObject$1(val)) {
    return Object.assign({}, val)
  }
  // we do not clone primitives
  // TODO: is that ok?
  return val
}

function forEach(iteratee, func) {
  if (!iteratee) return
  if (iteratee.constructor.prototype.forEach) {
    iteratee.forEach(func);
  } else {
    Object.keys(iteratee).forEach(function(key) {
      return func(iteratee[key], key)
    });
  }
}

/**
  @module

  Platform utilities such as browser detection etc.

  @example

  ```js
  import platform from 'substance/util/platform'
  ```
*/
const platform = {

  inBrowser: false,

  /**
    True if user agent is Internet Explorer or Microsoft Edge.
  */
  isIE: false,
  /**
    True if user agent is Firefox
  */

  isFF: false,

  isWebkit: false,

  /*
    Major version

    ATTENTION: at the moment only extracted for IE
  */
  version: -1,

  // TODO: make sure that this is implemented correctly

  isWindows: false,

  isMac: false,

  // in tests we change the state of this to emulate execuatio under certain conditions
  // to reset to defaults we call this function
  _reset: detect
};

function detect() {

  if (typeof window !== 'undefined') {
    platform.inBrowser = true;

    // Detect Internet Explorer / Edge
    const ua = window.navigator.userAgent;
    const msie = ua.indexOf('MSIE ');
    const trident = ua.indexOf('Trident/');
    const edge = ua.indexOf('Edge/');

    if (msie > 0) {
      // IE 10 or older => return version number
      platform.isIE = true;
      platform.version = 10;
      // TODO: if we need someday, this would be the exact version number
      // parseInt(ua.substring(msie + 5, ua.indexOf('.', msie)), 10)
    } else if (trident > 0) {
      // IE 11 => return version number
      platform.isIE = true;
      platform.version = 11;
      platform.isTrident = true;
      // TODO: if we need someday, this would be the exact version number
      // var rv = ua.indexOf('rv:')
      // parseInt(ua.substring(rv + 3, ua.indexOf('.', rv)), 10)
    } else if (edge > 0) {
      // IE 12 => return version number
      platform.isIE = true;
      platform.isEdge = true;
      platform.version = 12;
      // TODO: if we need someday, this would be the exact version number
      parseInt(ua.substring(edge + 5, ua.indexOf('.', edge)), 10);
    }

    // Detect Firefox
    platform.isFF = window.navigator.userAgent.toLowerCase().indexOf('firefox') > -1;

    // TODO: explicit detection of Webkit&/Blink
    platform.isWebkit = !platform.isFF && !platform.isIE;
  } else {
    platform.inBrowser = false;
  }

  if (platform.inBrowser) {
    platform.isWindows = (window.navigator !== undefined && window.navigator.appVersion && window.navigator.appVersion.indexOf("Win") !== -1);
    platform.isMac = (window.navigator !== undefined && window.navigator.platform.indexOf('Mac') >= 0);
  }

  // TOOD: is there a more reliable way to detect NodeJS?
  if (typeof process !== 'undefined') {
    platform.inNodeJS = true;
  }
}

detect();

function cloneDeep(val) {
  if (isArray$1(val)) {
    return _cloneArrayDeep(val);
  }
  // HACK: should we clone Files?
  // ATM we only use it when creating FileNodes
  // where the File instance is actually not serialized
  if (platform.inBrowser && val instanceof window.File) {
    return val
  }
  if (isObject$1(val)) {
    return _cloneObjectDeep(val)
  }
  // primitives don't need to be cloned
  // TODO: is that ok?
  return val
}

function _cloneObjectDeep(obj) {
  let res = {};
  forEach(obj, (val, key) => {
    res[key] = cloneDeep(val);
  });
  return res
}

function _cloneArrayDeep(arr) {
  return arr.map(cloneDeep)
}

function createCountingIdGenerator() {
  var counters = {};
  return function uuid(prefix) {
    if (!counters.hasOwnProperty(prefix)) {
      counters[prefix] = 1;
    }
    var result = [prefix, '-', counters[prefix]++].join('');
    return result
  }
}

function isString$1(s) {
  return typeof s === 'string'
}

function levenshtein(a, b){
  let m = [];
  for(let i = 0; i <= b.length; i++) {
    m[i] = [i];
    if(i === 0) continue;
    let ib = i-1;
    for(let j = 0; j <= a.length; j++){
      m[0][j] = j;
      if(j === 0) continue;
      let jb = j-1;
      m[i][j] = b.charAt(ib) === a.charAt(jb) ? m[ib][jb] : Math.min(
        m[ib][jb]+1,
        m[i][jb]+1,
        m[ib][j]+1
      );
    }
  }
  return m
}

/*
  Determines a list of changes to transform String a into String b.

  @param {String} a
  @param {String} b
*/
function diff(a, b, offset) {
  if (!isString$1(a) || !isString$1(b)) {
    throw new Error('Illegal arguments.')
  }
  offset = offset || 0;
  let changes = [];
  if (a || b) {
    if (!a && b) {
      changes.push({ type:'insert', start:offset, text:b });
    } else if (a && !b) {
      changes.push({ type:'delete', start:offset, end:offset+a.length });
    } else {
      let m = levenshtein(a, b);
      changes = _diff(a, b, m, offset);
    }
  }
  return changes
}

function _diff(a, b, m, offset) {
  let i = b.length;
  let j = a.length;
  let changes = [];
  let current;
  while (i>0 && j>0) {
    _next();
  }
  _commit();
  return changes

  function _next() {
    let d = m[i][j];
    let ib = i-1;
    let jb = j-1;
    // substitute
    if (m[ib][jb]<d) {
      if (current && current.type === 'replace') {
        current.start--;
        current.text.unshift(b.charAt(ib));
      } else {
        _commit();
        current = { type:'replace', start:jb, end:j, text:[b.charAt(ib)] };
      }
      i--;
      j--;
    }
    // insert
    else if (m[ib][j]<d) {
      if (current && current.type === 'insert') {
        current.start--;
        current.text.unshift(b.charAt(ib));
      } else {
        _commit();
        current = { type:'insert', start:jb, text:[b.charAt(ib)] };
      }
      i--;
    }
    // delete char
    else if (m[i][jb]<d) {
      if (current && current.type === 'delete') {
        current.start--;
      } else {
        _commit();
        current = { type:'delete', start:jb, end:j };
      }
      j--;
    }
    // preserve
    else {
      _commit();
      i--;
      j--;
    }
  }

  function _commit() {
    if (current) {
      switch (current.type) {
        case 'insert':
          current.start += offset;
          current.text = current.text.join('');
          break
        case 'delete':
          current.start += offset;
          current.end += offset;
          break
        case 'replace':
          current.start += offset;
          current.end += offset;
          current.text = current.text.join('');
          break
        default:
          throw new Error('Invalid state')
      }
      changes.push(current);
      current = null;
    }
  }

}

/*
  Escape XML Entities

  HACK: this is just a cheap implementation to escape XML entities
*/
function encodeXMLEntities(str) {
  return String(str)
    .replace(/&/g, '&amp;')
    .replace(/</g, '&lt;')
    .replace(/>/g, '&gt;')
    .replace(/"/g, '&quot;')
}

// for debugging
const DEBUG = false;
let count = 0;
const COUNT_MSG = '%s listeners registered in the whole system.';

/**
  Event support.
*/
class EventEmitter {

  /**
    Emit an event.

    @param {String} event
    @param ...arguments
    @return true if a listener was notified, false otherwise.
   */
  emit(event) {
    if (event in this.__events__) {
      // console.log("Emitting event %s (%d listeners) on", event, this.__events__[event].length, this)
      // Clone the list of bindings so that handlers can remove or add handlers during the call.
      var bindings = this.__events__[event].slice();
      var args = Array.prototype.slice.call(arguments, 1);
      for (var i = 0, len = bindings.length; i < len; i++) {
        var binding = bindings[i];
        // console.log("- triggering %s on %s", event, binding.context.constructor.name)
        binding.method.apply(binding.context, args);
      }
      return true
    }
    return false
  }

  /**
    Subscribe a listener to an event.

    Optionally, a `priority` can be provided to control the order
    of all bindings. The default priority is 0. All listeners with the
    same priority remain in order of registration.
    A lower priority will make the listener be called later, a higher
    priority earlier.

    @param {String} event
    @param {Function} method
    @param {Object} context
   */
  on(event, method, context) {
    // TODO: we could add options like 'once'
    _on.call(this, event, method, context);
  }

  /**
    Unsubscribe a listener from an event.

    @param {String} event
    @param {Function} method
    @param {Object} context
   */
  off(event, method, context) { // eslint-disable-line no-unused-vars
    if (arguments.length === 1 && isObject$1(arguments[0])) {
      _disconnect.call(this, arguments[0]);
    } else {
      _off.apply(this, arguments);
    }
  }

  _debugEvents() {
    /* eslint-disable no-console */
    console.log('### EventEmitter: ', this);
    forEach(this.__events__, (handlers, name) => {
      console.log("- %s listeners for %s: ", handlers.length, name, handlers);
    });
    /* eslint-enable no-console */
  }

  get __events__() {
    if (!this.___events___) {
      this.___events___ = {};
    }
    return this.___events___
  }

}

/*
  Internal implementation for registering a listener.

  @param {String} event
  @param {Function} method
  @param {Object} context
 */
function _on(event, method, context) {
  /* eslint-disable no-invalid-this */
  var bindings;
  validateMethod( method, context );
  if (this.__events__.hasOwnProperty(event)) {
    bindings = this.__events__[event];
  } else {
    // Auto-initialize bindings list
    bindings = this.__events__[event] = [];
  }
  // Add binding
  bindings.push({
    method: method,
    context: context || null
  });
  if (DEBUG) {
    count++;
    console.info('_on()', event, method.name, context, this);
    console.info(COUNT_MSG, count);
  }
  return this
  /*eslint-enable no-invalid-this */
}

/*
  Remove a listener.

  @param {String} event
  @param {Function} method
  @param {Object} context
 */
function _off(event, method, context) {
  /* eslint-disable no-invalid-this */
  if (arguments.length === 0) {
    if (DEBUG) {
      forEach(this.__events__, (bindings) => {
        bindings.forEach((b) => {
          console.info('_off()', b.method.name, b.context, this);
        });
        count -= bindings.length;
      });
      console.info(COUNT_MSG, count);
    }
    this.___events___ = {};
    return this
  }
  if (arguments.length === 1) {
    // Remove all bindings for event
    if (DEBUG) {
      count -= (this.__events__[event] || []).length;
      console.info(COUNT_MSG, count);
    }
    delete this.__events__[event];
    return this
  }
  validateMethod(method, context);
  if (!(event in this.__events__) || !this.__events__[event].length) {
    if (DEBUG) console.info('NO MATCHING BINDINGS');
    // No matching bindings
    return this
  }
  // Default to null context
  if (arguments.length < 3) {
    context = null;
  }
  // Remove matching handlers
  let bindings = this.__events__[event];
  for (let i = bindings.length-1; i >= 0; i--) {
    const b = bindings[i];
    if (b.method === method && b.context === context) {
      bindings.splice(i, 1);
      if (DEBUG) count--;
    }
  }
  // Cleanup if now empty
  if (bindings.length === 0) {
    delete this.__events__[event];
  }
  if (DEBUG) console.info(COUNT_MSG, count);
  return this
  /* eslint-enable no-invalid-this */
}

// removes a listener from all events
function _disconnect(context) {
  /* eslint-disable no-invalid-this */
  // Remove all connections to the context
  forEach(this.__events__, (bindings, event) => {
    for (let i = bindings.length-1; i>=0; i--) {
      // bindings[i] may have been removed by the previous steps
      // so check it still exists
      if (bindings[i] && bindings[i].context === context) {
        _off.call(this, event, bindings[i].method, context);
      }
    }
  });
  return this
  /* eslint-enable no-invalid-this */
}

function validateMethod(method, context) {
  // Validate method and context
  if (typeof method === 'string') {
    // Validate method
    if (context === undefined || context === null) {
      throw new Error( 'Method name "' + method + '" has no context.' )
    }
    if (!(method in context)) {
      // Technically the method does not need to exist yet: it could be
      // added before call time. But this probably signals a typo.
      throw new Error( 'Method not found: "' + method + '"' )
    }
    if (typeof context[method] !== 'function') {
      // Technically the property could be replaced by a function before
      // call time. But this probably signals a typo.
      throw new Error( 'Property "' + method + '" is not a function' )
    }
  } else if (typeof method !== 'function') {
    throw new Error( 'Invalid callback. Function or method name expected.' )
  }
}

function extend(...args) {
  return Object.assign(...args)
}

// just as a reference to detect name collisions
// with native Object properties
const PLAINOBJ = {};

/*
 * Simple registry implementation.
 *
 * @class Registry
 * @private
 */
class Registry {
  constructor(entries, validator) {
    this.entries = {};
    this.names = [];
    this.validator = validator;

    if (entries) {
      forEach(entries, function(entry, name) {
        this.add(name, entry);
      }.bind(this));
    }
  }

  /**
   * Check if an entry is registered for a given name.
   *
   * @param {String} name
   * @method contains
   * @memberof module:Basics.Registry.prototype
   */
  contains(name) {
    return this.entries.hasOwnProperty(name)
  }

  /**
   * Add an entry to the registry.
   *
   * @param {String} name
   * @param {Object} entry
   * @method add
   * @memberof module:Basics.Registry.prototype
   */
  add(name, entry) {
    if (this.validator) {
      this.validator(entry);
    }
    if (PLAINOBJ[name]) {
      throw new Error('Illegal key: "'+name+'" is a property of Object which is thus not allowed as a key.')
    }
    if (this.contains(name)) {
      this.remove(name);
    }
    this.entries[name] = entry;
    this.names.push(name);
  }

  /**
   * Remove an entry from the registry.
   *
   * @param {String} name
   * @method remove
   * @memberof module:Basics.Registry.prototype
   */
  remove(name) {
    let pos = this.names.indexOf(name);
    if (pos >= 0) {
      this.names.splice(pos, 1);
    }
    delete this.entries[name];
  }

  /**
   * @method clear
   * @memberof module:Basics.Registry.prototype
   */
  clear() {
    this.names = [];
    this.entries = {};
  }

  /**
   * Get the entry registered for a given name.
   *
   * @param {String} name
   * @return The registered entry
   * @method get
   * @memberof module:Basics.Registry.prototype
   */
  get(name, strict) {
    let result = this.entries[name];
    if (strict && !result) {
      throw new Error('No entry registered for name '+name)
    }
    return result
  }

  /*
    Iterate all registered entries in the order they were registered.

    @param {Function} callback with signature function(entry, name)
   */
  forEach(callback) {
    for (let i = 0; i < this.names.length; i++) {
      let name = this.names[i];
      let _continue = callback(this.entries[name], name);
      if (_continue === false) {
        break
      }
    }
  }

  map(callback) {
    let result = [];
    this.forEach((entry, name) => {
      result.push(callback(entry, name));
    });
    return result
  }

  filter(callback) {
    let result = [];
    this.forEach(function(entry, name) {
      if (callback(entry, name)) {
        result.push(entry);
      }
    });
    return result
  }

  values() {
    return this.filter(() => { return true })
  }
}

Registry.prototype._isRegistry = true;

function isFunction$1(f) {
  return typeof f === 'function'
}

function filter(iteratee, fn) {
  if (!iteratee) return []
  if (iteratee.constructor.prototype.filter && isFunction$1(iteratee.constructor.prototype.filter)) {
    return iteratee.filter(fn)
  }
  let result = [];
  forEach(iteratee, (val, key) => {
    if (fn(val, key)) {
      result.push(val);
    }
  });
  return result
}

function findIndex$1(arr, predicate) {
  if (!isFunction$1(predicate)) return arr.indexOf(predicate)
  for (let i = 0; i < arr.length; i++) {
    if (predicate(arr[i])) return i
  }
  return -1
}

function flatten(arr) {
  return Array.prototype.concat.apply([], arr)
}

function flattenOften(arr, max) {
  if (!(max > 0)) throw new Error("'max' must be a positive number")
  let l = arr.length;
  arr = flatten(arr);
  let round = 1;
  while (round < max && l < arr.length) {
    l = arr.length;
    arr = flatten(arr);
    round++;
  }
  return arr
}

function map(iteratee, func) {
  if (!iteratee) return []
  if (!func) func = function(item) { return item };
  if (Array.isArray(iteratee)) {
    return iteratee.map(func)
  } else {
    return Object.keys(iteratee).map(function(key) {
      return func(iteratee[key], key)
    })
  }
}

/**
  Get bounding rectangle relative to a given parent element. Allows multiple
  elements being passed (we need this for selections that consist of multiple
  selection fragments). Takes a relative parent element that is used as a
  reference point, instead of the browser's viewport.

  @param {Array} els elements to compute the bounding rectangle for
  @param {DOMElement} containerEl relative parent used as a reference point
  @return {object} rectangle description with left, top, right, bottom, width and height
*/
function getRelativeBoundingRect(els, containerEl) {
  if (els.length === undefined) {
    els = [els];
  }
  var elRects = map(els, function(el) {
    return _getBoundingOffsetsRect(el, containerEl)
  });

  var elsRect = _getBoundingRect(elRects);
  var containerElRect = containerEl.getBoundingClientRect();
  return {
    left: elsRect.left,
    top: elsRect.top,
    right: containerElRect.width - elsRect.left - elsRect.width,
    bottom: containerElRect.height - elsRect.top - elsRect.height,
    width: elsRect.width,
    height: elsRect.height
  }
}

/*
  Calculate a bounding rectangle for a set of rectangles.

  Note: Here, `bounds.right` and `bounds.bottom` are relative to
  the left top of the viewport.
*/
function _getBoundingRect(rects) {
  var bounds = {
    left: Number.POSITIVE_INFINITY,
    top: Number.POSITIVE_INFINITY,
    right: Number.NEGATIVE_INFINITY,
    bottom: Number.NEGATIVE_INFINITY,
    width: Number.NaN,
    height: Number.NaN
  };

  forEach(rects, function(rect) {
    if (rect.left < bounds.left) {
      bounds.left = rect.left;
    }
    if (rect.top < bounds.top) {
      bounds.top = rect.top;
    }
    if (rect.left + rect.width > bounds.right) {
      bounds.right = rect.left + rect.width;
    }
    if (rect.top + rect.height > bounds.bottom) {
      bounds.bottom = rect.top + rect.height;
    }
  });
  bounds.width = bounds.right - bounds.left;
  bounds.height = bounds.bottom - bounds.top;
  return bounds
}

/*
  Calculate the bounding rect of a single element relative to a parent.

  The rectangle dimensions are calculated as the union of the given elements
  clientRects. A selection fragment, for example, may appear as a multi-line span
  element that consists of a single client rect per line of text in variable widths.
*/
function _getBoundingOffsetsRect(el, relativeParentEl) {
  var relativeParentElRect = relativeParentEl.getBoundingClientRect();
  var elRect = _getBoundingRect(el.getClientRects());

  var left = elRect.left - relativeParentElRect.left;
  var top = elRect.top - relativeParentElRect.top;
  return {
    left: left,
    top: top,
    right: relativeParentElRect.width - left - elRect.width,
    bottom: relativeParentElRect.height - top - elRect.height,
    width: elRect.width,
    height: elRect.height
  }
}

/**
  Get bounding bounds for a given mouse event, relative parent element.

  @param {MouseEvent} mouseEvent the source mouse event
  @param {DOMElement} containerEl used as a reference point to calculate position
  @return {object} bound description with left, top, right, bottom
*/
function getRelativeMouseBounds(mouseEvent, containerEl) {
  let containerElRect = containerEl.getBoundingClientRect();
  let left = mouseEvent.clientX - containerElRect.left;
  let top = mouseEvent.clientY - containerElRect.top;
  let res = {
    left: left,
    right: containerElRect.width - left,
    top: top,
    bottom: containerElRect.height - top
  };
  return res;
}

var inBrowser = platform.inBrowser;

function isArrayEqual(arr1, arr2) {
  if (arr1 === arr2) return true
  if (!isArray$1(arr1) || !isArray$1(arr2)) return false
  if (arr1.length !== arr2.length) return false
  let L = arr1.length;
  for (var i = 0; i < L; i++) {
    if (arr1[i] !== arr2[i]) return false
  }
  return true
}

function isBoolean(val) {
  return (val === true || val === false || (val && val.constructor === Boolean) )
}

function isPlainObject$1(o) {
  return Boolean(o) && o.constructor === {}.constructor
}

function isEqual(a, b) {
  if (a === b) return true
  if (isArray$1(a) && isArray$1(b)) {
    if (a.length !== b.length) return false
    for (let i = 0; i < a.length; i++) {
      if (!isEqual(a[i], b[i])) return false
    }
    return true
  }
  if (isPlainObject$1(a) && isPlainObject$1(b)) {
    let akeys = Object.keys(a).sort();
    let bkeys = Object.keys(b).sort();
    if (!isEqual(akeys, bkeys)) return false
    for (let i = 0; i < akeys.length; i++) {
      let key = akeys[i];
      if (!isEqual(a[key], b[key])) return false
    }
    return true
  }
  return false
}

function isNil(o) {
  return o === null || o === undefined
}

function isNumber(n) {
  return typeof n === 'number'
}

/* eslint-disable strict */
var keys$1 = {
  UNDEFINED: 0,
  BACKSPACE: 8,
  DELETE: 46,
  LEFT: 37,
  RIGHT: 39,
  UP: 38,
  DOWN: 40,
  ENTER: 13,
  END: 35,
  HOME: 36,
  TAB: 9,
  PAGEUP: 33,
  PAGEDOWN: 34,
  ESCAPE: 27,
  SHIFT: 16,
  SPACE: 32
};

function parseKeyEvent(event, onlyModifiers) {
  let frags = [];
  if (event.altKey) {
    if (event.code === 'AltRight') {
      frags.push('ALTGR');
    } else {
      frags.push('ALT');
    }
  }
  if (event.ctrlKey) frags.push('CTRL');
  if (event.metaKey) frags.push('META');
  if (event.shiftKey) frags.push('SHIFT');
  if (!onlyModifiers) {
    frags.push(event.keyCode);
  }
  return frags.join('+')
}

function last$1(arr) {
  return arr[arr.length-1]
}

/*
  An object that can be access via path API.

  @example

  var obj = new PathObject({a: "aVal", b: {b1: 'b1Val', b2: 'b2Val'}})
*/

class PathObject {

  /*
    @param {object} [root] An object to operate on
  */
  constructor(root) {
    if (root) {
      this.__root__ = root;
    }
  }

  contains(id) {
    return Boolean(this.getRoot()[id])
  }

  getRoot() {
    if (this.__root__) {
      return this.__root__
    } else {
      return this
    }
  }

  /**
    Get value at path

    @return {object} The value stored for a given path

    @example

    obj.get(['b', 'b1'])
    => b1Val
  */
  get(path) {
    if (!path) {
      return undefined
    }
    if (isString$1(path)) {
      return this.getRoot()[path]
    }
    if (arguments.length > 1) {
      path = Array.prototype.slice(arguments, 0);
    }
    if (!isArray$1(path)) {
      throw new Error('Illegal argument for PathObject.get()')
    }
    return get(this.getRoot(), path)
  }

  set(path, value) {
    if (!path) {
      throw new Error('Illegal argument: PathObject.set(>path<, value) - path is mandatory.')
    }
    if (isString$1(path)) {
      this.getRoot()[path] = value;
    } else {
      setWith(this.getRoot(), path, value);
    }
  }

  delete(path) {
    if (isString$1(path)) {
      delete this.getRoot()[path];
    } else if (path.length === 1) {
      delete this.getRoot()[path[0]];
    } else {
      var success = unset(this.getRoot(), path);
      if (!success) {
        throw new Error('Could not delete property at path' + path)
      }
    }
  }

  clear() {
    var root = this.getRoot();
    for (var key in root) {
      if (root.hasOwnProperty(key)) {
        delete root[key];
      }
    }
  }

}

PathObject.prototype._isPathObject = true;

function startsWith(str, prefix) {
  if (!isString$1(str)) return false
  if (str.startsWith) return str.startsWith(prefix)
  if (!isString$1(prefix)) prefix = String(prefix);
  return str.slice(0, prefix.length) === prefix
}

/**
  A place to store global variables.
*/
const _global = (typeof global !== 'undefined') ? global : window;
const substanceGlobals = _global.hasOwnProperty('Substance') ? _global.Substance : _global.Substance = {
  DEBUG_RENDERING: true
};

function times(num, fn, ctx) {
  for (let i=0; i<num; i++) {
    fn.call(ctx);
  }
}

class TreeNode {}

/*
 * A tree-structure for indexes.
 *
 * @class TreeIndex
 * @param {object} [obj] An object to operate on
 * @memberof module:Basics
 * @example
 *
 * var index = new TreeIndex({a: "aVal", b: {b1: 'b1Val', b2: 'b2Val'}});
 */

class TreeIndex {
  /**
   * Get value at path.
   *
   * @return {object} The value stored for a given path
   *
   * @example
   *
   * obj.get(['b', 'b1']);
   * => b1Val
   */
  get(path) {
    if (arguments.length > 1) {
      path = Array.prototype.slice(arguments, 0);
    }
    if (isString$1(path)) {
      path = [path];
    }
    return get(this, path);
  }

  getAll(path) {
    if (arguments.length > 1) {
      path = Array.prototype.slice(arguments, 0);
    }
    if (isString$1(path)) {
      path = [path];
    }
    if (!isArray$1(path)) {
      throw new Error('Illegal argument for TreeIndex.get()');
    }
    var node = get(this, path);
    return this._collectValues(node);
  }

  set(path, value) {
    if (isString$1(path)) {
      path = [path];
    }
    setWith(this, path, value, function(val) {
      if (!val) return new TreeNode();
    });
  }

  delete(path) {
    if (isString$1(path)) {
      delete this[path];
    } else if(path.length === 1) {
      delete this[path[0]];
    } else {
      var key = path[path.length-1];
      path = path.slice(0, -1);
      var parent = get(this, path);
      if (parent) {
        delete parent[key];
      }
    }
  }

  clear() {
    var root = this;
    for (var key in root) {
      if (root.hasOwnProperty(key)) {
        delete root[key];
      }
    }
  }

  traverse(fn) {
    this._traverse(this, [], fn);
  }

  forEach(...args) {
    this.traverse(...args);
  }

  _traverse(root, path, fn) {
    var id;
    for (id in root) {
      if (!root.hasOwnProperty(id)) continue;
      var child = root[id];
      var childPath = path.concat([id]);
      if (child instanceof TreeNode) {
        this._traverse(child, childPath, fn);
      } else {
        fn(child, childPath);
      }
    }
  }

  _collectValues(root) {
    // TODO: don't know if this is the best solution
    // We use this only for indexes, e.g., to get all annotation on one node
    var vals = {};
    this._traverse(root, [], function(val, path) {
      var key = path[path.length-1];
      vals[key] = val;
    });
    return vals;
  }
}

class TreeIndexArrays extends TreeIndex {

  contains(path) {
    let val = super.get(path);
    return Boolean(val)
  }

  get(path) {
    let val = super.get(path);
    if (val instanceof TreeNode) {
      val = val.__values__ || [];
    }
    return val;
  }

  set(path, arr) {
    let val = super.get(path);
    val.__values__ = arr;
  }

  add(path, value) {
    if (isString$1(path)) {
      path = [path];
    }
    if (!isArray$1(path)) {
      throw new Error('Illegal arguments.');
    }
    var arr;

    // We are using setWith, as it allows us to create nodes on the way down
    // setWith can be controlled via a hook called for each key in the path
    // If val is not defined, a new node must be created and returned.
    // If val is defined, then we must return undefined to keep the original tree node
    // __dummy__ is necessary as setWith is used to set a value, but we want
    // to append to the array
    setWith(this, path.concat(['__values__','__dummy__']), undefined, function(val, key) {
      if (key === '__values__') {
        if (!val) val = [];
        arr = val;
      } else if (!val) {
        val = new TreeNode();
      }
      return val;
    });
    delete arr.__dummy__;
    arr.push(value);
  }

  remove(path, value) {
    var arr = get(this, path);
    if (arr instanceof TreeNode) {
      if (arguments.length === 1) {
        delete arr.__values__;
      } else {
        deleteFromArray(arr.__values__, value);
      }
    }
  }

  _collectValues(root) {
    var vals = [];
    this._traverse(root, [], function(val) {
      vals.push(val);
    });
    vals = Array.prototype.concat.apply([], vals);
    return vals
  }
}

TreeIndex.Arrays = TreeIndexArrays;

/*!
Math.uuid.js (v1.4)
http://www.broofa.com
mailto:robert@broofa.com
Copyright (c) 2010 Robert Kieffer
Dual licensed under the MIT and GPL licenses.
*/

/**
 * Generates a unique id.
 *
 * @param {String} [prefix] if provided the UUID will be prefixed.
 * @param {Number} [len] if provided a UUID with given length will be created.
 * @return A generated uuid.
 */
function uuid(prefix, len) {
  if (prefix && prefix[prefix.length-1] !== "-") {
    prefix = prefix.concat("-");
  }
  var chars = '0123456789abcdefghijklmnopqrstuvwxyz'.split('');
  var uuid = [];
  var radix = 16;
  var idx;
  len = len || 32;
  if (len) {
    // Compact form
    for (idx = 0; idx < len; idx++) uuid[idx] = chars[0 | Math.random()*radix];
  } else {
    // rfc4122, version 4 form
    var r;
    // rfc4122 requires these characters
    uuid[8] = uuid[13] = uuid[18] = uuid[23] = '-';
    uuid[14] = '4';
    // Fill in random data.  At i==19 set the high bits of clock sequence as
    // per rfc4122, sec. 4.1.5
    for (idx = 0; idx < 36; idx++) {
      if (!uuid[idx]) {
        r = 0 | Math.random()*16;
        uuid[idx] = chars[(idx === 19) ? (r & 0x3) | 0x8 : r];
      }
    }
  }
  return (prefix ? prefix : "") + uuid.join('');
}

function getDOMRangeFromEvent(evt) {
  let range, x = evt.clientX, y = evt.clientY;
  // Try the simple IE way first
  if (document.body.createTextRange) {
    range = document.body.createTextRange();
    range.moveToPoint(x, y);
  }
  else if (!isNil(document.createRange)) {
    // Try Mozilla's rangeOffset and rangeParent properties,
    // which are exactly what we want
    if (!isNil(evt.rangeParent)) {
      range = document.createRange();
      range.setStart(evt.rangeParent, evt.rangeOffset);
      range.collapse(true);
    }
    // Try the standards-based way next
    else if (document.caretPositionFromPoint) {
      let pos = document.caretPositionFromPoint(x, y);
      range = document.createRange();
      range.setStart(pos.offsetNode, pos.offset);
      range.collapse(true);
    }
    // Next, the WebKit way
    else if (document.caretRangeFromPoint) {
      range = document.caretRangeFromPoint(x, y);
    }
  }
  return range
}

/*
  Get selection rectangle relative to panel content element
*/
function getSelectionRect(parentRect) {
  if (platform.inBrowser) {
    const wsel = window.getSelection();
    if (wsel.rangeCount === 0) return
    const wrange = wsel.getRangeAt(0);
    let contentRect = parentRect;
    let selectionRect = wrange.getBoundingClientRect();
    if (selectionRect.top === 0 && selectionRect.bottom === 0) {
      selectionRect = _fixForCursorRectBug();
    }
    return getRelativeRect(contentRect, selectionRect)
  }
}

function _fixForCursorRectBug() {
  let wsel = window.getSelection();
  let el = wsel.anchorNode;
  if (!el) return
  while (el && el.nodeType !== 1) {
    el = el.parentNode;
  }
  let rects = el.getClientRects();
  let rect = rects[0];
  return {
    left: rect.left,
    top: rect.top,
    width: 0,
    height: rect.height,
    right: rect.width,
    bottom: rect.bottom
  }
}

function getRelativeRect(parentRect, childRect) {
  var left = childRect.left - parentRect.left;
  var top = childRect.top - parentRect.top;
  return {
    left: left,
    top: top,
    right: parentRect.width - left - childRect.width,
    bottom: parentRect.height - top - childRect.height,
    width: childRect.width,
    height: childRect.height
  }
}

function isMouseInsideDOMSelection(e) {
  let wsel = window.getSelection();
  if (wsel.rangeCount === 0) {
    return false
  }
  let wrange = wsel.getRangeAt(0);
  let selectionRect = wrange.getBoundingClientRect();
  return e.clientX >= selectionRect.left &&
         e.clientX <= selectionRect.right &&
         e.clientY >= selectionRect.top &&
         e.clientY <= selectionRect.bottom
}

/**
  A document selection. Refers to a Substance document model, not to the DOM.
*/
class Selection {

  constructor() {
    // Internal stuff
    var _internal = {};
    Object.defineProperty(this, "_internal", {
      enumerable: false,
      value: _internal
    });
      // set when attached to document
    _internal.doc = null;
  }

  clone() {
    var newSel = this._clone();
    if (this._internal.doc) {
      newSel.attach(this._internal.doc);
    }
    return newSel
  }

  /**
    @returns {Document} The attached document instance
  */
  getDocument() {
    var doc = this._internal.doc;
    if (!doc) {
      throw new Error('Selection is not attached to a document.')
    }
    return doc
  }

  isAttached() {
    return Boolean(this._internal.doc)
  }

  /**
    Attach document to the selection.

    @internal
    @param {Document} doc document to attach
    @returns {this}
  */
  attach(doc) {
    this._internal.doc = doc;
    return this
  }

  /**
    @returns {Boolean} true when selection is null.
  */
  isNull() { return false; }

  /**
    @returns {Boolean} true for property selections
  */
  isPropertySelection() { return false; }

  /**
    @returns {Boolean} true if selection is a {@link model/ContainerSelection}
  */
  isContainerSelection() { return false; }

  /**
    @returns {Boolean} true if selection is a {@link model/NodeSelection}
  */
  isNodeSelection() { return false; }

  isCustomSelection() { return false; }

  /**
    @returns {Boolean} true when selection is collapsed
  */
  isCollapsed() { return true; }

  /**
    @returns {Boolean} true if startOffset < endOffset
  */
  isReverse() { return false; }

  getType() {
    throw new Error('Selection.getType() is abstract.')
  }

  get type() {
    return this.getType()
  }

  /**
    @returns {Boolean} true if selection equals `other` selection
  */
  equals(other) {
    if (this === other) {
      return true
    } else if (!other) {
      return false
    } else if (this.isNull() !== other.isNull()) {
      return false
    } else if (this.getType() !== other.getType()) {
      return false
    } else {
      // Note: returning true here, so that sub-classes
      // can call this as a predicate in their expression
      return true
    }
  }

  /**
    @returns {String} This selection as human readable string.
  */
  toString() {
    return "null"
  }

  /**
    Convert container selection to JSON.

    @abstract
    @returns {Object}
  */
  toJSON() {
    throw new Error('This method is abstract.')
  }

  createWith(update) {
    let SelectionClass = this.constructor;
    let data = this.toJSON();
    Object.assign(data, update);
    return SelectionClass.fromJSON(data)
  }
}

// for duck-typed instanceof
Selection.prototype._isSelection = true;

/*
  Class to represent null selections.

  @internal
*/
class NullSelection extends Selection {

  isNull() {
    return true
  }

  getType() {
    return 'null'
  }

  toJSON() {
    return null
  }

  clone() {
    return this
  }
}

/**
  We use a singleton to represent NullSelections.

  @type {model/Selection}
*/

Selection.nullSelection = Object.freeze(new NullSelection());

/**
  @internal
*/
class Coordinate {

  /**
   @param {Array} path the address of a property, such as ['text_1', 'content']
   @param {int} offset the position in the property
  */
  constructor(path, offset) {
    // HACK: to allow this class be inherited but without calling this ctor
    if (arguments[0] === 'SKIP') return
    if (arguments.length === 1) {
      let data = arguments[0];
      this.path = data.path;
      this.offset = data.offset;
    } else {
      this.path = path;
      this.offset = offset;
    }
    if (!isArray$1(this.path)) {
      throw new Error('Invalid arguments: path should be an array.')
    }
    if (!isNumber(this.offset) || this.offset < 0) {
      throw new Error('Invalid arguments: offset must be a positive number.')
    }
  }

  equals(other) {
    return (other === this ||
      (isArrayEqual(other.path, this.path) && other.offset === this.offset) )
  }

  withCharPos(offset) {
    return new Coordinate(this.path, offset)
  }

  getNodeId() {
    return this.path[0]
  }

  getPath() {
    return this.path
  }

  getOffset() {
    return this.offset
  }

  toJSON() {
    return {
      path: this.path.slice(),
      offset: this.offset
    }
  }

  toString() {
    return "(" + this.path.join('.') + ", " + this.offset + ")"
  }

  isPropertyCoordinate() {
    return this.path.length > 1
  }

  isNodeCoordinate() {
    return this.path.length === 1
  }

  hasSamePath(other) {
    return isArrayEqual(this.path, other.path)
  }
}

Coordinate.prototype._isCoordinate = true;

/**
  A selection which is bound to a property. Implements {@link model/Selection}.

  @example

  ```js
  var propSel = doc.createSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 3,
    endOffset: 6
  })
*/
class PropertySelection extends Selection {

  /**
    @param {array} path
    @param {int} startOffset
    @param {int} endOffset
    @param {bool} reverse
    @param {string} [containerId]
    @param {string} [surfaceId]
  */
  constructor(path, startOffset, endOffset, reverse, containerId, surfaceId) {
    super();

    if (arguments.length === 1) {
      let data = arguments[0];
      path = data.path;
      startOffset = data.startOffset;
      endOffset = data.endOffset;
      reverse = data.reverse;
      containerId = data.containerId;
      surfaceId = data.surfaceId;
    }

    if (!path || !isNumber(startOffset)) {
      throw new Error('Invalid arguments: `path` and `startOffset` are mandatory');
    }

    this.start = new Coordinate(path, startOffset);
    this.end = new Coordinate(path, isNumber(endOffset) ? endOffset : startOffset);

    /**
      Selection direction.
      @type {Boolean}
    */
    this.reverse = Boolean(reverse);

    this.containerId = containerId;

    /**
      Identifier of the surface this selection should be active in.
      @type {String}
    */
    this.surfaceId = surfaceId;
  }

  get path() {
    return this.start.path
  }

  get startOffset() {
    console.warn('DEPRECATED: Use sel.start.offset instead');
    return this.start.offset
  }

  get endOffset() {
    console.warn('DEPRECATED: Use sel.end.offset instead');
    return this.end.offset
  }

  /**
    Convert container selection to JSON.

    @returns {Object}
  */
  toJSON() {
    return {
      type: 'property',
      path: this.start.path,
      startOffset: this.start.offset,
      endOffset: this.end.offset,
      reverse: this.reverse,
      containerId: this.containerId,
      surfaceId: this.surfaceId
    }
  }

  isPropertySelection() {
    return true
  }

  getType() {
    return 'property'
  }

  isNull() {
    return false
  }

  isCollapsed() {
    return this.start.offset === this.end.offset;
  }

  isReverse() {
    return this.reverse
  }

  equals(other) {
    return (
      Selection.prototype.equals.call(this, other) &&
      (this.start.equals(other.start) && this.end.equals(other.end))
    )
  }

  toString() {
    /* istanbul ignore next */
    return [
      "PropertySelection(", JSON.stringify(this.path), ", ",
      this.start.offset, " -> ", this.end.offset,
      (this.reverse?", reverse":""),
      (this.surfaceId?(", "+this.surfaceId):""),
      ")"
    ].join('')
  }

  /**
    Collapse a selection to chosen direction.

    @param {String} direction either left of right
    @returns {PropertySelection}
  */
  collapse(direction) {
    var offset;
    if (direction === 'left') {
      offset = this.start.offset;
    } else {
      offset = this.end.offset;
    }
    return this.createWithNewRange(offset, offset)
  }

  // Helper Methods
  // ----------------------

  /**
    Get path of a selection, e.g. target property where selected data is stored.

    @returns {String[]} path
  */
  getPath() {
    return this.start.path;
  }

  getNodeId() {
    return this.start.path[0];
  }

  /**
    Checks if this selection is inside another one.

    @param {Selection} other
    @param {Boolean} [strict] true if should check that it is strictly inside the other
    @returns {Boolean}
  */
  isInsideOf(other, strict) {
    if (other.isNull()) return false
    if (other.isContainerSelection()) {
      return other.contains(this, strict)
    }
    if (strict) {
      return (isArrayEqual(this.path, other.path) &&
        this.start.offset > other.start.offset &&
        this.end.offset < other.end.offset);
    } else {
      return (isArrayEqual(this.path, other.path) &&
        this.start.offset >= other.start.offset &&
        this.end.offset <= other.end.offset);
    }
  }

  /**
    Checks if this selection contains another one.

    @param {Selection} other
    @param {Boolean} [strict] true if should check that it is strictly contains the other
    @returns {Boolean}
  */
  contains(other, strict) {
    if (other.isNull()) return false
    return other.isInsideOf(this, strict)
  }

  /**
    Checks if this selection overlaps another one.

    @param {Selection} other
    @param {Boolean} [strict] true if should check that it is strictly overlaps the other
    @returns {Boolean}
  */
  overlaps(other, strict) {
    if (other.isNull()) return false
    if (other.isContainerSelection()) {
      // console.log('PropertySelection.overlaps: delegating to ContainerSelection.overlaps...')
      return other.overlaps(this)
    }
    if (!isArrayEqual(this.path, other.path)) return false
    if (strict) {
      return (! (this.start.offset>=other.end.offset||this.end.offset<=other.start.offset) );
    } else {
      return (! (this.start.offset>other.end.offset||this.end.offset<other.start.offset) );
    }
  }

  /**
    Checks if this selection has the right boundary in common with another one.

    @param {Selection} other
    @returns {Boolean}
  */
  isRightAlignedWith(other) {
    if (other.isNull()) return false
    if (other.isContainerSelection()) {
      // console.log('PropertySelection.isRightAlignedWith: delegating to ContainerSelection.isRightAlignedWith...')
      return other.isRightAlignedWith(this)
    }
    return (isArrayEqual(this.path, other.path) &&
      this.end.offset === other.end.offset);
  }

  /**
    Checks if this selection has the left boundary in common with another one.

    @param {Selection} other
    @returns {Boolean}
  */
  isLeftAlignedWith(other) {
    if (other.isNull()) return false
    if (other.isContainerSelection()) {
      // console.log('PropertySelection.isLeftAlignedWith: delegating to ContainerSelection.isLeftAlignedWith...')
      return other.isLeftAlignedWith(this)
    }
    return (isArrayEqual(this.path, other.path) &&
      this.start.offset === other.start.offset);
  }

  /**
    Expands selection to include another selection.

    @param {Selection} other
    @returns {Selection} a new selection
  */
  expand(other) {
    if (other.isNull()) return this

    // if the other is a ContainerSelection
    // we delegate to that implementation as it is more complex
    // and can deal with PropertySelections, too
    if (other.isContainerSelection()) {
      return other.expand(this)
    }
    if (!isArrayEqual(this.path, other.path)) {
      throw new Error('Can not expand PropertySelection to a different property.')
    }
    var newStartOffset = Math.min(this.start.offset, other.start.offset);
    var newEndOffset = Math.max(this.end.offset, other.end.offset);
    return this.createWithNewRange(newStartOffset, newEndOffset);
  }

  /**
    Creates a new selection by truncating this one by another selection.

    @param {Selection} other
    @returns {Selection} a new selection
  */
  truncateWith(other) {
    if (other.isNull()) return this
    if (other.isInsideOf(this, 'strict')) {
      // the other selection should overlap only on one side
      throw new Error('Can not truncate with a contained selections')
    }
    if (!this.overlaps(other)) {
      return this
    }
    var otherStartOffset, otherEndOffset;
    if (other.isPropertySelection()) {
      otherStartOffset = other.start.offset;
      otherEndOffset = other.end.offset;
    } else if (other.isContainerSelection()) {
      // either the startPath or the endPath must be the same
      if (isArrayEqual(other.start.path, this.start.path)) {
        otherStartOffset = other.start.offset;
      } else {
        otherStartOffset = this.start.offset;
      }
      if (isArrayEqual(other.end.path, this.start.path)) {
        otherEndOffset = other.end.offset;
      } else {
        otherEndOffset = this.end.offset;
      }
    } else {
      return this
    }

    var newStartOffset;
    var newEndOffset;
    if (this.start.offset > otherStartOffset && this.end.offset > otherEndOffset) {
      newStartOffset = otherEndOffset;
      newEndOffset = this.end.offset;
    } else if (this.start.offset < otherStartOffset && this.end.offset < otherEndOffset) {
      newStartOffset = this.start.offset;
      newEndOffset = otherStartOffset;
    } else if (this.start.offset === otherStartOffset) {
      if (this.end.offset <= otherEndOffset) {
        return Selection.nullSelection;
      } else {
        newStartOffset = otherEndOffset;
        newEndOffset = this.end.offset;
      }
    } else if (this.end.offset === otherEndOffset) {
      if (this.start.offset >= otherStartOffset) {
        return Selection.nullSelection;
      } else {
        newStartOffset = this.start.offset;
        newEndOffset = otherStartOffset;
      }
    } else if (other.contains(this)) {
      return Selection.nullSelection
    } else {
      // FIXME: if this happens, we have a bug somewhere above
      throw new Error('Illegal state.')
    }
    return this.createWithNewRange(newStartOffset, newEndOffset)
  }

  /**
    Creates a new selection with given range and same path.

    @param {Number} startOffset
    @param {Number} endOffset
    @returns {Selection} a new selection
  */
  createWithNewRange(startOffset, endOffset) {
    var sel = new PropertySelection(this.path, startOffset, endOffset, false, this.containerId, this.surfaceId);
    var doc = this._internal.doc;
    if (doc) {
      sel.attach(doc);
    }
    return sel
  }

  _clone() {
    return new PropertySelection(this.start.path, this.start.offset, this.end.offset, this.reverse, this.containerId, this.surfaceId);
  }

}

PropertySelection.fromJSON = function(json) {
  return new PropertySelection(json)
};

/**
  A selection spanning multiple nodes.

  @class
  @extends PropertySelection

  @example

  ```js
  let containerSel = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 5,
    endPath: ['p3', 'content'],
    endOffset: 4,
  })
  ```
*/
class ContainerSelection extends Selection {

  constructor(containerId, startPath, startOffset, endPath, endOffset, reverse, surfaceId) {
    super();

    if (arguments.length === 1) {
      let data = arguments[0];
      containerId = data.containerId;
      startPath = data.startPath;
      startOffset = data.startOffset;
      endPath = data.endPath;
      endOffset = data.endOffset;
      reverse = data.reverse;
      surfaceId = data.surfaceId;
    }

    /**
      @type {String}
    */
    this.containerId = containerId;
    if (!this.containerId) throw new Error('Invalid arguments: `containerId` is mandatory')

    this.start = new Coordinate(startPath, startOffset);
    this.end = new Coordinate(isNil(endPath) ? startPath : endPath, isNil(endOffset) ? startOffset : endOffset);

    this.reverse = Boolean(reverse);

    this.surfaceId = surfaceId;
  }

  /* istanbul ignore start */

  get startPath() {
    console.warn('DEPRECATED: use sel.start.path instead.');
    return this.start.path
  }

  get startOffset() {
    console.warn('DEPRECATED: use sel.start.offset instead.');
    return this.start.offset
  }

  get endPath() {
    console.warn('DEPRECATED: use sel.end.path instead.');
    return this.end.path
  }

  get endOffset() {
    console.warn('DEPRECATED: use sel.end.offset instead.');
    return this.end.offset
  }

  /* istanbul ignore end */

  toJSON() {
    return {
      type: 'container',
      containerId: this.containerId,
      startPath: this.start.path,
      startOffset: this.start.offset,
      endPath: this.end.path,
      endOffset: this.end.offset,
      reverse: this.reverse,
      surfaceId: this.surfaceId
    }
  }

  isContainerSelection() {
    return true
  }

  getType() {
    return 'container'
  }

  isNull() {
    return false
  }

  isCollapsed() {
    return this.start.equals(this.end)
  }

  isReverse() {
    return this.reverse
  }

  equals(other) {
    return (
      Selection.prototype.equals.call(this, other) &&
      this.containerId === other.containerId &&
      (this.start.equals(other.start) && this.end.equals(other.end))
    )
  }

  toString() {
    /* istanbul ignore next */
    return [
      "ContainerSelection(",
      this.containerId, ", ",
      JSON.stringify(this.start.path), ", ", this.start.offset,
      " -> ",
      JSON.stringify(this.end.path), ", ", this.end.offset,
      (this.reverse?", reverse":""),
      (this.surfaceId?(", "+this.surfaceId):""),
      ")"
    ].join('')
  }

  /**
    @return {model/Container} The container node instance for this selection.
  */
  getContainer() {
    if (!this._internal.container) {
      this._internal.container = this.getDocument().get(this.containerId);
    }
    return this._internal.container
  }

  isInsideOf(other, strict) {
    // Note: this gets called from PropertySelection.contains()
    // because this implementation can deal with mixed selection types.
    if (other.isNull()) return false
    strict = Boolean(strict);
    let r1 = this._range(this);
    let r2 = this._range(other);
    return (r2.start.isBefore(r1.start, strict) &&
      r1.end.isBefore(r2.end, strict))
  }

  contains(other, strict) {
    // Note: this gets called from PropertySelection.isInsideOf()
    // because this implementation can deal with mixed selection types.
    if (other.isNull()) return false
    strict = Boolean(strict);
    let r1 = this._range(this);
    let r2 = this._range(other);
    return (r1.start.isBefore(r2.start, strict) &&
      r2.end.isBefore(r1.end, strict))
  }

  containsNode(nodeId, strict) {
    const container = this.getContainer();
    if (!container.contains(nodeId)) return false
    const coor = new Coordinate([nodeId], 0);
    const address = container.getAddress(coor);
    const r = this._range(this);
    // console.log('ContainerSelection.containsNode()', address, 'is within', r.start, '->', r.end, '?')
    let contained = r.start.isBefore(address, strict);
    if (contained) {
      address.offset = 1;
      contained = r.end.isAfter(address, strict);
    }
    return contained
  }

  overlaps(other) {
    let r1 = this._range(this);
    let r2 = this._range(other);
    // it overlaps if they are not disjunct
    return !(r1.end.isBefore(r2.start, false) ||
      r2.end.isBefore(r1.start, false))
  }

  isLeftAlignedWith(other) {
    let r1 = this._range(this);
    let r2 = this._range(other);
    return r1.start.isEqual(r2.start)
  }

  isRightAlignedWith(other) {
    let r1 = this._range(this);
    let r2 = this._range(other);
    return r1.end.isEqual(r2.end)
  }

  /**
    Collapse a selection to chosen direction.

    @param {String} direction either left of right
    @returns {PropertySelection}
  */
  collapse(direction) {
    let coor;
    if (direction === 'left') {
      coor = this.start;
    } else {
      coor = this.end;
    }
    return _createNewSelection(this, coor, coor)
  }

  expand(other) {
    let r1 = this._range(this);
    let r2 = this._range(other);
    let start;
    let end;

    if (r1.start.isEqual(r2.start)) {
      start = new Coordinate(this.start.path, Math.min(this.start.offset, other.start.offset));
    } else if (r1.start.isAfter(r2.start)) {
      start = new Coordinate(other.start.path, other.start.offset);
    } else {
      start = this.start;
    }
    if (r1.end.isEqual(r2.end)) {
      end = new Coordinate(this.end.path, Math.max(this.end.offset, other.end.offset));
    } else if (r1.end.isBefore(r2.end, false)) {
      end = new Coordinate(other.end.path, other.end.offset);
    } else {
      end = this.end;
    }

    return _createNewSelection(this, start, end)
  }

  truncateWith(other) {
    if (other.isInsideOf(this, 'strict')) {
      // the other selection should overlap only on one side
      throw new Error('Can not truncate with a contained selections')
    }
    if (!this.overlaps(other)) {
      return this
    }
    let r1 = this._range(this);
    let r2 = this._range(other);
    let start, end;
    if (r2.start.isBefore(r1.start, 'strict') && r2.end.isBefore(r1.end, 'strict')) {
      start = other.end;
      end = this.end;
    } else if (r1.start.isBefore(r2.start, 'strict') && r1.end.isBefore(r2.end, 'strict')) {
      start = this.start;
      end = other.start;
    } else if (r1.start.isEqual(r2.start)) {
      if (r2.end.isBefore(r1.end, 'strict')) {
        start = other.end;
        end = this.end;
      } else {
        // the other selection is larger which eliminates this one
        return Selection.nullSelection
      }
    } else if (r1.end.isEqual(r2.end)) {
      if (r1.start.isBefore(r2.start, 'strict')) {
        start = this.start;
        end = other.start;
      } else {
        // the other selection is larger which eliminates this one
        return Selection.nullSelection
      }
    } else if (this.isInsideOf(other)) {
      return Selection.nullSelection
    } else {
      throw new Error('Could not determine coordinates for truncate. Check input')
    }
    return _createNewSelection(this, start, end)
  }

  /**
    Get the node ids covered by this selection.

    @returns {String[]} an array of ids
  */
  getNodeIds() {
    const container = this.getContainer();
    const startPos = container.getPosition(this.start.path[0]);
    const endPos = container.getPosition(this.end.path[0]);
    return container.nodes.slice(startPos, endPos+1)
  }

  /**
    Splits a container selection into property selections.

    @returns {PropertySelection[]}
  */
  splitIntoPropertySelections() {
    let sels = [];
    let fragments = this.getFragments();
    fragments.forEach(function(fragment) {
      if (fragment instanceof Selection.Fragment) {
        sels.push(
          new PropertySelection(fragment.path, fragment.startOffset,
            fragment.endOffset, false, this.containerId, this.surfaceId)
        );
      }
    }.bind(this));
    return sels
  }

  _clone() {
    return new ContainerSelection(this)
  }

  _range(sel) {
    // EXPERIMENTAL: caching the internal address based range
    // as we use it very often.
    // However, this is dangerous as this data can get invalid by a change
    if (sel._internal.addressRange) {
      return sel._internal.addressRange
    }

    let container = this.getContainer();
    let startAddress = container.getAddress(sel.start);
    let endAddress;
    if (sel.isCollapsed()) {
      endAddress = startAddress;
    } else {
      endAddress = container.getAddress(sel.end);
    }
    let addressRange = {
      start: startAddress,
      end: endAddress
    };
    if (sel._isContainerSelection) {
      sel._internal.addressRange = addressRange;
    }
    return addressRange
  }

  get path() {
    throw new Error('ContainerSelection has no path property. Use startPath and endPath instead')
  }

}

ContainerSelection.prototype._isContainerSelection = true;

ContainerSelection.fromJSON = function(properties) {
  let sel = new ContainerSelection(properties);
  return sel
};

function _createNewSelection(containerSel, start, end) {
  let newSel;

  if (start === end) {
    newSel = new PropertySelection({
      path: start.path,
      startOffset: start.offset,
      endOffset: start.offset,
      containerId: containerSel.containerId,
      surfaceId: containerSel.surfaceId
    });
  } else {
    newSel = new ContainerSelection(containerSel.containerId,
    start.path, start.offset, end.path, end.offset, false, containerSel.surfaceId);
  }
  // we need to attach the new selection
  const doc = containerSel._internal.doc;
  if (doc) {
    newSel.attach(doc);
  }
  return newSel
}

class NodeSelection extends Selection {

  constructor(containerId, nodeId, mode, reverse, surfaceId) {
    super();

    if (arguments.length === 1) {
      let data = arguments[0];
      containerId = data.containerId;
      nodeId = data.nodeId;
      mode = data.mode;
      reverse = data.reverse;
      surfaceId = data.surfaceId;
    }

    if (!isString$1(containerId)) {
      throw new Error("'containerId' is mandatory.");
    }
    if (!isString$1(nodeId)) {
      throw new Error("'nodeId' is mandatory.");
    }
    mode = mode || "full";

    this.containerId = containerId;
    this.nodeId = nodeId;
    this.mode = mode;
    this.reverse = Boolean(reverse);
    this.surfaceId = surfaceId;

    this.start = new Coordinate([nodeId], 0);
    this.end = new Coordinate([nodeId], 1);
  }

  equals(other) {
    return (
      super.equals(other) &&
      this.nodeId === other.nodeId &&
      this.mode === other.mode
    )
  }

  isNodeSelection() {
    return true;
  }

  getType() {
    return 'node';
  }

  getNodeId() {
    return this.nodeId;
  }

  isFull() {
    return this.mode === 'full';
  }

  isBefore() {
    return this.mode === 'before';
  }

  isAfter() {
    return this.mode === 'after';
  }

  isCollapsed() {
    return this.mode !== 'full';
  }

  toJSON() {
    return {
      type: 'node',
      nodeId: this.nodeId,
      mode: this.mode,
      reverse: this.reverse,
      containerId: this.containerId,
      surfaceId: this.surfaceId
    };
  }

  toString() {
    /* istanbul ignore next */
    return [
      "NodeSelection(",
      this.containerId, ".", this.nodeId, ", ",
      this.mode, ", ",
      (this.reverse?", reverse":""),
      (this.surfaceId?(", "+this.surfaceId):""),
      ")"
    ].join('');
  }

  collapse(direction) {
    if (direction === 'left') {
      if (this.isBefore()) {
        return this;
      } else {
        return new NodeSelection(this.containerId, this.nodeId, 'before', this.reverse, this.surfaceId);
      }
    } else if (direction === 'right') {
      if (this.isAfter()) {
        return this;
      } else {
        return new NodeSelection(this.containerId, this.nodeId, 'after', this.reverse, this.surfaceId);
      }
    } else {
      throw new Error("'direction' must be either 'left' or 'right'");
    }
  }

  _getCoordinate() {
    if (this.mode === 'before') {
      return new Coordinate([this.nodeId], 0);
    } else if (this.mode === 'after') {
      return new Coordinate([this.nodeId], 1);
    }
  }

  _clone() {
    return new NodeSelection(this);
  }
}

NodeSelection.prototype._isNodeSelection = true;

NodeSelection.fromJSON = function(json) {
  return new NodeSelection(json);
};

// TODO: is this used?
NodeSelection._createFromCoordinate = function(coor) {
  var containerId = coor.containerId;
  var nodeId = coor.getNodeId();
  var mode = coor.offset === 0 ? 'before' : 'after';
  return new NodeSelection(containerId, nodeId, mode, false);
};

class CustomSelection extends Selection {

  constructor(customType, data, surfaceId) {
    super();

    if (arguments.length === 1) {
      let _data = arguments[0];
      customType = _data.customType;
      data = _data.data;
      surfaceId = _data.surfaceId;
    }

    this.customType = customType;
    this.data = data || {};
    this.surfaceId = surfaceId;
  }

  isCustomSelection() {
    return true;
  }

  getType() {
    return 'custom';
  }

  getCustomType() {
    return this.customType;
  }

  toJSON() {
    return {
      type: 'custom',
      customType: this.customType,
      data: cloneDeep(this.data),
      surfaceId: this.surfaceId
    };
  }

  toString() {
    /* istanbul ignore next */
    return [
      'CustomSelection(',
      this.customType,', ',
      JSON.stringify(this.data),
      ")"
    ].join('');
  }

  equals(other) {
    return (
      Selection.prototype.equals.call(this, other) &&
      other.isCustomSelection() &&
      isEqual(this.data, other.data)
    );
  }

  _clone() {
    return new CustomSelection(this)
  }
}

CustomSelection.prototype._isCustomSelection = true;

CustomSelection.fromJSON = function(json) {
  return new CustomSelection(json);
};

function fromJSON(json) {
  if (!json) return Selection.nullSelection
  var type = json.type;
  switch(type) {
    case 'property':
      return PropertySelection.fromJSON(json)
    case 'container':
      return ContainerSelection.fromJSON(json)
    case 'node':
      return NodeSelection.fromJSON(json)
    case 'custom':
      return CustomSelection.fromJSON(json)
    default:
      // console.error('Selection.fromJSON(): unsupported selection data', json)
      return Selection.nullSelection
  }
}

/*
  Helper to check if a coordinate is the first position of a node.
*/
function isFirst(doc, coor) {
  if (coor.isNodeCoordinate() && coor.offset === 0) return true
  let node = doc.get(coor.path[0]).getRoot();
  if (node.isText() && coor.offset === 0) return true
  if (node.isList()) {
    let itemId = coor.path[0];
    if (node.items[0] === itemId && coor.offset === 0) return true
  }
}

/*
  Helper to check if a coordinate is the last position of a node.
*/
function isLast(doc, coor) {
  if (coor.isNodeCoordinate() && coor.offset > 0) return true
  let node = doc.get(coor.path[0]).getRoot();
  if (node.isText() && coor.offset >= node.getLength()) return true
  if (node.isList()) {
    let itemId = coor.path[0];
    let item = doc.get(itemId);
    if (last$1(node.items) === itemId && coor.offset === item.getLength()) return true
  }
}

function isEntirelySelected(doc, node, start, end) {
  let { isEntirelySelected } = getRangeInfo(doc, node, start, end);
  return isEntirelySelected
}

function getRangeInfo(doc, node, start, end) {
  let isFirst = true;
  let isLast = true;
  if (node.isText()) {
    if (start && start.offset !== 0) isFirst = false;
    if (end && end.offset < node.getLength()) isLast = false;
  } else if (node.isList()) {
    if (start) {
      let itemId = start.path[0];
      let itemPos = node.getItemPosition(itemId);
      if (itemPos > 0 || start.offset !== 0) isFirst = false;
    }
    if (end) {
      let itemId = end.path[0];
      let itemPos = node.getItemPosition(itemId);
      let item = doc.get(itemId);
      if (itemPos < node.items.length-1 || end.offset < item.getLength()) isLast = false;
    }
  }
  let isEntirelySelected = isFirst && isLast;
  return {isFirst, isLast, isEntirelySelected}
}

function setCursor(tx, node, containerId, mode) {
  if (node.isText()) {
    let offset = 0;
    if (mode === 'after') {
      let text = node.getText();
      offset = text.length;
    }
    tx.setSelection({
      type: 'property',
      path: node.getTextPath(),
      startOffset: offset,
      containerId: containerId
    });
  } else if (node.isList()) {
    let item, offset;
    if (mode === 'after') {
      item = node.getLastItem();
      offset = item.getLength();
    } else {
      item = node.getFirstItem();
      offset = 0;
    }
    tx.setSelection({
      type: 'property',
      path: item.getTextPath(),
      startOffset: offset,
      containerId: containerId
    });
  } else {
    tx.setSelection({
      type: 'node',
      containerId: containerId,
      nodeId: node.id,
      // NOTE: ATM we mostly use 'full' NodeSelections
      // Still, they are supported internally
      // mode: mode
    });
  }
}

function selectNode(tx, nodeId, containerId) {
  tx.setSelection(createNodeSelection({ doc: tx, nodeId, containerId }));
}

function createNodeSelection({ doc, nodeId, containerId, mode, reverse, surfaceId}) {
  let node = doc.get(nodeId);
  if (!node) return Selection.nullSelection
  node = node.getRoot();
  if (node.isText()) {
    return new PropertySelection({
      path: node.getTextPath(),
      startOffset: mode === 'after' ? node.getLength() : 0,
      endOffset: mode === 'before' ? 0 : node.getLength(),
      reverse: reverse,
      containerId: containerId,
      surfaceId: surfaceId
    })
  } else if (node.isList() && node.getLength()>0) {
    let first = node.getFirstItem();
    let last$$1 = node.getLastItem();
    let start = {
      path: first.getTextPath(),
      offset: 0
    };
    let end = {
      path: last$$1.getTextPath(),
      offset: last$$1.getLength()
    };
    if (mode === 'after') start = end;
    else if (mode === 'before') end = start;
    return new ContainerSelection({
      startPath: start.path,
      startOffset: start.offset,
      endPath: end.path,
      endOffset: end.offset,
      reverse: reverse,
      containerId: containerId,
      surfaceId: surfaceId
    })
  } else {
    return new NodeSelection({ nodeId, mode, reverse, containerId, surfaceId })
  }
}

function stepIntoIsolatedNode(editorSession, comp) {
  // this succeeds if the content component provides
  // a grabFocus() implementation
  if (comp.grabFocus()) return true

  // otherwise we try to find the first surface
  let surface = comp.find('.sc-surface');
  if (surface) {
    if (surface._isTextPropertyEditor) {
      const doc = editorSession.getDocument();
      const path = surface.getPath();
      const text = doc.get(path, 'strict');
      editorSession.setSelection({
        type: 'property',
        path: path,
        startOffset: text.length,
        surfaceId: surface.id
      });
      return true
    } else if (surface._isContainerEditor) {
      let container = surface.getContainer();
      if (container.length > 0) {
        let first = container.getChildAt(0);
        setCursor(editorSession, first, container.id, 'after');
      }
      return true
    }
  }
  return false
}


var selectionHelpers = Object.freeze({
	fromJSON: fromJSON,
	isFirst: isFirst,
	isLast: isLast,
	isEntirelySelected: isEntirelySelected,
	getRangeInfo: getRangeInfo,
	setCursor: setCursor,
	selectNode: selectNode,
	createNodeSelection: createNodeSelection,
	stepIntoIsolatedNode: stepIntoIsolatedNode
});

class Conflict extends Error {
  constructor(a, b) {
    super("Conflict: " + JSON.stringify(a) +" vs " + JSON.stringify(b));
    this.a = a;
    this.b = b;
  }
}

const INSERT = "insert";
const DELETE$1 = "delete";

class TextOperation {

  constructor(data) {
    if (!data || data.type === undefined || data.pos === undefined || data.str === undefined) {
      throw new Error("Illegal argument: insufficient data.")
    }
    // 'insert' or 'delete'
    this.type = data.type;
    // the position where to apply the operation
    this.pos = data.pos;
    // the string to delete or insert
    this.str = data.str;
    // sanity checks
    if(!this.isInsert() && !this.isDelete()) {
      throw new Error("Illegal type.")
    }
    if (!isString$1(this.str)) {
      throw new Error("Illegal argument: expecting string.")
    }
    if (!isNumber(this.pos) || this.pos < 0) {
      throw new Error("Illegal argument: expecting positive number as pos.")
    }
  }

  apply(str) {
    if (this.isEmpty()) return str
    if (this.type === INSERT) {
      if (str.length < this.pos) {
        throw new Error("Provided string is too short.")
      }
      if (str.splice) {
        return str.splice(this.pos, 0, this.str)
      } else {
        return str.slice(0, this.pos).concat(this.str).concat(str.slice(this.pos))
      }
    }
    else /* if (this.type === DELETE) */ {
      if (str.length < this.pos + this.str.length) {
        throw new Error("Provided string is too short.")
      }
      if (str.splice) {
        return str.splice(this.pos, this.str.length)
      } else {
        return str.slice(0, this.pos).concat(str.slice(this.pos + this.str.length))
      }
    }
  }

  clone() {
    return new TextOperation(this)
  }

  isNOP() {
    return this.type === "NOP" || this.str.length === 0
  }

  isInsert() {
    return this.type === INSERT
  }

  isDelete() {
    return this.type === DELETE$1
  }

  getLength() {
    return this.str.length
  }

  invert() {
    var data = {
      type: this.isInsert() ? DELETE$1 : INSERT,
      pos: this.pos,
      str: this.str
    };
    return new TextOperation(data)
  }

  hasConflict(other) {
    return _hasConflict(this, other)
  }

  isEmpty() {
    return this.str.length === 0
  }

  toJSON() {
    return {
      type: this.type,
      pos: this.pos,
      str: this.str
    }
  }

  toString() {
    return ["(", (this.isInsert() ? INSERT : DELETE$1), ",", this.pos, ",'", this.str, "')"].join('')
  }
}

TextOperation.prototype._isOperation = true;
TextOperation.prototype._isTextOperation = true;

function _hasConflict(a, b) {
  // Insert vs Insert:
  //
  // Insertions are conflicting iff their insert position is the same.
  if (a.type === INSERT && b.type === INSERT) return (a.pos === b.pos)
  // Delete vs Delete:
  //
  // Deletions are conflicting if their ranges overlap.
  if (a.type === DELETE$1 && b.type === DELETE$1) {
    // to have no conflict, either `a` should be after `b` or `b` after `a`, otherwise.
    return !(a.pos >= b.pos + b.str.length || b.pos >= a.pos + a.str.length)
  }
  // Delete vs Insert:
  //
  // A deletion and an insertion are conflicting if the insert position is within the deleted range.
  var del, ins;
  if (a.type === DELETE$1) {
    del = a; ins = b;
  } else {
    del = b; ins = a;
  }
  return (ins.pos >= del.pos && ins.pos < del.pos + del.str.length)
}

// Transforms two Insertions
// --------

function transform_insert_insert(a, b) {
  if (a.pos === b.pos) {
    b.pos += a.str.length;
  }
  else if (a.pos < b.pos) {
    b.pos += a.str.length;
  }
  else {
    a.pos += b.str.length;
  }
}

// Transform two Deletions
// --------
//

function transform_delete_delete$1(a, b, first) {
  // reduce to a normalized case
  if (a.pos > b.pos) {
    return transform_delete_delete$1(b, a, !first)
  }
  if (a.pos === b.pos && a.str.length > b.str.length) {
    return transform_delete_delete$1(b, a, !first)
  }
  // take out overlapping parts
  if (b.pos < a.pos + a.str.length) {
    var s = b.pos - a.pos;
    var s1 = a.str.length - s;
    var s2 = s + b.str.length;
    a.str = a.str.slice(0, s) + a.str.slice(s2);
    b.str = b.str.slice(s1);
    b.pos -= s;
  } else {
    b.pos -= a.str.length;
  }
}

// Transform Insert and Deletion
// --------
//

function transform_insert_delete(a, b) {
  if (a.type === DELETE$1) {
    return transform_insert_delete(b, a)
  }
  // we can assume, that a is an insertion and b is a deletion
  // a is before b
  if (a.pos <= b.pos) {
    b.pos += a.str.length;
  }
  // a is after b
  else if (a.pos >= b.pos + b.str.length) {
    a.pos -= b.str.length;
  }
  // Note: this is a conflict case the user should be noticed about
  // If applied still, the deletion takes precedence
  // a.pos > b.pos && <= b.pos + b.length
  else {
    var s = a.pos - b.pos;
    b.str = b.str.slice(0, s) + a.str + b.str.slice(s);
    a.str = "";
  }
}

function transform$1(a, b, options) {
  options = options || {};
  if (options["no-conflict"] && _hasConflict(a, b)) {
    throw new Conflict(a, b)
  }
  if (!options.inplace) {
    a = a.clone();
    b = b.clone();
  }
  if (a.type === INSERT && b.type === INSERT) {
    transform_insert_insert(a, b);
  }
  else if (a.type === DELETE$1 && b.type === DELETE$1) {
    transform_delete_delete$1(a, b, true);
  }
  else {
    transform_insert_delete(a,b);
  }
  return [a, b]
}

TextOperation.transform = function() {
  return transform$1.apply(null, arguments)
};

/* Factories */

TextOperation.Insert = function(pos, str) {
  return new TextOperation({ type: INSERT, pos: pos, str: str })
};

TextOperation.Delete = function(pos, str) {
  return new TextOperation({ type: DELETE$1, pos: pos, str: str })
};

TextOperation.INSERT = INSERT;
TextOperation.DELETE = DELETE$1;

TextOperation.fromJSON = function(data) {
  return new TextOperation(data)
};

const NOP$1 = "NOP";
const DELETE$2 = "delete";
const INSERT$1 = "insert";

class ArrayOperation {

  constructor(data) {
    if (!data || !data.type) {
      throw new Error("Illegal argument: insufficient data.")
    }
    this.type = data.type;
    if (this.type === NOP$1) return

    if (this.type !== INSERT$1 && this.type !== DELETE$2) {
      throw new Error("Illegal type.")
    }
    // the position where to apply the operation
    this.pos = data.pos;
    // the value to insert or delete
    this.val = data.val;
    if (!isNumber(this.pos) || this.pos < 0) {
      throw new Error("Illegal argument: expecting positive number as pos.")
    }
  }

  apply(array) {
    if (this.type === NOP$1) {
      return array
    }
    if (this.type === INSERT$1) {
      if (array.length < this.pos) {
        throw new Error("Provided array is too small.")
      }
      array.splice(this.pos, 0, this.val);
      return array
    }
    // Delete
    else /* if (this.type === DELETE) */ {
      if (array.length < this.pos) {
        throw new Error("Provided array is too small.")
      }
      if (!isEqual(array[this.pos], this.val)) {
        throw Error("Unexpected value at position " + this.pos + ". Expected " + this.val + ", found " + array[this.pos])
      }
      array.splice(this.pos, 1);
      return array
    }
  }

  clone() {
    var data = {
      type: this.type,
      pos: this.pos,
      val: cloneDeep(this.val)
    };
    return new ArrayOperation(data)
  }

  invert() {
    var data = this.toJSON();
    if (this.type === NOP$1) data.type = NOP$1;
    else if (this.type === INSERT$1) data.type = DELETE$2;
    else /* if (this.type === DELETE) */ data.type = INSERT$1;
    return new ArrayOperation(data)
  }

  hasConflict(other) {
    return ArrayOperation.hasConflict(this, other)
  }

  toJSON() {
    var result = {
      type: this.type,
    };
    if (this.type === NOP$1) return result
    result.pos = this.pos;
    result.val = cloneDeep(this.val);
    return result
  }

  isInsert() {
    return this.type === INSERT$1
  }

  isDelete() {
    return this.type === DELETE$2
  }

  getOffset() {
    return this.pos
  }

  getValue() {
    return this.val
  }

  isNOP() {
    return this.type === NOP$1
  }

  toString() {
    return ["(", (this.isInsert() ? INSERT$1 : DELETE$2), ",", this.getOffset(), ",'", this.getValue(), "')"].join('')
  }
}

ArrayOperation.prototype._isOperation = true;
ArrayOperation.prototype._isArrayOperation = true;

function hasConflict$1(a, b) {
  if (a.type === NOP$1 || b.type === NOP$1) return false
  if (a.type === INSERT$1 && b.type === INSERT$1) {
    return a.pos === b.pos
  } else {
    return false
  }
}

function transform_insert_insert$1(a, b) {
  if (a.pos === b.pos) {
    b.pos += 1;
  }
  // a before b
  else if (a.pos < b.pos) {
    b.pos += 1;
  }
  // a after b
  else {
    a.pos += 1;
  }
}

function transform_delete_delete$2(a, b) {
  // turn the second of two concurrent deletes into a NOP
  if (a.pos === b.pos) {
    b.type = NOP$1;
    a.type = NOP$1;
    return
  }
  if (a.pos < b.pos) {
    b.pos -= 1;
  } else {
    a.pos -= 1;
  }
}

function transform_insert_delete$1(a, b) {
  // reduce to a normalized case
  if (a.type === DELETE$2) {
    var tmp = a;
    a = b;
    b = tmp;
  }
  if (a.pos <= b.pos) {
    b.pos += 1;
  } else {
    a.pos -= 1;
  }
}

var transform$2 = function(a, b, options) {
  options = options || {};
  // enable conflicts when you want to notify the user of potential problems
  // Note that even in these cases, there is a defined result.
  if (options['no-conflict'] && hasConflict$1(a, b)) {
    throw new Conflict(a, b)
  }
  // this is used internally only as optimization, e.g., when rebasing an operation
  if (!options.inplace) {
    a = a.clone();
    b = b.clone();
  }
  if (a.type === NOP$1 || b.type === NOP$1) {
    // nothing to transform
  }
  else if (a.type === INSERT$1 && b.type === INSERT$1) {
    transform_insert_insert$1(a, b);
  }
  else if (a.type === DELETE$2 && b.type === DELETE$2) {
    transform_delete_delete$2(a, b);
  }
  else {
    transform_insert_delete$1(a, b);
  }
  return [a, b]
};

ArrayOperation.transform = transform$2;
ArrayOperation.hasConflict = hasConflict$1;

/* Factories */

ArrayOperation.Insert = function(pos, val) {
  return new ArrayOperation({type:INSERT$1, pos: pos, val: val})
};

ArrayOperation.Delete = function(pos, val) {
  return new ArrayOperation({ type:DELETE$2, pos: pos, val: val })
};

ArrayOperation.fromJSON = function(data) {
  return new ArrayOperation(data)
};

ArrayOperation.NOP = NOP$1;
ArrayOperation.DELETE = DELETE$2;
ArrayOperation.INSERT = INSERT$1;

const SHIFT = 'shift';

class CoordinateOperation {

  constructor(data) {
    if (!data || data.type === undefined) {
      throw new Error("Illegal argument: insufficient data.")
    }
    // 'shift'
    this.type = data.type;
    // the position where to apply the operation
    this.val = data.val;
    // sanity checks
    if(!this.isShift()) {
      throw new Error("Illegal type.")
    }
    if (!isNumber(this.val)) {
      throw new Error("Illegal argument: expecting number as shift value.")
    }
  }

  apply(coor) {
    coor.offset = coor.offset + this.val;
  }

  isShift() {
    return this.type === SHIFT
  }

  isNOP() {
    switch (this.type) {
      case SHIFT: {
        return this.val === 0
      }
      default:
        return false
    }
  }

  clone() {
    return new CoordinateOperation(this)
  }

  invert() {
    let data;
    switch (this.type) {
      case SHIFT:
        data = {
          type: SHIFT,
          val: -this.val
        };
        break
      default:
        throw new Error('Invalid type.')
    }
    return new CoordinateOperation(data)
  }

  hasConflict() {
    // TODO: support conflict detection?
    return false
  }

  toJSON() {
    return {
      type: this.type,
      val: this.val
    }
  }

  toString() {
    return ["(", (this.type), ",", this.val, "')"].join('')
  }

}

CoordinateOperation.prototype._isOperation = true;
CoordinateOperation.prototype._isCoordinateOperation = true;

function transform_shift_shift(a, b) {
  a.val += b.val;
  b.val += a.val;
}

function transform$3(a, b, options) {
  options = options || {};
  // TODO: support conflict detection?
  if (!options.inplace) {
    a = a.clone();
    b = b.clone();
  }
  if (a.type === SHIFT && b.type === SHIFT) {
    transform_shift_shift(a, b);
  }
  else {
    throw new Error('Illegal type')
  }
  return [a, b]
}

CoordinateOperation.transform = function(...args) {
  return transform$3(...args)
};

CoordinateOperation.fromJSON = function(json) {
  return new CoordinateOperation(json)
};

CoordinateOperation.Shift = function(val) {
  return new CoordinateOperation({
    type: SHIFT,
    val: val
  })
};

const NOP = "NOP";
const CREATE = "create";
const DELETE = 'delete';
const UPDATE = 'update';
const SET = 'set';

class ObjectOperation {

  constructor(data) {
    /* istanbul ignore next */
    if (!data) {
      throw new Error('Data of ObjectOperation is missing.')
    }
    /* istanbul ignore next */
    if (!data.type) {
      throw new Error('Invalid data: type is mandatory.')
    }
    this.type = data.type;
    if (data.type === NOP) {
      return
    }
    this.path = data.path;
    if (!data.path) {
      throw new Error('Invalid data: path is mandatory.')
    }
    if (this.type === CREATE || this.type === DELETE) {
      if (!data.val) {
        throw new Error('Invalid data: value is missing.')
      }
      this.val = data.val;
    }
    else if (this.type === UPDATE) {
      if (data.diff) {
        this.diff = data.diff;
        if (data.diff._isTextOperation) {
          this.propertyType = 'string';
        } else if (data.diff._isArrayOperation) {
          this.propertyType = 'array';
        } else if (data.diff._isCoordinateOperation) {
          this.propertyType = 'coordinate';
        } else {
          throw new Error('Invalid data: diff must be a TextOperation or an ArrayOperation.')
        }
      } else {
        throw new Error("Invalid data: diff is mandatory for update operation.")
      }
    }
    else if (this.type === SET) {
      this.val = data.val;
      this.original = data.original;
    } else {
      throw new Error('Invalid type: '+ data.type)
    }
  }

  apply(obj) {
    if (this.type === NOP) return obj
    var adapter;
    if (obj._isPathObject) {
      adapter = obj;
    } else {
      adapter = new PathObject(obj);
    }
    if (this.type === CREATE) {
      adapter.set(this.path, cloneDeep(this.val));
      return obj
    }
    if (this.type === DELETE) {
      adapter.delete(this.path, "strict");
    }
    else if (this.type === UPDATE) {
      var diff$$1 = this.diff;
      switch (this.propertyType) {
        case 'array': {
          let arr = adapter.get(this.path);
          diff$$1.apply(arr);
          break
        }
        case 'string': {
          let str = adapter.get(this.path);
          if (isNil(str)) str = '';
          str = diff$$1.apply(str);
          adapter.set(this.path, str);
          break
        }
        case 'coordinate': {
          let coor = adapter.get(this.path);
          if (!coor) throw new Error('No coordinate with path '+this.path)
          diff$$1.apply(coor);
          break
        }
        default:
          throw new Error('Invalid state.')
      }
    }
    else if (this.type === SET) {
      // clone here as the operations value must not be changed
      adapter.set(this.path, cloneDeep(this.val));
    }
    else {
      throw new Error('Invalid type.')
    }
    return obj
  }

  clone() {
    var data = {
      type: this.type,
      path: this.path,
    };
    if (this.val) {
      data.val = cloneDeep(this.val);
    }
    if (this.diff) {
      data.diff = this.diff.clone();
    }
    return new ObjectOperation(data)
  }

  isNOP() {
    if (this.type === NOP) return true
    else if (this.type === UPDATE) return this.diff.isNOP()
  }

  isCreate() {
    return this.type === CREATE
  }

  isDelete() {
    return this.type === DELETE
  }

  isUpdate(propertyType) {
    if (propertyType) {
      return (this.type === UPDATE && this.propertyType === propertyType)
    } else {
      return this.type === UPDATE
    }
  }

  isSet() {
    return this.type === SET
  }

  invert() {
    if (this.type === NOP) {
      return new ObjectOperation({ type: NOP })
    }
    var result = new ObjectOperation(this);
    if (this.type === CREATE) {
      result.type = DELETE;
    }
    else if (this.type === DELETE) {
      result.type = CREATE;
    }
    else if (this.type === UPDATE) {
      var invertedDiff;
      if (this.diff._isTextOperation) {
        invertedDiff = TextOperation.fromJSON(this.diff.toJSON()).invert();
      } else if (this.diff._isArrayOperation) {
        invertedDiff = ArrayOperation.fromJSON(this.diff.toJSON()).invert();
      } else if (this.diff._isCoordinateOperation) {
        invertedDiff = CoordinateOperation.fromJSON(this.diff.toJSON()).invert();
      } else {
        throw new Error('Illegal type')
      }
      result.diff = invertedDiff;
    }
    else /* if (this.type === SET) */ {
      result.val = this.original;
      result.original = this.val;
    }
    return result
  }

  hasConflict(other) {
    return ObjectOperation.hasConflict(this, other)
  }

  toJSON() {
    if (this.type === NOP) {
      return { type: NOP }
    }
    var data = {
      type: this.type,
      path: this.path,
    };
    if (this.type === CREATE || this.type === DELETE) {
      data.val = this.val;
    }
    else if (this.type === UPDATE) {
      if (this.diff._isTextOperation) {
        data.propertyType = "string";
      } else if (this.diff._isArrayOperation) {
        data.propertyType = "array";
      } else if (this.diff._isCoordinateOperation) {
        data.propertyType = "coordinate";
      } else {
        throw new Error('Invalid property type.')
      }
      data.diff = this.diff.toJSON();
    }
    else /* if (this.type === SET) */ {
      data.val = this.val;
      data.original = this.original;
    }
    return data
  }

  getType() {
    return this.type
  }

  getPath() {
    return this.path
  }

  getValue() {
    return this.val
  }

  getOldValue() {
    return this.original
  }

  getValueOp() {
    return this.diff
  }

  /* istanbul ignore next */
  toString() {
    switch (this.type) {
      case CREATE:
        return ["(+,", JSON.stringify(this.path), JSON.stringify(this.val), ")"].join('')
      case DELETE:
        return ["(-,", JSON.stringify(this.path), JSON.stringify(this.val), ")"].join('')
      case UPDATE:
        return ["(>>,", JSON.stringify(this.path), this.propertyType, this.diff.toString(), ")"].join('')
      case SET:
        return ["(=,", JSON.stringify(this.path), this.val, this.original, ")"].join('')
      case NOP:
        return "NOP"
      default:
        throw new Error('Invalid type')
    }
  }
}

ObjectOperation.prototype._isOperation = true;
ObjectOperation.prototype._isObjectOperation = true;

/* Low level implementation */

function hasConflict(a, b) {
  if (a.type === NOP || b.type === NOP) return false
  return isEqual(a.path, b.path)
}

function transform_delete_delete(a, b) {
  // both operations have the same effect.
  // the transformed operations are turned into NOPs
  a.type = NOP;
  b.type = NOP;
}

function transform_create_create() {
  throw new Error("Can not transform two concurring creates of the same property")
}

function transform_delete_create() {
  throw new Error('Illegal state: can not create and delete a value at the same time.')
}

function transform_delete_update(a, b, flipped) {
  if (a.type !== DELETE) {
    return transform_delete_update(b, a, true)
  }
  var op;
  switch (b.propertyType) {
    case 'string':
      op = TextOperation.fromJSON(b.diff);
      break
    case 'array':
      op = ArrayOperation.fromJSON(b.diff);
      break
    case 'coordinate':
      op = CoordinateOperation.fromJSON(b.diff);
      break
    default:
      throw new Error('Illegal type')
  }
  // (DELETE, UPDATE) is transformed into (DELETE, CREATE)
  if (!flipped) {
    a.type = NOP;
    b.type = CREATE;
    b.val = op.apply(a.val);
  }
  // (UPDATE, DELETE): the delete is updated to delete the updated value
  else {
    a.val = op.apply(a.val);
    b.type = NOP;
  }
}

function transform_create_update() {
  // it is not possible to reasonably transform this.
  throw new Error("Can not transform a concurring create and update of the same property")
}

function transform_update_update(a, b) {
  // Note: this is a conflict the user should know about
  var op_a, op_b, t;
  switch(b.propertyType) {
    case 'string':
      op_a = TextOperation.fromJSON(a.diff);
      op_b = TextOperation.fromJSON(b.diff);
      t = TextOperation.transform(op_a, op_b, {inplace: true});
      break
    case 'array':
      op_a = ArrayOperation.fromJSON(a.diff);
      op_b = ArrayOperation.fromJSON(b.diff);
      t = ArrayOperation.transform(op_a, op_b, {inplace: true});
      break
    case 'coordinate':
      op_a = CoordinateOperation.fromJSON(a.diff);
      op_b = CoordinateOperation.fromJSON(b.diff);
      t = CoordinateOperation.transform(op_a, op_b, {inplace: true});
      break
    default:
      throw new Error('Illegal type')
  }
  a.diff = t[0];
  b.diff = t[1];
}

function transform_create_set() {
  throw new Error('Illegal state: can not create and set a value at the same time.')
}

function transform_delete_set(a, b, flipped) {
  if (a.type !== DELETE) return transform_delete_set(b, a, true)
  if (!flipped) {
    a.type = NOP;
    b.type = CREATE;
    b.original = undefined;
  } else {
    a.val = b.val;
    b.type = NOP;
  }
}

function transform_update_set() {
  throw new Error("Unresolvable conflict: update + set.")
}

function transform_set_set(a, b) {
  a.type = NOP;
  b.original = a.val;
}

const _NOP = 0;
const _CREATE = 1;
const _DELETE = 2;
const _UPDATE = 4;
const _SET = 8;

const CODE = (() => {
  const c = {};
  c[NOP] =_NOP;
  c[CREATE] = _CREATE;
  c[DELETE] = _DELETE;
  c[UPDATE] = _UPDATE;
  c[SET] = _SET;
  return c
})();

const __transform__ = (() => {
  /* eslint-disable no-multi-spaces */
  const t = {};
  t[_DELETE | _DELETE] = transform_delete_delete;
  t[_DELETE | _CREATE] = transform_delete_create;
  t[_DELETE | _UPDATE] = transform_delete_update;
  t[_CREATE | _CREATE] = transform_create_create;
  t[_CREATE | _UPDATE] = transform_create_update;
  t[_UPDATE | _UPDATE] = transform_update_update;
  t[_CREATE | _SET   ] = transform_create_set;
  t[_DELETE | _SET   ] = transform_delete_set;
  t[_UPDATE | _SET   ] = transform_update_set;
  t[_SET    | _SET   ] = transform_set_set;
  /* eslint-enable no-multi-spaces */
  return t
})();

function transform(a, b, options) {
  options = options || {};
  if (options['no-conflict'] && hasConflict(a, b)) {
    throw new Conflict(a, b)
  }
  if (!options.inplace) {
    a = a.clone();
    b = b.clone();
  }
  if (a.isNOP() || b.isNOP()) {
    return [a, b]
  }
  var sameProp = isEqual(a.path, b.path);
  // without conflict: a' = a, b' = b
  if (sameProp) {
    __transform__[CODE[a.type] | CODE[b.type]](a,b);
  }
  return [a, b]
}

ObjectOperation.transform = transform;
ObjectOperation.hasConflict = hasConflict;

/* Factories */

ObjectOperation.Create = function(idOrPath, val) {
  var path;
  if (isString$1(idOrPath)) {
    path = [idOrPath];
  } else {
    path = idOrPath;
  }
  return new ObjectOperation({type: CREATE, path: path, val: val})
};

ObjectOperation.Delete = function(idOrPath, val) {
  var path;
  if (isString$1(idOrPath)) {
    path = [idOrPath];
  } else {
    path = idOrPath;
  }
  return new ObjectOperation({type: DELETE, path: path, val: val})
};

ObjectOperation.Update = function(path, op) {
  return new ObjectOperation({
    type: UPDATE,
    path: path,
    diff: op
  })
};

ObjectOperation.Set = function(path, oldVal, newVal) {
  return new ObjectOperation({
    type: SET,
    path: path,
    val: cloneDeep(newVal),
    original: cloneDeep(oldVal)
  })
};

ObjectOperation.fromJSON = function(data) {
  data = cloneDeep(data);
  if (data.type === "update") {
    switch (data.propertyType) {
      case "string":
        data.diff = TextOperation.fromJSON(data.diff);
        break
      case "array":
        data.diff = ArrayOperation.fromJSON(data.diff);
        break
      case "coordinate":
        data.diff = CoordinateOperation.fromJSON(data.diff);
        break
      default:
        throw new Error("Unsupported update diff:" + JSON.stringify(data.diff))
    }
  }
  var op = new ObjectOperation(data);
  return op
};

ObjectOperation.NOP = NOP;
ObjectOperation.CREATE = CREATE;
ObjectOperation.DELETE = DELETE;
ObjectOperation.UPDATE = UPDATE;
ObjectOperation.SET = SET;

/*
  Transforms change A with B, as if A was done before B.
  A' and B' can be used to update two clients to get to the
  same document content.

     / A - B' \
  v_n          v_n+1
     \ B - A' /
*/
function transformDocumentChange(A, B) {
  _transformInplaceBatch(A, B);
}

function transformSelection(sel, a) {
  let newSel = sel.clone();
  let hasChanged = _transformSelectionInplace(newSel, a);
  if (hasChanged) {
    return newSel
  } else {
    return sel
  }
}

function _transformInplaceSingle(a, b) {
  for (let i = 0; i < a.ops.length; i++) {
    let a_op = a.ops[i];
    for (let j = 0; j < b.ops.length; j++) {
      let b_op = b.ops[j];
      // ATTENTION: order of arguments is important.
      // First argument is the dominant one, i.e. it is treated as if it was applied before
      ObjectOperation.transform(a_op, b_op, {inplace: true});
    }
  }
  if (a.before) {
    _transformSelectionInplace(a.before.selection, b);
  }
  if (a.after) {
    _transformSelectionInplace(a.after.selection, b);
  }
  if (b.before) {
    _transformSelectionInplace(b.before.selection, a);
  }
  if (b.after) {
    _transformSelectionInplace(b.after.selection, a);
  }
}

function _transformInplaceBatch(A, B) {
  if (!isArray$1(A)) {
    A = [A];
  }
  if (!isArray$1(B)) {
    B = [B];
  }
  for (let i = 0; i < A.length; i++) {
    let a = A[i];
    for (let j = 0; j < B.length; j++) {
      let b = B[j];
      _transformInplaceSingle(a,b);
    }
  }
}

function _transformSelectionInplace(sel, a) {
  if (!sel || (!sel.isPropertySelection() && !sel.isContainerSelection()) ) {
    return false
  }
  let ops = a.ops;
  let hasChanged = false;
  let isCollapsed = sel.isCollapsed();
  for(let i=0; i<ops.length; i++) {
    let op = ops[i];
    hasChanged |= _transformCoordinateInplace(sel.start, op);
    if (!isCollapsed) {
      hasChanged |= _transformCoordinateInplace(sel.end, op);
    } else {
      if (sel.isContainerSelection()) {
        sel.end.path = sel.start.path;
      }
      sel.end.offset = sel.start.offset;
    }
  }
  return hasChanged
}

function _transformCoordinateInplace(coor, op) {
  if (!isEqual(op.path, coor.path)) return false
  let hasChanged = false;
  if (op.type === 'update' && op.propertyType === 'string') {
    let diff$$1 = op.diff;
    let newOffset;
    if (diff$$1.isInsert() && diff$$1.pos <= coor.offset) {
      newOffset = coor.offset + diff$$1.str.length;
      // console.log('Transforming coordinate after inserting %s chars:', diff.str.length, coor.toString(), '->', newOffset)
      coor.offset = newOffset;
      hasChanged = true;
    } else if (diff$$1.isDelete() && diff$$1.pos <= coor.offset) {
      newOffset = Math.max(diff$$1.pos, coor.offset - diff$$1.str.length);
      // console.log('Transforming coordinate after deleting %s chars:', diff.str.length, coor.toString(), '->', newOffset)
      coor.offset = newOffset;
      hasChanged = true;
    }
  }
  return hasChanged
}


var operationHelpers = Object.freeze({
	transformDocumentChange: transformDocumentChange,
	transformSelection: transformSelection
});

/*
  Internal helper class used by model/data/Node.
*/
class Property {

  constructor(definition) {
    this.definition = definition;
  }

  isArray() {
    return isArray$1(this.definition.type)
  }

  isReference() {
    if (this.isArray()) {
      return last$1(this.definition.type) === 'id'
    } else {
      return this.definition.type === 'id'
    }
  }

  isText() {
    return Boolean(this.definition._isText)
  }

  isOwned() {
    return Boolean(this.definition.owned)
  }

  isOptional() {
    return Boolean(this.definition.optional)
  }

  isNotNull() {
    return Boolean(this.definition.notNull)
  }

  hasDefault() {
    return this.definition.hasOwnProperty('default')
  }

  getDefault() {
    return this.definition.default
  }

  createDefaultValue() {
    if (isArray$1(this.definition.type)) {
      return []
    }
    switch(this.definition.type) {
      case 'object':
        return {}
      case 'number':
        return -1
      case 'coordinate':
        return new Coordinate([], 0)
      case 'boolean':
        return false
      case 'id':
        return null
      case 'string':
        return ''
      default:
        return null
    }
  }

  get type() {
    return this.definition.type
  }

  get name() {
    return this.definition.name
  }
}

/*
  Base node implementation.

  @prop {String} id an id that is unique within this data
 */
class Node extends EventEmitter {

  /**
    @param {Object} properties
  */
  constructor() {
    super();

    // NOTE: this indirection allows us to implement a overridable initializer
    // For instance, DocumentNode sets the document instance and the props
    this._initialize.apply(this, arguments);
  }

  _initialize(data) {
    const NodeClass = this.constructor;

    let schema = NodeClass.schema;
    for (var name in schema) {
      if (!schema.hasOwnProperty(name)) continue
      let prop = schema[name];
      // check integrity of provided data, such as type correctness,
      // and mandatory properties
      const propIsGiven = (data[name] !== undefined);
      const hasDefault = prop.hasDefault();
      const isOptional = prop.isOptional();
      if ( (!isOptional && !hasDefault) && !propIsGiven) {
        throw new Error('Property ' + name + ' is mandatory for node type ' + this.type)
      }
      if (propIsGiven) {
        this[name] = _checked(prop, data[name]);
      } else if (hasDefault) {
        this[name] = cloneDeep(_checked(prop, prop.getDefault()));
      } else {
        // property is optional
      }
    }
  }

  dispose() {
    // nothing by default
  }

  /**
    Check if the node is of a given type.

    @param {String} typeName
    @returns {Boolean} true if the node has a parent with given type, false otherwise.
  */
  isInstanceOf(typeName) {
    return Node.isInstanceOf(this.constructor, typeName)
  }

  getSchema() {
    return this.constructor.schema
  }

  /**
    Get a the list of all polymorphic types.

    @returns {String[]} An array of type names.
   */
  getTypeNames() {
    var typeNames = [];
    var NodeClass = this.constructor;
    while (NodeClass.type !== "node") {
      typeNames.push(NodeClass.type);
      NodeClass = Object.getPrototypeOf(NodeClass);
    }
    return typeNames
  }

  /**
   * Get the type of a property.
   *
   * @param {String} propertyName
   * @returns The property's type.
   */
  getPropertyType(propertyName) {
    var schema = this.constructor.schema;
    return schema[propertyName].type
  }

  /**
    Convert node to JSON.

    @returns {Object} JSON representation of node.
   */
  toJSON() {
    var data = {
      type: this.type
    };
    const schema = this.getSchema();
    forEach(schema, (prop, name) => {
      let val = this[name];
      if (prop.isOptional() && val === undefined) return
      if (isArray$1(val) || isObject$1(val)) {
        val = cloneDeep(val);
      }
      data[prop.name] = val;
    });
    return data
  }

  get type() {
    return this.constructor.type
  }

}

Node.prototype._isNode = true;

// NOTE: this code and its deps will always be included in the bundle as rollup considers this as global side-effect
Object.defineProperty(Node, 'schema', {
  get() { return this._schema },
  set(schema) {
    let NodeClass = this;
    // TODO: discuss if we want this. Is a bit more convenient
    // ATM we transfer 'type' to the static property
    if (schema.type) {
      NodeClass.type = schema.type;
    }
    // collects a full schema considering the schemas of parent class
    // we will use the unfolded schema, check integrity of the given props (mandatory, readonly)
    // or fill in default values for undefined properties.
    NodeClass._schema = compileSchema(NodeClass, schema);
  }
});

Node.define = Node.defineSchema = function define(schema) {
  this.schema = schema;
};

Node.schema = {
  type: "node",
  id: 'string'
};

/**
  Internal implementation of Node.prototype.isInstanceOf.

  @static
  @private
  @returns {Boolean}
 */
Node.isInstanceOf = function(NodeClass, typeName) {
  var type = NodeClass.type;
  while (type !== "node") {
    if (type === typeName) return true
    var _super = Object.getPrototypeOf(NodeClass.prototype).constructor;
    if (_super && _super.type) {
      NodeClass = _super;
      type = NodeClass.type;
    } else {
      break
    }
  }
  return false
};

// ### Internal implementation

function compileSchema(NodeClass, schema) {
  let compiledSchema = _compileSchema(schema);
  let schemas = [compiledSchema];
  let clazz = NodeClass;
  while(clazz) {
    var parentProto = Object.getPrototypeOf(clazz.prototype);
    if (!parentProto) {
      break
    }
    clazz = parentProto.constructor;
    if (clazz && clazz._schema) {
      schemas.unshift(clazz._schema);
    }
  }
  schemas.unshift({});
  return Object.assign.apply(null, schemas)
}

function _compileSchema(schema) {
  let compiledSchema = {};
  forEach(schema, function(definition, name) {
    // skip 'type'
    if (name === 'type') {
      return
    }
    if (isString$1(definition) || isArray$1(definition)) {
      definition = { type: definition };
    }
    definition = _compileDefintion(definition);
    definition.name = name;
    compiledSchema[name] = new Property(definition);
  });
  return compiledSchema
}

function _compileDefintion(definition) {
  let result = definition;
  if (isArray$1(definition.type) && definition.type[0] !== "array") {
    definition.type = [ "array", definition.type[0] ];
  } else if (definition.type === 'text') {
    result = {
      type: "string",
      default: '',
      _isText: true
    };
  }
  return result
}

function _checked(prop, value) {
  let type;
  let name = prop.name;
  if (prop.isArray()) {
    type = "array";
  } else {
    type = prop.type;
  }
  if (value === null) {
    if (prop.isNotNull()) {
      throw new Error('Value for property ' + name + ' is null.')
    } else {
      return value
    }
  }
  if (value === undefined) {
    throw new Error('Value for property ' + name + ' is undefined.')
  }
  if (type === "string" && !isString$1(value) ||
      type === "boolean" && !isBoolean(value) ||
      type === "number" && !isNumber(value) ||
      type === "array" && !isArray$1(value) ||
      type === "id" && !isString$1(value) ||
      type === "object" && !isObject$1(value)) {
    throw new Error('Illegal value type for property ' + name + ': expected ' + type + ', was ' + (typeof value))
  }
  return value
}

/**
  Base node type for document nodes.

  @example

  The following example shows how a new node type is defined.

  ```js
  class Todo extends TextBlock {}
  Todo.schema = {
    type: 'todo',
    content: 'text',
    done: { type: 'bool', default: false }
  }
  ```

  The following data types are supported:

  - `string` bare metal string data type
  - `text` a string that carries annotations
  - `number` numeric values
  - `bool` boolean values
  - `id` a node id referencing another node in the document
*/
class DocumentNode extends Node {

  /**
    @param {Document} doc A document instance
    @param {object} node properties
  */
  constructor(doc, props) {
    super(doc, props);
  }

  _initialize(doc, props) {
    this.document = doc;
    super._initialize(props);
  }

  /**
    Get the Document instance.

    @returns {Document}
  */
  getDocument() {
    return this.document
  }

  /**
    Whether this node has a parent.

    `parent` is a built-in property for implementing nested nodes.

    @returns {Boolean}
  */
  hasParent() {
    return Boolean(this.parent)
  }

  /**
    @returns {DocumentNode} the parent node
  */
  getParent() {
    return this.document.get(this.parent)
  }

  /**
    Get the root node.

    The root node is the last ancestor returned
    by a sequence of `getParent()` calls.

    @returns {DocumentNode}
  */
  getRoot() {
    let node = this;
    while(node.parent) {
      node = node.parent;
    }
    return node
  }

  /**
    Checks whether this node has children.

    @returns {Boolean} default: false
  */
  hasChildren() {
    return false
  }

  /**
    Get the index of a given child.

    @returns {Number} default: -1
  */
  getChildIndex(child) { // eslint-disable-line
    return -1
  }

  /**
    Get a child node at a given position.

    @returns {DocumentNode} default: null
  */
  getChildAt(idx) { // eslint-disable-line
    return null
  }

  /**
    Get the number of children nodes.

    @returns {Number} default: 0
  */
  getChildCount() {
    return 0
  }

  // TODO: should this really be here?
  // volatile property necessary to render highlighted node differently
  // TODO: We should get this out here
  setHighlighted(highlighted, scope) {
    if (this.highlighted !== highlighted) {
      this.highlightedScope = scope;
      this.highlighted = highlighted;
      this.emit('highlighted', highlighted);
    }
  }

  // Experimental: we are working on a simpler API replacing the
  // rather inconvenient EventProxy API.
  on(eventName, handler, ctx) {
    var match = _matchPropertyEvent(eventName);
    if (match) {
      var propertyName = match[1];
      if (this.constructor.schema[propertyName]) {
        var doc = this.getDocument();
        doc.getEventProxy('path')
          .on([this.id, propertyName], handler, ctx);
      }
    }
    EventEmitter.prototype.on.apply(this, arguments);
  }

  off(ctx, eventName, handler) {
    var doc = this.getDocument();
    var match = false;
    if (!eventName) {
      doc.getEventProxy('path').off(ctx);
    } else {
      match = _matchPropertyEvent(eventName);
    }
    if (match) {
      var propertyName = match[1];
      doc.getEventProxy('path')
        .off(ctx, [this.id, propertyName], handler);
    }
    EventEmitter.prototype.off.apply(this, arguments);
  }

  _onPropertyChange(propertyName) {
    var args = [propertyName + ':changed']
      .concat(Array.prototype.slice.call(arguments, 1));
    this.emit.apply(this, args);
  }

  // Node categories
  // --------------------

  /**
    @returns {Boolean} true if node is a block node (e.g. Paragraph, Figure, List, Table)
  */
  isBlock() {
    return Boolean(this.constructor.isBlock)
  }

  /**
    @returns {Boolean} true if node is a text node (e.g. Paragraph, Codebock)
  */
  isText() {
    return Boolean(this.constructor.isText)
  }

  /**
    @returns {Boolean} true if node is an inline node (e.g. Citation)
  */
  isInline() {
    return Boolean(this.constructor.isInline)
  }

  isList() {
    return Boolean(this.constructor.isList)
  }

  isIsolatedNode() {
    return !this.isText() && !this.isList()
  }

}

DocumentNode.prototype._isDocumentNode = true;

/**
  Declares a node to be treated as block-type node.

  BlockNodes are considers the direct descendant of `Container` nodes.
  @type {Boolean} default: false
*/
DocumentNode.isBlock = false;

/**
  Declares a node to be treated as text-ish node.

  @type {Boolean} default: false
*/
DocumentNode.isText = false;

/**
  Declares a node to be treated as {@link model/PropertyAnnotation}.

  @type {Boolean} default: false
*/
DocumentNode.isPropertyAnnotation = false;

/**
  Declares a node to be treated as {@link model/ContainerAnnotation}.

  @type {Boolean} default: false
*/
DocumentNode.isContainerAnnotation = false;

/**
  Declares a node to be treated as {@link model/InlineNode}.

  @type {Boolean} default: false
*/
DocumentNode.isInline = false;

function _matchPropertyEvent(eventName) {
  return /([a-zA-Z_0-9]+):changed/.exec(eventName)
}

/**
  Index for Nodes.

  Node indexes are first-class citizens in {@link model/data/Data}.
  I.e., they are updated after each operation, and before any other listener is notified.

  @abstract

 */
class NodeIndex {

  /**
    Check if a node should be indexed.

    Used internally only. Override this in subclasses to achieve a custom behavior.

    @private
    @param {Node}
    @returns {Boolean} true if the given node should be added to the index.
   */
  select(node) { // eslint-disable-line no-unused-vars
    throw new Error('This method is abstract.')
  }

  /**
    Called when a node has been created.

    @param {Node} node
   */
  create(node) { // eslint-disable-line no-unused-vars
    throw new Error('This method is abstract.')
  }

  /**
    Called when a node has been deleted.

    @param {model/data/Node} node
   */
  delete(node) { // eslint-disable-line no-unused-vars
    throw new Error('This method is abstract.')
  }

  set(node, path, newValue, oldValue) {
    this.update(node, path, newValue, oldValue);
  }

  /**
    Called when a property has been updated.

    @private
    @param {Node} node
   */
  update(node, path, newValue, oldValue) { // eslint-disable-line no-unused-vars
    throw new Error('This method is abstract.')
  }

  /**
    Reset the index using a Data instance.

    @private
   */
  reset(data) {
    this._clear();
    this._initialize(data);
  }

  /**
    Clone this index.

    @return A cloned NodeIndex.
   */
  clone() {
    var NodeIndexClass = this.constructor;
    var clone$$1 = new NodeIndexClass();
    return clone$$1
  }

  _clear() {
    throw new Error('This method is abstract')
  }

  _initialize(data) {
    forEach(data.getNodes(), function(node) {
      if (this.select(node)) {
        this.create(node);
      }
    }.bind(this));
  }
}

/**
  Create a new NodeIndex using the given prototype as mixin.

  @param {Object} prototype
  @returns {NodeIndex} A customized NodeIndex.
 */
NodeIndex.create = function(prototype) {
  var index = Object.assign(new NodeIndex(), prototype);
  index.clone = function() {
    return NodeIndex.create(prototype)
  };
  return index
};

/**
  Create a filter to filter nodes by type.

  @param {String} type
  @returns {function}
 */
NodeIndex.filterByType = function(type) {
  return function(node) {
    return node.isInstanceOf(type)
  }
};

class DocumentIndex extends NodeIndex {}

// A collection of methods to update annotations
// --------
//
// As we treat annotations as overlay of plain text we need to keep them up-to-date during editing.

var annotationHelpers = {
  insertedText,
  deletedText,
  transferAnnotations,
  expandAnnotation,
  fuseAnnotation,
  truncateAnnotation
};

function insertedText(doc, coordinate, length) {
  if (!length) return;
  var index = doc.getIndex('annotations');
  var annotations = index.get(coordinate.path);
  for (let i = 0; i < annotations.length; i++) {
    let anno = annotations[i];
    var pos = coordinate.offset;
    var start = anno.start.offset;
    var end = anno.end.offset;
    var newStart = start;
    var newEnd = end;
    if ( (pos < start) ||
         (pos === start) ) {
      newStart += length;
    }
    // inline nodes do not expand automatically
    if ( (pos < end) ||
         (pos === end && !anno.isInline()) ) {
      newEnd += length;
    }
    // TODO: Use coordintate ops!
    if (newStart !== start) {
      doc.set([anno.id, 'start', 'offset'], newStart);
    }
    if (newEnd !== end) {
      doc.set([anno.id, 'end', 'offset'], newEnd);
    }
  }

  // TODO: fix support for container annotations
  // // same for container annotation anchors
  // index = doc.getIndex('container-annotation-anchors');
  // var anchors = index.get(coordinate.path);
  // forEach(anchors, function(anchor) {
  //   var pos = coordinate.offset;
  //   var start = anchor.offset;
  //   var changed = false;
  //   if ( (pos < start) ||
  //        (pos === start && !coordinate.after) ) {
  //     start += length;
  //     changed = true;
  //   }
  //   if (changed) {
  //     let coor = (anchor.isStart?'start':'end');
  //     // TODO: Use coordintate ops!
  //     doc.set([anchor.id, coor, 'offset'], start);
  //   }
  // });
}

function deletedText(doc, path, startOffset, endOffset) {
  if (startOffset === endOffset) return;
  var index = doc.getIndex('annotations');
  var annotations = index.get(path);
  var length = endOffset - startOffset;
  for (let i = 0; i < annotations.length; i++) {
    let anno = annotations[i];
    var pos1 = startOffset;
    var pos2 = endOffset;
    var start = anno.start.offset;
    var end = anno.end.offset;
    var newStart = start;
    var newEnd = end;
    if (pos2 <= start) {
      newStart -= length;
      newEnd -= length;
      doc.set([anno.id, 'start', 'offset'], newStart);
      doc.set([anno.id, 'end', 'offset'], newEnd);
    } else {
      if (pos1 <= start) {
        newStart = start - Math.min(pos2-pos1, start-pos1);
      }
      if (pos1 <= end) {
        newEnd = end - Math.min(pos2-pos1, end-pos1);
      }
      // delete the annotation if it has collapsed by this delete
      if (start !== end && newStart === newEnd) {
        doc.delete(anno.id);
      } else {
        // TODO: Use coordintate ops!
        if (start !== newStart) {
          doc.set([anno.id, 'start', 'offset'], newStart);
        }
        if (end !== newEnd) {
          doc.set([anno.id, 'end', 'offset'], newEnd);
        }
      }
    }
  }
  // TODO: fix support for container annotations
  // // same for container annotation anchors
  // index = doc.getIndex('container-annotation-anchors');
  // var anchors = index.get(path);
  // var containerAnnoIds = [];
  // forEach(anchors, function(anchor) {
  //   containerAnnoIds.push(anchor.id);
  //   var pos1 = startOffset;
  //   var pos2 = endOffset;
  //   var start = anchor.offset;
  //   var changed = false;
  //   if (pos2 <= start) {
  //     start -= length;
  //     changed = true;
  //   } else {
  //     if (pos1 <= start) {
  //       var newStart = start - Math.min(pos2-pos1, start-pos1);
  //       if (start !== newStart) {
  //         start = newStart;
  //         changed = true;
  //       }
  //     }
  //   }
  //   if (changed) {
  //     // TODO: Use coordintate ops!
  //     let coor = (anchor.isStart?'start':'end');
  //     doc.set([anchor.id, coor, 'offset'], start);
  //   }
  // });
  // // check all anchors after that if they have collapsed and remove the annotation in that case
  // forEach(uniq(containerAnnoIds), function(id) {
  //   var anno = doc.get(id);
  //   var annoSel = anno.getSelection();
  //   if(annoSel.isCollapsed()) {
  //     // console.log("...deleting container annotation because it has collapsed" + id);
  //     doc.delete(id);
  //   }
  // });
}

// used when breaking a node to transfer annotations to the new property
function transferAnnotations(doc, path, offset, newPath, newOffset) {
  var index = doc.getIndex('annotations');
  var annotations = index.get(path, offset);
  for (let i = 0; i < annotations.length; i++) {
    let a = annotations[i];
    var isInside = (offset > a.start.offset && offset < a.end.offset);
    var start = a.start.offset;
    var end = a.end.offset;
    // 1. if the cursor is inside an annotation it gets either split or truncated
    if (isInside) {
      // create a new annotation if the annotation is splittable
      if (a.canSplit()) {
        let newAnno = a.toJSON();
        newAnno.id = uuid(a.type + "_");
        newAnno.start.path = newPath;
        newAnno.start.offset = newOffset;
        newAnno.end.path = newPath;
        newAnno.end.offset = newOffset + a.end.offset - offset;
        doc.create(newAnno);
      }
      // in either cases truncate the first part
      let newStartOffset = a.start.offset;
      let newEndOffset = offset;
      // if after truncate the anno is empty, delete it
      if (newEndOffset === newStartOffset) {
        doc.delete(a.id);
      }
      // ... otherwise update the range
      else {
        // TODO: Use coordintate ops!
        if (newStartOffset !== start) {
          doc.set([a.id, 'start', 'offset'], newStartOffset);
        }
        if (newEndOffset !== end) {
          doc.set([a.id, 'end', 'offset'], newEndOffset);
        }
      }
    }
    // 2. if the cursor is before an annotation then simply transfer the annotation to the new node
    else if (a.start.offset >= offset) {
      // TODO: Use coordintate ops!
      // Note: we are preserving the annotation so that anything which is connected to the annotation
      // remains valid.
      doc.set([a.id, 'start', 'path'], newPath);
      doc.set([a.id, 'start', 'offset'], newOffset + a.start.offset - offset);
      doc.set([a.id, 'end', 'path'], newPath);
      doc.set([a.id, 'end', 'offset'], newOffset + a.end.offset - offset);
    }
  }

  // TODO: fix support for container annotations
  // // same for container annotation anchors
  // index = doc.getIndex('container-annotation-anchors');
  // var anchors = index.get(path);
  // var containerAnnoIds = [];
  // forEach(anchors, function(anchor) {
  //   containerAnnoIds.push(anchor.id);
  //   var start = anchor.offset;
  //   if (offset <= start) {
  //     // TODO: Use coordintate ops!
  //     let coor = anchor.isStart?'start':'end'
  //     doc.set([anchor.id, coor, 'path'], newPath);
  //     doc.set([anchor.id, coor, 'offset'], newOffset + anchor.offset - offset);
  //   }
  // });
  // // check all anchors after that if they have collapsed and remove the annotation in that case
  // forEach(uniq(containerAnnoIds), function(id) {
  //   var anno = doc.get(id);
  //   var annoSel = anno.getSelection();
  //   if(annoSel.isCollapsed()) {
  //     // console.log("...deleting container annotation because it has collapsed" + id);
  //     doc.delete(id);
  //   }
  // });
}

/*
 @param {model/Document} tx
 @param {model/PropertyAnnotation} args.anno annotation which should be expanded
 @param {model/Selection}  args.selection selection to which to expand
*/
function truncateAnnotation(tx, anno, sel) {
  if (!sel || !sel._isSelection) throw new Error('Argument "selection" is required.')
  if (!anno || !anno._isAnnotation) throw new Error('Argument "anno" is required.')
  let annoSel = anno.getSelection();
  let newAnnoSel = annoSel.truncateWith(sel);
  anno._updateRange(tx, newAnnoSel);
  return anno
}

/*
 @param {model/Document} tx
 @param {model/PropertyAnnotation} args.anno annotation which should be expanded
 @param {model/Selection}  args.selection selection to which to expand
*/
function expandAnnotation(tx, anno, sel) {
  if (!sel || !sel._isSelection) throw new Error('Argument "selection" is required.')
  if (!anno || !anno._isAnnotation) throw new Error('Argument "anno" is required.')
  let annoSel = anno.getSelection();
  let newAnnoSel = annoSel.expand(sel);
  anno._updateRange(tx, newAnnoSel);
  return anno
}

/*
 @param {model/Document} tx
 @param {model/PropertyAnnotation[]} args.annos annotations which should be fused
*/
function fuseAnnotation(tx, annos) {
  if (!isArray$1(annos) || annos.length < 2) {
    throw new Error('fuseAnnotation(): at least two annotations are necessary.')
  }
  let sel, annoType;
  annos.forEach(function(anno, idx) {
    if (idx === 0) {
      sel = anno.getSelection();
      annoType = anno.type;
    } else {
      if (anno.type !== annoType) {
        throw new Error('fuseAnnotation(): all annotations must be of the same type.')
      }
      sel = sel.expand(anno.getSelection());
    }
  });
  // expand the first and delete the others
  for (var i = 1; i < annos.length; i++) {
    tx.delete(annos[i].id);
  }
  expandAnnotation(tx, annos[0], sel);
  tx.setSelection(sel);
}

class PropertyIndex extends NodeIndex {

  constructor(property) {
    super();

    this._property = property || 'id';
    this.index = new TreeIndex();
  }

  /**
    Get all indexed nodes for a given path.

    @param {Array<String>} path
    @returns A node or an object with ids and nodes as values.
   */
  get(path) {
    return this.index.get(path) || {}
  }

  /**
    Collects nodes recursively.

    @returns An object with ids as keys and nodes as values.
   */
  getAll(path) {
    return this.index.getAll(path)
  }

  /**
    Check if a node should be indexed.

    Used internally only. Override this in subclasses to achieve a custom behavior.

    @private
    @param {Node}
    @returns {Boolean} true if the given node should be added to the index.
   */
  select(node) { // eslint-disable-line
    return true
  }

  /**
    Called when a node has been created.

    Override this in subclasses for customization.

    @private
    @param {Node} node
   */
  create(node) {
    var values = node[this._property];
    if (!isArray$1(values)) {
      values = [values];
    }
    forEach(values, function(value) {
      this.index.set([value, node.id], node);
    }.bind(this));
  }

  /**
   * Called when a node has been deleted.
   *
   * Override this in subclasses for customization.
   *
   * @private
   * @param {model/data/Node} node
   */
  delete(node) {
    var values = node[this._property];
    if (!isArray$1(values)) {
      values = [values];
    }
    forEach(values, function(value) {
      this.index.delete([value, node.id]);
    }.bind(this));
  }

  /**
    Called when a property has been updated.

    Override this in subclasses for customization.

    @private
    @param {Node} node
   */
  update(node, path, newValue, oldValue) {
    if (!this.select(node) || path[1] !== this._property) return
    var values = oldValue;
    if (!isArray$1(values)) {
      values = [values];
    }
    forEach(values, function(value) {
      this.index.delete([value, node.id]);
    }.bind(this));
    values = newValue;
    if (!isArray$1(values)) {
      values = [values];
    }
    forEach(values, function(value) {
      this.index.set([value, node.id], node);
    }.bind(this));
  }

  set(node, path, newValue, oldValue) {
    this.update(node, path, newValue, oldValue);
  }

  _clear() {
    this.index.clear();
  }

  _initialize(data) {
    forEach(data.getNodes(), function(node) {
      if (this.select(node)) {
        this.create(node);
      }
    }.bind(this));
  }
}

/*
  Index for Annotations.

  @example
  Lets us look up existing annotations by path and type

  To get all annotations for the content of a text node

    var aIndex = doc.annotationIndex
    aIndex.get(["text_1", "content"])

  You can also scope for a specific range

    aIndex.get(["text_1", "content"], 23, 45)
*/
class AnnotationIndex extends DocumentIndex {

  constructor() {
    super();

    this.byPath = new TreeIndex();
    this.byType = new TreeIndex();
  }

  select(node) {
    return Boolean(node._isPropertyAnnotation)
  }

  reset(data) {
    this.byPath.clear();
    this.byType.clear();
    this._initialize(data);
  }

  // TODO: use object interface? so we can combine filters (path and type)
  get(path, start, end, type) {
    var annotations;
    if (isString$1(path) || path.length === 1) {
      annotations = this.byPath.getAll(path) || {};
    } else {
      annotations = this.byPath.get(path);
    }
    annotations = map(annotations);
    if (isNumber(start)) {
      annotations = filter(annotations, AnnotationIndex.filterByRange(start, end));
    }
    if (type) {
      annotations = filter(annotations, DocumentIndex.filterByType(type));
    }
    return annotations
  }

  create(anno) {
    this.byType.set([anno.type, anno.id], anno);
    this.byPath.set(anno.start.path.concat([anno.id]), anno);
  }

  delete(anno) {
    this._delete(anno.type, anno.id, anno.start.path);
  }

  _delete(type, id, path) {
    this.byType.delete([type, id]);
    this.byPath.delete(path.concat([id]));
  }

  update(node, path, newValue, oldValue) {
    // TODO: this should better be a coordinate op
    if (this.select(node) && path[1] === 'start' && path[2] === "path") {
      this._delete(node.type, node.id, oldValue);
      this.create(node);
    }
  }
}

AnnotationIndex.filterByRange = function(start, end) {
  return function(anno) {
    var aStart = anno.start.offset;
    var aEnd = anno.end.offset;
    var overlap = (aEnd >= start);
    // Note: it is allowed to omit the end part
    if (isNumber(end)) {
      overlap = overlap && (aStart <= end);
    }
    return overlap
  }
};

class ContainerAnnotationIndex extends DocumentIndex {

  constructor() {
    super();
    this.byId = new TreeIndex();
  }

  select(node) {
    return Boolean(node._isContainerAnnotation)
  }

  reset(data) {
    this.byId.clear();
    this._initialize(data);
  }

  get(containerId, type) {
    var annotations = map(this.byId.get(containerId));
    if (isString$1(type)) {
      annotations = filter(annotations, DocumentIndex.filterByType);
    }
    return annotations
  }

  create(anno) {
    this.byId.set([anno.containerId, anno.id], anno);
  }

  delete(anno) {
    this.byId.delete([anno.containerId, anno.id]);
  }

  update(node, path, newValue, oldValue) { // eslint-disable-line
    // TODO should we support moving a container anno from one container to another?
  }

}

/*
  Specification:

  - create:
    ```
    'c <JSON.stringify(data)>'
    'c { id: "1123", type: "paragraph", content: ""}'
    ```
  - delete:
    ```
    'd <JSON.stringify(data)>'
    'd { id: "1123", type: "paragraph", content: ""}'
    ```
  - set a property
    ```
    's <property path> <value> <old value>'
    's p1.content foo'
    ```
  - update a property
    ```
    'u <property path> <primitive op>'
    'u p1.content t+ 4 foo'
    ```

Primitive type operations:

  - insert text
    ```
    't+ <pos> <string>'
    't+ 4 foo'
    ```
  - delete text
    ```
    't- <pos> <string>'
    't- 4 foo'
    ```
  - insert value into array
    ```
    'a+ <pos> <value>'
    'a+ 0 p1'
    ```
  - delete value from array
    ```
    'a- <pos> <value>'
    'a- 0 p1'
    ```
*/

class OperationSerializer{

  constructor() {
    this.SEPARATOR = '\t';
  }

  serialize(op) {
    var out = [];
    switch (op.type) {
      case 'create':
        out.push('c');
        out.push(op.val.id);
        out.push(op.val);
        break
      case 'delete':
        out.push('d');
        out.push(op.val.id);
        out.push(op.val);
        break
      case 'set':
        out.push('s');
        out.push(op.path.join('.'));
        out.push(op.val);
        out.push(op.original);
        break
      case 'update':
        out.push('u');
        out.push(op.path.join('.'));
        Array.prototype.push.apply(out, this.serializePrimitiveOp(op.diff));
        break
      default:
        throw new Error('Unsupported operation type.')
    }
    return out
  }

  serializePrimitiveOp(op) {
    var out = [];
    if (op._isTextOperation) {
      if (op.isInsert()) {
        out.push('t+');
      } else if (op.isDelete()) {
        out.push('t-');
      }
      out.push(op.pos);
      out.push(op.str);
    } else if (op._isArrayOperation) {
      if (op.isInsert()) {
        out.push('a+');
      } else if (op.isDelete()) {
        out.push('a-');
      }
      out.push(op.pos);
      out.push(op.val);
    } else if (op._isCoordinateOperation) {
      if (op.isShift()) {
        out.push('c>>');
      } else {
        throw new Error('Unsupported CoordinateOperation type.')
      }
      out.push(op.pos);
      out.push(op.val);
    } else {
      throw new Error('Unsupported operation type.')
    }
    return out
  }

  deserialize(str, tokenizer) {
    if (!tokenizer) {
      tokenizer = new Tokenizer(str, this.SEPARATOR);
    }
    var type = tokenizer.getString();
    var op, path, val, oldVal, diff$$1;
    switch (type) {
      case 'c':
        path = tokenizer.getPath();
        val = tokenizer.getObject();
        op = ObjectOperation.Create(path, val);
        break
      case 'd':
        path = tokenizer.getPath();
        val = tokenizer.getObject();
        op = ObjectOperation.Delete(path, val);
        break
      case 's':
        path = tokenizer.getPath();
        val = tokenizer.getAny();
        oldVal = tokenizer.getAny();
        op = ObjectOperation.Set(path, oldVal, val);
        break
      case 'u':
        path = tokenizer.getPath();
        diff$$1 = this.deserializePrimitiveOp(str, tokenizer);
        op = ObjectOperation.Update(path, diff$$1);
        break
      default:
        throw new Error('Illegal type for ObjectOperation: '+ type)
    }
    return op
  }

  deserializePrimitiveOp(str, tokenizer) {
    if (!tokenizer) {
      tokenizer = new Tokenizer(str, this.SEPARATOR);
    }
    var type = tokenizer.getString();
    var op, pos, val;
    switch (type) {
      case 't+':
        pos = tokenizer.getNumber();
        val = tokenizer.getString();
        op = TextOperation.Insert(pos, val);
        break
      case 't-':
        pos = tokenizer.getNumber();
        val = tokenizer.getString();
        op = TextOperation.Delete(pos, val);
        break
      case 'a+':
        pos = tokenizer.getNumber();
        val = tokenizer.getAny();
        op = ArrayOperation.Insert(pos, val);
        break
      case 'a-':
        pos = tokenizer.getNumber();
        val = tokenizer.getAny();
        op = ArrayOperation.Delete(pos, val);
        break
      case 'c>>':
        val = tokenizer.getNumber();
        op = CoordinateOperation.Shift(val);
        break
      default:
        throw new Error('Unsupported operation type: ' + type)
    }
    return op
  }
}

class Tokenizer {
  constructor(str, sep) {
    if (isArray$1(arguments[0])) {
      this.tokens = arguments[0];
    } else {
      this.tokens = str.split(sep);
    }
    this.pos = -1;
  }

  error(msg) {
    throw new Error('Parsing error: ' + msg + '\n' + this.tokens[this.pos])
  }

  getString() {
    this.pos++;
    var str = this.tokens[this.pos];
    if (str[0] === '"') {
      str = str.slice(1, -1);
    }
    return str
  }

  getNumber() {
    this.pos++;
    var number;
    var token = this.tokens[this.pos];
    try {
      if (isNumber(token)) {
        number = token;
      } else {
        number = parseInt(this.tokens[this.pos], 10);
      }
      return number
    } catch (err) {
      this.error('expected number');
    }
  }

  getObject() {
    this.pos++;
    var obj;
    var token = this.tokens[this.pos];
    try {
      if (isObject$1(token)) {
        obj = token;
      } else {
        obj = JSON.parse(this.tokens[this.pos]);
      }
      return obj
    } catch (err) {
      this.error('expected object');
    }
  }

  getAny() {
    this.pos++;
    var token = this.tokens[this.pos];
    return token
  }

  getPath() {
    var str = this.getString();
    return str.split('.')
  }
}

OperationSerializer.Tokenizer = Tokenizer;

class DocumentChange {

  constructor(ops, before, after) {
    if (arguments.length === 1 && isPlainObject$1(arguments[0])) {
      let data = arguments[0];
      // a unique id for the change
      this.sha = data.sha;
      // when the change has been applied
      this.timestamp = data.timestamp;
      // application state before the change was applied
      this.before = data.before || {};
      // array of operations
      this.ops = data.ops;
      this.info = data.info; // custom change info
      // application state after the change was applied
      this.after = data.after || {};
    } else if (arguments.length === 3) {
      this.sha = uuid();
      this.info = {};
      this.timestamp = Date.now();
      this.ops = ops.slice(0);
      this.before = before || {};
      this.after = after || {};
    } else {
      throw new Error('Illegal arguments.')
    }
    // a hash with all updated properties
    this.updated = null;
    // a hash with all created nodes
    this.created = null;
    // a hash with all deleted nodes
    this.deleted = null;
  }

  /*
    Extract aggregated information about which nodes and properties have been affected.
    This gets called by Document after applying the change.
  */
  _extractInformation(doc) {
    let ops = this.ops;
    let created = {};
    let deleted = {};
    let updated = {};
    let affectedContainerAnnos = [];

    // TODO: we will introduce a special operation type for coordinates
    function _checkAnnotation(op) {
      switch (op.type) {
        case "create":
        case "delete": {
          let node = op.val;
          if (node.hasOwnProperty('start')) {
            updated[node.start.path] = true;
          }
          if (node.hasOwnProperty('end')) {
            updated[node.end.path] = true;
          }
          break
        }
        case "update":
        case "set": {
          // HACK: detecting annotation changes in an opportunistic way
          let node = doc.get(op.path[0]);
          if (node) {
            if (node._isPropertyAnnotation) {
              updated[node.start.path] = true;
            } else if (node._isContainerAnnotation) {
              affectedContainerAnnos.push(node);
            }
          }
          break
        }
        default:
          /* istanbul ignore next */
          throw new Error('Illegal state')
      }
    }

    for (let i = 0; i < ops.length; i++) {
      let op = ops[i];
      if (op.type === "create") {
        created[op.val.id] = op.val;
        delete deleted[op.val.id];
      }
      if (op.type === "delete") {
        delete created[op.val.id];
        deleted[op.val.id] = op.val;
      }
      if (op.type === "set" || op.type === "update") {
        updated[op.path] = true;
        // also mark the node itself as dirty
        updated[op.path[0]] = true;
      }
      _checkAnnotation(op);
    }

    affectedContainerAnnos.forEach(function(anno) {
      let container = doc.get(anno.containerId, 'strict');
      let startPos = container.getPosition(anno.start.path[0]);
      let endPos = container.getPosition(anno.end.path[0]);
      for (let pos = startPos; pos <= endPos; pos++) {
        let node = container.getChildAt(pos);
        let path;
        if (node.isText()) {
          path = [node.id, 'content'];
        } else {
          path = [node.id];
        }
        if (!deleted[node.id]) {
          updated[path] = true;
        }
      }
    });

    // remove all deleted nodes from updated
    if(Object.keys(deleted).length > 0) {
      forEach(updated, function(_, key) {
        let nodeId = key.split(',')[0];
        if (deleted[nodeId]) {
          delete updated[key];
        }
      });
    }

    this.created = created;
    this.deleted = deleted;
    this.updated = updated;
  }

  invert() {
    // shallow cloning this
    let copy = this.toJSON();
    copy.ops = [];
    // swapping before and after
    let tmp = copy.before;
    copy.before = copy.after;
    copy.after = tmp;
    let inverted = DocumentChange.fromJSON(copy);
    let ops = [];
    for (let i = this.ops.length - 1; i >= 0; i--) {
      ops.push(this.ops[i].invert());
    }
    inverted.ops = ops;
    return inverted
  }

  /* istanbul ignore start */
  isAffected(path) {
    console.error('DEPRECATED: use change.hasUpdated() instead');
    return this.hasUpdated(path)
  }

  isUpdated(path) {
    console.error('DEPRECATED: use change.hasUpdated() instead');
    return this.hasUpdated(path)
  }
  /* istanbul ignore end */

  hasUpdated(path) {
    return this.updated[path]
  }

  hasDeleted(id) {
    return this.deleted[id]
  }

  serialize() {
    // TODO serializers and deserializers should allow
    // for application data in 'after' and 'before'

    let opSerializer = new OperationSerializer();
    let data = this.toJSON();
    data.ops = this.ops.map(function(op) {
      return opSerializer.serialize(op)
    });
    return JSON.stringify(data)
  }

  clone() {
    return DocumentChange.fromJSON(this.toJSON())
  }

  toJSON() {
    let data = {
      // to identify this change
      sha: this.sha,
      // before state
      before: clone(this.before),
      ops: map(this.ops, function(op) {
        return op.toJSON()
      }),
      info: this.info,
      // after state
      after: clone(this.after),
    };

    // Just to make sure rich selection objects don't end up
    // in the JSON result
    data.after.selection = undefined;
    data.before.selection = undefined;

    let sel = this.before.selection;
    if (sel && sel._isSelection) {
      data.before.selection = sel.toJSON();
    }
    sel = this.after.selection;
    if (sel && sel._isSelection) {
      data.after.selection = sel.toJSON();
    }
    return data
  }
}

DocumentChange.deserialize = function(str) {
  let opSerializer = new OperationSerializer();
  let data = JSON.parse(str);
  data.ops = data.ops.map(function(opData) {
    return opSerializer.deserialize(opData)
  });
  if (data.before.selection) {
    data.before.selection = fromJSON(data.before.selection);
  }
  if (data.after.selection) {
    data.after.selection = fromJSON(data.after.selection);
  }
  return new DocumentChange(data)
};

DocumentChange.fromJSON = function(data) {
  // Don't write to original object on deserialization
  let change = cloneDeep(data);
  change.ops = data.ops.map(function(opData) {
    return ObjectOperation.fromJSON(opData)
  });
  change.before.selection = fromJSON(data.before.selection);
  change.after.selection = fromJSON(data.after.selection);
  return new DocumentChange(change)
};

class PathEventProxy {

  constructor(doc) {
    this.listeners = new TreeIndex.Arrays();
    this._list = [];
    this.doc = doc;
  }

  on(path, method, context) {
    this._add(context, path, method);
  }

  // proxy.off(this)
  // proxy.off(this, path)
  // proxy.off(this, path, this.onPropertyUpdate)
  off(context, path, method) {
    this._remove(context, path, method);
  }

  connect(listener, path, method) {
    console.warn('DEPRECATED: use proxy.on(path, this.onPropertyChange, this) instead');
    this.on(path, method, listener);
  }

  disconnect(listener) {
    console.warn('DEPRECATED: use proxy.off(this) instead');
    this.off(listener);
  }

  onDocumentChanged(change, info, doc) {
    // stop if no listeners registered
    if (this._list.length === 0) {
      return
    }
    var listeners = this.listeners;
    forEach(change.updated, function(_, pathStr) {
      var scopedListeners = listeners.get(pathStr.split(','));
      if (isArray$1(scopedListeners)) scopedListeners = scopedListeners.slice(0);
      forEach(scopedListeners, function(entry) {
        entry.method.call(entry.listener, change, info, doc);
      });
    });
  }

  _add(listener, path, method) {
    if (!method) {
      throw new Error('Invalid argument: expected function but got ' + method)
    }
    var entry = { listener: listener, path: path, method: method };
    this.listeners.add(path, entry);
    this._list.push(entry);
  }

  _remove(listener, path, method) {
    for (var i = 0; i < this._list.length; i++) {
      var item = this._list[i];
      var match = (
        (!path || isEqual(item.path, path)) &&
        (!listener || item.listener === listener) &&
        (!method || item.method !== method)
      );
      if (match) {
        var entry = this._list[i];
        this._list.splice(i, 1);
        this.listeners.remove(entry.path, entry);
      }
    }
  }
}

/*
  A data storage implemention that supports data defined via a {@link Schema},
  and incremental updates which are backed by a OT library.

  It forms the underlying implementation for {@link Document}.
 */
class Data extends EventEmitter {

  /**
    @param {Schema} schema
    @param {Object} [options]
  */
  constructor(schema, nodeFactory) {
    super();

    /* istanbul ignore start */
    if (!schema) {
      throw new Error('schema is mandatory')
    }
    if (!nodeFactory) {
      throw new Error('nodeFactory is mandatory')
    }
   /* istanbul ignore end */

    this.schema = schema;
    this.nodeFactory = nodeFactory;
    this.nodes = {};
    this.indexes = {};

    // Sometimes necessary to resolve issues with updating indexes in presence
    // of cyclic dependencies
    this.__QUEUE_INDEXING__ = false;
    this.queue = [];
  }

  /**
    Check if this storage contains a node with given id.

    @returns {bool} `true` if a node with id exists, `false` otherwise.
   */
  contains(id) {
    return Boolean(this.nodes[id])
  }

  /**
    Get a node or value via path.

    @param {String|String[]} path node id or path to property.
    @returns {Node|Object|Primitive|undefined} a Node instance, a value or undefined if not found.
   */
  get(path, strict) {
    let result = this._get(path);
    if (strict && result === undefined) {
      if (isString$1(path)) {
        throw new Error("Could not find node with id '"+path+"'.")
      } else if (!this.contains(path[0])) {
        throw new Error("Could not find node with id '"+path[0]+"'.")
      } else {
        throw new Error("Property for path '"+path+"' us undefined.")
      }
    }
    return result
  }

  _get(path) {
    if (!path) return undefined
    let result;
    if (isString$1(path)) {
      result = this.nodes[path];
    } else if (path.length === 1) {
      result = this.nodes[path[0]];
    } else if (path.length > 1) {
      let context = this.nodes[path[0]];
      for (let i = 1; i < path.length-1; i++) {
        if (!context) return undefined
        context = context[path[i]];
      }
      if (!context) return undefined
      result = context[path[path.length-1]];
    }
    return result
  }

  /**
    Get the internal storage for nodes.

    @return The internal node storage.
   */
  getNodes() {
    return this.nodes
  }

  /**
    Create a node from the given data.

    @return {Node} The created node.
   */
  create(nodeData) {
    var node = this.nodeFactory.create(nodeData.type, nodeData);
    if (!node) {
      throw new Error('Illegal argument: could not create node for data:', nodeData)
    }
    if (this.contains(node.id)) {
      throw new Error("Node already exists: " + node.id)
    }
    if (!node.id || !node.type) {
      throw new Error("Node id and type are mandatory.")
    }
    this.nodes[node.id] = node;

    var change = {
      type: 'create',
      node: node
    };

    if (this.__QUEUE_INDEXING__) {
      this.queue.push(change);
    } else {
      this._updateIndexes(change);
    }

    return node
  }

  /**
    Delete the node with given id.

    @param {String} nodeId
    @returns {Node} The deleted node.
   */
  delete(nodeId) {
    var node = this.nodes[nodeId];
    if (!node) return
    node.dispose();
    delete this.nodes[nodeId];

    var change = {
      type: 'delete',
      node: node,
    };

    if (this.__QUEUE_INDEXING__) {
      this.queue.push(change);
    } else {
      this._updateIndexes(change);
    }

    return node
  }

  /**
    Set a property to a new value.

    @param {Array} property path
    @param {Object} newValue
    @returns {Node} The deleted node.
   */
  set(path, newValue) {
    let node = this.get(path[0]);
    let oldValue = this._set(path, newValue);
    var change = {
      type: 'set',
      node: node,
      path: path,
      newValue: newValue,
      oldValue: oldValue
    };
    if (this.__QUEUE_INDEXING__) {
      this.queue.push(change);
    } else {
      this._updateIndexes(change);
    }
    return oldValue
  }

  _set(path, newValue) {
    let oldValue = _setValue(this.nodes, path, newValue);
    return oldValue
  }

  /**
    Update a property incrementally.

    @param {Array} property path
    @param {Object} diff
    @returns {any} The value before applying the update.
  */
  update(path, diff$$1) {
    var realPath = this.getRealPath(path);
    if (!realPath) {
      console.error('Could not resolve path', path);
      return
    }
    let node = this.get(realPath[0]);
    let oldValue = this._get(realPath);
    let newValue;
    if (diff$$1.isOperation) {
      newValue = diff$$1.apply(oldValue);
    } else {
      diff$$1 = this._normalizeDiff(oldValue, diff$$1);
      if (isString$1(oldValue)) {
        switch (diff$$1.type) {
          case 'delete': {
            newValue = oldValue.split('').splice(diff$$1.start, diff$$1.end-diff$$1.start).join('');
            break
          }
          case 'insert': {
            newValue = [oldValue.substring(0, diff$$1.start), diff$$1.text, oldValue.substring(diff$$1.start)].join('');
            break
          }
          default:
            throw new Error('Unknown diff type')
        }
      } else if (isArray$1(oldValue)) {
        newValue = oldValue.slice(0);
        switch (diff$$1.type) {
          case 'delete': {
            newValue.splice(diff$$1.pos, 1);
            break
          }
          case 'insert': {
            newValue.splice(diff$$1.pos, 0, diff$$1.value);
            break
          }
          default:
            throw new Error('Unknown diff type')
        }
      } else if (oldValue._isCoordinate) {
        switch (diff$$1.type) {
          case 'shift': {
            // ATTENTION: in this case we do not want to create a new value
            oldValue = { path: oldValue.path, offset: oldValue.offset };
            newValue = oldValue;
            newValue.offset += diff$$1.value;
            break
          }
          default:
            throw new Error('Unknown diff type')
        }
      } else {
        throw new Error('Diff is not supported:', JSON.stringify(diff$$1))
      }
    }
    this._set(realPath, newValue);

    var change = {
      type: 'update',
      node: node,
      path: realPath,
      newValue: newValue,
      oldValue: oldValue
    };

    if (this.__QUEUE_INDEXING__) {
      this.queue.push(change);
    } else {
      this._updateIndexes(change);
    }

    return oldValue
  }

  // normalize to support legacy formats
  _normalizeDiff(value, diff$$1) {
    if (isString$1(value)) {
      // legacy
      if (diff$$1['delete']) {
        console.warn('DEPRECATED: use doc.update(path, {type:"delete", start:s, end: e}) instead');
        diff$$1 = {
          type: 'delete',
          start: diff$$1['delete'].start,
          end: diff$$1['delete'].end
        };
      } else if (diff$$1['insert']) {
        console.warn('DEPRECATED: use doc.update(path, {type:"insert", start:s, text: t}) instead');
        diff$$1 = {
          type: 'insert',
          start: diff$$1['insert'].offset,
          text: diff$$1['insert'].value
        };
      }
    } else if (isArray$1(value)) {
      // legacy
      if (diff$$1['delete']) {
        console.warn('DEPRECATED: use doc.update(path, {type:"delete", pos:1}) instead');
        diff$$1 = {
          type: 'delete',
          pos: diff$$1['delete'].offset
        };
      } else if (diff$$1['insert']) {
        console.warn('DEPRECATED: use doc.update(path, {type:"insert", pos:1, value: "foo"}) instead');
        diff$$1 = {
          type: 'insert',
          pos: diff$$1['insert'].offset,
          value: diff$$1['insert'].value
        };
      }
    } else if (value._isCoordinate) {
      if (diff$$1.hasOwnProperty('shift')) {
        console.warn('DEPRECATED: use doc.update(path, {type:"shift", value:2}) instead');
        diff$$1 = {
          type: 'shift',
          value: diff$$1['shift']
        };
      }
    }
    return diff$$1
  }

  /*
    DEPRECATED: We moved away from having JSON as first-class exchange format.
    We will remove this soon.

    @internal
    @deprecated
   */
  toJSON() {
    let nodes = {};
    forEach(this.nodes, (node)=>{
      nodes[node.id] = node.toJSON();
    });
    return {
      schema: [this.schema.id, this.schema.version],
      nodes: nodes
    }
  }

  /**
    Clear nodes.

    @internal
   */
  reset() {
    this.nodes.clear();
  }

  /**
    Add a node index.

    @param {String} name
    @param {NodeIndex} index
   */
  addIndex(name, index) {
    if (this.indexes[name]) {
      console.error('Index with name %s already exists.', name);
    }
    index.reset(this);
    this.indexes[name] = index;
    return index
  }

  /**
    Get the node index with given name.

    @param {String} name
    @returns {NodeIndex} The node index.
   */
  getIndex(name) {
    return this.indexes[name]
  }

  /**
    Update a node index by providing of change object.

    @param {Object} change
   */
  _updateIndexes(change) {
    if (!change || this.__QUEUE_INDEXING__) return
    forEach(this.indexes, function(index) {
      if (index.select(change.node)) {
        if (!index[change.type]) {
          console.error('Contract: every NodeIndex must implement ' + change.type);
        }
        index[change.type](change.node, change.path, change.newValue, change.oldValue);
      }
    });
  }

  /**
    Stops indexing process, all changes will be collected in indexing queue.

    @private
  */
  _stopIndexing() {
    this.__QUEUE_INDEXING__ = true;
  }

  /**
    Update all index changes from indexing queue.

    @private
  */
  _startIndexing() {
    this.__QUEUE_INDEXING__ = false;
    while(this.queue.length >0) {
      var change = this.queue.shift();
      this._updateIndexes(change);
    }
  }

}

function _setValue(root, path, newValue) {
  let ctx = root;
  let L = path.length;
  for (let i = 0; i < L-1; i++) {
    ctx = ctx[path[i]];
    if (!ctx) throw new Error('Can not set value.')
  }
  let oldValue = ctx[path[L-1]];
  ctx[path[L-1]] = newValue;
  return oldValue
}

/**
  Incremental data storage implemention.

  @internal
 */
class IncrementalData extends Data {

  /**
    Create a new node.

    @param {object} nodeData
    @returns {ObjectOperation} The applied operation.
   */
  create(nodeData) {
    if (nodeData._isNode) {
      nodeData = nodeData.toJSON();
    }
    let op = ObjectOperation.Create([nodeData.id], nodeData);
    this.apply(op);
    return op
  }

  /**
    Delete a node.

    @param {String} nodeId
    @returns {ObjectOperation} The applied operation.
   */
  delete(nodeId) {
    var op = null;
    var node = this.get(nodeId);
    if (node) {
      var nodeData = node.toJSON();
      op = ObjectOperation.Delete([nodeId], nodeData);
      this.apply(op);
    }
    return op
  }

  /**
    Update a property incrementally.

    The diff can be of the following forms (depending on the updated property type):
      - String:
        - `{ insert: { offset: Number, value: Object } }`
        - `{ delete: { start: Number, end: Number } }`
      - Array:
        - `{ insert: { offset: Number, value: Object } }`
        - `{ delete: { offset: Number } }`

    @param {array} path
    @param {object} diff
    @returns {ObjectOperation} The applied operation.
  */
  update(path, diff$$1) {
    var diffOp = this._getDiffOp(path, diff$$1);
    var op = ObjectOperation.Update(path, diffOp);
    this.apply(op);
    return op
  }

  /**
    Set a property to a new value

    @param {Array} path
    @param {Object} newValue
    @returns {ObjectOperation} The applied operation.
   */
  set(path, newValue) {
    var oldValue = this.get(path);
    var op = ObjectOperation.Set(path, oldValue, newValue);
    this.apply(op);
    return op
  }

  /**
    Apply a given operation.

    @param {ObjectOperation} op
   */
  apply(op) {
    if (op.type === ObjectOperation.NOP) return
    else if (op.type === ObjectOperation.CREATE) {
      // clone here as the operations value must not be changed
      super.create(cloneDeep(op.val));
    } else if (op.type === ObjectOperation.DELETE) {
      super.delete(op.val.id);
    } else if (op.type === ObjectOperation.UPDATE) {
      var oldVal = this.get(op.path);
      var diff$$1 = op.diff;
      if (op.propertyType === 'array') {
        if (! (diff$$1._isArrayOperation) ) {
          diff$$1 = ArrayOperation.fromJSON(diff$$1);
        }
        // array ops work inplace
        diff$$1.apply(oldVal);
      } else if (op.propertyType === 'string') {
        if (!(diff$$1._isTextOperation) ) {
          diff$$1 = TextOperation.fromJSON(diff$$1);
        }
        var newVal = diff$$1.apply(oldVal);
        super.set(op.path, newVal);
      } else if (op.propertyType === 'coordinate') {
        if (!(diff$$1._isCoordinateOperation) ) {
          diff$$1 = CoordinateOperation.fromJSON(diff$$1);
        }
        diff$$1.apply(oldVal);
      } else {
        throw new Error("Unsupported type for operational update.")
      }
    } else if (op.type === ObjectOperation.SET) {
      super.set(op.path, op.val);
    } else {
      throw new Error("Illegal state.")
    }
    this.emit('operation:applied', op, this);
  }

  /**
    Creates proper operation based on provided node path and diff.

    @param {Array} path
    @param {Object} diff
    @returns {ObjectOperation} operation.
    @private
  */
  _getDiffOp(path, diff$$1) {
    var diffOp = null;
    if (diff$$1.isOperation) {
      diffOp = diff$$1;
    } else {
      var value = this.get(path);
      diff$$1 = this._normalizeDiff(value, diff$$1);
      if (value === null || value === undefined) {
        throw new Error('Property has not been initialized: ' + JSON.stringify(path))
      } else if (isString$1(value)) {
        switch (diff$$1.type) {
          case 'delete': {
            diffOp = TextOperation.Delete(diff$$1.start, value.substring(diff$$1.start, diff$$1.end));
            break
          }
          case 'insert': {
            diffOp = TextOperation.Insert(diff$$1.start, diff$$1.text);
            break
          }
          default:
            throw new Error('Unknown diff type')
        }
      } else if (isArray$1(value)) {
        switch (diff$$1.type) {
          case 'delete': {
            diffOp = ArrayOperation.Delete(diff$$1.pos, value[diff$$1.pos]);
            break
          }
          case 'insert': {
            diffOp = ArrayOperation.Insert(diff$$1.pos, diff$$1.value);
            break
          }
          default:
            throw new Error('Unknown diff type')
        }
      } else if (value._isCoordinate) {
        switch (diff$$1.type) {
          case 'shift': {
            diffOp = CoordinateOperation.Shift(diff$$1.value);
            break
          }
          default:
            throw new Error('Unknown diff type')
        }
      }
    }
    if (!diffOp) {
      throw new Error('Unsupported diff: ' + JSON.stringify(diff$$1))
    }
    return diffOp
  }

}

class DocumentNodeFactory {

  constructor(doc) {
    this.doc = doc;
  }

  create(nodeType, nodeData) {
    var NodeClass = this.doc.schema.getNodeClass(nodeType);
    if (!NodeClass) {
      throw new Error('No node registered by that name: ' + nodeType)
    }
    return new NodeClass(this.doc, nodeData)
  }
}

class JSONConverter {

  importDocument(doc, json) {
    if (!json.nodes) {
      throw new Error('Invalid JSON format.')
    }
    var schema = doc.getSchema();
    if (json.schema && schema.name !== json.schema.name) {
      throw new Error('Incompatible schema.')
    }
    // the json should just be an array of nodes
    var nodes = json.nodes;
    // import data in a block with deactivated indexers and listeners
    // as the data contains cyclic references which
    // cause problems.
    doc.import(function(tx) {
      forEach(nodes, function(node) {
        // overwrite existing nodes
        if (tx.get(node.id)) {
          tx.delete(node.id);
        }
        tx.create(node);
      });
    });
    return doc
  }

  exportDocument(doc) {
    var schema = doc.getSchema();
    var json = {
      schema: {
        name: schema.name
      },
      nodes: {}
    };
    forEach(doc.getNodes(), function(node) {
      if (node._isDocumentNode) {
        json.nodes[node.id] = node.toJSON();
      }
    });
    return json
  }
}

/*
  This is an experiment trying to have better support for data types with a hierarchical
  nature, such as Lists, Tables etc.

  Our data model is inherently flat, and does not have any special support for hierarchical data types.
  The flat data model is essential for a simple OT implementation. Hierarchy is achieved by storing ids
  to reference child nodes.

  After longer discussions we agreed that we are very happy with the overlay nature
  of annotations, i.e., text is essentially modelled as plain text, and annotations are attached to it.
  This way we can map complex text manipulation to primitive object operations.
  For many other content types it would often be extremly helpful being able to traverse the structure in both
  directions, from parent to children (which is already possible), and back from children to parent.

  We do not want to store a the id of a parent node into the children, as this would be redundant, and would increase the amount of necessary operations.
  Instead we want to establish a link dynamically on the Node instance when the id is set in the parent (during construction or when updated).

  The most in-obstrusive implementation is to add an 'operation:applied' hook, watching for such changes
  and setting the reference. First we will apply this only for specific node types.
  Later this will be derived from the schema.
  With Texture we want to investigate a further option: replacing the node model with a DOM.
*/

class ParentNodeHook {

  constructor(doc) {
    this.doc = doc;
    this.table = {};
    doc.data.on('operation:applied', this._onOperationApplied, this);
  }

  _onOperationApplied(op) {
    const doc = this.doc;
    const table = this.table;
    let node = doc.get(op.path[0]);
    // TODO: instead of hard coding this here we should compile a matcher
    // based on the document schema
    switch(op.type) {
      case 'create': {
        switch (node.type) {
          case 'list':
            _setParent(node, node.items);
            break
          case 'list-item': {
            _setRegisteredParent(node);
            break
          }
          case 'table':
            _setParent(node, node.cells);
            break
          case 'table-cell': {
            _setRegisteredParent(node);
            break
          }
          default:
            //
        }
        break
      }
      case 'update': {
        // ATTENTION: we only set parents but don't remove when they are deleted
        // assuming that if the parent gets deleted, the children get deleted too
        let update = op.diff;
        switch(node.type) {
          case 'list':
            if (op.path[1] === 'items') {
              if (update.isInsert()) {
                _setParent(node, update.getValue());
              }
            }
            break
          case 'table':
            if (op.path[1] === 'cells') {
              if (update.isInsert()) {
                _setParent(node, update.getValue());
              }
            }
            break
          default:
            //
        }
        break
      }
      case 'set': {
        switch(node.type) {
          case 'list':
            if (op.path[1] === 'items') {
              _setParent(node, op.getValue());
            }
            break
          case 'table':
            if (op.path[1] === 'cells') {
              _setParent(node, op.getValue());
            }
            break
          default:
            //
        }
        break
      }
      default:
        //
    }

    function _setParent(parent, ids) {
      if (ids) {
        if (isArray$1(ids)) {
          ids.forEach(_set);
        } else {
          _set(ids);
        }
      }
      function _set(id) {
        // Note: it can happen, e.g. during deserialization, that the child node
        // is created later than the parent node
        // so we store the parent for later
        table[id] = parent;
        let child = doc.get(id);
        if (child) {
          child.parent = parent;
        }
      }
    }
    function _setRegisteredParent(child) {
      let parent = table[child.id];
      if (parent) {
        child.parent = parent;
      }
    }
  }
}

ParentNodeHook.register = function(doc) {
  return new ParentNodeHook(doc)
};

const converter = new JSONConverter();

/**
  Basic implementation of a Document.

  @example

  ```js
  import { Document } from 'substance'

  class MyArticle extends Document {
    constructor(...args) {
      super(...args)

      this.addIndex('foo', FooIndex)
    }
  }
  ```
*/

class Document extends EventEmitter {

  /**
    @param {DocumentSchema} schema The document schema.
  */
  constructor(schema) {
    super();

    // HACK: to be able to inherit but not execute this ctor
    if (arguments[0] === 'SKIP') return

    this.__id__ = uuid();

    /* istanbul ignore next */
    if (!schema) {
      throw new Error('A document needs a schema for reflection.')
    }

    this.schema = schema;
    this.nodeFactory = new DocumentNodeFactory(this);
    this.data = new IncrementalData(schema, this.nodeFactory);

    // all by type
    this.addIndex('type', new PropertyIndex('type'));

    // special index for (property-scoped) annotations
    this.addIndex('annotations', new AnnotationIndex());

    // TODO: these are only necessary if there is a container annotation
    // in the schema
    // special index for (container-scoped) annotations
    this.addIndex('container-annotations', new ContainerAnnotationIndex());

    // change event proxies are triggered after a document change has been applied
    // before the regular document:changed event is fired.
    // They serve the purpose of making the event notification more efficient
    // In earlier days all observers such as node views where listening on the same event 'operation:applied'.
    // This did not scale with increasing number of nodes, as on every operation all listeners where notified.
    // The proxies filter the document change by interest and then only notify a small set of observers.
    // Example: NotifyByPath notifies only observers which are interested in changes to a certain path.
    this.eventProxies = {
      'path': new PathEventProxy(this),
    };
    this.on('document:changed', this._updateEventProxies, this);
    // TODO: maybe we want to have a generalized concept for such low-level hooks
    // e.g. indexes are similar
    ParentNodeHook.register(this);
  }

  dispose() {
    this.off();
    this.data.off();
  }

  get id() {
    return this.__id__
  }

  /**
    @returns {model/DocumentSchema} the document's schema.
  */
  getSchema() {
    return this.schema
  }

  /**
    Check if this storage contains a node with given id.

    @returns {Boolean} `true` if a node with id exists, `false` otherwise.
  */
  contains(id) {
    return this.data.contains(id)
  }

  /**
    Get a node or value via path.

    @param {String|String[]} path node id or path to property.
    @returns {DocumentNode|any|undefined} a Node instance, a value or undefined if not found.
  */
  get(path, strict) {
    return this.data.get(path, strict)
  }

  /**
    @return {Object} A hash of {@link model/DocumentNode} instances.
  */
  getNodes() {
    return this.data.getNodes()
  }

  /**
    Creates a context like a transaction for importing nodes.
    This is important in presence of cyclic dependencies.
    Indexes will not be updated during the import but will afterwards
    when all nodes are have been created.

    @private
    @param {Function} importer a `function(doc)`, where with `doc` is a `model/AbstractDocument`

    @example

    Consider the following example from our documentation generator:
    We want to have a member index, which keeps track of members of namespaces, modules, and classes.
    grouped by type, and in the case of classes, also grouped by 'instance' and 'class'.

    ```
    ui
      - class
        - ui/Component
    ui/Component
      - class
        - method
          - mount
      - instance
        - method
          - render
    ```

    To decide which grouping to apply, the parent type of a member needs to be considered.
    Using an incremental approach, this leads to the problem, that the parent must exist
    before the child. At the same time, e.g. when deserializing, the parent has already
    a field with all children ids. This cyclic dependency is best address, by turning
    off all listeners (such as indexes) until the data is consistent.

  */
  import(importer) {
    try {
      this.data._stopIndexing();
      importer(this);
      this.data._startIndexing();
    } finally {
      this.data.queue = [];
      this.data._startIndexing();
    }
  }

  /**
    Create a node from the given data.

    @param {Object} plain node data.
    @return {DocumentNode} The created node.

    @example

    ```js
    doc.transaction(function(tx) {
      tx.create({
        id: 'p1',
        type: 'paragraph',
        content: 'Hi I am a Substance paragraph.'
      })
    })
    ```
  */
  create(nodeData) {
    if (!nodeData.id) {
      nodeData.id = uuid(nodeData.type);
    }
    var op = this._create(nodeData);
    var change = new DocumentChange([op], {}, {});
    change._extractInformation(this);
    this._notifyChangeListeners(change, { hidden: true });
    return this.data.get(nodeData.id)
  }

  /**
    Delete the node with given id.

    @param {String} nodeId
    @returns {DocumentNode} The deleted node.

    @example

    ```js
    doc.transaction(function(tx) {
      tx.delete('p1')
    })
    ```
  */
  delete(nodeId) {
    var node = this.get(nodeId);
    var op = this._delete(nodeId);
    var change = new DocumentChange([op], {}, {});
    change._extractInformation(this);
    this._notifyChangeListeners(change, { hidden: true });
    return node
  }

  /**
    Set a property to a new value.

    @param {String[]} property path
    @param {any} newValue
    @returns {DocumentNode} The deleted node.

    @example

    ```js
    doc.transaction(function(tx) {
      tx.set(['p1', 'content'], "Hello there! I'm a new paragraph.")
    })
    ```
  */
  set(path, value) {
    var oldValue = this.get(path);
    var op = this._set(path, value);
    var change = new DocumentChange([op], {}, {});
    change._extractInformation(this);
    this._notifyChangeListeners(change, { hidden: true });
    return oldValue
  }

  /**
    Update a property incrementally.

    @param {Array} property path
    @param {Object} diff
    @returns {any} The value before applying the update.

    @example


    Inserting text into a string property:
    ```
    doc.update(['p1', 'content'], { insert: {offset: 3, value: "fee"} })
    ```
    would turn "Foobar" into "Foofeebar".

    Deleting text from a string property:
    ```
    doc.update(['p1', 'content'], { delete: {start: 0, end: 3} })
    ```
    would turn "Foobar" into "bar".

    Inserting into an array:
    ```
    doc.update(['p1', 'content'], { insert: {offset: 2, value: 0} })
    ```
    would turn `[1,2,3,4]` into `[1,2,0,3,4]`.

    Deleting from an array:
    ```
    doc.update(['body', 'nodes'], { delete: 2 })
    ```
    would turn `[1,2,3,4]` into `[1,2,4]`.
  */
  update(path, diff$$1) {
    var op = this._update(path, diff$$1);
    var change = new DocumentChange([op], {}, {});
    change._extractInformation(this);
    this._notifyChangeListeners(change, { hidden: true });
    return op
  }

  /**
    Add a document index.

    @param {String} name
    @param {DocumentIndex} index
  */
  addIndex(name, index) {
    return this.data.addIndex(name, index)
  }

  /**
    @param {String} name
    @returns {DocumentIndex} the node index with given name.
  */
  getIndex(name) {
    return this.data.getIndex(name)
  }

  /**
    Creates a selection which is attached to this document.
    Every selection implementation provides its own
    parameter format which is basically a JSON representation.

    @param {model/Selection} sel An object describing the selection.

    @example

    Creating a PropertySelection:

    ```js
    doc.createSelection({
      type: 'property',
      path: [ 'text1', 'content'],
      startOffset: 10,
      endOffset: 20,
      containerId: 'body'
    })
    ```

    Creating a ContainerSelection:

    ```js
    doc.createSelection({
      type: 'container',
      containerId: 'body',
      startPath: [ 'p1', 'content'],
      startOffset: 10,
      endPath: [ 'p2', 'content'],
      endOffset: 20
    })
    ```

    Creating a NullSelection:

    ```js
    doc.createSelection(null)
    ```
  */
  createSelection(data) {
    let sel;
    if (isNil(data)) return Selection.nullSelection
    if (arguments.length !== 1 || !isPlainObject$1(data)) {
      throw new Error('Illegal argument: call createSelection({ type: ... }')
    } else {
      switch (data.type) {
        case 'property': {
          if (isNil(data.endOffset)) {
            data.endOffset = data.startOffset;
          }
          if (!data.hasOwnProperty('reverse')) {
            if (data.startOffset>data.endOffset) {
              [data.startOffset, data.endOffset] = [data.endOffset, data.startOffset];
              data.reverse = !data.reverse;
            }
          }
          // integrity checks:
          let text = this.get(data.path, 'strict');
          if (data.startOffset < 0 || data.startOffset > text.length) {
            throw new Error('Invalid startOffset: target property has length '+text.length+', given startOffset is ' + data.startOffset)
          }
          if (data.endOffset < 0 || data.endOffset > text.length) {
            throw new Error('Invalid startOffset: target property has length '+text.length+', given endOffset is ' + data.endOffset)
          }
          sel = new PropertySelection(data);
          break
        }
        case 'container': {
          let container = this.get(data.containerId, 'strict');
          if (!container) throw new Error('Can not create ContainerSelection: container "'+data.containerId+'" does not exist.')
          let start = this._normalizeCoor({ path: data.startPath, offset: data.startOffset});
          let end = this._normalizeCoor({ path: data.endPath, offset: data.endOffset});
          let startAddress = container.getAddress(start);
          let endAddress = container.getAddress(end);
          if (!startAddress) {
            throw new Error('Invalid arguments for ContainerSelection: ', start.toString())
          }
          if (!endAddress) {
            throw new Error('Invalid arguments for ContainerSelection: ', end.toString())
          }
          if (!data.hasOwnProperty('reverse')) {
            if (endAddress.isBefore(startAddress, 'strict')) {
              [start, end] = [end, start];
              data.reverse = true;
            }
          }
          sel = new ContainerSelection(container.id, start.path, start.offset, end.path, end.offset, data.reverse, data.surfaceId);
          break
        }
        case 'node': {
          sel = createNodeSelection({
            doc: this,
            nodeId: data.nodeId,
            mode: data.mode,
            containerId: data.containerId,
            reverse: data.reverse,
            surfaceId: data.surfaceId
          });
          break
        }
        case 'custom': {
          sel = CustomSelection.fromJSON(data);
          break
        }
        default:
          throw new Error('Illegal selection type', data)
      }
    }
    if (!sel.isNull()) {
      sel.attach(this);
    }
    return sel
  }

  getEventProxy(name) {
    return this.eventProxies[name]
  }

  newInstance() {
    var DocumentClass = this.constructor;
    return new DocumentClass(this.schema)
  }

  // useful in combination with paste transformation
  createSnippet() {
    var snippet = this.newInstance();
    var snippetContainer = snippet.create({
      type: 'container',
      id: Document.SNIPPET_ID
    });
    snippet.getContainer = function() {
      return snippetContainer
    };
    return snippet
  }

  _apply(documentChange) {
    forEach(documentChange.ops, function(op) {
      this.data.apply(op);
      this.emit('operation:applied', op);
    }.bind(this));
    // extract aggregated information, such as which property has been affected etc.
    documentChange._extractInformation(this);
  }

  _notifyChangeListeners(change, info) {
    info = info || {};
    this.emit('document:changed', change, info, this);
  }

  _updateEventProxies(change, info) {
    forEach(this.eventProxies, function(proxy) {
      proxy.onDocumentChanged(change, info, this);
    }.bind(this));
  }

  createFromDocument(doc) {
    let nodes = doc.getNodes();
    let annotations = [];
    let contentNodes = [];
    let containers = [];
    forEach(nodes, (node) => {
      if (node._isAnnotation) {
        annotations.push(node);
      } else if (node._isContainer) {
        containers.push(node);
      } else {
        contentNodes.push(node);
      }
    });
    contentNodes.concat(annotations).concat(containers).forEach(n=>this.create(n));
  }

  /**
    Convert to JSON.

    @returns {Object} Plain content.
  */
  toJSON() {
    return converter.exportDocument(this)
  }

  getAnnotations(path) {
    return this.getIndex('annotations').get(path)
  }

  _create(nodeData) {
    var op = this.data.create(nodeData);
    return op
  }

  _delete(nodeId) {
    var op = this.data.delete(nodeId);
    return op
  }

  _update(path, diff$$1) {
    var op = this.data.update(path, diff$$1);
    return op
  }

  _set(path, value) {
    var op = this.data.set(path, value);
    return op
  }

  // NOTE: this is still here because DOMSelection is using it
  _createSelectionFromRange(range) {
    if (!range) return Selection.nullSelection
    let inOneNode = isEqual(range.start.path, range.end.path);
    if (inOneNode) {
      if (range.start.isNodeCoordinate()) {
        // ATTENTION: we only create full NodeSelections
        // when mapping from the DOM to Model  return new NodeSelection(range.containerId, range.start.getNodeId(), mode, range.reverse, range.surfaceId)
        return new NodeSelection(range.containerId, range.start.getNodeId(), 'full', range.reverse, range.surfaceId)
      } else {
        return this.createSelection({
          type: 'property',
          path: range.start.path,
          startOffset: range.start.offset,
          endOffset: range.end.offset,
          reverse: range.reverse,
          containerId: range.containerId,
          surfaceId: range.surfaceId
        })
      }
    } else {
      return this.createSelection({
        type: 'container',
        startPath: range.start.path,
        startOffset: range.start.offset,
        endPath: range.end.path,
        endOffset: range.end.offset,
        reverse: range.reverse,
        containerId: range.containerId,
        surfaceId: range.surfaceId
      })
    }
  }

  _normalizeCoor({ path, offset }) {
    // NOTE: normalizing so that a node coordinate is used only for 'isolated nodes'
    if (path.length === 1) {
      let node = this.get(path[0]).getRoot();
      if (node.isText()) {
        // console.warn("DEPRECATED: don't use node coordinates for TextNodes. Use selectionHelpers instead to set cursor at first or last position conveniently.")
        return new Coordinate(node.getTextPath(), offset === 0 ? 0 : node.getLength())
      } else if (node.isList()) {
        // console.warn("DEPRECATED: don't use node coordinates for ListNodes. Use selectionHelpers instead to set cursor at first or last position conveniently.")
        if (offset === 0) {
          let item = node.getItemAt(0);
          return new Coordinate(item.getTextPath(), 0)
        } else {
          let item = this.get(last$1(node.items));
          return new Coordinate(item.getTextPath(), item.getLength())
        }
      }
    }
    return new Coordinate(path, offset)
  }

}

Document.prototype._isDocument = true;

// used by transforms copy, paste
// and by ClipboardImporter/Exporter
Document.SNIPPET_ID = "snippet";

Document.TEXT_SNIPPET_ID = "text-snippet";

/**
  A {@link Document} instance that is used during transaction.

  During editing a TransactionDocument is kept up-to-date with the real one.
  Whenever a transaction is started on the document, a TransactionDocument is used to
  record changes, which are applied en-bloc when the transaction is saved.

  The transaction document is the common way to manipulate the document.
  It provides a 'turtle-graphics' style API, i.e., it has a state

  @example

  To start a transaction run

  ```
  doc.transaction(function(tx) {
    // use tx to record changes
  })
  ```
*/
class TransactionDocument extends Document {

  /**
    @param {model/Document} document a document instance
  */
  constructor(document) {
    super('SKIP');

    this.schema = document.schema;
    this.nodeFactory = new DocumentNodeFactory(this);
    this.data = new IncrementalData(this.schema, this.nodeFactory);

    this.document = document;

    // ops recorded since transaction start
    this.ops = [];
    this.lastOp = null;

    // copy all indexes
    forEach(document.data.indexes, function(index, name) {
      this.data.addIndex(name, index.clone());
    }.bind(this));

    // ATTENTION: this must before loading the seed
    ParentNodeHook.register(this);

    this.createFromDocument(document);

    // make sure that we mirror all changes that are done outside of transactions
    document.on('document:changed', this._onDocumentChanged, this);
    this._skipNextDocumentChange = false;
  }

  dispose() {
    this.document.off(this);
    this.data.off();
  }

  create(nodeData) {
    if (!nodeData.id) {
      nodeData.id = uuid(nodeData.type);
    }
    if (!nodeData.type) {
      throw new Error('No node type provided')
    }
    this.lastOp = this.data.create(nodeData);
    if (this.lastOp) {
      this.ops.push(this.lastOp);
      return this.data.get(nodeData.id)
    }
  }

  createDefaultTextNode(text, dir) {
    return this.create({
      type: this.getSchema().getDefaultTextType(),
      content: text || '',
      direction: dir
    })
  }

  delete(nodeId) {
    this.lastOp = this.data.delete(nodeId);
    if (this.lastOp) {
      this.ops.push(this.lastOp);
    }
  }

  set(path, value) {
    this.lastOp = this.data.set(path, value);
    if (this.lastOp) {
      this.ops.push(this.lastOp);
    }
  }

  update(path, diffOp) {
    let op = this.lastOp = this.data.update(path, diffOp);
    if (op) {
      this.ops.push(op);
      return op
    }
  }

  _onDocumentChanged(change) {
    if (this._skipNextDocumentChange) {
      this._skipNextDocumentChange = false;
    } else {
      // NOTE: this is hooked to document:changed (low-level), to make sure that we
      // update the transaction document too when the document is manipulated directly, e.g. using `document.create(...)`
      this._apply(change);
    }
  }

  _apply(documentChange) {
    documentChange.ops.forEach(function(op) {
      this.data.apply(op);
    }.bind(this));
  }

  _reset() {
    this.ops = [];
    this.lastOp = null;
  }

  _rollback() {
    for (var i = this.ops.length - 1; i >= 0; i--) {
      this.data.apply(this.ops[i].invert());
    }
    this.ops = [];
    this.lastOp = null;
  }

  newInstance() {
    return this.document.newInstance()
  }
}

TransactionDocument.prototype._isTransactionDocument = true;

/**
  Creates a new document instance containing only the selected content

  @param {Object} args object with `selection`
  @return {Object} with a `doc` property that has a fresh doc with the copied content
*/

function copySelection(doc, selection) {
  if (!selection) throw new Error("'selection' is mandatory.")
  let copy = null;
  if (!selection.isNull() && !selection.isCollapsed()) {
    // return a simplified version if only a piece of text is selected
    if (selection.isPropertySelection()) {
      copy = _copyPropertySelection(doc, selection);
    }
    else if (selection.isContainerSelection()) {
      copy = _copyContainerSelection(doc, selection);
    }
    else if (selection.isNodeSelection()) {
      copy = _copyNodeSelection(doc, selection);
    }
    else {
      console.error('Copy is not yet supported for selection type.');
    }
  }
  return copy
}

function _copyPropertySelection(doc, selection) {
  let path = selection.start.path;
  let offset = selection.start.offset;
  let endOffset = selection.end.offset;
  let text = doc.get(path);
  let snippet = doc.createSnippet();
  let containerNode = snippet.getContainer();
  snippet.create({
    type: doc.schema.getDefaultTextType(),
    id: Document.TEXT_SNIPPET_ID,
    content: text.substring(offset, endOffset)
  });
  containerNode.show(Document.TEXT_SNIPPET_ID);
  let annotations = doc.getIndex('annotations').get(path, offset, endOffset);
  forEach(annotations, function(anno) {
    let data = cloneDeep(anno.toJSON());
    let path = [Document.TEXT_SNIPPET_ID, 'content'];
    data.start = {
      path: path,
      offset: Math.max(offset, anno.start.offset)-offset
    };
    data.end = {
      path: path,
      offset: Math.min(endOffset, anno.end.offset)-offset
    };
    snippet.create(data);
  });
  return snippet
}

function _copyContainerSelection(tx, sel) {
  let snippet = tx.createSnippet();
  let container = snippet.getContainer();

  let nodeIds = sel.getNodeIds();
  let L = nodeIds.length;
  if (L === 0) return snippet

  let start = sel.start;
  let end = sel.end;

  let skippedFirst = false;
  let skippedLast = false;

  // First copy the whole covered nodes
  let created = {};
  for(let i = 0; i<L; i++) {
    let id = nodeIds[i];
    let node = tx.get(id);
    // skip NIL selections, such as cursor at the end of first node or cursor at the start of last node.
    if (i===0 && isLast(tx, start)) {
      skippedFirst = true;
      continue
    }
    if (i===L-1 && isFirst(tx, end)) {
      skippedLast = true;
      continue
    }
    if (!created[id]) {
      documentHelpers.copyNode(node).forEach((nodeData) => {
        let copy = snippet.create(nodeData);
        created[copy.id] = true;
      });
      container.show(id);
    }
  }
  if (!skippedFirst) {
    // ATTENTION: we need the root node here, e.g. the list, not the list items
    let startNode = snippet.get(start.getNodeId()).getRoot();
    if (startNode.isText()) {
      documentHelpers.deleteTextRange(snippet, null, start);
    } else if (startNode.isList()) {
      documentHelpers.deleteListRange(snippet, startNode, null, start);
    }
  }
  if (!skippedLast) {
    // ATTENTION: we need the root node here, e.g. the list, not the list items
    let endNode = snippet.get(end.getNodeId()).getRoot();
    if (endNode.isText()) {
      documentHelpers.deleteTextRange(snippet, end, null);
    } else if (endNode.isList()) {
      documentHelpers.deleteListRange(snippet, endNode, end, null);
    }
  }
  return snippet
}

function _copyNodeSelection(doc, selection) {
  let snippet = doc.createSnippet();
  let containerNode = snippet.getContainer();
  let nodeId = selection.getNodeId();
  let node = doc.get(nodeId);
  documentHelpers.copyNode(node).forEach((nodeData) => {
    snippet.create(nodeData);
  });
  containerNode.show(node.id);
  return snippet
}

/**
  Pastes clipboard content at the current selection

  @param {Object} args object with `selection` and `doc` for Substance content or
  `text` for external HTML content
  @return {Object} with updated `selection`
*/

function paste(tx, args) {
  let sel = tx.selection;
  if (!sel || sel.isNull()) {
    throw new Error("Can not paste, without selection.")
  }
  args = args || {};
  args.text = args.text || '';
  let pasteDoc = args.doc;
  // TODO: is there a better way to detect that this paste is happening within a
  // container?
  let inContainer = Boolean(sel.containerId);

  // when we are in a container, we interpret line-breaks
  // and create a document with multiple paragraphs
  // in a PropertyEditor we paste the text as is
  if (!pasteDoc && !inContainer) {
    tx.insertText(args.text);
    return
  }
  if (!pasteDoc) {
    pasteDoc = _convertPlainTextToDocument(tx, args);
  }
  if (!sel.isCollapsed()) {
    tx.deleteSelection();
  }
  let nodes = pasteDoc.get(Document.SNIPPET_ID).nodes;
  let schema = tx.getSchema();
  if (nodes.length > 0) {
    let first = pasteDoc.get(nodes[0]);
    if (schema.isInstanceOf(first.type, 'text')) {
      _pasteAnnotatedText(tx, pasteDoc);
      // HACK: this changes the container's nodes array.
      // We do this, to be able to call _pasteDocument inserting the remaining nodes
      nodes.shift();
    }
    // if still nodes left > 0
    if (nodes.length > 0) {
      _pasteDocument(tx, pasteDoc);
    }
  }
  return args
}

/*
  Splits plain text by lines and puts them into paragraphs.
*/
function _convertPlainTextToDocument(tx, args) {
  let lines = args.text.split(/\s*\n\s*\n/);
  let pasteDoc = tx.getDocument().newInstance();
  let defaultTextType = pasteDoc.getSchema().getDefaultTextType();
  let container = pasteDoc.create({
    type: 'container',
    id: Document.SNIPPET_ID,
    nodes: []
  });
  let node;
  if (lines.length === 1) {
    node = pasteDoc.create({
      id: Document.TEXT_SNIPPET_ID,
      type: defaultTextType,
      content: lines[0]
    });
    container.show(node.id);
  } else {
    for (let i = 0; i < lines.length; i++) {
      node = pasteDoc.create({
        id: uuid(defaultTextType),
        type: defaultTextType,
        content: lines[i]
      });
      container.show(node.id);
    }
  }
  return pasteDoc
}

function _pasteAnnotatedText(tx, copy) {
  let sel = tx.selection;
  let nodes = copy.get(Document.SNIPPET_ID).nodes;
  let textPath = [nodes[0], 'content'];
  let text = copy.get(textPath);
  let annotations = copy.getIndex('annotations').get(textPath);
  // insert plain text
  let path = sel.start.path;
  let offset = sel.start.offset;
  tx.insertText(text);
  // copy annotations
  forEach(annotations, function(anno) {
    let data = anno.toJSON();
    data.start.path = path.slice(0);
    data.start.offset += offset;
    data.end.offset += offset;
    // create a new uuid if a node with the same id exists already
    if (tx.get(data.id)) data.id = uuid(data.type);
    tx.create(data);
  });
}

function _pasteDocument(tx, pasteDoc) {
  let sel = tx.selection;
  let containerId = sel.containerId;
  let container = tx.get(containerId);
  let insertPos;
  if (sel.isPropertySelection()) {
    let startPath = sel.start.path;
    let nodeId = sel.start.getNodeId();
    let startPos = container.getPosition(nodeId, 'strict');
    let text = tx.get(startPath);
    // Break, unless we are at the last character of a node,
    // then we can simply insert after the node
    if (text.length === 0) {
      insertPos = startPos;
      container.hide(nodeId);
      documentHelpers.deleteNode(tx, tx.get(nodeId));
    } else if ( text.length === sel.start.offset ) {
      insertPos = startPos + 1;
    } else {
      tx.break();
      insertPos = startPos + 1;
    }
  } else if (sel.isNodeSelection()) {
    let nodePos = container.getPosition(sel.getNodeId(), 'strict');
    if (sel.isBefore()) {
      insertPos = nodePos;
    } else if (sel.isAfter()) {
      insertPos = nodePos+1;
    } else {
      throw new Error('Illegal state: the selection should be collapsed.')
    }
  }
  // transfer nodes from content document
  let nodeIds = pasteDoc.get(Document.SNIPPET_ID).nodes;
  let insertedNodes = [];
  let visited = {};
  for (let i = 0; i < nodeIds.length; i++) {
    let node = pasteDoc.get(nodeIds[i]);
    // Note: this will on the one hand make sure
    // node ids are changed to avoid collisions in
    // the target doc
    // Plus, it uses reflection to create owned nodes recursively,
    // and to transfer attached annotations.
    let newId = _transferWithDisambiguatedIds(node.getDocument(), tx, node.id, visited);
    // get the node in the targetDocument
    node = tx.get(newId);
    container.show(newId, insertPos++);
    insertedNodes.push(node);
  }

  if (insertedNodes.length > 0) {
    let lastNode = last$1(insertedNodes);
    setCursor(tx, lastNode, containerId, 'after');
  }
}

// We need to disambiguate ids if the target document
// contains a node with the same id.
// Unfortunately, this can be difficult in some cases,
// e.g. other nodes that have a reference to the re-named node
// We only fix annotations for now.
function _transferWithDisambiguatedIds(sourceDoc, targetDoc, id, visited) {
  if (visited[id]) throw new Error('FIXME: dont call me twice')
  const node = sourceDoc.get(id, 'strict');
  let oldId = node.id;
  let newId;
  if (targetDoc.contains(node.id)) {
    // change the node id
    newId = uuid(node.type);
    node.id = newId;
  }
  visited[id] = node.id;
  const annotationIndex = sourceDoc.getIndex('annotations');
  const nodeSchema = node.getSchema();
  // collect annotations so that we can create them in the target doc afterwards
  let annos = [];
  // now we iterate all properties of the node schema,
  // to see if there are owned references, which need to be created recursively,
  // and if there are text properties, where annotations could be attached to
  for (let key in nodeSchema) {
    if (key === 'id' || key === 'type' || !nodeSchema.hasOwnProperty(key)) continue
    const prop = nodeSchema[key];
    const name = prop.name;
    // Look for references to owned children and create recursively
    if ((prop.isReference() && prop.isOwned()) || (prop.type === 'file')) {
      // NOTE: we need to recurse directly here, so that we can
      // update renamed references
      let val = node[prop.name];
      if (prop.isArray()) {
        _transferArrayOfReferences(sourceDoc, targetDoc, val, visited);
      } else {
        let id = val;
        if (!visited[id]) {
          node[name] = _transferWithDisambiguatedIds(sourceDoc, targetDoc, id, visited);
        }
      }
    }
    // Look for text properties and create annotations in the target doc accordingly
    else if (prop.isText()) {
      let _annos = annotationIndex.get([node.id]);
      for (let i = 0; i < _annos.length; i++) {
        let anno = _annos[i];
        if (anno.start.path[0] === oldId) {
          anno.start.path[0] = newId;
        }
        if (anno.end.path[0] === oldId) {
          anno.end.path[0] = newId;
        }
        annos.push(anno);
      }
    }
  }
  targetDoc.create(node);
  for (let i = 0; i < annos.length; i++) {
    _transferWithDisambiguatedIds(sourceDoc, targetDoc, annos[i].id, visited);
  }
  return node.id
}

function _transferArrayOfReferences(sourceDoc, targetDoc, arr, visited) {
  for (let i = 0; i < arr.length; i++) {
    let val = arr[i];
    // multi-dimensional
    if (isArray$1(val)) {
      _transferArrayOfReferences(sourceDoc, targetDoc, val, visited);
    } else {
      let id = val;
      if (id && !visited[id]) {
        arr[i] = _transferWithDisambiguatedIds(sourceDoc, targetDoc, id, visited);
      }
    }
  }
}

/**
  Core editing implementation, that controls meta behavior
  such as deleting a selection, merging nodes, etc.

  Some of the implementation are then delegated to specific editing behaviors,
  such as manipulating content of a text-property, merging or breaking text nodes

  Note: this is pretty much the same what we did with transforms before.
        We decided to move this here, to switch to a stateful editor implementation (aka turtle-graphics-style)
 */
class Editing {

  // create an annotation for the current selection using the given data
  annotate(tx, annotation) {
    let sel = tx.selection;
    let schema = tx.getSchema();
    let AnnotationClass = schema.getNodeClass(annotation.type);
    if (!AnnotationClass) throw new Error('Unknown annotation type', annotation)
    let start = sel.start;
    let end = sel.end;
    let containerId = sel.containerId;
    let nodeData = { start, end, containerId };
    /* istanbul ignore else  */
    if (sel.isPropertySelection()) {
      if (!AnnotationClass.prototype._isAnnotation) {
        throw new Error('Annotation can not be created for a selection.')
      }
    } else if (sel.isContainerSelection()) {
      if (AnnotationClass.prototype._isPropertyAnnotation) {
        console.warn('NOT SUPPORTED YET: creating property annotations for a non collapsed container selection.');
      }
    }
    Object.assign(nodeData, annotation);
    return tx.create(nodeData)
  }

  break(tx) {
    let sel = tx.selection;
    if (sel.isNodeSelection()) {
      let containerId = sel.containerId;
      let container = tx.get(containerId);
      let nodeId = sel.getNodeId();
      let nodePos = container.getPosition(nodeId, 'strict');
      let textNode = tx.createDefaultTextNode();
      if (sel.isBefore()) {
        tx.update(container.getContentPath(), { type: 'insert', pos: nodePos, value: textNode.id });
        // leave selection as is
      } else {
        tx.update(container.getContentPath(), { type: 'insert', pos: nodePos+1, value: textNode.id });
        setCursor(tx, textNode, containerId, 'before');
      }
    }
    else if (sel.isCustomSelection()) {
      // TODO: what to do with custom selections?
    }
    else if (sel.isCollapsed() || sel.isPropertySelection()) {
      let containerId = sel.containerId;
      if (!sel.isCollapsed()) {
        // delete the selection
        this._deletePropertySelection(tx, sel);
        tx.setSelection(sel.collapse('left'));
      }
      // then break the node
      if (containerId) {
        let container = tx.get(containerId);
        let nodeId = sel.start.path[0];
        let node = tx.get(nodeId);
        this._breakNode(tx, node, sel.start, container);
      }
    }
    else if (sel.isContainerSelection()) {
      let start = sel.start;
      let containerId = sel.containerId;
      let container = tx.get(containerId);
      let startNodeId = start.path[0];
      let nodePos = container.getPosition(startNodeId, 'strict');
      this._deleteContainerSelection(tx, sel, {noMerge: true });
      setCursor(tx, container.getNodeAt(nodePos+1), containerId, 'before');
    }
  }

  delete(tx, direction) {
    let sel = tx.selection;
    // special implementation for node selections:
    // either delete the node (replacing with an empty text node)
    // or just move the cursor
    /* istanbul ignore else  */
    if (sel.isNodeSelection()) {
      this._deleteNodeSelection(tx, sel, direction);
    }
    // TODO: what to do with custom selections?
    else if (sel.isCustomSelection()) {}
    // if the selection is collapsed this is the classical one-character deletion
    // either backwards (backspace) or forward (delete)
    else if (sel.isCollapsed()) {
      // Deletion of a single character leads to a merge
      // if cursor is at a text boundary (TextNode, ListItem)
      // and direction is towards that boundary
      let path = sel.start.path;
      let node = tx.get(path[0]);
      let text = tx.get(path);
      let offset = sel.start.offset;
      let needsMerge = (sel.containerId && (
        (offset === 0 && direction === 'left') ||
        (offset === text.length && direction === 'right')
      ));
      if (needsMerge) {
        // ATTENTION: deviation from standard implementation
        // for list items: Word and GDoc toggle a list item
        // when doing a BACKSPACE at the first position
        let root = node.getRoot();
        if (root.isList() && offset === 0 && direction === 'left') {
          return this.toggleList(tx)
        } else {
          let container = tx.get(sel.containerId);
          this._merge(tx, root, sel.start, direction, container);
        }
      } else {
        let startOffset = (direction === 'left') ? offset-1 : offset;
        let endOffset = startOffset+1;
        let start = { path: path, offset: startOffset };
        let end = { path: path, offset: endOffset };
        documentHelpers.deleteTextRange(tx, start, end);
        tx.setSelection({
          type: 'property',
          path: path,
          startOffset: startOffset,
          containerId: sel.containerId
        });
      }
    }
    // deleting a range of characters within a text property
    else if (sel.isPropertySelection()) {
      documentHelpers.deleteTextRange(tx, sel.start, sel.end);
      tx.setSelection(sel.collapse('left'));
    }
    // deleting a range within a container (across multiple nodes)
    else if (sel.isContainerSelection()) {
      this._deleteContainerSelection(tx, sel);
    }
    else {
      console.warn('Unsupported case: tx.delete(%)', direction, sel);
    }
  }

  _deleteNodeSelection(tx, sel, direction) {
    let nodeId = sel.getNodeId();
    let container = tx.get(sel.containerId);
    let nodePos = container.getPosition(nodeId, 'strict');
    if (sel.isFull() ||
        sel.isBefore() && direction === 'right' ||
        sel.isAfter() && direction === 'left' ) {
      // replace the node with default text node
      let contentPath = container.getContentPath();
      tx.update(contentPath, { type: 'delete', pos: nodePos });
      documentHelpers.deleteNode(tx, tx.get(nodeId));
      let newNode = tx.createDefaultTextNode();
      tx.update(contentPath, { type: 'insert', pos: nodePos, value: newNode.id });
      tx.setSelection({
        type: 'property',
        path: newNode.getTextPath(),
        startOffset: 0,
        containerId: container.id,
      });
    } else {
      /* istanbul ignore else  */
      if (sel.isBefore() && direction === 'left') {
        if (nodePos > 0) {
          let previous = container.getNodeAt(nodePos-1);
          if (previous.isText()) {
            tx.setSelection({
              type: 'property',
              path: previous.getTextPath(),
              startOffset: previous.getLength()
            });
            this.delete(tx, direction);
          } else {
            tx.setSelection({
              type: 'node',
              nodeId: previous.id,
              containerId: container.id
            });
          }
        } else {
          // nothing to do
        }
      } else if (sel.isAfter() && direction === 'right') {
        if (nodePos < container.getLength()-1) {
          let next = container.getNodeAt(nodePos+1);
          if (next.isText()) {
            tx.setSelection({
              type: 'property',
              path: next.getTextPath(),
              startOffset: 0
            });
            this.delete(tx, direction);
          } else {
            tx.setSelection({
              type: 'node',
              nodeId: next.id,
              containerId: container.id
            });
          }
        } else {
          // nothing to do
        }
      } else {
        console.warn('Unsupported case: delete(%s)', direction, sel);
      }
    }
  }

  _deletePropertySelection(tx, sel) {
    let path = sel.start.path;
    let start = sel.start.offset;
    let end = sel.end.offset;
    tx.update(path, { type: 'delete', start: start, end: end });
    annotationHelpers.deletedText(tx, path, start, end);
  }

  // deletes all inner nodes and 'truncates' start and end node
  _deleteContainerSelection(tx, sel, options = {}) {
    let containerId = sel.containerId;
    let container = tx.get(containerId);
    let start = sel.start;
    let end = sel.end;
    let startId = start.getNodeId();
    let endId = end.getNodeId();
    let startPos = container.getPosition(startId, 'strict');
    let endPos = container.getPosition(endId, 'strict');

    // special case: selection within one node
    if (startPos === endPos) {
      // ATTENTION: we need the root node here e.g. the list, not the list-item
      let node = tx.get(startId).getRoot();
      /* istanbul ignore else  */
      if (node.isText()) {
        documentHelpers.deleteTextRange(tx, start, end);
      } else if (node.isList()) {
        documentHelpers.deleteListRange(tx, node, start, end);
      } else {
        throw new Error('Not supported yet.')
      }
      tx.setSelection(sel.collapse('left'));
      return
    }

    // TODO: document the algorithm

    let firstNode = tx.get(start.getNodeId());
    let lastNode = tx.get(end.getNodeId());
    let firstEntirelySelected = isEntirelySelected(tx, firstNode, start, null);
    let lastEntirelySelected = isEntirelySelected(tx, lastNode, null, end);

    // delete or truncate last node
    if (lastEntirelySelected) {
      tx.update([container.id, 'nodes'], { type: 'delete', pos: endPos });
      documentHelpers.deleteNode(tx, lastNode);
    } else {
      // ATTENTION: we need the root node here e.g. the list, not the list-item
      let node = lastNode.getRoot();
      /* istanbul ignore else  */
      if (node.isText()) {
        documentHelpers.deleteTextRange(tx, null, end);
      } else if (node.isList()) {
        documentHelpers.deleteListRange(tx, node, null, end);
      } else {
        // IsolatedNodes can not be selected partially
      }
    }

    // delete inner nodes
    for (let i = endPos-1; i > startPos; i--) {
      let nodeId = container.nodes[i];
      tx.update([container.id, 'nodes'], { type: 'delete', pos: i });
      documentHelpers.deleteNode(tx, tx.get(nodeId));
    }

    // delete or truncate the first node
    if (firstEntirelySelected) {
      tx.update([container.id, 'nodes'], { type: 'delete', pos: startPos });
      documentHelpers.deleteNode(tx, firstNode);
    } else {
      // ATTENTION: we need the root node here e.g. the list, not the list-item
      let node = firstNode.getRoot();
      /* istanbul ignore else  */
      if (node.isText()) {
        documentHelpers.deleteTextRange(tx, start, null);
      } else if (node.isList()) {
        documentHelpers.deleteListRange(tx, node, start, null);
      } else {
        // IsolatedNodes can not be selected partially
      }
    }

    // insert a new TextNode if all has been deleted
    if (firstEntirelySelected && lastEntirelySelected) {
      // insert a new paragraph
      let textNode = tx.createDefaultTextNode();
      tx.update([container.id, 'nodes'], { type: 'insert', pos: startPos, value: textNode.id });
      tx.setSelection({
        type: 'property',
        path: textNode.getTextPath(),
        startOffset: 0,
        containerId: containerId
      });
    } else if (!firstEntirelySelected && !lastEntirelySelected) {
      if (!options.noMerge) {
        this._merge(tx, firstNode, sel.start, 'right', container);
      }
      tx.setSelection(sel.collapse('left'));
    } else if (firstEntirelySelected) {
      setCursor(tx, lastNode, container.id, 'before');
    } else {
      setCursor(tx, firstNode, container.id, 'after');
    }
  }

  insertInlineNode(tx, nodeData) {
    let sel = tx.selection;
    let text = "\uFEFF";
    this.insertText(tx, text);
    sel = tx.selection;
    let endOffset = tx.selection.end.offset;
    let startOffset = endOffset - text.length;
    nodeData = Object.assign({}, nodeData, {
      start: {
        path: sel.path,
        offset: startOffset
      },
      end: {
        path: sel.path,
        offset: endOffset
      }
    });
    return tx.create(nodeData)
  }

  insertBlockNode(tx, nodeData) {
    let sel = tx.selection;
    // don't create the node if it already exists
    let blockNode;
    if (!nodeData._isNode || !tx.get(nodeData.id)) {
      blockNode = tx.create(nodeData);
    } else {
      blockNode = tx.get(nodeData.id);
    }
    /* istanbul ignore else  */
    if (sel.isNodeSelection()) {
      let containerId = sel.containerId;
      let container = tx.get(containerId);
      let nodeId = sel.getNodeId();
      let nodePos = container.getPosition(nodeId, 'strict');
      // insert before
      if (sel.isBefore()) {
        tx.update(container.getContentPath(), { type: 'insert', pos: nodePos, value: blockNode.id });
      }
      // insert after
      else if (sel.isAfter()) {
        tx.update(container.getContentPath(), { type: 'insert', pos: nodePos+1, value: blockNode.id });
        tx.setSelection({
          type: 'node',
          containerId: containerId,
          nodeId: blockNode.id,
          mode: 'after'
        });
      } else {
        tx.update(container.getContentPath(), { type: 'delete', pos: nodePos });
        documentHelpers.deleteNode(tx, tx.get(nodeId));
        tx.update([container.id, 'nodes'], { type: 'insert', pos: nodePos, value: blockNode.id });
        tx.setSelection({
          type: 'node',
          containerId: containerId,
          nodeId: blockNode.id,
          mode: 'after'
        });
      }
    } else if (sel.isPropertySelection()) {
      /* istanbul ignore next */
      if (!sel.containerId) throw new Error('insertBlockNode can only be used within a container.')
      let container = tx.get(sel.containerId);
      if (!sel.isCollapsed()) {
        this._deletePropertySelection(tx, sel);
        tx.setSelection(sel.collapse('left'));
      }
      let node = tx.get(sel.path[0]);
      /* istanbul ignore next */
      if (!node) throw new Error('Invalid selection.')
      let nodePos = container.getPosition(node.id, 'strict');
      /* istanbul ignore else  */
      if (node.isText()) {
        let text = node.getText();
        // replace node
        if (text.length === 0) {
          tx.update(container.getContentPath(), { type: 'delete', pos: nodePos });
          documentHelpers.deleteNode(tx, node);
          tx.update([container.id, 'nodes'], { type: 'insert', pos: nodePos, value: blockNode.id });
          setCursor(tx, blockNode, container.id, 'after');
        }
        // insert before
        else if (sel.start.offset === 0) {
          tx.update(container.getContentPath(), { type: 'insert', pos: nodePos, value: blockNode.id });
        }
        // insert after
        else if (sel.start.offset === text.length) {
          tx.update(container.getContentPath(), { type: 'insert', pos: nodePos+1, value: blockNode.id });
          setCursor(tx, blockNode, container.id, 'before');
        }
        // break
        else {
          this.break(tx);
          tx.update(container.getContentPath(), { type: 'insert', pos: nodePos+1, value: blockNode.id });
          setCursor(tx, blockNode, container.id, 'after');
        }
      } else {
        console.error('Not supported: insertBlockNode() on a custom node');
      }
    } else if (sel.isContainerSelection()) {
      if (sel.isCollapsed()) {
        let start = sel.start;
        /* istanbul ignore else  */
        if (start.isPropertyCoordinate()) {
          tx.setSelection({
            type: 'property',
            path: start.path,
            startOffset: start.offset,
            containerId: sel.containerId,
          });
        } else if (start.isNodeCoordinate()) {
          tx.setSelection({
            type: 'node',
            containerId: sel.containerId,
            nodeId: start.path[0],
            mode: start.offset === 0 ? 'before' : 'after',
          });
        } else {
          throw new Error('Unsupported selection for insertBlockNode')
        }
        return this.insertBlockNode(tx, blockNode)
      } else {
        this.break(tx);
        return this.insertBlockNode(tx, blockNode)
      }
    }
    return blockNode
  }

  insertText(tx, text) {
    let sel = tx.selection;
    // type over a selected node or insert a paragraph before
    // or after
    /* istanbul ignore else  */
    if (sel.isNodeSelection()) {
      let containerId = sel.containerId;
      let container = tx.get(containerId);
      let nodeId = sel.getNodeId();
      let nodePos = container.getPosition(nodeId, 'strict');
      let textNode = tx.createDefaultTextNode(text);
      if (sel.isBefore()) {
        container.show(textNode, nodePos);
      } else if (sel.isAfter()) {
        container.show(textNode, nodePos+1);
      } else {
        container.hide(nodeId);
        documentHelpers.deleteNode(tx, tx.get(nodeId));
        container.show(textNode, nodePos);
      }
      setCursor(tx, textNode, sel.containerId, 'after');
    } else if (sel.isCustomSelection()) {
      // TODO: what to do with custom selections?
    } else if (sel.isCollapsed() || sel.isPropertySelection()) {
      // console.log('#### before', sel.toString())
      this._insertText(tx, sel, text);
      // console.log('### setting selection after typing: ', tx.selection.toString())
    } else if (sel.isContainerSelection()) {
      this._deleteContainerSelection(tx, sel);
      this.insertText(tx, text);
    }
  }

  paste(tx, content) {
    if (!content) return
    /* istanbul ignore else  */
    if (isString$1(content)) {
      paste(tx, {text: content});
    } else if (content._isDocument) {
      paste(tx, {doc: content});
    } else {
      throw new Error('Illegal content for paste.')
    }
  }

  /**
    Switch text type for a given node. E.g. from `paragraph` to `heading`.

    @param {Object} args object with `selection`, `containerId` and `data` with new node data
    @return {Object} object with updated `selection`

    @example

    ```js
    switchTextType(tx, {
      selection: bodyEditor.getSelection(),
      containerId: bodyEditor.getContainerId(),
      data: {
        type: 'heading',
        level: 2
      }
    })
    ```
  */
  switchTextType(tx, data) {
    let sel = tx.selection;
    /* istanbul ignore next */
    if (!sel.isPropertySelection()) {
      throw new Error("Selection must be a PropertySelection.")
    }
    let containerId = sel.containerId;
    /* istanbul ignore next */
    if (!containerId) {
      throw new Error("Selection must be within a container.")
    }
    let path = sel.path;
    let nodeId = path[0];
    let node = tx.get(nodeId);
    /* istanbul ignore next */
    if (!(node.isInstanceOf('text'))) {
      throw new Error('Trying to use switchTextType on a non text node.')
    }
    // create a new node and transfer annotations
    let newNode = Object.assign({
      id: uuid(data.type),
      type: data.type,
      content: node.content,
      direction: node.direction
    }, data);
    let newPath = [newNode.id, 'content'];
    newNode = tx.create(newNode);
    annotationHelpers.transferAnnotations(tx, path, 0, newPath, 0);

    // hide and delete the old one, show the new node
    let container = tx.get(sel.containerId);
    let pos = container.getPosition(nodeId, 'strict');
    container.hide(nodeId);
    documentHelpers.deleteNode(tx, node);
    container.show(newNode.id, pos);

    tx.setSelection({
      type: 'property',
      path: newPath,
      startOffset: sel.start.offset,
      endOffset: sel.end.offset,
      containerId: containerId
    });

    return newNode
  }

  toggleList(tx, params) {
    let sel = tx.selection;
    let container = tx.get(sel.containerId);
    /* istanbul ignore next */
    if (!container) {
      throw new Error("Selection must be within a container.")
    }
    if (sel.isPropertySelection()) {
      let nodeId = sel.start.path[0];
      // ATTENTION: we need the root node here e.g. the list, not the list-item
      let node = tx.get(nodeId).getRoot();
      let nodePos = container.getPosition(node.id, 'strict');
      /* istanbul ignore else  */
      if (node.isText()) {
        tx.update([container.id, 'nodes'], { type: 'delete', pos: nodePos });
        // TODO: what if this should create a different list-item type?
        let newItem = tx.create({
          type: 'list-item',
          content: node.getText(),
        });
        annotationHelpers.transferAnnotations(tx, node.getTextPath(), 0, newItem.getTextPath(), 0);
        let newList = tx.create(Object.assign({
          type: 'list',
          items: [newItem.id]
        }, params));
        documentHelpers.deleteNode(tx, node);
        tx.update([container.id, 'nodes'], { type: 'insert', pos: nodePos, value: newList.id });
        tx.setSelection({
          type: 'property',
          path: newItem.getTextPath(),
          startOffset: sel.start.offset,
          containerId: sel.containerId
        });
      } else if (node.isList()) {
        let itemId = sel.start.path[0];
        let itemPos = node.getItemPosition(itemId);
        let item = node.getItemAt(itemPos);
        let newTextNode = tx.createDefaultTextNode(item.getText());
        annotationHelpers.transferAnnotations(tx, item.getTextPath(), 0, newTextNode.getTextPath(), 0);
        // take the item out of the list
        node.removeItemAt(itemPos);
        if (node.isEmpty()) {
          tx.update([container.id, 'nodes'], { type: 'delete', pos: nodePos });
          documentHelpers.deleteNode(tx, node);
          tx.update([container.id, 'nodes'], { type: 'insert', pos: nodePos, value: newTextNode.id });
        } else if (itemPos === 0) {
          tx.update([container.id, 'nodes'], { type: 'insert', pos: nodePos, value: newTextNode.id });
        } else if (node.getLength() <= itemPos){
          tx.update([container.id, 'nodes'], { type: 'insert', pos: nodePos+1, value: newTextNode.id });
        } else {
          //split the
          let tail = [];
          const items = node.items.slice();
          const L = items.length;
          for (let i = L-1; i >= itemPos; i--) {
            tail.unshift(items[i]);
            node.removeItemAt(i);
          }
          let newList = tx.create({
            type: 'list',
            items: tail,
            ordered: node.ordered
          });
          tx.update([container.id, 'nodes'], { type: 'insert', pos: nodePos+1, value: newTextNode.id });
          tx.update([container.id, 'nodes'], { type: 'insert', pos: nodePos+2, value: newList.id });
        }
        tx.setSelection({
          type: 'property',
          path: newTextNode.getTextPath(),
          startOffset: sel.start.offset,
          containerId: sel.containerId
        });
      } else {
        // unsupported node type
      }
    } else if (sel.isContainerSelection()) {
      console.error('TODO: support toggleList with ContainerSelection');
    }
  }

  indent(tx) {
    let sel = tx.selection;
    if (sel.isPropertySelection()) {
      let nodeId = sel.start.getNodeId();
      // ATTENTION: we need the root node here, e.g. the list, not the list items
      let node = tx.get(nodeId).getRoot();
      if (node.isList()) {
        let itemId = sel.start.path[0];
        let item = tx.get(itemId);
        // Note: allowing only 3 levels
        if (item && item.level<3) {
          tx.set([itemId, 'level'], item.level+1);
        }
      }
    } else if (sel.isContainerSelection()) {
      console.error('TODO: support toggleList with ContainerSelection');
    }
  }

  dedent(tx) {
    let sel = tx.selection;
    if (sel.isPropertySelection()) {
      let nodeId = sel.start.getNodeId();
      // ATTENTION: we need the root node here, e.g. the list, not the list items
      let node = tx.get(nodeId).getRoot();
      if (node.isList()) {
        let itemId = sel.start.path[0];
        let item = tx.get(itemId);
        if (item && item.level>1) {
          tx.set([itemId, 'level'], item.level-1);
        }
      }
    } else if (sel.isContainerSelection()) {
      console.error('TODO: support toggleList with ContainerSelection');
    }
  }

  /*
    <-->: anno
    |--|: area of change
    I: <--> |--|     :   nothing
    II: |--| <-->    :   move both by total span+L
    III: |-<-->-|    :   delete anno
    IV: |-<-|->      :   move start by diff to start+L, and end by total span+L
    V: <-|->-|       :   move end by diff to start+L
    VI: <-->|--|     :   noting if !anno.autoExpandRight
    VII: <-|--|->    :   move end by total span+L
  */
  _insertText(tx, sel, text) {
    let start = sel.start;
    let end = sel.end;
    /* istanbul ignore next  */
    if (!isArrayEqual(start.path, end.path)) {
      throw new Error('Unsupported state: range should be on one property')
    }
    let path = start.path;
    let startOffset = start.offset;
    let endOffset = end.offset;
    let typeover = !sel.isCollapsed();
    let L = text.length;
    // delete selected text
    if (typeover) {
      tx.update(path, { type: 'delete', start: startOffset, end: endOffset });
    }
    // insert new text
    tx.update(path, { type: 'insert', start: startOffset, text: text });
    // update annotations
    let annos = tx.getAnnotations(path);
    annos.forEach(function(anno) {
      let annoStart = anno.start.offset;
      let annoEnd = anno.end.offset;

      /* istanbul ignore else  */
      // I anno is before
      if (annoEnd<startOffset) {
        return
      }
      // II anno is after
      else if (annoStart>=endOffset) {
        tx.update([anno.id, 'start'], { type: 'shift', value: startOffset-endOffset+L });
        tx.update([anno.id, 'end'], { type: 'shift', value: startOffset-endOffset+L });
      }
      // III anno is deleted
      // NOTE: InlineNodes only have a length of one character
      // so they are always 'covered', and as they can not expand
      // they are deleted
      else if (
        (annoStart>=startOffset && annoEnd<endOffset) ||
        (anno._isInlineNode && annoStart>=startOffset && annoEnd<=endOffset)
      ) {
        tx.delete(anno.id);
      }
      // IV anno.start between and anno.end after
      else if (annoStart>=startOffset && annoEnd>=endOffset) {
        // do not move start if typing over
        if (annoStart>startOffset || !typeover) {
          tx.update([anno.id, 'start'], { type: 'shift', value: startOffset-annoStart+L });
        }
        tx.update([anno.id, 'end'], { type: 'shift', value: startOffset-endOffset+L });
      }
      // V anno.start before and anno.end between
      else if (annoStart<startOffset && annoEnd<endOffset) {
        // NOTE: here the anno gets expanded (that's the common way)
        tx.update([anno.id, 'end'], { type: 'shift', value: startOffset-annoEnd+L });
      }
      // VI
      else if (annoEnd === startOffset && !anno.constructor.autoExpandRight) {
          // skip
      }
      // VII anno.start before and anno.end after
      else if (annoStart<startOffset && annoEnd>=endOffset) {
        if (anno._isInlineNode) {
          // skip
        } else {
          tx.update([anno.id, 'end'], { type: 'shift', value: startOffset-endOffset+L });
        }
      }
      else {
        console.warn('TODO: handle annotation update case.');
      }
    });
    let offset = startOffset + text.length;
    tx.setSelection({
      type: 'property',
      path: start.path,
      startOffset: offset,
      containerId: sel.containerId,
      surfaceId: sel.surfaceId
    });
  }

  _breakNode(tx, node, coor, container) {
    // ATTENTION: we need the root here, e.g. a list, not the list-item
    node = node.getRoot();
    /* istanbul ignore else  */
    if (node.isText()) {
      this._breakTextNode(tx, node, coor, container);
    } else if (node.isList()) {
      this._breakListNode(tx, node, coor, container);
    } else {
      throw new Error('Not supported')
    }
  }

  _breakTextNode(tx, node, coor, container) {
    let path = coor.path;
    let offset = coor.offset;
    let nodePos = container.getPosition(node.id, 'strict');
    let text = node.getText();

    // when breaking at the first position, a new node of the same
    // type will be inserted.
    if (offset === 0) {
      let newNode = tx.create({
        type: node.type,
        content: ""
      });
      // show the new node
      container.show(newNode.id, nodePos);
      tx.setSelection({
        type: 'property',
        path: path,
        startOffset: 0,
        containerId: container.id
      });
    }
    // otherwise split the text property and create a new paragraph node with trailing text and annotations transferred
    else {
      let newNode = node.toJSON();
      delete newNode.id;
      newNode.content = text.substring(offset);
      // if at the end insert a default text node no matter in which text node we are
      if (offset === text.length) {
        newNode.type = tx.getSchema().getDefaultTextType();
      }
      newNode = tx.create(newNode);
      // Now we need to transfer annotations
      if (offset < text.length) {
        // transfer annotations which are after offset to the new node
        annotationHelpers.transferAnnotations(tx, path, offset, newNode.getTextPath(), 0);
        // truncate the original property
        tx.update(path, { type: 'delete', start: offset, end: text.length });
      }
      // show the new node
      container.show(newNode.id, nodePos+1);
      // update the selection
      tx.setSelection({
        type: 'property',
        path: newNode.getTextPath(),
        startOffset: 0,
        containerId: container.id
      });
    }
  }

  _breakListNode(tx, node, coor, container) {
    let path = coor.path;
    let offset = coor.offset;
    let listItem = tx.get(path[0]);

    let L = node.length;
    let itemPos = node.getItemPosition(listItem.id);
    let text = listItem.getText();
    let newItem = listItem.toJSON();
    delete newItem.id;
    if (offset === 0) {
      // if breaking at an empty list item, then the list is split into two
      if (!text) {
        // if it is the first or last item, a default text node is inserted before or after, and the item is removed
        // if the list has only one element, it is removed
        let nodePos = container.getPosition(node.id, 'strict');
        let newTextNode = tx.createDefaultTextNode();
        // if the list is empty, replace it with a paragraph
        if (L < 2) {
          container.hide(node.id);
          documentHelpers.deleteNode(tx, node);
          container.show(newTextNode.id, nodePos);
        }
        // if at the first list item, remove the item
        else if (itemPos === 0) {
          node.remove(listItem.id);
          documentHelpers.deleteNode(tx, listItem);
          container.show(newTextNode.id, nodePos);
        }
        // if at the last list item, remove the item and append the paragraph
        else if (itemPos >= L-1) {
          node.remove(listItem.id);
          documentHelpers.deleteNode(tx, listItem);
          container.show(newTextNode.id, nodePos+1);
        }
        // otherwise create a new list
        else {
          let tail = [];
          const items = node.items.slice();
          for (let i = L-1; i > itemPos; i--) {
            tail.unshift(items[i]);
            node.remove(items[i]);
          }
          node.remove(items[itemPos]);
          let newList = tx.create({
            type: 'list',
            items: tail,
            ordered: node.ordered
          });
          container.show(newTextNode.id, nodePos+1);
          container.show(newList.id, nodePos+2);
        }
        tx.setSelection({
          type: 'property',
          path: newTextNode.getTextPath(),
          startOffset: 0
        });
      }
      // insert a new paragraph above the current one
      else {
        newItem.content = "";
        newItem = tx.create(newItem);
        node.insertItemAt(itemPos, newItem.id);
        tx.setSelection({
          type: 'property',
          path: listItem.getTextPath(),
          startOffset: 0
        });
      }
    }
    // otherwise split the text property and create a new paragraph node with trailing text and annotations transferred
    else {
      newItem.content = text.substring(offset);
      newItem = tx.create(newItem);
      // Now we need to transfer annotations
      if (offset < text.length) {
        // transfer annotations which are after offset to the new node
        annotationHelpers.transferAnnotations(tx, path, offset, [newItem.id,'content'], 0);
        // truncate the original property
        tx.update(path, { type: 'delete', start: offset, end: text.length });
      }
      node.insertItemAt(itemPos+1, newItem.id);
      tx.setSelection({
        type: 'property',
        path: newItem.getTextPath(),
        startOffset: 0
      });
    }
  }

  _merge(tx, node, coor, direction, container) {
    // detect cases where list items get merged
    // within a single list node
    if (node.isList()) {
      let list = node;
      let itemId = coor.path[0];
      let itemPos = list.getItemPosition(itemId);
      let withinListNode = (
        (direction === 'left' && itemPos > 0) ||
        (direction === 'right' && itemPos<list.items.length-1)
      );
      if (withinListNode) {
        itemPos = (direction === 'left') ? itemPos-1 : itemPos;
        let target = list.getItemAt(itemPos);
        let targetLength = target.getLength();
        documentHelpers.mergeListItems(tx, list.id, itemPos);
        tx.setSelection({
          type: 'property',
          path: target.getTextPath(),
          startOffset: targetLength,
          containerId: container.id
        });
        return
      }
    }
    // in all other cases merge is done across node boundaries
    let nodePos = container.getPosition(node, 'strict');
    if (direction === 'left' && nodePos > 0) {
      this._mergeNodes(tx, container, nodePos-1, direction);
    } else if (direction === 'right' && nodePos<container.getLength()-1) {
      this._mergeNodes(tx, container, nodePos, direction);
    }
  }

  _mergeNodes(tx, container, pos, direction) {
    let first = container.getChildAt(pos);
    let second = container.getChildAt(pos+1);
    if (first.isText()) {
      // Simplification for empty nodes
      if (first.isEmpty()) {
        container.hide(first.id);
        documentHelpers.deleteNode(tx, first);
        // TODO: need to clear where to handle
        // selections ... probably better not to do it here
        setCursor(tx, second, container.id, 'before');
        return
      }
      let target = first;
      let targetPath = target.getTextPath();
      let targetLength = target.getLength();
      if (second.isText()) {
        let source = second;
        let sourcePath = source.getTextPath();
        container.hide(source.id);
        // append the text
        tx.update(targetPath, { type: 'insert', start: targetLength, text: source.getText() });
        // transfer annotations
        annotationHelpers.transferAnnotations(tx, sourcePath, 0, targetPath, targetLength);
        documentHelpers.deleteNode(tx, source);
        tx.setSelection({
          type: 'property',
          path: targetPath,
          startOffset: targetLength,
          containerId: container.id
        });
      } else if (second.isList()) {
        let list = second;
        if (!second.isEmpty()) {
          let source = list.getFirstItem();
          let sourcePath = source.getTextPath();
          // remove merged item from list
          list.removeItemAt(0);
          // append the text
          tx.update(targetPath, { type: 'insert', start: targetLength, text: source.getText() });
          // transfer annotations
          annotationHelpers.transferAnnotations(tx, sourcePath, 0, targetPath, targetLength);
          // delete item and prune empty list
          documentHelpers.deleteNode(tx, source);
        }
        if (list.isEmpty()) {
          container.hide(list.id);
          documentHelpers.deleteNode(tx, list);
        }
        tx.setSelection({
          type: 'property',
          path: targetPath,
          startOffset: targetLength,
          containerId: container.id
        });
      } else {
        selectNode(tx, direction === 'left' ? first.id : second.id, container.id);
      }
    } else if (first.isList()) {
      if (second.isText()) {
        let source = second;
        let sourcePath = source.getTextPath();
        let target = first.getLastItem();
        let targetPath = target.getTextPath();
        let targetLength = target.getLength();
        // hide source
        container.hide(source.id);
        // append the text
        tx.update(targetPath, { type: 'insert', start: targetLength, text: source.getText() });
        // transfer annotations
        annotationHelpers.transferAnnotations(tx, sourcePath, 0, targetPath, targetLength);
        documentHelpers.deleteNode(tx, source);
        tx.setSelection({
          type: 'property',
          path: target.getTextPath(),
          startOffset: targetLength,
          containerId: container.id
        });
      } else if (second.isList()) {
        /* istanbul ignore next */
        if (direction !== 'right') {
          // ATTENTION: merging two lists by using BACKSPACE is not possible,
          // as BACKSPACE will first turn the list into a paragraph
          throw new Error('Illegal state')
        }
        container.hide(second.id);
        let firstItems = first.items.slice();
        let secondItems = second.items.slice();
        for (let i=0; i<secondItems.length;i++) {
          second.removeItemAt(0);
          first.appendItem(secondItems[i]);
        }
        documentHelpers.deleteNode(tx, second);
        let item = tx.get(last$1(firstItems));
        tx.setSelection({
          type: 'property',
          path: item.getTextPath(),
          startOffset: item.getLength(),
          containerId: container.id
        });
      } else {
        selectNode(tx, direction === 'left' ? first.id : second.id, container.id);
      }
    } else {
      if (second.isText() && second.isEmpty()) {
        container.hide(second.id);
        documentHelpers.deleteNode(tx, second);
        setCursor(tx, first, container.id, 'after');
      } else {
        selectNode(tx, direction === 'left' ? first.id : second.id, container.id);
      }
    }
  }
}

/*
  Abstract base class for document editor APIs such as Transaction.

  It implements a turtle-graphics way of editing by maintaining a selection state
  and providing an interface for low- and high-level manipulation.

  Low-level manipulations are dispatched to the edited document instance.
  Higher-level manipulations involve complex manipulations keeping the selection in a correct state.
 */
class EditingInterface {

  constructor(doc) {
    this._document = doc;
    this._selection = null;
    // TODO: allow for custom editing implementation
    this._impl = new Editing();
    this._direction = null;
  }

  getDocument() {
    return this._document
  }

  /* low-level API */

  get(...args) {
    return this._document.get(...args)
  }

  contains(id) {
    return this._document.contains(id)
  }

  create(nodeData) {
    return this._document.create(nodeData)
  }

  createDefaultTextNode(content) {
    return this._document.createDefaultTextNode(content, this._direction)
  }

  delete(nodeId) {
    return this._document.delete(nodeId)
  }

  set(path, value) {
    return this._document.set(path, value)
  }

  update(path, diffOp) {
    return this._document.update(path, diffOp)
  }

  /* Selection API */

  createSelection(selData) {
    // TODO: this is questionable
    // I'd like to convenience first for the 99% use cases
    // which means that I copy over containerId and surfaceId
    // if possible and not explicitly set
    const oldSel = this._selection;
    if (selData && oldSel) {
      if (!selData.containerId && oldSel.containerId) {
        selData.containerId = oldSel.containerId;
      }
      if (!selData.surfaceId && oldSel.surfaceId) {
        selData.surfaceId = oldSel.surfaceId;
      }
    }
    return this._document.createSelection(selData)
  }

  setSelection(sel) {
    if (!sel) sel = Selection.nullSelection;
    else if (isPlainObject$1(sel)) {
      sel = this.createSelection(sel);
    }
    let oldSel = this._selection;
    if (oldSel && sel && !sel.isNull()) {
      if (!sel.containerId) {
        sel.containerId = oldSel.containerId;
      }
    }
    this._selection = sel;
    return sel
  }

  getSelection() {
    return this._selection
  }

  get selection() {
    return this._selection
  }

  set selection(sel) {
    this.setSelection(sel);
  }

  /*
    ATTENTION/TODO: text direction could be different on different paragraphs
    I.e. it should probably be a TextNode property
  */
  get textDirection() {
    return this._direction
  }

  set textDirection(dir) {
    this._direction = dir;
  }

  /* High-level editing */

  annotate(annotationData) {
    const sel = this._selection;
    if (sel && (sel.isPropertySelection() || sel.isContainerSelection())) {
      return this._impl.annotate(this, annotationData)
    }
  }

  break() {
    if (this._selection && !this._selection.isNull()) {
      this._impl.break(this);
    }
  }

  copySelection() {
    const sel = this._selection;
    if (sel && !sel.isNull() && !sel.isCollapsed()) {
      return copySelection(this.getDocument(), this._selection)
    }
  }

  deleteSelection(options) {
    const sel = this._selection;
    if (sel && !sel.isNull() && !sel.isCollapsed()) {
      this._impl.delete(this, 'right', options);
    }
  }

  deleteCharacter(direction) {
    const sel = this._selection;
    if (!sel || sel.isNull()) {
      // nothing
    } else if (!sel.isCollapsed()) {
      this.deleteSelection();
    } else {
      this._impl.delete(this, direction);
    }
  }

  insertText(text) {
    const sel = this._selection;
    if (sel && !sel.isNull()) {
      this._impl.insertText(this, text);
    }
  }

  // insert an inline node with given data at the current selection
  insertInlineNode(inlineNode) {
    const sel = this._selection;
    if (sel && !sel.isNull() && sel.isPropertySelection()) {
      return this._impl.insertInlineNode(this, inlineNode)
    }
  }

  insertBlockNode(blockNode) {
    const sel = this._selection;
    if (sel && !sel.isNull()) {
      return this._impl.insertBlockNode(this, blockNode)
    }
  }

  paste(content) {
    const sel = this._selection;
    if (sel && !sel.isNull()) {
      return this._impl.paste(this, content)
    }
  }

  switchTextType(nodeData) {
    const sel = this._selection;
    if (sel && !sel.isNull()) {
      return this._impl.switchTextType(this, nodeData)
    }
  }

  toggleList(params) {
    const sel = this._selection;
    if (sel && !sel.isNull()) {
      return this._impl.toggleList(this, params)
    }
  }

  indent() {
    const sel = this._selection;
    if (sel && !sel.isNull()) {
      return this._impl.indent(this)
    }
  }

  dedent() {
    const sel = this._selection;
    if (sel && !sel.isNull()) {
      return this._impl.dedent(this)
    }
  }

  /* Legacy low-level API */

  getIndex(...args) {
    return this._document.getIndex(...args)
  }

  getAnnotations(...args) {
    return this._document.getAnnotations(...args)
  }

  getSchema() {
    return this._document.getSchema()
  }

  createSnippet() {
    return this._document.createSnippet()
  }

}

class ChangeRecorder extends EditingInterface {

  constructor(doc) {
    super(new TransactionDocument(doc));
  }

  generateChange() {
    const ops = this._document.ops.slice();
    this._document.ops.length = 0;
    let change = new DocumentChange(ops, {}, {});
    change._extractInformation(this._document);
    return change
  }

}

/**
  Some helpers for working with Documents.

  @module
  @example

  ```js
  import { documentHelpers } from 'substance'
  documentHelpers.getPropertyAnnotationsForSelection(doc, sel)
  ```
*/
var documentHelpers = {
  getPropertyAnnotationsForSelection,
  getContainerAnnotationsForSelection,
  getTextForSelection,
  getMarkersForSelection,
  getChangeFromDocument,
  copyNode,
  deleteNode,
  deleteTextRange,
  deleteListRange,
  mergeListItems,
  isContainerAnnotation
};

/**
  For a given selection get all property annotations

  @param {Document} doc
  @param {Selection} sel
  @return {PropertyAnnotation[]} An array of property annotations.
          Returns an empty array when selection is a container selection.
*/
function getPropertyAnnotationsForSelection(doc, sel, options) {
  options = options || {};
  if (!sel.isPropertySelection()) {
    return []
  }
  let path = sel.getPath();
  let annotations = doc.getIndex('annotations').get(path, sel.start.offset, sel.end.offset);
  if (options.type) {
    annotations = filter(annotations, DocumentIndex.filterByType(options.type));
  }
  return annotations
}

/**
  For a given selection get all container annotations

  @param {Document} doc
  @param {Selection} sel
  @param {String} containerId
  @param {String} options.type provides only annotations of that type
  @return {Array} An array of container annotations
*/
function getContainerAnnotationsForSelection(doc, sel, containerId, options) {
  // ATTENTION: looking for container annotations is not as efficient as property
  // selections, as we do not have an index that has notion of the spatial extend
  // of an annotation. Opposed to that, common annotations are bound
  // to properties which make it easy to lookup.
  /* istanbul ignore next */
  if (!containerId) {
    throw new Error("'containerId' is required.")
  }
  options = options || {};
  let index = doc.getIndex('container-annotations');
  let annotations = index.get(containerId, options.type);
  annotations = filter(annotations, function(anno) {
    return sel.overlaps(anno.getSelection())
  });
  return annotations
}

/**
  @param {Document} doc
  @param {String} type
  @return {Boolean} `true` if given type is a {@link ContainerAnnotation}
*/
function isContainerAnnotation(doc, type) {
  let schema = doc.getSchema();
  return schema.isInstanceOf(type, 'container-annotation')
}

/**
  For a given selection, get the corresponding text string

  @param {Document} doc
  @param {Selection} sel
  @return {string} text enclosed by the annotation
*/
function getTextForSelection(doc, sel) {
  if (!sel || sel.isNull()) {
    return ""
  } else if (sel.isPropertySelection()) {
    let text = doc.get(sel.start.path);
    return text.substring(sel.start.offset, sel.end.offset)
  } else if (sel.isContainerSelection()) {
    let result = [];
    let nodeIds = sel.getNodeIds();
    let L = nodeIds.length;
    for (let i = 0; i < L; i++) {
      let id = nodeIds[i];
      let node = doc.get(id);
      if (node.isText()) {
        let text = node.getText();
        if (i === L-1) {
          text = text.slice(0, sel.end.offset);
        }
        if (i === 0) {
          text = text.slice(sel.start.offset);
        }
        result.push(text);
      }
    }
    return result.join('\n')
  }
}

function getMarkersForSelection(doc, sel) {
  // only PropertySelections are supported right now
  if (!sel || !sel.isPropertySelection()) return []
  const path = sel.getPath();
  // markers are stored as one hash for each path, grouped by marker key
  let markers = doc.getIndex('markers').get(path);
  const filtered = filter(markers, function(m) {
    return m.containsSelection(sel)
  });
  return filtered
}

function getChangeFromDocument(doc) {
  let recorder = new ChangeRecorder(doc);
  return recorder.generateChange()
}

/*
  Deletes a node and its children and attached annotations
  and removes it from a given container
*/
function deleteNode(doc, node) {
  /* istanbul ignore next */
  if (!node) {
    console.warn('Invalid arguments');
    return
  }
  // TODO: bring back support for container annotations
  if (node.isText()) {
    // remove all associated annotations
    let annos = doc.getIndex('annotations').get(node.id);
    for (let i = 0; i < annos.length; i++) {
      doc.delete(annos[i].id);
    }
  }
  // delete recursively
  // ATM we do a cascaded delete if the property has type 'id' or ['array', 'id'] and property 'owned' set,
  // or if it 'file'
  let nodeSchema = node.getSchema();
  forEach(nodeSchema, (prop) => {
    if ((prop.isReference() && prop.isOwned()) || (prop.type === 'file')) {
      if (prop.isArray()) {
        let ids = node[prop.name];
        ids.forEach((id) => {
          deleteNode(doc, doc.get(id));
        });
      } else {
        deleteNode(doc, doc.get(node[prop.name]));
      }
    }
  });
  doc.delete(node.id);
}

/*
  Creates a 'deep' JSON copy of a node returning an array of JSON objects
  that can be used to create the object tree owned by the given root node.

  @param {DocumentNode} node
*/
function copyNode(node) {
  let nodes = [];
  // EXPERIMENTAL: using schema reflection to determine whether to do a 'deep' copy or just shallow
  let nodeSchema = node.getSchema();
  let doc = node.getDocument();
  forEach(nodeSchema, (prop) => {
    // ATM we do a cascaded copy if the property has type 'id', ['array', 'id'] and is owned by the node,
    // or it is of type 'file'
    if ((prop.isReference() && prop.isOwned()) || (prop.type === 'file')) {
      let val = node[prop.name];
      nodes.push(_copyChildren(val));
    }
  });
  nodes.push(node.toJSON());
  let annotationIndex = node.getDocument().getIndex('annotations');
  let annotations = annotationIndex.get([node.id]);
  forEach(annotations, function(anno) {
    nodes.push(anno.toJSON());
  });
  let result = flatten(nodes).filter(Boolean);
  // console.log('copyNode()', node, result)
  return result

  function _copyChildren(val) {
    if (!val) return null
    if (isArray$1(val)) {
      return flatten(val.map(_copyChildren))
    } else {
      let id = val;
      if (!id) return null
      let child = doc.get(id);
      if (!child) return
      return copyNode(child)
    }
  }
}

/*
  <-->: anno
  |--|: area of change
  I: <--> |--|     :   nothing
  II: |--| <-->    :   move both by total span
  III: |-<-->-|    :   delete anno
  IV: |-<-|->      :   move start by diff to start, and end by total span
  V: <-|->-|       :   move end by diff to start
  VI: <-|--|->     :   move end by total span
*/
function deleteTextRange(doc, start, end) {
  if (!start) {
    start = {
      path: end.path,
      offset: 0
    };
  }
  let path = start.path;
  let text = doc.get(path);
  if (!end) {
    end = {
      path: start.path,
      offset: text.length
    };
  }
  /* istanbul ignore next */
  if (!isArrayEqual(start.path, end.path)) {
    throw new Error('start and end must be on one property')
  }
  let startOffset = start.offset;
  let endOffset = end.offset;
  doc.update(path, { type: 'delete', start: startOffset, end: endOffset });
  // update annotations
  let annos = doc.getAnnotations(path);
  annos.forEach(function(anno) {
    let annoStart = anno.start.offset;
    let annoEnd = anno.end.offset;
    // I anno is before
    if (annoEnd<=startOffset) {
      return
    }
    // II anno is after
    else if (annoStart>=endOffset) {
      doc.update([anno.id, 'start'], { type: 'shift', value: startOffset-endOffset });
      doc.update([anno.id, 'end'], { type: 'shift', value: startOffset-endOffset });
    }
    // III anno is deleted
    else if (annoStart>=startOffset && annoEnd<=endOffset) {
      doc.delete(anno.id);
    }
    // IV anno.start between and anno.end after
    else if (annoStart>=startOffset && annoEnd>=endOffset) {
      if (annoStart>startOffset) {
        doc.update([anno.id, 'start'], { type: 'shift', value: startOffset-annoStart });
      }
      doc.update([anno.id, 'end'], { type: 'shift', value: startOffset-endOffset });
    }
    // V anno.start before and anno.end between
    else if (annoStart<=startOffset && annoEnd<=endOffset) {
      doc.update([anno.id, 'end'], { type: 'shift', value: startOffset-annoEnd });
    }
    // VI anno.start before and anno.end after
    else if (annoStart<startOffset && annoEnd >= endOffset) {
      doc.update([anno.id, 'end'], { type: 'shift', value: startOffset-endOffset });
    }
    else {
      console.warn('TODO: handle annotation update case.');
    }
  });
}

function deleteListRange(doc, list, start, end) {
  if (doc !== list.getDocument()) {
    list = doc.get(list.id);
  }
  if (!start) {
    start = {
      path: list.getItemAt(0).getTextPath(),
      offset: 0
    };
  }
  if (!end) {
    let item = list.getLastItem();
    end = {
      path: item.getTextPath(),
      offset: item.getLength()
    };
  }
  let startId = start.path[0];
  let startPos = list.getItemPosition(startId);
  let endId = end.path[0];
  let endPos = list.getItemPosition(endId);
  // range within the same item
  if (startPos === endPos) {
    deleteTextRange(doc, start, end);
    return
  }
  // normalize the range if it is 'reverse'
  if (startPos > endPos) {
    [start, end] = [end, start];
    [startPos, endPos] = [endPos, startPos];
    [startId, endId] = [endId, startId];
  }
  let firstItem = doc.get(startId);
  let lastItem = doc.get(endId);
  let firstEntirelySelected = isEntirelySelected(doc, firstItem, start, null);
  let lastEntirelySelected = isEntirelySelected(doc, lastItem, null, end);

  // delete or truncate last node
  if (lastEntirelySelected) {
    list.removeItemAt(endPos);
    deleteNode(doc, lastItem);
  } else {
    deleteTextRange(doc, null, end);
  }

  // delete inner nodes
  for (let i = endPos-1; i > startPos; i--) {
    let itemId = list.items[i];
    list.removeItemAt(i);
    deleteNode(doc, doc.get(itemId));
  }

  // delete or truncate the first node
  if (firstEntirelySelected) {
    list.removeItemAt(startPos);
    deleteNode(doc, firstItem);
  } else {
    deleteTextRange(doc, start, null);
  }

  if (!firstEntirelySelected && !lastEntirelySelected) {
    mergeListItems(doc, list.id, startPos);
  }
}

function mergeListItems(doc, listId, itemPos) {
  // HACK: make sure that the list is really from the doc
  let list = doc.get(listId);
  let target = list.getItemAt(itemPos);
  let targetPath = target.getTextPath();
  let targetLength = target.getLength();
  let source = list.getItemAt(itemPos+1);
  let sourcePath = source.getTextPath();
  // hide source
  list.removeItemAt(itemPos+1);
  // append the text
  doc.update(targetPath, { type: 'insert', start: targetLength, text: source.getText() });
  // transfer annotations
  annotationHelpers.transferAnnotations(doc, sourcePath, 0, targetPath, targetLength);
  doc.delete(source.id);
}

class Annotation extends DocumentNode {

  constructor(doc, props) {
    super(doc, _normalizedProps(props));

    // wrap coordinates
    this.start = new Coordinate(this.start);
    this.end = new Coordinate(this.end);
  }

  /* istanbul ignore start */

  get startPath() {
    console.warn('DEPRECATED: use Annotation.start.path instead.');
    return this.start.path
  }

  set startPath(path) {
    console.warn('DEPRECATED: use Annotation.start.path instead.');
    this.start.path = path;
  }

  get startOffset() {
    console.warn('DEPRECATED: use Annotation.start.offset instead.');
    return this.start.offset
  }

  set startOffset(offset) {
    console.warn('DEPRECATED: use Annotation.start.offset instead.');
    this.start.offset = offset;
  }

  get endPath() {
    console.warn('DEPRECATED: use Annotation.end.path instead.');
    return this.end.path
  }

  set endPath(path) {
    console.warn('DEPRECATED: use Annotation.end.path instead.');
    this.end.path = path;
  }

  get endOffset() {
    console.warn('DEPRECATED: use Annotation.end.offset instead.');
    return this.end.offset
  }

  set endOffset(offset) {
    console.warn('DEPRECATED: use Annotation.end.offset instead.');
    this.end.offset = offset;
  }

  /* istanbul ignore end */

  /**
    Get the plain text spanned by this annotation.

    @return {String}
  */
  getText() {
    var doc = this.getDocument();
    if (!doc) {
      console.warn('Trying to use a Annotation which is not attached to the document.');
      return ""
    }
    return documentHelpers.getTextForSelection(doc, this.getSelection())
  }

  /**
    Determines if an annotation can be split e.g., when breaking a node.

    In these cases, a new annotation will be created attached to the created node.

    For certain annotation types,you may want to the annotation truncated
    rather than split, where you need to override this method returning `false`.
  */
  canSplit() {
    return true
  }

  /**
    If this annotation is an Anchor.

    Anchors are annotations with a zero width.
    For instance, ContainerAnnotation have a start and an end anchor,
    or rendered cursors are modeled as anchors.

    @returns {Boolean}
  */
  isAnchor() {
    return false
  }
}

Annotation.define({
  type: "annotation",
  start: "coordinate",
  end: "coordinate"
});

Annotation.prototype._isAnnotation = true;

function _normalizedProps(props) {
  if (!props.hasOwnProperty('start')) {
    /*
      Instead of
        { path: [...], startOffset: 0, endOffset: 10 }
      use
        { start: { path: [], offset: 0 }, end: { path: [], offset: 10 } }
    */
    // console.warn('DEPRECATED: create Annotation with "start" and "end" coordinate instead.')
    props = Object.assign({}, props);
    props.start = {
      path: props.startPath || props.path,
      offset: props.startOffset
    };
    props.end = {};
    if (props.hasOwnProperty('endPath')) {
      props.end.path = props.endPath;
    } else {
      props.end.path = props.start.path;
    }
    if (props.hasOwnProperty('endOffset')) {
      props.end.offset = props.endOffset;
    } else {
      props.end.offset = props.start.offset;
    }
    delete props.path;
    delete props.startPath;
    delete props.endPath;
    delete props.startOffset;
    delete props.endOffset;
  } else if (props.hasOwnProperty('end') && !props.end.path) {
    props.end.path = props.start.path;
  }
  return props
}

class BlockNode extends DocumentNode {}

BlockNode.isBlock = true;

class ChangeHistory {

  constructor() {
    // undo list
    this.doneChanges = [];
    // redo list
    this.undoneChanges = [];
    // last change for accumlation
    this.lastChange = null;
  }

  canUndo() {
    return this.doneChanges.length > 0
  }

  canRedo() {
    return this.undoneChanges.length > 0
  }

  push(change) {
    this.doneChanges.push(change);
    this.undoneChanges = [];
  }

}

class ContainerAddress {

  constructor(pos, offset) {
    this.pos = pos;
    this.offset = offset;
  }

  isBefore(other, strict) {
    strict = Boolean(strict);
    if (this.pos < other.pos) {
      return true
    } else if (this.pos > other.pos) {
      return false
    } else if (this.offset < other.offset) {
      return true
    } else if (this.offset > other.offset) {
      return false
    }
    if (strict) {
      return false
    } else {
      return true
    }
  }

  isAfter(other, strict) {
    return other.isBefore(this, strict)
  }

  isEqual(other) {
    return (this.pos === other.pos && this.offset === other.offset)
  }

  toString() {
    return [this.pos,'.',this.offset].join('')
  }
}

/*
  A Container represents a list of nodes.

  While most editing occurs on a property level (such as editing text),
  other things happen on a node level, e.g., breaking or mergin nodes,
  or spanning annotations so called ContainerAnnotations.
*/
class Container extends DocumentNode {

  constructor(...args) {
    super(...args);

    // NOTE: we are caching positions as they are queried very often,
    // whereas the number of changes to a container are quite rare.
    // The cache gets invalidated whenever the container is changed.
    this._enableCaching();
  }

  dispose() {
    this.document.off(this);
  }

  getContentPath() {
    return [this.id, 'nodes']
  }

  getContent() {
    return this.nodes
  }

  contains(nodeId) {
    return this.getPosition(nodeId) >= 0
  }

  getPosition(node, strict) {
    if (isString$1(node)) {
      node = this.document.get(node);
    }
    if (!node) return -1
    let pos = this._getPosition(node);
    if (strict && pos < 0) {
      throw new Error('Node is not within this container: ' + node.id)
    }
    return pos
  }

  getNodeAt(idx) {
    let content = this.getContent();
    if (idx < 0 || idx >= content.length) {
      throw new Error('Array index out of bounds: ' + idx + ", " + content.length)
    }
    return this.getDocument().get(content[idx])
  }

  getNodes() {
    const doc = this.getDocument();
    return this.getContent().map(id => doc.get(id)).filter(Boolean)
  }

  show(nodeId, pos) {
    var doc = this.getDocument();
    var arg1 = arguments[0];
    if (!isString$1(arg1)) {
      if (arg1._isNode) {
        nodeId = arg1.id;
      }
    }
    if (!isNumber(pos)) {
      pos = this.getLength();
    }
    doc.update(this.getContentPath(), { type: 'insert', pos: pos, value: nodeId });
  }

  hide(nodeId) {
    var doc = this.getDocument();
    var pos = this.getPosition(nodeId);
    if (pos >= 0) {
      doc.update(this.getContentPath(), { type: 'delete', pos: pos });
    }
  }

  getAddress(coor) {
    if (!coor._isCoordinate) {
      // we have broken with an earlier version of this API
      throw new Error('Illegal argument: Container.getAddress(coor) expects a Coordinate instance.')
    }
    var nodeId = coor.path[0];
    var nodePos = this.getPosition(nodeId);
    var offset;
    if (coor.isNodeCoordinate()) {
      if (coor.offset > 0) {
        offset = Number.MAX_VALUE;
      } else {
        offset = 0;
      }
    } else {
      offset = coor.offset;
    }
    return new ContainerAddress(nodePos, offset)
  }

  getLength() {
    return this.getContent().length
  }

  get length() {
    return this.getLength()
  }

  _getPosition(node) {
    if (this._isCaching) {
      return this._getCachedPosition(node)
    } else {
      return this._lookupPosition(node)
    }
  }

  _getCachedPosition(node) {
    let cache = this._cachedPositions || this._fillCache();
    let nodeId = node.id;
    let pos = -1;
    if (cache.hasOwnProperty(nodeId)) {
      pos = cache[nodeId];
    } else {
      pos = this._lookupPosition(node);
      cache[nodeId] = pos;
    }
    return pos
  }

  _fillCache() {
    let positions = {};
    this.nodes.forEach((id, pos) => {
      positions[id] = pos;
    });
    this._cachedPositions = positions;
    return positions
  }

  _invalidateCache() {
    this._cachedPositions = null;
  }

  _lookupPosition(node) {
    if (node.hasParent()) {
      node = node.getRoot();
    }
    return this.getContent().indexOf(node.id)
  }

  _enableCaching() {
    // this hook is used to invalidate cached positions
    // caching is done only in the 'real' document, not in a TransactionDocument
    if (this.document) {
      this.document.data.on('operation:applied', this._onOperationApplied, this);
      this._isCaching = true;
    }
  }

  _onOperationApplied(op) {
    if (op.type === 'set' || op.type === 'update') {
      if (op.path[0] === this.id) {
        this._invalidateCache();
      }
    }
  }

  _onDocumentChange(change) {
    if (change.hasUpdated(this.getContentPath())) {
      this._invalidateCache();
    }
  }

  // NOTE: this has been in ParentNodeMixin before
  // TODO: try to get rid of this

  hasChildren() {
    return this.nodes.length > 0
  }

  getChildIndex(child) {
    return this.nodes.indexOf(child.id)
  }

  getChildren() {
    var doc = this.getDocument();
    var childrenIds = this.nodes;
    return childrenIds.map(function(id) {
      return doc.get(id)
    })
  }

  getChildAt(idx) {
    var childrenIds = this.nodes;
    if (idx < 0 || idx >= childrenIds.length) {
      throw new Error('Array index out of bounds: ' + idx + ", " + childrenIds.length)
    }
    return this.getDocument().get(childrenIds[idx])
  }

  getChildCount() {
    return this.nodes.length
  }

}

Container.prototype._isContainer = true;

Container.schema = {
  type: 'container',
  nodes: { type: ['array', 'id'], default: [] }
};

/*
  Describes an annotation sticking on a container that can span over multiple
  nodes.

  @class

  @example

  ```js
  {
    "id": "subject_reference_1",
    "type": "subject_reference",
    "containerId": "content",
    "start": {
      "path": ["text_2", "content"],
      "offset": 100,
    },
    "end": {
      "path": ["text_4", "content"],
      "offset": 40
    }
  }
  ```
 */

class ContainerAnnotation extends Annotation {

  /**
    Provides a selection which has the same range as this annotation.

    @return {model/ContainerSelection}
  */
  getSelection() {
    var doc = this.getDocument();
    // Guard: when this is called while this node has been detached already.
    if (!doc) {
      console.warn('Trying to use a ContainerAnnotation which is not attached to the document.');
      return Selection.nullSelection()
    }
    return doc.createSelection({
      type: "container",
      containerId: this.containerId,
      startPath: this.start.path,
      startOffset: this.start.offset,
      endPath: this.end.path,
      endOffset: this.end.offset
    })
  }

  setHighlighted(highlighted, scope) {
    if (this.highlighted !== highlighted) {
      this.highlighted = highlighted;
      this.highlightedScope = scope;
      this.emit('highlighted', highlighted, scope);
      forEach(this.fragments, function(frag) {
        frag.emit('highlighted', highlighted, scope);
      });
    }
  }

  _updateRange(tx, sel) {
    if (!sel.isContainerSelection()) {
      throw new Error('Invalid argument.')
    }
    // TODO: use coordinate ops
    if (!isEqual(this.start.path, sel.start.path)) {
      tx.set([this.id, 'start', 'path'], sel.start.path);
    }
    if (this.start.offset !== sel.start.offset) {
      tx.set([this.id, 'start', 'offset'], sel.start.offset);
    }
    if (!isEqual(this.end.path, sel.end.path)) {
      tx.set([this.id, 'end', 'path'], sel.end.path);
    }
    if (this.end.offset !== sel.end.offset) {
      tx.set([this.id, 'end', 'offset'], sel.end.offset);
    }
  }
}

ContainerAnnotation.schema = {
  type: "container-annotation",
  containerId: "string"
};

ContainerAnnotation.prototype._isContainerAnnotation = true;

/*
  A FileProxy represents a proxy from a FileNode to the real resource.
  As real resources may need to be fetched, the FileProxy typically has
  multiple internal states.
*/
class FileProxy {
  constructor(fileNode, context) {
    this.fileNode = fileNode;
    this.context = context;
    fileNode.setProxy(this);
  }

  get id() {
    return this.fileNode.id
  }

  /*
    Fires a property update on the file node
  */
  triggerUpdate() {
    let fileId = this.fileNode.id;
    this.context.editorSession.transaction((tx) => {
      tx.set([fileId, '__changed__'], '');
    }, { history: false });
  }

  getUrl() {
    return ''
  }

  sync() {
    return Promise.reject(new Error('sync method not implemented'))
  }
}

FileProxy.match = function(fileNode, context) { // eslint-disable-line
  return false
};

/*
  Generic implementation of a file proxy that works with
  any file type
*/
class DefaultFileProxy extends FileProxy {

  constructor(fileNode, context) {
    super(fileNode, context);

    // used locally e.g. after drop or file dialog
    this.file = fileNode.sourceFile;
    if (this.file) {
      this._fileUrl = URL.createObjectURL(this.file);
    }
    this.url = fileNode.url;
  }

  getUrl() {
    // if we have fetched the url already, just serve it here
    if (this.url) {
      return this.url
    }
    // if we have a local file, use it's data URL
    if (this._fileUrl) {
      return this._fileUrl
    }
    // no URL available
    return ''
  }

  sync() {
    if (!this.url) {
      console.info('Simulating file upload. Creating blob url instead.', this._fileUrl);
      this.url = this._fileUrl;
    }
    return Promise.resolve()
  }
}

/*
  Registry for Nodes.

  @class NodeRegistry
  @extends util/Registry
 */
class NodeRegistry extends Registry {
  /**
    Register a Node class.

    @param {Class} nodeClass
   */
  register(nodeClazz) {
    var type = nodeClazz.prototype.type;
    if ( typeof type !== 'string' || type === '' ) {
      console.error('#### nodeClazz', nodeClazz);
      throw new Error( 'Node names must be strings and must not be empty')
    }
    if ( !( nodeClazz.prototype._isNode) ) {
      throw new Error( 'Nodes must be subclasses of Substance.Data.Node' )
    }
    if (this.contains(type)) {
      throw new Error('Node class is already registered: ' + type)
    }
    this.add(type, nodeClazz);
  }

}

/**
  Schema for Data Objects.

  @internal
 */
class Schema {

  /**
    @param {String} name
    @param {String} version
  */
  constructor(name, version) {
    /**
      @type {String}
    */
    this.name = name;
    /**
      @type {String}
    */
    this.version = version;
    /**
      @type {NodeRegistry}
      @private
    */
    this.nodeRegistry = new NodeRegistry();
    /**
      @type {Array} all Node classes which have `Node.tocType = true`
      @private
    */
    this.tocTypes = [];

    // add built-in node classes
    this.addNodes(this.getBuiltIns());
  }

  /**
    Add nodes to the schema.

    @param {Array} nodes Array of Node classes
  */
  addNodes(nodes) {
    if (!nodes) return
    forEach(nodes, function(NodeClass) {
      if (!NodeClass.prototype._isNode) {
        console.error('Illegal node class: ', NodeClass);
      } else {
        this.addNode(NodeClass);
      }
    }.bind(this));
  }

  addNode(NodeClass) {
    this.nodeRegistry.register(NodeClass);
    if (NodeClass.tocType) {
      this.tocTypes.push(NodeClass.type);
    }
  }

  /**
    Get the node class for a type name.

    @param {String} name
    @returns {Class}
  */
  getNodeClass(name) {
    return this.nodeRegistry.get(name)
  }

  /**
    Provide all built-in node classes.

    @private
    @returns {Node[]} An array of Node classes.
  */
  getBuiltIns() {
    return []
  }

  /**
    Checks if a given type is of given parent type.

    @param {String} type
    @param {String} parentType
    @returns {Boolean} true if type is and instance of parentType.
  */
  isInstanceOf(type, parentType) {
    var NodeClass = this.getNodeClass(type);
    if (NodeClass) {
      return Node.isInstanceOf(NodeClass, parentType)
    }
    return false
  }

  /**
    Iterate over all registered node classes.

    See {@link util/Registry#each}

    @param {Function} callback
    @param {Object} context
  */
  each() {
    this.nodeRegistry.each.apply(this.nodeRegistry, arguments);
  }

  /**
    @returns {Node[]} list of types that should appear in a TOC
  */
  getTocTypes() {
    return this.tocTypes
  }

  /**
    @returns {String} the name of the default textish node (e.g. 'paragraph')
  */
  getDefaultTextType() {
    throw new Error('Schmema.prototype.getDefaultTextType() must be overridden.')
  }

  getNodeSchema(type) {
    var NodeClass = this.getNodeClass(type);
    if (!NodeClass) {
      console.error('Unknown node type ', type);
      return null
    }
    return NodeClass.schema
  }
}

/**
  A property annotation can be used to overlay text and give it a special meaning.
  PropertyAnnotations only work on text properties. If you want to annotate multiple
  nodes you have to use a {@link model/ContainerAnnotation}.

  @prop {String[]} path Identifies a text property in the document (e.g. `['text_1', 'content']`)
  @prop {Number} startOffset the character where the annoation starts
  @prop {Number} endOffset: the character where the annoation starts

  @example

  Here's how a **strong** annotation is created. In Substance annotations are stored
  separately from the text. Annotations are just regular nodes in the document.
  They refer to a certain range (`startOffset, endOffset`) in a text property (`path`).

  ```js
  doc.transaction(function(tx) {
    tx.create({
      id: 's1',
      type: 'strong',
      start: {
        path: ['p1', 'content'],
        offset: 10
      },
      end: {
        offset
      }
      path: ['p1', 'content'],
      "startOffset": 10,
      "endOffset": 19
    })
  })
  ```
*/
class PropertyAnnotation extends Annotation {

  get path() {
    return this.start.path
  }

  getPath() {
    return this.start.path
  }

  getSelection() {
    return this.getDocument().createSelection({
      type: 'property',
      path: this.path,
      startOffset: this.start.offset,
      endOffset: this.end.offset
    })
  }

  // used by annotationHelpers
  _updateRange(tx, sel) {
    if (!sel.isPropertySelection()) {
      throw new Error('Invalid argument: PropertyAnnotation._updateRange() requires a PropertySelection.')
    }
    if (!isArrayEqual(this.start.path, sel.start.path)) {
      tx.set([this.id, 'path'], sel.start.path);
    }
    // TODO: these should be Coordinate ops
    if (this.start.offset !== sel.start.offset) {
      tx.set([this.id, 'start', 'offset'], sel.start.offset);
    }
    if (this.end.offset !== sel.end.offset) {
      tx.set([this.id, 'end', 'offset'], sel.end.offset);
    }
  }

  get startPath() {
    return this.path
  }

  get endPath() {
    return this.path
  }
}

PropertyAnnotation.prototype._isAnnotation = true;
PropertyAnnotation.prototype._isPropertyAnnotation = true;

PropertyAnnotation.isPropertyAnnotation = true;
PropertyAnnotation.autoExpandRight = true;

PropertyAnnotation.schema = {
  type: "annotation",
  start: "coordinate",
  end: "coordinate",
  // this is only used when an annotation is used 'stand-alone'
  // i.e. not attached to a property
  _content: { type: "string", optional: true}
};

class DocumentSchema extends Schema {

  constructor({ name, DocumentClass, defaultTextType='text', version='0.0.0' }) {
    super(name, version);

    /* istanbul ignore next */
    if (!DocumentClass) {
      throw new Error('DocumentClass is mandatory')
    }

    this.DocumentClass = DocumentClass;
    this.defaultTextType = defaultTextType;
  }

  getDocumentClass() {
    return this.DocumentClass
  }

  /*
    @override
  */
  getDefaultTextType() {
    return this.defaultTextType
  }

  /*
    @override
  */
  getBuiltIns() {
    return [DocumentNode, PropertyAnnotation, Container, ContainerAnnotation]
  }

}

const ENTER = 1;
const EXIT = -1;
const ANCHOR = -2;

// Fragmenter
// --------
//
// An algorithm that is used to fragment overlapping structure elements
// following a priority rule set.
// E.g., we use this for creating DOM elements for annotations. The annotations
// can partially be overlapping. However this is not allowed in general for DOM elements
// or other hierarchical structures.
//
// Example: For the annotation use case consider a 'comment' spanning partially
// over an 'emphasis' annotation.
// 'The <comment>quick brown <bold>fox</comment> jumps over</bold> the lazy dog.'
// We want to be able to create a valid XML structure:
// 'The <comment>quick brown <bold>fox</bold></comment><bold> jumps over</bold> the lazy dog.'
//
// For that one would choose
//
//     {
//        'comment': 0,
//        'bold': 1
//     }
//
// as priority levels.
// In case of structural violations as in the example, elements with a higher level
// would be fragmented and those with lower levels would be preserved as one piece.
//
// TODO: If a violation for nodes of the same level occurs an Error should be thrown.
// Currently, in such cases the first element that is opened earlier is preserved.

class Fragmenter {

  constructor(options) {
    Object.assign(this, options);
  }

  start(rootContext, text, annotations) {
    if (!isString$1(text)) {
      throw new Error("Illegal argument: 'text' must be a String, but was " + text)
    }
    this._start(rootContext, text, annotations);
  }

  onText(context, text, entry) { // eslint-disable-line
  }

  // should return the created user context
  onEnter(entry, parentContext) { // eslint-disable-line
    return null
  }

  onExit(entry, context, parentContext) { // eslint-disable-line
  }

  _enter(entry, parentContext) {
    entry.counter++;
    return this.onEnter(entry, parentContext)
  }

  _exit(entry, context, parentContext) {
    this.onExit(entry, context, parentContext);
  }

  _createText(context, text, entry) {
    this.onText(context, text, entry);
  }

  _start(rootContext, text, annotations) {
    var entries = _extractEntries.call(this, annotations);
    var stack = [{context: rootContext, entry: null}];

    var pos = 0;
    for (var i = 0; i < entries.length; i++) {
      var entry = entries[i];
      var textFragment = text.substring(pos, entry.pos);
      if (textFragment) {
        // add the last text to the current element
        this._createText(stack[stack.length-1].context, textFragment, entry);
      }

      pos = entry.pos;
      var stackLevel, idx, _entry;
      if (entry.mode === ENTER || entry.mode === ANCHOR) {
        // find the correct position and insert an entry
        for (stackLevel = 1; stackLevel < stack.length; stackLevel++) {
          if (entry.level < stack[stackLevel].entry.level) {
            break
          }
        }
        // create elements which are open, and are now stacked ontop of the
        // entered entry
        for (idx = stack.length-1; idx >= stackLevel; idx--) {
          _entry = stack[idx].entry;
          // compute number of characters since last 'enter'
          _entry.length = pos - _entry.pos;
          this._exit(_entry, stack[idx].context, stack[idx-1].context);
        }
        stack.splice(stackLevel, 0, {entry: entry});
        // create new elements for all lower entries
        for (idx = stackLevel; idx < stack.length; idx++) {
          _entry = stack[idx].entry;
          // bump 'enter' pos
          _entry.pos = pos;
          stack[idx].context = this._enter(_entry, stack[idx-1].context);
        }
      }
      if (entry.mode === EXIT || entry.mode === ANCHOR) {
        // find the according entry and remove it from the stack
        for (stackLevel = 1; stackLevel < stack.length; stackLevel++) {
          if (stack[stackLevel].entry.node === entry.node) {
            break
          }
        }
        for (idx = stack.length-1; idx >= stackLevel; idx--) {
          _entry = stack[idx].entry;
          // compute number of characters since last 'enter'
          _entry.length = pos - _entry.pos;
          this._exit(_entry, stack[idx].context, stack[idx-1].context);
        }
        stack.splice(stackLevel, 1);
        // create new elements for all lower entries
        for (idx = stackLevel; idx < stack.length; idx++) {
          _entry = stack[idx].entry;
          // bump 'enter' pos
          _entry.pos = pos;
          stack[idx].context = this._enter(_entry, stack[idx-1].context);
        }
      }
    }

    // Finally append a trailing text node
    var trailingText = text.substring(pos);
    if (trailingText) {
      this._createText(rootContext, trailingText);
    }
  }

}

Fragmenter.SHOULD_NOT_SPLIT = 0;
Fragmenter.NORMAL = 10;
Fragmenter.ANY = 100;
Fragmenter.ALWAYS_ON_TOP = Number.MAX_VALUE;

// This is a sweep algorithm wich uses a set of ENTER/EXIT entries
// to manage a stack of active elements.
// Whenever a new element is entered it will be appended to its parent element.
// The stack is ordered by the annotation types.
//
// Examples:
//
// - simple case:
//
//       [top] -> ENTER(idea1) -> [top, idea1]
//
//   Creates a new 'idea' element and appends it to 'top'
//
// - stacked ENTER:
//
//       [top, idea1] -> ENTER(bold1) -> [top, idea1, bold1]
//
//   Creates a new 'bold' element and appends it to 'idea1'
//
// - simple EXIT:
//
//       [top, idea1] -> EXIT(idea1) -> [top]
//
//   Removes 'idea1' from stack.
//
// - reordering ENTER:
//
//       [top, bold1] -> ENTER(idea1) -> [top, idea1, bold1]
//
//   Inserts 'idea1' at 2nd position, creates a new 'bold1', and appends itself to 'top'
//
// - reordering EXIT
//
//       [top, idea1, bold1] -> EXIT(idea1)) -> [top, bold1]
//
//   Removes 'idea1' from stack and creates a new 'bold1'
//

function _extractEntries(annotations) {
  var openers = [];
  var closers = [];
  forEach(annotations, function(a) {
    var isAnchor = (a.isAnchor ? a.isAnchor() : false);
    // special treatment for zero-width annos such as ContainerAnnotation.Anchors
    if (isAnchor) {
      openers.push({
        mode: ANCHOR,
        pos: a.offset,
        id: a.id,
        level: Fragmenter.ALWAYS_ON_TOP,
        type: 'anchor',
        node: a,
        counter: -1,
        length: 0
      });
    } else {
      // TODO better naming, `Node.level` does not say enough
      // Better would be `Node.fragmentation = Fragmenter.SHOULD_NOT_SPLIT;`
      // meaning, that the fragmenter should try to render the fragment as one single
      // element, and not splitting it up on different stack levels.
      // E.g. When bold an link are overlapping
      // the fragmenter should not split the link element such as A<b>B<a>CD</a></b><a>EF</a>GH
      // but should instead A<b>B</b><a><b>CD</b><a>EF</a>GH

      // use a weak default level when not given
      var l = Fragmenter.NORMAL;
      var isInline = (a.isInline ? a.isInline() : false);
      if (isInline) {
        l = Number.MAX_VALUE;
      } else if (a.constructor.hasOwnProperty('fragmentation')) {
        l = a.constructor.fragmentation;
      } else if (a.hasOwnProperty('fragmentationHint')) {
        l = a.fragmentationHint;
      }
      var startOffset = Math.min(a.start.offset, a.end.offset);
      var endOffset = Math.max(a.start.offset, a.end.offset);
      var opener = {
        pos: startOffset,
        mode: ENTER,
        level: l,
        id: a.id,
        type: a.type,
        node: a,
        length: 0,
        counter: -1,
      };
      openers.push(opener);
      closers.push({
        pos: endOffset,
        mode: EXIT,
        level: l,
        id: a.id,
        type: a.type,
        node: a,
        opener: opener
      });
    }
  });

  // sort the openers
  openers.sort(_compareOpeners);
  // store indexes for openers
  for (var i = openers.length - 1; i >= 0; i--) {
    openers[i].idx = i;
  }
  closers.sort(_compareClosers);
  // merge openers and closers, sorted by pos
  var entries = new Array(openers.length+closers.length);
  var idx = 0;
  var idx1 = 0;
  var idx2 = 0;
  var opener = openers[idx1];
  var closer = closers[idx2];
  while(opener || closer) {
    if (opener && closer) {
      // close before open
      if (closer.pos <= opener.pos && closer.opener !== opener) {
        entries[idx] = closer;
        idx2++;
      } else {
        entries[idx] = opener;
        idx1++;
      }
    } else if (opener) {
      entries[idx] = opener;
      idx1++;
    } else if (closer) {
      entries[idx] = closer;
      idx2++;
    }
    opener = openers[idx1];
    closer = closers[idx2];
    idx++;
  }
  return entries
}

function _compareOpeners(a, b) {
  if (a.pos < b.pos) return -1
  if (a.pos > b.pos) return 1
  if (a.mode < b.mode) return -1
  if (a.mode > b.mode) return 1
  if (a.mode === b.mode) {
    if (a.level < b.level) return -1
    if (a.level > b.level) return 1
  }
  return 0
}

// sort in inverse order of openers
function _compareClosers(a, b) {
  if (a.pos < b.pos) return -1
  if (a.pos > b.pos) return 1
  // this makes closer be sorted in inverse order of openers
  // to reduce stack sice
  // HACK: a bit trial error. When we have to collapsed annotations
  // at the same position then we want the closers in the same order
  // as the openers.
  if (a.pos === a.opener.pos && b.pos === b.opener.pos) {
    if (a.opener.idx < b.opener.idx) {
      return -1
    } else {
      return 1
    }
  }
  if (a.opener.idx > b.opener.idx) return -1
  if (a.opener.idx < b.opener.idx) return 1
  return 0
}

class DOMExporter {

  constructor(config, context) {
    this.context = context || {};
    if (!config.converters) {
      throw new Error('config.converters is mandatory')
    }
    if (!config.converters._isRegistry) {
      this.converters = new Registry();
      config.converters.forEach(function(Converter) {
        let converter = isFunction$1(Converter) ? new Converter() : Converter;
        if (!converter.type) {
          console.error('Converter must provide the type of the associated node.', converter);
          return
        }
        this.converters.add(converter.type, converter);
      }.bind(this));
    } else {
      this.converters = config.converters;
    }

    this.state = {
      doc: null
    };
    this.config = config;
    // NOTE: Subclasses (HTMLExporter and XMLExporter) must initialize this
    // with a proper DOMElement instance which is used to create new elements.
    this._elementFactory = config.elementFactory;
    if (!this._elementFactory) {
      throw new Error("'elementFactory' is mandatory")
    }
    this.$$ = this.createElement.bind(this);
  }

  exportDocument(doc) {
    // TODO: this is no left without much functionality
    // still, it would be good to have a consistent top-level API
    // i.e. converter.importDocument(el) and converter.exportDocument(doc)
    // On the other side, the 'internal' API methods are named this.convert*.
    return this.convertDocument(doc)
  }

  /**
    @param {Document}
    @returns {DOMElement|DOMElement[]} The exported document as DOM or an array of elements
             if exported as partial, which depends on the actual implementation
             of `this.convertDocument()`.

    @abstract
    @example

    convertDocument(doc) {
      var elements = this.convertContainer(doc, this.state.containerId)
      var out = elements.map(function(el) {
        return el.outerHTML
      })
      return out.join('')
    }
  */
  convertDocument(doc) { // eslint-disable-line
    throw new Error('This method is abstract')
  }

  convertContainer(container) {
    if (!container) {
      throw new Error('Illegal arguments: container is mandatory.')
    }
    var doc = container.getDocument();
    this.state.doc = doc;
    var elements = [];
    container.nodes.forEach(function(id) {
      var node = doc.get(id);
      var nodeEl = this.convertNode(node);
      elements.push(nodeEl);
    }.bind(this));
    return elements
  }

  convertNode(node) {
    if (isString$1(node)) {
      // Assuming this.state.doc has been set by convertDocument
      node = this.state.doc.get(node);
    } else {
      this.state.doc = node.getDocument();
    }
    var converter = this.getNodeConverter(node);
    // special treatment for annotations, i.e. if someone calls
    // `exporter.convertNode(anno)`
    if (node._isPropertyAnnotation && (!converter || !converter.export)) {
      return this._convertPropertyAnnotation(node)
    }
    if (!converter) {
      converter = this.getDefaultBlockConverter();
    }
    var el;
    if (converter.tagName) {
      el = this.$$(converter.tagName);
    } else {
      el = this.$$('div');
    }
    el.attr(this.config.idAttribute, node.id);
    if (converter.export) {
      el = converter.export(node, el, this) || el;
    } else {
      el = this.getDefaultBlockConverter().export(node, el, this) || el;
    }
    return el
  }

  convertProperty(doc, path, options) {
    this.initialize(doc, options);
    var wrapper = this.$$('div')
      .append(this.annotatedText(path));
    return wrapper.innerHTML
  }

  annotatedText(path) {
    var doc = this.state.doc;
    var text = doc.get(path);
    var annotations = doc.getIndex('annotations').get(path);
    return this._annotatedText(text, annotations)
  }

  getNodeConverter(node) {
    return this.converters.get(node.type)
  }

  getDefaultBlockConverter() {
    throw new Error('This method is abstract.')
  }

  getDefaultPropertyAnnotationConverter() {
    throw new Error('This method is abstract.')
  }

  getDocument() {
    return this.state.doc
  }

  createElement(str) {
    return this._elementFactory.createElement(str)
  }

  _annotatedText(text, annotations) {
    var self = this;

    var annotator = new Fragmenter();
    annotator.onText = function(context, text) {
      context.children.push(encodeXMLEntities(text));
    };
    annotator.onEnter = function(fragment) {
      var anno = fragment.node;
      return {
        annotation: anno,
        children: []
      }
    };
    annotator.onExit = function(fragment, context, parentContext) {
      var anno = context.annotation;
      var converter = self.getNodeConverter(anno);
      if (!converter) {
        converter = self.getDefaultPropertyAnnotationConverter();
      }
      var el;
      if (converter.tagName) {
        el = this.$$(converter.tagName);
      } else {
        el = this.$$('span');
      }
      el.attr(this.config.idAttribute, anno.id);
      el.append(context.children);
      if (converter.export) {
        el = converter.export(anno, el, self) || el;
      }
      parentContext.children.push(el);
    }.bind(this);
    var wrapper = { children: [] };
    annotator.start(wrapper, text, annotations);
    return wrapper.children
  }

  /*
    This is used when someone calls `exporter.convertNode(anno)`
    Usually, annotations are converted by calling exporter.annotatedText(path).
    Still it makes sense to be able to export just a fragment containing just
    the annotation element.
  */
  _convertPropertyAnnotation(anno) {
    // take only the annotations within the range of the anno
    var wrapper = this.$$('div').append(this.annotatedText(anno.path));
    var el = wrapper.find('['+this.config.idAttribute+'="'+anno.id+'"]');
    return el
  }

}

const WS_LEFT = /^\s+/g;
const WS_LEFT_ALL = /^\s*/g;
const WS_RIGHT = /\s+$/g;
const WS_ALL = /\s+/g;
// var ALL_WS_NOTSPACE_LEFT = /^[\t\n]+/g
// var ALL_WS_NOTSPACE_RIGHT = /[\t\n]+$/g
const SPACE = " ";
const TABS_OR_NL = /[\t\n\r]+/g;

const INVISIBLE_CHARACTER = "\u200B";

/**
  A generic base implementation for XML/HTML importers.

  @param {Object} config
  @param {DocumentSchema} config.schema
  @param {object[]} config.converters
 */
class DOMImporter {

  constructor(config, context) {
    this.context = context || {};

    if (!config.schema) {
      throw new Error('"config.schema" is mandatory')
    }
    if (!config.converters) {
      throw new Error('"config.converters" is mandatory')
    }

    this.config = Object.assign({ idAttribute: 'id' }, config);
    this.schema = config.schema;
    this.converters = config.converters;
    this.state = null;

    this._defaultBlockConverter = null;
    this._allConverters = [];
    this._blockConverters = [];
    this._propertyAnnotationConverters = [];

    this.state = new DOMImporter.State();

    this._initialize();
  }

  /*
    Goes through all converters, checks their consistency
    and registers them depending on the type in different sets.
  */
  _initialize() {
    const schema = this.schema;
    const defaultTextType = schema.getDefaultTextType();
    const converters = this.converters;
    for (let i = 0; i < converters.length; i++) {
      let converter;
      if (typeof converters[i] === 'function') {
        const Converter = converters[i];
        converter = new Converter();
      } else {
        converter = converters[i];
      }
      if (!converter.type) {
        throw new Error('Converter must provide the type of the associated node.')
      }
      if (!converter.matchElement && !converter.tagName) {
        throw new Error('Converter must provide a matchElement function or a tagName property.')
      }
      if (!converter.matchElement) {
        converter.matchElement = this._defaultElementMatcher.bind(converter);
      }
      const NodeClass = schema.getNodeClass(converter.type);
      if (!NodeClass) {
        throw new Error('No node type defined for converter')
      }
      if (!this._defaultBlockConverter && defaultTextType === converter.type) {
        this._defaultBlockConverter = converter;
      }
      this._allConverters.push(converter);
      // Defaults to _blockConverters
      if (NodeClass.prototype._isPropertyAnnotation) {
        this._propertyAnnotationConverters.push(converter);
      } else {
        this._blockConverters.push(converter);
      }
    }
    if (!this._defaultBlockConverter) {
      throw new Error(`No converter for defaultTextType ${defaultTextType}`)
    }
  }

  dispose() {
    if (this.state.doc) {
      this.state.doc.dispose();
    }
  }

  /**
    Resets this importer.

    Make sure to either create a new importer instance or call this method
    when you want to generate nodes belonging to different documents.
  */
  reset() {
    if (this.state.doc) {
      this.state.doc.dispose();
    }
    this.state.reset();
    this.state.doc = this._createDocument();
  }

  getDocument() {
    return this.state.doc
  }

  /**
    Converts all children of a given element and creates a Container node.

    @param {DOMElement[]} elements All elements that should be converted into the container.
    @param {String} containerId The id of the target container node.
    @returns {Container} the container node
   */
  convertContainer(elements, containerId) {
    if (!this.state.doc) this.reset();
    const state = this.state;
    const iterator = new ArrayIterator(elements);
    const nodeIds = [];
    while(iterator.hasNext()) {
      const el = iterator.next();
      let node;
      const blockTypeConverter = this._getConverterForElement(el, 'block');
      if (blockTypeConverter) {
        state.pushContext(el.tagName, blockTypeConverter);
        let nodeData = this._createNodeData(el, blockTypeConverter.type);
        nodeData = blockTypeConverter.import(el, nodeData, this) || nodeData;
        node = this._createNode(nodeData);
        let context = state.popContext();
        context.annos.forEach((a) => {
          this._createNode(a);
        });
      } else if (el.isCommentNode()) {
        continue
      } else {
        // skip empty text nodes
        if (el.isTextNode() && /^\s*$/.exec(el.textContent)) continue
        // If we find text nodes on the block level we wrap
        // it into a paragraph element (or what is configured as default block level element)
        iterator.back();
        node = this._wrapInlineElementsIntoBlockElement(iterator);
      }
      if (node) {
        nodeIds.push(node.id);
      }
    }
    return this._createNode({
      type: 'container',
      id: containerId,
      nodes: nodeIds
    })
  }

  /**
    Converts a single HTML element and creates a node in the current document.

    @param {ui/DOMElement} el the HTML element
    @returns {object} the created node as JSON
   */
  convertElement(el) {
    if (!this.state.doc) this.reset();
    let isTopLevel = !this.state.isConverting;
    if (isTopLevel) {
      this.state.isConverting = true;
    }

    let nodeData, annos;
    const converter = this._getConverterForElement(el);
    if (converter) {
      const NodeClass = this.schema.getNodeClass(converter.type);
      nodeData = this._createNodeData(el, converter.type);
      this.state.pushContext(el.tagName, converter);
      // Note: special treatment for property annotations and inline nodes
      // i.e. if someone calls `importer.convertElement(annoEl)`
      // usually, annotations are imported in the course of `importer.annotatedText(..)`
      // The peculiarity here is that in such a case, it is not
      // not clear, which property the annotation is attached to
      if (NodeClass.isInline) {
        nodeData = this._convertInlineNode(el, nodeData, converter);
      }
      else if (NodeClass.prototype._isPropertyAnnotation) {
        nodeData = this._convertPropertyAnnotation(el, nodeData);
      } else {
        nodeData = converter.import(el, nodeData, this) || nodeData;
      }
      let context = this.state.popContext();
      annos = context.annos;
    } else {
      throw new Error('No converter found for '+el.tagName)
    }
    // create the node
    const node = this._createNode(nodeData);
    // and all annos which have been created during this call
    annos.forEach((a) => {
      this._createNode(a);
    });

    // HACK: to allow using an importer stand-alone
    // i.e. creating detached elements
    if (this.config["stand-alone"] && isTopLevel) {
      this.state.isConverting = false;
      this.reset();
    }
    return node
  }

  /**
    Convert annotated text. You should call this method only for elements
    containing rich-text.

    @param {DOMElement} el
    @param {String[]} path The target property where the extracted text (plus annotations) should be stored.
    @param {Object} options
    @param {Boolean} options.preserveWhitespace when true will preserve whitespace. Default: false.
    @returns {String} The converted text as plain-text

    @example

    ```
    p.content = converter.annotatedText(pEl, [p.id, 'content'])
    ```
   */
  annotatedText(el, path, options={}) {
    if (!path) {
      throw new Error('path is mandatory')
    }
    const state = this.state;
    const context = last$1(state.contexts);
    // NOTE: this API is meant for node converters, which have been triggered
    // via convertElement().
    if (!context) {
      throw new Error('This should be called from within an element converter.')
    }
    // TODO: are there more options?
    const oldPreserveWhitespace = state.preserveWhitespace;
    if (options.preserveWhitespace) {
      state.preserveWhitespace = true;
    }
    state.stack.push({ path: path, offset: 0, text: "", annos: []});
    // IMO we should reset the last char, as it is only relevant within one
    // annotated text property. This feature is mainly used to eat up
    // whitespace in XML/HTML at tag boundaries, produced by pretty-printed XML/HTML.
    this.state.lastChar = '';
    const iterator = el.getChildNodeIterator();
    const text = this._annotatedText(iterator);
    // now we can create all annotations which have been created during this
    // call of annotatedText
    const top = state.stack.pop();
    context.annos = context.annos.concat(top.annos);

    // reset state
    state.preserveWhitespace = oldPreserveWhitespace;

    return text
  }

  /**
    Converts the given element as plain-text.

    @param {ui/DOMElement} el
    @returns {String} The plain text
   */
  plainText(el) {
    var state = this.state;
    var text = el.textContent;
    if (state.stack.length > 0) {
      var context = last$1(state.stack);
      context.offset += text.length;
      context.text += context.text.concat(text);
    }
    return text
  }

  /*
    Tells the converter to insert custom text.

    During conversion of annotatedText this is used to insert different
    text than taken from the DOM. E.g., for inline nodes we insert an invisible
    character instead of the inner content.

    @private
    @param {String}
   */
  _customText(text) {
    var state = this.state;
    if (state.stack.length > 0) {
      var context = last$1(state.stack);
      context.offset += text.length;
      context.text += context.text.concat(text);
    }
    return text
  }

  /**
    Generates an id. The generated id is unique with respect to all ids generated so far.

    @param {String} a prefix
    @return {String} the generated id
   */
  nextId(prefix) {
    // TODO: we could create more beautiful ids?
    // however we would need to be careful as there might be another
    // element in the HTML coming with that id
    // For now we use shas
    return this.state.uuid(prefix)
  }

  _getNextId(dom, type) {
    let id = this.nextId(type);
    while (this.state.ids[id] || dom.find('#'+id)) {
      id = this.nextId(type);
    }
    return id
  }

  _getIdForElement(el, type) {
    let id = el.getAttribute(this.config.idAttribute);
    if (id && !this.state.ids[id]) return id
    return this._getNextId(el.getOwnerDocument(), type)
  }


  // Note: this is e.g. shared by ClipboardImporter which has a different
  // implementation of this.createDocument()
  _createDocument() {
    // create an empty document and initialize the container if not present
    const schema = this.config.schema;
    const DocumentClass = schema.getDocumentClass();
    return new DocumentClass(schema)
  }

  _convertPropertyAnnotation(el, nodeData) {
    const path = [nodeData.id, '_content'];
    // if there is no context, this is called stand-alone
    // i.e., user tries to convert an annotation element
    // directly, not part of a block element, such as a paragraph
    nodeData._content = this.annotatedText(el, path);
    nodeData.start = { path, offset: 0 };
    nodeData.end = { offset: nodeData._content.length };
    return nodeData
  }

  _convertInlineNode(el, nodeData, converter) {
    const path = [nodeData.id, '_content'];
    if (converter.import) {
      nodeData = converter.import(el, nodeData, this) || nodeData;
    }
    nodeData._content = '$';
    nodeData.start = { path, offset: 0 };
    nodeData.end = { offset: 1 };
    return nodeData
  }

  _createNodeData(el, type) {
    if (!type) {
      throw new Error('type is mandatory.')
    }
    let nodeData = {
      type,
      id: this._getIdForElement(el, type)
    };
    this.state.ids[nodeData.id] = true;
    return nodeData
  }

  _createNode(nodeData) {
    let doc = this.state.doc;
    // NOTE: if your Document implementation adds default nodes in the constructor
    // and you have exported the node, we need to remove the default version first
    // TODO: alternatively we could just update the existing one. For now we remove the old one.
    let node = doc.get(nodeData.id);
    if (node) {
      // console.warn('Node with same it already exists.', node)
      doc.delete(node.id);
    }
    return doc.create(nodeData)
  }

  _defaultElementMatcher(el) {
    return el.is(this.tagName)
  }

  /*
    Internal function for parsing annotated text
  */
  _annotatedText(iterator) {
    const state = this.state;
    const context = last$1(state.stack);
    /* istanbul ignore next */
    if (!context) {
      throw new Error('Illegal state: context is null.')
    }
    while(iterator.hasNext()) {
      var el = iterator.next();
      var text = "";
      /* istanbul ignore else */
      // Plain text nodes...
      if (el.isTextNode()) {
        text = this._prepareText(el.textContent);
        if (text.length) {
          // Note: text is not merged into the reentrant state
          // so that we are able to return for this reentrant call
          context.text = context.text.concat(text);
          context.offset += text.length;
        }
      } else if (el.isCommentNode()) {
        // skip comment nodes
        continue
      } else if (el.isElementNode()) {
        const annoConverter = this._getConverterForElement(el, 'inline');
        // if no inline converter is found we just traverse deeper
        if (!annoConverter) {
          /* istanbul ignore next */
          if (!this.IGNORE_DEFAULT_WARNINGS) {
            console.warn('Unsupported inline element. We will not create an annotation for it, but process its children to extract annotated text.', el.outerHTML);
          }
          // this descends into children elements without introducing a new stack frame
          // and without creating an element.
          const iterator = el.getChildNodeIterator();
          this._annotatedText(iterator);
          continue
        }
        // reentrant: we delegate the conversion to the inline node class
        // it will either call us back (this.annotatedText) or give us a finished
        // node instantly (self-managed)
        var startOffset = context.offset;
        const annoType = annoConverter.type;
        const AnnoClass = this.schema.getNodeClass(annoType);
        let annoData = this._createNodeData(el, annoType);
        // push a new context so we can deal with reentrant calls
        let stackFrame = {
          path: context.path,
          offset: startOffset,
          text: "",
          annos: []
        };
        state.stack.push(stackFrame);
        // with custom import
        if (annoConverter.import) {
          state.pushContext(el.tagName, annoConverter);
          annoData = annoConverter.import(el, annoData, this) || annoData;
          state.popContext();
        }
        // As opposed to earlier implementations we do not rely on
        // let the content be converted by custom implementations
        // as they do not own the content
        // TODO: we should make sure to throw when the user tries to
        if (AnnoClass.isInline) {
          this._customText(INVISIBLE_CHARACTER);
        } else {
          // We call this to descent into the element
          // which could be 'forgotten' otherwise.
          // TODO: what if the converter has processed the element already?
          const iterator = el.getChildNodeIterator();
          this._annotatedText(iterator);
        }
        // ... and transfer the result into the current context
        state.stack.pop();
        context.offset = stackFrame.offset;
        context.text = context.text.concat(stackFrame.text);
        // in the mean time the offset will probably have changed to reentrant calls
        const endOffset = context.offset;
        annoData.start = {
          path: context.path.slice(0),
          offset: startOffset
        };
        annoData.end = {
          offset: endOffset
        };
        // merge annos into parent stack frame
        let parentFrame = last$1(state.stack);
        parentFrame.annos = parentFrame.annos.concat(stackFrame.annos, annoData);
      } else {
        console.warn('Unknown element type. Taking plain text.', el.outerHTML);
        text = this._prepareText(el.textContent);
        context.text = context.text.concat(text);
        context.offset += text.length;
      }
    }
    // return the plain text collected during this reentrant call
    return context.text
  }

  _getConverterForElement(el, mode) {
    var converters;
    if (mode === "block") {
      if (!el.tagName) return null
      converters = this._blockConverters;
    } else if (mode === "inline") {
      converters = this._propertyAnnotationConverters;
    } else {
      converters = this._allConverters;
    }
    var converter = null;
    for (var i = 0; i < converters.length; i++) {
      if (this._converterCanBeApplied(converters[i], el)) {
        converter = converters[i];
        break
      }
    }
    return converter
  }

  _converterCanBeApplied(converter, el) {
    return converter.matchElement(el, this)
  }

  /*
    Wraps the remaining (inline) elements of a node iterator into a default
    block node.

    @param {DOMImporter.ChildIterator} childIterator
    @returns {object} node data
   */
  _wrapInlineElementsIntoBlockElement(childIterator) {
    if (!childIterator.hasNext()) return
    let dom = childIterator.peek().getOwnerDocument();
    let wrapper = dom.createElement('wrapper');
    while(childIterator.hasNext()) {
      const el = childIterator.next();
      // if there is a block node we finish this wrapper
      const blockTypeConverter = this._getConverterForElement(el, 'block');
      if (blockTypeConverter) {
        childIterator.back();
        break
      }
      wrapper.append(el.clone());
    }
    const type = this.schema.getDefaultTextType();
    const id = this._getNextId(dom, type);
    const converter = this._defaultBlockConverter;
    let nodeData = { type, id };
    this.state.pushContext('wrapper', converter);
    nodeData = converter.import(wrapper, nodeData, this) || nodeData;
    let context = this.state.popContext();
    let annos = context.annos;
    // create the node
    const node = this._createNode(nodeData);
    // and all annos which have been created during this call
    annos.forEach((a) => {
      this._createNode(a);
    });
    return node
  }

  // TODO: this needs to be tested and documented
  _prepareText(text) {
    const state = this.state;
    if (state.preserveWhitespace) {
      return text
    }
    var repl = SPACE;
    // replace multiple tabs and new-lines by one space
    text = text.replace(TABS_OR_NL, '');
    // TODO: the last char handling is only necessary for for nested calls
    // i.e., when processing the content of an annotation, for instance
    // we need to work out how we could control this with an inner state
    if (state.lastChar === SPACE) {
      text = text.replace(WS_LEFT_ALL, repl);
    } else {
      text = text.replace(WS_LEFT, repl);
    }
    text = text.replace(WS_RIGHT, repl);
    // EXPERIMENTAL: also remove white-space within
    // this happens if somebody treats the text more like it would be done in Markdown
    // i.e. introducing line-breaks
    if (this.config.REMOVE_INNER_WS || state.removeInnerWhitespace) {
      text = text.replace(WS_ALL, SPACE);
    }
    state.lastChar = text[text.length-1] || state.lastChar;
    return text
  }

  /*
    Removes any leading and trailing whitespaces from the content
    within the given element.
    Attention: this is not yet implemented fully. Atm, trimming is only done
    on the first and last text node (if they exist).
   */
  _trimTextContent(el) {
    var nodes = el.getChildNodes();
    var firstNode = nodes[0];
    var lastNode = last$1(nodes);
    var text, trimmed;
      // trim the first and last text
    if (firstNode && firstNode.isTextNode()) {
      text = firstNode.textContent;
      trimmed = this._trimLeft(text);
      firstNode.textContent = trimmed;
    }
    if (lastNode && lastNode.isTextNode()) {
      text = lastNode.textContent;
      trimmed = this._trimRight(text);
      lastNode.textContent = trimmed;
    }
    return el
  }

  _trimLeft(text) {
    return text.replace(WS_LEFT, "")
  }

  _trimRight(text) {
    return text.replace(WS_RIGHT, "")
  }

}

class DOMImporterState {

  constructor() {
    this.reset();
  }

  reset() {
    this.preserveWhitespace = false;
    this.nodes = [];
    this.annotations = [];
    this.containerId = null;
    this.container = [];
    this.ids = {};
    // stack for reentrant calls into convertElement()
    this.contexts = [];
    // stack for reentrant calls into annotatedText()
    this.stack = [];
    this.lastChar = "";
    this.skipTypes = {};
    this.ignoreAnnotations = false;
    this.isConverting = false;

    // experimental: trying to generate simpler ids during import
    // this.uuid = uuid
    this.uuid = createCountingIdGenerator();
  }

  pushContext(tagName, converter) {
    this.contexts.push({ tagName: tagName, converter: converter, annos: []});
  }

  popContext() {
    return this.contexts.pop()
  }

  getCurrentContext() {
    return last$1(this.contexts)
  }

}

DOMImporter.State = DOMImporterState;

DOMImporter.INVISIBLE_CHARACTER = INVISIBLE_CHARACTER;

class EditingBehavior {

  constructor() {
    this._merge = {};
    this._mergeComponents = {};
    this._break = {};
  }

  defineMerge(firstType, secondType, impl) {
    if (!this._merge[firstType]) {
      this._merge[firstType] = {};
    }
    this._merge[firstType][secondType] = impl;
    return this
  }

  canMerge(firstType, secondType) {
    return (this._merge[firstType] && this._merge[firstType][secondType])
  }

  getMerger(firstType, secondType) {
    return this._merge[firstType][secondType]
  }

  defineComponentMerge(nodeType, impl) {
    this._mergeComponents[nodeType] = impl;
  }

  canMergeComponents(nodeType) {
    return this._mergeComponents[nodeType]
  }

  getComponentMerger(nodeType) {
    return this._mergeComponents[nodeType]
  }

  defineBreak(nodeType, impl) {
    this._break[nodeType] = impl;
    return this
  }

  canBreak(nodeType) {
    return this._break[nodeType]
  }

  getBreaker(nodeType) {
    return this._break[nodeType]
  }

}

class FileNode extends DocumentNode {

  constructor(...args) {
    super(...args);
  }

  getUrl() {
    if (this.proxy) {
      return this.proxy.getUrl()
    } else {
      // this happens if no FileProxy is attached
      console.warn('No file proxy attached to ', this.id);
      return ''
    }
  }

  setProxy(proxy) {
    this.proxy = proxy;
  }
}

FileNode.type = 'file';

FileNode.schema = {
  url: { type: 'string', optional: true },
  fileType: { type: 'string', optional: true },
  mimeType: { type: 'string', optional: true },
  sourceFile: { type: 'object', optional: true }
};

FileNode.prototype._isFileNode = true;
FileNode._isFileNode = true;

/*
  Internal implementation used to store event bindings.
*/
class DOMEventListener {

  constructor(eventName, handler, options) {
    /* istanbul ignore next */
    if (!isString$1(eventName) || !isFunction$1(handler)) {
      throw new Error("Illegal arguments: 'eventName' must be a String, and 'handler' must be a Function.")
    }
    options = options || {};
    var origHandler = handler;
    var context = options.context;
    var capture = Boolean(options.capture);

    if (context) {
      handler = handler.bind(context);
    }
    if (options.once === true) {
      handler = _once(this, handler);
    }

    this.eventName = eventName;
    this.originalHandler = origHandler;
    this.handler = handler;
    this.capture = capture;
    this.context = context;
    this.options = options;
    // set when this gets attached to a DOM element
    this._el = null;
  }

}

DOMEventListener.prototype._isDOMEventListener = true;

DOMEventListener.findIndex = function(eventListeners, eventName, handler) {
  var idx = -1;
  if (arguments[1]._isDOMEventListener) {
    idx = eventListeners.indexOf(arguments[1]);
  } else {
    idx = findIndex$1(eventListeners,
      _matches.bind(null, {
        eventName: eventName,
        originalHandler: handler
      })
    );
  }
  return idx
};

function _matches(l1, l2) {
  return l1.eventName === l2.eventName && l1.originalHandler === l2.originalHandler
}

function _once(listener, handler) {
  return function(event) {
    handler(event);
    listener._el.removeEventListener(listener);
  }
}

const NOT_IMPLEMENTED = 'This method is not implemented.';

/**
  A unified interface for DOM elements used by Substance.

  @abstract
*/
class DOMElement {

  /*
    The element's id.
    @property {String} DOMElement#id
  */

  /*
    The element's tag name in lower case.
    @property {String} DOMElement#tagName
  */

  /*
    @property {String} DOMElement#textContent
   */

  /*
    The inner HTML string.

    @property {String} DOMElement#innerHTML
   */

  /*
    The outer HTML string.

    @property {String} DOMElement#outerHTML
   */

  /*
    An array of child nodes, including nodes such as TextNodes.

    @property {Array<DOMElement>} DOMElement#childNodes
   */

  /*
    An array of child elements.

    @property {Array<DOMElement>} DOMElement#children children
   */

  /*
    The computed height.

    @property {Array<DOMElement>} DOMElement#height
   */

  /*
    The computed width.

    @property {Array<DOMElement>} DOMElement#width
   */

  /**
    @returns the native element
  */
  getNativeElement() {
    /* istanbul ignore next */
    throw new Error(NOT_IMPLEMENTED)
  }

  /**
    Checks if the element is a TextNode.

    @returns {Boolean} true if the element is of type `Node.TEXT_NODE`
   */
  isTextNode() {
    /* istanbul ignore next */
    return false
  }

  /**
    Checks if the element is actually an element as opposed to a node.

    @returns {Boolean} true if the element is of type `Node.ELEMENT_NODE`
   */
  isElementNode() {
    /* istanbul ignore next */
    return false
  }

  /**
    Checks if the element is a CommentNode.

    @returns {Boolean} true if the element is of type `Node.COMMENT_NODE`
   */
  isCommentNode() {
    /* istanbul ignore next */
    return false
  }

  /**
    Checks if the element is a DocumentNode.

    @returns {Boolean} true if the element is of type `Node.DOCUMENT_NODE`
   */
  isDocumentNode() {
    /* istanbul ignore next */
    return false
  }

  /**
    Get the tagName of this element.

    @private
    @note Considered as private API, in favor of the property {DOMElement.prototype.tagName}
    @returns {String} the tag name in lower-case.
   */
  getTagName() {
    /* istanbul ignore next */
    throw new Error(NOT_IMPLEMENTED)
  }

  /**
    Set the tagName of this element.

    @private
    @note Considered as private API, in favor of the property {DOMElement.prototype.tagName}
    @param {String} tagName the new tag name
    @returns {this}
  */
  setTagName(tagName) { // eslint-disable-line no-unused-vars
    /* istanbul ignore next */
    throw new Error(NOT_IMPLEMENTED)
  }

  /**
    Get the id of this element.

    @private
    @note Considered as private API, in favor of the property {DOMElement.prototype.id}
    @returns {String} the id.
   */
  getId() {
    return this.getAttribute('id')
  }

  /**
    Set the id of this element.

    @private
    @note Considered as private API, in favor of the property {DOMElement.prototype.id}
    @param {String} id the new id
    @returns {this}
  */
  setId(id) {
    this.setAttribute('id', id);
  }

  /**
    Checks if a CSS class is set.

    @param {String} className
    @returns {Boolean} true if the CSS class is set
  */
  hasClass(className) { // eslint-disable-line no-unused-vars
    /* istanbul ignore next */
    throw new Error(NOT_IMPLEMENTED)
  }

  /**
    Adds a CSS class.

    @param {String} classString A space-separated string with CSS classes
    @returns {this}
  */
  addClass(classString) { // eslint-disable-line no-unused-vars
    /* istanbul ignore next */
    throw new Error(NOT_IMPLEMENTED)
  }

  /**
    Removes a CSS class.

    @param {String} classString A space-separated string with CSS classes
    @returns {this}
  */
  removeClass(classString) { // eslint-disable-line no-unused-vars
    /* istanbul ignore next */
    throw new Error(NOT_IMPLEMENTED)
  }

  hasAttribute(name) {
    return Boolean(this.getAttribute(name))
  }

  /**
    jQuery style getter and setter for attributes.

    @param {String} name
    @param {String} [value] if present the attribute will be set
    @returns {String|this} if used as getter the attribute value, otherwise this element for chaining
   */
  attr() {
    if (arguments.length === 1) {
      if (isString$1(arguments[0])) {
        return this.getAttribute(arguments[0])
      } else if (isObject$1(arguments[0])) {
        forEach(arguments[0], function(value, name) {
          this.setAttribute(name, value);
        }.bind(this));
      }
    } else if (arguments.length === 2) {
      this.setAttribute(arguments[0], arguments[1]);
    }
    return this
  }

  /**
    Removes an attribute.

    @param {String} name
    @returns {this}
  */
  removeAttr(name) {
    var names = name.split(/\s+/);
    if (names.length === 1) {
      this.removeAttribute(name);
    } else {
      names.forEach(function(name) {
        this.removeAttribute(name);
      }.bind(this));
    }
    return this
  }

  /**
    Get the attribute with a given name.

    @returns {String} the attribute's value.
  */
  getAttribute(name) { // eslint-disable-line no-unused-vars
    /* istanbul ignore next */
    throw new Error(NOT_IMPLEMENTED)
  }

  /**
    Set the attribute with a given name.

    @param {String} the attribute's value.
    @returns {this}
  */
  setAttribute(name, value) { // eslint-disable-line no-unused-vars
    /* istanbul ignore next */
    throw new Error(NOT_IMPLEMENTED)
  }

  removeAttribute(name) { // eslint-disable-line no-unused-vars
    /* istanbul ignore next */
    throw new Error(NOT_IMPLEMENTED)
  }

  getAttributes() {
    /* istanbul ignore next */
    throw new Error(NOT_IMPLEMENTED)
  }

  /**
    jQuery style getter and setter for HTML element properties.

    @param {String} name
    @param {String} [value] if present the property will be set
    @returns {String|this} if used as getter the property value, otherwise this element for chaining
   */
  htmlProp() {
    if (arguments.length === 1) {
      if (isString$1(arguments[0])) {
        return this.getProperty(arguments[0])
      } else if (isObject$1(arguments[0])) {
        forEach(arguments[0], function(value, name) {
          this.setProperty(name, value);
        }.bind(this));
      }
    } else if (arguments.length === 2) {
      this.setProperty(arguments[0], arguments[1]);
    }
    return this
  }

  getProperty(name) { // eslint-disable-line no-unused-vars
    /* istanbul ignore next */
    throw new Error(NOT_IMPLEMENTED)
  }

  setProperty(name, value) { // eslint-disable-line no-unused-vars
    /* istanbul ignore next */
    throw new Error(NOT_IMPLEMENTED)
  }

  /**
    jQuery style getter and setter for the *value* of an element.

    @param {String} [value] The value to set.
    @returns {String|this} the value if used as a getter, `this` otherwise
  */
  val(value) {
    if (arguments.length === 0) {
      return this.getValue()
    } else {
      this.setValue(value);
      return this
    }
  }

  getValue() {
    return this.getProperty('value')
  }

  setValue(value) {
    this.setProperty('value', value);
    return this
  }

  /**
    jQuery style method to set or get inline CSS styles.

    @param {String} name the style name
    @param {String} [value] the style value
    @returns {String|this} the style value or this if used as a setter
  */
  css() {
    /* istanbul ignore else */
    if (arguments.length === 1) {
      /* istanbul ignore else */
      if (isString$1(arguments[0])) {
        return this.getStyle(arguments[0])
      } else if (isObject$1(arguments[0])) {
        forEach(arguments[0], function(value, name) {
          this.setStyle(name, value);
        }.bind(this));
      } else {
        throw new Error('Illegal arguments.')
      }
    } else if (arguments.length === 2) {
      this.setStyle(arguments[0], arguments[1]);
    } else {
      throw new Error('Illegal arguments.')
    }
    return this
  }

  getStyle(name) { // eslint-disable-line no-unused-vars
    /* istanbul ignore next */
    throw new Error(NOT_IMPLEMENTED)
  }

  setStyle(name, value) { // eslint-disable-line no-unused-vars
    /* istanbul ignore next */
    throw new Error(NOT_IMPLEMENTED)
  }

  /**
    Gets or sets the text content of an element.

    @param {String} [text] The text content to set.
    @returns {String|this} The text content if used as a getter, `this` otherwise
  */
  text(text) {
    if (arguments.length === 0) {
      return this.getTextContent()
    } else {
      this.setTextContent(text);
    }
    return this
  }

  /**
    Get the textContent of this element.

    @private
    @note Considered as private API, in favor of the property {DOMElement.prototype.innerHTML}
    @returns {String}
  */
  getTextContent() {
    /* istanbul ignore next */
    throw new Error(NOT_IMPLEMENTED)
  }

  /**
    Set the textContent of this element.

    @private
    @note Considered as private API, in favor of the property {DOMElement.prototype.innerHTML}
    @param {String} text the new text content
    @returns {this}
  */
  setTextContent(text) { // eslint-disable-line no-unused-vars
    /* istanbul ignore next */
    throw new Error(NOT_IMPLEMENTED)
  }

  /**
    jQuery style getter and setter for the innerHTML of an element.

    @param {String} [html] The html to set.
    @returns {String|this} the inner html if used as a getter, `this` otherwise
   */
  html(html) {
    if (arguments.length === 0) {
      return this.getInnerHTML()
    } else {
      this.setInnerHTML(html);
    }
    return this
  }

  /**
    Get the innerHTML of this element.

    @private
    @note Considered as private API, in favor of the property {@link DOMElement.prototype.innerHTML}
    @returns {String}
  */
  getInnerHTML() {
    /* istanbul ignore next */
    throw new Error(NOT_IMPLEMENTED)
  }

  /**
    Set the innerHTML of this element.

    @private
    @note Considered as private API, in favor of the property {@link DOMElement.prototype.innerHTML}
    @param {String} text the new text content
    @returns {this}
  */
  setInnerHTML(html) { // eslint-disable-line no-unused-vars
    /* istanbul ignore next */
    throw new Error(NOT_IMPLEMENTED)
  }

  /**
    Get the outerHTML of this element.

    @private
    @note Considered as private API, in favor of the property {@link DOMElement.prototype.outerHTML}
    @returns {String}
  */
  getOuterHTML() {
    /* istanbul ignore next */
    throw new Error(NOT_IMPLEMENTED)
  }

  /**
    Registers an Element event handler.

    @param {String} event The event name.
    @param {Function} handler The handler function.
    @param {Object} [context] context where the function should be bound to
    @param {Object} [options]
    @param {Object} [options.capture] to register the event in the event's capture phase (bubbling top-down)
    @returns {this}
  */
  on(eventName, handler, context, options) {
    /* istanbul ignore next */
    if (!isString$1(eventName)) {
      throw new Error('Illegal argument: "event" must be a String.')
    }
    options = options || {};
    if (context) {
      options.context = context;
    }
    /* istanbul ignore next */
    if (!handler || !isFunction$1(handler)) {
      throw new Error('Illegal argument: invalid handler function for event ' + eventName)
    }
    this.addEventListener(eventName, handler, options);
    return this
  }

  /**
    Unregisters the handler of a given event.

    @param {String} event The event name.
    @returns {this}
  */
  off(eventName, handler) {
    // el.off(this): disconnect all listeners bound to the given context
    if (arguments.length === 1 && !isString$1(eventName)) {
      let context = arguments[0];
      this.getEventListeners().filter(function(l) {
        return l.context === context
      }).forEach(function(l) {
        this.removeEventListener(l);
      }.bind(this));
    } else {
      this.removeEventListener(eventName, handler);
    }
    return this
  }

  addEventListener(eventName, handler, options = {}) {
    let listener;
    if (arguments.length === 1 && arguments[0]) {
      listener = arguments[0];
    } else {
      listener = this._createEventListener(eventName, handler, options);
    }
    if (!this.eventListeners) {
      this.eventListeners = [];
    }
    listener._el = this;
    this.eventListeners.push(listener);
    this._addEventListenerNative(listener);
    return this
  }

  _createEventListener(eventName, handler, options) {
    return new DOMEventListener(eventName, handler, options)
  }

  _addEventListenerNative(listener) {} // eslint-disable-line

  removeEventListener(eventName, handler) {
    if (!this.eventListeners) return
    // console.log('removing event listener', eventName, handler);
    let listener = null, idx = -1;
    idx = DOMEventListener.findIndex(this.eventListeners, eventName, handler);
    listener = this.eventListeners[idx];
    if (idx > -1) {
      this.eventListeners.splice(idx, 1);
      // console.log('BrowserDOMElement.removeEventListener:', eventName, this.eventListeners.length);
      listener._el = null;
      this._removeEventListenerNative(listener);
    }
    return this
  }

  _removeEventListenerNative(listener) {} // eslint-disable-line

  removeAllEventListeners() {
    if (!this.eventListeners) return
    for (let i = 0; i < this.eventListeners.length; i++) {
      let listener = this.eventListeners[i];
      // console.log('BrowserDOMElement.removeEventListener:', eventName, this.eventListeners.length);
      listener._el = null;
      this._removeEventListenerNative(listener);
    }
    delete this.eventListeners;
  }

  getEventListeners() {
    return this.eventListeners || []
  }

  /**
    Gets the type of this element in lower-case.

    @private
    @note Considered as private API, in favor of the property {@link DOMElement.prototype.nodeType}
    @returns {String}
  */
  getNodeType() {
    /* istanbul ignore next */
    throw new Error(NOT_IMPLEMENTED)
  }

  getContentType() {
    /* istanbul ignore next */
    throw new Error(NOT_IMPLEMENTED)
  }

  getChildCount() {
    /* istanbul ignore next */
    throw new Error(NOT_IMPLEMENTED)
  }

  /**
    Get child nodes of this element.

    This method provides a new array with wrapped native elements.
    Better use getChildAt().

    @private Considered as private API, in favor of the property {DOMElement.prototype.childNodes}
    @returns {Array<DOMElement>}
   */
  getChildNodes() {
    /* istanbul ignore next */
    throw new Error(NOT_IMPLEMENTED)
  }

  /**
    Get child elements of this element.

    This method provides a new array with wrapped native elements.
    Better use getChildAt().

    @private Considered as private API, in favor of the property {DOMElement.prototype.children}
    @returns {Array<DOMElement>}
   */
  getChildren() {
    /* istanbul ignore next */
    throw new Error(NOT_IMPLEMENTED)
  }

  getChildAt(pos) { // eslint-disable-line no-unused-vars
    /* istanbul ignore next */
    throw new Error(NOT_IMPLEMENTED)
  }

  getChildIndex(child) { // eslint-disable-line no-unused-vars
    /* istanbul ignore next */
    throw new Error(NOT_IMPLEMENTED)
  }

  getChildNodeIterator() {
    return new ArrayIterator(this.getChildNodes())
  }

  getLastChild() {
    /* istanbul ignore next */
    throw new Error(NOT_IMPLEMENTED)
  }

  getFirstChild() {
    /* istanbul ignore next */
    throw new Error(NOT_IMPLEMENTED)
  }

  getNextSibling() {
    /* istanbul ignore next */
    throw new Error(NOT_IMPLEMENTED)
  }

  getPreviousSibling() {
    /* istanbul ignore next */
    throw new Error(NOT_IMPLEMENTED)
  }

  /**
    Creates a clone of the current element.

    @returns {DOMElement} A clone of this element.
  */
  clone() {
    /* istanbul ignore next */
    throw new Error(NOT_IMPLEMENTED)
  }

  /**
    Creates a DOMElement.

    @param {String} str a tag name or an HTML element as string.
    @returns {DOMElement}
  */
  createElement(str) { // eslint-disable-line no-unused-vars
    /* istanbul ignore next */
    throw new Error(NOT_IMPLEMENTED)
  }

  createTextNode(text) { // eslint-disable-line no-unused-vars
    /* istanbul ignore next */
    throw new Error(NOT_IMPLEMENTED)
  }

  createComment(data) { // eslint-disable-line no-unused-vars
    /* istanbul ignore next */
    throw new Error(NOT_IMPLEMENTED)
  }

  createProcessingInstruction(name, data) { // eslint-disable-line no-unused-vars
    /* istanbul ignore next */
    throw new Error(NOT_IMPLEMENTED)
  }

  createCDATASection(data) { // eslint-disable-line no-unused-vars
    /* istanbul ignore next */
    throw new Error(NOT_IMPLEMENTED)
  }

  /**
    Checks if a given CSS selector matches for this element.

    **Attention**
    This method is currently not implemented for {VirtualElement}.
    This means you should use it only in importer implementations.

    @param {String} cssSelector
    @returns {Boolean}
   */
  is(cssSelector) { // eslint-disable-line no-unused-vars
    /* istanbul ignore next */
    throw new Error(NOT_IMPLEMENTED)
  }

  /**
    Get the parent element of this element.

    @returns {DOMElement} the parent element
   */
  getParent() {
    /* istanbul ignore next */
    throw new Error(NOT_IMPLEMENTED)
  }

  /**
    Get the ownerDocument of this element.

    @returns {DOMElement} the document element
  */
  getOwnerDocument() {
    /* istanbul ignore next */
    throw new Error(NOT_IMPLEMENTED)
  }

  /**
    Find the first descendant element matching the given CSS selector.
    Note this differs from jQuery.find() that it returns only one element.

    **Attention**
    This method is currently not implemented for {VirtualElement}.
    This means you can use it only in importer implementations, but not in render or exporter implementations.

    @param {String} cssSelector
    @returns {DOMElement} found element
   */
  find(cssSelector) { // eslint-disable-line no-unused-vars
    /* istanbul ignore next */
    throw new Error(NOT_IMPLEMENTED)
  }

  /**
    Find all descendant elements matching the given CSS selector.

    **Attention**
    This method is currently not implemented for {VirtualElement}.
    This means you can use it only in importer implementations, but not in render or exporter implementations.

    @param {String} cssSelector
    @returns {Array<DOMElement>} found elements
   */
  findAll(cssSelector) { // eslint-disable-line no-unused-vars
    /* istanbul ignore next */
    throw new Error(NOT_IMPLEMENTED)
  }

  /**
    Append a child element.

    @param {DOMElement|String} child An element or text to append
    @returns {this}
   */
  append(child) {
    var children;
    if (arguments.length === 1) {
      if (isArray$1(child)) {
        children = child;
      } else {
        this.appendChild(child);
        return this
      }
    } else {
      children = arguments;
    }
    if (children) {
      Array.prototype.forEach.call(children, this.appendChild.bind(this));
    }
    return this
  }

  appendChild(child) { // eslint-disable-line no-unused-vars
    /* istanbul ignore next */
    throw new Error(NOT_IMPLEMENTED)
  }

  /**
    Insert a child element at a given position.

    @param {Number} pos insert position
    @param {DOMElement|String} child The child element or text to insert.
    @returns {this}
  */
  insertAt(pos, child) { // eslint-disable-line no-unused-vars
    /* istanbul ignore next */
    throw new Error(NOT_IMPLEMENTED)
  }

  insertBefore(newChild, before) { // eslint-disable-line no-unused-vars
    /* istanbul ignore next */
    throw new Error(NOT_IMPLEMENTED)
  }

  /**
    Remove the child at a given position.

    @param {Number} pos
    @returns {this}
  */
  removeAt(pos) { // eslint-disable-line no-unused-vars
    /* istanbul ignore next */
    throw new Error(NOT_IMPLEMENTED)
  }

  removeChild(child) { // eslint-disable-line no-unused-vars
    /* istanbul ignore next */
    throw new Error(NOT_IMPLEMENTED)
  }

  replaceChild(oldChild, newChild) { // eslint-disable-line no-unused-vars
    /* istanbul ignore next */
    throw new Error(NOT_IMPLEMENTED)
  }

  /**
    Removes this element from its parent.
    @returns {this}
  */
  remove() {
    var parent = this.getParent();
    if (parent) {
      parent.removeChild(this);
    }
  }

  /**
    Removes all child nodes from this element.

    @returns {this}
  */
  empty() {
    /* istanbul ignore next */
    throw new Error(NOT_IMPLEMENTED)
  }

  serialize() {
    return this.getOuterHTML()
  }

  isInDocument() {
    let el = this;
    while(el) {
      if (el.isDocumentNode()) {
        return true
      }
      el = el.getParent();
    }
  }

  /**
    Focusses this element.

    **Attention: this makes only sense for elements which are rendered in the browser**

  */
  focus() {
    /* istanbul ignore next */
    return this
  }

  /**
    Blur this element.
  */
  blur() {
    /* istanbul ignore next */
    return this
  }

  /**
    Trigger a click event on this element.
  */
  click() {
    /* istanbul ignore next */
    return this
  }

  /* API to retrieve layout information */

  getWidth() {
    /* istanbul ignore next */
    return 0
  }

  getHeight() {
    /* istanbul ignore next */
    return 0
  }

  /**
    Outer height as provided by $.outerHeight(withMargin)
  */
  getOuterHeight(withMargin) { // eslint-disable-line no-unused-vars
    /* istanbul ignore next */
    return 0
  }

  /**
    Offset values as provided by $.offset()
  */
  getOffset() {
    /* istanbul ignore next */
    return { top: 0, left: 0 }
  }

  /**
    Position values as provided by $.position()
  */
  getPosition() {
    /* istanbul ignore next */
    return { top: 0, left: 0 }
  }

  /**
    Get element factory conveniently

    @example

    var $$ = el.getElementFactory()
    $$('div').append('bla')
  */
  getElementFactory() {
    return this.createElement.bind(this)
  }

  /**
    Triggers a custom event.

    @param {String} name
    @param {Object} data
  */
  emit(name, data) { // eslint-disable-line
    /* istanbul ignore next */
    throw new Error(NOT_IMPLEMENTED)
  }

  // properties

  get id() {
    return this.getId()
  }

  set id(id) {
    this.setId(id);
  }

  get tagName() {
    return this.getTagName()
  }

  set tagName(tagName) {
    this.setTagName(tagName);
  }

  get nodeName() {
    return this.getTagName()
  }

  get nodeType() {
    return this.getNodeType()
  }

  get className() {
    return this.getAttribute('class')
  }

  set className(className) {
    this.setAttribute('class', className);
  }

  get textContent() {
    return this.getTextContent()
  }

  set textContent(text) {
    this.setTextContent(text);
  }

  get innerHTML() {
    return this.getInnerHTML()
  }

  set innerHTML(html) {
    this.setInnerHTML(html);
  }

  get outerHTML() {
    return this.getOuterHTML()
  }

  get firstChild() {
    return this.getFirstChild()
  }

  get lastChild() {
    return this.getLastChild()
  }

  get nextSibling() {
    return this.getNextSibling()
  }

  get previousSibling() {
    return this.getPreviousSibling()
  }

  get parentNode() {
    return this.getParent()
  }

  get height() {
    return this.getHeight()
  }

  get width() {
    return this.getWidth()
  }

  get value() {
    return this.getValue()
  }

  set value(value) {
    return this.setValue(value)
  }
}

DOMElement.prototype._isDOMElement = true;

DOMElement.pxStyles = {
  top: true,
  bottom: true,
  left: true,
  right: true,
  height: true,
  width: true
};

DOMElement.EMPTY_HTML = '<html><head></head><body></body></html>';

// using a dynamic signature to store the wrapper on the native element
// this way, we avoid to inadvertently use a wrong API created by another
// DOMElement instance on the element
const SIGNATURE = uuid('_BrowserDOMElement');

function _attach(nativeEl, browserDOMElement) {
  nativeEl[SIGNATURE] = browserDOMElement;
}

function _detach(nativeEl) {
  delete nativeEl[SIGNATURE];
}

function _unwrap(nativeEl) {
  return nativeEl[SIGNATURE]
}

class BrowserDOMElement extends DOMElement {

  constructor(el) {
    super();
    console.assert(el instanceof window.Node, "Expecting native DOM node.");
    this.el = el;
    // creating a backlink so we can move between the native DOM API and
    // the DOMElement API
    _attach(el, this);
  }

  getNativeElement() {
    return this.el
  }

  getNodeType() {
    switch(this.el.nodeType) {
      case window.Node.TEXT_NODE:
        return "text"
      case window.Node.ELEMENT_NODE:
        return 'element'
      case window.Node.DOCUMENT_NODE:
        return 'document'
      case window.Node.COMMENT_NODE:
        return 'comment'
      case window.Node.PROCESSING_INSTRUCTION_NODE:
        return 'directive'
      case window.Node.CDATA_SECTION_NODE:
        return 'cdata'
      default:
        //
    }
  }

  isTextNode() {
    return (this.el.nodeType === window.Node.TEXT_NODE)
  }

  isElementNode() {
    return (this.el.nodeType === window.Node.ELEMENT_NODE)
  }

  isCommentNode() {
    return (this.el.nodeType === window.Node.COMMENT_NODE)
  }

  isDocumentNode() {
    return (this.el.nodeType === window.Node.DOCUMENT_NODE)
  }

  hasClass(className) {
    return this.el.classList.contains(className)
  }

  addClass(className) {
    this.el.classList.add(className);
    return this
  }

  removeClass(className) {
    this.el.classList.remove(className);
    return this
  }

  getAttribute(name) {
    return this.el.getAttribute(name)
  }

  setAttribute(name, value) {
    this.el.setAttribute(name, String(value));
    return this
  }

  removeAttribute(name) {
    this.el.removeAttribute(name);
    return this
  }

  getAttributes() {
    if (!this.el.attributes._mapAdapter) {
      this.el.attributes._mapAdapter = new AttributesMapAdapter(this.el.attributes);
    }
    return this.el.attributes._mapAdapter
  }

  getProperty(name) {
    return this.el[name]
  }

  setProperty(name, value) {
    // ATTENTION: element properties are only used on HTML elements, such as the 'value' of an <input> element
    // In XML there are only attributes
    if (this._isXML()) throw new Error('setProperty() is only supported for HTML elements.')
    if (!this._changedProperties) this._changedProperties = new Set();
    // NOTE: there is no removeProperty()
    // thus we are clearing our property when value=undefined
    if (value === undefined) {
      this._changedProperties.delete(name);
    } else {
      this._changedProperties.add(name);
    }
    this.el[name] = value;
    return this
  }

  getTagName() {
    // it is convenient in HTML mode to always use tagName in lower case
    // however, in XML this is not allowed, as tag names are case sensitive there
    if (this._isXML()) {
      return this.el.tagName
    } else if (this.el.tagName) {
      return this.el.tagName.toLowerCase()
    }
  }

  setTagName(tagName) {
    let newEl = this.createElement(tagName);
    let attributes = this.el.attributes;
    let l = attributes.length;
    let i;
    for(i = 0; i < l; i++) {
      let attr = attributes.item(i);
      newEl.setAttribute(attr.name, attr.value);
    }
    // NOTE: it does not make sense to set properties as they have a dynamic nature
    // which depends strongly on the type they are defined on
    // if (!this._changedProperties) this._changedProperties = new Set()
    // this._changedProperties.forEach((name)=>{
    //   newEl[name] = this.el[name]
    // })
    if (this.eventListeners) {
      this.eventListeners.forEach(function(listener) {
        newEl.addEventListener(listener.eventName, listener.handler, listener.capture);
      });
    }
    newEl.append(this.getChildNodes());

    this._replaceNativeEl(newEl.getNativeElement());
    return this
  }

  getId() {
    return this.el.id
  }

  setId(id) {
    this.el.id = id;
    return this
  }

  getStyle(name) {
    // NOTE: important to provide computed style, otherwise we don't get inherited styles
    let style = this.getComputedStyle();
    return style[name] || this.el.style[name]
  }

  getComputedStyle() {
    return window.getComputedStyle(this.el)
  }

  setStyle(name, value) {
    if (DOMElement.pxStyles[name] && isNumber(value)) value = value + 'px';
    this.el.style[name] = value;
    return this
  }

  getTextContent() {
    return this.el.textContent
  }

  setTextContent(text) {
    this.el.textContent = text;
    return this
  }

  getInnerHTML() {
    if (this._isXML()) {
      let xs = new window.XMLSerializer();
      let result = Array.prototype.map.call(this.el.childNodes, c => xs.serializeToString(c));
      return result.join('')
    } else {
      return this.el.innerHTML
    }
  }

  setInnerHTML(html) {
    // TODO: if in some cases we need to use XMLSerializer to get the innerHTML
    // then we probably need to use DOMParser here accordingly
    this.el.innerHTML = html;
    return this
  }

  getOuterHTML() {
    // NOTE: this was necessary in some browsers, which did not provide
    // el.outerHTML for XML elements
    if (this._isXML()) {
      let xs = new window.XMLSerializer();
      return xs.serializeToString(this.el)
    } else {
      return this.el.outerHTML
    }
  }

  _addEventListenerNative(listener) {
    this.el.addEventListener(listener.eventName, listener.handler, listener.capture);
  }

  _removeEventListenerNative(listener) {
    this.el.removeEventListener(listener.eventName, listener.handler);
  }

  getEventListeners() {
    return this.eventListeners || []
  }

  getChildCount() {
    return this.el.childNodes.length
  }

  getChildNodes() {
    let childNodes = [];
    for (let node = this.el.firstChild; node; node = node.nextSibling) {
      childNodes.push(BrowserDOMElement.wrap(node));
    }
    return childNodes
  }

  get childNodes() {
    return this.getChildNodes()
  }

  getChildren() {
    // Some browsers don't filter elements here and also include text nodes,
    // that why we can't use el.children
    let children = [];
    for (let node = this.el.firstChild; node; node = node.nextSibling) {
      if (node.nodeType === window.Node.ELEMENT_NODE) {
        children.push(BrowserDOMElement.wrap(node));
      }
    }
    return children
  }

  get children() {
    return this.getChildren()
  }

  getChildAt(pos) {
    return BrowserDOMElement.wrap(this.el.childNodes[pos])
  }

  getChildIndex(child) {
    /* istanbul ignore next */
    if (!child._isBrowserDOMElement) {
      throw new Error('Expecting a BrowserDOMElement instance.')
    }
    return Array.prototype.indexOf.call(this.el.childNodes, child.el)
  }

  getFirstChild() {
    let firstChild = this.el.firstChild;
    /* istanbul ignore else */
    if (firstChild) {
      return BrowserDOMElement.wrap(firstChild)
    } else {
      return null
    }
  }

  getLastChild() {
    var lastChild = this.el.lastChild;
    /* istanbul ignore else */
    if (lastChild) {
      return BrowserDOMElement.wrap(lastChild)
    } else {
      return null
    }
  }

  getNextSibling() {
    let next = this.el.nextSibling;
    /* istanbul ignore else */
    if (next) {
      return BrowserDOMElement.wrap(next)
    } else {
      return null
    }
  }

  getPreviousSibling() {
    let previous = this.el.previousSibling;
    /* istanbul ignore else */
    if (previous) {
      return BrowserDOMElement.wrap(previous)
    } else {
      return null
    }
  }

  clone() {
    let clone$$1 = this.el.cloneNode(true);
    return BrowserDOMElement.wrap(clone$$1)
  }

  createDocument(format) {
    return BrowserDOMElement.createDocument(format)
  }

  createElement(tagName) {
    let doc = this._getNativeOwnerDocument();
    let el = doc.createElement(tagName);
    return BrowserDOMElement.wrap(el)
  }

  createTextNode(text) {
    let doc = this._getNativeOwnerDocument();
    let el = doc.createTextNode(text);
    return BrowserDOMElement.wrap(el)
  }

  createComment(data) {
    let doc = this._getNativeOwnerDocument();
    let el = doc.createComment(data);
    return BrowserDOMElement.wrap(el)
  }

  createProcessingInstruction(name, data) {
    let doc = this._getNativeOwnerDocument();
    let el = doc.createProcessingInstruction(name, data);
    return BrowserDOMElement.wrap(el)
  }

  createCDATASection(data) {
    let doc = this._getNativeOwnerDocument();
    let el = doc.createCDATASection(data);
    return BrowserDOMElement.wrap(el)
  }

  is(cssSelector) {
    // ATTENTION: looking at https://developer.mozilla.org/en/docs/Web/API/Element/matches
    // Element.matches might not be supported by some mobile browsers
    let el = this.el;
    /* istanbul ignore else */
    if (this.isElementNode()) {
      return matches(el, cssSelector)
    } else {
      return false
    }
  }

  getParent() {
    let parent = this.el.parentNode;
    /* istanbul ignore else */
    if (parent) {
      return BrowserDOMElement.wrap(parent)
    } else {
      return null
    }
  }

  getOwnerDocument() {
    return BrowserDOMElement.wrap(this._getNativeOwnerDocument())
  }

  get ownerDocument() {
    return this.getOwnerDocument()
  }

  _getNativeOwnerDocument() {
    return (this.isDocumentNode() ? this.el : this.el.ownerDocument)
  }

  find(cssSelector) {
    let result = null;
    if (this.el.querySelector) {
      result = this.el.querySelector(cssSelector);
    }
    if (result) {
      return BrowserDOMElement.wrap(result)
    } else {
      return null
    }
  }

  findAll(cssSelector) {
    let result = [];
    if (this.el.querySelectorAll) {
      result = this.el.querySelectorAll(cssSelector);
    }
    return Array.prototype.map.call(result, function(el) {
      return BrowserDOMElement.wrap(el)
    })
  }

  _normalizeChild(child) {
    if (child instanceof window.Node) {
      child = BrowserDOMElement.wrap(child);
    }
    // Note: element is owned by a different implementation.
    // Probably you are using two different versions of Substance on the same element.
    // Can't tell if this is bad. For now we continue by wrapping it again
    else if (child._isBrowserDOMElement && ! (child instanceof BrowserDOMElement)) {
      child = BrowserDOMElement.wrap(child);
    } else if (isString$1(child) || isNumber(child)) {
      child = this.createTextNode(child);
    }
    /* istanbul ignore next */
    if (!child || !child._isBrowserDOMElement) {
      throw new Error('Illegal child type.')
    }
    console.assert(_unwrap(child.el) === child, "The backlink to the wrapper should be consistent");
    return child.getNativeElement()
  }

  appendChild(child) {
    let nativeChild = this._normalizeChild(child);
    this.el.appendChild(nativeChild);
    return this
  }

  insertAt(pos, child) {
    let nativeChild = this._normalizeChild(child);
    let childNodes = this.el.childNodes;
    if (pos >= childNodes.length) {
      this.el.appendChild(nativeChild);
    } else {
      this.el.insertBefore(nativeChild, childNodes[pos]);
    }
    return this
  }

  insertBefore(child, before) {
    /* istanbul ignore next */
    if (!before || !before._isBrowserDOMElement) {
      throw new Error('insertBefore(): Illegal arguments. "before" must be a BrowserDOMElement instance.')
    }
    var nativeChild = this._normalizeChild(child);
    this.el.insertBefore(nativeChild, before.el);
    return this
  }

  removeAt(pos) {
    this.el.removeChild(this.el.childNodes[pos]);
    return this;
  }

  removeChild(child) {
    /* istanbul ignore next */
    if (!child || !child._isBrowserDOMElement) {
      throw new Error('removeChild(): Illegal arguments. Expecting a BrowserDOMElement instance.')
    }
    this.el.removeChild(child.el);
    return this
  }

  replaceChild(oldChild, newChild) {
    /* istanbul ignore next */
    if (!newChild || !oldChild ||
        !newChild._isBrowserDOMElement || !oldChild._isBrowserDOMElement) {
      throw new Error('replaceChild(): Illegal arguments. Expecting BrowserDOMElement instances.')
    }
    // Attention: Node.replaceChild has weird semantics
    this.el.replaceChild(newChild.el, oldChild.el);
    return this
  }

  empty() {
    let el = this.el;
    while (el.lastChild) {
      el.removeChild(el.lastChild);
    }
    return this
  }

  remove() {
    if (this.el.parentNode) {
      this.el.parentNode.removeChild(this.el);
    }
    return this
  }

  serialize() {
    let outerHTML = this.el.outerHTML;
    if (isString$1(outerHTML)) {
      return outerHTML
    } else {
      let xs = new window.XMLSerializer();
      return xs.serializeToString(this.el)
    }
  }

  isInDocument() {
    let el = this.el;
    while(el) {
      if (el.nodeType === window.Node.DOCUMENT_NODE) {
        return true
      }
      el = el.parentNode;
    }
  }

  _replaceNativeEl(newEl) {
    console.assert(newEl instanceof window.Node, "Expecting a native element.");
    let oldEl = this.el;
    let parentNode = oldEl.parentNode;
    if (parentNode) {
      parentNode.replaceChild(newEl, oldEl);
    }
    this.el = newEl;
    _detach(oldEl);
    _attach(newEl, this);
  }

  _getChildNodeCount() {
    return this.el.childNodes.length
  }

  focus() {
    this.el.focus();
    return this
  }

  blur() {
    this.el.focus();
    return this
  }

  click() {
    this.el.click();
    return this
  }

  getWidth() {
    let rect = this.el.getClientRects()[0];
    if (rect) {
      return rect.width
    } else {
      return 0
    }
  }

  getHeight() {
    let rect = this.el.getClientRects()[0];
    if (rect) {
      return rect.height
    } else {
      return 0
    }
  }

  getOffset() {
    let rect = this.el.getBoundingClientRect();
    return {
      top: rect.top + document.body.scrollTop,
      left: rect.left + document.body.scrollLeft
    }
  }

  getPosition() {
    return {left: this.el.offsetLeft, top: this.el.offsetTop}
  }

  getOuterHeight(withMargin) {
    let outerHeight = this.el.offsetHeight;
    if (withMargin) {
      let style = this.getComputedStyle();
      outerHeight += parseInt(style.marginTop, 10) + parseInt(style.marginBottom, 10);
    }
    return outerHeight
  }

  getContentType() {
    return this._getNativeOwnerDocument().contentType
  }

  _isXML() {
    return this.getContentType() === 'application/xml'
  }

  emit(name, data) {
    let event;
    if (data) {
      event = new window.CustomEvent(name, { detail: data });
    } else {
      event = new window.Event(name);
    }
    this.el.dispatchEvent(event);
  }
}

BrowserDOMElement.prototype._isBrowserDOMElement = true;

// TODO: flesh out how options should look like (e.g. XML namespaceURI etc.)
BrowserDOMElement.createDocument = function(format) {
  let doc;
  if (format === 'xml') {
    // HACK: didn't find a way to create an empty XML doc without a root element
    doc = window.document.implementation.createDocument(null, 'dummy');
    // remove the
    doc.removeChild(doc.firstChild);
  } else {
    doc = (new window.DOMParser()).parseFromString(DOMElement.EMPTY_HTML, 'text/html');
  }
  return BrowserDOMElement.wrap(doc)
};

BrowserDOMElement.parseMarkup = function(str, format, options={}) {
  if (!str) {
    return BrowserDOMElement.createDocument(format)
  }
  if (options.snippet) {
    str = `<div id='__snippet__'>${str}</div>`;
  }
  let doc;
  let parser = new window.DOMParser();
  if (format === 'html') {
    doc = BrowserDOMElement.wrap(
      _check(
        parser.parseFromString(str, 'text/html')
      )
    );
  } else if (format === 'xml') {
    doc = BrowserDOMElement.wrap(
      _check(
        parser.parseFromString(str, 'text/xml')
      )
    );
  }
  if (options.snippet) {
    let childNodes = doc.find('#__snippet__').childNodes;
    if (childNodes.length === 1) {
      return childNodes[0]
    } else {
      return childNodes
    }
  } else {
    return doc
  }

  function _check(doc) {
    if (doc) {
      let parserError = doc.querySelector('parsererror');
      if (parserError) {
        throw new Error("ParserError: " + parserError)
      }
    }
    return doc
  }
};

BrowserDOMElement.wrap =
BrowserDOMElement.wrapNativeElement = function(el) {
  if (el) {
    let _el = _unwrap(el);
    if (_el) {
      return _el
    } else if (el instanceof window.Node) {
      return new BrowserDOMElement(el)
    } else if (el._isBrowserDOMElement) {
      return new BrowserDOMElement(el.getNativeElement())
    } else if (el === window) {
      return BrowserDOMElement.getBrowserWindow()
    }
  } else {
    return null
  }
};

BrowserDOMElement.unwrap = function(nativeEl) {
  return _unwrap(nativeEl)
};

/*
  Wrapper for the window element exposing DOMElement's EventListener API.
*/
class BrowserWindow {

  constructor() {
    // Note: not
    this.el = window;
    window.__BrowserDOMElementWrapper__ = this;
  }

}

BrowserWindow.prototype.on = BrowserDOMElement.prototype.on;

BrowserWindow.prototype.off = BrowserDOMElement.prototype.off;

BrowserWindow.prototype.addEventListener = BrowserDOMElement.prototype.addEventListener;

BrowserWindow.prototype.removeEventListener = BrowserDOMElement.prototype.removeEventListener;

BrowserWindow.prototype.getEventListeners = BrowserDOMElement.prototype.getEventListeners;

BrowserDOMElement.getBrowserWindow = function() {
  if (window[SIGNATURE]) return window[SIGNATURE]
  return new BrowserWindow(window)
};

BrowserDOMElement.isReverse = function(anchorNode, anchorOffset, focusNode, focusOffset) {
  // the selection is reversed when the focus propertyEl is before
  // the anchor el or the computed charPos is in reverse order
  if (focusNode && anchorNode) {
    if (!BrowserDOMElement.isReverse._r1) {
      BrowserDOMElement.isReverse._r1 = window.document.createRange();
      BrowserDOMElement.isReverse._r2 = window.document.createRange();
    }
    const _r1 = BrowserDOMElement.isReverse._r1;
    const _r2 = BrowserDOMElement.isReverse._r2;
    _r1.setStart(anchorNode.getNativeElement(), anchorOffset);
    _r2.setStart(focusNode.getNativeElement(), focusOffset);
    let cmp = _r1.compareBoundaryPoints(window.Range.START_TO_START, _r2);
    if (cmp === 1) {
      return true
    }
  }
  return false
};

BrowserDOMElement.getWindowSelection = function() {
  let nativeSel = window.getSelection();
  let result = {
    anchorNode: BrowserDOMElement.wrap(nativeSel.anchorNode),
    anchorOffset: nativeSel.anchorOffset,
    focusNode: BrowserDOMElement.wrap(nativeSel.focusNode),
    focusOffset: nativeSel.focusOffset
  };
  return result
};


function matches(el, selector) {
  let elProto = window.Element.prototype;
  let _matches = (
    elProto.matches || elProto.matchesSelector ||
    elProto.msMatchesSelector || elProto.webkitMatchesSelector
  );
  return _matches.call(el, selector)
}

class AttributesMapAdapter {

  constructor(attributes) {
    this.attributes = attributes;
  }

  get size() {
    return this.attributes.length
  }

  get(name) {
    let item = this.attributes.getNamedItem(name);
    if (item) {
      return item.value
    }
  }

  entries() {
    let entries = [];
    let S = this.size;
    for (let i = 0; i < S; i++) {
      const item = this.attributes.item(i);
      entries.push([item.name, item.value]);
    }
    return entries
  }

  set() {
    throw new Error('This is a read-only map.')
  }
}

//Types of elements found in the DOM
var index = {
	Text: "text", //Text
	Directive: "directive", //<? ... ?>
	Comment: "comment", //<!-- ... -->
	Script: "script", //<script> tags
	Style: "style", //<style> tags
	Tag: "tag", //Any tag
	CDATA: "cdata", //<![CDATA[ ... ]]>
	Doctype: "doctype",

	isTag: function(elem){
		return elem.type === "tag" || elem.type === "script" || elem.type === "style";
	}
};

var amp = "&";
var apos = "'";
var gt = ">";
var lt = "<";
var quot = "\"";
var xmlJSON = {
	amp: amp,
	apos: apos,
	gt: gt,
	lt: lt,
	quot: quot
};

var xml = Object.freeze({
	amp: amp,
	apos: apos,
	gt: gt,
	lt: lt,
	quot: quot,
	default: xmlJSON
});

var Aacute = "Á";
var aacute = "á";
var Abreve = "Ă";
var abreve = "ă";
var ac = "∾";
var acd = "∿";
var acE = "∾̳";
var Acirc = "Â";
var acirc = "â";
var acute = "´";
var Acy = "А";
var acy = "а";
var AElig = "Æ";
var aelig = "æ";
var af = "⁡";
var Afr = "𝔄";
var afr = "𝔞";
var Agrave = "À";
var agrave = "à";
var alefsym = "ℵ";
var aleph = "ℵ";
var Alpha = "Α";
var alpha = "α";
var Amacr = "Ā";
var amacr = "ā";
var amalg = "⨿";
var amp$1 = "&";
var AMP = "&";
var andand = "⩕";
var And = "⩓";
var and = "∧";
var andd = "⩜";
var andslope = "⩘";
var andv = "⩚";
var ang = "∠";
var ange = "⦤";
var angle = "∠";
var angmsdaa = "⦨";
var angmsdab = "⦩";
var angmsdac = "⦪";
var angmsdad = "⦫";
var angmsdae = "⦬";
var angmsdaf = "⦭";
var angmsdag = "⦮";
var angmsdah = "⦯";
var angmsd = "∡";
var angrt = "∟";
var angrtvb = "⊾";
var angrtvbd = "⦝";
var angsph = "∢";
var angst = "Å";
var angzarr = "⍼";
var Aogon = "Ą";
var aogon = "ą";
var Aopf = "𝔸";
var aopf = "𝕒";
var apacir = "⩯";
var ap = "≈";
var apE = "⩰";
var ape = "≊";
var apid = "≋";
var apos$1 = "'";
var ApplyFunction = "⁡";
var approx = "≈";
var approxeq = "≊";
var Aring = "Å";
var aring = "å";
var Ascr = "𝒜";
var ascr = "𝒶";
var Assign = "≔";
var ast = "*";
var asymp = "≈";
var asympeq = "≍";
var Atilde = "Ã";
var atilde = "ã";
var Auml = "Ä";
var auml = "ä";
var awconint = "∳";
var awint = "⨑";
var backcong = "≌";
var backepsilon = "϶";
var backprime = "‵";
var backsim = "∽";
var backsimeq = "⋍";
var Backslash = "∖";
var Barv = "⫧";
var barvee = "⊽";
var barwed = "⌅";
var Barwed = "⌆";
var barwedge = "⌅";
var bbrk = "⎵";
var bbrktbrk = "⎶";
var bcong = "≌";
var Bcy = "Б";
var bcy = "б";
var bdquo = "„";
var becaus = "∵";
var because = "∵";
var Because = "∵";
var bemptyv = "⦰";
var bepsi = "϶";
var bernou = "ℬ";
var Bernoullis = "ℬ";
var Beta = "Β";
var beta = "β";
var beth = "ℶ";
var between = "≬";
var Bfr = "𝔅";
var bfr = "𝔟";
var bigcap = "⋂";
var bigcirc = "◯";
var bigcup = "⋃";
var bigodot = "⨀";
var bigoplus = "⨁";
var bigotimes = "⨂";
var bigsqcup = "⨆";
var bigstar = "★";
var bigtriangledown = "▽";
var bigtriangleup = "△";
var biguplus = "⨄";
var bigvee = "⋁";
var bigwedge = "⋀";
var bkarow = "⤍";
var blacklozenge = "⧫";
var blacksquare = "▪";
var blacktriangle = "▴";
var blacktriangledown = "▾";
var blacktriangleleft = "◂";
var blacktriangleright = "▸";
var blank = "␣";
var blk12 = "▒";
var blk14 = "░";
var blk34 = "▓";
var block = "█";
var bne = "=⃥";
var bnequiv = "≡⃥";
var bNot = "⫭";
var bnot = "⌐";
var Bopf = "𝔹";
var bopf = "𝕓";
var bot = "⊥";
var bottom = "⊥";
var bowtie = "⋈";
var boxbox = "⧉";
var boxdl = "┐";
var boxdL = "╕";
var boxDl = "╖";
var boxDL = "╗";
var boxdr = "┌";
var boxdR = "╒";
var boxDr = "╓";
var boxDR = "╔";
var boxh = "─";
var boxH = "═";
var boxhd = "┬";
var boxHd = "╤";
var boxhD = "╥";
var boxHD = "╦";
var boxhu = "┴";
var boxHu = "╧";
var boxhU = "╨";
var boxHU = "╩";
var boxminus = "⊟";
var boxplus = "⊞";
var boxtimes = "⊠";
var boxul = "┘";
var boxuL = "╛";
var boxUl = "╜";
var boxUL = "╝";
var boxur = "└";
var boxuR = "╘";
var boxUr = "╙";
var boxUR = "╚";
var boxv = "│";
var boxV = "║";
var boxvh = "┼";
var boxvH = "╪";
var boxVh = "╫";
var boxVH = "╬";
var boxvl = "┤";
var boxvL = "╡";
var boxVl = "╢";
var boxVL = "╣";
var boxvr = "├";
var boxvR = "╞";
var boxVr = "╟";
var boxVR = "╠";
var bprime = "‵";
var breve = "˘";
var Breve = "˘";
var brvbar = "¦";
var bscr = "𝒷";
var Bscr = "ℬ";
var bsemi = "⁏";
var bsim = "∽";
var bsime = "⋍";
var bsolb = "⧅";
var bsol = "\\";
var bsolhsub = "⟈";
var bull = "•";
var bullet = "•";
var bump = "≎";
var bumpE = "⪮";
var bumpe = "≏";
var Bumpeq = "≎";
var bumpeq = "≏";
var Cacute = "Ć";
var cacute = "ć";
var capand = "⩄";
var capbrcup = "⩉";
var capcap = "⩋";
var cap = "∩";
var Cap = "⋒";
var capcup = "⩇";
var capdot = "⩀";
var CapitalDifferentialD = "ⅅ";
var caps = "∩︀";
var caret = "⁁";
var caron = "ˇ";
var Cayleys = "ℭ";
var ccaps = "⩍";
var Ccaron = "Č";
var ccaron = "č";
var Ccedil = "Ç";
var ccedil = "ç";
var Ccirc = "Ĉ";
var ccirc = "ĉ";
var Cconint = "∰";
var ccups = "⩌";
var ccupssm = "⩐";
var Cdot = "Ċ";
var cdot = "ċ";
var cedil = "¸";
var Cedilla = "¸";
var cemptyv = "⦲";
var cent = "¢";
var centerdot = "·";
var CenterDot = "·";
var cfr = "𝔠";
var Cfr = "ℭ";
var CHcy = "Ч";
var chcy = "ч";
var check = "✓";
var checkmark = "✓";
var Chi = "Χ";
var chi = "χ";
var circ = "ˆ";
var circeq = "≗";
var circlearrowleft = "↺";
var circlearrowright = "↻";
var circledast = "⊛";
var circledcirc = "⊚";
var circleddash = "⊝";
var CircleDot = "⊙";
var circledR = "®";
var circledS = "Ⓢ";
var CircleMinus = "⊖";
var CirclePlus = "⊕";
var CircleTimes = "⊗";
var cir = "○";
var cirE = "⧃";
var cire = "≗";
var cirfnint = "⨐";
var cirmid = "⫯";
var cirscir = "⧂";
var ClockwiseContourIntegral = "∲";
var CloseCurlyDoubleQuote = "”";
var CloseCurlyQuote = "’";
var clubs = "♣";
var clubsuit = "♣";
var colon = ":";
var Colon = "∷";
var Colone = "⩴";
var colone = "≔";
var coloneq = "≔";
var comma = ",";
var commat = "@";
var comp = "∁";
var compfn = "∘";
var complement = "∁";
var complexes = "ℂ";
var cong = "≅";
var congdot = "⩭";
var Congruent = "≡";
var conint = "∮";
var Conint = "∯";
var ContourIntegral = "∮";
var copf = "𝕔";
var Copf = "ℂ";
var coprod = "∐";
var Coproduct = "∐";
var copy = "©";
var COPY = "©";
var copysr = "℗";
var CounterClockwiseContourIntegral = "∳";
var crarr = "↵";
var cross = "✗";
var Cross = "⨯";
var Cscr = "𝒞";
var cscr = "𝒸";
var csub = "⫏";
var csube = "⫑";
var csup = "⫐";
var csupe = "⫒";
var ctdot = "⋯";
var cudarrl = "⤸";
var cudarrr = "⤵";
var cuepr = "⋞";
var cuesc = "⋟";
var cularr = "↶";
var cularrp = "⤽";
var cupbrcap = "⩈";
var cupcap = "⩆";
var CupCap = "≍";
var cup = "∪";
var Cup = "⋓";
var cupcup = "⩊";
var cupdot = "⊍";
var cupor = "⩅";
var cups = "∪︀";
var curarr = "↷";
var curarrm = "⤼";
var curlyeqprec = "⋞";
var curlyeqsucc = "⋟";
var curlyvee = "⋎";
var curlywedge = "⋏";
var curren = "¤";
var curvearrowleft = "↶";
var curvearrowright = "↷";
var cuvee = "⋎";
var cuwed = "⋏";
var cwconint = "∲";
var cwint = "∱";
var cylcty = "⌭";
var dagger = "†";
var Dagger = "‡";
var daleth = "ℸ";
var darr = "↓";
var Darr = "↡";
var dArr = "⇓";
var dash = "‐";
var Dashv = "⫤";
var dashv = "⊣";
var dbkarow = "⤏";
var dblac = "˝";
var Dcaron = "Ď";
var dcaron = "ď";
var Dcy = "Д";
var dcy = "д";
var ddagger = "‡";
var ddarr = "⇊";
var DD = "ⅅ";
var dd = "ⅆ";
var DDotrahd = "⤑";
var ddotseq = "⩷";
var deg = "°";
var Del = "∇";
var Delta = "Δ";
var delta = "δ";
var demptyv = "⦱";
var dfisht = "⥿";
var Dfr = "𝔇";
var dfr = "𝔡";
var dHar = "⥥";
var dharl = "⇃";
var dharr = "⇂";
var DiacriticalAcute = "´";
var DiacriticalDot = "˙";
var DiacriticalDoubleAcute = "˝";
var DiacriticalGrave = "`";
var DiacriticalTilde = "˜";
var diam = "⋄";
var diamond = "⋄";
var Diamond = "⋄";
var diamondsuit = "♦";
var diams = "♦";
var die = "¨";
var DifferentialD = "ⅆ";
var digamma = "ϝ";
var disin = "⋲";
var div = "÷";
var divide = "÷";
var divideontimes = "⋇";
var divonx = "⋇";
var DJcy = "Ђ";
var djcy = "ђ";
var dlcorn = "⌞";
var dlcrop = "⌍";
var dollar = "$";
var Dopf = "𝔻";
var dopf = "𝕕";
var Dot = "¨";
var dot = "˙";
var DotDot = "⃜";
var doteq = "≐";
var doteqdot = "≑";
var DotEqual = "≐";
var dotminus = "∸";
var dotplus = "∔";
var dotsquare = "⊡";
var doublebarwedge = "⌆";
var DoubleContourIntegral = "∯";
var DoubleDot = "¨";
var DoubleDownArrow = "⇓";
var DoubleLeftArrow = "⇐";
var DoubleLeftRightArrow = "⇔";
var DoubleLeftTee = "⫤";
var DoubleLongLeftArrow = "⟸";
var DoubleLongLeftRightArrow = "⟺";
var DoubleLongRightArrow = "⟹";
var DoubleRightArrow = "⇒";
var DoubleRightTee = "⊨";
var DoubleUpArrow = "⇑";
var DoubleUpDownArrow = "⇕";
var DoubleVerticalBar = "∥";
var DownArrowBar = "⤓";
var downarrow = "↓";
var DownArrow = "↓";
var Downarrow = "⇓";
var DownArrowUpArrow = "⇵";
var DownBreve = "̑";
var downdownarrows = "⇊";
var downharpoonleft = "⇃";
var downharpoonright = "⇂";
var DownLeftRightVector = "⥐";
var DownLeftTeeVector = "⥞";
var DownLeftVectorBar = "⥖";
var DownLeftVector = "↽";
var DownRightTeeVector = "⥟";
var DownRightVectorBar = "⥗";
var DownRightVector = "⇁";
var DownTeeArrow = "↧";
var DownTee = "⊤";
var drbkarow = "⤐";
var drcorn = "⌟";
var drcrop = "⌌";
var Dscr = "𝒟";
var dscr = "𝒹";
var DScy = "Ѕ";
var dscy = "ѕ";
var dsol = "⧶";
var Dstrok = "Đ";
var dstrok = "đ";
var dtdot = "⋱";
var dtri = "▿";
var dtrif = "▾";
var duarr = "⇵";
var duhar = "⥯";
var dwangle = "⦦";
var DZcy = "Џ";
var dzcy = "џ";
var dzigrarr = "⟿";
var Eacute = "É";
var eacute = "é";
var easter = "⩮";
var Ecaron = "Ě";
var ecaron = "ě";
var Ecirc = "Ê";
var ecirc = "ê";
var ecir = "≖";
var ecolon = "≕";
var Ecy = "Э";
var ecy = "э";
var eDDot = "⩷";
var Edot = "Ė";
var edot = "ė";
var eDot = "≑";
var ee = "ⅇ";
var efDot = "≒";
var Efr = "𝔈";
var efr = "𝔢";
var eg = "⪚";
var Egrave = "È";
var egrave = "è";
var egs = "⪖";
var egsdot = "⪘";
var el = "⪙";
var Element = "∈";
var elinters = "⏧";
var ell = "ℓ";
var els = "⪕";
var elsdot = "⪗";
var Emacr = "Ē";
var emacr = "ē";
var empty = "∅";
var emptyset = "∅";
var EmptySmallSquare = "◻";
var emptyv = "∅";
var EmptyVerySmallSquare = "▫";
var emsp13 = " ";
var emsp14 = " ";
var emsp = " ";
var ENG = "Ŋ";
var eng = "ŋ";
var ensp = " ";
var Eogon = "Ę";
var eogon = "ę";
var Eopf = "𝔼";
var eopf = "𝕖";
var epar = "⋕";
var eparsl = "⧣";
var eplus = "⩱";
var epsi = "ε";
var Epsilon = "Ε";
var epsilon = "ε";
var epsiv = "ϵ";
var eqcirc = "≖";
var eqcolon = "≕";
var eqsim = "≂";
var eqslantgtr = "⪖";
var eqslantless = "⪕";
var Equal = "⩵";
var equals = "=";
var EqualTilde = "≂";
var equest = "≟";
var Equilibrium = "⇌";
var equiv = "≡";
var equivDD = "⩸";
var eqvparsl = "⧥";
var erarr = "⥱";
var erDot = "≓";
var escr = "ℯ";
var Escr = "ℰ";
var esdot = "≐";
var Esim = "⩳";
var esim = "≂";
var Eta = "Η";
var eta = "η";
var ETH = "Ð";
var eth = "ð";
var Euml = "Ë";
var euml = "ë";
var euro = "€";
var excl = "!";
var exist = "∃";
var Exists = "∃";
var expectation = "ℰ";
var exponentiale = "ⅇ";
var ExponentialE = "ⅇ";
var fallingdotseq = "≒";
var Fcy = "Ф";
var fcy = "ф";
var female = "♀";
var ffilig = "ﬃ";
var fflig = "ﬀ";
var ffllig = "ﬄ";
var Ffr = "𝔉";
var ffr = "𝔣";
var filig = "ﬁ";
var FilledSmallSquare = "◼";
var FilledVerySmallSquare = "▪";
var fjlig = "fj";
var flat = "♭";
var fllig = "ﬂ";
var fltns = "▱";
var fnof = "ƒ";
var Fopf = "𝔽";
var fopf = "𝕗";
var forall = "∀";
var ForAll = "∀";
var fork = "⋔";
var forkv = "⫙";
var Fouriertrf = "ℱ";
var fpartint = "⨍";
var frac12 = "½";
var frac13 = "⅓";
var frac14 = "¼";
var frac15 = "⅕";
var frac16 = "⅙";
var frac18 = "⅛";
var frac23 = "⅔";
var frac25 = "⅖";
var frac34 = "¾";
var frac35 = "⅗";
var frac38 = "⅜";
var frac45 = "⅘";
var frac56 = "⅚";
var frac58 = "⅝";
var frac78 = "⅞";
var frasl = "⁄";
var frown = "⌢";
var fscr = "𝒻";
var Fscr = "ℱ";
var gacute = "ǵ";
var Gamma = "Γ";
var gamma = "γ";
var Gammad = "Ϝ";
var gammad = "ϝ";
var gap = "⪆";
var Gbreve = "Ğ";
var gbreve = "ğ";
var Gcedil = "Ģ";
var Gcirc = "Ĝ";
var gcirc = "ĝ";
var Gcy = "Г";
var gcy = "г";
var Gdot = "Ġ";
var gdot = "ġ";
var ge = "≥";
var gE = "≧";
var gEl = "⪌";
var gel = "⋛";
var geq = "≥";
var geqq = "≧";
var geqslant = "⩾";
var gescc = "⪩";
var ges = "⩾";
var gesdot = "⪀";
var gesdoto = "⪂";
var gesdotol = "⪄";
var gesl = "⋛︀";
var gesles = "⪔";
var Gfr = "𝔊";
var gfr = "𝔤";
var gg = "≫";
var Gg = "⋙";
var ggg = "⋙";
var gimel = "ℷ";
var GJcy = "Ѓ";
var gjcy = "ѓ";
var gla = "⪥";
var gl = "≷";
var glE = "⪒";
var glj = "⪤";
var gnap = "⪊";
var gnapprox = "⪊";
var gne = "⪈";
var gnE = "≩";
var gneq = "⪈";
var gneqq = "≩";
var gnsim = "⋧";
var Gopf = "𝔾";
var gopf = "𝕘";
var grave = "`";
var GreaterEqual = "≥";
var GreaterEqualLess = "⋛";
var GreaterFullEqual = "≧";
var GreaterGreater = "⪢";
var GreaterLess = "≷";
var GreaterSlantEqual = "⩾";
var GreaterTilde = "≳";
var Gscr = "𝒢";
var gscr = "ℊ";
var gsim = "≳";
var gsime = "⪎";
var gsiml = "⪐";
var gtcc = "⪧";
var gtcir = "⩺";
var gt$1 = ">";
var GT = ">";
var Gt = "≫";
var gtdot = "⋗";
var gtlPar = "⦕";
var gtquest = "⩼";
var gtrapprox = "⪆";
var gtrarr = "⥸";
var gtrdot = "⋗";
var gtreqless = "⋛";
var gtreqqless = "⪌";
var gtrless = "≷";
var gtrsim = "≳";
var gvertneqq = "≩︀";
var gvnE = "≩︀";
var Hacek = "ˇ";
var hairsp = " ";
var half = "½";
var hamilt = "ℋ";
var HARDcy = "Ъ";
var hardcy = "ъ";
var harrcir = "⥈";
var harr = "↔";
var hArr = "⇔";
var harrw = "↭";
var Hat = "^";
var hbar = "ℏ";
var Hcirc = "Ĥ";
var hcirc = "ĥ";
var hearts = "♥";
var heartsuit = "♥";
var hellip = "…";
var hercon = "⊹";
var hfr = "𝔥";
var Hfr = "ℌ";
var HilbertSpace = "ℋ";
var hksearow = "⤥";
var hkswarow = "⤦";
var hoarr = "⇿";
var homtht = "∻";
var hookleftarrow = "↩";
var hookrightarrow = "↪";
var hopf = "𝕙";
var Hopf = "ℍ";
var horbar = "―";
var HorizontalLine = "─";
var hscr = "𝒽";
var Hscr = "ℋ";
var hslash = "ℏ";
var Hstrok = "Ħ";
var hstrok = "ħ";
var HumpDownHump = "≎";
var HumpEqual = "≏";
var hybull = "⁃";
var hyphen = "‐";
var Iacute = "Í";
var iacute = "í";
var ic = "⁣";
var Icirc = "Î";
var icirc = "î";
var Icy = "И";
var icy = "и";
var Idot = "İ";
var IEcy = "Е";
var iecy = "е";
var iexcl = "¡";
var iff = "⇔";
var ifr = "𝔦";
var Ifr = "ℑ";
var Igrave = "Ì";
var igrave = "ì";
var ii = "ⅈ";
var iiiint = "⨌";
var iiint = "∭";
var iinfin = "⧜";
var iiota = "℩";
var IJlig = "Ĳ";
var ijlig = "ĳ";
var Imacr = "Ī";
var imacr = "ī";
var image = "ℑ";
var ImaginaryI = "ⅈ";
var imagline = "ℐ";
var imagpart = "ℑ";
var imath = "ı";
var Im = "ℑ";
var imof = "⊷";
var imped = "Ƶ";
var Implies = "⇒";
var incare = "℅";
var infin = "∞";
var infintie = "⧝";
var inodot = "ı";
var intcal = "⊺";
var int = "∫";
var Int = "∬";
var integers = "ℤ";
var Integral = "∫";
var intercal = "⊺";
var Intersection = "⋂";
var intlarhk = "⨗";
var intprod = "⨼";
var InvisibleComma = "⁣";
var InvisibleTimes = "⁢";
var IOcy = "Ё";
var iocy = "ё";
var Iogon = "Į";
var iogon = "į";
var Iopf = "𝕀";
var iopf = "𝕚";
var Iota = "Ι";
var iota = "ι";
var iprod = "⨼";
var iquest = "¿";
var iscr = "𝒾";
var Iscr = "ℐ";
var isin = "∈";
var isindot = "⋵";
var isinE = "⋹";
var isins = "⋴";
var isinsv = "⋳";
var isinv = "∈";
var it = "⁢";
var Itilde = "Ĩ";
var itilde = "ĩ";
var Iukcy = "І";
var iukcy = "і";
var Iuml = "Ï";
var iuml = "ï";
var Jcirc = "Ĵ";
var jcirc = "ĵ";
var Jcy = "Й";
var jcy = "й";
var Jfr = "𝔍";
var jfr = "𝔧";
var jmath = "ȷ";
var Jopf = "𝕁";
var jopf = "𝕛";
var Jscr = "𝒥";
var jscr = "𝒿";
var Jsercy = "Ј";
var jsercy = "ј";
var Jukcy = "Є";
var jukcy = "є";
var Kappa = "Κ";
var kappa = "κ";
var kappav = "ϰ";
var Kcedil = "Ķ";
var kcedil = "ķ";
var Kcy = "К";
var kcy = "к";
var Kfr = "𝔎";
var kfr = "𝔨";
var kgreen = "ĸ";
var KHcy = "Х";
var khcy = "х";
var KJcy = "Ќ";
var kjcy = "ќ";
var Kopf = "𝕂";
var kopf = "𝕜";
var Kscr = "𝒦";
var kscr = "𝓀";
var lAarr = "⇚";
var Lacute = "Ĺ";
var lacute = "ĺ";
var laemptyv = "⦴";
var lagran = "ℒ";
var Lambda = "Λ";
var lambda = "λ";
var lang = "⟨";
var Lang = "⟪";
var langd = "⦑";
var langle = "⟨";
var lap = "⪅";
var Laplacetrf = "ℒ";
var laquo = "«";
var larrb = "⇤";
var larrbfs = "⤟";
var larr = "←";
var Larr = "↞";
var lArr = "⇐";
var larrfs = "⤝";
var larrhk = "↩";
var larrlp = "↫";
var larrpl = "⤹";
var larrsim = "⥳";
var larrtl = "↢";
var latail = "⤙";
var lAtail = "⤛";
var lat = "⪫";
var late = "⪭";
var lates = "⪭︀";
var lbarr = "⤌";
var lBarr = "⤎";
var lbbrk = "❲";
var lbrace = "{";
var lbrack = "[";
var lbrke = "⦋";
var lbrksld = "⦏";
var lbrkslu = "⦍";
var Lcaron = "Ľ";
var lcaron = "ľ";
var Lcedil = "Ļ";
var lcedil = "ļ";
var lceil = "⌈";
var lcub = "{";
var Lcy = "Л";
var lcy = "л";
var ldca = "⤶";
var ldquo = "“";
var ldquor = "„";
var ldrdhar = "⥧";
var ldrushar = "⥋";
var ldsh = "↲";
var le = "≤";
var lE = "≦";
var LeftAngleBracket = "⟨";
var LeftArrowBar = "⇤";
var leftarrow = "←";
var LeftArrow = "←";
var Leftarrow = "⇐";
var LeftArrowRightArrow = "⇆";
var leftarrowtail = "↢";
var LeftCeiling = "⌈";
var LeftDoubleBracket = "⟦";
var LeftDownTeeVector = "⥡";
var LeftDownVectorBar = "⥙";
var LeftDownVector = "⇃";
var LeftFloor = "⌊";
var leftharpoondown = "↽";
var leftharpoonup = "↼";
var leftleftarrows = "⇇";
var leftrightarrow = "↔";
var LeftRightArrow = "↔";
var Leftrightarrow = "⇔";
var leftrightarrows = "⇆";
var leftrightharpoons = "⇋";
var leftrightsquigarrow = "↭";
var LeftRightVector = "⥎";
var LeftTeeArrow = "↤";
var LeftTee = "⊣";
var LeftTeeVector = "⥚";
var leftthreetimes = "⋋";
var LeftTriangleBar = "⧏";
var LeftTriangle = "⊲";
var LeftTriangleEqual = "⊴";
var LeftUpDownVector = "⥑";
var LeftUpTeeVector = "⥠";
var LeftUpVectorBar = "⥘";
var LeftUpVector = "↿";
var LeftVectorBar = "⥒";
var LeftVector = "↼";
var lEg = "⪋";
var leg = "⋚";
var leq = "≤";
var leqq = "≦";
var leqslant = "⩽";
var lescc = "⪨";
var les = "⩽";
var lesdot = "⩿";
var lesdoto = "⪁";
var lesdotor = "⪃";
var lesg = "⋚︀";
var lesges = "⪓";
var lessapprox = "⪅";
var lessdot = "⋖";
var lesseqgtr = "⋚";
var lesseqqgtr = "⪋";
var LessEqualGreater = "⋚";
var LessFullEqual = "≦";
var LessGreater = "≶";
var lessgtr = "≶";
var LessLess = "⪡";
var lesssim = "≲";
var LessSlantEqual = "⩽";
var LessTilde = "≲";
var lfisht = "⥼";
var lfloor = "⌊";
var Lfr = "𝔏";
var lfr = "𝔩";
var lg = "≶";
var lgE = "⪑";
var lHar = "⥢";
var lhard = "↽";
var lharu = "↼";
var lharul = "⥪";
var lhblk = "▄";
var LJcy = "Љ";
var ljcy = "љ";
var llarr = "⇇";
var ll = "≪";
var Ll = "⋘";
var llcorner = "⌞";
var Lleftarrow = "⇚";
var llhard = "⥫";
var lltri = "◺";
var Lmidot = "Ŀ";
var lmidot = "ŀ";
var lmoustache = "⎰";
var lmoust = "⎰";
var lnap = "⪉";
var lnapprox = "⪉";
var lne = "⪇";
var lnE = "≨";
var lneq = "⪇";
var lneqq = "≨";
var lnsim = "⋦";
var loang = "⟬";
var loarr = "⇽";
var lobrk = "⟦";
var longleftarrow = "⟵";
var LongLeftArrow = "⟵";
var Longleftarrow = "⟸";
var longleftrightarrow = "⟷";
var LongLeftRightArrow = "⟷";
var Longleftrightarrow = "⟺";
var longmapsto = "⟼";
var longrightarrow = "⟶";
var LongRightArrow = "⟶";
var Longrightarrow = "⟹";
var looparrowleft = "↫";
var looparrowright = "↬";
var lopar = "⦅";
var Lopf = "𝕃";
var lopf = "𝕝";
var loplus = "⨭";
var lotimes = "⨴";
var lowast = "∗";
var lowbar = "_";
var LowerLeftArrow = "↙";
var LowerRightArrow = "↘";
var loz = "◊";
var lozenge = "◊";
var lozf = "⧫";
var lpar = "(";
var lparlt = "⦓";
var lrarr = "⇆";
var lrcorner = "⌟";
var lrhar = "⇋";
var lrhard = "⥭";
var lrm = "‎";
var lrtri = "⊿";
var lsaquo = "‹";
var lscr = "𝓁";
var Lscr = "ℒ";
var lsh = "↰";
var Lsh = "↰";
var lsim = "≲";
var lsime = "⪍";
var lsimg = "⪏";
var lsqb = "[";
var lsquo = "‘";
var lsquor = "‚";
var Lstrok = "Ł";
var lstrok = "ł";
var ltcc = "⪦";
var ltcir = "⩹";
var lt$1 = "<";
var LT = "<";
var Lt = "≪";
var ltdot = "⋖";
var lthree = "⋋";
var ltimes = "⋉";
var ltlarr = "⥶";
var ltquest = "⩻";
var ltri = "◃";
var ltrie = "⊴";
var ltrif = "◂";
var ltrPar = "⦖";
var lurdshar = "⥊";
var luruhar = "⥦";
var lvertneqq = "≨︀";
var lvnE = "≨︀";
var macr = "¯";
var male = "♂";
var malt = "✠";
var maltese = "✠";
var map$1 = "↦";
var mapsto = "↦";
var mapstodown = "↧";
var mapstoleft = "↤";
var mapstoup = "↥";
var marker = "▮";
var mcomma = "⨩";
var Mcy = "М";
var mcy = "м";
var mdash = "—";
var mDDot = "∺";
var measuredangle = "∡";
var MediumSpace = " ";
var Mellintrf = "ℳ";
var Mfr = "𝔐";
var mfr = "𝔪";
var mho = "℧";
var micro = "µ";
var midast = "*";
var midcir = "⫰";
var mid = "∣";
var middot = "·";
var minusb = "⊟";
var minus = "−";
var minusd = "∸";
var minusdu = "⨪";
var MinusPlus = "∓";
var mlcp = "⫛";
var mldr = "…";
var mnplus = "∓";
var models = "⊧";
var Mopf = "𝕄";
var mopf = "𝕞";
var mp = "∓";
var mscr = "𝓂";
var Mscr = "ℳ";
var mstpos = "∾";
var Mu = "Μ";
var mu = "μ";
var multimap = "⊸";
var mumap = "⊸";
var nabla = "∇";
var Nacute = "Ń";
var nacute = "ń";
var nang = "∠⃒";
var nap = "≉";
var napE = "⩰̸";
var napid = "≋̸";
var napos = "ŉ";
var napprox = "≉";
var natural = "♮";
var naturals = "ℕ";
var natur = "♮";
var nbsp = " ";
var nbump = "≎̸";
var nbumpe = "≏̸";
var ncap = "⩃";
var Ncaron = "Ň";
var ncaron = "ň";
var Ncedil = "Ņ";
var ncedil = "ņ";
var ncong = "≇";
var ncongdot = "⩭̸";
var ncup = "⩂";
var Ncy = "Н";
var ncy = "н";
var ndash = "–";
var nearhk = "⤤";
var nearr = "↗";
var neArr = "⇗";
var nearrow = "↗";
var ne = "≠";
var nedot = "≐̸";
var NegativeMediumSpace = "​";
var NegativeThickSpace = "​";
var NegativeThinSpace = "​";
var NegativeVeryThinSpace = "​";
var nequiv = "≢";
var nesear = "⤨";
var nesim = "≂̸";
var NestedGreaterGreater = "≫";
var NestedLessLess = "≪";
var NewLine = "\n";
var nexist = "∄";
var nexists = "∄";
var Nfr = "𝔑";
var nfr = "𝔫";
var ngE = "≧̸";
var nge = "≱";
var ngeq = "≱";
var ngeqq = "≧̸";
var ngeqslant = "⩾̸";
var nges = "⩾̸";
var nGg = "⋙̸";
var ngsim = "≵";
var nGt = "≫⃒";
var ngt = "≯";
var ngtr = "≯";
var nGtv = "≫̸";
var nharr = "↮";
var nhArr = "⇎";
var nhpar = "⫲";
var ni = "∋";
var nis = "⋼";
var nisd = "⋺";
var niv = "∋";
var NJcy = "Њ";
var njcy = "њ";
var nlarr = "↚";
var nlArr = "⇍";
var nldr = "‥";
var nlE = "≦̸";
var nle = "≰";
var nleftarrow = "↚";
var nLeftarrow = "⇍";
var nleftrightarrow = "↮";
var nLeftrightarrow = "⇎";
var nleq = "≰";
var nleqq = "≦̸";
var nleqslant = "⩽̸";
var nles = "⩽̸";
var nless = "≮";
var nLl = "⋘̸";
var nlsim = "≴";
var nLt = "≪⃒";
var nlt = "≮";
var nltri = "⋪";
var nltrie = "⋬";
var nLtv = "≪̸";
var nmid = "∤";
var NoBreak = "⁠";
var NonBreakingSpace = " ";
var nopf = "𝕟";
var Nopf = "ℕ";
var Not = "⫬";
var not = "¬";
var NotCongruent = "≢";
var NotCupCap = "≭";
var NotDoubleVerticalBar = "∦";
var NotElement = "∉";
var NotEqual = "≠";
var NotEqualTilde = "≂̸";
var NotExists = "∄";
var NotGreater = "≯";
var NotGreaterEqual = "≱";
var NotGreaterFullEqual = "≧̸";
var NotGreaterGreater = "≫̸";
var NotGreaterLess = "≹";
var NotGreaterSlantEqual = "⩾̸";
var NotGreaterTilde = "≵";
var NotHumpDownHump = "≎̸";
var NotHumpEqual = "≏̸";
var notin = "∉";
var notindot = "⋵̸";
var notinE = "⋹̸";
var notinva = "∉";
var notinvb = "⋷";
var notinvc = "⋶";
var NotLeftTriangleBar = "⧏̸";
var NotLeftTriangle = "⋪";
var NotLeftTriangleEqual = "⋬";
var NotLess = "≮";
var NotLessEqual = "≰";
var NotLessGreater = "≸";
var NotLessLess = "≪̸";
var NotLessSlantEqual = "⩽̸";
var NotLessTilde = "≴";
var NotNestedGreaterGreater = "⪢̸";
var NotNestedLessLess = "⪡̸";
var notni = "∌";
var notniva = "∌";
var notnivb = "⋾";
var notnivc = "⋽";
var NotPrecedes = "⊀";
var NotPrecedesEqual = "⪯̸";
var NotPrecedesSlantEqual = "⋠";
var NotReverseElement = "∌";
var NotRightTriangleBar = "⧐̸";
var NotRightTriangle = "⋫";
var NotRightTriangleEqual = "⋭";
var NotSquareSubset = "⊏̸";
var NotSquareSubsetEqual = "⋢";
var NotSquareSuperset = "⊐̸";
var NotSquareSupersetEqual = "⋣";
var NotSubset = "⊂⃒";
var NotSubsetEqual = "⊈";
var NotSucceeds = "⊁";
var NotSucceedsEqual = "⪰̸";
var NotSucceedsSlantEqual = "⋡";
var NotSucceedsTilde = "≿̸";
var NotSuperset = "⊃⃒";
var NotSupersetEqual = "⊉";
var NotTilde = "≁";
var NotTildeEqual = "≄";
var NotTildeFullEqual = "≇";
var NotTildeTilde = "≉";
var NotVerticalBar = "∤";
var nparallel = "∦";
var npar = "∦";
var nparsl = "⫽⃥";
var npart = "∂̸";
var npolint = "⨔";
var npr = "⊀";
var nprcue = "⋠";
var nprec = "⊀";
var npreceq = "⪯̸";
var npre = "⪯̸";
var nrarrc = "⤳̸";
var nrarr = "↛";
var nrArr = "⇏";
var nrarrw = "↝̸";
var nrightarrow = "↛";
var nRightarrow = "⇏";
var nrtri = "⋫";
var nrtrie = "⋭";
var nsc = "⊁";
var nsccue = "⋡";
var nsce = "⪰̸";
var Nscr = "𝒩";
var nscr = "𝓃";
var nshortmid = "∤";
var nshortparallel = "∦";
var nsim = "≁";
var nsime = "≄";
var nsimeq = "≄";
var nsmid = "∤";
var nspar = "∦";
var nsqsube = "⋢";
var nsqsupe = "⋣";
var nsub = "⊄";
var nsubE = "⫅̸";
var nsube = "⊈";
var nsubset = "⊂⃒";
var nsubseteq = "⊈";
var nsubseteqq = "⫅̸";
var nsucc = "⊁";
var nsucceq = "⪰̸";
var nsup = "⊅";
var nsupE = "⫆̸";
var nsupe = "⊉";
var nsupset = "⊃⃒";
var nsupseteq = "⊉";
var nsupseteqq = "⫆̸";
var ntgl = "≹";
var Ntilde = "Ñ";
var ntilde = "ñ";
var ntlg = "≸";
var ntriangleleft = "⋪";
var ntrianglelefteq = "⋬";
var ntriangleright = "⋫";
var ntrianglerighteq = "⋭";
var Nu = "Ν";
var nu = "ν";
var num = "#";
var numero = "№";
var numsp = " ";
var nvap = "≍⃒";
var nvdash = "⊬";
var nvDash = "⊭";
var nVdash = "⊮";
var nVDash = "⊯";
var nvge = "≥⃒";
var nvgt = ">⃒";
var nvHarr = "⤄";
var nvinfin = "⧞";
var nvlArr = "⤂";
var nvle = "≤⃒";
var nvlt = "<⃒";
var nvltrie = "⊴⃒";
var nvrArr = "⤃";
var nvrtrie = "⊵⃒";
var nvsim = "∼⃒";
var nwarhk = "⤣";
var nwarr = "↖";
var nwArr = "⇖";
var nwarrow = "↖";
var nwnear = "⤧";
var Oacute = "Ó";
var oacute = "ó";
var oast = "⊛";
var Ocirc = "Ô";
var ocirc = "ô";
var ocir = "⊚";
var Ocy = "О";
var ocy = "о";
var odash = "⊝";
var Odblac = "Ő";
var odblac = "ő";
var odiv = "⨸";
var odot = "⊙";
var odsold = "⦼";
var OElig = "Œ";
var oelig = "œ";
var ofcir = "⦿";
var Ofr = "𝔒";
var ofr = "𝔬";
var ogon = "˛";
var Ograve = "Ò";
var ograve = "ò";
var ogt = "⧁";
var ohbar = "⦵";
var ohm = "Ω";
var oint = "∮";
var olarr = "↺";
var olcir = "⦾";
var olcross = "⦻";
var oline = "‾";
var olt = "⧀";
var Omacr = "Ō";
var omacr = "ō";
var Omega = "Ω";
var omega = "ω";
var Omicron = "Ο";
var omicron = "ο";
var omid = "⦶";
var ominus = "⊖";
var Oopf = "𝕆";
var oopf = "𝕠";
var opar = "⦷";
var OpenCurlyDoubleQuote = "“";
var OpenCurlyQuote = "‘";
var operp = "⦹";
var oplus = "⊕";
var orarr = "↻";
var Or = "⩔";
var or = "∨";
var ord = "⩝";
var order = "ℴ";
var orderof = "ℴ";
var ordf = "ª";
var ordm = "º";
var origof = "⊶";
var oror = "⩖";
var orslope = "⩗";
var orv = "⩛";
var oS = "Ⓢ";
var Oscr = "𝒪";
var oscr = "ℴ";
var Oslash = "Ø";
var oslash = "ø";
var osol = "⊘";
var Otilde = "Õ";
var otilde = "õ";
var otimesas = "⨶";
var Otimes = "⨷";
var otimes = "⊗";
var Ouml = "Ö";
var ouml = "ö";
var ovbar = "⌽";
var OverBar = "‾";
var OverBrace = "⏞";
var OverBracket = "⎴";
var OverParenthesis = "⏜";
var para = "¶";
var parallel = "∥";
var par = "∥";
var parsim = "⫳";
var parsl = "⫽";
var part = "∂";
var PartialD = "∂";
var Pcy = "П";
var pcy = "п";
var percnt = "%";
var period = ".";
var permil = "‰";
var perp = "⊥";
var pertenk = "‱";
var Pfr = "𝔓";
var pfr = "𝔭";
var Phi = "Φ";
var phi = "φ";
var phiv = "ϕ";
var phmmat = "ℳ";
var phone = "☎";
var Pi = "Π";
var pi = "π";
var pitchfork = "⋔";
var piv = "ϖ";
var planck = "ℏ";
var planckh = "ℎ";
var plankv = "ℏ";
var plusacir = "⨣";
var plusb = "⊞";
var pluscir = "⨢";
var plus = "+";
var plusdo = "∔";
var plusdu = "⨥";
var pluse = "⩲";
var PlusMinus = "±";
var plusmn = "±";
var plussim = "⨦";
var plustwo = "⨧";
var pm = "±";
var Poincareplane = "ℌ";
var pointint = "⨕";
var popf = "𝕡";
var Popf = "ℙ";
var pound = "£";
var prap = "⪷";
var Pr = "⪻";
var pr = "≺";
var prcue = "≼";
var precapprox = "⪷";
var prec = "≺";
var preccurlyeq = "≼";
var Precedes = "≺";
var PrecedesEqual = "⪯";
var PrecedesSlantEqual = "≼";
var PrecedesTilde = "≾";
var preceq = "⪯";
var precnapprox = "⪹";
var precneqq = "⪵";
var precnsim = "⋨";
var pre = "⪯";
var prE = "⪳";
var precsim = "≾";
var prime = "′";
var Prime = "″";
var primes = "ℙ";
var prnap = "⪹";
var prnE = "⪵";
var prnsim = "⋨";
var prod = "∏";
var Product = "∏";
var profalar = "⌮";
var profline = "⌒";
var profsurf = "⌓";
var prop = "∝";
var Proportional = "∝";
var Proportion = "∷";
var propto = "∝";
var prsim = "≾";
var prurel = "⊰";
var Pscr = "𝒫";
var pscr = "𝓅";
var Psi = "Ψ";
var psi = "ψ";
var puncsp = " ";
var Qfr = "𝔔";
var qfr = "𝔮";
var qint = "⨌";
var qopf = "𝕢";
var Qopf = "ℚ";
var qprime = "⁗";
var Qscr = "𝒬";
var qscr = "𝓆";
var quaternions = "ℍ";
var quatint = "⨖";
var quest = "?";
var questeq = "≟";
var quot$1 = "\"";
var QUOT = "\"";
var rAarr = "⇛";
var race = "∽̱";
var Racute = "Ŕ";
var racute = "ŕ";
var radic = "√";
var raemptyv = "⦳";
var rang = "⟩";
var Rang = "⟫";
var rangd = "⦒";
var range = "⦥";
var rangle = "⟩";
var raquo = "»";
var rarrap = "⥵";
var rarrb = "⇥";
var rarrbfs = "⤠";
var rarrc = "⤳";
var rarr = "→";
var Rarr = "↠";
var rArr = "⇒";
var rarrfs = "⤞";
var rarrhk = "↪";
var rarrlp = "↬";
var rarrpl = "⥅";
var rarrsim = "⥴";
var Rarrtl = "⤖";
var rarrtl = "↣";
var rarrw = "↝";
var ratail = "⤚";
var rAtail = "⤜";
var ratio = "∶";
var rationals = "ℚ";
var rbarr = "⤍";
var rBarr = "⤏";
var RBarr = "⤐";
var rbbrk = "❳";
var rbrace = "}";
var rbrack = "]";
var rbrke = "⦌";
var rbrksld = "⦎";
var rbrkslu = "⦐";
var Rcaron = "Ř";
var rcaron = "ř";
var Rcedil = "Ŗ";
var rcedil = "ŗ";
var rceil = "⌉";
var rcub = "}";
var Rcy = "Р";
var rcy = "р";
var rdca = "⤷";
var rdldhar = "⥩";
var rdquo = "”";
var rdquor = "”";
var rdsh = "↳";
var real = "ℜ";
var realine = "ℛ";
var realpart = "ℜ";
var reals = "ℝ";
var Re = "ℜ";
var rect = "▭";
var reg = "®";
var REG = "®";
var ReverseElement = "∋";
var ReverseEquilibrium = "⇋";
var ReverseUpEquilibrium = "⥯";
var rfisht = "⥽";
var rfloor = "⌋";
var rfr = "𝔯";
var Rfr = "ℜ";
var rHar = "⥤";
var rhard = "⇁";
var rharu = "⇀";
var rharul = "⥬";
var Rho = "Ρ";
var rho = "ρ";
var rhov = "ϱ";
var RightAngleBracket = "⟩";
var RightArrowBar = "⇥";
var rightarrow = "→";
var RightArrow = "→";
var Rightarrow = "⇒";
var RightArrowLeftArrow = "⇄";
var rightarrowtail = "↣";
var RightCeiling = "⌉";
var RightDoubleBracket = "⟧";
var RightDownTeeVector = "⥝";
var RightDownVectorBar = "⥕";
var RightDownVector = "⇂";
var RightFloor = "⌋";
var rightharpoondown = "⇁";
var rightharpoonup = "⇀";
var rightleftarrows = "⇄";
var rightleftharpoons = "⇌";
var rightrightarrows = "⇉";
var rightsquigarrow = "↝";
var RightTeeArrow = "↦";
var RightTee = "⊢";
var RightTeeVector = "⥛";
var rightthreetimes = "⋌";
var RightTriangleBar = "⧐";
var RightTriangle = "⊳";
var RightTriangleEqual = "⊵";
var RightUpDownVector = "⥏";
var RightUpTeeVector = "⥜";
var RightUpVectorBar = "⥔";
var RightUpVector = "↾";
var RightVectorBar = "⥓";
var RightVector = "⇀";
var ring = "˚";
var risingdotseq = "≓";
var rlarr = "⇄";
var rlhar = "⇌";
var rlm = "‏";
var rmoustache = "⎱";
var rmoust = "⎱";
var rnmid = "⫮";
var roang = "⟭";
var roarr = "⇾";
var robrk = "⟧";
var ropar = "⦆";
var ropf = "𝕣";
var Ropf = "ℝ";
var roplus = "⨮";
var rotimes = "⨵";
var RoundImplies = "⥰";
var rpar = ")";
var rpargt = "⦔";
var rppolint = "⨒";
var rrarr = "⇉";
var Rrightarrow = "⇛";
var rsaquo = "›";
var rscr = "𝓇";
var Rscr = "ℛ";
var rsh = "↱";
var Rsh = "↱";
var rsqb = "]";
var rsquo = "’";
var rsquor = "’";
var rthree = "⋌";
var rtimes = "⋊";
var rtri = "▹";
var rtrie = "⊵";
var rtrif = "▸";
var rtriltri = "⧎";
var RuleDelayed = "⧴";
var ruluhar = "⥨";
var rx = "℞";
var Sacute = "Ś";
var sacute = "ś";
var sbquo = "‚";
var scap = "⪸";
var Scaron = "Š";
var scaron = "š";
var Sc = "⪼";
var sc = "≻";
var sccue = "≽";
var sce = "⪰";
var scE = "⪴";
var Scedil = "Ş";
var scedil = "ş";
var Scirc = "Ŝ";
var scirc = "ŝ";
var scnap = "⪺";
var scnE = "⪶";
var scnsim = "⋩";
var scpolint = "⨓";
var scsim = "≿";
var Scy = "С";
var scy = "с";
var sdotb = "⊡";
var sdot = "⋅";
var sdote = "⩦";
var searhk = "⤥";
var searr = "↘";
var seArr = "⇘";
var searrow = "↘";
var sect = "§";
var semi = ";";
var seswar = "⤩";
var setminus = "∖";
var setmn = "∖";
var sext = "✶";
var Sfr = "𝔖";
var sfr = "𝔰";
var sfrown = "⌢";
var sharp = "♯";
var SHCHcy = "Щ";
var shchcy = "щ";
var SHcy = "Ш";
var shcy = "ш";
var ShortDownArrow = "↓";
var ShortLeftArrow = "←";
var shortmid = "∣";
var shortparallel = "∥";
var ShortRightArrow = "→";
var ShortUpArrow = "↑";
var shy = "­";
var Sigma = "Σ";
var sigma = "σ";
var sigmaf = "ς";
var sigmav = "ς";
var sim = "∼";
var simdot = "⩪";
var sime = "≃";
var simeq = "≃";
var simg = "⪞";
var simgE = "⪠";
var siml = "⪝";
var simlE = "⪟";
var simne = "≆";
var simplus = "⨤";
var simrarr = "⥲";
var slarr = "←";
var SmallCircle = "∘";
var smallsetminus = "∖";
var smashp = "⨳";
var smeparsl = "⧤";
var smid = "∣";
var smile = "⌣";
var smt = "⪪";
var smte = "⪬";
var smtes = "⪬︀";
var SOFTcy = "Ь";
var softcy = "ь";
var solbar = "⌿";
var solb = "⧄";
var sol = "/";
var Sopf = "𝕊";
var sopf = "𝕤";
var spades = "♠";
var spadesuit = "♠";
var spar = "∥";
var sqcap = "⊓";
var sqcaps = "⊓︀";
var sqcup = "⊔";
var sqcups = "⊔︀";
var Sqrt = "√";
var sqsub = "⊏";
var sqsube = "⊑";
var sqsubset = "⊏";
var sqsubseteq = "⊑";
var sqsup = "⊐";
var sqsupe = "⊒";
var sqsupset = "⊐";
var sqsupseteq = "⊒";
var square = "□";
var Square = "□";
var SquareIntersection = "⊓";
var SquareSubset = "⊏";
var SquareSubsetEqual = "⊑";
var SquareSuperset = "⊐";
var SquareSupersetEqual = "⊒";
var SquareUnion = "⊔";
var squarf = "▪";
var squ = "□";
var squf = "▪";
var srarr = "→";
var Sscr = "𝒮";
var sscr = "𝓈";
var ssetmn = "∖";
var ssmile = "⌣";
var sstarf = "⋆";
var Star = "⋆";
var star = "☆";
var starf = "★";
var straightepsilon = "ϵ";
var straightphi = "ϕ";
var strns = "¯";
var sub = "⊂";
var Sub = "⋐";
var subdot = "⪽";
var subE = "⫅";
var sube = "⊆";
var subedot = "⫃";
var submult = "⫁";
var subnE = "⫋";
var subne = "⊊";
var subplus = "⪿";
var subrarr = "⥹";
var subset = "⊂";
var Subset = "⋐";
var subseteq = "⊆";
var subseteqq = "⫅";
var SubsetEqual = "⊆";
var subsetneq = "⊊";
var subsetneqq = "⫋";
var subsim = "⫇";
var subsub = "⫕";
var subsup = "⫓";
var succapprox = "⪸";
var succ = "≻";
var succcurlyeq = "≽";
var Succeeds = "≻";
var SucceedsEqual = "⪰";
var SucceedsSlantEqual = "≽";
var SucceedsTilde = "≿";
var succeq = "⪰";
var succnapprox = "⪺";
var succneqq = "⪶";
var succnsim = "⋩";
var succsim = "≿";
var SuchThat = "∋";
var sum = "∑";
var Sum = "∑";
var sung = "♪";
var sup1 = "¹";
var sup2 = "²";
var sup3 = "³";
var sup = "⊃";
var Sup = "⋑";
var supdot = "⪾";
var supdsub = "⫘";
var supE = "⫆";
var supe = "⊇";
var supedot = "⫄";
var Superset = "⊃";
var SupersetEqual = "⊇";
var suphsol = "⟉";
var suphsub = "⫗";
var suplarr = "⥻";
var supmult = "⫂";
var supnE = "⫌";
var supne = "⊋";
var supplus = "⫀";
var supset = "⊃";
var Supset = "⋑";
var supseteq = "⊇";
var supseteqq = "⫆";
var supsetneq = "⊋";
var supsetneqq = "⫌";
var supsim = "⫈";
var supsub = "⫔";
var supsup = "⫖";
var swarhk = "⤦";
var swarr = "↙";
var swArr = "⇙";
var swarrow = "↙";
var swnwar = "⤪";
var szlig = "ß";
var Tab = "\t";
var target = "⌖";
var Tau = "Τ";
var tau = "τ";
var tbrk = "⎴";
var Tcaron = "Ť";
var tcaron = "ť";
var Tcedil = "Ţ";
var tcedil = "ţ";
var Tcy = "Т";
var tcy = "т";
var tdot = "⃛";
var telrec = "⌕";
var Tfr = "𝔗";
var tfr = "𝔱";
var there4 = "∴";
var therefore = "∴";
var Therefore = "∴";
var Theta = "Θ";
var theta = "θ";
var thetasym = "ϑ";
var thetav = "ϑ";
var thickapprox = "≈";
var thicksim = "∼";
var ThickSpace = "  ";
var ThinSpace = " ";
var thinsp = " ";
var thkap = "≈";
var thksim = "∼";
var THORN = "Þ";
var thorn = "þ";
var tilde = "˜";
var Tilde = "∼";
var TildeEqual = "≃";
var TildeFullEqual = "≅";
var TildeTilde = "≈";
var timesbar = "⨱";
var timesb = "⊠";
var times$1 = "×";
var timesd = "⨰";
var tint = "∭";
var toea = "⤨";
var topbot = "⌶";
var topcir = "⫱";
var top = "⊤";
var Topf = "𝕋";
var topf = "𝕥";
var topfork = "⫚";
var tosa = "⤩";
var tprime = "‴";
var trade = "™";
var TRADE = "™";
var triangle = "▵";
var triangledown = "▿";
var triangleleft = "◃";
var trianglelefteq = "⊴";
var triangleq = "≜";
var triangleright = "▹";
var trianglerighteq = "⊵";
var tridot = "◬";
var trie = "≜";
var triminus = "⨺";
var TripleDot = "⃛";
var triplus = "⨹";
var trisb = "⧍";
var tritime = "⨻";
var trpezium = "⏢";
var Tscr = "𝒯";
var tscr = "𝓉";
var TScy = "Ц";
var tscy = "ц";
var TSHcy = "Ћ";
var tshcy = "ћ";
var Tstrok = "Ŧ";
var tstrok = "ŧ";
var twixt = "≬";
var twoheadleftarrow = "↞";
var twoheadrightarrow = "↠";
var Uacute = "Ú";
var uacute = "ú";
var uarr = "↑";
var Uarr = "↟";
var uArr = "⇑";
var Uarrocir = "⥉";
var Ubrcy = "Ў";
var ubrcy = "ў";
var Ubreve = "Ŭ";
var ubreve = "ŭ";
var Ucirc = "Û";
var ucirc = "û";
var Ucy = "У";
var ucy = "у";
var udarr = "⇅";
var Udblac = "Ű";
var udblac = "ű";
var udhar = "⥮";
var ufisht = "⥾";
var Ufr = "𝔘";
var ufr = "𝔲";
var Ugrave = "Ù";
var ugrave = "ù";
var uHar = "⥣";
var uharl = "↿";
var uharr = "↾";
var uhblk = "▀";
var ulcorn = "⌜";
var ulcorner = "⌜";
var ulcrop = "⌏";
var ultri = "◸";
var Umacr = "Ū";
var umacr = "ū";
var uml = "¨";
var UnderBar = "_";
var UnderBrace = "⏟";
var UnderBracket = "⎵";
var UnderParenthesis = "⏝";
var Union = "⋃";
var UnionPlus = "⊎";
var Uogon = "Ų";
var uogon = "ų";
var Uopf = "𝕌";
var uopf = "𝕦";
var UpArrowBar = "⤒";
var uparrow = "↑";
var UpArrow = "↑";
var Uparrow = "⇑";
var UpArrowDownArrow = "⇅";
var updownarrow = "↕";
var UpDownArrow = "↕";
var Updownarrow = "⇕";
var UpEquilibrium = "⥮";
var upharpoonleft = "↿";
var upharpoonright = "↾";
var uplus = "⊎";
var UpperLeftArrow = "↖";
var UpperRightArrow = "↗";
var upsi = "υ";
var Upsi = "ϒ";
var upsih = "ϒ";
var Upsilon = "Υ";
var upsilon = "υ";
var UpTeeArrow = "↥";
var UpTee = "⊥";
var upuparrows = "⇈";
var urcorn = "⌝";
var urcorner = "⌝";
var urcrop = "⌎";
var Uring = "Ů";
var uring = "ů";
var urtri = "◹";
var Uscr = "𝒰";
var uscr = "𝓊";
var utdot = "⋰";
var Utilde = "Ũ";
var utilde = "ũ";
var utri = "▵";
var utrif = "▴";
var uuarr = "⇈";
var Uuml = "Ü";
var uuml = "ü";
var uwangle = "⦧";
var vangrt = "⦜";
var varepsilon = "ϵ";
var varkappa = "ϰ";
var varnothing = "∅";
var varphi = "ϕ";
var varpi = "ϖ";
var varpropto = "∝";
var varr = "↕";
var vArr = "⇕";
var varrho = "ϱ";
var varsigma = "ς";
var varsubsetneq = "⊊︀";
var varsubsetneqq = "⫋︀";
var varsupsetneq = "⊋︀";
var varsupsetneqq = "⫌︀";
var vartheta = "ϑ";
var vartriangleleft = "⊲";
var vartriangleright = "⊳";
var vBar = "⫨";
var Vbar = "⫫";
var vBarv = "⫩";
var Vcy = "В";
var vcy = "в";
var vdash = "⊢";
var vDash = "⊨";
var Vdash = "⊩";
var VDash = "⊫";
var Vdashl = "⫦";
var veebar = "⊻";
var vee = "∨";
var Vee = "⋁";
var veeeq = "≚";
var vellip = "⋮";
var verbar = "|";
var Verbar = "‖";
var vert = "|";
var Vert = "‖";
var VerticalBar = "∣";
var VerticalLine = "|";
var VerticalSeparator = "❘";
var VerticalTilde = "≀";
var VeryThinSpace = " ";
var Vfr = "𝔙";
var vfr = "𝔳";
var vltri = "⊲";
var vnsub = "⊂⃒";
var vnsup = "⊃⃒";
var Vopf = "𝕍";
var vopf = "𝕧";
var vprop = "∝";
var vrtri = "⊳";
var Vscr = "𝒱";
var vscr = "𝓋";
var vsubnE = "⫋︀";
var vsubne = "⊊︀";
var vsupnE = "⫌︀";
var vsupne = "⊋︀";
var Vvdash = "⊪";
var vzigzag = "⦚";
var Wcirc = "Ŵ";
var wcirc = "ŵ";
var wedbar = "⩟";
var wedge = "∧";
var Wedge = "⋀";
var wedgeq = "≙";
var weierp = "℘";
var Wfr = "𝔚";
var wfr = "𝔴";
var Wopf = "𝕎";
var wopf = "𝕨";
var wp = "℘";
var wr = "≀";
var wreath = "≀";
var Wscr = "𝒲";
var wscr = "𝓌";
var xcap = "⋂";
var xcirc = "◯";
var xcup = "⋃";
var xdtri = "▽";
var Xfr = "𝔛";
var xfr = "𝔵";
var xharr = "⟷";
var xhArr = "⟺";
var Xi = "Ξ";
var xi = "ξ";
var xlarr = "⟵";
var xlArr = "⟸";
var xmap = "⟼";
var xnis = "⋻";
var xodot = "⨀";
var Xopf = "𝕏";
var xopf = "𝕩";
var xoplus = "⨁";
var xotime = "⨂";
var xrarr = "⟶";
var xrArr = "⟹";
var Xscr = "𝒳";
var xscr = "𝓍";
var xsqcup = "⨆";
var xuplus = "⨄";
var xutri = "△";
var xvee = "⋁";
var xwedge = "⋀";
var Yacute = "Ý";
var yacute = "ý";
var YAcy = "Я";
var yacy = "я";
var Ycirc = "Ŷ";
var ycirc = "ŷ";
var Ycy = "Ы";
var ycy = "ы";
var yen = "¥";
var Yfr = "𝔜";
var yfr = "𝔶";
var YIcy = "Ї";
var yicy = "ї";
var Yopf = "𝕐";
var yopf = "𝕪";
var Yscr = "𝒴";
var yscr = "𝓎";
var YUcy = "Ю";
var yucy = "ю";
var yuml = "ÿ";
var Yuml = "Ÿ";
var Zacute = "Ź";
var zacute = "ź";
var Zcaron = "Ž";
var zcaron = "ž";
var Zcy = "З";
var zcy = "з";
var Zdot = "Ż";
var zdot = "ż";
var zeetrf = "ℨ";
var ZeroWidthSpace = "​";
var Zeta = "Ζ";
var zeta = "ζ";
var zfr = "𝔷";
var Zfr = "ℨ";
var ZHcy = "Ж";
var zhcy = "ж";
var zigrarr = "⇝";
var zopf = "𝕫";
var Zopf = "ℤ";
var Zscr = "𝒵";
var zscr = "𝓏";
var zwj = "‍";
var zwnj = "‌";
var entitiesJSON = {
	Aacute: Aacute,
	aacute: aacute,
	Abreve: Abreve,
	abreve: abreve,
	ac: ac,
	acd: acd,
	acE: acE,
	Acirc: Acirc,
	acirc: acirc,
	acute: acute,
	Acy: Acy,
	acy: acy,
	AElig: AElig,
	aelig: aelig,
	af: af,
	Afr: Afr,
	afr: afr,
	Agrave: Agrave,
	agrave: agrave,
	alefsym: alefsym,
	aleph: aleph,
	Alpha: Alpha,
	alpha: alpha,
	Amacr: Amacr,
	amacr: amacr,
	amalg: amalg,
	amp: amp$1,
	AMP: AMP,
	andand: andand,
	And: And,
	and: and,
	andd: andd,
	andslope: andslope,
	andv: andv,
	ang: ang,
	ange: ange,
	angle: angle,
	angmsdaa: angmsdaa,
	angmsdab: angmsdab,
	angmsdac: angmsdac,
	angmsdad: angmsdad,
	angmsdae: angmsdae,
	angmsdaf: angmsdaf,
	angmsdag: angmsdag,
	angmsdah: angmsdah,
	angmsd: angmsd,
	angrt: angrt,
	angrtvb: angrtvb,
	angrtvbd: angrtvbd,
	angsph: angsph,
	angst: angst,
	angzarr: angzarr,
	Aogon: Aogon,
	aogon: aogon,
	Aopf: Aopf,
	aopf: aopf,
	apacir: apacir,
	ap: ap,
	apE: apE,
	ape: ape,
	apid: apid,
	apos: apos$1,
	ApplyFunction: ApplyFunction,
	approx: approx,
	approxeq: approxeq,
	Aring: Aring,
	aring: aring,
	Ascr: Ascr,
	ascr: ascr,
	Assign: Assign,
	ast: ast,
	asymp: asymp,
	asympeq: asympeq,
	Atilde: Atilde,
	atilde: atilde,
	Auml: Auml,
	auml: auml,
	awconint: awconint,
	awint: awint,
	backcong: backcong,
	backepsilon: backepsilon,
	backprime: backprime,
	backsim: backsim,
	backsimeq: backsimeq,
	Backslash: Backslash,
	Barv: Barv,
	barvee: barvee,
	barwed: barwed,
	Barwed: Barwed,
	barwedge: barwedge,
	bbrk: bbrk,
	bbrktbrk: bbrktbrk,
	bcong: bcong,
	Bcy: Bcy,
	bcy: bcy,
	bdquo: bdquo,
	becaus: becaus,
	because: because,
	Because: Because,
	bemptyv: bemptyv,
	bepsi: bepsi,
	bernou: bernou,
	Bernoullis: Bernoullis,
	Beta: Beta,
	beta: beta,
	beth: beth,
	between: between,
	Bfr: Bfr,
	bfr: bfr,
	bigcap: bigcap,
	bigcirc: bigcirc,
	bigcup: bigcup,
	bigodot: bigodot,
	bigoplus: bigoplus,
	bigotimes: bigotimes,
	bigsqcup: bigsqcup,
	bigstar: bigstar,
	bigtriangledown: bigtriangledown,
	bigtriangleup: bigtriangleup,
	biguplus: biguplus,
	bigvee: bigvee,
	bigwedge: bigwedge,
	bkarow: bkarow,
	blacklozenge: blacklozenge,
	blacksquare: blacksquare,
	blacktriangle: blacktriangle,
	blacktriangledown: blacktriangledown,
	blacktriangleleft: blacktriangleleft,
	blacktriangleright: blacktriangleright,
	blank: blank,
	blk12: blk12,
	blk14: blk14,
	blk34: blk34,
	block: block,
	bne: bne,
	bnequiv: bnequiv,
	bNot: bNot,
	bnot: bnot,
	Bopf: Bopf,
	bopf: bopf,
	bot: bot,
	bottom: bottom,
	bowtie: bowtie,
	boxbox: boxbox,
	boxdl: boxdl,
	boxdL: boxdL,
	boxDl: boxDl,
	boxDL: boxDL,
	boxdr: boxdr,
	boxdR: boxdR,
	boxDr: boxDr,
	boxDR: boxDR,
	boxh: boxh,
	boxH: boxH,
	boxhd: boxhd,
	boxHd: boxHd,
	boxhD: boxhD,
	boxHD: boxHD,
	boxhu: boxhu,
	boxHu: boxHu,
	boxhU: boxhU,
	boxHU: boxHU,
	boxminus: boxminus,
	boxplus: boxplus,
	boxtimes: boxtimes,
	boxul: boxul,
	boxuL: boxuL,
	boxUl: boxUl,
	boxUL: boxUL,
	boxur: boxur,
	boxuR: boxuR,
	boxUr: boxUr,
	boxUR: boxUR,
	boxv: boxv,
	boxV: boxV,
	boxvh: boxvh,
	boxvH: boxvH,
	boxVh: boxVh,
	boxVH: boxVH,
	boxvl: boxvl,
	boxvL: boxvL,
	boxVl: boxVl,
	boxVL: boxVL,
	boxvr: boxvr,
	boxvR: boxvR,
	boxVr: boxVr,
	boxVR: boxVR,
	bprime: bprime,
	breve: breve,
	Breve: Breve,
	brvbar: brvbar,
	bscr: bscr,
	Bscr: Bscr,
	bsemi: bsemi,
	bsim: bsim,
	bsime: bsime,
	bsolb: bsolb,
	bsol: bsol,
	bsolhsub: bsolhsub,
	bull: bull,
	bullet: bullet,
	bump: bump,
	bumpE: bumpE,
	bumpe: bumpe,
	Bumpeq: Bumpeq,
	bumpeq: bumpeq,
	Cacute: Cacute,
	cacute: cacute,
	capand: capand,
	capbrcup: capbrcup,
	capcap: capcap,
	cap: cap,
	Cap: Cap,
	capcup: capcup,
	capdot: capdot,
	CapitalDifferentialD: CapitalDifferentialD,
	caps: caps,
	caret: caret,
	caron: caron,
	Cayleys: Cayleys,
	ccaps: ccaps,
	Ccaron: Ccaron,
	ccaron: ccaron,
	Ccedil: Ccedil,
	ccedil: ccedil,
	Ccirc: Ccirc,
	ccirc: ccirc,
	Cconint: Cconint,
	ccups: ccups,
	ccupssm: ccupssm,
	Cdot: Cdot,
	cdot: cdot,
	cedil: cedil,
	Cedilla: Cedilla,
	cemptyv: cemptyv,
	cent: cent,
	centerdot: centerdot,
	CenterDot: CenterDot,
	cfr: cfr,
	Cfr: Cfr,
	CHcy: CHcy,
	chcy: chcy,
	check: check,
	checkmark: checkmark,
	Chi: Chi,
	chi: chi,
	circ: circ,
	circeq: circeq,
	circlearrowleft: circlearrowleft,
	circlearrowright: circlearrowright,
	circledast: circledast,
	circledcirc: circledcirc,
	circleddash: circleddash,
	CircleDot: CircleDot,
	circledR: circledR,
	circledS: circledS,
	CircleMinus: CircleMinus,
	CirclePlus: CirclePlus,
	CircleTimes: CircleTimes,
	cir: cir,
	cirE: cirE,
	cire: cire,
	cirfnint: cirfnint,
	cirmid: cirmid,
	cirscir: cirscir,
	ClockwiseContourIntegral: ClockwiseContourIntegral,
	CloseCurlyDoubleQuote: CloseCurlyDoubleQuote,
	CloseCurlyQuote: CloseCurlyQuote,
	clubs: clubs,
	clubsuit: clubsuit,
	colon: colon,
	Colon: Colon,
	Colone: Colone,
	colone: colone,
	coloneq: coloneq,
	comma: comma,
	commat: commat,
	comp: comp,
	compfn: compfn,
	complement: complement,
	complexes: complexes,
	cong: cong,
	congdot: congdot,
	Congruent: Congruent,
	conint: conint,
	Conint: Conint,
	ContourIntegral: ContourIntegral,
	copf: copf,
	Copf: Copf,
	coprod: coprod,
	Coproduct: Coproduct,
	copy: copy,
	COPY: COPY,
	copysr: copysr,
	CounterClockwiseContourIntegral: CounterClockwiseContourIntegral,
	crarr: crarr,
	cross: cross,
	Cross: Cross,
	Cscr: Cscr,
	cscr: cscr,
	csub: csub,
	csube: csube,
	csup: csup,
	csupe: csupe,
	ctdot: ctdot,
	cudarrl: cudarrl,
	cudarrr: cudarrr,
	cuepr: cuepr,
	cuesc: cuesc,
	cularr: cularr,
	cularrp: cularrp,
	cupbrcap: cupbrcap,
	cupcap: cupcap,
	CupCap: CupCap,
	cup: cup,
	Cup: Cup,
	cupcup: cupcup,
	cupdot: cupdot,
	cupor: cupor,
	cups: cups,
	curarr: curarr,
	curarrm: curarrm,
	curlyeqprec: curlyeqprec,
	curlyeqsucc: curlyeqsucc,
	curlyvee: curlyvee,
	curlywedge: curlywedge,
	curren: curren,
	curvearrowleft: curvearrowleft,
	curvearrowright: curvearrowright,
	cuvee: cuvee,
	cuwed: cuwed,
	cwconint: cwconint,
	cwint: cwint,
	cylcty: cylcty,
	dagger: dagger,
	Dagger: Dagger,
	daleth: daleth,
	darr: darr,
	Darr: Darr,
	dArr: dArr,
	dash: dash,
	Dashv: Dashv,
	dashv: dashv,
	dbkarow: dbkarow,
	dblac: dblac,
	Dcaron: Dcaron,
	dcaron: dcaron,
	Dcy: Dcy,
	dcy: dcy,
	ddagger: ddagger,
	ddarr: ddarr,
	DD: DD,
	dd: dd,
	DDotrahd: DDotrahd,
	ddotseq: ddotseq,
	deg: deg,
	Del: Del,
	Delta: Delta,
	delta: delta,
	demptyv: demptyv,
	dfisht: dfisht,
	Dfr: Dfr,
	dfr: dfr,
	dHar: dHar,
	dharl: dharl,
	dharr: dharr,
	DiacriticalAcute: DiacriticalAcute,
	DiacriticalDot: DiacriticalDot,
	DiacriticalDoubleAcute: DiacriticalDoubleAcute,
	DiacriticalGrave: DiacriticalGrave,
	DiacriticalTilde: DiacriticalTilde,
	diam: diam,
	diamond: diamond,
	Diamond: Diamond,
	diamondsuit: diamondsuit,
	diams: diams,
	die: die,
	DifferentialD: DifferentialD,
	digamma: digamma,
	disin: disin,
	div: div,
	divide: divide,
	divideontimes: divideontimes,
	divonx: divonx,
	DJcy: DJcy,
	djcy: djcy,
	dlcorn: dlcorn,
	dlcrop: dlcrop,
	dollar: dollar,
	Dopf: Dopf,
	dopf: dopf,
	Dot: Dot,
	dot: dot,
	DotDot: DotDot,
	doteq: doteq,
	doteqdot: doteqdot,
	DotEqual: DotEqual,
	dotminus: dotminus,
	dotplus: dotplus,
	dotsquare: dotsquare,
	doublebarwedge: doublebarwedge,
	DoubleContourIntegral: DoubleContourIntegral,
	DoubleDot: DoubleDot,
	DoubleDownArrow: DoubleDownArrow,
	DoubleLeftArrow: DoubleLeftArrow,
	DoubleLeftRightArrow: DoubleLeftRightArrow,
	DoubleLeftTee: DoubleLeftTee,
	DoubleLongLeftArrow: DoubleLongLeftArrow,
	DoubleLongLeftRightArrow: DoubleLongLeftRightArrow,
	DoubleLongRightArrow: DoubleLongRightArrow,
	DoubleRightArrow: DoubleRightArrow,
	DoubleRightTee: DoubleRightTee,
	DoubleUpArrow: DoubleUpArrow,
	DoubleUpDownArrow: DoubleUpDownArrow,
	DoubleVerticalBar: DoubleVerticalBar,
	DownArrowBar: DownArrowBar,
	downarrow: downarrow,
	DownArrow: DownArrow,
	Downarrow: Downarrow,
	DownArrowUpArrow: DownArrowUpArrow,
	DownBreve: DownBreve,
	downdownarrows: downdownarrows,
	downharpoonleft: downharpoonleft,
	downharpoonright: downharpoonright,
	DownLeftRightVector: DownLeftRightVector,
	DownLeftTeeVector: DownLeftTeeVector,
	DownLeftVectorBar: DownLeftVectorBar,
	DownLeftVector: DownLeftVector,
	DownRightTeeVector: DownRightTeeVector,
	DownRightVectorBar: DownRightVectorBar,
	DownRightVector: DownRightVector,
	DownTeeArrow: DownTeeArrow,
	DownTee: DownTee,
	drbkarow: drbkarow,
	drcorn: drcorn,
	drcrop: drcrop,
	Dscr: Dscr,
	dscr: dscr,
	DScy: DScy,
	dscy: dscy,
	dsol: dsol,
	Dstrok: Dstrok,
	dstrok: dstrok,
	dtdot: dtdot,
	dtri: dtri,
	dtrif: dtrif,
	duarr: duarr,
	duhar: duhar,
	dwangle: dwangle,
	DZcy: DZcy,
	dzcy: dzcy,
	dzigrarr: dzigrarr,
	Eacute: Eacute,
	eacute: eacute,
	easter: easter,
	Ecaron: Ecaron,
	ecaron: ecaron,
	Ecirc: Ecirc,
	ecirc: ecirc,
	ecir: ecir,
	ecolon: ecolon,
	Ecy: Ecy,
	ecy: ecy,
	eDDot: eDDot,
	Edot: Edot,
	edot: edot,
	eDot: eDot,
	ee: ee,
	efDot: efDot,
	Efr: Efr,
	efr: efr,
	eg: eg,
	Egrave: Egrave,
	egrave: egrave,
	egs: egs,
	egsdot: egsdot,
	el: el,
	Element: Element,
	elinters: elinters,
	ell: ell,
	els: els,
	elsdot: elsdot,
	Emacr: Emacr,
	emacr: emacr,
	empty: empty,
	emptyset: emptyset,
	EmptySmallSquare: EmptySmallSquare,
	emptyv: emptyv,
	EmptyVerySmallSquare: EmptyVerySmallSquare,
	emsp13: emsp13,
	emsp14: emsp14,
	emsp: emsp,
	ENG: ENG,
	eng: eng,
	ensp: ensp,
	Eogon: Eogon,
	eogon: eogon,
	Eopf: Eopf,
	eopf: eopf,
	epar: epar,
	eparsl: eparsl,
	eplus: eplus,
	epsi: epsi,
	Epsilon: Epsilon,
	epsilon: epsilon,
	epsiv: epsiv,
	eqcirc: eqcirc,
	eqcolon: eqcolon,
	eqsim: eqsim,
	eqslantgtr: eqslantgtr,
	eqslantless: eqslantless,
	Equal: Equal,
	equals: equals,
	EqualTilde: EqualTilde,
	equest: equest,
	Equilibrium: Equilibrium,
	equiv: equiv,
	equivDD: equivDD,
	eqvparsl: eqvparsl,
	erarr: erarr,
	erDot: erDot,
	escr: escr,
	Escr: Escr,
	esdot: esdot,
	Esim: Esim,
	esim: esim,
	Eta: Eta,
	eta: eta,
	ETH: ETH,
	eth: eth,
	Euml: Euml,
	euml: euml,
	euro: euro,
	excl: excl,
	exist: exist,
	Exists: Exists,
	expectation: expectation,
	exponentiale: exponentiale,
	ExponentialE: ExponentialE,
	fallingdotseq: fallingdotseq,
	Fcy: Fcy,
	fcy: fcy,
	female: female,
	ffilig: ffilig,
	fflig: fflig,
	ffllig: ffllig,
	Ffr: Ffr,
	ffr: ffr,
	filig: filig,
	FilledSmallSquare: FilledSmallSquare,
	FilledVerySmallSquare: FilledVerySmallSquare,
	fjlig: fjlig,
	flat: flat,
	fllig: fllig,
	fltns: fltns,
	fnof: fnof,
	Fopf: Fopf,
	fopf: fopf,
	forall: forall,
	ForAll: ForAll,
	fork: fork,
	forkv: forkv,
	Fouriertrf: Fouriertrf,
	fpartint: fpartint,
	frac12: frac12,
	frac13: frac13,
	frac14: frac14,
	frac15: frac15,
	frac16: frac16,
	frac18: frac18,
	frac23: frac23,
	frac25: frac25,
	frac34: frac34,
	frac35: frac35,
	frac38: frac38,
	frac45: frac45,
	frac56: frac56,
	frac58: frac58,
	frac78: frac78,
	frasl: frasl,
	frown: frown,
	fscr: fscr,
	Fscr: Fscr,
	gacute: gacute,
	Gamma: Gamma,
	gamma: gamma,
	Gammad: Gammad,
	gammad: gammad,
	gap: gap,
	Gbreve: Gbreve,
	gbreve: gbreve,
	Gcedil: Gcedil,
	Gcirc: Gcirc,
	gcirc: gcirc,
	Gcy: Gcy,
	gcy: gcy,
	Gdot: Gdot,
	gdot: gdot,
	ge: ge,
	gE: gE,
	gEl: gEl,
	gel: gel,
	geq: geq,
	geqq: geqq,
	geqslant: geqslant,
	gescc: gescc,
	ges: ges,
	gesdot: gesdot,
	gesdoto: gesdoto,
	gesdotol: gesdotol,
	gesl: gesl,
	gesles: gesles,
	Gfr: Gfr,
	gfr: gfr,
	gg: gg,
	Gg: Gg,
	ggg: ggg,
	gimel: gimel,
	GJcy: GJcy,
	gjcy: gjcy,
	gla: gla,
	gl: gl,
	glE: glE,
	glj: glj,
	gnap: gnap,
	gnapprox: gnapprox,
	gne: gne,
	gnE: gnE,
	gneq: gneq,
	gneqq: gneqq,
	gnsim: gnsim,
	Gopf: Gopf,
	gopf: gopf,
	grave: grave,
	GreaterEqual: GreaterEqual,
	GreaterEqualLess: GreaterEqualLess,
	GreaterFullEqual: GreaterFullEqual,
	GreaterGreater: GreaterGreater,
	GreaterLess: GreaterLess,
	GreaterSlantEqual: GreaterSlantEqual,
	GreaterTilde: GreaterTilde,
	Gscr: Gscr,
	gscr: gscr,
	gsim: gsim,
	gsime: gsime,
	gsiml: gsiml,
	gtcc: gtcc,
	gtcir: gtcir,
	gt: gt$1,
	GT: GT,
	Gt: Gt,
	gtdot: gtdot,
	gtlPar: gtlPar,
	gtquest: gtquest,
	gtrapprox: gtrapprox,
	gtrarr: gtrarr,
	gtrdot: gtrdot,
	gtreqless: gtreqless,
	gtreqqless: gtreqqless,
	gtrless: gtrless,
	gtrsim: gtrsim,
	gvertneqq: gvertneqq,
	gvnE: gvnE,
	Hacek: Hacek,
	hairsp: hairsp,
	half: half,
	hamilt: hamilt,
	HARDcy: HARDcy,
	hardcy: hardcy,
	harrcir: harrcir,
	harr: harr,
	hArr: hArr,
	harrw: harrw,
	Hat: Hat,
	hbar: hbar,
	Hcirc: Hcirc,
	hcirc: hcirc,
	hearts: hearts,
	heartsuit: heartsuit,
	hellip: hellip,
	hercon: hercon,
	hfr: hfr,
	Hfr: Hfr,
	HilbertSpace: HilbertSpace,
	hksearow: hksearow,
	hkswarow: hkswarow,
	hoarr: hoarr,
	homtht: homtht,
	hookleftarrow: hookleftarrow,
	hookrightarrow: hookrightarrow,
	hopf: hopf,
	Hopf: Hopf,
	horbar: horbar,
	HorizontalLine: HorizontalLine,
	hscr: hscr,
	Hscr: Hscr,
	hslash: hslash,
	Hstrok: Hstrok,
	hstrok: hstrok,
	HumpDownHump: HumpDownHump,
	HumpEqual: HumpEqual,
	hybull: hybull,
	hyphen: hyphen,
	Iacute: Iacute,
	iacute: iacute,
	ic: ic,
	Icirc: Icirc,
	icirc: icirc,
	Icy: Icy,
	icy: icy,
	Idot: Idot,
	IEcy: IEcy,
	iecy: iecy,
	iexcl: iexcl,
	iff: iff,
	ifr: ifr,
	Ifr: Ifr,
	Igrave: Igrave,
	igrave: igrave,
	ii: ii,
	iiiint: iiiint,
	iiint: iiint,
	iinfin: iinfin,
	iiota: iiota,
	IJlig: IJlig,
	ijlig: ijlig,
	Imacr: Imacr,
	imacr: imacr,
	image: image,
	ImaginaryI: ImaginaryI,
	imagline: imagline,
	imagpart: imagpart,
	imath: imath,
	Im: Im,
	imof: imof,
	imped: imped,
	Implies: Implies,
	incare: incare,
	infin: infin,
	infintie: infintie,
	inodot: inodot,
	intcal: intcal,
	int: int,
	Int: Int,
	integers: integers,
	Integral: Integral,
	intercal: intercal,
	Intersection: Intersection,
	intlarhk: intlarhk,
	intprod: intprod,
	InvisibleComma: InvisibleComma,
	InvisibleTimes: InvisibleTimes,
	IOcy: IOcy,
	iocy: iocy,
	Iogon: Iogon,
	iogon: iogon,
	Iopf: Iopf,
	iopf: iopf,
	Iota: Iota,
	iota: iota,
	iprod: iprod,
	iquest: iquest,
	iscr: iscr,
	Iscr: Iscr,
	isin: isin,
	isindot: isindot,
	isinE: isinE,
	isins: isins,
	isinsv: isinsv,
	isinv: isinv,
	it: it,
	Itilde: Itilde,
	itilde: itilde,
	Iukcy: Iukcy,
	iukcy: iukcy,
	Iuml: Iuml,
	iuml: iuml,
	Jcirc: Jcirc,
	jcirc: jcirc,
	Jcy: Jcy,
	jcy: jcy,
	Jfr: Jfr,
	jfr: jfr,
	jmath: jmath,
	Jopf: Jopf,
	jopf: jopf,
	Jscr: Jscr,
	jscr: jscr,
	Jsercy: Jsercy,
	jsercy: jsercy,
	Jukcy: Jukcy,
	jukcy: jukcy,
	Kappa: Kappa,
	kappa: kappa,
	kappav: kappav,
	Kcedil: Kcedil,
	kcedil: kcedil,
	Kcy: Kcy,
	kcy: kcy,
	Kfr: Kfr,
	kfr: kfr,
	kgreen: kgreen,
	KHcy: KHcy,
	khcy: khcy,
	KJcy: KJcy,
	kjcy: kjcy,
	Kopf: Kopf,
	kopf: kopf,
	Kscr: Kscr,
	kscr: kscr,
	lAarr: lAarr,
	Lacute: Lacute,
	lacute: lacute,
	laemptyv: laemptyv,
	lagran: lagran,
	Lambda: Lambda,
	lambda: lambda,
	lang: lang,
	Lang: Lang,
	langd: langd,
	langle: langle,
	lap: lap,
	Laplacetrf: Laplacetrf,
	laquo: laquo,
	larrb: larrb,
	larrbfs: larrbfs,
	larr: larr,
	Larr: Larr,
	lArr: lArr,
	larrfs: larrfs,
	larrhk: larrhk,
	larrlp: larrlp,
	larrpl: larrpl,
	larrsim: larrsim,
	larrtl: larrtl,
	latail: latail,
	lAtail: lAtail,
	lat: lat,
	late: late,
	lates: lates,
	lbarr: lbarr,
	lBarr: lBarr,
	lbbrk: lbbrk,
	lbrace: lbrace,
	lbrack: lbrack,
	lbrke: lbrke,
	lbrksld: lbrksld,
	lbrkslu: lbrkslu,
	Lcaron: Lcaron,
	lcaron: lcaron,
	Lcedil: Lcedil,
	lcedil: lcedil,
	lceil: lceil,
	lcub: lcub,
	Lcy: Lcy,
	lcy: lcy,
	ldca: ldca,
	ldquo: ldquo,
	ldquor: ldquor,
	ldrdhar: ldrdhar,
	ldrushar: ldrushar,
	ldsh: ldsh,
	le: le,
	lE: lE,
	LeftAngleBracket: LeftAngleBracket,
	LeftArrowBar: LeftArrowBar,
	leftarrow: leftarrow,
	LeftArrow: LeftArrow,
	Leftarrow: Leftarrow,
	LeftArrowRightArrow: LeftArrowRightArrow,
	leftarrowtail: leftarrowtail,
	LeftCeiling: LeftCeiling,
	LeftDoubleBracket: LeftDoubleBracket,
	LeftDownTeeVector: LeftDownTeeVector,
	LeftDownVectorBar: LeftDownVectorBar,
	LeftDownVector: LeftDownVector,
	LeftFloor: LeftFloor,
	leftharpoondown: leftharpoondown,
	leftharpoonup: leftharpoonup,
	leftleftarrows: leftleftarrows,
	leftrightarrow: leftrightarrow,
	LeftRightArrow: LeftRightArrow,
	Leftrightarrow: Leftrightarrow,
	leftrightarrows: leftrightarrows,
	leftrightharpoons: leftrightharpoons,
	leftrightsquigarrow: leftrightsquigarrow,
	LeftRightVector: LeftRightVector,
	LeftTeeArrow: LeftTeeArrow,
	LeftTee: LeftTee,
	LeftTeeVector: LeftTeeVector,
	leftthreetimes: leftthreetimes,
	LeftTriangleBar: LeftTriangleBar,
	LeftTriangle: LeftTriangle,
	LeftTriangleEqual: LeftTriangleEqual,
	LeftUpDownVector: LeftUpDownVector,
	LeftUpTeeVector: LeftUpTeeVector,
	LeftUpVectorBar: LeftUpVectorBar,
	LeftUpVector: LeftUpVector,
	LeftVectorBar: LeftVectorBar,
	LeftVector: LeftVector,
	lEg: lEg,
	leg: leg,
	leq: leq,
	leqq: leqq,
	leqslant: leqslant,
	lescc: lescc,
	les: les,
	lesdot: lesdot,
	lesdoto: lesdoto,
	lesdotor: lesdotor,
	lesg: lesg,
	lesges: lesges,
	lessapprox: lessapprox,
	lessdot: lessdot,
	lesseqgtr: lesseqgtr,
	lesseqqgtr: lesseqqgtr,
	LessEqualGreater: LessEqualGreater,
	LessFullEqual: LessFullEqual,
	LessGreater: LessGreater,
	lessgtr: lessgtr,
	LessLess: LessLess,
	lesssim: lesssim,
	LessSlantEqual: LessSlantEqual,
	LessTilde: LessTilde,
	lfisht: lfisht,
	lfloor: lfloor,
	Lfr: Lfr,
	lfr: lfr,
	lg: lg,
	lgE: lgE,
	lHar: lHar,
	lhard: lhard,
	lharu: lharu,
	lharul: lharul,
	lhblk: lhblk,
	LJcy: LJcy,
	ljcy: ljcy,
	llarr: llarr,
	ll: ll,
	Ll: Ll,
	llcorner: llcorner,
	Lleftarrow: Lleftarrow,
	llhard: llhard,
	lltri: lltri,
	Lmidot: Lmidot,
	lmidot: lmidot,
	lmoustache: lmoustache,
	lmoust: lmoust,
	lnap: lnap,
	lnapprox: lnapprox,
	lne: lne,
	lnE: lnE,
	lneq: lneq,
	lneqq: lneqq,
	lnsim: lnsim,
	loang: loang,
	loarr: loarr,
	lobrk: lobrk,
	longleftarrow: longleftarrow,
	LongLeftArrow: LongLeftArrow,
	Longleftarrow: Longleftarrow,
	longleftrightarrow: longleftrightarrow,
	LongLeftRightArrow: LongLeftRightArrow,
	Longleftrightarrow: Longleftrightarrow,
	longmapsto: longmapsto,
	longrightarrow: longrightarrow,
	LongRightArrow: LongRightArrow,
	Longrightarrow: Longrightarrow,
	looparrowleft: looparrowleft,
	looparrowright: looparrowright,
	lopar: lopar,
	Lopf: Lopf,
	lopf: lopf,
	loplus: loplus,
	lotimes: lotimes,
	lowast: lowast,
	lowbar: lowbar,
	LowerLeftArrow: LowerLeftArrow,
	LowerRightArrow: LowerRightArrow,
	loz: loz,
	lozenge: lozenge,
	lozf: lozf,
	lpar: lpar,
	lparlt: lparlt,
	lrarr: lrarr,
	lrcorner: lrcorner,
	lrhar: lrhar,
	lrhard: lrhard,
	lrm: lrm,
	lrtri: lrtri,
	lsaquo: lsaquo,
	lscr: lscr,
	Lscr: Lscr,
	lsh: lsh,
	Lsh: Lsh,
	lsim: lsim,
	lsime: lsime,
	lsimg: lsimg,
	lsqb: lsqb,
	lsquo: lsquo,
	lsquor: lsquor,
	Lstrok: Lstrok,
	lstrok: lstrok,
	ltcc: ltcc,
	ltcir: ltcir,
	lt: lt$1,
	LT: LT,
	Lt: Lt,
	ltdot: ltdot,
	lthree: lthree,
	ltimes: ltimes,
	ltlarr: ltlarr,
	ltquest: ltquest,
	ltri: ltri,
	ltrie: ltrie,
	ltrif: ltrif,
	ltrPar: ltrPar,
	lurdshar: lurdshar,
	luruhar: luruhar,
	lvertneqq: lvertneqq,
	lvnE: lvnE,
	macr: macr,
	male: male,
	malt: malt,
	maltese: maltese,
	map: map$1,
	mapsto: mapsto,
	mapstodown: mapstodown,
	mapstoleft: mapstoleft,
	mapstoup: mapstoup,
	marker: marker,
	mcomma: mcomma,
	Mcy: Mcy,
	mcy: mcy,
	mdash: mdash,
	mDDot: mDDot,
	measuredangle: measuredangle,
	MediumSpace: MediumSpace,
	Mellintrf: Mellintrf,
	Mfr: Mfr,
	mfr: mfr,
	mho: mho,
	micro: micro,
	midast: midast,
	midcir: midcir,
	mid: mid,
	middot: middot,
	minusb: minusb,
	minus: minus,
	minusd: minusd,
	minusdu: minusdu,
	MinusPlus: MinusPlus,
	mlcp: mlcp,
	mldr: mldr,
	mnplus: mnplus,
	models: models,
	Mopf: Mopf,
	mopf: mopf,
	mp: mp,
	mscr: mscr,
	Mscr: Mscr,
	mstpos: mstpos,
	Mu: Mu,
	mu: mu,
	multimap: multimap,
	mumap: mumap,
	nabla: nabla,
	Nacute: Nacute,
	nacute: nacute,
	nang: nang,
	nap: nap,
	napE: napE,
	napid: napid,
	napos: napos,
	napprox: napprox,
	natural: natural,
	naturals: naturals,
	natur: natur,
	nbsp: nbsp,
	nbump: nbump,
	nbumpe: nbumpe,
	ncap: ncap,
	Ncaron: Ncaron,
	ncaron: ncaron,
	Ncedil: Ncedil,
	ncedil: ncedil,
	ncong: ncong,
	ncongdot: ncongdot,
	ncup: ncup,
	Ncy: Ncy,
	ncy: ncy,
	ndash: ndash,
	nearhk: nearhk,
	nearr: nearr,
	neArr: neArr,
	nearrow: nearrow,
	ne: ne,
	nedot: nedot,
	NegativeMediumSpace: NegativeMediumSpace,
	NegativeThickSpace: NegativeThickSpace,
	NegativeThinSpace: NegativeThinSpace,
	NegativeVeryThinSpace: NegativeVeryThinSpace,
	nequiv: nequiv,
	nesear: nesear,
	nesim: nesim,
	NestedGreaterGreater: NestedGreaterGreater,
	NestedLessLess: NestedLessLess,
	NewLine: NewLine,
	nexist: nexist,
	nexists: nexists,
	Nfr: Nfr,
	nfr: nfr,
	ngE: ngE,
	nge: nge,
	ngeq: ngeq,
	ngeqq: ngeqq,
	ngeqslant: ngeqslant,
	nges: nges,
	nGg: nGg,
	ngsim: ngsim,
	nGt: nGt,
	ngt: ngt,
	ngtr: ngtr,
	nGtv: nGtv,
	nharr: nharr,
	nhArr: nhArr,
	nhpar: nhpar,
	ni: ni,
	nis: nis,
	nisd: nisd,
	niv: niv,
	NJcy: NJcy,
	njcy: njcy,
	nlarr: nlarr,
	nlArr: nlArr,
	nldr: nldr,
	nlE: nlE,
	nle: nle,
	nleftarrow: nleftarrow,
	nLeftarrow: nLeftarrow,
	nleftrightarrow: nleftrightarrow,
	nLeftrightarrow: nLeftrightarrow,
	nleq: nleq,
	nleqq: nleqq,
	nleqslant: nleqslant,
	nles: nles,
	nless: nless,
	nLl: nLl,
	nlsim: nlsim,
	nLt: nLt,
	nlt: nlt,
	nltri: nltri,
	nltrie: nltrie,
	nLtv: nLtv,
	nmid: nmid,
	NoBreak: NoBreak,
	NonBreakingSpace: NonBreakingSpace,
	nopf: nopf,
	Nopf: Nopf,
	Not: Not,
	not: not,
	NotCongruent: NotCongruent,
	NotCupCap: NotCupCap,
	NotDoubleVerticalBar: NotDoubleVerticalBar,
	NotElement: NotElement,
	NotEqual: NotEqual,
	NotEqualTilde: NotEqualTilde,
	NotExists: NotExists,
	NotGreater: NotGreater,
	NotGreaterEqual: NotGreaterEqual,
	NotGreaterFullEqual: NotGreaterFullEqual,
	NotGreaterGreater: NotGreaterGreater,
	NotGreaterLess: NotGreaterLess,
	NotGreaterSlantEqual: NotGreaterSlantEqual,
	NotGreaterTilde: NotGreaterTilde,
	NotHumpDownHump: NotHumpDownHump,
	NotHumpEqual: NotHumpEqual,
	notin: notin,
	notindot: notindot,
	notinE: notinE,
	notinva: notinva,
	notinvb: notinvb,
	notinvc: notinvc,
	NotLeftTriangleBar: NotLeftTriangleBar,
	NotLeftTriangle: NotLeftTriangle,
	NotLeftTriangleEqual: NotLeftTriangleEqual,
	NotLess: NotLess,
	NotLessEqual: NotLessEqual,
	NotLessGreater: NotLessGreater,
	NotLessLess: NotLessLess,
	NotLessSlantEqual: NotLessSlantEqual,
	NotLessTilde: NotLessTilde,
	NotNestedGreaterGreater: NotNestedGreaterGreater,
	NotNestedLessLess: NotNestedLessLess,
	notni: notni,
	notniva: notniva,
	notnivb: notnivb,
	notnivc: notnivc,
	NotPrecedes: NotPrecedes,
	NotPrecedesEqual: NotPrecedesEqual,
	NotPrecedesSlantEqual: NotPrecedesSlantEqual,
	NotReverseElement: NotReverseElement,
	NotRightTriangleBar: NotRightTriangleBar,
	NotRightTriangle: NotRightTriangle,
	NotRightTriangleEqual: NotRightTriangleEqual,
	NotSquareSubset: NotSquareSubset,
	NotSquareSubsetEqual: NotSquareSubsetEqual,
	NotSquareSuperset: NotSquareSuperset,
	NotSquareSupersetEqual: NotSquareSupersetEqual,
	NotSubset: NotSubset,
	NotSubsetEqual: NotSubsetEqual,
	NotSucceeds: NotSucceeds,
	NotSucceedsEqual: NotSucceedsEqual,
	NotSucceedsSlantEqual: NotSucceedsSlantEqual,
	NotSucceedsTilde: NotSucceedsTilde,
	NotSuperset: NotSuperset,
	NotSupersetEqual: NotSupersetEqual,
	NotTilde: NotTilde,
	NotTildeEqual: NotTildeEqual,
	NotTildeFullEqual: NotTildeFullEqual,
	NotTildeTilde: NotTildeTilde,
	NotVerticalBar: NotVerticalBar,
	nparallel: nparallel,
	npar: npar,
	nparsl: nparsl,
	npart: npart,
	npolint: npolint,
	npr: npr,
	nprcue: nprcue,
	nprec: nprec,
	npreceq: npreceq,
	npre: npre,
	nrarrc: nrarrc,
	nrarr: nrarr,
	nrArr: nrArr,
	nrarrw: nrarrw,
	nrightarrow: nrightarrow,
	nRightarrow: nRightarrow,
	nrtri: nrtri,
	nrtrie: nrtrie,
	nsc: nsc,
	nsccue: nsccue,
	nsce: nsce,
	Nscr: Nscr,
	nscr: nscr,
	nshortmid: nshortmid,
	nshortparallel: nshortparallel,
	nsim: nsim,
	nsime: nsime,
	nsimeq: nsimeq,
	nsmid: nsmid,
	nspar: nspar,
	nsqsube: nsqsube,
	nsqsupe: nsqsupe,
	nsub: nsub,
	nsubE: nsubE,
	nsube: nsube,
	nsubset: nsubset,
	nsubseteq: nsubseteq,
	nsubseteqq: nsubseteqq,
	nsucc: nsucc,
	nsucceq: nsucceq,
	nsup: nsup,
	nsupE: nsupE,
	nsupe: nsupe,
	nsupset: nsupset,
	nsupseteq: nsupseteq,
	nsupseteqq: nsupseteqq,
	ntgl: ntgl,
	Ntilde: Ntilde,
	ntilde: ntilde,
	ntlg: ntlg,
	ntriangleleft: ntriangleleft,
	ntrianglelefteq: ntrianglelefteq,
	ntriangleright: ntriangleright,
	ntrianglerighteq: ntrianglerighteq,
	Nu: Nu,
	nu: nu,
	num: num,
	numero: numero,
	numsp: numsp,
	nvap: nvap,
	nvdash: nvdash,
	nvDash: nvDash,
	nVdash: nVdash,
	nVDash: nVDash,
	nvge: nvge,
	nvgt: nvgt,
	nvHarr: nvHarr,
	nvinfin: nvinfin,
	nvlArr: nvlArr,
	nvle: nvle,
	nvlt: nvlt,
	nvltrie: nvltrie,
	nvrArr: nvrArr,
	nvrtrie: nvrtrie,
	nvsim: nvsim,
	nwarhk: nwarhk,
	nwarr: nwarr,
	nwArr: nwArr,
	nwarrow: nwarrow,
	nwnear: nwnear,
	Oacute: Oacute,
	oacute: oacute,
	oast: oast,
	Ocirc: Ocirc,
	ocirc: ocirc,
	ocir: ocir,
	Ocy: Ocy,
	ocy: ocy,
	odash: odash,
	Odblac: Odblac,
	odblac: odblac,
	odiv: odiv,
	odot: odot,
	odsold: odsold,
	OElig: OElig,
	oelig: oelig,
	ofcir: ofcir,
	Ofr: Ofr,
	ofr: ofr,
	ogon: ogon,
	Ograve: Ograve,
	ograve: ograve,
	ogt: ogt,
	ohbar: ohbar,
	ohm: ohm,
	oint: oint,
	olarr: olarr,
	olcir: olcir,
	olcross: olcross,
	oline: oline,
	olt: olt,
	Omacr: Omacr,
	omacr: omacr,
	Omega: Omega,
	omega: omega,
	Omicron: Omicron,
	omicron: omicron,
	omid: omid,
	ominus: ominus,
	Oopf: Oopf,
	oopf: oopf,
	opar: opar,
	OpenCurlyDoubleQuote: OpenCurlyDoubleQuote,
	OpenCurlyQuote: OpenCurlyQuote,
	operp: operp,
	oplus: oplus,
	orarr: orarr,
	Or: Or,
	or: or,
	ord: ord,
	order: order,
	orderof: orderof,
	ordf: ordf,
	ordm: ordm,
	origof: origof,
	oror: oror,
	orslope: orslope,
	orv: orv,
	oS: oS,
	Oscr: Oscr,
	oscr: oscr,
	Oslash: Oslash,
	oslash: oslash,
	osol: osol,
	Otilde: Otilde,
	otilde: otilde,
	otimesas: otimesas,
	Otimes: Otimes,
	otimes: otimes,
	Ouml: Ouml,
	ouml: ouml,
	ovbar: ovbar,
	OverBar: OverBar,
	OverBrace: OverBrace,
	OverBracket: OverBracket,
	OverParenthesis: OverParenthesis,
	para: para,
	parallel: parallel,
	par: par,
	parsim: parsim,
	parsl: parsl,
	part: part,
	PartialD: PartialD,
	Pcy: Pcy,
	pcy: pcy,
	percnt: percnt,
	period: period,
	permil: permil,
	perp: perp,
	pertenk: pertenk,
	Pfr: Pfr,
	pfr: pfr,
	Phi: Phi,
	phi: phi,
	phiv: phiv,
	phmmat: phmmat,
	phone: phone,
	Pi: Pi,
	pi: pi,
	pitchfork: pitchfork,
	piv: piv,
	planck: planck,
	planckh: planckh,
	plankv: plankv,
	plusacir: plusacir,
	plusb: plusb,
	pluscir: pluscir,
	plus: plus,
	plusdo: plusdo,
	plusdu: plusdu,
	pluse: pluse,
	PlusMinus: PlusMinus,
	plusmn: plusmn,
	plussim: plussim,
	plustwo: plustwo,
	pm: pm,
	Poincareplane: Poincareplane,
	pointint: pointint,
	popf: popf,
	Popf: Popf,
	pound: pound,
	prap: prap,
	Pr: Pr,
	pr: pr,
	prcue: prcue,
	precapprox: precapprox,
	prec: prec,
	preccurlyeq: preccurlyeq,
	Precedes: Precedes,
	PrecedesEqual: PrecedesEqual,
	PrecedesSlantEqual: PrecedesSlantEqual,
	PrecedesTilde: PrecedesTilde,
	preceq: preceq,
	precnapprox: precnapprox,
	precneqq: precneqq,
	precnsim: precnsim,
	pre: pre,
	prE: prE,
	precsim: precsim,
	prime: prime,
	Prime: Prime,
	primes: primes,
	prnap: prnap,
	prnE: prnE,
	prnsim: prnsim,
	prod: prod,
	Product: Product,
	profalar: profalar,
	profline: profline,
	profsurf: profsurf,
	prop: prop,
	Proportional: Proportional,
	Proportion: Proportion,
	propto: propto,
	prsim: prsim,
	prurel: prurel,
	Pscr: Pscr,
	pscr: pscr,
	Psi: Psi,
	psi: psi,
	puncsp: puncsp,
	Qfr: Qfr,
	qfr: qfr,
	qint: qint,
	qopf: qopf,
	Qopf: Qopf,
	qprime: qprime,
	Qscr: Qscr,
	qscr: qscr,
	quaternions: quaternions,
	quatint: quatint,
	quest: quest,
	questeq: questeq,
	quot: quot$1,
	QUOT: QUOT,
	rAarr: rAarr,
	race: race,
	Racute: Racute,
	racute: racute,
	radic: radic,
	raemptyv: raemptyv,
	rang: rang,
	Rang: Rang,
	rangd: rangd,
	range: range,
	rangle: rangle,
	raquo: raquo,
	rarrap: rarrap,
	rarrb: rarrb,
	rarrbfs: rarrbfs,
	rarrc: rarrc,
	rarr: rarr,
	Rarr: Rarr,
	rArr: rArr,
	rarrfs: rarrfs,
	rarrhk: rarrhk,
	rarrlp: rarrlp,
	rarrpl: rarrpl,
	rarrsim: rarrsim,
	Rarrtl: Rarrtl,
	rarrtl: rarrtl,
	rarrw: rarrw,
	ratail: ratail,
	rAtail: rAtail,
	ratio: ratio,
	rationals: rationals,
	rbarr: rbarr,
	rBarr: rBarr,
	RBarr: RBarr,
	rbbrk: rbbrk,
	rbrace: rbrace,
	rbrack: rbrack,
	rbrke: rbrke,
	rbrksld: rbrksld,
	rbrkslu: rbrkslu,
	Rcaron: Rcaron,
	rcaron: rcaron,
	Rcedil: Rcedil,
	rcedil: rcedil,
	rceil: rceil,
	rcub: rcub,
	Rcy: Rcy,
	rcy: rcy,
	rdca: rdca,
	rdldhar: rdldhar,
	rdquo: rdquo,
	rdquor: rdquor,
	rdsh: rdsh,
	real: real,
	realine: realine,
	realpart: realpart,
	reals: reals,
	Re: Re,
	rect: rect,
	reg: reg,
	REG: REG,
	ReverseElement: ReverseElement,
	ReverseEquilibrium: ReverseEquilibrium,
	ReverseUpEquilibrium: ReverseUpEquilibrium,
	rfisht: rfisht,
	rfloor: rfloor,
	rfr: rfr,
	Rfr: Rfr,
	rHar: rHar,
	rhard: rhard,
	rharu: rharu,
	rharul: rharul,
	Rho: Rho,
	rho: rho,
	rhov: rhov,
	RightAngleBracket: RightAngleBracket,
	RightArrowBar: RightArrowBar,
	rightarrow: rightarrow,
	RightArrow: RightArrow,
	Rightarrow: Rightarrow,
	RightArrowLeftArrow: RightArrowLeftArrow,
	rightarrowtail: rightarrowtail,
	RightCeiling: RightCeiling,
	RightDoubleBracket: RightDoubleBracket,
	RightDownTeeVector: RightDownTeeVector,
	RightDownVectorBar: RightDownVectorBar,
	RightDownVector: RightDownVector,
	RightFloor: RightFloor,
	rightharpoondown: rightharpoondown,
	rightharpoonup: rightharpoonup,
	rightleftarrows: rightleftarrows,
	rightleftharpoons: rightleftharpoons,
	rightrightarrows: rightrightarrows,
	rightsquigarrow: rightsquigarrow,
	RightTeeArrow: RightTeeArrow,
	RightTee: RightTee,
	RightTeeVector: RightTeeVector,
	rightthreetimes: rightthreetimes,
	RightTriangleBar: RightTriangleBar,
	RightTriangle: RightTriangle,
	RightTriangleEqual: RightTriangleEqual,
	RightUpDownVector: RightUpDownVector,
	RightUpTeeVector: RightUpTeeVector,
	RightUpVectorBar: RightUpVectorBar,
	RightUpVector: RightUpVector,
	RightVectorBar: RightVectorBar,
	RightVector: RightVector,
	ring: ring,
	risingdotseq: risingdotseq,
	rlarr: rlarr,
	rlhar: rlhar,
	rlm: rlm,
	rmoustache: rmoustache,
	rmoust: rmoust,
	rnmid: rnmid,
	roang: roang,
	roarr: roarr,
	robrk: robrk,
	ropar: ropar,
	ropf: ropf,
	Ropf: Ropf,
	roplus: roplus,
	rotimes: rotimes,
	RoundImplies: RoundImplies,
	rpar: rpar,
	rpargt: rpargt,
	rppolint: rppolint,
	rrarr: rrarr,
	Rrightarrow: Rrightarrow,
	rsaquo: rsaquo,
	rscr: rscr,
	Rscr: Rscr,
	rsh: rsh,
	Rsh: Rsh,
	rsqb: rsqb,
	rsquo: rsquo,
	rsquor: rsquor,
	rthree: rthree,
	rtimes: rtimes,
	rtri: rtri,
	rtrie: rtrie,
	rtrif: rtrif,
	rtriltri: rtriltri,
	RuleDelayed: RuleDelayed,
	ruluhar: ruluhar,
	rx: rx,
	Sacute: Sacute,
	sacute: sacute,
	sbquo: sbquo,
	scap: scap,
	Scaron: Scaron,
	scaron: scaron,
	Sc: Sc,
	sc: sc,
	sccue: sccue,
	sce: sce,
	scE: scE,
	Scedil: Scedil,
	scedil: scedil,
	Scirc: Scirc,
	scirc: scirc,
	scnap: scnap,
	scnE: scnE,
	scnsim: scnsim,
	scpolint: scpolint,
	scsim: scsim,
	Scy: Scy,
	scy: scy,
	sdotb: sdotb,
	sdot: sdot,
	sdote: sdote,
	searhk: searhk,
	searr: searr,
	seArr: seArr,
	searrow: searrow,
	sect: sect,
	semi: semi,
	seswar: seswar,
	setminus: setminus,
	setmn: setmn,
	sext: sext,
	Sfr: Sfr,
	sfr: sfr,
	sfrown: sfrown,
	sharp: sharp,
	SHCHcy: SHCHcy,
	shchcy: shchcy,
	SHcy: SHcy,
	shcy: shcy,
	ShortDownArrow: ShortDownArrow,
	ShortLeftArrow: ShortLeftArrow,
	shortmid: shortmid,
	shortparallel: shortparallel,
	ShortRightArrow: ShortRightArrow,
	ShortUpArrow: ShortUpArrow,
	shy: shy,
	Sigma: Sigma,
	sigma: sigma,
	sigmaf: sigmaf,
	sigmav: sigmav,
	sim: sim,
	simdot: simdot,
	sime: sime,
	simeq: simeq,
	simg: simg,
	simgE: simgE,
	siml: siml,
	simlE: simlE,
	simne: simne,
	simplus: simplus,
	simrarr: simrarr,
	slarr: slarr,
	SmallCircle: SmallCircle,
	smallsetminus: smallsetminus,
	smashp: smashp,
	smeparsl: smeparsl,
	smid: smid,
	smile: smile,
	smt: smt,
	smte: smte,
	smtes: smtes,
	SOFTcy: SOFTcy,
	softcy: softcy,
	solbar: solbar,
	solb: solb,
	sol: sol,
	Sopf: Sopf,
	sopf: sopf,
	spades: spades,
	spadesuit: spadesuit,
	spar: spar,
	sqcap: sqcap,
	sqcaps: sqcaps,
	sqcup: sqcup,
	sqcups: sqcups,
	Sqrt: Sqrt,
	sqsub: sqsub,
	sqsube: sqsube,
	sqsubset: sqsubset,
	sqsubseteq: sqsubseteq,
	sqsup: sqsup,
	sqsupe: sqsupe,
	sqsupset: sqsupset,
	sqsupseteq: sqsupseteq,
	square: square,
	Square: Square,
	SquareIntersection: SquareIntersection,
	SquareSubset: SquareSubset,
	SquareSubsetEqual: SquareSubsetEqual,
	SquareSuperset: SquareSuperset,
	SquareSupersetEqual: SquareSupersetEqual,
	SquareUnion: SquareUnion,
	squarf: squarf,
	squ: squ,
	squf: squf,
	srarr: srarr,
	Sscr: Sscr,
	sscr: sscr,
	ssetmn: ssetmn,
	ssmile: ssmile,
	sstarf: sstarf,
	Star: Star,
	star: star,
	starf: starf,
	straightepsilon: straightepsilon,
	straightphi: straightphi,
	strns: strns,
	sub: sub,
	Sub: Sub,
	subdot: subdot,
	subE: subE,
	sube: sube,
	subedot: subedot,
	submult: submult,
	subnE: subnE,
	subne: subne,
	subplus: subplus,
	subrarr: subrarr,
	subset: subset,
	Subset: Subset,
	subseteq: subseteq,
	subseteqq: subseteqq,
	SubsetEqual: SubsetEqual,
	subsetneq: subsetneq,
	subsetneqq: subsetneqq,
	subsim: subsim,
	subsub: subsub,
	subsup: subsup,
	succapprox: succapprox,
	succ: succ,
	succcurlyeq: succcurlyeq,
	Succeeds: Succeeds,
	SucceedsEqual: SucceedsEqual,
	SucceedsSlantEqual: SucceedsSlantEqual,
	SucceedsTilde: SucceedsTilde,
	succeq: succeq,
	succnapprox: succnapprox,
	succneqq: succneqq,
	succnsim: succnsim,
	succsim: succsim,
	SuchThat: SuchThat,
	sum: sum,
	Sum: Sum,
	sung: sung,
	sup1: sup1,
	sup2: sup2,
	sup3: sup3,
	sup: sup,
	Sup: Sup,
	supdot: supdot,
	supdsub: supdsub,
	supE: supE,
	supe: supe,
	supedot: supedot,
	Superset: Superset,
	SupersetEqual: SupersetEqual,
	suphsol: suphsol,
	suphsub: suphsub,
	suplarr: suplarr,
	supmult: supmult,
	supnE: supnE,
	supne: supne,
	supplus: supplus,
	supset: supset,
	Supset: Supset,
	supseteq: supseteq,
	supseteqq: supseteqq,
	supsetneq: supsetneq,
	supsetneqq: supsetneqq,
	supsim: supsim,
	supsub: supsub,
	supsup: supsup,
	swarhk: swarhk,
	swarr: swarr,
	swArr: swArr,
	swarrow: swarrow,
	swnwar: swnwar,
	szlig: szlig,
	Tab: Tab,
	target: target,
	Tau: Tau,
	tau: tau,
	tbrk: tbrk,
	Tcaron: Tcaron,
	tcaron: tcaron,
	Tcedil: Tcedil,
	tcedil: tcedil,
	Tcy: Tcy,
	tcy: tcy,
	tdot: tdot,
	telrec: telrec,
	Tfr: Tfr,
	tfr: tfr,
	there4: there4,
	therefore: therefore,
	Therefore: Therefore,
	Theta: Theta,
	theta: theta,
	thetasym: thetasym,
	thetav: thetav,
	thickapprox: thickapprox,
	thicksim: thicksim,
	ThickSpace: ThickSpace,
	ThinSpace: ThinSpace,
	thinsp: thinsp,
	thkap: thkap,
	thksim: thksim,
	THORN: THORN,
	thorn: thorn,
	tilde: tilde,
	Tilde: Tilde,
	TildeEqual: TildeEqual,
	TildeFullEqual: TildeFullEqual,
	TildeTilde: TildeTilde,
	timesbar: timesbar,
	timesb: timesb,
	times: times$1,
	timesd: timesd,
	tint: tint,
	toea: toea,
	topbot: topbot,
	topcir: topcir,
	top: top,
	Topf: Topf,
	topf: topf,
	topfork: topfork,
	tosa: tosa,
	tprime: tprime,
	trade: trade,
	TRADE: TRADE,
	triangle: triangle,
	triangledown: triangledown,
	triangleleft: triangleleft,
	trianglelefteq: trianglelefteq,
	triangleq: triangleq,
	triangleright: triangleright,
	trianglerighteq: trianglerighteq,
	tridot: tridot,
	trie: trie,
	triminus: triminus,
	TripleDot: TripleDot,
	triplus: triplus,
	trisb: trisb,
	tritime: tritime,
	trpezium: trpezium,
	Tscr: Tscr,
	tscr: tscr,
	TScy: TScy,
	tscy: tscy,
	TSHcy: TSHcy,
	tshcy: tshcy,
	Tstrok: Tstrok,
	tstrok: tstrok,
	twixt: twixt,
	twoheadleftarrow: twoheadleftarrow,
	twoheadrightarrow: twoheadrightarrow,
	Uacute: Uacute,
	uacute: uacute,
	uarr: uarr,
	Uarr: Uarr,
	uArr: uArr,
	Uarrocir: Uarrocir,
	Ubrcy: Ubrcy,
	ubrcy: ubrcy,
	Ubreve: Ubreve,
	ubreve: ubreve,
	Ucirc: Ucirc,
	ucirc: ucirc,
	Ucy: Ucy,
	ucy: ucy,
	udarr: udarr,
	Udblac: Udblac,
	udblac: udblac,
	udhar: udhar,
	ufisht: ufisht,
	Ufr: Ufr,
	ufr: ufr,
	Ugrave: Ugrave,
	ugrave: ugrave,
	uHar: uHar,
	uharl: uharl,
	uharr: uharr,
	uhblk: uhblk,
	ulcorn: ulcorn,
	ulcorner: ulcorner,
	ulcrop: ulcrop,
	ultri: ultri,
	Umacr: Umacr,
	umacr: umacr,
	uml: uml,
	UnderBar: UnderBar,
	UnderBrace: UnderBrace,
	UnderBracket: UnderBracket,
	UnderParenthesis: UnderParenthesis,
	Union: Union,
	UnionPlus: UnionPlus,
	Uogon: Uogon,
	uogon: uogon,
	Uopf: Uopf,
	uopf: uopf,
	UpArrowBar: UpArrowBar,
	uparrow: uparrow,
	UpArrow: UpArrow,
	Uparrow: Uparrow,
	UpArrowDownArrow: UpArrowDownArrow,
	updownarrow: updownarrow,
	UpDownArrow: UpDownArrow,
	Updownarrow: Updownarrow,
	UpEquilibrium: UpEquilibrium,
	upharpoonleft: upharpoonleft,
	upharpoonright: upharpoonright,
	uplus: uplus,
	UpperLeftArrow: UpperLeftArrow,
	UpperRightArrow: UpperRightArrow,
	upsi: upsi,
	Upsi: Upsi,
	upsih: upsih,
	Upsilon: Upsilon,
	upsilon: upsilon,
	UpTeeArrow: UpTeeArrow,
	UpTee: UpTee,
	upuparrows: upuparrows,
	urcorn: urcorn,
	urcorner: urcorner,
	urcrop: urcrop,
	Uring: Uring,
	uring: uring,
	urtri: urtri,
	Uscr: Uscr,
	uscr: uscr,
	utdot: utdot,
	Utilde: Utilde,
	utilde: utilde,
	utri: utri,
	utrif: utrif,
	uuarr: uuarr,
	Uuml: Uuml,
	uuml: uuml,
	uwangle: uwangle,
	vangrt: vangrt,
	varepsilon: varepsilon,
	varkappa: varkappa,
	varnothing: varnothing,
	varphi: varphi,
	varpi: varpi,
	varpropto: varpropto,
	varr: varr,
	vArr: vArr,
	varrho: varrho,
	varsigma: varsigma,
	varsubsetneq: varsubsetneq,
	varsubsetneqq: varsubsetneqq,
	varsupsetneq: varsupsetneq,
	varsupsetneqq: varsupsetneqq,
	vartheta: vartheta,
	vartriangleleft: vartriangleleft,
	vartriangleright: vartriangleright,
	vBar: vBar,
	Vbar: Vbar,
	vBarv: vBarv,
	Vcy: Vcy,
	vcy: vcy,
	vdash: vdash,
	vDash: vDash,
	Vdash: Vdash,
	VDash: VDash,
	Vdashl: Vdashl,
	veebar: veebar,
	vee: vee,
	Vee: Vee,
	veeeq: veeeq,
	vellip: vellip,
	verbar: verbar,
	Verbar: Verbar,
	vert: vert,
	Vert: Vert,
	VerticalBar: VerticalBar,
	VerticalLine: VerticalLine,
	VerticalSeparator: VerticalSeparator,
	VerticalTilde: VerticalTilde,
	VeryThinSpace: VeryThinSpace,
	Vfr: Vfr,
	vfr: vfr,
	vltri: vltri,
	vnsub: vnsub,
	vnsup: vnsup,
	Vopf: Vopf,
	vopf: vopf,
	vprop: vprop,
	vrtri: vrtri,
	Vscr: Vscr,
	vscr: vscr,
	vsubnE: vsubnE,
	vsubne: vsubne,
	vsupnE: vsupnE,
	vsupne: vsupne,
	Vvdash: Vvdash,
	vzigzag: vzigzag,
	Wcirc: Wcirc,
	wcirc: wcirc,
	wedbar: wedbar,
	wedge: wedge,
	Wedge: Wedge,
	wedgeq: wedgeq,
	weierp: weierp,
	Wfr: Wfr,
	wfr: wfr,
	Wopf: Wopf,
	wopf: wopf,
	wp: wp,
	wr: wr,
	wreath: wreath,
	Wscr: Wscr,
	wscr: wscr,
	xcap: xcap,
	xcirc: xcirc,
	xcup: xcup,
	xdtri: xdtri,
	Xfr: Xfr,
	xfr: xfr,
	xharr: xharr,
	xhArr: xhArr,
	Xi: Xi,
	xi: xi,
	xlarr: xlarr,
	xlArr: xlArr,
	xmap: xmap,
	xnis: xnis,
	xodot: xodot,
	Xopf: Xopf,
	xopf: xopf,
	xoplus: xoplus,
	xotime: xotime,
	xrarr: xrarr,
	xrArr: xrArr,
	Xscr: Xscr,
	xscr: xscr,
	xsqcup: xsqcup,
	xuplus: xuplus,
	xutri: xutri,
	xvee: xvee,
	xwedge: xwedge,
	Yacute: Yacute,
	yacute: yacute,
	YAcy: YAcy,
	yacy: yacy,
	Ycirc: Ycirc,
	ycirc: ycirc,
	Ycy: Ycy,
	ycy: ycy,
	yen: yen,
	Yfr: Yfr,
	yfr: yfr,
	YIcy: YIcy,
	yicy: yicy,
	Yopf: Yopf,
	yopf: yopf,
	Yscr: Yscr,
	yscr: yscr,
	YUcy: YUcy,
	yucy: yucy,
	yuml: yuml,
	Yuml: Yuml,
	Zacute: Zacute,
	zacute: zacute,
	Zcaron: Zcaron,
	zcaron: zcaron,
	Zcy: Zcy,
	zcy: zcy,
	Zdot: Zdot,
	zdot: zdot,
	zeetrf: zeetrf,
	ZeroWidthSpace: ZeroWidthSpace,
	Zeta: Zeta,
	zeta: zeta,
	zfr: zfr,
	Zfr: Zfr,
	ZHcy: ZHcy,
	zhcy: zhcy,
	zigrarr: zigrarr,
	zopf: zopf,
	Zopf: Zopf,
	Zscr: Zscr,
	zscr: zscr,
	zwj: zwj,
	zwnj: zwnj,
	"in": "∈",
	"Map": "⤅"
};

var entities = Object.freeze({
	Aacute: Aacute,
	aacute: aacute,
	Abreve: Abreve,
	abreve: abreve,
	ac: ac,
	acd: acd,
	acE: acE,
	Acirc: Acirc,
	acirc: acirc,
	acute: acute,
	Acy: Acy,
	acy: acy,
	AElig: AElig,
	aelig: aelig,
	af: af,
	Afr: Afr,
	afr: afr,
	Agrave: Agrave,
	agrave: agrave,
	alefsym: alefsym,
	aleph: aleph,
	Alpha: Alpha,
	alpha: alpha,
	Amacr: Amacr,
	amacr: amacr,
	amalg: amalg,
	amp: amp$1,
	AMP: AMP,
	andand: andand,
	And: And,
	and: and,
	andd: andd,
	andslope: andslope,
	andv: andv,
	ang: ang,
	ange: ange,
	angle: angle,
	angmsdaa: angmsdaa,
	angmsdab: angmsdab,
	angmsdac: angmsdac,
	angmsdad: angmsdad,
	angmsdae: angmsdae,
	angmsdaf: angmsdaf,
	angmsdag: angmsdag,
	angmsdah: angmsdah,
	angmsd: angmsd,
	angrt: angrt,
	angrtvb: angrtvb,
	angrtvbd: angrtvbd,
	angsph: angsph,
	angst: angst,
	angzarr: angzarr,
	Aogon: Aogon,
	aogon: aogon,
	Aopf: Aopf,
	aopf: aopf,
	apacir: apacir,
	ap: ap,
	apE: apE,
	ape: ape,
	apid: apid,
	apos: apos$1,
	ApplyFunction: ApplyFunction,
	approx: approx,
	approxeq: approxeq,
	Aring: Aring,
	aring: aring,
	Ascr: Ascr,
	ascr: ascr,
	Assign: Assign,
	ast: ast,
	asymp: asymp,
	asympeq: asympeq,
	Atilde: Atilde,
	atilde: atilde,
	Auml: Auml,
	auml: auml,
	awconint: awconint,
	awint: awint,
	backcong: backcong,
	backepsilon: backepsilon,
	backprime: backprime,
	backsim: backsim,
	backsimeq: backsimeq,
	Backslash: Backslash,
	Barv: Barv,
	barvee: barvee,
	barwed: barwed,
	Barwed: Barwed,
	barwedge: barwedge,
	bbrk: bbrk,
	bbrktbrk: bbrktbrk,
	bcong: bcong,
	Bcy: Bcy,
	bcy: bcy,
	bdquo: bdquo,
	becaus: becaus,
	because: because,
	Because: Because,
	bemptyv: bemptyv,
	bepsi: bepsi,
	bernou: bernou,
	Bernoullis: Bernoullis,
	Beta: Beta,
	beta: beta,
	beth: beth,
	between: between,
	Bfr: Bfr,
	bfr: bfr,
	bigcap: bigcap,
	bigcirc: bigcirc,
	bigcup: bigcup,
	bigodot: bigodot,
	bigoplus: bigoplus,
	bigotimes: bigotimes,
	bigsqcup: bigsqcup,
	bigstar: bigstar,
	bigtriangledown: bigtriangledown,
	bigtriangleup: bigtriangleup,
	biguplus: biguplus,
	bigvee: bigvee,
	bigwedge: bigwedge,
	bkarow: bkarow,
	blacklozenge: blacklozenge,
	blacksquare: blacksquare,
	blacktriangle: blacktriangle,
	blacktriangledown: blacktriangledown,
	blacktriangleleft: blacktriangleleft,
	blacktriangleright: blacktriangleright,
	blank: blank,
	blk12: blk12,
	blk14: blk14,
	blk34: blk34,
	block: block,
	bne: bne,
	bnequiv: bnequiv,
	bNot: bNot,
	bnot: bnot,
	Bopf: Bopf,
	bopf: bopf,
	bot: bot,
	bottom: bottom,
	bowtie: bowtie,
	boxbox: boxbox,
	boxdl: boxdl,
	boxdL: boxdL,
	boxDl: boxDl,
	boxDL: boxDL,
	boxdr: boxdr,
	boxdR: boxdR,
	boxDr: boxDr,
	boxDR: boxDR,
	boxh: boxh,
	boxH: boxH,
	boxhd: boxhd,
	boxHd: boxHd,
	boxhD: boxhD,
	boxHD: boxHD,
	boxhu: boxhu,
	boxHu: boxHu,
	boxhU: boxhU,
	boxHU: boxHU,
	boxminus: boxminus,
	boxplus: boxplus,
	boxtimes: boxtimes,
	boxul: boxul,
	boxuL: boxuL,
	boxUl: boxUl,
	boxUL: boxUL,
	boxur: boxur,
	boxuR: boxuR,
	boxUr: boxUr,
	boxUR: boxUR,
	boxv: boxv,
	boxV: boxV,
	boxvh: boxvh,
	boxvH: boxvH,
	boxVh: boxVh,
	boxVH: boxVH,
	boxvl: boxvl,
	boxvL: boxvL,
	boxVl: boxVl,
	boxVL: boxVL,
	boxvr: boxvr,
	boxvR: boxvR,
	boxVr: boxVr,
	boxVR: boxVR,
	bprime: bprime,
	breve: breve,
	Breve: Breve,
	brvbar: brvbar,
	bscr: bscr,
	Bscr: Bscr,
	bsemi: bsemi,
	bsim: bsim,
	bsime: bsime,
	bsolb: bsolb,
	bsol: bsol,
	bsolhsub: bsolhsub,
	bull: bull,
	bullet: bullet,
	bump: bump,
	bumpE: bumpE,
	bumpe: bumpe,
	Bumpeq: Bumpeq,
	bumpeq: bumpeq,
	Cacute: Cacute,
	cacute: cacute,
	capand: capand,
	capbrcup: capbrcup,
	capcap: capcap,
	cap: cap,
	Cap: Cap,
	capcup: capcup,
	capdot: capdot,
	CapitalDifferentialD: CapitalDifferentialD,
	caps: caps,
	caret: caret,
	caron: caron,
	Cayleys: Cayleys,
	ccaps: ccaps,
	Ccaron: Ccaron,
	ccaron: ccaron,
	Ccedil: Ccedil,
	ccedil: ccedil,
	Ccirc: Ccirc,
	ccirc: ccirc,
	Cconint: Cconint,
	ccups: ccups,
	ccupssm: ccupssm,
	Cdot: Cdot,
	cdot: cdot,
	cedil: cedil,
	Cedilla: Cedilla,
	cemptyv: cemptyv,
	cent: cent,
	centerdot: centerdot,
	CenterDot: CenterDot,
	cfr: cfr,
	Cfr: Cfr,
	CHcy: CHcy,
	chcy: chcy,
	check: check,
	checkmark: checkmark,
	Chi: Chi,
	chi: chi,
	circ: circ,
	circeq: circeq,
	circlearrowleft: circlearrowleft,
	circlearrowright: circlearrowright,
	circledast: circledast,
	circledcirc: circledcirc,
	circleddash: circleddash,
	CircleDot: CircleDot,
	circledR: circledR,
	circledS: circledS,
	CircleMinus: CircleMinus,
	CirclePlus: CirclePlus,
	CircleTimes: CircleTimes,
	cir: cir,
	cirE: cirE,
	cire: cire,
	cirfnint: cirfnint,
	cirmid: cirmid,
	cirscir: cirscir,
	ClockwiseContourIntegral: ClockwiseContourIntegral,
	CloseCurlyDoubleQuote: CloseCurlyDoubleQuote,
	CloseCurlyQuote: CloseCurlyQuote,
	clubs: clubs,
	clubsuit: clubsuit,
	colon: colon,
	Colon: Colon,
	Colone: Colone,
	colone: colone,
	coloneq: coloneq,
	comma: comma,
	commat: commat,
	comp: comp,
	compfn: compfn,
	complement: complement,
	complexes: complexes,
	cong: cong,
	congdot: congdot,
	Congruent: Congruent,
	conint: conint,
	Conint: Conint,
	ContourIntegral: ContourIntegral,
	copf: copf,
	Copf: Copf,
	coprod: coprod,
	Coproduct: Coproduct,
	copy: copy,
	COPY: COPY,
	copysr: copysr,
	CounterClockwiseContourIntegral: CounterClockwiseContourIntegral,
	crarr: crarr,
	cross: cross,
	Cross: Cross,
	Cscr: Cscr,
	cscr: cscr,
	csub: csub,
	csube: csube,
	csup: csup,
	csupe: csupe,
	ctdot: ctdot,
	cudarrl: cudarrl,
	cudarrr: cudarrr,
	cuepr: cuepr,
	cuesc: cuesc,
	cularr: cularr,
	cularrp: cularrp,
	cupbrcap: cupbrcap,
	cupcap: cupcap,
	CupCap: CupCap,
	cup: cup,
	Cup: Cup,
	cupcup: cupcup,
	cupdot: cupdot,
	cupor: cupor,
	cups: cups,
	curarr: curarr,
	curarrm: curarrm,
	curlyeqprec: curlyeqprec,
	curlyeqsucc: curlyeqsucc,
	curlyvee: curlyvee,
	curlywedge: curlywedge,
	curren: curren,
	curvearrowleft: curvearrowleft,
	curvearrowright: curvearrowright,
	cuvee: cuvee,
	cuwed: cuwed,
	cwconint: cwconint,
	cwint: cwint,
	cylcty: cylcty,
	dagger: dagger,
	Dagger: Dagger,
	daleth: daleth,
	darr: darr,
	Darr: Darr,
	dArr: dArr,
	dash: dash,
	Dashv: Dashv,
	dashv: dashv,
	dbkarow: dbkarow,
	dblac: dblac,
	Dcaron: Dcaron,
	dcaron: dcaron,
	Dcy: Dcy,
	dcy: dcy,
	ddagger: ddagger,
	ddarr: ddarr,
	DD: DD,
	dd: dd,
	DDotrahd: DDotrahd,
	ddotseq: ddotseq,
	deg: deg,
	Del: Del,
	Delta: Delta,
	delta: delta,
	demptyv: demptyv,
	dfisht: dfisht,
	Dfr: Dfr,
	dfr: dfr,
	dHar: dHar,
	dharl: dharl,
	dharr: dharr,
	DiacriticalAcute: DiacriticalAcute,
	DiacriticalDot: DiacriticalDot,
	DiacriticalDoubleAcute: DiacriticalDoubleAcute,
	DiacriticalGrave: DiacriticalGrave,
	DiacriticalTilde: DiacriticalTilde,
	diam: diam,
	diamond: diamond,
	Diamond: Diamond,
	diamondsuit: diamondsuit,
	diams: diams,
	die: die,
	DifferentialD: DifferentialD,
	digamma: digamma,
	disin: disin,
	div: div,
	divide: divide,
	divideontimes: divideontimes,
	divonx: divonx,
	DJcy: DJcy,
	djcy: djcy,
	dlcorn: dlcorn,
	dlcrop: dlcrop,
	dollar: dollar,
	Dopf: Dopf,
	dopf: dopf,
	Dot: Dot,
	dot: dot,
	DotDot: DotDot,
	doteq: doteq,
	doteqdot: doteqdot,
	DotEqual: DotEqual,
	dotminus: dotminus,
	dotplus: dotplus,
	dotsquare: dotsquare,
	doublebarwedge: doublebarwedge,
	DoubleContourIntegral: DoubleContourIntegral,
	DoubleDot: DoubleDot,
	DoubleDownArrow: DoubleDownArrow,
	DoubleLeftArrow: DoubleLeftArrow,
	DoubleLeftRightArrow: DoubleLeftRightArrow,
	DoubleLeftTee: DoubleLeftTee,
	DoubleLongLeftArrow: DoubleLongLeftArrow,
	DoubleLongLeftRightArrow: DoubleLongLeftRightArrow,
	DoubleLongRightArrow: DoubleLongRightArrow,
	DoubleRightArrow: DoubleRightArrow,
	DoubleRightTee: DoubleRightTee,
	DoubleUpArrow: DoubleUpArrow,
	DoubleUpDownArrow: DoubleUpDownArrow,
	DoubleVerticalBar: DoubleVerticalBar,
	DownArrowBar: DownArrowBar,
	downarrow: downarrow,
	DownArrow: DownArrow,
	Downarrow: Downarrow,
	DownArrowUpArrow: DownArrowUpArrow,
	DownBreve: DownBreve,
	downdownarrows: downdownarrows,
	downharpoonleft: downharpoonleft,
	downharpoonright: downharpoonright,
	DownLeftRightVector: DownLeftRightVector,
	DownLeftTeeVector: DownLeftTeeVector,
	DownLeftVectorBar: DownLeftVectorBar,
	DownLeftVector: DownLeftVector,
	DownRightTeeVector: DownRightTeeVector,
	DownRightVectorBar: DownRightVectorBar,
	DownRightVector: DownRightVector,
	DownTeeArrow: DownTeeArrow,
	DownTee: DownTee,
	drbkarow: drbkarow,
	drcorn: drcorn,
	drcrop: drcrop,
	Dscr: Dscr,
	dscr: dscr,
	DScy: DScy,
	dscy: dscy,
	dsol: dsol,
	Dstrok: Dstrok,
	dstrok: dstrok,
	dtdot: dtdot,
	dtri: dtri,
	dtrif: dtrif,
	duarr: duarr,
	duhar: duhar,
	dwangle: dwangle,
	DZcy: DZcy,
	dzcy: dzcy,
	dzigrarr: dzigrarr,
	Eacute: Eacute,
	eacute: eacute,
	easter: easter,
	Ecaron: Ecaron,
	ecaron: ecaron,
	Ecirc: Ecirc,
	ecirc: ecirc,
	ecir: ecir,
	ecolon: ecolon,
	Ecy: Ecy,
	ecy: ecy,
	eDDot: eDDot,
	Edot: Edot,
	edot: edot,
	eDot: eDot,
	ee: ee,
	efDot: efDot,
	Efr: Efr,
	efr: efr,
	eg: eg,
	Egrave: Egrave,
	egrave: egrave,
	egs: egs,
	egsdot: egsdot,
	el: el,
	Element: Element,
	elinters: elinters,
	ell: ell,
	els: els,
	elsdot: elsdot,
	Emacr: Emacr,
	emacr: emacr,
	empty: empty,
	emptyset: emptyset,
	EmptySmallSquare: EmptySmallSquare,
	emptyv: emptyv,
	EmptyVerySmallSquare: EmptyVerySmallSquare,
	emsp13: emsp13,
	emsp14: emsp14,
	emsp: emsp,
	ENG: ENG,
	eng: eng,
	ensp: ensp,
	Eogon: Eogon,
	eogon: eogon,
	Eopf: Eopf,
	eopf: eopf,
	epar: epar,
	eparsl: eparsl,
	eplus: eplus,
	epsi: epsi,
	Epsilon: Epsilon,
	epsilon: epsilon,
	epsiv: epsiv,
	eqcirc: eqcirc,
	eqcolon: eqcolon,
	eqsim: eqsim,
	eqslantgtr: eqslantgtr,
	eqslantless: eqslantless,
	Equal: Equal,
	equals: equals,
	EqualTilde: EqualTilde,
	equest: equest,
	Equilibrium: Equilibrium,
	equiv: equiv,
	equivDD: equivDD,
	eqvparsl: eqvparsl,
	erarr: erarr,
	erDot: erDot,
	escr: escr,
	Escr: Escr,
	esdot: esdot,
	Esim: Esim,
	esim: esim,
	Eta: Eta,
	eta: eta,
	ETH: ETH,
	eth: eth,
	Euml: Euml,
	euml: euml,
	euro: euro,
	excl: excl,
	exist: exist,
	Exists: Exists,
	expectation: expectation,
	exponentiale: exponentiale,
	ExponentialE: ExponentialE,
	fallingdotseq: fallingdotseq,
	Fcy: Fcy,
	fcy: fcy,
	female: female,
	ffilig: ffilig,
	fflig: fflig,
	ffllig: ffllig,
	Ffr: Ffr,
	ffr: ffr,
	filig: filig,
	FilledSmallSquare: FilledSmallSquare,
	FilledVerySmallSquare: FilledVerySmallSquare,
	fjlig: fjlig,
	flat: flat,
	fllig: fllig,
	fltns: fltns,
	fnof: fnof,
	Fopf: Fopf,
	fopf: fopf,
	forall: forall,
	ForAll: ForAll,
	fork: fork,
	forkv: forkv,
	Fouriertrf: Fouriertrf,
	fpartint: fpartint,
	frac12: frac12,
	frac13: frac13,
	frac14: frac14,
	frac15: frac15,
	frac16: frac16,
	frac18: frac18,
	frac23: frac23,
	frac25: frac25,
	frac34: frac34,
	frac35: frac35,
	frac38: frac38,
	frac45: frac45,
	frac56: frac56,
	frac58: frac58,
	frac78: frac78,
	frasl: frasl,
	frown: frown,
	fscr: fscr,
	Fscr: Fscr,
	gacute: gacute,
	Gamma: Gamma,
	gamma: gamma,
	Gammad: Gammad,
	gammad: gammad,
	gap: gap,
	Gbreve: Gbreve,
	gbreve: gbreve,
	Gcedil: Gcedil,
	Gcirc: Gcirc,
	gcirc: gcirc,
	Gcy: Gcy,
	gcy: gcy,
	Gdot: Gdot,
	gdot: gdot,
	ge: ge,
	gE: gE,
	gEl: gEl,
	gel: gel,
	geq: geq,
	geqq: geqq,
	geqslant: geqslant,
	gescc: gescc,
	ges: ges,
	gesdot: gesdot,
	gesdoto: gesdoto,
	gesdotol: gesdotol,
	gesl: gesl,
	gesles: gesles,
	Gfr: Gfr,
	gfr: gfr,
	gg: gg,
	Gg: Gg,
	ggg: ggg,
	gimel: gimel,
	GJcy: GJcy,
	gjcy: gjcy,
	gla: gla,
	gl: gl,
	glE: glE,
	glj: glj,
	gnap: gnap,
	gnapprox: gnapprox,
	gne: gne,
	gnE: gnE,
	gneq: gneq,
	gneqq: gneqq,
	gnsim: gnsim,
	Gopf: Gopf,
	gopf: gopf,
	grave: grave,
	GreaterEqual: GreaterEqual,
	GreaterEqualLess: GreaterEqualLess,
	GreaterFullEqual: GreaterFullEqual,
	GreaterGreater: GreaterGreater,
	GreaterLess: GreaterLess,
	GreaterSlantEqual: GreaterSlantEqual,
	GreaterTilde: GreaterTilde,
	Gscr: Gscr,
	gscr: gscr,
	gsim: gsim,
	gsime: gsime,
	gsiml: gsiml,
	gtcc: gtcc,
	gtcir: gtcir,
	gt: gt$1,
	GT: GT,
	Gt: Gt,
	gtdot: gtdot,
	gtlPar: gtlPar,
	gtquest: gtquest,
	gtrapprox: gtrapprox,
	gtrarr: gtrarr,
	gtrdot: gtrdot,
	gtreqless: gtreqless,
	gtreqqless: gtreqqless,
	gtrless: gtrless,
	gtrsim: gtrsim,
	gvertneqq: gvertneqq,
	gvnE: gvnE,
	Hacek: Hacek,
	hairsp: hairsp,
	half: half,
	hamilt: hamilt,
	HARDcy: HARDcy,
	hardcy: hardcy,
	harrcir: harrcir,
	harr: harr,
	hArr: hArr,
	harrw: harrw,
	Hat: Hat,
	hbar: hbar,
	Hcirc: Hcirc,
	hcirc: hcirc,
	hearts: hearts,
	heartsuit: heartsuit,
	hellip: hellip,
	hercon: hercon,
	hfr: hfr,
	Hfr: Hfr,
	HilbertSpace: HilbertSpace,
	hksearow: hksearow,
	hkswarow: hkswarow,
	hoarr: hoarr,
	homtht: homtht,
	hookleftarrow: hookleftarrow,
	hookrightarrow: hookrightarrow,
	hopf: hopf,
	Hopf: Hopf,
	horbar: horbar,
	HorizontalLine: HorizontalLine,
	hscr: hscr,
	Hscr: Hscr,
	hslash: hslash,
	Hstrok: Hstrok,
	hstrok: hstrok,
	HumpDownHump: HumpDownHump,
	HumpEqual: HumpEqual,
	hybull: hybull,
	hyphen: hyphen,
	Iacute: Iacute,
	iacute: iacute,
	ic: ic,
	Icirc: Icirc,
	icirc: icirc,
	Icy: Icy,
	icy: icy,
	Idot: Idot,
	IEcy: IEcy,
	iecy: iecy,
	iexcl: iexcl,
	iff: iff,
	ifr: ifr,
	Ifr: Ifr,
	Igrave: Igrave,
	igrave: igrave,
	ii: ii,
	iiiint: iiiint,
	iiint: iiint,
	iinfin: iinfin,
	iiota: iiota,
	IJlig: IJlig,
	ijlig: ijlig,
	Imacr: Imacr,
	imacr: imacr,
	image: image,
	ImaginaryI: ImaginaryI,
	imagline: imagline,
	imagpart: imagpart,
	imath: imath,
	Im: Im,
	imof: imof,
	imped: imped,
	Implies: Implies,
	incare: incare,
	infin: infin,
	infintie: infintie,
	inodot: inodot,
	intcal: intcal,
	int: int,
	Int: Int,
	integers: integers,
	Integral: Integral,
	intercal: intercal,
	Intersection: Intersection,
	intlarhk: intlarhk,
	intprod: intprod,
	InvisibleComma: InvisibleComma,
	InvisibleTimes: InvisibleTimes,
	IOcy: IOcy,
	iocy: iocy,
	Iogon: Iogon,
	iogon: iogon,
	Iopf: Iopf,
	iopf: iopf,
	Iota: Iota,
	iota: iota,
	iprod: iprod,
	iquest: iquest,
	iscr: iscr,
	Iscr: Iscr,
	isin: isin,
	isindot: isindot,
	isinE: isinE,
	isins: isins,
	isinsv: isinsv,
	isinv: isinv,
	it: it,
	Itilde: Itilde,
	itilde: itilde,
	Iukcy: Iukcy,
	iukcy: iukcy,
	Iuml: Iuml,
	iuml: iuml,
	Jcirc: Jcirc,
	jcirc: jcirc,
	Jcy: Jcy,
	jcy: jcy,
	Jfr: Jfr,
	jfr: jfr,
	jmath: jmath,
	Jopf: Jopf,
	jopf: jopf,
	Jscr: Jscr,
	jscr: jscr,
	Jsercy: Jsercy,
	jsercy: jsercy,
	Jukcy: Jukcy,
	jukcy: jukcy,
	Kappa: Kappa,
	kappa: kappa,
	kappav: kappav,
	Kcedil: Kcedil,
	kcedil: kcedil,
	Kcy: Kcy,
	kcy: kcy,
	Kfr: Kfr,
	kfr: kfr,
	kgreen: kgreen,
	KHcy: KHcy,
	khcy: khcy,
	KJcy: KJcy,
	kjcy: kjcy,
	Kopf: Kopf,
	kopf: kopf,
	Kscr: Kscr,
	kscr: kscr,
	lAarr: lAarr,
	Lacute: Lacute,
	lacute: lacute,
	laemptyv: laemptyv,
	lagran: lagran,
	Lambda: Lambda,
	lambda: lambda,
	lang: lang,
	Lang: Lang,
	langd: langd,
	langle: langle,
	lap: lap,
	Laplacetrf: Laplacetrf,
	laquo: laquo,
	larrb: larrb,
	larrbfs: larrbfs,
	larr: larr,
	Larr: Larr,
	lArr: lArr,
	larrfs: larrfs,
	larrhk: larrhk,
	larrlp: larrlp,
	larrpl: larrpl,
	larrsim: larrsim,
	larrtl: larrtl,
	latail: latail,
	lAtail: lAtail,
	lat: lat,
	late: late,
	lates: lates,
	lbarr: lbarr,
	lBarr: lBarr,
	lbbrk: lbbrk,
	lbrace: lbrace,
	lbrack: lbrack,
	lbrke: lbrke,
	lbrksld: lbrksld,
	lbrkslu: lbrkslu,
	Lcaron: Lcaron,
	lcaron: lcaron,
	Lcedil: Lcedil,
	lcedil: lcedil,
	lceil: lceil,
	lcub: lcub,
	Lcy: Lcy,
	lcy: lcy,
	ldca: ldca,
	ldquo: ldquo,
	ldquor: ldquor,
	ldrdhar: ldrdhar,
	ldrushar: ldrushar,
	ldsh: ldsh,
	le: le,
	lE: lE,
	LeftAngleBracket: LeftAngleBracket,
	LeftArrowBar: LeftArrowBar,
	leftarrow: leftarrow,
	LeftArrow: LeftArrow,
	Leftarrow: Leftarrow,
	LeftArrowRightArrow: LeftArrowRightArrow,
	leftarrowtail: leftarrowtail,
	LeftCeiling: LeftCeiling,
	LeftDoubleBracket: LeftDoubleBracket,
	LeftDownTeeVector: LeftDownTeeVector,
	LeftDownVectorBar: LeftDownVectorBar,
	LeftDownVector: LeftDownVector,
	LeftFloor: LeftFloor,
	leftharpoondown: leftharpoondown,
	leftharpoonup: leftharpoonup,
	leftleftarrows: leftleftarrows,
	leftrightarrow: leftrightarrow,
	LeftRightArrow: LeftRightArrow,
	Leftrightarrow: Leftrightarrow,
	leftrightarrows: leftrightarrows,
	leftrightharpoons: leftrightharpoons,
	leftrightsquigarrow: leftrightsquigarrow,
	LeftRightVector: LeftRightVector,
	LeftTeeArrow: LeftTeeArrow,
	LeftTee: LeftTee,
	LeftTeeVector: LeftTeeVector,
	leftthreetimes: leftthreetimes,
	LeftTriangleBar: LeftTriangleBar,
	LeftTriangle: LeftTriangle,
	LeftTriangleEqual: LeftTriangleEqual,
	LeftUpDownVector: LeftUpDownVector,
	LeftUpTeeVector: LeftUpTeeVector,
	LeftUpVectorBar: LeftUpVectorBar,
	LeftUpVector: LeftUpVector,
	LeftVectorBar: LeftVectorBar,
	LeftVector: LeftVector,
	lEg: lEg,
	leg: leg,
	leq: leq,
	leqq: leqq,
	leqslant: leqslant,
	lescc: lescc,
	les: les,
	lesdot: lesdot,
	lesdoto: lesdoto,
	lesdotor: lesdotor,
	lesg: lesg,
	lesges: lesges,
	lessapprox: lessapprox,
	lessdot: lessdot,
	lesseqgtr: lesseqgtr,
	lesseqqgtr: lesseqqgtr,
	LessEqualGreater: LessEqualGreater,
	LessFullEqual: LessFullEqual,
	LessGreater: LessGreater,
	lessgtr: lessgtr,
	LessLess: LessLess,
	lesssim: lesssim,
	LessSlantEqual: LessSlantEqual,
	LessTilde: LessTilde,
	lfisht: lfisht,
	lfloor: lfloor,
	Lfr: Lfr,
	lfr: lfr,
	lg: lg,
	lgE: lgE,
	lHar: lHar,
	lhard: lhard,
	lharu: lharu,
	lharul: lharul,
	lhblk: lhblk,
	LJcy: LJcy,
	ljcy: ljcy,
	llarr: llarr,
	ll: ll,
	Ll: Ll,
	llcorner: llcorner,
	Lleftarrow: Lleftarrow,
	llhard: llhard,
	lltri: lltri,
	Lmidot: Lmidot,
	lmidot: lmidot,
	lmoustache: lmoustache,
	lmoust: lmoust,
	lnap: lnap,
	lnapprox: lnapprox,
	lne: lne,
	lnE: lnE,
	lneq: lneq,
	lneqq: lneqq,
	lnsim: lnsim,
	loang: loang,
	loarr: loarr,
	lobrk: lobrk,
	longleftarrow: longleftarrow,
	LongLeftArrow: LongLeftArrow,
	Longleftarrow: Longleftarrow,
	longleftrightarrow: longleftrightarrow,
	LongLeftRightArrow: LongLeftRightArrow,
	Longleftrightarrow: Longleftrightarrow,
	longmapsto: longmapsto,
	longrightarrow: longrightarrow,
	LongRightArrow: LongRightArrow,
	Longrightarrow: Longrightarrow,
	looparrowleft: looparrowleft,
	looparrowright: looparrowright,
	lopar: lopar,
	Lopf: Lopf,
	lopf: lopf,
	loplus: loplus,
	lotimes: lotimes,
	lowast: lowast,
	lowbar: lowbar,
	LowerLeftArrow: LowerLeftArrow,
	LowerRightArrow: LowerRightArrow,
	loz: loz,
	lozenge: lozenge,
	lozf: lozf,
	lpar: lpar,
	lparlt: lparlt,
	lrarr: lrarr,
	lrcorner: lrcorner,
	lrhar: lrhar,
	lrhard: lrhard,
	lrm: lrm,
	lrtri: lrtri,
	lsaquo: lsaquo,
	lscr: lscr,
	Lscr: Lscr,
	lsh: lsh,
	Lsh: Lsh,
	lsim: lsim,
	lsime: lsime,
	lsimg: lsimg,
	lsqb: lsqb,
	lsquo: lsquo,
	lsquor: lsquor,
	Lstrok: Lstrok,
	lstrok: lstrok,
	ltcc: ltcc,
	ltcir: ltcir,
	lt: lt$1,
	LT: LT,
	Lt: Lt,
	ltdot: ltdot,
	lthree: lthree,
	ltimes: ltimes,
	ltlarr: ltlarr,
	ltquest: ltquest,
	ltri: ltri,
	ltrie: ltrie,
	ltrif: ltrif,
	ltrPar: ltrPar,
	lurdshar: lurdshar,
	luruhar: luruhar,
	lvertneqq: lvertneqq,
	lvnE: lvnE,
	macr: macr,
	male: male,
	malt: malt,
	maltese: maltese,
	map: map$1,
	mapsto: mapsto,
	mapstodown: mapstodown,
	mapstoleft: mapstoleft,
	mapstoup: mapstoup,
	marker: marker,
	mcomma: mcomma,
	Mcy: Mcy,
	mcy: mcy,
	mdash: mdash,
	mDDot: mDDot,
	measuredangle: measuredangle,
	MediumSpace: MediumSpace,
	Mellintrf: Mellintrf,
	Mfr: Mfr,
	mfr: mfr,
	mho: mho,
	micro: micro,
	midast: midast,
	midcir: midcir,
	mid: mid,
	middot: middot,
	minusb: minusb,
	minus: minus,
	minusd: minusd,
	minusdu: minusdu,
	MinusPlus: MinusPlus,
	mlcp: mlcp,
	mldr: mldr,
	mnplus: mnplus,
	models: models,
	Mopf: Mopf,
	mopf: mopf,
	mp: mp,
	mscr: mscr,
	Mscr: Mscr,
	mstpos: mstpos,
	Mu: Mu,
	mu: mu,
	multimap: multimap,
	mumap: mumap,
	nabla: nabla,
	Nacute: Nacute,
	nacute: nacute,
	nang: nang,
	nap: nap,
	napE: napE,
	napid: napid,
	napos: napos,
	napprox: napprox,
	natural: natural,
	naturals: naturals,
	natur: natur,
	nbsp: nbsp,
	nbump: nbump,
	nbumpe: nbumpe,
	ncap: ncap,
	Ncaron: Ncaron,
	ncaron: ncaron,
	Ncedil: Ncedil,
	ncedil: ncedil,
	ncong: ncong,
	ncongdot: ncongdot,
	ncup: ncup,
	Ncy: Ncy,
	ncy: ncy,
	ndash: ndash,
	nearhk: nearhk,
	nearr: nearr,
	neArr: neArr,
	nearrow: nearrow,
	ne: ne,
	nedot: nedot,
	NegativeMediumSpace: NegativeMediumSpace,
	NegativeThickSpace: NegativeThickSpace,
	NegativeThinSpace: NegativeThinSpace,
	NegativeVeryThinSpace: NegativeVeryThinSpace,
	nequiv: nequiv,
	nesear: nesear,
	nesim: nesim,
	NestedGreaterGreater: NestedGreaterGreater,
	NestedLessLess: NestedLessLess,
	NewLine: NewLine,
	nexist: nexist,
	nexists: nexists,
	Nfr: Nfr,
	nfr: nfr,
	ngE: ngE,
	nge: nge,
	ngeq: ngeq,
	ngeqq: ngeqq,
	ngeqslant: ngeqslant,
	nges: nges,
	nGg: nGg,
	ngsim: ngsim,
	nGt: nGt,
	ngt: ngt,
	ngtr: ngtr,
	nGtv: nGtv,
	nharr: nharr,
	nhArr: nhArr,
	nhpar: nhpar,
	ni: ni,
	nis: nis,
	nisd: nisd,
	niv: niv,
	NJcy: NJcy,
	njcy: njcy,
	nlarr: nlarr,
	nlArr: nlArr,
	nldr: nldr,
	nlE: nlE,
	nle: nle,
	nleftarrow: nleftarrow,
	nLeftarrow: nLeftarrow,
	nleftrightarrow: nleftrightarrow,
	nLeftrightarrow: nLeftrightarrow,
	nleq: nleq,
	nleqq: nleqq,
	nleqslant: nleqslant,
	nles: nles,
	nless: nless,
	nLl: nLl,
	nlsim: nlsim,
	nLt: nLt,
	nlt: nlt,
	nltri: nltri,
	nltrie: nltrie,
	nLtv: nLtv,
	nmid: nmid,
	NoBreak: NoBreak,
	NonBreakingSpace: NonBreakingSpace,
	nopf: nopf,
	Nopf: Nopf,
	Not: Not,
	not: not,
	NotCongruent: NotCongruent,
	NotCupCap: NotCupCap,
	NotDoubleVerticalBar: NotDoubleVerticalBar,
	NotElement: NotElement,
	NotEqual: NotEqual,
	NotEqualTilde: NotEqualTilde,
	NotExists: NotExists,
	NotGreater: NotGreater,
	NotGreaterEqual: NotGreaterEqual,
	NotGreaterFullEqual: NotGreaterFullEqual,
	NotGreaterGreater: NotGreaterGreater,
	NotGreaterLess: NotGreaterLess,
	NotGreaterSlantEqual: NotGreaterSlantEqual,
	NotGreaterTilde: NotGreaterTilde,
	NotHumpDownHump: NotHumpDownHump,
	NotHumpEqual: NotHumpEqual,
	notin: notin,
	notindot: notindot,
	notinE: notinE,
	notinva: notinva,
	notinvb: notinvb,
	notinvc: notinvc,
	NotLeftTriangleBar: NotLeftTriangleBar,
	NotLeftTriangle: NotLeftTriangle,
	NotLeftTriangleEqual: NotLeftTriangleEqual,
	NotLess: NotLess,
	NotLessEqual: NotLessEqual,
	NotLessGreater: NotLessGreater,
	NotLessLess: NotLessLess,
	NotLessSlantEqual: NotLessSlantEqual,
	NotLessTilde: NotLessTilde,
	NotNestedGreaterGreater: NotNestedGreaterGreater,
	NotNestedLessLess: NotNestedLessLess,
	notni: notni,
	notniva: notniva,
	notnivb: notnivb,
	notnivc: notnivc,
	NotPrecedes: NotPrecedes,
	NotPrecedesEqual: NotPrecedesEqual,
	NotPrecedesSlantEqual: NotPrecedesSlantEqual,
	NotReverseElement: NotReverseElement,
	NotRightTriangleBar: NotRightTriangleBar,
	NotRightTriangle: NotRightTriangle,
	NotRightTriangleEqual: NotRightTriangleEqual,
	NotSquareSubset: NotSquareSubset,
	NotSquareSubsetEqual: NotSquareSubsetEqual,
	NotSquareSuperset: NotSquareSuperset,
	NotSquareSupersetEqual: NotSquareSupersetEqual,
	NotSubset: NotSubset,
	NotSubsetEqual: NotSubsetEqual,
	NotSucceeds: NotSucceeds,
	NotSucceedsEqual: NotSucceedsEqual,
	NotSucceedsSlantEqual: NotSucceedsSlantEqual,
	NotSucceedsTilde: NotSucceedsTilde,
	NotSuperset: NotSuperset,
	NotSupersetEqual: NotSupersetEqual,
	NotTilde: NotTilde,
	NotTildeEqual: NotTildeEqual,
	NotTildeFullEqual: NotTildeFullEqual,
	NotTildeTilde: NotTildeTilde,
	NotVerticalBar: NotVerticalBar,
	nparallel: nparallel,
	npar: npar,
	nparsl: nparsl,
	npart: npart,
	npolint: npolint,
	npr: npr,
	nprcue: nprcue,
	nprec: nprec,
	npreceq: npreceq,
	npre: npre,
	nrarrc: nrarrc,
	nrarr: nrarr,
	nrArr: nrArr,
	nrarrw: nrarrw,
	nrightarrow: nrightarrow,
	nRightarrow: nRightarrow,
	nrtri: nrtri,
	nrtrie: nrtrie,
	nsc: nsc,
	nsccue: nsccue,
	nsce: nsce,
	Nscr: Nscr,
	nscr: nscr,
	nshortmid: nshortmid,
	nshortparallel: nshortparallel,
	nsim: nsim,
	nsime: nsime,
	nsimeq: nsimeq,
	nsmid: nsmid,
	nspar: nspar,
	nsqsube: nsqsube,
	nsqsupe: nsqsupe,
	nsub: nsub,
	nsubE: nsubE,
	nsube: nsube,
	nsubset: nsubset,
	nsubseteq: nsubseteq,
	nsubseteqq: nsubseteqq,
	nsucc: nsucc,
	nsucceq: nsucceq,
	nsup: nsup,
	nsupE: nsupE,
	nsupe: nsupe,
	nsupset: nsupset,
	nsupseteq: nsupseteq,
	nsupseteqq: nsupseteqq,
	ntgl: ntgl,
	Ntilde: Ntilde,
	ntilde: ntilde,
	ntlg: ntlg,
	ntriangleleft: ntriangleleft,
	ntrianglelefteq: ntrianglelefteq,
	ntriangleright: ntriangleright,
	ntrianglerighteq: ntrianglerighteq,
	Nu: Nu,
	nu: nu,
	num: num,
	numero: numero,
	numsp: numsp,
	nvap: nvap,
	nvdash: nvdash,
	nvDash: nvDash,
	nVdash: nVdash,
	nVDash: nVDash,
	nvge: nvge,
	nvgt: nvgt,
	nvHarr: nvHarr,
	nvinfin: nvinfin,
	nvlArr: nvlArr,
	nvle: nvle,
	nvlt: nvlt,
	nvltrie: nvltrie,
	nvrArr: nvrArr,
	nvrtrie: nvrtrie,
	nvsim: nvsim,
	nwarhk: nwarhk,
	nwarr: nwarr,
	nwArr: nwArr,
	nwarrow: nwarrow,
	nwnear: nwnear,
	Oacute: Oacute,
	oacute: oacute,
	oast: oast,
	Ocirc: Ocirc,
	ocirc: ocirc,
	ocir: ocir,
	Ocy: Ocy,
	ocy: ocy,
	odash: odash,
	Odblac: Odblac,
	odblac: odblac,
	odiv: odiv,
	odot: odot,
	odsold: odsold,
	OElig: OElig,
	oelig: oelig,
	ofcir: ofcir,
	Ofr: Ofr,
	ofr: ofr,
	ogon: ogon,
	Ograve: Ograve,
	ograve: ograve,
	ogt: ogt,
	ohbar: ohbar,
	ohm: ohm,
	oint: oint,
	olarr: olarr,
	olcir: olcir,
	olcross: olcross,
	oline: oline,
	olt: olt,
	Omacr: Omacr,
	omacr: omacr,
	Omega: Omega,
	omega: omega,
	Omicron: Omicron,
	omicron: omicron,
	omid: omid,
	ominus: ominus,
	Oopf: Oopf,
	oopf: oopf,
	opar: opar,
	OpenCurlyDoubleQuote: OpenCurlyDoubleQuote,
	OpenCurlyQuote: OpenCurlyQuote,
	operp: operp,
	oplus: oplus,
	orarr: orarr,
	Or: Or,
	or: or,
	ord: ord,
	order: order,
	orderof: orderof,
	ordf: ordf,
	ordm: ordm,
	origof: origof,
	oror: oror,
	orslope: orslope,
	orv: orv,
	oS: oS,
	Oscr: Oscr,
	oscr: oscr,
	Oslash: Oslash,
	oslash: oslash,
	osol: osol,
	Otilde: Otilde,
	otilde: otilde,
	otimesas: otimesas,
	Otimes: Otimes,
	otimes: otimes,
	Ouml: Ouml,
	ouml: ouml,
	ovbar: ovbar,
	OverBar: OverBar,
	OverBrace: OverBrace,
	OverBracket: OverBracket,
	OverParenthesis: OverParenthesis,
	para: para,
	parallel: parallel,
	par: par,
	parsim: parsim,
	parsl: parsl,
	part: part,
	PartialD: PartialD,
	Pcy: Pcy,
	pcy: pcy,
	percnt: percnt,
	period: period,
	permil: permil,
	perp: perp,
	pertenk: pertenk,
	Pfr: Pfr,
	pfr: pfr,
	Phi: Phi,
	phi: phi,
	phiv: phiv,
	phmmat: phmmat,
	phone: phone,
	Pi: Pi,
	pi: pi,
	pitchfork: pitchfork,
	piv: piv,
	planck: planck,
	planckh: planckh,
	plankv: plankv,
	plusacir: plusacir,
	plusb: plusb,
	pluscir: pluscir,
	plus: plus,
	plusdo: plusdo,
	plusdu: plusdu,
	pluse: pluse,
	PlusMinus: PlusMinus,
	plusmn: plusmn,
	plussim: plussim,
	plustwo: plustwo,
	pm: pm,
	Poincareplane: Poincareplane,
	pointint: pointint,
	popf: popf,
	Popf: Popf,
	pound: pound,
	prap: prap,
	Pr: Pr,
	pr: pr,
	prcue: prcue,
	precapprox: precapprox,
	prec: prec,
	preccurlyeq: preccurlyeq,
	Precedes: Precedes,
	PrecedesEqual: PrecedesEqual,
	PrecedesSlantEqual: PrecedesSlantEqual,
	PrecedesTilde: PrecedesTilde,
	preceq: preceq,
	precnapprox: precnapprox,
	precneqq: precneqq,
	precnsim: precnsim,
	pre: pre,
	prE: prE,
	precsim: precsim,
	prime: prime,
	Prime: Prime,
	primes: primes,
	prnap: prnap,
	prnE: prnE,
	prnsim: prnsim,
	prod: prod,
	Product: Product,
	profalar: profalar,
	profline: profline,
	profsurf: profsurf,
	prop: prop,
	Proportional: Proportional,
	Proportion: Proportion,
	propto: propto,
	prsim: prsim,
	prurel: prurel,
	Pscr: Pscr,
	pscr: pscr,
	Psi: Psi,
	psi: psi,
	puncsp: puncsp,
	Qfr: Qfr,
	qfr: qfr,
	qint: qint,
	qopf: qopf,
	Qopf: Qopf,
	qprime: qprime,
	Qscr: Qscr,
	qscr: qscr,
	quaternions: quaternions,
	quatint: quatint,
	quest: quest,
	questeq: questeq,
	quot: quot$1,
	QUOT: QUOT,
	rAarr: rAarr,
	race: race,
	Racute: Racute,
	racute: racute,
	radic: radic,
	raemptyv: raemptyv,
	rang: rang,
	Rang: Rang,
	rangd: rangd,
	range: range,
	rangle: rangle,
	raquo: raquo,
	rarrap: rarrap,
	rarrb: rarrb,
	rarrbfs: rarrbfs,
	rarrc: rarrc,
	rarr: rarr,
	Rarr: Rarr,
	rArr: rArr,
	rarrfs: rarrfs,
	rarrhk: rarrhk,
	rarrlp: rarrlp,
	rarrpl: rarrpl,
	rarrsim: rarrsim,
	Rarrtl: Rarrtl,
	rarrtl: rarrtl,
	rarrw: rarrw,
	ratail: ratail,
	rAtail: rAtail,
	ratio: ratio,
	rationals: rationals,
	rbarr: rbarr,
	rBarr: rBarr,
	RBarr: RBarr,
	rbbrk: rbbrk,
	rbrace: rbrace,
	rbrack: rbrack,
	rbrke: rbrke,
	rbrksld: rbrksld,
	rbrkslu: rbrkslu,
	Rcaron: Rcaron,
	rcaron: rcaron,
	Rcedil: Rcedil,
	rcedil: rcedil,
	rceil: rceil,
	rcub: rcub,
	Rcy: Rcy,
	rcy: rcy,
	rdca: rdca,
	rdldhar: rdldhar,
	rdquo: rdquo,
	rdquor: rdquor,
	rdsh: rdsh,
	real: real,
	realine: realine,
	realpart: realpart,
	reals: reals,
	Re: Re,
	rect: rect,
	reg: reg,
	REG: REG,
	ReverseElement: ReverseElement,
	ReverseEquilibrium: ReverseEquilibrium,
	ReverseUpEquilibrium: ReverseUpEquilibrium,
	rfisht: rfisht,
	rfloor: rfloor,
	rfr: rfr,
	Rfr: Rfr,
	rHar: rHar,
	rhard: rhard,
	rharu: rharu,
	rharul: rharul,
	Rho: Rho,
	rho: rho,
	rhov: rhov,
	RightAngleBracket: RightAngleBracket,
	RightArrowBar: RightArrowBar,
	rightarrow: rightarrow,
	RightArrow: RightArrow,
	Rightarrow: Rightarrow,
	RightArrowLeftArrow: RightArrowLeftArrow,
	rightarrowtail: rightarrowtail,
	RightCeiling: RightCeiling,
	RightDoubleBracket: RightDoubleBracket,
	RightDownTeeVector: RightDownTeeVector,
	RightDownVectorBar: RightDownVectorBar,
	RightDownVector: RightDownVector,
	RightFloor: RightFloor,
	rightharpoondown: rightharpoondown,
	rightharpoonup: rightharpoonup,
	rightleftarrows: rightleftarrows,
	rightleftharpoons: rightleftharpoons,
	rightrightarrows: rightrightarrows,
	rightsquigarrow: rightsquigarrow,
	RightTeeArrow: RightTeeArrow,
	RightTee: RightTee,
	RightTeeVector: RightTeeVector,
	rightthreetimes: rightthreetimes,
	RightTriangleBar: RightTriangleBar,
	RightTriangle: RightTriangle,
	RightTriangleEqual: RightTriangleEqual,
	RightUpDownVector: RightUpDownVector,
	RightUpTeeVector: RightUpTeeVector,
	RightUpVectorBar: RightUpVectorBar,
	RightUpVector: RightUpVector,
	RightVectorBar: RightVectorBar,
	RightVector: RightVector,
	ring: ring,
	risingdotseq: risingdotseq,
	rlarr: rlarr,
	rlhar: rlhar,
	rlm: rlm,
	rmoustache: rmoustache,
	rmoust: rmoust,
	rnmid: rnmid,
	roang: roang,
	roarr: roarr,
	robrk: robrk,
	ropar: ropar,
	ropf: ropf,
	Ropf: Ropf,
	roplus: roplus,
	rotimes: rotimes,
	RoundImplies: RoundImplies,
	rpar: rpar,
	rpargt: rpargt,
	rppolint: rppolint,
	rrarr: rrarr,
	Rrightarrow: Rrightarrow,
	rsaquo: rsaquo,
	rscr: rscr,
	Rscr: Rscr,
	rsh: rsh,
	Rsh: Rsh,
	rsqb: rsqb,
	rsquo: rsquo,
	rsquor: rsquor,
	rthree: rthree,
	rtimes: rtimes,
	rtri: rtri,
	rtrie: rtrie,
	rtrif: rtrif,
	rtriltri: rtriltri,
	RuleDelayed: RuleDelayed,
	ruluhar: ruluhar,
	rx: rx,
	Sacute: Sacute,
	sacute: sacute,
	sbquo: sbquo,
	scap: scap,
	Scaron: Scaron,
	scaron: scaron,
	Sc: Sc,
	sc: sc,
	sccue: sccue,
	sce: sce,
	scE: scE,
	Scedil: Scedil,
	scedil: scedil,
	Scirc: Scirc,
	scirc: scirc,
	scnap: scnap,
	scnE: scnE,
	scnsim: scnsim,
	scpolint: scpolint,
	scsim: scsim,
	Scy: Scy,
	scy: scy,
	sdotb: sdotb,
	sdot: sdot,
	sdote: sdote,
	searhk: searhk,
	searr: searr,
	seArr: seArr,
	searrow: searrow,
	sect: sect,
	semi: semi,
	seswar: seswar,
	setminus: setminus,
	setmn: setmn,
	sext: sext,
	Sfr: Sfr,
	sfr: sfr,
	sfrown: sfrown,
	sharp: sharp,
	SHCHcy: SHCHcy,
	shchcy: shchcy,
	SHcy: SHcy,
	shcy: shcy,
	ShortDownArrow: ShortDownArrow,
	ShortLeftArrow: ShortLeftArrow,
	shortmid: shortmid,
	shortparallel: shortparallel,
	ShortRightArrow: ShortRightArrow,
	ShortUpArrow: ShortUpArrow,
	shy: shy,
	Sigma: Sigma,
	sigma: sigma,
	sigmaf: sigmaf,
	sigmav: sigmav,
	sim: sim,
	simdot: simdot,
	sime: sime,
	simeq: simeq,
	simg: simg,
	simgE: simgE,
	siml: siml,
	simlE: simlE,
	simne: simne,
	simplus: simplus,
	simrarr: simrarr,
	slarr: slarr,
	SmallCircle: SmallCircle,
	smallsetminus: smallsetminus,
	smashp: smashp,
	smeparsl: smeparsl,
	smid: smid,
	smile: smile,
	smt: smt,
	smte: smte,
	smtes: smtes,
	SOFTcy: SOFTcy,
	softcy: softcy,
	solbar: solbar,
	solb: solb,
	sol: sol,
	Sopf: Sopf,
	sopf: sopf,
	spades: spades,
	spadesuit: spadesuit,
	spar: spar,
	sqcap: sqcap,
	sqcaps: sqcaps,
	sqcup: sqcup,
	sqcups: sqcups,
	Sqrt: Sqrt,
	sqsub: sqsub,
	sqsube: sqsube,
	sqsubset: sqsubset,
	sqsubseteq: sqsubseteq,
	sqsup: sqsup,
	sqsupe: sqsupe,
	sqsupset: sqsupset,
	sqsupseteq: sqsupseteq,
	square: square,
	Square: Square,
	SquareIntersection: SquareIntersection,
	SquareSubset: SquareSubset,
	SquareSubsetEqual: SquareSubsetEqual,
	SquareSuperset: SquareSuperset,
	SquareSupersetEqual: SquareSupersetEqual,
	SquareUnion: SquareUnion,
	squarf: squarf,
	squ: squ,
	squf: squf,
	srarr: srarr,
	Sscr: Sscr,
	sscr: sscr,
	ssetmn: ssetmn,
	ssmile: ssmile,
	sstarf: sstarf,
	Star: Star,
	star: star,
	starf: starf,
	straightepsilon: straightepsilon,
	straightphi: straightphi,
	strns: strns,
	sub: sub,
	Sub: Sub,
	subdot: subdot,
	subE: subE,
	sube: sube,
	subedot: subedot,
	submult: submult,
	subnE: subnE,
	subne: subne,
	subplus: subplus,
	subrarr: subrarr,
	subset: subset,
	Subset: Subset,
	subseteq: subseteq,
	subseteqq: subseteqq,
	SubsetEqual: SubsetEqual,
	subsetneq: subsetneq,
	subsetneqq: subsetneqq,
	subsim: subsim,
	subsub: subsub,
	subsup: subsup,
	succapprox: succapprox,
	succ: succ,
	succcurlyeq: succcurlyeq,
	Succeeds: Succeeds,
	SucceedsEqual: SucceedsEqual,
	SucceedsSlantEqual: SucceedsSlantEqual,
	SucceedsTilde: SucceedsTilde,
	succeq: succeq,
	succnapprox: succnapprox,
	succneqq: succneqq,
	succnsim: succnsim,
	succsim: succsim,
	SuchThat: SuchThat,
	sum: sum,
	Sum: Sum,
	sung: sung,
	sup1: sup1,
	sup2: sup2,
	sup3: sup3,
	sup: sup,
	Sup: Sup,
	supdot: supdot,
	supdsub: supdsub,
	supE: supE,
	supe: supe,
	supedot: supedot,
	Superset: Superset,
	SupersetEqual: SupersetEqual,
	suphsol: suphsol,
	suphsub: suphsub,
	suplarr: suplarr,
	supmult: supmult,
	supnE: supnE,
	supne: supne,
	supplus: supplus,
	supset: supset,
	Supset: Supset,
	supseteq: supseteq,
	supseteqq: supseteqq,
	supsetneq: supsetneq,
	supsetneqq: supsetneqq,
	supsim: supsim,
	supsub: supsub,
	supsup: supsup,
	swarhk: swarhk,
	swarr: swarr,
	swArr: swArr,
	swarrow: swarrow,
	swnwar: swnwar,
	szlig: szlig,
	Tab: Tab,
	target: target,
	Tau: Tau,
	tau: tau,
	tbrk: tbrk,
	Tcaron: Tcaron,
	tcaron: tcaron,
	Tcedil: Tcedil,
	tcedil: tcedil,
	Tcy: Tcy,
	tcy: tcy,
	tdot: tdot,
	telrec: telrec,
	Tfr: Tfr,
	tfr: tfr,
	there4: there4,
	therefore: therefore,
	Therefore: Therefore,
	Theta: Theta,
	theta: theta,
	thetasym: thetasym,
	thetav: thetav,
	thickapprox: thickapprox,
	thicksim: thicksim,
	ThickSpace: ThickSpace,
	ThinSpace: ThinSpace,
	thinsp: thinsp,
	thkap: thkap,
	thksim: thksim,
	THORN: THORN,
	thorn: thorn,
	tilde: tilde,
	Tilde: Tilde,
	TildeEqual: TildeEqual,
	TildeFullEqual: TildeFullEqual,
	TildeTilde: TildeTilde,
	timesbar: timesbar,
	timesb: timesb,
	times: times$1,
	timesd: timesd,
	tint: tint,
	toea: toea,
	topbot: topbot,
	topcir: topcir,
	top: top,
	Topf: Topf,
	topf: topf,
	topfork: topfork,
	tosa: tosa,
	tprime: tprime,
	trade: trade,
	TRADE: TRADE,
	triangle: triangle,
	triangledown: triangledown,
	triangleleft: triangleleft,
	trianglelefteq: trianglelefteq,
	triangleq: triangleq,
	triangleright: triangleright,
	trianglerighteq: trianglerighteq,
	tridot: tridot,
	trie: trie,
	triminus: triminus,
	TripleDot: TripleDot,
	triplus: triplus,
	trisb: trisb,
	tritime: tritime,
	trpezium: trpezium,
	Tscr: Tscr,
	tscr: tscr,
	TScy: TScy,
	tscy: tscy,
	TSHcy: TSHcy,
	tshcy: tshcy,
	Tstrok: Tstrok,
	tstrok: tstrok,
	twixt: twixt,
	twoheadleftarrow: twoheadleftarrow,
	twoheadrightarrow: twoheadrightarrow,
	Uacute: Uacute,
	uacute: uacute,
	uarr: uarr,
	Uarr: Uarr,
	uArr: uArr,
	Uarrocir: Uarrocir,
	Ubrcy: Ubrcy,
	ubrcy: ubrcy,
	Ubreve: Ubreve,
	ubreve: ubreve,
	Ucirc: Ucirc,
	ucirc: ucirc,
	Ucy: Ucy,
	ucy: ucy,
	udarr: udarr,
	Udblac: Udblac,
	udblac: udblac,
	udhar: udhar,
	ufisht: ufisht,
	Ufr: Ufr,
	ufr: ufr,
	Ugrave: Ugrave,
	ugrave: ugrave,
	uHar: uHar,
	uharl: uharl,
	uharr: uharr,
	uhblk: uhblk,
	ulcorn: ulcorn,
	ulcorner: ulcorner,
	ulcrop: ulcrop,
	ultri: ultri,
	Umacr: Umacr,
	umacr: umacr,
	uml: uml,
	UnderBar: UnderBar,
	UnderBrace: UnderBrace,
	UnderBracket: UnderBracket,
	UnderParenthesis: UnderParenthesis,
	Union: Union,
	UnionPlus: UnionPlus,
	Uogon: Uogon,
	uogon: uogon,
	Uopf: Uopf,
	uopf: uopf,
	UpArrowBar: UpArrowBar,
	uparrow: uparrow,
	UpArrow: UpArrow,
	Uparrow: Uparrow,
	UpArrowDownArrow: UpArrowDownArrow,
	updownarrow: updownarrow,
	UpDownArrow: UpDownArrow,
	Updownarrow: Updownarrow,
	UpEquilibrium: UpEquilibrium,
	upharpoonleft: upharpoonleft,
	upharpoonright: upharpoonright,
	uplus: uplus,
	UpperLeftArrow: UpperLeftArrow,
	UpperRightArrow: UpperRightArrow,
	upsi: upsi,
	Upsi: Upsi,
	upsih: upsih,
	Upsilon: Upsilon,
	upsilon: upsilon,
	UpTeeArrow: UpTeeArrow,
	UpTee: UpTee,
	upuparrows: upuparrows,
	urcorn: urcorn,
	urcorner: urcorner,
	urcrop: urcrop,
	Uring: Uring,
	uring: uring,
	urtri: urtri,
	Uscr: Uscr,
	uscr: uscr,
	utdot: utdot,
	Utilde: Utilde,
	utilde: utilde,
	utri: utri,
	utrif: utrif,
	uuarr: uuarr,
	Uuml: Uuml,
	uuml: uuml,
	uwangle: uwangle,
	vangrt: vangrt,
	varepsilon: varepsilon,
	varkappa: varkappa,
	varnothing: varnothing,
	varphi: varphi,
	varpi: varpi,
	varpropto: varpropto,
	varr: varr,
	vArr: vArr,
	varrho: varrho,
	varsigma: varsigma,
	varsubsetneq: varsubsetneq,
	varsubsetneqq: varsubsetneqq,
	varsupsetneq: varsupsetneq,
	varsupsetneqq: varsupsetneqq,
	vartheta: vartheta,
	vartriangleleft: vartriangleleft,
	vartriangleright: vartriangleright,
	vBar: vBar,
	Vbar: Vbar,
	vBarv: vBarv,
	Vcy: Vcy,
	vcy: vcy,
	vdash: vdash,
	vDash: vDash,
	Vdash: Vdash,
	VDash: VDash,
	Vdashl: Vdashl,
	veebar: veebar,
	vee: vee,
	Vee: Vee,
	veeeq: veeeq,
	vellip: vellip,
	verbar: verbar,
	Verbar: Verbar,
	vert: vert,
	Vert: Vert,
	VerticalBar: VerticalBar,
	VerticalLine: VerticalLine,
	VerticalSeparator: VerticalSeparator,
	VerticalTilde: VerticalTilde,
	VeryThinSpace: VeryThinSpace,
	Vfr: Vfr,
	vfr: vfr,
	vltri: vltri,
	vnsub: vnsub,
	vnsup: vnsup,
	Vopf: Vopf,
	vopf: vopf,
	vprop: vprop,
	vrtri: vrtri,
	Vscr: Vscr,
	vscr: vscr,
	vsubnE: vsubnE,
	vsubne: vsubne,
	vsupnE: vsupnE,
	vsupne: vsupne,
	Vvdash: Vvdash,
	vzigzag: vzigzag,
	Wcirc: Wcirc,
	wcirc: wcirc,
	wedbar: wedbar,
	wedge: wedge,
	Wedge: Wedge,
	wedgeq: wedgeq,
	weierp: weierp,
	Wfr: Wfr,
	wfr: wfr,
	Wopf: Wopf,
	wopf: wopf,
	wp: wp,
	wr: wr,
	wreath: wreath,
	Wscr: Wscr,
	wscr: wscr,
	xcap: xcap,
	xcirc: xcirc,
	xcup: xcup,
	xdtri: xdtri,
	Xfr: Xfr,
	xfr: xfr,
	xharr: xharr,
	xhArr: xhArr,
	Xi: Xi,
	xi: xi,
	xlarr: xlarr,
	xlArr: xlArr,
	xmap: xmap,
	xnis: xnis,
	xodot: xodot,
	Xopf: Xopf,
	xopf: xopf,
	xoplus: xoplus,
	xotime: xotime,
	xrarr: xrarr,
	xrArr: xrArr,
	Xscr: Xscr,
	xscr: xscr,
	xsqcup: xsqcup,
	xuplus: xuplus,
	xutri: xutri,
	xvee: xvee,
	xwedge: xwedge,
	Yacute: Yacute,
	yacute: yacute,
	YAcy: YAcy,
	yacy: yacy,
	Ycirc: Ycirc,
	ycirc: ycirc,
	Ycy: Ycy,
	ycy: ycy,
	yen: yen,
	Yfr: Yfr,
	yfr: yfr,
	YIcy: YIcy,
	yicy: yicy,
	Yopf: Yopf,
	yopf: yopf,
	Yscr: Yscr,
	yscr: yscr,
	YUcy: YUcy,
	yucy: yucy,
	yuml: yuml,
	Yuml: Yuml,
	Zacute: Zacute,
	zacute: zacute,
	Zcaron: Zcaron,
	zcaron: zcaron,
	Zcy: Zcy,
	zcy: zcy,
	Zdot: Zdot,
	zdot: zdot,
	zeetrf: zeetrf,
	ZeroWidthSpace: ZeroWidthSpace,
	Zeta: Zeta,
	zeta: zeta,
	zfr: zfr,
	Zfr: Zfr,
	ZHcy: ZHcy,
	zhcy: zhcy,
	zigrarr: zigrarr,
	zopf: zopf,
	Zopf: Zopf,
	Zscr: Zscr,
	zscr: zscr,
	zwj: zwj,
	zwnj: zwnj,
	default: entitiesJSON
});

var require$$0$1 = ( xml && xmlJSON ) || xml;

var require$$1 = ( entities && entitiesJSON ) || entities;

var inverseXML = getInverseObj(require$$0$1);
var xmlReplacer = getInverseReplacer(inverseXML);

var XML = getInverse(inverseXML, xmlReplacer);

var inverseHTML = getInverseObj(require$$1);
var htmlReplacer = getInverseReplacer(inverseHTML);

var HTML = getInverse(inverseHTML, htmlReplacer);

function getInverseObj(obj){
	return Object.keys(obj).sort().reduce(function(inverse, name){
		inverse[obj[name]] = "&" + name + ";";
		return inverse;
	}, {});
}

function getInverseReplacer(inverse){
	var single = [],
	    multiple = [];

	Object.keys(inverse).forEach(function(k){
		if(k.length === 1){
			single.push("\\" + k);
		} else {
			multiple.push(k);
		}
	});

	
	multiple.unshift("[" + single.join("") + "]");

	return new RegExp(multiple.join("|"), "g");
}

var re_nonASCII = /[^\0-\x7F]/g;
var re_astralSymbols = /[\uD800-\uDBFF][\uDC00-\uDFFF]/g;

function singleCharReplacer(c){
	return "&#x" + c.charCodeAt(0).toString(16).toUpperCase() + ";";
}

function astralReplacer(c){
	
	var high = c.charCodeAt(0);
	var low  = c.charCodeAt(1);
	var codePoint = (high - 0xD800) * 0x400 + low - 0xDC00 + 0x10000;
	return "&#x" + codePoint.toString(16).toUpperCase() + ";";
}

function getInverse(inverse, re){
	function func(name){
		return inverse[name];
	}

	return function(data){
		return data
				.replace(re, func)
				.replace(re_astralSymbols, astralReplacer)
				.replace(re_nonASCII, singleCharReplacer);
	};
}

var re_xmlChars = getInverseReplacer(inverseXML);

function escapeXML(data){
	return data
			.replace(re_xmlChars, singleCharReplacer)
			.replace(re_astralSymbols, astralReplacer)
			.replace(re_nonASCII, singleCharReplacer);
}

var escape = escapeXML;

var encode = {
	XML: XML,
	HTML: HTML,
	escape: escape
};

var decode = {
	"0": 65533,
	"128": 8364,
	"130": 8218,
	"131": 402,
	"132": 8222,
	"133": 8230,
	"134": 8224,
	"135": 8225,
	"136": 710,
	"137": 8240,
	"138": 352,
	"139": 8249,
	"140": 338,
	"142": 381,
	"145": 8216,
	"146": 8217,
	"147": 8220,
	"148": 8221,
	"149": 8226,
	"150": 8211,
	"151": 8212,
	"152": 732,
	"153": 8482,
	"154": 353,
	"155": 8250,
	"156": 339,
	"158": 382,
	"159": 376
};

var decode$1 = Object.freeze({
	default: decode
});

var require$$0$1$1 = ( decode$1 && decode ) || decode$1;

var decodeMap = require$$0$1$1;

var decode_codepoint = decodeCodePoint;


function decodeCodePoint(codePoint){

	if((codePoint >= 0xD800 && codePoint <= 0xDFFF) || codePoint > 0x10FFFF){
		return "\uFFFD";
	}

	if(codePoint in decodeMap){
		codePoint = decodeMap[codePoint];
	}

	var output = "";

	if(codePoint > 0xFFFF){
		codePoint -= 0x10000;
		output += String.fromCharCode(codePoint >>> 10 & 0x3FF | 0xD800);
		codePoint = 0xDC00 | codePoint & 0x3FF;
	}

	output += String.fromCharCode(codePoint);
	return output;
}

var Aacute$1 = "Á";
var aacute$1 = "á";
var Acirc$1 = "Â";
var acirc$1 = "â";
var acute$1 = "´";
var AElig$1 = "Æ";
var aelig$1 = "æ";
var Agrave$1 = "À";
var agrave$1 = "à";
var amp$2 = "&";
var AMP$1 = "&";
var Aring$1 = "Å";
var aring$1 = "å";
var Atilde$1 = "Ã";
var atilde$1 = "ã";
var Auml$1 = "Ä";
var auml$1 = "ä";
var brvbar$1 = "¦";
var Ccedil$1 = "Ç";
var ccedil$1 = "ç";
var cedil$1 = "¸";
var cent$1 = "¢";
var copy$1 = "©";
var COPY$1 = "©";
var curren$1 = "¤";
var deg$1 = "°";
var divide$1 = "÷";
var Eacute$1 = "É";
var eacute$1 = "é";
var Ecirc$1 = "Ê";
var ecirc$1 = "ê";
var Egrave$1 = "È";
var egrave$1 = "è";
var ETH$1 = "Ð";
var eth$1 = "ð";
var Euml$1 = "Ë";
var euml$1 = "ë";
var frac12$1 = "½";
var frac14$1 = "¼";
var frac34$1 = "¾";
var gt$2 = ">";
var GT$1 = ">";
var Iacute$1 = "Í";
var iacute$1 = "í";
var Icirc$1 = "Î";
var icirc$1 = "î";
var iexcl$1 = "¡";
var Igrave$1 = "Ì";
var igrave$1 = "ì";
var iquest$1 = "¿";
var Iuml$1 = "Ï";
var iuml$1 = "ï";
var laquo$1 = "«";
var lt$2 = "<";
var LT$1 = "<";
var macr$1 = "¯";
var micro$1 = "µ";
var middot$1 = "·";
var nbsp$1 = " ";
var not$1 = "¬";
var Ntilde$1 = "Ñ";
var ntilde$1 = "ñ";
var Oacute$1 = "Ó";
var oacute$1 = "ó";
var Ocirc$1 = "Ô";
var ocirc$1 = "ô";
var Ograve$1 = "Ò";
var ograve$1 = "ò";
var ordf$1 = "ª";
var ordm$1 = "º";
var Oslash$1 = "Ø";
var oslash$1 = "ø";
var Otilde$1 = "Õ";
var otilde$1 = "õ";
var Ouml$1 = "Ö";
var ouml$1 = "ö";
var para$1 = "¶";
var plusmn$1 = "±";
var pound$1 = "£";
var quot$2 = "\"";
var QUOT$1 = "\"";
var raquo$1 = "»";
var reg$1 = "®";
var REG$1 = "®";
var sect$1 = "§";
var shy$1 = "­";
var sup1$1 = "¹";
var sup2$1 = "²";
var sup3$1 = "³";
var szlig$1 = "ß";
var THORN$1 = "Þ";
var thorn$1 = "þ";
var times$1$1 = "×";
var Uacute$1 = "Ú";
var uacute$1 = "ú";
var Ucirc$1 = "Û";
var ucirc$1 = "û";
var Ugrave$1 = "Ù";
var ugrave$1 = "ù";
var uml$1 = "¨";
var Uuml$1 = "Ü";
var uuml$1 = "ü";
var Yacute$1 = "Ý";
var yacute$1 = "ý";
var yen$1 = "¥";
var yuml$1 = "ÿ";
var legacyJSON = {
	Aacute: Aacute$1,
	aacute: aacute$1,
	Acirc: Acirc$1,
	acirc: acirc$1,
	acute: acute$1,
	AElig: AElig$1,
	aelig: aelig$1,
	Agrave: Agrave$1,
	agrave: agrave$1,
	amp: amp$2,
	AMP: AMP$1,
	Aring: Aring$1,
	aring: aring$1,
	Atilde: Atilde$1,
	atilde: atilde$1,
	Auml: Auml$1,
	auml: auml$1,
	brvbar: brvbar$1,
	Ccedil: Ccedil$1,
	ccedil: ccedil$1,
	cedil: cedil$1,
	cent: cent$1,
	copy: copy$1,
	COPY: COPY$1,
	curren: curren$1,
	deg: deg$1,
	divide: divide$1,
	Eacute: Eacute$1,
	eacute: eacute$1,
	Ecirc: Ecirc$1,
	ecirc: ecirc$1,
	Egrave: Egrave$1,
	egrave: egrave$1,
	ETH: ETH$1,
	eth: eth$1,
	Euml: Euml$1,
	euml: euml$1,
	frac12: frac12$1,
	frac14: frac14$1,
	frac34: frac34$1,
	gt: gt$2,
	GT: GT$1,
	Iacute: Iacute$1,
	iacute: iacute$1,
	Icirc: Icirc$1,
	icirc: icirc$1,
	iexcl: iexcl$1,
	Igrave: Igrave$1,
	igrave: igrave$1,
	iquest: iquest$1,
	Iuml: Iuml$1,
	iuml: iuml$1,
	laquo: laquo$1,
	lt: lt$2,
	LT: LT$1,
	macr: macr$1,
	micro: micro$1,
	middot: middot$1,
	nbsp: nbsp$1,
	not: not$1,
	Ntilde: Ntilde$1,
	ntilde: ntilde$1,
	Oacute: Oacute$1,
	oacute: oacute$1,
	Ocirc: Ocirc$1,
	ocirc: ocirc$1,
	Ograve: Ograve$1,
	ograve: ograve$1,
	ordf: ordf$1,
	ordm: ordm$1,
	Oslash: Oslash$1,
	oslash: oslash$1,
	Otilde: Otilde$1,
	otilde: otilde$1,
	Ouml: Ouml$1,
	ouml: ouml$1,
	para: para$1,
	plusmn: plusmn$1,
	pound: pound$1,
	quot: quot$2,
	QUOT: QUOT$1,
	raquo: raquo$1,
	reg: reg$1,
	REG: REG$1,
	sect: sect$1,
	shy: shy$1,
	sup1: sup1$1,
	sup2: sup2$1,
	sup3: sup3$1,
	szlig: szlig$1,
	THORN: THORN$1,
	thorn: thorn$1,
	times: times$1$1,
	Uacute: Uacute$1,
	uacute: uacute$1,
	Ucirc: Ucirc$1,
	ucirc: ucirc$1,
	Ugrave: Ugrave$1,
	ugrave: ugrave$1,
	uml: uml$1,
	Uuml: Uuml$1,
	uuml: uuml$1,
	Yacute: Yacute$1,
	yacute: yacute$1,
	yen: yen$1,
	yuml: yuml$1
};

var _entities = {
  encodeXML: encode.XML,
  decodeCodepoint: decode_codepoint,
  entitiesJSON,
  legacyJSON,
  xmlJSON
};

function createCommonjsModule(fn, module) {
	return module = { exports: {} }, fn(module, module.exports), module.exports;
}

var ElementType$1 = index;
var entities$1 = _entities;

const booleanAttributes = {
  __proto__: null,
  allowfullscreen: true,
  async: true,
  autofocus: true,
  autoplay: true,
  checked: true,
  controls: true,
  default: true,
  defer: true,
  disabled: true,
  hidden: true,
  ismap: true,
  loop: true,
  multiple: true,
  muted: true,
  open: true,
  readonly: true,
  required: true,
  reversed: true,
  scoped: true,
  seamless: true,
  selected: true,
  typemustmatch: true
};

const unencodedElements = {
  __proto__: null,
  style: true,
  script: true,
  xmp: true,
  iframe: true,
  noembed: true,
  noframes: true,
  plaintext: true,
  noscript: true
};

const singleTag = {
  __proto__: null,
  area: true,
  base: true,
  basefont: true,
  br: true,
  col: true,
  command: true,
  embed: true,
  frame: true,
  hr: true,
  img: true,
  input: true,
  isindex: true,
  keygen: true,
  link: true,
  meta: true,
  param: true,
  source: true,
  track: true,
  wbr: true,
};

function formatAttribs(el, opts) {
  let output = [];
  const attributes = el.attributes;
  
  attributes.forEach((value, key) => {
    if (!value && booleanAttributes[key]) {
      output.push(key);
    } else {
      output.push(key + '="' + (opts.decodeEntities ? entities$1.encodeXML(value) : value) + '"');
    }
  });
  return output.join(' ')
}

function render(dom, opts) {
  if (!Array.isArray(dom)) dom = [dom];
  opts = opts || {};

  let output = [];

  for(var i = 0; i < dom.length; i++){
    let elem = dom[i];

    if (elem.type === 'root') {
      output.push(render(elem.childNodes, opts));
    } else if (ElementType$1.isTag(elem)) {
      output.push(renderTag(elem, opts));
    } else if (elem.type === ElementType$1.Directive) {
      output.push(renderDirective(elem));
    } else if (elem.type === ElementType$1.Comment) {
      output.push(renderComment(elem));
    } else if (elem.type === ElementType$1.CDATA) {
      output.push(renderCdata(elem));
    } else {
      output.push(renderText(elem, opts));
    }
  }

  return output.join('')
}

function renderTag(elem, opts) {
  
  if (elem.name === "svg") opts = {decodeEntities: opts.decodeEntities, xmlMode: true};

  let tag = '<' + elem.name;
  let attribs = formatAttribs(elem, opts);

  if (attribs) {
    tag += ' ' + attribs;
  }

  if (
    opts.xmlMode
    && (!elem.childNodes || elem.childNodes.length === 0)
  ) {
    tag += '/>';
  } else {
    tag += '>';
    if (elem.childNodes) {
      tag += render(elem.childNodes, opts);
    }

    if (!singleTag[elem.name] || opts.xmlMode) {
      tag += '</' + elem.name + '>';
    }
  }

  return tag
}

function renderDirective(elem) {
  return '<' + elem.data + '>'
}

function renderText(elem, opts) {
  var data = elem.data || '';
  
  if (opts.decodeEntities && !(elem.parent && elem.parent.name in unencodedElements)) {
    data = entities$1.encodeXML(data);
  }
  return data
}

function renderCdata(elem) {
  return '<![CDATA[' + elem.childNodes[0].data + ']]>'
}

function renderComment(elem) {
  return '<!--' + elem.data + '-->'
}

var serialize$1 = render;

let ElementType$1$1 = index;
let serialize = serialize$1;
var stringify = {
	getInnerHTML: getInnerHTML,
	getOuterHTML: serialize,
	getText: getText
};

function getInnerHTML(elem, opts){
	return elem.childNodes ? elem.childNodes.map(function(elem){
		return serialize(elem, opts);
	}).join("") : "";
}

function getText(elem){
	if(Array.isArray(elem)) return elem.map(getText).join("");
	switch(elem.type) {
		case ElementType$1$1.Tag:
		case ElementType$1$1.Script:
		case ElementType$1$1.Style:
			return getText(elem.childNodes)
		case ElementType$1$1.Text:
		case ElementType$1$1.Comment:
		case ElementType$1$1.CDATA:
			return elem.data
		default:
			return ""
	}
}

var traversal = createCommonjsModule(function (module, exports) {
var getChildren = exports.getChildren = function(elem){
	return elem.childNodes;
};

var getParent = exports.getParent = function(elem){
	return elem.parent;
};

exports.getSiblings = function(elem){
	var parent = getParent(elem);
	return parent ? getChildren(parent) : [elem];
};

exports.getAttributeValue = function(elem, name){
	return elem.getAttribute(name);
};

exports.hasAttrib = function(elem, name){
	return elem.hasAttribute(name);
};

exports.getName = function(elem){
	return elem.name
};

exports.getNameWithoutNS = function(elem){
  return elem.nameWithoutNS
};
});

var manipulation = createCommonjsModule(function (module, exports) {
function removeElement(elem){
	if(elem.prev) elem.prev.next = elem.next;
	if(elem.next) elem.next.prev = elem.prev;
	if(elem.parent){
		var childs = elem.parent.childNodes;
		let pos = childs.lastIndexOf(elem);
		if (pos < 0) throw new Error('Invalid state')
		childs.splice(pos, 1);
		elem.parent = null;
	}
}

function replaceElement(elem, replacement){
	if (replacement.parent) exports.removeElement(replacement);
	var prev = replacement.prev = elem.prev;
	if(prev){
		prev.next = replacement;
	}

	var next = replacement.next = elem.next;
	if(next){
		next.prev = replacement;
	}

	var parent = replacement.parent = elem.parent;
	if(parent){
		var childs = parent.childNodes;
		let pos = childs.lastIndexOf(elem);
		if (pos < 0) throw new Error('Invalid state')
		childs[pos] = replacement;
	}
}

function appendChild(elem, child){
	if (child.parent) removeElement(child);
	child.parent = elem;

	if(elem.childNodes.push(child) !== 1){
		var sibling = elem.childNodes[elem.childNodes.length - 2];
		sibling.next = child;
		child.prev = sibling;
		child.next = null;
	}
}

function append(elem, next){
	if (next.parent) removeElement(next);
	var parent = elem.parent,
		currNext = elem.next;

	next.next = currNext;
	next.prev = elem;
	elem.next = next;
	next.parent = parent;

	if(currNext){
		currNext.prev = next;
		if(parent){
			var childs = parent.childNodes;
			let pos = childs.lastIndexOf(currNext);
			if (pos < 0) throw new Error('Invalid state')
			childs.splice(pos, 0, next);
		}
	} else if(parent){
		parent.childNodes.push(next);
	}
}

function prepend(elem, prev){
	if (prev.parent) removeElement(prev);
	var parent = elem.parent;
	if(parent){
		var childs = parent.childNodes;
		let pos = childs.lastIndexOf(elem);
		if (pos < 0) throw new Error('Invalid state')
		childs.splice(pos, 0, prev);
	}

	if(elem.prev){
		elem.prev.next = prev;
	}

	prev.parent = parent;
	prev.prev = elem.prev;
	prev.next = elem;
	elem.prev = prev;
}

exports.removeElement = removeElement;
exports.replaceElement = replaceElement;
exports.appendChild = appendChild;
exports.append = append;
exports.prepend = prepend;
});

var isTag$1 = index.isTag;

var querying = {
	filter: filter$1,
	find: find$1,
	findOneChild: findOneChild,
	findOne: findOne,
	existsOne: existsOne,
	findAll: findAll
};

function filter$1(test, element, recurse, limit){
	if(!Array.isArray(element)) element = [element];

	if(typeof limit !== "number" || !isFinite(limit)){
		limit = Infinity;
	}
	return find$1(test, element, recurse !== false, limit);
}

function find$1(test, elems, recurse, limit){
	var result = [], childs;

	for(var i = 0, j = elems.length; i < j; i++){
		if(test(elems[i])){
			result.push(elems[i]);
			if(--limit <= 0) break;
		}

		childs = elems[i].childNodes;
		if(recurse && childs && childs.length > 0){
			childs = find$1(test, childs, recurse, limit);
			result = result.concat(childs);
			limit -= childs.length;
			if(limit <= 0) break;
		}
	}

	return result;
}

function findOneChild(test, elems){
	for(var i = 0, l = elems.length; i < l; i++){
		if(test(elems[i])) return elems[i];
	}

	return null;
}

function findOne(test, elems){
	var elem = null;

	for(var i = 0, l = elems.length; i < l && !elem; i++){
		const child = elems[i];
		if(!isTag$1(child)){
			continue;
		} else if(test(child)){
			elem = child;
		} else if(child.childNodes.length > 0){
			elem = findOne(test, child.childNodes);
		}
	}

	return elem;
}

function existsOne(test, elems){
	for(var i = 0, l = elems.length; i < l; i++){
		if(
			isTag$1(elems[i]) && (
				test(elems[i]) || (
					elems[i].childNodes.length > 0 &&
					existsOne(test, elems[i].childNodes)
				)
			)
		){
			return true;
		}
	}

	return false;
}

function findAll(test, elems){
	var result = [];
	for(var i = 0, j = elems.length; i < j; i++){
		if(!isTag$1(elems[i])) continue;
		if(test(elems[i])) result.push(elems[i]);

		if(elems[i].childNodes.length > 0){
			result = result.concat(findAll(test, elems[i].childNodes));
		}
	}
	return result;
}

var legacy = createCommonjsModule(function (module, exports) {
var ElementType = index;
var isTag = exports.isTag = ElementType.isTag;

exports.testElement = function(options, element){
	for(var key in options){
		if(!options.hasOwnProperty(key));
		else if(key === "tag_name"){
			if(!isTag(element) || !options.tag_name(element.name)){
				return false;
			}
		} else if(key === "tag_type"){
			if(!options.tag_type(element.type)) return false;
		} else if(key === "tag_contains"){
			if(isTag(element) || !options.tag_contains(element.data)){
				return false;
			}
		} else if(!element.attributes || !options[key](element.getAttribute(key))) {
			return false;
		}
	}
	return true;
};

var Checks = {
	tag_name: function(name){
		if(typeof name === "function"){
			return function(elem){ return isTag(elem) && name(elem.name); };
		} else if(name === "*"){
			return isTag;
		} else {
			return function(elem){ return isTag(elem) && elem.name === name; };
		}
	},
	tag_type: function(type){
		if(typeof type === "function"){
			return function(elem){ return type(elem.type); };
		} else {
			return function(elem){ return elem.type === type; };
		}
	},
	tag_contains: function(data){
		if(typeof data === "function"){
			return function(elem){ return !isTag(elem) && data(elem.data); };
		} else {
			return function(elem){ return !isTag(elem) && elem.data === data; };
		}
	}
};

function getAttribCheck(attrib, value){
	if(typeof value === "function"){
		return function(elem){ return value(elem.getAttribute(attrib)); };
	} else {
		return function(elem){ return elem.getAttribute(attrib) === value; };
	}
}

function combineFuncs(a, b){
	return function(elem){
		return a(elem) || b(elem);
	};
}

exports.getElements = function(options, element, recurse, limit){
	var funcs = Object.keys(options).map(function(key){
		var value = options[key];
		return key in Checks ? Checks[key](value) : getAttribCheck(key, value);
	});

	return funcs.length === 0 ? [] : this.filter(
		funcs.reduce(combineFuncs),
		element, recurse, limit
	);
};

exports.getElementById = function(id, element, recurse){
	if(!Array.isArray(element)) element = [element];
	return this.findOne(getAttribCheck("id", id), element, recurse !== false);
};

exports.getElementsByTagName = function(name, element, recurse, limit){
	return this.filter(Checks.tag_name(name), element, recurse, limit);
};

exports.getElementsByTagType = function(type, element, recurse, limit){
	return this.filter(Checks.tag_type(type), element, recurse, limit);
};
});

var helpers = createCommonjsModule(function (module, exports) {


exports.removeSubsets = function(nodes) {
	var idx = nodes.length, node, ancestor, replace;

	
	
	while (--idx > -1) {
		node = ancestor = nodes[idx];

		
		nodes[idx] = null;
		replace = true;

		while (ancestor) {
			if (nodes.indexOf(ancestor) > -1) {
				replace = false;
				nodes.splice(idx, 1);
				break;
			}
			ancestor = ancestor.parent;
		}

		
		if (replace) {
			nodes[idx] = node;
		}
	}

	return nodes;
};


var POSITION = {
	DISCONNECTED: 1,
	PRECEDING: 2,
	FOLLOWING: 4,
	CONTAINS: 8,
	CONTAINED_BY: 16
};























var comparePos = exports.compareDocumentPosition = function(nodeA, nodeB) {
	var aParents = [];
	var bParents = [];
	var current, sharedParent, siblings, aSibling, bSibling, idx;

	if (nodeA === nodeB) {
		return 0;
	}

	current = nodeA;
	while (current) {
		aParents.unshift(current);
		current = current.parent;
	}
	current = nodeB;
	while (current) {
		bParents.unshift(current);
		current = current.parent;
	}

	idx = 0;
	while (aParents[idx] === bParents[idx]) {
		idx++;
	}

	if (idx === 0) {
		return POSITION.DISCONNECTED;
	}

	sharedParent = aParents[idx - 1];
	siblings = sharedParent.childNodes;
	aSibling = aParents[idx];
	bSibling = bParents[idx];

	if (siblings.indexOf(aSibling) > siblings.indexOf(bSibling)) {
		if (sharedParent === nodeB) {
			return POSITION.FOLLOWING | POSITION.CONTAINED_BY;
		}
		return POSITION.FOLLOWING;
	} else {
		if (sharedParent === nodeA) {
			return POSITION.PRECEDING | POSITION.CONTAINS;
		}
		return POSITION.PRECEDING;
	}
};








exports.uniqueSort = function(nodes) {
	var idx = nodes.length, node, position;

	nodes = nodes.slice();

	while (--idx > -1) {
		node = nodes[idx];
		position = nodes.indexOf(node);
		if (position > -1 && position < idx) {
			nodes.splice(idx, 1);
		}
	}
	nodes.sort(function(a, b) {
		var relative = comparePos(a, b);
		if (relative & POSITION.PRECEDING) {
			return -1;
		} else if (relative & POSITION.FOLLOWING) {
			return 1;
		}
		return 0;
	});

	return nodes;
};
});

var index$2 = createCommonjsModule(function (module) {
var DomUtils = module.exports;

[
	stringify,
	traversal,
	manipulation,
	querying,
	legacy,
	helpers
].forEach(function(ext){
	Object.keys(ext).forEach(function(key){
		DomUtils[key] = ext[key].bind(DomUtils);
	});
});
});

var index$3 = {
	trueFunc: function trueFunc(){
		return true;
	},
	falseFunc: function falseFunc(){
		return false;
	}
};

var index$4 = parse$1;

var re_name = /^(?:\\.|[\w\-\u00c0-\uFFFF])+/;
var re_escape = /\\([\da-f]{1,6}\s?|(\s)|.)/ig;
var re_attr = /^\s*((?:\\.|[\w\u00c0-\uFFFF\-])+)\s*(?:(\S?)=\s*(?:(['"])(.*?)\3|(#?(?:\\.|[\w\u00c0-\uFFFF\-])*)|)|)\s*(i)?\]/;

var actionTypes = {
	__proto__: null,
	"undefined": "exists",
	"":  "equals",
	"~": "element",
	"^": "start",
	"$": "end",
	"*": "any",
	"!": "not",
	"|": "hyphen"
};

var simpleSelectors = {
	__proto__: null,
	">": "child",
	"<": "parent",
	"~": "sibling",
	"+": "adjacent"
};

var attribSelectors = {
	__proto__: null,
	"#": ["id", "equals"],
	".": ["class", "element"]
};

//pseudos, whose data-property is parsed as well
var unpackPseudos = {
	__proto__: null,
	"has": true,
	"not": true,
	"matches": true
};

var stripQuotesFromPseudos = {
	__proto__: null,
	"contains": true,
	"icontains": true
};

var quotes = {
	__proto__: null,
	"\"": true,
	"'": true
};

//unescape function taken from https://github.com/jquery/sizzle/blob/master/src/sizzle.js#L139
function funescape( _, escaped, escapedWhitespace ) {
	var high = "0x" + escaped - 0x10000;
	// NaN means non-codepoint
	// Support: Firefox
	// Workaround erroneous numeric interpretation of +"0x"
	return high !== high || escapedWhitespace ?
		escaped :
		// BMP codepoint
		high < 0 ?
			String.fromCharCode( high + 0x10000 ) :
			// Supplemental Plane codepoint (surrogate pair)
			String.fromCharCode( high >> 10 | 0xD800, high & 0x3FF | 0xDC00 );
}

function unescapeCSS(str){
	return str.replace(re_escape, funescape);
}

function isWhitespace(c){
	return c === " " || c === "\n" || c === "\t" || c === "\f" || c === "\r";
}

function parse$1(selector, options){
	var subselects = [];

	selector = parseSelector(subselects, selector + "", options);

	if(selector !== ""){
		throw new SyntaxError("Unmatched selector: " + selector);
	}

	return subselects;
}

function parseSelector(subselects, selector, options){
	var tokens = [],
		sawWS = false,
		data, firstChar, name, quot;

	function getName(){
		var sub = selector.match(re_name)[0];
		selector = selector.substr(sub.length);
		return unescapeCSS(sub);
	}

	function stripWhitespace(start){
		while(isWhitespace(selector.charAt(start))) start++;
		selector = selector.substr(start);
	}

	stripWhitespace(0);

	while(selector !== ""){
		firstChar = selector.charAt(0);

		if(isWhitespace(firstChar)){
			sawWS = true;
			stripWhitespace(1);
		} else if(firstChar in simpleSelectors){
			tokens.push({type: simpleSelectors[firstChar]});
			sawWS = false;

			stripWhitespace(1);
		} else if(firstChar === ","){
			if(tokens.length === 0){
				throw new SyntaxError("empty sub-selector");
			}
			subselects.push(tokens);
			tokens = [];
			sawWS = false;
			stripWhitespace(1);
		} else {
			if(sawWS){
				if(tokens.length > 0){
					tokens.push({type: "descendant"});
				}
				sawWS = false;
			}

			if(firstChar === "*"){
				selector = selector.substr(1);
				tokens.push({type: "universal"});
			} else if(firstChar in attribSelectors){
				selector = selector.substr(1);
				tokens.push({
					type: "attribute",
					name: attribSelectors[firstChar][0],
					action: attribSelectors[firstChar][1],
					value: getName(),
					ignoreCase: false
				});
			} else if(firstChar === "["){
				selector = selector.substr(1);
				data = selector.match(re_attr);
				if(!data){
					throw new SyntaxError("Malformed attribute selector: " + selector);
				}
				selector = selector.substr(data[0].length);
				name = unescapeCSS(data[1]);

				if(
					!options || (
						"lowerCaseAttributeNames" in options ?
							options.lowerCaseAttributeNames :
							!options.xmlMode
					)
				){
					name = name.toLowerCase();
				}

				tokens.push({
					type: "attribute",
					name: name,
					action: actionTypes[data[2]],
					value: unescapeCSS(data[4] || data[5] || ""),
					ignoreCase: !!data[6]
				});

			} else if(firstChar === ":"){
				if(selector.charAt(1) === ":"){
					selector = selector.substr(2);
					tokens.push({type: "pseudo-element", name: getName().toLowerCase()});
					continue;
				}

				selector = selector.substr(1);

				name = getName().toLowerCase();
				data = null;

				if(selector.charAt(0) === "("){
					if(name in unpackPseudos){
						quot = selector.charAt(1);
						var quoted = quot in quotes;

						selector = selector.substr(quoted + 1);

						data = [];
						selector = parseSelector(data, selector, options);

						if(quoted){
							if(selector.charAt(0) !== quot){
								throw new SyntaxError("unmatched quotes in :" + name);
							} else {
								selector = selector.substr(1);
							}
						}

						if(selector.charAt(0) !== ")"){
							throw new SyntaxError("missing closing parenthesis in :" + name + " " + selector);
						}

						selector = selector.substr(1);
					} else {
						var pos = 1, counter = 1;

						for(; counter > 0 && pos < selector.length; pos++){
							if(selector.charAt(pos) === "(") counter++;
							else if(selector.charAt(pos) === ")") counter--;
						}

						if(counter){
							throw new SyntaxError("parenthesis not matched");
						}

						data = selector.substr(1, pos - 2);
						selector = selector.substr(pos);

						if(name in stripQuotesFromPseudos){
							quot = data.charAt(0);

							if(quot === data.slice(-1) && quot in quotes){
								data = data.slice(1, -1);
							}

							data = unescapeCSS(data);
						}
					}
				}

				tokens.push({type: "pseudo", name: name, data: data});
			} else if(re_name.test(selector)){
				name = getName();

				if(!options || ("lowerCaseTags" in options ? options.lowerCaseTags : !options.xmlMode)){
					name = name.toLowerCase();
				}

				tokens.push({type: "tag", name: name});
			} else {
				if(tokens.length && tokens[tokens.length - 1].type === "descendant"){
					tokens.pop();
				}
				addToken(subselects, tokens);
				return selector;
			}
		}
	}

	addToken(subselects, tokens);

	return selector;
}

function addToken(subselects, tokens){
	if(subselects.length > 0 && tokens.length === 0){
		throw new SyntaxError("empty sub-selector");
	}

	subselects.push(tokens);
}

var parse_1$1 = parse$2;

//following http://www.w3.org/TR/css3-selectors/#nth-child-pseudo

//[ ['-'|'+']? INTEGER? {N} [ S* ['-'|'+'] S* INTEGER ]?
var re_nthElement = /^([+\-]?\d*n)?\s*(?:([+\-]?)\s*(\d+))?$/;

/*
	parses a nth-check formula, returns an array of two numbers
*/
function parse$2(formula){
	formula = formula.trim().toLowerCase();

	if(formula === "even"){
		return [2, 0];
	} else if(formula === "odd"){
		return [2, 1];
	} else {
		var parsed = formula.match(re_nthElement);

		if(!parsed){
			throw new SyntaxError("n-th rule couldn't be parsed ('" + formula + "')");
		}

		var a;

		if(parsed[1]){
			a = parseInt(parsed[1], 10);
			if(isNaN(a)){
				if(parsed[1].charAt(0) === "-") a = -1;
				else a = 1;
			}
		} else a = 0;

		return [
			a,
			parsed[3] ? parseInt((parsed[2] || "") + parsed[3], 10) : 0
		];
	}
}

var compile_1$1 = compile$1;

var trueFunc$1$1  = index$3.trueFunc;
var falseFunc$1$1 = index$3.falseFunc;

/*
	returns a function that checks if an elements index matches the given rule
	highly optimized to return the fastest solution
*/
function compile$1(parsed){
	var a = parsed[0],
	    b = parsed[1] - 1;

	//when b <= 0, a*n won't be possible for any matches when a < 0
	//besides, the specification says that no element is matched when a and b are 0
	if(b < 0 && a <= 0) return falseFunc$1$1;

	//when a is in the range -1..1, it matches any element (so only b is checked)
	if(a ===-1) return function(pos){ return pos <= b; };
	if(a === 0) return function(pos){ return pos === b; };
	//when b <= 0 and a === 1, they match any element
	if(a === 1) return b < 0 ? trueFunc$1$1 : function(pos){ return pos >= b; };

	//when a > 0, modulo can be used to check if there is a match
	var bMod = b % a;
	if(bMod < 0) bMod += a;

	if(a > 1){
		return function(pos){
			return pos >= b && pos % a === bMod;
		};
	}

	a *= -1; //make `a` positive

	return function(pos){
		return pos <= b && pos % a === bMod;
	};
}

var index$5 = function nthCheck(formula){
	return compile_1$1(parse_1$1(formula));
};

var parse_1 = parse_1$1;
var compile_1 = compile_1$1;

index$5.parse = parse_1;
index$5.compile = compile_1;

var universal = 50;
var tag = 30;
var attribute = 1;
var pseudo = 0;
var descendant = -1;
var child = -1;
var parent$1 = -1;
var sibling = -1;
var adjacent = -1;
var procedure$2 = {
	universal: universal,
	tag: tag,
	attribute: attribute,
	pseudo: pseudo,
	descendant: descendant,
	child: child,
	parent: parent$1,
	sibling: sibling,
	adjacent: adjacent
};

var procedure$3 = Object.freeze({
	universal: universal,
	tag: tag,
	attribute: attribute,
	pseudo: pseudo,
	descendant: descendant,
	child: child,
	parent: parent$1,
	sibling: sibling,
	adjacent: adjacent,
	default: procedure$2
});

var require$$0 = ( procedure$3 && procedure$2 ) || procedure$3;

var sort = sortByProcedure;



var procedure$1 = require$$0;

var attributes = {
	__proto__: null,
	exists: 10,
	equals: 8,
	not: 7,
	start: 6,
	end: 6,
	any: 5,
	hyphen: 4,
	element: 4
};

function sortByProcedure(arr){
	var procs = arr.map(getProcedure);
	for(var i = 1; i < arr.length; i++){
		var procNew = procs[i];

		if(procNew < 0) continue;

		for(var j = i - 1; j >= 0 && procNew < procs[j]; j--){
			var token = arr[j + 1];
			arr[j + 1] = arr[j];
			arr[j] = token;
			procs[j + 1] = procs[j];
			procs[j] = procNew;
		}
	}
}

function getProcedure(token){
	var proc = procedure$1[token.type];

	if(proc === procedure$1.attribute){
		proc = attributes[token.action];

		if(proc === attributes.equals && token.name === "id"){
			
			proc = 9;
		}

		if(token.ignoreCase){
			
			
			proc >>= 1;
		}
	} else if(proc === procedure$1.pseudo){
		if(!token.data){
			proc = 3;
		} else if(token.name === "has" || token.name === "contains"){
			proc = 0; 
		} else if(token.name === "matches" || token.name === "not"){
			proc = 0;
			for(var i = 0; i < token.data.length; i++){
				
				if(token.data[i].length !== 1) continue;
				var cur = getProcedure(token.data[i][0]);
				
				if(cur === 0){
					proc = 0;
					break;
				}
				if(cur > proc) proc = cur;
			}
			if(token.data.length > 1 && proc > 0) proc -= 1;
		} else {
			proc = 1;
		}
	}
	return proc;
}

var falseFunc$2 = index$3.falseFunc;


var reChars = /[-[\]{}()*+?.,\\^$|#\s]/g;

function factory(adapter){
	
	var attributeRules = {
		__proto__: null,
		equals: function(next, data){
			var name  = data.name,
				value = data.value;

			if(data.ignoreCase){
				value = value.toLowerCase();

				return function equalsIC(elem){
					var attr = adapter.getAttributeValue(elem, name);
					return attr != null && attr.toLowerCase() === value && next(elem);
				};
			}

			return function equals(elem){
				return adapter.getAttributeValue(elem, name) === value && next(elem);
			};
		},
		hyphen: function(next, data){
			var name  = data.name,
				value = data.value,
				len = value.length;

			if(data.ignoreCase){
				value = value.toLowerCase();

				return function hyphenIC(elem){
					var attr = adapter.getAttributeValue(elem, name);
					return attr != null &&
							(attr.length === len || attr.charAt(len) === "-") &&
							attr.substr(0, len).toLowerCase() === value &&
							next(elem);
				};
			}

			return function hyphen(elem){
				var attr = adapter.getAttributeValue(elem, name);
				return attr != null &&
						attr.substr(0, len) === value &&
						(attr.length === len || attr.charAt(len) === "-") &&
						next(elem);
			};
		},
		element: function(next, data){
			var name = data.name,
				value = data.value;
			if (data.name === 'class') {
				let value = data.value;
				if (/\s/.test(value)) return function() { return false }
				return function clazz(elem) {
					let classes = elem.classes;
					return classes && classes.has(value) && next(elem)
				}
			} else {
				if(/\s/.test(value)){
					return falseFunc$2;
				}

				value = value.replace(reChars, "\\$&");

				var pattern = "(?:^|\\s)" + value + "(?:$|\\s)",
					flags = data.ignoreCase ? "i" : "",
					regex = new RegExp(pattern, flags);

				return function element(elem){
					var attr = adapter.getAttributeValue(elem, name);
					return attr != null && regex.test(attr) && next(elem);
				};
			}
		},
		exists: function(next, data){
			var name = data.name;
			return function exists(elem){
				return adapter.hasAttrib(elem, name) && next(elem);
			};
		},
		start: function(next, data){
			var name  = data.name,
				value = data.value,
				len = value.length;

			if(len === 0){
				return falseFunc$2;
			}

			if(data.ignoreCase){
				value = value.toLowerCase();

				return function startIC(elem){
					var attr = adapter.getAttributeValue(elem, name);
					return attr != null && attr.substr(0, len).toLowerCase() === value && next(elem);
				};
			}

			return function start(elem){
				var attr = adapter.getAttributeValue(elem, name);
				return attr != null && attr.substr(0, len) === value && next(elem);
			};
		},
		end: function(next, data){
			var name  = data.name,
				value = data.value,
				len   = -value.length;

			if(len === 0){
				return falseFunc$2;
			}

			if(data.ignoreCase){
				value = value.toLowerCase();

				return function endIC(elem){
					var attr = adapter.getAttributeValue(elem, name);
					return attr != null && attr.substr(len).toLowerCase() === value && next(elem);
				};
			}

			return function end(elem){
				var attr = adapter.getAttributeValue(elem, name);
				return attr != null && attr.substr(len) === value && next(elem);
			};
		},
		any: function(next, data){
			var name  = data.name,
				value = data.value;

			if(value === ""){
				return falseFunc$2;
			}

			if(data.ignoreCase){
				var regex = new RegExp(value.replace(reChars, "\\$&"), "i");

				return function anyIC(elem){
					var attr = adapter.getAttributeValue(elem, name);
					return attr != null && regex.test(attr) && next(elem);
				};
			}

			return function any(elem){
				var attr = adapter.getAttributeValue(elem, name);
				return attr != null && attr.indexOf(value) >= 0 && next(elem);
			};
		},
		not: function(next, data){
			var name  = data.name,
				value = data.value;

			if(value === ""){
				return function notEmpty(elem){
					return !!adapter.getAttributeValue(elem, name) && next(elem);
				};
			} else if(data.ignoreCase){
				value = value.toLowerCase();

				return function notIC(elem){
					var attr = adapter.getAttributeValue(elem, name);
					return attr != null && attr.toLowerCase() !== value && next(elem);
				};
			}

			return function not(elem){
				return adapter.getAttributeValue(elem, name) !== value && next(elem);
			};
		}
	};

	return {
		compile: function(next, data, options){
			if(options && options.strict && (
				data.ignoreCase || data.action === "not"
			)) throw new Error("Unsupported attribute selector");
			return attributeRules[data.action](next, data);
		},
		rules: attributeRules
	};
}

var attributes$1 = factory;

var attributeFactory = attributes$1;

function generalFactory(adapter, Pseudos){
	
	return {
		__proto__: null,

		attribute: attributeFactory(adapter).compile,
		pseudo: Pseudos.compile,

		
		tag: function(next, data){
			var name = data.name;
			return function tag(elem){
				return adapter.getNameWithoutNS(elem) === name && next(elem);
			}
		},

		
		descendant: function(next){
			return function descendant(elem){

				var found = false;

				while(!found && (elem = adapter.getParent(elem))){
					found = next(elem);
				}

				return found;
			};
		},
		_flexibleDescendant: function(next){
			
			return function descendant(elem){

				var found = next(elem);

				while(!found && (elem = adapter.getParent(elem))){
					found = next(elem);
				}

				return found;
			};
		},
		parent: function(next, data, options){
			if(options && options.strict) throw new Error("Parent selector isn't part of CSS3");

			return function parent(elem){
				return adapter.getChildren(elem).some(test);
			};

			function test(elem){
				return adapter.isTag(elem) && next(elem);
			}
		},
		child: function(next){
			return function child(elem){
				var parent = adapter.getParent(elem);
				return !!parent && next(parent);
			};
		},
		sibling: function(next){
			return function sibling(elem){
				var siblings = adapter.getSiblings(elem);

				for(var i = 0; i < siblings.length; i++){
					if(adapter.isTag(siblings[i])){
						if(siblings[i] === elem) break;
						if(next(siblings[i])) return true;
					}
				}

				return false;
			};
		},
		adjacent: function(next){
			return function adjacent(elem){
				var siblings = adapter.getSiblings(elem),
					lastElement;

				for(var i = 0; i < siblings.length; i++){
					if(adapter.isTag(siblings[i])){
						if(siblings[i] === elem) break;
						lastElement = siblings[i];
					}
				}

				return !!lastElement && next(lastElement);
			};
		},
		universal: function(next){
			return next;
		}
	};
}

var general = generalFactory;

var getNCheck         = index$5;
var BaseFuncs$1         = index$3;
var attributesFactory = attributes$1;
var trueFunc$1          = BaseFuncs$1.trueFunc;
var falseFunc$3         = BaseFuncs$1.falseFunc;

function filtersFactory(adapter){
	var attributes  = attributesFactory(adapter),
		checkAttrib = attributes.rules.equals;

	
	function equals(a, b){
		if(typeof adapter.equals === "function") return adapter.equals(a, b);

		return a === b;
	}

	function getAttribFunc(name, value){
		var data = {name: name, value: value};
		return function attribFunc(next){
			return checkAttrib(next, data);
		};
	}

	function getChildFunc(next){
		return function(elem){
			return !!adapter.getParent(elem) && next(elem);
		};
	}

	var filters = {
		contains: function(next, text){
			return function contains(elem){
				return next(elem) && adapter.getText(elem).indexOf(text) >= 0;
			};
		},
		icontains: function(next, text){
			var itext = text.toLowerCase();
			return function icontains(elem){
				return next(elem) &&
					adapter.getText(elem).toLowerCase().indexOf(itext) >= 0;
			};
		},

		
		"nth-child": function(next, rule){
			var func = getNCheck(rule);

			if(func === falseFunc$3) return func;
			if(func === trueFunc$1)  return getChildFunc(next);

			return function nthChild(elem){
				var siblings = adapter.getSiblings(elem);

				for(var i = 0, pos = 0; i < siblings.length; i++){
					if(adapter.isTag(siblings[i])){
						if(siblings[i] === elem) break;
						else pos++;
					}
				}

				return func(pos) && next(elem);
			};
		},
		"nth-last-child": function(next, rule){
			var func = getNCheck(rule);

			if(func === falseFunc$3) return func;
			if(func === trueFunc$1)  return getChildFunc(next);

			return function nthLastChild(elem){
				var siblings = adapter.getSiblings(elem);

				for(var pos = 0, i = siblings.length - 1; i >= 0; i--){
					if(adapter.isTag(siblings[i])){
						if(siblings[i] === elem) break;
						else pos++;
					}
				}

				return func(pos) && next(elem);
			};
		},
		"nth-of-type": function(next, rule){
			var func = getNCheck(rule);

			if(func === falseFunc$3) return func;
			if(func === trueFunc$1)  return getChildFunc(next);

			return function nthOfType(elem){
				var siblings = adapter.getSiblings(elem);

				for(var pos = 0, i = 0; i < siblings.length; i++){
					if(adapter.isTag(siblings[i])){
						if(siblings[i] === elem) break;
						if(adapter.getName(siblings[i]) === adapter.getName(elem)) pos++;
					}
				}

				return func(pos) && next(elem);
			};
		},
		"nth-last-of-type": function(next, rule){
			var func = getNCheck(rule);

			if(func === falseFunc$3) return func;
			if(func === trueFunc$1)  return getChildFunc(next);

			return function nthLastOfType(elem){
				var siblings = adapter.getSiblings(elem);

				for(var pos = 0, i = siblings.length - 1; i >= 0; i--){
					if(adapter.isTag(siblings[i])){
						if(siblings[i] === elem) break;
						if(adapter.getName(siblings[i]) === adapter.getName(elem)) pos++;
					}
				}

				return func(pos) && next(elem);
			};
		},

		
		root: function(next){
			return function(elem){
				return !adapter.getParent(elem) && next(elem);
			};
		},

		scope: function(next, rule, options, context){
			if(!context || context.length === 0){
				
				return filters.root(next);
			}

			if(context.length === 1){
				
				return function(elem){
					return equals(context[0], elem) && next(elem);
				};
			}

			return function(elem){
				return context.indexOf(elem) >= 0 && next(elem);
			};
		},

		
		checkbox: getAttribFunc("type", "checkbox"),
		file: getAttribFunc("type", "file"),
		password: getAttribFunc("type", "password"),
		radio: getAttribFunc("type", "radio"),
		reset: getAttribFunc("type", "reset"),
		image: getAttribFunc("type", "image"),
		submit: getAttribFunc("type", "submit")
	};
	return filters;
}

function pseudosFactory$1(adapter){
	
	function getFirstElement(elems){
		for(var i = 0; elems && i < elems.length; i++){
			if(adapter.isTag(elems[i])) return elems[i];
		}
	}

	
	var pseudos = {
		empty: function(elem){
			return !adapter.getChildren(elem).some(function(elem){
				return adapter.isTag(elem) || elem.type === "text";
			});
		},

		"first-child": function(elem){
			return getFirstElement(adapter.getSiblings(elem)) === elem;
		},
		"last-child": function(elem){
			var siblings = adapter.getSiblings(elem);

			for(var i = siblings.length - 1; i >= 0; i--){
				if(siblings[i] === elem) return true;
				if(adapter.isTag(siblings[i])) break;
			}

			return false;
		},
		"first-of-type": function(elem){
			var siblings = adapter.getSiblings(elem);

			for(var i = 0; i < siblings.length; i++){
				if(adapter.isTag(siblings[i])){
					if(siblings[i] === elem) return true;
					if(adapter.getName(siblings[i]) === adapter.getName(elem)) break;
				}
			}

			return false;
		},
		"last-of-type": function(elem){
			var siblings = adapter.getSiblings(elem);

			for(var i = siblings.length - 1; i >= 0; i--){
				if(adapter.isTag(siblings[i])){
					if(siblings[i] === elem) return true;
					if(adapter.getName(siblings[i]) === adapter.getName(elem)) break;
				}
			}

			return false;
		},
		"only-of-type": function(elem){
			var siblings = adapter.getSiblings(elem);

			for(var i = 0, j = siblings.length; i < j; i++){
				if(adapter.isTag(siblings[i])){
					if(siblings[i] === elem) continue;
					if(adapter.getName(siblings[i]) === adapter.getName(elem)) return false;
				}
			}

			return true;
		},
		"only-child": function(elem){
			var siblings = adapter.getSiblings(elem);

			for(var i = 0; i < siblings.length; i++){
				if(adapter.isTag(siblings[i]) && siblings[i] !== elem) return false;
			}

			return true;
		},

		
		link: function(elem){
			return adapter.hasAttrib(elem, "href");
		},
		visited: falseFunc$3, 
		

		
		

		
		selected: function(elem){
			if(adapter.hasAttrib(elem, "selected")) return true;
			else if(adapter.getName(elem) !== "option") return false;

			
			var parent = adapter.getParent(elem);

			if(
				!parent ||
				adapter.getName(parent) !== "select" ||
				adapter.hasAttrib(parent, "multiple")
			) return false;

			var siblings = adapter.getChildren(parent),
				sawElem  = false;

			for(var i = 0; i < siblings.length; i++){
				if(adapter.isTag(siblings[i])){
					if(siblings[i] === elem){
						sawElem = true;
					} else if(!sawElem){
						return false;
					} else if(adapter.hasAttrib(siblings[i], "selected")){
						return false;
					}
				}
			}

			return sawElem;
		},
		
		
		
		
		
		
		disabled: function(elem){
			return adapter.hasAttrib(elem, "disabled");
		},
		enabled: function(elem){
			return !adapter.hasAttrib(elem, "disabled");
		},
		
		checked: function(elem){
			return adapter.hasAttrib(elem, "checked") || pseudos.selected(elem);
		},
		
		required: function(elem){
			return adapter.hasAttrib(elem, "required");
		},
		
		optional: function(elem){
			return !adapter.hasAttrib(elem, "required");
		},

		

		
		parent: function(elem){
			return !pseudos.empty(elem);
		},
		
		header: function(elem){
			var name = adapter.getName(elem);
			return name === "h1" ||
					name === "h2" ||
					name === "h3" ||
					name === "h4" ||
					name === "h5" ||
					name === "h6";
		},

		
		button: function(elem){
			var name = adapter.getName(elem);
			return name === "button" ||
					name === "input" &&
					adapter.getAttributeValue(elem, "type") === "button";
		},
		
		input: function(elem){
			var name = adapter.getName(elem);
			return name === "input" ||
					name === "textarea" ||
					name === "select" ||
					name === "button";
		},
		
		text: function(elem){
			var attr;
			return adapter.getName(elem) === "input" && (
				!(attr = adapter.getAttributeValue(elem, "type")) ||
				attr.toLowerCase() === "text"
			);
		}
	};

	return pseudos;
}

function verifyArgs(func, name, subselect){
	if(subselect === null){
		if(func.length > 1 && name !== "scope"){
			throw new Error("pseudo-selector :" + name + " requires an argument");
		}
	} else {
		if(func.length === 1){
			throw new Error("pseudo-selector :" + name + " doesn't have any arguments");
		}
	}
}


var re_CSS3 = /^(?:(?:nth|last|first|only)-(?:child|of-type)|root|empty|(?:en|dis)abled|checked|not)$/;

function factory$1(adapter){
	var pseudos = pseudosFactory$1(adapter);
	var filters = filtersFactory(adapter);

	return {
		compile: function(next, data, options, context){
			var name = data.name,
				subselect = data.data;

			if(options && options.strict && !re_CSS3.test(name)){
				throw new Error(":" + name + " isn't part of CSS3");
			}

			if(typeof filters[name] === "function"){
				verifyArgs(filters[name], name,  subselect);
				return filters[name](next, subselect, options, context);
			} else if(typeof pseudos[name] === "function"){
				var func = pseudos[name];
				verifyArgs(func, name, subselect);

				if(next === trueFunc$1) return func;

				return function pseudoArgs(elem){
					return func(elem, subselect) && next(elem);
				};
			} else {
				throw new Error("unmatched pseudo-class :" + name);
			}
		},
		filters: filters,
		pseudos: pseudos
	};
}

var pseudos = factory$1;

var compile$$1 = compileFactory$1;

var parse$$1          = index$4;
var BaseFuncs      = index$3;
var sortRules      = sort;
var procedure      = require$$0;
var rulesFactory   = general;
var pseudosFactory = pseudos;
var trueFunc       = BaseFuncs.trueFunc;
var falseFunc$1      = BaseFuncs.falseFunc;

function compileFactory$1(adapter){
	var Pseudos     = pseudosFactory(adapter),
		filters     = Pseudos.filters,
		Rules 			= rulesFactory(adapter, Pseudos);

	function compile$$1(selector, options, context){
		var next = compileUnsafe(selector, options, context);
		return wrap(next);
	}

	function wrap(next){
		return function base(elem){
			return adapter.isTag(elem) && next(elem);
		};
	}

	function compileUnsafe(selector, options, context){
		var token = parse$$1(selector, options);
		return compileToken(token, options, context);
	}

	function includesScopePseudo(t){
		return t.type === "pseudo" && (
			t.name === "scope" || (
				Array.isArray(t.data) &&
				t.data.some(function(data){
					return data.some(includesScopePseudo);
				})
			)
		);
	}

	var DESCENDANT_TOKEN = {type: "descendant"},
		FLEXIBLE_DESCENDANT_TOKEN = {type: "_flexibleDescendant"},
		SCOPE_TOKEN = {type: "pseudo", name: "scope"},
		PLACEHOLDER_ELEMENT = {};

	
	
	function absolutize(token, context){
		
		var hasContext = !!context && !!context.length && context.every(function(e){
			return e === PLACEHOLDER_ELEMENT || !!adapter.getParent(e);
		});


		token.forEach(function(t){
			if(t.length > 0 && isTraversal(t[0]) && t[0].type !== "descendant"){
				
			} else if(hasContext && !includesScopePseudo(t)){
				t.unshift(DESCENDANT_TOKEN);
			} else {
				return;
			}

			t.unshift(SCOPE_TOKEN);
		});
	}

	function compileToken(token, options, context){
		token = token.filter(function(t){ return t.length > 0; });

		token.forEach(sortRules);

		var isArrayContext = Array.isArray(context);

		context = (options && options.context) || context;

		if(context && !isArrayContext) context = [context];

		absolutize(token, context);

		var shouldTestNextSiblings = false;

		var query = token
			.map(function(rules){
				if(rules[0] && rules[1] && rules[0].name === "scope"){
					var ruleType = rules[1].type;
					if(isArrayContext && ruleType === "descendant") rules[1] = FLEXIBLE_DESCENDANT_TOKEN;
					else if(ruleType === "adjacent" || ruleType === "sibling") shouldTestNextSiblings = true;
				}
				return compileRules(rules, options, context);
			})
			.reduce(reduceRules, falseFunc$1);

		query.shouldTestNextSiblings = shouldTestNextSiblings;

		return query;
	}

	function isTraversal(t){
		return procedure[t.type] < 0;
	}

	function compileRules(rules, options, context){
		return rules.reduce(function(func, rule){
			if(func === falseFunc$1) return func;
			return Rules[rule.type](func, rule, options, context);
		}, options && options.rootFunc || trueFunc);
	}

	function reduceRules(a, b){
		if(b === falseFunc$1 || a === trueFunc){
			return a;
		}
		if(a === falseFunc$1 || b === trueFunc){
			return b;
		}

		return function combine(elem){
			return a(elem) || b(elem);
		};
	}

	function containsTraversal(t){
		return t.some(isTraversal);
	}

	
	
	
	filters.not = function(next, token, options, context){
		var opts = {
			xmlMode: !!(options && options.xmlMode),
			strict: !!(options && options.strict)
		};

		if(opts.strict){
			if(token.length > 1 || token.some(containsTraversal)){
				throw new Error("complex selectors in :not aren't allowed in strict mode");
			}
		}

		var func = compileToken(token, opts, context);

		if(func === falseFunc$1) return next;
		if(func === trueFunc)  return falseFunc$1;

		return function(elem){
			return !func(elem) && next(elem);
		};
	};

	filters.has = function(next, token, options){
		var opts = {
			xmlMode: !!(options && options.xmlMode),
			strict: !!(options && options.strict)
		};

		
		var context = token.some(containsTraversal) ? [PLACEHOLDER_ELEMENT] : null;

		var func = compileToken(token, opts, context);

		if(func === falseFunc$1) return falseFunc$1;
		if(func === trueFunc){
			return function(elem){
				return adapter.getChildren(elem).some(adapter.isTag) && next(elem);
			};
		}

		func = wrap(func);

		if(context){
			return function has(elem){
				return next(elem) && (
					(context[0] = elem), adapter.existsOne(func, adapter.getChildren(elem))
				);
			};
		}

		return function has(elem){
			return next(elem) && adapter.existsOne(func, adapter.getChildren(elem));
		};
	};

	filters.matches = function(next, token, options, context){
		var opts = {
			xmlMode: !!(options && options.xmlMode),
			strict: !!(options && options.strict),
			rootFunc: next
		};

		return compileToken(token, opts, context);
	};

	compile$$1.compileToken = compileToken;
	compile$$1.compileUnsafe = compileUnsafe;
	compile$$1.Pseudos = Pseudos;

	return compile$$1;
}

var index$1 = CSSselect$1;

var DomUtils       = index$2;
var falseFunc      = index$3.falseFunc;
var compileFactory = compile$$1;
var defaultCompile = compileFactory(DomUtils);

function adapterCompile(adapter){
	return adapter === DomUtils ? defaultCompile : compileFactory(adapter);
}

function getSelectorFunc(searchFunc){
	return function select(query, elems, options){
		options = options || {};
		options.adapter = options.adapter || DomUtils;
		var compile$$1 = adapterCompile(options.adapter);

		if(typeof query !== "function") query = compile$$1.compileUnsafe(query, options, elems);
		if(query.shouldTestNextSiblings) elems = appendNextSiblings((options && options.context) || elems, options.adapter);
		if(!Array.isArray(elems)) elems = options.adapter.getChildren(elems);
		else elems = options.adapter.removeSubsets(elems);
		return searchFunc(query, elems, options);
	};
}

function getNextSiblings(elem, adapter){
	var siblings = adapter.getSiblings(elem);
	if(!Array.isArray(siblings)) return [];
	siblings = siblings.slice(0);
	while(siblings.shift() !== elem);
	return siblings;
}

function appendNextSiblings(elems, adapter){
	
	if(!Array.isArray(elems)) elems = [elems];
	var newElems = elems.slice(0);

	for(var i = 0, len = elems.length; i < len; i++){
		var nextSiblings = getNextSiblings(newElems[i], adapter);
		newElems.push.apply(newElems, nextSiblings);
	}
	return newElems;
}

var selectAll = getSelectorFunc(function selectAll(query, elems, options){
	return (query === falseFunc || !elems || elems.length === 0) ? [] : options.adapter.findAll(query, elems);
});

var selectOne = getSelectorFunc(function selectOne(query, elems, options){
	return (query === falseFunc || !elems || elems.length === 0) ? null : options.adapter.findOne(query, elems);
});

function is(elem, query, options){
	options = options || {};
	options.adapter = options.adapter || DomUtils;
	var compile$$1 = adapterCompile(options.adapter);
	return (typeof query === "function" ? query : compile$$1(query, options))(elem);
}


function CSSselect$1(query, elems, options){
	return selectAll(query, elems, options);
}

CSSselect$1.compile = defaultCompile;
CSSselect$1.filters = defaultCompile.Pseudos.filters;
CSSselect$1.pseudos = defaultCompile.Pseudos.pseudos;

CSSselect$1.selectAll = selectAll;
CSSselect$1.selectOne = selectOne;

CSSselect$1.is = is;


CSSselect$1.parse = defaultCompile;
CSSselect$1.iterate = selectAll;


CSSselect$1._compileUnsafe = defaultCompile.compileUnsafe;
CSSselect$1._compileToken = defaultCompile.compileToken;

var amp$1$1 = "&";
var apos$1$1 = "'";
var gt$1$1 = ">";
var lt$1$1 = "<";
var quot$1$1 = "\"";
var xmlJSON$1 = {
	amp: amp$1$1,
	apos: apos$1$1,
	gt: gt$1$1,
	lt: lt$1$1,
	quot: quot$1$1
};

var xml$1 = Object.freeze({
	amp: amp$1$1,
	apos: apos$1$1,
	gt: gt$1$1,
	lt: lt$1$1,
	quot: quot$1$1,
	default: xmlJSON$1
});

var Aacute$1$1 = "Á";
var aacute$1$1 = "á";
var Abreve$1 = "Ă";
var abreve$1 = "ă";
var ac$1 = "∾";
var acd$1 = "∿";
var acE$1 = "∾̳";
var Acirc$1$1 = "Â";
var acirc$1$1 = "â";
var acute$1$1 = "´";
var Acy$1 = "А";
var acy$1 = "а";
var AElig$1$1 = "Æ";
var aelig$1$1 = "æ";
var af$1 = "⁡";
var Afr$1 = "𝔄";
var afr$1 = "𝔞";
var Agrave$1$1 = "À";
var agrave$1$1 = "à";
var alefsym$1 = "ℵ";
var aleph$1 = "ℵ";
var Alpha$1 = "Α";
var alpha$1 = "α";
var Amacr$1 = "Ā";
var amacr$1 = "ā";
var amalg$1 = "⨿";
var amp$1$2 = "&";
var AMP$1$1 = "&";
var andand$1 = "⩕";
var And$1 = "⩓";
var and$1 = "∧";
var andd$1 = "⩜";
var andslope$1 = "⩘";
var andv$1 = "⩚";
var ang$1 = "∠";
var ange$1 = "⦤";
var angle$1 = "∠";
var angmsdaa$1 = "⦨";
var angmsdab$1 = "⦩";
var angmsdac$1 = "⦪";
var angmsdad$1 = "⦫";
var angmsdae$1 = "⦬";
var angmsdaf$1 = "⦭";
var angmsdag$1 = "⦮";
var angmsdah$1 = "⦯";
var angmsd$1 = "∡";
var angrt$1 = "∟";
var angrtvb$1 = "⊾";
var angrtvbd$1 = "⦝";
var angsph$1 = "∢";
var angst$1 = "Å";
var angzarr$1 = "⍼";
var Aogon$1 = "Ą";
var aogon$1 = "ą";
var Aopf$1 = "𝔸";
var aopf$1 = "𝕒";
var apacir$1 = "⩯";
var ap$1 = "≈";
var apE$1 = "⩰";
var ape$1 = "≊";
var apid$1 = "≋";
var apos$1$2 = "'";
var ApplyFunction$1 = "⁡";
var approx$1 = "≈";
var approxeq$1 = "≊";
var Aring$1$1 = "Å";
var aring$1$1 = "å";
var Ascr$1 = "𝒜";
var ascr$1 = "𝒶";
var Assign$1 = "≔";
var ast$1 = "*";
var asymp$1 = "≈";
var asympeq$1 = "≍";
var Atilde$1$1 = "Ã";
var atilde$1$1 = "ã";
var Auml$1$1 = "Ä";
var auml$1$1 = "ä";
var awconint$1 = "∳";
var awint$1 = "⨑";
var backcong$1 = "≌";
var backepsilon$1 = "϶";
var backprime$1 = "‵";
var backsim$1 = "∽";
var backsimeq$1 = "⋍";
var Backslash$1 = "∖";
var Barv$1 = "⫧";
var barvee$1 = "⊽";
var barwed$1 = "⌅";
var Barwed$1 = "⌆";
var barwedge$1 = "⌅";
var bbrk$1 = "⎵";
var bbrktbrk$1 = "⎶";
var bcong$1 = "≌";
var Bcy$1 = "Б";
var bcy$1 = "б";
var bdquo$1 = "„";
var becaus$1 = "∵";
var because$1 = "∵";
var Because$1 = "∵";
var bemptyv$1 = "⦰";
var bepsi$1 = "϶";
var bernou$1 = "ℬ";
var Bernoullis$1 = "ℬ";
var Beta$1 = "Β";
var beta$1 = "β";
var beth$1 = "ℶ";
var between$1 = "≬";
var Bfr$1 = "𝔅";
var bfr$1 = "𝔟";
var bigcap$1 = "⋂";
var bigcirc$1 = "◯";
var bigcup$1 = "⋃";
var bigodot$1 = "⨀";
var bigoplus$1 = "⨁";
var bigotimes$1 = "⨂";
var bigsqcup$1 = "⨆";
var bigstar$1 = "★";
var bigtriangledown$1 = "▽";
var bigtriangleup$1 = "△";
var biguplus$1 = "⨄";
var bigvee$1 = "⋁";
var bigwedge$1 = "⋀";
var bkarow$1 = "⤍";
var blacklozenge$1 = "⧫";
var blacksquare$1 = "▪";
var blacktriangle$1 = "▴";
var blacktriangledown$1 = "▾";
var blacktriangleleft$1 = "◂";
var blacktriangleright$1 = "▸";
var blank$1 = "␣";
var blk12$1 = "▒";
var blk14$1 = "░";
var blk34$1 = "▓";
var block$1 = "█";
var bne$1 = "=⃥";
var bnequiv$1 = "≡⃥";
var bNot$1 = "⫭";
var bnot$1 = "⌐";
var Bopf$1 = "𝔹";
var bopf$1 = "𝕓";
var bot$1 = "⊥";
var bottom$1 = "⊥";
var bowtie$1 = "⋈";
var boxbox$1 = "⧉";
var boxdl$1 = "┐";
var boxdL$1 = "╕";
var boxDl$1 = "╖";
var boxDL$1 = "╗";
var boxdr$1 = "┌";
var boxdR$1 = "╒";
var boxDr$1 = "╓";
var boxDR$1 = "╔";
var boxh$1 = "─";
var boxH$1 = "═";
var boxhd$1 = "┬";
var boxHd$1 = "╤";
var boxhD$1 = "╥";
var boxHD$1 = "╦";
var boxhu$1 = "┴";
var boxHu$1 = "╧";
var boxhU$1 = "╨";
var boxHU$1 = "╩";
var boxminus$1 = "⊟";
var boxplus$1 = "⊞";
var boxtimes$1 = "⊠";
var boxul$1 = "┘";
var boxuL$1 = "╛";
var boxUl$1 = "╜";
var boxUL$1 = "╝";
var boxur$1 = "└";
var boxuR$1 = "╘";
var boxUr$1 = "╙";
var boxUR$1 = "╚";
var boxv$1 = "│";
var boxV$1 = "║";
var boxvh$1 = "┼";
var boxvH$1 = "╪";
var boxVh$1 = "╫";
var boxVH$1 = "╬";
var boxvl$1 = "┤";
var boxvL$1 = "╡";
var boxVl$1 = "╢";
var boxVL$1 = "╣";
var boxvr$1 = "├";
var boxvR$1 = "╞";
var boxVr$1 = "╟";
var boxVR$1 = "╠";
var bprime$1 = "‵";
var breve$1 = "˘";
var Breve$1 = "˘";
var brvbar$1$1 = "¦";
var bscr$1 = "𝒷";
var Bscr$1 = "ℬ";
var bsemi$1 = "⁏";
var bsim$1 = "∽";
var bsime$1 = "⋍";
var bsolb$1 = "⧅";
var bsol$1 = "\\";
var bsolhsub$1 = "⟈";
var bull$1 = "•";
var bullet$1 = "•";
var bump$1 = "≎";
var bumpE$1 = "⪮";
var bumpe$1 = "≏";
var Bumpeq$1 = "≎";
var bumpeq$1 = "≏";
var Cacute$1 = "Ć";
var cacute$1 = "ć";
var capand$1 = "⩄";
var capbrcup$1 = "⩉";
var capcap$1 = "⩋";
var cap$1 = "∩";
var Cap$1 = "⋒";
var capcup$1 = "⩇";
var capdot$1 = "⩀";
var CapitalDifferentialD$1 = "ⅅ";
var caps$1 = "∩︀";
var caret$1 = "⁁";
var caron$1 = "ˇ";
var Cayleys$1 = "ℭ";
var ccaps$1 = "⩍";
var Ccaron$1 = "Č";
var ccaron$1 = "č";
var Ccedil$1$1 = "Ç";
var ccedil$1$1 = "ç";
var Ccirc$1 = "Ĉ";
var ccirc$1 = "ĉ";
var Cconint$1 = "∰";
var ccups$1 = "⩌";
var ccupssm$1 = "⩐";
var Cdot$1 = "Ċ";
var cdot$1 = "ċ";
var cedil$1$1 = "¸";
var Cedilla$1 = "¸";
var cemptyv$1 = "⦲";
var cent$1$1 = "¢";
var centerdot$1 = "·";
var CenterDot$1 = "·";
var cfr$1 = "𝔠";
var Cfr$1 = "ℭ";
var CHcy$1 = "Ч";
var chcy$1 = "ч";
var check$1 = "✓";
var checkmark$1 = "✓";
var Chi$1 = "Χ";
var chi$1 = "χ";
var circ$1 = "ˆ";
var circeq$1 = "≗";
var circlearrowleft$1 = "↺";
var circlearrowright$1 = "↻";
var circledast$1 = "⊛";
var circledcirc$1 = "⊚";
var circleddash$1 = "⊝";
var CircleDot$1 = "⊙";
var circledR$1 = "®";
var circledS$1 = "Ⓢ";
var CircleMinus$1 = "⊖";
var CirclePlus$1 = "⊕";
var CircleTimes$1 = "⊗";
var cir$1 = "○";
var cirE$1 = "⧃";
var cire$1 = "≗";
var cirfnint$1 = "⨐";
var cirmid$1 = "⫯";
var cirscir$1 = "⧂";
var ClockwiseContourIntegral$1 = "∲";
var CloseCurlyDoubleQuote$1 = "”";
var CloseCurlyQuote$1 = "’";
var clubs$1 = "♣";
var clubsuit$1 = "♣";
var colon$1 = ":";
var Colon$1 = "∷";
var Colone$1 = "⩴";
var colone$1 = "≔";
var coloneq$1 = "≔";
var comma$1 = ",";
var commat$1 = "@";
var comp$1 = "∁";
var compfn$1 = "∘";
var complement$1 = "∁";
var complexes$1 = "ℂ";
var cong$1 = "≅";
var congdot$1 = "⩭";
var Congruent$1 = "≡";
var conint$1 = "∮";
var Conint$1 = "∯";
var ContourIntegral$1 = "∮";
var copf$1 = "𝕔";
var Copf$1 = "ℂ";
var coprod$1 = "∐";
var Coproduct$1 = "∐";
var copy$1$1 = "©";
var COPY$1$1 = "©";
var copysr$1 = "℗";
var CounterClockwiseContourIntegral$1 = "∳";
var crarr$1 = "↵";
var cross$1 = "✗";
var Cross$1 = "⨯";
var Cscr$1 = "𝒞";
var cscr$1 = "𝒸";
var csub$1 = "⫏";
var csube$1 = "⫑";
var csup$1 = "⫐";
var csupe$1 = "⫒";
var ctdot$1 = "⋯";
var cudarrl$1 = "⤸";
var cudarrr$1 = "⤵";
var cuepr$1 = "⋞";
var cuesc$1 = "⋟";
var cularr$1 = "↶";
var cularrp$1 = "⤽";
var cupbrcap$1 = "⩈";
var cupcap$1 = "⩆";
var CupCap$1 = "≍";
var cup$1 = "∪";
var Cup$1 = "⋓";
var cupcup$1 = "⩊";
var cupdot$1 = "⊍";
var cupor$1 = "⩅";
var cups$1 = "∪︀";
var curarr$1 = "↷";
var curarrm$1 = "⤼";
var curlyeqprec$1 = "⋞";
var curlyeqsucc$1 = "⋟";
var curlyvee$1 = "⋎";
var curlywedge$1 = "⋏";
var curren$1$1 = "¤";
var curvearrowleft$1 = "↶";
var curvearrowright$1 = "↷";
var cuvee$1 = "⋎";
var cuwed$1 = "⋏";
var cwconint$1 = "∲";
var cwint$1 = "∱";
var cylcty$1 = "⌭";
var dagger$1 = "†";
var Dagger$1 = "‡";
var daleth$1 = "ℸ";
var darr$1 = "↓";
var Darr$1 = "↡";
var dArr$1 = "⇓";
var dash$1 = "‐";
var Dashv$1 = "⫤";
var dashv$1 = "⊣";
var dbkarow$1 = "⤏";
var dblac$1 = "˝";
var Dcaron$1 = "Ď";
var dcaron$1 = "ď";
var Dcy$1 = "Д";
var dcy$1 = "д";
var ddagger$1 = "‡";
var ddarr$1 = "⇊";
var DD$1 = "ⅅ";
var dd$1 = "ⅆ";
var DDotrahd$1 = "⤑";
var ddotseq$1 = "⩷";
var deg$1$1 = "°";
var Del$1 = "∇";
var Delta$1 = "Δ";
var delta$1 = "δ";
var demptyv$1 = "⦱";
var dfisht$1 = "⥿";
var Dfr$1 = "𝔇";
var dfr$1 = "𝔡";
var dHar$1 = "⥥";
var dharl$1 = "⇃";
var dharr$1 = "⇂";
var DiacriticalAcute$1 = "´";
var DiacriticalDot$1 = "˙";
var DiacriticalDoubleAcute$1 = "˝";
var DiacriticalGrave$1 = "`";
var DiacriticalTilde$1 = "˜";
var diam$1 = "⋄";
var diamond$1 = "⋄";
var Diamond$1 = "⋄";
var diamondsuit$1 = "♦";
var diams$1 = "♦";
var die$1 = "¨";
var DifferentialD$1 = "ⅆ";
var digamma$1 = "ϝ";
var disin$1 = "⋲";
var div$1 = "÷";
var divide$1$1 = "÷";
var divideontimes$1 = "⋇";
var divonx$1 = "⋇";
var DJcy$1 = "Ђ";
var djcy$1 = "ђ";
var dlcorn$1 = "⌞";
var dlcrop$1 = "⌍";
var dollar$1 = "$";
var Dopf$1 = "𝔻";
var dopf$1 = "𝕕";
var Dot$1 = "¨";
var dot$1 = "˙";
var DotDot$1 = "⃜";
var doteq$1 = "≐";
var doteqdot$1 = "≑";
var DotEqual$1 = "≐";
var dotminus$1 = "∸";
var dotplus$1 = "∔";
var dotsquare$1 = "⊡";
var doublebarwedge$1 = "⌆";
var DoubleContourIntegral$1 = "∯";
var DoubleDot$1 = "¨";
var DoubleDownArrow$1 = "⇓";
var DoubleLeftArrow$1 = "⇐";
var DoubleLeftRightArrow$1 = "⇔";
var DoubleLeftTee$1 = "⫤";
var DoubleLongLeftArrow$1 = "⟸";
var DoubleLongLeftRightArrow$1 = "⟺";
var DoubleLongRightArrow$1 = "⟹";
var DoubleRightArrow$1 = "⇒";
var DoubleRightTee$1 = "⊨";
var DoubleUpArrow$1 = "⇑";
var DoubleUpDownArrow$1 = "⇕";
var DoubleVerticalBar$1 = "∥";
var DownArrowBar$1 = "⤓";
var downarrow$1 = "↓";
var DownArrow$1 = "↓";
var Downarrow$1 = "⇓";
var DownArrowUpArrow$1 = "⇵";
var DownBreve$1 = "̑";
var downdownarrows$1 = "⇊";
var downharpoonleft$1 = "⇃";
var downharpoonright$1 = "⇂";
var DownLeftRightVector$1 = "⥐";
var DownLeftTeeVector$1 = "⥞";
var DownLeftVectorBar$1 = "⥖";
var DownLeftVector$1 = "↽";
var DownRightTeeVector$1 = "⥟";
var DownRightVectorBar$1 = "⥗";
var DownRightVector$1 = "⇁";
var DownTeeArrow$1 = "↧";
var DownTee$1 = "⊤";
var drbkarow$1 = "⤐";
var drcorn$1 = "⌟";
var drcrop$1 = "⌌";
var Dscr$1 = "𝒟";
var dscr$1 = "𝒹";
var DScy$1 = "Ѕ";
var dscy$1 = "ѕ";
var dsol$1 = "⧶";
var Dstrok$1 = "Đ";
var dstrok$1 = "đ";
var dtdot$1 = "⋱";
var dtri$1 = "▿";
var dtrif$1 = "▾";
var duarr$1 = "⇵";
var duhar$1 = "⥯";
var dwangle$1 = "⦦";
var DZcy$1 = "Џ";
var dzcy$1 = "џ";
var dzigrarr$1 = "⟿";
var Eacute$1$1 = "É";
var eacute$1$1 = "é";
var easter$1 = "⩮";
var Ecaron$1 = "Ě";
var ecaron$1 = "ě";
var Ecirc$1$1 = "Ê";
var ecirc$1$1 = "ê";
var ecir$1 = "≖";
var ecolon$1 = "≕";
var Ecy$1 = "Э";
var ecy$1 = "э";
var eDDot$1 = "⩷";
var Edot$1 = "Ė";
var edot$1 = "ė";
var eDot$1 = "≑";
var ee$1 = "ⅇ";
var efDot$1 = "≒";
var Efr$1 = "𝔈";
var efr$1 = "𝔢";
var eg$1 = "⪚";
var Egrave$1$1 = "È";
var egrave$1$1 = "è";
var egs$1 = "⪖";
var egsdot$1 = "⪘";
var el$1 = "⪙";
var Element$1 = "∈";
var elinters$1 = "⏧";
var ell$1 = "ℓ";
var els$1 = "⪕";
var elsdot$1 = "⪗";
var Emacr$1 = "Ē";
var emacr$1 = "ē";
var empty$1 = "∅";
var emptyset$1 = "∅";
var EmptySmallSquare$1 = "◻";
var emptyv$1 = "∅";
var EmptyVerySmallSquare$1 = "▫";
var emsp13$1 = " ";
var emsp14$1 = " ";
var emsp$1 = " ";
var ENG$1 = "Ŋ";
var eng$1 = "ŋ";
var ensp$1 = " ";
var Eogon$1 = "Ę";
var eogon$1 = "ę";
var Eopf$1 = "𝔼";
var eopf$1 = "𝕖";
var epar$1 = "⋕";
var eparsl$1 = "⧣";
var eplus$1 = "⩱";
var epsi$1 = "ε";
var Epsilon$1 = "Ε";
var epsilon$1 = "ε";
var epsiv$1 = "ϵ";
var eqcirc$1 = "≖";
var eqcolon$1 = "≕";
var eqsim$1 = "≂";
var eqslantgtr$1 = "⪖";
var eqslantless$1 = "⪕";
var Equal$1 = "⩵";
var equals$1 = "=";
var EqualTilde$1 = "≂";
var equest$1 = "≟";
var Equilibrium$1 = "⇌";
var equiv$1 = "≡";
var equivDD$1 = "⩸";
var eqvparsl$1 = "⧥";
var erarr$1 = "⥱";
var erDot$1 = "≓";
var escr$1 = "ℯ";
var Escr$1 = "ℰ";
var esdot$1 = "≐";
var Esim$1 = "⩳";
var esim$1 = "≂";
var Eta$1 = "Η";
var eta$1 = "η";
var ETH$1$1 = "Ð";
var eth$1$1 = "ð";
var Euml$1$1 = "Ë";
var euml$1$1 = "ë";
var euro$1 = "€";
var excl$1 = "!";
var exist$1 = "∃";
var Exists$1 = "∃";
var expectation$1 = "ℰ";
var exponentiale$1 = "ⅇ";
var ExponentialE$1 = "ⅇ";
var fallingdotseq$1 = "≒";
var Fcy$1 = "Ф";
var fcy$1 = "ф";
var female$1 = "♀";
var ffilig$1 = "ﬃ";
var fflig$1 = "ﬀ";
var ffllig$1 = "ﬄ";
var Ffr$1 = "𝔉";
var ffr$1 = "𝔣";
var filig$1 = "ﬁ";
var FilledSmallSquare$1 = "◼";
var FilledVerySmallSquare$1 = "▪";
var fjlig$1 = "fj";
var flat$1 = "♭";
var fllig$1 = "ﬂ";
var fltns$1 = "▱";
var fnof$1 = "ƒ";
var Fopf$1 = "𝔽";
var fopf$1 = "𝕗";
var forall$1 = "∀";
var ForAll$1 = "∀";
var fork$1 = "⋔";
var forkv$1 = "⫙";
var Fouriertrf$1 = "ℱ";
var fpartint$1 = "⨍";
var frac12$1$1 = "½";
var frac13$1 = "⅓";
var frac14$1$1 = "¼";
var frac15$1 = "⅕";
var frac16$1 = "⅙";
var frac18$1 = "⅛";
var frac23$1 = "⅔";
var frac25$1 = "⅖";
var frac34$1$1 = "¾";
var frac35$1 = "⅗";
var frac38$1 = "⅜";
var frac45$1 = "⅘";
var frac56$1 = "⅚";
var frac58$1 = "⅝";
var frac78$1 = "⅞";
var frasl$1 = "⁄";
var frown$1 = "⌢";
var fscr$1 = "𝒻";
var Fscr$1 = "ℱ";
var gacute$1 = "ǵ";
var Gamma$1 = "Γ";
var gamma$1 = "γ";
var Gammad$1 = "Ϝ";
var gammad$1 = "ϝ";
var gap$1 = "⪆";
var Gbreve$1 = "Ğ";
var gbreve$1 = "ğ";
var Gcedil$1 = "Ģ";
var Gcirc$1 = "Ĝ";
var gcirc$1 = "ĝ";
var Gcy$1 = "Г";
var gcy$1 = "г";
var Gdot$1 = "Ġ";
var gdot$1 = "ġ";
var ge$1 = "≥";
var gE$1 = "≧";
var gEl$1 = "⪌";
var gel$1 = "⋛";
var geq$1 = "≥";
var geqq$1 = "≧";
var geqslant$1 = "⩾";
var gescc$1 = "⪩";
var ges$1 = "⩾";
var gesdot$1 = "⪀";
var gesdoto$1 = "⪂";
var gesdotol$1 = "⪄";
var gesl$1 = "⋛︀";
var gesles$1 = "⪔";
var Gfr$1 = "𝔊";
var gfr$1 = "𝔤";
var gg$1 = "≫";
var Gg$1 = "⋙";
var ggg$1 = "⋙";
var gimel$1 = "ℷ";
var GJcy$1 = "Ѓ";
var gjcy$1 = "ѓ";
var gla$1 = "⪥";
var gl$1 = "≷";
var glE$1 = "⪒";
var glj$1 = "⪤";
var gnap$1 = "⪊";
var gnapprox$1 = "⪊";
var gne$1 = "⪈";
var gnE$1 = "≩";
var gneq$1 = "⪈";
var gneqq$1 = "≩";
var gnsim$1 = "⋧";
var Gopf$1 = "𝔾";
var gopf$1 = "𝕘";
var grave$1 = "`";
var GreaterEqual$1 = "≥";
var GreaterEqualLess$1 = "⋛";
var GreaterFullEqual$1 = "≧";
var GreaterGreater$1 = "⪢";
var GreaterLess$1 = "≷";
var GreaterSlantEqual$1 = "⩾";
var GreaterTilde$1 = "≳";
var Gscr$1 = "𝒢";
var gscr$1 = "ℊ";
var gsim$1 = "≳";
var gsime$1 = "⪎";
var gsiml$1 = "⪐";
var gtcc$1 = "⪧";
var gtcir$1 = "⩺";
var gt$1$2 = ">";
var GT$1$1 = ">";
var Gt$1 = "≫";
var gtdot$1 = "⋗";
var gtlPar$1 = "⦕";
var gtquest$1 = "⩼";
var gtrapprox$1 = "⪆";
var gtrarr$1 = "⥸";
var gtrdot$1 = "⋗";
var gtreqless$1 = "⋛";
var gtreqqless$1 = "⪌";
var gtrless$1 = "≷";
var gtrsim$1 = "≳";
var gvertneqq$1 = "≩︀";
var gvnE$1 = "≩︀";
var Hacek$1 = "ˇ";
var hairsp$1 = " ";
var half$1 = "½";
var hamilt$1 = "ℋ";
var HARDcy$1 = "Ъ";
var hardcy$1 = "ъ";
var harrcir$1 = "⥈";
var harr$1 = "↔";
var hArr$1 = "⇔";
var harrw$1 = "↭";
var Hat$1 = "^";
var hbar$1 = "ℏ";
var Hcirc$1 = "Ĥ";
var hcirc$1 = "ĥ";
var hearts$1 = "♥";
var heartsuit$1 = "♥";
var hellip$1 = "…";
var hercon$1 = "⊹";
var hfr$1 = "𝔥";
var Hfr$1 = "ℌ";
var HilbertSpace$1 = "ℋ";
var hksearow$1 = "⤥";
var hkswarow$1 = "⤦";
var hoarr$1 = "⇿";
var homtht$1 = "∻";
var hookleftarrow$1 = "↩";
var hookrightarrow$1 = "↪";
var hopf$1 = "𝕙";
var Hopf$1 = "ℍ";
var horbar$1 = "―";
var HorizontalLine$1 = "─";
var hscr$1 = "𝒽";
var Hscr$1 = "ℋ";
var hslash$1 = "ℏ";
var Hstrok$1 = "Ħ";
var hstrok$1 = "ħ";
var HumpDownHump$1 = "≎";
var HumpEqual$1 = "≏";
var hybull$1 = "⁃";
var hyphen$1 = "‐";
var Iacute$1$1 = "Í";
var iacute$1$1 = "í";
var ic$1 = "⁣";
var Icirc$1$1 = "Î";
var icirc$1$1 = "î";
var Icy$1 = "И";
var icy$1 = "и";
var Idot$1 = "İ";
var IEcy$1 = "Е";
var iecy$1 = "е";
var iexcl$1$1 = "¡";
var iff$1 = "⇔";
var ifr$1 = "𝔦";
var Ifr$1 = "ℑ";
var Igrave$1$1 = "Ì";
var igrave$1$1 = "ì";
var ii$1 = "ⅈ";
var iiiint$1 = "⨌";
var iiint$1 = "∭";
var iinfin$1 = "⧜";
var iiota$1 = "℩";
var IJlig$1 = "Ĳ";
var ijlig$1 = "ĳ";
var Imacr$1 = "Ī";
var imacr$1 = "ī";
var image$1 = "ℑ";
var ImaginaryI$1 = "ⅈ";
var imagline$1 = "ℐ";
var imagpart$1 = "ℑ";
var imath$1 = "ı";
var Im$1 = "ℑ";
var imof$1 = "⊷";
var imped$1 = "Ƶ";
var Implies$1 = "⇒";
var incare$1 = "℅";
var infin$1 = "∞";
var infintie$1 = "⧝";
var inodot$1 = "ı";
var intcal$1 = "⊺";
var int$1 = "∫";
var Int$1 = "∬";
var integers$1 = "ℤ";
var Integral$1 = "∫";
var intercal$1 = "⊺";
var Intersection$1 = "⋂";
var intlarhk$1 = "⨗";
var intprod$1 = "⨼";
var InvisibleComma$1 = "⁣";
var InvisibleTimes$1 = "⁢";
var IOcy$1 = "Ё";
var iocy$1 = "ё";
var Iogon$1 = "Į";
var iogon$1 = "į";
var Iopf$1 = "𝕀";
var iopf$1 = "𝕚";
var Iota$1 = "Ι";
var iota$1 = "ι";
var iprod$1 = "⨼";
var iquest$1$1 = "¿";
var iscr$1 = "𝒾";
var Iscr$1 = "ℐ";
var isin$1 = "∈";
var isindot$1 = "⋵";
var isinE$1 = "⋹";
var isins$1 = "⋴";
var isinsv$1 = "⋳";
var isinv$1 = "∈";
var it$1 = "⁢";
var Itilde$1 = "Ĩ";
var itilde$1 = "ĩ";
var Iukcy$1 = "І";
var iukcy$1 = "і";
var Iuml$1$1 = "Ï";
var iuml$1$1 = "ï";
var Jcirc$1 = "Ĵ";
var jcirc$1 = "ĵ";
var Jcy$1 = "Й";
var jcy$1 = "й";
var Jfr$1 = "𝔍";
var jfr$1 = "𝔧";
var jmath$1 = "ȷ";
var Jopf$1 = "𝕁";
var jopf$1 = "𝕛";
var Jscr$1 = "𝒥";
var jscr$1 = "𝒿";
var Jsercy$1 = "Ј";
var jsercy$1 = "ј";
var Jukcy$1 = "Є";
var jukcy$1 = "є";
var Kappa$1 = "Κ";
var kappa$1 = "κ";
var kappav$1 = "ϰ";
var Kcedil$1 = "Ķ";
var kcedil$1 = "ķ";
var Kcy$1 = "К";
var kcy$1 = "к";
var Kfr$1 = "𝔎";
var kfr$1 = "𝔨";
var kgreen$1 = "ĸ";
var KHcy$1 = "Х";
var khcy$1 = "х";
var KJcy$1 = "Ќ";
var kjcy$1 = "ќ";
var Kopf$1 = "𝕂";
var kopf$1 = "𝕜";
var Kscr$1 = "𝒦";
var kscr$1 = "𝓀";
var lAarr$1 = "⇚";
var Lacute$1 = "Ĺ";
var lacute$1 = "ĺ";
var laemptyv$1 = "⦴";
var lagran$1 = "ℒ";
var Lambda$1 = "Λ";
var lambda$1 = "λ";
var lang$1 = "⟨";
var Lang$1 = "⟪";
var langd$1 = "⦑";
var langle$1 = "⟨";
var lap$1 = "⪅";
var Laplacetrf$1 = "ℒ";
var laquo$1$1 = "«";
var larrb$1 = "⇤";
var larrbfs$1 = "⤟";
var larr$1 = "←";
var Larr$1 = "↞";
var lArr$1 = "⇐";
var larrfs$1 = "⤝";
var larrhk$1 = "↩";
var larrlp$1 = "↫";
var larrpl$1 = "⤹";
var larrsim$1 = "⥳";
var larrtl$1 = "↢";
var latail$1 = "⤙";
var lAtail$1 = "⤛";
var lat$1 = "⪫";
var late$1 = "⪭";
var lates$1 = "⪭︀";
var lbarr$1 = "⤌";
var lBarr$1 = "⤎";
var lbbrk$1 = "❲";
var lbrace$1 = "{";
var lbrack$1 = "[";
var lbrke$1 = "⦋";
var lbrksld$1 = "⦏";
var lbrkslu$1 = "⦍";
var Lcaron$1 = "Ľ";
var lcaron$1 = "ľ";
var Lcedil$1 = "Ļ";
var lcedil$1 = "ļ";
var lceil$1 = "⌈";
var lcub$1 = "{";
var Lcy$1 = "Л";
var lcy$1 = "л";
var ldca$1 = "⤶";
var ldquo$1 = "“";
var ldquor$1 = "„";
var ldrdhar$1 = "⥧";
var ldrushar$1 = "⥋";
var ldsh$1 = "↲";
var le$1 = "≤";
var lE$1 = "≦";
var LeftAngleBracket$1 = "⟨";
var LeftArrowBar$1 = "⇤";
var leftarrow$1 = "←";
var LeftArrow$1 = "←";
var Leftarrow$1 = "⇐";
var LeftArrowRightArrow$1 = "⇆";
var leftarrowtail$1 = "↢";
var LeftCeiling$1 = "⌈";
var LeftDoubleBracket$1 = "⟦";
var LeftDownTeeVector$1 = "⥡";
var LeftDownVectorBar$1 = "⥙";
var LeftDownVector$1 = "⇃";
var LeftFloor$1 = "⌊";
var leftharpoondown$1 = "↽";
var leftharpoonup$1 = "↼";
var leftleftarrows$1 = "⇇";
var leftrightarrow$1 = "↔";
var LeftRightArrow$1 = "↔";
var Leftrightarrow$1 = "⇔";
var leftrightarrows$1 = "⇆";
var leftrightharpoons$1 = "⇋";
var leftrightsquigarrow$1 = "↭";
var LeftRightVector$1 = "⥎";
var LeftTeeArrow$1 = "↤";
var LeftTee$1 = "⊣";
var LeftTeeVector$1 = "⥚";
var leftthreetimes$1 = "⋋";
var LeftTriangleBar$1 = "⧏";
var LeftTriangle$1 = "⊲";
var LeftTriangleEqual$1 = "⊴";
var LeftUpDownVector$1 = "⥑";
var LeftUpTeeVector$1 = "⥠";
var LeftUpVectorBar$1 = "⥘";
var LeftUpVector$1 = "↿";
var LeftVectorBar$1 = "⥒";
var LeftVector$1 = "↼";
var lEg$1 = "⪋";
var leg$1 = "⋚";
var leq$1 = "≤";
var leqq$1 = "≦";
var leqslant$1 = "⩽";
var lescc$1 = "⪨";
var les$1 = "⩽";
var lesdot$1 = "⩿";
var lesdoto$1 = "⪁";
var lesdotor$1 = "⪃";
var lesg$1 = "⋚︀";
var lesges$1 = "⪓";
var lessapprox$1 = "⪅";
var lessdot$1 = "⋖";
var lesseqgtr$1 = "⋚";
var lesseqqgtr$1 = "⪋";
var LessEqualGreater$1 = "⋚";
var LessFullEqual$1 = "≦";
var LessGreater$1 = "≶";
var lessgtr$1 = "≶";
var LessLess$1 = "⪡";
var lesssim$1 = "≲";
var LessSlantEqual$1 = "⩽";
var LessTilde$1 = "≲";
var lfisht$1 = "⥼";
var lfloor$1 = "⌊";
var Lfr$1 = "𝔏";
var lfr$1 = "𝔩";
var lg$1 = "≶";
var lgE$1 = "⪑";
var lHar$1 = "⥢";
var lhard$1 = "↽";
var lharu$1 = "↼";
var lharul$1 = "⥪";
var lhblk$1 = "▄";
var LJcy$1 = "Љ";
var ljcy$1 = "љ";
var llarr$1 = "⇇";
var ll$1 = "≪";
var Ll$1 = "⋘";
var llcorner$1 = "⌞";
var Lleftarrow$1 = "⇚";
var llhard$1 = "⥫";
var lltri$1 = "◺";
var Lmidot$1 = "Ŀ";
var lmidot$1 = "ŀ";
var lmoustache$1 = "⎰";
var lmoust$1 = "⎰";
var lnap$1 = "⪉";
var lnapprox$1 = "⪉";
var lne$1 = "⪇";
var lnE$1 = "≨";
var lneq$1 = "⪇";
var lneqq$1 = "≨";
var lnsim$1 = "⋦";
var loang$1 = "⟬";
var loarr$1 = "⇽";
var lobrk$1 = "⟦";
var longleftarrow$1 = "⟵";
var LongLeftArrow$1 = "⟵";
var Longleftarrow$1 = "⟸";
var longleftrightarrow$1 = "⟷";
var LongLeftRightArrow$1 = "⟷";
var Longleftrightarrow$1 = "⟺";
var longmapsto$1 = "⟼";
var longrightarrow$1 = "⟶";
var LongRightArrow$1 = "⟶";
var Longrightarrow$1 = "⟹";
var looparrowleft$1 = "↫";
var looparrowright$1 = "↬";
var lopar$1 = "⦅";
var Lopf$1 = "𝕃";
var lopf$1 = "𝕝";
var loplus$1 = "⨭";
var lotimes$1 = "⨴";
var lowast$1 = "∗";
var lowbar$1 = "_";
var LowerLeftArrow$1 = "↙";
var LowerRightArrow$1 = "↘";
var loz$1 = "◊";
var lozenge$1 = "◊";
var lozf$1 = "⧫";
var lpar$1 = "(";
var lparlt$1 = "⦓";
var lrarr$1 = "⇆";
var lrcorner$1 = "⌟";
var lrhar$1 = "⇋";
var lrhard$1 = "⥭";
var lrm$1 = "‎";
var lrtri$1 = "⊿";
var lsaquo$1 = "‹";
var lscr$1 = "𝓁";
var Lscr$1 = "ℒ";
var lsh$1 = "↰";
var Lsh$1 = "↰";
var lsim$1 = "≲";
var lsime$1 = "⪍";
var lsimg$1 = "⪏";
var lsqb$1 = "[";
var lsquo$1 = "‘";
var lsquor$1 = "‚";
var Lstrok$1 = "Ł";
var lstrok$1 = "ł";
var ltcc$1 = "⪦";
var ltcir$1 = "⩹";
var lt$1$2 = "<";
var LT$1$1 = "<";
var Lt$1 = "≪";
var ltdot$1 = "⋖";
var lthree$1 = "⋋";
var ltimes$1 = "⋉";
var ltlarr$1 = "⥶";
var ltquest$1 = "⩻";
var ltri$1 = "◃";
var ltrie$1 = "⊴";
var ltrif$1 = "◂";
var ltrPar$1 = "⦖";
var lurdshar$1 = "⥊";
var luruhar$1 = "⥦";
var lvertneqq$1 = "≨︀";
var lvnE$1 = "≨︀";
var macr$1$1 = "¯";
var male$1 = "♂";
var malt$1 = "✠";
var maltese$1 = "✠";
var map$2 = "↦";
var mapsto$1 = "↦";
var mapstodown$1 = "↧";
var mapstoleft$1 = "↤";
var mapstoup$1 = "↥";
var marker$1 = "▮";
var mcomma$1 = "⨩";
var Mcy$1 = "М";
var mcy$1 = "м";
var mdash$1 = "—";
var mDDot$1 = "∺";
var measuredangle$1 = "∡";
var MediumSpace$1 = " ";
var Mellintrf$1 = "ℳ";
var Mfr$1 = "𝔐";
var mfr$1 = "𝔪";
var mho$1 = "℧";
var micro$1$1 = "µ";
var midast$1 = "*";
var midcir$1 = "⫰";
var mid$1 = "∣";
var middot$1$1 = "·";
var minusb$1 = "⊟";
var minus$1 = "−";
var minusd$1 = "∸";
var minusdu$1 = "⨪";
var MinusPlus$1 = "∓";
var mlcp$1 = "⫛";
var mldr$1 = "…";
var mnplus$1 = "∓";
var models$1 = "⊧";
var Mopf$1 = "𝕄";
var mopf$1 = "𝕞";
var mp$1 = "∓";
var mscr$1 = "𝓂";
var Mscr$1 = "ℳ";
var mstpos$1 = "∾";
var Mu$1 = "Μ";
var mu$1 = "μ";
var multimap$1 = "⊸";
var mumap$1 = "⊸";
var nabla$1 = "∇";
var Nacute$1 = "Ń";
var nacute$1 = "ń";
var nang$1 = "∠⃒";
var nap$1 = "≉";
var napE$1 = "⩰̸";
var napid$1 = "≋̸";
var napos$1 = "ŉ";
var napprox$1 = "≉";
var natural$1 = "♮";
var naturals$1 = "ℕ";
var natur$1 = "♮";
var nbsp$1$1 = " ";
var nbump$1 = "≎̸";
var nbumpe$1 = "≏̸";
var ncap$1 = "⩃";
var Ncaron$1 = "Ň";
var ncaron$1 = "ň";
var Ncedil$1 = "Ņ";
var ncedil$1 = "ņ";
var ncong$1 = "≇";
var ncongdot$1 = "⩭̸";
var ncup$1 = "⩂";
var Ncy$1 = "Н";
var ncy$1 = "н";
var ndash$1 = "–";
var nearhk$1 = "⤤";
var nearr$1 = "↗";
var neArr$1 = "⇗";
var nearrow$1 = "↗";
var ne$1 = "≠";
var nedot$1 = "≐̸";
var NegativeMediumSpace$1 = "​";
var NegativeThickSpace$1 = "​";
var NegativeThinSpace$1 = "​";
var NegativeVeryThinSpace$1 = "​";
var nequiv$1 = "≢";
var nesear$1 = "⤨";
var nesim$1 = "≂̸";
var NestedGreaterGreater$1 = "≫";
var NestedLessLess$1 = "≪";
var NewLine$1 = "\n";
var nexist$1 = "∄";
var nexists$1 = "∄";
var Nfr$1 = "𝔑";
var nfr$1 = "𝔫";
var ngE$1 = "≧̸";
var nge$1 = "≱";
var ngeq$1 = "≱";
var ngeqq$1 = "≧̸";
var ngeqslant$1 = "⩾̸";
var nges$1 = "⩾̸";
var nGg$1 = "⋙̸";
var ngsim$1 = "≵";
var nGt$1 = "≫⃒";
var ngt$1 = "≯";
var ngtr$1 = "≯";
var nGtv$1 = "≫̸";
var nharr$1 = "↮";
var nhArr$1 = "⇎";
var nhpar$1 = "⫲";
var ni$1 = "∋";
var nis$1 = "⋼";
var nisd$1 = "⋺";
var niv$1 = "∋";
var NJcy$1 = "Њ";
var njcy$1 = "њ";
var nlarr$1 = "↚";
var nlArr$1 = "⇍";
var nldr$1 = "‥";
var nlE$1 = "≦̸";
var nle$1 = "≰";
var nleftarrow$1 = "↚";
var nLeftarrow$1 = "⇍";
var nleftrightarrow$1 = "↮";
var nLeftrightarrow$1 = "⇎";
var nleq$1 = "≰";
var nleqq$1 = "≦̸";
var nleqslant$1 = "⩽̸";
var nles$1 = "⩽̸";
var nless$1 = "≮";
var nLl$1 = "⋘̸";
var nlsim$1 = "≴";
var nLt$1 = "≪⃒";
var nlt$1 = "≮";
var nltri$1 = "⋪";
var nltrie$1 = "⋬";
var nLtv$1 = "≪̸";
var nmid$1 = "∤";
var NoBreak$1 = "⁠";
var NonBreakingSpace$1 = " ";
var nopf$1 = "𝕟";
var Nopf$1 = "ℕ";
var Not$1 = "⫬";
var not$1$1 = "¬";
var NotCongruent$1 = "≢";
var NotCupCap$1 = "≭";
var NotDoubleVerticalBar$1 = "∦";
var NotElement$1 = "∉";
var NotEqual$1 = "≠";
var NotEqualTilde$1 = "≂̸";
var NotExists$1 = "∄";
var NotGreater$1 = "≯";
var NotGreaterEqual$1 = "≱";
var NotGreaterFullEqual$1 = "≧̸";
var NotGreaterGreater$1 = "≫̸";
var NotGreaterLess$1 = "≹";
var NotGreaterSlantEqual$1 = "⩾̸";
var NotGreaterTilde$1 = "≵";
var NotHumpDownHump$1 = "≎̸";
var NotHumpEqual$1 = "≏̸";
var notin$1 = "∉";
var notindot$1 = "⋵̸";
var notinE$1 = "⋹̸";
var notinva$1 = "∉";
var notinvb$1 = "⋷";
var notinvc$1 = "⋶";
var NotLeftTriangleBar$1 = "⧏̸";
var NotLeftTriangle$1 = "⋪";
var NotLeftTriangleEqual$1 = "⋬";
var NotLess$1 = "≮";
var NotLessEqual$1 = "≰";
var NotLessGreater$1 = "≸";
var NotLessLess$1 = "≪̸";
var NotLessSlantEqual$1 = "⩽̸";
var NotLessTilde$1 = "≴";
var NotNestedGreaterGreater$1 = "⪢̸";
var NotNestedLessLess$1 = "⪡̸";
var notni$1 = "∌";
var notniva$1 = "∌";
var notnivb$1 = "⋾";
var notnivc$1 = "⋽";
var NotPrecedes$1 = "⊀";
var NotPrecedesEqual$1 = "⪯̸";
var NotPrecedesSlantEqual$1 = "⋠";
var NotReverseElement$1 = "∌";
var NotRightTriangleBar$1 = "⧐̸";
var NotRightTriangle$1 = "⋫";
var NotRightTriangleEqual$1 = "⋭";
var NotSquareSubset$1 = "⊏̸";
var NotSquareSubsetEqual$1 = "⋢";
var NotSquareSuperset$1 = "⊐̸";
var NotSquareSupersetEqual$1 = "⋣";
var NotSubset$1 = "⊂⃒";
var NotSubsetEqual$1 = "⊈";
var NotSucceeds$1 = "⊁";
var NotSucceedsEqual$1 = "⪰̸";
var NotSucceedsSlantEqual$1 = "⋡";
var NotSucceedsTilde$1 = "≿̸";
var NotSuperset$1 = "⊃⃒";
var NotSupersetEqual$1 = "⊉";
var NotTilde$1 = "≁";
var NotTildeEqual$1 = "≄";
var NotTildeFullEqual$1 = "≇";
var NotTildeTilde$1 = "≉";
var NotVerticalBar$1 = "∤";
var nparallel$1 = "∦";
var npar$1 = "∦";
var nparsl$1 = "⫽⃥";
var npart$1 = "∂̸";
var npolint$1 = "⨔";
var npr$1 = "⊀";
var nprcue$1 = "⋠";
var nprec$1 = "⊀";
var npreceq$1 = "⪯̸";
var npre$1 = "⪯̸";
var nrarrc$1 = "⤳̸";
var nrarr$1 = "↛";
var nrArr$1 = "⇏";
var nrarrw$1 = "↝̸";
var nrightarrow$1 = "↛";
var nRightarrow$1 = "⇏";
var nrtri$1 = "⋫";
var nrtrie$1 = "⋭";
var nsc$1 = "⊁";
var nsccue$1 = "⋡";
var nsce$1 = "⪰̸";
var Nscr$1 = "𝒩";
var nscr$1 = "𝓃";
var nshortmid$1 = "∤";
var nshortparallel$1 = "∦";
var nsim$1 = "≁";
var nsime$1 = "≄";
var nsimeq$1 = "≄";
var nsmid$1 = "∤";
var nspar$1 = "∦";
var nsqsube$1 = "⋢";
var nsqsupe$1 = "⋣";
var nsub$1 = "⊄";
var nsubE$1 = "⫅̸";
var nsube$1 = "⊈";
var nsubset$1 = "⊂⃒";
var nsubseteq$1 = "⊈";
var nsubseteqq$1 = "⫅̸";
var nsucc$1 = "⊁";
var nsucceq$1 = "⪰̸";
var nsup$1 = "⊅";
var nsupE$1 = "⫆̸";
var nsupe$1 = "⊉";
var nsupset$1 = "⊃⃒";
var nsupseteq$1 = "⊉";
var nsupseteqq$1 = "⫆̸";
var ntgl$1 = "≹";
var Ntilde$1$1 = "Ñ";
var ntilde$1$1 = "ñ";
var ntlg$1 = "≸";
var ntriangleleft$1 = "⋪";
var ntrianglelefteq$1 = "⋬";
var ntriangleright$1 = "⋫";
var ntrianglerighteq$1 = "⋭";
var Nu$1 = "Ν";
var nu$1 = "ν";
var num$1 = "#";
var numero$1 = "№";
var numsp$1 = " ";
var nvap$1 = "≍⃒";
var nvdash$1 = "⊬";
var nvDash$1 = "⊭";
var nVdash$1 = "⊮";
var nVDash$1 = "⊯";
var nvge$1 = "≥⃒";
var nvgt$1 = ">⃒";
var nvHarr$1 = "⤄";
var nvinfin$1 = "⧞";
var nvlArr$1 = "⤂";
var nvle$1 = "≤⃒";
var nvlt$1 = "<⃒";
var nvltrie$1 = "⊴⃒";
var nvrArr$1 = "⤃";
var nvrtrie$1 = "⊵⃒";
var nvsim$1 = "∼⃒";
var nwarhk$1 = "⤣";
var nwarr$1 = "↖";
var nwArr$1 = "⇖";
var nwarrow$1 = "↖";
var nwnear$1 = "⤧";
var Oacute$1$1 = "Ó";
var oacute$1$1 = "ó";
var oast$1 = "⊛";
var Ocirc$1$1 = "Ô";
var ocirc$1$1 = "ô";
var ocir$1 = "⊚";
var Ocy$1 = "О";
var ocy$1 = "о";
var odash$1 = "⊝";
var Odblac$1 = "Ő";
var odblac$1 = "ő";
var odiv$1 = "⨸";
var odot$1 = "⊙";
var odsold$1 = "⦼";
var OElig$1 = "Œ";
var oelig$1 = "œ";
var ofcir$1 = "⦿";
var Ofr$1 = "𝔒";
var ofr$1 = "𝔬";
var ogon$1 = "˛";
var Ograve$1$1 = "Ò";
var ograve$1$1 = "ò";
var ogt$1 = "⧁";
var ohbar$1 = "⦵";
var ohm$1 = "Ω";
var oint$1 = "∮";
var olarr$1 = "↺";
var olcir$1 = "⦾";
var olcross$1 = "⦻";
var oline$1 = "‾";
var olt$1 = "⧀";
var Omacr$1 = "Ō";
var omacr$1 = "ō";
var Omega$1 = "Ω";
var omega$1 = "ω";
var Omicron$1 = "Ο";
var omicron$1 = "ο";
var omid$1 = "⦶";
var ominus$1 = "⊖";
var Oopf$1 = "𝕆";
var oopf$1 = "𝕠";
var opar$1 = "⦷";
var OpenCurlyDoubleQuote$1 = "“";
var OpenCurlyQuote$1 = "‘";
var operp$1 = "⦹";
var oplus$1 = "⊕";
var orarr$1 = "↻";
var Or$1 = "⩔";
var or$1 = "∨";
var ord$1 = "⩝";
var order$1 = "ℴ";
var orderof$1 = "ℴ";
var ordf$1$1 = "ª";
var ordm$1$1 = "º";
var origof$1 = "⊶";
var oror$1 = "⩖";
var orslope$1 = "⩗";
var orv$1 = "⩛";
var oS$1 = "Ⓢ";
var Oscr$1 = "𝒪";
var oscr$1 = "ℴ";
var Oslash$1$1 = "Ø";
var oslash$1$1 = "ø";
var osol$1 = "⊘";
var Otilde$1$1 = "Õ";
var otilde$1$1 = "õ";
var otimesas$1 = "⨶";
var Otimes$1 = "⨷";
var otimes$1 = "⊗";
var Ouml$1$1 = "Ö";
var ouml$1$1 = "ö";
var ovbar$1 = "⌽";
var OverBar$1 = "‾";
var OverBrace$1 = "⏞";
var OverBracket$1 = "⎴";
var OverParenthesis$1 = "⏜";
var para$1$1 = "¶";
var parallel$1 = "∥";
var par$1 = "∥";
var parsim$1 = "⫳";
var parsl$1 = "⫽";
var part$1 = "∂";
var PartialD$1 = "∂";
var Pcy$1 = "П";
var pcy$1 = "п";
var percnt$1 = "%";
var period$1 = ".";
var permil$1 = "‰";
var perp$1 = "⊥";
var pertenk$1 = "‱";
var Pfr$1 = "𝔓";
var pfr$1 = "𝔭";
var Phi$1 = "Φ";
var phi$1 = "φ";
var phiv$1 = "ϕ";
var phmmat$1 = "ℳ";
var phone$1 = "☎";
var Pi$1 = "Π";
var pi$1 = "π";
var pitchfork$1 = "⋔";
var piv$1 = "ϖ";
var planck$1 = "ℏ";
var planckh$1 = "ℎ";
var plankv$1 = "ℏ";
var plusacir$1 = "⨣";
var plusb$1 = "⊞";
var pluscir$1 = "⨢";
var plus$1 = "+";
var plusdo$1 = "∔";
var plusdu$1 = "⨥";
var pluse$1 = "⩲";
var PlusMinus$1 = "±";
var plusmn$1$1 = "±";
var plussim$1 = "⨦";
var plustwo$1 = "⨧";
var pm$1 = "±";
var Poincareplane$1 = "ℌ";
var pointint$1 = "⨕";
var popf$1 = "𝕡";
var Popf$1 = "ℙ";
var pound$1$1 = "£";
var prap$1 = "⪷";
var Pr$1 = "⪻";
var pr$1 = "≺";
var prcue$1 = "≼";
var precapprox$1 = "⪷";
var prec$1 = "≺";
var preccurlyeq$1 = "≼";
var Precedes$1 = "≺";
var PrecedesEqual$1 = "⪯";
var PrecedesSlantEqual$1 = "≼";
var PrecedesTilde$1 = "≾";
var preceq$1 = "⪯";
var precnapprox$1 = "⪹";
var precneqq$1 = "⪵";
var precnsim$1 = "⋨";
var pre$1 = "⪯";
var prE$1 = "⪳";
var precsim$1 = "≾";
var prime$1 = "′";
var Prime$1 = "″";
var primes$1 = "ℙ";
var prnap$1 = "⪹";
var prnE$1 = "⪵";
var prnsim$1 = "⋨";
var prod$1 = "∏";
var Product$1 = "∏";
var profalar$1 = "⌮";
var profline$1 = "⌒";
var profsurf$1 = "⌓";
var prop$1 = "∝";
var Proportional$1 = "∝";
var Proportion$1 = "∷";
var propto$1 = "∝";
var prsim$1 = "≾";
var prurel$1 = "⊰";
var Pscr$1 = "𝒫";
var pscr$1 = "𝓅";
var Psi$1 = "Ψ";
var psi$1 = "ψ";
var puncsp$1 = " ";
var Qfr$1 = "𝔔";
var qfr$1 = "𝔮";
var qint$1 = "⨌";
var qopf$1 = "𝕢";
var Qopf$1 = "ℚ";
var qprime$1 = "⁗";
var Qscr$1 = "𝒬";
var qscr$1 = "𝓆";
var quaternions$1 = "ℍ";
var quatint$1 = "⨖";
var quest$1 = "?";
var questeq$1 = "≟";
var quot$1$2 = "\"";
var QUOT$1$1 = "\"";
var rAarr$1 = "⇛";
var race$1 = "∽̱";
var Racute$1 = "Ŕ";
var racute$1 = "ŕ";
var radic$1 = "√";
var raemptyv$1 = "⦳";
var rang$1 = "⟩";
var Rang$1 = "⟫";
var rangd$1 = "⦒";
var range$1 = "⦥";
var rangle$1 = "⟩";
var raquo$1$1 = "»";
var rarrap$1 = "⥵";
var rarrb$1 = "⇥";
var rarrbfs$1 = "⤠";
var rarrc$1 = "⤳";
var rarr$1 = "→";
var Rarr$1 = "↠";
var rArr$1 = "⇒";
var rarrfs$1 = "⤞";
var rarrhk$1 = "↪";
var rarrlp$1 = "↬";
var rarrpl$1 = "⥅";
var rarrsim$1 = "⥴";
var Rarrtl$1 = "⤖";
var rarrtl$1 = "↣";
var rarrw$1 = "↝";
var ratail$1 = "⤚";
var rAtail$1 = "⤜";
var ratio$1 = "∶";
var rationals$1 = "ℚ";
var rbarr$1 = "⤍";
var rBarr$1 = "⤏";
var RBarr$1 = "⤐";
var rbbrk$1 = "❳";
var rbrace$1 = "}";
var rbrack$1 = "]";
var rbrke$1 = "⦌";
var rbrksld$1 = "⦎";
var rbrkslu$1 = "⦐";
var Rcaron$1 = "Ř";
var rcaron$1 = "ř";
var Rcedil$1 = "Ŗ";
var rcedil$1 = "ŗ";
var rceil$1 = "⌉";
var rcub$1 = "}";
var Rcy$1 = "Р";
var rcy$1 = "р";
var rdca$1 = "⤷";
var rdldhar$1 = "⥩";
var rdquo$1 = "”";
var rdquor$1 = "”";
var rdsh$1 = "↳";
var real$1 = "ℜ";
var realine$1 = "ℛ";
var realpart$1 = "ℜ";
var reals$1 = "ℝ";
var Re$1 = "ℜ";
var rect$1 = "▭";
var reg$1$1 = "®";
var REG$1$1 = "®";
var ReverseElement$1 = "∋";
var ReverseEquilibrium$1 = "⇋";
var ReverseUpEquilibrium$1 = "⥯";
var rfisht$1 = "⥽";
var rfloor$1 = "⌋";
var rfr$1 = "𝔯";
var Rfr$1 = "ℜ";
var rHar$1 = "⥤";
var rhard$1 = "⇁";
var rharu$1 = "⇀";
var rharul$1 = "⥬";
var Rho$1 = "Ρ";
var rho$1 = "ρ";
var rhov$1 = "ϱ";
var RightAngleBracket$1 = "⟩";
var RightArrowBar$1 = "⇥";
var rightarrow$1 = "→";
var RightArrow$1 = "→";
var Rightarrow$1 = "⇒";
var RightArrowLeftArrow$1 = "⇄";
var rightarrowtail$1 = "↣";
var RightCeiling$1 = "⌉";
var RightDoubleBracket$1 = "⟧";
var RightDownTeeVector$1 = "⥝";
var RightDownVectorBar$1 = "⥕";
var RightDownVector$1 = "⇂";
var RightFloor$1 = "⌋";
var rightharpoondown$1 = "⇁";
var rightharpoonup$1 = "⇀";
var rightleftarrows$1 = "⇄";
var rightleftharpoons$1 = "⇌";
var rightrightarrows$1 = "⇉";
var rightsquigarrow$1 = "↝";
var RightTeeArrow$1 = "↦";
var RightTee$1 = "⊢";
var RightTeeVector$1 = "⥛";
var rightthreetimes$1 = "⋌";
var RightTriangleBar$1 = "⧐";
var RightTriangle$1 = "⊳";
var RightTriangleEqual$1 = "⊵";
var RightUpDownVector$1 = "⥏";
var RightUpTeeVector$1 = "⥜";
var RightUpVectorBar$1 = "⥔";
var RightUpVector$1 = "↾";
var RightVectorBar$1 = "⥓";
var RightVector$1 = "⇀";
var ring$1 = "˚";
var risingdotseq$1 = "≓";
var rlarr$1 = "⇄";
var rlhar$1 = "⇌";
var rlm$1 = "‏";
var rmoustache$1 = "⎱";
var rmoust$1 = "⎱";
var rnmid$1 = "⫮";
var roang$1 = "⟭";
var roarr$1 = "⇾";
var robrk$1 = "⟧";
var ropar$1 = "⦆";
var ropf$1 = "𝕣";
var Ropf$1 = "ℝ";
var roplus$1 = "⨮";
var rotimes$1 = "⨵";
var RoundImplies$1 = "⥰";
var rpar$1 = ")";
var rpargt$1 = "⦔";
var rppolint$1 = "⨒";
var rrarr$1 = "⇉";
var Rrightarrow$1 = "⇛";
var rsaquo$1 = "›";
var rscr$1 = "𝓇";
var Rscr$1 = "ℛ";
var rsh$1 = "↱";
var Rsh$1 = "↱";
var rsqb$1 = "]";
var rsquo$1 = "’";
var rsquor$1 = "’";
var rthree$1 = "⋌";
var rtimes$1 = "⋊";
var rtri$1 = "▹";
var rtrie$1 = "⊵";
var rtrif$1 = "▸";
var rtriltri$1 = "⧎";
var RuleDelayed$1 = "⧴";
var ruluhar$1 = "⥨";
var rx$1 = "℞";
var Sacute$1 = "Ś";
var sacute$1 = "ś";
var sbquo$1 = "‚";
var scap$1 = "⪸";
var Scaron$1 = "Š";
var scaron$1 = "š";
var Sc$1 = "⪼";
var sc$1 = "≻";
var sccue$1 = "≽";
var sce$1 = "⪰";
var scE$1 = "⪴";
var Scedil$1 = "Ş";
var scedil$1 = "ş";
var Scirc$1 = "Ŝ";
var scirc$1 = "ŝ";
var scnap$1 = "⪺";
var scnE$1 = "⪶";
var scnsim$1 = "⋩";
var scpolint$1 = "⨓";
var scsim$1 = "≿";
var Scy$1 = "С";
var scy$1 = "с";
var sdotb$1 = "⊡";
var sdot$1 = "⋅";
var sdote$1 = "⩦";
var searhk$1 = "⤥";
var searr$1 = "↘";
var seArr$1 = "⇘";
var searrow$1 = "↘";
var sect$1$1 = "§";
var semi$1 = ";";
var seswar$1 = "⤩";
var setminus$1 = "∖";
var setmn$1 = "∖";
var sext$1 = "✶";
var Sfr$1 = "𝔖";
var sfr$1 = "𝔰";
var sfrown$1 = "⌢";
var sharp$1 = "♯";
var SHCHcy$1 = "Щ";
var shchcy$1 = "щ";
var SHcy$1 = "Ш";
var shcy$1 = "ш";
var ShortDownArrow$1 = "↓";
var ShortLeftArrow$1 = "←";
var shortmid$1 = "∣";
var shortparallel$1 = "∥";
var ShortRightArrow$1 = "→";
var ShortUpArrow$1 = "↑";
var shy$1$1 = "­";
var Sigma$1 = "Σ";
var sigma$1 = "σ";
var sigmaf$1 = "ς";
var sigmav$1 = "ς";
var sim$1 = "∼";
var simdot$1 = "⩪";
var sime$1 = "≃";
var simeq$1 = "≃";
var simg$1 = "⪞";
var simgE$1 = "⪠";
var siml$1 = "⪝";
var simlE$1 = "⪟";
var simne$1 = "≆";
var simplus$1 = "⨤";
var simrarr$1 = "⥲";
var slarr$1 = "←";
var SmallCircle$1 = "∘";
var smallsetminus$1 = "∖";
var smashp$1 = "⨳";
var smeparsl$1 = "⧤";
var smid$1 = "∣";
var smile$1 = "⌣";
var smt$1 = "⪪";
var smte$1 = "⪬";
var smtes$1 = "⪬︀";
var SOFTcy$1 = "Ь";
var softcy$1 = "ь";
var solbar$1 = "⌿";
var solb$1 = "⧄";
var sol$1 = "/";
var Sopf$1 = "𝕊";
var sopf$1 = "𝕤";
var spades$1 = "♠";
var spadesuit$1 = "♠";
var spar$1 = "∥";
var sqcap$1 = "⊓";
var sqcaps$1 = "⊓︀";
var sqcup$1 = "⊔";
var sqcups$1 = "⊔︀";
var Sqrt$1 = "√";
var sqsub$1 = "⊏";
var sqsube$1 = "⊑";
var sqsubset$1 = "⊏";
var sqsubseteq$1 = "⊑";
var sqsup$1 = "⊐";
var sqsupe$1 = "⊒";
var sqsupset$1 = "⊐";
var sqsupseteq$1 = "⊒";
var square$1 = "□";
var Square$1 = "□";
var SquareIntersection$1 = "⊓";
var SquareSubset$1 = "⊏";
var SquareSubsetEqual$1 = "⊑";
var SquareSuperset$1 = "⊐";
var SquareSupersetEqual$1 = "⊒";
var SquareUnion$1 = "⊔";
var squarf$1 = "▪";
var squ$1 = "□";
var squf$1 = "▪";
var srarr$1 = "→";
var Sscr$1 = "𝒮";
var sscr$1 = "𝓈";
var ssetmn$1 = "∖";
var ssmile$1 = "⌣";
var sstarf$1 = "⋆";
var Star$1 = "⋆";
var star$1 = "☆";
var starf$1 = "★";
var straightepsilon$1 = "ϵ";
var straightphi$1 = "ϕ";
var strns$1 = "¯";
var sub$1 = "⊂";
var Sub$1 = "⋐";
var subdot$1 = "⪽";
var subE$1 = "⫅";
var sube$1 = "⊆";
var subedot$1 = "⫃";
var submult$1 = "⫁";
var subnE$1 = "⫋";
var subne$1 = "⊊";
var subplus$1 = "⪿";
var subrarr$1 = "⥹";
var subset$1 = "⊂";
var Subset$1 = "⋐";
var subseteq$1 = "⊆";
var subseteqq$1 = "⫅";
var SubsetEqual$1 = "⊆";
var subsetneq$1 = "⊊";
var subsetneqq$1 = "⫋";
var subsim$1 = "⫇";
var subsub$1 = "⫕";
var subsup$1 = "⫓";
var succapprox$1 = "⪸";
var succ$1 = "≻";
var succcurlyeq$1 = "≽";
var Succeeds$1 = "≻";
var SucceedsEqual$1 = "⪰";
var SucceedsSlantEqual$1 = "≽";
var SucceedsTilde$1 = "≿";
var succeq$1 = "⪰";
var succnapprox$1 = "⪺";
var succneqq$1 = "⪶";
var succnsim$1 = "⋩";
var succsim$1 = "≿";
var SuchThat$1 = "∋";
var sum$1 = "∑";
var Sum$1 = "∑";
var sung$1 = "♪";
var sup1$1$1 = "¹";
var sup2$1$1 = "²";
var sup3$1$1 = "³";
var sup$1 = "⊃";
var Sup$1 = "⋑";
var supdot$1 = "⪾";
var supdsub$1 = "⫘";
var supE$1 = "⫆";
var supe$1 = "⊇";
var supedot$1 = "⫄";
var Superset$1 = "⊃";
var SupersetEqual$1 = "⊇";
var suphsol$1 = "⟉";
var suphsub$1 = "⫗";
var suplarr$1 = "⥻";
var supmult$1 = "⫂";
var supnE$1 = "⫌";
var supne$1 = "⊋";
var supplus$1 = "⫀";
var supset$1 = "⊃";
var Supset$1 = "⋑";
var supseteq$1 = "⊇";
var supseteqq$1 = "⫆";
var supsetneq$1 = "⊋";
var supsetneqq$1 = "⫌";
var supsim$1 = "⫈";
var supsub$1 = "⫔";
var supsup$1 = "⫖";
var swarhk$1 = "⤦";
var swarr$1 = "↙";
var swArr$1 = "⇙";
var swarrow$1 = "↙";
var swnwar$1 = "⤪";
var szlig$1$1 = "ß";
var Tab$1 = "\t";
var target$1 = "⌖";
var Tau$1 = "Τ";
var tau$1 = "τ";
var tbrk$1 = "⎴";
var Tcaron$1 = "Ť";
var tcaron$1 = "ť";
var Tcedil$1 = "Ţ";
var tcedil$1 = "ţ";
var Tcy$1 = "Т";
var tcy$1 = "т";
var tdot$1 = "⃛";
var telrec$1 = "⌕";
var Tfr$1 = "𝔗";
var tfr$1 = "𝔱";
var there4$1 = "∴";
var therefore$1 = "∴";
var Therefore$1 = "∴";
var Theta$1 = "Θ";
var theta$1 = "θ";
var thetasym$1 = "ϑ";
var thetav$1 = "ϑ";
var thickapprox$1 = "≈";
var thicksim$1 = "∼";
var ThickSpace$1 = "  ";
var ThinSpace$1 = " ";
var thinsp$1 = " ";
var thkap$1 = "≈";
var thksim$1 = "∼";
var THORN$1$1 = "Þ";
var thorn$1$1 = "þ";
var tilde$1 = "˜";
var Tilde$1 = "∼";
var TildeEqual$1 = "≃";
var TildeFullEqual$1 = "≅";
var TildeTilde$1 = "≈";
var timesbar$1 = "⨱";
var timesb$1 = "⊠";
var times$2 = "×";
var timesd$1 = "⨰";
var tint$1 = "∭";
var toea$1 = "⤨";
var topbot$1 = "⌶";
var topcir$1 = "⫱";
var top$1 = "⊤";
var Topf$1 = "𝕋";
var topf$1 = "𝕥";
var topfork$1 = "⫚";
var tosa$1 = "⤩";
var tprime$1 = "‴";
var trade$1 = "™";
var TRADE$1 = "™";
var triangle$1 = "▵";
var triangledown$1 = "▿";
var triangleleft$1 = "◃";
var trianglelefteq$1 = "⊴";
var triangleq$1 = "≜";
var triangleright$1 = "▹";
var trianglerighteq$1 = "⊵";
var tridot$1 = "◬";
var trie$1 = "≜";
var triminus$1 = "⨺";
var TripleDot$1 = "⃛";
var triplus$1 = "⨹";
var trisb$1 = "⧍";
var tritime$1 = "⨻";
var trpezium$1 = "⏢";
var Tscr$1 = "𝒯";
var tscr$1 = "𝓉";
var TScy$1 = "Ц";
var tscy$1 = "ц";
var TSHcy$1 = "Ћ";
var tshcy$1 = "ћ";
var Tstrok$1 = "Ŧ";
var tstrok$1 = "ŧ";
var twixt$1 = "≬";
var twoheadleftarrow$1 = "↞";
var twoheadrightarrow$1 = "↠";
var Uacute$1$1 = "Ú";
var uacute$1$1 = "ú";
var uarr$1 = "↑";
var Uarr$1 = "↟";
var uArr$1 = "⇑";
var Uarrocir$1 = "⥉";
var Ubrcy$1 = "Ў";
var ubrcy$1 = "ў";
var Ubreve$1 = "Ŭ";
var ubreve$1 = "ŭ";
var Ucirc$1$1 = "Û";
var ucirc$1$1 = "û";
var Ucy$1 = "У";
var ucy$1 = "у";
var udarr$1 = "⇅";
var Udblac$1 = "Ű";
var udblac$1 = "ű";
var udhar$1 = "⥮";
var ufisht$1 = "⥾";
var Ufr$1 = "𝔘";
var ufr$1 = "𝔲";
var Ugrave$1$1 = "Ù";
var ugrave$1$1 = "ù";
var uHar$1 = "⥣";
var uharl$1 = "↿";
var uharr$1 = "↾";
var uhblk$1 = "▀";
var ulcorn$1 = "⌜";
var ulcorner$1 = "⌜";
var ulcrop$1 = "⌏";
var ultri$1 = "◸";
var Umacr$1 = "Ū";
var umacr$1 = "ū";
var uml$1$1 = "¨";
var UnderBar$1 = "_";
var UnderBrace$1 = "⏟";
var UnderBracket$1 = "⎵";
var UnderParenthesis$1 = "⏝";
var Union$1 = "⋃";
var UnionPlus$1 = "⊎";
var Uogon$1 = "Ų";
var uogon$1 = "ų";
var Uopf$1 = "𝕌";
var uopf$1 = "𝕦";
var UpArrowBar$1 = "⤒";
var uparrow$1 = "↑";
var UpArrow$1 = "↑";
var Uparrow$1 = "⇑";
var UpArrowDownArrow$1 = "⇅";
var updownarrow$1 = "↕";
var UpDownArrow$1 = "↕";
var Updownarrow$1 = "⇕";
var UpEquilibrium$1 = "⥮";
var upharpoonleft$1 = "↿";
var upharpoonright$1 = "↾";
var uplus$1 = "⊎";
var UpperLeftArrow$1 = "↖";
var UpperRightArrow$1 = "↗";
var upsi$1 = "υ";
var Upsi$1 = "ϒ";
var upsih$1 = "ϒ";
var Upsilon$1 = "Υ";
var upsilon$1 = "υ";
var UpTeeArrow$1 = "↥";
var UpTee$1 = "⊥";
var upuparrows$1 = "⇈";
var urcorn$1 = "⌝";
var urcorner$1 = "⌝";
var urcrop$1 = "⌎";
var Uring$1 = "Ů";
var uring$1 = "ů";
var urtri$1 = "◹";
var Uscr$1 = "𝒰";
var uscr$1 = "𝓊";
var utdot$1 = "⋰";
var Utilde$1 = "Ũ";
var utilde$1 = "ũ";
var utri$1 = "▵";
var utrif$1 = "▴";
var uuarr$1 = "⇈";
var Uuml$1$1 = "Ü";
var uuml$1$1 = "ü";
var uwangle$1 = "⦧";
var vangrt$1 = "⦜";
var varepsilon$1 = "ϵ";
var varkappa$1 = "ϰ";
var varnothing$1 = "∅";
var varphi$1 = "ϕ";
var varpi$1 = "ϖ";
var varpropto$1 = "∝";
var varr$1 = "↕";
var vArr$1 = "⇕";
var varrho$1 = "ϱ";
var varsigma$1 = "ς";
var varsubsetneq$1 = "⊊︀";
var varsubsetneqq$1 = "⫋︀";
var varsupsetneq$1 = "⊋︀";
var varsupsetneqq$1 = "⫌︀";
var vartheta$1 = "ϑ";
var vartriangleleft$1 = "⊲";
var vartriangleright$1 = "⊳";
var vBar$1 = "⫨";
var Vbar$1 = "⫫";
var vBarv$1 = "⫩";
var Vcy$1 = "В";
var vcy$1 = "в";
var vdash$1 = "⊢";
var vDash$1 = "⊨";
var Vdash$1 = "⊩";
var VDash$1 = "⊫";
var Vdashl$1 = "⫦";
var veebar$1 = "⊻";
var vee$1 = "∨";
var Vee$1 = "⋁";
var veeeq$1 = "≚";
var vellip$1 = "⋮";
var verbar$1 = "|";
var Verbar$1 = "‖";
var vert$1 = "|";
var Vert$1 = "‖";
var VerticalBar$1 = "∣";
var VerticalLine$1 = "|";
var VerticalSeparator$1 = "❘";
var VerticalTilde$1 = "≀";
var VeryThinSpace$1 = " ";
var Vfr$1 = "𝔙";
var vfr$1 = "𝔳";
var vltri$1 = "⊲";
var vnsub$1 = "⊂⃒";
var vnsup$1 = "⊃⃒";
var Vopf$1 = "𝕍";
var vopf$1 = "𝕧";
var vprop$1 = "∝";
var vrtri$1 = "⊳";
var Vscr$1 = "𝒱";
var vscr$1 = "𝓋";
var vsubnE$1 = "⫋︀";
var vsubne$1 = "⊊︀";
var vsupnE$1 = "⫌︀";
var vsupne$1 = "⊋︀";
var Vvdash$1 = "⊪";
var vzigzag$1 = "⦚";
var Wcirc$1 = "Ŵ";
var wcirc$1 = "ŵ";
var wedbar$1 = "⩟";
var wedge$1 = "∧";
var Wedge$1 = "⋀";
var wedgeq$1 = "≙";
var weierp$1 = "℘";
var Wfr$1 = "𝔚";
var wfr$1 = "𝔴";
var Wopf$1 = "𝕎";
var wopf$1 = "𝕨";
var wp$1 = "℘";
var wr$1 = "≀";
var wreath$1 = "≀";
var Wscr$1 = "𝒲";
var wscr$1 = "𝓌";
var xcap$1 = "⋂";
var xcirc$1 = "◯";
var xcup$1 = "⋃";
var xdtri$1 = "▽";
var Xfr$1 = "𝔛";
var xfr$1 = "𝔵";
var xharr$1 = "⟷";
var xhArr$1 = "⟺";
var Xi$1 = "Ξ";
var xi$1 = "ξ";
var xlarr$1 = "⟵";
var xlArr$1 = "⟸";
var xmap$1 = "⟼";
var xnis$1 = "⋻";
var xodot$1 = "⨀";
var Xopf$1 = "𝕏";
var xopf$1 = "𝕩";
var xoplus$1 = "⨁";
var xotime$1 = "⨂";
var xrarr$1 = "⟶";
var xrArr$1 = "⟹";
var Xscr$1 = "𝒳";
var xscr$1 = "𝓍";
var xsqcup$1 = "⨆";
var xuplus$1 = "⨄";
var xutri$1 = "△";
var xvee$1 = "⋁";
var xwedge$1 = "⋀";
var Yacute$1$1 = "Ý";
var yacute$1$1 = "ý";
var YAcy$1 = "Я";
var yacy$1 = "я";
var Ycirc$1 = "Ŷ";
var ycirc$1 = "ŷ";
var Ycy$1 = "Ы";
var ycy$1 = "ы";
var yen$1$1 = "¥";
var Yfr$1 = "𝔜";
var yfr$1 = "𝔶";
var YIcy$1 = "Ї";
var yicy$1 = "ї";
var Yopf$1 = "𝕐";
var yopf$1 = "𝕪";
var Yscr$1 = "𝒴";
var yscr$1 = "𝓎";
var YUcy$1 = "Ю";
var yucy$1 = "ю";
var yuml$1$1 = "ÿ";
var Yuml$1 = "Ÿ";
var Zacute$1 = "Ź";
var zacute$1 = "ź";
var Zcaron$1 = "Ž";
var zcaron$1 = "ž";
var Zcy$1 = "З";
var zcy$1 = "з";
var Zdot$1 = "Ż";
var zdot$1 = "ż";
var zeetrf$1 = "ℨ";
var ZeroWidthSpace$1 = "​";
var Zeta$1 = "Ζ";
var zeta$1 = "ζ";
var zfr$1 = "𝔷";
var Zfr$1 = "ℨ";
var ZHcy$1 = "Ж";
var zhcy$1 = "ж";
var zigrarr$1 = "⇝";
var zopf$1 = "𝕫";
var Zopf$1 = "ℤ";
var Zscr$1 = "𝒵";
var zscr$1 = "𝓏";
var zwj$1 = "‍";
var zwnj$1 = "‌";
var entitiesJSON$1 = {
	Aacute: Aacute$1$1,
	aacute: aacute$1$1,
	Abreve: Abreve$1,
	abreve: abreve$1,
	ac: ac$1,
	acd: acd$1,
	acE: acE$1,
	Acirc: Acirc$1$1,
	acirc: acirc$1$1,
	acute: acute$1$1,
	Acy: Acy$1,
	acy: acy$1,
	AElig: AElig$1$1,
	aelig: aelig$1$1,
	af: af$1,
	Afr: Afr$1,
	afr: afr$1,
	Agrave: Agrave$1$1,
	agrave: agrave$1$1,
	alefsym: alefsym$1,
	aleph: aleph$1,
	Alpha: Alpha$1,
	alpha: alpha$1,
	Amacr: Amacr$1,
	amacr: amacr$1,
	amalg: amalg$1,
	amp: amp$1$2,
	AMP: AMP$1$1,
	andand: andand$1,
	And: And$1,
	and: and$1,
	andd: andd$1,
	andslope: andslope$1,
	andv: andv$1,
	ang: ang$1,
	ange: ange$1,
	angle: angle$1,
	angmsdaa: angmsdaa$1,
	angmsdab: angmsdab$1,
	angmsdac: angmsdac$1,
	angmsdad: angmsdad$1,
	angmsdae: angmsdae$1,
	angmsdaf: angmsdaf$1,
	angmsdag: angmsdag$1,
	angmsdah: angmsdah$1,
	angmsd: angmsd$1,
	angrt: angrt$1,
	angrtvb: angrtvb$1,
	angrtvbd: angrtvbd$1,
	angsph: angsph$1,
	angst: angst$1,
	angzarr: angzarr$1,
	Aogon: Aogon$1,
	aogon: aogon$1,
	Aopf: Aopf$1,
	aopf: aopf$1,
	apacir: apacir$1,
	ap: ap$1,
	apE: apE$1,
	ape: ape$1,
	apid: apid$1,
	apos: apos$1$2,
	ApplyFunction: ApplyFunction$1,
	approx: approx$1,
	approxeq: approxeq$1,
	Aring: Aring$1$1,
	aring: aring$1$1,
	Ascr: Ascr$1,
	ascr: ascr$1,
	Assign: Assign$1,
	ast: ast$1,
	asymp: asymp$1,
	asympeq: asympeq$1,
	Atilde: Atilde$1$1,
	atilde: atilde$1$1,
	Auml: Auml$1$1,
	auml: auml$1$1,
	awconint: awconint$1,
	awint: awint$1,
	backcong: backcong$1,
	backepsilon: backepsilon$1,
	backprime: backprime$1,
	backsim: backsim$1,
	backsimeq: backsimeq$1,
	Backslash: Backslash$1,
	Barv: Barv$1,
	barvee: barvee$1,
	barwed: barwed$1,
	Barwed: Barwed$1,
	barwedge: barwedge$1,
	bbrk: bbrk$1,
	bbrktbrk: bbrktbrk$1,
	bcong: bcong$1,
	Bcy: Bcy$1,
	bcy: bcy$1,
	bdquo: bdquo$1,
	becaus: becaus$1,
	because: because$1,
	Because: Because$1,
	bemptyv: bemptyv$1,
	bepsi: bepsi$1,
	bernou: bernou$1,
	Bernoullis: Bernoullis$1,
	Beta: Beta$1,
	beta: beta$1,
	beth: beth$1,
	between: between$1,
	Bfr: Bfr$1,
	bfr: bfr$1,
	bigcap: bigcap$1,
	bigcirc: bigcirc$1,
	bigcup: bigcup$1,
	bigodot: bigodot$1,
	bigoplus: bigoplus$1,
	bigotimes: bigotimes$1,
	bigsqcup: bigsqcup$1,
	bigstar: bigstar$1,
	bigtriangledown: bigtriangledown$1,
	bigtriangleup: bigtriangleup$1,
	biguplus: biguplus$1,
	bigvee: bigvee$1,
	bigwedge: bigwedge$1,
	bkarow: bkarow$1,
	blacklozenge: blacklozenge$1,
	blacksquare: blacksquare$1,
	blacktriangle: blacktriangle$1,
	blacktriangledown: blacktriangledown$1,
	blacktriangleleft: blacktriangleleft$1,
	blacktriangleright: blacktriangleright$1,
	blank: blank$1,
	blk12: blk12$1,
	blk14: blk14$1,
	blk34: blk34$1,
	block: block$1,
	bne: bne$1,
	bnequiv: bnequiv$1,
	bNot: bNot$1,
	bnot: bnot$1,
	Bopf: Bopf$1,
	bopf: bopf$1,
	bot: bot$1,
	bottom: bottom$1,
	bowtie: bowtie$1,
	boxbox: boxbox$1,
	boxdl: boxdl$1,
	boxdL: boxdL$1,
	boxDl: boxDl$1,
	boxDL: boxDL$1,
	boxdr: boxdr$1,
	boxdR: boxdR$1,
	boxDr: boxDr$1,
	boxDR: boxDR$1,
	boxh: boxh$1,
	boxH: boxH$1,
	boxhd: boxhd$1,
	boxHd: boxHd$1,
	boxhD: boxhD$1,
	boxHD: boxHD$1,
	boxhu: boxhu$1,
	boxHu: boxHu$1,
	boxhU: boxhU$1,
	boxHU: boxHU$1,
	boxminus: boxminus$1,
	boxplus: boxplus$1,
	boxtimes: boxtimes$1,
	boxul: boxul$1,
	boxuL: boxuL$1,
	boxUl: boxUl$1,
	boxUL: boxUL$1,
	boxur: boxur$1,
	boxuR: boxuR$1,
	boxUr: boxUr$1,
	boxUR: boxUR$1,
	boxv: boxv$1,
	boxV: boxV$1,
	boxvh: boxvh$1,
	boxvH: boxvH$1,
	boxVh: boxVh$1,
	boxVH: boxVH$1,
	boxvl: boxvl$1,
	boxvL: boxvL$1,
	boxVl: boxVl$1,
	boxVL: boxVL$1,
	boxvr: boxvr$1,
	boxvR: boxvR$1,
	boxVr: boxVr$1,
	boxVR: boxVR$1,
	bprime: bprime$1,
	breve: breve$1,
	Breve: Breve$1,
	brvbar: brvbar$1$1,
	bscr: bscr$1,
	Bscr: Bscr$1,
	bsemi: bsemi$1,
	bsim: bsim$1,
	bsime: bsime$1,
	bsolb: bsolb$1,
	bsol: bsol$1,
	bsolhsub: bsolhsub$1,
	bull: bull$1,
	bullet: bullet$1,
	bump: bump$1,
	bumpE: bumpE$1,
	bumpe: bumpe$1,
	Bumpeq: Bumpeq$1,
	bumpeq: bumpeq$1,
	Cacute: Cacute$1,
	cacute: cacute$1,
	capand: capand$1,
	capbrcup: capbrcup$1,
	capcap: capcap$1,
	cap: cap$1,
	Cap: Cap$1,
	capcup: capcup$1,
	capdot: capdot$1,
	CapitalDifferentialD: CapitalDifferentialD$1,
	caps: caps$1,
	caret: caret$1,
	caron: caron$1,
	Cayleys: Cayleys$1,
	ccaps: ccaps$1,
	Ccaron: Ccaron$1,
	ccaron: ccaron$1,
	Ccedil: Ccedil$1$1,
	ccedil: ccedil$1$1,
	Ccirc: Ccirc$1,
	ccirc: ccirc$1,
	Cconint: Cconint$1,
	ccups: ccups$1,
	ccupssm: ccupssm$1,
	Cdot: Cdot$1,
	cdot: cdot$1,
	cedil: cedil$1$1,
	Cedilla: Cedilla$1,
	cemptyv: cemptyv$1,
	cent: cent$1$1,
	centerdot: centerdot$1,
	CenterDot: CenterDot$1,
	cfr: cfr$1,
	Cfr: Cfr$1,
	CHcy: CHcy$1,
	chcy: chcy$1,
	check: check$1,
	checkmark: checkmark$1,
	Chi: Chi$1,
	chi: chi$1,
	circ: circ$1,
	circeq: circeq$1,
	circlearrowleft: circlearrowleft$1,
	circlearrowright: circlearrowright$1,
	circledast: circledast$1,
	circledcirc: circledcirc$1,
	circleddash: circleddash$1,
	CircleDot: CircleDot$1,
	circledR: circledR$1,
	circledS: circledS$1,
	CircleMinus: CircleMinus$1,
	CirclePlus: CirclePlus$1,
	CircleTimes: CircleTimes$1,
	cir: cir$1,
	cirE: cirE$1,
	cire: cire$1,
	cirfnint: cirfnint$1,
	cirmid: cirmid$1,
	cirscir: cirscir$1,
	ClockwiseContourIntegral: ClockwiseContourIntegral$1,
	CloseCurlyDoubleQuote: CloseCurlyDoubleQuote$1,
	CloseCurlyQuote: CloseCurlyQuote$1,
	clubs: clubs$1,
	clubsuit: clubsuit$1,
	colon: colon$1,
	Colon: Colon$1,
	Colone: Colone$1,
	colone: colone$1,
	coloneq: coloneq$1,
	comma: comma$1,
	commat: commat$1,
	comp: comp$1,
	compfn: compfn$1,
	complement: complement$1,
	complexes: complexes$1,
	cong: cong$1,
	congdot: congdot$1,
	Congruent: Congruent$1,
	conint: conint$1,
	Conint: Conint$1,
	ContourIntegral: ContourIntegral$1,
	copf: copf$1,
	Copf: Copf$1,
	coprod: coprod$1,
	Coproduct: Coproduct$1,
	copy: copy$1$1,
	COPY: COPY$1$1,
	copysr: copysr$1,
	CounterClockwiseContourIntegral: CounterClockwiseContourIntegral$1,
	crarr: crarr$1,
	cross: cross$1,
	Cross: Cross$1,
	Cscr: Cscr$1,
	cscr: cscr$1,
	csub: csub$1,
	csube: csube$1,
	csup: csup$1,
	csupe: csupe$1,
	ctdot: ctdot$1,
	cudarrl: cudarrl$1,
	cudarrr: cudarrr$1,
	cuepr: cuepr$1,
	cuesc: cuesc$1,
	cularr: cularr$1,
	cularrp: cularrp$1,
	cupbrcap: cupbrcap$1,
	cupcap: cupcap$1,
	CupCap: CupCap$1,
	cup: cup$1,
	Cup: Cup$1,
	cupcup: cupcup$1,
	cupdot: cupdot$1,
	cupor: cupor$1,
	cups: cups$1,
	curarr: curarr$1,
	curarrm: curarrm$1,
	curlyeqprec: curlyeqprec$1,
	curlyeqsucc: curlyeqsucc$1,
	curlyvee: curlyvee$1,
	curlywedge: curlywedge$1,
	curren: curren$1$1,
	curvearrowleft: curvearrowleft$1,
	curvearrowright: curvearrowright$1,
	cuvee: cuvee$1,
	cuwed: cuwed$1,
	cwconint: cwconint$1,
	cwint: cwint$1,
	cylcty: cylcty$1,
	dagger: dagger$1,
	Dagger: Dagger$1,
	daleth: daleth$1,
	darr: darr$1,
	Darr: Darr$1,
	dArr: dArr$1,
	dash: dash$1,
	Dashv: Dashv$1,
	dashv: dashv$1,
	dbkarow: dbkarow$1,
	dblac: dblac$1,
	Dcaron: Dcaron$1,
	dcaron: dcaron$1,
	Dcy: Dcy$1,
	dcy: dcy$1,
	ddagger: ddagger$1,
	ddarr: ddarr$1,
	DD: DD$1,
	dd: dd$1,
	DDotrahd: DDotrahd$1,
	ddotseq: ddotseq$1,
	deg: deg$1$1,
	Del: Del$1,
	Delta: Delta$1,
	delta: delta$1,
	demptyv: demptyv$1,
	dfisht: dfisht$1,
	Dfr: Dfr$1,
	dfr: dfr$1,
	dHar: dHar$1,
	dharl: dharl$1,
	dharr: dharr$1,
	DiacriticalAcute: DiacriticalAcute$1,
	DiacriticalDot: DiacriticalDot$1,
	DiacriticalDoubleAcute: DiacriticalDoubleAcute$1,
	DiacriticalGrave: DiacriticalGrave$1,
	DiacriticalTilde: DiacriticalTilde$1,
	diam: diam$1,
	diamond: diamond$1,
	Diamond: Diamond$1,
	diamondsuit: diamondsuit$1,
	diams: diams$1,
	die: die$1,
	DifferentialD: DifferentialD$1,
	digamma: digamma$1,
	disin: disin$1,
	div: div$1,
	divide: divide$1$1,
	divideontimes: divideontimes$1,
	divonx: divonx$1,
	DJcy: DJcy$1,
	djcy: djcy$1,
	dlcorn: dlcorn$1,
	dlcrop: dlcrop$1,
	dollar: dollar$1,
	Dopf: Dopf$1,
	dopf: dopf$1,
	Dot: Dot$1,
	dot: dot$1,
	DotDot: DotDot$1,
	doteq: doteq$1,
	doteqdot: doteqdot$1,
	DotEqual: DotEqual$1,
	dotminus: dotminus$1,
	dotplus: dotplus$1,
	dotsquare: dotsquare$1,
	doublebarwedge: doublebarwedge$1,
	DoubleContourIntegral: DoubleContourIntegral$1,
	DoubleDot: DoubleDot$1,
	DoubleDownArrow: DoubleDownArrow$1,
	DoubleLeftArrow: DoubleLeftArrow$1,
	DoubleLeftRightArrow: DoubleLeftRightArrow$1,
	DoubleLeftTee: DoubleLeftTee$1,
	DoubleLongLeftArrow: DoubleLongLeftArrow$1,
	DoubleLongLeftRightArrow: DoubleLongLeftRightArrow$1,
	DoubleLongRightArrow: DoubleLongRightArrow$1,
	DoubleRightArrow: DoubleRightArrow$1,
	DoubleRightTee: DoubleRightTee$1,
	DoubleUpArrow: DoubleUpArrow$1,
	DoubleUpDownArrow: DoubleUpDownArrow$1,
	DoubleVerticalBar: DoubleVerticalBar$1,
	DownArrowBar: DownArrowBar$1,
	downarrow: downarrow$1,
	DownArrow: DownArrow$1,
	Downarrow: Downarrow$1,
	DownArrowUpArrow: DownArrowUpArrow$1,
	DownBreve: DownBreve$1,
	downdownarrows: downdownarrows$1,
	downharpoonleft: downharpoonleft$1,
	downharpoonright: downharpoonright$1,
	DownLeftRightVector: DownLeftRightVector$1,
	DownLeftTeeVector: DownLeftTeeVector$1,
	DownLeftVectorBar: DownLeftVectorBar$1,
	DownLeftVector: DownLeftVector$1,
	DownRightTeeVector: DownRightTeeVector$1,
	DownRightVectorBar: DownRightVectorBar$1,
	DownRightVector: DownRightVector$1,
	DownTeeArrow: DownTeeArrow$1,
	DownTee: DownTee$1,
	drbkarow: drbkarow$1,
	drcorn: drcorn$1,
	drcrop: drcrop$1,
	Dscr: Dscr$1,
	dscr: dscr$1,
	DScy: DScy$1,
	dscy: dscy$1,
	dsol: dsol$1,
	Dstrok: Dstrok$1,
	dstrok: dstrok$1,
	dtdot: dtdot$1,
	dtri: dtri$1,
	dtrif: dtrif$1,
	duarr: duarr$1,
	duhar: duhar$1,
	dwangle: dwangle$1,
	DZcy: DZcy$1,
	dzcy: dzcy$1,
	dzigrarr: dzigrarr$1,
	Eacute: Eacute$1$1,
	eacute: eacute$1$1,
	easter: easter$1,
	Ecaron: Ecaron$1,
	ecaron: ecaron$1,
	Ecirc: Ecirc$1$1,
	ecirc: ecirc$1$1,
	ecir: ecir$1,
	ecolon: ecolon$1,
	Ecy: Ecy$1,
	ecy: ecy$1,
	eDDot: eDDot$1,
	Edot: Edot$1,
	edot: edot$1,
	eDot: eDot$1,
	ee: ee$1,
	efDot: efDot$1,
	Efr: Efr$1,
	efr: efr$1,
	eg: eg$1,
	Egrave: Egrave$1$1,
	egrave: egrave$1$1,
	egs: egs$1,
	egsdot: egsdot$1,
	el: el$1,
	Element: Element$1,
	elinters: elinters$1,
	ell: ell$1,
	els: els$1,
	elsdot: elsdot$1,
	Emacr: Emacr$1,
	emacr: emacr$1,
	empty: empty$1,
	emptyset: emptyset$1,
	EmptySmallSquare: EmptySmallSquare$1,
	emptyv: emptyv$1,
	EmptyVerySmallSquare: EmptyVerySmallSquare$1,
	emsp13: emsp13$1,
	emsp14: emsp14$1,
	emsp: emsp$1,
	ENG: ENG$1,
	eng: eng$1,
	ensp: ensp$1,
	Eogon: Eogon$1,
	eogon: eogon$1,
	Eopf: Eopf$1,
	eopf: eopf$1,
	epar: epar$1,
	eparsl: eparsl$1,
	eplus: eplus$1,
	epsi: epsi$1,
	Epsilon: Epsilon$1,
	epsilon: epsilon$1,
	epsiv: epsiv$1,
	eqcirc: eqcirc$1,
	eqcolon: eqcolon$1,
	eqsim: eqsim$1,
	eqslantgtr: eqslantgtr$1,
	eqslantless: eqslantless$1,
	Equal: Equal$1,
	equals: equals$1,
	EqualTilde: EqualTilde$1,
	equest: equest$1,
	Equilibrium: Equilibrium$1,
	equiv: equiv$1,
	equivDD: equivDD$1,
	eqvparsl: eqvparsl$1,
	erarr: erarr$1,
	erDot: erDot$1,
	escr: escr$1,
	Escr: Escr$1,
	esdot: esdot$1,
	Esim: Esim$1,
	esim: esim$1,
	Eta: Eta$1,
	eta: eta$1,
	ETH: ETH$1$1,
	eth: eth$1$1,
	Euml: Euml$1$1,
	euml: euml$1$1,
	euro: euro$1,
	excl: excl$1,
	exist: exist$1,
	Exists: Exists$1,
	expectation: expectation$1,
	exponentiale: exponentiale$1,
	ExponentialE: ExponentialE$1,
	fallingdotseq: fallingdotseq$1,
	Fcy: Fcy$1,
	fcy: fcy$1,
	female: female$1,
	ffilig: ffilig$1,
	fflig: fflig$1,
	ffllig: ffllig$1,
	Ffr: Ffr$1,
	ffr: ffr$1,
	filig: filig$1,
	FilledSmallSquare: FilledSmallSquare$1,
	FilledVerySmallSquare: FilledVerySmallSquare$1,
	fjlig: fjlig$1,
	flat: flat$1,
	fllig: fllig$1,
	fltns: fltns$1,
	fnof: fnof$1,
	Fopf: Fopf$1,
	fopf: fopf$1,
	forall: forall$1,
	ForAll: ForAll$1,
	fork: fork$1,
	forkv: forkv$1,
	Fouriertrf: Fouriertrf$1,
	fpartint: fpartint$1,
	frac12: frac12$1$1,
	frac13: frac13$1,
	frac14: frac14$1$1,
	frac15: frac15$1,
	frac16: frac16$1,
	frac18: frac18$1,
	frac23: frac23$1,
	frac25: frac25$1,
	frac34: frac34$1$1,
	frac35: frac35$1,
	frac38: frac38$1,
	frac45: frac45$1,
	frac56: frac56$1,
	frac58: frac58$1,
	frac78: frac78$1,
	frasl: frasl$1,
	frown: frown$1,
	fscr: fscr$1,
	Fscr: Fscr$1,
	gacute: gacute$1,
	Gamma: Gamma$1,
	gamma: gamma$1,
	Gammad: Gammad$1,
	gammad: gammad$1,
	gap: gap$1,
	Gbreve: Gbreve$1,
	gbreve: gbreve$1,
	Gcedil: Gcedil$1,
	Gcirc: Gcirc$1,
	gcirc: gcirc$1,
	Gcy: Gcy$1,
	gcy: gcy$1,
	Gdot: Gdot$1,
	gdot: gdot$1,
	ge: ge$1,
	gE: gE$1,
	gEl: gEl$1,
	gel: gel$1,
	geq: geq$1,
	geqq: geqq$1,
	geqslant: geqslant$1,
	gescc: gescc$1,
	ges: ges$1,
	gesdot: gesdot$1,
	gesdoto: gesdoto$1,
	gesdotol: gesdotol$1,
	gesl: gesl$1,
	gesles: gesles$1,
	Gfr: Gfr$1,
	gfr: gfr$1,
	gg: gg$1,
	Gg: Gg$1,
	ggg: ggg$1,
	gimel: gimel$1,
	GJcy: GJcy$1,
	gjcy: gjcy$1,
	gla: gla$1,
	gl: gl$1,
	glE: glE$1,
	glj: glj$1,
	gnap: gnap$1,
	gnapprox: gnapprox$1,
	gne: gne$1,
	gnE: gnE$1,
	gneq: gneq$1,
	gneqq: gneqq$1,
	gnsim: gnsim$1,
	Gopf: Gopf$1,
	gopf: gopf$1,
	grave: grave$1,
	GreaterEqual: GreaterEqual$1,
	GreaterEqualLess: GreaterEqualLess$1,
	GreaterFullEqual: GreaterFullEqual$1,
	GreaterGreater: GreaterGreater$1,
	GreaterLess: GreaterLess$1,
	GreaterSlantEqual: GreaterSlantEqual$1,
	GreaterTilde: GreaterTilde$1,
	Gscr: Gscr$1,
	gscr: gscr$1,
	gsim: gsim$1,
	gsime: gsime$1,
	gsiml: gsiml$1,
	gtcc: gtcc$1,
	gtcir: gtcir$1,
	gt: gt$1$2,
	GT: GT$1$1,
	Gt: Gt$1,
	gtdot: gtdot$1,
	gtlPar: gtlPar$1,
	gtquest: gtquest$1,
	gtrapprox: gtrapprox$1,
	gtrarr: gtrarr$1,
	gtrdot: gtrdot$1,
	gtreqless: gtreqless$1,
	gtreqqless: gtreqqless$1,
	gtrless: gtrless$1,
	gtrsim: gtrsim$1,
	gvertneqq: gvertneqq$1,
	gvnE: gvnE$1,
	Hacek: Hacek$1,
	hairsp: hairsp$1,
	half: half$1,
	hamilt: hamilt$1,
	HARDcy: HARDcy$1,
	hardcy: hardcy$1,
	harrcir: harrcir$1,
	harr: harr$1,
	hArr: hArr$1,
	harrw: harrw$1,
	Hat: Hat$1,
	hbar: hbar$1,
	Hcirc: Hcirc$1,
	hcirc: hcirc$1,
	hearts: hearts$1,
	heartsuit: heartsuit$1,
	hellip: hellip$1,
	hercon: hercon$1,
	hfr: hfr$1,
	Hfr: Hfr$1,
	HilbertSpace: HilbertSpace$1,
	hksearow: hksearow$1,
	hkswarow: hkswarow$1,
	hoarr: hoarr$1,
	homtht: homtht$1,
	hookleftarrow: hookleftarrow$1,
	hookrightarrow: hookrightarrow$1,
	hopf: hopf$1,
	Hopf: Hopf$1,
	horbar: horbar$1,
	HorizontalLine: HorizontalLine$1,
	hscr: hscr$1,
	Hscr: Hscr$1,
	hslash: hslash$1,
	Hstrok: Hstrok$1,
	hstrok: hstrok$1,
	HumpDownHump: HumpDownHump$1,
	HumpEqual: HumpEqual$1,
	hybull: hybull$1,
	hyphen: hyphen$1,
	Iacute: Iacute$1$1,
	iacute: iacute$1$1,
	ic: ic$1,
	Icirc: Icirc$1$1,
	icirc: icirc$1$1,
	Icy: Icy$1,
	icy: icy$1,
	Idot: Idot$1,
	IEcy: IEcy$1,
	iecy: iecy$1,
	iexcl: iexcl$1$1,
	iff: iff$1,
	ifr: ifr$1,
	Ifr: Ifr$1,
	Igrave: Igrave$1$1,
	igrave: igrave$1$1,
	ii: ii$1,
	iiiint: iiiint$1,
	iiint: iiint$1,
	iinfin: iinfin$1,
	iiota: iiota$1,
	IJlig: IJlig$1,
	ijlig: ijlig$1,
	Imacr: Imacr$1,
	imacr: imacr$1,
	image: image$1,
	ImaginaryI: ImaginaryI$1,
	imagline: imagline$1,
	imagpart: imagpart$1,
	imath: imath$1,
	Im: Im$1,
	imof: imof$1,
	imped: imped$1,
	Implies: Implies$1,
	incare: incare$1,
	infin: infin$1,
	infintie: infintie$1,
	inodot: inodot$1,
	intcal: intcal$1,
	int: int$1,
	Int: Int$1,
	integers: integers$1,
	Integral: Integral$1,
	intercal: intercal$1,
	Intersection: Intersection$1,
	intlarhk: intlarhk$1,
	intprod: intprod$1,
	InvisibleComma: InvisibleComma$1,
	InvisibleTimes: InvisibleTimes$1,
	IOcy: IOcy$1,
	iocy: iocy$1,
	Iogon: Iogon$1,
	iogon: iogon$1,
	Iopf: Iopf$1,
	iopf: iopf$1,
	Iota: Iota$1,
	iota: iota$1,
	iprod: iprod$1,
	iquest: iquest$1$1,
	iscr: iscr$1,
	Iscr: Iscr$1,
	isin: isin$1,
	isindot: isindot$1,
	isinE: isinE$1,
	isins: isins$1,
	isinsv: isinsv$1,
	isinv: isinv$1,
	it: it$1,
	Itilde: Itilde$1,
	itilde: itilde$1,
	Iukcy: Iukcy$1,
	iukcy: iukcy$1,
	Iuml: Iuml$1$1,
	iuml: iuml$1$1,
	Jcirc: Jcirc$1,
	jcirc: jcirc$1,
	Jcy: Jcy$1,
	jcy: jcy$1,
	Jfr: Jfr$1,
	jfr: jfr$1,
	jmath: jmath$1,
	Jopf: Jopf$1,
	jopf: jopf$1,
	Jscr: Jscr$1,
	jscr: jscr$1,
	Jsercy: Jsercy$1,
	jsercy: jsercy$1,
	Jukcy: Jukcy$1,
	jukcy: jukcy$1,
	Kappa: Kappa$1,
	kappa: kappa$1,
	kappav: kappav$1,
	Kcedil: Kcedil$1,
	kcedil: kcedil$1,
	Kcy: Kcy$1,
	kcy: kcy$1,
	Kfr: Kfr$1,
	kfr: kfr$1,
	kgreen: kgreen$1,
	KHcy: KHcy$1,
	khcy: khcy$1,
	KJcy: KJcy$1,
	kjcy: kjcy$1,
	Kopf: Kopf$1,
	kopf: kopf$1,
	Kscr: Kscr$1,
	kscr: kscr$1,
	lAarr: lAarr$1,
	Lacute: Lacute$1,
	lacute: lacute$1,
	laemptyv: laemptyv$1,
	lagran: lagran$1,
	Lambda: Lambda$1,
	lambda: lambda$1,
	lang: lang$1,
	Lang: Lang$1,
	langd: langd$1,
	langle: langle$1,
	lap: lap$1,
	Laplacetrf: Laplacetrf$1,
	laquo: laquo$1$1,
	larrb: larrb$1,
	larrbfs: larrbfs$1,
	larr: larr$1,
	Larr: Larr$1,
	lArr: lArr$1,
	larrfs: larrfs$1,
	larrhk: larrhk$1,
	larrlp: larrlp$1,
	larrpl: larrpl$1,
	larrsim: larrsim$1,
	larrtl: larrtl$1,
	latail: latail$1,
	lAtail: lAtail$1,
	lat: lat$1,
	late: late$1,
	lates: lates$1,
	lbarr: lbarr$1,
	lBarr: lBarr$1,
	lbbrk: lbbrk$1,
	lbrace: lbrace$1,
	lbrack: lbrack$1,
	lbrke: lbrke$1,
	lbrksld: lbrksld$1,
	lbrkslu: lbrkslu$1,
	Lcaron: Lcaron$1,
	lcaron: lcaron$1,
	Lcedil: Lcedil$1,
	lcedil: lcedil$1,
	lceil: lceil$1,
	lcub: lcub$1,
	Lcy: Lcy$1,
	lcy: lcy$1,
	ldca: ldca$1,
	ldquo: ldquo$1,
	ldquor: ldquor$1,
	ldrdhar: ldrdhar$1,
	ldrushar: ldrushar$1,
	ldsh: ldsh$1,
	le: le$1,
	lE: lE$1,
	LeftAngleBracket: LeftAngleBracket$1,
	LeftArrowBar: LeftArrowBar$1,
	leftarrow: leftarrow$1,
	LeftArrow: LeftArrow$1,
	Leftarrow: Leftarrow$1,
	LeftArrowRightArrow: LeftArrowRightArrow$1,
	leftarrowtail: leftarrowtail$1,
	LeftCeiling: LeftCeiling$1,
	LeftDoubleBracket: LeftDoubleBracket$1,
	LeftDownTeeVector: LeftDownTeeVector$1,
	LeftDownVectorBar: LeftDownVectorBar$1,
	LeftDownVector: LeftDownVector$1,
	LeftFloor: LeftFloor$1,
	leftharpoondown: leftharpoondown$1,
	leftharpoonup: leftharpoonup$1,
	leftleftarrows: leftleftarrows$1,
	leftrightarrow: leftrightarrow$1,
	LeftRightArrow: LeftRightArrow$1,
	Leftrightarrow: Leftrightarrow$1,
	leftrightarrows: leftrightarrows$1,
	leftrightharpoons: leftrightharpoons$1,
	leftrightsquigarrow: leftrightsquigarrow$1,
	LeftRightVector: LeftRightVector$1,
	LeftTeeArrow: LeftTeeArrow$1,
	LeftTee: LeftTee$1,
	LeftTeeVector: LeftTeeVector$1,
	leftthreetimes: leftthreetimes$1,
	LeftTriangleBar: LeftTriangleBar$1,
	LeftTriangle: LeftTriangle$1,
	LeftTriangleEqual: LeftTriangleEqual$1,
	LeftUpDownVector: LeftUpDownVector$1,
	LeftUpTeeVector: LeftUpTeeVector$1,
	LeftUpVectorBar: LeftUpVectorBar$1,
	LeftUpVector: LeftUpVector$1,
	LeftVectorBar: LeftVectorBar$1,
	LeftVector: LeftVector$1,
	lEg: lEg$1,
	leg: leg$1,
	leq: leq$1,
	leqq: leqq$1,
	leqslant: leqslant$1,
	lescc: lescc$1,
	les: les$1,
	lesdot: lesdot$1,
	lesdoto: lesdoto$1,
	lesdotor: lesdotor$1,
	lesg: lesg$1,
	lesges: lesges$1,
	lessapprox: lessapprox$1,
	lessdot: lessdot$1,
	lesseqgtr: lesseqgtr$1,
	lesseqqgtr: lesseqqgtr$1,
	LessEqualGreater: LessEqualGreater$1,
	LessFullEqual: LessFullEqual$1,
	LessGreater: LessGreater$1,
	lessgtr: lessgtr$1,
	LessLess: LessLess$1,
	lesssim: lesssim$1,
	LessSlantEqual: LessSlantEqual$1,
	LessTilde: LessTilde$1,
	lfisht: lfisht$1,
	lfloor: lfloor$1,
	Lfr: Lfr$1,
	lfr: lfr$1,
	lg: lg$1,
	lgE: lgE$1,
	lHar: lHar$1,
	lhard: lhard$1,
	lharu: lharu$1,
	lharul: lharul$1,
	lhblk: lhblk$1,
	LJcy: LJcy$1,
	ljcy: ljcy$1,
	llarr: llarr$1,
	ll: ll$1,
	Ll: Ll$1,
	llcorner: llcorner$1,
	Lleftarrow: Lleftarrow$1,
	llhard: llhard$1,
	lltri: lltri$1,
	Lmidot: Lmidot$1,
	lmidot: lmidot$1,
	lmoustache: lmoustache$1,
	lmoust: lmoust$1,
	lnap: lnap$1,
	lnapprox: lnapprox$1,
	lne: lne$1,
	lnE: lnE$1,
	lneq: lneq$1,
	lneqq: lneqq$1,
	lnsim: lnsim$1,
	loang: loang$1,
	loarr: loarr$1,
	lobrk: lobrk$1,
	longleftarrow: longleftarrow$1,
	LongLeftArrow: LongLeftArrow$1,
	Longleftarrow: Longleftarrow$1,
	longleftrightarrow: longleftrightarrow$1,
	LongLeftRightArrow: LongLeftRightArrow$1,
	Longleftrightarrow: Longleftrightarrow$1,
	longmapsto: longmapsto$1,
	longrightarrow: longrightarrow$1,
	LongRightArrow: LongRightArrow$1,
	Longrightarrow: Longrightarrow$1,
	looparrowleft: looparrowleft$1,
	looparrowright: looparrowright$1,
	lopar: lopar$1,
	Lopf: Lopf$1,
	lopf: lopf$1,
	loplus: loplus$1,
	lotimes: lotimes$1,
	lowast: lowast$1,
	lowbar: lowbar$1,
	LowerLeftArrow: LowerLeftArrow$1,
	LowerRightArrow: LowerRightArrow$1,
	loz: loz$1,
	lozenge: lozenge$1,
	lozf: lozf$1,
	lpar: lpar$1,
	lparlt: lparlt$1,
	lrarr: lrarr$1,
	lrcorner: lrcorner$1,
	lrhar: lrhar$1,
	lrhard: lrhard$1,
	lrm: lrm$1,
	lrtri: lrtri$1,
	lsaquo: lsaquo$1,
	lscr: lscr$1,
	Lscr: Lscr$1,
	lsh: lsh$1,
	Lsh: Lsh$1,
	lsim: lsim$1,
	lsime: lsime$1,
	lsimg: lsimg$1,
	lsqb: lsqb$1,
	lsquo: lsquo$1,
	lsquor: lsquor$1,
	Lstrok: Lstrok$1,
	lstrok: lstrok$1,
	ltcc: ltcc$1,
	ltcir: ltcir$1,
	lt: lt$1$2,
	LT: LT$1$1,
	Lt: Lt$1,
	ltdot: ltdot$1,
	lthree: lthree$1,
	ltimes: ltimes$1,
	ltlarr: ltlarr$1,
	ltquest: ltquest$1,
	ltri: ltri$1,
	ltrie: ltrie$1,
	ltrif: ltrif$1,
	ltrPar: ltrPar$1,
	lurdshar: lurdshar$1,
	luruhar: luruhar$1,
	lvertneqq: lvertneqq$1,
	lvnE: lvnE$1,
	macr: macr$1$1,
	male: male$1,
	malt: malt$1,
	maltese: maltese$1,
	map: map$2,
	mapsto: mapsto$1,
	mapstodown: mapstodown$1,
	mapstoleft: mapstoleft$1,
	mapstoup: mapstoup$1,
	marker: marker$1,
	mcomma: mcomma$1,
	Mcy: Mcy$1,
	mcy: mcy$1,
	mdash: mdash$1,
	mDDot: mDDot$1,
	measuredangle: measuredangle$1,
	MediumSpace: MediumSpace$1,
	Mellintrf: Mellintrf$1,
	Mfr: Mfr$1,
	mfr: mfr$1,
	mho: mho$1,
	micro: micro$1$1,
	midast: midast$1,
	midcir: midcir$1,
	mid: mid$1,
	middot: middot$1$1,
	minusb: minusb$1,
	minus: minus$1,
	minusd: minusd$1,
	minusdu: minusdu$1,
	MinusPlus: MinusPlus$1,
	mlcp: mlcp$1,
	mldr: mldr$1,
	mnplus: mnplus$1,
	models: models$1,
	Mopf: Mopf$1,
	mopf: mopf$1,
	mp: mp$1,
	mscr: mscr$1,
	Mscr: Mscr$1,
	mstpos: mstpos$1,
	Mu: Mu$1,
	mu: mu$1,
	multimap: multimap$1,
	mumap: mumap$1,
	nabla: nabla$1,
	Nacute: Nacute$1,
	nacute: nacute$1,
	nang: nang$1,
	nap: nap$1,
	napE: napE$1,
	napid: napid$1,
	napos: napos$1,
	napprox: napprox$1,
	natural: natural$1,
	naturals: naturals$1,
	natur: natur$1,
	nbsp: nbsp$1$1,
	nbump: nbump$1,
	nbumpe: nbumpe$1,
	ncap: ncap$1,
	Ncaron: Ncaron$1,
	ncaron: ncaron$1,
	Ncedil: Ncedil$1,
	ncedil: ncedil$1,
	ncong: ncong$1,
	ncongdot: ncongdot$1,
	ncup: ncup$1,
	Ncy: Ncy$1,
	ncy: ncy$1,
	ndash: ndash$1,
	nearhk: nearhk$1,
	nearr: nearr$1,
	neArr: neArr$1,
	nearrow: nearrow$1,
	ne: ne$1,
	nedot: nedot$1,
	NegativeMediumSpace: NegativeMediumSpace$1,
	NegativeThickSpace: NegativeThickSpace$1,
	NegativeThinSpace: NegativeThinSpace$1,
	NegativeVeryThinSpace: NegativeVeryThinSpace$1,
	nequiv: nequiv$1,
	nesear: nesear$1,
	nesim: nesim$1,
	NestedGreaterGreater: NestedGreaterGreater$1,
	NestedLessLess: NestedLessLess$1,
	NewLine: NewLine$1,
	nexist: nexist$1,
	nexists: nexists$1,
	Nfr: Nfr$1,
	nfr: nfr$1,
	ngE: ngE$1,
	nge: nge$1,
	ngeq: ngeq$1,
	ngeqq: ngeqq$1,
	ngeqslant: ngeqslant$1,
	nges: nges$1,
	nGg: nGg$1,
	ngsim: ngsim$1,
	nGt: nGt$1,
	ngt: ngt$1,
	ngtr: ngtr$1,
	nGtv: nGtv$1,
	nharr: nharr$1,
	nhArr: nhArr$1,
	nhpar: nhpar$1,
	ni: ni$1,
	nis: nis$1,
	nisd: nisd$1,
	niv: niv$1,
	NJcy: NJcy$1,
	njcy: njcy$1,
	nlarr: nlarr$1,
	nlArr: nlArr$1,
	nldr: nldr$1,
	nlE: nlE$1,
	nle: nle$1,
	nleftarrow: nleftarrow$1,
	nLeftarrow: nLeftarrow$1,
	nleftrightarrow: nleftrightarrow$1,
	nLeftrightarrow: nLeftrightarrow$1,
	nleq: nleq$1,
	nleqq: nleqq$1,
	nleqslant: nleqslant$1,
	nles: nles$1,
	nless: nless$1,
	nLl: nLl$1,
	nlsim: nlsim$1,
	nLt: nLt$1,
	nlt: nlt$1,
	nltri: nltri$1,
	nltrie: nltrie$1,
	nLtv: nLtv$1,
	nmid: nmid$1,
	NoBreak: NoBreak$1,
	NonBreakingSpace: NonBreakingSpace$1,
	nopf: nopf$1,
	Nopf: Nopf$1,
	Not: Not$1,
	not: not$1$1,
	NotCongruent: NotCongruent$1,
	NotCupCap: NotCupCap$1,
	NotDoubleVerticalBar: NotDoubleVerticalBar$1,
	NotElement: NotElement$1,
	NotEqual: NotEqual$1,
	NotEqualTilde: NotEqualTilde$1,
	NotExists: NotExists$1,
	NotGreater: NotGreater$1,
	NotGreaterEqual: NotGreaterEqual$1,
	NotGreaterFullEqual: NotGreaterFullEqual$1,
	NotGreaterGreater: NotGreaterGreater$1,
	NotGreaterLess: NotGreaterLess$1,
	NotGreaterSlantEqual: NotGreaterSlantEqual$1,
	NotGreaterTilde: NotGreaterTilde$1,
	NotHumpDownHump: NotHumpDownHump$1,
	NotHumpEqual: NotHumpEqual$1,
	notin: notin$1,
	notindot: notindot$1,
	notinE: notinE$1,
	notinva: notinva$1,
	notinvb: notinvb$1,
	notinvc: notinvc$1,
	NotLeftTriangleBar: NotLeftTriangleBar$1,
	NotLeftTriangle: NotLeftTriangle$1,
	NotLeftTriangleEqual: NotLeftTriangleEqual$1,
	NotLess: NotLess$1,
	NotLessEqual: NotLessEqual$1,
	NotLessGreater: NotLessGreater$1,
	NotLessLess: NotLessLess$1,
	NotLessSlantEqual: NotLessSlantEqual$1,
	NotLessTilde: NotLessTilde$1,
	NotNestedGreaterGreater: NotNestedGreaterGreater$1,
	NotNestedLessLess: NotNestedLessLess$1,
	notni: notni$1,
	notniva: notniva$1,
	notnivb: notnivb$1,
	notnivc: notnivc$1,
	NotPrecedes: NotPrecedes$1,
	NotPrecedesEqual: NotPrecedesEqual$1,
	NotPrecedesSlantEqual: NotPrecedesSlantEqual$1,
	NotReverseElement: NotReverseElement$1,
	NotRightTriangleBar: NotRightTriangleBar$1,
	NotRightTriangle: NotRightTriangle$1,
	NotRightTriangleEqual: NotRightTriangleEqual$1,
	NotSquareSubset: NotSquareSubset$1,
	NotSquareSubsetEqual: NotSquareSubsetEqual$1,
	NotSquareSuperset: NotSquareSuperset$1,
	NotSquareSupersetEqual: NotSquareSupersetEqual$1,
	NotSubset: NotSubset$1,
	NotSubsetEqual: NotSubsetEqual$1,
	NotSucceeds: NotSucceeds$1,
	NotSucceedsEqual: NotSucceedsEqual$1,
	NotSucceedsSlantEqual: NotSucceedsSlantEqual$1,
	NotSucceedsTilde: NotSucceedsTilde$1,
	NotSuperset: NotSuperset$1,
	NotSupersetEqual: NotSupersetEqual$1,
	NotTilde: NotTilde$1,
	NotTildeEqual: NotTildeEqual$1,
	NotTildeFullEqual: NotTildeFullEqual$1,
	NotTildeTilde: NotTildeTilde$1,
	NotVerticalBar: NotVerticalBar$1,
	nparallel: nparallel$1,
	npar: npar$1,
	nparsl: nparsl$1,
	npart: npart$1,
	npolint: npolint$1,
	npr: npr$1,
	nprcue: nprcue$1,
	nprec: nprec$1,
	npreceq: npreceq$1,
	npre: npre$1,
	nrarrc: nrarrc$1,
	nrarr: nrarr$1,
	nrArr: nrArr$1,
	nrarrw: nrarrw$1,
	nrightarrow: nrightarrow$1,
	nRightarrow: nRightarrow$1,
	nrtri: nrtri$1,
	nrtrie: nrtrie$1,
	nsc: nsc$1,
	nsccue: nsccue$1,
	nsce: nsce$1,
	Nscr: Nscr$1,
	nscr: nscr$1,
	nshortmid: nshortmid$1,
	nshortparallel: nshortparallel$1,
	nsim: nsim$1,
	nsime: nsime$1,
	nsimeq: nsimeq$1,
	nsmid: nsmid$1,
	nspar: nspar$1,
	nsqsube: nsqsube$1,
	nsqsupe: nsqsupe$1,
	nsub: nsub$1,
	nsubE: nsubE$1,
	nsube: nsube$1,
	nsubset: nsubset$1,
	nsubseteq: nsubseteq$1,
	nsubseteqq: nsubseteqq$1,
	nsucc: nsucc$1,
	nsucceq: nsucceq$1,
	nsup: nsup$1,
	nsupE: nsupE$1,
	nsupe: nsupe$1,
	nsupset: nsupset$1,
	nsupseteq: nsupseteq$1,
	nsupseteqq: nsupseteqq$1,
	ntgl: ntgl$1,
	Ntilde: Ntilde$1$1,
	ntilde: ntilde$1$1,
	ntlg: ntlg$1,
	ntriangleleft: ntriangleleft$1,
	ntrianglelefteq: ntrianglelefteq$1,
	ntriangleright: ntriangleright$1,
	ntrianglerighteq: ntrianglerighteq$1,
	Nu: Nu$1,
	nu: nu$1,
	num: num$1,
	numero: numero$1,
	numsp: numsp$1,
	nvap: nvap$1,
	nvdash: nvdash$1,
	nvDash: nvDash$1,
	nVdash: nVdash$1,
	nVDash: nVDash$1,
	nvge: nvge$1,
	nvgt: nvgt$1,
	nvHarr: nvHarr$1,
	nvinfin: nvinfin$1,
	nvlArr: nvlArr$1,
	nvle: nvle$1,
	nvlt: nvlt$1,
	nvltrie: nvltrie$1,
	nvrArr: nvrArr$1,
	nvrtrie: nvrtrie$1,
	nvsim: nvsim$1,
	nwarhk: nwarhk$1,
	nwarr: nwarr$1,
	nwArr: nwArr$1,
	nwarrow: nwarrow$1,
	nwnear: nwnear$1,
	Oacute: Oacute$1$1,
	oacute: oacute$1$1,
	oast: oast$1,
	Ocirc: Ocirc$1$1,
	ocirc: ocirc$1$1,
	ocir: ocir$1,
	Ocy: Ocy$1,
	ocy: ocy$1,
	odash: odash$1,
	Odblac: Odblac$1,
	odblac: odblac$1,
	odiv: odiv$1,
	odot: odot$1,
	odsold: odsold$1,
	OElig: OElig$1,
	oelig: oelig$1,
	ofcir: ofcir$1,
	Ofr: Ofr$1,
	ofr: ofr$1,
	ogon: ogon$1,
	Ograve: Ograve$1$1,
	ograve: ograve$1$1,
	ogt: ogt$1,
	ohbar: ohbar$1,
	ohm: ohm$1,
	oint: oint$1,
	olarr: olarr$1,
	olcir: olcir$1,
	olcross: olcross$1,
	oline: oline$1,
	olt: olt$1,
	Omacr: Omacr$1,
	omacr: omacr$1,
	Omega: Omega$1,
	omega: omega$1,
	Omicron: Omicron$1,
	omicron: omicron$1,
	omid: omid$1,
	ominus: ominus$1,
	Oopf: Oopf$1,
	oopf: oopf$1,
	opar: opar$1,
	OpenCurlyDoubleQuote: OpenCurlyDoubleQuote$1,
	OpenCurlyQuote: OpenCurlyQuote$1,
	operp: operp$1,
	oplus: oplus$1,
	orarr: orarr$1,
	Or: Or$1,
	or: or$1,
	ord: ord$1,
	order: order$1,
	orderof: orderof$1,
	ordf: ordf$1$1,
	ordm: ordm$1$1,
	origof: origof$1,
	oror: oror$1,
	orslope: orslope$1,
	orv: orv$1,
	oS: oS$1,
	Oscr: Oscr$1,
	oscr: oscr$1,
	Oslash: Oslash$1$1,
	oslash: oslash$1$1,
	osol: osol$1,
	Otilde: Otilde$1$1,
	otilde: otilde$1$1,
	otimesas: otimesas$1,
	Otimes: Otimes$1,
	otimes: otimes$1,
	Ouml: Ouml$1$1,
	ouml: ouml$1$1,
	ovbar: ovbar$1,
	OverBar: OverBar$1,
	OverBrace: OverBrace$1,
	OverBracket: OverBracket$1,
	OverParenthesis: OverParenthesis$1,
	para: para$1$1,
	parallel: parallel$1,
	par: par$1,
	parsim: parsim$1,
	parsl: parsl$1,
	part: part$1,
	PartialD: PartialD$1,
	Pcy: Pcy$1,
	pcy: pcy$1,
	percnt: percnt$1,
	period: period$1,
	permil: permil$1,
	perp: perp$1,
	pertenk: pertenk$1,
	Pfr: Pfr$1,
	pfr: pfr$1,
	Phi: Phi$1,
	phi: phi$1,
	phiv: phiv$1,
	phmmat: phmmat$1,
	phone: phone$1,
	Pi: Pi$1,
	pi: pi$1,
	pitchfork: pitchfork$1,
	piv: piv$1,
	planck: planck$1,
	planckh: planckh$1,
	plankv: plankv$1,
	plusacir: plusacir$1,
	plusb: plusb$1,
	pluscir: pluscir$1,
	plus: plus$1,
	plusdo: plusdo$1,
	plusdu: plusdu$1,
	pluse: pluse$1,
	PlusMinus: PlusMinus$1,
	plusmn: plusmn$1$1,
	plussim: plussim$1,
	plustwo: plustwo$1,
	pm: pm$1,
	Poincareplane: Poincareplane$1,
	pointint: pointint$1,
	popf: popf$1,
	Popf: Popf$1,
	pound: pound$1$1,
	prap: prap$1,
	Pr: Pr$1,
	pr: pr$1,
	prcue: prcue$1,
	precapprox: precapprox$1,
	prec: prec$1,
	preccurlyeq: preccurlyeq$1,
	Precedes: Precedes$1,
	PrecedesEqual: PrecedesEqual$1,
	PrecedesSlantEqual: PrecedesSlantEqual$1,
	PrecedesTilde: PrecedesTilde$1,
	preceq: preceq$1,
	precnapprox: precnapprox$1,
	precneqq: precneqq$1,
	precnsim: precnsim$1,
	pre: pre$1,
	prE: prE$1,
	precsim: precsim$1,
	prime: prime$1,
	Prime: Prime$1,
	primes: primes$1,
	prnap: prnap$1,
	prnE: prnE$1,
	prnsim: prnsim$1,
	prod: prod$1,
	Product: Product$1,
	profalar: profalar$1,
	profline: profline$1,
	profsurf: profsurf$1,
	prop: prop$1,
	Proportional: Proportional$1,
	Proportion: Proportion$1,
	propto: propto$1,
	prsim: prsim$1,
	prurel: prurel$1,
	Pscr: Pscr$1,
	pscr: pscr$1,
	Psi: Psi$1,
	psi: psi$1,
	puncsp: puncsp$1,
	Qfr: Qfr$1,
	qfr: qfr$1,
	qint: qint$1,
	qopf: qopf$1,
	Qopf: Qopf$1,
	qprime: qprime$1,
	Qscr: Qscr$1,
	qscr: qscr$1,
	quaternions: quaternions$1,
	quatint: quatint$1,
	quest: quest$1,
	questeq: questeq$1,
	quot: quot$1$2,
	QUOT: QUOT$1$1,
	rAarr: rAarr$1,
	race: race$1,
	Racute: Racute$1,
	racute: racute$1,
	radic: radic$1,
	raemptyv: raemptyv$1,
	rang: rang$1,
	Rang: Rang$1,
	rangd: rangd$1,
	range: range$1,
	rangle: rangle$1,
	raquo: raquo$1$1,
	rarrap: rarrap$1,
	rarrb: rarrb$1,
	rarrbfs: rarrbfs$1,
	rarrc: rarrc$1,
	rarr: rarr$1,
	Rarr: Rarr$1,
	rArr: rArr$1,
	rarrfs: rarrfs$1,
	rarrhk: rarrhk$1,
	rarrlp: rarrlp$1,
	rarrpl: rarrpl$1,
	rarrsim: rarrsim$1,
	Rarrtl: Rarrtl$1,
	rarrtl: rarrtl$1,
	rarrw: rarrw$1,
	ratail: ratail$1,
	rAtail: rAtail$1,
	ratio: ratio$1,
	rationals: rationals$1,
	rbarr: rbarr$1,
	rBarr: rBarr$1,
	RBarr: RBarr$1,
	rbbrk: rbbrk$1,
	rbrace: rbrace$1,
	rbrack: rbrack$1,
	rbrke: rbrke$1,
	rbrksld: rbrksld$1,
	rbrkslu: rbrkslu$1,
	Rcaron: Rcaron$1,
	rcaron: rcaron$1,
	Rcedil: Rcedil$1,
	rcedil: rcedil$1,
	rceil: rceil$1,
	rcub: rcub$1,
	Rcy: Rcy$1,
	rcy: rcy$1,
	rdca: rdca$1,
	rdldhar: rdldhar$1,
	rdquo: rdquo$1,
	rdquor: rdquor$1,
	rdsh: rdsh$1,
	real: real$1,
	realine: realine$1,
	realpart: realpart$1,
	reals: reals$1,
	Re: Re$1,
	rect: rect$1,
	reg: reg$1$1,
	REG: REG$1$1,
	ReverseElement: ReverseElement$1,
	ReverseEquilibrium: ReverseEquilibrium$1,
	ReverseUpEquilibrium: ReverseUpEquilibrium$1,
	rfisht: rfisht$1,
	rfloor: rfloor$1,
	rfr: rfr$1,
	Rfr: Rfr$1,
	rHar: rHar$1,
	rhard: rhard$1,
	rharu: rharu$1,
	rharul: rharul$1,
	Rho: Rho$1,
	rho: rho$1,
	rhov: rhov$1,
	RightAngleBracket: RightAngleBracket$1,
	RightArrowBar: RightArrowBar$1,
	rightarrow: rightarrow$1,
	RightArrow: RightArrow$1,
	Rightarrow: Rightarrow$1,
	RightArrowLeftArrow: RightArrowLeftArrow$1,
	rightarrowtail: rightarrowtail$1,
	RightCeiling: RightCeiling$1,
	RightDoubleBracket: RightDoubleBracket$1,
	RightDownTeeVector: RightDownTeeVector$1,
	RightDownVectorBar: RightDownVectorBar$1,
	RightDownVector: RightDownVector$1,
	RightFloor: RightFloor$1,
	rightharpoondown: rightharpoondown$1,
	rightharpoonup: rightharpoonup$1,
	rightleftarrows: rightleftarrows$1,
	rightleftharpoons: rightleftharpoons$1,
	rightrightarrows: rightrightarrows$1,
	rightsquigarrow: rightsquigarrow$1,
	RightTeeArrow: RightTeeArrow$1,
	RightTee: RightTee$1,
	RightTeeVector: RightTeeVector$1,
	rightthreetimes: rightthreetimes$1,
	RightTriangleBar: RightTriangleBar$1,
	RightTriangle: RightTriangle$1,
	RightTriangleEqual: RightTriangleEqual$1,
	RightUpDownVector: RightUpDownVector$1,
	RightUpTeeVector: RightUpTeeVector$1,
	RightUpVectorBar: RightUpVectorBar$1,
	RightUpVector: RightUpVector$1,
	RightVectorBar: RightVectorBar$1,
	RightVector: RightVector$1,
	ring: ring$1,
	risingdotseq: risingdotseq$1,
	rlarr: rlarr$1,
	rlhar: rlhar$1,
	rlm: rlm$1,
	rmoustache: rmoustache$1,
	rmoust: rmoust$1,
	rnmid: rnmid$1,
	roang: roang$1,
	roarr: roarr$1,
	robrk: robrk$1,
	ropar: ropar$1,
	ropf: ropf$1,
	Ropf: Ropf$1,
	roplus: roplus$1,
	rotimes: rotimes$1,
	RoundImplies: RoundImplies$1,
	rpar: rpar$1,
	rpargt: rpargt$1,
	rppolint: rppolint$1,
	rrarr: rrarr$1,
	Rrightarrow: Rrightarrow$1,
	rsaquo: rsaquo$1,
	rscr: rscr$1,
	Rscr: Rscr$1,
	rsh: rsh$1,
	Rsh: Rsh$1,
	rsqb: rsqb$1,
	rsquo: rsquo$1,
	rsquor: rsquor$1,
	rthree: rthree$1,
	rtimes: rtimes$1,
	rtri: rtri$1,
	rtrie: rtrie$1,
	rtrif: rtrif$1,
	rtriltri: rtriltri$1,
	RuleDelayed: RuleDelayed$1,
	ruluhar: ruluhar$1,
	rx: rx$1,
	Sacute: Sacute$1,
	sacute: sacute$1,
	sbquo: sbquo$1,
	scap: scap$1,
	Scaron: Scaron$1,
	scaron: scaron$1,
	Sc: Sc$1,
	sc: sc$1,
	sccue: sccue$1,
	sce: sce$1,
	scE: scE$1,
	Scedil: Scedil$1,
	scedil: scedil$1,
	Scirc: Scirc$1,
	scirc: scirc$1,
	scnap: scnap$1,
	scnE: scnE$1,
	scnsim: scnsim$1,
	scpolint: scpolint$1,
	scsim: scsim$1,
	Scy: Scy$1,
	scy: scy$1,
	sdotb: sdotb$1,
	sdot: sdot$1,
	sdote: sdote$1,
	searhk: searhk$1,
	searr: searr$1,
	seArr: seArr$1,
	searrow: searrow$1,
	sect: sect$1$1,
	semi: semi$1,
	seswar: seswar$1,
	setminus: setminus$1,
	setmn: setmn$1,
	sext: sext$1,
	Sfr: Sfr$1,
	sfr: sfr$1,
	sfrown: sfrown$1,
	sharp: sharp$1,
	SHCHcy: SHCHcy$1,
	shchcy: shchcy$1,
	SHcy: SHcy$1,
	shcy: shcy$1,
	ShortDownArrow: ShortDownArrow$1,
	ShortLeftArrow: ShortLeftArrow$1,
	shortmid: shortmid$1,
	shortparallel: shortparallel$1,
	ShortRightArrow: ShortRightArrow$1,
	ShortUpArrow: ShortUpArrow$1,
	shy: shy$1$1,
	Sigma: Sigma$1,
	sigma: sigma$1,
	sigmaf: sigmaf$1,
	sigmav: sigmav$1,
	sim: sim$1,
	simdot: simdot$1,
	sime: sime$1,
	simeq: simeq$1,
	simg: simg$1,
	simgE: simgE$1,
	siml: siml$1,
	simlE: simlE$1,
	simne: simne$1,
	simplus: simplus$1,
	simrarr: simrarr$1,
	slarr: slarr$1,
	SmallCircle: SmallCircle$1,
	smallsetminus: smallsetminus$1,
	smashp: smashp$1,
	smeparsl: smeparsl$1,
	smid: smid$1,
	smile: smile$1,
	smt: smt$1,
	smte: smte$1,
	smtes: smtes$1,
	SOFTcy: SOFTcy$1,
	softcy: softcy$1,
	solbar: solbar$1,
	solb: solb$1,
	sol: sol$1,
	Sopf: Sopf$1,
	sopf: sopf$1,
	spades: spades$1,
	spadesuit: spadesuit$1,
	spar: spar$1,
	sqcap: sqcap$1,
	sqcaps: sqcaps$1,
	sqcup: sqcup$1,
	sqcups: sqcups$1,
	Sqrt: Sqrt$1,
	sqsub: sqsub$1,
	sqsube: sqsube$1,
	sqsubset: sqsubset$1,
	sqsubseteq: sqsubseteq$1,
	sqsup: sqsup$1,
	sqsupe: sqsupe$1,
	sqsupset: sqsupset$1,
	sqsupseteq: sqsupseteq$1,
	square: square$1,
	Square: Square$1,
	SquareIntersection: SquareIntersection$1,
	SquareSubset: SquareSubset$1,
	SquareSubsetEqual: SquareSubsetEqual$1,
	SquareSuperset: SquareSuperset$1,
	SquareSupersetEqual: SquareSupersetEqual$1,
	SquareUnion: SquareUnion$1,
	squarf: squarf$1,
	squ: squ$1,
	squf: squf$1,
	srarr: srarr$1,
	Sscr: Sscr$1,
	sscr: sscr$1,
	ssetmn: ssetmn$1,
	ssmile: ssmile$1,
	sstarf: sstarf$1,
	Star: Star$1,
	star: star$1,
	starf: starf$1,
	straightepsilon: straightepsilon$1,
	straightphi: straightphi$1,
	strns: strns$1,
	sub: sub$1,
	Sub: Sub$1,
	subdot: subdot$1,
	subE: subE$1,
	sube: sube$1,
	subedot: subedot$1,
	submult: submult$1,
	subnE: subnE$1,
	subne: subne$1,
	subplus: subplus$1,
	subrarr: subrarr$1,
	subset: subset$1,
	Subset: Subset$1,
	subseteq: subseteq$1,
	subseteqq: subseteqq$1,
	SubsetEqual: SubsetEqual$1,
	subsetneq: subsetneq$1,
	subsetneqq: subsetneqq$1,
	subsim: subsim$1,
	subsub: subsub$1,
	subsup: subsup$1,
	succapprox: succapprox$1,
	succ: succ$1,
	succcurlyeq: succcurlyeq$1,
	Succeeds: Succeeds$1,
	SucceedsEqual: SucceedsEqual$1,
	SucceedsSlantEqual: SucceedsSlantEqual$1,
	SucceedsTilde: SucceedsTilde$1,
	succeq: succeq$1,
	succnapprox: succnapprox$1,
	succneqq: succneqq$1,
	succnsim: succnsim$1,
	succsim: succsim$1,
	SuchThat: SuchThat$1,
	sum: sum$1,
	Sum: Sum$1,
	sung: sung$1,
	sup1: sup1$1$1,
	sup2: sup2$1$1,
	sup3: sup3$1$1,
	sup: sup$1,
	Sup: Sup$1,
	supdot: supdot$1,
	supdsub: supdsub$1,
	supE: supE$1,
	supe: supe$1,
	supedot: supedot$1,
	Superset: Superset$1,
	SupersetEqual: SupersetEqual$1,
	suphsol: suphsol$1,
	suphsub: suphsub$1,
	suplarr: suplarr$1,
	supmult: supmult$1,
	supnE: supnE$1,
	supne: supne$1,
	supplus: supplus$1,
	supset: supset$1,
	Supset: Supset$1,
	supseteq: supseteq$1,
	supseteqq: supseteqq$1,
	supsetneq: supsetneq$1,
	supsetneqq: supsetneqq$1,
	supsim: supsim$1,
	supsub: supsub$1,
	supsup: supsup$1,
	swarhk: swarhk$1,
	swarr: swarr$1,
	swArr: swArr$1,
	swarrow: swarrow$1,
	swnwar: swnwar$1,
	szlig: szlig$1$1,
	Tab: Tab$1,
	target: target$1,
	Tau: Tau$1,
	tau: tau$1,
	tbrk: tbrk$1,
	Tcaron: Tcaron$1,
	tcaron: tcaron$1,
	Tcedil: Tcedil$1,
	tcedil: tcedil$1,
	Tcy: Tcy$1,
	tcy: tcy$1,
	tdot: tdot$1,
	telrec: telrec$1,
	Tfr: Tfr$1,
	tfr: tfr$1,
	there4: there4$1,
	therefore: therefore$1,
	Therefore: Therefore$1,
	Theta: Theta$1,
	theta: theta$1,
	thetasym: thetasym$1,
	thetav: thetav$1,
	thickapprox: thickapprox$1,
	thicksim: thicksim$1,
	ThickSpace: ThickSpace$1,
	ThinSpace: ThinSpace$1,
	thinsp: thinsp$1,
	thkap: thkap$1,
	thksim: thksim$1,
	THORN: THORN$1$1,
	thorn: thorn$1$1,
	tilde: tilde$1,
	Tilde: Tilde$1,
	TildeEqual: TildeEqual$1,
	TildeFullEqual: TildeFullEqual$1,
	TildeTilde: TildeTilde$1,
	timesbar: timesbar$1,
	timesb: timesb$1,
	times: times$2,
	timesd: timesd$1,
	tint: tint$1,
	toea: toea$1,
	topbot: topbot$1,
	topcir: topcir$1,
	top: top$1,
	Topf: Topf$1,
	topf: topf$1,
	topfork: topfork$1,
	tosa: tosa$1,
	tprime: tprime$1,
	trade: trade$1,
	TRADE: TRADE$1,
	triangle: triangle$1,
	triangledown: triangledown$1,
	triangleleft: triangleleft$1,
	trianglelefteq: trianglelefteq$1,
	triangleq: triangleq$1,
	triangleright: triangleright$1,
	trianglerighteq: trianglerighteq$1,
	tridot: tridot$1,
	trie: trie$1,
	triminus: triminus$1,
	TripleDot: TripleDot$1,
	triplus: triplus$1,
	trisb: trisb$1,
	tritime: tritime$1,
	trpezium: trpezium$1,
	Tscr: Tscr$1,
	tscr: tscr$1,
	TScy: TScy$1,
	tscy: tscy$1,
	TSHcy: TSHcy$1,
	tshcy: tshcy$1,
	Tstrok: Tstrok$1,
	tstrok: tstrok$1,
	twixt: twixt$1,
	twoheadleftarrow: twoheadleftarrow$1,
	twoheadrightarrow: twoheadrightarrow$1,
	Uacute: Uacute$1$1,
	uacute: uacute$1$1,
	uarr: uarr$1,
	Uarr: Uarr$1,
	uArr: uArr$1,
	Uarrocir: Uarrocir$1,
	Ubrcy: Ubrcy$1,
	ubrcy: ubrcy$1,
	Ubreve: Ubreve$1,
	ubreve: ubreve$1,
	Ucirc: Ucirc$1$1,
	ucirc: ucirc$1$1,
	Ucy: Ucy$1,
	ucy: ucy$1,
	udarr: udarr$1,
	Udblac: Udblac$1,
	udblac: udblac$1,
	udhar: udhar$1,
	ufisht: ufisht$1,
	Ufr: Ufr$1,
	ufr: ufr$1,
	Ugrave: Ugrave$1$1,
	ugrave: ugrave$1$1,
	uHar: uHar$1,
	uharl: uharl$1,
	uharr: uharr$1,
	uhblk: uhblk$1,
	ulcorn: ulcorn$1,
	ulcorner: ulcorner$1,
	ulcrop: ulcrop$1,
	ultri: ultri$1,
	Umacr: Umacr$1,
	umacr: umacr$1,
	uml: uml$1$1,
	UnderBar: UnderBar$1,
	UnderBrace: UnderBrace$1,
	UnderBracket: UnderBracket$1,
	UnderParenthesis: UnderParenthesis$1,
	Union: Union$1,
	UnionPlus: UnionPlus$1,
	Uogon: Uogon$1,
	uogon: uogon$1,
	Uopf: Uopf$1,
	uopf: uopf$1,
	UpArrowBar: UpArrowBar$1,
	uparrow: uparrow$1,
	UpArrow: UpArrow$1,
	Uparrow: Uparrow$1,
	UpArrowDownArrow: UpArrowDownArrow$1,
	updownarrow: updownarrow$1,
	UpDownArrow: UpDownArrow$1,
	Updownarrow: Updownarrow$1,
	UpEquilibrium: UpEquilibrium$1,
	upharpoonleft: upharpoonleft$1,
	upharpoonright: upharpoonright$1,
	uplus: uplus$1,
	UpperLeftArrow: UpperLeftArrow$1,
	UpperRightArrow: UpperRightArrow$1,
	upsi: upsi$1,
	Upsi: Upsi$1,
	upsih: upsih$1,
	Upsilon: Upsilon$1,
	upsilon: upsilon$1,
	UpTeeArrow: UpTeeArrow$1,
	UpTee: UpTee$1,
	upuparrows: upuparrows$1,
	urcorn: urcorn$1,
	urcorner: urcorner$1,
	urcrop: urcrop$1,
	Uring: Uring$1,
	uring: uring$1,
	urtri: urtri$1,
	Uscr: Uscr$1,
	uscr: uscr$1,
	utdot: utdot$1,
	Utilde: Utilde$1,
	utilde: utilde$1,
	utri: utri$1,
	utrif: utrif$1,
	uuarr: uuarr$1,
	Uuml: Uuml$1$1,
	uuml: uuml$1$1,
	uwangle: uwangle$1,
	vangrt: vangrt$1,
	varepsilon: varepsilon$1,
	varkappa: varkappa$1,
	varnothing: varnothing$1,
	varphi: varphi$1,
	varpi: varpi$1,
	varpropto: varpropto$1,
	varr: varr$1,
	vArr: vArr$1,
	varrho: varrho$1,
	varsigma: varsigma$1,
	varsubsetneq: varsubsetneq$1,
	varsubsetneqq: varsubsetneqq$1,
	varsupsetneq: varsupsetneq$1,
	varsupsetneqq: varsupsetneqq$1,
	vartheta: vartheta$1,
	vartriangleleft: vartriangleleft$1,
	vartriangleright: vartriangleright$1,
	vBar: vBar$1,
	Vbar: Vbar$1,
	vBarv: vBarv$1,
	Vcy: Vcy$1,
	vcy: vcy$1,
	vdash: vdash$1,
	vDash: vDash$1,
	Vdash: Vdash$1,
	VDash: VDash$1,
	Vdashl: Vdashl$1,
	veebar: veebar$1,
	vee: vee$1,
	Vee: Vee$1,
	veeeq: veeeq$1,
	vellip: vellip$1,
	verbar: verbar$1,
	Verbar: Verbar$1,
	vert: vert$1,
	Vert: Vert$1,
	VerticalBar: VerticalBar$1,
	VerticalLine: VerticalLine$1,
	VerticalSeparator: VerticalSeparator$1,
	VerticalTilde: VerticalTilde$1,
	VeryThinSpace: VeryThinSpace$1,
	Vfr: Vfr$1,
	vfr: vfr$1,
	vltri: vltri$1,
	vnsub: vnsub$1,
	vnsup: vnsup$1,
	Vopf: Vopf$1,
	vopf: vopf$1,
	vprop: vprop$1,
	vrtri: vrtri$1,
	Vscr: Vscr$1,
	vscr: vscr$1,
	vsubnE: vsubnE$1,
	vsubne: vsubne$1,
	vsupnE: vsupnE$1,
	vsupne: vsupne$1,
	Vvdash: Vvdash$1,
	vzigzag: vzigzag$1,
	Wcirc: Wcirc$1,
	wcirc: wcirc$1,
	wedbar: wedbar$1,
	wedge: wedge$1,
	Wedge: Wedge$1,
	wedgeq: wedgeq$1,
	weierp: weierp$1,
	Wfr: Wfr$1,
	wfr: wfr$1,
	Wopf: Wopf$1,
	wopf: wopf$1,
	wp: wp$1,
	wr: wr$1,
	wreath: wreath$1,
	Wscr: Wscr$1,
	wscr: wscr$1,
	xcap: xcap$1,
	xcirc: xcirc$1,
	xcup: xcup$1,
	xdtri: xdtri$1,
	Xfr: Xfr$1,
	xfr: xfr$1,
	xharr: xharr$1,
	xhArr: xhArr$1,
	Xi: Xi$1,
	xi: xi$1,
	xlarr: xlarr$1,
	xlArr: xlArr$1,
	xmap: xmap$1,
	xnis: xnis$1,
	xodot: xodot$1,
	Xopf: Xopf$1,
	xopf: xopf$1,
	xoplus: xoplus$1,
	xotime: xotime$1,
	xrarr: xrarr$1,
	xrArr: xrArr$1,
	Xscr: Xscr$1,
	xscr: xscr$1,
	xsqcup: xsqcup$1,
	xuplus: xuplus$1,
	xutri: xutri$1,
	xvee: xvee$1,
	xwedge: xwedge$1,
	Yacute: Yacute$1$1,
	yacute: yacute$1$1,
	YAcy: YAcy$1,
	yacy: yacy$1,
	Ycirc: Ycirc$1,
	ycirc: ycirc$1,
	Ycy: Ycy$1,
	ycy: ycy$1,
	yen: yen$1$1,
	Yfr: Yfr$1,
	yfr: yfr$1,
	YIcy: YIcy$1,
	yicy: yicy$1,
	Yopf: Yopf$1,
	yopf: yopf$1,
	Yscr: Yscr$1,
	yscr: yscr$1,
	YUcy: YUcy$1,
	yucy: yucy$1,
	yuml: yuml$1$1,
	Yuml: Yuml$1,
	Zacute: Zacute$1,
	zacute: zacute$1,
	Zcaron: Zcaron$1,
	zcaron: zcaron$1,
	Zcy: Zcy$1,
	zcy: zcy$1,
	Zdot: Zdot$1,
	zdot: zdot$1,
	zeetrf: zeetrf$1,
	ZeroWidthSpace: ZeroWidthSpace$1,
	Zeta: Zeta$1,
	zeta: zeta$1,
	zfr: zfr$1,
	Zfr: Zfr$1,
	ZHcy: ZHcy$1,
	zhcy: zhcy$1,
	zigrarr: zigrarr$1,
	zopf: zopf$1,
	Zopf: Zopf$1,
	Zscr: Zscr$1,
	zscr: zscr$1,
	zwj: zwj$1,
	zwnj: zwnj$1,
	"in": "∈",
	"Map": "⤅"
};

var entities$2 = Object.freeze({
	Aacute: Aacute$1$1,
	aacute: aacute$1$1,
	Abreve: Abreve$1,
	abreve: abreve$1,
	ac: ac$1,
	acd: acd$1,
	acE: acE$1,
	Acirc: Acirc$1$1,
	acirc: acirc$1$1,
	acute: acute$1$1,
	Acy: Acy$1,
	acy: acy$1,
	AElig: AElig$1$1,
	aelig: aelig$1$1,
	af: af$1,
	Afr: Afr$1,
	afr: afr$1,
	Agrave: Agrave$1$1,
	agrave: agrave$1$1,
	alefsym: alefsym$1,
	aleph: aleph$1,
	Alpha: Alpha$1,
	alpha: alpha$1,
	Amacr: Amacr$1,
	amacr: amacr$1,
	amalg: amalg$1,
	amp: amp$1$2,
	AMP: AMP$1$1,
	andand: andand$1,
	And: And$1,
	and: and$1,
	andd: andd$1,
	andslope: andslope$1,
	andv: andv$1,
	ang: ang$1,
	ange: ange$1,
	angle: angle$1,
	angmsdaa: angmsdaa$1,
	angmsdab: angmsdab$1,
	angmsdac: angmsdac$1,
	angmsdad: angmsdad$1,
	angmsdae: angmsdae$1,
	angmsdaf: angmsdaf$1,
	angmsdag: angmsdag$1,
	angmsdah: angmsdah$1,
	angmsd: angmsd$1,
	angrt: angrt$1,
	angrtvb: angrtvb$1,
	angrtvbd: angrtvbd$1,
	angsph: angsph$1,
	angst: angst$1,
	angzarr: angzarr$1,
	Aogon: Aogon$1,
	aogon: aogon$1,
	Aopf: Aopf$1,
	aopf: aopf$1,
	apacir: apacir$1,
	ap: ap$1,
	apE: apE$1,
	ape: ape$1,
	apid: apid$1,
	apos: apos$1$2,
	ApplyFunction: ApplyFunction$1,
	approx: approx$1,
	approxeq: approxeq$1,
	Aring: Aring$1$1,
	aring: aring$1$1,
	Ascr: Ascr$1,
	ascr: ascr$1,
	Assign: Assign$1,
	ast: ast$1,
	asymp: asymp$1,
	asympeq: asympeq$1,
	Atilde: Atilde$1$1,
	atilde: atilde$1$1,
	Auml: Auml$1$1,
	auml: auml$1$1,
	awconint: awconint$1,
	awint: awint$1,
	backcong: backcong$1,
	backepsilon: backepsilon$1,
	backprime: backprime$1,
	backsim: backsim$1,
	backsimeq: backsimeq$1,
	Backslash: Backslash$1,
	Barv: Barv$1,
	barvee: barvee$1,
	barwed: barwed$1,
	Barwed: Barwed$1,
	barwedge: barwedge$1,
	bbrk: bbrk$1,
	bbrktbrk: bbrktbrk$1,
	bcong: bcong$1,
	Bcy: Bcy$1,
	bcy: bcy$1,
	bdquo: bdquo$1,
	becaus: becaus$1,
	because: because$1,
	Because: Because$1,
	bemptyv: bemptyv$1,
	bepsi: bepsi$1,
	bernou: bernou$1,
	Bernoullis: Bernoullis$1,
	Beta: Beta$1,
	beta: beta$1,
	beth: beth$1,
	between: between$1,
	Bfr: Bfr$1,
	bfr: bfr$1,
	bigcap: bigcap$1,
	bigcirc: bigcirc$1,
	bigcup: bigcup$1,
	bigodot: bigodot$1,
	bigoplus: bigoplus$1,
	bigotimes: bigotimes$1,
	bigsqcup: bigsqcup$1,
	bigstar: bigstar$1,
	bigtriangledown: bigtriangledown$1,
	bigtriangleup: bigtriangleup$1,
	biguplus: biguplus$1,
	bigvee: bigvee$1,
	bigwedge: bigwedge$1,
	bkarow: bkarow$1,
	blacklozenge: blacklozenge$1,
	blacksquare: blacksquare$1,
	blacktriangle: blacktriangle$1,
	blacktriangledown: blacktriangledown$1,
	blacktriangleleft: blacktriangleleft$1,
	blacktriangleright: blacktriangleright$1,
	blank: blank$1,
	blk12: blk12$1,
	blk14: blk14$1,
	blk34: blk34$1,
	block: block$1,
	bne: bne$1,
	bnequiv: bnequiv$1,
	bNot: bNot$1,
	bnot: bnot$1,
	Bopf: Bopf$1,
	bopf: bopf$1,
	bot: bot$1,
	bottom: bottom$1,
	bowtie: bowtie$1,
	boxbox: boxbox$1,
	boxdl: boxdl$1,
	boxdL: boxdL$1,
	boxDl: boxDl$1,
	boxDL: boxDL$1,
	boxdr: boxdr$1,
	boxdR: boxdR$1,
	boxDr: boxDr$1,
	boxDR: boxDR$1,
	boxh: boxh$1,
	boxH: boxH$1,
	boxhd: boxhd$1,
	boxHd: boxHd$1,
	boxhD: boxhD$1,
	boxHD: boxHD$1,
	boxhu: boxhu$1,
	boxHu: boxHu$1,
	boxhU: boxhU$1,
	boxHU: boxHU$1,
	boxminus: boxminus$1,
	boxplus: boxplus$1,
	boxtimes: boxtimes$1,
	boxul: boxul$1,
	boxuL: boxuL$1,
	boxUl: boxUl$1,
	boxUL: boxUL$1,
	boxur: boxur$1,
	boxuR: boxuR$1,
	boxUr: boxUr$1,
	boxUR: boxUR$1,
	boxv: boxv$1,
	boxV: boxV$1,
	boxvh: boxvh$1,
	boxvH: boxvH$1,
	boxVh: boxVh$1,
	boxVH: boxVH$1,
	boxvl: boxvl$1,
	boxvL: boxvL$1,
	boxVl: boxVl$1,
	boxVL: boxVL$1,
	boxvr: boxvr$1,
	boxvR: boxvR$1,
	boxVr: boxVr$1,
	boxVR: boxVR$1,
	bprime: bprime$1,
	breve: breve$1,
	Breve: Breve$1,
	brvbar: brvbar$1$1,
	bscr: bscr$1,
	Bscr: Bscr$1,
	bsemi: bsemi$1,
	bsim: bsim$1,
	bsime: bsime$1,
	bsolb: bsolb$1,
	bsol: bsol$1,
	bsolhsub: bsolhsub$1,
	bull: bull$1,
	bullet: bullet$1,
	bump: bump$1,
	bumpE: bumpE$1,
	bumpe: bumpe$1,
	Bumpeq: Bumpeq$1,
	bumpeq: bumpeq$1,
	Cacute: Cacute$1,
	cacute: cacute$1,
	capand: capand$1,
	capbrcup: capbrcup$1,
	capcap: capcap$1,
	cap: cap$1,
	Cap: Cap$1,
	capcup: capcup$1,
	capdot: capdot$1,
	CapitalDifferentialD: CapitalDifferentialD$1,
	caps: caps$1,
	caret: caret$1,
	caron: caron$1,
	Cayleys: Cayleys$1,
	ccaps: ccaps$1,
	Ccaron: Ccaron$1,
	ccaron: ccaron$1,
	Ccedil: Ccedil$1$1,
	ccedil: ccedil$1$1,
	Ccirc: Ccirc$1,
	ccirc: ccirc$1,
	Cconint: Cconint$1,
	ccups: ccups$1,
	ccupssm: ccupssm$1,
	Cdot: Cdot$1,
	cdot: cdot$1,
	cedil: cedil$1$1,
	Cedilla: Cedilla$1,
	cemptyv: cemptyv$1,
	cent: cent$1$1,
	centerdot: centerdot$1,
	CenterDot: CenterDot$1,
	cfr: cfr$1,
	Cfr: Cfr$1,
	CHcy: CHcy$1,
	chcy: chcy$1,
	check: check$1,
	checkmark: checkmark$1,
	Chi: Chi$1,
	chi: chi$1,
	circ: circ$1,
	circeq: circeq$1,
	circlearrowleft: circlearrowleft$1,
	circlearrowright: circlearrowright$1,
	circledast: circledast$1,
	circledcirc: circledcirc$1,
	circleddash: circleddash$1,
	CircleDot: CircleDot$1,
	circledR: circledR$1,
	circledS: circledS$1,
	CircleMinus: CircleMinus$1,
	CirclePlus: CirclePlus$1,
	CircleTimes: CircleTimes$1,
	cir: cir$1,
	cirE: cirE$1,
	cire: cire$1,
	cirfnint: cirfnint$1,
	cirmid: cirmid$1,
	cirscir: cirscir$1,
	ClockwiseContourIntegral: ClockwiseContourIntegral$1,
	CloseCurlyDoubleQuote: CloseCurlyDoubleQuote$1,
	CloseCurlyQuote: CloseCurlyQuote$1,
	clubs: clubs$1,
	clubsuit: clubsuit$1,
	colon: colon$1,
	Colon: Colon$1,
	Colone: Colone$1,
	colone: colone$1,
	coloneq: coloneq$1,
	comma: comma$1,
	commat: commat$1,
	comp: comp$1,
	compfn: compfn$1,
	complement: complement$1,
	complexes: complexes$1,
	cong: cong$1,
	congdot: congdot$1,
	Congruent: Congruent$1,
	conint: conint$1,
	Conint: Conint$1,
	ContourIntegral: ContourIntegral$1,
	copf: copf$1,
	Copf: Copf$1,
	coprod: coprod$1,
	Coproduct: Coproduct$1,
	copy: copy$1$1,
	COPY: COPY$1$1,
	copysr: copysr$1,
	CounterClockwiseContourIntegral: CounterClockwiseContourIntegral$1,
	crarr: crarr$1,
	cross: cross$1,
	Cross: Cross$1,
	Cscr: Cscr$1,
	cscr: cscr$1,
	csub: csub$1,
	csube: csube$1,
	csup: csup$1,
	csupe: csupe$1,
	ctdot: ctdot$1,
	cudarrl: cudarrl$1,
	cudarrr: cudarrr$1,
	cuepr: cuepr$1,
	cuesc: cuesc$1,
	cularr: cularr$1,
	cularrp: cularrp$1,
	cupbrcap: cupbrcap$1,
	cupcap: cupcap$1,
	CupCap: CupCap$1,
	cup: cup$1,
	Cup: Cup$1,
	cupcup: cupcup$1,
	cupdot: cupdot$1,
	cupor: cupor$1,
	cups: cups$1,
	curarr: curarr$1,
	curarrm: curarrm$1,
	curlyeqprec: curlyeqprec$1,
	curlyeqsucc: curlyeqsucc$1,
	curlyvee: curlyvee$1,
	curlywedge: curlywedge$1,
	curren: curren$1$1,
	curvearrowleft: curvearrowleft$1,
	curvearrowright: curvearrowright$1,
	cuvee: cuvee$1,
	cuwed: cuwed$1,
	cwconint: cwconint$1,
	cwint: cwint$1,
	cylcty: cylcty$1,
	dagger: dagger$1,
	Dagger: Dagger$1,
	daleth: daleth$1,
	darr: darr$1,
	Darr: Darr$1,
	dArr: dArr$1,
	dash: dash$1,
	Dashv: Dashv$1,
	dashv: dashv$1,
	dbkarow: dbkarow$1,
	dblac: dblac$1,
	Dcaron: Dcaron$1,
	dcaron: dcaron$1,
	Dcy: Dcy$1,
	dcy: dcy$1,
	ddagger: ddagger$1,
	ddarr: ddarr$1,
	DD: DD$1,
	dd: dd$1,
	DDotrahd: DDotrahd$1,
	ddotseq: ddotseq$1,
	deg: deg$1$1,
	Del: Del$1,
	Delta: Delta$1,
	delta: delta$1,
	demptyv: demptyv$1,
	dfisht: dfisht$1,
	Dfr: Dfr$1,
	dfr: dfr$1,
	dHar: dHar$1,
	dharl: dharl$1,
	dharr: dharr$1,
	DiacriticalAcute: DiacriticalAcute$1,
	DiacriticalDot: DiacriticalDot$1,
	DiacriticalDoubleAcute: DiacriticalDoubleAcute$1,
	DiacriticalGrave: DiacriticalGrave$1,
	DiacriticalTilde: DiacriticalTilde$1,
	diam: diam$1,
	diamond: diamond$1,
	Diamond: Diamond$1,
	diamondsuit: diamondsuit$1,
	diams: diams$1,
	die: die$1,
	DifferentialD: DifferentialD$1,
	digamma: digamma$1,
	disin: disin$1,
	div: div$1,
	divide: divide$1$1,
	divideontimes: divideontimes$1,
	divonx: divonx$1,
	DJcy: DJcy$1,
	djcy: djcy$1,
	dlcorn: dlcorn$1,
	dlcrop: dlcrop$1,
	dollar: dollar$1,
	Dopf: Dopf$1,
	dopf: dopf$1,
	Dot: Dot$1,
	dot: dot$1,
	DotDot: DotDot$1,
	doteq: doteq$1,
	doteqdot: doteqdot$1,
	DotEqual: DotEqual$1,
	dotminus: dotminus$1,
	dotplus: dotplus$1,
	dotsquare: dotsquare$1,
	doublebarwedge: doublebarwedge$1,
	DoubleContourIntegral: DoubleContourIntegral$1,
	DoubleDot: DoubleDot$1,
	DoubleDownArrow: DoubleDownArrow$1,
	DoubleLeftArrow: DoubleLeftArrow$1,
	DoubleLeftRightArrow: DoubleLeftRightArrow$1,
	DoubleLeftTee: DoubleLeftTee$1,
	DoubleLongLeftArrow: DoubleLongLeftArrow$1,
	DoubleLongLeftRightArrow: DoubleLongLeftRightArrow$1,
	DoubleLongRightArrow: DoubleLongRightArrow$1,
	DoubleRightArrow: DoubleRightArrow$1,
	DoubleRightTee: DoubleRightTee$1,
	DoubleUpArrow: DoubleUpArrow$1,
	DoubleUpDownArrow: DoubleUpDownArrow$1,
	DoubleVerticalBar: DoubleVerticalBar$1,
	DownArrowBar: DownArrowBar$1,
	downarrow: downarrow$1,
	DownArrow: DownArrow$1,
	Downarrow: Downarrow$1,
	DownArrowUpArrow: DownArrowUpArrow$1,
	DownBreve: DownBreve$1,
	downdownarrows: downdownarrows$1,
	downharpoonleft: downharpoonleft$1,
	downharpoonright: downharpoonright$1,
	DownLeftRightVector: DownLeftRightVector$1,
	DownLeftTeeVector: DownLeftTeeVector$1,
	DownLeftVectorBar: DownLeftVectorBar$1,
	DownLeftVector: DownLeftVector$1,
	DownRightTeeVector: DownRightTeeVector$1,
	DownRightVectorBar: DownRightVectorBar$1,
	DownRightVector: DownRightVector$1,
	DownTeeArrow: DownTeeArrow$1,
	DownTee: DownTee$1,
	drbkarow: drbkarow$1,
	drcorn: drcorn$1,
	drcrop: drcrop$1,
	Dscr: Dscr$1,
	dscr: dscr$1,
	DScy: DScy$1,
	dscy: dscy$1,
	dsol: dsol$1,
	Dstrok: Dstrok$1,
	dstrok: dstrok$1,
	dtdot: dtdot$1,
	dtri: dtri$1,
	dtrif: dtrif$1,
	duarr: duarr$1,
	duhar: duhar$1,
	dwangle: dwangle$1,
	DZcy: DZcy$1,
	dzcy: dzcy$1,
	dzigrarr: dzigrarr$1,
	Eacute: Eacute$1$1,
	eacute: eacute$1$1,
	easter: easter$1,
	Ecaron: Ecaron$1,
	ecaron: ecaron$1,
	Ecirc: Ecirc$1$1,
	ecirc: ecirc$1$1,
	ecir: ecir$1,
	ecolon: ecolon$1,
	Ecy: Ecy$1,
	ecy: ecy$1,
	eDDot: eDDot$1,
	Edot: Edot$1,
	edot: edot$1,
	eDot: eDot$1,
	ee: ee$1,
	efDot: efDot$1,
	Efr: Efr$1,
	efr: efr$1,
	eg: eg$1,
	Egrave: Egrave$1$1,
	egrave: egrave$1$1,
	egs: egs$1,
	egsdot: egsdot$1,
	el: el$1,
	Element: Element$1,
	elinters: elinters$1,
	ell: ell$1,
	els: els$1,
	elsdot: elsdot$1,
	Emacr: Emacr$1,
	emacr: emacr$1,
	empty: empty$1,
	emptyset: emptyset$1,
	EmptySmallSquare: EmptySmallSquare$1,
	emptyv: emptyv$1,
	EmptyVerySmallSquare: EmptyVerySmallSquare$1,
	emsp13: emsp13$1,
	emsp14: emsp14$1,
	emsp: emsp$1,
	ENG: ENG$1,
	eng: eng$1,
	ensp: ensp$1,
	Eogon: Eogon$1,
	eogon: eogon$1,
	Eopf: Eopf$1,
	eopf: eopf$1,
	epar: epar$1,
	eparsl: eparsl$1,
	eplus: eplus$1,
	epsi: epsi$1,
	Epsilon: Epsilon$1,
	epsilon: epsilon$1,
	epsiv: epsiv$1,
	eqcirc: eqcirc$1,
	eqcolon: eqcolon$1,
	eqsim: eqsim$1,
	eqslantgtr: eqslantgtr$1,
	eqslantless: eqslantless$1,
	Equal: Equal$1,
	equals: equals$1,
	EqualTilde: EqualTilde$1,
	equest: equest$1,
	Equilibrium: Equilibrium$1,
	equiv: equiv$1,
	equivDD: equivDD$1,
	eqvparsl: eqvparsl$1,
	erarr: erarr$1,
	erDot: erDot$1,
	escr: escr$1,
	Escr: Escr$1,
	esdot: esdot$1,
	Esim: Esim$1,
	esim: esim$1,
	Eta: Eta$1,
	eta: eta$1,
	ETH: ETH$1$1,
	eth: eth$1$1,
	Euml: Euml$1$1,
	euml: euml$1$1,
	euro: euro$1,
	excl: excl$1,
	exist: exist$1,
	Exists: Exists$1,
	expectation: expectation$1,
	exponentiale: exponentiale$1,
	ExponentialE: ExponentialE$1,
	fallingdotseq: fallingdotseq$1,
	Fcy: Fcy$1,
	fcy: fcy$1,
	female: female$1,
	ffilig: ffilig$1,
	fflig: fflig$1,
	ffllig: ffllig$1,
	Ffr: Ffr$1,
	ffr: ffr$1,
	filig: filig$1,
	FilledSmallSquare: FilledSmallSquare$1,
	FilledVerySmallSquare: FilledVerySmallSquare$1,
	fjlig: fjlig$1,
	flat: flat$1,
	fllig: fllig$1,
	fltns: fltns$1,
	fnof: fnof$1,
	Fopf: Fopf$1,
	fopf: fopf$1,
	forall: forall$1,
	ForAll: ForAll$1,
	fork: fork$1,
	forkv: forkv$1,
	Fouriertrf: Fouriertrf$1,
	fpartint: fpartint$1,
	frac12: frac12$1$1,
	frac13: frac13$1,
	frac14: frac14$1$1,
	frac15: frac15$1,
	frac16: frac16$1,
	frac18: frac18$1,
	frac23: frac23$1,
	frac25: frac25$1,
	frac34: frac34$1$1,
	frac35: frac35$1,
	frac38: frac38$1,
	frac45: frac45$1,
	frac56: frac56$1,
	frac58: frac58$1,
	frac78: frac78$1,
	frasl: frasl$1,
	frown: frown$1,
	fscr: fscr$1,
	Fscr: Fscr$1,
	gacute: gacute$1,
	Gamma: Gamma$1,
	gamma: gamma$1,
	Gammad: Gammad$1,
	gammad: gammad$1,
	gap: gap$1,
	Gbreve: Gbreve$1,
	gbreve: gbreve$1,
	Gcedil: Gcedil$1,
	Gcirc: Gcirc$1,
	gcirc: gcirc$1,
	Gcy: Gcy$1,
	gcy: gcy$1,
	Gdot: Gdot$1,
	gdot: gdot$1,
	ge: ge$1,
	gE: gE$1,
	gEl: gEl$1,
	gel: gel$1,
	geq: geq$1,
	geqq: geqq$1,
	geqslant: geqslant$1,
	gescc: gescc$1,
	ges: ges$1,
	gesdot: gesdot$1,
	gesdoto: gesdoto$1,
	gesdotol: gesdotol$1,
	gesl: gesl$1,
	gesles: gesles$1,
	Gfr: Gfr$1,
	gfr: gfr$1,
	gg: gg$1,
	Gg: Gg$1,
	ggg: ggg$1,
	gimel: gimel$1,
	GJcy: GJcy$1,
	gjcy: gjcy$1,
	gla: gla$1,
	gl: gl$1,
	glE: glE$1,
	glj: glj$1,
	gnap: gnap$1,
	gnapprox: gnapprox$1,
	gne: gne$1,
	gnE: gnE$1,
	gneq: gneq$1,
	gneqq: gneqq$1,
	gnsim: gnsim$1,
	Gopf: Gopf$1,
	gopf: gopf$1,
	grave: grave$1,
	GreaterEqual: GreaterEqual$1,
	GreaterEqualLess: GreaterEqualLess$1,
	GreaterFullEqual: GreaterFullEqual$1,
	GreaterGreater: GreaterGreater$1,
	GreaterLess: GreaterLess$1,
	GreaterSlantEqual: GreaterSlantEqual$1,
	GreaterTilde: GreaterTilde$1,
	Gscr: Gscr$1,
	gscr: gscr$1,
	gsim: gsim$1,
	gsime: gsime$1,
	gsiml: gsiml$1,
	gtcc: gtcc$1,
	gtcir: gtcir$1,
	gt: gt$1$2,
	GT: GT$1$1,
	Gt: Gt$1,
	gtdot: gtdot$1,
	gtlPar: gtlPar$1,
	gtquest: gtquest$1,
	gtrapprox: gtrapprox$1,
	gtrarr: gtrarr$1,
	gtrdot: gtrdot$1,
	gtreqless: gtreqless$1,
	gtreqqless: gtreqqless$1,
	gtrless: gtrless$1,
	gtrsim: gtrsim$1,
	gvertneqq: gvertneqq$1,
	gvnE: gvnE$1,
	Hacek: Hacek$1,
	hairsp: hairsp$1,
	half: half$1,
	hamilt: hamilt$1,
	HARDcy: HARDcy$1,
	hardcy: hardcy$1,
	harrcir: harrcir$1,
	harr: harr$1,
	hArr: hArr$1,
	harrw: harrw$1,
	Hat: Hat$1,
	hbar: hbar$1,
	Hcirc: Hcirc$1,
	hcirc: hcirc$1,
	hearts: hearts$1,
	heartsuit: heartsuit$1,
	hellip: hellip$1,
	hercon: hercon$1,
	hfr: hfr$1,
	Hfr: Hfr$1,
	HilbertSpace: HilbertSpace$1,
	hksearow: hksearow$1,
	hkswarow: hkswarow$1,
	hoarr: hoarr$1,
	homtht: homtht$1,
	hookleftarrow: hookleftarrow$1,
	hookrightarrow: hookrightarrow$1,
	hopf: hopf$1,
	Hopf: Hopf$1,
	horbar: horbar$1,
	HorizontalLine: HorizontalLine$1,
	hscr: hscr$1,
	Hscr: Hscr$1,
	hslash: hslash$1,
	Hstrok: Hstrok$1,
	hstrok: hstrok$1,
	HumpDownHump: HumpDownHump$1,
	HumpEqual: HumpEqual$1,
	hybull: hybull$1,
	hyphen: hyphen$1,
	Iacute: Iacute$1$1,
	iacute: iacute$1$1,
	ic: ic$1,
	Icirc: Icirc$1$1,
	icirc: icirc$1$1,
	Icy: Icy$1,
	icy: icy$1,
	Idot: Idot$1,
	IEcy: IEcy$1,
	iecy: iecy$1,
	iexcl: iexcl$1$1,
	iff: iff$1,
	ifr: ifr$1,
	Ifr: Ifr$1,
	Igrave: Igrave$1$1,
	igrave: igrave$1$1,
	ii: ii$1,
	iiiint: iiiint$1,
	iiint: iiint$1,
	iinfin: iinfin$1,
	iiota: iiota$1,
	IJlig: IJlig$1,
	ijlig: ijlig$1,
	Imacr: Imacr$1,
	imacr: imacr$1,
	image: image$1,
	ImaginaryI: ImaginaryI$1,
	imagline: imagline$1,
	imagpart: imagpart$1,
	imath: imath$1,
	Im: Im$1,
	imof: imof$1,
	imped: imped$1,
	Implies: Implies$1,
	incare: incare$1,
	infin: infin$1,
	infintie: infintie$1,
	inodot: inodot$1,
	intcal: intcal$1,
	int: int$1,
	Int: Int$1,
	integers: integers$1,
	Integral: Integral$1,
	intercal: intercal$1,
	Intersection: Intersection$1,
	intlarhk: intlarhk$1,
	intprod: intprod$1,
	InvisibleComma: InvisibleComma$1,
	InvisibleTimes: InvisibleTimes$1,
	IOcy: IOcy$1,
	iocy: iocy$1,
	Iogon: Iogon$1,
	iogon: iogon$1,
	Iopf: Iopf$1,
	iopf: iopf$1,
	Iota: Iota$1,
	iota: iota$1,
	iprod: iprod$1,
	iquest: iquest$1$1,
	iscr: iscr$1,
	Iscr: Iscr$1,
	isin: isin$1,
	isindot: isindot$1,
	isinE: isinE$1,
	isins: isins$1,
	isinsv: isinsv$1,
	isinv: isinv$1,
	it: it$1,
	Itilde: Itilde$1,
	itilde: itilde$1,
	Iukcy: Iukcy$1,
	iukcy: iukcy$1,
	Iuml: Iuml$1$1,
	iuml: iuml$1$1,
	Jcirc: Jcirc$1,
	jcirc: jcirc$1,
	Jcy: Jcy$1,
	jcy: jcy$1,
	Jfr: Jfr$1,
	jfr: jfr$1,
	jmath: jmath$1,
	Jopf: Jopf$1,
	jopf: jopf$1,
	Jscr: Jscr$1,
	jscr: jscr$1,
	Jsercy: Jsercy$1,
	jsercy: jsercy$1,
	Jukcy: Jukcy$1,
	jukcy: jukcy$1,
	Kappa: Kappa$1,
	kappa: kappa$1,
	kappav: kappav$1,
	Kcedil: Kcedil$1,
	kcedil: kcedil$1,
	Kcy: Kcy$1,
	kcy: kcy$1,
	Kfr: Kfr$1,
	kfr: kfr$1,
	kgreen: kgreen$1,
	KHcy: KHcy$1,
	khcy: khcy$1,
	KJcy: KJcy$1,
	kjcy: kjcy$1,
	Kopf: Kopf$1,
	kopf: kopf$1,
	Kscr: Kscr$1,
	kscr: kscr$1,
	lAarr: lAarr$1,
	Lacute: Lacute$1,
	lacute: lacute$1,
	laemptyv: laemptyv$1,
	lagran: lagran$1,
	Lambda: Lambda$1,
	lambda: lambda$1,
	lang: lang$1,
	Lang: Lang$1,
	langd: langd$1,
	langle: langle$1,
	lap: lap$1,
	Laplacetrf: Laplacetrf$1,
	laquo: laquo$1$1,
	larrb: larrb$1,
	larrbfs: larrbfs$1,
	larr: larr$1,
	Larr: Larr$1,
	lArr: lArr$1,
	larrfs: larrfs$1,
	larrhk: larrhk$1,
	larrlp: larrlp$1,
	larrpl: larrpl$1,
	larrsim: larrsim$1,
	larrtl: larrtl$1,
	latail: latail$1,
	lAtail: lAtail$1,
	lat: lat$1,
	late: late$1,
	lates: lates$1,
	lbarr: lbarr$1,
	lBarr: lBarr$1,
	lbbrk: lbbrk$1,
	lbrace: lbrace$1,
	lbrack: lbrack$1,
	lbrke: lbrke$1,
	lbrksld: lbrksld$1,
	lbrkslu: lbrkslu$1,
	Lcaron: Lcaron$1,
	lcaron: lcaron$1,
	Lcedil: Lcedil$1,
	lcedil: lcedil$1,
	lceil: lceil$1,
	lcub: lcub$1,
	Lcy: Lcy$1,
	lcy: lcy$1,
	ldca: ldca$1,
	ldquo: ldquo$1,
	ldquor: ldquor$1,
	ldrdhar: ldrdhar$1,
	ldrushar: ldrushar$1,
	ldsh: ldsh$1,
	le: le$1,
	lE: lE$1,
	LeftAngleBracket: LeftAngleBracket$1,
	LeftArrowBar: LeftArrowBar$1,
	leftarrow: leftarrow$1,
	LeftArrow: LeftArrow$1,
	Leftarrow: Leftarrow$1,
	LeftArrowRightArrow: LeftArrowRightArrow$1,
	leftarrowtail: leftarrowtail$1,
	LeftCeiling: LeftCeiling$1,
	LeftDoubleBracket: LeftDoubleBracket$1,
	LeftDownTeeVector: LeftDownTeeVector$1,
	LeftDownVectorBar: LeftDownVectorBar$1,
	LeftDownVector: LeftDownVector$1,
	LeftFloor: LeftFloor$1,
	leftharpoondown: leftharpoondown$1,
	leftharpoonup: leftharpoonup$1,
	leftleftarrows: leftleftarrows$1,
	leftrightarrow: leftrightarrow$1,
	LeftRightArrow: LeftRightArrow$1,
	Leftrightarrow: Leftrightarrow$1,
	leftrightarrows: leftrightarrows$1,
	leftrightharpoons: leftrightharpoons$1,
	leftrightsquigarrow: leftrightsquigarrow$1,
	LeftRightVector: LeftRightVector$1,
	LeftTeeArrow: LeftTeeArrow$1,
	LeftTee: LeftTee$1,
	LeftTeeVector: LeftTeeVector$1,
	leftthreetimes: leftthreetimes$1,
	LeftTriangleBar: LeftTriangleBar$1,
	LeftTriangle: LeftTriangle$1,
	LeftTriangleEqual: LeftTriangleEqual$1,
	LeftUpDownVector: LeftUpDownVector$1,
	LeftUpTeeVector: LeftUpTeeVector$1,
	LeftUpVectorBar: LeftUpVectorBar$1,
	LeftUpVector: LeftUpVector$1,
	LeftVectorBar: LeftVectorBar$1,
	LeftVector: LeftVector$1,
	lEg: lEg$1,
	leg: leg$1,
	leq: leq$1,
	leqq: leqq$1,
	leqslant: leqslant$1,
	lescc: lescc$1,
	les: les$1,
	lesdot: lesdot$1,
	lesdoto: lesdoto$1,
	lesdotor: lesdotor$1,
	lesg: lesg$1,
	lesges: lesges$1,
	lessapprox: lessapprox$1,
	lessdot: lessdot$1,
	lesseqgtr: lesseqgtr$1,
	lesseqqgtr: lesseqqgtr$1,
	LessEqualGreater: LessEqualGreater$1,
	LessFullEqual: LessFullEqual$1,
	LessGreater: LessGreater$1,
	lessgtr: lessgtr$1,
	LessLess: LessLess$1,
	lesssim: lesssim$1,
	LessSlantEqual: LessSlantEqual$1,
	LessTilde: LessTilde$1,
	lfisht: lfisht$1,
	lfloor: lfloor$1,
	Lfr: Lfr$1,
	lfr: lfr$1,
	lg: lg$1,
	lgE: lgE$1,
	lHar: lHar$1,
	lhard: lhard$1,
	lharu: lharu$1,
	lharul: lharul$1,
	lhblk: lhblk$1,
	LJcy: LJcy$1,
	ljcy: ljcy$1,
	llarr: llarr$1,
	ll: ll$1,
	Ll: Ll$1,
	llcorner: llcorner$1,
	Lleftarrow: Lleftarrow$1,
	llhard: llhard$1,
	lltri: lltri$1,
	Lmidot: Lmidot$1,
	lmidot: lmidot$1,
	lmoustache: lmoustache$1,
	lmoust: lmoust$1,
	lnap: lnap$1,
	lnapprox: lnapprox$1,
	lne: lne$1,
	lnE: lnE$1,
	lneq: lneq$1,
	lneqq: lneqq$1,
	lnsim: lnsim$1,
	loang: loang$1,
	loarr: loarr$1,
	lobrk: lobrk$1,
	longleftarrow: longleftarrow$1,
	LongLeftArrow: LongLeftArrow$1,
	Longleftarrow: Longleftarrow$1,
	longleftrightarrow: longleftrightarrow$1,
	LongLeftRightArrow: LongLeftRightArrow$1,
	Longleftrightarrow: Longleftrightarrow$1,
	longmapsto: longmapsto$1,
	longrightarrow: longrightarrow$1,
	LongRightArrow: LongRightArrow$1,
	Longrightarrow: Longrightarrow$1,
	looparrowleft: looparrowleft$1,
	looparrowright: looparrowright$1,
	lopar: lopar$1,
	Lopf: Lopf$1,
	lopf: lopf$1,
	loplus: loplus$1,
	lotimes: lotimes$1,
	lowast: lowast$1,
	lowbar: lowbar$1,
	LowerLeftArrow: LowerLeftArrow$1,
	LowerRightArrow: LowerRightArrow$1,
	loz: loz$1,
	lozenge: lozenge$1,
	lozf: lozf$1,
	lpar: lpar$1,
	lparlt: lparlt$1,
	lrarr: lrarr$1,
	lrcorner: lrcorner$1,
	lrhar: lrhar$1,
	lrhard: lrhard$1,
	lrm: lrm$1,
	lrtri: lrtri$1,
	lsaquo: lsaquo$1,
	lscr: lscr$1,
	Lscr: Lscr$1,
	lsh: lsh$1,
	Lsh: Lsh$1,
	lsim: lsim$1,
	lsime: lsime$1,
	lsimg: lsimg$1,
	lsqb: lsqb$1,
	lsquo: lsquo$1,
	lsquor: lsquor$1,
	Lstrok: Lstrok$1,
	lstrok: lstrok$1,
	ltcc: ltcc$1,
	ltcir: ltcir$1,
	lt: lt$1$2,
	LT: LT$1$1,
	Lt: Lt$1,
	ltdot: ltdot$1,
	lthree: lthree$1,
	ltimes: ltimes$1,
	ltlarr: ltlarr$1,
	ltquest: ltquest$1,
	ltri: ltri$1,
	ltrie: ltrie$1,
	ltrif: ltrif$1,
	ltrPar: ltrPar$1,
	lurdshar: lurdshar$1,
	luruhar: luruhar$1,
	lvertneqq: lvertneqq$1,
	lvnE: lvnE$1,
	macr: macr$1$1,
	male: male$1,
	malt: malt$1,
	maltese: maltese$1,
	map: map$2,
	mapsto: mapsto$1,
	mapstodown: mapstodown$1,
	mapstoleft: mapstoleft$1,
	mapstoup: mapstoup$1,
	marker: marker$1,
	mcomma: mcomma$1,
	Mcy: Mcy$1,
	mcy: mcy$1,
	mdash: mdash$1,
	mDDot: mDDot$1,
	measuredangle: measuredangle$1,
	MediumSpace: MediumSpace$1,
	Mellintrf: Mellintrf$1,
	Mfr: Mfr$1,
	mfr: mfr$1,
	mho: mho$1,
	micro: micro$1$1,
	midast: midast$1,
	midcir: midcir$1,
	mid: mid$1,
	middot: middot$1$1,
	minusb: minusb$1,
	minus: minus$1,
	minusd: minusd$1,
	minusdu: minusdu$1,
	MinusPlus: MinusPlus$1,
	mlcp: mlcp$1,
	mldr: mldr$1,
	mnplus: mnplus$1,
	models: models$1,
	Mopf: Mopf$1,
	mopf: mopf$1,
	mp: mp$1,
	mscr: mscr$1,
	Mscr: Mscr$1,
	mstpos: mstpos$1,
	Mu: Mu$1,
	mu: mu$1,
	multimap: multimap$1,
	mumap: mumap$1,
	nabla: nabla$1,
	Nacute: Nacute$1,
	nacute: nacute$1,
	nang: nang$1,
	nap: nap$1,
	napE: napE$1,
	napid: napid$1,
	napos: napos$1,
	napprox: napprox$1,
	natural: natural$1,
	naturals: naturals$1,
	natur: natur$1,
	nbsp: nbsp$1$1,
	nbump: nbump$1,
	nbumpe: nbumpe$1,
	ncap: ncap$1,
	Ncaron: Ncaron$1,
	ncaron: ncaron$1,
	Ncedil: Ncedil$1,
	ncedil: ncedil$1,
	ncong: ncong$1,
	ncongdot: ncongdot$1,
	ncup: ncup$1,
	Ncy: Ncy$1,
	ncy: ncy$1,
	ndash: ndash$1,
	nearhk: nearhk$1,
	nearr: nearr$1,
	neArr: neArr$1,
	nearrow: nearrow$1,
	ne: ne$1,
	nedot: nedot$1,
	NegativeMediumSpace: NegativeMediumSpace$1,
	NegativeThickSpace: NegativeThickSpace$1,
	NegativeThinSpace: NegativeThinSpace$1,
	NegativeVeryThinSpace: NegativeVeryThinSpace$1,
	nequiv: nequiv$1,
	nesear: nesear$1,
	nesim: nesim$1,
	NestedGreaterGreater: NestedGreaterGreater$1,
	NestedLessLess: NestedLessLess$1,
	NewLine: NewLine$1,
	nexist: nexist$1,
	nexists: nexists$1,
	Nfr: Nfr$1,
	nfr: nfr$1,
	ngE: ngE$1,
	nge: nge$1,
	ngeq: ngeq$1,
	ngeqq: ngeqq$1,
	ngeqslant: ngeqslant$1,
	nges: nges$1,
	nGg: nGg$1,
	ngsim: ngsim$1,
	nGt: nGt$1,
	ngt: ngt$1,
	ngtr: ngtr$1,
	nGtv: nGtv$1,
	nharr: nharr$1,
	nhArr: nhArr$1,
	nhpar: nhpar$1,
	ni: ni$1,
	nis: nis$1,
	nisd: nisd$1,
	niv: niv$1,
	NJcy: NJcy$1,
	njcy: njcy$1,
	nlarr: nlarr$1,
	nlArr: nlArr$1,
	nldr: nldr$1,
	nlE: nlE$1,
	nle: nle$1,
	nleftarrow: nleftarrow$1,
	nLeftarrow: nLeftarrow$1,
	nleftrightarrow: nleftrightarrow$1,
	nLeftrightarrow: nLeftrightarrow$1,
	nleq: nleq$1,
	nleqq: nleqq$1,
	nleqslant: nleqslant$1,
	nles: nles$1,
	nless: nless$1,
	nLl: nLl$1,
	nlsim: nlsim$1,
	nLt: nLt$1,
	nlt: nlt$1,
	nltri: nltri$1,
	nltrie: nltrie$1,
	nLtv: nLtv$1,
	nmid: nmid$1,
	NoBreak: NoBreak$1,
	NonBreakingSpace: NonBreakingSpace$1,
	nopf: nopf$1,
	Nopf: Nopf$1,
	Not: Not$1,
	not: not$1$1,
	NotCongruent: NotCongruent$1,
	NotCupCap: NotCupCap$1,
	NotDoubleVerticalBar: NotDoubleVerticalBar$1,
	NotElement: NotElement$1,
	NotEqual: NotEqual$1,
	NotEqualTilde: NotEqualTilde$1,
	NotExists: NotExists$1,
	NotGreater: NotGreater$1,
	NotGreaterEqual: NotGreaterEqual$1,
	NotGreaterFullEqual: NotGreaterFullEqual$1,
	NotGreaterGreater: NotGreaterGreater$1,
	NotGreaterLess: NotGreaterLess$1,
	NotGreaterSlantEqual: NotGreaterSlantEqual$1,
	NotGreaterTilde: NotGreaterTilde$1,
	NotHumpDownHump: NotHumpDownHump$1,
	NotHumpEqual: NotHumpEqual$1,
	notin: notin$1,
	notindot: notindot$1,
	notinE: notinE$1,
	notinva: notinva$1,
	notinvb: notinvb$1,
	notinvc: notinvc$1,
	NotLeftTriangleBar: NotLeftTriangleBar$1,
	NotLeftTriangle: NotLeftTriangle$1,
	NotLeftTriangleEqual: NotLeftTriangleEqual$1,
	NotLess: NotLess$1,
	NotLessEqual: NotLessEqual$1,
	NotLessGreater: NotLessGreater$1,
	NotLessLess: NotLessLess$1,
	NotLessSlantEqual: NotLessSlantEqual$1,
	NotLessTilde: NotLessTilde$1,
	NotNestedGreaterGreater: NotNestedGreaterGreater$1,
	NotNestedLessLess: NotNestedLessLess$1,
	notni: notni$1,
	notniva: notniva$1,
	notnivb: notnivb$1,
	notnivc: notnivc$1,
	NotPrecedes: NotPrecedes$1,
	NotPrecedesEqual: NotPrecedesEqual$1,
	NotPrecedesSlantEqual: NotPrecedesSlantEqual$1,
	NotReverseElement: NotReverseElement$1,
	NotRightTriangleBar: NotRightTriangleBar$1,
	NotRightTriangle: NotRightTriangle$1,
	NotRightTriangleEqual: NotRightTriangleEqual$1,
	NotSquareSubset: NotSquareSubset$1,
	NotSquareSubsetEqual: NotSquareSubsetEqual$1,
	NotSquareSuperset: NotSquareSuperset$1,
	NotSquareSupersetEqual: NotSquareSupersetEqual$1,
	NotSubset: NotSubset$1,
	NotSubsetEqual: NotSubsetEqual$1,
	NotSucceeds: NotSucceeds$1,
	NotSucceedsEqual: NotSucceedsEqual$1,
	NotSucceedsSlantEqual: NotSucceedsSlantEqual$1,
	NotSucceedsTilde: NotSucceedsTilde$1,
	NotSuperset: NotSuperset$1,
	NotSupersetEqual: NotSupersetEqual$1,
	NotTilde: NotTilde$1,
	NotTildeEqual: NotTildeEqual$1,
	NotTildeFullEqual: NotTildeFullEqual$1,
	NotTildeTilde: NotTildeTilde$1,
	NotVerticalBar: NotVerticalBar$1,
	nparallel: nparallel$1,
	npar: npar$1,
	nparsl: nparsl$1,
	npart: npart$1,
	npolint: npolint$1,
	npr: npr$1,
	nprcue: nprcue$1,
	nprec: nprec$1,
	npreceq: npreceq$1,
	npre: npre$1,
	nrarrc: nrarrc$1,
	nrarr: nrarr$1,
	nrArr: nrArr$1,
	nrarrw: nrarrw$1,
	nrightarrow: nrightarrow$1,
	nRightarrow: nRightarrow$1,
	nrtri: nrtri$1,
	nrtrie: nrtrie$1,
	nsc: nsc$1,
	nsccue: nsccue$1,
	nsce: nsce$1,
	Nscr: Nscr$1,
	nscr: nscr$1,
	nshortmid: nshortmid$1,
	nshortparallel: nshortparallel$1,
	nsim: nsim$1,
	nsime: nsime$1,
	nsimeq: nsimeq$1,
	nsmid: nsmid$1,
	nspar: nspar$1,
	nsqsube: nsqsube$1,
	nsqsupe: nsqsupe$1,
	nsub: nsub$1,
	nsubE: nsubE$1,
	nsube: nsube$1,
	nsubset: nsubset$1,
	nsubseteq: nsubseteq$1,
	nsubseteqq: nsubseteqq$1,
	nsucc: nsucc$1,
	nsucceq: nsucceq$1,
	nsup: nsup$1,
	nsupE: nsupE$1,
	nsupe: nsupe$1,
	nsupset: nsupset$1,
	nsupseteq: nsupseteq$1,
	nsupseteqq: nsupseteqq$1,
	ntgl: ntgl$1,
	Ntilde: Ntilde$1$1,
	ntilde: ntilde$1$1,
	ntlg: ntlg$1,
	ntriangleleft: ntriangleleft$1,
	ntrianglelefteq: ntrianglelefteq$1,
	ntriangleright: ntriangleright$1,
	ntrianglerighteq: ntrianglerighteq$1,
	Nu: Nu$1,
	nu: nu$1,
	num: num$1,
	numero: numero$1,
	numsp: numsp$1,
	nvap: nvap$1,
	nvdash: nvdash$1,
	nvDash: nvDash$1,
	nVdash: nVdash$1,
	nVDash: nVDash$1,
	nvge: nvge$1,
	nvgt: nvgt$1,
	nvHarr: nvHarr$1,
	nvinfin: nvinfin$1,
	nvlArr: nvlArr$1,
	nvle: nvle$1,
	nvlt: nvlt$1,
	nvltrie: nvltrie$1,
	nvrArr: nvrArr$1,
	nvrtrie: nvrtrie$1,
	nvsim: nvsim$1,
	nwarhk: nwarhk$1,
	nwarr: nwarr$1,
	nwArr: nwArr$1,
	nwarrow: nwarrow$1,
	nwnear: nwnear$1,
	Oacute: Oacute$1$1,
	oacute: oacute$1$1,
	oast: oast$1,
	Ocirc: Ocirc$1$1,
	ocirc: ocirc$1$1,
	ocir: ocir$1,
	Ocy: Ocy$1,
	ocy: ocy$1,
	odash: odash$1,
	Odblac: Odblac$1,
	odblac: odblac$1,
	odiv: odiv$1,
	odot: odot$1,
	odsold: odsold$1,
	OElig: OElig$1,
	oelig: oelig$1,
	ofcir: ofcir$1,
	Ofr: Ofr$1,
	ofr: ofr$1,
	ogon: ogon$1,
	Ograve: Ograve$1$1,
	ograve: ograve$1$1,
	ogt: ogt$1,
	ohbar: ohbar$1,
	ohm: ohm$1,
	oint: oint$1,
	olarr: olarr$1,
	olcir: olcir$1,
	olcross: olcross$1,
	oline: oline$1,
	olt: olt$1,
	Omacr: Omacr$1,
	omacr: omacr$1,
	Omega: Omega$1,
	omega: omega$1,
	Omicron: Omicron$1,
	omicron: omicron$1,
	omid: omid$1,
	ominus: ominus$1,
	Oopf: Oopf$1,
	oopf: oopf$1,
	opar: opar$1,
	OpenCurlyDoubleQuote: OpenCurlyDoubleQuote$1,
	OpenCurlyQuote: OpenCurlyQuote$1,
	operp: operp$1,
	oplus: oplus$1,
	orarr: orarr$1,
	Or: Or$1,
	or: or$1,
	ord: ord$1,
	order: order$1,
	orderof: orderof$1,
	ordf: ordf$1$1,
	ordm: ordm$1$1,
	origof: origof$1,
	oror: oror$1,
	orslope: orslope$1,
	orv: orv$1,
	oS: oS$1,
	Oscr: Oscr$1,
	oscr: oscr$1,
	Oslash: Oslash$1$1,
	oslash: oslash$1$1,
	osol: osol$1,
	Otilde: Otilde$1$1,
	otilde: otilde$1$1,
	otimesas: otimesas$1,
	Otimes: Otimes$1,
	otimes: otimes$1,
	Ouml: Ouml$1$1,
	ouml: ouml$1$1,
	ovbar: ovbar$1,
	OverBar: OverBar$1,
	OverBrace: OverBrace$1,
	OverBracket: OverBracket$1,
	OverParenthesis: OverParenthesis$1,
	para: para$1$1,
	parallel: parallel$1,
	par: par$1,
	parsim: parsim$1,
	parsl: parsl$1,
	part: part$1,
	PartialD: PartialD$1,
	Pcy: Pcy$1,
	pcy: pcy$1,
	percnt: percnt$1,
	period: period$1,
	permil: permil$1,
	perp: perp$1,
	pertenk: pertenk$1,
	Pfr: Pfr$1,
	pfr: pfr$1,
	Phi: Phi$1,
	phi: phi$1,
	phiv: phiv$1,
	phmmat: phmmat$1,
	phone: phone$1,
	Pi: Pi$1,
	pi: pi$1,
	pitchfork: pitchfork$1,
	piv: piv$1,
	planck: planck$1,
	planckh: planckh$1,
	plankv: plankv$1,
	plusacir: plusacir$1,
	plusb: plusb$1,
	pluscir: pluscir$1,
	plus: plus$1,
	plusdo: plusdo$1,
	plusdu: plusdu$1,
	pluse: pluse$1,
	PlusMinus: PlusMinus$1,
	plusmn: plusmn$1$1,
	plussim: plussim$1,
	plustwo: plustwo$1,
	pm: pm$1,
	Poincareplane: Poincareplane$1,
	pointint: pointint$1,
	popf: popf$1,
	Popf: Popf$1,
	pound: pound$1$1,
	prap: prap$1,
	Pr: Pr$1,
	pr: pr$1,
	prcue: prcue$1,
	precapprox: precapprox$1,
	prec: prec$1,
	preccurlyeq: preccurlyeq$1,
	Precedes: Precedes$1,
	PrecedesEqual: PrecedesEqual$1,
	PrecedesSlantEqual: PrecedesSlantEqual$1,
	PrecedesTilde: PrecedesTilde$1,
	preceq: preceq$1,
	precnapprox: precnapprox$1,
	precneqq: precneqq$1,
	precnsim: precnsim$1,
	pre: pre$1,
	prE: prE$1,
	precsim: precsim$1,
	prime: prime$1,
	Prime: Prime$1,
	primes: primes$1,
	prnap: prnap$1,
	prnE: prnE$1,
	prnsim: prnsim$1,
	prod: prod$1,
	Product: Product$1,
	profalar: profalar$1,
	profline: profline$1,
	profsurf: profsurf$1,
	prop: prop$1,
	Proportional: Proportional$1,
	Proportion: Proportion$1,
	propto: propto$1,
	prsim: prsim$1,
	prurel: prurel$1,
	Pscr: Pscr$1,
	pscr: pscr$1,
	Psi: Psi$1,
	psi: psi$1,
	puncsp: puncsp$1,
	Qfr: Qfr$1,
	qfr: qfr$1,
	qint: qint$1,
	qopf: qopf$1,
	Qopf: Qopf$1,
	qprime: qprime$1,
	Qscr: Qscr$1,
	qscr: qscr$1,
	quaternions: quaternions$1,
	quatint: quatint$1,
	quest: quest$1,
	questeq: questeq$1,
	quot: quot$1$2,
	QUOT: QUOT$1$1,
	rAarr: rAarr$1,
	race: race$1,
	Racute: Racute$1,
	racute: racute$1,
	radic: radic$1,
	raemptyv: raemptyv$1,
	rang: rang$1,
	Rang: Rang$1,
	rangd: rangd$1,
	range: range$1,
	rangle: rangle$1,
	raquo: raquo$1$1,
	rarrap: rarrap$1,
	rarrb: rarrb$1,
	rarrbfs: rarrbfs$1,
	rarrc: rarrc$1,
	rarr: rarr$1,
	Rarr: Rarr$1,
	rArr: rArr$1,
	rarrfs: rarrfs$1,
	rarrhk: rarrhk$1,
	rarrlp: rarrlp$1,
	rarrpl: rarrpl$1,
	rarrsim: rarrsim$1,
	Rarrtl: Rarrtl$1,
	rarrtl: rarrtl$1,
	rarrw: rarrw$1,
	ratail: ratail$1,
	rAtail: rAtail$1,
	ratio: ratio$1,
	rationals: rationals$1,
	rbarr: rbarr$1,
	rBarr: rBarr$1,
	RBarr: RBarr$1,
	rbbrk: rbbrk$1,
	rbrace: rbrace$1,
	rbrack: rbrack$1,
	rbrke: rbrke$1,
	rbrksld: rbrksld$1,
	rbrkslu: rbrkslu$1,
	Rcaron: Rcaron$1,
	rcaron: rcaron$1,
	Rcedil: Rcedil$1,
	rcedil: rcedil$1,
	rceil: rceil$1,
	rcub: rcub$1,
	Rcy: Rcy$1,
	rcy: rcy$1,
	rdca: rdca$1,
	rdldhar: rdldhar$1,
	rdquo: rdquo$1,
	rdquor: rdquor$1,
	rdsh: rdsh$1,
	real: real$1,
	realine: realine$1,
	realpart: realpart$1,
	reals: reals$1,
	Re: Re$1,
	rect: rect$1,
	reg: reg$1$1,
	REG: REG$1$1,
	ReverseElement: ReverseElement$1,
	ReverseEquilibrium: ReverseEquilibrium$1,
	ReverseUpEquilibrium: ReverseUpEquilibrium$1,
	rfisht: rfisht$1,
	rfloor: rfloor$1,
	rfr: rfr$1,
	Rfr: Rfr$1,
	rHar: rHar$1,
	rhard: rhard$1,
	rharu: rharu$1,
	rharul: rharul$1,
	Rho: Rho$1,
	rho: rho$1,
	rhov: rhov$1,
	RightAngleBracket: RightAngleBracket$1,
	RightArrowBar: RightArrowBar$1,
	rightarrow: rightarrow$1,
	RightArrow: RightArrow$1,
	Rightarrow: Rightarrow$1,
	RightArrowLeftArrow: RightArrowLeftArrow$1,
	rightarrowtail: rightarrowtail$1,
	RightCeiling: RightCeiling$1,
	RightDoubleBracket: RightDoubleBracket$1,
	RightDownTeeVector: RightDownTeeVector$1,
	RightDownVectorBar: RightDownVectorBar$1,
	RightDownVector: RightDownVector$1,
	RightFloor: RightFloor$1,
	rightharpoondown: rightharpoondown$1,
	rightharpoonup: rightharpoonup$1,
	rightleftarrows: rightleftarrows$1,
	rightleftharpoons: rightleftharpoons$1,
	rightrightarrows: rightrightarrows$1,
	rightsquigarrow: rightsquigarrow$1,
	RightTeeArrow: RightTeeArrow$1,
	RightTee: RightTee$1,
	RightTeeVector: RightTeeVector$1,
	rightthreetimes: rightthreetimes$1,
	RightTriangleBar: RightTriangleBar$1,
	RightTriangle: RightTriangle$1,
	RightTriangleEqual: RightTriangleEqual$1,
	RightUpDownVector: RightUpDownVector$1,
	RightUpTeeVector: RightUpTeeVector$1,
	RightUpVectorBar: RightUpVectorBar$1,
	RightUpVector: RightUpVector$1,
	RightVectorBar: RightVectorBar$1,
	RightVector: RightVector$1,
	ring: ring$1,
	risingdotseq: risingdotseq$1,
	rlarr: rlarr$1,
	rlhar: rlhar$1,
	rlm: rlm$1,
	rmoustache: rmoustache$1,
	rmoust: rmoust$1,
	rnmid: rnmid$1,
	roang: roang$1,
	roarr: roarr$1,
	robrk: robrk$1,
	ropar: ropar$1,
	ropf: ropf$1,
	Ropf: Ropf$1,
	roplus: roplus$1,
	rotimes: rotimes$1,
	RoundImplies: RoundImplies$1,
	rpar: rpar$1,
	rpargt: rpargt$1,
	rppolint: rppolint$1,
	rrarr: rrarr$1,
	Rrightarrow: Rrightarrow$1,
	rsaquo: rsaquo$1,
	rscr: rscr$1,
	Rscr: Rscr$1,
	rsh: rsh$1,
	Rsh: Rsh$1,
	rsqb: rsqb$1,
	rsquo: rsquo$1,
	rsquor: rsquor$1,
	rthree: rthree$1,
	rtimes: rtimes$1,
	rtri: rtri$1,
	rtrie: rtrie$1,
	rtrif: rtrif$1,
	rtriltri: rtriltri$1,
	RuleDelayed: RuleDelayed$1,
	ruluhar: ruluhar$1,
	rx: rx$1,
	Sacute: Sacute$1,
	sacute: sacute$1,
	sbquo: sbquo$1,
	scap: scap$1,
	Scaron: Scaron$1,
	scaron: scaron$1,
	Sc: Sc$1,
	sc: sc$1,
	sccue: sccue$1,
	sce: sce$1,
	scE: scE$1,
	Scedil: Scedil$1,
	scedil: scedil$1,
	Scirc: Scirc$1,
	scirc: scirc$1,
	scnap: scnap$1,
	scnE: scnE$1,
	scnsim: scnsim$1,
	scpolint: scpolint$1,
	scsim: scsim$1,
	Scy: Scy$1,
	scy: scy$1,
	sdotb: sdotb$1,
	sdot: sdot$1,
	sdote: sdote$1,
	searhk: searhk$1,
	searr: searr$1,
	seArr: seArr$1,
	searrow: searrow$1,
	sect: sect$1$1,
	semi: semi$1,
	seswar: seswar$1,
	setminus: setminus$1,
	setmn: setmn$1,
	sext: sext$1,
	Sfr: Sfr$1,
	sfr: sfr$1,
	sfrown: sfrown$1,
	sharp: sharp$1,
	SHCHcy: SHCHcy$1,
	shchcy: shchcy$1,
	SHcy: SHcy$1,
	shcy: shcy$1,
	ShortDownArrow: ShortDownArrow$1,
	ShortLeftArrow: ShortLeftArrow$1,
	shortmid: shortmid$1,
	shortparallel: shortparallel$1,
	ShortRightArrow: ShortRightArrow$1,
	ShortUpArrow: ShortUpArrow$1,
	shy: shy$1$1,
	Sigma: Sigma$1,
	sigma: sigma$1,
	sigmaf: sigmaf$1,
	sigmav: sigmav$1,
	sim: sim$1,
	simdot: simdot$1,
	sime: sime$1,
	simeq: simeq$1,
	simg: simg$1,
	simgE: simgE$1,
	siml: siml$1,
	simlE: simlE$1,
	simne: simne$1,
	simplus: simplus$1,
	simrarr: simrarr$1,
	slarr: slarr$1,
	SmallCircle: SmallCircle$1,
	smallsetminus: smallsetminus$1,
	smashp: smashp$1,
	smeparsl: smeparsl$1,
	smid: smid$1,
	smile: smile$1,
	smt: smt$1,
	smte: smte$1,
	smtes: smtes$1,
	SOFTcy: SOFTcy$1,
	softcy: softcy$1,
	solbar: solbar$1,
	solb: solb$1,
	sol: sol$1,
	Sopf: Sopf$1,
	sopf: sopf$1,
	spades: spades$1,
	spadesuit: spadesuit$1,
	spar: spar$1,
	sqcap: sqcap$1,
	sqcaps: sqcaps$1,
	sqcup: sqcup$1,
	sqcups: sqcups$1,
	Sqrt: Sqrt$1,
	sqsub: sqsub$1,
	sqsube: sqsube$1,
	sqsubset: sqsubset$1,
	sqsubseteq: sqsubseteq$1,
	sqsup: sqsup$1,
	sqsupe: sqsupe$1,
	sqsupset: sqsupset$1,
	sqsupseteq: sqsupseteq$1,
	square: square$1,
	Square: Square$1,
	SquareIntersection: SquareIntersection$1,
	SquareSubset: SquareSubset$1,
	SquareSubsetEqual: SquareSubsetEqual$1,
	SquareSuperset: SquareSuperset$1,
	SquareSupersetEqual: SquareSupersetEqual$1,
	SquareUnion: SquareUnion$1,
	squarf: squarf$1,
	squ: squ$1,
	squf: squf$1,
	srarr: srarr$1,
	Sscr: Sscr$1,
	sscr: sscr$1,
	ssetmn: ssetmn$1,
	ssmile: ssmile$1,
	sstarf: sstarf$1,
	Star: Star$1,
	star: star$1,
	starf: starf$1,
	straightepsilon: straightepsilon$1,
	straightphi: straightphi$1,
	strns: strns$1,
	sub: sub$1,
	Sub: Sub$1,
	subdot: subdot$1,
	subE: subE$1,
	sube: sube$1,
	subedot: subedot$1,
	submult: submult$1,
	subnE: subnE$1,
	subne: subne$1,
	subplus: subplus$1,
	subrarr: subrarr$1,
	subset: subset$1,
	Subset: Subset$1,
	subseteq: subseteq$1,
	subseteqq: subseteqq$1,
	SubsetEqual: SubsetEqual$1,
	subsetneq: subsetneq$1,
	subsetneqq: subsetneqq$1,
	subsim: subsim$1,
	subsub: subsub$1,
	subsup: subsup$1,
	succapprox: succapprox$1,
	succ: succ$1,
	succcurlyeq: succcurlyeq$1,
	Succeeds: Succeeds$1,
	SucceedsEqual: SucceedsEqual$1,
	SucceedsSlantEqual: SucceedsSlantEqual$1,
	SucceedsTilde: SucceedsTilde$1,
	succeq: succeq$1,
	succnapprox: succnapprox$1,
	succneqq: succneqq$1,
	succnsim: succnsim$1,
	succsim: succsim$1,
	SuchThat: SuchThat$1,
	sum: sum$1,
	Sum: Sum$1,
	sung: sung$1,
	sup1: sup1$1$1,
	sup2: sup2$1$1,
	sup3: sup3$1$1,
	sup: sup$1,
	Sup: Sup$1,
	supdot: supdot$1,
	supdsub: supdsub$1,
	supE: supE$1,
	supe: supe$1,
	supedot: supedot$1,
	Superset: Superset$1,
	SupersetEqual: SupersetEqual$1,
	suphsol: suphsol$1,
	suphsub: suphsub$1,
	suplarr: suplarr$1,
	supmult: supmult$1,
	supnE: supnE$1,
	supne: supne$1,
	supplus: supplus$1,
	supset: supset$1,
	Supset: Supset$1,
	supseteq: supseteq$1,
	supseteqq: supseteqq$1,
	supsetneq: supsetneq$1,
	supsetneqq: supsetneqq$1,
	supsim: supsim$1,
	supsub: supsub$1,
	supsup: supsup$1,
	swarhk: swarhk$1,
	swarr: swarr$1,
	swArr: swArr$1,
	swarrow: swarrow$1,
	swnwar: swnwar$1,
	szlig: szlig$1$1,
	Tab: Tab$1,
	target: target$1,
	Tau: Tau$1,
	tau: tau$1,
	tbrk: tbrk$1,
	Tcaron: Tcaron$1,
	tcaron: tcaron$1,
	Tcedil: Tcedil$1,
	tcedil: tcedil$1,
	Tcy: Tcy$1,
	tcy: tcy$1,
	tdot: tdot$1,
	telrec: telrec$1,
	Tfr: Tfr$1,
	tfr: tfr$1,
	there4: there4$1,
	therefore: therefore$1,
	Therefore: Therefore$1,
	Theta: Theta$1,
	theta: theta$1,
	thetasym: thetasym$1,
	thetav: thetav$1,
	thickapprox: thickapprox$1,
	thicksim: thicksim$1,
	ThickSpace: ThickSpace$1,
	ThinSpace: ThinSpace$1,
	thinsp: thinsp$1,
	thkap: thkap$1,
	thksim: thksim$1,
	THORN: THORN$1$1,
	thorn: thorn$1$1,
	tilde: tilde$1,
	Tilde: Tilde$1,
	TildeEqual: TildeEqual$1,
	TildeFullEqual: TildeFullEqual$1,
	TildeTilde: TildeTilde$1,
	timesbar: timesbar$1,
	timesb: timesb$1,
	times: times$2,
	timesd: timesd$1,
	tint: tint$1,
	toea: toea$1,
	topbot: topbot$1,
	topcir: topcir$1,
	top: top$1,
	Topf: Topf$1,
	topf: topf$1,
	topfork: topfork$1,
	tosa: tosa$1,
	tprime: tprime$1,
	trade: trade$1,
	TRADE: TRADE$1,
	triangle: triangle$1,
	triangledown: triangledown$1,
	triangleleft: triangleleft$1,
	trianglelefteq: trianglelefteq$1,
	triangleq: triangleq$1,
	triangleright: triangleright$1,
	trianglerighteq: trianglerighteq$1,
	tridot: tridot$1,
	trie: trie$1,
	triminus: triminus$1,
	TripleDot: TripleDot$1,
	triplus: triplus$1,
	trisb: trisb$1,
	tritime: tritime$1,
	trpezium: trpezium$1,
	Tscr: Tscr$1,
	tscr: tscr$1,
	TScy: TScy$1,
	tscy: tscy$1,
	TSHcy: TSHcy$1,
	tshcy: tshcy$1,
	Tstrok: Tstrok$1,
	tstrok: tstrok$1,
	twixt: twixt$1,
	twoheadleftarrow: twoheadleftarrow$1,
	twoheadrightarrow: twoheadrightarrow$1,
	Uacute: Uacute$1$1,
	uacute: uacute$1$1,
	uarr: uarr$1,
	Uarr: Uarr$1,
	uArr: uArr$1,
	Uarrocir: Uarrocir$1,
	Ubrcy: Ubrcy$1,
	ubrcy: ubrcy$1,
	Ubreve: Ubreve$1,
	ubreve: ubreve$1,
	Ucirc: Ucirc$1$1,
	ucirc: ucirc$1$1,
	Ucy: Ucy$1,
	ucy: ucy$1,
	udarr: udarr$1,
	Udblac: Udblac$1,
	udblac: udblac$1,
	udhar: udhar$1,
	ufisht: ufisht$1,
	Ufr: Ufr$1,
	ufr: ufr$1,
	Ugrave: Ugrave$1$1,
	ugrave: ugrave$1$1,
	uHar: uHar$1,
	uharl: uharl$1,
	uharr: uharr$1,
	uhblk: uhblk$1,
	ulcorn: ulcorn$1,
	ulcorner: ulcorner$1,
	ulcrop: ulcrop$1,
	ultri: ultri$1,
	Umacr: Umacr$1,
	umacr: umacr$1,
	uml: uml$1$1,
	UnderBar: UnderBar$1,
	UnderBrace: UnderBrace$1,
	UnderBracket: UnderBracket$1,
	UnderParenthesis: UnderParenthesis$1,
	Union: Union$1,
	UnionPlus: UnionPlus$1,
	Uogon: Uogon$1,
	uogon: uogon$1,
	Uopf: Uopf$1,
	uopf: uopf$1,
	UpArrowBar: UpArrowBar$1,
	uparrow: uparrow$1,
	UpArrow: UpArrow$1,
	Uparrow: Uparrow$1,
	UpArrowDownArrow: UpArrowDownArrow$1,
	updownarrow: updownarrow$1,
	UpDownArrow: UpDownArrow$1,
	Updownarrow: Updownarrow$1,
	UpEquilibrium: UpEquilibrium$1,
	upharpoonleft: upharpoonleft$1,
	upharpoonright: upharpoonright$1,
	uplus: uplus$1,
	UpperLeftArrow: UpperLeftArrow$1,
	UpperRightArrow: UpperRightArrow$1,
	upsi: upsi$1,
	Upsi: Upsi$1,
	upsih: upsih$1,
	Upsilon: Upsilon$1,
	upsilon: upsilon$1,
	UpTeeArrow: UpTeeArrow$1,
	UpTee: UpTee$1,
	upuparrows: upuparrows$1,
	urcorn: urcorn$1,
	urcorner: urcorner$1,
	urcrop: urcrop$1,
	Uring: Uring$1,
	uring: uring$1,
	urtri: urtri$1,
	Uscr: Uscr$1,
	uscr: uscr$1,
	utdot: utdot$1,
	Utilde: Utilde$1,
	utilde: utilde$1,
	utri: utri$1,
	utrif: utrif$1,
	uuarr: uuarr$1,
	Uuml: Uuml$1$1,
	uuml: uuml$1$1,
	uwangle: uwangle$1,
	vangrt: vangrt$1,
	varepsilon: varepsilon$1,
	varkappa: varkappa$1,
	varnothing: varnothing$1,
	varphi: varphi$1,
	varpi: varpi$1,
	varpropto: varpropto$1,
	varr: varr$1,
	vArr: vArr$1,
	varrho: varrho$1,
	varsigma: varsigma$1,
	varsubsetneq: varsubsetneq$1,
	varsubsetneqq: varsubsetneqq$1,
	varsupsetneq: varsupsetneq$1,
	varsupsetneqq: varsupsetneqq$1,
	vartheta: vartheta$1,
	vartriangleleft: vartriangleleft$1,
	vartriangleright: vartriangleright$1,
	vBar: vBar$1,
	Vbar: Vbar$1,
	vBarv: vBarv$1,
	Vcy: Vcy$1,
	vcy: vcy$1,
	vdash: vdash$1,
	vDash: vDash$1,
	Vdash: Vdash$1,
	VDash: VDash$1,
	Vdashl: Vdashl$1,
	veebar: veebar$1,
	vee: vee$1,
	Vee: Vee$1,
	veeeq: veeeq$1,
	vellip: vellip$1,
	verbar: verbar$1,
	Verbar: Verbar$1,
	vert: vert$1,
	Vert: Vert$1,
	VerticalBar: VerticalBar$1,
	VerticalLine: VerticalLine$1,
	VerticalSeparator: VerticalSeparator$1,
	VerticalTilde: VerticalTilde$1,
	VeryThinSpace: VeryThinSpace$1,
	Vfr: Vfr$1,
	vfr: vfr$1,
	vltri: vltri$1,
	vnsub: vnsub$1,
	vnsup: vnsup$1,
	Vopf: Vopf$1,
	vopf: vopf$1,
	vprop: vprop$1,
	vrtri: vrtri$1,
	Vscr: Vscr$1,
	vscr: vscr$1,
	vsubnE: vsubnE$1,
	vsubne: vsubne$1,
	vsupnE: vsupnE$1,
	vsupne: vsupne$1,
	Vvdash: Vvdash$1,
	vzigzag: vzigzag$1,
	Wcirc: Wcirc$1,
	wcirc: wcirc$1,
	wedbar: wedbar$1,
	wedge: wedge$1,
	Wedge: Wedge$1,
	wedgeq: wedgeq$1,
	weierp: weierp$1,
	Wfr: Wfr$1,
	wfr: wfr$1,
	Wopf: Wopf$1,
	wopf: wopf$1,
	wp: wp$1,
	wr: wr$1,
	wreath: wreath$1,
	Wscr: Wscr$1,
	wscr: wscr$1,
	xcap: xcap$1,
	xcirc: xcirc$1,
	xcup: xcup$1,
	xdtri: xdtri$1,
	Xfr: Xfr$1,
	xfr: xfr$1,
	xharr: xharr$1,
	xhArr: xhArr$1,
	Xi: Xi$1,
	xi: xi$1,
	xlarr: xlarr$1,
	xlArr: xlArr$1,
	xmap: xmap$1,
	xnis: xnis$1,
	xodot: xodot$1,
	Xopf: Xopf$1,
	xopf: xopf$1,
	xoplus: xoplus$1,
	xotime: xotime$1,
	xrarr: xrarr$1,
	xrArr: xrArr$1,
	Xscr: Xscr$1,
	xscr: xscr$1,
	xsqcup: xsqcup$1,
	xuplus: xuplus$1,
	xutri: xutri$1,
	xvee: xvee$1,
	xwedge: xwedge$1,
	Yacute: Yacute$1$1,
	yacute: yacute$1$1,
	YAcy: YAcy$1,
	yacy: yacy$1,
	Ycirc: Ycirc$1,
	ycirc: ycirc$1,
	Ycy: Ycy$1,
	ycy: ycy$1,
	yen: yen$1$1,
	Yfr: Yfr$1,
	yfr: yfr$1,
	YIcy: YIcy$1,
	yicy: yicy$1,
	Yopf: Yopf$1,
	yopf: yopf$1,
	Yscr: Yscr$1,
	yscr: yscr$1,
	YUcy: YUcy$1,
	yucy: yucy$1,
	yuml: yuml$1$1,
	Yuml: Yuml$1,
	Zacute: Zacute$1,
	zacute: zacute$1,
	Zcaron: Zcaron$1,
	zcaron: zcaron$1,
	Zcy: Zcy$1,
	zcy: zcy$1,
	Zdot: Zdot$1,
	zdot: zdot$1,
	zeetrf: zeetrf$1,
	ZeroWidthSpace: ZeroWidthSpace$1,
	Zeta: Zeta$1,
	zeta: zeta$1,
	zfr: zfr$1,
	Zfr: Zfr$1,
	ZHcy: ZHcy$1,
	zhcy: zhcy$1,
	zigrarr: zigrarr$1,
	zopf: zopf$1,
	Zopf: Zopf$1,
	Zscr: Zscr$1,
	zscr: zscr$1,
	zwj: zwj$1,
	zwnj: zwnj$1,
	default: entitiesJSON$1
});

var require$$0$2 = ( xml$1 && xmlJSON$1 ) || xml$1;

var require$$1$1 = ( entities$2 && entitiesJSON$1 ) || entities$2;

var inverseXML$1 = getInverseObj$1(require$$0$2);
var xmlReplacer$1 = getInverseReplacer$1(inverseXML$1);

var XML$1 = getInverse$1(inverseXML$1, xmlReplacer$1);

var inverseHTML$1 = getInverseObj$1(require$$1$1);
var htmlReplacer$1 = getInverseReplacer$1(inverseHTML$1);

var HTML$1 = getInverse$1(inverseHTML$1, htmlReplacer$1);

function getInverseObj$1(obj){
	return Object.keys(obj).sort().reduce(function(inverse, name){
		inverse[obj[name]] = "&" + name + ";";
		return inverse;
	}, {});
}

function getInverseReplacer$1(inverse){
	var single = [],
	    multiple = [];

	Object.keys(inverse).forEach(function(k){
		if(k.length === 1){
			single.push("\\" + k);
		} else {
			multiple.push(k);
		}
	});

	
	multiple.unshift("[" + single.join("") + "]");

	return new RegExp(multiple.join("|"), "g");
}

var re_nonASCII$1 = /[^\0-\x7F]/g;
var re_astralSymbols$1 = /[\uD800-\uDBFF][\uDC00-\uDFFF]/g;

function singleCharReplacer$1(c){
	return "&#x" + c.charCodeAt(0).toString(16).toUpperCase() + ";";
}

function astralReplacer$1(c){
	
	var high = c.charCodeAt(0);
	var low  = c.charCodeAt(1);
	var codePoint = (high - 0xD800) * 0x400 + low - 0xDC00 + 0x10000;
	return "&#x" + codePoint.toString(16).toUpperCase() + ";";
}

function getInverse$1(inverse, re){
	function func(name){
		return inverse[name];
	}

	return function(data){
		return data
				.replace(re, func)
				.replace(re_astralSymbols$1, astralReplacer$1)
				.replace(re_nonASCII$1, singleCharReplacer$1);
	};
}

var re_xmlChars$1 = getInverseReplacer$1(inverseXML$1);

function escapeXML$1(data){
	return data
			.replace(re_xmlChars$1, singleCharReplacer$1)
			.replace(re_astralSymbols$1, astralReplacer$1)
			.replace(re_nonASCII$1, singleCharReplacer$1);
}

var escape$1 = escapeXML$1;

var encode$1 = {
	XML: XML$1,
	HTML: HTML$1,
	escape: escape$1
};

var decode$1$1 = {
	"0": 65533,
	"128": 8364,
	"130": 8218,
	"131": 402,
	"132": 8222,
	"133": 8230,
	"134": 8224,
	"135": 8225,
	"136": 710,
	"137": 8240,
	"138": 352,
	"139": 8249,
	"140": 338,
	"142": 381,
	"145": 8216,
	"146": 8217,
	"147": 8220,
	"148": 8221,
	"149": 8226,
	"150": 8211,
	"151": 8212,
	"152": 732,
	"153": 8482,
	"154": 353,
	"155": 8250,
	"156": 339,
	"158": 382,
	"159": 376
};

var decode$1$2 = Object.freeze({
	default: decode$1$1
});

var require$$0$1$2 = ( decode$1$2 && decode$1$1 ) || decode$1$2;

var decodeMap$1 = require$$0$1$2;

var decode_codepoint$1 = decodeCodePoint$1;


function decodeCodePoint$1(codePoint){

	if((codePoint >= 0xD800 && codePoint <= 0xDFFF) || codePoint > 0x10FFFF){
		return "\uFFFD";
	}

	if(codePoint in decodeMap$1){
		codePoint = decodeMap$1[codePoint];
	}

	var output = "";

	if(codePoint > 0xFFFF){
		codePoint -= 0x10000;
		output += String.fromCharCode(codePoint >>> 10 & 0x3FF | 0xD800);
		codePoint = 0xDC00 | codePoint & 0x3FF;
	}

	output += String.fromCharCode(codePoint);
	return output;
}

var Aacute$1$2 = "Á";
var aacute$1$2 = "á";
var Acirc$1$2 = "Â";
var acirc$1$2 = "â";
var acute$1$2 = "´";
var AElig$1$2 = "Æ";
var aelig$1$2 = "æ";
var Agrave$1$2 = "À";
var agrave$1$2 = "à";
var amp$2$1 = "&";
var AMP$1$2 = "&";
var Aring$1$2 = "Å";
var aring$1$2 = "å";
var Atilde$1$2 = "Ã";
var atilde$1$2 = "ã";
var Auml$1$2 = "Ä";
var auml$1$2 = "ä";
var brvbar$1$2 = "¦";
var Ccedil$1$2 = "Ç";
var ccedil$1$2 = "ç";
var cedil$1$2 = "¸";
var cent$1$2 = "¢";
var copy$1$2 = "©";
var COPY$1$2 = "©";
var curren$1$2 = "¤";
var deg$1$2 = "°";
var divide$1$2 = "÷";
var Eacute$1$2 = "É";
var eacute$1$2 = "é";
var Ecirc$1$2 = "Ê";
var ecirc$1$2 = "ê";
var Egrave$1$2 = "È";
var egrave$1$2 = "è";
var ETH$1$2 = "Ð";
var eth$1$2 = "ð";
var Euml$1$2 = "Ë";
var euml$1$2 = "ë";
var frac12$1$2 = "½";
var frac14$1$2 = "¼";
var frac34$1$2 = "¾";
var gt$2$1 = ">";
var GT$1$2 = ">";
var Iacute$1$2 = "Í";
var iacute$1$2 = "í";
var Icirc$1$2 = "Î";
var icirc$1$2 = "î";
var iexcl$1$2 = "¡";
var Igrave$1$2 = "Ì";
var igrave$1$2 = "ì";
var iquest$1$2 = "¿";
var Iuml$1$2 = "Ï";
var iuml$1$2 = "ï";
var laquo$1$2 = "«";
var lt$2$1 = "<";
var LT$1$2 = "<";
var macr$1$2 = "¯";
var micro$1$2 = "µ";
var middot$1$2 = "·";
var nbsp$1$2 = " ";
var not$1$2 = "¬";
var Ntilde$1$2 = "Ñ";
var ntilde$1$2 = "ñ";
var Oacute$1$2 = "Ó";
var oacute$1$2 = "ó";
var Ocirc$1$2 = "Ô";
var ocirc$1$2 = "ô";
var Ograve$1$2 = "Ò";
var ograve$1$2 = "ò";
var ordf$1$2 = "ª";
var ordm$1$2 = "º";
var Oslash$1$2 = "Ø";
var oslash$1$2 = "ø";
var Otilde$1$2 = "Õ";
var otilde$1$2 = "õ";
var Ouml$1$2 = "Ö";
var ouml$1$2 = "ö";
var para$1$2 = "¶";
var plusmn$1$2 = "±";
var pound$1$2 = "£";
var quot$2$1 = "\"";
var QUOT$1$2 = "\"";
var raquo$1$2 = "»";
var reg$1$2 = "®";
var REG$1$2 = "®";
var sect$1$2 = "§";
var shy$1$2 = "­";
var sup1$1$2 = "¹";
var sup2$1$2 = "²";
var sup3$1$2 = "³";
var szlig$1$2 = "ß";
var THORN$1$2 = "Þ";
var thorn$1$2 = "þ";
var times$1$2 = "×";
var Uacute$1$2 = "Ú";
var uacute$1$2 = "ú";
var Ucirc$1$2 = "Û";
var ucirc$1$2 = "û";
var Ugrave$1$2 = "Ù";
var ugrave$1$2 = "ù";
var uml$1$2 = "¨";
var Uuml$1$2 = "Ü";
var uuml$1$2 = "ü";
var Yacute$1$2 = "Ý";
var yacute$1$2 = "ý";
var yen$1$2 = "¥";
var yuml$1$2 = "ÿ";
var legacyJSON$1 = {
	Aacute: Aacute$1$2,
	aacute: aacute$1$2,
	Acirc: Acirc$1$2,
	acirc: acirc$1$2,
	acute: acute$1$2,
	AElig: AElig$1$2,
	aelig: aelig$1$2,
	Agrave: Agrave$1$2,
	agrave: agrave$1$2,
	amp: amp$2$1,
	AMP: AMP$1$2,
	Aring: Aring$1$2,
	aring: aring$1$2,
	Atilde: Atilde$1$2,
	atilde: atilde$1$2,
	Auml: Auml$1$2,
	auml: auml$1$2,
	brvbar: brvbar$1$2,
	Ccedil: Ccedil$1$2,
	ccedil: ccedil$1$2,
	cedil: cedil$1$2,
	cent: cent$1$2,
	copy: copy$1$2,
	COPY: COPY$1$2,
	curren: curren$1$2,
	deg: deg$1$2,
	divide: divide$1$2,
	Eacute: Eacute$1$2,
	eacute: eacute$1$2,
	Ecirc: Ecirc$1$2,
	ecirc: ecirc$1$2,
	Egrave: Egrave$1$2,
	egrave: egrave$1$2,
	ETH: ETH$1$2,
	eth: eth$1$2,
	Euml: Euml$1$2,
	euml: euml$1$2,
	frac12: frac12$1$2,
	frac14: frac14$1$2,
	frac34: frac34$1$2,
	gt: gt$2$1,
	GT: GT$1$2,
	Iacute: Iacute$1$2,
	iacute: iacute$1$2,
	Icirc: Icirc$1$2,
	icirc: icirc$1$2,
	iexcl: iexcl$1$2,
	Igrave: Igrave$1$2,
	igrave: igrave$1$2,
	iquest: iquest$1$2,
	Iuml: Iuml$1$2,
	iuml: iuml$1$2,
	laquo: laquo$1$2,
	lt: lt$2$1,
	LT: LT$1$2,
	macr: macr$1$2,
	micro: micro$1$2,
	middot: middot$1$2,
	nbsp: nbsp$1$2,
	not: not$1$2,
	Ntilde: Ntilde$1$2,
	ntilde: ntilde$1$2,
	Oacute: Oacute$1$2,
	oacute: oacute$1$2,
	Ocirc: Ocirc$1$2,
	ocirc: ocirc$1$2,
	Ograve: Ograve$1$2,
	ograve: ograve$1$2,
	ordf: ordf$1$2,
	ordm: ordm$1$2,
	Oslash: Oslash$1$2,
	oslash: oslash$1$2,
	Otilde: Otilde$1$2,
	otilde: otilde$1$2,
	Ouml: Ouml$1$2,
	ouml: ouml$1$2,
	para: para$1$2,
	plusmn: plusmn$1$2,
	pound: pound$1$2,
	quot: quot$2$1,
	QUOT: QUOT$1$2,
	raquo: raquo$1$2,
	reg: reg$1$2,
	REG: REG$1$2,
	sect: sect$1$2,
	shy: shy$1$2,
	sup1: sup1$1$2,
	sup2: sup2$1$2,
	sup3: sup3$1$2,
	szlig: szlig$1$2,
	THORN: THORN$1$2,
	thorn: thorn$1$2,
	times: times$1$2,
	Uacute: Uacute$1$2,
	uacute: uacute$1$2,
	Ucirc: Ucirc$1$2,
	ucirc: ucirc$1$2,
	Ugrave: Ugrave$1$2,
	ugrave: ugrave$1$2,
	uml: uml$1$2,
	Uuml: Uuml$1$2,
	uuml: uuml$1$2,
	Yacute: Yacute$1$2,
	yacute: yacute$1$2,
	yen: yen$1$2,
	yuml: yuml$1$2
};

var _entities$1 = {
  encodeXML: encode$1.XML,
  decodeCodepoint: decode_codepoint$1,
  entitiesJSON: entitiesJSON$1,
  legacyJSON: legacyJSON$1,
  xmlJSON: xmlJSON$1
};

var _entities_decodeCodepoint = _entities$1.decodeCodepoint;

var _entities_decodeCodepoint$1 = Object.freeze({
	default: _entities_decodeCodepoint
});

var _entities_entitiesJSON = _entities$1.entitiesJSON;


var _entities_entitiesJSON$1 = Object.freeze({
	default: _entities_entitiesJSON
});

var _entities_legacyJSON = _entities$1.legacyJSON;

var _entities_legacyJSON$1 = Object.freeze({
	default: _entities_legacyJSON
});

var _entities_xmlJSON = _entities$1.xmlJSON;

var _entities_xmlJSON$1 = Object.freeze({
	default: _entities_xmlJSON
});

var require$$0$1$1$1 = ( _entities_decodeCodepoint$1 && _entities_decodeCodepoint ) || _entities_decodeCodepoint$1;

var require$$1$1$1 = ( _entities_entitiesJSON$1 && _entities_entitiesJSON ) || _entities_entitiesJSON$1;

var require$$2 = ( _entities_legacyJSON$1 && _entities_legacyJSON ) || _entities_legacyJSON$1;

var require$$3 = ( _entities_xmlJSON$1 && _entities_xmlJSON ) || _entities_xmlJSON$1;

var Tokenizer_1 = Tokenizer$1;

var decodeCodePoint$1$1 = require$$0$1$1$1;
var entityMap = require$$1$1$1;
var legacyMap = require$$2;
var xmlMap    = require$$3;
var i = 0;
var TEXT                      = i++;
var BEFORE_TAG_NAME           = i++;
var IN_TAG_NAME               = i++;
var IN_SELF_CLOSING_TAG       = i++;
var BEFORE_CLOSING_TAG_NAME   = i++;
var IN_CLOSING_TAG_NAME       = i++;
var AFTER_CLOSING_TAG_NAME    = i++;
var BEFORE_ATTRIBUTE_NAME     = i++;
var IN_ATTRIBUTE_NAME         = i++;
var AFTER_ATTRIBUTE_NAME      = i++;
var BEFORE_ATTRIBUTE_VALUE    = i++;
var IN_ATTRIBUTE_VALUE_DQ     = i++;
var IN_ATTRIBUTE_VALUE_SQ     = i++;
var IN_ATTRIBUTE_VALUE_NQ     = i++;
var BEFORE_DECLARATION        = i++;
var IN_DECLARATION            = i++;
var IN_PROCESSING_INSTRUCTION = i++;
var BEFORE_COMMENT            = i++;
var IN_COMMENT                = i++;
var AFTER_COMMENT_1           = i++;
var AFTER_COMMENT_2           = i++;
var BEFORE_CDATA_1            = i++;
var BEFORE_CDATA_2            = i++;
var BEFORE_CDATA_3            = i++;
var BEFORE_CDATA_4            = i++;
var BEFORE_CDATA_5            = i++;
var BEFORE_CDATA_6            = i++;
var IN_CDATA                  = i++;
var AFTER_CDATA_1             = i++;
var AFTER_CDATA_2             = i++;
var BEFORE_SPECIAL            = i++;
var BEFORE_SPECIAL_END        = i++;
var BEFORE_SCRIPT_1           = i++;
var BEFORE_SCRIPT_2           = i++;
var BEFORE_SCRIPT_3           = i++;
var BEFORE_SCRIPT_4           = i++;
var BEFORE_SCRIPT_5           = i++;
var AFTER_SCRIPT_1            = i++;
var AFTER_SCRIPT_2            = i++;
var AFTER_SCRIPT_3            = i++;
var AFTER_SCRIPT_4            = i++;
var AFTER_SCRIPT_5            = i++;
var BEFORE_STYLE_1            = i++;
var BEFORE_STYLE_2            = i++;
var BEFORE_STYLE_3            = i++;
var BEFORE_STYLE_4            = i++;
var AFTER_STYLE_1             = i++;
var AFTER_STYLE_2             = i++;
var AFTER_STYLE_3             = i++;
var AFTER_STYLE_4             = i++;
var BEFORE_ENTITY             = i++;
var BEFORE_NUMERIC_ENTITY     = i++;
var IN_NAMED_ENTITY           = i++;
var IN_NUMERIC_ENTITY         = i++;
var IN_HEX_ENTITY             = i++;
var j = 0;
var SPECIAL_NONE              = j++;
var SPECIAL_SCRIPT            = j++;
var SPECIAL_STYLE             = j++;

function whitespace(c){
	return c === " " || c === "\n" || c === "\t" || c === "\f" || c === "\r";
}

function characterState(char, SUCCESS){
	return function(c){
		if(c === char) this._state = SUCCESS;
	};
}

function ifElseState(upper, SUCCESS, FAILURE){
	var lower = upper.toLowerCase();

	if(upper === lower){
		return function(c){
			if(c === lower){
				this._state = SUCCESS;
			} else {
				this._state = FAILURE;
				this._index--;
			}
		};
	} else {
		return function(c){
			if(c === lower || c === upper){
				this._state = SUCCESS;
			} else {
				this._state = FAILURE;
				this._index--;
			}
		};
	}
}

function consumeSpecialNameChar(upper, NEXT_STATE){
	var lower = upper.toLowerCase();

	return function(c){
		if(c === lower || c === upper){
			this._state = NEXT_STATE;
		} else {
			this._state = IN_TAG_NAME;
			this._index--; 
		}
	};
}

function Tokenizer$1(options, cbs){
	this._state = TEXT;
	this._buffer = "";
	this._sectionStart = 0;
	this._index = 0;
	this._bufferOffset = 0; 
	this._baseState = TEXT;
	this._special = SPECIAL_NONE;
	this._cbs = cbs;
	this._running = true;
	this._ended = false;
	this._xmlMode = !!(options && options.xmlMode);
	this._decodeEntities = !!(options && options.decodeEntities);
}

Tokenizer$1.prototype._stateText = function(c){
	if(c === "<"){
		if(this._index > this._sectionStart){
			this._cbs.ontext(this._getSection());
		}
		this._state = BEFORE_TAG_NAME;
		this._sectionStart = this._index;
	} else if(this._decodeEntities && this._special === SPECIAL_NONE && c === "&"){
		if(this._index > this._sectionStart){
			this._cbs.ontext(this._getSection());
		}
		this._baseState = TEXT;
		this._state = BEFORE_ENTITY;
		this._sectionStart = this._index;
	}
};

Tokenizer$1.prototype._stateBeforeTagName = function(c){
	if(c === "/"){
		this._state = BEFORE_CLOSING_TAG_NAME;
	} else if(c === "<"){
		this._cbs.ontext(this._getSection());
		this._sectionStart = this._index;
	} else if(c === ">" || this._special !== SPECIAL_NONE || whitespace(c)) {
		this._state = TEXT;
	} else if(c === "!"){
		this._state = BEFORE_DECLARATION;
		this._sectionStart = this._index + 1;
	} else if(c === "?"){
		this._state = IN_PROCESSING_INSTRUCTION;
		this._sectionStart = this._index + 1;
	} else {
		this._state = (!this._xmlMode && (c === "s" || c === "S")) ?
						BEFORE_SPECIAL : IN_TAG_NAME;
		this._sectionStart = this._index;
	}
};

Tokenizer$1.prototype._stateInTagName = function(c){
	if(c === "/" || c === ">" || whitespace(c)){
		this._emitToken("onopentagname");
		this._state = BEFORE_ATTRIBUTE_NAME;
		this._index--;
	}
};

Tokenizer$1.prototype._stateBeforeCloseingTagName = function(c){
	if(whitespace(c));
	else if(c === ">"){
		this._state = TEXT;
	} else if(this._special !== SPECIAL_NONE){
		if(c === "s" || c === "S"){
			this._state = BEFORE_SPECIAL_END;
		} else {
			this._state = TEXT;
			this._index--;
		}
	} else {
		this._state = IN_CLOSING_TAG_NAME;
		this._sectionStart = this._index;
	}
};

Tokenizer$1.prototype._stateInCloseingTagName = function(c){
	if(c === ">" || whitespace(c)){
		this._emitToken("onclosetag");
		this._state = AFTER_CLOSING_TAG_NAME;
		this._index--;
	}
};

Tokenizer$1.prototype._stateAfterCloseingTagName = function(c){
	
	if(c === ">"){
		this._state = TEXT;
		this._sectionStart = this._index + 1;
	}
};

Tokenizer$1.prototype._stateBeforeAttributeName = function(c){
	if(c === ">"){
		this._cbs.onopentagend();
		this._state = TEXT;
		this._sectionStart = this._index + 1;
	} else if(c === "/"){
		this._state = IN_SELF_CLOSING_TAG;
	} else if(!whitespace(c)){
		this._state = IN_ATTRIBUTE_NAME;
		this._sectionStart = this._index;
	}
};

Tokenizer$1.prototype._stateInSelfClosingTag = function(c){
	if(c === ">"){
		this._cbs.onselfclosingtag();
		this._state = TEXT;
		this._sectionStart = this._index + 1;
	} else if(!whitespace(c)){
		this._state = BEFORE_ATTRIBUTE_NAME;
		this._index--;
	}
};

Tokenizer$1.prototype._stateInAttributeName = function(c){
	if(c === "=" || c === "/" || c === ">" || whitespace(c)){
		this._cbs.onattribname(this._getSection());
		this._sectionStart = -1;
		this._state = AFTER_ATTRIBUTE_NAME;
		this._index--;
	}
};

Tokenizer$1.prototype._stateAfterAttributeName = function(c){
	if(c === "="){
		this._state = BEFORE_ATTRIBUTE_VALUE;
	} else if(c === "/" || c === ">"){
		this._cbs.onattribend();
		this._state = BEFORE_ATTRIBUTE_NAME;
		this._index--;
	} else if(!whitespace(c)){
		this._cbs.onattribend();
		this._state = IN_ATTRIBUTE_NAME;
		this._sectionStart = this._index;
	}
};

Tokenizer$1.prototype._stateBeforeAttributeValue = function(c){
	if(c === "\""){
		this._state = IN_ATTRIBUTE_VALUE_DQ;
		this._sectionStart = this._index + 1;
	} else if(c === "'"){
		this._state = IN_ATTRIBUTE_VALUE_SQ;
		this._sectionStart = this._index + 1;
	} else if(!whitespace(c)){
		this._state = IN_ATTRIBUTE_VALUE_NQ;
		this._sectionStart = this._index;
		this._index--; 
	}
};

Tokenizer$1.prototype._stateInAttributeValueDoubleQuotes = function(c){
	if(c === "\""){
		this._emitToken("onattribdata");
		this._cbs.onattribend();
		this._state = BEFORE_ATTRIBUTE_NAME;
	} else if(this._decodeEntities && c === "&"){
		this._emitToken("onattribdata");
		this._baseState = this._state;
		this._state = BEFORE_ENTITY;
		this._sectionStart = this._index;
	}
};

Tokenizer$1.prototype._stateInAttributeValueSingleQuotes = function(c){
	if(c === "'"){
		this._emitToken("onattribdata");
		this._cbs.onattribend();
		this._state = BEFORE_ATTRIBUTE_NAME;
	} else if(this._decodeEntities && c === "&"){
		this._emitToken("onattribdata");
		this._baseState = this._state;
		this._state = BEFORE_ENTITY;
		this._sectionStart = this._index;
	}
};

Tokenizer$1.prototype._stateInAttributeValueNoQuotes = function(c){
	if(whitespace(c) || c === ">"){
		this._emitToken("onattribdata");
		this._cbs.onattribend();
		this._state = BEFORE_ATTRIBUTE_NAME;
		this._index--;
	} else if(this._decodeEntities && c === "&"){
		this._emitToken("onattribdata");
		this._baseState = this._state;
		this._state = BEFORE_ENTITY;
		this._sectionStart = this._index;
	}
};

Tokenizer$1.prototype._stateBeforeDeclaration = function(c){
	this._state = c === "[" ? BEFORE_CDATA_1 :
					c === "-" ? BEFORE_COMMENT :
						IN_DECLARATION;
};

Tokenizer$1.prototype._stateInDeclaration = function(c){
	if(c === ">"){
		this._cbs.ondeclaration(this._getSection());
		this._state = TEXT;
		this._sectionStart = this._index + 1;
	}
};

Tokenizer$1.prototype._stateInProcessingInstruction = function(c){
	if(c === ">"){
		this._cbs.onprocessinginstruction(this._getSection());
		this._state = TEXT;
		this._sectionStart = this._index + 1;
	}
};

Tokenizer$1.prototype._stateBeforeComment = function(c){
	if(c === "-"){
		this._state = IN_COMMENT;
		this._sectionStart = this._index + 1;
	} else {
		this._state = IN_DECLARATION;
	}
};

Tokenizer$1.prototype._stateInComment = function(c){
	if(c === "-") this._state = AFTER_COMMENT_1;
};

Tokenizer$1.prototype._stateAfterComment1 = function(c){
	if(c === "-"){
		this._state = AFTER_COMMENT_2;
	} else {
		this._state = IN_COMMENT;
	}
};

Tokenizer$1.prototype._stateAfterComment2 = function(c){
	if(c === ">"){
		
		this._cbs.oncomment(this._buffer.substring(this._sectionStart, this._index - 2));
		this._state = TEXT;
		this._sectionStart = this._index + 1;
	} else if(c !== "-"){
		this._state = IN_COMMENT;
	}
	
};

Tokenizer$1.prototype._stateBeforeCdata1 = ifElseState("C", BEFORE_CDATA_2, IN_DECLARATION);
Tokenizer$1.prototype._stateBeforeCdata2 = ifElseState("D", BEFORE_CDATA_3, IN_DECLARATION);
Tokenizer$1.prototype._stateBeforeCdata3 = ifElseState("A", BEFORE_CDATA_4, IN_DECLARATION);
Tokenizer$1.prototype._stateBeforeCdata4 = ifElseState("T", BEFORE_CDATA_5, IN_DECLARATION);
Tokenizer$1.prototype._stateBeforeCdata5 = ifElseState("A", BEFORE_CDATA_6, IN_DECLARATION);

Tokenizer$1.prototype._stateBeforeCdata6 = function(c){
	if(c === "["){
		this._state = IN_CDATA;
		this._sectionStart = this._index + 1;
	} else {
		this._state = IN_DECLARATION;
		this._index--;
	}
};

Tokenizer$1.prototype._stateInCdata = function(c){
	if(c === "]") this._state = AFTER_CDATA_1;
};

Tokenizer$1.prototype._stateAfterCdata1 = characterState("]", AFTER_CDATA_2);

Tokenizer$1.prototype._stateAfterCdata2 = function(c){
	if(c === ">"){
		
		this._cbs.oncdata(this._buffer.substring(this._sectionStart, this._index - 2));
		this._state = TEXT;
		this._sectionStart = this._index + 1;
	} else if(c !== "]") {
		this._state = IN_CDATA;
	}
	
};

Tokenizer$1.prototype._stateBeforeSpecial = function(c){
	if(c === "c" || c === "C"){
		this._state = BEFORE_SCRIPT_1;
	} else if(c === "t" || c === "T"){
		this._state = BEFORE_STYLE_1;
	} else {
		this._state = IN_TAG_NAME;
		this._index--; 
	}
};

Tokenizer$1.prototype._stateBeforeSpecialEnd = function(c){
	if(this._special === SPECIAL_SCRIPT && (c === "c" || c === "C")){
		this._state = AFTER_SCRIPT_1;
	} else if(this._special === SPECIAL_STYLE && (c === "t" || c === "T")){
		this._state = AFTER_STYLE_1;
	}
	else this._state = TEXT;
};

Tokenizer$1.prototype._stateBeforeScript1 = consumeSpecialNameChar("R", BEFORE_SCRIPT_2);
Tokenizer$1.prototype._stateBeforeScript2 = consumeSpecialNameChar("I", BEFORE_SCRIPT_3);
Tokenizer$1.prototype._stateBeforeScript3 = consumeSpecialNameChar("P", BEFORE_SCRIPT_4);
Tokenizer$1.prototype._stateBeforeScript4 = consumeSpecialNameChar("T", BEFORE_SCRIPT_5);

Tokenizer$1.prototype._stateBeforeScript5 = function(c){
	if(c === "/" || c === ">" || whitespace(c)){
		this._special = SPECIAL_SCRIPT;
	}
	this._state = IN_TAG_NAME;
	this._index--; 
};

Tokenizer$1.prototype._stateAfterScript1 = ifElseState("R", AFTER_SCRIPT_2, TEXT);
Tokenizer$1.prototype._stateAfterScript2 = ifElseState("I", AFTER_SCRIPT_3, TEXT);
Tokenizer$1.prototype._stateAfterScript3 = ifElseState("P", AFTER_SCRIPT_4, TEXT);
Tokenizer$1.prototype._stateAfterScript4 = ifElseState("T", AFTER_SCRIPT_5, TEXT);

Tokenizer$1.prototype._stateAfterScript5 = function(c){
	if(c === ">" || whitespace(c)){
		this._special = SPECIAL_NONE;
		this._state = IN_CLOSING_TAG_NAME;
		this._sectionStart = this._index - 6;
		this._index--; 
	}
	else this._state = TEXT;
};

Tokenizer$1.prototype._stateBeforeStyle1 = consumeSpecialNameChar("Y", BEFORE_STYLE_2);
Tokenizer$1.prototype._stateBeforeStyle2 = consumeSpecialNameChar("L", BEFORE_STYLE_3);
Tokenizer$1.prototype._stateBeforeStyle3 = consumeSpecialNameChar("E", BEFORE_STYLE_4);

Tokenizer$1.prototype._stateBeforeStyle4 = function(c){
	if(c === "/" || c === ">" || whitespace(c)){
		this._special = SPECIAL_STYLE;
	}
	this._state = IN_TAG_NAME;
	this._index--; 
};

Tokenizer$1.prototype._stateAfterStyle1 = ifElseState("Y", AFTER_STYLE_2, TEXT);
Tokenizer$1.prototype._stateAfterStyle2 = ifElseState("L", AFTER_STYLE_3, TEXT);
Tokenizer$1.prototype._stateAfterStyle3 = ifElseState("E", AFTER_STYLE_4, TEXT);

Tokenizer$1.prototype._stateAfterStyle4 = function(c){
	if(c === ">" || whitespace(c)){
		this._special = SPECIAL_NONE;
		this._state = IN_CLOSING_TAG_NAME;
		this._sectionStart = this._index - 5;
		this._index--; 
	}
	else this._state = TEXT;
};

Tokenizer$1.prototype._stateBeforeEntity = ifElseState("#", BEFORE_NUMERIC_ENTITY, IN_NAMED_ENTITY);
Tokenizer$1.prototype._stateBeforeNumericEntity = ifElseState("X", IN_HEX_ENTITY, IN_NUMERIC_ENTITY);


Tokenizer$1.prototype._parseNamedEntityStrict = function(){
	
	if(this._sectionStart + 1 < this._index){
		var entity = this._buffer.substring(this._sectionStart + 1, this._index),
		    map = this._xmlMode ? xmlMap : entityMap;

		if(map.hasOwnProperty(entity)){
			this._emitPartial(map[entity]);
			this._sectionStart = this._index + 1;
		}
	}
};



Tokenizer$1.prototype._parseLegacyEntity = function(){
	var start = this._sectionStart + 1,
	    limit = this._index - start;

	if(limit > 6) limit = 6; 

	while(limit >= 2){ 
		var entity = this._buffer.substr(start, limit);

		if(legacyMap.hasOwnProperty(entity)){
			this._emitPartial(legacyMap[entity]);
			this._sectionStart += limit + 1;
			return;
		} else {
			limit--;
		}
	}
};

Tokenizer$1.prototype._stateInNamedEntity = function(c){
	if(c === ";"){
		this._parseNamedEntityStrict();
		if(this._sectionStart + 1 < this._index && !this._xmlMode){
			this._parseLegacyEntity();
		}
		this._state = this._baseState;
	} else if((c < "a" || c > "z") && (c < "A" || c > "Z") && (c < "0" || c > "9")){
		if(this._xmlMode);
		else if(this._sectionStart + 1 === this._index);
		else if(this._baseState !== TEXT){
			if(c !== "="){
				this._parseNamedEntityStrict();
			}
		} else {
			this._parseLegacyEntity();
		}

		this._state = this._baseState;
		this._index--;
	}
};

Tokenizer$1.prototype._decodeNumericEntity = function(offset, base){
	var sectionStart = this._sectionStart + offset;

	if(sectionStart !== this._index){
		
		var entity = this._buffer.substring(sectionStart, this._index);
		var parsed = parseInt(entity, base);

		this._emitPartial(decodeCodePoint$1$1(parsed));
		this._sectionStart = this._index;
	} else {
		this._sectionStart--;
	}

	this._state = this._baseState;
};

Tokenizer$1.prototype._stateInNumericEntity = function(c){
	if(c === ";"){
		this._decodeNumericEntity(2, 10);
		this._sectionStart++;
	} else if(c < "0" || c > "9"){
		if(!this._xmlMode){
			this._decodeNumericEntity(2, 10);
		} else {
			this._state = this._baseState;
		}
		this._index--;
	}
};

Tokenizer$1.prototype._stateInHexEntity = function(c){
	if(c === ";"){
		this._decodeNumericEntity(3, 16);
		this._sectionStart++;
	} else if((c < "a" || c > "f") && (c < "A" || c > "F") && (c < "0" || c > "9")){
		if(!this._xmlMode){
			this._decodeNumericEntity(3, 16);
		} else {
			this._state = this._baseState;
		}
		this._index--;
	}
};

Tokenizer$1.prototype._cleanup = function (){
	if(this._sectionStart < 0){
		this._buffer = "";
		this._bufferOffset += this._index;
		this._index = 0;
	} else if(this._running){
		if(this._state === TEXT){
			if(this._sectionStart !== this._index){
				this._cbs.ontext(this._buffer.substr(this._sectionStart));
			}
			this._buffer = "";
			this._bufferOffset += this._index;
			this._index = 0;
		} else if(this._sectionStart === this._index){
			
			this._buffer = "";
			this._bufferOffset += this._index;
			this._index = 0;
		} else {
			
			this._buffer = this._buffer.substr(this._sectionStart);
			this._index -= this._sectionStart;
			this._bufferOffset += this._sectionStart;
		}

		this._sectionStart = 0;
	}
};


Tokenizer$1.prototype.write = function(chunk){
	if(this._ended) this._cbs.onerror(Error(".write() after done!"));

	this._buffer += chunk;
	this._parse();
};

Tokenizer$1.prototype._parse = function(){
	while(this._index < this._buffer.length && this._running){
		var c = this._buffer.charAt(this._index);
		if(this._state === TEXT) {
			this._stateText(c);
		} else if(this._state === BEFORE_TAG_NAME){
			this._stateBeforeTagName(c);
		} else if(this._state === IN_TAG_NAME) {
			this._stateInTagName(c);
		} else if(this._state === BEFORE_CLOSING_TAG_NAME){
			this._stateBeforeCloseingTagName(c);
		} else if(this._state === IN_CLOSING_TAG_NAME){
			this._stateInCloseingTagName(c);
		} else if(this._state === AFTER_CLOSING_TAG_NAME){
			this._stateAfterCloseingTagName(c);
		} else if(this._state === IN_SELF_CLOSING_TAG){
			this._stateInSelfClosingTag(c);
		}

		
		else if(this._state === BEFORE_ATTRIBUTE_NAME){
			this._stateBeforeAttributeName(c);
		} else if(this._state === IN_ATTRIBUTE_NAME){
			this._stateInAttributeName(c);
		} else if(this._state === AFTER_ATTRIBUTE_NAME){
			this._stateAfterAttributeName(c);
		} else if(this._state === BEFORE_ATTRIBUTE_VALUE){
			this._stateBeforeAttributeValue(c);
		} else if(this._state === IN_ATTRIBUTE_VALUE_DQ){
			this._stateInAttributeValueDoubleQuotes(c);
		} else if(this._state === IN_ATTRIBUTE_VALUE_SQ){
			this._stateInAttributeValueSingleQuotes(c);
		} else if(this._state === IN_ATTRIBUTE_VALUE_NQ){
			this._stateInAttributeValueNoQuotes(c);
		}

		
		else if(this._state === BEFORE_DECLARATION){
			this._stateBeforeDeclaration(c);
		} else if(this._state === IN_DECLARATION){
			this._stateInDeclaration(c);
		}

		
		else if(this._state === IN_PROCESSING_INSTRUCTION){
			this._stateInProcessingInstruction(c);
		}

		
		else if(this._state === BEFORE_COMMENT){
			this._stateBeforeComment(c);
		} else if(this._state === IN_COMMENT){
			this._stateInComment(c);
		} else if(this._state === AFTER_COMMENT_1){
			this._stateAfterComment1(c);
		} else if(this._state === AFTER_COMMENT_2){
			this._stateAfterComment2(c);
		}

		
		else if(this._state === BEFORE_CDATA_1){
			this._stateBeforeCdata1(c);
		} else if(this._state === BEFORE_CDATA_2){
			this._stateBeforeCdata2(c);
		} else if(this._state === BEFORE_CDATA_3){
			this._stateBeforeCdata3(c);
		} else if(this._state === BEFORE_CDATA_4){
			this._stateBeforeCdata4(c);
		} else if(this._state === BEFORE_CDATA_5){
			this._stateBeforeCdata5(c);
		} else if(this._state === BEFORE_CDATA_6){
			this._stateBeforeCdata6(c);
		} else if(this._state === IN_CDATA){
			this._stateInCdata(c);
		} else if(this._state === AFTER_CDATA_1){
			this._stateAfterCdata1(c);
		} else if(this._state === AFTER_CDATA_2){
			this._stateAfterCdata2(c);
		}

		
		else if(this._state === BEFORE_SPECIAL){
			this._stateBeforeSpecial(c);
		} else if(this._state === BEFORE_SPECIAL_END){
			this._stateBeforeSpecialEnd(c);
		}

		
		else if(this._state === BEFORE_SCRIPT_1){
			this._stateBeforeScript1(c);
		} else if(this._state === BEFORE_SCRIPT_2){
			this._stateBeforeScript2(c);
		} else if(this._state === BEFORE_SCRIPT_3){
			this._stateBeforeScript3(c);
		} else if(this._state === BEFORE_SCRIPT_4){
			this._stateBeforeScript4(c);
		} else if(this._state === BEFORE_SCRIPT_5){
			this._stateBeforeScript5(c);
		}

		else if(this._state === AFTER_SCRIPT_1){
			this._stateAfterScript1(c);
		} else if(this._state === AFTER_SCRIPT_2){
			this._stateAfterScript2(c);
		} else if(this._state === AFTER_SCRIPT_3){
			this._stateAfterScript3(c);
		} else if(this._state === AFTER_SCRIPT_4){
			this._stateAfterScript4(c);
		} else if(this._state === AFTER_SCRIPT_5){
			this._stateAfterScript5(c);
		}

		
		else if(this._state === BEFORE_STYLE_1){
			this._stateBeforeStyle1(c);
		} else if(this._state === BEFORE_STYLE_2){
			this._stateBeforeStyle2(c);
		} else if(this._state === BEFORE_STYLE_3){
			this._stateBeforeStyle3(c);
		} else if(this._state === BEFORE_STYLE_4){
			this._stateBeforeStyle4(c);
		}

		else if(this._state === AFTER_STYLE_1){
			this._stateAfterStyle1(c);
		} else if(this._state === AFTER_STYLE_2){
			this._stateAfterStyle2(c);
		} else if(this._state === AFTER_STYLE_3){
			this._stateAfterStyle3(c);
		} else if(this._state === AFTER_STYLE_4){
			this._stateAfterStyle4(c);
		}

		
		else if(this._state === BEFORE_ENTITY){
			this._stateBeforeEntity(c);
		} else if(this._state === BEFORE_NUMERIC_ENTITY){
			this._stateBeforeNumericEntity(c);
		} else if(this._state === IN_NAMED_ENTITY){
			this._stateInNamedEntity(c);
		} else if(this._state === IN_NUMERIC_ENTITY){
			this._stateInNumericEntity(c);
		} else if(this._state === IN_HEX_ENTITY){
			this._stateInHexEntity(c);
		}

		else {
			this._cbs.onerror(Error("unknown _state"), this._state);
		}

		this._index++;
	}

	this._cleanup();
};

Tokenizer$1.prototype.pause = function(){
	this._running = false;
};
Tokenizer$1.prototype.resume = function(){
	this._running = true;

	if(this._index < this._buffer.length){
		this._parse();
	}
	if(this._ended){
		this._finish();
	}
};

Tokenizer$1.prototype.end = function(chunk){
	if(this._ended) this._cbs.onerror(Error(".end() after done!"));
	if(chunk) this.write(chunk);

	this._ended = true;

	if(this._running) this._finish();
};

Tokenizer$1.prototype._finish = function(){
	
	if(this._sectionStart < this._index){
		this._handleTrailingData();
	}

	this._cbs.onend();
};

Tokenizer$1.prototype._handleTrailingData = function(){
	var data = this._buffer.substr(this._sectionStart);

	if(this._state === IN_CDATA || this._state === AFTER_CDATA_1 || this._state === AFTER_CDATA_2){
		this._cbs.oncdata(data);
	} else if(this._state === IN_COMMENT || this._state === AFTER_COMMENT_1 || this._state === AFTER_COMMENT_2){
		this._cbs.oncomment(data);
	} else if(this._state === IN_NAMED_ENTITY && !this._xmlMode){
		this._parseLegacyEntity();
		if(this._sectionStart < this._index){
			this._state = this._baseState;
			this._handleTrailingData();
		}
	} else if(this._state === IN_NUMERIC_ENTITY && !this._xmlMode){
		this._decodeNumericEntity(2, 10);
		if(this._sectionStart < this._index){
			this._state = this._baseState;
			this._handleTrailingData();
		}
	} else if(this._state === IN_HEX_ENTITY && !this._xmlMode){
		this._decodeNumericEntity(3, 16);
		if(this._sectionStart < this._index){
			this._state = this._baseState;
			this._handleTrailingData();
		}
	} else if(
		this._state !== IN_TAG_NAME &&
		this._state !== BEFORE_ATTRIBUTE_NAME &&
		this._state !== BEFORE_ATTRIBUTE_VALUE &&
		this._state !== AFTER_ATTRIBUTE_NAME &&
		this._state !== IN_ATTRIBUTE_NAME &&
		this._state !== IN_ATTRIBUTE_VALUE_SQ &&
		this._state !== IN_ATTRIBUTE_VALUE_DQ &&
		this._state !== IN_ATTRIBUTE_VALUE_NQ &&
		this._state !== IN_CLOSING_TAG_NAME
	){
		this._cbs.ontext(data);
	}
	
	
};

Tokenizer$1.prototype.reset = function(){
	Tokenizer$1.call(this, {xmlMode: this._xmlMode, decodeEntities: this._decodeEntities}, this._cbs);
};

Tokenizer$1.prototype.getAbsoluteIndex = function(){
	return this._bufferOffset + this._index;
};

Tokenizer$1.prototype._getSection = function(){
	return this._buffer.substring(this._sectionStart, this._index);
};

Tokenizer$1.prototype._emitToken = function(name){
	this._cbs[name](this._getSection());
	this._sectionStart = -1;
};

Tokenizer$1.prototype._emitPartial = function(value){
	if(this._baseState !== TEXT){
		this._cbs.onattribdata(value); 
	} else {
		this._cbs.ontext(value);
	}
};

var _inherits = function() {};

var _inherits$1 = Object.freeze({
	default: _inherits
});

var _resolve_empty = {};

var _resolve_empty$1 = Object.freeze({
	default: _resolve_empty
});

var require$$1$2 = ( _inherits$1 && _inherits ) || _inherits$1;

var require$$2$1 = ( _resolve_empty$1 && _resolve_empty ) || _resolve_empty$1;

var Tokenizer$1$1 = Tokenizer_1;





var formTags = {
	input: true,
	option: true,
	optgroup: true,
	select: true,
	button: true,
	datalist: true,
	textarea: true
};

var openImpliesClose = {
	tr      : { tr:true, th:true, td:true },
	th      : { th:true },
	td      : { thead:true, th:true, td:true },
	body    : { head:true, link:true, script:true },
	li      : { li:true },
	p       : { p:true },
	h1      : { p:true },
	h2      : { p:true },
	h3      : { p:true },
	h4      : { p:true },
	h5      : { p:true },
	h6      : { p:true },
	select  : formTags,
	input   : formTags,
	output  : formTags,
	button  : formTags,
	datalist: formTags,
	textarea: formTags,
	option  : { option:true },
	optgroup: { optgroup:true }
};

var voidElements = {
	__proto__: null,
	area: true,
	base: true,
	basefont: true,
	br: true,
	col: true,
	command: true,
	embed: true,
	frame: true,
	hr: true,
	img: true,
	input: true,
	isindex: true,
	keygen: true,
	link: true,
	meta: true,
	param: true,
	source: true,
	track: true,
	wbr: true,

	
	path: true,
	circle: true,
	ellipse: true,
	line: true,
	rect: true,
	use: true,
	stop: true,
	polyline: true,
	polygon: true
};

var re_nameEnd = /\s|\//;

function Parser$1(cbs, options){
	this._options = options || {};
	this._cbs = cbs || {};

	this._tagname = "";
	this._attribname = "";
	this._attribvalue = "";
	this._attribs = null;
	this._stack = [];

	this.startIndex = 0;
	this.endIndex = null;

	this._lowerCaseTagNames = "lowerCaseTags" in this._options ?
									!!this._options.lowerCaseTags :
									!this._options.xmlMode;
	this._lowerCaseAttributeNames = "lowerCaseAttributeNames" in this._options ?
									!!this._options.lowerCaseAttributeNames :
									!this._options.xmlMode;

	if(this._options.Tokenizer) {
		Tokenizer$1$1 = this._options.Tokenizer;
	}
	this._tokenizer = new Tokenizer$1$1(this._options, this);

	if(this._cbs.onparserinit) this._cbs.onparserinit(this);
}

require$$1$2(Parser$1, require$$2$1.EventEmitter);

Parser$1.prototype._updatePosition = function(initialOffset){
	if(this.endIndex === null){
		if(this._tokenizer._sectionStart <= initialOffset){
			this.startIndex = 0;
		} else {
			this.startIndex = this._tokenizer._sectionStart - initialOffset;
		}
	}
	else this.startIndex = this.endIndex + 1;
	this.endIndex = this._tokenizer.getAbsoluteIndex();
};


Parser$1.prototype.ontext = function(data){
	this._updatePosition(1);
	this.endIndex--;

	if(this._cbs.ontext) this._cbs.ontext(data);
};

Parser$1.prototype.onopentagname = function(name){
	if(this._lowerCaseTagNames){
		name = name.toLowerCase();
	}

	this._tagname = name;

	if(!this._options.xmlMode && name in openImpliesClose) {
		for(
			var el;
			(el = this._stack[this._stack.length - 1]) in openImpliesClose[name];
			this.onclosetag(el)
		);
	}

	if(this._options.xmlMode || !(name in voidElements)){
		this._stack.push(name);
	}

	if(this._cbs.onopentagname) this._cbs.onopentagname(name);
	if(this._cbs.onopentag) this._attribs = {};
};

Parser$1.prototype.onopentagend = function(){
	this._updatePosition(1);

	if(this._attribs){
		if(this._cbs.onopentag) this._cbs.onopentag(this._tagname, this._attribs);
		this._attribs = null;
	}

	if(!this._options.xmlMode && this._cbs.onclosetag && this._tagname in voidElements){
		this._cbs.onclosetag(this._tagname);
	}

	this._tagname = "";
};

Parser$1.prototype.onclosetag = function(name){
	this._updatePosition(1);
	if(this._lowerCaseTagNames){
		name = name.toLowerCase();
	}
	
	
	
	if(this._options.xmlMode) {
		const stack = this._stack;
		let last = stack.pop();
		while(last !== name) {
			if(this._cbs.onerror) {
				this._cbs.onerror("Unclosed tag <"+last+">");
			}
			last = stack.pop();
		}
		this.onopentagend();
		if(this._cbs.onclosetag) {
			this._cbs.onclosetag(last);
		}
	} else {
		if(this._stack.length && (!(name in voidElements))) {
			let pos = this._stack.lastIndexOf(name);
			if(pos !== -1){
				if(this._cbs.onclosetag){
					pos = this._stack.length - pos;
					while(pos--) this._cbs.onclosetag(this._stack.pop());
				}
				else this._stack.length = pos;
			} else if(name === "p"){
				this.onopentagname(name);
				this._closeCurrentTag();
			}
		} else if(name === "br" || name === "p"){
			this.onopentagname(name);
			this._closeCurrentTag();
		}
	}
};

Parser$1.prototype.onselfclosingtag = function(){
	if(this._options.xmlMode || this._options.recognizeSelfClosing){
		this._closeCurrentTag();
	} else {
		this.onopentagend();
	}
};

Parser$1.prototype._closeCurrentTag = function(){
	var name = this._tagname;

	this.onopentagend();

	
	
	if(this._stack[this._stack.length - 1] === name){
		if(this._cbs.onclosetag){
			this._cbs.onclosetag(name);
		}
		this._stack.pop();
	}
};

Parser$1.prototype.onattribname = function(name){
	if(this._lowerCaseAttributeNames){
		name = name.toLowerCase();
	}
	this._attribname = name;
};

Parser$1.prototype.onattribdata = function(value){
	this._attribvalue += value;
};

Parser$1.prototype.onattribend = function(){
	if(this._cbs.onattribute) this._cbs.onattribute(this._attribname, this._attribvalue);
	if(
		this._attribs &&
		!Object.prototype.hasOwnProperty.call(this._attribs, this._attribname)
	){
		this._attribs[this._attribname] = this._attribvalue;
	}
	this._attribname = "";
	this._attribvalue = "";
};

Parser$1.prototype._getInstructionName = function(value){
	var idx = value.search(re_nameEnd),
	    name = idx < 0 ? value : value.substr(0, idx);

	if(this._lowerCaseTagNames){
		name = name.toLowerCase();
	}

	return name;
};

Parser$1.prototype.ondeclaration = function(value){
	if(this._cbs.onprocessinginstruction){
		var name = this._getInstructionName(value);
		this._cbs.onprocessinginstruction("!" + name, "!" + value);
	}
};

Parser$1.prototype.onprocessinginstruction = function(value){
	if(this._cbs.onprocessinginstruction){
		var name = this._getInstructionName(value);
		this._cbs.onprocessinginstruction("?" + name, "?" + value);
	}
};

Parser$1.prototype.oncomment = function(value){
	this._updatePosition(4);

	if(this._cbs.oncomment) this._cbs.oncomment(value);
	if(this._cbs.oncommentend) this._cbs.oncommentend();
};

Parser$1.prototype.oncdata = function(value){
	this._updatePosition(1);

	if(this._options.xmlMode || this._options.recognizeCDATA){
		if(this._cbs.oncdatastart) this._cbs.oncdatastart();
		if(this._cbs.ontext) this._cbs.ontext(value);
		if(this._cbs.oncdataend) this._cbs.oncdataend();
	} else {
		this.oncomment("[CDATA[" + value + "]]");
	}
};

Parser$1.prototype.onerror = function(err){
	if(this._cbs.onerror) this._cbs.onerror(err);
};

Parser$1.prototype.onend = function(){
	if(this._cbs.onclosetag){
		for(
			var i = this._stack.length;
			i > 0;
			this._cbs.onclosetag(this._stack[--i])
		);
	}
	if(this._cbs.onend) this._cbs.onend();
};



Parser$1.prototype.reset = function(){
	if(this._cbs.onreset) this._cbs.onreset();
	this._tokenizer.reset();

	this._tagname = "";
	this._attribname = "";
	this._attribs = null;
	this._stack = [];

	if(this._cbs.onparserinit) this._cbs.onparserinit(this);
};


Parser$1.prototype.parseComplete = function(data){
	this.reset();
	this.end(data);
};

Parser$1.prototype.write = function(chunk){
	this._tokenizer.write(chunk);
};

Parser$1.prototype.end = function(chunk){
	this._tokenizer.end(chunk);
};

Parser$1.prototype.pause = function(){
	this._tokenizer.pause();
};

Parser$1.prototype.resume = function(){
	this._tokenizer.resume();
};


Parser$1.prototype.parseChunk = Parser$1.prototype.write;
Parser$1.prototype.done = Parser$1.prototype.end;

var Parser_1 = Parser$1;

Parser_1.prototype.oncdata = function(value){
  this._updatePosition(1);

  if(this._options.xmlMode || this._options.recognizeCDATA){
    if(this._cbs.oncdatastart) this._cbs.oncdatastart(value);
    
    if(this._cbs.oncdataend) this._cbs.oncdataend();
  } else {
    this.oncomment("[CDATA[" + value + "]]");
  }
};

/*
  Parses HTML or XML

  Options:
  - format: 'html' or 'xml'
  - ownerDocument: an MemoryDOMElement instance of type 'document'
*/
function parseMarkup(markup, options) {
  let format = options.ownerDocument ? options.ownerDocument.format : options.format;
  /* istanbul ignore next */
  if (!format) {
    throw new Error("Either 'ownerDocument' or 'format' must be set.")
  }
  let parserOptions = {
    xmlMode : (format === 'xml')
  };
  let handler = new DomHandler({ format });
  let parser = new Parser_1(handler, parserOptions);
  parser.end(markup);
  return handler.document
}


const re_whitespace = /\s+/g;

/*
  Customized implementation of [DomHandler](https://github.com/fb55/domhandler).
*/
class DomHandler {

  constructor(options = {}) {
    this.options = options;
    this.document = null;
    this._tagStack = [];
  }

  // called directly after construction of Parser and at the end of Parser.reset()
  onparserinit(){
    this.document = new MemoryDOMElement('document', { format: this.options.format });
    this._tagStack = [this.document];
  }

  onend(){
    // TODO: would be nice to generate a good error message
    if (this._tagStack.length>1) {
      throw new Error(`Unexpected EOF. Tag was opened but not closed.`)
    }
  }

  onerror(error) {
    throw new Error(error)
  }

  onclosetag() {
    this._tagStack.pop();
  }

  _addDomElement(element) {
    let parent = this._tagStack[this._tagStack.length - 1];
    if (!parent.childNodes) parent.childNodes = [];
    let siblings = parent.childNodes;

    let previousSibling = siblings[siblings.length - 1];
    // set up next/previous link
    element.next = null;
    if(previousSibling){
      element.prev = previousSibling;
      previousSibling.next = element;
    } else {
      element.prev = null;
    }
    // either push the element to the current open tag's children, or keep a reference as top-level element
    siblings.push(element);
    element.parent = parent || null;
  }

  onopentag(name, attributes) {
    let element = this.document.createElement(name);
    forEach(attributes, (val, key) => {
      element.setAttribute(key, val);
    });
    this._addDomElement(element);
    this._tagStack.push(element);
  }

  ontext(text) {
    if (this.options.normalizeWhitespace) {
      text = text.replace(re_whitespace, " ");
    }
    let lastTag;
    let _top = this._tagStack[this._tagStack.length - 1];
    if (_top && _top.childNodes) lastTag = _top.childNodes[_top.childNodes.length - 1];
    if (lastTag && lastTag.type === index.Text) {
      lastTag.data += text;
    } else {
      let element = this.document.createTextNode(text);
      this._addDomElement(element);
    }
  }

  oncomment(data) {
    var lastTag = this._tagStack[this._tagStack.length - 1];
    if(lastTag && lastTag.type === index.Comment){
      lastTag.data += data;
    } else {
      let element = this.document.createComment(data);
      this._addDomElement(element);
      this._tagStack.push(element);
    }
  }

  oncommentend() {
    this._tagStack.pop();
  }

  oncdatastart(data) {
    let element = this.document.createCDATASection(data);
    this._addDomElement(element);
    this._tagStack.push(element);
  }

  oncdataend() {
    this._tagStack.pop();
  }

  onprocessinginstruction(name, data) {
    let element = this.document.createProcessingInstruction(name, data);
    this._addDomElement(element);
  }

}

class MemoryDOMElement extends DOMElement {

  constructor(type, args = {}) {
    super();

    this.type = type;
    if (!type) throw new Error("'type' is mandatory")

    this.ownerDocument = args.ownerDocument;
    /* istanbul ignore next */
    if (type !== 'document' && !this.ownerDocument) {
      throw new Error("'ownerDocument' is mandatory")
    }

    // NOTE: there are some properties which are named so that this
    // can be used together with htmlparser2 and css-select
    // but which could have a better naming, e.g., name -> tagName

    switch(type) {
      case index.Tag: {
        if (!args.name) throw new Error("'name' is mandatory.")
        this.name = this._normalizeName(args.name);
        this.nameWithoutNS = nameWithoutNS(this.name);
        this.properties = new Map();
        this.attributes = new Map();
        this.classes = new Set();
        this.styles = new Map();
        this.eventListeners = [];
        this.childNodes = args.children || args.childNodes || [];
        this._assign(args);
        break
      }
      case index.Text:
      case index.Comment: {
        this.data = args.data || '';
        break
      }
      case index.CDATA: {
        this.data = args.data || '';
        break
      }
      case index.Directive: {
        if (!args.name) throw new Error("'name' is mandatory.")
        this.name = this._normalizeName(args.name);
        this.nameWithoutNS = nameWithoutNS(this.name);
        this.data = args.data;
        break
      }
      case 'document': {
        let format = args.format;
        this.format = format;
        if (!format) throw new Error("'format' is mandatory.")
        this.childNodes = args.children || args.childNodes || [];
        switch(format) {
          case 'xml':
            this.contentType = 'application/xml';
            break
          case 'html':
            this.contentType = 'text/html';
            break
          default:
            throw new Error('Unsupported format ' + format)
        }
        break
      }
      default:
        this.name = null;
        this.properties = new Map();
        this.attributes = new Map();
        this.classes = new Set();
        this.styles = new Map();
        this.eventListeners = [];
        this.childNodes = args.children || args.childNodes || [];
    }
  }

  getNativeElement() {
    return this
  }

  getNodeType() {
    switch(this.type) {
      case index.Tag:
      case index.Script:
      case index.Style:
        return 'element'
      default:
        return this.type
    }
  }

  isTextNode() {
    return this.type === "text"
  }

  isElementNode() {
    return this.type === "tag" || this.type === "script"
  }

  isCommentNode() {
    return this.type === "comment"
  }

  isDocumentNode() {
    return this.type === "document"
  }

  isComponentNode() {
    return this.type === "component"
  }

  clone(deep) {
    let clone$$1 = new MemoryDOMElement(this.type, this);
    if (this.childNodes) {
      clone$$1.childNodes.length = 0;
      if (deep) {
        this.childNodes.forEach((child) => {
          clone$$1.appendChild(child.clone(deep));
        });
      }
    }
    return clone$$1
  }

  get tagName() {
    return this.getTagName()
  }

  set tagName(tagName) {
    this.setTagName(tagName);
  }

  getTagName() {
    return this.name
  }

  setTagName(tagName) {
    if (this._isXML()) {
      this.name = String(tagName);
    } else {
      this.name = String(tagName).toLowerCase();
    }
    return this
  }

  hasAttribute(name) {
    return this.attributes.has(name)
  }

  getAttribute(name) {
    return this.attributes.get(name)
  }

  setAttribute(name, value) {
    value = String(value);
    // Note: keeping the Set version of classes and styles in sync
    switch(name) {
      case 'class':
        parseClasses(this.classes, value);
        break
      case 'style':
        parseStyles(this.styles, value);
        break
      default:
        //
    }
    this.attributes.set(name, value);
    if (this._isHTML()) {
      deriveHTMLPropertyFromAttribute(this, name, value);
    }
    return this
  }

  removeAttribute(name) {
    switch(name) {
      case 'class':
        this.classes = new Set();
        break
      case 'style':
        this.styles = new Map();
        break
      default:
        //
    }
    this.attributes.delete(name);
    return this
  }

  getAttributes() {
    return this.attributes
  }

  getProperty(name) {
    if (this.properties) {
      return this.properties.get(name)
    }
  }

  setProperty(name, value) {
    if (this.properties) {
      if (this._isXML()) {
        throw new Error('setProperty() is only be used on HTML elements')
      }
      _setHTMLPropertyValue(this, name, value);
    }
    return this
  }

  hasClass(name) {
    if (this.classes) {
      return this.classes.has(name)
    }
  }

  addClass(name) {
    this.classes.add(name);
    this.attributes.set('class', stringifyClasses(this.classes));
    return this
  }

  removeClass(name) {
    if (this.classes && this.classes.has(name)) {
      this.classes.delete(name);
      this.attributes.set('class', stringifyClasses(this.classes));
    }
    return this
  }

  getContentType() {
    return this.getOwnerDocument().contentType
  }

  getInnerHTML() {
    return index$2.getInnerHTML(this)
  }

  // TODO: parse html using settings from el,
  // clear old childNodes and append new childNodes
  setInnerHTML(html) {
    if (this.childNodes) {
      let _doc = parseMarkup(html, {
        ownerDocument: this.getOwnerDocument()
      });
      this.empty();
      // ATTENTION: important to copy the childNodes array first
      // as appendChild removes from parent
      _doc.childNodes.slice(0).forEach((child) => {
        this.appendChild(child);
      });
    }
    return this
  }

  getOuterHTML() {
    return index$2.getOuterHTML(this, { xmlMode: this._isXML() })
  }

  getTextContent() {
    return index$2.getText(this)
  }

  setTextContent(text) {
    switch(this.type) {
      case index.Text:
      case index.Comment:
      case index.CDATA: {
        this.data = text;
        break
      }
      default: {
        if (this.childNodes) {
          let child = this.createTextNode(text);
          this.empty();
          this.appendChild(child);
        }
      }
    }
    return this
  }

  getStyle(name) {
    if (this.styles) {
      return this.styles.get(name)
    }
  }

  setStyle(name, value) {
    if (this.styles) {
      if (DOMElement.pxStyles[name] && isNumber(value)) {
        value = value + "px";
      }
      this.styles.set(name, value);
      this.attributes.set('style', stringifyStyles(this.styles));
    }
    return this
  }

  is(cssSelector) {
    return index$1.is(this, cssSelector, { xmlMode: this._isXML() })
  }

  find(cssSelector) {
    return index$1.selectOne(cssSelector, this, { xmlMode: this._isXML() })
  }

  findAll(cssSelector) {
    return index$1.selectAll(cssSelector, this, { xmlMode: this._isXML() })
  }

  getChildCount() {
    if (this.childNodes) {
      return this.childNodes.length
    } else {
      return 0
    }
  }

  getChildNodes() {
    return this.childNodes.slice(0)
  }

  getChildren() {
    return this.childNodes.filter(function(node) {
      return node.type === "tag"
    })
  }

  get children() {
    return this.getChildren()
  }

  getChildAt(pos) {
    if (this.childNodes) {
      return this.childNodes[pos]
    }
  }

  getChildIndex(child) {
    if (this.childNodes) {
      return this.childNodes.indexOf(child)
    }
  }

  getLastChild() {
    if (this.childNodes) {
      return last$1(this.childNodes)
    }
  }

  getFirstChild() {
    if (this.childNodes) {
      return this.childNodes[0]
    }
  }

  getNextSibling() {
    return this.next
  }

  getPreviousSibling() {
    return this.prev
  }

  getParent() {
    return this.parent
  }

  getOwnerDocument() {
    return (this.type === 'document') ? this : this.ownerDocument
  }

  getFormat() {
    return this.getOwnerDocument().format
  }

  createDocument(format) {
    return MemoryDOMElement.createDocument(format)
  }

  createElement(tagName) {
    return new MemoryDOMElement(index.Tag, { name: tagName, ownerDocument: this.getOwnerDocument() })
  }

  createTextNode(text) {
    return new MemoryDOMElement(index.Text, { data: text, ownerDocument: this.getOwnerDocument() })
  }

  createComment(data) {
    return new MemoryDOMElement(index.Comment, { data: data, ownerDocument: this.getOwnerDocument() })
  }

  createProcessingInstruction(name, data) {
    return new MemoryDOMElement(index.Directive, { name: name, data: data, ownerDocument: this.getOwnerDocument() })
  }

  createCDATASection(data) {
    return new MemoryDOMElement(index.CDATA, { data: data, ownerDocument: this.getOwnerDocument() })
  }

  appendChild(child) {
    if (this.childNodes && !isNil(child)) {
      child = this._normalizeChild(child);
      if (!child) return this
      index$2.appendChild(this, child);
      child.ownerDocument = this.getOwnerDocument();
    }
    return this
  }

  removeChild(child) {
    if (child.parentNode === this) {
      child.remove();
    }
  }

  insertAt(pos, child) {
    child = this._normalizeChild(child);
    if (!child) return this
    let childNodes = this.childNodes;
    if (childNodes) {
      // NOTE: manipulating htmlparser's internal children array
      if (pos >= childNodes.length) {
        index$2.appendChild(this, child);
      } else {
        index$2.prepend(childNodes[pos], child);
      }
      child.ownerDocument = this.getOwnerDocument();
    }
    return this
  }

  insertBefore(newChild, before) {
    if (this.childNodes) {
      var pos = this.childNodes.indexOf(before);
      if (pos > -1) {
        index$2.prepend(before, newChild);
        newChild.ownerDocument = this.getOwnerDocument();
      } else {
        throw new Error('insertBefore(): reference node is not a child of this element.')
      }
    }
    return this
  }

  removeAt(pos) {
    let childNodes = this.childNodes;
    if (childNodes) {
      let child = childNodes[pos];
      child.remove();
    }
    return this
  }

  empty() {
    let childNodes = this.childNodes;
    if (childNodes) {
      childNodes.forEach((child) => {
        child.next = child.prev = child.parent = null;
      });
      childNodes.length = 0;
    }
    return this
  }

  remove() {
    index$2.removeElement(this);
    return this
  }

  replaceChild(oldChild, newChild) {
    if (oldChild.parent === this) {
      oldChild.replaceWith(newChild);
    }
    return this
  }

  replaceWith(newEl) {
    newEl = this._normalizeChild(newEl);
    index$2.replaceElement(this, newEl);
    newEl.ownerDocument = this.getOwnerDocument();
    return this
  }

  getEventListeners() {
    return this.eventListeners
  }

  click() {
    this.emit('click', { target: this });
    return this
  }

  emit(name, data) {
    this._propagateEvent(new MemoryDOMElementEvent(name, this, data));
  }

  _propagateEvent(event) {
    let listeners = this.eventListeners;
    if (listeners) {
      let listener = listeners.find((l) => {
        return l.eventName === event.type
      });
      if (listener) listener.handler(event);
      if (event.stopped) return
      let p = this.parentNode;
      if (p) p._propagateEvent(event);
    }
  }

  removeAllEventListeners() {
    this.eventListeners = [];
    return this
  }

  _assign(other) {
    if (other.name) this.name = other.name;
    if (this.classes && other.classes) {
      other.classes.forEach((val) => {
        this.classes.add(val);
      });
    }
    if (this.styles && other.styles) {
      forEach(other.styles, (val, name) => {
        this.styles.set(name, val);
      });
    }
    // TODO: while it is 'smart' to deal with 'style' and 'class'
    // implicitly, it introduces some confusion here
    let otherAttributes = other.attributes || other.attribs;
    if (this.attributes && otherAttributes) {
      forEach(otherAttributes, (val, name) => {
        switch (name) {
          case 'class': {
            parseClasses(this.classes, val);
            break
          }
          case 'style': {
            parseStyles(this.styles, val);
            break
          }
          default:
            this.attributes.set(name, val);
        }
      });
    }
    if (this.eventListeners && other.eventListeners) {
      this.eventListeners = this.eventListeners.concat(other.eventListeners);
    }
  }

  _normalizeChild(child) {
    if (isString$1(child)) {
      child = this.createTextNode(child);
    }
    /* istanbul ignore next */
    if (!child || !child._isMemoryDOMElement) {
      throw new Error('Illegal argument: only String and MemoryDOMElement instances are valid.')
    }
    return child
  }

  _normalizeName(name) {
    if (this._isXML()) {
      return name
    } else {
      return name.toLowerCase()
    }
  }

  _isHTML() {
    return this.getFormat() === 'html'
  }

  _isXML() {
    return this.getFormat() === 'xml'
  }

}

MemoryDOMElement.prototype._isMemoryDOMElement = true;

MemoryDOMElement.createDocument = function(format) {
  if (format === 'xml') {
    return new MemoryDOMElement('document', { format: format })
  } else {
    return MemoryDOMElement.parseMarkup(DOMElement.EMPTY_HTML, 'html')
  }
};

MemoryDOMElement.parseMarkup = function(str, format, options={}) {
  if (!str) {
    return MemoryDOMElement.createDocument(format)
  }
  if (options.snippet) {
    str = `<__snippet__>${str}</__snippet__>`;
  }
  let doc;
  if (format === 'html') {
    doc = parseMarkup(str, { format: format });
    _sanitizeHTMLStructure(doc);
  } else if (format === 'xml') {
    doc = parseMarkup(str, { format: format });
  }
  if (options.snippet) {
    let childNodes = doc.find('__snippet__').childNodes;
    if (childNodes.length === 1) {
      return childNodes[0]
    } else {
      return childNodes
    }
  } else {
    return doc
  }
};

MemoryDOMElement.wrap =
MemoryDOMElement.wrapNativeElement = function(el) {
  if (inBrowser) {
    // HACK: at many places we have an `isBrowser` check
    // to skip code that uses window or window.document
    // To be able to test such code together with the memory DOM implementation
    // we stub out window and document
    if (el === window || el === window.document) {
      return new DOMElementStub()
    }
    // HACK: additionally, if a window.document.Node or a BrowserDOMElement is given
    // as it happens when trying to mount onto t.sandbox with DefaultDOMElement using MemoryDOMElement as default
    // we just return a new root element
    else if (el instanceof window.Node || el._isBrowserDOMElement) {
      // return MemoryDOMElement.createDocument('html').createElement('div')
    }
  }
  /* istanbul ignore next */
  if (!el._isMemoryDOMElement) {
    throw new Error('Illegal argument: expected MemoryDOMElement instance')
  }
  return el
};

MemoryDOMElement.unwrap = function(el) {
  /* istanbul ignore next */
  if (!el._isMemoryDOMElement) {
    throw new Error('Illegal argument: expected MemoryDOMElement instance')
  }
  return el
};

// TODO: this is used only in browser to determine if
// a selection  is reverse.
/* istanbul ignore next */
MemoryDOMElement.isReverse = function() {
  return false
};

function parseClasses(classes, classStr) {
  classStr.split(/\s+/).forEach((name) => {
    classes.add(name);
  });
}

function stringifyClasses(classes) {
  return Array.from(classes).join(' ')
}

function parseStyles(styles, styleStr) {
  styleStr = (styleStr || '').trim();
  if (!styleStr) return
  styleStr.split(';').forEach((style) => {
    let n = style.indexOf(':');
    // skip if there is no :, or if it is the first/last character
    if (n < 1 || n === style.length-1) return
    let name = style.slice(0,n).trim();
    let val = style.slice(n+1).trim();
    styles.set(name, val);
  });
}

function stringifyStyles(styles) {
  if (!styles) return ''
  let str = Object.keys(styles).map((name) => {
    return name + ':' + styles[name]
  }).join(';');
  if (str.length > 0) str += ';';
  return str
}

const BUILTIN_EVENTS = [
  'keydown', 'keyup', 'keypress',
  'mousedown', 'mouseup', 'mouseover', 'click', 'dblclick'
].reduce((m, k)=>{m[k]=true;return m}, {});

class MemoryDOMElementEvent {

  constructor(type, target, detail) {
    this.type = type;
    this.timeStamp = Date.now();
    this.target = target;

    if (BUILTIN_EVENTS[type]) {
      // TODO: dunno if this is the best way of doing it
      if (detail) {
        Object.assign(this, detail);
      }
    } else {
      this.detail = detail;
    }
  }

  stopPropagation() {
    this.stopped = true;
  }

  preventDefault() {
    this.defaultPrevented = true;
  }
}

class DOMElementStub {
  on() {}
  off(){}
}

function nameWithoutNS(name) {
  const idx = name.indexOf(':');
  if (idx > 0) {
    return name.slice(idx+1)
  } else {
    return name
  }
}

// Note: some attributes are used to initialize an
// element property
const ATTR_TO_PROPS = {
  "input": {
    "value": true,
    "checked": (el, name, value) => {
      const checked = (value !== 'off');
      el.setProperty('checked', checked);
    }
  }
};

function deriveHTMLPropertyFromAttribute(el, name, value) {
  const mappings = ATTR_TO_PROPS[el.tagName];
  if (mappings) {
    let mapper = mappings[name];
    if (mapper === true) {
      el.setProperty(name, value);
    } else if (mapper) {
      mapper(el, name, value);
    }
  }
}

const PROPERTY_TRANSFORMATIONS = {
  "input": {
    "checked": (el, name, value) => {
      if (value === true) {
        el.properties.set(name, true);
        el.properties.set('value', 'on');
      } else {
        el.properties.set(name, false);
        el.properties.set('value', 'off');
      }
    },
    "value": (el, name, value) => {
      let type = el.getAttribute('type');
      switch(type) {
        case 'checkbox':
          if (value === 'on') {
            el.properties.set(name, true);
            el.properties.set('value', 'on');
          } else {
            el.properties.set(name, false);
            el.properties.set('value', 'off');
          }
          break
        default:
          _setProperty(el, name, value);
      }
    }
  }
};

function _setProperty(el, name, value) {
  if (value === undefined) {
    el.properties.delete(name);
  } else {
    el.properties.set(name, String(value));
  }
}

function _setHTMLPropertyValue(el, name, value) {
  const trafos = PROPERTY_TRANSFORMATIONS[el.tagName];
  if (trafos) {
    let mapper = trafos[name];
    if (mapper) {
      mapper(el, name, value);
      return
    }
  }
  _setProperty(el, name, value);
}

function _sanitizeHTMLStructure(doc) {
  // as opposed to DOMParser in the browser
  // htmlparser2 does not create <head> and <body> per se
  // thus we need to make sure that everything is working
  // similar as in the browser
  let htmlEl = doc.find('html');
  if (!htmlEl) {
    // remove head and nodes which must go into the head
    // so they do not go into the body
    let headEl = doc.find('head');
    let titleEl = doc.find('title');
    let metaEls = doc.findAll('meta');
    let bodyEl = doc.find('body');
    if (headEl) headEl.remove();
    if (titleEl) titleEl.remove();
    metaEls.forEach(e => e.remove());
    if (bodyEl) bodyEl.remove();

    // keep the remaining content nodes,
    // we will add them to the body
    let contentNodes = doc.childNodes.slice();
    contentNodes.forEach((c)=>{c.parent = null;});
    doc.childNodes.length = 0;

    htmlEl = doc.createElement('html');
    // if not there create a <head> and
    // add all the elements that are supposed to
    // go there
    if (!headEl) {
      headEl = doc.createElement('head');
      headEl.appendChild(titleEl);
      headEl.append(metaEls);
      htmlEl.appendChild(headEl);
    }
    if (!bodyEl) {
      bodyEl = doc.createElement('body');
      bodyEl.append(contentNodes);
    }
    htmlEl.appendChild(bodyEl);

    doc.append(htmlEl);
  }
}

/**
  A Bridge to the default DOMElement implementation, either BrowserDOMElement, or MemoryDOMElement.
*/
let DefaultDOMElement = {};

DefaultDOMElement.createDocument = function(format) {
  return _getDefaultImpl().createDocument(format)
};

/* istanbul ignore next */
DefaultDOMElement.createElement = function(tagName) {
  console.error("DEPRECATED: every element should have an ownerDocument. Use DefaultDOMElement.createDocument() to create a document first");
  let doc = DefaultDOMElement.createDocument('html');
  return doc.createElement(tagName)
};

/* istanbul ignore next */
DefaultDOMElement.createTextNode = function(text) {
  console.error("DEPRECATED: every element should have a ownerDocument. Use DefaultDOMElement.createDocument() to create a document first");
  let doc = DefaultDOMElement.createDocument('html');
  return doc.createTextNode(text)
};

/*
  A wrapper for Browser's `window` providing
  the DOMElement's eventlistener API.
*/
DefaultDOMElement.getBrowserWindow = function() {
  return _getDefaultImpl().getBrowserWindow()
};

/*
  @param {String} html
  @returns {DOMElement|DOMElement[]}
*/
DefaultDOMElement.parseHTML = function(html, options) {
  return _getDefaultImpl().parseMarkup(html, 'html', options)
};

/*
  @param {String} xml
  @returns {DOMElement|DOMElement[]}
*/
DefaultDOMElement.parseXML = function(xml, options) {
  return _getDefaultImpl().parseMarkup(xml, 'xml', options)
};

DefaultDOMElement.parseSnippet = function(str, format) {
  return _getDefaultImpl().parseMarkup(str, format, {snippet: true})
};

DefaultDOMElement.wrap =
DefaultDOMElement.wrapNativeElement = function(nativeEl) {
  if (!nativeEl) throw new Error('Illegal argument')
  return _getDefaultImpl().wrap(nativeEl)
};

DefaultDOMElement.unwrap = function(nativeEl) {
  if (!nativeEl) throw new Error('Illegal argument')
  return _getDefaultImpl().unwrap(nativeEl)
};

//TODO: this should not be part of DefaultDOMElement
/* istanbul ignore next */
DefaultDOMElement.isReverse = function(anchorNode, anchorOffset, focusNode, focusOffset) {
  return _getDefaultImpl().isReverse(anchorNode, anchorOffset, focusNode, focusOffset)
};

function _getDefaultImpl() {
  if (platform.inBrowser && !platform.inNodeJS) {
    return BrowserDOMElement
  } else {
    return MemoryDOMElement
  }
}

/*
  Base class for custom HTML exporters. If you want to use XML as your
  exchange format see {@link model/XMLExporter}.
*/

class HTMLExporter extends DOMExporter {

  constructor(config, context) {
    super(_defaultConfig(config), context);
  }

  exportDocument(doc) {
    let htmlEl = DefaultDOMElement.parseHTML('<html><head></head><body></body></html>');
    return this.convertDocument(doc, htmlEl)
  }

  getDefaultBlockConverter() {
    return defaultBlockConverter // eslint-disable-line no-use-before-define
  }

  getDefaultPropertyAnnotationConverter() {
    return defaultAnnotationConverter // eslint-disable-line no-use-before-define
  }

}

function _defaultConfig(config) {
  config = Object.assign({
    idAttribute: 'data-id'
  }, config);
  if (!config.elementFactory) {
    config.elementFactory = DefaultDOMElement.createDocument('html');
  }
  return config
}


const defaultAnnotationConverter = {
  tagName: 'span',
  export: function(node, el) {
    el.tagName = 'span';
    el.attr('data-type', node.type);
    var properties = node.toJSON();
    forEach(properties, function(value, name) {
      if (name === 'id' || name === 'type') return
      if (isString$1(value) || isNumber(value) || isBoolean(value)) {
        el.attr('data-'+name, value);
      }
    });
  }
};

const defaultBlockConverter = {
  export: function(node, el, converter) {
    el.attr('data-type', node.type);
    var properties = node.toJSON();
    forEach(properties, function(value, name) {
      if (name === 'id' || name === 'type') {
        return
      }
      var prop = converter.$$('div').attr('property', name);
      if (node.getPropertyType(name) === 'string' && value) {
        prop.append(converter.annotatedText([node.id, name]));
      } else {
        prop.text(value);
      }
      el.append(prop);
    });
  }
};

/*
  Base class for custom HTML importers. If you want to use XML as your
  exchange format see {@link model/XMLImporter}.

  @class
  @abstract
*/
class HTMLImporter extends DOMImporter {

  constructor(config) {
    super(Object.assign({ idAttribute: 'data-id' }, config));

    // only used internally for creating wrapper elements
    this._el = DefaultDOMElement.parseHTML('<html></html>');
  }

  importDocument(html) {
    this.reset();
    var parsed = DefaultDOMElement.parseHTML(html);
    this.convertDocument(parsed);
    return this.state.doc
  }

  /**
    Orchestrates conversion of a whole document.

    This method should be overridden by custom importers to reflect the
    structure of a custom HTML document or fragment, and to control where
    things go to within the document.

    @abstract
    @param {ui/DOMElement} documentEl the document element.

    @example

    When a fragment `<h1>Foo</h1><p></Bar</p>` is imported the implementation
    looks like this.

    ```js
      convertDocument(els) {
        this.convertContainer(els, 'body')
      }
    ```

    If a full document `<html><body><p>A</p><p>B</p></body></html>` is imported
    you get the `<html>` element instead of a node array.

    ```js
      convertDocument(htmlEl) {
        var bodyEl = htmlEl.find('body')
        this.convertContainer(bodyEl.children, 'body')
      }
    ```
  */
  convertDocument(documentEl) { // eslint-disable-line
    throw new Error('This method is abstract')
  }

}

class InlineNode extends PropertyAnnotation {}

InlineNode.prototype._isInlineNode = true;

InlineNode.isInline = true;

class Marker extends PropertyAnnotation {
  invalidate() {}
  remove() {
    this.getDocument().data.delete(this.id);
  }

  // TODO: we should use the Coordinate comparison API here
  containsSelection(sel) {
    if (sel.isNull()) return false;
    if (sel.isPropertySelection()) {
      return (isArrayEqual(this.start.path, sel.start.path) &&
        this.start.offset <= sel.start.offset &&
        this.end.offset >= sel.end.offset)
    } else {
      console.warn('Marker.contains() does not support other selection types.');
    }
  }

}

// while having the same interface, Markers should still be treated differently, e.g. not go into the AnnotationIndex
Marker.prototype._isPropertyAnnotation = false;
Marker.prototype._isMarker = true;
Marker.autoExpandRight = false;

class Range {

  constructor(start, end, reverse, containerId, surfaceId) {
    // HACK: to allow this class be inherited but without calling this ctor
    if (arguments[0] === 'SKIP') return
    if (arguments.length === 1 && isPlainObject$1(arguments[0])) {
      let data = arguments[0];
      this.start = data.start;
      this.end = data.end;
      this.reverse = Boolean(data.reverse);
      this.containerId = data.containerId;
      this.surfaceId = data.surfaceId;
    } else {
      this.start = start;
      this.end = end;
      this.reverse = Boolean(reverse);
      this.containerId = containerId;
      this.surfaceId = surfaceId;
    }
  }

  isCollapsed() {
    return this.start.equals(this.end)
  }

  equals(other) {
    if (this === other) return true
    else {
      return (
        this.containerId === other.containerId &&
        this.start.equals(other.start) &&
        this.end.equals(other.end)
      )
    }
  }

  isReverse() {
    return this.reverse
  }

  toString() {
    let str = [this.start.toString(), '->', this.end.toString()];
    if (this.isReverse()) {
      str.push('[reverse]');
    }
    if (this.containerId) {
      str.push('[container='+this.containerId+']');
    }
    if (this.surfaceId) {
      str.push('[surface='+this.surfaceId+']');
    }
    return str.join('')
  }

}

Range.prototype._isRange = true;

// import printStacktrace from '../util/printStacktrace'

class SelectionState {

  constructor(doc) {
    this.document = doc;

    this.selection = Selection.nullSelection;
    this._state = {};
    this._resetState();
  }

  setSelection(sel) {
    // printStacktrace()
    if (!sel) {
      sel = Selection.nullSelection;
    } else {
      sel.attach(this.document);
    }
    // TODO: selection state is selection plus derived state,
    // thus we need to return false only if both did not change
    this._deriveState(sel);
    this.selection = sel;
    return true
  }

  getSelection() {
    return this.selection
  }

  getAnnotationsForType(type) {
    const state = this._state;
    if (state.annosByType) {
      return state.annosByType.get(type) || []
    }
    return []
  }

  getMarkers() {
    // returns markers under the current selection
    return this._state.markers || []
  }

  isInlineNodeSelection() {
    return this._state.isInlineNodeSelection
  }

  getContainer() {
    return this._state.container
  }

  getPreviousNode() {
    return this._state.previousNode
  }

  getNextNode() {
    return this._state.nextNode
  }

  /*
    @returns if the previous node is one char away
  */
  isFirst() {
    return Boolean(this._state.isFirst)
  }

  /*
    @returns if the next node is one char away
  */
  isLast() {
    return Boolean(this._state.isLast)
  }

  get(key) {
    return this._state[key]
  }

  // used to store custom states (e.g. IsolatedNodeComponent uses this)
  set(key, value) {
    if (this._state[key]) {
      throw new Error(`State ${key} is already set`)
    }
    this._state[key] = value;
  }

  _deriveState(sel) {
    this._resetState();
    this._deriveContainerSelectionState(sel);
    this._deriveAnnoState(sel);
    if (this.document.getIndex('markers')) {
      this._deriveMarkerState(sel);
    }
    // console.log('SelectionState:', this._state)
  }

  _deriveContainerSelectionState(sel) {
    let state = this._state;
    let doc = this.document;
    if (sel.containerId) {
      let container = doc.get(sel.containerId);
      state.container = container;
      let startId = sel.start.getNodeId();
      let endId = sel.end.getNodeId();
      let startNode = doc.get(startId).getRoot();
      let startPos = container.getPosition(startNode);
      if (startPos > 0) {
        state.previousNode = container.getNodeAt(startPos-1);
      }
      state.isFirst = isFirst(doc, sel.start);
      let endNode, endPos;
      if (endId === startId) {
        endNode = startNode;
        endPos = startPos;
      } else {
        endNode = doc.get(endId).getRoot();
        endPos = container.getPosition(endNode);
      }
      if (endPos < container.getLength()-1) {
        state.nextNode = container.getNodeAt(endPos+1);
      }
      state.isLast = isLast(doc, sel.end);
    }
  }

  _deriveAnnoState(sel) {
    const doc = this.document;
    const state = this._state;

    // create a mapping by type for the currently selected annotations
    let annosByType = new TreeIndex.Arrays();
    const propAnnos = documentHelpers.getPropertyAnnotationsForSelection(doc, sel);
    propAnnos.forEach(function(anno) {
      annosByType.add(anno.type, anno);
    });

    if (propAnnos.length === 1 && propAnnos[0].isInline()) {
      state.isInlineNodeSelection = propAnnos[0].getSelection().equals(sel);
    }

    const containerId = sel.containerId;
    if (containerId) {
      const containerAnnos = documentHelpers.getContainerAnnotationsForSelection(doc, sel, containerId);
      containerAnnos.forEach(function(anno) {
        annosByType.add(anno.type, anno);
      });
    }
    state.annosByType = annosByType;
  }

  _deriveMarkerState(sel) {
    const doc = this.document;
    let state = this._state;
    let markers = documentHelpers.getMarkersForSelection(doc, sel);
    state.markers = markers;
  }

  _resetState() {
    this._state = {
      // all annotations under the current selection
      annosByType: null,
      // markers under the current selection
      markers: null,
      // flags for inline nodes
      isInlineNodeSelection: false,
      // container information (only for ContainerSelection)
      container: null,
      previousNode: null,
      nextNode: null,
      // if the previous node is one char away
      isFirst: false,
      // if the next node is one char away
      isLast: false
    };
    return this._state
  }
}

/**
  A base class for all text-ish nodes, such as Paragraphs, Headings,
  Prerendered, etc.
*/
class TextNode extends DocumentNode {

  getTextPath() {
    // TODO: deprecate this
    // console.warn('DEPRECATED: use node.getPath()')
    return this.getPath()
  }

  getPath() {
    return [this.id, 'content']
  }

  getText() {
    return this.content
  }

  isEmpty() {
    return !this.content
  }

  getLength() {
    return this.content.length
  }

}

TextNode.isText = true;

TextNode.schema = {
  type: "text",
  content: "text",
  direction: { type: "string", optional: true }
};

class TextBlock extends TextNode {}

TextBlock.isBlock = true;

/*
  A transaction for editing a document in an EditorSession.

  Wherever you see `tx`, it is an instance of this class.

  The transaction is used to manipulate the document in a 'turtle-graphics' style.
  For that it maintains an internal state consisting of an array of operations, a selection, and
  the current surface.

  Usually, at the beginning of a transaction, one Surface is focused. This is used to initialize
  the transaction state. Depending on the type of Surface or the type of the current selection,
  some manipulations are allowed or others are not: it is not possible to create a ContainerAnnotation without
  a ContainerSelection, or pasting a list of nodes into a TextPropertyEditor will strip the structure and just
  the text content.

*/
class Transaction extends EditingInterface {

  /*
    @param {Document} doc
  */
  constructor(doc, editorSession) {
    super();

    // TransactionDocument is essentially a clone of the document used to apply a sequence of document operations
    // without touching the original document
    this._stageDoc = this._document = new TransactionDocument(doc, this);
    this._editorSession = editorSession;

    // internal state
    this._isTransacting = false;
    this._state = 'idle';
    this._surface = null;
  }

  dispose() {
    this._stageDoc.dispose();
  }

  setSelection(sel) {
    super.setSelection(sel);

    // NOTE: we might want to remove 'surfaceId' from selection, and instead
    // map surfaces to model paths. For the time being we keep it the old
    // way, but take it from the currently focused surface
    sel = this._selection;
    if (!sel.isNull()) {
      if (!sel.surfaceId) {
        // TODO: We could check if the selection is valid within the given surface
        let surface = this._editorSession.getFocusedSurface();
        if (surface) {
          sel.surfaceId = surface.id;
        } else {
          // TODO: instead of warning we could try to 'find' a suitable surface. However, this would also be a bit 'magical'
          console.warn('No focused surface. Selection will not be rendered.');
        }
      }
    }
  }

  // internal API

  // NOTE: ops are actually owned by TransactionDocument
  // we use the transaction document internally and not this instance
  get ops() {
    return this._stageDoc.ops
  }
  set ops(ops) {
    this._stageDoc.ops = ops;
  }

  rollback() {
    this._stageDoc._rollback();
  }

  _apply(...args) {
    this._stageDoc._apply(...args);
  }

  // _ensureStarted() {
  //   if (this._state !== 'started') throw new Error('Transaction has not been started, or cancelled or saved already.')
  // }

  /**
    Start a transaction to manipulate the document

    @param {function} transformation a function(tx) that performs actions on the transaction document tx

    @example

    ```js
    doc.transaction(function(tx, args) {
      tx.update(...)
      ...
      return {
        selection: newSelection
      }
    })
    ```
  */
  _recordChange(transformation, selection) {
    // TODO: we could get rid of isTransacting and use this._state instead
    if (this._isTransacting) throw new Error('Nested transactions are not supported.')
    if (!isFunction$1(transformation)) throw new Error('Document.transaction() requires a transformation function.')
    this._isTransacting = true;
    this._reset();
    this._state = 'started';
    let change;
    try {
      this.setSelection(selection);
      let selBefore = this.getSelection();
      transformation(this, {
        selection: selBefore
      });
      let ops = this.ops;
      if (ops.length > 0) {
        change = new DocumentChange(ops, this._before, this._after);
        change.before = { selection: selBefore };
        change.after = { selection: this.getSelection() };
      }
      this._state = 'finished';
    } finally {
      if (this._state !== 'finished') {
        this.rollback();
      }
      this._state = 'idle';
      this._isTransacting = false;
    }
    return change
  }

  _reset() {
    this._before = {};
    this._after = {};
    this._stageDoc._reset();
    this._info = {};
    this.setSelection(null);
  }
}

/*
  Base class for custom XML exporters. If you want to use HTML as your
  exchange format see {@link model/HTMLExporter}.
*/
class XMLExporter extends DOMExporter {

  constructor(config, context) {
    super(_defaultConfig$1(config), context);
  }

  getDefaultBlockConverter() {
    return defaultBlockConverter$1 // eslint-disable-line no-use-before-define
  }

  getDefaultPropertyAnnotationConverter() {
    return defaultAnnotationConverter$1 // eslint-disable-line no-use-before-define
  }

}

function _defaultConfig$1(config) {
  config = Object.assign({
    idAttribute: 'id'
  }, config);
  if (!config.elementFactory) {
    config.elementFactory = DefaultDOMElement.createDocument('xml');
  }
  return config
}

const defaultAnnotationConverter$1 = {
  tagName: 'annotation',
  export: function(node, el) {
    el.attr('type', node.type);
    const properties = node.toJSON();
    forEach(properties, function(value, name) {
      if (name === 'id' || name === 'type') return
      if (isString$1(value) || isNumber(value) || isBoolean(value)) {
        el.attr(name, value);
      }
    });
  }
};

const defaultBlockConverter$1 = {
  tagName: 'block',
  export: function(node, el, converter) {
    el.attr('type', node.type);
    const properties = node.toJSON();
    forEach(properties, function(value, name) {
      if (name === 'id' || name === 'type') {
        return
      }
      const prop = converter.$$(name);
      if (node.getPropertyType(name) === 'string') {
        prop.append(converter.annotatedText([node.id, name]));
      } else {
        prop.text(value);
      }
      el.append(prop);
    });
  }
};

/*
  Base class for custom XML importers. If you want to use HTML as your
  exchange format see {@link model/HTMLImporter}.

  TODO: provide example and activate reenable API docs
*/

class XMLImporter extends DOMImporter {

  constructor(config, context) {
    super(Object.assign({ idAttribute: 'id' }, config), context);
  }

  importDocument(xml) {
    this.reset();
    let dom = DefaultDOMElement.parseXML(xml);
    this.convertDocument(dom);
    return this.state.doc
  }

}

/*
  Compute snapshot based on a given doc (JSON) and a set of changes
*/
function computeSnapshot(jsonDoc, changeset) {
  // Clone the doc to make sure we don't manipulate in-place
  jsonDoc = cloneDeep(jsonDoc);
  let nodes = jsonDoc.nodes;
  changeset.forEach((change) => {
    change.ops.forEach((opData) => {
      try {
        let op = ObjectOperation.fromJSON(opData);
        op.apply(nodes);
      } catch (err) {
        console.error(err, opData);
      }
    });
  });
  return jsonDoc
}

const EMPTY_DOC = { nodes: {} };

/*
  API for creating and retrieving document snapshots
*/
class SnapshotEngine {
  constructor(config) {
    this.changeStore = config.changeStore;
    this.snapshotStore = config.snapshotStore;
  }

  /*
    Compute snapshot for given documentId and version

    Example: Let's assume we want to request a snapshot for a new version 20.
    Now getClosestSnapshot will give us version 15. This requires us to fetch
    the changes since version 16 and apply those, plus the very new change.
  */
  getSnapshot(documentId, version, cb) {
    let jsonDoc = EMPTY_DOC;
    this._getClosestSnapshot(documentId, version, (err, snapshot, closestVersion) => {
      if (err) {
        return cb(err)
      }
      if (snapshot && version === closestVersion) {
        // we alread have a snapshot for this version
        return cb(null, snapshot, version)
      }
      let knownVersion;
      if (snapshot) {
        knownVersion = closestVersion;
      } else {
        knownVersion = 0; // we need to fetch all changes
      }
      if (snapshot) {
        jsonDoc = snapshot;
      }
      // Now we get the remaining changes after the known version
      this.changeStore.getChanges(documentId, knownVersion, version, (err, changes) => {
        if (err) return cb(err)
        if (changes.length < (version - knownVersion)) {
          return cb('Changes missing for reconstructing version '+ version)
        }
        jsonDoc = computeSnapshot(jsonDoc, changes);
        cb(null, jsonDoc, version);
      });
    });
  }

  /*
    Creates a snapshot
  */
  createSnapshot(documentId, version, cb) {
    this.getSnapshot(documentId, version, (err, snapshot) => {
      if (err) return cb(err)
      this.snapshotStore.saveSnapshot(documentId, version, snapshot, cb);
    });
  }

  _getClosestSnapshot(documentId, version, cb) {
    let closestVersion;

    this.snapshotStore.getVersions(documentId, (err, versions) => {
      if (versions.indexOf(version) >= 0) {
        closestVersion = version;
      } else {
        // We don't have a snaphot for that requested version
        let smallerVersions = versions.filter(function(v) {
          return parseInt(v, 10) < version
        });
        // Take the closest version if there is any
        closestVersion = Math.max.apply(null, smallerVersions);
      }
      if (!closestVersion) {
        return cb(null, undefined)
      }
      this.snapshotStore.getSnapshot(documentId, version, cb);
    });
  }
}

/*
  DocumentEngine
*/
class DocumentEngine extends EventEmitter {
  constructor(config) {
    super();
    this.changeStore = config.changeStore;
    this.snapshotStore = config.snapshotStore;
    // Snapshot creation frequency (e.g. if it equals 15 then every
    // 15th version will be saved as snapshot)
    this.snapshotFrequency = config.snapshotFrequency || 1;
    this.snapshotEngine = new SnapshotEngine({
      changeStore: this.changeStore,
      snapshotStore: this.snapshotStore
    });
  }

  /*
    Creates a new document

    A valid document must have at least one valid change
  */
  createDocument(documentId, initialChange, cb) {
    this.addChange(documentId, initialChange, cb);
  }

  /*
    Get a document snapshot for a given version. If no version
    is provivded, the latest version is returned
  */
  getDocument(documentId, version, cb) {
    if (typeof version === 'function') {
      cb = version;
      version = undefined;
    }
    if (!documentId) {
      throw new Error('Invalid Arguments')
    }
    if (version === undefined) {
      this.getVersion(documentId, (err, version) => {
        if (err) return cb(err)
        this.snapshotEngine.getSnapshot(documentId, version, cb);
      });
    } else {
      this.snapshotEngine.getSnapshot(documentId, version, cb);
    }
  }

  /*
    Delete document by documentId
  */
  deleteDocument(documentId, cb) {
    this.changeStore.deleteChanges(documentId, (err) => {
      if (err) {
        return cb(new Error('Deleting changes failed'))
      }
    });
  }

  /*
    Check if a given document exists
  */
  documentExists(documentId, cb) {
    this.getVersion(documentId, (err, version) => {
      if (version >= 0) {
        cb(null, true);
      } else {
        cb(null, false);
      }
    });
  }

  /*
    Get changes based on documentId, sinceVersion
  */
  getChanges(documentId, sinceVersion, toVersion, cb) {
    this.changeStore.getChanges(documentId, sinceVersion, toVersion, cb);
  }

  /*
    Get version for given documentId
  */
  getVersion(documentId, cb) {
    this.changeStore.getVersion(documentId, cb);
  }

  /*
    Here the implementer decides whether a snapshot should be created or not.
    It may be a good strategy to only create a snaphot for every 10th version.
    However for now we will just snapshot each change to keep things simple.
  */
  requestSnapshot(documentId, version, cb) {
    if (version % this.snapshotFrequency === 0) {
      this.snapshotEngine.createSnapshot(documentId, version, cb);
    } else {
      cb(null); // do nothing
    }
  }

  /*
    Add change to a given documentId.

    Snapshot creation is requested on each change to be stored.
  */
  addChange(documentId, change, cb) {
    this.changeStore.addChange(documentId, change, (err, newVersion) => {
      if (err) return cb(err)

      this.requestSnapshot(documentId, newVersion, () => {
        // no matter if snaphot creation errored or not we will confirm change
        cb(null, newVersion);
      });
    });
  }
}

/*
  Implements Substance SnapshotStore API. This is just a dumb store.
  No integrity checks are made, as this is the task of SnapshotEngine
*/
class SnapshotStore {
  constructor(seed) {
    this._snapshots = seed || {};
  }

  /*
    Get all available versions for a document
  */
  getVersions(documentId, cb) {
    let versions = this._getVersions(documentId);
    cb(null, versions);
  }

  /*
    Get Snapshot by documentId and version.

    Returns snapshot data and snaphot version
  */
  getSnapshot(documentId, version, cb) {
    if (!arguments.length === 3) {
      throw new Error('Invalid Arguments')
    }
    let docEntry = this._snapshots[documentId];
    if (!docEntry) return cb(null, undefined)
    let snapshot = docEntry[version];
    if (snapshot) {
      cb(null, snapshot, version);
    } else {
      cb(null, undefined);
    }
  }

  /*
    Saves a snapshot for a given documentId and version.

    Please note that an existing snapshot will be overwritten.
  */
  saveSnapshot(documentId, version, data, cb) {
    if (!documentId || !version || !data) {
      throw new Error('Invalid arguments')
    }
    let docEntry = this._snapshots[documentId];
    if (!docEntry) {
      docEntry = this._snapshots[documentId] = {};
    }
    docEntry[version] = data;
    cb(null, docEntry[version]);
  }

  /*
    Removes a snapshot for a given documentId + version
  */
  deleteSnapshot(documentId, version, cb) {
    let docEntry = this._snapshots[documentId];
    if (!docEntry || !docEntry[version]) {
      return cb(new Error('Snapshot does not exist and can not be deleted'))
    }
    let snapshot = this._snapshots[documentId][version];
    delete this._snapshots[documentId][version];
    cb(null, snapshot);
  }

  /*
    Get versions for a given document
  */
  _getVersions(documentId) {
    let docEntry = this._snapshots[documentId];
    if (!docEntry) return [] // no versions available
    return Object.keys(docEntry)
  }

}

/**
  A virtual {@link DOMElement} which is used by the {@link Component} API.

  A VirtualElement is just a description of a DOM structure. It represents a
  virtual DOM mixed with Components. This virtual structure needs to be compiled
  to a {@link Component} to actually create a real DOM element,
  which is done by {@link RenderingEngine}
*/
class VirtualElement extends DOMElement {

  constructor(owner) {
    super();

    // set when this gets inserted into another virtual element
    this.parent = null;
    // set when created by RenderingContext
    this._owner = owner;
    // set when ref'd
    this._ref = null;
  }

  getParent() {
    return this.parent
  }

  get childNodes() {
    return this.getChildNodes()
  }

  /*
    Provides the component after this VirtualElement has been rendered.
  */
  getComponent() {
    return this._comp
  }

  /**
    Associates a reference identifier with this element.

    When rendered the corresponding component is stored in the owner using the given key.
    In addition to that, components with a reference are preserved when its parent is rerendered.

    > Attention: only the owner should use this method, as it only
      affects the owner's references

    @param {String} ref id for the compiled Component
  */
  ref(ref) {
    if (!ref) throw new Error('Illegal argument')
    /*
      Attention: only the owner can create a ref()
      If you run into this situation, e.g. when you pass down a virtual element
      to a component which wants to have a ref itself,
      then you have other options:

      1. via props:
      ```js
        this.props.content.getComponent()
      ```

      2. via Component.getChildAt or Component.find()
      ```
        this.getChildAt(0)
        this.find('.child')
      ```
    */
    if (this._ref) throw new Error('A VirtualElement can only be referenced once.')
    this._ref = ref;
    if (this._context) {
      const refs = this._context.refs;
      if(refs[ref]) {
        throw new Error('An item with reference "'+ref+'" already exists.')
      }
      refs[ref] = this;
    }
    return this
  }

  isInDocument() {
    return false
  }

}

VirtualElement.prototype._isVirtualElement = true;

/*
  A virtual HTML element.

  @private
  @class VirtualElement.VirtualHTMLElement
  @extends ui/VirtualElement
*/
class VirtualHTMLElement extends VirtualElement {

  constructor(tagName) {
    super();

    this._tagName = tagName;
    this.classNames = null;
    this.attributes = null;
    this.htmlProps = null;
    this.style = null;
    this.eventListeners = null;

    // TODO: this is semantically incorrect. It should be named childNodes
    this.children = [];
  }

  getTagName() {
    return this._tagName
  }

  setTagName(tagName) {
    this._tagName = tagName;
    return this
  }

  hasClass(className) {
    if (this.classNames) {
      return this.classNames.indexOf(className) > -1
    }
    return false
  }

  addClass(className) {
    if (!this.classNames) {
      this.classNames = [];
    }
    this.classNames.push(className);
    return this
  }

  removeClass(className) {
    if (this.classNames) {
      this.classNames = without(this.classNames, className);
    }
    return this
  }

  removeAttribute(name) {
    if (this.attributes) {
      delete this.attributes[name];
    }
    return this
  }

  getAttribute(name) {
    if (this.attributes) {
      return this.attributes[name]
    }
  }

  setAttribute(name, value) {
    if (!this.attributes) {
      this.attributes = {};
    }
    this.attributes[name] = value;
    return this
  }

  getAttributes() {
    // we are having separated storages for differet
    // kind of attributes which we now pull together
    // in the same way as a native DOM element has it
    var attributes = {};
    if (this.attributes) {
      Object.assign(attributes, this.attributes);
    }
    if (this.classNames) {
      attributes.class = this.classNames.join(' ');
    }
    if (this.style) {
      attributes.style = map(this.style, function(val, key) {
        return key + ":" + val
      }).join(';');
    }
    return attributes
  }

  getId() {
    return this.getAttribute('id')
  }

  setId(id) {
    this.setAttribute('id', id);
    return this
  }

  setTextContent(text) {
    text = String(text || '');
    this.empty();
    this.appendChild(text);
    return this
  }

  setInnerHTML(html) {
    html = html || '';
    this.empty();
    this._innerHTMLString = html;
    return this
  }

  getInnerHTML() {
    if (!this.hasOwnProperty('_innerHTMLString')) {
      throw new Error('Not supported.')
    } else {
      return this._innerHTMLString
    }
  }

  getValue() {
    return this.htmlProp('value')
  }

  setValue(value) {
    this.htmlProp('value', value);
    return this
  }

  getChildNodes() {
    return this.children
  }

  getChildren() {
    return this.children.filter(function(child) {
      return child.getNodeType() !== "text"
    })
  }

  isTextNode() {
    return false
  }

  isElementNode() {
    return true
  }

  isCommentNode() {
    return false
  }

  isDocumentNode() {
    return false
  }

  append() {
    if (this._innerHTMLString) {
      throw Error('It is not possible to mix $$.html() with $$.append(). You can call $$.empty() to reset this virtual element.')
    }
    this._append(this.children, arguments);
    return this
  }

  appendChild(child) {
    if (this._innerHTMLString) {
      throw Error('It is not possible to mix $$.html() with $$.append(). You can call $$.empty() to reset this virtual element.')
    }
    this._appendChild(this.children, child);
    return this
  }

  insertAt(pos, child) {
    child = this._normalizeChild(child);
    if (!child) {
      throw new Error('Illegal child: ' + child)
    }
    if (!child._isVirtualElement) {
      throw new Error('Illegal argument for $$.insertAt():' + child)
    }
    if (pos < 0 || pos > this.children.length) {
      throw new Error('insertAt(): index out of bounds.')
    }
    this._insertAt(this.children, pos, child);
    return this
  }

  insertBefore(child, before) {
    var pos = this.children.indexOf(before);
    if (pos > -1) {
      this.insertAt(pos, child);
    } else {
      throw new Error('insertBefore(): reference node is not a child of this element.')
    }
    return this
  }

  removeAt(pos) {
    if (pos < 0 || pos >= this.children.length) {
      throw new Error('removeAt(): Index out of bounds.')
    }
    this._removeAt(pos);
    return this
  }

  removeChild(child) {
    if (!child || !child._isVirtualElement) {
      throw new Error('removeChild(): Illegal arguments. Expecting a CheerioDOMElement instance.')
    }
    var idx = this.children.indexOf(child);
    if (idx < 0) {
      throw new Error('removeChild(): element is not a child.')
    }
    this.removeAt(idx);
    return this
  }

  replaceChild(oldChild, newChild) {
    if (!newChild || !oldChild ||
        !newChild._isVirtualElement || !oldChild._isVirtualElement) {
      throw new Error('replaceChild(): Illegal arguments. Expecting BrowserDOMElement instances.')
    }
    var idx = this.children.indexOf(oldChild);
    if (idx < 0) {
      throw new Error('replaceChild(): element is not a child.')
    }
    this.removeAt(idx);
    this.insertAt(idx, newChild);
    return this
  }

  empty() {
    var children = this.children;
    while (children.length) {
      var child = children.pop();
      child.parent = null;
    }
    delete this._innerHTMLString;
    return this
  }

  getProperty(name) {
    if (this.htmlProps) {
      return this.htmlProps[name]
    }
  }

  setProperty(name, value) {
    if (!this.htmlProps) {
      this.htmlProps = {};
    }
    this.htmlProps[name] = value;
    return this
  }

  removeProperty(name) {
    if (this.htmlProps) {
      delete this.htmlProps[name];
    }
    return this
  }

  getStyle(name) {
    if (this.style) {
      return this.style[name]
    }
  }

  setStyle(name, value) {
    if (!this.style) {
      this.style = {};
    }
    if (DOMElement.pxStyles[name] && isNumber(value)) value = value + 'px';
    this.style[name] = value;
    return this
  }

  _createEventListener(eventName, handler, options) {
    options.context = options.context || this._owner._comp;
    return super._createEventListener(eventName, handler, options)
  }

  getNodeType() {
    return "element"
  }

  hasInnerHTML() {
    return Boolean(this._innerHTMLString)
  }

  _normalizeChild(child) {
    if (isNil(child)) {
      return
    } else if (child._isVirtualElement) {
      return child
    } else if (isString$1(child) || isBoolean(child) || isNumber(child)) {
      return new VirtualTextNode(String(child))
    } else {
      throw new Error('Unsupported child type')
    }
  }

  _append(outlet, args) {
    if (args.length === 1 && !isArray$1(args[0])) {
      this._appendChild(outlet, args[0]);
      return
    }
    var children;
    if (isArray$1(args[0])) {
      children = args[0];
    } else if (arguments.length > 1) {
      children = Array.prototype.slice.call(args,0);
    } else {
      return
    }
    children.forEach(this._appendChild.bind(this, outlet));
  }

  _appendChild(outlet, child) {
    child = this._normalizeChild(child);
    // TODO: discuss. Having a bad feeling about this,
    // because it could obscure an implementation error
    if (!child) return
    outlet.push(child);
    this._attach(child);
    return child
  }

  _insertAt(outlet, pos, child) {
    if (!child) return
    outlet.splice(pos, 0, child);
    this._attach(child);
  }

  _removeAt(outlet, pos) {
    var child = outlet[pos];
    outlet.splice(pos, 1);
    this._detach(child);
  }

  _attach(child) {
    child.parent = this;
    if (this._context && child._owner !== this._owner && child._ref) {
      this._context.foreignRefs[child._ref] = child;
    }
  }

  _detach(child) {
    child.parent = null;
    if (this._context && child._owner !== this._owner && child._ref) {
      delete this.context.foreignRefs[child._ref];
    }
  }

  _mergeHTMLConfig(other) {
    if (other.classNames) {
      if (!this.classNames) {
        this.classNames = [];
      }
      this.classNames = this.classNames.concat(other.classNames);
    }
    if (other.attributes) {
      if (!this.attributes) {
        this.attributes = {};
      }
      Object.assign(this.attributes, other.attributes);
    }
    if (other.htmlProps) {
      if (!this.htmlProps) {
        this.htmlProps = {};
      }
      Object.assign(this.htmlProps, other.htmlProps);
    }
    if (other.style) {
      if (!this.style) {
        this.style = {};
      }
      Object.assign(this.style, other.style);
    }
    if (other.eventListeners) {
      if (!this.eventListeners) {
        this.eventListeners = [];
      }
      this.eventListeners = this.eventListeners.concat(other.eventListeners);
    }
  }
}

VirtualHTMLElement.prototype._isVirtualHTMLElement = true;


/*
  A virtual element which gets rendered by a custom component.

  @private
  @class VirtualElement.VirtualComponent
  @extends ui/VirtualElement
*/
class VirtualComponent extends VirtualHTMLElement {

  constructor(ComponentClass, props) {
    super();

    props = props || {};

    this.ComponentClass = ComponentClass;
    this.props = props;
    if (!props.children) {
      props.children = [];
    }
    this.children = props.children;
  }

  get _isVirtualComponent() { return true; }

  getComponent() {
    return this._comp
  }

  // Note: for VirtualComponentElement we put children into props
  // so that the render method of ComponentClass can place it.
  getChildren() {
    return this.props.children
  }

  getNodeType() {
    return 'component'
  }

  outlet(name) {
    return new Outlet(this, name)
  }

  _attach(child) {
    child._preliminaryParent = this;
  }

  _detach(child) {
    child._preliminaryParent = null;
  }

  _copyHTMLConfig() {
    return {
      classNames: clone(this.classNames),
      attributes: clone(this.attributes),
      htmlProps: clone(this.htmlProps),
      style: clone(this.style),
      eventListeners: clone(this.eventListeners)
    }
  }
}

class Outlet {
  constructor(virtualEl, name) {
    this.virtualEl = virtualEl;
    this.name = name;
    Object.freeze(this);
  }

  _getOutlet() {
    var outlet = this.virtualEl.props[this.name];
    if (!outlet) {
      outlet = [];
      this.virtualEl.props[this.name] = outlet;
    }
    return outlet
  }

  append() {
    var outlet = this._getOutlet();
    this.virtualEl._append(outlet, arguments);
    return this
  }

  empty() {
    var arr = this.virtualEl.props[this.name];
    arr.forEach(function(el) {
      this._detach(el);
    }.bind(this));
    arr.splice(0, arr.length);
    return this
  }
}


class VirtualTextNode extends VirtualElement {

  constructor(text) {
    super();
    this.text = text;
  }

  get _isVirtualTextNode() { return true; }
}

VirtualElement.Component = VirtualComponent;
VirtualElement.TextNode = VirtualTextNode;

/**
  Create a virtual DOM representation which is used by Component
  for differential/reactive rendering.

  @param elementType HTML tag name or Component class
  @param [props] a properties object for Component classes
  @return {VirtualElement} a virtual DOM node

  @example

  Create a virtual DOM Element

  ```
  $$('a').attr({href: './foo'}).addClass('se-nav-item')
  ```

  Create a virtual Component

  ```
  $$(HelloMessage, {name: 'John'})
  ```
*/
VirtualElement.createElement = function() {
  var content;
  var _first = arguments[0];
  var _second = arguments[1];
  var type;
  if (isString$1(_first)) {
    type = "element";
  } else if (isFunction$1(_first) && _first.prototype._isComponent) {
    type = "component";
  } else if (isNil(_first)) {
    throw new Error('$$(null): provided argument was null or undefined.')
  } else {
    throw new Error('Illegal usage of $$()')
  }
  // some props are mapped to built-ins
  var props = {};
  var classNames, ref;
  var eventHandlers = [];
  for(var key in _second) {
    if (!_second.hasOwnProperty(key)) continue
    var val = _second[key];
    switch(key) {
      case 'class':
        classNames = val;
        break
      case 'ref':
        ref = val;
        break
      default:
        props[key] = val;
    }
  }
  if (type === 'element') {
    content = new VirtualHTMLElement(_first);
    // remaining props are attributes
    // TODO: should we make sure that these are only string values?
    content.attr(props);
  } else {
    content = new VirtualComponent(_first, props);
  }
  // HACK: this is set to the current context by RenderingEngine
  // otherwise this will provide rubbish
  content._owner = this.owner;
  if (classNames) {
    content.addClass(classNames);
  }
  if (ref) {
    content.ref(ref);
  }
  eventHandlers.forEach(function(h) {
    if (isFunction$1(h.handler)) {
      content.on(h.name, h.handler);
    } else if (isPlainObject$1(h.handler)) {
      var params = h.handler;
      content.on(h.name, params.handler, params.context, params);
    } else {
      throw new Error('Illegal arguments for $$(_,{ on'+h.name+'})')
    }
  });
  // allow a notation similar to React.createElement
  // $$(MyComponent, {}, ...children)
  if (arguments.length > 2) {
    content.append(flattenOften(Array.prototype.slice.call(arguments, 2), 3));
  }
  return content
};

/*

  ## Rendering Algorithm

  TODO: document the algorithm

  ## Findings

  What makes our rendering algorithm so difficult?

  - Dependency Injection requires a (direct) parent to be allow constructor injection, i.e. that injected dependencies
    are available in the constructor already. As a consequence a component tree must to be constructed from top to down.

  - The earliest time to evaluate `$$(MyComponent)`, is when it has been attached to an existing component.
    I.e., to run `MyComponent.render()` an instance of `MyComponent` is needed, which can only be created with an existing
    parent component.

  - In general, it is *not* possible to have a naturally descending rendering algorithm, i.e. a simple recursion calling
    `render()` and creating or updating Components on the way, preserving a simple stack-trace.
    Instead, it requires calling `render()` on one level, then doing comparisons with the existing tree
    to be able to reuse components, and then descend into the sub-tree.

  - If components are passed down via props, things get even more difficult.
    For example, consider a situation where components are passed via props:
    ```
      render($$) {
        return $$('div').append(
          $$(Wrapper, {
            foo: $$(MyComponent)
          })
        )
      }
    ```
    At the time when this component gets rendered, `MyComponent` can not be instantiated, as it is not known what `Wrapper`
    actually does with it.
    While the general approach is working from top-to-down, in this case it has a bottom-to-up nature, i.e., the child needs
    to be rendered to know what to do with the passed component.

    Particularly, this is problematic when the passed component has a reference:
    ```
      render($$) {
        return $$('div').append(
          $$(Wrapper, {
            foo: $$(MyComponent).ref('foo')
          })
        )
      }
    ```
    As nothing is known at the time of descending about the content of `Wrapper` the rendering algorithm
    can not tell that it ought to be preserved. For now, the correct way to deal with this situation is
    to use a reference for the wrapper as well:
    ```
      render($$) {
        return $$('div').append(
          $$(Wrapper, {
            foo: $$(MyComponent).ref('foo')
          }).ref('wrapper')
        )
      }
    ```

  ## TODO

  There is a lot of hacks in the current implementation, which should be cleaned up when we have time.
  - remove means to change props of VirtualComponent dynamically: currently, it is possible to do something like
    ```
      $$(MyComponent).attr({ "data-id": "foo" }).append('foo')
    ```
    ATM, `attr()` and `append()` represent a means to change the `props` of the component.
    This is pretty inconsistent and creates confusion about the responsibility for rendering the element.
    Only the Component's `render()` should be responsible for that. If necessary, children or anything like that must
    be passed down via props.
  - remove outlets: outlets are just another way to change props.
  - try to fuse `virtualComponent._content` into virtualComponent: ATM, `VirtualComponent` uses a `VirtualHTMLElement`
    instance to store the result of `render()`. This makes understanding the virtual tree after rendering difficult,
    as there is another layer via `virtualComponent._content`.
  - Rethink strategy for 'reusing' components: ATM we consider refs as the key indicator whether to preserve a component
    or not. This makes sense essentially, but could be extended to an opportunistic strategy where components are reused
    implicitly when they are at the right place.
*/
class RenderingEngine {

  constructor(options = {}) {
    this.elementFactory = options.elementFactory || DefaultDOMElement.createDocument('html');
  }

  _render(comp, oldProps, oldState) {
    // var t0 = Date.now()
    var vel = _createWrappingVirtualComponent(comp);
    var state = new RenderingEngine.State(this.elementFactory);
    if (oldProps) {
      state.setOldProps(vel, oldProps);
    }
    if (oldState) {
      state.setOldState(vel, oldState);
    }
    try {
      // capture: this calls the render() method of components, creating a virtual DOM
      // console.log('### capturing...')
      // let t0 = Date.now()
      _capture(state, vel, 'forceCapture');
      // console.log('### ... finished in %s ms', Date.now()-t0)

      // console.log('### rendering...')
      // t0 = Date.now()
      _render(state, vel);
      // console.log('### ... finished in %s ms', Date.now()-t0)

      _triggerUpdate(state, vel);

    } finally {
      state.dispose();
    }
    // console.log("RenderingEngine: finished rendering in %s ms", Date.now()-t0)
  }

  // this is used together with the incremental Component API
  _renderChild(comp, vel) {
    // HACK: to make this work with the rest of the implementation
    // we ingest a fake parent
    var state = new RenderingEngine.State(this.elementFactory);
    vel.parent = { _comp: comp };
    try {
      _capture(state, vel);
      _render(state, vel);
      return vel._comp
    } finally {
      state.dispose();
    }
  }
}


function _create(state, vel) {
  var comp = vel._comp;
  console.assert(!comp, "Component instance should not exist when this method is used.");
  var parent = vel.parent._comp;
  // making sure the parent components have been instantiated
  if (!parent) {
    parent = _create(state, vel.parent);
  }
  if (vel._isVirtualComponent) {
    console.assert(parent, "A Component should have a parent.");
    comp = new vel.ComponentClass(parent, vel.props);
    // HACK: making sure that we have the right props
    vel.props = comp.props;
    comp.__htmlConfig__ = vel._copyHTMLConfig();
  } else if (vel._isVirtualHTMLElement) {
    comp = new Component.Element(parent, vel);
  } else if (vel._isVirtualTextNode) {
    comp = new Component.TextNode(parent, vel);
  }
  if (vel._ref) {
    comp._ref = vel._ref;
  }
  if (vel._owner) {
    comp._owner = vel._owner._comp;
  }
  vel._comp = comp;
  return comp
}

function _capture(state, vel, forceCapture) {
  if (state.isCaptured(vel)) {
    return vel
  }
  // a captured VirtualElement has a component instance attached
  var comp = vel._comp;
  if (!comp) {
    comp = _create(state, vel);
    state.setNew(vel);
  }
  if (vel._isVirtualComponent) {
    var needRerender;
    // NOTE: forceCapture is used for the first entrance
    // from this.render(comp) where we want to fource capturing
    // as it has already been cleared that a rerender is necessary
    if (forceCapture) {
      needRerender = true;
    } else {
      // NOTE: don't ask shouldRerender if no element is there yet
      needRerender = !comp.el || comp.shouldRerender(vel.props, comp.state);
      comp.__htmlConfig__ = vel._copyHTMLConfig();
      state.setOldProps(vel, comp.props);
      state.setOldState(vel, comp.state);
      // updates prop triggering willReceiveProps
      comp._setProps(vel.props);
      if (!state.isNew(vel)) {
        state.setUpdated(vel);
      }
    }
    if (needRerender) {
      var context = new CaptureContext(vel);
      var content = comp.render(context.$$);
      if (!content || !content._isVirtualHTMLElement) {
        throw new Error("Component.render must return VirtualHTMLElement")
      }

      if (comp.__htmlConfig__) {
        content._mergeHTMLConfig(comp.__htmlConfig__);
      }
      content._comp = comp;
      vel._content = content;
      if (!state.isNew(vel) && comp.isMounted()) {
        state.setUpdated(vel);
      }
      // Mapping: map virtual elements to existing components based on refs
      _prepareVirtualComponent(state, comp, content);
      // Descending
      // TODO: only do this in DEBUG mode
      if (substanceGlobals.DEBUG_RENDERING) {
        // in this case we use the render() function as iterating function, where
        // $$ is a function which creates components and renders them recursively.
        // first we can create all element components that can be reached
        // without recursion
        var stack = content.children.slice(0);
        while (stack.length) {
          var child = stack.shift();
          if (state.isCaptured(child)) continue
          // virtual components are addressed via recursion, not captured here
          if (child._isVirtualComponent) continue
          if (!child._comp) {
            _create(state, child);
          }
          if (child._isVirtualHTMLElement && child.children.length > 0) {
            stack = stack.concat(child.children);
          }
          state.setCaptured(child);
        }
        state.setCaptured(content);
        // then we run comp.render($$) with a special $$ that captures VirtualComponent's
        // recursively
        var descendingContext = new DescendingContext(state, context);
        while (descendingContext.hasPendingCaptures()) {
          descendingContext.reset();
          comp.render(descendingContext.$$);
        }
      } else {
        // a VirtualComponent has its content as a VirtualHTMLElement
        // which needs to be captured recursively
        _capture(state, vel._content);
      }
    } else {
      state.setSkipped(vel);
    }
  } else if (vel._isVirtualHTMLElement) {
    for (var i = 0; i < vel.children.length; i++) {
      _capture(state, vel.children[i]);
    }
  }
  state.setCaptured(vel);
  return vel
}

function _render(state, vel) {
  if (state.isSkipped(vel)) return
  // console.log('... rendering', vel._ref)

  // before changes can be applied, a VirtualElement must have been captured
  // FIXME: with DEBUG_RENDERING we are having troubles with this assumption.
  // It happens when the rerendered component is having children injected from its parent.
  // Then the parent is no rerendered this, these injected components are not recaptured, and this assertion does not hold.
  // However, it seems not to be critical, as these components don't need to be rerendered
  // Still we should find a consistent way

  let comp = vel._comp;
  console.assert(comp && comp._isComponent, "A captured VirtualElement must have a component instance attached.");

  // VirtualComponents apply changes to its content element
  if (vel._isVirtualComponent) {
    _render(state, vel._content);

    // store refs and foreignRefs
    const context = vel._content._context;
    let refs = {};
    let foreignRefs = {};
    forEach(context.refs, (vel, ref) => {
      refs[ref] = vel._comp;
    });
    forEach(context.foreignRefs, (vel, ref) => {
      foreignRefs[ref] = vel._comp;
    });
    comp.refs = refs;
    comp.__foreignRefs__ = foreignRefs;
    return
  }

  // render the element
  if (!comp.el) {
    comp.el = _createElement(state, vel);
    comp.el._comp = comp;
  }
  _updateElement(comp, vel);

  // structural updates are necessary only for HTML elements (without innerHTML set)
  if (vel._isVirtualHTMLElement && !vel.hasInnerHTML()) {
    var newChildren = vel.children;
    var oldComp, virtualComp, newComp;
    var pos1 = 0; var pos2 = 0;

    // HACK: removing all childNodes that are not owned by a component
    // this happened in Edge every 1s. Don't know why.
    // With this implementation all external DOM mutations will be eliminated
    var oldChildren = [];
    comp.el.getChildNodes().forEach(function(node) {
      var childComp = node._comp;

      // TODO: to allow mounting a prerendered DOM element
      // we would need to allow to 'take ownership' instead of removing
      // the element. This being a special situation, we should only
      // do that when mounting

      // remove orphaned nodes and relocated components
      if (!childComp || state.isRelocated(childComp)) {
        comp.el.removeChild(node);
      } else {
        oldChildren.push(childComp);
      }
    });

    while(pos1 < oldChildren.length || pos2 < newChildren.length) {
      // skip detached components
      // Note: components get detached when preserved nodes
      // are found in a swapped order. Then the only way is
      // to detach one of them from the DOM, and reinsert it later at the new position
      do {
        oldComp = oldChildren[pos1++];
      } while (oldComp && (state.isDetached(oldComp)))

      virtualComp = newChildren[pos2++];
      // remove remaining old ones if no new one is left
      if (oldComp && !virtualComp) {
        while (oldComp) {
          _removeChild(state, comp, oldComp);
          oldComp = oldChildren[pos1++];
        }
        break
      }

      // Try to reuse TextNodes to avoid unnecesary DOM manipulations
      if (oldComp && oldComp.el.isTextNode() &&
          virtualComp && virtualComp._isVirtualTextNode &&
          oldComp.el.textContent === virtualComp.text ) {
        continue
      }

      if (!state.isRendered(virtualComp)) {
        _render(state, virtualComp);
      }

      newComp = virtualComp._comp;

      // ATTENTION: relocating a component does not update its context
      if (state.isRelocated(newComp)) {
        newComp._setParent(comp);
      }

      console.assert(newComp, 'Component instance should now be available.');
      // append remaining new ones if no old one is left
      if (virtualComp && !oldComp) {
        _appendChild(state, comp, newComp);
        continue
      }
      // Differential update
      else if (state.isMapped(virtualComp)) {
        // identity
        if (newComp === oldComp) {
          // no structural change
        } else {
          // the order of elements with ref has changed
          state.setDetached(oldComp);
          _removeChild(state, comp, oldComp);
          pos2--;
        }
      }
      else if (state.isMapped(oldComp)) {
        _insertChildBefore(state, comp, newComp, oldComp);
        pos1--;
      }
      else {
        // both elements are not mapped
        // TODO: we could try to reuse components if they are of same type
        // However, this needs a better mapping strategy, not only
        // based on refs.
        _replaceChild(state, comp, oldComp, newComp);
      }
    }
  }

  state.setRendered(vel);
}

function _triggerUpdate(state, vel) {
  if (vel._isVirtualComponent) {
    if (!state.isSkipped(vel)) {
      vel._content.children.forEach(_triggerUpdate.bind(null, state));
    }
    if (state.isUpdated(vel)) {
      vel._comp.didUpdate(state.getOldProps(vel), state.getOldState(vel));
    }
  } else if (vel._isVirtualHTMLElement) {
    vel.children.forEach(_triggerUpdate.bind(null, state));
  }
}

function _appendChild(state, parent, child) {
  parent.el.appendChild(child.el);
  _triggerDidMount(state, parent, child);
}

function _replaceChild(state, parent, oldChild, newChild) {
  parent.el.replaceChild(oldChild.el, newChild.el);
  if (!state.isDetached(oldChild)) {
    oldChild.triggerDispose();
  }
  _triggerDidMount(state, parent, newChild);
}

function _insertChildBefore(state, parent, child, before) {
  parent.el.insertBefore(child.el, before.el);
  _triggerDidMount(state, parent, child);
}

function _removeChild(state, parent, child) {
  parent.el.removeChild(child.el);
  if (!state.isDetached(child)) {
    child.triggerDispose();
  }
}

function _triggerDidMount(state, parent, child) {
  if (!state.isDetached(child) &&
      parent.isMounted() && !child.isMounted()) {
    child.triggerDidMount(true);
  }
}

/*
  Prepares a new virtual component by comparing it with
  the old version.

  It sets the _comp references in the new version where its ancestors
  can be mapped to corresponding virtual components in the old version.
*/
function _prepareVirtualComponent(state, comp, vc) {
  var newRefs = {};
  var foreignRefs = {};
  // TODO: iron this out. refs are stored on the context
  // though, it would be cleaner if they were on the VirtualComponent
  // Where vc._owner would need to be a VirtualComponent and not a
  // component.
  if (vc._context) {
    newRefs = vc._context.refs;
    foreignRefs = vc._context.foreignRefs;
  }
  var oldRefs = comp.refs;
  var oldForeignRefs = comp.__foreignRefs__;
  // map virtual components to existing ones
  forEach(newRefs, function(vc, ref) {
    var comp = oldRefs[ref];
    if (comp) _mapComponents(state, comp, vc);
  });
  forEach(foreignRefs, function(vc, ref) {
    var comp = oldForeignRefs[ref];
    if (comp) _mapComponents(state, comp, vc);
  });
}

/*
  This tries to map the virtual component to existing component instances
  by looking at the old and new refs, making sure that the element type is
  compatible.
  This is then applied to the ancestors leading to an implicit
  mapping of parent elements, which makes
*/

function _mapComponents(state, comp, vc) {
  if (!comp && !vc) return true
  if (!comp || !vc) return false
  // Stop if one them has been mapped already
  // or the virtual element has its own component already
  // or if virtual element and component do not match semantically
  // Note: the owner component is mapped at very first, so this
  // recursion will stop at the owner at the latest.
  if (state.isMapped(vc) || state.isMapped(comp)) {
    return vc._comp === comp
  }
  if (vc._comp) {
    if (vc._comp === comp) {
      state.setMapped(vc);
      state.setMapped(comp);
      return true
    } else {
      return false
    }
  }
  if (!_isOfSameType(comp, vc)) {
    return false
  }

  vc._comp = comp;
  state.setMapped(vc);
  state.setMapped(comp);

  var canMapParent;
  var parent = comp.getParent();
  if (vc.parent) {
    canMapParent = _mapComponents(state, parent, vc.parent);
  }
  // to be able to support implicit retaining of elements
  // we need to propagate mapping through the 'preliminary' parent chain
  // i.e. not taking the real parents as rendered, but the Components into which
  // we have passed children (via vel.append() or vel.outlet().append())
  else if (vc._preliminaryParent) {
    while (parent && parent._isElementComponent) {
      parent = parent.getParent();
    }
    canMapParent = _mapComponents(state, parent, vc._preliminaryParent);
  }
  if (!canMapParent) {
    state.setRelocated(vc);
    state.setRelocated(comp);
  }
  return canMapParent
}

function _isOfSameType(comp, vc) {
  return (
    (comp._isElementComponent && vc._isVirtualHTMLElement) ||
    (comp._isComponent && vc._isVirtualComponent && comp.constructor === vc.ComponentClass) ||
    (comp._isTextNodeComponent && vc._isVirtualTextNode)
  )
}

function _createElement(state, vel) {
  var el;
  if (vel._isVirtualTextNode) {
    el = state.elementFactory.createTextNode(vel.text);
  } else {
    el = state.elementFactory.createElement(vel.tagName);
  }
  return el
}

function _updateElement(comp, vel) {
  if (comp._isTextNodeComponent) {
    comp.setTextContent(vel.text);
    return
  }
  var el = comp.el;
  console.assert(el, "Component's element should exist at this point.");
  var tagName = el.getTagName();
  if (vel.tagName.toLowerCase() !== tagName) {
    el.setTagName(vel.tagName);
  }
  _updateHash({
    oldHash: el.getAttributes(),
    newHash: vel.getAttributes(),
    update: function(key, val) {
      el.setAttribute(key, val);
    },
    remove: function(key) {
      el.removeAttribute(key);
    }
  });
  _updateHash({
    oldHash: el.htmlProps,
    newHash: vel.htmlProps,
    update: function(key, val) {
      el.setProperty(key, val);
    },
    remove: function(key) {
      el.removeProperty(key);
    }
  });
  _updateListeners({
    el: el,
    oldListeners: el.getEventListeners(),
    newListeners: vel.getEventListeners()
  });

  // special treatment of HTML elements having custom innerHTML
  if (vel.hasInnerHTML()) {
    if (!el._hasInnerHTML) {
      el.empty();
      el.setInnerHTML(vel.getInnerHTML());
    } else {
      var oldInnerHTML = el.getInnerHTML();
      var newInnerHTML = vel.getInnerHTML();
      if (oldInnerHTML !== newInnerHTML) {
        el.setInnerHTML(newInnerHTML);
      }
    }
    el._hasInnerHTML = true;
  }
}

function _updateHash(args) {
  var newHash = args.newHash;
  var oldHash = args.oldHash || {};
  var updatedKeys = {};
  var update = args.update;
  var remove = args.remove;
  var key;
  for (key in newHash) {
    if (newHash.hasOwnProperty(key)) {
      var oldVal = oldHash[key];
      var newVal = newHash[key];
      updatedKeys[key] = true;
      if (oldVal !== newVal) {
        update(key, newVal);
      }
    }
  }
  for (key in oldHash) {
    if (oldHash.hasOwnProperty(key) && !updatedKeys[key]) {
      remove(key);
    }
  }
}

function _updateListeners(args) {
  var el = args.el;
  // NOTE: considering the low number of listeners
  // it is quicker to just remove all
  // and add again instead of computing the minimal update
  var newListeners = args.newListeners || [];
  el.removeAllEventListeners();
  for (var i=0; i<newListeners.length;i++) {
    el.addEventListener(newListeners[i]);
  }
}


/*
  Descending Context Used by RenderingEngine
*/

class DescendingContext {
  constructor(state, captureContext) {
    this.state = state;
    this.owner = captureContext.owner;
    this.refs = {};
    this.foreignRefs = {};
    this.elements = captureContext.elements;
    this.pos = 0;
    this.updates = captureContext.components.length;
    this.remaining = this.updates;

    this.$$ = this._createComponent.bind(this);
  }

  _createComponent() {
    var state = this.state;
    var vel = this.elements[this.pos++];
    // only capture VirtualComponent's with a captured parent
    // all others have been captured at this point already
    // or will either be captured by a different owner
    if (!state.isCaptured(vel) && vel._isVirtualComponent &&
         vel.parent && state.isCaptured(vel.parent)) {
      _capture(state, vel);
      this.updates++;
      this.remaining--;
    }
    // Note: we return a new VirtualElement so that the render method does work
    // as expected.
    // TODO: instead of creating a new VirtualElement each time, we could return
    // an immutable wrapper for the already recorded element.
    vel = VirtualElement.createElement.apply(this, arguments);
    // these variables need to be set make the 'ref()' API work
    vel._context = this;
    vel._owner = this.owner;
    // Note: important to deactivate these methods as otherwise the captured
    // element will be damaged when calling el.append()
    vel._attach = function() {};
    vel._detach = function() {};
    return vel
  }

  hasPendingCaptures() {
    return this.updates > 0 && this.remaining > 0
  }

  reset() {
    this.pos = 0;
    this.updates = 0;
    this.refs = {};
  }

  _ancestorsReady(vel) {
    while (vel) {
      if (this.state.isCaptured(vel) ||
          // TODO: iron this out
          vel === this.owner || vel === this.owner._content) {
        return true
      }
      vel = vel.parent;
    }
    return false
  }


}

RenderingEngine._internal = {
  _capture: _capture,
  _wrap: _createWrappingVirtualComponent,
};

class CaptureContext {
  constructor(owner) {
    this.owner = owner;
    this.refs = {};
    this.foreignRefs = {};
    this.elements = [];
    this.components = [];
    this.$$ = this._createComponent.bind(this);
    this.$$.capturing = true;
  }

  _createComponent() {
    var vel = VirtualElement.createElement.apply(this, arguments);
    vel._context = this;
    vel._owner = this.owner;
    if (vel._isVirtualComponent) {
      // virtual components need to be captured recursively
      this.components.push(vel);
    }
    this.elements.push(vel);
    return vel
  }
}


function _createWrappingVirtualComponent(comp) {
  var vel = new VirtualElement.Component(comp.constructor);
  vel._comp = comp;
  if (comp.__htmlConfig__) {
    vel._mergeHTMLConfig(comp.__htmlConfig__);
  }
  return vel
}

RenderingEngine.createContext = function(comp) {
  var vel = _createWrappingVirtualComponent(comp);
  return new CaptureContext(vel)
};


class RenderingState {

  constructor(elementFactory) {
    this.elementFactory = elementFactory;
    this.poluted = [];
    this.id = "__"+uuid();
  }

  dispose() {
    var id = this.id;
    this.poluted.forEach(function(obj) {
      delete obj[id];
    });
  }

  set(obj, key, val) {
    var info = obj[this.id];
    if (!info) {
      info = {};
      obj[this.id] = info;
      this.poluted.push(obj);
    }
    info[key] = val;
  }

  get(obj, key) {
    var info = obj[this.id];
    if (info) {
      return info[key]
    }
  }

  setMapped(c) {
    this.set(c, 'mapped', true);
  }


  isMapped(c) {
    return Boolean(this.get(c, 'mapped'))
  }

  setRelocated(c) {
    this.set(c, 'relocated', true);
  }

  isRelocated(c) {
    return Boolean(this.get(c, 'relocated'))
  }

  setDetached(c) {
    this.set(c, 'detached', true);
  }

  isDetached(c) {
    return Boolean(this.get(c, 'detached'))
  }

  setCaptured(vc) {
    this.set(vc, 'captured', true);
  }

  isCaptured(vc) {
    return Boolean(this.get(vc, 'captured'))
  }

  setNew(vc) {
    this.set(vc, 'created', true);
  }

  isNew(vc) {
    return Boolean(this.get(vc, 'created'))
  }

  setUpdated(vc) {
    this.set(vc, 'updated', true);
  }

  isUpdated(vc) {
    return Boolean(this.get(vc, 'updated'))
  }

  setSkipped(vc) {
    this.set(vc, 'skipped', true);
  }

  isSkipped(vc) {
    return Boolean(this.get(vc, 'skipped'))
  }

  setRendered(vc) {
    this.set(vc, 'rendered', true);
  }

  isRendered(vc) {
    return Boolean(this.get(vc, 'rendered'))
  }

  setOldProps(vc, oldProps) {
    this.set(vc, 'oldProps', oldProps);
  }

  getOldProps(vc) {
    return this.get(vc, 'oldProps')
  }

  setOldState(vc, oldState) {
    this.set(vc, 'oldState', oldState);
  }

  getOldState(vc) {
    return this.get(vc, 'oldState')
  }
}

RenderingEngine.State = RenderingState;

/**
  A light-weight component implementation inspired by
  [React](https://facebook.github.io/react/) and [Ember](http://emberjs.com/).
  In contrast to the large frameworks it does much less things automagically in
  favour of synchronous rendering and a minimalistic life-cycle. It also
  provides *up-tree* communication and *dependency injection*.

  ### Why synchronous rendering?

  Synchronous rendering, while it may *seem* less performant, is necessary
  because substance must render the model, after it has changed before the next
  change is triggered by the user.

  Asynchronous rendering as it exists in React means that the UI will
  eventually *catch* up to changes in the model. This is not acceptable in
  substance because substance plays with contenteditable and thus, cursor
  positions, etc are maintained in the browser's DOM. If we went the async way,
  the cursor in the DOM would be briefly inconsistent with the cursor in the
  model. In this brief window, changes triggered by the user would be impossible
  to apply.

  ### Concepts:

  - `props` are provided by a parent component.  An initial set of properties is provided
  via constructor. After that, the parent component can call `setProps` or `extendProps`
  to update these properties which triggers rerendering if the properties change.

  - `state` is a set of flags and values which are used to control how the component
  gets rendered given the current props. Using `setState` the component can change
  its internal state, which leads to a rerendering if the state changes.
  Prefer using `extendState` rather than `setState`.

    Normally, a component maintains its own state. It isn't recommended that a
  parent pass in or update state. If you find the need for this, you should be
  looking at `props`.

    State would be useful in situations where the component itself controls some
  aspect of rendering. Eg. whether a dropdown is open or not could be a state
  within the dropdown component itself since no other component needs to know
  it.

  - A child component with a `ref` id will be reused on rerender. All others will be
  wiped and rerender from scratch. If you want to preserve a grand-child (or lower), then
  make sure that all anchestors have a ref id. After rendering the child will be
  accessible via `this.refs[ref]`.

  - A component can send actions via `send` which are bubbled up through all parent
  components until one handles it. A component declares that it can handle an
  action by calling the `handleActions` method on itself in the constructor or
  the `didUpdate` lifecycle hook.

  ### Lifecycle hooks

  The {@link RenderingEngine} triggers a set of hooks for you to define behavior
  in various stages of the rendering cycle. The names are pretty self
  explanatory. If in doubt, please check out the method documentation below.

  1. {@link Component#didMount}
  1. {@link Component#didUpdate}
  1. {@link Component#dispose}
  1. {@link Component#willReceiveProps}
  1. {@link Component#willUpdateState}

  @implements EventEmitter

  @example

  Define a component:

  ```
  class HelloMessage extends Component {
    render() {
      return $$('div').append(
        'Hello ',
        this.props.name
      )
    }
  }
  ```

  And mount it to a DOM Element:

  ```
  HelloMessage.mount({name: 'John'}, document.body)
  ```
*/
class Component extends EventEmitter {
  /**
    Construcutor is only used internally.

    @param {Component} parent The parent component
    @param {Object} props     Properties against which this class must
                              be rendered the first time.
  */
  constructor(parent, props = {}, options = {}) {
    super();

    // TODO: it turned out that the signature is sub-optimal
    // w.r.t. `parent`. Creating a root component allowing for manual dependency injection
    // we could change to `new Component(props, options)`
    // with options `parent` and `context`
    // Also, the parent-child relation could be inconsistent with the actual elements, which should be checked.

    this.parent = (parent && parent._isComponent) ? parent : null;

    // EXPERIMENTAL: used for attaching to existing/pre-rendered element
    this.el = options.el;

    // context from parent (dependency injection) or if given via options
    // the latter is a rather EXPERIMENTAL feature only used
    let context = options.context ? options.context : this._getContext() || {};
    this.context = context;
    Object.freeze(this.context);

    // used for rerendering and can be used by components for incremental rendering
    // Note: usually it is inherited from the parent. In case of root components
    // it can be provided via context or options
    this.renderingEngine = (parent && parent.renderingEngine) || context.renderingEngine || options.renderingEngine || new RenderingEngine();

    // HACK: to allow that ElementComponent and TextComponent can derive from Component
    // we need to skip the initialization of the rest
    if (this._SKIP_COMPONENT_INIT) return

    this.__id__ = uuid();

    // store for ref'd child components
    this.refs = {};
    // HACK: a temporary solution to handle refs owned by an ancestor
    // is to store them here as well, so that we can map virtual components
    // efficiently
    this.__foreignRefs__ = {};

    // action handlers added via `handleAction()` are stored here
    this._actionHandlers = {};

    // setting props without triggering willReceiveProps
    this.props = props;
    Object.freeze(this.props);

    // initializing state
    this.state = this.getInitialState() || {};
    Object.freeze(this.state);
  }

  getId() {
    return this.__id__
  }

  setId() {
    throw new Error("'id' is readonly")
  }

  /**
    Provides the context which is delivered to every child component. Override
    if you want to provide your own child context. Child context is available to
    all components rendered from within this component's render method as
    `this.context`.

    @example

    ```
    class A extends Component {
    ...
      getChildContext() {
        // Optional, but useful to merge super's context
        return Object.assign({}, super.getChildContext(), {foo: 'bar'})
      }

      render($$) {
        return $$(B)
      }
    }

    class B extends Component {
      render($$) {
        // this.context.foo is available here
      }
    }
    ```
    Component

    @return {Object} the child context
  */
  getChildContext() {
    return this.childContext || {}
  }

  /**
    Override this within your component to provide the initial state for the
    component. This method is internally called by the
    {@link RenderingEngine} and the state defined here is made available to
    the {@link Component#render} method as this.state.

    @return {Object} the initial state
  */
  getInitialState() {
    return {}
  }

  /**
    Provides the parent of this component.

    @return {Component} the parent component or null if this component does not have a parent.
  */
  getParent() {
    return this.parent
  }

  /**
    Get the top-most Component. This the component mounted using
    {@link ui/Component.mount}
    @return {Component} The root component
  */
  getRoot() {
    var comp = this;
    var parent = comp;
    while (parent) {
      comp = parent;
      parent = comp.getParent();
    }
    return comp
  }

  getNativeElement() {
    return this.el.getNativeElement()
  }

  /**
    Short hand for using labelProvider API

    @example

    ```
    render($$) {
      let el = $$('div').addClass('sc-my-component')
      el.append(this.getLabel('first-name'))
      return el
    }
    ```
  */
  getLabel(name) {
    let labelProvider = this.context.labelProvider;
    if (!labelProvider) throw new Error('Missing labelProvider.')
    return labelProvider.getLabel(name)
  }

  /**
    Get a component class for the component name provided. Use this within the
    render method to render children nodes.

    @example

    ```
    render($$) {
      let el = $$('div').addClass('sc-my-component')
      let caption = this.props.node.getCaption() // some method that returns a node
      let CaptionClass = this.getComponent(caption.type)
      el.append($$(CaptionClass, {node: caption}))
      return el
    }
    ```

    @param  {String} componentName The component's registration name
    @param  {Boolean} maybe if `true` then does not throw when no Component is found
    @return {Class}                The ComponentClass
  */
  getComponent(componentName, maybe) {
    let componentRegistry = this.getComponentRegistry();
    if (!componentRegistry) throw new Error('Missing componentRegistry.')
    const ComponentClass = componentRegistry.get(componentName);
    if (!maybe && !ComponentClass) {
      throw new Error('No Component registered with name ' + componentName)
    }
    return ComponentClass
  }

  getComponentRegistry() {
    return this.props.componentRegistry || this.context.componentRegistry
  }

  getFlow() {
    return this.context.flow
  }

  /**
    Render the component.

    ATTENTION: this does not create a DOM presentation but
    a virtual representation which is compiled into a DOM element later.

    Every Component should override this method.

    @param {Function} $$ method to create components
    @return {VirtualElement} VirtualElement created using {@param $$}
  */
  render($$) {
    /* istanbul ignore next */
    return $$('div')
  }

  /**
    Mount a component to the DOM.

    @example

    ```
    var app = Texture.mount({
      configurator: configurator,
      documentId: 'elife-15278'
    }, document.body)
    ```
  */
  mount(el) {
    if (!el) {
      throw new Error('Element is required.')
    }
    el = DefaultDOMElement.wrap(el);
    // Makes sure a new element is created for the component
    this.el = null;
    this.renderingEngine = new RenderingEngine({ elementFactory: el.getOwnerDocument() });
    this._render();
    el.appendChild(this.el);
    if (el.isInDocument()) {
      this.triggerDidMount(true);
    }
    return this
  }

  /**
    Determines if Component should be rendered again using {@link ui/Component#rerender}
    after changing props. For comparisons, you can use `this.props` and
    `newProps`.

    The default implementation simply returns true.

    @param {Object} newProps The new props being applied to this component.
    @param {Object} newState The new state being applied
    @return a boolean indicating whether rerender() should be run.
  */
  shouldRerender(newProps, newState) { // eslint-disable-line
    return true
  }

  /**
    Rerenders the component.

    Call this to manually trigger a rerender.
  */
  rerender() {
    this._rerender(this.props, this.state);
  }

  _rerender(oldProps, oldState) {
    this._render(oldProps, oldState);
    // when this component is not mounted still trigger didUpdate()
    if (!this.isMounted()) {
      this.didUpdate(oldProps, oldState);
    }
  }

  _render(oldProps, oldState) {
    if (this.__isRendering__) {
      throw new Error('Component is rendering already.')
    }
    this.__isRendering__ = true;
    try {
      this.renderingEngine._render(this, oldProps, oldState);
    } finally {
      delete this.__isRendering__;
    }
  }

  /**
    Triggers didMount handlers recursively.

    Gets called when using `component.mount(el)` on an element being
    in the DOM already. Typically this is done for a root component.

    If this is not possible because you want to do things differently, make sure
    you call 'component.triggerDidMount()' on root components.

    @param isMounted an optional param for optimization, it's used mainly internally
    @private
    @example

    ```
    var frag = document.createDocumentFragment()
    var comp = MyComponent.mount(frag)
    ...
    $('body').append(frag)
    comp.triggerDidMount()
    ```
  */
  triggerDidMount() {
    // TODO: discuss and leave the decision here as a statement
    // while rendering is done bottom up, IMO  it makes
    // sense to propagate didMount() from top to down.

    // To prevent from multiple calls to didMount, which can happen under
    // specific circumstances we use a guard.
    if (!this.__isMounted__) {
      this.__isMounted__ = true;
      this.didMount();
    }
    // Trigger didMount for the children first
    this.getChildren().forEach(function(child) {
      // We pass isMounted=true to save costly calls to Component.isMounted
      // for each child / grandchild
      child.triggerDidMount(true);
    });
  }

  /**
    Called when the element is inserted into the DOM. Typically, you can use
    this to set up subscriptions to changes in the document or in a node of
    your interest.

    Remember to unsubscribe from all changes in the {@link ui/Component#dispose}
    method otherwise listeners you have attached may be called without a context.

    @example

    ```javascript
    class Foo extends Component {
      didMount() {
        this.context.editorSession.onRender('document', this.rerender, this, {
          path: [this.props.node.id, 'label']
        })
      }

      dispose() {
        this.context.editorSession.off(this)
      }
    }
    ```

    Make sure that you call `component.mount(el)` using an element
    which is already in the DOM.

    ```javascript
    var component = new MyComponent()
    component.mount($('body')[0])
    ```
  */
  didMount() {}

  /**
    Hook which is called after state or props have been updated and the implied
    rerender is completed.
  */
  didUpdate() {}

  /**
    @return {boolean} indicating if this component has been mounted
   */
  isMounted() {
    return this.__isMounted__
  }

  /**
   Triggers dispose handlers recursively.

   @private
  */
  triggerDispose() {
    this.getChildren().forEach(function(child) {
      child.triggerDispose();
    });
    this.dispose();
    this.__isMounted__ = false;
  }

  /**
    A hook which is called when the component is unmounted, i.e. removed from DOM,
    hence disposed. See {@link ui/Component#didMount} for example usage.

    Remember to unsubscribe all change listeners here.
  */
  dispose() {}

  /*
    Attention: this is used when a preserved component is relocated.
    E.g., rendered with a new parent.
  */
  _setParent(newParent) {
    this.parent = newParent;
    this.context = this._getContext() || {};
    Object.freeze(this.context);
  }

  /**
    Send an action request to the parent component, bubbling up the component
    hierarchy until an action handler is found.

    @param action the name of the action
    @param ... arbitrary number of arguments
    @returns {Boolean} true if the action was handled, false otherwise
    @example
  */
  send(action) {
    var comp = this;
    while(comp) {
      if (comp._actionHandlers && comp._actionHandlers[action]) {
        comp._actionHandlers[action].apply(comp, Array.prototype.slice.call(arguments, 1));
        return true
      }
      comp = comp.getParent();
    }
    console.warn('Action', action, 'was not handled.');
    return false
  }

  /**
    Define action handlers. Call this during construction/initialization of a component.
    @param {Object} actionHandler  An object where the keys define the handled
        actions and the values define the handler to be invoked.

    These handlers are automatically removed once the Component is disposed, so
    there is no need to unsubscribe these handlers in the {@link ui/Component#dispose}
    hook.

    @example

    ```
    class MyComponent extends Component {
      constructor(...args) {
        super(...args)
        this.handleActions({
         'openPrompt': this.openPrompt,
         'closePrompt': this.closePrompt
        })
      }
    }
    ```
  */
  handleActions(actionHandlers) {
    forEach(actionHandlers, function(method, actionName) {
      this.handleAction(actionName, method);
    }.bind(this));
    return this
  }

  /**
    Define an action handler. Call this during construction/initialization of a component.

    @param {String} action name
    @param {Functon} a function of this component.
  */
  handleAction(name, handler) {
    if (!name || !handler || !isFunction$1(handler)) {
      throw new Error('Illegal arguments.')
    }
    handler = handler.bind(this);
    this._actionHandlers[name] = handler;
  }

  /**
    Get the current component state

    @return {Object} the current state
  */
  getState() {
    return this.state
  }

  /**
    Sets the state of this component, potentially leading to a rerender. It is
    better practice to use {@link ui/Component#extendState}. That way, the code
    which updates state only updates part relevant to it.

    Eg. If you have a Component that has a dropdown open state flag and another
    enabled/disabled state flag for a node in the dropdown, you want to isolate
    the pieces of your code making the two changes. The part of your code
    opening and closing the dropdown should not also automatically change or
    remove the enabled flag.

    Note: Usually this is used by the component itself.
    @param {object} newState an object with a partial update.
  */
  setState(newState) {
    var oldProps = this.props;
    var oldState = this.state;
    // Note: while setting props it is allowed to call this.setState()
    // which will not lead to an extra rerender
    var needRerender = !this.__isSettingProps__ &&
      this.shouldRerender(this.getProps(), newState);
    // triggering this to provide a possibility to look at old before it is changed
    this.willUpdateState(newState);
    this.state = newState || {};
    Object.freeze(this.state);
    if (needRerender) {
      this._rerender(oldProps, oldState);
    } else if (!this.__isSettingProps__) {
      this.didUpdate(oldProps, oldState);
    }
  }

  /**
    This is similar to `setState()` but extends the existing state instead of
    replacing it.

    @param {object} newState an object with a partial update.
  */
  extendState(newState) {
    newState = extend({}, this.state, newState);
    this.setState(newState);
  }

  /**
    Called before state is changed.
  */
  willUpdateState(newState) { // eslint-disable-line
  }

  /**
    Get the current properties

    @return {Object} the current state
  */
  getProps() {
    return this.props
  }

  /**
    Sets the properties of this component, potentially leading to a rerender.

    @param {object} an object with properties
  */
  setProps(newProps) {
    var oldProps = this.props;
    var oldState = this.state;
    var needRerender = this.shouldRerender(newProps, this.state);
    this._setProps(newProps);
    if (needRerender) {
      this._rerender(oldProps, oldState);
    } else {
      this.didUpdate(oldProps, oldState);
    }
  }

  _setProps(newProps) {
    newProps = newProps || {};
    // set a flag so that this.setState() can omit triggering render
    this.__isSettingProps__ = true;
    try {
      this.willReceiveProps(newProps);
      this.props = newProps || {};
      Object.freeze(newProps);
    } finally {
      delete this.__isSettingProps__;
    }
  }

  /**
    Extends the properties of the component, without necessarily leading to a
    rerender.

    @param {object} an object with properties
  */
  extendProps(updatedProps) {
    var newProps = extend({}, this.props, updatedProps);
    this.setProps(newProps);
  }

  /**
    Hook which is called before properties are updated. Use this to dispose objects which will be replaced when properties change.

    For example you can use this to derive state from props.
    @param {object} newProps
  */
  willReceiveProps(newProps) { // eslint-disable-line
  }

  getTextContent() {
    if (this.el) {
      return this.el.textContent
    }
  }

  get textContent() {
    return this.getTextContent()
  }

  getInnerHTML() {
    if (this.el) {
      return this.el.getInnerHTML()
    }
  }

  get innerHTML() {
    return this.getInnerHTML()
  }

  getOuterHTML() {
    if (this.el) {
      return this.el.getOuterHTML()
    }
  }

  get outerHTML() {
    return this.getOuterHTML()
  }

  getAttribute(name) {
    if (this.el) {
      return this.el.getAttribute(name)
    }
  }

  setAttribute(name, val) {
    if (this.el) {
      this.el.setAttribute(name, val);
    }
    return this
  }

  getProperty(name) {
    if (this.el) {
      return this.el.getProperty(name)
    }
  }

  setProperty(name, val) {
    if (this.el) {
      this.el.setProperty(name, val);
    }
    return this
  }

  hasClass(name) {
    if (this.el) {
      return this.el.hasClass(name)
    }
  }

  addClass(name) {
    if (this.el) {
      this.el.addClass(name);
    }
    return this
  }

  removeClass(name) {
    if (this.el) {
      this.el.removeClass(name);
    }
    return this
  }

  getStyle(name) {
    if (this.el) {
      return this.el.getStyle(name)
    }
  }

  setStyle(name, val) {
    if (this.el) {
      return this.el.setStyle(name, val)
    }
    return this
  }

  getValue() {
    if (this.el) {
      return this.el.getValue()
    }
  }

  setValue(val) {
    if (this.el) {
      this.el.setValue(val);
    }
    return this
  }

  getChildCount() {
    if (!this.el) return 0
    return this.el.getChildCount()
  }

  get childNodes() {
    return this.getChildNodes()
  }

  getChildNodes() {
    if (!this.el) return []
    var childNodes = this.el.getChildNodes();
    childNodes = childNodes.map(_unwrapComp).filter(Boolean);
    return childNodes
  }

  getChildren() {
    if (!this.el) return []
    var children = this.el.getChildren();
    children = children.map(_unwrapComp).filter(Boolean);
    return children
  }

  getChildAt(pos) {
    var node = this.el.getChildAt(pos);
    return _unwrapCompStrict(node)
  }

  find(cssSelector) {
    var el = this.el.find(cssSelector);
    return _unwrapComp(el)
  }

  findAll(cssSelector) {
    var els = this.el.findAll(cssSelector);
    return els.map(_unwrapComp).filter(Boolean)
  }

  appendChild(child) {
    this.insertAt(this.getChildCount(), child);
  }

  insertAt(pos, childEl) {
    if (isString$1(childEl)) {
      childEl = new VirtualElement.TextNode(childEl);
    }
    if (!childEl._isVirtualElement) {
      throw new Error('Invalid argument: "child" must be a VirtualElement.')
    }
    var child = this.renderingEngine._renderChild(this, childEl);
    this.el.insertAt(pos, child.el);
    _mountChild(this, child);
  }

  removeAt(pos) {
    var childEl = this.el.getChildAt(pos);
    if (childEl) {
      var child = _unwrapCompStrict(childEl);
      _disposeChild(child);
      this.el.removeAt(pos);
    }
  }

  removeChild(child) {
    if (!child || !child._isComponent) {
      throw new Error('removeChild(): Illegal arguments. Expecting a Component instance.')
    }
    // TODO: remove ref from owner
    _disposeChild(child);
    this.el.removeChild(child.el);
  }

  replaceChild(oldChild, newChild) {
    if (!newChild || !oldChild ||
        !newChild._isComponent || !oldChild._isComponent) {
      throw new Error('replaceChild(): Illegal arguments. Expecting BrowserDOMElement instances.')
    }
    // Attention: Node.replaceChild has weird semantics
    _disposeChild(oldChild);
    this.el.replaceChild(newChild.el, oldChild.el);
    if (this.isMounted()) {
      newChild.triggerDidMount(true);
    }
  }

  // ATTENTION: we had problems here, that using
  // component.el.empty() instead of component.empty()
  // did cause the children not to dispose(), which is maybe
  // impossible to achieve.
  // TODO: Thus we may consider to rename it, or take
  // other measure to warn the the user about this problem
  empty() {
    this._clear();
    return this
  }

  _clear() {
    let el = this.el;
    if (el) {
      this.getChildNodes().forEach(function(child) {
        _disposeChild(child);
      });
      el.empty();
    }
    this.refs = {};
    this.__foreignRefs__ = {};
  }

  remove() {
    _disposeChild(this);
    this.el.remove();
  }

  addEventListener() {
    throw new Error("Not supported.")
  }

  removeEventListener() {
    throw new Error("Not supported.")
  }

  insertBefore() {
    throw new Error("Not supported.")
  }

  click() {
    if (this.el) {
      this.el.click();
    }
  }

  getComponentPath() {
    let path = [];
    let comp = this;
    while (comp) {
      path.unshift(comp);
      comp = comp.getParent();
    }
    return path
  }

  _getContext() {
    var context = {};
    var parent = this.getParent();
    if (parent) {
      context = extend(context, parent.context);
      if (parent.getChildContext) {
        return extend(context, parent.getChildContext())
      }
    }
    return context
  }

}

Component.prototype._isComponent = true;

Component.prototype.attr = DOMElement.prototype.attr;

Component.prototype.htmlProp = DOMElement.prototype.htmlProp;

Component.prototype.val = DOMElement.prototype.val;

Component.prototype.css = DOMElement.prototype.css;

Component.prototype.text = DOMElement.prototype.text;

Component.prototype.append = DOMElement.prototype.append;

Component.unwrap = _unwrapComp;

Component.render = function(props) {
  props = props || {};
  var ComponentClass = this;
  var comp = new ComponentClass(null, props);
  comp._render();
  return comp
};

Component.mount = function(props, el) {
  if (arguments.length === 1) {
    el = props;
    props = {};
  }
  if (!el) throw new Error("'el' is required.")
  if (isString$1(el)) {
    var selector = el;
    if (platform.inBrowser) {
      el = window.document.querySelector(selector);
    } else {
      throw new Error("This selector is not supported on server side.")
    }
  }
  el = new DefaultDOMElement.wrap(el);
  const ComponentClass = this;
  let comp = new ComponentClass(null, props);
  comp.mount(el);
  return comp
};

Component.getComponentForDOMElement = function(el) {
  return _unwrapComp(el)
};

Component.unwrapDOMElement = function(el) {
  console.warn('DEPRECATED: Use Component.getComponentForDOMElement');
  return Component.getComponentForDOMElement(el)
};

Component.getComponentFromNativeElement = function(nativeEl) {
  // while it sounds strange to wrap a native element
  // first, it makes sense after all, as DefaultDOMElement.wrap()
  // provides the DOMElement instance of a previously wrapped native element.
  return _unwrapComp(DefaultDOMElement.wrap(nativeEl))
};

// NOTE: this is used for incremental updates only
function _disposeChild(child) {
  child.triggerDispose();
  if (child._owner && child._ref) {
    console.assert(child._owner.refs[child._ref] === child, "Owner's ref should point to this child instance.");
    delete child._owner.refs[child._ref];
  }
}

// NOTE: this is used for incremental updates only
function _mountChild(parent, child) {
  if (parent.isMounted()) {
    child.triggerDidMount(true);
  }
  if (child._owner && child._ref) {
    child._owner.refs[child._ref] = child;
  }
}

// NOTE: we keep a reference to the component in all DOMElement instances
function _unwrapComp(el) {
  if (el) {
    if (!el._isDOMElement) el = DefaultDOMElement.unwrap(el);
    if (el) return el._comp
  }
}

function _unwrapCompStrict(el) {
  let comp = _unwrapComp(el);
  if (!comp) throw new Error("Expecting a back-link to the component instance.")
  return comp
}


class ElementComponent extends Component {

  constructor(parent) {
    super(parent);
  }

}

ElementComponent.prototype._isElementComponent = true;
ElementComponent.prototype._SKIP_COMPONENT_INIT = true;

class TextNodeComponent extends Component {

  constructor(parent) {
    super(parent);
  }

  setTextContent(text) {
    if (!this.el) {
      throw new Error('Component must be rendered first.')
    }
    if (this.el.textContent !== text) {
      this.el.textContent = text;
    }
  }

  getChildNodes() {
    return []
  }

  getChildren() {
    return []
  }

}

TextNodeComponent.prototype._isTextNodeComponent = true;
TextNodeComponent.prototype._SKIP_COMPONENT_INIT = true;

Component.Element = ElementComponent;
Component.TextNode = TextNodeComponent;

/*
  Each time a resource node is created an automatic fetch is triggered. This
  happens only once. We may want to adapt other fetching strategies as well.

  TODO: Resources would be problematic in a realtime scenario atm, as all
  collaborators would trigger a fetch of the resource, when it should only
  be done by the user who created the resource explicitly
*/

class ResourceManager {

  constructor(editorSession, context) {
    this.editorSession = editorSession;
    this.context = context;
    this.editorSession.onRender('document', this._onDocumentChange, this);
  }

  dispose() {
    this.editorSession.off(this);
  }

  /*
    Trigger fetch of a given resource
  */
  triggerFetch(resource) {
    resource.fetchPayload(this.context, (err, props) => {
      if (err) {
        this._updateNode(resource.id, {
          errorMessage: err.toString()
        });
      } else {
        this._updateNode(resource.id, props);
      }
    });
  }

  _onDocumentChange(change) {
    let doc = this.editorSession.getDocument();
    forEach(change.created, (node) => {
      node = doc.get(node.id);
      if (node.constructor.isResource) {
        setTimeout(() => {
          this.triggerFetch(node);
        });
      }
    });
  }

  /*
      Fill in node payload
  */
  _updateNode(nodeId, props) {
    let editorSession = this.editorSession;
    editorSession.transaction((tx) => {
      forEach(props, (val, key) => {
        tx.set([nodeId, key], val);
      });
    });
  }
}

/**
  Renders an annotation. Used internally by different components (e.g. ui/AnnotatedTextComponent)

  @class
  @component
  @extends ui/Component

  @prop {Object} doc document
  @prop {Object} node node which describes annotation

  @example

  ```js
  $$(AnnotationComponent, {
    doc: doc,
    node: node
  })
  ```
*/

class AnnotationComponent extends Component {

  // TODO: we should avoid to have a didMount hook on an abstract base class
  didMount() {
    let node = this.props.node;
    node.on('highlighted', this.onHighlightedChanged, this);
  }

  // TODO: we should avoid to have a didMount hook on an abstract base class
  dispose() {
    let node = this.props.node;
    node.off(this);
  }

  render($$) {
    let el = $$(this.getTagName())
      .attr("data-id", this.props.node.id)
      .addClass(this.getClassNames());
    if (this.props.node.highlighted) {
      el.addClass('sm-highlighted');
    }
    el.append(this.props.children);
    return el
  }

  getClassNames() {
    return 'sc-'+this.props.node.type
  }

  onHighlightedChanged() {
    if (this.props.node.highlighted) {
      this.el.addClass('sm-highlighted');
    } else {
      this.el.removeClass('sm-highlighted');
    }
  }

  getTagName() {
    return 'span'
  }

}

class AbstractIsolatedNodeComponent extends Component {

  constructor(...args) {
    super(...args);

    this.name = this.props.node.id;
    this._id = this.context.surface.id +'/'+this.name;
    this._state = {
      selectionFragment: null
    };

    this.handleAction('escape', this.escape);
    this.ContentClass = this._getContentClass(this.props.node) || Component;

    // NOTE: FF does not allow to navigate contenteditable isles
    let useBlocker = platform.isFF || !this.ContentClass.noBlocker;
    this.blockingMode = useBlocker ? 'closed' : 'open';
  }

  getChildContext() {
    return {
      isolatedNodeComponent: this,
      // TODO: we should clear 'surface' here
      // so that we know that we are not controlled by a surface
      surface: undefined
    }
  }

  getInitialState() {
    let selState = this.context.editorSession.getSelectionState();
    return this._deriveStateFromSelectionState(selState)
  }

  didMount() {
    super.didMount();

    let editorSession = this.context.editorSession;
    editorSession.onRender('selection', this._onSelectionChanged, this);
  }

  dispose() {
    super.dispose.call(this);

    let editorSession = this.context.editorSession;
    editorSession.off(this);
  }

  renderContent($$, node, options = {}) {
    let ComponentClass = this.ContentClass;
    if (!ComponentClass) {
      console.error('Could not resolve a component for type: ' + node.type);
      return $$(this.__elementTag)
    } else {
      let props = Object.assign({
        disabled: this.props.disabled,
        node: node,
        isolatedNodeState: this.state.mode,
        focused: (this.state.mode === 'focused')
      }, options);
      return $$(ComponentClass, props)
    }
  }

  getId() {
    return this._id
  }

  get id() { return this.getId() }

  getMode() {
    return this.state.mode
  }

  isOpen() {
    return this.blockingMode === 'open'
  }

  isClosed() {
    return this.blockingMode === 'closed'
  }

  isNotSelected() {
    return !this.state.mode
  }

  isSelected() {
    return this.state.mode === 'selected'
  }

  isCoSelected() {
    return this.state.mode === 'co-selected'
  }

  isFocused() {
    return this.state.mode === 'focused'
  }

  isCoFocused() {
    return this.state.mode === 'co-focused'
  }

  getParentSurface() {
    return this.context.surface
  }

  escape() {
    // console.log('Escaping from IsolatedNode', this.id)
    this.selectNode();
  }

  _onSelectionChanged() {
    let editorSession = this.context.editorSession;
    let newState = this._deriveStateFromSelectionState(editorSession.getSelectionState());
    if (!newState && this.state.mode) {
      this.extendState({ mode: null });
    } else if (newState && newState.mode !== this.state.mode) {
      this.extendState(newState);
    }
  }

  onKeydown(event) {
    // console.log('####', event.keyCode, event.metaKey, event.ctrlKey, event.shiftKey);
    // TODO: while this works when we have an isolated node with input or CE,
    // there is no built-in way of receiving key events in other cases
    // We need a global event listener for keyboard events which dispatches to the current isolated node
    if (event.keyCode === keys$1.ESCAPE && this.state.mode === 'focused') {
      event.stopPropagation();
      event.preventDefault();
      this.escape();
    }
  }

  _getContentClass(node) {
    let componentRegistry = this.context.componentRegistry;
    let ComponentClass = componentRegistry.get(node.type);
    return ComponentClass
  }

  _getSurface(selState) {
    let surface = selState.get('surface');
    if (surface === undefined) {
      let sel = selState.getSelection();
      if (sel && sel.surfaceId) {
        let surfaceManager = this.context.surfaceManager;
        surface = surfaceManager.getSurface(sel.surfaceId);
      } else {
        surface = null;
      }
      selState.set('surface', surface);
    }
    return surface
  }

  // compute the list of surfaces and isolated nodes
  // for the given selection
  _getIsolatedNodes(selState) {
    let isolatedNodes = selState.get('isolatedNodes');
    if (!isolatedNodes) {
      let sel = selState.getSelection();
      isolatedNodes = [];
      if (sel && sel.surfaceId) {
        let surfaceManager = this.context.surfaceManager;
        let surface = surfaceManager.getSurface(sel.surfaceId);
        isolatedNodes = surface.getComponentPath().filter(comp => comp._isAbstractIsolatedNodeComponent);
      }
      selState.set('isolatedNodes', isolatedNodes);
    }
    return isolatedNodes
  }

  _shouldConsumeEvent(event) {
    let comp = Component.unwrap(event.target);
    return (comp && (comp === this || comp.context.isolatedNodeComponent === this))
  }

}

AbstractIsolatedNodeComponent.prototype._isAbstractIsolatedNodeComponent = true;

class InlineNodeComponent extends AbstractIsolatedNodeComponent {

  render($$) {
    let node = this.props.node;
    let ContentClass = this.ContentClass;

    let el = $$('span');
    el.addClass(this.getClassNames())
      .addClass('sc-inline-node')
      .addClass('sm-'+this.props.node.type)
      .attr("data-id", node.id)
      .attr('data-inline', '1');

    let disabled = this.isDisabled();

    if (this.state.mode) {
      el.addClass('sm-'+this.state.mode);
    } else {
      el.addClass('sm-not-selected');
    }

    if (!ContentClass.noStyle) {
      el.addClass('sm-default-style');
    }

    // shadowing handlers of the parent surface
    // TODO: extract this into a helper so that we can reuse it anywhere where we want
    // to prevent propagation to the parent surface
    el.on('keydown', this.onKeydown);

    el.append(
      this.renderContent($$, node)
        .ref('content')
        .addClass('se-content')
    );

    if (disabled) {
      el.addClass('sm-disabled')
         .attr('contenteditable', false)
         .on('click', this.onClick);
    }

    el.attr('draggable', true);
    return el
  }

  isDisabled() {
    return !this.state.mode || ['co-selected', 'cursor'].indexOf(this.state.mode) > -1;
  }

  getClassNames() {
    return ''
  }

  onClick(event) {
    if (!this._shouldConsumeEvent(event)) return
    this.selectNode();
  }

  selectNode() {
    // console.log('IsolatedNodeComponent: selecting node.');
    let editorSession = this.context.editorSession;
    let surface = this.context.surface;
    let node = this.props.node;
    editorSession.setSelection({
      type: 'property',
      path: node.start.path,
      startOffset: node.start.offset,
      endOffset: node.end.offset,
      containerId: surface.getContainerId(),
      surfaceId: surface.id
    });
  }

  // TODO: this is almost the same as in InlineNodeComponent
  // We should try to consolidate this
  _deriveStateFromSelectionState(selState) {
    let surface = this._getSurface(selState);
    if (!surface) return null
    // detect cases where this node is selected or co-selected by inspecting the selection
    if (surface === this.context.surface) {
      let sel = selState.getSelection();
      let node = this.props.node;
      if (sel.isPropertySelection() && !sel.isCollapsed() && isEqual(sel.path, node.path)) {
        let nodeSel = node.getSelection();
        if(nodeSel.equals(sel)) {
          return { mode: 'selected' }
        }
        if (sel.contains(nodeSel)) {
          return { mode: 'co-selected' }
        }
      }
    }
    let isolatedNodeComponent = surface.context.isolatedNodeComponent;
    if (!isolatedNodeComponent) return null
    if (isolatedNodeComponent === this) {
      return { mode: 'focused' }
    }
    let isolatedNodes = this._getIsolatedNodes(selState);
    if (isolatedNodes.indexOf(this) > -1) {
      return { mode: 'co-focused' }
    }
    return null
  }

}

InlineNodeComponent.prototype._isInlineNodeComponent = true;

/**
  Renders an anotated text. Used internally by {@link ui/TextPropertyComponent}.

  @class
  @component
  @extends ui/Component

  @prop {String[]} path The property to be rendered.
*/

class AnnotatedTextComponent extends Component {

  render($$) {
    let el = this._renderContent($$)
      .addClass('sc-annotated-text')
      .css({ whiteSpace: "pre-wrap" });
    return el
  }

  getText() {
    return this.getDocument().get(this.props.path) || ''
  }

  getAnnotations() {
    return this.getDocument().getIndex('annotations').get(this.props.path)
  }

  _getTagName() {
    return this.props.tagName
  }

  _onDocumentChange(update) {
    if (update.change && update.change.updated[this.getPath()]) {
      this.rerender();
    }
  }

  _renderContent($$) {
    let text = this.getText();
    let annotations = this.getAnnotations();
    let el = $$(this._getTagName() || 'span');
    if (annotations && annotations.length > 0) {
      let fragmenter = new Fragmenter({
        onText: this._renderTextNode.bind(this),
        onEnter: this._renderFragment.bind(this, $$),
        onExit: this._finishFragment.bind(this)
      });
      fragmenter.start(el, text, annotations);
    } else {
      el.append(text);
    }
    return el
  }

  _renderTextNode(context, text) {
    if (text && text.length > 0) {
      context.append(text);
    }
  }

  _renderFragment($$, fragment) {
    let doc = this.getDocument();
    let componentRegistry = this.getComponentRegistry();
    let node = fragment.node;
    // TODO: fix support for container annotations
    if (node.type === "container-annotation-fragment") {
      // return $$(AnnotationComponent, { doc: doc, node: node })
      //   .addClass("se-annotation-fragment")
      //   .addClass(node.anno.getTypeNames().join(' ').replace(/_/g, "-"));
    } else if (node.type === "container-annotation-anchor") {
      // return $$(AnnotationComponent, { doc: doc, node: node })
      //   .addClass("se-anchor")
      //   .addClass(node.anno.getTypeNames().join(' ').replace(/_/g, "-"))
      //   .addClass(node.isStart?"start-anchor":"end-anchor")
    } else {
      let ComponentClass = componentRegistry.get(node.type) || AnnotationComponent;
      if (node.constructor.isInline &&
          // also no extra wrapping if the node is already an inline node
          !ComponentClass.prototype._isInlineNodeComponent &&
          // opt-out for custom implementations
          !ComponentClass.isCustom) {
        ComponentClass = InlineNodeComponent;
      }
      let el = $$(ComponentClass, { doc: doc, node: node });
      return el
    }
  }

  _finishFragment(fragment, context, parentContext) {
    parentContext.append(context);
  }

  /**
    Gets document instance.

    @return {Document} The document instance
   */
  getDocument() {
    return this.props.doc || this.context.doc
  }

}

class CursorComponent extends Component {

  render($$) {
    // TODO: we should rename se-cursor to sc-cursor
    let el = $$('span').addClass('se-cursor');
    // Add zero-width character. Since we have a non-empty element, the
    // outline style set on the cursor would not be visible in certain
    // scenarios (e.g. when cursor is at the very beginning of a text.
    el.append("\uFEFF");
    el.append($$('div').addClass('se-cursor-inner'));

    if (this.props.collaborator) {
      let collaboratorIndex = this.props.collaborator.colorIndex;
      el.addClass('sm-collaborator-'+collaboratorIndex);
    } else {
      el.addClass('sm-local-user');
    }

    return el
  }

}

class SelectionFragmentComponent extends Component {

  render($$) {
    // TODO: we should rename se-cursor to sc-cursor
    let el = $$('span').addClass('se-selection-fragment');
    if (this.props.collaborator) {
      let collaboratorIndex = this.props.collaborator.colorIndex;
      el.addClass('sm-collaborator-'+collaboratorIndex);
    } else {
      el.addClass('sm-local-user');
    }
    el.append(this.props.children);
    return el
  }

}

/**
  Renders a text property. Used internally by different components to render
  editable text.

  @class
  @component
  @extends ui/AnnotatedTextComponent

  @prop {String[]} path path to a text property
  @prop {String} [tagName] specifies which tag should be used - defaults to `div`

  @example

  ```js
  $$(TextProperty, {
    path: [ 'paragraph-1', 'content']
  })
  ```
*/

class TextPropertyComponent extends AnnotatedTextComponent {

  getInitialState() {
    const markersManager = this.context.markersManager;
    let path = this.props.path;
    let markers;
    if (markersManager) {
      // register and get initial set of markers
      markersManager.register(this);
      markers = markersManager.getMarkers(path, {
        surfaceId: this.getSurfaceId(),
        containerId: this.getContainerId()
      });
    } else {
      const doc = this.getDocument();
      markers = doc.getAnnotations(path);
    }
    return {
      markers: markers
    }
  }

  didMount() {
    if (this.context.surface && this.context.surface.hasNativeSpellcheck()) {
      this.domObserver = new window.MutationObserver(this._onDomMutations.bind(this));
      this.domObserver.observe(this.el.getNativeElement(), { subtree: true, characterData: true, characterDataOldValue: true });
    }
  }

  dispose() {
    if (this.context.markersManager) {
      this.context.markersManager.deregister(this);
    }
  }

  render($$) {
    let path = this.getPath();

    let el = this._renderContent($$)
      .addClass('sc-text-property')
      .attr({
        'data-path': path.join('.')
      })
      .css({
        'white-space': 'pre-wrap'
      });

    if (!this.props.withoutBreak) {
      el.append($$('br'));
    }
    return el
  }

  getAnnotations() {
    if (this.props.markers) {
      return this.state.markers.concat(this.props.markers)
    } else {
      return this.state.markers
    }
  }

  _renderFragment($$, fragment) {
    let node = fragment.node;
    let id = node.id;
    let el;
    if (node.type === 'cursor') {
      el = $$(CursorComponent, { collaborator: node.collaborator });
    } else if (node.type === 'selection-fragment') {
      el = $$(SelectionFragmentComponent, { collaborator: node.collaborator });
    } else {
      el = super._renderFragment.apply(this, arguments);
      if (id) {
        el.ref(id + '@' + fragment.counter);
      }
    }
    el.attr('data-offset', fragment.pos);
    return el
  }

  _onDomMutations(mutations) {
    // HACK: only detecting mutations that are coming from the native spell-correction
    if (mutations.length === 2 && mutations[0].target === mutations[1].target) {
      let textEl = DefaultDOMElement.unwrap(mutations[0].target);
      if (textEl) {
        this._applyTextMutation(textEl, mutations[0].oldValue);
        return
      }
    }
    // in all other cases, revert the change by rerendering
    this.rerender();
  }

  _applyTextMutation(textEl, oldText) {
    // find the offset
    let offset = _getCharPos(textEl, 0);
    let newText = textEl.textContent;
    let changes = diff(oldText, newText, offset);

    let editorSession = this.context.editorSession;
    let path = this.getPath();
    editorSession.transaction(function(tx) {
      changes.forEach(function(change) {
        // NOTE: atomic text replace is not supported currently
        if (change.type === 'replace') {
          tx.update(path, { type: 'delete', start: change.start, end: change.end });
          tx.update(path, { type: 'insert', start: change.start, text: change.text });
        } else {
          tx.update(path, change);
        }
      });
    });
  }

  getPath() {
    return this.props.path
  }

  getText() {
    return this.getDocument().get(this.getPath())
  }

  getDocument() {
    return this.props.doc ||this.context.doc
  }

  getSurface() {
    return this.props.surface || this.context.surface
  }

  getSurfaceId() {
    let surface = this.getSurface();
    return surface ? surface.id : null
  }

  getContainerId() {
    let surface = this.getSurface();
    return surface ? surface.getContainerId() : null
  }

  isEditable() {
    return this.getSurface().isEditable()
  }

  isReadonly() {
    return this.getSurface().isReadonly()
  }

  getDOMCoordinate(charPos) {
    return this._getDOMCoordinate(this.el, charPos)
  }


  _finishFragment(fragment, context, parentContext) {
    context.attr('data-length', fragment.length);
    parentContext.append(context);
  }

  _getDOMCoordinate(el, charPos) {
    let l;
    let idx = 0;
    if (charPos === 0) {
      return {
        container: el.getNativeElement(),
        offset: 0
      }
    }
    for (let child = el.getFirstChild(); child; child = child.getNextSibling(), idx++) {
      if (child.isTextNode()) {
        l = child.textContent.length;
        if (l >= charPos) {
          return {
            container: child.getNativeElement(),
            offset: charPos
          }
        } else {
          charPos -= l;
        }
      } else if (child.isElementNode()) {
        let length = child.getAttribute('data-length');
        if (length) {
          l = parseInt(length, 10);
          if (l >= charPos) {
            // special handling for InlineNodes
            if (child.attr('data-inline')) {
              let nextSibling = child.getNextSibling();
              if (nextSibling && nextSibling.isTextNode()) {
                return {
                  container: nextSibling.getNativeElement(),
                  offset: 0
                }
              } else {
                return {
                  container: el.getNativeElement(),
                  offset: el.getChildIndex(child) + 1
                }
              }
            }
            return this._getDOMCoordinate(child, charPos, idx)
          } else {
            charPos -= l;
          }
        } else {
          console.error('FIXME: Can not map to DOM coordinates.');
          return null
        }
      }
    }
  }

}

TextPropertyComponent.prototype._isTextPropertyComponent = true;


// Helpers for DOM selection mapping

/*
  Used to map from DOM to model.
  Given a root element and a DOM element - which is typically somewhere inside a surface
  it tries to find the next TextProperty by walking up the DOM.
  If found it computes the character position, counting chars and using the hints, data-length and data-offset,
  rendered by the TextPropertyComponent
*/
TextPropertyComponent.getCoordinate = function(root, el, offset) {
  let context = _getPropertyContext(root, el, offset);
  if (!context) {
    return null
  }
  // in some cases we need to normalize the DOM coordinate
  // before we can use it for retrieving charPos
  // E.g. observed with #273
  let charPos = _getCharPos(context.node, context.offset);
  if (isNumber(charPos)) {
    let coor = new Coordinate(context.path, charPos);
    coor._comp = context.comp;
    return coor
  } else {
    return null
  }
};

function _getPropertyContext(root, node, offset) {
  let result = {
    comp: null,
    el: null,
    path: null,
    node: node,
    offset: offset
  };
  while (node && node !== root) {
    if (node.isElementNode()) {
      let comp = Component.unwrap(node);
      if (comp && comp._isTextPropertyComponent) {
        result.comp = comp;
        result.el = node;
        result.path = comp.getPath();
        return result;
      }
      // we need to normalize situations where the DOM coordinate
      // is inside an inline node, which we have observed
      // can actually happen.
      if (node.getAttribute('data-inline')) {
        result.node = node;
        if (offset > 0) {
          result.offset = 1;
        }
      }
    }
    node = node.getParent();
  }
  return null
}

function _getCharPos(node, offset) {
  let charPos = offset;
  let parent, childIdx;

  /*
    In the following implementation we are exploiting two facts
    for optimization:
    - an element with data-path is assumed to be the text property element
    - an element with data-offset is assumed to be an annotation element

    Particularly, the data-offset property is helpful to get the character position
    in just one iteration.
  */

  parent = node.getParent();
  if (node.isTextNode()) {
    // TextNode is first child
    if (node === parent.firstChild) {
      // ... we can stop if parent is text property
      let parentPath = parent.getAttribute('data-path');
      let parentOffset = parent.getAttribute('data-offset');
      if (parentPath) {
        charPos = offset;
      }
      // ... and we can stop if parent has an offset hint
      else if (parentOffset) {
        charPos = parseInt(parentOffset, 10) + offset;
      }
      // ... otherwise we count the charPos by recursing up-tree
      else {
        charPos = _getCharPos(parent, 0) + offset;
      }
    } else {
      // the node has a predecessor so we can apply recurse using the child index
      childIdx = parent.getChildIndex(node);
      charPos = _getCharPos(parent, childIdx) + offset;
    }
  } else if (node.isElementNode()) {
    let pathStr = node.getAttribute('data-path');
    let offsetStr = node.getAttribute('data-offset');
    // if node is the element of a text property, then offset is a child index
    // up to which we need to sum up all lengths
    if (pathStr) {
      charPos = _countCharacters(node, offset);
    }
    // similar if node is the element of an annotation, and we can use the
    // element's offset
    else if (offsetStr) {
      childIdx = parent.getChildIndex(node);
      charPos = parseInt(offsetStr, 10) + _countCharacters(node, offset);
    }
    // for other elements we need to count characters in the child tree
    // adding the offset of this element which needs to be computed by recursing up-tree
    else {
      childIdx = parent.getChildIndex(node);
      charPos = _getCharPos(parent, childIdx) + _countCharacters(node, offset);
    }
  } else {
    // Unsupported case
    return null
  }
  return charPos;
}

function _countCharacters(el, maxIdx) {
  let charPos = 0;
  // inline elements have a length of 1
  if (el.getAttribute('data-inline')) {
    return maxIdx === 0 ? 0 : 1;
  }
  let l = el.getChildCount();
  if (arguments.length === 1) {
    maxIdx = l;
  }
  maxIdx = Math.min(l, maxIdx);
  for (let i=0, child = el.getFirstChild(); i < maxIdx; child = child.getNextSibling(), i++) {
    if (child.isTextNode()) {
      charPos += child.getTextContent().length;
    } else if (child.isElementNode()) {
      let length = child.getAttribute('data-length');
      if (child.getAttribute('data-inline')) {
        charPos += 1;
      } else if (length) {
        charPos += parseInt(length, 10);
      } else {
        charPos += _countCharacters(child);
      }
    }
  }
  return charPos
}

const BRACKET = 'X';

/*
  Isolation Strategies:
    - default: IsolatedNode renders a blocker the content gets enabled by a double-click.
    - open: No blocker. Content is enabled when parent surface is.

    > Notes:

      The blocker is used to shield the inner UI and not to interfer with general editing gestures.
      In some cases however, e.g. a figure (with image and caption), it feels better if the content directly accessible.
      In this case, the content component must provide a means to drag the node, e.g. set `<img draggable=true>`.
      This works only in browsers that are able to deal with 'contenteditable' isles,
      i.e. a structure where the isolated node is contenteditable=false, and inner elements have contenteditable=true
      Does not work in Edge. Works in Chrome, Safari

      The the default unblocking gesture requires the content to implement a grabFocus() method, which should set the selection
      into one of the surfaces, or set a CustomSelection.
*/
class IsolatedNodeComponent extends AbstractIsolatedNodeComponent {

  constructor(...args) {
    super(...args);
  }

  render($$) {
    let node = this.props.node;
    let ContentClass = this.ContentClass;
    let disabled = this.props.disabled;

    // console.log('##### IsolatedNodeComponent.render()', $$.capturing);
    let el = $$('div');
    el.addClass(this.getClassNames())
      .addClass('sc-isolated-node')
      .addClass('sm-'+this.props.node.type)
      .attr("data-id", node.id);
    if (disabled) {
      el.addClass('sm-disabled');
    }
    if (this.state.mode) {
      el.addClass('sm-'+this.state.mode);
    }
    if (!ContentClass.noStyle) {
      el.addClass('sm-default-style');
    }
    // always handle ESCAPE
    el.on('keydown', this.onKeydown);

    // console.log('##### rendering IsolatedNode', this.id)
    let shouldRenderBlocker = (
      this.blockingMode === 'closed' &&
      !this.state.unblocked
    );

    // HACK: we need something 'editable' where we can put DOM selection into,
    // otherwise native cursor navigation gets broken
    el.append(
      $$('div').addClass('se-bracket sm-left').ref('left')
        .append(BRACKET)
    );

    let content = this.renderContent($$, node, {
      disabled: this.props.disabled || shouldRenderBlocker
    }).ref('content');
    content.attr('contenteditable', false);

    el.append(content);
    el.append($$(Blocker).ref('blocker'));
    el.append(
      $$('div').addClass('se-bracket sm-right').ref('right')
        .append(BRACKET)
    );

    if (!shouldRenderBlocker) {
      el.addClass('sm-no-blocker');
      el.on('click', this.onClick)
        .on('dblclick', this.onDblClick);
    }
    el.on('mousedown', this._reserveMousedown, this);

    return el
  }

  getClassNames() {
    return ''
  }

  getContent() {
    return this.refs.content
  }

  selectNode() {
    // console.log('IsolatedNodeComponent: selecting node.');
    let editorSession = this.context.editorSession;
    let surface = this.context.surface;
    let nodeId = this.props.node.id;
    editorSession.setSelection({
      type: 'node',
      nodeId: nodeId,
      containerId: surface.getContainerId(),
      surfaceId: surface.id
    });
  }

  // EXPERIMENTAL: trying to catch clicks not handler by the
  // content when this is unblocked
  onClick(event) {
    // console.log('### Clicked on IsolatedNode', this.id, event.target)
    event.stopPropagation();
  }

  onDblClick(event) {
    // console.log('### DblClicked on IsolatedNode', this.id, event.target)
    event.stopPropagation();
  }

  grabFocus(event) {
    let content = this.refs.content;
    if (content.grabFocus) {
      content.grabFocus(event);
      return true
    }
  }

  // EXPERIMENTAL: Surface and IsolatedNodeComponent communicate via flag on the mousedown event
  // and only reacting on click or mouseup when the mousedown has been reserved
  _reserveMousedown(event) {
    if (event.__reserved__) {
      // console.log('%s: mousedown already reserved by %s', this.id, event.__reserved__.id)
      return
    } else {
      // console.log('%s: taking mousedown ', this.id)
      event.__reserved__ = this;
    }
  }

  _deriveStateFromSelectionState(selState) {
    let surface = this._getSurface(selState);
    let newState = { mode: null, unblocked: null};
    if (!surface) return newState
    // detect cases where this node is selected or co-selected by inspecting the selection
    if (surface === this.context.surface) {
      let sel = selState.getSelection();
      let nodeId = this.props.node.id;
      if (sel.isNodeSelection() && sel.getNodeId() === nodeId) {
        if (sel.isFull()) {
          newState.mode = 'selected';
          newState.unblocked = true;
        } else if (sel.isBefore()) {
          newState.mode = 'cursor';
          newState.position = 'before';
        } else if (sel.isAfter()) {
          newState.mode = 'cursor';
          newState.position = 'after';
        }
      }
      if (sel.isContainerSelection() && sel.containsNode(nodeId)) {
        newState.mode = 'co-selected';
      }
    } else {
      let isolatedNodeComponent = surface.context.isolatedNodeComponent;
      if (isolatedNodeComponent) {
        if (isolatedNodeComponent === this) {
          newState.mode = 'focused';
          newState.unblocked = true;
        } else {
          let isolatedNodes = this._getIsolatedNodes(selState);
          if (isolatedNodes.indexOf(this) > -1) {
            newState.mode = 'co-focused';
            newState.unblocked = true;
          }
        }
      }
    }
    return newState
  }

}

IsolatedNodeComponent.prototype._isIsolatedNodeComponent = true;

IsolatedNodeComponent.prototype._isDisabled = IsolatedNodeComponent.prototype.isDisabled;

IsolatedNodeComponent.getDOMCoordinate = function(comp, coor) {
  let { start, end } = IsolatedNodeComponent.getDOMCoordinates(comp);
  if (coor.offset === 0) return start
  else return end
};

IsolatedNodeComponent.getDOMCoordinates = function(comp) {
  const left = comp.refs.left;
  const right = comp.refs.right;
  return {
    start: {
      container: left.getNativeElement(),
      offset: 0
    },
    end: {
      container: right.getNativeElement(),
      offset: right.getChildCount()
    }
  }
};

IsolatedNodeComponent.getCoordinate = function(nodeEl, options) {
  let comp = Component.unwrap(nodeEl, 'strict').context.isolatedNodeComponent;
  let offset = null;
  if (options.direction === 'left' || nodeEl === comp.refs.left.el) {
    offset = 0;
  } else if (options.direction === 'right' || nodeEl === comp.refs.right.el) {
    offset = 1;
  }
  let coor;
  if (offset !== null) {
    coor = new Coordinate([comp.props.node.id], offset);
    coor._comp = comp;
  }
  return coor
};

class Blocker extends Component {

  render($$) {
    return $$('div').addClass('sc-isolated-node-blocker')
      .attr('draggable', true)
      .attr('contenteditable', false)
      .on('click', this.onClick)
      .on('dblclick', this.onDblClick)
  }

  onClick(event) {
    if (event.target !== this.getNativeElement()) return
    // console.log('Clicked on Blocker of %s', this._getIsolatedNodeComponent().id, event)
    event.stopPropagation();
    const comp = this._getIsolatedNodeComponent();
    comp.extendState({ mode: 'selected', unblocked: true });
    comp.selectNode();
  }

  onDblClick(event) {
    // console.log('DblClicked on Blocker of %s', this.getParent().id, event)
    event.stopPropagation();
  }

  _getIsolatedNodeComponent() {
    return this.context.isolatedNodeComponent
  }

}

const DEBUG$1 = false;

/*
  A class that maps DOM selections to model selections.

  There are some difficulties with mapping model selections:
  1. DOM selections can not model discontinuous selections.
  2. Not all positions reachable via ContentEditable can be mapped to model selections. For instance,
     there are extra positions before and after non-editable child elements.
  3. Some native cursor behaviors need to be overidden.

  @param {Editor} Editor component
 */
class DOMSelection {

  constructor(editor) {
    this.editor = editor;
    if (platform.inBrowser) {
      this.wRange = window.document.createRange();
    }
    // keeping the last DOM and Model coordinates
    // TODO: why are we doing this?
    this.state = { dom: null, model: null };
  }

  /**
    Create a model selection by mapping the current DOM selection
    to model coordinates.

    @param {object} options
      - `direction`: `left` or `right`; a hint for disambiguations, used by Surface during cursor navigation.
    @returns {model/Selection}
  */
  getSelection(options) {
    // HACK: ignore this if not Browser (e.g. when running the test suite in node)
    if (!platform.inBrowser) return
    let range = this.mapDOMSelection(options);
    let doc = this.editor.getDocument();
    // TODO: consolidate
    return doc._createSelectionFromRange(range)
  }

  getSelectionForDOMRange(wrange) {
    let range = this.mapDOMRange(wrange);
    let doc = this.editor.getDocument();
    return doc._createSelectionFromRange(range)
  }

  /*
    Maps the current DOM selection to a model range.

    @param {object} [options]
      - `direction`: `left` or `right`; a hint for disambiguations, used by Surface during cursor navigation.
    @returns {model/Range}
  */
  mapDOMSelection(options) {
    let wSel = window.getSelection();
    let state = this.state;
    let range;
    // Use this log whenever the mapping goes wrong to analyze what
    // is actually being provided by the browser
    if (DEBUG$1) console.info('DOM->Model: ', wSel.anchorNode, wSel.anchorOffset, wSel.focusNode, wSel.focusOffset);
    if (wSel.rangeCount === 0) return _null()
    let anchorNode = DefaultDOMElement.wrapNativeElement(wSel.anchorNode);
    if (wSel.isCollapsed) {
      let coor = this._getCoordinate(anchorNode, wSel.anchorOffset, options);
      if (!coor) return _null()
      range = _createRange({
        start: coor,
        end: coor
      });
    }
    else {
      let focusNode = DefaultDOMElement.wrapNativeElement(wSel.focusNode);
      range = this._getRange(anchorNode, wSel.anchorOffset, focusNode, wSel.focusOffset, options);
    }
    if (DEBUG$1) console.info('DOM->Model: range ', range ? range.toString() : null);
    state.model = range;
    return range

    function _null() {
      state.dom = null;
      state.model = null;
      return null
    }
  }

  /**
    Transfer a given model selection into the DOM.

    @param {model/Selection} sel
  */
  setSelection(sel) {
    // HACK: ignore this if not Browser (e.g. when running the test suite in node)
    if (!platform.inBrowser) return
    let state = this.state;
    let wSel = window.getSelection();
    let wRange = this.wRange;
    if (!sel || sel.isNull()) return this.clear()
    // console.log('### DOMSelection: setting selection', sel.toString());
    let {start, end} = this.mapModelToDOMCoordinates(sel);
    if (!start) return this.clear()
    if (sel.isReverse()) {
      [start, end] = [end, start];
    }
    state.dom = {
      anchorNode: start.container,
      anchorOffset: start.offset,
      focusNode: end.container,
      focusOffset: end.offset
    };
    _set(state.dom);

    function _set({anchorNode, anchorOffset, focusNode, focusOffset}) {
      wSel.removeAllRanges();
      wRange.setStart(anchorNode, anchorOffset);
      wRange.setEnd(anchorNode, anchorOffset);
      wSel.addRange(wRange);
      if (focusNode !== anchorOffset || focusOffset !== anchorOffset) {
        wSel.extend(focusNode, focusOffset);
      }
    }
  }

  mapModelToDOMCoordinates(sel) {
    if (DEBUG$1) console.info('Model->DOM: sel =', sel.toString());
    let rootEl;
    let surface = this.editor.surfaceManager.getSurface(sel.surfaceId);
    if (!surface) {
      console.warn('Selection should have "surfaceId" set.');
      rootEl = this.editor.el;
    } else {
      rootEl = surface.el;
    }
    if (sel.isNull() || sel.isCustomSelection()) {
      return {}
    }

    let start, end;
    if (sel.isPropertySelection() || sel.isContainerSelection()) {
      start = this._getDOMCoordinate(rootEl, sel.start);
      if (!start) {
        console.warn('FIXME: selection seems to be invalid.');
        return {}
      }
      if (sel.isCollapsed()) {
        end = start;
      } else {
        end = this._getDOMCoordinate(rootEl, sel.end);
        if (!end) {
          console.warn('FIXME: selection seems to be invalid.');
          return {}
        }
      }
    } else if (sel.isNodeSelection()) {
      let comp = Component.unwrap(rootEl.find('*[data-id="'+sel.getNodeId()+'"]'));
      if (!comp) {
        console.error('Could not find component with id', sel.getNodeId());
        return {}
      }
      if (comp._isIsolatedNodeComponent) {
        let coors = IsolatedNodeComponent.getDOMCoordinates(comp, sel);
        start = coors.start;
        end = coors.end;
        // Note: ATM we do not render collapsed NodeSelections differently
        // if (sel.isAfter()) start = end
        // else if (sel.isBefore()) end = start
      } else {
        let _nodeEl = comp.el;
        start = {
          container: _nodeEl.getNativeElement(),
          offset: 0
        };
        end = {
          container: _nodeEl.getNativeElement(),
          offset: _nodeEl.getChildCount()
        };
      }
    }
    if (DEBUG$1) console.info('Model->DOM:', start.container, start.offset, end.container, end.offset, 'isReverse?', sel.isReverse());
    return {start,end}
  }

  _getDOMCoordinate(rootEl, coor) {
    let comp, domCoor = null;
    if (coor.isNodeCoordinate()) {
      comp = Component.unwrap(rootEl.find('*[data-id="'+coor.getNodeId()+'"]'));
      if (comp) {
        if (comp._isIsolatedNodeComponent) {
          domCoor = IsolatedNodeComponent.getDOMCoordinate(comp, coor);
        } else {
          let domOffset = 0;
          if (coor.offset > 0) {
            domOffset = comp.getChildCount();
          }
          domCoor = {
            container: comp.getNativeElement(),
            offset: domOffset
          };
        }
      }
    } else {
      comp = Component.unwrap(rootEl.find('.sc-text-property[data-path="'+coor.path.join('.')+'"]'));
      if (comp) {
        domCoor = comp.getDOMCoordinate(coor.offset);
      }
    }
    return domCoor
  }

  /*
    Map a DOM range to a model range.

    @param {Range} range
    @returns {model/Range}
  */
  mapDOMRange(wRange, options) {
    return this._getRange(
      DefaultDOMElement.wrapNativeElement(wRange.startContainer),
      wRange.startOffset,
      DefaultDOMElement.wrapNativeElement(wRange.endContainer),
      wRange.endOffset, options)
  }

  /*
    Clear the DOM selection.
  */
  clear() {
    window.getSelection().removeAllRanges();
    this.state.dom = null;
    this.state.model = null;
  }

  collapse(dir) {
    let wSel = window.getSelection();
    let wRange;
    if (wSel.rangeCount > 0) {
      wRange = wSel.getRangeAt(0);
      wRange.collapse(dir === 'left');
      wSel.removeAllRanges();
      wSel.addRange(wRange);
    }
  }

  select(el) {
    let wSel = window.getSelection();
    let wRange = window.document.createRange();
    wRange.selectNode(el.getNativeElement());
    wSel.removeAllRanges();
    wSel.addRange(wRange);
  }

  extend(el, offset) {
    let wSel = window.getSelection();
    wSel.extend(el.getNativeElement(), offset);
  }

  setCursor(el, offset) {
    let wSel = window.getSelection();
    let wRange = window.document.createRange();
    wRange.setStart(el.getNativeElement(), offset);
    wSel.removeAllRanges();
    wSel.addRange(wRange);
  }

  /*
    Extract a model range from given DOM elements.

    @param {Node} anchorNode
    @param {number} anchorOffset
    @param {Node} focusNode
    @param {number} focusOffset
    @returns {model/Range}
  */
  _getRange(anchorNode, anchorOffset, focusNode, focusOffset, options = {}) {
    let isReverse = DefaultDOMElement.isReverse(anchorNode, anchorOffset, focusNode, focusOffset);
    let isCollapsed = (anchorNode === focusNode && anchorOffset === focusOffset);
    let start, end;
    if (isCollapsed) {
      start = end = this._getCoordinate(anchorNode, anchorOffset, options);
    } else {
      start = this._getCoordinate(anchorNode, anchorOffset, { direction: isReverse ? 'right' : 'left' });
      end = this._getCoordinate(focusNode, focusOffset, options);
    }
    if (start && end) {
      return _createRange({ start, end, isReverse })
    } else {
      return null
    }
  }

  /*
    Map a DOM coordinate to a model coordinate.

    @param {Node} node
    @param {number} offset
    @param {object} options
    @param {object} [options]
      - `direction`: `left` or `right`; a hint for disambiguation.
    @returns {model/Coordinate}

    @info

    `options.direction` can be used to control the result when this function is called
    after cursor navigation. The root problem is that we are using ContentEditable on
    Container level (as opposed to TextProperty level). The native ContentEditable allows
    cursor positions which do not make sense in the model sense.

    For example,

    ```
    <div contenteditable=true>
      <p data-path="p1.content">foo</p>
      <img>
      <p data-path="p1.content">bar</p>
    </div>
    ```
    would allow to set the cursor directly before or after the image, which
    we want to prevent, as it is not a valid insert position for text.
    Instead, if we find the DOM selection in such a situation, then we map it to the
    closest valid model address. And this depends on the direction of movement.
    Moving `left` would provide the previous address, `right` would provide the next address.
    The default direction is `right`.
  */
  _getCoordinate(nodeEl, offset, options={}) {
    let coor = null;
    // this deals with a cursor in a TextProperty
    if (!coor) {
      coor = TextPropertyComponent.getCoordinate(this.editor.el, nodeEl, offset);
    }
    let comp = Component.unwrap(nodeEl);
    if (!coor && comp) {
      // let IsolatedNodeComponent figure out where the selection is
      if (comp.context.isolatedNodeComponent) {
        coor = IsolatedNodeComponent.getCoordinate(nodeEl, options);
      }
    }
    // Edge-cases: These handlers are hacked so that the case is covered,
    // not solved 'elegantly'
    if (!coor) {
      // as in #354: sometimes anchor or focus is the surface itself
      if (comp && comp._isContainerEditor) {
        let childIdx = (offset === 0) ? 0 : offset-1;
        let isBefore = (offset === 0);
        let container = comp.getContainer();
        let childNode = container.getNodeAt(childIdx);
        let childComp = comp.getChildAt(childIdx);
        coor = new Coordinate([childNode.id], isBefore?0:1 );
        coor._comp = childComp;
      }
      // sometimes anchor or focus is a Node component with TextPropertyComponents as children (all TextNode Components)
      else if (nodeEl.isElementNode() && nodeEl.getChildCount() > 0) {
        let child = (offset > 0) ? nodeEl.getChildAt(offset-1) : nodeEl.firstChild;
        let prop;
        let childComp = Component.unwrap(child);
        if (childComp && childComp._isTextPropertyComponent) {
          prop = child;
        }
        // let prop = last(child.findAll('data-path'))
        if (prop) {
          coor = TextPropertyComponent.getCoordinate(nodeEl, prop, (offset > 0) ? prop.getChildCount() : 0);
        }
      }
    }
    return coor
  }

}

/*
 Helper for creating a model range correctly
 as for model/Range start should be before end.

 In contrast to that, DOM selections are described with anchor and focus coordinates,
 i.e. bearing the information of direction implicitly.
 To simplify the implementation we treat anchor and focus equally
 and only at the end exploit the fact deriving an isReverse flag
 and bringing start and end in the correct order.
*/
function _createRange({start, end, isReverse}) {
  if (isReverse) {
    [start, end] = [end, start];
  }
  if (!start._comp || !end._comp) {
    console.error('FIXME: getCoordinate() should provide a component instance');
    return null
  }
  let surface = start._comp.context.surface;
  if (!surface) {
    console.error('FIXME: Editable components should have their surface in the context');
    return null
  }
  if (surface !== end._comp.context.surface) {
    console.error('Coordinates are within two different surfaces. Can not create a selection.');
    return null
  }
  return new Range(start, end, isReverse, surface.getContainerId(), surface.id)
}

/**
  Reusable abstract editor implementation.

  @example

  ```js
  class SimpleWriter extends AbstractEditor {
    render($$) {
      // render editor
    }
  }
  ```
*/
class AbstractEditor extends Component {

  constructor(...args) {
    super(...args);
    this._initialize(this.props);
  }

  _initialize(props) {
    if (!props.editorSession) {
      throw new Error('EditorSession instance required');
    }
    this.editorSession = props.editorSession;
    this.doc = this.editorSession.getDocument();

    let configurator = this.editorSession.getConfigurator();
    this.componentRegistry = configurator.getComponentRegistry();
    this.toolGroups = configurator.getToolGroups();
    this.labelProvider = configurator.getLabelProvider();
    this.iconProvider = configurator.getIconProvider();

    // legacy
    this.surfaceManager = this.editorSession.surfaceManager;
    this.commandManager = this.editorSession.commandManager;
    this.dragManager = this.editorSession.dragManager;
    this.macroManager = this.editorSession.macroManager;
    this.converterRegistry = this.editorSession.converterRegistry;
    this.globalEventHandler = this.editorSession.globalEventHandler;
    this.editingBehavior = this.editorSession.editingBehavior;
    this.markersManager = this.editorSession.markersManager;

    this.resourceManager = new ResourceManager(this.editorSession, this.getChildContext());

    this.domSelection = new DOMSelection(this);
  }

  willReceiveProps(nextProps) {
    let newSession = nextProps.editorSession;
    let shouldDispose = newSession && newSession !== this.editorSession;
    if (shouldDispose) {
      this._dispose();
      this._initialize(nextProps);
    }
  }

  dispose() {
    this._dispose();
  }

  _dispose() {
    // Note: we need to clear everything, as the childContext
    // changes which is immutable
    this.empty();
    // not necessary
    // this.domSelection.dispose()
    this.resourceManager.dispose();
  }

  getChildContext() {
    return {
      editor: this,
      editorSession: this.editorSession,
      doc: this.doc, // TODO: remove in favor of editorSession
      componentRegistry: this.componentRegistry,
      surfaceManager: this.surfaceManager,
      domSelection: this.domSelection,
      commandManager: this.commandManager,
      markersManager: this.markersManager,
      converterRegistry: this.converterRegistry,
      dragManager: this.dragManager,
      editingBehavior: this.editingBehavior,
      globalEventHandler: this.globalEventHandler,
      iconProvider: this.iconProvider,
      labelProvider: this.labelProvider,
      resourceManager: this.resourceManager,
      // ATTENTION: this is a map of tool target names to maps of tool names to tools
      // i.e. a declarative way to map tools to tool groups
      toolGroups: this.toolGroups,
    }
  }

  getDocument() {
    return this.editorSession.getDocument()
  }

  getConfigurator() {
    return this.editorSession.getConfigurator()
  }

  getEditorSession() {
    return this.editorSession
  }

  getComponentRegistry() {
    return this.componentRegistry
  }
}

class AbstractScrollPane$$1 extends Component {

  /*
    Expose scrollPane as a child context
  */
  getChildContext() {
    return {
      scrollPane: this
    }
  }

  didMount() {
    if (platform.inBrowser) {
      this.windowEl = DefaultDOMElement.wrapNativeElement(window);
      this.windowEl.on('resize', this.onSelectionPositioned, this);
    }
  }

  dispose() {
    if (this.windowEl) {
      this.windowEl.off(this);
    }
  }

  getName() {
    return this.props.name
  }

  /*
    Determine selection rectangle relative to content element
    and emit a selection:positioned event with positioning hints
  */
  onSelectionPositioned() {
    let contentRect = this._getContentRect();
    let selectionRect = this._getSelectionRect();
    if (!selectionRect) return
    let hints = {
      contentRect,
      selectionRect
    };
    this._emitSelectionPositioned(hints);
    this._scrollSelectionIntoView(selectionRect);
  }

  _emitSelectionPositioned(hints) {
    // Allows overlays to do positioning relative to the current
    // selection bounding rectangle.
    this.emit('selection:positioned', hints);
    // TODO: Remove legacy support
    this.emit('dom-selection:rendered', hints);
  }

  /*
    Determine mouse bounds relative to content element
    and emit context-menu:opened event with positioning hints
  */
  _onContextMenu(e) {
    e.preventDefault();
    let mouseBounds = this._getMouseBounds(e);
    this.emit('context-menu:opened', {
      mouseBounds: mouseBounds
    });
  }

  _scrollSelectionIntoView(selectionRect) {
    let upperBound = this.getScrollPosition();
    let lowerBound = upperBound + this.getHeight();
    let selTop = selectionRect.top;
    let selBottom = selectionRect.top + selectionRect.height;
    if ((selTop < upperBound && selBottom < upperBound) ||
        (selTop > lowerBound && selBottom > lowerBound)) {
      this.setScrollPosition(selTop);
    }
  }

  /**
    Returns the height of scrollPane (inner content overflows)
  */
  getHeight() {
    throw new Error('Abstract method')
  }

  /**
    Returns the cumulated height of a panel's content
  */
  getContentHeight() {
    throw new Error('Abstract method')
  }

  getContentElement() {
    // TODO: should be wrapped in DefaultDOMElement
    throw new Error('Abstract method')
  }

  // /**
  //   Get the `.se-scrollable` element
  // */
  getScrollableElement() {
    throw new Error('Abstract method')
  }

  /**
    Get current scroll position (scrollTop) of `.se-scrollable` element
  */
  getScrollPosition() {
    throw new Error('Abstract method')
  }

  setScrollPosition() {
    throw new Error('Abstract method')
  }

  /**
    Get offset relative to `.se-content`.

    @param {DOMNode} el DOM node that lives inside the
  */
  getPanelOffsetForElement(el) { // eslint-disable-line
    throw new Error('Abstract method')
  }

  /**
    Scroll to a given sub component.

    @param {String} componentId component id, must be present in data-id attribute
  */
  scrollTo(componentId, onlyIfNotVisible) { // eslint-disable-line
    throw new Error('Abstract method')
  }

  _getContentRect() {
    return this.getContentElement().getNativeElement().getBoundingClientRect()
  }

  /*
    Get selection rectangle relative to panel content element
  */
  _getSelectionRect() {
    return getSelectionRect(this._getContentRect())
  }

  _getMouseBounds(e) {
    return getRelativeMouseBounds(e, this.getContentElement().getNativeElement())
  }

}

/**
  Commands are used to perform UI triggered actions on the document. For instance the
  {@link ui/AnnotationCommand} takes care of creating, expanding, truncating and
  deleting annotations such as strong and emphasis. It does so by determining a
  commandState by inspecting the current selection, which is used to parametrize
  the corresponding tool component. E.g. the strong tool gets active and clickable
  in create mode when a word in the text is selected. Triggered by a click on the tool,
  or a keyboard shortcut, the command gets executed by running the code specified in the
  execute method.

  @class Command
  @abstract

  @example

  ```
  class MyCommand extends Command {
    getCommandState(params, context) {
      // determine commandState based on params and context
    }

    execute(params, context) {
      // perform operations on the document
    }
  }
  ```
*/
class Command {

  /**
    Construcutor is only used internally.

    @constructor
    @param {Object} config    Config provided during command registration
  */
  constructor(config) {
    this.config = config || {};
    this.name = this.config.name;
    if (!this.name) {
      throw new Error("'name' is required");
    }
  }

  get isAsync() {
    return false
  }

  /**
    Get the command name specified at command registration. See
    {@link Configurator#addCommand}
  */
  getName() {
    return this.name
  }

  /**
    Determines command state, based on passed params and context. The command
    state is usually used as props for tool components.

    @example

    This shows the implementation of {@link EditAnnotationCommand#getCommandState}

    ```
    getCommandState(params) {
      const sel = this._getSelection(params)
      const annos = params.selectionState.getAnnotationsForType(this.config.nodeType)
      const newState = {
        disabled: true,
      }

      if (annos.length === 1 && sel.isPropertySelection()) {
        newState.disabled = false
        newState.node = annos[0]
      }
    }
    ```

    @param {Object} params      Provides editorSession, selectionState, surface, selection
    @param {Object} context     Provides app-specific context.
  */
  getCommandState(params, context) { // eslint-disable-line
    throw new Error('Command.getCommandState() is abstract.')
  }

  /**
    Execute command and perform operations on the document

    @param {Object} params      Provides commandState, editorSession, selectionState, surface, selection
    @param {Object} context     Provides app-specific context.

    @return {Object} info object with execution details
  */
  execute(params, context) { // eslint-disable-line
    throw new Error('Command.execute() is abstract.')
  }

  _getEditorSession(params, context) {
    let editorSession = params.editorSession || context.editorSession;
    if (!editorSession) {
      throw new Error("'editorSession' is required.")
    }
    return editorSession
  }

  _getSelection(params) {
    let sel = params.selection || params.selectionState.getSelection();
    if (!sel) {
      throw new Error("'selection' is required.")
    }
    return sel
  }

}

Command.prototype._isCommand = true;

/**
  A class for commands intended to be executed on the annotations.

  See the example below to learn how to register an `AnnotationCommand`
  for a strong annotation.

  @class AnnotationCommand
  @extends ui/Command

  @example

  ```js
  import { AnnotationCommand } from 'substance'

  config.addCommand('strong', AnnotationCommand, {nodeType: 'strong'})
  // Disable, when cursor is collapsed
  config.addCommand('strong', AnnotationCommand, {
    nodeType: 'strong',
    disableCollapsedCursor: true
  })
  ```
*/

class AnnotationCommand extends Command {

  constructor(...args) {
    super(...args);

    if (!this.config.nodeType) {
      throw new Error("'nodeType' is required")
    }
  }

  /**
    Get the type of an annotation.

    @returns {String} The annotation's type.
   */
  getAnnotationType() {
    return this.config.nodeType
  }

  /**
    Get the annotation's data.

    @returns {Object} The annotation's data.
   */
  getAnnotationData() {
    return {}
  }

  /**
    Checks if command couldn't be executed with current selection.

    @param {Array} annos annotations
    @param {Object} sel selection

    @returns {Boolean} Whether or not command could be executed.
   */
  isDisabled(sel, params) {
    let selectionState = params.selectionState;
    // TODO: Container selections should be valid if the annotation type
    // is a container annotation. Currently we only allow property selections.
    if (!sel || sel.isNull() || !sel.isAttached() || sel.isCustomSelection()||
        sel.isNodeSelection() || sel.isContainerSelection() || selectionState.isInlineNodeSelection()) {
      return true
    }

    if (this.config.disableCollapsedCursor && sel.isCollapsed()) {
      return true
    }

    return false
  }

  /**
    Checks if new annotations could be created.
    There should be no annotation overlapping, selection must be not collapsed.

    @param {Array} annos annotations
    @param {Object} sel selection

    @returns {Boolean} Whether or not annotation could be created.
   */
  // When there's no existing annotation overlapping, we create a new one.
  canCreate(annos, sel) {
    return (annos.length === 0 && !sel.isCollapsed())
  }

  /**
    Checks if annotations could be fused.
    There should be more than one annotation overlaped by current selection.

    @param {Array} annos annotations
    @param {Object} sel selection

    @returns {Boolean} Whether or not annotations could be fused.
   */
  canFuse(annos, sel) {
    // When more than one annotation overlaps with the current selection
    return (annos.length >= 2 && !sel.isCollapsed())
  }

  /**
    Checks if annotation could be deleted.
    Cursor or selection must be inside an existing annotation.

    @param {Array} annos annotations
    @param {Object} sel selection

    @returns {Boolean} Whether or not annotation could be deleted.
   */
  canDelete(annos, sel) {
    // When the cursor or selection is inside an existing annotation
    if (annos.length !== 1) return false
    let annoSel = annos[0].getSelection();
    return sel.isInsideOf(annoSel)
  }

  /**
    Checks if annotation could be expanded.
    There should be overlap with only a single annotation,
    selection should be also outside of this annotation.

    @param {Array} annos annotations
    @param {Object} sel selection

    @returns {Boolean} Whether or not annotation could be expanded.
   */
  canExpand(annos, sel) {
    // When there's some overlap with only a single annotation we do an expand
    if (annos.length !== 1) return false
    let annoSel = annos[0].getSelection();
    return sel.overlaps(annoSel) && !sel.isInsideOf(annoSel)
  }

  /**
    Checks if annotation could be truncated.
    There should be overlap with only a single annotation,
    selection should also have boundary in common with this annotation.

    @param {Array} annos annotations
    @param {Object} sel selection

    @returns {Boolean} Whether or not annotation could be truncated.
   */
  canTruncate(annos, sel) {
    if (annos.length !== 1) return false
    let annoSel = annos[0].getSelection();

    return (sel.isLeftAlignedWith(annoSel) || sel.isRightAlignedWith(annoSel)) &&
           !sel.contains(annoSel) &&
           !sel.isCollapsed()
  }

  /**
    Gets command state object.

    @param {Object} state.selection the current selection
    @returns {Object} info object with command details.
  */
  getCommandState(params) { // eslint-disable-line

    let sel = this._getSelection(params);
    // We can skip all checking if a disabled condition is met
    // E.g. we don't allow toggling of property annotations when current
    // selection is a container selection
    if (this.isDisabled(sel, params)) {
      return {
        disabled: true
      }
    }
    let annos = this._getAnnotationsForSelection(params);
    let newState = {
      disabled: false,
      active: false,
      mode: null
    };
    if (this.canCreate(annos, sel)) {
      newState.mode = 'create';
    } else if (this.canFuse(annos, sel)) {
      newState.mode = 'fuse';
    } else if (this.canTruncate(annos, sel)) {
      newState.active = true;
      newState.mode = 'truncate';
    } else if (this.canExpand(annos, sel)) {
      newState.mode = 'expand';
    } else if (this.canDelete(annos, sel)) {
      newState.active = true;
      newState.mode = 'delete';
    } else {
      newState.disabled = true;
    }
    return newState
  }

  /**
    Execute command and trigger transformation.

    @returns {Object} info object with execution details.
  */
  // Execute command and trigger transformations
  execute(params) {
    // Disabled the next line as I believe it is
    // always passed via params already
    // params.selection = this._getSelection(params)
    let commandState = params.commandState;

    if (commandState.disabled) return false
    switch(commandState.mode) {
      case 'create':
        return this.executeCreate(params)
      case 'fuse':
        return this.executeFuse(params)
      case 'truncate':
        return this.executeTruncate(params)
      case 'expand':
        return this.executeExpand(params)
      case 'delete':
        return this.executeDelete(params)
      default:
        console.warn('Command.execute(): unknown mode', commandState.mode);
        return false
    }
  }

  executeCreate(params) {
    let annos = this._getAnnotationsForSelection(params);
    this._checkPrecondition(params, annos, this.canCreate);
    let editorSession = this._getEditorSession(params);
    let annoData = this.getAnnotationData();
    annoData.type = this.getAnnotationType();
    let anno;
    editorSession.transaction((tx) => {
      anno = tx.annotate(annoData);
    });
    return {
      mode: 'create',
      anno: anno
    }
  }

  executeFuse(params) {
    let annos = this._getAnnotationsForSelection(params);
    this._checkPrecondition(params, annos, this.canFuse);
    this._applyTransform(params, function(tx) {
      annotationHelpers.fuseAnnotation(tx, annos);
    });
    return {
      mode: 'fuse',
      anno: annos[0]
    }
  }

  executeTruncate(params) {
    let annos = this._getAnnotationsForSelection(params);
    let anno = annos[0];
    this._checkPrecondition(params, annos, this.canTruncate);
    this._applyTransform(params, function(tx) {
      annotationHelpers.truncateAnnotation(tx, anno, params.selection);
    });
    return {
      mode: 'truncate',
      anno: anno
    }
  }

  executeExpand(params) {
    let annos = this._getAnnotationsForSelection(params);
    let anno = annos[0];
    this._checkPrecondition(params, annos, this.canExpand);
    this._applyTransform(params, function(tx) {
      annotationHelpers.expandAnnotation(tx, anno, params.selection);
    });
    return {
      mode: 'expand',
      anno: anno
    }
  }

  executeDelete(params) {
    let annos = this._getAnnotationsForSelection(params);
    let anno = annos[0];
    this._checkPrecondition(params, annos, this.canDelete);
    this._applyTransform(params, function(tx) {
      return tx.delete(anno.id)
    });
    return {
      mode: 'delete',
      annoId: anno.id
    }
  }

  _checkPrecondition(params, annos, checker) {
    let sel = this._getSelection(params);
    if (!checker.call(this, annos, sel)) {
      throw new Error("AnnotationCommand: can't execute command for selection " + sel.toString())
    }
  }

  _getAnnotationsForSelection(params) {
    return params.selectionState.getAnnotationsForType(this.getAnnotationType())
  }

  /**
    Apply an annotation transformation.

    @returns {Object} transformed annotations.
   */
  _applyTransform(params, transformFn) {
    let sel = this._getSelection(params);
    if (sel.isNull()) return

    let editorSession = this._getEditorSession(params);
    let result; // to store transform result
    editorSession.setSelection(sel);
    editorSession.transaction(function(tx) {
      let out = transformFn(tx, params);
      if (out) result = out.result;
    });
    return result
  }

}

/**
  Default Tool implementation

  A tool must be associated with a Command, which holds all the logic, while the tool
  is just the visual representation of the command state. You can use this component
  for simple button-like tools, or extend it to create your own UI.

  @class
  @component

  @example

  Usually instantiated in a Toolbar or an Overlay. Usage:

  ```
  $$(Tool, {
    icon: 'strong',
    label: 'strong',
    style: 'outline',
    active: false,
    disabled: false
  })
  ```


  ```
  config.addCommand('strong', AnnotationCommand, { nodeType: 'strong' })
  config.addTool('strong', AnnotationTool, {
    target: 'annotations'
  })
  ```
*/
class Tool extends Component {

  get _isTool() {
    return true
  }

  /**
    Default tool rendering. You can override this method to provide your custom markup
  */
  render($$) {
    let el = $$('div')
      .addClass('se-tool');

    let customClassNames = this.getClassNames();
    if (customClassNames) {
      el.addClass(customClassNames);
    }

    let title = this.getTitle();
    if (title) {
      el.attr('title', title);
      el.attr('aria-label', title);
    }

    el.append(
      this.renderButton($$)
    );
    return el
  }

  renderButton($$) {
    let Button = this.getComponent('button');
    let btn = $$(Button, {
      icon: this.props.showIcon ? this.props.name : null,
      label: this.props.showLabel ? this.props.name : null,
      hint: this.props.showHint ? this.props.name : null,
      active: this.props.active,
      disabled: this.props.disabled,
      style: this.props.style
    }).on('click', this.onClick);
    return btn
  }

  getClassNames() {
    return ''
  }

  getTitle() {
    let labelProvider = this.context.labelProvider;
    let title = this.props.title || labelProvider.getLabel(this.getName());
    // Used only by annotation tool so far
    if (this.props.mode) {
      title = [capitalize(this.props.mode), title].join(' ');
    }
    return title
  }

  /*
    For now always same as tool name
  */
  getCommandName() {
    return this.getName()
  }

  getName() {
    return this.props.name
  }

  onClick(e) {
    e.preventDefault();
    e.stopPropagation();
    if (!this.props.disabled) this.executeCommand();
  }

  /**
    Executes the associated command
  */
  executeCommand(props) {
    props = Object.assign({ mode: this.props.mode }, props);
    this.context.commandManager.executeCommand(this.getCommandName(), props);
  }
}

/**

  Reuseable AnnotationTool component. Can be used without modification
  for pure marker annotations that don't carry data. E.g. strong, emphasis,
  superscript, etc.

  @class
  @component

  @example

  ```
  config.addTool('strong', AnnotationTool, {
    target: 'annotations'
  })
  ```
*/

class AnnotationTool extends Tool {

  render($$) {
    let el = super.render.call(this, $$);
    el.addClass('sm-annotation-tool');
    return el
  }

  renderButton($$) {
    let el = super.renderButton.call(this, $$);
    el.append(this.renderMode($$));
    return el
  }

  /*
    Renders a small hint for the mode (expand, truncate, edit, etc)
  */
  renderMode($$) {
    let mode = this.props.mode;
    let el = $$('div').addClass('se-mode');

    let iconEl = this.context.iconProvider.renderIcon($$, mode);
    if (iconEl) {
      el.append(iconEl);
    }
    return el
  }
}

class NodeComponent extends Component {

  render($$) {
    let tagName = this.getTagName();
    let el = $$(tagName)
      .attr('data-id', this.props.node.id)
      .addClass(this.getClassNames());
    return el
  }

  getTagName() {
    return 'div'
  }

  getClassNames() {
    return ''
  }

}

const INLINENODES = ['a','b','big','i','small','tt','abbr','acronym','cite','code','dfn','em','kbd','strong','samp','time','var','bdo','br','img','map','object','q','script','span','sub','sup','button','input','label','select','textarea'].reduce((m,n)=>{m[n]=true;return m}, {});

/**
  Import HTML from clipboard. Used for inter-application copy'n'paste.

  @internal
*/
class ClipboardImporter extends HTMLImporter {

  constructor(config) {
    super(_withCatchAllConverter(config));
    // disabling warnings about default importers
    this.IGNORE_DEFAULT_WARNINGS = true;

    Object.assign(config, {
      trimWhitespaces: true,
      REMOVE_INNER_WS: true
    });

    // ATTENTION: this is only here so we can enfore windows conversion
    // mode from within tests
    this._isWindows = platform.isWindows;
  }

  /**
    Parses HTML and applies some sanitization/normalization.
  */
  importDocument(html) {
    if (this._isWindows) {
      // Under windows we can exploit <!--StartFragment--> and <!--EndFragment-->
      // to have an easier life
      let match = /<!--StartFragment-->(.*)<!--EndFragment-->/.exec(html);
      if (match) {
        html = match[1];
      }
    }

    // when copying from a substance editor we store JSON in a script tag in the head
    // If the import fails e.g. because the schema is incompatible
    // we fall back to plain HTML import
    if (html.search(/script id=.substance-clipboard./)>=0) {
      let htmlDoc = DefaultDOMElement.parseHTML(html);
      let substanceData = htmlDoc.find('#substance-clipboard');
      if (substanceData) {
        let jsonStr = substanceData.textContent;
        try {
          return this.importFromJSON(jsonStr)
        } finally {
          // nothing
        }
      }
    }

    let htmlDoc = DefaultDOMElement.parseHTML(html);
    let body = htmlDoc.find('body');
    body = this._sanitizeBody(body);
    if (!body) {
      console.warn('Invalid HTML.');
      return null
    }
    this._wrapIntoParagraph(body);
    this.reset();
    this.convertBody(body);
    const doc = this.state.doc;
    return doc
  }

  _sanitizeBody(body) {
    body = this._fixupGoogleDocsBody(body);
    // Remove <meta> element
    body.findAll('meta').forEach(el => el.remove());
    return body
  }

  _fixupGoogleDocsBody(body) {
    if (!body) return
    // Google Docs has a strange convention to use a bold tag as
    // container for the copied elements
    // HACK: we exploit the fact that this element has an id with a
    // specific format, e.g., id="docs-internal-guid-5bea85da-43dc-fb06-e327-00c1c6576cf7"
    let bold = body.find('b');
    if (bold && /^docs-internal/.exec(bold.id)) {
      return bold
    }
    return body
  }

  _wrapIntoParagraph(body) {
    let childNodes = body.getChildNodes();
    let shouldWrap = false;
    for (let i = 0; i < childNodes.length; i++) {
      const c = childNodes[i];
      if (c.isTextNode()) {
        if (!(/^\s+$/.exec(c.textContent))) {
          shouldWrap = true;
          break
        }
      } else if (INLINENODES[c.tagName]) {
        shouldWrap = true;
        break
      }
    }
    if (shouldWrap) {
      let p = body.createElement('p');
      p.append(childNodes);
      body.append(p);
    }
  }

  importFromJSON(jsonStr) {
    this.reset();
    let doc = this.getDocument();
    let jsonData = JSON.parse(jsonStr);
    let converter = new JSONConverter();
    converter.importDocument(doc, jsonData);
    return doc
  }

  /**
    Converts all children of a given body element.

    @param {String} body body element of given HTML document
  */
  convertBody(body) {
    this.convertContainer(body.childNodes, Document.SNIPPET_ID);
  }

  /**
    Creates substance document to paste.

    @return {Document} the document instance
  */
  _createDocument() {
    let emptyDoc = super._createDocument();
    return emptyDoc.createSnippet()
  }
}

function _withCatchAllConverter(config) {
  let defaultTextType = config.schema.getDefaultTextType();
  config.converters = config.converters.concat([{
    type: defaultTextType,
    matchElement: function(el) { return el.is('div') },
    import: function(el, node, converter) {
      node.content = converter.annotatedText(el, [node.id, 'content']);
    }
  }]);
  return config
}

/**
  Export HTML from clipboard. Used for inter-application copy'n'paste.

  @internal
*/
class ClipboardExporter extends HTMLExporter {

  /**
    Exports document in html format.

    @param {Document} doc document to export

    @return {String} html representation of given document
  */
  exportDocument(doc) {
    this.state.doc = doc;
    let html;
    let elements = this.convertDocument(doc);
    // special treatment for a text snippet
    if (elements.length === 1 && elements[0].attr('data-id') === Document.TEXT_SNIPPET_ID) {
      html = elements[0].innerHTML;
    } else {
      html = elements.map(function(el) {
        return el.outerHTML
      }).join('');
    }
    let jsonConverter = new JSONConverter();
    let jsonStr = JSON.stringify(jsonConverter.exportDocument(doc));
    let substanceContent = `<script id="substance-clipboard" type="application/json">${jsonStr}</script>`;
    return '<html><head>' +substanceContent+ '</head><body>' + html + '</body></html>'
  }

  /**
    Coverts document to set of DOM elements.

    @param {Document} doc document to convert

    @return {Array} array of DOM elements each represented single node
  */
  convertDocument(doc) {
    let content = doc.get(Document.SNIPPET_ID);
    if (!content) {
      throw new Error('Illegal clipboard document: could not find container "' + Document.SNIPPET_ID + '"')
    }
    return this.convertContainer(content)
  }

}

/**
  The Clipboard is a Component which should be rendered as a sibling component
  of one or multiple Surfaces.

  It uses the JSONImporter and JSONExporter for internal copy'n'pasting,
  i.e., within one window or between two instances with the same DocumentSchema.

  For inter-application copy'n'paste, the ClipboardImporter and ClipboardExporter is used.

  @internal
*/
class Clipboard {

  constructor(editorSession, config) {
    this.editorSession = editorSession;
    let doc = editorSession.getDocument();
    let schema = doc.getSchema();

    let htmlConverters = [];
    if (config.converterRegistry) {
      htmlConverters = config.converterRegistry.get('html').values() || [];
    }
    let _config = {
      schema: schema,
      DocumentClass: doc.constructor,
      converters: htmlConverters
    };

    this.htmlImporter = new ClipboardImporter(_config);
    this.htmlExporter = new ClipboardExporter(_config);
  }

  dispose() {
    this.htmlImporter.dispose();
    // does not need to be disposed
    // this.htmlExporter.dispose()
  }

  getEditorSession() {
    return this.editorSession
  }

  /*
    Called by to enable clipboard handling on a given root element.
  */
  attach(el) {
    el.on('copy', this.onCopy, this);
    el.on('cut', this.onCut, this);
    el.on('paste', this.onPaste, this);
  }

  /*
    Called by to disable clipboard handling.
  */
  detach(el) {
    el.off(this);
  }

  /*
    Called when copy event fired.

    @param {Event} event
  */
  onCopy(event) {
    // console.log("Clipboard.onCopy", arguments);
    let clipboardData = this._copy();
    substanceGlobals._clipboardData = event.clipboardData;

    if (event.clipboardData && clipboardData.doc) {
      event.preventDefault();
      // store as plain text and html
      event.clipboardData.setData('text/plain', clipboardData.text);
      // WORKAROUND: under IE and Edge it is not permitted to set 'text/html'
      if (!platform.isIE && !platform.isEdge) {
        event.clipboardData.setData('text/html', clipboardData.html);
      }
    }
  }

  /*
    Called when cut event fired.

    @param {Event} event
  */
  onCut(event) {
    // preventing default behavior to avoid that contenteditable
    // destroys our DOM
    event.preventDefault();
    // console.log("Clipboard.onCut", arguments);
    this.onCopy(event);
    let editorSession = this.getEditorSession();
    editorSession.transaction((tx)=>{
      tx.deleteSelection();
    });
  }

  /*
    Called when paste event fired.

    @param {Event} event
  */
  // Works on Safari/Chrome/FF
  onPaste(event) {
    let clipboardData = event.clipboardData;

    let types = {};
    for (let i = 0; i < clipboardData.types.length; i++) {
      types[clipboardData.types[i]] = true;
    }
    // console.log('onPaste(): received content types', types);

    event.preventDefault();
    event.stopPropagation();

    let plainText;
    let html;
    if (types['text/plain']) {
      plainText = clipboardData.getData('text/plain');
    }
    if (types['text/html']) {
      html = clipboardData.getData('text/html');
    }

    // HACK: to allow at least in app copy and paste under Edge (which blocks HTML)
    // we guess by comparing the old and new plain text
    if (platform.isEdge &&
        substanceGlobals.clipboardData &&
        substanceGlobals.clipboardData.text === plainText) {
      html = substanceGlobals.clipboardData.html;
    } else {
      substanceGlobals.clipboardData = {
        text: plainText,
        html: html
      };
    }

    // console.log('onPaste(): html = ', html);

    // WORKAROUND: FF does not provide HTML coming in from other applications
    // so fall back to pasting plain text
    if (platform.isFF && !html) {
      this._pastePlainText(plainText);
      return
    }

    // if we have content given as HTML we let the importer assess the quality first
    // and fallback to plain text import if it's bad
    if (html) {
      if (!this._pasteHtml(html, plainText)) {
        this._pastePlainText(plainText);
      }
    } else {
      this._pastePlainText(plainText);
    }
  }

  /*
    Pastes a given plain text into the surface.

    @param {String} plainText plain text
  */
  _pastePlainText(plainText) {
    let editorSession = this.getEditorSession();
    editorSession.transaction(function(tx) {
      tx.paste(plainText);
    }, { action: 'paste' });
  }

  /*
    Copies selected content from document to clipboard.
  */
  _copy() {
    let editorSession = this.getEditorSession();
    let sel = editorSession.getSelection();
    let doc = editorSession.getDocument();
    let clipboardDoc = null;
    let clipboardText = "";
    let clipboardHtml = "";
    if (!sel.isCollapsed()) {
      clipboardText = documentHelpers.getTextForSelection(doc, sel) || "";
      clipboardDoc = copySelection(doc, sel);
      clipboardHtml = this.htmlExporter.exportDocument(clipboardDoc);
    }
    return {
      doc: clipboardDoc,
      html: clipboardHtml,
      text: clipboardText
    }
  }

  /*
    Pastes a given parsed html document into the surface.

    @param {ui/DOMElement} docElement
    @param {String} text plain text representation used as a fallback
  */
  _pasteHtml(html, text) {
    let content = this.htmlImporter.importDocument(html);
    this.paste(content, text);
    return true
  }

  /*
    Takes a clipboard document and pastes it at the current
    cursor position.

    Used by PasteCommand
  */
  paste(doc, text) {
    let content = doc || text;
    let editorSession = this.getEditorSession();
    if (content) {
      editorSession.transaction((tx) => {
        tx.paste(content);
      }, { action: 'paste' });
    }
  }

}

/*
  Listens to changes on the document and selection and updates the commandStates
  accordingly.

  @class CommandManager
*/
class CommandManager {

  constructor(context, commands) {
    if (!context.editorSession) {
      throw new Error('EditorSession required.')
    }

    this.editorSession = context.editorSession;
    this.doc = this.editorSession.getDocument();
    this.context = Object.assign({}, context, {
      // for convenienve we provide access to the doc directly
      doc: this.doc
    });
    // Set up command registry
    this.commandRegistry = new Registry();
    forEach(commands, function(command) {
      if(!command._isCommand) {
        throw new Error("Expecting instances of ui/Command.")
      }
      this.commandRegistry.add(command.name, command);
    }.bind(this));

    this.editorSession.onUpdate(this.onSessionUpdate, this);
    this.updateCommandStates(this.editorSession);
  }

  dispose() {
    this.editorSession.off(this);
  }

  onSessionUpdate(editorSession) {
    if (editorSession.hasChanged('change') || editorSession.hasChanged('selection')) {
      this.updateCommandStates(editorSession);
    }
  }

  /*
    Compute new command states object
  */
  updateCommandStates(editorSession) {
    const commandContext = this.getCommandContext();
    const params = this._getCommandParams();
    const surface = params.surface;
    const commandRegistry = this.commandRegistry;

    // EXPERIMENTAL:
    // We want to control which commands are available
    // in each surface
    // Trying out a white-list and a black list
    // TODO: discuss, and maybe think about optimizing this
    // by caching the result...
    let commandNames = commandRegistry.names.slice();
    if (surface) {
      let included = surface.props.commands;
      let excluded = surface.props.excludedCommands;
      if (included) {
        commandNames = included;
      } else if (excluded) {
        excluded = excluded.slice(0);
        for (let i = commandNames.length - 1; i >= 0; i--) {
          let idx = excluded.indexOf(commandNames[i]);
          if (idx >= 0) {
            excluded.splice(idx, 1);
            commandNames.splice(i, 1);
          }
        }
      }
    }
    const commands = commandNames.map(name => commandRegistry.get(name));
    let commandStates = {};
    commands.forEach((cmd) => {
      if (cmd) {
        commandStates[cmd.getName()] = cmd.getCommandState(params, commandContext);
      }
    });
    // NOTE: We previously did a check if commandStates were actually changed
    // before updating them. However, we currently have complex objects
    // in the command state (e.g. EditInlineNodeCommand) so we had to remove it.
    // See Issue #1004
    this.commandStates = commandStates;
    editorSession.setCommandStates(commandStates);
  }

  /*
    Execute a command, given a context and arguments.

    Commands are run async if cmd.isAsync() returns true.
  */
  executeCommand(commandName, userParams, cb) {
    let cmd = this.commandRegistry.get(commandName);
    if (!cmd) {
      console.warn('command', commandName, 'not registered');
      return
    }
    let commandState = this.commandStates[commandName];
    let params = Object.assign(this._getCommandParams(), userParams, {
      commandState: commandState
    });

    if (cmd.isAsync) {
      // TODO: Request UI lock here
      this.editorSession.lock();
      cmd.execute(params, this.getCommandContext(), (err, info) => {
        if (err) {
          if (cb) {
            cb(err);
          } else {
            console.error(err);
          }
        } else {
          if (cb) cb(null, info);
        }
        this.editorSession.unlock();
      });
    } else {
      let info = cmd.execute(params, this.getCommandContext());
      return info
    }
  }

  /*
    Exposes the current commandStates object
  */
  getCommandStates() {
    return this.commandStates
  }

  getCommandContext() {
    return this.context
  }

  // TODO: while we need it here this should go into the flow thingie later
  _getCommandParams() {
    let editorSession = this.context.editorSession;
    let selectionState = editorSession.getSelectionState();
    let sel = selectionState.getSelection();
    let surface = this.context.surfaceManager.getFocusedSurface();
    return {
      editorSession: editorSession,
      selectionState: selectionState,
      surface: surface,
      selection: sel,
    }
  }
}

class ComponentRegistry extends Registry {
  constructor(entries) {
    super(entries, function(ComponentClass) {
      if (!ComponentClass.prototype._isComponent) {
        throw new Error('Component registry: wrong type. Expected a ComponentClass. Was: ' + String(ComponentClass))
      }
    });
  }
}

class FontAwesomeIcon extends Component {
  constructor(...args) {
    super(...args);
  }

  render($$) {
    return $$('i').addClass('fa ' + this.props.icon)
  }

}

class FontAwesomeIconProvider {

  constructor(icons) {
    this.faMap = {};
    this.textMap = {};
    forEach(icons, function(config, name) {
      let faClass = config['fontawesome'];
      if (faClass) {
        this.addFAIcon(name, faClass);
      }
      let text = config['text'];
      if (text) {
        this.addTextIcon(name, text);
      }
    }.bind(this));
  }

  renderIcon($$, name) {
    let faClass = this.faMap[name];
    let text = this.textMap[name];
    if (faClass) {
      return $$(FontAwesomeIcon, { icon: faClass })
    } else if (text) {
      return text
    }
  }

  addFAIcon(name, faClass) {
    this.faMap[name] = faClass;
  }

  addTextIcon(name, text) {
    this.textMap[name] = text;
  }
}

/*
 Default label provider implementation
*/
class DefaultLabelProvider {
  constructor(labels, lang) {
    this.lang = lang || 'en';
    this.labels = labels;
  }

  getLabel(name) {
    let labels = this.labels[this.lang];
    if (!labels) return name
    return labels[name] || name
  }
}

class ToolGroup extends Component {

  render($$) {

    let tools = this.props.tools;
    let el = $$('div').addClass('sc-tool-group');
    el.addClass('sm-target-'+this.props.name);
    if (this.props.layout) {
      el.addClass('sm-layout-'+this.props.layout);
    }

    tools.forEach((tool) => {
      let toolProps = Object.assign({}, tool.toolProps, {
        showIcon: this.props.showIcons,
        showLabel: this.props.showLabels,
        style: this.props.toolStyle
      });

      el.append(
        $$(tool.Class, toolProps).ref(tool.name)
      );
    });
    return el
  }
}

class SaveHandlerStub {

  /*
    Saving a document involves two steps.

    - syncing files (e.g. images) with a backend
    - storing a snapshot of the document's content (e.g. a XML serialization)
  */
  saveDocument({fileManager}) {
    console.info('Simulating save ...');

    return fileManager.sync()
    .then(() => {
      // Here you would run a converter (HTML/XML) usually
      // and send the result to a REST endpoint.
      console.info('Creating document snapshot...');
    })

  }
}

/**
  Default Configurator for Substance editors. It provides an API for
  adding nodes to the schema, components, commands and tools etc.

  @class

  @example

  ```js
  let configurator = new Configurator()

  configurator.addNode(Heading)
  configurator.addComponent('heading', HeadingComponent)
  ```

  To modularize configuration, package definitions can be imported.

  ```js
  configurator.import(ParagraphPackage)
  ```

  You can create your own extensions that way.

  ```js

  const AlienPackage = {
    name: 'alien'
    configure: function(config) {
      config.addNode(AlienNode)
      config.addComponent('alien', AlienComponent)
      config.addCommand('add-alien', AddAlienCommand)
      config.addTool('add-alien', AddAlienTool)
    }
  }

  ```

  From within a package, another package can be imported. This provides
  a simple mechanism to model dependencies between packages. Just make
  sure you don't run into cyclic dependencies as there is no checking for
  that at the moment.
*/
class Configurator {
  constructor() {
    this.config = {
      schema: {},
      nodes: {},
      components: {},
      converters: {},
      importers: {},
      exporters: {},
      fileProxies: [],
      commands: {},
      tools: new Map(),
      toolGroups: new Map(),
      textTypes: [],
      editingBehaviors: [],
      macros: [],
      dropHandlers: [],
      keyboardShortcuts: [],
      icons: {},
      labels: {},
      lang: 'en_US',
      SaveHandlerClass: null
    };
  }

  // Record phase API
  // ------------------------

  /**
    Defines the document schema for this configuration.

    @param  {DocumentSchema} schema A schema to be used for articles created
        from this configuration.
   */
  defineSchema(schema) {
    if (schema.ArticleClass) {
      console.warn('DEPRECATED: schema.ArticleClass is now called schema.DocumentClass');
      schema.DocumentClass = schema.ArticleClass;
    }
    if (!schema.DocumentClass) {
      throw new Error('schema.DocumentClass is mandatory')
    }
    this.config.schema = schema;
  }

  /**
    Adds a node to this configuration. Later, when you use
    {@link Configurator#getSchema()}, this node will be added to that schema.
    Usually, used within a package to add its own nodes to the schema.

    @param {Node} NodeClass
   */
  addNode(NodeClass) {
    var type = NodeClass.type;
    if (!type) {
      throw new Error('A NodeClass must have a type.')
    }
    if (this.config.nodes[type]) {
      throw new Error('NodeClass with this type name is already registered: ' + type)
    }
    this.config.nodes[type] = NodeClass;
  }

  /**
    Adds a converter for a conversion format.

    @param {string} type      a conversion format type, eg. 'html', 'xml', 'json'
    @param {Object} converter a converter for that format.
   */
  addConverter(type, converter) {
    var converters = this.config.converters[type];
    if (!converters) {
      converters = {};
      this.config.converters[type] = converters;
    }
    if (!converter.type) {
      throw new Error('A converter needs an associated type.')
    }
    converters[converter.type] = converter;
  }

  /**
    Add importer for a conversion format.

    @param {string} type          a conversion format type. eg. 'html', 'xml'
    @param {Object} ImporterClass an importer for the conversion format.
   */
  addImporter(type, ImporterClass) {
    this.config.importers[type] = ImporterClass;
  }

  /**
    Add exporter for a conversion format.

    @param {string} type          a conversion format type. eg. 'html', 'xml'
    @param {Object} ExporterClass an exporter for the conversion format.
   */
  addExporter(type, ExporterClass) {
    this.config.exporters[type] = ExporterClass;
  }

  /**
    Add a component for a node type. Components ({@link Component}) are the
    ui representation of a node for rendering and manipulation. This is usually
    used within a package to add representations for nodes added by that
    package.

    A component can be added once per nodeType. If you provide two components
    for the same node type, Substance can't figure out which one to use.

    @param {String} nodeType       the type attribute of the node for which this
                                   component is to be used.
    @param {Class} ComponentClass  A subclass of {@link Component} for nodes
                                   of nodeType.
   */
  addComponent(nodeType, ComponentClass, force) {
    if (!force && this.config.components[nodeType]) {
      throw new Error(nodeType+' already registered')
    }
    if (!ComponentClass) {
      throw new Error('Provided nil for component '+nodeType)
    }
    if (!ComponentClass.prototype._isComponent) {
      throw new Error('ComponentClass must be a subclass of ui/Component.')
    }
    this.config.components[nodeType] = ComponentClass;
  }

  addCommand(name, CommandClass, options) {
    if (!isString$1(name)) {
      throw new Error("Expecting 'name' to be a String")
    }
    if (!CommandClass) {
      throw new Error('Provided nil for command '+name)
    }
    if (!CommandClass.prototype._isCommand) {
      throw new Error("Expecting 'CommandClass' to be of type ui/Command.")
    }
    this.config.commands[name] = {
      name: name,
      CommandClass: CommandClass,
      options: options || {}
    };
  }

  addToolGroup(name, ToolGroupClass, options) {
    options = options || {};
    ToolGroupClass = ToolGroupClass || ToolGroup;

    this.config.toolGroups.set(name, {
      name: name,
      tools: new Map(),
      Class: ToolGroupClass,
      options: options
    });
  }

  addTool(name, ToolClass, options) {
    options = options || {};

    if (options.target) {
      console.warn('DEPRECATED: please use `toolGroup` instead of `target`', name);
    }
    let toolGroupNames = options.toolGroup || options.target;
    if (isString$1(toolGroupNames)) {
      toolGroupNames = [ toolGroupNames ];
    }

    if (!toolGroupNames && options.overlay) {
      toolGroupNames = [ 'overlay' ];
    } else if (!toolGroupNames) {
      toolGroupNames = [ 'default' ];
    }

    if (!isString$1(name)) {
      throw new Error("Expecting 'name' to be a String")
    }
    if (!ToolClass) {
      throw new Error('Provided nil for tool '+name)
    }
    if (!ToolClass || !ToolClass.prototype._isTool) {
      throw new Error("Expecting 'ToolClass' to be of type ui/Tool. name:")
    }

    toolGroupNames.forEach((toolGroupName) => {
      let toolGroup = this.config.toolGroups.get(toolGroupName);
      if (!toolGroup) {
        console.error(`No toolGroup registered with name: ${toolGroupName}`);
        return
      }
      toolGroup.tools.set(name, {
        name: name,
        Class: ToolClass,
        options: options || {}
      });
    });
  }

  /**
    Adds an icon to the configuration which can be later retrieved via the
    iconProvider.

    @param {string} iconName name or key for retrieving the icon
    @param {Object} options  your custom method of representing the icon as a
        JSON object. Enables plugging in your own IconProvider.
   */
  addIcon(iconName, options) {
    var iconConfig = this.config.icons[iconName];
    if (!iconConfig) {
      iconConfig = {};
      this.config.icons[iconName] = iconConfig;
    }
    Object.assign(iconConfig, options);
  }

  /**
    Define a new label
    Label is either a string or a hash with translations.
    If string is provided 'en' is used as the language.

    @param {String} labelName name of label.
    @param {String} label label.

    @example

    ```
    // Using english only.
    config.addLabel('paragraph.content', 'Paragraph')

    // Using multiple languages
    config.addLabel('superscript', {
      en: 'Superscript',
      de: 'Hochgestellt'
    })

    .
    .
    // Usage within other code
    let labels = this.context.labelProvider
    $$('span').append(labels.getLabel('superscript'))
    ```
  */
  addLabel(labelName, label) {
    if (isString$1(label)) {
      if(!this.config.labels['en']) {
        this.config.labels['en'] = {};
      }
      this.config.labels['en'][labelName] = label;
    } else {
      forEach(label, function(label, lang) {
        if (!this.config.labels[lang]) {
          this.config.labels[lang] = {};
        }
        this.config.labels[lang][labelName] = label;
      }.bind(this));
    }
  }

  /**
    Replaces the seed function for this configuration.

    Use a seed function to create the empty state for your document. This should
    be used only once per configuration. You shouldn't call this within package
    config methods.

    You can use {@link Configurator#getSeed} method to get this seed and
    apply it on your document {@link Document} class.

    @param {function} seed   A transaction function that creates the seed
        document from an empty document.

    @example

    ```js
    var seedFn = function(tx) {
      var body = tx.get('body')

      tx.create({
        id: 'p1',
        type: 'paragraph',
        content: 'This is your new paragraph!'
      })
      body.show('p1')
    }

    config.addSeed(seedFn)
    ```
  */
  addSeed(seed) {
    this.config.seed = seed;
  }

  addTextType(textType, options) {
    this.config.textTypes.push({
      spec: textType,
      options: options || {}
    });
  }

  /**
    Adds an editing behavior to this configuration. {@link EditingBehavior}
    for more.

    @param {EditingBehavior} editingBehavior.
   */
  addEditingBehavior(editingBehavior) {
    this.config.editingBehaviors.push(editingBehavior);
  }

  addMacro(macro) {
    this.config.macros.push(macro);
  }

  addDragAndDrop(DragAndDropHandlerClass) {
    // we deprecated this after it became more clear what
    // we actually needed to solve
    console.warn('DEPRECATED: Use addDropHandler() instead');
    if (!DragAndDropHandlerClass.prototype._isDragAndDropHandler) {
      throw new Error('Only instances of DragAndDropHandler are allowed.')
    }
    this.addDropHandler(new DragAndDropHandlerClass());
  }

  addDropHandler(dropHandler) {
    // legacy
    if (dropHandler._isDragAndDropHandler) {
      dropHandler.type = dropHandler.type || 'drop-asset';
    }
    this.config.dropHandlers.push(dropHandler);
  }

  addKeyboardShortcut(combo, spec) {
    let entry = {
      key: combo,
      spec: spec
    };
    this.config.keyboardShortcuts.push(entry);
  }

  addFileProxy(FileProxyClass) {
    this.config.fileProxies.push(FileProxyClass);
  }

  getFileAdapters() {
    return this.config.fileProxies.slice(0)
  }

  /**
    Configure this instance of configuration for provided package.
    @param  {Object} pkg     Object should contain a `configure` method that
                             takes a Configurator instance as the first method.
    @param  {Object} options Additional options to pass to the
                             package.`configure` method

    @return {configurator}   returns the configurator instance to make it easy
                             to chain calls to import.
   */
  import(pkg, options) {
    pkg.configure(this, options || {});
    return this
  }

  // Config Interpreter APIs
  // ------------------------

  getConfig() {
    return this.config
  }

  getStyles() {
    return this.config.styles
  }

  getSchema() {
    if (!this.schema) {
      this.schema = new DocumentSchema(this.config.schema);
      this.schema.addNodes(this.config.nodes);
    }
    return this.schema
  }

  getDocumentClass() {
    return this.config.schema.DocumentClass
  }

  createArticle(seed) {
    const schema = this.getSchema();
    const DocumentClass = schema.getDocumentClass();
    let doc = new DocumentClass(schema);
    if (seed) {
      seed(doc);
    }
    return doc
  }

  createImporter(type, context, options = {}) {
    var ImporterClass = this.config.importers[type];
    var config = Object.assign({
      schema: this.getSchema(),
      converters: this.getConverterRegistry().get(type).values(),
    }, options);
    return new ImporterClass(config, context)
  }

  createExporter(type, context, options = {}) {
    var ExporterClass = this.config.exporters[type];
    var config = Object.assign({
      schema: this.getSchema(),
      converters: this.getConverterRegistry().get(type).values()
    }, options);
    return new ExporterClass(config, context)
  }

  getToolGroups() {
    return this.config.toolGroups
  }

  getTools(toolGroupName) {
    return this.config.toolGroups.get(toolGroupName).tools
  }

  getComponentRegistry() {
    var componentRegistry = new ComponentRegistry();
    forEach(this.config.components, function(ComponentClass, name) {
      componentRegistry.add(name, ComponentClass);
    });
    return componentRegistry
  }

  getCommands() {
    return map(this.config.commands, function(item, name) {
      return new item.CommandClass(Object.assign({name: name}, item.options))
    })
  }

  getSurfaceCommandNames() {
    var commands = this.getCommands();
    var commandNames = commands.map(function(C) {
      return C.type
    });
    return commandNames
  }

  /*
    A converter registry is a registry by file type and then by node type

    `configurator.getConverterRegistry().get('html').get('paragraph')` provides
    a HTML converter for Paragraphs.
  */
  getConverterRegistry() {
    if (!this.converterRegistry) {
      var converterRegistry = new Registry();
      forEach(this.config.converters, function(converters, name) {
        converterRegistry.add(name, new Registry(converters));
      });
      this.converterRegistry = converterRegistry;
    }
    return this.converterRegistry
  }

  getDropHandlers() {
    return this.config.dropHandlers.slice(0)
  }

  getSeed() {
    return this.config.seed
  }

  getTextTypes() {
    return this.config.textTypes.map(function(t) {
      return t.spec
    })
  }

  getIconProvider() {
    return new FontAwesomeIconProvider(this.config.icons)
  }

  getLabelProvider() {
    return new DefaultLabelProvider(this.config.labels)
  }

  getEditingBehavior() {
    var editingBehavior = new EditingBehavior();
    this.config.editingBehaviors.forEach(function(behavior) {
      behavior.register(editingBehavior);
    });
    return editingBehavior
  }

  getMacros() {
    return this.config.macros
  }

  getKeyboardShortcuts() {
    return this.config.keyboardShortcuts
  }

  setDefaultLanguage(lang) {
    this.config.lang = lang;
  }

  getDefaultLanguage() {
    return this.config.lang || 'en_US'
  }

  setSaveHandlerClass(SaveHandlerClass) {
    this.config.SaveHandlerClass = SaveHandlerClass;
  }

  getSaveHandler() {
    let SaveHandler = this.config.SaveHandlerClass || SaveHandlerStub;
    return new SaveHandler()
  }
}

class UnsupportedNodeComponent extends Component {

  render($$) {
    return $$('pre')
      .addClass('content-node unsupported')
      .attr({
        'data-id': this.props.node.id,
        contentEditable: false
      })
      .append(
        JSON.stringify(this.props.node.properties, null, 2)
      )
  }
}

/**
   Abstract interface for editing components.
   Dances with contenteditable, so you don't have to.
*/
class Surface extends Component {

  constructor(...args) {
    super(...args);

    // EditorSession instance must be provided either as a prop
    // or via dependency-injection
    this.editorSession = this.props.editorSession || this.context.editorSession;
    if (!this.editorSession) {
      throw new Error('No EditorSession provided')
    }
    this.name = this.props.name;
    if (!this.name) {
      throw new Error('Surface must have a name.')
    }
    if (this.name.indexOf('/') > -1) {
      // because we are using '/' to deal with nested surfaces (isolated nodes)
      throw new Error("Surface.name must not contain '/'")
    }
    // this path is an identifier unique for this surface
    // considering nesting in IsolatedNodes
    this._surfaceId = createSurfaceId(this);

    this.clipboard = new Clipboard(this.editorSession, {
      converterRegistry: this.context.converterRegistry
    });

    this.domSelection = this.context.domSelection;
    if (!this.domSelection) throw new Error('DOMSelection instance must be provided via context.')

    this.domObserver = null;

    // HACK: we need to listen to mousup on document
    // to catch events outside the surface
    if (platform.inBrowser) {
      this.documentEl = DefaultDOMElement.wrapNativeElement(window.document);
    }

    // set when editing is enabled
    this.undoEnabled = true;

    // a registry for TextProperties which allows us to dispatch changes
    this._textProperties = {};

    this._state = {
      // true if the document session's selection is addressing this surface
      skipNextFocusEvent: false
    };
  }

  getChildContext() {
    return {
      surface: this,
      doc: this.getDocument(),
      // HACK: clearing isolatedNodeComponent so that we can easily know
      // if this surface is within an isolated node
      isolatedNodeComponent: null
    }
  }

  didMount() {
    if (this.context.surfaceManager) {
      this.context.surfaceManager.registerSurface(this);
    }
    this.editorSession.onRender('selection', this._onSelectionChanged, this);
  }


  dispose() {
    this.editorSession.off(this);
    this.clipboard.dispose();
    if (this.domObserver) {
      this.domObserver.disconnect();
    }
    if (this.context.surfaceManager) {
      this.context.surfaceManager.unregisterSurface(this);
    }
  }

  didUpdate() {
    this._updateContentEditableState();
  }

  render($$) {
    let tagName = this.props.tagName || 'div';
    let el = $$(tagName)
      .addClass('sc-surface')
      .attr('tabindex', 2)
      .attr('data-surface-id', this.id);

    if (!this.isDisabled()) {
      if (this.isEditable()) {
        // Keyboard Events
        el.on('keydown', this.onKeyDown);
        // OSX specific handling of dead-keys
        if (!platform.isIE) {
          el.on('compositionstart', this.onCompositionStart);
        }
        // Note: TextEvent in Chrome/Webkit is the easiest for us
        // as it contains the actual inserted string.
        // Though, it is not available in FF and not working properly in IE
        // where we fall back to a ContentEditable backed implementation.
        if (platform.inBrowser && window.TextEvent && !platform.isIE) {
          el.on('textInput', this.onTextInput);
        } else {
          el.on('keypress', this.onTextInputShim);
        }
      }
      if (!this.isReadonly()) {
        // Mouse Events
        el.on('mousedown', this.onMouseDown);
        el.on('contextmenu', this.onContextMenu);
        // disable drag'n'drop
        // we will react on this to render a custom selection
        el.on('focus', this.onNativeFocus);
        el.on('blur', this.onNativeBlur);
        // activate the clipboard
        this.clipboard.attach(el);
      }

    }
    return el
  }

  renderNode($$, node) {
    let doc = this.getDocument();
    let componentRegistry = this.getComponentRegistry();
    let ComponentClass = componentRegistry.get(node.type);
    if (!ComponentClass) {
      console.error('Could not resolve a component for type: ' + node.type);
      ComponentClass = UnsupportedNodeComponent;
    }
    return $$(ComponentClass, {
      doc: doc,
      node: node
    }).ref(node.id)
  }

  getComponentRegistry() {
    return this.context.componentRegistry || this.props.componentRegistry
  }

  getName() {
    return this.name
  }

  getId() {
    return this._surfaceId
  }

  isDisabled() {
    return this.props.disabled
  }

  isEditable() {
    return (this.props.editing === "full" || this.props.editing === undefined)
  }

  isSelectable() {
    return (this.props.editing === "selection" || this.props.editing === "full")
  }

  isReadonly() {
    return this.props.editing === "readonly"
  }

  getElement() {
    return this.el
  }

  getDocument() {
    return this.editorSession.getDocument()
  }

  getEditorSession() {
    return this.editorSession
  }

  isEnabled() {
    return !this.state.disabled
  }

  isContainerEditor() {
    return false
  }

  hasNativeSpellcheck() {
    return this.props.spellcheck === 'native'
  }

  getContainerId() {
    return null
  }

  blur() {
    if (this.el) {
      this.el.blur();
    }
  }

  focus() {
    if (this.isDisabled()) return
    // console.log('Focusing surface %s explicitly with Surface.focus()', this.getId());
    // NOTE: FF is causing problems with dynamically activated contenteditables
    // and focusing
    if (platform.isFF) {
      this.domSelection.clear();
      this.el.getNativeElement().blur();
    }
    this._focus();
  }

  // As the DOMSelection is owned by the Editor now, rerendering could now be done by someone else, e.g. the SurfaceManager?
  rerenderDOMSelection() {
    if (this.isDisabled()) return
    if (platform.inBrowser) {
      // console.log('Surface.rerenderDOMSelection', this.__id__);
      let sel = this.editorSession.getSelection();
      if (sel.surfaceId === this.getId()) {
        this.domSelection.setSelection(sel);
        // HACK: accessing the scrollpane directly
        // TODO: this should be done in a different way
        // this will let the scrollpane know that the DOM selection is ready
        const scrollPane = this.context.scrollPane;
        if (scrollPane) {
          this.context.scrollPane.onSelectionPositioned();
        }
      }
    }
  }

  getDomNodeForId(nodeId) {
    return this.el.getNativeElement().querySelector('*[data-id="'+nodeId+'"]')
  }

  /* Event handlers */

  /*
   * Handle document key down events.
   */
  onKeyDown(event) {
    if (!this._shouldConsumeEvent(event)) return
    // console.log('Surface.onKeyDown()', this.getId());

    // ignore fake IME events (emitted in IE and Chromium)
    if ( event.key === 'Dead' ) return

    // keyboard shortcuts
    let custom = this.editorSession.keyboardManager.onKeydown(event);
    if (!custom) {
      // core handlers for cursor movements and editor interactions
      switch ( event.keyCode ) {
        // Cursor movements
        case keys$1.LEFT:
        case keys$1.RIGHT:
          return this._handleLeftOrRightArrowKey(event)
        case keys$1.UP:
        case keys$1.DOWN:
          return this._handleUpOrDownArrowKey(event)
        case keys$1.HOME:
        case keys$1.END:
          return this._handleHomeOrEndKey(event)
        case keys$1.PAGEUP:
        case keys$1.PAGEDOWN:
          return this._handlePageUpOrDownKey(event)
        // Input (together with text-input)
        case keys$1.ENTER:
          return this._handleEnterKey(event)
        case keys$1.TAB:
          return this._handleTabKey(event)
        case keys$1.BACKSPACE:
        case keys$1.DELETE:
          return this._handleDeleteKey(event)
        default:
          break
      }
    }
  }

  onTextInput(event) {
    if (!this._shouldConsumeEvent(event)) return
    // console.log("TextInput:", event);
    event.preventDefault();
    event.stopPropagation();
    if (!event.data) return

    let text = event.data;
    if (!this.editorSession.keyboardManager.onTextInput(text)) {
      this.editorSession.transaction((tx) => {
        tx.insertText(text);
      }, { action: 'type' });
    }
  }

  // Handling Dead-keys under OSX
  onCompositionStart(event) {
    if (!this._shouldConsumeEvent(event)) return
  }

  // TODO: do we need this anymore?
  onTextInputShim(event) {
    if (!this._shouldConsumeEvent(event)) return
    // Filter out non-character keys
    if (
      // Catches most keys that don't produce output (charCode === 0, thus no character)
      event.which === 0 || event.charCode === 0 ||
      // Opera 12 doesn't always adhere to that convention
      event.keyCode === keys$1.TAB || event.keyCode === keys$1.ESCAPE ||
      // prevent combinations with meta keys, but not alt-graph which is represented as ctrl+alt
      Boolean(event.metaKey) || (Boolean(event.ctrlKey)^Boolean(event.altKey))
    ) {
      return
    }
    let character = String.fromCharCode(event.which);
    if (!event.shiftKey) {
      character = character.toLowerCase();
    }
    event.preventDefault();
    event.stopPropagation();
    if (!this.editorSession.keyboardManager.onTextInput(character)) {
      if (character.length>0) {
        this.editorSession.transaction((tx) => {
          tx.insertText(character);
        }, { action: 'type' });
      }
    }
  }

  // TODO: the whole mouse event based selection mechanism needs
  // to be redesigned. The current implementation works basically
  // though, there are some things which do not work well cross-browser
  // particularly, double- and triple clicks.
  // also it turned out to be problematic to react on mouse down instantly
  onMouseDown(event) {
    if (!this._shouldConsumeEvent(event)) {
      // console.log('skipping mousedown', this.id)
      return
    }

    // EXPERIMENTAL: trying to 'reserve' a mousedown event
    // so that parents know that they shouldn't react
    // This is similar to event.stopPropagation() but without
    // side-effects.
    // Note: some browsers do not do clicks, selections etc. on children if propagation is stopped
    if (event.__reserved__) {
      // console.log('%s: mousedown already reserved by %s', this.id, event.__reserved__.id)
      return
    } else {
      // console.log('%s: taking mousedown ', this.id)
      event.__reserved__ = this;
    }

    // NOTE: this is here to make sure that this surface is contenteditable
    // For instance, IsolatedNodeComponent sets contenteditable=false on this element
    // to achieve selection isolation
    if (this.isEditable()) {
      this.el.setAttribute('contenteditable', true);
    }

    // TODO: what is this exactly?
    if ( event.button !== 0 ) {
      return
    }

    // special treatment for triple clicks
    if (!(platform.isIE && platform.version<12) && event.detail >= 3) {
      let sel = this.getEditorSession().getSelection();
      if (sel.isPropertySelection()) {
        this._selectProperty(sel.path);
        event.preventDefault();
        event.stopPropagation();
        return
      } else if (sel.isContainerSelection()) {
        this._selectProperty(sel.startPath);
        event.preventDefault();
        event.stopPropagation();
        return
      }
    }
    // 'mouseDown' is triggered before 'focus' so we tell
    // our focus handler that we are already dealing with it
    // The opposite situation, when the surface gets focused e.g. using keyboard
    // then the handler needs to kick in and recover a persisted selection or such
    this._state.skipNextFocusEvent = true;

    // Bind mouseup to the whole document in case of dragging out of the surface
    if (this.documentEl) {
      // TODO: we should handle mouse up only if we started a drag (and the selection has really changed)
      this.documentEl.on('mouseup', this.onMouseUp, this, { once: true });
    }
  }

  onMouseUp(e) {
    // console.log('Surface.onMouseup', this.id);
    // ATTENTION: filtering events does not make sense here,
    // as we need to make sure that pick the selection even
    // when the mouse is released outside the surface
    // if (!this._shouldConsumeEvent(e)) return
    e.stopPropagation();
    // ATTENTION: this delay is necessary for cases the user clicks
    // into an existing selection. In this case the window selection still
    // holds the old value, and is set to the correct selection after this
    // being called.
    setTimeout(function() {
      let sel = this.domSelection.getSelection();
      this._setSelection(sel);
    }.bind(this));
  }

  // When a user right clicks the DOM selection is updated (in Chrome the nearest
  // word gets selected). Like we do with the left mouse clicks we need to sync up
  // our model selection.
  onContextMenu(event) {
    if (!this._shouldConsumeEvent(event)) return
    let sel = this.domSelection.getSelection();
    this._setSelection(sel);
  }

  onNativeBlur() {
    // console.log('Native blur on surface', this.getId());
    let _state = this._state;
    _state.hasNativeFocus = false;
  }

  onNativeFocus() {
    // console.log('Native focus on surface', this.getId());
    let _state = this._state;
    _state.hasNativeFocus = true;
  }

  // Internal implementations


  _onSelectionChanged(selection) {
    let newMode = this._deriveModeFromSelection(selection);
    if (this.state.mode !== newMode) {
      this.extendState({
        mode: newMode
      });
    }
  }

  // helper to manage surface mode which is derived from the current selection
  _deriveModeFromSelection(sel) {
    if (!sel) return null
    let surfaceId = sel.surfaceId;
    let id = this.getId();
    let mode;
    if (startsWith(surfaceId, id)) {
      if (surfaceId.length === id.length) {
        mode = 'focused';
      } else {
        mode = 'co-focused';
      }
    }
    return mode
  }

  _updateContentEditableState() {
    // NOTE: managing contenteditable is difficult in
    // order to achieve a correct behavior for IsolatedNodes
    // For 'closed' isolated nodes it is important that the parents'
    // contenteditables are all false. Otherwise, the cursor
    // can leave the isolated area.
    let enableContenteditable = false;
    if (this.isEditable() && !this.props.disabled) {
      enableContenteditable = true;
      if (this.state.mode === 'co-focused') {
        let selState = this.context.editorSession.getSelectionState();
        let sel = selState.getSelection();
        let surface = this.context.surfaceManager.getSurface(sel.surfaceId);
        if (surface) {
          let isolatedNodeComponent = surface.context.isolatedNodeComponent;
          if (isolatedNodeComponent) {
            enableContenteditable = isolatedNodeComponent.isOpen();
          }
        }
      }
    }
    if (enableContenteditable) {
      this.el.setAttribute('contenteditable', true);
    } else {
      // TODO: find out what is better
      this.el.removeAttribute('contenteditable');
    }
  }

  _focus() {
    this._state.hasNativeFocus = true;
    // HACK: we must not focus explicitly in Chrome/Safari
    // as otherwise we get a crazy auto-scroll
    // Still, this is ok, as everything is working fine
    // there, without that (as opposed to FF/Edge)
    if (this.el && !platform.isWebkit) {
      this._state.skipNextFocusEvent = true;
      // ATTENTION: unfortunately, focusing the contenteditable does lead to auto-scrolling
      // in some browsers
      this.el.focus();
      this._state.skipNextFocusEvent = false;
    }
  }

  _handleLeftOrRightArrowKey(event) {
    event.stopPropagation();
    let direction = (event.keyCode === keys$1.LEFT) ? 'left' : 'right';
    // Note: we need this timeout so that CE updates the DOM selection first
    // before we map it to the model
    window.setTimeout(function() {
      this._updateModelSelection({direction});
    }.bind(this));
  }

  _handleUpOrDownArrowKey(event) {
    event.stopPropagation();
    // Note: we need this timeout so that CE updates the DOM selection first
    // before we map it to the model
    window.setTimeout(function() {
      let options = {
        direction: (event.keyCode === keys$1.UP) ? 'left' : 'right'
      };
      this._updateModelSelection(options);
    }.bind(this));
  }

  _handleHomeOrEndKey(event) {
    event.stopPropagation();
    // Note: we need this timeout so that CE updates the DOM selection first
    // before we map it to the model
    window.setTimeout(function() {
      let options = {
        direction: (event.keyCode === keys$1.HOME) ? 'left' : 'right'
      };
      this._updateModelSelection(options);
    }.bind(this));
  }

  _handlePageUpOrDownKey(event) {
    event.stopPropagation();
    // Note: we need this timeout so that CE updates the DOM selection first
    // before we map it to the model
    window.setTimeout(function() {
      let options = {
        direction: (event.keyCode === keys$1.PAGEUP) ? 'left' : 'right'
      };
      this._updateModelSelection(options);
    }.bind(this));
  }

  _handleTabKey(event) {
    event.stopPropagation();
    if (this.props.handleTab === false) {
      event.preventDefault();
      this.el.emit('tab', {
        altKey: event.altKey,
        ctrlKey: event.ctrlKey,
        metaKey: event.metaKey,
        shiftKey: event.shiftKey,
        code: event.code
      });
    } else {
      window.setTimeout(()=>{
        this._updateModelSelection();
      });
    }
  }

  _handleEnterKey(event) {
    event.preventDefault();
    event.stopPropagation();
    this.editorSession.transaction((tx) => {
      tx.break();
    }, { action: 'break' });
  }

  _handleDeleteKey(event) {
    event.preventDefault();
    event.stopPropagation();
    let direction = (event.keyCode === keys$1.BACKSPACE) ? 'left' : 'right';
    this.editorSession.transaction((tx) => {
      tx.deleteCharacter(direction);
    }, { action: 'delete' });
  }

  _hasNativeFocus() {
    return Boolean(this._state.hasNativeFocus)
  }

  _setSelection(sel) {
    // Since we allow the surface be blurred natively when clicking
    // on tools we now need to make sure that the element is focused natively
    // when we set the selection
    // This is actually only a problem on FF, other browsers set the focus implicitly
    // when a new DOM selection is set.
    // ATTENTION: in FF 44 this was causing troubles, making the CE unselectable
    // until the next native blur.
    // TODO: check if this is still necessary
    if (!sel.isNull() && sel.surfaceId === this.id && platform.isFF) {
      this._focus();
    }
    this.editorSession.setSelection(sel);
  }

  _updateModelSelection(options) {
    let sel = this.domSelection.getSelection(options);
    // console.log('Surface: updating model selection', sel.toString());
    // NOTE: this will also lead to a rerendering of the selection
    // via session.on('update')
    this._setSelection(sel);
  }

  _selectProperty(path) {
    let doc = this.getDocument();
    let text = doc.get(path);
    this._setSelection(doc.createSelection({
      type: 'property',
      path: path,
      startOffset: 0,
      endOffset: text.length
    }));
  }

  // internal API for TextProperties to enable dispatching
  // TextProperty components are registered via path
  _registerTextProperty(textPropertyComponent) {
    let path = textPropertyComponent.getPath();
    this._textProperties[path] = textPropertyComponent;
  }

  _unregisterTextProperty(textPropertyComponent) {
    let path = textPropertyComponent.getPath();
    if (this._textProperties[path] === textPropertyComponent) {
      delete this._textProperties[path];
    }
  }

  _getTextPropertyComponent(path) {
    return this._textProperties[path]
  }

  // TODO: we could integrate container node rendering into this helper
  // TODO: this helper should be available also in non surface context
  _renderNode($$, nodeId) {
    let doc = this.getDocument();
    let node = doc.get(nodeId);
    let componentRegistry = this.context.componentRegistry || this.props.componentRegistry;
    let ComponentClass = componentRegistry.get(node.type);
    if (!ComponentClass) {
      console.error('Could not resolve a component for type: ' + node.type);
      ComponentClass = UnsupportedNodeComponent;
    }
    return $$(ComponentClass, {
      doc: doc,
      node: node
    })
  }

  // only take care of events which are emitted on targets which belong to this surface
  _shouldConsumeEvent(event) {
    // console.log('should consume?', event.target, this.id)
    let comp = Component.unwrap(event.target);
    return (comp && (comp === this || comp.context.surface === this))
  }

  // Experimental: used by DragManager
  getSelectionFromEvent(event) {
    let domRange = getDOMRangeFromEvent(event);
    let sel = this.domSelection.getSelectionForDOMRange(domRange);
    sel.surfaceId = this.getId();
    return sel;
  }

  setSelectionFromEvent(event) {
    let sel = this.getSelectionFromEvent(event);
    if (sel) {
      this._state.skipNextFocusEvent = true;
      this._setSelection(sel);
    } else {
      console.error('Could not create a selection from event.');
    }
  }

  get id() {
    return this._surfaceId
  }

}

Surface.prototype._isSurface = true;

/*
  Computes the id of a surface

  With IsolatedNodes, surfaces can be nested.
  In this case the id can be seen as a path from the top-most to the nested ones

  @examples

  - top-level surface: 'body'
  - table cell: 'body/t1/t1-A1.content'
  - figure caption: 'body/fig1/fig1-caption.content'
  - nested containers: 'body/section1'
*/
function createSurfaceId(surface) {
  let isolatedNodeComponent = surface.context.isolatedNodeComponent;
  if (isolatedNodeComponent) {
    let parentSurface = isolatedNodeComponent.context.surface;
    // nested containers
    if (surface.isContainerEditor()) {
      if (isolatedNodeComponent._isInlineNodeComponent) {
        return parentSurface.id + '/' + isolatedNodeComponent.props.node.id + '/' + surface.name
      } else {
        return parentSurface.id + '/' + surface.name
      }
    }
    // other isolated nodes such as tables, figures, etc.
    else {
      return parentSurface.id + '/' + isolatedNodeComponent.props.node.id + '/' + surface.name
    }
  } else {
    return surface.name
  }
}

/**
  Represents an editor for content rendered in a flow, such as a manuscript.

  @prop {String} name unique editor name
  @prop {String} containerId container id

  @example

  Create a full-fledged `ContainerEditor` for the `body` container of a document.
  Allow Strong and Emphasis annotations and to switch text types between paragraph
  and heading at level 1.

  ```js
  $$(ContainerEditor, {
    name: 'bodyEditor',
    containerId: 'body'
  })
  ```
*/

class ContainerEditor extends Surface {

  constructor(parent, props, el) {
    // default props derived from the given props
    props.containerId = props.containerId || props.node.id;
    props.name = props.name || props.containerId || props.node.id;

    super(parent, props, el);

    this.containerId = this.props.containerId;
    if (!isString$1(this.containerId)) {
      throw new Error("Property 'containerId' is mandatory.")
    }
    let doc = this.getDocument();
    this.container = doc.get(this.containerId);
    if (!this.container) {
      throw new Error('Container with id ' + this.containerId + ' does not exist.')
    }

    this.textTypes = this.props.textTypes || [];

    this.editingBehavior = this.context.editingBehavior || new EditingBehavior();

    this._deriveInternalState(this.props);
  }

  // Note: this component is self managed
  shouldRerender(newProps) {
    if (newProps.disabled !== this.props.disabled) return true
    // TODO: we should still detect when the document has changed,
    // see https://github.com/substance/substance/issues/543
    return false
  }

  willReceiveProps(newProps) {
    super.willReceiveProps.apply(this, arguments);
    this._deriveInternalState(newProps);
  }

  didMount() {
    super.didMount.apply(this, arguments);
    let editorSession = this.getEditorSession();
    editorSession.onUpdate('document', this._onContainerChanged, this, {
      path: [this.getContainerId(), 'nodes']
    });

  }

  dispose() {
    super.dispose.apply(this, arguments);
    let editorSession = this.getEditorSession();
    editorSession.off(this);
  }

  render($$) {
    let el = super.render($$);

    let doc = this.getDocument();
    let containerId = this.getContainerId();
    let containerNode = doc.get(containerId);
    if (!containerNode) {
      console.warn('No container node found for ', containerId);
    }
    el.addClass('sc-container-editor container-node ' + containerId)
      .attr("data-id", containerId);

    // native spellcheck
    el.attr('spellcheck', this.props.spellcheck === 'native');

    containerNode.getNodes().forEach(function(node) {
      el.append(this._renderNode($$, node));
    }.bind(this));

    // No editing if disabled by user or container is empty
    if (!this.props.disabled && !this.isEmpty()) {
      el.addClass('sm-enabled');
      el.setAttribute('contenteditable', true);
    }

    return el
  }

  _renderNode($$, node) {
    if (!node) throw new Error('Illegal argument')
    if (node.isText()) {
      return super.renderNode($$, node)
    } else {
      let componentRegistry = this.context.componentRegistry;
      let ComponentClass = componentRegistry.get(node.type);
      if (ComponentClass.prototype._isCustomNodeComponent || ComponentClass.prototype._isIsolatedNodeComponent) {
        return $$(ComponentClass, { node: node }).ref(node.id)
      } else {
        return $$(IsolatedNodeComponent, { node: node }).ref(node.id)
      }
    }
  }

  _deriveInternalState(props) {
    let _state = this._state;
    if (!props.hasOwnProperty('enabled') || props.enabled) {
      _state.enabled = true;
    } else {
      _state.enabled = false;
    }
  }

  _selectNextIsolatedNode(direction) {
    let selState = this.getEditorSession().getSelectionState();
    let node = (direction === 'left') ? selState.getPreviousNode() : selState.getNextNode();
    if (!node || !node.isIsolatedNode()) return false
    if (
      (direction === 'left' && selState.isFirst()) ||
      (direction === 'right' && selState.isLast())
    ) {
      this.getEditorSession().setSelection({
        type: 'node',
        nodeId: node.id,
        containerId: selState.getContainer().id,
        surfaceId: this.id
      });
      return true
    }
    return false
  }

  _handleLeftOrRightArrowKey(event) {
    event.stopPropagation();
    const doc = this.getDocument();
    const sel = this.getEditorSession().getSelection();
    const left = (event.keyCode === keys$1.LEFT);
    const right = !left;
    const direction = left ? 'left' : 'right';

    if (sel && !sel.isNull()) {
      const container = doc.get(sel.containerId, 'strict');

      // Don't react if we are at the boundary of the document
      if (sel.isNodeSelection()) {
        let nodePos = container.getPosition(doc.get(sel.getNodeId()));
        if ((left && nodePos === 0) || (right && nodePos === container.length-1)) {
          event.preventDefault();
          return
        }
      }

      if (sel.isNodeSelection() && !event.shiftKey) {
        this.domSelection.collapse(direction);
      }
    }

    window.setTimeout(() => {
      this._updateModelSelection({ direction });
    });
  }

  _handleUpOrDownArrowKey(event) {
    event.stopPropagation();
    const doc = this.getDocument();
    const sel = this.getEditorSession().getSelection();
    const up = (event.keyCode === keys$1.UP);
    const down = !up;
    const direction = up ? 'left' : 'right';

    if (sel && !sel.isNull()) {
      const container = doc.get(sel.containerId, 'strict');
      // Don't react if we are at the boundary of the document
      if (sel.isNodeSelection()) {
        let nodePos = container.getPosition(doc.get(sel.getNodeId()));
        if ((up && nodePos === 0) || (down && nodePos === container.length-1)) {
          event.preventDefault();
          return
        }
        // Unfortunately we need to navigate out of an isolated node
        // manually, as even Chrome on Win is not able to do it.
        let editorSession = this.getEditorSession();
        // TODO the following fixes the mentioned problem for
        // regular UP/DOWN (non expanding)
        // For SHIFT+DOWN it happens to work, and only SHIFT-UP when started as NodeSelection needs to be fixed
        if (!event.shiftKey) {
          event.preventDefault();
          if (up) {
            let prev = container.getChildAt(nodePos-1);
            setCursor(editorSession, prev, sel.containerId, 'after');
            return
          } else {
            let next = container.getChildAt(nodePos+1);
            setCursor(editorSession, next, sel.containerId, 'before');
            return
          }
        }
      }
    }

    window.setTimeout(() => {
      this._updateModelSelection({ direction });
    });
  }

  _handleTabKey(event) {
    const editorSession = this.getEditorSession();
    const sel = editorSession.getSelection();
    if (sel.isNodeSelection() && sel.isFull()) {
      const comp = this.refs[sel.getNodeId()];
      if (comp && stepIntoIsolatedNode(editorSession, comp)) {
        event.preventDefault();
        event.stopPropagation();
        return
      }
    }
    super._handleTabKey(event);
  }

  // Used by Clipboard
  isContainerEditor() {
    return true
  }

  /**
    Returns the containerId the editor is bound to
  */
  getContainerId() {
    return this.containerId
  }

  getContainer() {
    return this.getDocument().get(this.getContainerId())
  }

  isEmpty() {
    let containerNode = this.getContainer();
    return (containerNode && containerNode.nodes.length === 0)
  }

  isEditable() {
    return super.isEditable.call(this) && !this.isEmpty()
  }

  getTextTypes() {
    return this.textTypes || []
  }

  // Used by SwitchTextTypeTool
  // TODO: Filter by enabled commands for this Surface
  // TODO: rethink
  getTextCommands() {
    var textCommands = {};
    this.commandRegistry.each(function(cmd) {
      if (cmd.constructor.textTypeName) {
        textCommands[cmd.getName()] = cmd;
      }
    });
    return textCommands
  }

  // called by flow when subscribed resources have been updated
  _onContainerChanged(change) {
    let doc = this.getDocument();
    // first update the container
    let renderContext = RenderingEngine.createContext(this);
    let $$ = renderContext.$$;
    let container = this.getContainer();
    let path = container.getContentPath();
    for (let i = 0; i < change.ops.length; i++) {
      let op = change.ops[i];
      if (op.type === "update" && op.path[0] === path[0]) {
        let diff$$1 = op.diff;
        if (diff$$1.type === "insert") {
          let nodeId = diff$$1.getValue();
          let node = doc.get(nodeId);
          let nodeEl;
          if (node) {
            nodeEl = this._renderNode($$, node);
          } else {
            // node does not exist anymore
            // so we insert a stub element, so that the number of child
            // elements is consistent
            nodeEl = $$('div');
          }
          this.insertAt(diff$$1.getOffset(), nodeEl);
        } else if (diff$$1.type === "delete") {
          this.removeAt(diff$$1.getOffset());
        }
      }
    }
  }
}

ContainerEditor.prototype._isContainerEditor = true;

function setDOMSelection(startNode, startOffset, endNode, endOffset) {
  let wsel = window.getSelection();
  let wrange = window.document.createRange();
  if (startNode._isDOMElement) {
    startNode = startNode.getNativeElement();
  }
  if (!endNode) {
    endNode = startNode;
    endOffset = startOffset;
  }
  if (endNode._isDOMElement) {
    endNode = endNode.getNativeElement();
  }
  wrange.setStart(startNode, startOffset);
  wrange.setEnd(endNode, endOffset);
  wsel.removeAllRanges();
  wsel.addRange(wrange);
}

class DragAndDropHandler {

  match(dragState, context) { // eslint-disable-line
    return false
  }

  drop(dragState, context) { // eslint-disable-line
    // nothing
  }

  get _isDragAndDropHandler() {
    return true
  }

}

class DragManager extends EventEmitter {

  constructor(customDropHandlers, context) {
    super();

    this.context = context;

    let dropAssetHandlers = [];
    let moveInlineHandlers = [];
    customDropHandlers.forEach((h) => {
      // legacy: default type = 'asset'
      let type = h.type || 'drop-asset';
      switch (type) {
        case 'drop-asset': {
          dropAssetHandlers.push(h);
          break
        }
        case 'move-inline': {
          moveInlineHandlers.push(h);
          break
        }
        default:
          console.warn('Unknown type of drop handler.', h);
      }
    });

    // TODO: This could live in the configurator at some point
    this.dropHandlers = [
      // source is a PropertySelection, target is a property
      new MoveInline(moveInlineHandlers),
      // source is a NodeSelection, target is a container position
      new MoveBlockNode(),
      // drop external files
      new InsertNodes(dropAssetHandlers, this.context),
      // dynamic custom handler, activated via custom dropzone
      // not via configuration
      new CustomHandler(),
    ];
    if (platform.inBrowser) {
      this.el = DefaultDOMElement.wrapNativeElement(document);
      this.el.on('dragstart', this.onDragStart, this);
      // this.el.on('dragend', this.onDragEnd, this)
      this.el.on('drop', this.onDragEnd, this);
      this.el.on('dragenter', this.onDragEnter, this);
      this.el.on('dragexit', this.onDragExit, this);
      this.el.on('mousedown', this.onMousedown, this);
    }
  }

  dispose() {
    if (this.el) {
      this.el.off(this);
    }
  }

  onDragStart(e) {
    // console.log('#### DragManager.onDragStart')
    this._initDrag(e, { external: false });
    // Ensure we have a small dragIcon, so dragged content does not eat up
    // all screen space.
    var img = document.createElement("img");
    img.src = "data:image/gif;base64,R0lGODlhAQABAAAAACH5BAEKAAEALAAAAAABAAEAAAICTAEAOw==";
    e.dataTransfer.setDragImage(img, 0, 0);
    // TODO: the following might probably not work in FF as it disallows setting the drag data
    // Note: setData('text/html', ... ) is necessary so that the browser shows
    // the target caret while dragging, the content must be allowed to drop inline
    // console.log('#### dragState.mode', )
    // TODO: at this point we should do it similar as in Clipboard
    // i.e. copy the selection and export it to HTML
    if (this.dragState.mode === 'inline') {
      e.dataTransfer.setData('text/html', img.outerHTML);
    } else {
      // otherwise we clear the data trying to make the caret
      // invisible this way
      e.dataTransfer.setData('text/html', '<div></div>');
    }
    // console.log('####', this.dragState)
  }

  /*
    When drag starts externally, e.g. draggin a file into the workspace
  */
  onDragEnter(e) {
    if (!this.dragState) {
      // console.log('onDragEnter(e)', e)
      this._initDrag(e, {external: true});
    }
  }

  onDragEnd(event) {
    if (event.__reserved__) return
    // console.log('onDragEnd', event)
    if (this.dragState) {
      event.stopPropagation();
      event.preventDefault();
      // HACK: there is no way to know if Dropzones wants to
      // extend state, as it can only do it on drop
      this._onDragEnd(event);
    }
  }

  onDragExit(e) {
    if (platform.isFF) {
      // FF fires this quite rapidly
    } else {
      // TODO: OTOH, we need to find out if this
      // is really necessary in the other browsers
      // console.log('onDragExit', e)
      this._onDragEnd(e);
    }
  }

  extendDragState(extState) {
    Object.assign(this.dragState, extState);
  }

  // used to at least reset on the next mousedown
  // TODO: figure out if we could make this only 'once'
  onMousedown(event) {
    if (this.dragState) {
      this.dragState = null;
      this._onDragEnd(event);
    }
  }

  _onDragEnd(event) {
    if (!this.dragState) {
      // TODO: There are cases where _onDragEnd is called manually via
      // handleDrop and another time via the native dragend event. check
      // why this happens and how it can be avoided
      console.warn('Not in a valid drag state.');
    } else {
      this._handleDrop(event);
    }
    this.emit('drag:finished');
    this.dragState = null;
  }

  /*
    Called by Dropzones component after drop received
  */
  _handleDrop(e) {
    let dragState = this.dragState;
    let i, handler;
    let match = false;
    dragState.event = e;
    dragState.data = this._getData(e);
    // Run through drop handlers and call the first that matches
    for (i = 0; i < this.dropHandlers.length && !match; i++) {
      handler = this.dropHandlers[i];
      match = handler.match(dragState);
    }
    if (match) {
      let editorSession = this.context.editorSession;
      editorSession.transaction((tx) => {
        handler.drop(tx, dragState);
      });
    } else {
      console.error('No drop handler could be found.');
    }
  }

  /*
    Initializes dragState, which encapsulate state through the whole
    drag + drop operation.

    ATTENTION: This can not be debugged properly in Chrome
  */
  _initDrag(event, options) {
    // TODO: we need to figure out how to enable dragging cursors
    // e.g., when dragging an inline node containing an img, it looks
    // nice, showing the target caret and a dragging cursor.
    // Doing the same just with text content does show the forbidden symbol

    // console.log('_initDrag')
    let sel = this._getSelection();
    let dragState = Object.assign({ startEvent: event }, options);
    this.dragState = dragState;

    // external drag
    // Note: we only consider drops on the block-level or with custom dropzones
    if (dragState.external) {
      dragState.selectionDrag = false;
      dragState.sourceSelection = null;
      dragState.scrollPanes = this._getSurfacesGroupedByScrollPane();
      this.emit('drag:started', dragState);
      return
    }

    // Note: selection drags are always without drop-zones,
    // but using the native cursor
    let isSelectionDrag = (
      (sel.isPropertySelection() || sel.isContainerSelection()) &&
      isMouseInsideDOMSelection(event)
    );
    if (isSelectionDrag) {
      // TODO: we do not support dragging of ContainerSelection yet
      if (sel.isContainerSelection()) {
        console.warn('Dragging of ContainerSelection is not supported yet.');
        return _stop()
      }
      // console.log('DragManager: starting a selection drag', sel.toString())
      dragState.inline = true;
      dragState.selectionDrag = true;
      dragState.sourceSelection = sel;
      // TODO: should we emit for dropzones?
      return
    }
    let comp = Component.unwrap(event.target);
    if (!comp) return _stop()
    let isolatedNodeComponent;
    if (comp._isInlineNodeComponent) {
      isolatedNodeComponent = comp;
      dragState.inline = true;
      dragState.sourceNode = comp.props.node;
    } else {
      isolatedNodeComponent = comp.context.isolatedNodeComponent;
    }
    if (!isolatedNodeComponent) return _stop()
    let surface = isolatedNodeComponent.context.surface;
    // dragging an InlineNode
    if(isolatedNodeComponent._isInlineNodeComponent) {
      let inlineNode = isolatedNodeComponent.props.node;
      dragState.inline = true;
      dragState.selectionDrag = true;
      dragState.sourceSelection = {
        type: 'property',
        path: inlineNode.start.path,
        startOffset: inlineNode.start.offset,
        endOffset: inlineNode.end.offset,
        containerId: surface.getContainerId(),
        surfaceId: surface.id
      };
      return
    }
    // dragging an IsolatedNode
    // console.log('DragManager: started dragging a node or from external')
    dragState.selectionDrag = false;
    dragState.nodeDrag = true;
    dragState.sourceSelection = {
      type: 'node',
      nodeId: isolatedNodeComponent.props.node.id,
      containerId: surface.getContainerId(),
      surfaceId: surface.id
    };
    // We store the scrollPanes in dragState so the Dropzones component
    // can use it to compute dropzones per scrollpane for each contained
    // surface
    dragState.scrollPanes = this._getSurfacesGroupedByScrollPane();
    // console.log('... emitting dragstart for Dropzones')
    this.emit('drag:started', dragState);

    function _stop() {
      // console.log('.... NOPE')
      event.preventDefault();
      event.stopPropagation();
    }
  }

  _getSurfacesGroupedByScrollPane() {
    // We need to determine all ContainerEditors and their scrollPanes; those have the drop
    // zones attached
    let surfaces = this.context.surfaceManager.getSurfaces();
    let scrollPanes = {};
    surfaces.forEach((surface) => {
      // Skip for everything but container editors
      if (!surface.isContainerEditor()) return
      let scrollPane = surface.context.scrollPane;
      let scrollPaneName = scrollPane.getName();
      let surfaceName = surface.getName();
      if (!scrollPanes[scrollPaneName]) {
        let surfaces = {};
        surfaces[surfaceName] = surface;
        scrollPanes[scrollPaneName] = { scrollPane, surfaces };
      } else {
        scrollPanes[scrollPaneName].surfaces[surfaceName] = surface;
      }
    });
    return scrollPanes
  }

  _getSelection() {
    return this.context.editorSession.getSelection()
  }

  _getComponents(targetEl) {
    let res = [];
    let curr = targetEl;
    while (curr) {
      let comp = Component.getComponentForDOMElement(curr);
      if (comp) {
        res.unshift(comp);
        if(comp._isSurface) {
          return res
        }
      }
      curr = curr.parentNode;
    }
    return null
  }

  _getIsolatedNodeOrContainerChild(targetEl) {
    let parent, current;
    current = targetEl;
    parent = current.parentNode;
    while(parent) {
      if (parent._comp && parent._comp._isContainerEditor) {
        return current._comp
      } else if (current._comp && current._comp._isIsolatedNode) {
        return current._comp
      }
      current = parent;
      parent = current.parentNode;
    }
  }

  /*
    Following best practice from Mozilla for URI extraction

    See: https://developer.mozilla.org/en-US/docs/Web/Guide/HTML/Recommended_Drag_Types#link
  */
  _extractUris(dataTransfer) {
    let uris = [];
    let rawUriList = dataTransfer.getData('text/uri-list');
    if (rawUriList) {
      uris = rawUriList.split('\n').filter(function(item) {
        return !item.startsWith('#')
      });
    }
    return uris
  }

  /*
    Extracts information from e.dataTransfer (files, uris, text, html)
  */
  _getData(e) {
    let dataTransfer = e.dataTransfer;
    if (dataTransfer) {
      return {
        files: Array.prototype.slice.call(dataTransfer.files),
        uris: this._extractUris(dataTransfer),
        text: dataTransfer.getData('text/plain'),
        html: dataTransfer.getData('text/html')
      }
    }
  }
}

/*
  Moves a selected node to a new location.

  Used as a drop handler for internal drags with NodeSelections.
*/
class MoveBlockNode extends DragAndDropHandler {

  match(dragState) {
    let {insertPos} = dragState.dropParams;
    // - sourceSeletion must be a NodeSelection
    return (!dragState.external && dragState.nodeDrag &&
      dragState.dropType === 'place' && insertPos >= 0
    )
  }

  drop(tx, dragState) {
    // - remember current selection (node that is dragged)
    // - delete current selection (removes node from original position)
    // - determine node selection based on given insertPos
    // - paste node at new insert position
    let { insertPos } = dragState.dropParams;
    tx.setSelection(dragState.sourceSelection);
    let copy = tx.copySelection();
    // just clear, but don't merge or don't insert a new node
    tx.deleteSelection({ clear: true });
    let containerId = dragState.targetSurface.getContainerId();
    let surfaceId = dragState.targetSurface.getName();
    let container = tx.get(containerId);
    let targetNode = container.nodes[insertPos];
    let insertMode = 'before';
    if (!targetNode) {
      targetNode = container.nodes[insertPos-1];
      insertMode = 'after';
    }
    tx.setSelection({
      type: 'node',
      nodeId: targetNode,
      mode: insertMode,
      containerId: containerId,
      surfaceId: surfaceId
    });
    tx.paste(copy);
  }
}

class MoveInline extends DragAndDropHandler {

  match(dragState) {
    return !dragState.external && dragState.inline
  }

  drop(tx, dragState) {
    let event = dragState.event;
    let sourceSel = dragState.sourceSelection;
    let wrange = getDOMRangeFromEvent(event);
    if (!wrange) return
    let comp = Component.unwrap(event.target);
    if (!comp) return
    let domSelection = comp.context.domSelection;
    if (!domSelection) return
    let range = domSelection.mapDOMRange(wrange);
    if (!range) return
    let targetSel = tx.getDocument()._createSelectionFromRange(range);

    // TODO: iterate custom move-inline handlers
    tx.selection = sourceSel;
    let snippet = tx.copySelection();
    tx.deleteSelection();
    tx.selection = transformSelection(targetSel, tx);
    tx.paste(snippet);
  }
}

class InsertNodes extends DragAndDropHandler {
  constructor(assetHandlers, context) {
    super();
    this.assetHandlers = assetHandlers;
    this.context = context;
  }

  match(dragState) {
    return dragState.dropType === 'place' && dragState.external
  }

  drop(tx, dragState) {
    let { insertPos } = dragState.dropParams;
    let files = dragState.data.files;
    let uris = dragState.data.uris;
    let containerId = dragState.targetSurface.getContainerId();
    let surfaceId = dragState.targetSurface.id;
    let container = tx.get(containerId);
    let targetNode = container.nodes[insertPos];
    let insertMode = 'before';
    if (!targetNode) {
      targetNode = container.nodes[insertPos-1];
      insertMode = 'after';
    }
    tx.setSelection({
      type: 'node',
      nodeId: targetNode,
      mode: insertMode,
      containerId: containerId,
      surfaceId: surfaceId
    });
    if (files.length > 0) {
      files.forEach((file) => {
        this._callHandlers(tx, {
          file: file,
          type: 'file'
        });
      });
    } else if (uris.length > 0) {
      uris.forEach((uri) => {
        this._callHandlers(tx, {
          uri: uri,
          type: 'uri'
        });
      });
    } else {
      console.info('TODO: implement html/text drop here');
    }
  }

  _callHandlers(tx, params) {
    let i, handler;
    for (i = 0; i < this.assetHandlers.length; i++) {
      handler = this.assetHandlers[i];

      let match = handler.match(params, this.context);
      if (match) {
        handler.drop(tx, params, this.context);
        break
      }
    }
  }
}

/*
  Built-in handler that calls a custom handler, specified
  on the component (e.g. see ImageComponent).
*/
class CustomHandler extends DragAndDropHandler {

  match(dragState) {
    return dragState.dropType === 'custom'
  }

  drop(tx, dragState) {
    // Delegate handling to component which set up the custom dropzone
    dragState.component.handleDrop(tx, dragState);
  }
}

/**
  Used for edit tools or property annotations (e.g. EditLinkTool)

  @class
*/
class EditAnnotationCommand extends Command {

  constructor(...args) {
    super(...args);

    if (!this.config.nodeType) {
      throw new Error("'nodeType' is required")
    }
  }

  /**
    Get command state

    @return {Object} object with `disabled` and `node` properties
  */
  getCommandState(params) {
    let sel = this._getSelection(params);
    let annos = this._getAnnotationsForSelection(params);
    let newState = {
      disabled: true,
    };
    if (annos.length === 1 && sel.isPropertySelection() && sel.isCollapsed()) {
      newState.disabled = false;
      newState.node = annos[0];
    }
    return newState
  }

  execute(params) { } // eslint-disable-line

  _getAnnotationsForSelection(params) {
    return params.selectionState.getAnnotationsForType(this.config.nodeType)
  }
}

/*
  Storage for files such as images and other assets.

  This accumulates files created or loaded during a session
  so that it is possible to pick up a file after undo+redo.
*/
class FileManager {

  constructor(editorSession, extensions, context) {
    this.editorSession = editorSession;
    this.extensions = extensions;
    this.proxies = {};
    this.context = context;

    // adapt all existing files
    forEach(editorSession.getDocument().getNodes(), (node) => {
      if (node._isFileNode) this.storeFile(node);
    });

    this.editorSession.onUpdate('document', this._onDocumentChange, this);
  }

  dispose() {
    this.editorSession.off(this);
  }

  storeFile(fileNode) {
    let proxy = this.proxies[fileNode.id];
    // don't adapt the file if we already have it
    if (!proxy) {
      proxy = this.createFileProxy(fileNode);
      if (proxy) {
        this.proxies[fileNode.id] = proxy;
      }
    }
    fileNode.proxy = proxy;
    return proxy
  }

  createFileProxy(fileNode) { // eslint-disable-line
    let context = this.context;
    for (var i = 0; i < this.extensions.length; i++) {
      let ExtClass = this.extensions[i];
      if (ExtClass.match(fileNode, context)) {
        return new ExtClass(fileNode, context)
      }
    }
    console.error('No file adapter found for ', fileNode);
  }

  getProxy(fileNode) {
    return this.proxies[fileNode.id]
  }

  sync() {
    // Note: potentially this could be a bi-directional sync
    // ATM, we only consider upload
    let promises = map(this.proxies, (proxy) => {
      return proxy.sync()
    });
    return Promise.all(promises)
  }

  _onDocumentChange(change) {
    let doc = this.editorSession.getDocument();
    forEach(change.created, (nodeData) => {
      // we would need the real node to support inheritance
      let node = doc.get(nodeData.id);
      if (node._isFileNode) {
        this.storeFile(node);
      }
    });
  }
}

/*
  Experimental

  We are seeking for a solution providing access to global DOM events
  while considering the current app state ~ document session state.

  This implementation is just a prototype and might change with the next releases.
*/

/*
  TODO: to be 100% safe we would need to introduce a hidden contenteditable
  where we put the selection in case of non-surface situations
  so that we are still able to receive events such as 'copy' -- actually only Edge is not dispatching
  to window.document.
*/

const events = [ 'keydown', 'keyup', 'keypress', 'mousedown', 'mouseup' , 'copy' ];

class GlobalEventHandler {

  constructor(editorSession, surfaceManager) {
    this.editorSession = editorSession;
    this.surfaceManager = surfaceManager;
    this.listeners = [];
    this.initialize();
  }

  initialize() {
    if (platform.inBrowser) {
      let document = DefaultDOMElement.wrapNativeElement(window.document);
      events.forEach(function(name) {
        document.on(name, this._dispatch.bind(this, name), this);
      }.bind(this));
    }
  }

  dispose() {
    if (platform.inBrowser) {
      let document = DefaultDOMElement.wrapNativeElement(window.document);
      document.off(this);
    }
  }

  addEventListener(eventName, handler, options) {
    if (!options.id) {
      throw new Error("GlobalEventHandler can only be used with option 'id'")
    }
    let listener = new DOMEventListener(eventName, handler, options);
    this.listeners.push(listener);
  }

  removeEventListener(listener) {
    let idx = this.listeners.indexOf(listener);
    if (idx > -1) {
      this.listeners.splice(idx, 1);
    }
  }

  getEventListeners() {
    return this.listeners
  }

  _getActiveListener(eventName) {
    let editorSession = this.editorSession;
    let sel = editorSession.getSelection();
    if (sel) {
      let surfaceId = sel.surfaceId;
      for (let i = 0; i < this.listeners.length; i++) {
        let listener = this.listeners[i];
        if (listener.eventName === eventName && listener.options.id === surfaceId) {
          return listener
        }
      }
    }
  }

  _dispatch(eventName, e) {
    let listener = this._getActiveListener(eventName);
    if (listener) {
      listener.handler(e);
    }
  }
}

GlobalEventHandler.prototype.on = DOMElement.prototype.on;
GlobalEventHandler.prototype.off = DOMElement.prototype.off;

class ExecuteCommandHandler {
  constructor(editorSession, commandName) {
    this.editorSession = editorSession;
    this.commandName = commandName;
  }
  execute(params) {
    let commandState = params.editorSession.getCommandStates()[this.commandName];
    if (!commandState || commandState.disabled) return false
    this.editorSession.executeCommand(this.commandName, params);
    return true
  }
}

class KeyboardManager {

  constructor(editorSession, bindings, options) {
    this.editorSession = editorSession;

    this.context = options.context || {};

    this.keydownBindings = {};
    this.textinputBindings = {};

    bindings.forEach(({ key, spec }) => {
      // default combos are evaluated on keydown
      let type = spec.type || 'keydown';
      if(spec.command) {
        let handler = new ExecuteCommandHandler(editorSession, spec.command);
        let hook = handler.execute.bind(handler);
        if (type === 'keydown') {
          this.keydownBindings[parseCombo(key)] = hook;
        } else if (type === 'textinput') {
          this.textinputBindings[key] = hook;
        }
      } else {
        throw new Error('Keyboard binding not supported', spec)
      }
    });
  }

  onKeydown(event) {
    let key = parseKeyEvent(event);
    let hook = this.keydownBindings[key];
    if (hook) {
      event.preventDefault();
      event.stopPropagation();
      let params = this._getParams();
      hook(params, this.context);
      return true
    }
  }

  onTextInput(text) {
    let hook = this.textinputBindings[text];
    if (hook) {
      let params = this._getParams();
      return hook(params, this.context)
    }
  }

  _getParams() {
    let editorSession = this.editorSession;
    let selectionState = editorSession.getSelectionState();
    let sel = selectionState.getSelection();
    let surface = this.context.surfaceManager.getFocusedSurface();
    return {
      editorSession: editorSession,
      selectionState: selectionState,
      surface: surface,
      selection: sel,
    }
  }

}

function parseCombo(combo) {
  let frags = combo.split('+');
  let data = {
    keyCode: -1
  };
  for (var i = 0; i < frags.length; i++) {
    let frag = frags[i].toUpperCase();
    switch(frag) {
      case 'ALT': {
        data.altKey = true;
        break
      }
      case 'ALTGR': {
        data.altKey = true;
        data.code = 'AltRight';
        break
      }
      case 'CMD': {
        data.metaKey = true;
        break
      }
      case 'CTRL': {
        data.ctrlKey = true;
        break
      }
      case 'SHIFT': {
        data.shiftKey = true;
        break
      }
      default:
        if (frag.length === 1) {
          data.keyCode = frag.charCodeAt(0);
        } else if (keys$1.hasOwnProperty(frag)) {
          data.keyCode = keys$1[frag];
        } else {
          throw new Error('Unsupported keyboard command: '+ combo)
        }
    }
  }
  return parseKeyEvent(data)
}

KeyboardManager.parseCombo = parseCombo;

class MacroManager {

  constructor(context, macros) {
    this.context = context;
    this.macros = macros;
    this.context.editorSession.onFinalize('document', this.onDocumentChanged, this);
  }

  dispose() {
    this.context.editorSession.off(this);
  }

  onDocumentChanged(change, info) {
    this.executeMacros(change, info);
  }

  executeMacros(change, info) {
    let doc = this.context.editorSession.getDocument();
    let nodeId, node, text, start, end;
    let path;
    // HACK: we exploit the information of the internal structure
    // of this document changes
    switch(info.action) {
      case 'type': {
        let op = change.ops[0];
        if (op.type === 'update' && op.diff._isTextOperation) {
          path = op.path;
          nodeId = path[0];
          node = doc.get(nodeId);
          text = doc.get(path);
          start = op.diff.pos;
          end = start+op.diff.getLength();
        }
        break
      }
      case 'break': {
        // FIXME: this impl turned out to be fragile,
        // as the selection before the transaction
        // can show to nodes not being there anymore

        // We interpret a 'break' as kind of confirmation
        // of the current node
        // so we take the original selection
        // to determine the original node
        let sel = change.before.selection;
        if (!sel.isPropertySelection()) return
        path = sel.path;
        nodeId = path[0];
        node = doc.get(nodeId);
        // HACK: fragile - before.selection is not a good reference
        if (!node || !node.isText()) return
        text = node.getText();
        start = sel.start.offset;
        end = start;
        break
      }
      case 'paste': {
        // HACK: just support primitive plain-text paste
        if (change.ops.length === 1) {
          let op = change.ops[0];
          if (op.type === 'update' && op.propertyType === 'string') {
            path = op.path;
            nodeId = path[0];
            node = doc.get(nodeId);
            if (!node.isText()) return
            text = node.getText();
            start = op.diff.pos;
            end = start+op.diff.getLength();
          }
        }
        break
      }
      default:
        return
    }

    let props = {
      action: info.action,
      node: node,
      path: path,
      text: text,
      start: start,
      end: end,
      editorSession: this.context.editorSession,
      selection: this.context.editorSession.getSelection()
    };

    setTimeout(() => {
      for (let i = 0; i < this.macros.length; i++) {
        let macro = this.macros[i];
        let executed = macro.execute(props, this.context);
        if (executed) {
          break
        }
      }
    });

  }
}

class MarkersManager {

  constructor(editorSession) {
    this.editorSession = editorSession;

    // registry
    this._textProperties = {};
    this._dirtyProps = {};

    this._markers = new MarkersIndex(editorSession);

    editorSession.onUpdate(this._onChange, this);
    editorSession.onRender(this._updateProperties, this);
  }

  dispose() {
    this.editorSession.off(this);
    this._markers.dispose();
  }

  register(textProperyComponent) {
    let path = String(textProperyComponent.getPath());
    // console.log('registering property', path)
    let textProperties = this._textProperties[path];
    if (!textProperties) {
      textProperties = this._textProperties[path] = [];
    }
    textProperties.push(textProperyComponent);
  }

  deregister(textProperyComponent) {
    let path = String(textProperyComponent.getPath());
    // console.log('deregistering property', path)
    let textProperties = this._textProperties[path];
    if (!textProperties) {
      // FIXME: happens in test suite
      return
    }
    deleteFromArray(this._textProperties[path], textProperyComponent);
    if (textProperties.length === 0) {
      delete this._textProperties[path];
    }
  }

  getMarkers(path, opts) {
    opts = opts || {};
    let doc = this.editorSession.getDocument();
    let annos = doc.getAnnotations(path) || [];
    let markers = this._markers.get(path, opts.surfaceId, opts.containerId);
    return annos.concat(markers)
  }

  _onChange(editorSession) {
    if (editorSession.hasDocumentChanged()) {
      // mark all updated props per se as dirty
      if (editorSession.hasDocumentChanged()) {
        let change = editorSession.getChange();
        forEach(change.updated, (val, id) => {
          this._dirtyProps[id] = true;
        });
      }
      Object.assign(this._dirtyProps, this._markers._collectDirtyProps());
    }
  }

  _updateProperties() {
    Object.keys(this._dirtyProps).forEach((path) => {
      let textProperties = this._textProperties[path];
      if (textProperties) {
        textProperties.forEach(this._updateTextProperty.bind(this));
      }
    });
    this._dirtyProps = {};
  }

  _updateTextProperty(textPropertyComponent) {
    let path = textPropertyComponent.getPath();
    let markers = this.getMarkers(path, {
      surfaceId: textPropertyComponent.getSurfaceId(),
      containerId: textPropertyComponent.getContainerId()
    });
    // console.log('## providing %s markers for %s', markers.length, path)
    textPropertyComponent.setState({
      markers: markers
    });
  }

}

/*
  A DocumentIndex implementation for keeping track of markers
*/
class MarkersIndex {

  constructor(editorSession) {
    this.editorSession = editorSession;
    this.document = editorSession.getDocument();
    this._documentMarkers = new ArrayTree();
    this._surfaceMarkers = {};
    this._containerMarkers = {};

    this._dirtyProps = {};

    this.document.addIndex('markers', this);
    editorSession.onUpdate('document', this._onDocumentChange, this);
  }

  dispose() {
    // TODO: add an API to remove a custom index
    // and we should add a test which disposes the editor session
    delete this.document.data.indexes['markers'];
    this.editorSession.off(this);
  }

  reset() {
    this._documentMarkers = new ArrayTree();
    this._surfaceMarkers = {};
    this._containerMarkers = {};
    let doc = this.document;
    forEach(doc.getNodes(), (node) => {
      if (this.select(node)) {
        this.create(node);
      }
    });
  }

  select(node) {
    return node._isMarker
  }

  create(marker) {
    // console.log('Indexing marker', marker)
    switch (marker.constructor.scope) {
      case 'document': {
        this._dirtyProps[marker.path] = true;
        this._documentMarkers.add(marker.path, marker);
        break
      }
      case 'surface': {
        if (!this._surfaceMarkers[marker.surfaceId]) {
          this._surfaceMarkers[marker.surfaceId] = new ArrayTree();
        }
        this._dirtyProps[marker.path] = true;
        this._surfaceMarkers[marker.surfaceId].add(marker.path, marker);
        break
      }
      case 'container': {
        console.warn('Container scoped markers are not supported yet');
        break
      }
      default:
        console.error('Invalid marker scope.');
    }
  }

  delete(marker) {
    switch (marker.constructor.scope) {
      case 'document': {
        this._dirtyProps[marker.path] = true;
        this._documentMarkers.remove(marker.path, marker);
        break
      }
      case 'surface': {
        if (!this._surfaceMarkers[marker.surfaceId]) {
          this._surfaceMarkers[marker.surfaceId] = new ArrayTree();
        }
        this._dirtyProps[marker.path] = true;
        this._surfaceMarkers[marker.surfaceId].remove(marker.path, marker);
        break
      }
      case 'container': {
        console.warn('Container scoped markers are not supported yet');
        break
      }
      default:
        console.error('Invalid marker scope.');
    }
  }

  get(path, surfaceId) {
    let markers = this._documentMarkers[path] || [];
    if (surfaceId && this._surfaceMarkers[surfaceId]) {
      let surfaceMarkers = this._surfaceMarkers[surfaceId][path];
      if (surfaceMarkers) markers = markers.concat(surfaceMarkers);
    }
    // TODO support container scoped markers
    return markers
  }

  _collectDirtyProps() {
    let dirtyProps = this._dirtyProps;
    this._dirtyProps = {};
    return dirtyProps
  }

  // used for applying transformations
  _getAllCustomMarkers(path) {
    let markers = this._documentMarkers[path] || [];
    for(let surfaceId in this._surfaceMarkers) {
      if (!this._surfaceMarkers.hasOwnProperty(surfaceId)) continue
      let surfaceMarkers = this._surfaceMarkers[surfaceId][path];
      if (surfaceMarkers) markers = markers.concat(surfaceMarkers);
    }
    // TODO: support container markers
    return markers
  }

  _onDocumentChange(change) {
    let doc = this.doc;
    change.ops.forEach((op) => {
      let markers = this._getAllCustomMarkers(op.path);
      if (op.type === 'update' && op.diff._isTextOperation) {
        let diff$$1 = op.diff;
        switch (diff$$1.type) {
          case 'insert':
            this._transformInsert(doc, markers, diff$$1);
            break
          case 'delete':
            this._transformDelete(doc, markers, diff$$1);
            break
          default:
            //
        }
      }
    });
  }

  _transformInsert(doc, markers, op) {
    const pos = op.pos;
    const length = op.str.length;
    if (length === 0) return
    markers.forEach(function(marker) {
      // console.log('Transforming marker after insert')
      var start = marker.start.offset;
      var end = marker.end.offset;
      var newStart = start;
      var newEnd = end;
      if (pos >= end) return
      if (pos <= start) {
        newStart += length;
        newEnd += length;
        marker.start.offset = newStart;
        marker.end.offset = newEnd;
        return
      }
      if (pos < end) {
        newEnd += length;
        marker.end.offset = newEnd;
        if (marker.invalidate) marker.invalidate();
      }
    });
  }

  _transformDelete(doc, markers, op) {
    const pos1 = op.pos;
    const length = op.str.length;
    const pos2 = pos1 + length;
    if (pos1 === pos2) return
    markers.forEach((marker) => {
      var start = marker.start.offset;
      var end = marker.end.offset;
      var newStart = start;
      var newEnd = end;
      if (pos2 <= start) {
        newStart -= length;
        newEnd -= length;
        marker.start.offset = newStart;
        marker.end.offset = newEnd;
      } else if (pos1 >= end) {
        // nothing
      }
      // the marker needs to be changed
      // now, there might be cases where the marker gets invalid, such as a spell-correction
      else {
        if (pos1 <= start) {
          newStart = start - Math.min(pos2-pos1, start-pos1);
        }
        if (pos1 <= end) {
          newEnd = end - Math.min(pos2-pos1, end-pos1);
        }
        // TODO: we should do something special when the change occurred inside the marker
        if (start !== end && newStart === newEnd) {
          marker.remove();
          return
        }
        if (start !== newStart) {
          marker.start.offset = newStart;
        }
        if (end !== newEnd) {
          marker.end.offset = newEnd;
        }
        if (marker.invalidate) marker.invalidate();
      }
    });
  }

}

class SurfaceManager {

  constructor(editorSession) {
    this.editorSession = editorSession;
    this.surfaces = {};
    this._state = {
      focusedSurfaceId: null,
      selection: null,
    };
    editorSession.onUpdate('selection', this._onSelectionChanged, this);
    editorSession.onPostRender(this._recoverDOMSelection, this);
  }

  dispose() {
    this.editorSession.off(this);
  }

  /**
   * Get Surface instance
   *
   * @param {String} name Name under which the surface is registered
   * @return {ui/Surface} The surface instance
   */
  getSurface(name) {
    if (name) {
      return this.surfaces[name]
    }
  }

  /**
   * Get the currently focused Surface.
   *
   * @return {ui/Surface} Surface instance
   */
  getFocusedSurface() {
    if (this._state.focusedSurfaceId) {
      return this.getSurface(this._state.focusedSurfaceId)
    }
  }

  getSurfaces() {
    // HACK: not yet. we would need a polyfill
    // return Object.values(this.surfaces)
    return Object.keys(this.surfaces).map(key => this.surfaces[key])
  }

  /**
   * Register a surface
   *
   * @param surface {ui/Surface} A new surface instance to register
   */
  registerSurface(surface) {
    this.surfaces[surface.getId()] = surface;
  }

  /**
   * Unregister a surface
   *
   * @param surface {ui/Surface} A surface instance to unregister
   */
  unregisterSurface(surface) {
    surface.off(this);
    let surfaceId = surface.getId();
    // TODO: this is not working, has side-effects
    // with inline-nodes (see #985)
    // if (surface === this.getFocusedSurface()) {
    //   this.editorSession.setSelection(null)
    // }
    let registeredSurface = this.surfaces[surfaceId];
    if (registeredSurface === surface) {
      let focusedSurface = this.getFocusedSurface();
      delete this.surfaces[surfaceId];
      if (surface && focusedSurface === surface) {
        this._state.focusedSurfaceId = null;
      }
    }
  }

  _onSelectionChanged(selection) {
    const state = this._state;
    state.selection = selection;
    state.focusedSurfaceId = selection.surfaceId;
    // HACK: removing DOM selection *and* blurring when having a CustomSelection
    // otherwise we will receive events on the wrong surface
    // instead of bubbling up to GlobalEventManager
    if (selection && selection.isCustomSelection() && platform.inBrowser) {
      window.getSelection().removeAllRanges();
      window.document.activeElement.blur();
    }
  }

  _recoverDOMSelection() {
    // at the end of the update flow, make sure the surface is focused
    // and displays the right DOM selection.
    let info = this.editorSession.getChangeInfo() || {};
    if (info.skipSelectionRerender) return

    let focusedSurface = this.getFocusedSurface();
    if (focusedSurface && !focusedSurface.isDisabled()) {
      // console.log('Rendering selection on surface', focusedSurface.getId(), this.editorSession.getSelection().toString());
      focusedSurface.focus();
      focusedSurface.rerenderDOMSelection();
    }
  }
}

class EditorSession extends EventEmitter {

  constructor(doc, options) {
    super();

    options = options || {};

    this.document = doc;
    if (!options.configurator) {
      throw new Error('No configurator provided.')
    }
    this.configurator = options.configurator;

    this._transaction = new Transaction(doc, this);

    this._history = new ChangeHistory();
    // used for change accumulation (in a collab environment)
    this._currentChange = null;

    // TODO: while it is good to have these selection
    // related derived state informations separated
    // it would feel better to have the selection itself
    // as a property of this session
    this._selectionState = new SelectionState(doc);

    this._commandStates = [];

    // the session exposes these resources, and keeps track of changes
    this._resources = ['document', 'selection', 'commandStates'];
    // flags to keep track which resources have changed since the last 'flow'
    this._dirtyFlags = {};
    // set during a change
    this._change = null;
    this._info = null;

    this._flowStages = ['update', 'pre-render', 'render', 'post-render', 'position', 'finalize'];
    // to get something executed directly after a flow
    this._postponed = [];

    this._observers = {};

    this._lang = options.lang || this.configurator.getDefaultLanguage();
    this._dir = options.dir || 'ltr';

    // Managers
    // --------

    // surface manager takes care of surfaces, keeps track of the currently focused surface
    // and makes sure the DOM selection is rendered properly at the end of a flow
    this.surfaceManager = new SurfaceManager(this);
    // this context is provided to commands, tools, etc.
    this._context = {
      editorSession: this,
      //legacy
      surfaceManager: this.surfaceManager,
    };
    // to expose custom context just provide optios.context
    if (options.context) {
      Object.assign(this._context, options.context);
    }

    let configurator = this.configurator;
    let commands = configurator.getCommands();
    let dropHandlers = configurator.getDropHandlers();
    let macros = configurator.getMacros();
    let converterRegistry = configurator.getConverterRegistry();
    let editingBehavior = configurator.getEditingBehavior();

    this.fileManager = options.fileManager || new FileManager(this, configurator.getFileAdapters(), this._context);

    // Handling of saving
    this._hasUnsavedChanges = false;
    this._isSaving = false;

    if (options.saveHandler) {
      this.saveHandler = options.saveHandler;
    } else {
      this.saveHandler = configurator.getSaveHandler();
    }

    // The command manager keeps the commandStates up-to-date
    this.commandManager = new CommandManager(this._context, commands);
    // The drag manager dispatches drag requests to registered drag handlers
    // TODO: after consolidating the API of this class, we probably need a less diverse context
    this.dragManager = new DragManager(dropHandlers, Object.assign({}, this._context, {
      commandManager: this.commandManager
    }));
    // The macro manager dispatches to macro detectors at the end of the flow
    this.macroManager = new MacroManager(this._context, macros);
    this.globalEventHandler = new GlobalEventHandler(this, this.surfaceManager);
    this.markersManager = new MarkersManager(this);
    this.keyboardManager = new KeyboardManager(this, configurator.getKeyboardShortcuts(), {
      context: this._context
    });

    // TODO: see how we want to expose these
    this.converterRegistry = converterRegistry;
    this.editingBehavior = editingBehavior;
  }

  dispose() {
    this._transaction.dispose();
    this.surfaceManager.dispose();
    this.fileManager.dispose();
    this.commandManager.dispose();
    this.dragManager.dispose();
    this.macroManager.dispose();
    this.globalEventHandler.dispose();
    this.markersManager.dispose();
  }

  hasChanged(resource) {
    return this._dirtyFlags[resource]
  }

  hasDocumentChanged() {
    return this.hasChanged('document')
  }

  hasSelectionChanged() {
    return this.hasChanged('selection')
  }

  hasCommandStatesChanged() {
    return this.hasChanged('commandStates')
  }

  hasLanguageChanged() {
    return this.hasChanged('lang')
  }

  hasTextDirectionChanged() {
    return this.hasChanged('dir')
  }

  get(resourceName) {
    switch(resourceName) {
      case 'document':
        return this.getDocument()
      case 'selection':
        return this.getSelection()
      case 'commandStates':
        return this.getCommandStates()
      case 'change':
        return this.getChange()
      default:
        throw new Error('Unknown resource: ' + resourceName)
    }
  }

  getConfigurator() {
    return this.configurator
  }

  getContext() {
    return this._context
  }

  getDocument() {
    return this.document
  }

  getSelection() {
    return this.getSelectionState().getSelection()
  }

  getSelectionState() {
    return this._selectionState
  }

  getCommandStates() {
    return this._commandStates
  }

  getChange() {
    return this._change
  }

  getChangeInfo() {
    return this._info
  }

  getFocusedSurface() {
    return this.surfaceManager.getFocusedSurface()
  }

  getSurface(surfaceId) {
    return this.surfaceManager.getSurface(surfaceId)
  }

  getLanguage() {
    return this._lang
  }

  getTextDirection() {
    return this._dir
  }

  canUndo() {
    return this._history.canUndo()
  }

  canRedo() {
    return this._history.canRedo()
  }

  executeCommand(...args) {
    this.commandManager.executeCommand(...args);
  }

  setSelection(sel) {
    // console.log('EditorSession.setSelection()', sel)
    if (sel && isPlainObject$1(sel)) {
      sel = this.getDocument().createSelection(sel);
    }
    if (sel && !sel.isNull() && !sel.surfaceId) {
      let fs = this.getFocusedSurface();
      if (fs) {
        sel.surfaceId = fs.id;
      }
    }
    if (this._setSelection(sel)) {
      this.startFlow();
    }
    return sel
  }

  selectNode(nodeId) {
    let surface = this.getFocusedSurface();
    this.setSelection({
      type: 'node',
      nodeId: nodeId,
      containerId: surface.getContainerId(),
      surfaceId: surface.id
    });
  }

  setCommandStates(commandStates) {
    this._commandStates = commandStates;
    this._setDirty('commandStates');
  }

  setLanguage(lang) {
    if (this._lang !== lang) {
      this._lang = lang;
      this._setDirty('lang');
      this.startFlow();
    }
  }

  setTextDirection(dir) {
    if (this._dir !== dir) {
      this._dir = dir;
      this._setDirty('dir');
      this.startFlow();
    }
  }

  createSelection() {
    const doc = this.getDocument();
    return doc.createSelection.apply(doc, arguments)
  }

  getCollaborators() {
    return null
  }

  /*
    Set saveHandler via API

    E.g. if saveHandler not available at construction
  */
  setSaveHandler(saveHandler) {
    this.saveHandler = saveHandler;
  }

  /**
    Start a transaction to manipulate the document

    @param {function} transformation a function(tx) that performs actions on the transaction document tx

    @example

    ```js
    doc.transaction(function(tx, args) {
      tx.update(...)
      ...
      tx.setSelection(newSelection)
    })
    ```
  */
  transaction(transformation, info) {
    info = info || {};
    let change = this._transaction._recordChange(transformation, this.getSelection(), this.getFocusedSurface());
    if (change) {
      this._commit(change, info);
    } else {
      // if no changes, at least update the selection
      this._setSelection(this._transaction.getSelection());
      this.startFlow();
    }
    return change
  }

  undo() {
    this._undoRedo('undo');
  }

  redo() {
    this._undoRedo('redo');
  }

/* eslint-disable no-invalid-this*/

  on(...args) {
    let name = args[0];
    if (this._flowStages.indexOf(name) >= 0) {
      // remove the stage name from the args
      args.shift();
      let options = args[2] || {};
      let resource = options.resource;
      if (resource) {
        delete options.resource;
        args.unshift(resource);
      }
      this._registerObserver(name, args);
    } else {
      EventEmitter.prototype.on.apply(this, args);
    }
  }

  off(...args) {
    if (args.length === 1) {
      super.off(...args);
    } else {
      const stage = args[0];
      const method = args[1];
      const observer = args[2];
      this._deregisterObserver(stage, method, observer);
    }
  }

  /**
    Registers a hook for the `update` phase.

    During `update` data should be derived necessary for rendering.

    This is mainly used by extensions of the EditorSession to
    derive extra state information.

    @param {string} [resource] the name of the resource
    @param {Function} handler the function handler
    @param {Object} context owner of the handler
    @param {Object} [options] options for the resource handler

  */
  onUpdate(...args) {
    return this._registerObserver('update', args)
  }

  onPreRender(...args) {
    return this._registerObserver('pre-render', args)
  }

  /**
    Registers a hook for the 'render' phase.

    During `render`, components should be rerendered.

    @param {string} [resource] the name of the resource
    @param {Function} handler the function handler
    @param {Object} context owner of the handler
    @param {Object} [options] options for the resource handler

    @example

    This typically used by components that render node content.

    ```js
    class ImageComponent extends Component {
      didMount() {
        this.context.editorSession.onRender('document', this.rerender, this, {
          path: [this.props.node.id, 'src']
        })
      }
      dispose() {
        this.context.editorSession.off(this)
      }
      render($$) {
        ...
      }
    }
    ```
  */
  onRender(...args) {
    return this._registerObserver('render', args)
  }

  /**
    Registers a hook for the 'post-render' phase.

    ATM, this phase is used internally only, for recovering the DOM selection
    which typically gets destroyed due to rerendering

    @internal

    @param {string} [resource] the name of the resource
    @param {Function} handler the function handler
    @param {Object} context owner of the handler
    @param {Object} [options] options for the resource handler
  */
  onPostRender(...args) {
    return this._registerObserver('post-render', args)
  }

  /**
    Registers a hook for the 'position' phase.

    During `position`, components such as Overlays, for instance, should be positioned.
    At this stage, it is guaranteed that all content is rendered, and the DOM selection
    is set.

    @param {string} [resource] the name of the resource
    @param {Function} handler the function handler
    @param {Object} context owner of the handler
    @param {Object} [options] options for the resource handler

  */
  onPosition(...args) {
    return this._registerObserver('position', args)
  }

  onFinalize(...args) {
    return this._registerObserver('finalize', args)
  }

  _setSelection(sel) {
    let hasChanged = this.getSelectionState().setSelection(sel);
    if (hasChanged) this._setDirty('selection');
    return hasChanged
  }

  _undoRedo(which) {
    const doc = this.getDocument();
    var from, to;
    if (which === 'redo') {
      from = this._history.undoneChanges;
      to = this._history.doneChanges;
    } else {
      from = this._history.doneChanges;
      to = this._history.undoneChanges;
    }
    var change = from.pop();
    if (change) {
      this._applyChange(change, {});
      // move change to the opposite change list (undo <-> redo)
      to.push(change.invert());
      // use selection from change
      let sel = change.after.selection;
      if (sel) sel.attach(doc);
      this._setSelection(sel);
      // finally trigger the flow
      this.startFlow();
    } else {
      console.warn('No change can be %s.', (which === 'undo'? 'undone':'redone'));
    }
  }

  _transformLocalChangeHistory(externalChange) {
    // Transform the change history
    // Note: using a clone as the transform is done inplace
    // which is ok for the changes in the undo history, but not
    // for the external change
    var clone$$1 = {
      ops: externalChange.ops.map(function(op) { return op.clone(); })
    };
    transformDocumentChange(clone$$1, this._history.doneChanges);
    transformDocumentChange(clone$$1, this._history.undoneChanges);
  }

  _transformSelection(change) {
    var oldSelection = this.getSelection();
    var newSelection = transformSelection(oldSelection, change);
    // console.log('Transformed selection', change, oldSelection.toString(), newSelection.toString())
    return newSelection
  }

  _commit(change, info) {
    this._commitChange(change, info);
    // TODO: Not sure this is the best place to mark the session dirty
    this._hasUnsavedChanges = true;
    this.startFlow();
  }

  _commitChange(change, info) {
    change.timestamp = Date.now();
    this._transaction._document._skipNextDocumentChange = true;
    this._applyChange(change, info);
    if (info['history'] !== false && !info['hidden']) {
      this._history.push(change.invert());
    }
    var newSelection = change.after.selection || Selection.nullSelection;
    // HACK injecting the surfaceId here...
    // TODO: we should find out where the best place is to do this
    if (!newSelection.isNull() && !newSelection.surfaceId) {
      newSelection.surfaceId = change.after.surfaceId;
    }
    this._setSelection(newSelection);
  }

  _applyChange(change, info) {
    if (!change) {
      console.error('FIXME: change is null.');
      return
    }
    const doc = this.getDocument();
    doc._apply(change);
    doc._notifyChangeListeners(change, info);
    this._setDirty('document');
    this._change = change;
    this._info = info;
  }

  /*
    Are there unsaved changes?
  */
  hasUnsavedChanges() {
    return this._hasUnsavedChanges
  }

  /*
    Save session / document
  */
  save() {
    var saveHandler = this.saveHandler;

    if (this._hasUnsavedChanges && !this._isSaving) {
      this._isSaving = true;
      // Pass saving logic to the user defined callback if available
      if (saveHandler) {
        let saveParams = {
          editorSession: this,
          fileManager: this.fileManager
        };
        return saveHandler.saveDocument(saveParams)
        .then(() => {
          this._hasUnsavedChanges = false;
          // We update the selection, just so a selection update flow is
          // triggered (which will update the save tool)
          // TODO: model this kind of update more explicitly. It could be an 'update' to the
          // document resource (hasChanges was modified)
          this.setSelection(this.getSelection());
        })
        .catch((err) => {
          console.error('Error during save', err);
        }).then(() => { // finally
          this._isSaving = false;
        })
      } else {
        console.error('Document saving is not handled at the moment. Make sure saveHandler instance provided to editorSession');
        return Promise.reject()
      }
    }
  }

  /*
    Starts the flow.

    This is necessary when changing resources managed by the session.
    To be able to change multiple resources at the same time,
    this is not done automatically, but needs to be called
    by the implementation.

    @internal
  */
  startFlow() {
    if (this._flowing) {
      throw new Error('Already in a flow. You need to postpone the update.')
    }
    this._flowing = true;
    try {
      this.performFlow();
    } finally {
      this._resetFlow();
      this._flowing = false;
    }
    // Note: postponing is ATM used only by Macros
    // HACK: to avoid having multiple flows at the same time
    // we are running this deferred
    const postponed = this._postponed;
    const self = this;
    this._postponed = [];
    setTimeout(function() {
      postponed.forEach(function(fn) {
        fn(self);
      });
    }, 0);
  }

  /*
    Emits the phases in the correct order.

    @internal
  */
  performFlow() {
    this._flowStages.forEach((stage) => {
      this._notifyObservers(stage);
    });
  }

  postpone(fn) {
    this._postponed.push(fn);
  }

  _parseObserverArgs(args) {
    let params = { resource: null, handler: null, context: null, options: {} };
    // first can be a string
    let idx = 0;
    let arg = args[idx];
    if (isString$1(arg)) {
      params.resource = arg;
      idx++;
      arg = args[idx];
    }
    if (!arg) {
      throw new Error('Provided handler function was nil.')
    }
    if (!isFunction$1(arg)) {
      throw new Error('Expecting a handler Function.')
    }
    params.handler = arg;
    idx++;
    arg = args[idx];
    if (arg) {
      params.context = arg;
      idx++;
      arg = args[idx];
    }
    if (arg) {
      params.options = arg;
    }
    return params
  }

  _registerObserver(stage, args) {
    let observer = this._parseObserverArgs(args);
    let observers = this._observers[stage];
    if (!observers) {
      observers = this._observers[stage] = [];
    }
    observers.push(observer);
  }

  _deregisterObserver(stage, method, observer) {
    let self = this; // eslint-disable-line no-invalid-this
    if (arguments.length === 1) {
      // TODO: we should optimize this, as ATM this needs to traverse
      // a lot of registered listeners
      forEach(self._observers, (observers) => {
        for (let i = observers.length-1; i >=0 ; i--) {
          const o = observers[i];
          if (o.context === observer) {
            observers.splice(i, 1);
            o._deregistered = true;
          }
        }
      });
    } else {
      let observers = self._observers[stage];
      // if no observers are registered, then this might not
      // be a deregistration for a stage, but a regular event
      if (!observers) {
        EventEmitter.prototype.off.apply(self, arguments);
      } else {
        for (let i = observers.length-1; i >= 0; i--) {
          let o = observers[i];
          if (o.handler === method && o.context === observer) {
            observers.splice(i, 1);
            o._deregistered = true;
          }
        }
      }
    }
  }

  _notifyObservers(stage) {
    // TODO: this is not hierarchical anymore
    // i.e. probably we have to expect degradation of performance
    // with huuuge documents, as the number of listeners is
    // We could optimize this by 'compiling' a list of observers for
    // each configuration, maybe lazily
    // for now we accept this circumstance
    let _observers = this._observers[stage];
    if (!_observers) return
    // Make a copy so that iteration does not get confused, when listeners deregister
    // TODO: we could improve this by using a custom data structure that allows
    // manipulation during iteration
    let observers = _observers.slice();
    for (let i = 0; i < observers.length; i++) {
      let o = observers[i];
      // an observer might have been deregistered while this iteration was going on
      if (o._deregistered) continue
      if (!o.resource) {
        o.handler.call(o.context, this);
      } else if (o.resource === 'document') {
        if (!this.hasDocumentChanged()) continue
        const change = this.getChange();
        const info = this.getChangeInfo();
        const path = o.options.path;
        if (!path) {
          o.handler.call(o.context, change, info, this);
        } else if (change.isAffected(path)) {
          o.handler.call(o.context, change, info, this);
        }
      } else {
        if (!this.hasChanged(o.resource)) continue
        const resource = this.get(o.resource);
        o.handler.call(o.context, resource, this);
      }
    }
  }

  _setDirty(resource) {
    this._dirtyFlags[resource] = true;
  }

  _resetFlow() {
    Object.keys(this._dirtyFlags).forEach((resource) => {
      this._dirtyFlags[resource] = false;
    });
    this._change = null;
    this._info = null;
  }
}

class InsertNodeCommand extends Command {

  getCommandState(params) {
    let sel = params.selection;
    let newState = {
      disabled: true,
      active: false
    };
    if (sel && !sel.isNull() && !sel.isCustomSelection() && sel.containerId) {
      newState.disabled = false;
    }
    return newState
  }

  execute(params, context) {
    var state = params.commandState;
    if (state.disabled) return
    let editorSession = this._getEditorSession(params, context);
    editorSession.transaction((tx) => {
      let nodeData = this.createNodeData(tx, params, context);
      let node = tx.insertBlockNode(nodeData);
      this.setSelection(tx, node);
    });
  }

  createNodeData(tx, params, context) {
    const type = params.type;
    if (!type) throw new Error("'type' is mandatory")
    const doc = context.editorSession.getDocument();
    const nodeSchema = doc.getSchema().getNodeSchema(type);
    let nodeData = {type};
    forEach(nodeSchema, (key) => {
      if (params.hasOwnProperty(key)) {
        nodeData[key] = params[key];
      }
    });
    return nodeData
  }

  setSelection(tx, node) {
    if (node.isText()) {
      tx.selection = {
        type: 'property',
        path: node.getPath(),
        startOffset: 0
      };
    }
  }
}

class Toolbox extends Component {

  getInitialState() {
    return {
      activeToolGroups: this.getActiveToolGroups()
    }
  }

  // NOTE: We disable this, as its premature optimization and can cause
  // side effects. Rendering the toolbox always is not a big performance
  // issue atm.
  // shouldRerender(newProps, newState) {
  //   // poor-man's immutable style
  //   let hasChanged = !isEqual(this.props, newProps) || !isEqual(this.state.activeToolGroups, newState.activeToolGroups)
  //
  //   if (!hasChanged) {
  //     this.hide()
  //     return false
  //   }
  //   return true
  // }

  didMount() {
    // rerender the context menu after anything else has been updated
    this.context.editorSession.onRender(this._onCommandStatesChanged, this);
  }

  dispose() {
    this.context.editorSession.off(this);
  }

  _onCommandStatesChanged(editorSession) {
    if (editorSession.hasChanged('commandStates')) {
      this.setState({
        activeToolGroups: this.getActiveToolGroups()
      });
    }
  }

  /*
    Returns A map of tooltargets, the value being an array of tool objects
    containing Class and toolProps for rendering
  */
  getActiveToolGroups() {
    this._hasActiveTools = false;

    let toolGroups = this.context.toolGroups;
    let activeToolGroupNames = this.getActiveToolGroupNames();
    let activeToolGroups = new Map();

    activeToolGroupNames.forEach((toolGroupName) => {
      let toolGroup = toolGroups.get(toolGroupName);
      if (!toolGroup) return
      let tools = toolGroup.tools;
      let activeTools = this.getActiveTools(tools, toolGroupName);
      activeToolGroups.set(toolGroupName, {
        name: toolGroupName,
        Class: toolGroup.Class,
        tools: activeTools
      });
    });
    return activeToolGroups
  }

  /*
    For a given Map of tools and a toolGroupName determine
    a set of activeTools, based on the current commandStates
  */
  getActiveTools(tools, toolGroupName) { // eslint-disable-line
    let activeTools = new Map();
    let commandStates = this._getCommandStates();

    tools.forEach((tool, toolName) => {
      let toolProps = Object.assign(
        { disabled: true },
        commandStates[toolName],
        {
          name: toolName,
          label: toolName,
          // style hint only interprerted by generic Tool class
          style: this.getToolStyle(toolName)
        }
      );

      if (!toolProps.disabled || this.showDisabled()) {
        activeTools.set(tool.name, {
          name: tool.name,
          Class: tool.Class,
          toolProps: toolProps
        });
        this._hasActiveTools = true;
      }
    });

    return activeTools
  }

  hasActiveTools() {
    return Boolean(this._hasActiveTools)
  }

  getActiveToolGroupNames() {
    throw new Error('Abstract method')
  }

  showDisabled() {
    return false
  }

  hide() {
    // Optional hook for hiding the toolbox component
  }

  /*
    Override if you just want to use a different style
  */
  getToolStyle() {
    throw new Error('Abstract method')
  }

  _getCommandStates() {
    return this.context.commandManager.getCommandStates()
  }

}

/*
  A default implementation to render the content for the overlay (aka popup) tools.
*/
class Overlay extends Toolbox {

  didMount() {
    super.didMount();
    if (!this.context.scrollPane) {
      throw new Error('Requires scrollPane context')
    }
    this.context.scrollPane.on('selection:positioned', this._onSelectionPositioned, this);
  }

  dispose() {
    super.dispose();
    this.context.scrollPane.off(this);
  }

  render($$) {
    let el = $$('div').addClass(this.getClassNames());
    el.addClass('sm-hidden');
    el.addClass('sm-theme-'+this.getTheme());
    let activeToolGroups = this.state.activeToolGroups;
    let activeToolsEl = $$('div').addClass('se-active-tools');

    activeToolGroups.forEach((toolGroup) => {
      let toolGroupProps = Object.assign({}, toolGroup, {
        toolStyle: this.getToolStyle(),
        showIcons: true
      });
      activeToolsEl.append(
        $$(toolGroup.Class, toolGroupProps).ref(toolGroup.name)
      );
    });

    el.append(activeToolsEl);
    return el
  }

  getToolStyle() {
    return this.props.theme || 'plain-dark'
  }

  show(hints) {
    this.el.removeClass('sm-hidden');
    this._position(hints);
  }

  hide() {
    this.el.addClass('sm-hidden');
  }

  _onSelectionPositioned(hints) {
    if (this.hasActiveTools()) {
      this.el.removeClass('sm-hidden');
      let overlayWidth = this.el.htmlProp('offsetWidth');
      let selRect = hints.selectionRect;
      let selectionMaxWidth = selRect.width;
      // By default, Overlays are aligned center/bottom to the selection
      this.el.css('top', selRect.top + selRect.height);
      let leftPos = selRect.left + selectionMaxWidth/2 - overlayWidth/2;
      // Must not exceed left bound
      leftPos = Math.max(leftPos, 0);
      // Must not exceed right bound
      let maxLeftPos = selRect.left + selectionMaxWidth + selRect.right - overlayWidth;
      leftPos = Math.min(leftPos, maxLeftPos);
      this.el.css('left', leftPos);
    } else {
      this.el.addClass('sm-hidden');
    }
  }

  getClassNames() {
    return 'sc-overlay'
  }

  getTheme() {
    return 'dark'
  }

  getActiveToolGroupNames() {
    return this.props.toolGroups || ['overlay']
  }

}

class TextBlockComponent extends NodeComponent {

  render($$) {
    let el = super.render($$);
    el.addClass('sc-text-block');

    let node = this.props.node;
    // NOTE: we are not using the native text direction support as it changes the contenteditable behavior in a weird way
    // instead rtl text is supported on model level
    if (node.direction) {
      // el.attr('data-direction', node.direction)
      el.attr('dir', node.direction);
    }
    el.append($$(TextPropertyComponent, {
      path: node.getTextPath(),
      direction: node.direction
    }));
    return el
  }

}

/**
  Editor for a text property (annotated string). Needs to be
  instantiated inside a {@link ui/Controller} context.

  @class
  @component
  @extends ui/Surface

  @prop {String} name unique editor name
  @prop {String[]} path path to a text property
  @prop {ui/SurfaceCommand[]} commands array of command classes to be available

  @example

  Create a `TextPropertyEditor` for the `name` property of an author object. Allow emphasis annotations.

  ```js
  $$(TextPropertyEditor, {
    name: 'authorNameEditor',
    path: ['author_1', 'name'],
    commands: [EmphasisCommand]
  })
  ```
*/

class TextPropertyEditor extends Surface {

  constructor(parent, props) {
    // making props.name optional
    props.name = props.name || props.path.join('.');
    super(parent, props);

    if (!props.path) {
      throw new Error("Property 'path' is mandatory.")
    }
  }

  render($$) {
    let el = super.render.apply(this, arguments);
    el.addClass("sc-text-property-editor");

    if (!this.props.disabled) {
      el.addClass('sm-enabled');
      el.attr('contenteditable', true);
      // native spellcheck
      el.attr('spellcheck', this.props.spellcheck === 'native');
    }

    el.append(
      $$(TextPropertyComponent, {
        tagName: this.props.tagName || "div",
        path: this.props.path,
        markers: this.props.markers,
        withoutBreak: this.props.withoutBreak
      })
    );

    return el
  }

  _handleEnterKey(event) {
    event.preventDefault();
    event.stopPropagation();
    if (this.props.multiLine) {
      super._handleEnterKey(event);
    }
    this.el.emit('enter', {
      altKey: event.altKey,
      ctrlKey: event.ctrlKey,
      metaKey: event.metaKey,
      shiftKey: event.shiftKey,
      code: event.code
    });
  }

  getPath() {
    return this.props.path
  }
}

TextPropertyEditor.prototype._isTextPropertyEditor = true;

/*
  @example

  ```js
  $$(Button, {
    name: 'add-ref' // used to resolve icon and label
    label: 'Add reference' // optional if you want to set the label string explicity
  })
  ```
*/
class Button extends Component {
  render($$) {
    let el = $$('button')
      .addClass('sc-button');

    if (this.props.icon) {
      el.append(this.renderIcon($$));
    }
    if (this.props.label) {
      el.append(this.renderLabel($$));
    }
    if (this.props.active) {
      el.addClass('sm-active');
    }
    if (this.props.style) {
      el.addClass('sm-style-'+this.props.style);
    }
    if (this.props.disabled) {
      // make button inaccessible
      el.attr('tabindex', -1)
        .attr('disabled', true);
    } else {
      // make button accessible for tab-navigation
      el.attr('tabindex', 1);
    }

    // Ability to inject additional elements (should be avoided!)
    el.append(this.props.children);
    return el
  }

  renderIcon($$) {
    let iconEl = this.context.iconProvider.renderIcon($$, this.props.icon);
    return iconEl
  }

  renderLabel($$) {
    return $$('div').addClass('se-label').append(
      this.getLabel(this.props.label)
    )
  }

  renderHint($$) {
    return $$('div').addClass('se-hint').append(
      this.getLabel(this.props.hint+'-hint')
    )
  }

  getLabel(name) {
    let labelProvider = this.context.labelProvider;
    return labelProvider.getLabel(name)
  }
}

var ButtonPackage = {
  name: 'button',
  configure: function(config) {
    config.addComponent('button', Button);
  }
};

class ContextMenu extends Toolbox {

  didMount() {
    super.didMount();
    if (!this.context.scrollPane) {
      throw new Error('Requires a scrollPane context')
    }
    this.context.scrollPane.on('context-menu:opened', this._onContextMenuOpened, this);
  }

  dispose() {
    super.dispose();
    this.context.scrollPane.off(this);
  }

  /*
    Override with custom rendering
  */
  render($$) {
    let el = $$('div').addClass('sc-context-menu sm-hidden');
    let activeToolGroups = this.state.activeToolGroups;

    activeToolGroups.forEach((toolGroup) => {
      let toolGroupProps = Object.assign({}, toolGroup, {
        toolStyle: this.getToolStyle(),
        showLabels: true,
        // showHints: true
      });

      if (toolGroupProps.tools.size > 0) {
        let toolGroupEl = $$(toolGroup.Class, toolGroupProps);
        el.append(toolGroupEl);
      }
    });
    return el
  }

  getActiveToolGroupNames() {
    return ['context-menu-primary', 'context-menu-document']
  }

  showDisabled() {
    return true
  }

  /*
    Override if you just want to use a different style
  */
  getToolStyle() {
    return 'plain-dark'
  }

  hide() {
    this.el.addClass('sm-hidden');
  }

  /*
    Positions the content menu relative to the scrollPane
  */
  _onContextMenuOpened(hints) {
    let mouseBounds = hints.mouseBounds;
    this.el.removeClass('sm-hidden');
    let contextMenuWidth = this.el.htmlProp('offsetWidth');

    // By default, context menu are aligned left bottom to the mouse coordinate clicked
    this.el.css('top', mouseBounds.top);
    let leftPos = mouseBounds.left;
    // Must not exceed left bound
    leftPos = Math.max(leftPos, 0);
    // Must not exceed right bound
    let maxLeftPos = mouseBounds.left + mouseBounds.right - contextMenuWidth;
    leftPos = Math.min(leftPos, maxLeftPos);
    this.el.css('left', leftPos);
  }
}

var ContextMenuPackage = {
  name: 'context-menu',
  configure: function(config) {
    config.addComponent('context-menu', ContextMenu);
  }
};

/*
  Simple component for realizing grid layouts
*/
class Grid extends Component {
  render($$) {
    let el = $$('div').addClass('sc-grid');
    if (this.props.mobile) {
      el.addClass('sm-mobile');
    }
    el.append(this.props.children);
    return el
  }
}

/*
  A grid row
*/
class Row extends Component {
  render($$) {
    let el = $$('div').addClass('se-row');
    el.append(this.props.children);
    return el
  }
}

/*
  A grid cell
*/
class Cell extends Component {
  render($$) {
    let el = $$('div').addClass('se-cell');
    el.addClass('sm-column-'+this.props.columns);
    el.append(this.props.children);
    return el
  }
}

Grid.Row = Row;
Grid.Cell = Cell;

var GridPackage = {
  name: 'grid',
  configure: function(config) {
    config.addComponent('grid', Grid);
  }
};

/*
  A default implementation to render the content for the overlay (aka popup) tools.
*/
class Gutter extends Toolbox {

  render($$) {
    let el = $$('div').addClass(this.getClassNames());
    el.addClass('sm-hidden');
    el.addClass('sm-theme-'+this.getTheme());
    let activeToolGroups = this.state.activeToolGroups;
    let activeToolsEl = $$('div').addClass('se-active-tools');

    activeToolGroups.forEach((toolGroup) => {
      let toolGroupProps = Object.assign({}, toolGroup, {
        toolStyle: this.getToolStyle(),
        showIcons: true
      });
      activeToolsEl.append(
        $$(toolGroup.Class, toolGroupProps)
      );
    });

    el.append(activeToolsEl);
    return el
  }

  /*
    Override if you just want to use a different style
  */
  getToolStyle() {
    return 'outline-dark'
  }

  show(hints) {
    this.el.removeClass('sm-hidden');
    this._position(hints);
  }

  hide() {
    this.el.addClass('sm-hidden');
  }

  _position(hints) {
    if (hints) {
      // By default, gutter is centered (y-axis) and left of the scrollPane content (x-axis)
      this.el.css('top', hints.rectangle.top + hints.rectangle.height - hints.rectangle.height / 2);
      this.el.css('left', 0);
    }
  }

  getClassNames() {
    return 'sc-gutter'
  }

  getTheme() {
    return 'dark'
  }

  getActiveToolGroupNames() {
    return ['gutter']
  }

}

var GutterPackage = {
  name: 'gutter',
  configure: function(config) {
    config.addToolGroup('gutter');
    config.addComponent('gutter', Gutter);
  }
};

class Input extends Component {

  _onChange() {
    let editorSession = this.context.editorSession;
    let path = this.props.path;
    let newVal = this.el.val();

    editorSession.transaction(function(tx) {
      tx.set(path, newVal);
    });
  }

  render($$) {
    let val;

    if (this.props.path) {
      let editorSession = this.context.editorSession;
      let doc = editorSession.getDocument();
      val = doc.get(this.props.path);
    } else {
      val = this.props.value;
    }

    let el = $$('input').attr({
      value: val,
      type: this.props.type,
      placeholder: this.props.placeholder
    })
    .addClass('sc-input');

    if (this.props.path) {
      el.on('change', this._onChange);
    }

    if (this.props.centered) {
      el.addClass('sm-centered');
    }

    return el
  }
}

var InputPackage = {
  name: 'input',
  configure: function(config) {
    config.addComponent('input', Input);
  }
};

/**
  Layout component for simple layout tasks, without having to write CSS

  @class
  @component

  @prop {String} width 'small', 'medium', 'large' and 'full'
  @prop {String} [textAlign] 'center', 'left' or 'right'
  @prop {String} [noPadding] No padding around layout, will fill the whole space

  @example

  ```js
  var form = $$(Layout, {
    width: 'large',
    textAlign: 'center'
  });
  ```
*/
class Layout extends Component {

  render($$) {
    let el = $$('div').addClass('sc-layout');
    el.addClass('sm-width-'+this.props.width);
    if (this.props.textAlign) {
      el.addClass('sm-text-align-'+this.props.textAlign);
    }

    if (this.props.noPadding) {
      el.addClass('sm-no-padding');
    }

    el.append(this.props.children);
    return el
  }
}

var LayoutPackage = {
  name: 'layout',
  configure: function(config) {
    config.addComponent('layout', Layout);
  }
};

/**
  Modal dialog component

  @class
  @component

  @prop {String} width 'small', 'medium', 'large' and 'full'

  @example

  ```js
  var form = $$(Modal, {
    width: 'medium',
    textAlign: 'center'
  });
  ```
*/
class Modal extends Component {

  render($$) {
    let el = $$('div').addClass('sc-modal');

    // TODO: don't think that this is good enough. Right the modal is closed by any unhandled click.
    // Need to be discussed.
    el.on('click', this._closeModal);

    if (this.props.width) {
      el.addClass('sm-width-'+this.props.width);
    }

    el.append(
      $$('div').addClass('se-body').append(
        this.props.children
      )
    );
    return el
  }

  _closeModal(e) {
    let closeSurfaceClick = e.target.classList.contains('sc-modal');
    if (closeSurfaceClick) {
      this.send('closeModal');
    }
  }

}

var ModalPackage = {
  name: 'modal',
  configure: function(config) {
    config.addComponent('modal', Modal);
  }
};

var OverlayPackage = {
  name: 'overlay',
  configure: function(config) {
    config.addToolGroup('overlay');
    config.addComponent('overlay', Overlay);
  }
};

class Dropzones extends Component {

  didMount() {
    this.context.dragManager.on('drag:started', this.onDragStarted, this);
    this.context.dragManager.on('drag:finished', this.onDragFinished, this);
  }

  render($$) {
    let el = $$('div').addClass('sc-dropzones');

    if (this.state.dropzones) {
      el.on('dragenter', this.onDrag)
        .on('dragover', this.onDrag);

      // Dropzones are scoped by surfaceId
      forEach(this.state.dropzones, (dropzones, surfaceId) => {
        dropzones.forEach((dropzone, index) => {
          let dropType = dropzone.type;
          let dropzoneEl;
          if (dropType === 'place') {
            dropzoneEl = $$('div').addClass('se-dropzone')
              .attr({
                'data-dropzone-index': index,
                'data-dropzone-surface': surfaceId
              }).append(
                $$('div').addClass('se-drop-teaser').css({
                  top: dropzone.teaserPos
                })
              );
          } else if (dropType === 'custom') {
            dropzoneEl = $$('div').addClass('se-custom-dropzone').attr({
              'data-dropzone-index': index,
              'data-dropzone-surface': surfaceId
            }).append(
              // TODO: also provide se-custom-drop-teaser when custom
              // dropzone is provided
              $$('div').addClass('se-message').append(dropzone.message)
            );
          }
          if (dropzoneEl) {
            let shield = $$('div').addClass('se-drop-shield')
              .on('dragenter', this.onDragEnter)
              .on('dragleave', this.onDragLeave)
              .on('drop', this.onDrop)
              .on('mouseenter', this.onDragEnter)
              .on('mouseleave', this.onDragLeave)
              .on('mouseup', this.onDrop);
            dropzoneEl.append(shield);
            dropzoneEl.css({
              position: 'absolute',
              top: dropzone.top,
              left: dropzone.left,
              width: dropzone.width,
              height: dropzone.height
            });
            el.append(dropzoneEl);
          }
        });
      });
    } else {
      el.addClass('sm-hidden');
    }
    return el
  }

  // triggered by DragManager
  onDragStarted(dragState) {
    let dropzones = this._computeDropzones(dragState);
    setTimeout(() => {
      this.setState({
        dropzones: dropzones
      });
    }, 250);
  }

  // triggered by DragManager
  onDragFinished() {
    this.setState({});
  }

  onDragEnter(e) {
    // console.log('onDragEnter', e.target)
    e.target.parentNode.classList.add('sm-over');
  }

  onDragLeave(e) {
    // console.log('onDragLeave', e.target)
    e.target.parentNode.classList.remove('sm-over');
  }

  // just so that the teaser does not prevent dropping
  onDrag(e) { // eslint-disable-line
    // console.log('onDrag', e.target)
    e.preventDefault();
  }

  onDrop(e) {
    // console.log('Dropzones.onDrop()', e.target)
    // HACK: try if this is really necessary
    e.__reserved__ = true;
    e.preventDefault();
    e.stopPropagation();
    let dropzoneIndex = e.target.parentNode.dataset.dropzoneIndex;
    let dropzoneSurface = e.target.parentNode.dataset.dropzoneSurface;
    let dropzone = this.state.dropzones[dropzoneSurface][dropzoneIndex];
    let dropParams = dropzone.dropParams;
    let dropType = dropzone.type;
    // Determine target surface
    let targetSurface = this.context.surfaceManager.getSurface(dropzoneSurface);
    // Original component (e.g. img element)
    let component = dropzone.component;
    let dropzoneComponent = dropzone.dropzoneComponent;
    // HACK: extending the dragState here
    let dragManager = this.context.dragManager;
    dragManager.extendDragState({
      targetSurface,
      dropType,
      dropParams,
      component,
      dropzoneComponent
    });
    dragManager._onDragEnd(e);
  }

  /*
    Get bounding rect for a component (relative to scrollPane content element)
  */
  _getBoundingRect(comp) {
    let scrollPane = comp.context.scrollPane;
    let contentElement = scrollPane.getContentElement().getNativeElement();
    let rect = getRelativeBoundingRect(comp.getNativeElement(), contentElement);
    return rect
  }

  _computeDropzones(dragState) {
    let scrollPaneName = this.context.scrollPane.getName();
    let surfaces = dragState.scrollPanes[scrollPaneName].surfaces;
    let scopedDropzones = {};

    forEach(surfaces, (surface) => {
      let components = surface.childNodes;

      // e.g. 3 components = 4 drop zones (1 before, 1 after, 2 in-between)
      let numDropzones = components.length + 1;
      let dropzones = [];

      for (let i = 0; i < numDropzones; i++) {
        if (i === 0) {
          // First dropzone
          let firstComp = this._getBoundingRect(components[0]);
          dropzones.push({
            type: 'place',
            left: firstComp.left,
            top: firstComp.top,
            width: firstComp.width,
            height: firstComp.height / 2,
            teaserPos: 0,
            dropParams: {
              insertPos: i
            }
          });
        } else if (i === numDropzones - 1) {
          // Last dropzone
          let lastComp = this._getBoundingRect(components[i - 1]);
          dropzones.push({
            type: 'place',
            left: lastComp.left,
            top: lastComp.top + lastComp.height / 2,
            width: lastComp.width,
            height: lastComp.height / 2,
            teaserPos: lastComp.height / 2,
            dropParams: {
              insertPos: i
            }
          });
        } else {
          // Drop zone in between two components
          let upperComp = this._getBoundingRect(components[i-1]);
          let lowerComp = this._getBoundingRect(components[i]);
          let topBound = upperComp.top + upperComp.height / 2;
          let bottomBound = lowerComp.top + lowerComp.height / 2;

          dropzones.push({
            type: 'place',
            left: upperComp.left,
            top: topBound,
            width: upperComp.width,
            height: bottomBound - topBound,
            teaserPos: (upperComp.top + upperComp.height + lowerComp.top) / 2 - topBound,
            dropParams: {
              insertPos: i
            }
          });
        }

        if (i < numDropzones - 1) {
          let comp = components[i];
          // We get the isolated node wrapper and want to use the content element
          if (comp._isIsolatedNodeComponent) {
            comp = comp.getContent();
          }
          // If component has dropzones declared
          if (comp.getDropzoneSpecs) {
            let dropzoneSpecs = comp.getDropzoneSpecs();
            dropzoneSpecs.forEach((dropzoneSpec) => {
              let dropzoneComp = dropzoneSpec.component;
              let rect = this._getBoundingRect(dropzoneComp);
              dropzones.push({
                type: 'custom',
                component: comp,
                dropzoneComponent: dropzoneComp,
                left: rect.left,
                top: rect.top,
                width: rect.width,
                height: rect.height,
                message: dropzoneSpec.message,
                dropParams: dropzoneSpec.dropParams
              });
            });
          }
        }
      }
      scopedDropzones[surface.getName()] = dropzones;
    });
    return scopedDropzones
  }

  _renderDropTeaser(hints) {
    if (hints.visible) {
      this.el.removeClass('sm-hidden');
      this.el.css('top', hints.rect.top);
      this.el.css('left', hints.rect.left);
      this.el.css('right', hints.rect.right);
    } else {
      this.el.addClass('sm-hidden');
    }
  }

}

var DropzonesPackage = {
  name: 'dropzones',
  configure: function(config) {
    config.addComponent('dropzones', Dropzones);
  }
};

/**
  A rich scrollbar implementation that supports highlights.   Usually
  instantiated by {@link ScrollPane}, so you will likely not create it
  yourself.

  @class Scrollbar
  @component
  @private

  @prop {ui/ScrollPane} scrollPane scroll pane the scrollbar operates on
  @prop {object} highlights hightlights grouped by scope

  @example

  ```js
  $$(Scrollbar, {
    scrollPane: this,
    highlights: {
      'bib-items': ['bib-item-citation-1', 'bib-item-citation-2']
    }
  }).ref('scrollbar')
  ```
*/

class Scrollbar extends Component {

  didMount() {
    // do a full rerender when window gets resized
    DefaultDOMElement.getBrowserWindow().on('resize', this.onResize, this);
    // update the scroll handler on scroll
    this.props.scrollPane.on('scroll', this.onScroll, this);
    // TODO: why is this necessary here?
    setTimeout(function() {
      this.updatePositions();
    }.bind(this));
  }

  dispose() {
    DefaultDOMElement.getBrowserWindow().off(this);
    this.props.scrollPane.off(this);
  }

  didUpdate() {
    this.updatePositions();
  }

  render($$) {
    let el = $$('div')
      .addClass('sc-scrollbar')
      .on('mousedown', this.onMouseDown);

    if (this.props.highlights) {
      let highlightEls = [];

      forEach(this.props.highlights, function(highlights, scope) {
        forEach(highlights, function(h) {
          highlightEls.push(
            $$('div').ref(h).addClass('se-highlight sm-'+scope)
          );
        });
      });

      el.append(
        $$('div').ref('highlights')
          .addClass('se-highlights')
          .append(highlightEls)
      );
    }
    el.append($$('div').ref('thumb').addClass('se-thumb'));
    return el
  }

  updatePositions() {
    let scrollPane = this.props.scrollPane;
    let scrollableEl = scrollPane.getScrollableElement();
    let contentHeight = scrollPane.getContentHeight();
    let scrollPaneHeight = scrollPane.getHeight();
    let scrollTop = scrollPane.getScrollPosition();
    let contentEl = scrollPane.getContentElement();

    // Needed for scrollbar interaction
    this.factor = (contentHeight / scrollPaneHeight);

    if (this.factor <= 1) {
      this.el.addClass('sm-hide-thumb');
    } else {
      this.el.removeClass('sm-hide-thumb');
    }

    this.refs.thumb.css({
      top: scrollTop / this.factor,
      height: scrollPaneHeight / this.factor
    });

    // If we have highlights, update them as well
    if (this.props.highlights) {
      // Compute highlights
      forEach(this.props.highlights,function(highlights) {
        forEach(highlights, function(nodeId) {
          let nodeEl = scrollableEl.find('*[data-id="'+nodeId+'"]');

          if (!nodeEl) return

          // Compute bounding rect relative to scroll pane content element
          let rect = getRelativeBoundingRect(nodeEl.getNativeElement(), contentEl.getNativeElement());
          let top = rect.top / this.factor;
          let height = rect.height / this.factor;

          // Use specified minHeight for highlights
          if (height < Scrollbar.overlayMinHeight) {
            height = Scrollbar.overlayMinHeight;
          }

          let highlightEl = this.refs[nodeId];
          if (highlightEl) {
            this.refs[nodeId].css({
              top: top,
              height: height
            });
          } else {
            console.warn('no ref found for highlight', nodeId);
          }
        }.bind(this));
      }.bind(this));
    }
  }

  getScrollableElement() {
    return this.props.scrollPane.getScrollableElement()
  }

  onResize() {
    this.rerender();
  }

  onScroll() {
    this.updatePositions();
  }

  onMouseDown(e) {
    e.stopPropagation();
    e.preventDefault();
    this._mouseDown = true;

    // temporarily, we bind to events on window level
    // because could leave the this element's area while dragging
    let _window = DefaultDOMElement.getBrowserWindow();
    _window.on('mousemove', this.onMouseMove, this);
    _window.on('mouseup', this.onMouseUp, this);

    let scrollBarOffset = this.el.getOffset().top;
    let y = e.pageY - scrollBarOffset;
    let thumbEl = this.refs.thumb.el;
    if (e.target !== thumbEl.getNativeElement()) {
      // Jump to mousedown position
      this.offset = thumbEl.height / 2;
      this.onMouseMove(e);
    } else {
      this.offset = y - thumbEl.getPosition().top;
    }
  }

  // Handle Mouse Up
  onMouseUp() {
    this._mouseDown = false;
    let _window = DefaultDOMElement.getBrowserWindow();
    _window.off('mousemove', this.onMouseMove, this);
    _window.off('mouseup', this.onMouseUp, this);
  }

  onMouseMove(e) {
    if (this._mouseDown) {
      let scrollPane = this.props.scrollPane;
      let scrollableEl = scrollPane.getScrollableElement();
      let scrollBarOffset = this.el.getOffset().top;
      let y = e.pageY - scrollBarOffset;

      // find offset to visible-area.top
      let scroll = (y-this.offset)*this.factor;
      scrollableEl.setProperty('scrollTop', scroll);
    }
  }
}

Scrollbar.overlayMinHeight = 2;

var ScrollbarPackage = {
  name: 'scrollbar',
  configure: function(config) {
    config.addComponent('scrollbar', Scrollbar);
  }
};

/**
  Wraps content in a scroll pane.

  NOTE: It is best practice to put all overlays as direct childs of the ScrollPane
        to reduce the chance that positioning gets messed up (position: relative)

  @prop {String} scrollbarType 'native' or 'substance' for a more advanced visual scrollbar. Defaults to 'native'
  @prop {String} [scrollbarPosition] 'left' or 'right' only relevant when scrollBarType: 'substance'. Defaults to 'right'
  @prop {ui/Highlights} [highlights] object that maintains highlights and can be manipulated from different sources
  @prop {ui/TOCProvider} [tocProvider] object that maintains table of content entries

  @example

  ```js
  $$(ScrollPane, {
    scrollbarType: 'substance', // defaults to native
    scrollbarPosition: 'left', // defaults to right
    onScroll: this.onScroll.bind(this),
    highlights: this.contentHighlights,
    tocProvider: this.tocProvider
  })
  ```
*/
class ScrollPane extends AbstractScrollPane$$1 {

  didMount() {
    super.didMount();
    if (this.refs.scrollbar && this.props.highlights) {
      this.props.highlights.on('highlights:updated', this.onHighlightsUpdated, this);
    }
    if (this.refs.scrollbar) {
      this.domObserver = new window.MutationObserver(this._onContentChanged.bind(this));
      this.domObserver.observe(this.el.getNativeElement(), {
        subtree: true,
        attributes: true,
        characterData: true,
        childList: true,
      });
      this.context.editorSession.onPosition(this._onPosition, this);
    }
  }

  dispose() {
    super.dispose();
    if (this.props.highlights) {
      this.props.highlights.off(this);
    }
    this.context.editorSession.off(this);
    this.context.dragManager.off(this);
  }

  render($$) {
    let el = $$('div')
      .addClass('sc-scroll-pane');

    if (platform.isFF) {
      el.addClass('sm-firefox');
    }

    // When noStyle is provided we just use ScrollPane as a container, but without
    // any absolute positioned containers, leaving the body scrollable.
    if (!this.props.noStyle) {
      el.addClass('sm-default-style');
    }

    // Initialize Substance scrollbar (if enabled)
    if (this.props.scrollbarType === 'substance') {
      el.addClass('sm-substance-scrollbar');
      el.addClass('sm-scrollbar-position-' + this.props.scrollbarPosition);

      el.append(
        // TODO: is there a way to pass scrollbar highlights already
        // via props? Currently the are initialized with a delay
        $$(Scrollbar, {
          scrollPane: this
        }).ref('scrollbar')
          .attr('id', 'content-scrollbar')
      );

      // Scanline is debugging purposes, display: none by default.
      el.append(
        $$('div').ref("scanline").addClass('se-scanline')
      );
    }

    el.append(
      $$('div').ref('scrollable').addClass('se-scrollable').append(
        this.renderContent($$)
      ).on('scroll', this.onScroll)
    );
    return el
  }

  renderContent($$) {
    let contentEl = $$('div').ref('content').addClass('se-content');
    contentEl.append(this.props.children);
    if (this.props.contextMenu === 'custom') {
      contentEl.on('contextmenu', this._onContextMenu);
    }
    return contentEl
  }

  _onContentChanged() {
    this._contentChanged = true;
  }

  _onPosition() {
    if (this.refs.scrollbar && this._contentChanged) {
      this._contentChanged = false;
      this._updateScrollbar();
    }
  }

  _updateScrollbar() {
    if (this.refs.scrollbar) {
      this.refs.scrollbar.updatePositions();
    }
  }

  onHighlightsUpdated(highlights) {
    this.refs.scrollbar.extendProps({
      highlights: highlights
    });
  }

  onScroll() {
    let scrollPos = this.getScrollPosition();
    let scrollable = this.refs.scrollable;
    if (this.props.onScroll) {
      this.props.onScroll(scrollPos, scrollable);
    }
    // Update TOCProvider given
    if (this.props.tocProvider) {
      this.props.tocProvider.markActiveEntry(this);
    }
    this.emit('scroll', scrollPos, scrollable);
  }

  /**
    Returns the height of scrollPane (inner content overflows)
  */
  getHeight() {
    let scrollableEl = this.getScrollableElement();
    return scrollableEl.height
  }

  /**
    Returns the cumulated height of a panel's content
  */
  getContentHeight() {
    let contentEl = this.refs.content.el.getNativeElement();
    // Important to use scrollHeight here (e.g. to consider overflowing
    // content, that stretches the content area, such as an overlay or
    // a context menu)
    return contentEl.scrollHeight
  }

  /**
    Get the `.se-content` element
  */
  getContentElement() {
    return this.refs.content.el
  }

  /**
    Get the `.se-scrollable` element
  */
  getScrollableElement() {
    return this.refs.scrollable.el
  }

  /**
    Get current scroll position (scrollTop) of `.se-scrollable` element
  */
  getScrollPosition() {
    let scrollableEl = this.getScrollableElement();
    return scrollableEl.getProperty('scrollTop')
  }

  setScrollPosition(scrollPos) {
    let scrollableEl = this.getScrollableElement();
    scrollableEl.setProperty('scrollTop', scrollPos);
  }

  /**
    Get offset relative to `.se-content`.

    @param {DOMNode} el DOM node that lives inside the
  */
  getPanelOffsetForElement(el) {
    let nativeEl = el.getNativeElement();
    let contentContainerEl = this.refs.content.getNativeElement();
    let rect = getRelativeBoundingRect(nativeEl, contentContainerEl);
    return rect.top
  }

  /**
    Scroll to a given sub component.

    @param {String} componentId component id, must be present in data-id attribute
  */
  scrollTo(componentId, onlyIfNotVisible) {
    let scrollableEl = this.getScrollableElement();
    let targetNode = scrollableEl.find('*[data-id="'+componentId+'"]');
    if (targetNode) {
      const offset = this.getPanelOffsetForElement(targetNode);
      let shouldScroll = true;
      if (onlyIfNotVisible) {
        const height = scrollableEl.height;
        const oldOffset = scrollableEl.getProperty('scrollTop');
        shouldScroll = (offset < oldOffset || oldOffset+height<offset);
      }
      if (shouldScroll) {
        this.setScrollPosition(offset);
      }
    } else {
      console.warn(componentId, 'not found in scrollable container');
    }
  }

  /*
    Determines the selection bounding rectangle relative to the scrollpane's content.
  */
  onSelectionPositioned(...args) {
    super.onSelectionPositioned(...args);
    this._updateScrollbar();
  }

  _onContextMenu(e) {
    super._onContextMenu(e);
    this._updateScrollbar();
  }

}

var ScrollPanePackage = {
  name: 'scroll-pane',
  configure: function(config) {
    config.addComponent('scroll-pane', ScrollPane);
  }
};

/**
  Wraps content in a scroll pane.

  @class ScrollPane
  @component

  @prop {String} scrollbarType 'native' or 'substance' for a more advanced visual scrollbar. Defaults to 'native'
  @prop {String} [scrollbarPosition] 'left' or 'right' only relevant when scrollBarType: 'substance'. Defaults to 'right'
  @prop {ui/Highlights} [highlights] object that maintains highlights and can be manipulated from different sources
  @prop {ui/TOCProvider} [tocProvider] object that maintains table of content entries

  @example

  ```js
  $$(BodyScrollPane).append(
    content,
    $$(ContextMenu)
    $$(Overlay)
  )
  ```
*/
class BodyScrollPane extends AbstractScrollPane$$1 {

  /*
    Expose scrollPane as a child context
  */
  getChildContext() {
    return {
      scrollPane: this
    }
  }

  getName() {
    return 'body'
  }

  render($$) {
    let el = $$('div');
    if (this.props.contextMenu === 'custom') {
      el.on('contextmenu', this._onContextMenu);
    }
    el.append(this.props.children);
    return el
  }

  /**
    Returns the height of scrollPane (inner content overflows)
  */
  getHeight() {
    return window.innerHeight
  }

  /**
    Returns the cumulated height of a panel's content
  */
  getContentHeight() {
    return document.body.scrollHeight
  }

  getContentElement() {
    return DefaultDOMElement.wrapNativeElement(window.document.body)
  }

  // /**
  //   Get the `.se-scrollable` element
  // */
  getScrollableElement() {
    return document.body
  }

  /**
    Get current scroll position (scrollTop) of `.se-scrollable` element
  */
  getScrollPosition() {
    return document.body.scrollTop
  }

  setScrollPosition(scrollPos) {
    document.body.scrollTop = scrollPos;
  }

  /**
    Get offset relative to `.se-content`.

    @param {DOMNode} el DOM node that lives inside the
  */
  getPanelOffsetForElement(el) { // eslint-disable-line
    console.warn('TODO: implement');
  }

  /**
    Scroll to a given sub component.

    @param {String} componentId component id, must be present in data-id attribute
  */
  scrollTo(componentId, onlyIfNotVisible) { // eslint-disable-line
    console.warn('TODO: implement');
  }

}

var BodyScrollPanePackage = {
  name: 'body-scroll-pane',
  configure: function(config) {
    config.addComponent('body-scroll-pane', BodyScrollPane);
  }
};

/**
  A split view layout component. Takes properties for configuration and 2 children via append.

  @class SplitPane
  @component

  @prop {String} splitType either 'vertical' (default) or 'horizontal'.
  @prop {String} sizeA size of the first pane (A). '40%' or '100px' or 'inherit' are valid values.
  @prop {String} sizeB size of second pane. sizeA and sizeB can not be combined.

  @example

  ```js
  $$(SplitPane, {
    sizeA: '30%',
    splitType: 'horizontal'
  }).append(
    $$('div').append('Pane A'),
    $$('div').append('Pane B')
  )
  ```
*/

class SplitPane extends Component {

  render($$) {
    if (this.props.children.length !== 2) {
      throw new Error('SplitPane only works with exactly two child elements')
    }

    let el = $$('div').addClass('sc-split-pane');
    if (this.props.splitType === 'horizontal') {
      el.addClass('sm-horizontal');
    } else {
      el.addClass('sm-vertical');
    }

    let paneA = this.props.children[0];
    let paneB = this.props.children[1];

    // Apply configured size either to pane A or B.
    if (this.props.sizeB) {
      paneB.addClass('se-pane sm-sized');
      paneB.css(this.getSizedStyle(this.props.sizeB));
      paneA.addClass('se-pane sm-auto-fill');
    } else {
      paneA.addClass('se-pane sm-sized');
      paneA.css(this.getSizedStyle(this.props.sizeA));
      paneB.addClass('se-pane sm-auto-fill');
    }

    el.append(
      paneA,
      paneB
    );
    return el
  }

  // Accepts % and px units for size property
  getSizedStyle(size) {
    if (!size || size === 'inherit') return {}
    if (this.props.splitType === 'horizontal') {
      return {'height': size}
    } else {
      return {'width': size}
    }
  }

}

var SplitPanePackage = {
  name: 'split-pane',
  configure: function(config) {
    config.addComponent('split-pane', SplitPane);
  }
};

/*
  A tabbed pane layout component. The actual content is specified via append.

  @class TabbedPane
  @component

  @prop {Object[]} tabs an array of objects with id and name properties
  @prop {String} activeTab id of currently active tab

  @example

  ```js
  $$(TabbedPane, {
    tabs: [
      {id: 'tabA', 'A'},
      {id: 'tabB', 'B'},
    ],
    activeTab: 'tabA'
  }).ref('tabbedPane').append(
    tabAContent
  )
  ```
*/

class TabbedPane extends Component {

  render($$) {
    let el = $$('div').addClass('sc-tabbed-pane');
    let tabsEl = $$('div').addClass('se-tabs');
    forEach(this.props.tabs, function(tab) {
      let tabEl = $$('a')
        .addClass("se-tab")
        .attr({
          href: "#",
          "data-id": tab.id,
        })
        .on('click', this.onTabClicked);
      if (tab.id === this.props.activeTab) {
        tabEl.addClass("sm-active");
      }
      tabEl.append(
        $$('span').addClass('label').append(tab.name)
      );
      tabsEl.append(tabEl);
    }.bind(this));

    el.append(tabsEl);
    // Active content
    el.append(
      $$('div').addClass('se-tab-content').ref('tabContent').append(
        this.props.children
      )
    );
    return el
  }

  onTabClicked(e) {
    e.preventDefault();
    let tabId = e.currentTarget.dataset.id;
    this.send('switchTab', tabId);
  }
}

var TabbedPanePackage = {
  name: 'tabbed-pane',
  configure: function(config) {
    config.addComponent('tabbed-pane', TabbedPane);
  }
};

var FilePackage = {
  name: 'file',
  configure: function(config) {
    config.addNode(FileNode);
  }
};

class Undo extends Command {

  getCommandState(params) {
    let editorSession = params.editorSession;
    return {
      disabled: !editorSession.canUndo(),
      active: false
    }
  }

  execute(params) {
    let editorSession = params.editorSession;
    if (editorSession.canUndo()) {
      editorSession.undo();
      return true
    }
    return false
  }

}

class Redo extends Command {

  getCommandState(params) {
    let editorSession = params.editorSession;
    return {
      disabled: !editorSession.canRedo(),
      active: false
    }
  }

  execute(params) {
    let editorSession = params.editorSession;
    if (editorSession.canRedo()) {
      editorSession.redo();
      return true
    } else {
      return false
    }
  }

}

class SelectAll extends Command {

  getCommandState(params) {
    let editorSession = params.editorSession;
    return {
      disabled: editorSession.getSelection().isNull()
    }
  }

  execute(params) {
    let editorSession = params.editorSession;
    let doc = editorSession.getDocument();
    let surface = params.surface || editorSession.getFocusedSurface();
    if (surface) {
      let sel;
      // TODO: we should move the logic out of the surfaces
      if (surface._isContainerEditor) {
        let container = surface.getContainer();
        if (container.nodes.length === 0) {
          return false
        }
        let firstNodeId = container.nodes[0];
        let lastNodeId = last$1(container.nodes);
        sel = editorSession.createSelection({
          type: 'container',
          startPath: [firstNodeId],
          startOffset: 0,
          endPath: [lastNodeId],
          endOffset: 1,
          containerId: container.id,
          surfaceId: surface.id
        });
      } else if (surface._isTextPropertyEditor) {
        let path = surface.getPath();
        let text = doc.get(path);
        sel = editorSession.createSelection({
          type: 'property',
          path: path,
          startOffset: 0,
          endOffset: text.length,
          surfaceId: surface.id
        });
      }
      editorSession.setSelection(sel);
      return true
    }
    return false
  }
}

var BasePackage = {
  name: 'base',
  configure: function(config) {
    config.import(FilePackage);
    config.import(ScrollPanePackage);
    config.import(BodyScrollPanePackage);
    config.import(SplitPanePackage);
    config.import(TabbedPanePackage);
    config.import(ScrollbarPackage);
    config.import(GridPackage);
    config.import(ModalPackage);
    config.import(InputPackage);
    config.import(ButtonPackage);
    config.import(LayoutPackage);
    config.import(ContextMenuPackage);
    config.import(OverlayPackage);
    config.import(DropzonesPackage);
    config.import(GutterPackage);

    // Setup base toolgroups
    config.addToolGroup('document');
    config.addToolGroup('annotations');
    config.addToolGroup('default');
    config.addToolGroup('context-menu-primary');
    config.addToolGroup('context-menu-document');
    config.addToolGroup('insert');

    // Commands
    config.addCommand('undo', Undo);
    config.addCommand('redo', Redo);
    config.addCommand('select-all', SelectAll);

    config.addCommand('insert-node', InsertNodeCommand);

    // Tools
    config.addTool('undo', Tool, {toolGroup: ['document', 'context-menu-document']});
    config.addTool('redo', Tool, {toolGroup: ['document', 'context-menu-document']});
    config.addTool('select-all', Tool, {toolGroup: ['context-menu-document']});

    // Icons
    config.addIcon('undo', { 'fontawesome': 'fa-undo' });
    config.addIcon('redo', { 'fontawesome': 'fa-repeat' });
    config.addIcon('edit', { 'fontawesome': 'fa-cog' });
    config.addIcon('delete', { 'fontawesome': 'fa-times' });
    config.addIcon('expand', { 'fontawesome': 'fa-arrows-h' });
    config.addIcon('truncate', { 'fontawesome': 'fa-arrows-h' });

    // Labels
    config.addLabel('undo', {
      en: 'Undo',
      de: 'Rückgängig'
    });
    config.addLabel('redo', {
      en: 'Redo',
      de: 'Wiederherstellen'
    });
    config.addLabel('select-all', {
      en: 'Select All',
      de: 'Alles Auswählen'
    });
    config.addLabel('container-selection', {
      en: 'Container',
      de: 'Container'
    });
    config.addLabel('container', {
      en: 'Container',
      de: 'Container'
    });
    config.addLabel('insert-container', {
      en: 'Insert Container',
      de: 'Container einfügen'
    });

    if (platform.isMac) {
      config.addKeyboardShortcut('cmd+z', { command: 'undo' });
      config.addKeyboardShortcut('cmd+shift+z', { command: 'redo' });
      config.addKeyboardShortcut('cmd+a', { command: 'select-all' });
    } else {
      config.addKeyboardShortcut('ctrl+z', { command: 'undo' });
      config.addKeyboardShortcut('ctrl+shift+z', { command: 'redo' });
      config.addKeyboardShortcut('ctrl+a', { command: 'select-all' });
    }
  },
  UndoCommand: Undo,
  RedoCommand: Redo,
  SelectAllCommand: SelectAll
};

class Blockquote extends TextBlock {}

Blockquote.type = "blockquote";

class BlockquoteComponent extends TextBlockComponent {
  render($$) {
    let el = super.render.call(this, $$);
    return el.addClass('sc-blockquote')
  }
}

/*
 * HTML converter for Blockquote.
 */
var BlockquoteHTMLConverter = {

  type: 'blockquote',
  tagName: 'blockquote',

  import: function(el, node, converter) {
    node.content = converter.annotatedText(el, [node.id, 'content']);
  },

  export: function(node, el, converter) {
    el.append(
      converter.annotatedText([node.id, 'content'])
    );
  }

};

var BlockquotePackage = {
  name: 'blockquote',
  configure: function(config) {
    config.addNode(Blockquote);
    config.addComponent(Blockquote.type, BlockquoteComponent);
    config.addConverter('html', BlockquoteHTMLConverter);
    config.addConverter('xml', BlockquoteHTMLConverter);
    config.addTextType({
      name: 'blockquote',
      data: {type: 'blockquote'}
    });
    config.addLabel('blockquote', {
      en: 'Blockquote',
      de: 'Blockzitat'
    });
  },
  Blockquote: Blockquote,
  BlockquoteComponent: BlockquoteComponent,
  BlockquoteHTMLConverter: BlockquoteHTMLConverter
};

class Code extends PropertyAnnotation {}

Code.type = 'code';

var CodeHTMLConverter = {
  type: 'code',
  tagName: 'code'
};

var CodePackage = {
  name: 'code',
  configure: function(config, {toolGroup, disableCollapsedCursor}) {
    config.addNode(Code);
    config.addConverter('html', CodeHTMLConverter);
    config.addConverter('xml', CodeHTMLConverter);
    config.addComponent('code', AnnotationComponent);
    config.addCommand('code', AnnotationCommand, {
      disableCollapsedCursor,
      nodeType: Code.type
    });
    config.addTool('code', AnnotationTool, {
      toolGroup: toolGroup || 'annotations'
    });
    config.addIcon('code', { 'fontawesome': 'fa-code' });
    config.addLabel('code', {
      en: 'Code',
      de: 'Code'
    });
  },
  Code: Code,
  CodeHTMLConverter: CodeHTMLConverter
};

class Codeblock extends TextBlock {}

Codeblock.type = "codeblock";

class CodeblockComponent extends TextBlockComponent {
  render($$) {
    let el = super.render.call(this, $$);
    return el.addClass('sc-codeblock')
  }
}

var CodeblockHTMLConverter = {

  type: 'codeblock',
  tagName: 'pre',

  import: function(el, node, converter) {
    let codeEl = el.find('code');
    if (codeEl) {
      node.content = converter.annotatedText(codeEl, [node.id, 'content'], { preserveWhitespace: true });
    }
  },

  export: function(node, el, converter) {
    let $$ = converter.$$;
    el.append(
      $$('code').append(
        converter.annotatedText([node.id, 'content'])
      )
    );
  }
};

var CodeblockPackage = {
  name: 'codeblock',
  configure: function(config) {
    config.addNode(Codeblock);
    config.addComponent('codeblock', CodeblockComponent);
    config.addConverter('html', CodeblockHTMLConverter);
    config.addConverter('xml', CodeblockHTMLConverter);
    config.addTextType({
      name: 'codeblock',
      data: {type: 'codeblock'}
    });
    config.addLabel('codeblock', {
      en: 'Codeblock',
      de: 'Codeblock'
    });
  },
  Codeblock: Codeblock,
  CodeblockComponent: CodeblockComponent,
  CodeblockHTMLConverter: CodeblockHTMLConverter
};

class Emphasis extends PropertyAnnotation {}

Emphasis.type = "emphasis";

// hint for rendering in presence of overlapping annotations
Emphasis.fragmentation = Fragmenter.ANY;

var EmphasisHTMLConverter = {

  type: 'emphasis',
  tagName: 'em',

  matchElement: function(el) {
    return el.is('em, i')
  }

};

class EmphasisComponent extends AnnotationComponent {
  getTagName() {
    return 'em'
  }
}

var EmphasisPackage = {
  name: 'emphasis',
  configure: function(config, {toolGroup, disableCollapsedCursor}) {
    config.addNode(Emphasis);
    config.addConverter('html', EmphasisHTMLConverter);
    config.addConverter('xml', EmphasisHTMLConverter);
    config.addComponent('emphasis', EmphasisComponent);
    config.addCommand('emphasis', AnnotationCommand, {
      nodeType: Emphasis.type,
      disableCollapsedCursor
    });
    config.addTool('emphasis', AnnotationTool, {
      toolGroup: toolGroup || 'annotations'
    });
    config.addIcon('emphasis', { 'fontawesome': 'fa-italic' });
    config.addLabel('emphasis', {
      en: 'Emphasis',
      de: 'Betonung'
    });
    if (platform.isMac) {
      config.addKeyboardShortcut('cmd+i', { command: 'emphasis' });
    } else {
      config.addKeyboardShortcut('ctrl+i', { command: 'emphasis' });
    }
  },
  Emphasis,
  EmphasisComponent,
  EmphasisHTMLConverter
};

class Heading extends TextBlock {}

Heading.schema = {
  type: "heading",
  level: { type: "number", default: 1 }
};

class HeadingComponent extends TextBlockComponent {
  render($$) {
    let el = super.render($$);
    return el.addClass("sc-heading sm-level-"+this.props.node.level)
  }

  getTagName() {
    return 'h'+this.props.node.level
  }
}

var HeadingHTMLConverter = {

  type: "heading",

  matchElement: function(el) {
    return /^h\d$/.exec(el.tagName)
  },

  import: function(el, node, converter) {
    node.level = Number(el.tagName[1]);
    node.content = converter.annotatedText(el, [node.id, 'content']);
  },

  export: function(node, el, converter) {
    el.tagName = 'h'+node.level;
    el.append(
      converter.annotatedText([node.id, 'content'])
    );
  }

};

var HeadingPackage = {
  name: 'heading',
  configure: function(config) {
    config.addNode(Heading);
    config.addComponent(Heading.type, HeadingComponent);
    config.addConverter('html', HeadingHTMLConverter);
    config.addConverter('xml', HeadingHTMLConverter);
    config.addTextType({
      name: 'heading1',
      data: {type: 'heading', level: 1}
    });
    config.addTextType({
      name: 'heading2',
      data: {type: 'heading', level: 2}
    });
    config.addTextType({
      name: 'heading3',
      data: {type: 'heading', level: 3}
    });
    config.addLabel('heading1', {
      en: 'Heading 1',
      de: 'Überschrift 1'
    });
    config.addLabel('heading2', {
      en: 'Heading 2',
      de: 'Überschrift 2'
    });
    config.addLabel('heading3', {
      en: 'Heading 3',
      de: 'Überschrift 3'
    });
  },
  Heading,
  HeadingComponent,
  HeadingHTMLConverter
};

class ImageNode extends DocumentNode {

  getImageFile() {
    if (this.imageFile) {
      return this.document.get(this.imageFile)
    }
  }

  getUrl() {
    let imageFile = this.getImageFile();
    if (imageFile) {
      return imageFile.getUrl()
    }
  }
}

ImageNode.schema = {
  type: 'image',
  imageFile: { type: 'file' }
};

class ImageComponent extends NodeComponent {

  didMount() {
    super.didMount.call(this);
    this.context.editorSession.onRender('document', this._onDocumentChange, this);
  }

  dispose() {
    super.dispose.call(this);
    this.context.editorSession.off(this);
  }

  // TODO: verify if this check is correct and efficient
  _onDocumentChange(change) {
    if (change.isAffected(this.props.node.id) ||
      change.isAffected(this.props.node.imageFile)) {
      this.rerender();
    }
  }

  render($$) {
    let el = super.render($$);
    el.addClass('sc-image');
    el.append(
      $$('img').attr({
        src: this.props.node.getUrl(),
      }).ref('image')
    );
    return el
  }

  /* Custom dropzone protocol */
  getDropzoneSpecs() {
    return [
      {
        component: this.refs['image'],
        message: 'Replace Image',
        dropParams: {
          action: 'replace-image',
          nodeId: this.props.node.id,
        }
      }
    ]
  }

  handleDrop(tx, dragState) {
    let newImageFile = dragState.data.files[0];
    if (dragState.external) {
      let imageFile = tx.create({
        type: 'file',
        fileType: 'image',
        mimeType: newImageFile.type,
        url: URL.createObjectURL(newImageFile)
      });
      // TODO: we should delete the old image file if there are no
      // referenecs to it anymore
      tx.set([this.props.node.id, 'imageFile'], imageFile.id);
    } else {
      let nodeId = dragState.sourceSelection.nodeId;
      let node = tx.get(nodeId);
      if (node.type === 'image') {
        // Use the same filenode as the dragged source node
        tx.set([this.props.node.id, 'imageFile'], node.imageFile);
      }
    }


  }

}

var ImageHTMLConverter = {

  type: 'image',
  tagName: 'img',

  import: function(el, node, converter) {
    let imageFile = converter.getDocument().create({
      id: 'file-'+node.id,
      type: 'file',
      fileType: 'image',
      url: el.attr('src')
    });
    node.imageFile = imageFile.id;
  },

  export: function(node, el) {
    let imageFile = node.document.get(node.imageFile);
    el.attr('src', imageFile.getUrl());
  }
};

var insertImage = function(tx, file) {
  // Create file node for the image
  let imageFile = tx.create({
    type: 'file',
    fileType: 'image',
    mimeType: file.type,
    sourceFile: file
  });
  // Inserts image at current cursor pos
  tx.insertBlockNode({
    type: 'image',
    imageFile: imageFile.id
  });
};

class ImageCommand extends InsertNodeCommand {

  /*
    Inserts file and image nodes
  */
  execute(params) {
    let editorSession = params.editorSession;

    editorSession.transaction((tx) => {
      params.files.forEach((file) => {
        insertImage(tx, file);
      });
    });
  }

}

class InsertImageTool extends Tool {

  getClassNames() {
    return 'sc-insert-image-tool'
  }

  renderButton($$) {
    let button = super.renderButton($$);
    let input = $$('input').attr('type', 'file').ref('input')
      .on('change', this.onFileSelect);
    return [button, input]
  }

  onClick() {
    this.refs.input.click();
  }

  onFileSelect(e) {
    let files = e.currentTarget.files;
    this.executeCommand({
      files: Array.prototype.slice.call(files)
    });
  }

}

var DropImage = {
  type: 'drop-asset',
  match(params) {
    // Mime-type starts with 'image/'
    let isImage = params.file.type.indexOf('image/') === 0;
    return params.type === 'file' && isImage
  },
  drop(tx, params) {
    insertImage(tx, params.file);
  }
};

class ImageProxy extends DefaultFileProxy {}

// to detect that this class should take responsibility for a fileNode
ImageProxy.match = function(fileNode) {
  return fileNode.fileType === 'image'
};

var ImagePackage = {
  name: 'image',
  configure: function(config) {
    config.addNode(ImageNode);
    config.addComponent('image', ImageComponent);
    config.addConverter('html', ImageHTMLConverter);
    config.addConverter('xml', ImageHTMLConverter);
    config.addCommand('insert-image', ImageCommand);
    config.addTool('insert-image', InsertImageTool);
    config.addIcon('insert-image', { 'fontawesome': 'fa-image' });
    config.addLabel('image', {
      en: 'Image',
      de: 'Bild'
    });
    config.addLabel('insert-image', {
      en: 'Insert image',
      de: 'Bild einfügen'
    });

    config.addDropHandler(DropImage);
    config.addFileProxy(ImageProxy);
  },
  ImageNode: ImageNode,
  ImageComponent: ImageComponent,
  ImageHTMLConverter: ImageHTMLConverter,
  InsertImageCommand: ImageCommand,
  InsertImageTool: InsertImageTool,
  DropImage: DropImage
};

class InlineWrapper extends InlineNode {
  getWrappedNode() {
    return this.getDocument().get(this.wrappedNode)
  }
}

InlineWrapper.schema = {
  type: 'inline-wrapper',
  wrappedNode: 'id'
};

class InlineWrapperComponent extends InlineNodeComponent {

  getClassNames() {
    // ATTENTION: ATM it is necessary to add .sc-inline-node
    return 'sc-inline-wrapper sc-inline-node'
  }

  renderContent($$) {
    let node = this.props.node;
    let doc = node.getDocument();

    let wrappedNode = doc.get(node.wrappedNode);
    let el;
    if (wrappedNode) {
      let componentRegistry = this.context.componentRegistry;
      let ComponentClass = componentRegistry.get(wrappedNode.type);
      if (ComponentClass) {
        el = $$(ComponentClass, {
          disabled: this.props.disabled,
          node: wrappedNode,
        });
      } else {
        console.error('No component registered for node type' + wrappedNode.type);
      }
    } else {
      console.error('Could not find wrapped node: ' + node.wrappedNode);
    }
    return el
  }
}

var InlineWrapperConverter = {
  type: 'inline-wrapper',

  matchElement: function(el, converter) {
    var blockConverter = converter._getConverterForElement(el, 'block');
    return Boolean(blockConverter)
  },

  import: function(el, node, converter) {
    // HACK monkey patching the context
    node.id = converter.nextId('inline-wrapper');
    var state = converter.state;
    state.popElementContext();
    state.pushElementContext(state.getCurrentElementContext().tagName);
    node.wrappedNode = converter.convertElement(el).id;
  },

  export: function(node, el, converter) {
    return converter.convertNode(node.wrappedNode)
  }
};

/*
  This package adds a node to the model which can be used
  to use a block-level node within an inline context.

    The quick brown fox jumps over the lazy <fig><img src='./dog.jpg'/></fig>.

  To register the converter you must provide `config.converters` which is
  an array of names of the converters you want this to be registered in.
*/
var InlineWrapperPackage = {
  name: 'inline-wrapper',
  configure: function(config, options) {
    config.addNode(InlineWrapper);
    config.addComponent(InlineWrapper.type, InlineWrapperComponent);
    if (options.converters) {
      options.converters.forEach(function(name) {
        config.addConverter(name, InlineWrapperConverter);
      });
    }
  },
  InlineWrapper: InlineWrapper,
  InlineWrapperComponent: InlineWrapperComponent,
  InlineWrapperConverter: InlineWrapperConverter
};

class Link extends PropertyAnnotation {}

Link.schema = {
  type: "link",
  title: { type: 'string', optional: true },
  url: { type: 'string', 'default': ''}
};

// in presence of overlapping annotations will try to render this as one element
Link.fragmentation = Fragmenter.SHOULD_NOT_SPLIT;

Link.autoExpandRight = false;

class LinkComponent extends AnnotationComponent {

  didMount(...args) {
    super.didMount(...args);

    let node = this.props.node;
    this.context.editorSession.onRender('document', this.rerender, this, {
      path: [node.id, 'url']
    });
  }

  dispose(...args) {
    super.dispose(...args);

    this.context.editorSession.off(this);
  }

  render($$) { // eslint-disable-line
    let el = super.render($$);

    el.tagName = 'a';
    el.attr('href', this.props.node.url);

    let titleComps = [this.props.node.url];
    if (this.props.node.title) {
      titleComps.push(this.props.node.title);
    }

    return el.attr("title", titleComps.join(' | '))
  }

}

class LinkCommand extends AnnotationCommand {
  canFuse() { return false }

  /*
    On link creation we collapse the selection, as this is a condition for the
    EditLinkTool to be shown (see EditAnnotationCommand)
  */
  executeCreate(params) {
    let result = super.executeCreate(params);
    let editorSession = this._getEditorSession(params);
    editorSession.transaction((tx) => {
      tx.setSelection(tx.selection.collapse());
    });
    return result
  }
}

var LinkHTMLConverter = {

  type: "link",
  tagName: 'a',

  import: function(el, node) {
    node.url = el.attr('href');
    node.title = el.attr('title');
  },

  export: function(link, el) {
    el.attr({
      href: link.url,
      title: link.title
    });
  }

};

/**
  Tool to edit an existing link.

  Designed so that it can be used either in a toolbar, or within
  an overlay on the Surface.

  @component
*/
class EditLinkTool extends Tool {

  getUrlPath() {
    let propPath = this.constructor.urlPropertyPath;
    return [this.props.node.id].concat(propPath)
  }

  _openLink() {
    let doc = this.context.editorSession.getDocument();
    window.open(doc.get(this.getUrlPath()), '_blank');
  }

  render($$) {
    let Input = this.getComponent('input');
    let Button = this.getComponent('button');
    let el = $$('div').addClass('sc-edit-link-tool');

    // GUARD: Return if tool is disabled
    if (this.props.disabled) {
      console.warn('Tried to render EditLinkTool while disabled.');
      return el
    }

    let urlPath = this.getUrlPath();

    el.append(
      $$(Input, {
        type: 'url',
        path: urlPath,
        placeholder: 'Paste or type a link url'
      }),
      $$(Button, {
        icon: 'open-link',
        style: this.props.style
      }).attr('title', this.getLabel('open-link'))
        .on('click', this._openLink),

      $$(Button, {
        icon: 'delete',
        style: this.props.style
      }).attr('title', this.getLabel('delete-link'))
        .on('click', this.onDelete)
    );
    return el
  }

  onDelete(e) {
    e.preventDefault();
    let node = this.props.node;
    let sm = this.context.surfaceManager;
    let surface = sm.getFocusedSurface();
    if (!surface) {
      console.warn('No focused surface. Stopping command execution.');
      return
    }
    let editorSession = this.context.editorSession;
    editorSession.transaction(function(tx, args) {
      tx.delete(node.id);
      return args
    });
  }
}

EditLinkTool.urlPropertyPath = ['url'];

var LinkPackage = {
  name: 'link',
  configure: function(config, {
    toolGroup,
    editLinkToolGroup
  }) {
    config.addNode(Link);
    config.addComponent('link', LinkComponent);
    config.addConverter('html', LinkHTMLConverter);
    config.addConverter('xml', LinkHTMLConverter);
    config.addCommand('link', LinkCommand, {
      nodeType: 'link',
      disableCollapsedCursor: true
    });
    config.addCommand('edit-link', EditAnnotationCommand, {
      nodeType: 'link'
    });
    config.addTool('link', AnnotationTool, {
      toolGroup: toolGroup || 'annotations'
    });
    config.addTool('edit-link', EditLinkTool, {
      toolGroup: editLinkToolGroup || 'overlay'
    });
    config.addIcon('link', { 'fontawesome': 'fa-link'});
    config.addIcon('open-link', { 'fontawesome': 'fa-external-link' });
    config.addLabel('link', {
      en: 'Link',
      de: 'Link'
    });
    config.addLabel('open-link', {
      en: 'Open Link',
      de: 'Link öffnen'
    });
    config.addLabel('delete-link', {
      en: 'Remove Link',
      de: 'Link löschen'
    });
    if (platform.isMac) {
      config.addKeyboardShortcut('cmd+k', { command: 'link' });
    } else {
      config.addKeyboardShortcut('ctrl+k', { command: 'link' });
    }
  },
  Link,
  LinkComponent,
  LinkCommand,
  LinkHTMLConverter,
  EditLinkTool
};

class ListNode extends DocumentNode {

  getItemAt(idx) {
    return this.getDocument().get(this.items[idx])
  }

  getFirstItem() {
    return this.getItemAt(0)
  }

  getLastItem() {
    return this.getItemAt(this.getLength()-1)
  }

  getItems() {
    const doc = this.getDocument();
    return this.items.map((id) => {
      return doc.get(id)
    })
  }

  getItemPosition(itemId) {
    if (itemId._isNode) itemId = itemId.id;
    let pos = this.items.indexOf(itemId);
    if (pos < 0) throw new Error('Item is not within this list: ' + itemId)
    return pos
  }

  insertItemAt(pos, itemId) {
    const doc = this.getDocument();
    doc.update([this.id, 'items'], { type: 'insert', pos: pos, value: itemId });
  }

  appendItem(itemId) {
    this.insertItemAt(this.items.length, itemId);
  }

  removeItemAt(pos) {
    const doc = this.getDocument();
    doc.update([this.id, 'items'], { type: 'delete', pos: pos });
  }

  remove(itemId) {
    const doc = this.getDocument();
    const pos = this.getItemPosition(itemId);
    if (pos >= 0) {
      doc.update([this.id, 'items'], { type: 'delete', pos: pos });
    }
  }

  isEmpty() {
    return this.items.length === 0
  }

  getLength() {
    return this.items.length
  }

  get length() {
    return this.getLength()
  }
}

ListNode.isList = true;

ListNode.type = 'list';

ListNode.schema = {
  ordered: { type: 'boolean', default: false },
  // list-items are owned by the list
  items: { type: [ 'array', 'id' ], default: [], owned: true }
};

class ListItem extends TextNode {}

ListItem.type = 'list-item';

ListItem.schema = {
  level: { type: "number", default: 1 }
};

class ListItemComponent extends TextPropertyComponent {}

function getListTagName(node) {
  // TODO: we might want to have different types for different levels
  return node.ordered ? 'ol' : 'ul'
}

function renderListNode(node, rootEl, createElement) {
  let items = node.getItems();
  let stack = [rootEl];
  for (let i = 0; i < items.length; i++) {
    let item = items[i];
    if (item.level<stack.length) {
      for (let j = stack.length; j > item.level; j--) {
        stack.pop();
      }
    } else if (item.level>stack.length) {
      for (let j = stack.length; j < item.level; j++) {
        // Note: ATM all sublists have the same order type
        let sublist = createElement(getListTagName(node));
        last$1(stack).append(sublist);
        stack.push(sublist);
      }
    }
    console.assert(item.level === stack.length, 'item.level should now be the same as stack.length');
    last$1(stack).append(
      createElement(item)
    );
  }
  for(let j=stack.length; j>1;j--) {
    stack.pop();
  }
}

class ListComponent extends Component {

  didMount() {
    this.context.editorSession.onRender('document', this._onChange, this);
  }

  render($$) {
    let node = this.props.node;
    let el = $$(getListTagName(node))
      .addClass('sc-list')
      .attr('data-id', node.id);
    renderListNode(node, el, (arg) => {
      if (isString$1(arg)) {
        return $$(arg)
      } else if(arg.type === 'list-item') {
        let item = arg;
        return $$(ListItemComponent, {
          path: [item.id, 'content'],
          node: item,
          tagName: 'li'
        })
        // setting ref to preserve items when rerendering
        .ref(item.id)
      }
    });
    return el
  }

  _onChange(change) {
    const node = this.props.node;
    if (change.isAffected(node.id)) {
      return this.rerender()
    }
    // check if any of the list items are affected
    let itemIds = node.items;
    for (let i = 0; i < itemIds.length; i++) {
      if (change.isAffected([itemIds[i], 'level'])) {
        return this.rerender()
      }
    }
  }

}

// we need this ATM to prevent this being wrapped into an isolated node (see ContainerEditor._renderNode())
ListComponent.prototype._isCustomNodeComponent = true;

var ListHTMLConverter = {

  type: "list",

  matchElement: function(el) {
    return el.is('ul') || el.is('ol')
  },

  import: function(el, node, converter) {
    let self = this;
    this._santizeNestedLists(el);
    if (el.is('ol')) {
      node.ordered = true;
    }
    node.items = [];
    let itemEls = el.findAll('li');
    itemEls.forEach(function(li) {
      // ATTENTION: pulling out nested list elements here on-the-fly
      let listItem = converter.convertElement(li);
      listItem.level = _getLevel(li);
      node.items.push(listItem.id);
    });
    function _getLevel(li) {
      let _el = li;
      let level = 1;
      while(_el) {
        if (_el.parentNode === el) return level
        _el = _el.parentNode;
        if (self.matchElement(_el)) level++;
      }
    }
  },

  export: function(node, el, converter) {
    let $$ = converter.$$;
    el.tagName = getListTagName(node);
    renderListNode(node, el, (arg)=>{
      if (isString$1(arg)) {
        return $$(arg)
      } else {
        let item = arg;
        return $$('li').append(converter.annotatedText(item.getTextPath()))
      }
    });
    return el
  },

  _santizeNestedLists(root) {
    let nestedLists = root.findAll('ol,ul');
    nestedLists.forEach((el)=>{
      while (!el.parentNode.is('ol,ul')) {
        // pull it up
        let parent = el.parentNode;
        let grandParent = parent.parentNode;
        let pos = grandParent.getChildIndex(parent);
        grandParent.insertAt(pos+1, el);
      }
    });
  }
};

var ListItemHTMLConverter = {

  type: "list-item",

  matchElement: function(el) {
    return el.is('li')
  },

  import: function(el, node, converter) {
    node.content = converter.annotatedText(el, [node.id, 'content']);
  },

  export: function(node, el, converter) {
    el.append(converter.annotatedText(node.getTextPath()));
  }
};

class InsertListCommand extends Command {
  getCommandState (params) {
    let sel = this._getSelection(params);
    let selectionState = params.editorSession.getSelectionState();
    let commandState = {};
    let _disabledCollapsedCursor = this.config.disableCollapsedCursor && sel.isCollapsed();
    if (_disabledCollapsedCursor || !sel.isPropertySelection() || selectionState.isInlineNodeSelection()) {
      commandState.disabled = true;
    }
    return commandState
  }
  execute (params) {
    let ordered = this.config.ordered;
    let editorSession = params.editorSession;
    editorSession.transaction((tx) => {
      tx.toggleList({ ordered: ordered });
    });
  }
}

class InsertListTool extends Tool {
  getClassNames () {
    return 'sc-insert-list-tool'
  }
  renderButton ($$) {
    let button = super.renderButton($$);
    return [ button ]
  }
  onClick () {
    this.executeCommand({
      context: this.context
    });
  }
}

var ListPackage = {
  name: 'list',
  configure: function(config, {toolGroup, disableCollapsedCursor}) {
    config.addNode(ListNode);
    config.addNode(ListItem);
    config.addComponent('list', ListComponent);

    config.addCommand('insert-unordered-list', InsertListCommand, {
      nodeType: 'list',
      ordered: false,
      disableCollapsedCursor
    });
    config.addTool('insert-unordered-list', InsertListTool, { toolGroup });
    config.addLabel('insert-unordered-list', {
      en: 'Unordered list',
      de: 'Aufzählung'
    });
    config.addIcon('insert-unordered-list', { 'fontawesome': 'fa-list-ul' });

    config.addCommand('insert-ordered-list', InsertListCommand, {
      nodeType: 'list',
      ordered: true,
      disableCollapsedCursor
    });
    config.addTool('insert-ordered-list', InsertListTool, { toolGroup });
    config.addLabel('insert-ordered-list', {
      en: 'Ordered list',
      de: 'Nummerierte Liste'
    });
    config.addIcon('insert-ordered-list', { 'fontawesome': 'fa-list-ol' });

    config.addConverter('html', ListHTMLConverter);
    config.addConverter('html', ListItemHTMLConverter);
  }
};

class Paragraph extends TextBlock {}

Paragraph.type = "paragraph";

class ParagraphComponent extends TextBlockComponent {
  render($$) {
    let el = super.render($$);
    return el.addClass('sc-paragraph')
  }

  getTagName() {
    return 'p'
  }
}

var ParagraphHTMLConverter = {

  type: 'paragraph',
  tagName: 'p',

  import: function(el, node, converter) {
    node.content = converter.annotatedText(el, [node.id, 'content']);
  },

  export: function(node, el, converter) {
    el.append(converter.annotatedText([node.id, 'content']));
  }

};

var ParagraphPackage = {
  name: 'paragraph',
  configure: function(config) {
    config.addNode(Paragraph);
    config.addComponent(Paragraph.type, ParagraphComponent);
    config.addConverter('html', ParagraphHTMLConverter);
    config.addConverter('xml', ParagraphHTMLConverter);
    config.addTextType({
      name: 'paragraph',
      data: {type: 'paragraph'}
    });
    config.addLabel('paragraph', {
      en: 'Paragraph',
      de: 'Paragraph'
    });
    config.addLabel('paragraph.content', {
      en: 'Paragraph',
      de: 'Paragraph'
    });
  },
  Paragraph: Paragraph,
  ParagraphComponent: ParagraphComponent,
  ParagraphHTMLConverter: ParagraphHTMLConverter
};

class SaveCommand extends Command {
  constructor() {
    super({ name: 'save' });
  }

  getCommandState(params) {
    let dirty = params.editorSession.hasUnsavedChanges();
    return {
      disabled: !dirty,
      active: false
    }
  }

  execute(params) {
    let editorSession = params.editorSession;
    editorSession.save();
    return {
      status: 'saving-process-started'
    }
  }
}

var PersistencePackage = {
  name: 'persistence',
  configure: function(config) {
    config.addCommand('save', SaveCommand);
    config.addTool('save', Tool, {toolGroup: 'document'});
    config.addIcon('save', { 'fontawesome': 'fa-save' });
    config.addLabel('save', {
      en: 'Save',
      de: 'Speichern'
    });
  },
  SaveCommand: SaveCommand
};

class SwitchTextTypeCommand extends Command {
  // Available text types on the surface
  getTextTypes(params) {
    let surface = params.surface;
    if (surface && surface.isContainerEditor()) {
      return surface.getTextTypes()
    } else {
      return []
    }
  }

  getTextType(params) {
    let textTypes = this.getTextTypes(params);
    return find(textTypes, function(t) {
      return t.name === params.textType
    })
  }

  // Search which textType matches the current node
  // E.g. {type: 'heading', level: 1} => heading1
  getCurrentTextType(params, node) {
    let textTypes = this.getTextTypes(params);
    let currentTextType;
    textTypes.forEach(function(textType) {
      let nodeProps = clone(textType.data);
      delete nodeProps.type;
      if (isMatch(node, nodeProps) && node.type === textType.data.type) {
        currentTextType = textType;
      }
    });
    return currentTextType
  }

  isDisabled({selection, surface}) {
    if (!surface || !surface.isEnabled() || selection.isNull()) {
      return true
    }
    // When command is configured to be disabled for collapsed cursor
    if (selection && this.config.disableCollapsedCursor && selection.isCollapsed()) {
      return true
    }
    return false
  }

  getCommandState(params) {
    let doc = params.editorSession.getDocument();
    let sel = params.selection;

    let node;
    let newState = {
      disabled: false,
      textTypes: this.getTextTypes(params)
    };
    // Set disabled when not a property selection
    if (this.isDisabled(params)) {
      newState.disabled = true;
    } else if (sel.isContainerSelection()) {
      newState.disabled = true;
      newState.currentTextType = {name: 'container-selection'};
    } else if (sel.isPropertySelection()) {
      let path = sel.getPath();
      node = doc.get(path[0]);
      // There are cases where path points to an already deleted node,
      // so we need to guard node
      if (node) {
        if (node.isText() && node.isBlock()) {
          newState.currentTextType = this.getCurrentTextType(params, node);
        }
        if (!newState.currentTextType) {
          // We 'abuse' the currentTextType field by providing a property
          // identifier that is translated into a name using an default label set.
          // E.g. this.getLabel('figure.caption') -> Figure Caption
          newState.currentTextType = {name: [node.type, path[1]].join('.')};
          newState.disabled = true;
        }
      }
    } else if (sel.isNodeSelection()) {
      node = doc.get(sel.getNodeId());
      newState.currentTextType = {name: node.type};
      newState.disabled = true;
    } else if (sel.isCustomSelection()) {
      newState.currentTextType = {name: 'custom'};
      newState.disabled = true;
    }
    return newState
  }

  /**
    Trigger a switchTextType transaction
   */
  execute(params) {
    let textType = this.getTextType(params);
    let nodeData = textType.data;
    let surface = params.surface;
    let editorSession = params.editorSession;
    if (!surface) {
      console.warn('No focused surface. Stopping command execution.');
      return
    }
    editorSession.transaction(function(tx) {
      return tx.switchTextType(nodeData)
    });
    return nodeData
  }
}

/**
  SwitchTextTypeTool. Implements the SurfaceTool API.

  @class
  @component
*/
class SwitchTextTypeTool extends Tool {
  constructor(...args) {
    super(...args);

    // cursor for keyboard navigation
    this._navIdx = -1;
  }

  // UI Specific parts
  // ----------------

  didMount(...args) {
    super.didMount(...args);
    this._focusToggle();
  }

  render($$) {
    let labelProvider = this.context.labelProvider;
    let textTypeName = 'No selection';

    if (this.props.currentTextType) {
      textTypeName = this.props.currentTextType.name;
    }
    let el = $$('div').addClass('sc-switch-text-type');

    let toggleButton = $$('button').ref('toggle')
      .addClass('se-toggle')
      .attr('title', labelProvider.getLabel('switch_text'))
      .append(labelProvider.getLabel(textTypeName))
      .on('click', this.toggleAvailableTextTypes);

    if (this.props.disabled || !this.props.currentTextType) {
      el.addClass('sm-disabled');
      toggleButton.attr('tabindex', -1);
    } else {
      toggleButton.attr('tabindex', 1);
    }

    el.append(toggleButton);

    if (this.state.open) {
      el.addClass('sm-open');

      // dropdown options
      let options = $$('div').addClass("se-options").ref('options');
      forEach(this.props.textTypes, function(textType) {
        let button = $$('button')
            .addClass('se-option sm-'+textType.name)
            .attr('data-type', textType.name)
            .append(labelProvider.getLabel(textType.name))
            .on('click', this.handleClick);
        options.append(button);
      }.bind(this));
      el.append(options);
      el.on('keydown', this.onKeydown);
    }

    return el
  }

  didUpdate() {
    this._focusToggle();
  }

  _focusToggle() {
    if (this.state.open) {
      this.refs.toggle.el.focus();
    }
  }

  executeCommand(textType) {
    this.context.commandManager.executeCommand(this.getCommandName(), {
      textType: textType
    });
  }

  getTextCommands() {
    let surface = this.getSurface();
    if (!this.textCommands && surface) {
      this.textCommands = surface.getTextCommands();
    }
    return this.textCommands || {}
  }

  handleClick(e) {
    e.preventDefault();
    // Modifies the tool's state so that state.open is undefined, which is nice
    // because it means the dropdown will be closed automatically
    this.executeCommand(e.currentTarget.dataset.type);
  }

  onKeydown(event) {
    let handled = false;
    switch (event.keyCode) {
      case keys$1.UP:
        this._nav(-1);
        handled = true;
        break
      case keys$1.DOWN:
        this._nav(1);
        handled = true;
        break
      case keys$1.ESCAPE:
        this.toggleDropdown();
        handled = true;
        break
      default:
        // nothing
    }
    if (handled) {
      event.preventDefault();
      event.stopPropagation();
    }
  }

  toggleAvailableTextTypes(e) {
    e.preventDefault();
    e.stopPropagation();
    if (this.props.disabled) return

    // HACK: This only updates the view state state.open is not set on the tool itself
    // That way the dropdown automatically closes when the selection changes
    this.toggleDropdown();
  }

  toggleDropdown() {
    // reset index for keyboard navigation
    this._navIdx = -1;
    this.extendState({
      open: !this.state.open
    });
  }

  _nav(step) {
    this._navIdx += step;
    this._navIdx = Math.max(0, this._navIdx);
    this._navIdx = Math.min(this._getOptionsCount()-1, this._navIdx);

    if (this._navIdx >= 0) {
      let option = this.refs.options.children[this._navIdx];
      option.focus();
    }
  }

  _getOptionsCount() {
    return this.refs.options.children.length
  }

}

SwitchTextTypeTool.command = 'switch-text-type';

var SwitchTextTypePackage = {
  name: 'switch-text-type',
  configure: function(config, options) {
    config.addToolGroup('text');
    config.addCommand('switch-text-type', SwitchTextTypeCommand);
    config.addTool('switch-text-type', SwitchTextTypeTool, {toolGroup: options.toolGroup || 'text'});
  }
};

class Strong extends PropertyAnnotation {}

Strong.type = "strong";
Strong.fragmentation = Fragmenter.ANY;

var StrongHTMLConverter = {

  type: "strong",
  tagName: "strong",

  matchElement: function(el) {
    return el.is("strong, b")
  }

};

class StrongComponent extends AnnotationComponent {
  getTagName() {
    return 'strong'
  }
}

var StrongPackage = {
  name: 'strong',
  configure: function(config, {toolGroup, disableCollapsedCursor}) {
    config.addNode(Strong);
    config.addConverter('html', StrongHTMLConverter);
    config.addConverter('xml', StrongHTMLConverter);
    config.addComponent('strong', StrongComponent);

    config.addCommand('strong', AnnotationCommand, {
      nodeType: 'strong',
      disableCollapsedCursor
    });
    config.addTool('strong', AnnotationTool, {
      toolGroup: toolGroup || 'annotations'
    });
    config.addIcon('strong', { 'fontawesome': 'fa-bold' });
    config.addLabel('strong', {
      en: 'Strong',
      de: 'Fett'
    });
    if (platform.isMac) {
      config.addKeyboardShortcut('cmd+b', { command: 'strong' });
    } else {
      config.addKeyboardShortcut('ctrl+b', { command: 'strong' });
    }
  },
  Strong,
  StrongComponent,
  StrongHTMLConverter
};

class Subscript extends PropertyAnnotation {}

Subscript.type = 'subscript';

// hint for rendering in presence of overlapping annotations
Subscript.fragmentation = Fragmenter.ANY;

var SubscriptHTMLConverter = {
  type: 'subscript',
  tagName: 'sub'
};

var SubscriptPackage = {
  name: 'subscript',
  configure: function(config, {toolGroup, disableCollapsedCursor}) {
    config.addNode(Subscript);
    config.addConverter('html', SubscriptHTMLConverter);
    config.addConverter('xml', SubscriptHTMLConverter);
    config.addComponent('subscript', AnnotationComponent);
    config.addCommand('subscript', AnnotationCommand, {
      nodeType: 'subscript',
      disableCollapsedCursor
    });
    config.addTool('subscript', AnnotationTool, {toolGroup: toolGroup || 'annotations'});
    config.addIcon('subscript', { 'fontawesome': 'fa-subscript' });
    config.addLabel('subscript', {
      en: 'Subscript',
      de: 'Tiefgestellt'
    });
  },
  Subscript,
  SubscriptHTMLConverter
};

class Superscript extends PropertyAnnotation {}

Superscript.type = 'superscript';

// hint for rendering in presence of overlapping annotations
Superscript.fragmentation = Fragmenter.ANY;

var SuperscriptHTMLConverter = {
  type: 'superscript',
  tagName: 'sup'
};

var SuperscriptPackage = {
  name: 'superscript',
  configure: function(config, {toolGroup, disableCollapsedCursor}) {
    config.addNode(Superscript);
    config.addConverter('html', SuperscriptHTMLConverter);
    config.addConverter('xml', SuperscriptHTMLConverter);
    config.addComponent('superscript', AnnotationComponent);
    config.addCommand('superscript', AnnotationCommand, {
      nodeType: 'superscript',
      disableCollapsedCursor
    });
    config.addTool('superscript', AnnotationTool, {toolGroup: toolGroup || 'annotations'});
    config.addIcon('superscript', { 'fontawesome': 'fa-superscript' });
    config.addLabel('superscript', {
      en: 'Superscript',
      de: 'Hochgestellt'
    });
  },
  Superscript,
  SuperscriptHTMLConverter
};

const LEFT_QUOTE = "\u201C";
const RIGHT_QUOTE = "\u201D";

class InsertQuoteMarkCommand extends Command {

  getCommandState(params, context) { // eslint-disable-line
    // TODO: maybe only enable for specific selections?
    // let enabled = params.selection.isPropertySelection()
    return {
      disabled: false
    }
  }

  execute(params, context) { // eslint-disable-line
    let editorSession = params.editorSession;
    let sel = editorSession.getSelection();
    let doc = editorSession.getDocument();
    if (sel.isPropertySelection()) {
      let nodeId = sel.start.getNodeId();
      let node = doc.get(nodeId);
      if (node.isText()) {
        let text = node.getText();
        let offset = sel.start.offset;
        let mark;
        if (offset === 0 || /\s/.exec(text.slice(offset-1, offset))) {
          mark = LEFT_QUOTE;
        } else {
          mark = RIGHT_QUOTE;
        }
        editorSession.transaction((tx) => {
          tx.insertText(mark);
        });
        return true
      }
    }
    return false
  }

}

var QuoteMarksPackage = {

  name: 'quote-marks',

  configure: function(config) {
    config.addCommand('insertQuoteMark', InsertQuoteMarkCommand);
    config.addKeyboardShortcut('"', { type: 'textinput', command: 'insertQuoteMark' });
  }

};

class Table extends BlockNode {

  getRowCount() {
    return this.cells.length
  }

  getColCount() {
    if (this.cells.length > 0) {
      return this.cells[0].length
    } else {
      return 0
    }
  }

  getCellAt(row, col) {
    let cellId = this.cells[row][col];
    if (cellId) {
      return this.document.get(cellId)
    }
  }

}

Table.schema = {
  type: 'table',
  cells: { type: ['array', 'array', 'id'], default: [], owned: true }
};

class TableCell extends TextNode {}

TableCell.schema = {
  type: 'table-cell',
  rowspan: { type: 'number', default: 0 },
  colspan: { type: 'number', default: 0 }
};

class TableCellComponent extends Component {
  render($$) {
    let node = this.props.node;
    let el = $$('td').addClass('sc-table-cell');
    el.append(
      $$(TextPropertyEditor, {
        path: node.getTextPath(),
        disabled: this.props.disabled
      }).ref('editor')
    );
    if (node.rowspan > 0) {
      el.attr('rowspan', node.rowspan);
    }
    if (node.colspan > 0) {
      el.attr('colspan', node.colspan);
    }
    return el
  }

  grabFocus() {
    let node = this.props.node;
    this.context.editorSession.setSelection({
      type: 'property',
      path: node.getPath(),
      startOffset: node.getLength(),
      surfaceId: this.refs.editor.id
    });
  }
}

TableCellComponent.prototype._isTableCellComponent = true;

class TableComponent extends Component {

  render($$) {
    let el = $$('table').addClass('sc-table');
    let node = this.props.node;
    let doc = this.props.node.getDocument();
    let cells = this.props.node.cells;
    let rowCount = node.getRowCount();
    let colCount = node.getColCount();
    for (let i = 0; i < rowCount; i++) {
      let rowEl = $$('tr');
      for (let j = 0; j < colCount; j++) {
        let cellId = cells[i][j];
        // Merged cells (cellId is null) are skipped
        if (cellId) {
          let cellNode = doc.get(cellId);
          let cellEl = $$(TableCellComponent, {
            node: cellNode,
            disabled: this.props.disabled
          }).ref(cellNode.id);
          rowEl.append(cellEl);
        }
      }
      el.append(rowEl);
    }
    el.on('click', this.onClick);
    el.on('dblclick', this.onDblClick);
    return el
  }

  onClick(event) {
    event.stopPropagation();
    // console.log('Clicked on Table', this.props.node.id, event.target)
  }

  // TODO: this should only be used for the initial table state
  onDblClick(event) {
    event.stopPropagation();
    // console.log('DblClicked on Table', this.props.node.id, event.target)

    // HACK: assuming that if the event.target has a surface
    // it is a TextPropertyEditor of a cell
    let comp = Component.unwrap(event.target);
    if (comp) {
      let cellComp;
      if (comp._isTableCellComponent) {
        cellComp = comp;
      } else if (comp._isTextPropertyEditor) {
        cellComp = comp.getParent();
      } else if (comp._isTextPropertyComponent) {
        cellComp = comp.getParent().getParent();
      } else {
        console.warn('TODO: find the right cell');
      }
      if (cellComp) {
        cellComp.grabFocus();
      }
    }
  }

  grabFocus() {
    let cellId = this.props.node.cells[0][0];
    if (cellId) {
      let comp = this.refs[cellId];
      comp.grabFocus();
    }
  }

}

TableComponent.hasDropzones = true;

var TableHTMLConverter = {

  type: 'table',
  tagName: 'table',

  /*
    WARNING: This includes a pretty naive implementation for considering
    rowspans and colspans.

    TODO: Create test suite for this converter
  */
  import: function(el, node, converter) {
    let trs = el.find('tbody').getChildren();
    let colCount = 0;
    let cells = [];
    let rowspans = []; // we remember active rowspans here
    for (let i = 0; i < trs.length; i++) {
      let tds = trs[i].getChildren();
      let row = [];
      colCount = Math.max(tds.length, colCount);
      for (let j = 0; j < tds.length; j++) {
        let td = tds[j];
        // if there is an open rowspan
        if (rowspans[j] > 1) {
          row.push(null);
          rowspans[j] -= 1; // count down until exhausted
        }
        let tableCell = converter.convertElement(td);
        row.push(tableCell.id);
        if (tableCell.rowspan > 1) {
          rowspans[j] = tableCell.rowspan;
        }
        if (tableCell.colspan > 1) {
          // Add null values for colspans
          times(tableCell.colspan - 1, () => {
            row.push(null);
          });
        }
      }
      cells.push(row);
    }
    node.cells = cells;
  },

  export: function(node, el, converter) {
    let $$ = converter.$$;
    let rowCount = node.getRowCount();
    let colCount = node.getColCount();
    for (let i = 0; i < rowCount; i++) {
      let rowEl = $$('tr');
      for (let j = 0; j < colCount; j++) {
        let cellId = node.cells[i][j];
        // Merged cells (cellId is null) are skipped
        if (cellId) {
          let cellEl = converter.convertNode(cellId);
          rowEl.append(cellEl);
        }
      }
      el.append(rowEl);
    }
    return el
  }
};

var TableCellHTMLConverter = {

  type: 'table-cell',
  tagName: 'td',

  import: function(el, node, converter) {
    node.content = converter.annotatedText(el, [node.id, 'content']);
    let colspan = el.attr('colspan');
    let rowspan = el.attr('rowspan');
    if (colspan) {
      node.colspan = Number(colspan);
    }
    if (rowspan) {
      node.rowspan = Number(rowspan);
    }
  },

  export: function(node, el, converter) {
    el.append(converter.annotatedText([node.id, 'content']));
    if (node.rowspan > 0) {
      el.attr('rowspan', node.rowspan);
    }
    if (node.colspan > 0) {
      el.attr('colspan', node.colspan);
    }
    return el
  }
};

class InsertTableCommand extends InsertNodeCommand {
  createNodeData(tx) {

    // row-1
    let a1 = tx.create({ id: uuid('table-cell'), type: 'table-cell', content: "A1" });
    let b1 = tx.create({ id: uuid('table-cell'), type: 'table-cell', content: "B1" });
    // row-2
    let a2 = tx.create({ id: uuid('table-cell'), type: 'table-cell', content: "A2" });
    let b2 = tx.create({ id: uuid('table-cell'), type: 'table-cell', content: "B2" });

    return {
      id: uuid('table'),
      type: 'table',
      // null values mark merged cells
      cells: [
        [a1.id, b1.id ],
        [a2.id, b2.id]
      ]
    }

  }
}

var TablePackage = {
  name: 'table',
  configure: function(config) {
    config.addNode(Table);
    config.addNode(TableCell);
    config.addComponent('table', TableComponent);
    config.addConverter('html', TableHTMLConverter);
    config.addConverter('html', TableCellHTMLConverter);
    config.addConverter('xml', TableHTMLConverter);
    config.addConverter('xml', TableCellHTMLConverter);
    config.addCommand('insert-table', InsertTableCommand);
    config.addTool('insert-table', Tool, { toolGroup: 'insert' });
    config.addIcon('insert-table', { 'fontawesome': 'fa-table' });
    config.addLabel('insert-table', {
      en: 'Insert Table',
      de: 'Tabelle einfügen'
    });
    config.addLabel('table', {
      en: 'Table',
      de: 'Tabelle'
    });
    config.addLabel('table-cell.content', {
      en: 'Cell',
      de: 'Zelle'
    });
  }
};

class ProseArticle extends Document {
  constructor(schema) {
    super(schema);
    this._initialize();
  }

  _initialize() {
    this.create({
      type: 'container',
      id: 'body',
      nodes: []
    });
  }

}

var ProseEditorPackage = {
  name: 'prose-editor',
  configure: function(config) {
    config.defineSchema({
      name: 'prose-article',
      ArticleClass: ProseArticle,
      defaultTextType: 'paragraph'
    });
    // SwitchTextType, Undo/Redo etc.
    config.import(BasePackage);
    config.import(SwitchTextTypePackage);
    config.import(ParagraphPackage);
    config.import(HeadingPackage);
    config.import(CodeblockPackage);
    config.import(BlockquotePackage);
    config.import(EmphasisPackage);
    config.import(StrongPackage);
    config.import(SubscriptPackage);
    config.import(SuperscriptPackage);
    config.import(CodePackage);
    config.import(LinkPackage);
    config.import(QuoteMarksPackage);
    config.import(ListPackage);
    config.import(TablePackage);
  }
};

class SpellError extends Marker {
  invalidate() {
    this.remove();
  }
}

SpellError.schema = {
  type: 'spell-error',
  suggestions: { type: ['object'], default: [] }
};

SpellError.scope = 'document';

class CorrectionTool extends Tool {

  render($$) {
    let node = this.props.node;
    let Button = this.getComponent('button');
    let el = $$('div').addClass('sc-correction-tool');

    if (node && node.suggestions.length > 0) {
      node.suggestions.forEach((s) => {
        el.append(
          $$(Button, {
            label: s,
            style: this.props.style
          }).attr('title', this.getLabel('open-link'))
            .attr('data-correction', s)
            .on('click', this._applyCorrection.bind(this, s))
        );
      });
    } else {
      el.append(
        $$(Button, {
          label: 'No suggestions',
          style: this.props.style,
          disabled: true
        })
      );
    }
    return el
  }

  _applyCorrection(suggestion) {
    let editorSession = this.context.editorSession;
    let node = this.props.node;
    editorSession.transaction((tx) => {
      let sel = tx.getSelection();
      tx.setSelection({
        type: 'property',
        path: node.start.path,
        startOffset: node.start.offset,
        endOffset: node.end.offset,
        containerId: sel.containerId
      });
      tx.insertText(suggestion);
    });
  }
}

/**
  Used for edit tools or property annotations (e.g. EditLinkTool)

  @class
*/
class SpellCheckCommand extends Command {

  getCommandState(params) {

    let state = params.selectionState;
    let markers = state.getMarkers();
    if (markers.length === 0) {
      return {
        disabled: true
      }
    }
    markers = markers.filter(function(m) {
      return m.type === 'spell-error'
    });

    if (markers.length > 0) {
      return {
        disabled: false,
        active: false,
        mode: null,
        node: markers[0]
      }
    } else {
      return {
        disabled: true
      }
    }
  }
}

var SpellCheckPackage = {
  name: 'spell-check',
  configure: function(config) {
    config.addNode(SpellError);
    config.addCommand('correction', SpellCheckCommand);
    config.addTool('correction', CorrectionTool, {toolGroup: 'context-menu-primary'});
  }
};

const {UndoCommand: UndoCommand$1, RedoCommand, SelectAllCommand: SelectAllCommand$1} = BasePackage;

// TODO: maybe AbstractEditor is too heavy?
// still we need to do almost the same, so that nothing
// from the parent is leaking through to the children
class TextInput extends AbstractEditor {

  constructor(parent, props = {}) {
    super(parent, _createEditorSession(props));

    this.doc = this.editorSession.getDocument();
  }

  render($$) {
    let el = $$(this._getTagName()).addClass('sc-text-input');
    el.append(
      $$(TextInputEditor, {
        path: ['input', 'content']
      }).ref('input')
        .on('enter', this._onEnter)
        .on('escape', this._onEscape)
    );
    return el
  }

  didMount() {
    // set the cursor at the end of the content
    this.refs.input.selectLast();
  }

  dispose() {
    super.dispose();

    this.doc.dispose();
    this.editorSession.dispose();
  }

  // this component manages itself
  shouldRerender() {
    return false
  }

  getContent() {
    return this.getDocument().getContent()
  }

  _getDocument() {
    return this.context.editorSession.getDocument()
  }

  _getTagName() {
    return this.props.tagName || 'div'
  }

  _onEnter(event) {
    event.stopPropagation();
    this.el.emit('confirm');
  }

  _onEscape(event) {
    event.stopPropagation();
    this.el.emit('cancel');
  }

}

function _createEditorSession(props) {
  let config = new Configurator();
  config.addNode(TextNode);
  config.addToolGroup('annotations');
  config.addToolGroup('default');
  config.addCommand('undo', UndoCommand$1);
  config.addCommand('redo', RedoCommand);
  config.addCommand('select-all', SelectAllCommand$1);
  if (platform.isMac) {
    config.addKeyboardShortcut('cmd+z', { command: 'undo' });
    config.addKeyboardShortcut('cmd+shift+z', { command: 'redo' });
    config.addKeyboardShortcut('cmd+a', { command: 'select-all' });
  } else {
    config.addKeyboardShortcut('ctrl+z', { command: 'undo' });
    config.addKeyboardShortcut('ctrl+shift+z', { command: 'redo' });
    config.addKeyboardShortcut('ctrl+a', { command: 'select-all' });
  }

  config.defineSchema({
    name: 'text-input',
    // FIXME: this does not make sense here
    // as we do not have a container model
    defaultTextType: 'text',
    // FIXME: the name 'ArticleClass' is not general enough
    // plus: the configurator does not fail when this is not specified
    ArticleClass: TextInputDocument,
  });
  if (props.package) {
    config.import(props.package);
  }
  let doc = config.createArticle();
  if (props.content) {
    doc.set(['input', 'content'], props.content);
  }
  let editorSession = new EditorSession(doc, {
    configurator: config
  });
  return {
    editorSession
  }
}

class TextInputDocument extends Document {
  constructor(...args) {
    super(...args);

    this.create({
      type: 'text',
      id: 'input',
      content: ''
    });
  }
  getContentNode() {
    return this.get('input')
  }
  getContent() {
    return this.getContentNode().getText()
  }
}

// TODO: would be good if there were some events triggered by
// Surfaces
class TextInputEditor extends TextPropertyEditor {

  onKeyDown(event) {
    let handled = false;
    if (event.keyCode === 27) {
      handled = true;
      this.el.emit('escape');
    }
    if (handled) {
      event.stopPropagation();
      event.preventDefault();
    } else {
      super.onKeyDown(event);
    }
  }

  selectLast() {
    const doc = this.getDocument();
    const input = doc.getContentNode();
    this.editorSession.setSelection({
      type: 'property',
      path: input.getTextPath(),
      startOffset: input.getLength(),
      surfaceId: this.id
    });
  }

  selectAll() {
    const doc = this.getDocument();
    const input = doc.getContentNode();
    this.editorSession.setSelection({
      type: 'property',
      path: input.getTextPath(),
      startOffset: 0,
      endffset: input.getLength(),
      surfaceId: this.id
    });
  }

  _handleEnterKey(...args) {
    super._handleEnterKey(...args);
    this.el.emit('enter');
  }
}

var TextInputPackage = {
  name: 'text-input',
  configure(config) {
    config.addComponent('text-input', TextInput);
  },
  TextInput
};

class Toolbar extends Toolbox {

  render($$) {
    let el = $$('div').addClass(this.getClassNames());
    let activeToolGroups = this.state.activeToolGroups;

    activeToolGroups.forEach((toolGroup) => {
      let toolGroupProps = Object.assign({}, toolGroup, {
        toolStyle: this.getToolStyle(),
        layout: 'horizontal',
        showIcons: true
      });
      el.append(
        $$(toolGroup.Class, toolGroupProps)
      );
    });
    return el
  }

  getActiveToolGroupNames() {
    return this.props.toolGroups || ['text', 'document', 'annotations', 'default']
  }

  getClassNames() {
    return 'sc-toolbar';
  }

  getToolStyle() {
    return 'outline'
  }

  showDisabled() {
    return true
  }
}

class ToolDropdown extends Component {

  render($$) {
    let tools = this.props.tools;
    let commandStates = this.props.commandStates;
    let el = $$('div').addClass('sc-tool-dropdown');
    el.addClass('sm-target-'+this.props.name);
    el.append(
      this.renderButton($$)
    );
    if (this.state.open) {
      let optionEls = [];
      tools.forEach(function(tool, name) {
        let toolProps = Object.assign({}, commandStates[name]);
        toolProps.name = name;
        toolProps.label = name;
        toolProps.style = 'plain-dark'; // plain button style on dark bg will be used
        optionEls.push($$(tool.Class, toolProps));
      });
      el.append(
        $$('div').addClass('se-options').append(
          $$('div').addClass('se-arrow'),
          $$('div').addClass('se-content').append(
            optionEls
          )
        )
      );
    }
    return el
  }

  renderButton($$) {
    let btn = $$(Button, {
      label: this.props.name,
      active: this.state.open,
      disabled: this.props.disabled,
      style: this.props.style
    }).on('click', this.onClick);
    return btn
  }

  onClick() {
    let open = !this.state.open;
    this.setState({
      open: open
    });
  }
}

var ToolbarPackage = {
  name: 'toolbar',
  configure(config) {
    config.addComponent('toolbar', Toolbar);
    config.addComponent('tool-group', ToolGroup);
    config.addComponent('tool-dropdown', ToolDropdown);
  },
  Toolbar,
  ToolGroup,
  ToolDropdown
};

class TestEditor extends AbstractEditor {

  constructor(...args) {
    super(...args);
    this.handleActions({
      domSelectionRendered: function() {}
    });
  }

  render($$) {
    let doc = this.editorSession.getDocument();
    let el = $$('div');
    let body = $$(ContainerEditor, {
      node: doc.get('body')
    }).ref('surface');
    el.append(body);
    return el
  }
}

class TestNode extends DocumentNode {}

TestNode.type = 'test-node';

TestNode.schema = {
  boolVal: { type: "boolean", default: false },
  stringVal: { type: "string", default: "" },
  arrayVal: { type: ["array","string"], default: [] },
  objectVal: { type: "object", default: {} },
};

TestNode.isBlock = true;

class TestBlockNode extends BlockNode {}
TestBlockNode.type = 'test-block';

class TestContainerAnnotation extends ContainerAnnotation {}

TestContainerAnnotation.type = 'test-container-anno';

class StructuredNode extends DocumentNode {}

StructuredNode.schema = {
  type: "structured-node",
  title: "text",
  body: "text",
  caption: "text"
};

StructuredNode.isBlock = true;

class TestStructuredNodeComponent extends Component {

  render($$) {
    var node = this.props.node;
    var el = $$('div').addClass('sc-structured-node');
    el.append(
      $$(TextPropertyEditor, {
        disabled: this.props.disabled,
        path: [node.id, 'title']
      }).ref('titleEditor')
    );
    el.append(
      $$(TextPropertyEditor, {
        disabled: this.props.disabled,
        path: [node.id, 'body']
      }).ref('bodyEditor')
    );
    el.append(
      $$(TextPropertyEditor, {
        disabled: this.props.disabled,
        path: [node.id, 'caption']
      }).ref('captionEditor')
    );
    return el
  }
}

TestStructuredNodeComponent.fullWidth = true;

class TestInlineNode extends InlineNode {}

TestInlineNode.type = 'test-inline-node';

TestInlineNode.schema = {
  content: 'text'
};

class TestInlineNodeComponent extends InlineNodeComponent {
  didMount() {
    super.didMount();
    this.context.editorSession.onRender('document', this.rerender, this, {
      path: [this.props.node.id, 'content']
    });
  }

  dispose() {
    super.dispose();
    this.context.editorSession.off(this);
  }

  getClassNames() {
    // ATTENTION: ATM it is necessary to add .sc-inline-node
    return 'sc-test-inline sc-inline-node'
  }

  renderContent($$) {
    return $$('span').append(this.props.node.content)
  }
}

class TestContainerComponent extends Component {

  render($$) {
    var el = $$('div').addClass('sc-container');
    el.append(
      $$(ContainerEditor, {
        node: this.props.node
      }).ref('editor')
    );
    return el
  }
}

TestContainerComponent.fullWidth = true;

class TestMetaNode extends DocumentNode {}

TestMetaNode.schema = {
  type: 'meta',
  title: 'text'
};

var TestMetaNodeXMLConverter = {

  tagName: 'meta',
  type: 'meta',

  import: function(el, node, converter) {
    node.id = 'meta';
    var titleEl = el.find('title');
    if (titleEl) {
      node.title = converter.annotatedText(titleEl, ['meta', 'title']);
    } else {
      node.title = '';
    }
  },

  export: function(node, el, converter) {
    el.append(el.createElement('title').append(
      converter.annotatedText(['meta', 'title'])
    ));
  }

};

class TestHTMLImporter extends HTMLImporter {

  convertDocument(documentEl) {
    this.state.doc = this.createDocument();
    var bodyEl = documentEl.find('body');
    this.convertContainer(bodyEl.children, 'body');
  }

}

class TestXMLImporter extends XMLImporter {

  convertDocument(documentEl) {
    this.state.doc = this.createDocument();
    var bodyEl = documentEl.find('body');
    this.convertContainer(bodyEl.children, 'body');
  }

}

class TestHTMLExporter extends HTMLExporter {

  convertDocument(doc) {
    let el = this.createElement('div');
    el.append(
      this.convertContainer(doc.get('body'))
    );
    return el
  }

}

class TestXMLExporter extends XMLExporter {

  convertDocument(doc) {
    var articleEl = this.createElement('article');
    var body = doc.get('body');
    articleEl.append(
      this.convertContainer(body)
    );
    return articleEl
  }

}

class TestArticle extends Document {

  constructor(schema) {
    super(schema);
    this.create({
      type: "container",
      id: "body",
      nodes: []
    });
  }

}

const { InlineWrapper: InlineWrapper$1, InlineWrapperComponent: InlineWrapperComponent$1 } = InlineWrapperPackage;

function getTestConfig() {
  let config = new Configurator();
  config.addToolGroup('default');
  config.addToolGroup('annotations');
  config.addToolGroup('overlay');
  config.addToolGroup('insert');
  config.defineSchema({
    name: 'test-article',
    DocumentClass: TestArticle,
    defaultTextType: 'paragraph',
    version: 1.0
  });

  config.import(ParagraphPackage);
  config.import(HeadingPackage);
  config.import(StrongPackage);
  config.import(EmphasisPackage);
  config.import(LinkPackage);
  config.import(ListPackage);
  config.import(TablePackage);
  config.import(CodeblockPackage);
  config.import(FilePackage);
  config.import(ImagePackage);

  config.addComponent('container', TestContainerComponent);

  config.addNode(TestNode);

  config.addNode(TestBlockNode);
  config.addComponent(TestBlockNode.type, Component);

  config.addNode(TestContainerAnnotation);

  config.addNode(TestInlineNode);
  config.addComponent(TestInlineNode.type, TestInlineNodeComponent);

  config.addNode(StructuredNode);
  config.addComponent(StructuredNode.type, TestStructuredNodeComponent);

  config.addNode(TestMetaNode);
  config.addConverter('xml', TestMetaNodeXMLConverter);

  config.addNode(InlineWrapper$1);
  config.addComponent(InlineWrapper$1.type, InlineWrapperComponent$1);


  config.addImporter('html', TestHTMLImporter);
  config.addImporter('xml', TestXMLImporter);
  config.addExporter('html', TestHTMLExporter);
  config.addExporter('xml', TestXMLExporter);

  return config
}

function createEditorSession(...seeds) {
  let config = getTestConfig();
  let doc = new Document(config.getSchema());
  let body = doc.create({
    type: 'container',
    id: 'body'
  });
  seeds.forEach((seed)=>{
    seed(doc, body);
  });
  let editorSession = new EditorSession(doc, { configurator: config });
  let first = body.getNodeAt(0);
  if (first) {
    if (first.isText()) {
      editorSession.setSelection({
        type: 'property',
        path: first.getTextPath(),
        startOffset: 0,
        containerId: 'body',
        surfaceId: 'body'
      });
    } else if (first.isList()) {
      editorSession.setSelection({
        type: 'property',
        path: first.getItemAt(0).getTextPath(),
        startOffset: 0,
        containerId: 'body',
        surfaceId: 'body'
      });
    } else {
      editorSession.setSelection({
        type: 'node',
        nodeId: first.id,
        mode: 'before',
        containerId: 'body',
        surfaceId: 'body'
      });
    }
  }
  return editorSession
}

function getMountPoint(t) {
  let mountPoint;
  if (platform.inBrowser) {
    mountPoint = t.sandbox;
  } else {
    let htmlDoc = DefaultDOMElement.createDocument('html');
    mountPoint = htmlDoc.createElement('div');
    htmlDoc.appendChild(mountPoint);
  }
  return mountPoint
}

function setupEditor(t, ...f) {
  let editor = TestEditor.mount({ editorSession: createEditorSession(...f) }, getMountPoint(t));
  let editorSession = editor.editorSession;
  let doc = editorSession.getDocument();
  let surface = editor.refs.surface;
  return { editor, editorSession, doc, surface }
}

var containerAnnoSample = function(tx) {
  let body = tx.get('body');
  tx.create({
    type: 'paragraph',
    id: 'p1',
    content: '0123456789'
  });
  tx.create({
    type: 'paragraph',
    id: 'p2',
    content: '0123456789'
  });
  tx.create({
    type: 'paragraph',
    id: 'p3',
    content: '0123456789'
  });
  tx.create({
    type: 'paragraph',
    id: 'p4',
    content: '0123456789'
  });
  tx.create({
    type: 'test-container-anno',
    id: 'a1',
    start: {
      path: ['p1', 'content'],
      offset: 5,
    },
    end: {
      path: ['p3', 'content'],
      offset: 4
    },
    containerId: 'body'
  });
  tx.create({
    type: 'strong',
    id: 'a2',
    start: {
      path: ['p1', 'content'],
      offset: 0,
    },
    end: {
      offset: 2
    }
  });
  body.show('p1');
  body.show('p2');
  body.show('p3');
  body.show('p4');
  return tx
};

const test = substanceTest.module('AnnotationCommand');

test("can 'create' property annotation", function(t) {
  let { doc } = fixture(t);
  let selectionState = new SelectionState(doc);
  let cmd = new ToggleStrongCommand();
  let sel = doc.createSelection({
    type: 'property',
    path: ['p4', 'content'],
    startOffset: 1,
    endOffset: 6
  });
  selectionState.setSelection(sel);
  let cmdState = cmd.getCommandState({
    selectionState: selectionState
  });
  t.equal(cmdState.mode, 'create', "Mode should be correct.");
  t.end();
});

test("execute 'create' property annotation", function(t) {
  let { editorSession, doc } = fixture(t);
  let cmd = new ToggleStrongCommand();
  editorSession.setSelection({
    type: 'property',
    path: ['p4', 'content'],
    startOffset: 1,
    endOffset: 6
  });
  let res = cmd.execute({
    commandState: {
      mode: 'create'
    },
    editorSession: editorSession,
    selectionState: editorSession.getSelectionState()
  });
  let newAnno = res.anno;
  t.notNil(newAnno, 'A new anno should have been created');
  newAnno = doc.get(newAnno.id);
  t.equal(newAnno.type, 'strong', '.. of correct type');
  t.deepEqual(newAnno.start.path, ['p4', 'content'], ".. with correct path");
  t.equal(newAnno.start.offset, 1, '.. with correct startOffset');
  t.equal(newAnno.end.offset, 6, '.. with correct endOffset');
  t.end();
});

test("can 'delete' property annotation", function(t) {
  let { doc } = fixture(t);
  let selectionState = new SelectionState(doc);
  let cmd = new ToggleStrongCommand();
  let sel = doc.createSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 5,
    endOffset: 7
  });
  selectionState.setSelection(sel);
  let cmdState = cmd.getCommandState({
    selectionState: selectionState
  });
  t.equal(cmdState.mode, 'delete', "Mode should be correct.");
  t.end();
});

test("execute 'delete' property annotation", function(t) {
  let { doc } = fixture(t);
  let selectionState = new SelectionState(doc);
  let cmd = new ToggleStrongCommand();
  let sel = doc.createSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 5,
    endOffset: 7
  });
  selectionState.setSelection(sel);
  let cmdState = cmd.getCommandState({
    selectionState: selectionState
  });
  t.equal(cmdState.mode, 'delete', "Mode should be correct.");
  t.end();
});

test("can 'expand' property annotation", function(t) {
  let { editorSession, doc } = fixture(t);
  let cmd = new ToggleStrongCommand();
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 5,
    endOffset: 7
  });
  cmd.execute({
    commandState: {
      mode: 'delete'
    },
    editorSession: editorSession,
    selectionState: editorSession.getSelectionState()
  });
  t.isNil(doc.get('a3'), 'a3 should be gone.');
  t.end();
});

class ToggleStrongCommand extends AnnotationCommand {
  constructor() {
    super({ name: 'strong', nodeType: 'strong' });
  }
}

function fixture(t) {
  return setupEditor(t, containerAnnoSample, (tx) => {
    tx.create({
      id: 'a3',
      type: 'strong',
      start: { path: ['p1', 'content'], offset: 4 },
      end: { offset: 8 }
    });
  })
}

function createTestArticle(seedFn) {
  let config = getTestConfig();
  var doc = new TestArticle(config.getSchema());
  if (seedFn) {
    seedFn(doc);
  }
  return doc
}

/*
  Creates a document with the following content

  ```
  body:
    p1: '0123456789'
    p2: '0123456789'
    p3: '0123456789'
    p4: '0123456789'
  ```
*/
function simple(doc) {
  let body = doc.get('body');
  doc.create({
    type: 'paragraph',
    id: 'p1',
    content: '0123456789'
  });
  body.show('p1');
  doc.create({
    type: 'paragraph',
    id: 'p2',
    content: '0123456789'
  });
  body.show('p2');
  doc.create({
    type: 'paragraph',
    id: 'p3',
    content: '0123456789'
  });
  body.show('p3');
  doc.create({
    type: 'paragraph',
    id: 'p4',
    content: '0123456789'
  });
  body.show('p4');
}

const test$1 = substanceTest.module('annotationHelpers');

let truncateAnnotation$1 = annotationHelpers.truncateAnnotation;
let expandAnnotation$1 = annotationHelpers.expandAnnotation;
let fuseAnnotation$1 = annotationHelpers.fuseAnnotation;

test$1("Truncate property annotation with a given property selection", function(t) {
  let doc = fixture$1(A1);
  // Put cursor inside an the existing annotation
  let sel = doc.createSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 1,
    endOffset: 2
  });
  let anno = doc.get('a1');
  truncateAnnotation$1(doc, anno, sel);
  t.equal(anno.start.offset, 0, 'startOffset should be 0');
  t.equal(anno.end.offset, 1, 'endOffset should have changed from 2 to 1');
  t.end();
});

test$1("Truncate container annotation with a given property selection", function(t) {
  let doc = fixture$1(CA1);
  let sel = doc.createSelection({
    type: 'property',
    path: ['p3', 'content'],
    startOffset: 1,
    endOffset: 4
  });
  let anno = doc.get('ca1');
  truncateAnnotation$1(doc, anno, sel);
  t.equal(anno.end.offset, 1, 'endOffset should be 1');
  t.end();
});

test$1("Truncate container annotation with a given container selection", function(t) {
  let doc = fixture$1(CA1);
  let sel = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p2', 'content'],
    startOffset: 1,
    endPath: ['p3', 'content'],
    endOffset: 4,
  });
  let anno = doc.get('ca1');
  truncateAnnotation$1(doc, anno, sel);
  t.deepEqual(anno.end.path, ['p2', 'content'], 'endPath should be p2.content');
  t.equal(anno.end.offset, 1, 'endOffset should be 1');
  t.end();
});

test$1("Expand-right of property annotation for a given property selection", function(t) {
  let doc = fixture$1(A1);
  let sel = doc.createSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 1,
    endOffset: 6
  });
  let anno = doc.get('a1');
  expandAnnotation$1(doc, anno, sel);
  t.equal(anno.start.offset, 0, 'startOffset should be 0');
  t.equal(anno.end.offset, 6, 'endOffset should have changed from 2 to 1');
  t.end();
});

test$1("Expand container annotation for a given property selection (right expansion)", function(t) {
  let doc = fixture$1(CA1);
  let sel = doc.createSelection({
    type: 'property',
    path: ['p3', 'content'],
    startOffset: 1,
    endOffset: 6
  });
  let anno = doc.get('ca1');
  expandAnnotation$1(doc, anno, sel);
  t.equal(anno.end.offset, 6, 'endOffset should be 6');
  t.end();
});

test$1("Expand container annotation for a given container selection (expand right)", function(t) {
  let doc = fixture$1(CA1);
  let sel = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p2', 'content'],
    startOffset: 1,
    endPath: ['p3', 'content'],
    endOffset: 6,
  });
  let anno = doc.get('ca1');
  expandAnnotation$1(doc, anno, sel);
  t.deepEqual(anno.end.path, ['p3', 'content'], 'endPath should be p2.content');
  t.equal(anno.end.offset, 6, 'endOffset should be 6');
  t.end();
});

test$1("Fuse two property annotations for a given property selection", function(t) {
  let tx = new EditingInterface(fixture$1(A1, A2));
  // Put selection so that it touches both strong annotations
  tx.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 1,
    endOffset: 6
  });
  let a1 = tx.get('a1');
  let a2 = tx.get('a2');
  fuseAnnotation$1(tx, [a1, a2]);
  t.isNil(tx.get('a2'), 'a2 should be gone.');
  t.equal(a1.start.offset, 0, 'startOffset should be 0');
  t.equal(a1.end.offset, 6, 'endOffset should be 6');
  t.end();
});

test$1("Fuse two conatiner annotations for a given property selection", function(t) {
  let tx = new EditingInterface(fixture$1(CA1, CA2));
  tx.setSelection({
    type: 'property',
    path: ['p3', 'content'],
    startOffset: 3,
    endOffset: 8
  });
  let ca1 = tx.get('ca1');
  let ca2 = tx.get('ca2');
  fuseAnnotation$1(tx, [ca1, ca2]);
  t.isNil(tx.get('ca2'), 'ca2 should be gone.');
  t.deepEqual(ca1.start.path, ['p1', 'content'], 'start pPath should be p1.content');
  t.equal(ca1.start.offset, 5, 'start offset should be 5');
  t.deepEqual(ca1.end.path, ['p4', 'content'], 'end path should be p1.content');
  t.equal(ca1.end.offset, 9, 'end offset should be 9');
  t.end();
});

function fixture$1(...fns) {
  let doc = createTestArticle(simple);
  fns.forEach((fn)=>{
    fn(doc);
  });
  return doc
}

function A1(doc) {
  doc.create({
    type: 'strong',
    id: 'a1',
    start: {
      path: ['p1', 'content'],
      offset: 0,
    },
    end: {
      offset: 2
    }
  });
}

function A2(doc) {
  doc.create({
    id: 'a2',
    type: 'strong',
    start: {
      path: ['p1', 'content'],
      offset: 4
    },
    end: {
      offset: 6
    }
  });
}

function CA1(doc) {
  doc.create({
    type: 'test-container-anno',
    id: 'ca1',
    start: {
      path: ['p1', 'content'],
      offset: 5,
    },
    end: {
      path: ['p3', 'content'],
      offset: 4
    },
    containerId: 'body'
  });
}

function CA2(doc) {
  doc.create({
    type: 'test-container-anno',
    id: 'ca2',
    containerId: 'body',
    start: {
      path: ['p3', 'content'],
      offset: 7,
    },
    end: {
      path: ['p4', 'content'],
      offset: 9,
    }
  });
}

const test$2 = substanceTest.module('ArrayOperation');

function checkArrayOperationTransform(t, a, b, input, expected) {
  let ops = ArrayOperation.transform(a, b);
  let output = ops[1].apply(a.apply(input.slice(0)));
  t.deepEqual(output, expected, `(b' o a)('${JSON.stringify(input)}') == '${JSON.stringify(expected)}' with a=${a.toString()}, b'=${ops[1].toString()}`);
  output = ops[0].apply(b.apply(input.slice(0)));
  t.deepEqual(output, expected, `(a' o b)('${JSON.stringify(input)}') == '${JSON.stringify(expected)}' with b=${b.toString()}, a'=${ops[0].toString()}`);
}

test$2("Insert element", (t) => {
  let arr = [1,2,4];
  let expected = [1,2,3,4];
  let op = ArrayOperation.Insert(2, 3);
  op.apply(arr);
  t.deepEqual(arr, expected, 'Should insert element.');
  t.end();
});

test$2("Insert element after last position", (t) => {
  let arr = [1,2,3];
  let expected = [1,2,3,4];
  let op = ArrayOperation.Insert(arr.length, 4);
  op.apply(arr);
  t.deepEqual(arr, expected, 'Should append element.');
  t.end();
});

test$2("Delete element", (t) => {
  let arr = [1,2,3];
  let expected = [1,3];
  let op = ArrayOperation.Delete(1, 2);
  op.apply(arr);
  t.deepEqual(arr, expected, 'Should delete element.');
  t.end();
});

test$2("Create operation with invalid data", (t) => {
  t.throws(function() {
    new ArrayOperation();
  }, "Should throw if no data given.");
  t.throws(function() {
    new ArrayOperation({ type: 'foo' });
  }, "Should throw for invalid type.");
  t.throws(function() {
    new ArrayOperation({ type: ArrayOperation.INSERT, value: 1});
  }, "Should throw for missing position.");
  t.throws(function() {
    new ArrayOperation({ type: ArrayOperation.INSERT, pos: -1, value: 1});
  }, "Should throw for position < 0.");
  t.end();
});

test$2("Operation can be NOP", (t) => {
  let op = new ArrayOperation({type: ArrayOperation.NOP});
  t.ok(op.isNOP(), 'Operation should be NOP');
  t.end();
});

test$2("Apply operation on too short array.", (t) => {
  let arr = [1,2,3];
  let op = ArrayOperation.Insert(4, 5);
  t.throws(function() {
    op.apply(arr);
  }, "Should throw if applying insert operation out-of-bounds");
  op = ArrayOperation.Delete(4, 5);
  t.throws(function() {
    op.apply(arr);
  }, "Should throw if applying delete operation out-of-bounds");
  t.end();
});

// Note: it is better to fail in such cases, as this is an indicator for other greater problems.
test$2("Apply delete operation on wrong array.", (t) => {
  let arr = [1,2,3];
  let op = ArrayOperation.Delete(2, 4);
  t.throws(function() {
    op.apply(arr);
  }, "Should throw if applying delete operation with wrong value");
  t.end();
});

test$2("JSON de-/serialisation", (t) => {
  let op = ArrayOperation.Delete(1,2);
  let out = op.toJSON();
  t.equal(out.type, ArrayOperation.DELETE);
  t.equal(out.pos, 1);
  t.equal(out.val, 2);
  op = ArrayOperation.fromJSON(out);
  t.ok(op.isDelete());
  t.equal(op.getOffset(), 1);
  t.equal(op.getValue(), 2);
  op = new ArrayOperation({type: ArrayOperation.NOP});
  out = op.toJSON();
  t.deepEqual(out, {type: ArrayOperation.NOP});
  t.end();
});

// Insert-Insert Transformations
// --------
// Cases:
//  1. `a < b`:   operations should not be affected
//  2. `b < a`:   dito
//  3. `a == b`:  result depends on preference (first applied)

test$2("Transformation: a=Insert, b=Insert, a < b and b < a", (t) => {
  let input = [1,3,5];
  let expected = [1,2,3,4,5];
  let a = ArrayOperation.Insert(1, 2);
  let b = ArrayOperation.Insert(2, 4);
  checkArrayOperationTransform(t, a, b, input, expected);
  checkArrayOperationTransform(t, b, a, input, expected);
  t.end();
});

// Example:
//     A = [1,4], a = [+, 1, 2], b = [+, 1, 3]
//     A  - a ->  [1, 2, 4]   - b' ->   [1,2,3,4]     => b'= [+, 2, 3], transform(a, b) = [a, b']
//     A  - b ->  [1, 3, 4]   - a' ->   [1,3,2,4]     => a'= [+, 2, 2], transform(b, a) = [a', b]
test$2("Transformation: a=Insert, b=Insert, a == b", (t) => {
  let input = [1,4];
  let expected = [1,2,3,4];
  let expected_2 = [1,3,2,4];
  let a = ArrayOperation.Insert(1, 2);
  let b = ArrayOperation.Insert(1, 3);
  checkArrayOperationTransform(t, a, b, input, expected);
  checkArrayOperationTransform(t, b, a, input, expected_2);
  t.end();
});

// Delete-Delete Transformations
// --------
// Cases:
//  1. `a < b`:   operations should not be affected
//  2. `b < a`:   dito
//  3. `a == b`:  second operation should not have an effect
//                user should be noticed about conflict

test$2("Transformation: a=Delete, b=Delete (1,2), a < b and b < a", (t) => {
  let input = [1,2,3,4,5];
  let expected = [1,3,5];
  let a = ArrayOperation.Delete(1, 2);
  let b = ArrayOperation.Delete(3, 4);
  checkArrayOperationTransform(t, a, b, input, expected);
  checkArrayOperationTransform(t, b, a, input, expected);
  t.end();
});

test$2("Transformation: a=Delete, b=Delete (3), a == b", (t) => {
  let input = [1,2,3];
  let expected = [1,3];
  let a = ArrayOperation.Delete(1, 2);
  let b = ArrayOperation.Delete(1, 2);
  checkArrayOperationTransform(t, a, b, input, expected);
  checkArrayOperationTransform(t, b, a, input, expected);
  t.end();
});

// Insert-Delete Transformations
// --------
// Cases: (a = insertion, b = deletion)
//  1. `a < b`:   b must be shifted right
//  2. `b < a`:   a must be shifted left
//  3. `a == b`:  ???


// A = [1,3,4,5], a = [+, 1, 2], b = [-, 2, 4]
// A  - a ->  [1,2,3,4,5] - b' ->   [1,2,3,5]     => b'= [-, 3, 4]
// A  - b ->  [1,3,5]     - a' ->   [1,2,3,5]     => a'= [+, 1, 2] = a
test$2("Transformation: a=Insert, b=Delete (1), a < b", (t) => {
  let input = [1,3,4,5];
  let expected = [1,2,3,5];
  let a = ArrayOperation.Insert(1, 2);
  let b = ArrayOperation.Delete(2, 4);
  checkArrayOperationTransform(t, a, b, input, expected);
  checkArrayOperationTransform(t, b, a, input, expected);
  t.end();
});

// A = [1,2,3,5], a = [+,3,4], b = [-,1,2]
// A  - a ->  [1,2,3,4,5] - b' ->   [1,3,4,5]     => b'= [-,1,2] = b
// A  - b ->  [1,3,5]     - a' ->   [1,3,4,5]     => a'= [+,2,4]
test$2("Transformation: a=Insert, b=Delete (2), b < a", (t) => {
  let input = [1,2,3,5];
  let expected = [1,3,4,5];
  let a = ArrayOperation.Insert(3, 4);
  let b = ArrayOperation.Delete(1, 2);
  checkArrayOperationTransform(t, a, b, input, expected);
  checkArrayOperationTransform(t, b, a, input, expected);
  t.end();
});


// A = [1,2,3], a = [+,1,4], b = [-,1,2]
// A  - a ->  [1,4,2,3] - b' ->   [1,4,3]     => b'= [-,2,2]
// A  - b ->  [1,3]     - a' ->   [1,4,3]     => a'= [+,1,4] = a
test$2("Transformation: a=Insert, b=Delete (3), a == b", (t) => {
  let input = [1,2,3];
  let expected = [1,4,3];
  let a = ArrayOperation.Insert(1, 4);
  let b = ArrayOperation.Delete(1, 2);
  checkArrayOperationTransform(t, a, b, input, expected);
  checkArrayOperationTransform(t, b, a, input, expected);
  t.end();
});

test$2("Transformation: a=NOP || b=NOP", (t) => {
  let a = new ArrayOperation.Insert(1, 4);
  let b = new ArrayOperation({type: ArrayOperation.NOP});
  let tr = ArrayOperation.transform(a, b);
  t.deepEqual(tr[0].toJSON(), a.toJSON());
  t.deepEqual(tr[1].toJSON(), b.toJSON());
  tr = ArrayOperation.transform(b, a);
  t.deepEqual(tr[0].toJSON(), b.toJSON());
  t.deepEqual(tr[1].toJSON(), a.toJSON());
  t.end();
});

test$2("Inverting operations", (t) => {
  let op = ArrayOperation.Insert(1, 4);
  let inverse = op.invert();
  t.ok(inverse.isDelete(), 'Inverse of an insert op should be a delete op.');
  t.equal(inverse.getOffset(), op.getOffset(), 'Offset of inverted op should be the same.');
  t.equal(inverse.getValue(), op.getValue(), 'Value of inverted op should be the same.');
  op = ArrayOperation.Delete(2, 3);
  inverse = op.invert();
  t.ok(inverse.isInsert(), 'Inverse of a delete op should be an insert op.');
  t.equal(inverse.getOffset(), op.getOffset(), 'Offset of inverted op should be the same.');
  t.equal(inverse.getValue(), op.getValue(), 'Value of inverted op should be the same.');
  op = new ArrayOperation({type: ArrayOperation.NOP});
  inverse = op.invert();
  t.ok(inverse.isNOP(), 'Inverse of a nop is a nop.');
  t.end();
});

test$2("Transformations can be done inplace (optimzation for internal use)", (t) => {
  let a = ArrayOperation.Insert(2, 3);
  let b = ArrayOperation.Insert(2, 3);
  let tr = ArrayOperation.transform(a, b, {inplace: true});
  t.ok(a.getOffset() === tr[0].getOffset() && b.getOffset() === tr[1].getOffset(), "Transformation should be done inplace.");
  t.end();
});

test$2("With option 'no-conflict' conflicting operations can not be transformed.", (t) => {
  let a = ArrayOperation.Insert(2, 2);
  let b = ArrayOperation.Insert(2, 2);
  t.throws(function() {
    ArrayOperation.transform(a, b, { "no-conflict": true });
  }, 'Transforming conflicting ops should throw when option "no-conflict" is enabled.');
  t.end();
});

test$2("Conflicts: inserting at the same position", (t) => {
  // this is considered a conflict as a decision is needed to determine which element comes first.
  let a = ArrayOperation.Insert(2, 'a');
  let b = ArrayOperation.Insert(2, 'b');
  t.ok(a.hasConflict(b) && b.hasConflict(a), "Inserts at the same position are considered a conflict.");
  t.end();
});

test$2("Conflicts: inserting at different positions", (t) => {
  let a = ArrayOperation.Insert(2, 'a');
  let b = ArrayOperation.Insert(4, 'b');
  t.ok(!a.hasConflict(b) && !b.hasConflict(a), "Inserts at different positions should be fine.");
  t.end();
});

test$2("Conflicts: deleting at the same position", (t) => {
  // this is *not* considered a conflict as it is clear how the result should look like.
  let a = ArrayOperation.Delete(2, 'a');
  let b = ArrayOperation.Delete(2, 'a');
  t.ok(!a.hasConflict(b) && !b.hasConflict(a), "Deletes at the same position are not a conflict.");
  t.end();
});

test$2("Conflicts: inserting and deleting at the same position", (t) => {
  // this is *not* considered a conflict as it is clear how the result should look like.
  let a = ArrayOperation.Insert(2, 'a');
  let b = ArrayOperation.Delete(2, 'b');
  t.ok(!a.hasConflict(b) && !b.hasConflict(a), "Inserting and deleting at the same position is not a conflict.");
  t.end();
});

test$2("Conflicts: when NOP involved", (t) => {
  let a = ArrayOperation.Insert(2, 2);
  let b = new ArrayOperation({type: ArrayOperation.NOP});
  t.ok(!a.hasConflict(b) && !b.hasConflict(a), "NOPs should never conflict.");
  t.end();
});

var BrowserLinuxPLainTextFixture = "<meta http-equiv=\"content-type\" content=\"text/html; charset=utf-8\"><span style=\"color: rgb(37, 37, 37); font-family: sans-serif; font-size: 14px; font-style: normal; font-variant: normal; font-weight: normal; letter-spacing: normal; line-height: 22.3999996185303px; orphans: auto; text-align: start; text-indent: 0px; text-transform: none; white-space: normal; widows: 1; word-spacing: 0px; -webkit-text-stroke-width: 0px; display: inline !important; float: none; background-color: rgb(255, 255, 255);\">XXX</span>";

var BrowserLinuxAnnotatedTextFixture = "<meta http-equiv=\"content-type\" content=\"text/html; charset=utf-8\"><span style=\"color: rgb(37, 37, 37); font-family: sans-serif; font-size: 14px; font-style: normal; font-variant: normal; font-weight: normal; letter-spacing: normal; line-height: 22.3999996185303px; orphans: auto; text-align: start; text-indent: 0px; text-transform: none; white-space: normal; widows: 1; word-spacing: 0px; -webkit-text-stroke-width: 0px; display: inline !important; float: none; background-color: rgb(255, 255, 255);\">X</span><a href=\"https://en.wikipedia.org/wiki/United_States_Senate_election_in_Illinois,_2004\" title=\"United States Senate election in Illinois, 2004\" style=\"text-decoration: none; color: rgb(11, 0, 128); font-family: sans-serif; font-size: 14px; font-style: normal; font-variant: normal; font-weight: normal; letter-spacing: normal; line-height: 22.3999996185303px; orphans: auto; text-align: start; text-indent: 0px; text-transform: none; white-space: normal; widows: 1; word-spacing: 0px; -webkit-text-stroke-width: 0px; background: none rgb(255, 255, 255);\">X</a><span style=\"color: rgb(37, 37, 37); font-family: sans-serif; font-size: 14px; font-style: normal; font-variant: normal; font-weight: normal; letter-spacing: normal; line-height: 22.3999996185303px; orphans: auto; text-align: start; text-indent: 0px; text-transform: none; white-space: normal; widows: 1; word-spacing: 0px; -webkit-text-stroke-width: 0px; display: inline !important; float: none; background-color: rgb(255, 255, 255);\">X</span>";

var BrowserLinuxTwoParagraphsFixture = "<meta http-equiv=\"content-type\" content=\"text/html; charset=utf-8\"><p>AAA</p><p>BBB</p>\r\n";

var BrowserWindowsPlainTextFixture = "<html>\r\n<body>\r\n<!--StartFragment--><span style=\"color: rgb(37, 37, 37); font-family: sans-serif; font-size: 14px; font-style: normal; font-variant: normal; font-weight: normal; letter-spacing: normal; line-height: 12.8px; orphans: auto; text-align: start; text-indent: 0px; text-transform: none; white-space: normal; widows: 1; word-spacing: 0px; -webkit-text-stroke-width: 0px; display: inline !important; float: none; background-color: rgb(255, 255, 255);\">XXX</span><!--EndFragment-->\r\n</body>\r\n</html>\r\n";

var BrowserWindowsAnnotatedTextFixture = "<html>\r\n<body>\r\n<!--StartFragment--><span style=\"color: rgb(37, 37, 37); font-family: sans-serif; font-size: 14px; font-style: normal; font-variant: normal; font-weight: normal; letter-spacing: normal; line-height: 12.8px; orphans: auto; text-align: start; text-indent: 0px; text-transform: none; white-space: normal; widows: 1; word-spacing: 0px; -webkit-text-stroke-width: 0px; display: inline !important; float: none; background-color: rgb(255, 255, 255);\">X</span></span><a href=\"https://en.wikipedia.org/wiki/United_States_Senate_election_in_Illinois,_2004\" title=\"United States Senate election in Illinois, 2004\" style=\"text-decoration: none; color: rgb(11, 0, 128); font-family: sans-serif; font-size: 14px; font-style: normal; font-variant: normal; font-weight: normal; letter-spacing: normal; line-height: 12.8px; orphans: auto; text-align: start; text-indent: 0px; text-transform: none; white-space: normal; widows: 1; word-spacing: 0px; -webkit-text-stroke-width: 0px; background: none rgb(255, 255, 255);\">X</a><span style=\"color: rgb(37, 37, 37); font-family: sans-serif; font-size: 14px; font-style: normal; font-variant: normal; font-weight: normal; letter-spacing: normal; line-height: 12.8px; orphans: auto; text-align: start; text-indent: 0px; text-transform: none; white-space: normal; widows: 1; word-spacing: 0px; -webkit-text-stroke-width: 0px; display: inline !important; float: none; background-color: rgb(255, 255, 255);\">X</span></span><!--EndFragment-->\r\n</body>\r\n</html>";

var BrowserWindowsTwoParagraphsFixture = "<html>\r\n<body>\r\n<!--StartFragment--><p>AAA</p><p>BBB</p><!--EndFragment-->\r\n</body>\r\n</html>\r\n";

var BrowserLinuxFirefoxPlainTextFixture = "XXX";

var BrowserLinuxFirefoxAnnotatedTextFixture = "X<a href=\"https://en.wikipedia.org/wiki/United_States_Senate_election_in_Illinois,_2004\" title=\"United States Senate election in Illinois, 2004\">X</a>X";

var BrowserLinuxFirefoxTwoParagraphsFixture = "<p>AAA</p><p>BBB</p>";

var BrowserLinuxFirefoxWholePageFixture = "    <div id=\"mw-page-base\" class=\"noprint\"></div>\r\n    <div id=\"mw-head-base\" class=\"noprint\"></div>\r\n    <div id=\"content\" class=\"mw-body\" role=\"main\">\r\n      <a id=\"top\"></a>\r\n\r\n              <div id=\"siteNotice\"><div id=\"centralNotice\"></div></div>\r\n            <div class=\"mw-indicators\">\r\n</div>\r\n      <h1 id=\"firstHeading\" class=\"firstHeading\" lang=\"en\">Synonym</h1>\r\n                  <div id=\"bodyContent\" class=\"mw-body-content\">\r\n                  <div id=\"siteSub\">From Wikipedia, the free encyclopedia</div>\r\n                <div id=\"contentSub\"></div>\r\n                        </div></div>\r\n        <div id=\"mw-content-text\" dir=\"ltr\" class=\"mw-content-ltr\" lang=\"en\"><div class=\"hatnote\">This article is about the general meaning of \"synonym\".  For its use in biology, see <a href=\"https://en.wikipedia.org/wiki/Synonym_%28taxonomy%29\" title=\"Synonym (taxonomy)\">Synonym (taxonomy)</a>.</div>\r\n<table class=\"metadata plainlinks ambox ambox-content ambox-Refimprove\" role=\"presentation\">\r\n<tbody><tr>\r\n<td class=\"mbox-image\">\r\n<div style=\"width:52px\"><a href=\"https://en.wikipedia.org/wiki/File:Question_book-new.svg\" class=\"image\"><img alt=\"\" src=\"https://upload.wikimedia.org/wikipedia/en/thumb/9/99/Question_book-new.svg/50px-Question_book-new.svg.png\" srcset=\"//upload.wikimedia.org/wikipedia/en/thumb/9/99/Question_book-new.svg/75px-Question_book-new.svg.png 1.5x, //upload.wikimedia.org/wikipedia/en/thumb/9/99/Question_book-new.svg/100px-Question_book-new.svg.png 2x\" data-file-width=\"262\" data-file-height=\"204\" height=\"39\" width=\"50\"></a></div>\r\n</td>\r\n<td class=\"mbox-text\"><span class=\"mbox-text-span\">This article <b>needs additional citations for <a href=\"https://en.wikipedia.org/wiki/Wikipedia:Verifiability\" title=\"Wikipedia:Verifiability\">verification</a></b>. <span class=\"hide-when-compact\">Please help <a class=\"external text\" href=\"https://en.wikipedia.org/w/index.php?title=Synonym&amp;action=edit\">improve this article</a> by <a href=\"https://en.wikipedia.org/wiki/Help:Introduction_to_referencing/1\" title=\"Help:Introduction to referencing/1\" class=\"mw-redirect\">adding citations to reliable sources</a>. Unsourced material may be challenged and removed.</span> <small><i>(May 2014)</i></small></span></td>\r\n</tr>\r\n</tbody></table>\r\n<div class=\"thumb tright\">\r\n<div class=\"thumbinner\" style=\"width:222px;\"><a href=\"https://en.wikipedia.org/wiki/File:Library_of_Ashurbanipal_synonym_list_tablet.jpg\" class=\"image\"><img alt=\"\" src=\"https://upload.wikimedia.org/wikipedia/commons/thumb/6/64/Library_of_Ashurb…onym_list_tablet.jpg/220px-Library_of_Ashurbanipal_synonym_list_tablet.jpg\" class=\"thumbimage\" srcset=\"//upload.wikimedia.org/wikipedia/commons/thumb/6/64/Library_of_Ashurbanipal_synonym_list_tablet.jpg/330px-Library_of_Ashurbanipal_synonym_list_tablet.jpg 1.5x, //upload.wikimedia.org/wikipedia/commons/thumb/6/64/Library_of_Ashurbanipal_synonym_list_tablet.jpg/440px-Library_of_Ashurbanipal_synonym_list_tablet.jpg 2x\" data-file-width=\"1829\" data-file-height=\"2676\" height=\"322\" width=\"220\"></a>\r\n<div class=\"thumbcaption\">\r\n<div class=\"magnify\"></div></div></div></div></div><div class=\"thumb tright\"><div class=\"thumbinner\" style=\"width:222px;\"><div class=\"thumbcaption\">\r\nSynonym list in <a href=\"https://en.wikipedia.org/wiki/Cuneiform\" title=\"Cuneiform\">cuneiform</a> on a clay tablet, <a href=\"https://en.wikipedia.org/wiki/Neo-Assyrian\" title=\"Neo-Assyrian\" class=\"mw-redirect\">Neo-Assyrian</a> period. Ref:<a rel=\"nofollow\" class=\"external text\" href=\"http://www.britishmuseum.org/research/search_the_collection_database/search_object_details.aspx?objectid=308401&amp;partid=1\">K.4375</a> .</div>\r\n</div>\r\n</div>\r\n<p>A <b>synonym</b> is a word or phrase that means exactly or nearly the\r\n same as another word or phrase in the same language. Words that are\r\nsynonyms are said to be <b>synonymous</b>, and the state of being a synonym is called <b>synonymy</b>. The word comes from <a href=\"https://en.wikipedia.org/wiki/Ancient_Greek_language\" title=\"Ancient Greek language\" class=\"mw-redirect\">Ancient Greek</a> <i>syn</i> (<span xml:lang=\"grc\" lang=\"grc\"><a href=\"https://en.wiktionary.org/wiki/%CF%83%CF%8D%CE%BD\" class=\"extiw\" title=\"wikt:σύν\">σύν</a></span>) (\"with\") and <i>onoma</i> (<span xml:lang=\"grc\" lang=\"grc\"><a href=\"https://en.wiktionary.org/wiki/%E1%BD%84%CE%BD%CE%BF%CE%BC%CE%B1\" class=\"extiw\" title=\"wikt:ὄνομα\">ὄνομα</a></span>) (\"name\"). An example of synonyms are the words <i>begin</i>, <i>start</i>, and <i>commence</i>. Words can be synonymous when meant in certain <a href=\"https://en.wikipedia.org/wiki/Word_sense\" title=\"Word sense\">senses</a>, even if they are not synonymous in all of their senses. For example, if we talk about a <i>long time</i> or an <i>extended time</i>, <i>long</i> and <i>extended</i> are synonymous within that <a href=\"https://en.wikipedia.org/wiki/Context_%28language_use%29\" title=\"Context (language use)\">context</a>. Synonyms with exact interchangeability share a <a href=\"https://en.wikipedia.org/wiki/Seme_%28semantics%29\" title=\"Seme (semantics)\">seme</a> or denotational <a href=\"https://en.wikipedia.org/wiki/Sememe\" title=\"Sememe\">sememe</a>, whereas those with inexactly similar meanings share a broader denotational or connotational <a href=\"https://en.wikipedia.org/wiki/Sememe\" title=\"Sememe\">sememe</a> and thus overlap within a <a href=\"https://en.wikipedia.org/wiki/Semantic_field\" title=\"Semantic field\">semantic field</a>. Some academics call the former type <a href=\"https://en.wikipedia.org/wiki/Cognitive_synonymy\" title=\"Cognitive synonymy\">cognitive synonyms</a> to distinguish them from the latter type, which they call near-synonyms.<sup id=\"cite_ref-Stanojevi.C4.87_2009_1-0\" class=\"reference\"><a href=\"https://en.wikipedia.org/wiki/Synonym#cite_note-Stanojevi.C4.87_2009-1\"><span>[</span>1<span>]</span></a></sup></p>\r\n<p>In the figurative sense, two words are sometimes said to be synonymous if they have the same <a href=\"https://en.wikipedia.org/wiki/Connotation\" title=\"Connotation\">connotation</a>:</p>\r\n<blockquote class=\"templatequote\">\r\n<p>...a widespread impression that ... <a href=\"https://en.wikipedia.org/wiki/Hollywood\" title=\"Hollywood\">Hollywood</a> was synonymous with immorality...<sup id=\"cite_ref-2\" class=\"reference\"><a href=\"https://en.wikipedia.org/wiki/Synonym#cite_note-2\"><span>[</span>2<span>]</span></a></sup></p>\r\n<div class=\"templatequotecite\"><cite>— <a href=\"https://en.wikipedia.org/wiki/Doris_Kearns_Goodwin\" title=\"Doris Kearns Goodwin\">Doris Kearns Goodwin</a></cite></div>\r\n</blockquote>\r\n<p><a href=\"https://en.wikipedia.org/wiki/Metonymy\" title=\"Metonymy\">Metonymy</a> can sometimes be a form of synonymy, as when, for example, <i>the <a href=\"https://en.wikipedia.org/wiki/White_House\" title=\"White House\">White House</a></i> is used as a synonym of <i>the administration</i> in referring to the U.S. <a href=\"https://en.wikipedia.org/wiki/Executive_%28government%29\" title=\"Executive (government)\">executive branch</a> under a specific president. Thus a metonym is a type of synonym, and the word <i>metonym</i> is a <a href=\"https://en.wikipedia.org/wiki/Hyponymy_and_hypernymy\" title=\"Hyponymy and hypernymy\">hyponym</a> of the word <i>synonym</i>.</p>\r\n<p>The analysis of synonymy, <a href=\"https://en.wikipedia.org/wiki/Polysemy\" title=\"Polysemy\">polysemy</a>, and <a href=\"https://en.wikipedia.org/wiki/Hyponymy_and_hypernymy\" title=\"Hyponymy and hypernymy\">hyponymy and hypernymy</a> is vital to <a href=\"https://en.wikipedia.org/wiki/Taxonomy_%28general%29\" title=\"Taxonomy (general)\">taxonomy</a> and <a href=\"https://en.wikipedia.org/wiki/Ontology_%28information_science%29\" title=\"Ontology (information science)\">ontology</a> in the <a href=\"https://en.wikipedia.org/wiki/Information_science\" title=\"Information science\">information-science</a> senses of those terms. It has applications in <a href=\"https://en.wikipedia.org/wiki/Pedagogy\" title=\"Pedagogy\">pedagogy</a> and <a href=\"https://en.wikipedia.org/wiki/Machine_learning\" title=\"Machine learning\">machine learning</a>, because they rely on <a href=\"https://en.wikipedia.org/wiki/Word-sense_disambiguation\" title=\"Word-sense disambiguation\">word-sense disambiguation</a> and <a href=\"https://en.wiktionary.org/wiki/schema#Noun\" class=\"extiw\" title=\"wikt:schema\">schema</a>.</p>\r\n<p></p>\r\n<div id=\"toc\" class=\"toc\">\r\n<div id=\"toctitle\">\r\n<h2>Contents</h2>\r\n</div></div><div id=\"toc\" class=\"toc\">\r\n<ul>\r\n<li class=\"toclevel-1 tocsection-1\"><a href=\"https://en.wikipedia.org/wiki/Synonym#Examples\"><span class=\"tocnumber\">1</span> <span class=\"toctext\">Examples</span></a></li>\r\n<li class=\"toclevel-1 tocsection-2\"><a href=\"https://en.wikipedia.org/wiki/Synonym#Related_terms\"><span class=\"tocnumber\">2</span> <span class=\"toctext\">Related terms</span></a></li>\r\n<li class=\"toclevel-1 tocsection-3\"><a href=\"https://en.wikipedia.org/wiki/Synonym#See_also\"><span class=\"tocnumber\">3</span> <span class=\"toctext\">See also</span></a></li>\r\n<li class=\"toclevel-1 tocsection-4\"><a href=\"https://en.wikipedia.org/wiki/Synonym#References\"><span class=\"tocnumber\">4</span> <span class=\"toctext\">References</span></a></li>\r\n<li class=\"toclevel-1 tocsection-5\"><a href=\"https://en.wikipedia.org/wiki/Synonym#External_links\"><span class=\"tocnumber\">5</span> <span class=\"toctext\">External links</span></a></li>\r\n</ul>\r\n</div>\r\n<p></p>\r\n<h2><span class=\"mw-headline\" id=\"Examples\">Examples</span></h2>\r\n<p>Synonyms can be any <a href=\"https://en.wikipedia.org/wiki/Part_of_speech\" title=\"Part of speech\">part of speech</a> (such as <a href=\"https://en.wikipedia.org/wiki/Noun\" title=\"Noun\">nouns</a>, <a href=\"https://en.wikipedia.org/wiki/Verbs\" title=\"Verbs\" class=\"mw-redirect\">verbs</a>, <a href=\"https://en.wikipedia.org/wiki/Adjective\" title=\"Adjective\">adjectives</a>, <a href=\"https://en.wikipedia.org/wiki/Adverb\" title=\"Adverb\">adverbs</a> or <a href=\"https://en.wikipedia.org/wiki/Preposition\" title=\"Preposition\" class=\"mw-redirect\">prepositions</a>), as long as both words belong to the same part of speech. Examples:</p>\r\n<ul>\r\n<li>verb\r\n<ul>\r\n<li><i>buy</i> and <i>purchase</i></li>\r\n</ul>\r\n</li>\r\n<li>adjective\r\n<ul>\r\n<li><i>big</i> and <i>large</i></li>\r\n</ul>\r\n</li>\r\n<li>adverb\r\n<ul>\r\n<li><i>quickly</i> and <i>speedily</i></li>\r\n</ul>\r\n</li>\r\n<li>preposition\r\n<ul>\r\n<li><i>on</i> and <i>upon</i></li>\r\n</ul>\r\n</li>\r\n</ul>\r\n<p>Note that synonyms are defined with respect to certain senses of words; for instance, <i>pupil</i> as the <i>aperture in the iris of the eye</i> is not synonymous with <i>student</i>. Such like, <i>he expired</i> means the same as <i>he died</i>, yet <i>my passport has expired</i> cannot be replaced by <i>my passport has died</i>.</p>\r\n<p>In English, many synonyms emerged in the <a href=\"https://en.wikipedia.org/wiki/Middle_Ages\" title=\"Middle Ages\">Middle Ages</a>, after the <a href=\"https://en.wikipedia.org/wiki/Norman_conquest_of_England\" title=\"Norman conquest of England\">Norman conquest of England</a>. While <a href=\"https://en.wikipedia.org/wiki/England\" title=\"England\">England</a>'s new ruling class spoke <a href=\"https://en.wikipedia.org/wiki/Norman_French\" title=\"Norman French\" class=\"mw-redirect\">Norman French</a>, the lower classes continued to speak <a href=\"https://en.wikipedia.org/wiki/Old_English\" title=\"Old English\">Old English</a> (Anglo-Saxon). Thus, today we have synonyms like the Norman-derived <i>people</i>, <i>liberty</i> and <i>archer</i>, and the Saxon-derived <i>folk</i>, <i>freedom</i> and <i>bowman</i>. For more examples, see the <a href=\"https://en.wikipedia.org/wiki/List_of_Germanic_and_Latinate_equivalents_in_English\" title=\"List of Germanic and Latinate equivalents in English\">list of Germanic and Latinate equivalents in English</a>.</p>\r\n<p>Some <a href=\"https://en.wikipedia.org/wiki/Lexicographer\" title=\"Lexicographer\" class=\"mw-redirect\">lexicographers</a> claim that no synonyms have exactly the same meaning (in all contexts or social levels of language) because <a href=\"https://en.wikipedia.org/wiki/Etymology\" title=\"Etymology\">etymology</a>, <a href=\"https://en.wikipedia.org/wiki/Orthography\" title=\"Orthography\">orthography</a>,\r\n phonic qualities, ambiguous meanings, usage, etc. make them unique.\r\nDifferent words that are similar in meaning usually differ for a reason:\r\n <i>feline</i> is more formal than <i>cat</i>; <i>long</i> and <i>extended</i> are only synonyms in one usage and not in others (for example, a <i>long arm</i> is not the same as an <i>extended arm</i>). Synonyms are also a source of <a href=\"https://en.wikipedia.org/wiki/Euphemism\" title=\"Euphemism\">euphemisms</a>.</p>\r\n<p>The purpose of a <a href=\"https://en.wikipedia.org/wiki/Thesaurus\" title=\"Thesaurus\">thesaurus</a> is to offer the user a listing of similar or related words; these are often, but not always, synonyms.</p>\r\n<h2><span class=\"mw-headline\" id=\"Related_terms\">Related terms</span></h2>\r\n<ul>\r\n<li>The word <i><b>poecilonym</b></i> is a rare synonym of the word <i>synonym</i>. It is not entered in most major dictionaries and is a curiosity or piece of trivia for being an <a href=\"https://en.wikipedia.org/wiki/Autological_word\" title=\"Autological word\">autological word</a> because of its <a href=\"https://en.wikipedia.org/wiki/Meta\" title=\"Meta\">meta</a> quality as a synonym of <i>synonym</i>.</li>\r\n<li><b><a href=\"https://en.wikipedia.org/wiki/Antonym\" title=\"Antonym\" class=\"mw-redirect\">Antonyms</a></b> are words with opposite or nearly opposite meanings. For example: <i>hot</i> ↔ <i>cold</i>, <i>large</i> ↔ <i>small</i>, <i>thick</i> ↔ <i>thin</i>, <i>synonym</i> ↔ <i>antonym</i></li>\r\n<li><b><a href=\"https://en.wikipedia.org/wiki/Hypernym\" title=\"Hypernym\" class=\"mw-redirect\">Hypernyms</a></b> and <b><a href=\"https://en.wikipedia.org/wiki/Hyponym\" title=\"Hyponym\" class=\"mw-redirect\">hyponyms</a></b> are words that refer to, respectively, a general category and a specific instance of that category. For example, <i>vehicle</i> is a hypernym of <i>car</i>, and <i>car</i> is a hyponym of <i>vehicle</i>.</li>\r\n<li><b><a href=\"https://en.wikipedia.org/wiki/Homophone\" title=\"Homophone\">Homophones</a></b> are words that have the same pronunciation, but different meanings. For example, <i>witch</i> and <i>which</i> are homophones in most accents (because they are pronounced the same).</li>\r\n<li><b><a href=\"https://en.wikipedia.org/wiki/Homograph\" title=\"Homograph\">Homographs</a></b> are words that have the same spelling, but have different pronunciations. For example, one can <i>record</i> a song or keep a <i>record</i> of documents.</li>\r\n<li><b><a href=\"https://en.wikipedia.org/wiki/Homonym\" title=\"Homonym\">Homonyms</a></b> are words that have the same pronunciation and spelling, but have different meanings. For example, <i>rose</i> (a type of flower) and <i>rose</i> (past tense of <i>rise</i>) are homonyms.</li>\r\n</ul>\r\n<h2><span class=\"mw-headline\" id=\"See_also\">See also</span></h2>\r\n<ul>\r\n<li><a href=\"https://en.wikipedia.org/wiki/-onym\" title=\"-onym\">-onym</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Synonym_ring\" title=\"Synonym ring\">Synonym ring</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Cognitive_synonymy\" title=\"Cognitive synonymy\">Cognitive synonymy</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Elegant_variation\" title=\"Elegant variation\">Elegant variation</a>, the gratuitous use of a synonym in prose</li>\r\n</ul>\r\n<h2><span class=\"mw-headline\" id=\"References\">References</span></h2>\r\n<div class=\"reflist\" style=\"list-style-type: decimal;\">\r\n<ol class=\"references\">\r\n<li id=\"cite_note-Stanojevi.C4.87_2009-1\"></li></ol></div><li id=\"cite_note-Stanojevi.C4.87_2009-1\"> <span class=\"reference-text\"><cite id=\"CITEREFStanojevi.C4.872009\" class=\"citation\">Stanojević, Maja (2009), <a rel=\"nofollow\" class=\"external text\" href=\"http://facta.junis.ni.ac.rs/lal/lal200902/lal200902-05.pdf\">\"Cognitive synonymy: a general overview\"</a> <span style=\"font-size:85%;\">(PDF)</span>, <i>Facta Universitatis, Linguistics and Literature series</i> <b>7</b> (2): 193–200.</cite><span title=\"ctx_ver=Z39.88-2004&amp;rfr_id=info%3Asid%2Fen.wikipedia.org%3ASynonym&amp;rft.atitle=Cognitive+synonymy%3A+a+general+overview&amp;rft.aufirst=Maja&amp;rft.aulast=Stanojevi%C4%87&amp;rft.date=2009&amp;rft.genre=article&amp;rft_id=http%3A%2F%2Ffacta.junis.ni.ac.rs%2Flal%2Flal200902%2Flal200902-05.pdf&amp;rft.issue=2&amp;rft.jtitle=Facta+Universitatis%2C+Linguistics+and+Literature+series&amp;rft.pages=193-200&amp;rft_val_fmt=info%3Aofi%2Ffmt%3Akev%3Amtx%3Ajournal&amp;rft.volume=7\" class=\"Z3988\"></span></span></li>\r\n<li id=\"cite_note-2\"></li><div class=\"reflist\" style=\"list-style-type: decimal;\"><ol class=\"references\"><li value=\"2\" id=\"cite_note-2\"> <span class=\"reference-text\"><cite class=\"citation book\"><a rel=\"nofollow\" class=\"external text\" href=\"http://books.google.com/books?id=bGSSl9I4maEC&amp;pg=PA370&amp;dq=a+widespr…n%20that...%20Hollywood%20was%20synonymous%20with%20immorality&amp;f=false\"><i>The Fitzgeralds and the Kennedys</i></a>. Macmillan. 1991. p. 370. <a href=\"https://en.wikipedia.org/wiki/International_Standard_Book_Number\" title=\"International Standard Book Number\">ISBN</a> <a href=\"https://en.wikipedia.org/wiki/Special:BookSources/9780312063542\" title=\"Special:BookSources/9780312063542\">9780312063542</a><span class=\"reference-accessdate\">. Retrieved <span class=\"nowrap\">2014-05-27</span></span>.</cite><span title=\"ctx_ver=Z39.88-2004&amp;rfr_id=info%3Asid%2Fen.wikipedia.org%3ASynonym&amp;rft.btitle=The+Fitzgeralds+and+the+Kennedys&amp;rft.date=1991&amp;rft.genre=book&amp;rft_id=http%3A%2F%2Fbooks.google.com%2Fbooks%3Fid%3DbGSSl9I4maEC%26pg%3DPA370%26dq%3Da%2Bwidespread%2Bimpression%2Bthat...%2BHollywood%2Bwas%2Bsynonymous%2Bwith%2Bimmorality%26hl%3Den%26sa%3DX%26ei%3DIRSEU7roL--_sQSb7IGgCA%26ved%3D0CCsQ6AEwAA%23v%3Donepage%26q%3Da%2520widespread%2520impression%2520that...%2520Hollywood%2520was%2520synonymous%2520with%2520immorality%26f%3Dfalse&amp;rft.isbn=9780312063542&amp;rft.pages=370&amp;rft.pub=Macmillan&amp;rft_val_fmt=info%3Aofi%2Ffmt%3Akev%3Amtx%3Abook\" class=\"Z3988\"></span></span></li>\r\n</ol>\r\n</div>\r\n<h2><span class=\"mw-headline\" id=\"External_links\">External links</span></h2><div id=\"content\" class=\"mw-body\" role=\"main\"><div id=\"bodyContent\" class=\"mw-body-content\"><div id=\"mw-content-text\" dir=\"ltr\" class=\"mw-content-ltr\" lang=\"en\">\r\n<table class=\"mbox-small plainlinks sistersitebox\" style=\"border:1px solid #aaa;background-color:#f9f9f9\">\r\n<tbody><tr>\r\n<td class=\"mbox-image\"><img alt=\"\" src=\"https://upload.wikimedia.org/wikipedia/commons/thumb/f/f8/Wiktionary-logo-en.svg/37px-Wiktionary-logo-en.svg.png\" srcset=\"//upload.wikimedia.org/wikipedia/commons/thumb/f/f8/Wiktionary-logo-en.svg/55px-Wiktionary-logo-en.svg.png 1.5x, //upload.wikimedia.org/wikipedia/commons/thumb/f/f8/Wiktionary-logo-en.svg/73px-Wiktionary-logo-en.svg.png 2x\" data-file-width=\"1000\" data-file-height=\"1089\" height=\"40\" width=\"37\"></td>\r\n<td class=\"mbox-text plainlist\">Look up <i><b><a href=\"https://en.wiktionary.org/wiki/Special:Search/synonym\" class=\"extiw\" title=\"wiktionary:Special:Search/synonym\">synonym</a></b></i> in Wiktionary, the free dictionary.</td>\r\n</tr>\r\n</tbody></table>\r\n<p>Tools which graph words relations :</p>\r\n<ul>\r\n<li><a rel=\"nofollow\" class=\"external text\" href=\"http://graphwords.com/\">Graph Words</a> - Online tool for visualization word relations</li>\r\n<li><a rel=\"nofollow\" class=\"external text\" href=\"http://www.synonyms.net/\">Synonyms.net</a>\r\n - Online reference resource that provides instant synonyms and antonyms\r\n definitions including visualizations, voice pronunciations and\r\ntranslations</li>\r\n<li><a rel=\"nofollow\" class=\"external text\" href=\"http://dico.isc.cnrs.fr/en/index.html\">English/French Semantic Atlas</a>\r\n - Graph words relations in English, French and gives cross\r\nrepresentations for translations - offers 500 searches per user per day.</li>\r\n</ul>\r\n<p>Plain words synonyms finder :</p>\r\n<ul>\r\n<li><a rel=\"nofollow\" class=\"external text\" href=\"http://www.synonym-finder.com/\">Synonym Finder</a> - Synonym finder including hypernyms in search result</li>\r\n<li><a rel=\"nofollow\" class=\"external text\" href=\"http://www.how-to-say.net\">how to say</a> - Online Synonym finder</li>\r\n<li><a rel=\"nofollow\" class=\"external text\" href=\"http://synonymosum.com/\">Synonymosum</a> - Online Synonym Dictionary - words and their synonyms</li>\r\n<li><a rel=\"nofollow\" class=\"external text\" href=\"http://trovami.altervista.org/sinonimi/en\">Thesaurus</a> - Online synonyms in English, Italian, French and German</li>\r\n<li><a rel=\"nofollow\" class=\"external text\" href=\"http://synonyms.woxikon.com/\">Woxikon Synonyms</a> - Over 1 million synonyms - English, German, Spanish, French, Italian, Portuguese, Swedish and Dutch</li>\r\n<li><a rel=\"nofollow\" class=\"external text\" href=\"http://www.thefreedictionary.com/\">Freedictionary.com</a> Free online English thesaurus and dictionary containing synonyms, related words, antonyms, definitions, idioms and more</li>\r\n<li><a rel=\"nofollow\" class=\"external text\" href=\"http://www.powerthesaurus.org/\">Power Thesaurus</a> - Thesaurus with synonyms ordered by rating</li>\r\n<li><a rel=\"nofollow\" class=\"external text\" href=\"http://www.findmewords.com/synonyms.html\">FindMeWords Synonyms</a> - Online Synonym Dictionary with definitions</li>\r\n</ul>\r\n<table class=\"navbox\" style=\"border-spacing:0\">\r\n<tbody><tr>\r\n<td style=\"padding:2px\">\r\n<table id=\"collapsibleTable0\" class=\"nowraplinks collapsible autocollapse navbox-inner\" style=\"border-spacing:0;background:transparent;color:inherit\">\r\n<tbody><tr>\r\n<th scope=\"col\" class=\"navbox-title\" colspan=\"2\"><span class=\"collapseButton\">[<a href=\"https://en.wikipedia.org/wiki/Synonym#\" id=\"collapseButton0\">hide</a>]</span>\r\n<div class=\"plainlinks hlist navbar mini\">\r\n<ul>\r\n<li class=\"nv-view\"><a href=\"https://en.wikipedia.org/wiki/Template:Lexicology\" title=\"Template:Lexicology\"><abbr title=\"View this template\" style=\";;background:none transparent;border:none;\">v</abbr></a></li>\r\n<li class=\"nv-talk\"><a href=\"https://en.wikipedia.org/wiki/Template_talk:Lexicology\" title=\"Template talk:Lexicology\"><abbr title=\"Discuss this template\" style=\";;background:none transparent;border:none;\">t</abbr></a></li>\r\n<li class=\"nv-edit\"><a class=\"external text\" href=\"https://en.wikipedia.org/w/index.php?title=Template:Lexicology&amp;action=edit\"><abbr title=\"Edit this template\" style=\";;background:none transparent;border:none;\">e</abbr></a></li>\r\n</ul>\r\n</div>\r\n<div style=\"font-size:114%\"><a href=\"https://en.wikipedia.org/wiki/Lexicology\" title=\"Lexicology\">Lexicology</a></div>\r\n</th>\r\n</tr>\r\n<tr style=\"height:2px\">\r\n<td colspan=\"2\"></td>\r\n</tr>\r\n<tr>\r\n<th scope=\"row\" class=\"navbox-group\">Major terms</th>\r\n<td class=\"navbox-list navbox-odd hlist\" style=\"text-align:left;border-left-width:2px;border-left-style:solid;width:100%;padding:0px\">\r\n<div style=\"padding:0em 0.25em\">\r\n<ul>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Lexical_item\" title=\"Lexical item\">Lexical item</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Lexicon\" title=\"Lexicon\">Lexicon</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Lexis_%28linguistics%29\" title=\"Lexis (linguistics)\">Lexis</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Word\" title=\"Word\">Word</a></li>\r\n</ul>\r\n</div>\r\n</td>\r\n</tr>\r\n<tr style=\"height:2px\">\r\n<td colspan=\"2\"></td>\r\n</tr>\r\n<tr>\r\n<th scope=\"row\" class=\"navbox-group\">Elements</th>\r\n<td class=\"navbox-list navbox-even hlist\" style=\"text-align:left;border-left-width:2px;border-left-style:solid;width:100%;padding:0px\">\r\n<div style=\"padding:0em 0.25em\">\r\n<ul>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Cherology\" title=\"Cherology\">Chereme</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Glyph\" title=\"Glyph\">Glyphs</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Grapheme\" title=\"Grapheme\">Grapheme</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Lemma_%28psycholinguistics%29\" title=\"Lemma (psycholinguistics)\">Lemma</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Lexeme\" title=\"Lexeme\">Lexeme</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Meronymy\" title=\"Meronymy\">Meronymy</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Morpheme\" title=\"Morpheme\">Morpheme</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Phoneme\" title=\"Phoneme\">Phoneme</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Seme_%28semantics%29\" title=\"Seme (semantics)\">Seme</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Sememe\" title=\"Sememe\">Sememe</a></li>\r\n</ul>\r\n</div>\r\n</td>\r\n</tr>\r\n<tr style=\"height:2px\">\r\n<td colspan=\"2\"></td>\r\n</tr>\r\n<tr>\r\n<th scope=\"row\" class=\"navbox-group\"><a href=\"https://en.wikipedia.org/wiki/Semantics\" title=\"Semantics\">Semantic relations</a></th>\r\n<td class=\"navbox-list navbox-odd hlist\" style=\"text-align:left;border-left-width:2px;border-left-style:solid;width:100%;padding:0px\">\r\n<div style=\"padding:0em 0.25em\">\r\n<ul>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Opposite_%28semantics%29\" title=\"Opposite (semantics)\">Antonymy</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Holonymy\" title=\"Holonymy\">Holonymy</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Hyponymy_and_hypernymy\" title=\"Hyponymy and hypernymy\">Hyponymy and hypernymy</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Idiom\" title=\"Idiom\">Idiom</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Lexical_semantics\" title=\"Lexical semantics\">Lexical semantics</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Semantic_network\" title=\"Semantic network\">Semantic network</a></li>\r\n<li><strong class=\"selflink\">Synonym</strong></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Troponymy\" title=\"Troponymy\">Troponymy</a></li>\r\n</ul>\r\n</div>\r\n</td>\r\n</tr>\r\n<tr style=\"height:2px\">\r\n<td colspan=\"2\"></td>\r\n</tr>\r\n<tr>\r\n<th scope=\"row\" class=\"navbox-group\">Functions</th>\r\n<td class=\"navbox-list navbox-even hlist\" style=\"text-align:left;border-left-width:2px;border-left-style:solid;width:100%;padding:0px\">\r\n<div style=\"padding:0em 0.25em\">\r\n<ul>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Function_word\" title=\"Function word\">Function word</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Headword\" title=\"Headword\">Headword</a></li>\r\n</ul>\r\n</div>\r\n</td>\r\n</tr>\r\n<tr style=\"height:2px\">\r\n<td colspan=\"2\"></td>\r\n</tr>\r\n<tr>\r\n<th scope=\"row\" class=\"navbox-group\">Fields</th>\r\n<td class=\"navbox-list navbox-odd hlist\" style=\"text-align:left;border-left-width:2px;border-left-style:solid;width:100%;padding:0px\">\r\n<div style=\"padding:0em 0.25em\">\r\n<ul>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Controlled_vocabulary\" title=\"Controlled vocabulary\">Controlled vocabulary</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/English_lexicology_and_lexicography\" title=\"English lexicology and lexicography\">English lexicology and lexicography</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/International_scientific_vocabulary\" title=\"International scientific vocabulary\">International scientific vocabulary</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Lexicographic_error\" title=\"Lexicographic error\">Lexicographic error</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Lexicographic_information_cost\" title=\"Lexicographic information cost\">Lexicographic information cost</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Linguistic_prescription\" title=\"Linguistic prescription\">Linguistic prescription</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Morphology_%28linguistics%29\" title=\"Morphology (linguistics)\">Morphology</a></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Specialised_lexicography\" title=\"Specialised lexicography\">Specialised lexicography</a></li>\r\n</ul>\r\n</div>\r\n</td>\r\n</tr>\r\n</tbody></table>\r\n</td>\r\n</tr>\r\n</tbody></table>\r\n<table class=\"navbox\" style=\"border-spacing:0\">\r\n<tbody><tr>\r\n<td style=\"padding:2px\">\r\n<table class=\"nowraplinks hlist navbox-inner\" style=\"border-spacing:0;background:transparent;color:inherit\">\r\n<tbody><tr>\r\n<th scope=\"row\" class=\"navbox-group\"><a href=\"https://en.wikipedia.org/wiki/Help:Authority_control\" title=\"Help:Authority control\">Authority control</a></th>\r\n<td class=\"navbox-list navbox-odd\" style=\"text-align:left;border-left-width:2px;border-left-style:solid;width:100%;padding:0px\">\r\n<div style=\"padding:0em 0.25em\">\r\n<ul>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Library_of_Congress_Control_Number\" title=\"Library of Congress Control Number\">LCCN</a>: <span class=\"uid\"><a rel=\"nofollow\" class=\"external text\" href=\"http://id.loc.gov/authorities/subjects/sh85131642\">sh85131642</a></span></li>\r\n<li><a href=\"https://en.wikipedia.org/wiki/Integrated_Authority_File\" title=\"Integrated Authority File\">GND</a>: <span class=\"uid\"><a rel=\"nofollow\" class=\"external text\" href=\"http://d-nb.info/gnd/4058765-4\">4058765-4</a></span></li>\r\n</ul>\r\n</div>\r\n</td>\r\n</tr>\r\n</tbody></table>\r\n</td>\r\n</tr>\r\n</tbody></table>\r\n\r\n\r\n\r\n\r\n\r\n\r\n\r\n</div>\r\n        <div id=\"catlinks\" class=\"catlinks\"><div id=\"mw-normal-catlinks\" class=\"mw-normal-catlinks\"><a href=\"https://en.wikipedia.org/wiki/Help:Category\" title=\"Help:Category\">Categories</a>: <ul><li><a href=\"https://en.wikipedia.org/wiki/Category:Lexical_semantics\" title=\"Category:Lexical semantics\">Lexical semantics</a></li><li><a href=\"https://en.wikipedia.org/wiki/Category:Types_of_words\" title=\"Category:Types of words\">Types of words</a></li></ul></div></div>        <div class=\"visualClear\"></div>\r\n              </div>\r\n    </div>\r\n    <div id=\"mw-navigation\">\r\n      <h2>Navigation menu</h2>\r\n\r\n      <div id=\"mw-head\">\r\n                  <div id=\"p-personal\" role=\"navigation\" class=\"\" aria-labelledby=\"p-personal-label\">\r\n\r\n            <ul>\r\n              <li id=\"pt-createaccount\"><a href=\"https://en.wikipedia.org/w/index.php?title=Special:UserLogin&amp;returnto=Synonym&amp;type=signup\" title=\"You are encouraged to create an account and log in; however, it is not mandatory\">Create account</a></li><li id=\"pt-userpage\"><span style=\"color: gray;\">Not logged in</span></li><li id=\"pt-mytalk\"><a accesskey=\"n\" title=\"Your talk page [Alt+Shift+n]\" href=\"https://en.wikipedia.org/wiki/Special:MyTalk\">Talk</a></li><li id=\"pt-mycontris\"><a accesskey=\"y\" title=\"A list of your contributions [Alt+Shift+y]\" href=\"https://en.wikipedia.org/wiki/Special:MyContributions\">Contributions</a></li><li id=\"pt-login\"><a href=\"https://en.wikipedia.org/w/index.php?title=Special:UserLogin&amp;returnto=Synonym\" title=\"You're encouraged to log in; however, it's not mandatory. [Alt+Shift+o]\" accesskey=\"o\">Log in</a></li>            </ul>\r\n          </div>\r\n                  <div id=\"left-navigation\">\r\n                    <div id=\"p-namespaces\" role=\"navigation\" class=\"vectorTabs\" aria-labelledby=\"p-namespaces-label\">\r\n\r\n            <ul>\r\n                              <li id=\"ca-nstab-main\" class=\"selected\"><span><a href=\"https://en.wikipedia.org/wiki/Synonym\" title=\"View the content page [Alt+Shift+c]\" accesskey=\"c\">Article</a></span></li>\r\n                              <li id=\"ca-talk\"><span><a href=\"https://en.wikipedia.org/wiki/Talk:Synonym\" title=\"Discussion about the content page [Alt+Shift+t]\" accesskey=\"t\" rel=\"discussion\">Talk</a></span></li>\r\n                          </ul>\r\n          </div>\r\n\r\n                  </div>\r\n        <div id=\"right-navigation\">\r\n                    <div id=\"p-views\" role=\"navigation\" class=\"vectorTabs\" aria-labelledby=\"p-views-label\">\r\n\r\n            <ul>\r\n                              <li id=\"ca-view\" class=\"selected\"><span><a href=\"https://en.wikipedia.org/wiki/Synonym\">Read</a></span></li>\r\n                              <li id=\"ca-edit\"><span><a href=\"https://en.wikipedia.org/w/index.php?title=Synonym&amp;action=edit\" title=\"Edit this page [Alt+Shift+e]\" accesskey=\"e\">Edit</a></span></li>\r\n                              <li id=\"ca-history\" class=\"collapsible\"><span><a href=\"https://en.wikipedia.org/w/index.php?title=Synonym&amp;action=history\" title=\"Past revisions of this page [Alt+Shift+h]\" accesskey=\"h\">View history</a></span></li>\r\n                          </ul>\r\n          </div>\r\n\r\n                    <div id=\"p-search\" role=\"search\">\r\n\r\n\r\n            <form action=\"/w/index.php\" id=\"searchform\">\r\n              <div id=\"simpleSearch\">\r\n              <input autocomplete=\"off\" tabindex=\"1\" name=\"search\" placeholder=\"Search\" title=\"Search Wikipedia [Alt+Shift+f]\" accesskey=\"f\" id=\"searchInput\" type=\"search\"></div></form></div></div></div></div><div id=\"mw-navigation\"><div id=\"mw-head\"><div id=\"right-navigation\"><div id=\"p-search\" role=\"search\"><form action=\"/w/index.php\" id=\"searchform\"><div id=\"simpleSearch\">              </div>\r\n            </form>\r\n          </div>\r\n                  </div>\r\n      </div>\r\n      <div id=\"mw-panel\">\r\n        <div id=\"p-logo\" role=\"banner\"><a class=\"mw-wiki-logo\" href=\"https://en.wikipedia.org/wiki/Main_Page\" title=\"Visit the main page\"></a></div>\r\n            <div class=\"portal\" role=\"navigation\" id=\"p-navigation\" aria-labelledby=\"p-navigation-label\">\r\n\r\n\r\n      <div class=\"body\">\r\n                  <ul>\r\n            <li id=\"n-mainpage-description\"><a href=\"https://en.wikipedia.org/wiki/Main_Page\" title=\"Visit the main page [Alt+Shift+z]\" accesskey=\"z\">Main page</a></li><li id=\"n-contents\"><a href=\"https://en.wikipedia.org/wiki/Portal:Contents\" title=\"Guides to browsing Wikipedia\">Contents</a></li><li id=\"n-featuredcontent\"><a href=\"https://en.wikipedia.org/wiki/Portal:Featured_content\" title=\"Featured content – the best of Wikipedia\">Featured content</a></li><li id=\"n-currentevents\"><a href=\"https://en.wikipedia.org/wiki/Portal:Current_events\" title=\"Find background information on current events\">Current events</a></li><li id=\"n-randompage\"><a href=\"https://en.wikipedia.org/wiki/Special:Random\" title=\"Load a random article [Alt+Shift+x]\" accesskey=\"x\">Random article</a></li><li id=\"n-sitesupport\"><a href=\"https://donate.wikimedia.org/wiki/Special:FundraiserRedirector?utm_source=d…mp;utm_medium=sidebar&amp;utm_campaign=C13_en.wikipedia.org&amp;uselang=en\" title=\"Support us\">Donate to Wikipedia</a></li><li id=\"n-shoplink\"><a href=\"https://shop.wikimedia.org\" title=\"Visit the Wikipedia store\">Wikipedia store</a></li>          </ul>\r\n              </div>\r\n    </div>\r\n      <div class=\"portal\" role=\"navigation\" id=\"p-interaction\" aria-labelledby=\"p-interaction-label\">\r\n      <h3 id=\"p-interaction-label\">Interaction</h3>\r\n\r\n      <div class=\"body\">\r\n                  <ul>\r\n            <li id=\"n-help\"><a href=\"https://en.wikipedia.org/wiki/Help:Contents\" title=\"Guidance on how to use and edit Wikipedia\">Help</a></li><li id=\"n-aboutsite\"><a href=\"https://en.wikipedia.org/wiki/Wikipedia:About\" title=\"Find out about Wikipedia\">About Wikipedia</a></li><li id=\"n-portal\"><a href=\"https://en.wikipedia.org/wiki/Wikipedia:Community_portal\" title=\"About the project, what you can do, where to find things\">Community portal</a></li><li id=\"n-recentchanges\"><a href=\"https://en.wikipedia.org/wiki/Special:RecentChanges\" title=\"A list of recent changes in the wiki [Alt+Shift+r]\" accesskey=\"r\">Recent changes</a></li><li id=\"n-contactpage\"><a href=\"https://en.wikipedia.org/wiki/Wikipedia:Contact_us\" title=\"How to contact Wikipedia\">Contact page</a></li>          </ul>\r\n              </div>\r\n    </div>\r\n      <div class=\"portal\" role=\"navigation\" id=\"p-tb\" aria-labelledby=\"p-tb-label\">\r\n      <h3 id=\"p-tb-label\">Tools</h3>\r\n\r\n      <div class=\"body\">\r\n                  <ul>\r\n            <li id=\"t-whatlinkshere\"><a href=\"https://en.wikipedia.org/wiki/Special:WhatLinksHere/Synonym\" title=\"List of all English Wikipedia pages containing links to this page [Alt+Shift+j]\" accesskey=\"j\">What links here</a></li><li id=\"t-recentchangeslinked\"><a href=\"https://en.wikipedia.org/wiki/Special:RecentChangesLinked/Synonym\" title=\"Recent changes in pages linked from this page [Alt+Shift+k]\" accesskey=\"k\">Related changes</a></li><li id=\"t-upload\"><a href=\"https://en.wikipedia.org/wiki/Wikipedia:File_Upload_Wizard\" title=\"Upload files [Alt+Shift+u]\" accesskey=\"u\">Upload file</a></li><li id=\"t-specialpages\"><a href=\"https://en.wikipedia.org/wiki/Special:SpecialPages\" title=\"A list of all special pages [Alt+Shift+q]\" accesskey=\"q\">Special pages</a></li><li id=\"t-permalink\"><a href=\"https://en.wikipedia.org/w/index.php?title=Synonym&amp;oldid=690006505\" title=\"Permanent link to this revision of the page\">Permanent link</a></li><li id=\"t-info\"><a href=\"https://en.wikipedia.org/w/index.php?title=Synonym&amp;action=info\" title=\"More information about this page\">Page information</a></li><li id=\"t-wikibase\"><a href=\"https://www.wikidata.org/wiki/Q42106\" title=\"Link to connected data repository item [Alt+Shift+g]\" accesskey=\"g\">Wikidata item</a></li><li id=\"t-cite\"><a href=\"https://en.wikipedia.org/w/index.php?title=Special:CiteThisPage&amp;page=Synonym&amp;id=690006505\" title=\"Information on how to cite this page\">Cite this page</a></li>          </ul>\r\n              </div>\r\n    </div>\r\n      <div class=\"portal\" role=\"navigation\" id=\"p-coll-print_export\" aria-labelledby=\"p-coll-print_export-label\">\r\n      <h3 id=\"p-coll-print_export-label\">Print/export</h3>\r\n\r\n      <div class=\"body\">\r\n                  <ul>\r\n            <li id=\"coll-create_a_book\"><a href=\"https://en.wikipedia.org/w/index.php?title=Special:Book&amp;bookcmd=book_creator&amp;referer=Synonym\">Create a book</a></li><li id=\"coll-download-as-rdf2latex\"><a href=\"https://en.wikipedia.org/w/index.php?title=Special:Book&amp;bookcmd=render_…itle=Synonym&amp;returnto=Synonym&amp;oldid=690006505&amp;writer=rdf2latex\">Download as PDF</a></li><li id=\"t-print\"><a href=\"https://en.wikipedia.org/w/index.php?title=Synonym&amp;printable=yes\" title=\"Printable version of this page [Alt+Shift+p]\" accesskey=\"p\">Printable version</a></li>          </ul>\r\n              </div>\r\n    </div>\r\n      <div class=\"portal\" role=\"navigation\" id=\"p-lang\" aria-labelledby=\"p-lang-label\"><span aria-haspopup=\"true\" role=\"button\" tabindex=\"0\" title=\"Language settings\" class=\"uls-settings-trigger\"></span>\r\n      <h3 id=\"p-lang-label\">Languages</h3>\r\n\r\n      <div class=\"body\">\r\n                  <ul>\r\n            <li class=\"interlanguage-link interwiki-af\"><a href=\"https://af.wikipedia.org/wiki/Sinoniem\" title=\"Sinoniem – Afrikaans\" hreflang=\"af\" lang=\"af\">Afrikaans</a></li><li class=\"interlanguage-link interwiki-ar\"><a href=\"https://ar.wikipedia.org/wiki/%D8%AA%D8%B1%D8%A7%D8%AF%D9%81\" title=\"ترادف – Arabic\" hreflang=\"ar\" lang=\"ar\">العربية</a></li><li class=\"interlanguage-link interwiki-az\"><a href=\"https://az.wikipedia.org/wiki/Sinoniml%C9%99r\" title=\"Sinonimlər – Azerbaijani\" hreflang=\"az\" lang=\"az\">Azərbaycanca</a></li><li class=\"interlanguage-link interwiki-be\"><a href=\"https://be.wikipedia.org/wiki/%D0%A1%D1%96%D0%BD%D0%BE%D0%BD%D1%96%D0%BC%D1%8B\" title=\"Сінонімы – Belarusian\" hreflang=\"be\" lang=\"be\">Беларуская</a></li><li class=\"interlanguage-link interwiki-be-x-old\"><a href=\"https://be-x-old.wikipedia.org/wiki/%D0%A1%D1%8B%D0%BD%D0%BE%D0%BD%D1%96%D0%BC%D1%8B\" title=\"Сынонімы – беларуская (тарашкевіца)‎\" hreflang=\"be-x-old\" lang=\"be-x-old\">Беларуская (тарашкевіца)‎</a></li><li class=\"interlanguage-link interwiki-bg\"><a href=\"https://bg.wikipedia.org/wiki/%D0%A1%D0%B8%D0%BD%D0%BE%D0%BD%D0%B8%D0%BC\" title=\"Синоним – Bulgarian\" hreflang=\"bg\" lang=\"bg\">Български</a></li><li class=\"interlanguage-link interwiki-bar\"><a href=\"https://bar.wikipedia.org/wiki/Synonym\" title=\"Synonym – Bavarian\" hreflang=\"bar\" lang=\"bar\">Boarisch</a></li><li class=\"interlanguage-link interwiki-ca\"><a href=\"https://ca.wikipedia.org/wiki/Sin%C3%B2nim\" title=\"Sinònim – Catalan\" hreflang=\"ca\" lang=\"ca\">Català</a></li><li class=\"interlanguage-link interwiki-cv\"><a href=\"https://cv.wikipedia.org/wiki/%D0%A1%D0%B8%D0%BD%D0%BE%D0%BD%D0%B8%D0%BC%D1%81%D0%B5%D0%BC\" title=\"Синонимсем – Chuvash\" hreflang=\"cv\" lang=\"cv\">Чӑвашла</a></li><li class=\"interlanguage-link interwiki-cs\"><a href=\"https://cs.wikipedia.org/wiki/Synonymum\" title=\"Synonymum – Czech\" hreflang=\"cs\" lang=\"cs\">Čeština</a></li><li class=\"interlanguage-link interwiki-cy\"><a href=\"https://cy.wikipedia.org/wiki/Cyfystyr\" title=\"Cyfystyr – Welsh\" hreflang=\"cy\" lang=\"cy\">Cymraeg</a></li><li class=\"interlanguage-link interwiki-da\"><a href=\"https://da.wikipedia.org/wiki/Synonym\" title=\"Synonym – Danish\" hreflang=\"da\" lang=\"da\">Dansk</a></li><li class=\"interlanguage-link interwiki-de\"><a href=\"https://de.wikipedia.org/wiki/Synonym\" title=\"Synonym – German\" hreflang=\"de\" lang=\"de\">Deutsch</a></li><li class=\"interlanguage-link interwiki-et\"><a href=\"https://et.wikipedia.org/wiki/S%C3%BCnon%C3%BC%C3%BCm\" title=\"Sünonüüm – Estonian\" hreflang=\"et\" lang=\"et\">Eesti</a></li><li class=\"interlanguage-link interwiki-el\"><a href=\"https://el.wikipedia.org/wiki/%CE%A3%CF%85%CE%BD%CF%8E%CE%BD%CF%85%CE%BC%CE%BF\" title=\"Συνώνυμο – Greek\" hreflang=\"el\" lang=\"el\">Ελληνικά</a></li><li class=\"interlanguage-link interwiki-es\"><a href=\"https://es.wikipedia.org/wiki/Sinonimia_%28sem%C3%A1ntica%29\" title=\"Sinonimia (semántica) – Spanish\" hreflang=\"es\" lang=\"es\">Español</a></li><li class=\"interlanguage-link interwiki-eo\"><a href=\"https://eo.wikipedia.org/wiki/Sinonimo\" title=\"Sinonimo – Esperanto\" hreflang=\"eo\" lang=\"eo\">Esperanto</a></li><li class=\"interlanguage-link interwiki-eu\"><a href=\"https://eu.wikipedia.org/wiki/Sinonimo\" title=\"Sinonimo – Basque\" hreflang=\"eu\" lang=\"eu\">Euskara</a></li><li class=\"interlanguage-link interwiki-fa\"><a href=\"https://fa.wikipedia.org/wiki/%D9%85%D8%AA%D8%B1%D8%A7%D8%AF%D9%81\" title=\"مترادف – Persian\" hreflang=\"fa\" lang=\"fa\">فارسی</a></li><li class=\"interlanguage-link interwiki-fo\"><a href=\"https://fo.wikipedia.org/wiki/Samheiti\" title=\"Samheiti – Faroese\" hreflang=\"fo\" lang=\"fo\">Føroyskt</a></li><li class=\"interlanguage-link interwiki-fr\"><a href=\"https://fr.wikipedia.org/wiki/Synonymie\" title=\"Synonymie – French\" hreflang=\"fr\" lang=\"fr\">Français</a></li><li class=\"interlanguage-link interwiki-ga\"><a href=\"https://ga.wikipedia.org/wiki/Comhchiallach\" title=\"Comhchiallach – Irish\" hreflang=\"ga\" lang=\"ga\">Gaeilge</a></li><li class=\"interlanguage-link interwiki-gl\"><a href=\"https://gl.wikipedia.org/wiki/Sinonimia\" title=\"Sinonimia – Galician\" hreflang=\"gl\" lang=\"gl\">Galego</a></li><li class=\"interlanguage-link interwiki-ko\"><a href=\"https://ko.wikipedia.org/wiki/%EB%8F%99%EC%9D%98%EC%96%B4\" title=\"동의어 – Korean\" hreflang=\"ko\" lang=\"ko\">한국어</a></li><li class=\"interlanguage-link interwiki-hy\"><a href=\"https://hy.wikipedia.org/wiki/%D5%80%D5%B8%D5%B4%D5%A1%D5%B6%D5%AB%D5%B7\" title=\"Հոմանիշ – Armenian\" hreflang=\"hy\" lang=\"hy\">Հայերեն</a></li><li class=\"interlanguage-link interwiki-hi\"><a href=\"https://hi.wikipedia.org/wiki/%E0%A4%AA%E0%A4%B0%E0%A5%8D%E0%A4%AF%E0%A4%BE%E0%A4%AF%E0%A4%B5%E0%A4%BE%E0%A4%9A%E0%A5%80\" title=\"पर्यायवाची – Hindi\" hreflang=\"hi\" lang=\"hi\">हिन्दी</a></li><li class=\"interlanguage-link interwiki-hr\"><a href=\"https://hr.wikipedia.org/wiki/Sinonim\" title=\"Sinonim – Croatian\" hreflang=\"hr\" lang=\"hr\">Hrvatski</a></li><li class=\"interlanguage-link interwiki-io\"><a href=\"https://io.wikipedia.org/wiki/Sinonimo\" title=\"Sinonimo – Ido\" hreflang=\"io\" lang=\"io\">Ido</a></li><li class=\"interlanguage-link interwiki-id\"><a href=\"https://id.wikipedia.org/wiki/Sinonim\" title=\"Sinonim – Indonesian\" hreflang=\"id\" lang=\"id\">Bahasa Indonesia</a></li><li class=\"interlanguage-link interwiki-ia\"><a href=\"https://ia.wikipedia.org/wiki/Synonymo\" title=\"Synonymo – Interlingua\" hreflang=\"ia\" lang=\"ia\">Interlingua</a></li><li class=\"interlanguage-link interwiki-is\"><a href=\"https://is.wikipedia.org/wiki/Samheiti\" title=\"Samheiti – Icelandic\" hreflang=\"is\" lang=\"is\">Íslenska</a></li><li class=\"interlanguage-link interwiki-it\"><a href=\"https://it.wikipedia.org/wiki/Sinonimia\" title=\"Sinonimia – Italian\" hreflang=\"it\" lang=\"it\">Italiano</a></li><li class=\"interlanguage-link interwiki-he\"><a href=\"https://he.wikipedia.org/wiki/%D7%9E%D7%99%D7%9C%D7%94_%D7%A0%D7%A8%D7%93%D7%A4%D7%AA\" title=\"מילה נרדפת – Hebrew\" hreflang=\"he\" lang=\"he\">עברית</a></li><li class=\"interlanguage-link interwiki-jv\"><a href=\"https://jv.wikipedia.org/wiki/Dasanama\" title=\"Dasanama – Javanese\" hreflang=\"jv\" lang=\"jv\">Basa Jawa</a></li><li class=\"interlanguage-link interwiki-ka\"><a href=\"https://ka.wikipedia.org/wiki/%E1%83%A1%E1%83%98%E1%83%9C%E1%83%9D%E1%83%9C%E1%83%98%E1%83%9B%E1%83%94%E1%83%91%E1%83%98\" title=\"სინონიმები – Georgian\" hreflang=\"ka\" lang=\"ka\">ქართული</a></li><li class=\"interlanguage-link interwiki-kk\"><a href=\"https://kk.wikipedia.org/wiki/%D0%A1%D0%B8%D0%BD%D0%BE%D0%BD%D0%B8%D0%BC\" title=\"Синоним – Kazakh\" hreflang=\"kk\" lang=\"kk\">Қазақша</a></li><li class=\"interlanguage-link interwiki-ku\"><a href=\"https://ku.wikipedia.org/wiki/Hevwate\" title=\"Hevwate – Kurdish\" hreflang=\"ku\" lang=\"ku\">Kurdî</a></li><li class=\"interlanguage-link interwiki-la\"><a href=\"https://la.wikipedia.org/wiki/Synonymia\" title=\"Synonymia – Latin\" hreflang=\"la\" lang=\"la\">Latina</a></li><li class=\"interlanguage-link interwiki-lv\"><a href=\"https://lv.wikipedia.org/wiki/Sinon%C4%ABms\" title=\"Sinonīms – Latvian\" hreflang=\"lv\" lang=\"lv\">Latviešu</a></li><li class=\"interlanguage-link interwiki-lb\"><a href=\"https://lb.wikipedia.org/wiki/Synonymie\" title=\"Synonymie – Luxembourgish\" hreflang=\"lb\" lang=\"lb\">Lëtzebuergesch</a></li><li class=\"interlanguage-link interwiki-lt\"><a href=\"https://lt.wikipedia.org/wiki/Sinonimas\" title=\"Sinonimas – Lithuanian\" hreflang=\"lt\" lang=\"lt\">Lietuvių</a></li><li class=\"interlanguage-link interwiki-hu\"><a href=\"https://hu.wikipedia.org/wiki/Szinon%C3%ADmia\" title=\"Szinonímia – Hungarian\" hreflang=\"hu\" lang=\"hu\">Magyar</a></li><li class=\"interlanguage-link interwiki-mk\"><a href=\"https://mk.wikipedia.org/wiki/%D0%A1%D0%B8%D0%BD%D0%BE%D0%BD%D0%B8%D0%BC\" title=\"Синоним – Macedonian\" hreflang=\"mk\" lang=\"mk\">Македонски</a></li><li class=\"interlanguage-link interwiki-ml\"><a href=\"https://ml.wikipedia.org/wiki/%E0%B4%AA%E0%B4%B0%E0%B5%8D%E0%B4%AF%E0%B4%BE%E0%B4%AF%E0%B4%AA%E0%B4%A6%E0%B4%82\" title=\"പര്യായപദം – Malayalam\" hreflang=\"ml\" lang=\"ml\">മലയാളം</a></li><li class=\"interlanguage-link interwiki-ms\"><a href=\"https://ms.wikipedia.org/wiki/Sinonim\" title=\"Sinonim – Malay\" hreflang=\"ms\" lang=\"ms\">Bahasa Melayu</a></li><li class=\"interlanguage-link interwiki-nl\"><a href=\"https://nl.wikipedia.org/wiki/Synoniem_%28taalkunde%29\" title=\"Synoniem (taalkunde) – Dutch\" hreflang=\"nl\" lang=\"nl\">Nederlands</a></li><li class=\"interlanguage-link interwiki-ja\"><a href=\"https://ja.wikipedia.org/wiki/%E9%A1%9E%E7%BE%A9%E8%AA%9E\" title=\"類義語 – Japanese\" hreflang=\"ja\" lang=\"ja\">日本語</a></li><li class=\"interlanguage-link interwiki-ce\"><a href=\"https://ce.wikipedia.org/wiki/%D0%A1%D0%B8%D0%BD%D0%BE%D0%BD%D0%B8%D0%BC%D0%B0%D1%88\" title=\"Синонимаш – Chechen\" hreflang=\"ce\" lang=\"ce\">Нохчийн</a></li><li class=\"interlanguage-link interwiki-no\"><a href=\"https://no.wikipedia.org/wiki/Synonym\" title=\"Synonym – Norwegian\" hreflang=\"no\" lang=\"no\">Norsk bokmål</a></li><li class=\"interlanguage-link interwiki-nn\"><a href=\"https://nn.wikipedia.org/wiki/Synonym\" title=\"Synonym – Norwegian Nynorsk\" hreflang=\"nn\" lang=\"nn\">Norsk nynorsk</a></li><li class=\"interlanguage-link interwiki-uz\"><a href=\"https://uz.wikipedia.org/wiki/Sinonimlar\" title=\"Sinonimlar – Uzbek\" hreflang=\"uz\" lang=\"uz\">Oʻzbekcha/ўзбекча</a></li><li class=\"interlanguage-link interwiki-pfl\"><a href=\"https://pfl.wikipedia.org/wiki/Wort:Synonyme\" title=\"Wort:Synonyme – Palatine German\" hreflang=\"pfl\" lang=\"pfl\">Pälzisch</a></li><li class=\"interlanguage-link interwiki-pl\"><a href=\"https://pl.wikipedia.org/wiki/Synonim\" title=\"Synonim – Polish\" hreflang=\"pl\" lang=\"pl\">Polski</a></li><li class=\"interlanguage-link interwiki-pt\"><a href=\"https://pt.wikipedia.org/wiki/Sin%C3%B4nimo\" title=\"Sinônimo – Portuguese\" hreflang=\"pt\" lang=\"pt\">Português</a></li><li class=\"interlanguage-link interwiki-ro\"><a href=\"https://ro.wikipedia.org/wiki/Sinonim\" title=\"Sinonim – Romanian\" hreflang=\"ro\" lang=\"ro\">Română</a></li><li class=\"interlanguage-link interwiki-qu\"><a href=\"https://qu.wikipedia.org/wiki/Kaqlla_simi\" title=\"Kaqlla simi – Quechua\" hreflang=\"qu\" lang=\"qu\">Runa Simi</a></li><li class=\"interlanguage-link interwiki-ru\"><a href=\"https://ru.wikipedia.org/wiki/%D0%A1%D0%B8%D0%BD%D0%BE%D0%BD%D0%B8%D0%BC%D1%8B\" title=\"Синонимы – Russian\" hreflang=\"ru\" lang=\"ru\">Русский</a></li><li class=\"interlanguage-link interwiki-sco\"><a href=\"https://sco.wikipedia.org/wiki/Synonym\" title=\"Synonym – Scots\" hreflang=\"sco\" lang=\"sco\">Scots</a></li><li class=\"interlanguage-link interwiki-sq\"><a href=\"https://sq.wikipedia.org/wiki/Sinonimi\" title=\"Sinonimi – Albanian\" hreflang=\"sq\" lang=\"sq\">Shqip</a></li><li class=\"interlanguage-link interwiki-simple\"><a href=\"https://simple.wikipedia.org/wiki/Synonym\" title=\"Synonym – Simple English\" hreflang=\"simple\" lang=\"simple\">Simple English</a></li><li class=\"interlanguage-link interwiki-sk\"><a href=\"https://sk.wikipedia.org/wiki/Synonymum\" title=\"Synonymum – Slovak\" hreflang=\"sk\" lang=\"sk\">Slovenčina</a></li><li class=\"interlanguage-link interwiki-sl\"><a href=\"https://sl.wikipedia.org/wiki/Sopomenka\" title=\"Sopomenka – Slovenian\" hreflang=\"sl\" lang=\"sl\">Slovenščina</a></li><li class=\"interlanguage-link interwiki-sr\"><a href=\"https://sr.wikipedia.org/wiki/%D0%A1%D0%B8%D0%BD%D0%BE%D0%BD%D0%B8%D0%BC\" title=\"Синоним – Serbian\" hreflang=\"sr\" lang=\"sr\">Српски / srpski</a></li><li class=\"interlanguage-link interwiki-sh\"><a href=\"https://sh.wikipedia.org/wiki/Sinonim\" title=\"Sinonim – Serbo-Croatian\" hreflang=\"sh\" lang=\"sh\">Srpskohrvatski / српскохрватски</a></li><li class=\"interlanguage-link interwiki-su\"><a href=\"https://su.wikipedia.org/wiki/Sinonim\" title=\"Sinonim – Sundanese\" hreflang=\"su\" lang=\"su\">Basa Sunda</a></li><li class=\"interlanguage-link interwiki-fi\"><a href=\"https://fi.wikipedia.org/wiki/Synonymia\" title=\"Synonymia – Finnish\" hreflang=\"fi\" lang=\"fi\">Suomi</a></li><li class=\"interlanguage-link interwiki-sv\"><a href=\"https://sv.wikipedia.org/wiki/Synonym\" title=\"Synonym – Swedish\" hreflang=\"sv\" lang=\"sv\">Svenska</a></li><li class=\"interlanguage-link interwiki-ta\"><a href=\"https://ta.wikipedia.org/wiki/%E0%AE%92%E0%AE%A4%E0%AF%8D%E0%AE%A4%E0%AE%9A%E0%AF%8A%E0%AE%B2%E0%AF%8D\" title=\"ஒத்தசொல் – Tamil\" hreflang=\"ta\" lang=\"ta\">தமிழ்</a></li><li class=\"interlanguage-link interwiki-tt\"><a href=\"https://tt.wikipedia.org/wiki/%D0%A1%D0%B8%D0%BD%D0%BE%D0%BD%D0%B8%D0%BC\" title=\"Синоним – Tatar\" hreflang=\"tt\" lang=\"tt\">Татарча/tatarça</a></li><li class=\"interlanguage-link interwiki-te\"><a href=\"https://te.wikipedia.org/wiki/%E0%B0%AA%E0%B0%B0%E0%B1%8D%E0%B0%AF%E0%B0%BE%E0%B0%AF%E0%B0%AA%E0%B0%A6%E0%B0%82\" title=\"పర్యాయపదం – Telugu\" hreflang=\"te\" lang=\"te\">తెలుగు</a></li><li class=\"interlanguage-link interwiki-tr\"><a href=\"https://tr.wikipedia.org/wiki/E%C5%9Fanlaml%C4%B1\" title=\"Eşanlamlı – Turkish\" hreflang=\"tr\" lang=\"tr\">Türkçe</a></li><li class=\"interlanguage-link interwiki-uk\"><a href=\"https://uk.wikipedia.org/wiki/%D0%A1%D0%B8%D0%BD%D0%BE%D0%BD%D1%96%D0%BC\" title=\"Синонім – Ukrainian\" hreflang=\"uk\" lang=\"uk\">Українська</a></li><li class=\"interlanguage-link interwiki-ur\"><a href=\"https://ur.wikipedia.org/wiki/%D9%85%D8%AA%D8%B1%D8%A7%D8%AF%D9%81\" title=\"مترادف – Urdu\" hreflang=\"ur\" lang=\"ur\">اردو</a></li><li class=\"interlanguage-link interwiki-vi\"><a href=\"https://vi.wikipedia.org/wiki/T%E1%BB%AB_%C4%91%E1%BB%93ng_ngh%C4%A9a\" title=\"Từ đồng nghĩa – Vietnamese\" hreflang=\"vi\" lang=\"vi\">Tiếng Việt</a></li><li class=\"interlanguage-link interwiki-wa\"><a href=\"https://wa.wikipedia.org/wiki/Sinonimeye\" title=\"Sinonimeye – Walloon\" hreflang=\"wa\" lang=\"wa\">Walon</a></li><li class=\"interlanguage-link interwiki-zh-yue\"><a href=\"https://zh-yue.wikipedia.org/wiki/%E5%90%8C%E7%BE%A9%E8%A9%9E\" title=\"同義詞 – Cantonese\" hreflang=\"zh-yue\" lang=\"zh-yue\">粵語</a></li><li class=\"interlanguage-link interwiki-zh\"><a href=\"https://zh.wikipedia.org/wiki/%E5%90%8C%E4%B9%89%E8%AF%8D\" title=\"同义词 – Chinese\" hreflang=\"zh\" lang=\"zh\">中文</a></li>          </ul>\r\n        <div class=\"after-portlet after-portlet-lang\"><span class=\"wb-langlinks-edit wb-langlinks-link\"><a href=\"https://www.wikidata.org/wiki/Q42106#sitelinks-wikipedia\" title=\"Edit interlanguage links\" class=\"wbc-editpage\">Edit links</a></span></div>      </div>\r\n    </div>\r\n        </div>\r\n    </div>\r\n    <div id=\"footer\" role=\"contentinfo\">\r\n              <ul id=\"footer-info\">\r\n                      <li id=\"footer-info-lastmod\"> This page was last modified on 10 November 2015, at 18:44.</li>\r\n                      <li id=\"footer-info-copyright\">Text is available under the <a rel=\"license\" href=\"https://en.wikipedia.org/wiki/Wikipedia:Text_of_Creative_Commons_Attribution-ShareAlike_3.0_Unported_License\">Creative Commons Attribution-ShareAlike License</a>;\r\nadditional terms may apply.  By using this site, you agree to the <a href=\"https://wikimediafoundation.org/wiki/Terms_of_Use\">Terms of Use</a> and <a href=\"https://wikimediafoundation.org/wiki/Privacy_policy\">Privacy Policy</a>. Wikipedia® is a registered trademark of the <a href=\"https://www.wikimediafoundation.org/\">Wikimedia Foundation, Inc.</a>, a non-profit organization.</li>\r\n                  </ul>\r\n              <ul id=\"footer-places\">\r\n                      <li id=\"footer-places-privacy\"><a href=\"https://wikimediafoundation.org/wiki/Privacy_policy\" title=\"wmf:Privacy policy\">Privacy policy</a></li>\r\n                      <li id=\"footer-places-about\"><a href=\"https://en.wikipedia.org/wiki/Wikipedia:About\" title=\"Wikipedia:About\">About Wikipedia</a></li>\r\n                      <li id=\"footer-places-disclaimer\"><a href=\"https://en.wikipedia.org/wiki/Wikipedia:General_disclaimer\" title=\"Wikipedia:General disclaimer\">Disclaimers</a></li>\r\n                      <li id=\"footer-places-contact\"><a href=\"https://en.wikipedia.org/wiki/Wikipedia:Contact_us\">Contact Wikipedia</a></li>\r\n                      <li id=\"footer-places-developers\"><a href=\"https://www.mediawiki.org/wiki/Special:MyLanguage/How_to_contribute\">Developers</a></li>\r\n                      <li id=\"footer-places-mobileview\"><a href=\"https://en.m.wikipedia.org/w/index.php?title=Synonym&amp;mobileaction=toggle_view_mobile\" class=\"noprint stopMobileRedirectToggle\">Mobile view</a></li>\r\n                  </ul>\r\n                    <ul id=\"footer-icons\" class=\"noprint\">\r\n                      <li id=\"footer-copyrightico\">\r\n              <a href=\"https://wikimediafoundation.org/\"><img src=\"https://en.wikipedia.org/static/images/wikimedia-button.png\" srcset=\"/static/images/wikimedia-button-1.5x.png 1.5x, /static/images/wikimedia-button-2x.png 2x\" alt=\"Wikimedia Foundation\" height=\"31\" width=\"88\"></a>            </li>\r\n                      <li id=\"footer-poweredbyico\">\r\n              <a href=\"https://www.mediawiki.org/\"><img src=\"https://en.wikipedia.org/static/1.27.0-wmf.7/resources/assets/poweredby_mediawiki_88x31.png\" alt=\"Powered by MediaWiki\" srcset=\"/static/1.27.0-wmf.7/resources/assets/poweredby_mediawiki_132x47.png 1.5x, /static/1.27.0-wmf.7/resources/assets/poweredby_mediawiki_176x62.png 2x\" height=\"31\" width=\"88\"></a>            </li>\r\n                  </ul>\r\n            <div style=\"clear:both\"></div>\r\n    </div>\r\n";

var BrowserOSXFirefoxPlainTextFixture = "<html><head><meta http-equiv=\"content-type\" content=\"text/html; charset=utf-8\"></head><body>XXX</body></html>\r\n";

var BrowserOSXFirefoxAnnotatedTextFixture = "<html><head><meta http-equiv=\"content-type\" content=\"text/html; charset=utf-8\"></head><body>X<a href=\"https://en.wikipedia.org/wiki/United_States_Senate_election_in_Illinois,_2004\" title=\"United States Senate election in Illinois, 2004\">X</a>X</body></html>\r\n";

var BrowserOSXFirefoxTwoParagraphsFixture = "<html><head><meta http-equiv=\"content-type\" content=\"text/html; charset=utf-8\"></head><body><p>AAA</p><p>BBB</p></body></html>";

var BrowserWindowsFirefoxPlainTextFixture = "<html><body>\r\n<!--StartFragment-->XXX<!--EndFragment-->\r\n</body>\r\n</html>";

var BrowserWindowsFirefoxAnnotatedTextFixture = "<html><body>\r\n<!--StartFragment-->X<a href=\"https://en.wikipedia.org/wiki/United_States_Senate_election_in_Illinois,_2004\" title=\"United States Senate election in Illinois, 2004\">X</a>X<!--EndFragment-->\r\n</body>\r\n</html>";

var BrowserWindowsFirefoxTwoParagraphsFixture = "<html><body>\r\n<!--StartFragment--><p>AAA</p><p>BBB</p><!--EndFragment-->\r\n</body>\r\n</html>";

var BrowserWindowsEdgePlainTextFixture = "<HTML class=\"client-js ve-not-available\" lang=\"en\" dir=\"ltr\"><HEAD> <TITLE>Barack Obama - Wikipedia, the free encyclopedia</TITLE>\r\n</HEAD><BODY class=\"mediawiki ltr sitedir-ltr ns-0 ns-subject page-Barack_Obama skin-vector action-view\"><DIV class=\"mw-body\" id=\"content\" role=\"main\"><DIV class=\"mw-body-content\" id=\"bodyContent\"><DIV class=\"mw-content-ltr\" id=\"mw-content-text\" lang=\"en\" dir=\"ltr\"><P><!--StartFragment-->XXX<!--EndFragment--></P></DIV></DIV></DIV></BODY></HTML>";

var BrowserWindowsEdgeAnnotatedTextFixture = "<HTML class=\"client-js ve-not-available\" lang=\"en\" dir=\"ltr\"><HEAD> <TITLE>Barack Obama - Wikipedia, the free encyclopedia</TITLE>\r\n</HEAD><BODY class=\"mediawiki ltr sitedir-ltr ns-0 ns-subject page-Barack_Obama skin-vector action-view\"><DIV class=\"mw-body\" id=\"content\" role=\"main\"><DIV class=\"mw-body-content\" id=\"bodyContent\"><DIV class=\"mw-content-ltr\" id=\"mw-content-text\" lang=\"en\" dir=\"ltr\"><P><!--StartFragment-->X<A title=\"United States Senate election in Illinois, 2004\" href=\"https://en.wikipedia.org/wiki/United_States_Senate_election_in_Illinois,_2004\">X</A>X<!--EndFragment--></P></DIV></DIV></DIV></BODY></HTML>";

var BrowserWindowsEdgeTwoParagraphsFixture = "<HTML class=\"client-js ve-not-available\" lang=\"en\" dir=\"ltr\"><HEAD> <TITLE>Barack Obama - Wikipedia, the free encyclopedia</TITLE>\r\n</HEAD><BODY class=\"mediawiki ltr sitedir-ltr ns-0 ns-subject page-Barack_Obama skin-vector action-view\"><DIV class=\"mw-body\" id=\"content\" role=\"main\"><DIV class=\"mw-body-content\" id=\"bodyContent\"><DIV class=\"mw-content-ltr\" id=\"mw-content-text\" lang=\"en\" dir=\"ltr\"><!--StartFragment--><P>AAA</P><P>BBB</P><!--EndFragment--></DIV></DIV></DIV></BODY></HTML>";

var GDocsOSXLinuxChromePlainTextFixture = "<meta http-equiv=\"content-type\" content=\"text/html; charset=utf-8\"><meta charset=\"utf-8\"><b style=\"font-weight:normal;\" id=\"docs-internal-guid-5bea85da-43d8-f5c6-8657-4da8de7325c0\"><span style=\"font-size:14px;font-family:Arial;color:#252525;background-color:#ffffff;font-weight:400;font-style:normal;font-variant:normal;text-decoration:none;vertical-align:baseline;white-space:pre-wrap;\">XXX</span></b>";

var GDocsOSXLinuxChromeAnnotatedTextFixture = "<meta http-equiv=\"content-type\" content=\"text/html; charset=utf-8\"><meta charset=\"utf-8\"><b style=\"font-weight:normal;\" id=\"docs-internal-guid-5bea85da-43db-ac2a-ad43-e477b9e4cca9\"><span style=\"font-size:14px;font-family:Arial;color:#252525;background-color:#ffffff;font-weight:400;font-style:normal;font-variant:normal;text-decoration:none;vertical-align:baseline;white-space:pre-wrap;\">X</span><a href=\"https://en.wikipedia.org/wiki/United_States_Senate_election_in_Illinois,_2004\" style=\"text-decoration:none;\"><span style=\"font-size:14px;font-family:Arial;color:#0b0080;background-color:#ffffff;font-weight:400;font-style:normal;font-variant:normal;text-decoration:none;vertical-align:baseline;white-space:pre-wrap;\">X</span></a><span style=\"font-size:14px;font-family:Arial;color:#252525;background-color:#ffffff;font-weight:400;font-style:normal;font-variant:normal;text-decoration:none;vertical-align:baseline;white-space:pre-wrap;\">X</span></b>";

var GDocsOSXLinuxChromeTwoParagraphsFixture = "<meta http-equiv=\"content-type\" content=\"text/html; charset=utf-8\"><meta charset=\"utf-8\"><b style=\"font-weight:normal;\" id=\"docs-internal-guid-5bea85da-43dc-fb06-e327-00c1c6576cf7\"><p dir=\"ltr\" style=\"line-height:1.8327272415161155;margin-top:6pt;margin-bottom:6pt;\"><span style=\"font-size:14px;font-family:Arial;color:#252525;background-color:#ffffff;font-weight:400;font-style:normal;font-variant:normal;text-decoration:none;vertical-align:baseline;white-space:pre-wrap;\">AAA</span></p><p dir=\"ltr\" style=\"line-height:1.8327272415161155;margin-top:6pt;margin-bottom:6pt;\"><span style=\"font-size:14px;font-family:Arial;color:#252525;background-color:#ffffff;font-weight:400;font-style:normal;font-variant:normal;text-decoration:none;vertical-align:baseline;white-space:pre-wrap;\">BBB</span></p></b><br class=\"Apple-interchange-newline\">";

var GDocsLinuxFirefoxPlainTextFixture = "<meta charset=\"utf-8\"><span style=\"font-size:14px;font-family:Arial;color:#252525;background-color:#ffffff;font-weight:400;font-style:normal;font-variant:normal;text-decoration:none;vertical-align:baseline;\" id=\"docs-internal-guid-af43648c-43e0-a996-6b08-68384b28b868\">XXX</span>";

var GDocsLinuxFirefoxAnnotatedTextFixture = "<meta charset=\"utf-8\"><span style=\"font-size:14px;font-family:Arial;color:#252525;background-color:#ffffff;font-weight:400;font-style:normal;font-variant:normal;text-decoration:none;vertical-align:baseline;\" id=\"docs-internal-guid-af43648c-43e2-b6e6-7296-c5f0e1288023\">X</span><a href=\"https://en.wikipedia.org/wiki/United_States_Senate_election_in_Illinois,_2004\" style=\"text-decoration:none;\"><span style=\"font-size:14px;font-family:Arial;color:#0b0080;background-color:#ffffff;font-weight:400;font-style:normal;font-variant:normal;text-decoration:none;vertical-align:baseline;\">X</span></a><span style=\"font-size:14px;font-family:Arial;color:#252525;background-color:#ffffff;font-weight:400;font-style:normal;font-variant:normal;text-decoration:none;vertical-align:baseline;\">X</span>";

var GDocsOSXFirefoxPlainTextFixture = "<html><head><meta http-equiv=\"content-type\" content=\"text/html; charset=utf-8\"></head><body><meta charset=\"utf-8\"><span style=\"font-size:14px;font-family:Arial;color:#252525;background-color:#ffffff;font-weight:400;font-style:normal;font-variant:normal;text-decoration:none;vertical-align:baseline;\" id=\"docs-internal-guid-e0b4b2ba-4421-0487-8eea-908ebdaae584\">XXX</span></body></html>";

var LibreOfficeOSXPlainTextFixture = "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\">\r\n<html>\r\n<head>\r\n  <meta http-equiv=\"content-type\" content=\"text/html; charset=utf-8\"/>\r\n  <title></title>\r\n  <meta name=\"generator\" content=\"LibreOffice 5.0.3.2 (MacOSX)\"/>\r\n  <style type=\"text/css\">\r\n    @page { margin: 0.79in }\r\n    p { margin-bottom: 0.08in }\r\n    a:link { so-language: zxx }\r\n  </style>\r\n</head>\r\n<body lang=\"en\" dir=\"ltr\">\r\n<p style=\"margin-bottom: 0in\">XXX</p>\r\n</body>\r\n</html>";

var LibreOfficeOSXAnnotatedTextFixture = "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\">\r\n<html>\r\n<head>\r\n  <meta http-equiv=\"content-type\" content=\"text/html; charset=utf-8\"/>\r\n  <title></title>\r\n  <meta name=\"generator\" content=\"LibreOffice 5.0.3.2 (MacOSX)\"/>\r\n  <style type=\"text/css\">\r\n    @page { margin: 0.79in }\r\n    p { margin-bottom: 0.08in }\r\n    a:link { so-language: zxx }\r\n  </style>\r\n</head>\r\n<body lang=\"en\" dir=\"ltr\">\r\n<p style=\"margin-bottom: 0in\">X<a href=\"https://en.wikipedia.org/wiki/United_States_Senate_election_in_Illinois,_2004\">X</a>X</p>\r\n</body>\r\n</html>";

var LibreOfficeOSXTwoParagraphsFixture = "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\">\r\n<html>\r\n<head>\r\n  <meta http-equiv=\"content-type\" content=\"text/html; charset=utf-8\"/>\r\n  <title></title>\r\n  <meta name=\"generator\" content=\"LibreOffice 5.0.3.2 (MacOSX)\"/>\r\n  <style type=\"text/css\">\r\n    @page { margin: 0.79in }\r\n    p { margin-bottom: 0.08in }\r\n    a:link { so-language: zxx }\r\n  </style>\r\n</head>\r\n<body lang=\"en\" dir=\"ltr\">\r\n<p style=\"margin-bottom: 0in\">AAA</p>\r\n<p style=\"margin-bottom: 0in\">BBB</p>\r\n</body>\r\n</html>";

var MSW11OSXPlainTextFixture = "<html xmlns:o=\"urn:schemas-microsoft-com:office:office\"\r\nxmlns:w=\"urn:schemas-microsoft-com:office:word\"\r\nxmlns:m=\"http://schemas.microsoft.com/office/2004/12/omml\"\r\nxmlns=\"http://www.w3.org/TR/REC-html40\">\r\n\r\n<head>\r\n</head>\r\n\r\n<body bgcolor=white lang=EN-US style='tab-interval:36.0pt'>\r\n<!--StartFragment--><span lang=EN style='font-size:12.0pt;font-family:Cambria;\r\nmso-ascii-theme-font:minor-latin;mso-fareast-font-family:\"ＭＳ 明朝\";mso-fareast-theme-font:\r\nminor-fareast;mso-hansi-theme-font:minor-latin;mso-bidi-font-family:\"Times New Roman\";\r\nmso-bidi-theme-font:minor-bidi;mso-ansi-language:EN;mso-fareast-language:EN-US;\r\nmso-bidi-language:AR-SA'>XXX</span><!--EndFragment-->\r\n</body>\r\n\r\n</html>";

var MSW11OSXAnnotatedTextFixture = "<html xmlns:o=\"urn:schemas-microsoft-com:office:office\"\r\nxmlns:w=\"urn:schemas-microsoft-com:office:word\"\r\nxmlns:m=\"http://schemas.microsoft.com/office/2004/12/omml\"\r\nxmlns=\"http://www.w3.org/TR/REC-html40\">\r\n\r\n<head>\r\n</head>\r\n\r\n<body bgcolor=white lang=EN-US link=blue vlink=purple style='tab-interval:36.0pt'>\r\n<!--StartFragment--><span lang=EN style='font-size:12.0pt;font-family:Cambria;\r\nmso-ascii-theme-font:minor-latin;mso-fareast-font-family:\"ＭＳ 明朝\";mso-fareast-theme-font:\r\nminor-fareast;mso-hansi-theme-font:minor-latin;mso-bidi-font-family:\"Times New Roman\";\r\nmso-bidi-theme-font:minor-bidi;mso-ansi-language:EN;mso-fareast-language:EN-US;\r\nmso-bidi-language:AR-SA'>X<a\r\nhref=\"https://en.wikipedia.org/wiki/United_States_Senate_election_in_Illinois,_2004\">X</a>X</span><!--EndFragment-->\r\n</body>\r\n\r\n</html>\r\n";

var MSW11OSXTwoParagraphsFixture = "<html xmlns:o=\"urn:schemas-microsoft-com:office:office\"\r\nxmlns:w=\"urn:schemas-microsoft-com:office:word\"\r\nxmlns:m=\"http://schemas.microsoft.com/office/2004/12/omml\"\r\nxmlns=\"http://www.w3.org/TR/REC-html40\">\r\n\r\n<head>\r\n</head>\r\n\r\n<body bgcolor=white lang=EN-US link=blue vlink=purple style='tab-interval:36.0pt'>\r\n<!--StartFragment-->\r\n\r\n<p style='margin-bottom:0cm;margin-bottom:.0001pt'><span lang=EN\r\nstyle='mso-ansi-language:EN'>AAA</p>\r\n\r\n<p style='margin-bottom:0cm;margin-bottom:.0001pt'><span lang=EN\r\nstyle='mso-ansi-language:EN'>BBB</p>\r\n\r\n<!--EndFragment-->\r\n</body>\r\n\r\n</html>";

const ParagraphHTMLConverter$1 = ParagraphPackage.ParagraphHTMLConverter;
const HeadingHTMLConverter$1 = HeadingPackage.HeadingHTMLConverter;
const StrongHTMLConverter$1 = StrongPackage.StrongHTMLConverter;
const EmphasisHTMLConverter$1 = EmphasisPackage.EmphasisHTMLConverter;
const LinkHTMLConverter$1 = LinkPackage.LinkHTMLConverter;
const CodeblockHTMLConverter$1 = CodeblockPackage.CodeblockHTMLConverter;

ClipboardTests();

if (platform.inBrowser) {
  ClipboardTests('memory');
}

function ClipboardTests(memory) {

  const test = substanceTest.module('Clipboard' + (memory ? ' [memory]' : ''), {
    before: function() {
      if (memory) platform.inBrowser = false;
    },
    after: function() {
      platform._reset();
    }
  });

  test("Copying HTML, and plain text", function(t) {
    let { editorSession, clipboard } = _fixture(t, simple);
    editorSession.setSelection({ type: 'property', path: ['p1', 'content'], startOffset: 0, endOffset: 5 });
    let event = new ClipboardEvent();
    clipboard.onCopy(event);
    let clipboardData = event.clipboardData;
    t.notNil(clipboardData.data['text/plain'], "Clipboard should contain plain text data.");
    t.notNil(clipboardData.data['text/html'], "Clipboard should contain HTML data.");
    let htmlDoc = DefaultDOMElement.parseHTML(clipboardData.data['text/html']);
    let body = htmlDoc.find('body');
    t.notNil(body, 'The copied HTML should always be a full HTML document string, containing a body element.');
    t.end();
  });

  test("Copying a property selection", function(t) {
    let { editorSession, clipboard } = _fixture(t, simple);
    editorSession.setSelection({ type: 'property', path: ['p1', 'content'], startOffset: 0, endOffset: 5 });
    let TEXT = '01234';
    let event = new ClipboardEvent();
    clipboard.onCopy(event);

    let clipboardData = event.clipboardData;
    t.equal(clipboardData.data['text/plain'], TEXT, "Plain text should be correct.");

    let htmlDoc = DefaultDOMElement.parseHTML(clipboardData.data['text/html']);
    let body = htmlDoc.find('body');
    t.equal(body.text(), TEXT, "HTML text should be correct.");
    t.end();
  });

  test("Copying a container selection", function(t) {
    let { editorSession, clipboard } = _fixture(t, simple);
    editorSession.setSelection({
      type: 'container',
      containerId: 'body',
      startPath: ['p1', 'content'],
      startOffset: 1,
      endPath: ['p3', 'content'],
      endOffset: 5
    });
    let TEXT = [
      '123456789',
      '0123456789',
      '01234'
    ];

    let event = new ClipboardEvent();
    clipboard.onCopy(event);

    let clipboardData = event.clipboardData;
    t.equal(clipboardData.data['text/plain'], TEXT.join('\n'), "Plain text should be correct.");

    let htmlDoc = DefaultDOMElement.parseHTML(clipboardData.data['text/html']);
    let elements = htmlDoc.find('body').getChildren();
    t.equal(elements.length, 3, "HTML should consist of three elements.");
    let p1 = elements[0];
    t.equal(p1.attr('data-id'), 'p1', "First element should have correct data-id.");
    t.equal(p1.text(), TEXT[0], "First element should have correct text content.");
    let p2 = elements[1];
    t.equal(p2.attr('data-id'), 'p2', "Second element should have correct data-id.");
    t.equal(p2.text(), TEXT[1], "Second element should have correct text content.");
    let p3 = elements[2];
    t.equal(p3.attr('data-id'), 'p3', "Third element should have correct data-id.");
    t.equal(p3.text(), TEXT[2], "Third element should have correct text content.");
    t.end();
  });

  test("Pasting text into ContainerEditor using 'text/plain'.", function(t) {
    let { editorSession, clipboard, doc } = _fixture(t, simple);
    editorSession.setSelection({
      type: 'property',
      path: ['p1', 'content'],
      startOffset: 1,
      containerId: 'body'
    });
    let event = new ClipboardEvent();
    event.clipboardData.setData('text/plain', 'XXX');
    clipboard.onPaste(event);
    t.equal(doc.get(['p1', 'content']), '0XXX123456789', "Plain text should be correct.");
    t.end();
  });

  test("Pasting without any data given.", function(t) {
    let { editorSession, clipboard, doc } = _fixture(t, simple);
    editorSession.setSelection({
      type: 'property',
      path: ['p1', 'content'],
      startOffset: 1,
      containerId: 'body'
    });
    let event = new ClipboardEvent();
    clipboard.onPaste(event);
    t.equal(doc.get(['p1', 'content']), '0123456789', "Text should be still the same.");
    t.end();
  });

  test("Pasting text into ContainerEditor using 'text/html'.", function(t) {
    let { editorSession, clipboard, doc } = _fixture(t, simple);
    editorSession.setSelection({
      type: 'property',
      path: ['p1', 'content'],
      startOffset: 1,
      containerId: 'body'
    });
    let TEXT = 'XXX';
    let event = new ClipboardEvent();
    event.clipboardData.setData('text/plain', TEXT);
    event.clipboardData.setData('text/html', TEXT);
    clipboard.onPaste(event);
    t.equal(doc.get(['p1', 'content']), '0XXX123456789', "Plain text should be correct.");
    t.end();
  });

  // this test revealed #700: the problem was that in source code there where
  // `"` and `'` characters which did not survive the way through HTML correctly
  test("Copy and Pasting source code.", function(t) {
    let { editorSession, clipboard, doc } = _fixture(t, simple);
    let body = doc.get('body');
    let cb = doc.create({
      type: 'codeblock',
      id: 'cb1',
      content: [
        "function hello_world() {",
        "  alert('Hello World!');",
        "}"
      ].join("\n")
    });
    body.show(cb, body.getPosition('p1')+1);
    editorSession.setSelection(doc.createSelection({
      type: 'container',
      startPath: ['p1', 'content'],
      startOffset: 1,
      endPath: ['p2', 'content'],
      endOffset: 1,
      containerId: 'body',
    }));
    let event = new ClipboardEvent();
    clipboard.onCut(event);
    let cb1 = doc.get('cb1');
    t.isNil(cb1, "Codeblock should have been cutted.");
    clipboard.onPaste(event);
    cb1 = doc.get('cb1');
    t.notNil(cb1, "Codeblock should have been pasted.");
    t.deepEqual(cb1.toJSON(), cb.toJSON(), "Codeblock should have been pasted correctly.");
    t.end();
  });

  test("Browser - Chrome (OSX/Linux) - Plain Text", function(t) {
    _plainTextTest(t, BrowserLinuxPLainTextFixture);
  });

  test("Browser - Chrome (OSX/Linux) - Annotated Text", function(t) {
    _annotatedTextTest(t, BrowserLinuxAnnotatedTextFixture);
  });

  test("Browser - Chrome (OSX/Linux) - Two Paragraphs", function(t) {
    _twoParagraphsTest(t, BrowserLinuxTwoParagraphsFixture);
  });

  test("Browser - Chrome (Windows) - Plain Text", function(t) {
    _plainTextTest(t, BrowserWindowsPlainTextFixture, 'forceWindows');
  });

  test("Browser - Chrome (Windows) - Annotated Text", function(t) {
    _annotatedTextTest(t, BrowserWindowsAnnotatedTextFixture, 'forceWindows');
  });

  test("Browser - Chrome (Windows) - Two Paragraphs", function(t) {
    _twoParagraphsTest(t, BrowserWindowsTwoParagraphsFixture, 'forceWindows');
  });

  test("Browser - Firefox (Linux) - Plain Text", function(t) {
    _plainTextTest(t, BrowserLinuxFirefoxPlainTextFixture);
  });

  test("Browser - Firefox (Linux) - Annotated Text", function(t) {
    _annotatedTextTest(t, BrowserLinuxFirefoxAnnotatedTextFixture);
  });

  test("Browser - Firefox (Linux) - Two Paragraphs", function(t) {
    _twoParagraphsTest(t, BrowserLinuxFirefoxTwoParagraphsFixture);
  });

  test("Browser - Firefox (Linux) - Whole Page", function(t) {
    let html = BrowserLinuxFirefoxWholePageFixture;
    _fixtureTest(t, html, function(doc, clipboard) {
      let event = new ClipboardEvent();
      event.clipboardData.setData('text/plain', 'XXX');
      event.clipboardData.setData('text/html', html);
      clipboard.onPaste(event);
      // make sure HTML paste succeeded, by checking against the result of plain text insertion
      t.notOk(doc.get('p1').getText() === '0XXX123456789', "HTML conversion and paste should have been successful (not fall back to plain-text).");
      t.ok(doc.get('body').getLength() > 30, 'There should be a lot of paragraphs');
      t.end();
    });
  });

  test("Browser - Firefox (OSX) - Plain Text", function(t) {
    _plainTextTest(t, BrowserOSXFirefoxPlainTextFixture);
  });

  test("Browser - Firefox (OSX) - Annotated Text", function(t) {
    _annotatedTextTest(t, BrowserOSXFirefoxAnnotatedTextFixture);
  });

  test("Browser - Firefox (OSX) - Two Paragraphs", function(t) {
    _twoParagraphsTest(t, BrowserOSXFirefoxTwoParagraphsFixture);
  });

  test("Browser - Firefox (Windows) - Plain Text", function(t) {
    _plainTextTest(t, BrowserWindowsFirefoxPlainTextFixture, 'forceWindows');
  });

  test("Browser - Firefox (Windows) - Annotated Text", function(t) {
    _annotatedTextTest(t, BrowserWindowsFirefoxAnnotatedTextFixture, 'forceWindows');
  });

  test("Browser - Firefox (Windows) - Two Paragraphs", function(t) {
    _twoParagraphsTest(t, BrowserWindowsFirefoxTwoParagraphsFixture, 'forceWindows');
  });

  test("Browser - Edge (Windows) - Plain Text", function(t) {
    _plainTextTest(t, BrowserWindowsEdgePlainTextFixture, 'forceWindows');
  });

  test("Browser - Edge (Windows) - Annotated Text", function(t) {
    _annotatedTextTest(t, BrowserWindowsEdgeAnnotatedTextFixture, 'forceWindows');
  });

  test("Browser - Edge (Windows) - Two Paragraphs", function(t) {
    _twoParagraphsTest(t, BrowserWindowsEdgeTwoParagraphsFixture, 'forceWindows');
  });

  test("GoogleDocs - Chrome (OSX/Linux) - Plain Text", function(t) {
    _plainTextTest(t, GDocsOSXLinuxChromePlainTextFixture);
  });

  test("GoogleDocs - Chrome (OSX/Linux) - Annotated Text", function(t) {
    _annotatedTextTest(t, GDocsOSXLinuxChromeAnnotatedTextFixture);
  });

  test("GoogleDocs - Chrome (OSX/Linux) - Two Paragraphs", function(t) {
    _twoParagraphsTest(t, GDocsOSXLinuxChromeTwoParagraphsFixture);
  });

  test("GoogleDocs - Firefox (Linux) - Plain Text", function(t) {
    _plainTextTest(t, GDocsLinuxFirefoxPlainTextFixture);
  });

  test("GoogleDocs - Firefox (Linux) - Annotated Text", function(t) {
    _annotatedTextTest(t, GDocsLinuxFirefoxAnnotatedTextFixture);
  });

  test("GoogleDocs - Firefox (OSX) - Plain Text", function(t) {
    _plainTextTest(t, GDocsOSXFirefoxPlainTextFixture);
  });

  test("LibreOffice (OSX/Linux) - Plain Text", function(t) {
    _plainTextTest(t, LibreOfficeOSXPlainTextFixture);
  });

  test("LibreOffice (OSX/Linux) - Annotated Text", function(t) {
    _annotatedTextTest(t, LibreOfficeOSXAnnotatedTextFixture);
  });

  test("LibreOffice (OSX/Linux) - Two Paragraphs", function(t) {
    _twoParagraphsTest(t, LibreOfficeOSXTwoParagraphsFixture);
  });

  test("Microsoft Word 11 (OSX) - Plain Text", function(t) {
    _plainTextTest(t, MSW11OSXPlainTextFixture);
  });

  test("Microsoft Word 11 (OSX) - Annotated Text", function(t) {
    _annotatedTextTest(t, MSW11OSXAnnotatedTextFixture);
  });

  test("Microsoft Word 11 (OSX) - Two Paragraphs", function(t) {
    _twoParagraphsTest(t, MSW11OSXTwoParagraphsFixture);
  });

}


let converterRegistry = new Registry({
  "html": new Registry({
    "paragraph": ParagraphHTMLConverter$1,
    "heading": HeadingHTMLConverter$1,
    "strong": StrongHTMLConverter$1,
    "emphasis": EmphasisHTMLConverter$1,
    "link": LinkHTMLConverter$1,
    "codeblock": CodeblockHTMLConverter$1,
  })
});

let clipboardConfig = {
  converterRegistry: converterRegistry
};

class ClipboardEventData {
  constructor() {
    this.data = {};
  }

  getData(format) {
    return this.data[format]
  }

  setData(format, data) {
    this.data[format] = data;
  }

  get types() {
    return Object.keys(this.data)
  }
}

class ClipboardEvent {
  constructor() {
    this.clipboardData = new ClipboardEventData();
  }
  preventDefault() {}
  stopPropagation() {}
}

function _fixture(t, seed) {
  let { editorSession, doc } = setupEditor(t, seed);
  let clipboard = new Clipboard(editorSession, clipboardConfig);
  return { editorSession, doc, clipboard }
}

function _fixtureTest(t, html, impl, forceWindows) {
  let { editorSession, clipboard, doc } = _fixture(t, simple);
  if (forceWindows) {
    // NOTE: faking 'Windows' mode in importer so that
    // the correct implementation will be used
    clipboard.htmlImporter._isWindows = true;
  }
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 1,
    containerId: 'body'
  });
  impl(doc, clipboard);
}

function _plainTextTest(t, html, forceWindows) {
  _fixtureTest(t, html, function(doc, clipboard) {
    let event = new ClipboardEvent();
    event.clipboardData.setData('text/plain', '');
    event.clipboardData.setData('text/html', html);
    clipboard.onPaste(event);
    t.equal(doc.get(['p1', 'content']), '0XXX123456789', "Content should have been pasted correctly.");
    t.end();
  }, forceWindows);
}

function _annotatedTextTest(t, html, forceWindows) {
  _fixtureTest(t, html, function(doc, clipboard) {
    let event = new ClipboardEvent();
    event.clipboardData.setData('text/plain', '');
    event.clipboardData.setData('text/html', html);
    clipboard.onPaste(event);
    t.equal(doc.get(['p1', 'content']), '0XXX123456789', "Content should have been pasted correctly.");
    let annotations = doc.getIndex('annotations').get(['p1', 'content']);
    t.equal(annotations.length, 1, "There should be one annotation on the property now.");
    let anno = annotations[0];
    t.equal(anno.type, 'link', "The annotation should be a link.");
    t.end();
  }, forceWindows);
}

function _twoParagraphsTest(t, html, forceWindows) {
  _fixtureTest(t, html, function(doc, clipboard) {
    let event = new ClipboardEvent();
    event.clipboardData.setData('text/plain', '');
    event.clipboardData.setData('text/html', html);
    clipboard.onPaste(event);
    let body = doc.get('body');
    let p1 = body.getChildAt(0);
    t.equal(p1.content, '0AAA', "First paragraph should be truncated.");
    let p2 = body.getChildAt(1);
    t.equal(p2.content, 'BBB', "Second paragraph should contain 'BBB'.");
    let p3 = body.getChildAt(2);
    t.equal(p3.content, '123456789', "Remainder of original p1 should go into forth paragraph.");
    t.end();
  }, forceWindows);
}

let EXAMPLE_CHANGE = {
  ops: [{some: 'operation'}],
  info: {
    userId: 'testuser'
  }
};

/*
  Util for seeding
*/
function _addChange(store, change) {
  return function(cb) {
    store.addChange('test-doc', change, cb);
  }
}

function testChangeStore(createEmptyChangeStore, test) {

  /*
    Add change
  */

  test('Add a first change', (t) => {
    let changeStore = createEmptyChangeStore();
    changeStore.addChange('test-doc', EXAMPLE_CHANGE, (err, version) => {
      t.notOk(err, 'Should not error');
      t.equal(version, 1, 'Initial version should be one');
      t.end();
    });
  });

  test('Add a second change', (t) => {
    let changeStore = createEmptyChangeStore();
    async.series([
      _addChange(changeStore, EXAMPLE_CHANGE)
    ], () => {
      changeStore.addChange('test-doc', EXAMPLE_CHANGE, (err, version) => {
        t.notOk(err, 'Should not error');
        t.equal(version, 2, 'Incremened version should be two');
        t.end();
      });
    });
  });

  /*
    Get changes
  */

  test('Get all changes', (t) => {
    let changeStore = createEmptyChangeStore();
    async.series([
      _addChange(changeStore, EXAMPLE_CHANGE),
      _addChange(changeStore, EXAMPLE_CHANGE)
    ], () => {
      changeStore.getChanges('test-doc', (err, changes, version) => {
        t.equal(changes.length, 2, 'There should be two changes in the db');
        t.equal(version, 2, 'Latest version should be 2');
        t.end();
      });
    });
  });

  test('Get changes since version 1 (2 changes stored)', (t) => {
    let changeStore = createEmptyChangeStore();
    async.series([
      _addChange(changeStore, EXAMPLE_CHANGE),
      _addChange(changeStore, EXAMPLE_CHANGE)
    ], () => {
      changeStore.getChanges('test-doc', 1, (err, changes, version) => {
        t.equal(changes.length, 1, 'There should be two changes since version 1');
        t.equal(version, 2, 'Doc version should be 2');
        t.end();
      });
    });
  });

  test('Get changes since version 1 up to version 2 (3 changes stored)', (t) => {
    let changeStore = createEmptyChangeStore();
    async.series([
      _addChange(changeStore, EXAMPLE_CHANGE),
      _addChange(changeStore, EXAMPLE_CHANGE),
      _addChange(changeStore, EXAMPLE_CHANGE)
    ], () => {
      changeStore.getChanges('test-doc', 1, 2, (err, changes, version) => {
        t.equal(changes.length, 1, 'There should be two changes since version 1');
        t.equal(version, 3, 'Latest version should be 3');
        t.end();
      });
    });
  });

  test('Should return no changes if sinceVersion = actual version', (t) => {
    let changeStore = createEmptyChangeStore();
    async.series([
      _addChange(changeStore, EXAMPLE_CHANGE),
    ], () => {
      changeStore.getChanges('test-doc', 1, (err, changes, version) => {
        t.notOk(err, 'Should not error');
        t.equal(changes.length, 0, 'Should have zero changes');
        t.equal(version, 1, 'Document version should be 1');
        t.end();
      });
    });
  });

  /*
    Get version
  */

  test('Return version of test-doc', function(t) {
    let changeStore = createEmptyChangeStore();
    async.series([
      _addChange(changeStore, EXAMPLE_CHANGE),
    ], () => {
      changeStore.getVersion('test-doc', (err, version) => {
        t.notOk(err, 'Should not error');
        t.equal(version, 1, 'Document version should equal 1');
        t.end();
      });
    });
  });

  test("Return version=0 if no changes are found", (t) => {
    let changeStore = createEmptyChangeStore();
    changeStore.getVersion('not-existing-doc', (err, version) => {
      t.notOk(err, 'Should not error');
      t.equal(version, 0, 'Document version should equal 0');
      t.end();
    });
  });

  /*
    Delete changes
  */

  test('Delete all changes', (t) => {
    let changeStore = createEmptyChangeStore();
    async.series([
      _addChange(changeStore, EXAMPLE_CHANGE),
      _addChange(changeStore, EXAMPLE_CHANGE)
    ], () => {
      changeStore.deleteChanges('test-doc', function(err, changeCount) {
        t.notOk(err, 'Should not error');
        t.equal(changeCount, 2, 'There should be 2 deleted change');
        changeStore.getChanges('test-doc', (err, changes, version) => {
          t.notOk(err, 'Should not error');
          t.equal(changes.length, 0, 'There should not be changes anymore');
          t.equal(version, 0, 'Document version should be 0');
          t.end();
        });
      });
    });
  });

  test("Delete changes of not existing doc", (t) => {
    let changeStore = createEmptyChangeStore();
    changeStore.deleteChanges('not-existing-doc', (err, changeCount) => {
      t.notOk(err, 'Should not error');
      t.equal(changeCount, 0, 'There should be 0 deleted changes');
      t.end();
    });
  });
}

const test$3 = substanceTest.module('collab/ChangeStore');

function createChangeStore() {
  return new ChangeStore()
}

// Runs the offical backend test suite
testChangeStore(createChangeStore, test$3);

// NOTE: this is essentially a copy of ../../collab/computeSnapshot
function computeSnapshot$1(jsonDoc, changeset) {
  jsonDoc = cloneDeep(jsonDoc);
  let nodes = jsonDoc.nodes;
  changeset.forEach((change) => {
    change.ops.forEach((opData) => {
      try {
        let op = ObjectOperation.fromJSON(opData);
        op.apply(nodes);
      } catch (err) {
        console.error(err, opData);
      }
    });
  });
  return jsonDoc
}

/*
  Create a changeset

  Allows one or more functions to be passed, which will each represent a change
  in the result.
*/
function createChangeset(doc, fns) {
  if (!doc._isDocument) {
    throw new Error('Illegal arguments')
  }
  if (!Array.isArray(fns)) {
    fns = [ fns ];
  }

  let txDoc = new TransactionDocument(doc);
  let tx = new EditingInterface(txDoc);
  let opCount = 0;
  let changes = [];

  fns.forEach((fn) => {
    fn(tx);
    let opsForChange = txDoc.ops.slice(opCount);
    changes.push(new DocumentChange(opsForChange, {}, {}).toJSON());
    opCount = txDoc.ops.length;
  });
  txDoc.dispose();
  return changes
}

var twoParagraphs = function(tx) {
  var body = tx.get('body');
  tx.create({
    id: 'p1',
    type: 'paragraph',
    content: '0123456789'
  });
  body.show('p1');

  tx.create({
    id: 'p2',
    type: 'paragraph',
    content: '0123456789'
  });
  body.show('p2');
};

const EMPTY_DOC$1 = {nodes: {}};

/*
  Some transforms for paragraph (p1)
*/
function insertText(pos, text) {
  return function(tx) {
    tx.setSelection({ type: 'property', path: ['p1', 'content'], startOffset: pos });
    tx.insertText(text);
  }
}

/*
  'test-doc' with numChanges changes and available snapshots

  @param {Number} numChanges number of available changes
  @param {Number[]} snapshots an array of version numbers
*/
function makeStoresFixture(numChanges, snapshots) {
  // available snapshots e.g. [1,2,5]
  snapshots = snapshots || [];

  let changeFns = [];
  if (numChanges > 0) changeFns.push(twoParagraphs);
  for(var i=1; i < numChanges; i++) {
    changeFns.push(insertText(3, 'foo'));
  }
  let changes = createChangeset(createTestArticle(), changeFns);
  let changeStore = new ChangeStore({
    'test-doc': changes
  });
  let snapshotSeed = {'test-doc': {}};
  snapshots.forEach((version) => {
    snapshotSeed['test-doc'][version] = computeSnapshot$1(
      EMPTY_DOC$1,
      changes.slice(0,version)
    );
  });
  let snapshotStore = new SnapshotStore(snapshotSeed);
  return {
    changeStore: changeStore,
    snapshotStore: snapshotStore
  }
}

/* eslint-disable consistent-return */
const test$4 = substanceTest.module('collab/DocumentEngine');

/*
  TODO: Test documentEngine.deleteDocument, which removes clears all changes
        and snapshots
*/

test$4('Should allow creation of a new document', function(t) {
  let documentEngine = _fixture$1(0); // 0 changes, no snapshots
  let initialChange = _getChange(1);

  documentEngine.createDocument('test-doc', initialChange, function(err, newVersion) {
    t.notOk(err, 'Should not error');
    t.equal(newVersion, 1, 'Initial version should be 1');
    documentEngine.getDocument('test-doc', (err, snapshot, headVersion) => {
      t.ok(snapshot, 'Should have a snaphot');
      t.equal(headVersion, 1, 'Head version should be 1');
      t.end();
    });
  });
});

test$4('Should allow adding a new change', function(t) {
  let documentEngine = _fixture$1(2); // 0 changes, no snapshots
  let newChange = _getChange(3);

  documentEngine.addChange('test-doc', newChange, function(err, newVersion) {
    t.notOk(err, 'Should not error');
    t.equal(newVersion, 3, 'New version should be 3');
    documentEngine.getDocument('test-doc', (err, snapshot, headVersion) => {
      t.ok(snapshot, 'Should have a snaphot');
      t.equal(headVersion, 3, 'Latest snapshot version should be 3');
      t.end();
    });
  });
});


test$4('Should be able to retrieve changes', function(t) {
  let documentEngine = _fixture$1(2); // 0 changes, no snapshots
  documentEngine.getChanges('test-doc', function(err, changes) {
    t.equal(changes.length, 2, 'Should have two changes');
    t.end();
  });
});

/*
  'test-doc' with numChanges changes and available snapshots

  @param {Number} numChanges number of available changes
  @param {Number[]} snapshots an array of version numbers
*/
function _fixture$1(numChanges, snapshots) {
  let stores = makeStoresFixture(numChanges, snapshots);
  let documentEngine = new DocumentEngine({
    changeStore: stores.changeStore,
    snapshotStore: stores.snapshotStore,
    snapshotFrequency: 1 // store snapshots for each version
  });
  return documentEngine
}

/*
  Compute a valid change
*/
function _getChange(changeNum) {
  let stores = makeStoresFixture(changeNum);
  return stores.changeStore._changes['test-doc'][changeNum-1]
}

/* eslint-disable consistent-return */
const test$5 = substanceTest.module('collab/SnapshotEngine');

function _fixture$2(numChanges, snapshots) {
  let stores = makeStoresFixture(numChanges, snapshots);
  return new SnapshotEngine(stores)
}

test$5('Should error when getting snapshot that does not exist', function(t) {
  let snapshotEngine = _fixture$2(0); // 0 changes, no snapshots
  snapshotEngine.getSnapshot('test-doc', 1, function(err, snapshot) {
    t.ok(err, 'There should be an error');
    t.notOk(snapshot);
    t.end();
  });
});

test$5('Get existing snapshot (straight)', function(t) {
  let snapshotEngine = _fixture$2(1, [1]); // 1 change, snapshot for v1
  snapshotEngine.getSnapshot('test-doc', 1, function(err, snapshot) {
    t.notOk(err, 'There should be no error');
    t.ok(snapshot);
    t.end();
  });
});

test$5('Get existing snapshot (computed from scratch)', function(t) {
  // SnapshotEngine must compute snapshot on demand
  let snapshotEngine = _fixture$2(2, [1]); // 1 change, snapshot for v1
  snapshotEngine.getSnapshot('test-doc', 2, function(err, snapshot) {
    t.notOk(err, 'There should be no error');
    t.ok(snapshot);
    t.end();
  });
});

test$5('Get existing snapshot (computed)', function(t) {
  // SnapshotEngine takes existing snapshot v1 and fetches remaining
  // changes to compute v2.
  let snapshotEngine = _fixture$2(2, [1]); // 1 change, snapshot for v1
  snapshotEngine.getSnapshot('test-doc', 2, function(err, snapshot) {
    t.notOk(err, 'There should be no error');
    t.ok(snapshot);
    t.end();
  });
});

/*
  Run the functions in the tasks collection in series.

  NOTE: You can not access results of the executed functions
*/
function series$1(tasks, cb, i) {
  i = i || 0;
  tasks[i](function(err) {
    // Always stop execution on error
    if (err) return cb(err)
    if (i === tasks.length-1) {
      cb(...arguments); // we are done
    } else {
      series$1(tasks, cb, i + 1);
    }
  });
}

let EXAMPLE_SNAPSHOT = {test: 'test'};

function testSnapshotStore(createEmptySnapshotStore, test) {

  /*
    Save snapshot
  */
  test('Save a snapshot', (t) => {
    let snapshotStore = createEmptySnapshotStore();
    snapshotStore.saveSnapshot('test-doc', 1, EXAMPLE_SNAPSHOT, (err, snapshot) => {
      t.notOk(err, 'should not error');
      t.ok(snapshot, 'stored snapshot entry expected');
      t.end();
    });
  });

  /*
    Get snapshot
  */
  test('Retrieve snapshot for test-doc', (t) => {
    let snapshotStore = createEmptySnapshotStore();

    function _create(cb) {
      snapshotStore.saveSnapshot('test-doc', 3, EXAMPLE_SNAPSHOT, cb);
    }

    function _get(cb) {
      snapshotStore.getSnapshot('test-doc', 3, cb);
    }

    function _verify(err, snapshot) {
      t.notOk(err, 'should not error');
      t.equal(snapshot, EXAMPLE_SNAPSHOT);
      t.end();
    }

    series$1([_create, _get], _verify);
  });

  /*
    Get versions
  */
  test('Get all available versions for a document', (t) => {
    let snapshotStore = createEmptySnapshotStore();

    function _createV1(cb) {
      snapshotStore.saveSnapshot('test-doc', 1, EXAMPLE_SNAPSHOT, cb);
    }

    function _createV3(cb) {
      snapshotStore.saveSnapshot('test-doc', 3, EXAMPLE_SNAPSHOT, cb);
    }

    function _getVersions(cb) {
      snapshotStore.getVersions('test-doc', cb);
    }

    function _verify(err, versions) {
      t.notOk(err, 'should not error');
      t.deepEqual(versions, ['1', '3']);
      t.end();
    }

    series$1([_createV1, _createV3, _getVersions], _verify);
  });

  /*
    Delete snapshot
  */
  test('Delete snapshot', (t) => {
    let snapshotStore = createEmptySnapshotStore();

    function _create(cb) {
      snapshotStore.saveSnapshot('test-doc', 4, EXAMPLE_SNAPSHOT, cb);
    }

    function _delete(cb) {
      snapshotStore.deleteSnapshot('test-doc', 4, cb);
    }

    function _verify(err) {
      t.notOk(err, 'should not error');
      snapshotStore.getSnapshot('test-doc', 4, (err, snapshot) => {
        t.notOk(snapshot, 'snapshot should be undefined');
        t.notOk(err, 'should not error');
        t.end();
      });
    }

    series$1([_create, _delete], _verify);
  });
}

const test$6 = substanceTest.module('collab/SnapshotStore');

function createEmptySnapshotStore() {
  return new SnapshotStore()
}

// Runs the offical backend test suite
testSnapshotStore(createEmptySnapshotStore, test$6);

class TestComponent extends Component {

  constructor(...args) {
    super(...args);
    this._enableSpies();
  }

  _enableSpies() {
    ['didMount','didUpdate','dispose','shouldRerender','render'].forEach((name) => {
      substanceTest.spy(this, name);
    });
  }

  _disableSpies() {
    ['didMount','didUpdate','dispose','shouldRerender','render'].forEach((name) => {
      this[name].restore();
    });
  }
}

TestComponent.create = function(renderFunc, props) {
  const comp = new TestComponent(null, props);
  if (renderFunc) {
    comp.render = renderFunc;
  }
  if (props) {
    comp.setProps(props);
  } else {
    comp.rerender();
  }
  return comp
};

class SimpleComponent extends TestComponent {

  render($$) {
    var el = $$('div').addClass('simple-component');
    if (this.props.children) {
      el.append(this.props.children);
    }
    return el
  }
}

TestComponent.Simple = SimpleComponent;

const Simple = TestComponent.Simple;

// regular rendering using default DOM elements
ComponentTests();

// RenderingEngine in debug mode
ComponentTests('debug');

// in the browser do an extra run on memory DOM elements
if (platform.inBrowser) {
  ComponentTests(false, 'memory');
}

function ComponentTests(debug, memory) {

  const test = substanceTest.module('Component' + (debug ? ' [debug]' : '') + (memory ? ' [memory]' : ''), {
    before: function(t) {
      substanceGlobals.DEBUG_RENDERING = Boolean(debug);
      if (memory) platform.inBrowser = false;
      t._document = DefaultDOMElement.createDocument('html');
    },
    after: function() {
      platform._reset();
    }
  });

  test("Throw error when render method is not returning an element", function(t) {
    class MyComponent extends TestComponent {
      render() {}
    }
    t.throws(function() {
      MyComponent.render();
    }, "Should throw an exception when render does not return an element");
    t.end();
  });

  test("Mounting a component", function(t) {
    // Mounting a detached element
    let doc = t._document.createDocument('html');
    let el = doc.createElement('div');
    let comp = Simple.mount(el);
    t.equal(comp.didMount.callCount, 0, "didMount must not be called when mounting to detached elements");
    // Mounting an attached element
    comp = Simple.mount(doc.firstChild);
    t.equal(comp.didMount.callCount, 1, "didMount should have been called");
    t.end();
  });

  test("Render an HTML element", function(t) {
    let comp = TestComponent.create(function($$) {
      return $$('div')
    });
    t.equal(comp.el.tagName, 'div', 'Element should be a "div".');
    comp = TestComponent.create(function($$) {
      return $$('span')
    });
    t.equal(comp.el.tagName, 'span', 'Element should be a "span".');
    t.end();
  });

  test("Render an element with attributes", function(t) {
    let comp = TestComponent.create(function($$) {
      return $$('div').attr('data-id', 'foo')
    });
    t.equal(comp.el.attr('data-id'), 'foo', 'Element should be have data-id="foo".');
    t.end();
  });

  test("Render an element with css styles", function(t) {
    let comp = TestComponent.create(function($$) {
      return $$('div').css('width', '100px')
    });
    t.equal(comp.el.css('width'), '100px', 'Element should have a css width of 100px.');
    t.end();
  });

  test("Render an element with classes", function(t) {
    let comp = TestComponent.create(function($$) {
      return $$('div').addClass('test')
    });
    t.ok(comp.el.hasClass('test'), 'Element should have class "test".');
    t.end();
  });

  test("Render an element with value", function(t) {
    let comp = TestComponent.create(function($$) {
      return $$('input').attr('type', 'text').val('foo')
    });
    t.equal(comp.el.val(), 'foo', 'Value should be set.');
    t.end();
  });

  test("Render an element with plain text", function(t) {
    let comp = TestComponent.create(function($$) {
      return $$('div').text('foo')
    });
    t.equal(comp.el.textContent, 'foo','textContent should be set.');
    t.end();
  });

  test("Render an element with custom html", function(t) {
    let comp = TestComponent.create(function($$) {
      return $$('div').html('Hello <b>World</b>')
    });
    // ATTENTION: it is important to call find() on the element API
    // not on the Component API, as Component#find will only provide
    // elements which are Component instance.
    let b = comp.el.find('b');
    t.notNil(b, 'Element should have rendered HTML as content.');
    t.equal(b.textContent, 'World','Rendered element should have right content.');
    t.end();
  });

  test("Rendering an element with HTML attributes etc.", function(t) {
    let comp = TestComponent.create(function($$) {
      return $$('div')
        .addClass('foo')
        .attr('data-id', 'foo')
        .htmlProp('type', 'foo')
    });
    t.equal(comp.el.attr('data-id'), 'foo', 'Element should have data-id="foo".');
    t.ok(comp.el.hasClass('foo'), 'Element should have class "foo".');
    t.equal(comp.el.getProperty('type'), 'foo', 'Element should have type "foo".');
    t.end();
  });

  test("Rendering an input element with value", function(t) {
    let comp = TestComponent.create(function($$) {
      return $$('input').attr('type', 'text').val('foo')
    });
    t.equal(comp.el.val(), 'foo', 'Input field should have value "foo".');
    t.end();
  });

  test("Render a component", function(t) {
    let comp = Simple.render();
    t.equal(comp.el.tagName.toLowerCase(), 'div', 'Element should be a "div".');
    t.ok(comp.el.hasClass('simple-component'), 'Element should have class "simple-component".');
    t.end();
  });

  test("Rerender on setProps()", function(t) {
    let comp = Simple.render({ foo: 'bar '});
    comp.shouldRerender.reset();
    comp.render.reset();
    comp.setProps({ foo: 'baz' });
    t.ok(comp.shouldRerender.callCount > 0, "Component should have been asked whether to rerender.");
    t.ok(comp.render.callCount > 0, "Component should have been rerendered.");
    t.end();
  });

  test("Rerendering triggers didUpdate()", function(t) {
    let comp = Simple.render({ foo: 'bar '});
    substanceTest.spy(comp, 'didUpdate');
    comp.rerender();
    t.ok(comp.didUpdate.callCount === 1, "didUpdate() should have been called once.");
    t.end();
  });

  test("Setting props triggers willReceiveProps()", function(t) {
    let comp = Simple.render({ foo: 'bar '});
    substanceTest.spy(comp, 'willReceiveProps');
    comp.setProps({ foo: 'baz' });
    t.ok(comp.willReceiveProps.callCount === 1, "willReceiveProps() should have been called once.");
    t.end();
  });

  test("Rerender on setState()", function(t) {
    let comp = Simple.render();
    comp.shouldRerender.reset();
    comp.render.reset();
    comp.setState({ foo: 'baz' });
    t.ok(comp.shouldRerender.callCount > 0, "Component should have been asked whether to rerender.");
    t.ok(comp.render.callCount > 0, "Component should have been rerendered.");
    t.end();
  });

  test("Setting state triggers willUpdateState()", function(t) {
    let comp = Simple.render();
    substanceTest.spy(comp, 'willUpdateState');
    comp.setState({ foo: 'baz' });
    t.ok(comp.willUpdateState.callCount === 1, "willUpdateState() should have been called once.");
    t.end();
  });

  test("Trigger didUpdate() when state or props have changed even with shouldRerender() = false", function(t) {
    class A extends Component {
      shouldRerender() {
        return false
      }
      render($$) {
        return $$('div')
      }
    }
    let comp = A.render();
    substanceTest.spy(comp, 'didUpdate');
    // component will not rerender but still should trigger didUpdate()
    comp.setProps({foo: 'bar'});
    t.ok(comp.didUpdate.callCount === 1, "comp.didUpdate() should have been called once.");
    comp.didUpdate.reset();
    comp.setState({foo: 'bar'});
    t.ok(comp.didUpdate.callCount === 1, "comp.didUpdate() should have been called once.");
    t.end();
  });

  test("Dependency-Injection", function(t) {
    class Parent extends Component {
      getChildContext() {
        let childContext = {};
        if (this.props.name) {
          childContext[this.props.name] = this.props.name;
        }
        return childContext
      }
      render($$) {
        let el = $$('div');
        // direct child
        el.append($$(Child).ref('a'));
        // indirect child
        el.append($$('div').append(
          $$(Child).ref('b')
        ));
        el.append(
          $$(Wrapper, {
            name:'bar',
            // ingested grandchild
            children: [
              $$(Child).ref('c')
            ]
          })
        );
        return el
      }
    }
    class Child extends Component {
      render($$) {
        return $$('div')
      }
    }
    class Wrapper extends Component {
      render($$) {
        return $$('div').append(this.props.children)
      }
    }

    Wrapper.prototype.getChildContext = Parent.prototype.getChildContext;

    let comp = Parent.render({name: 'foo'});
    let a = comp.refs.a;
    let b = comp.refs.b;
    let c = comp.refs.c;
    t.notNil(a.context.foo, "'a' should have a property 'foo' in its context");
    t.isNil(a.context.bar, ".. but not 'bar'");
    t.notNil(b.context.foo, "'b' should have a property 'foo' in its context");
    t.isNil(b.context.bar, ".. but not 'bar'");
    t.notNil(c.context.foo, "'c' should have a property 'foo' in its context");
    t.notNil(c.context.bar, ".. and also 'bar'");
    t.end();
  });

  /* ##################### Rerendering ##########################*/

  test("Rerendering varying content", function(t) {
    class TestComponent$$1 extends Component {
      getInitialState() {
        return { mode: 0 }
      }
      render($$) {
        let el = $$('div');
        if (this.state.mode === 0) {
          el.append(
            "Foo",
            $$('br')
          );
        } else {
          el.append(
            "Bar",
            $$('span'),
            "Baz",
            $$('br')
          );
        }
        return el
      }
    }
    let comp = TestComponent$$1.render();
    let childNodes = comp.el.getChildNodes();
    t.equal(childNodes.length, 2, '# Component should have two children in mode 0');
    t.ok(childNodes[0].isTextNode(), '__first should be a TextNode');
    t.equal(childNodes[0].textContent, 'Foo', '____with proper text content');
    t.equal(childNodes[1].tagName, 'br', '__and second should be a <br>');

    comp.setState({ mode: 1 });
    childNodes = comp.el.getChildNodes();
    t.equal(childNodes.length, 4, '# Component should have 4 children in mode 1');
    t.ok(childNodes[0].isTextNode(), '__first should be a TextNode');
    t.equal(childNodes[0].textContent, 'Bar', '____with proper text content');
    t.equal(childNodes[1].tagName, 'span', '__second should be <span>');
    t.ok(childNodes[2].isTextNode(), '__third should be a TextNode');
    t.equal(childNodes[2].textContent, 'Baz', '____with proper text content');
    t.equal(childNodes[3].tagName, 'br', '__and last should be a <br>');
    t.end();
  });

  // events are not supported by cheerio
  test("Rendering an element with click handler", function(t) {

    class ClickableComponent extends Component {
      constructor(...args) {
        super(...args);
        this.value = 0;
      }
      render($$) {
        let el = $$('a').append('Click me');
        if (this.props.method === 'instance') {
          el.on('click', this.onClick);
        } else if (this.props.method === 'anonymous') {
          el.on('click', () => {
            this.value += 10;
          });
        }
        return el
      }
      onClick() {
        this.value += 1;
      }
    }

    // first render without a click handler
    let comp = ClickableComponent.render();

    comp.click();
    t.equal(comp.value, 0, 'Handler should not have been triggered');

    comp.value = 0;
    comp.setProps({method: 'instance'});
    comp.click();
    t.equal(comp.value, 1, 'Instance method should have been triggered');
    comp.rerender();
    comp.click();
    t.equal(comp.value, 2, 'Rerendering should not add multiple listeners.');

    comp.value = 0;
    comp.setProps({method: 'anonymous'});
    comp.click();
    t.equal(comp.value, 10, 'Anonymous handler should have been triggered');
    comp.rerender();
    comp.click();
    t.equal(comp.value, 20, 'Rerendering should not add multiple listeners.');
    t.end();
  });

  test("Rendering an element with once-click handler", function(t) {
    class ClickableComponent extends Component {
      constructor(...args) {
        super(...args);
        this.clicks = 0;
      }
      render($$) {
        return $$('a').append('Click me')
          .on('click', this.onClick, this, { once: true })
      }
      onClick() {
        this.clicks += 1;
      }
    }

    let comp = ClickableComponent.render();
    comp.click();
    t.equal(comp.clicks, 1, 'Handler should have been triggered');
    comp.click();
    t.equal(comp.clicks, 1, 'Handler should not have been triggered again');
    t.end();
  });

  /* ##################### Nested Elements/Components ##########################*/

  test("Render children elements", function(t) {
    let comp = TestComponent.create(function($$) {
      return $$('div').addClass('parent')
        .append($$('div').addClass('child1'))
        .append($$('div').addClass('child2'))
    });
    t.equal(comp.el.getChildCount(), 2, 'Component should have two children.');
    t.ok(comp.el.hasClass('parent'), 'Element should have class "parent".');
    t.ok(comp.el.getChildAt(0).hasClass('child1'), 'First child should have class "child1".');
    t.ok(comp.el.getChildAt(1).hasClass('child2'), 'Second child should have class "child2".');
    t.end();
  });

  test("Render children components", function(t) {
    let comp = TestComponent.create(function($$) {
      return $$('div').append(
        $$(Simple, {
          children: ['a']
        }),
        $$(Simple, {
          children: ['b']
        })
      )
    });
    t.equal(comp.getChildCount(), 2, "Component should have two children");
    let first = comp.getChildAt(0);
    let second = comp.getChildAt(1);
    t.ok(first instanceof Simple, 'First child should be a Simple');
    t.equal(first.el.textContent, 'a', '.. and should have text "a".');
    t.ok(second instanceof Simple, 'Second child should be a Simple');
    t.equal(second.el.textContent, 'b', '.. and should have text "b".');
    t.end();
  });

  test("Render grandchildren elements", function(t) {
    let comp = TestComponent.create(function($$) {
      return $$('div').append(
        $$('div').addClass('child').append(
          $$('div').addClass('a'),
          $$('div').addClass('b')
        )
      )
    });
    t.equal(comp.getChildCount(), 1, "Component should have 1 child");
    let child = comp.getChildAt(0);
    t.equal(child.getChildCount(), 2, ".. and two grandchildren");
    let first = child.getChildAt(0);
    let second = child.getChildAt(1);
    t.ok(first.el.hasClass('a'), 'First should have class "a".');
    t.ok(second.el.hasClass('b'), 'Second should have class "b".');
    t.end();
  });


  test("Render nested elements passed via props", function(t) {
    let comp = TestComponent.create(function($$) {
      return $$('div').append(
        $$(Simple, {
          children: [
            $$('div').addClass('a'),
            $$('div').addClass('b')
          ]
        })
      )
    });
    t.equal(comp.getChildCount(), 1, "Component should have 1 child");
    let child = comp.getChildAt(0);
    t.equal(child.getChildCount(), 2, ".. and two grandchildren");
    let first = child.getChildAt(0);
    let second = child.getChildAt(1);
    t.ok(first.el.hasClass('a'), 'First grandchild should have class "a".');
    t.ok(second.el.hasClass('b'), 'Second grandchild should have class "b".');
    t.end();
  });

  test("Call didMount once when mounted", function(t) {
    class Parent extends TestComponent {
      render($$) {
        return $$('div')
          .append($$(Child, {loading: true}).ref('child'))
      }
      didMount() {
        this.refs.child.setProps({ loading: false });
      }
    }
    class Child extends TestComponent {
      render($$) {
        if (this.props.loading) {
          return $$('div').append('Loading...')
        } else {
          return $$('div').append(
            $$(Simple).ref('child')
          )
        }
      }
    }

    let comp = Parent.mount(getMountPoint(t));
    let childComp = comp.refs.child;
    let grandChildComp = childComp.refs.child;
    t.equal(childComp.didMount.callCount, 1, "Child's didMount should have been called.");
    t.notNil(grandChildComp, 'Grandchild should have been rendered');
    // t.equal(grandChildComp.didMount.callCount, 1, "Grandchild's didMount should have been called too.")

    comp.empty();
    t.equal(childComp.dispose.callCount, 1, "Child's dispose should have been called once.");
    t.equal(grandChildComp.dispose.callCount, 1, "Grandchild's dispose should have been called once.");
    t.end();
  });

  test('Propagating properties to nested components', function(t) {
    class ItemComponent extends TestComponent {
      render($$) {
        return $$('div').append(this.props.name)
      }
    }
    class CompositeComponent extends TestComponent {
      render($$) {
        let el = $$('div').addClass('composite-component');
        for (let i = 0; i < this.props.items.length; i++) {
          let item = this.props.items[i];
          el.append($$(ItemComponent, item));
        }
        return el
      }
    }

    let comp = CompositeComponent.render({
      items: [ {name: 'A'}, {name: 'B'} ]
    });
    t.equal(comp.getChildCount(), 2, 'Component should have two children.');
    t.equal(comp.getChildAt(0).el.textContent, 'A', 'First child should have text A');
    t.equal(comp.getChildAt(1).el.textContent, 'B', 'First child should have text B');

    // Now we are gonna set new props
    comp.setProps({
      items: [ {name: 'X'}, {name: 'Y'} ]
    });
    t.equal(comp.getChildCount(), 2, 'Component should have two children.');
    t.equal(comp.getChildAt(0).el.textContent, 'X', 'First child should have text X');
    t.equal(comp.getChildAt(1).el.textContent, 'Y', 'First child should have text Y');
    t.end();
  });

  test("Special nesting situation", function(t) {
    // problem was observed in TOCPanel where components (tocEntry) are ingested via dependency-injection
    // and appended to a 'div' element (tocEntries) which then was ingested into a ScrollPane.
    // The order of _capturing must be determined correctly, i.e. first the ScrollPane needs to
    // be captured, so that the parent of the 'div' element (tocEntries) is known.
    // only then the tocEntry components can be captured.
    class Parent extends TestComponent {
      render($$) {
        let el = $$('div');
        // grandchildren wrapped into a 'div' element
        let grandchildren = $$('div').append(
          $$(GrandChild, { name: 'foo' }).ref('foo'),
          $$(GrandChild, { name: 'bar' }).ref('bar')
        );
        el.append(
          // grandchildren wrapper ingested into Child component
          $$(Child, {
            children: grandchildren
          })
        );
        return el
      }
    }
    class Child extends TestComponent {
      render($$) {
        return $$('div').append(this.props.children)
      }
    }
    class GrandChild extends TestComponent {
      render($$) {
        return $$('div').append(this.props.name)
      }
    }

    let comp = Parent.render();
    let foo = comp.refs.foo;
    let bar = comp.refs.bar;
    t.notNil(foo, "Component should have a ref 'foo'.");
    t.equal(foo.el.textContent, 'foo', "foo should have textContent 'foo'");
    t.notNil(bar, "Component should have a ref 'bar'.");
    t.equal(bar.el.textContent, 'bar', "bar should have textContent 'bar'");
    t.end();
  });

  test("Special nesting situation II", function(t) {
    class Parent extends Component {
      render($$) {
        return $$('div').addClass('parent').append(
          $$(Child, {
            children: [
              $$('div').addClass('grandchild-container').append(
                $$(Grandchild).ref('grandchild')
              )
            ]
          }).ref('child')
        )
      }
    }
    class Child extends Component {
      render($$) {
        let el = $$('div').addClass('child').append(
          this.props.children
        );
        return el
      }
    }
    class Grandchild extends Component {
      render($$) {
        return $$('div').addClass('grandchild')
      }
    }

    let comp = Parent.render();
    let child = comp.refs.child;
    let grandchild = comp.refs.grandchild;
    t.notNil(child, "Child should be referenced.");
    t.notNil(grandchild, "Grandchild should be referenced.");
    comp.rerender();
    t.ok(child === comp.refs.child, "Child should have been retained.");
    t.ok(grandchild === comp.refs.grandchild, "Grandchild should have been retained.");
    t.end();
  });

  // TODO: this test reveals that our rendering algorithm is not able
  // to preserve elements when ref'd components are passed down via props.
  // In such cases, the parent already
  test("Implicit retaining in 3-level nesting situation", function(t) {
    class Parent extends Component {
      render($$) {
        // Ideally, the 'wrapper' element and Child component would be preserved automatically
        // because of the ref'd component 'grandchild' passed via props.
        // However, ATM the rendering algorithm does not 'know' about the existence
        // of the ref'd component when rendering the top-level component.
        // This would be revealed during descent when the Child component
        // is rendered. A chicken egg problem: to decide to preserve the wrapper we need
        // to have it rendered already. We need to rethink the rendering algorithm.
        // For now, we need to ref the component which we pass the ref'd component into.
        return $$('div').addClass('parent').append(
          $$('div').addClass('wrapper').append(
            $$(Child, {
              children: [
                $$(Grandchild).ref('grandchild')
              ]
            })
            // disable the next line to reveal the problem
            .ref('child')
          )
        )
      }
    }
    class Child extends Component {
      render($$) {
        let el = $$('div').addClass('child').append(
          this.props.children
        );
        return el
      }
    }
    class Grandchild extends Component {
      render($$) {
        return $$('div').addClass('grandchild')
      }
    }
    let comp = Parent.render();
    let wrapper = comp.find('.wrapper');
    comp.rerender();
    let wrapper2 = comp.find('.wrapper');
    t.ok(wrapper.el === wrapper2.el, "wrapper element should have been retained.");
    t.end();
  });

  test("Edge case: unused children", function(t) {
    class Parent extends Component {
      render($$) {
        return $$('div').append(
          $$(Child, {
            // TODO: should this element be created at all?
            children: [$$('div').ref('unused')]
          })
        )
      }
    }
    class Child extends Component {
      render($$) {
        return $$('div')
      }
    }
    let comp = Parent.render();
    t.equal(comp.el.getChildCount(), 1, "Should have 1 child");
    t.equal(comp.el.textContent, '', "textContent should be empty");
    t.end();
  });

  test("Providing a ref'd child", function(t) {
    class Parent extends Component {
      render($$) {
        return $$('div').append(
          $$(Child, {
            children: [$$(Grandchild).ref('grandchild')]
          })
        )
      }
    }
    class Child extends Component {
      render($$) {
        return $$('div').append(this.props.children)
      }
    }
    class Grandchild extends Component {
      render($$) {
        return $$('div')
      }
    }
    let parent = Parent.render();
    t.equal(parent.getChildCount(), 1, "Should have 1 child");
    let child = parent.getChildAt(0);
    t.equal(child.getChildCount(), 1, "Should have 1 grandchild");
    let grandchild = child.getChildAt(0);
    t.ok(parent.refs.grandchild === grandchild, "Grandchild should be the same as the referenced component.");
    t.ok(child.props.children[0].getComponent() === grandchild, "Grandchild should be accessible via props of child.");
    t.end();
  });

  test("Implicit retaining should not override higher-level rules", function(t) {
    // If a child component has refs, itself should not be retained without
    // being ref'd by the parent
    class Parent extends Component {
      render($$) {
        // Child is not ref'd: this means the parent is not interested in keeping
        // this instance on rerender
        return $$('div').addClass('parent').append($$(Child))
      }
    }
    class Child extends Component {
      render($$) {
        // 'foo' is ref'd, so it should be retained when rerendering on this level
        let el = $$('div').addClass('child').append(
          $$('div').addClass('foo').ref('foo')
        );
        return el
      }
    }
    let comp = Parent.render();
    let child = comp.find('.child');
    t.notNil(child, "Child should exist.");
    let foo = child.refs.foo;
    child.rerender();
    t.ok(child.refs.foo === foo, "'foo' should have been retained.");
    comp.rerender();
    let child2 = comp.find('.child');
    t.ok(child !== child2, "Child should have been renewed.");
    t.ok(foo !== child2.refs.foo, "'foo' should be different as well.");
    t.end();
  });

  test("Eventlisteners on child element", function(t) {
    class Parent extends Component {
      render($$) {
        return $$('div').append($$(Child).ref('child'))
      }
    }
    class Child extends Component {
      constructor(...args) {
        super(...args);
        this.clicks = 0;
      }
      render($$) {
        return $$('a').append('Click me').on('click', this.onClick)
      }
      onClick() {
        this.clicks++;
      }
    }

    let comp = Parent.render();
    let child = comp.refs.child;
    child.click();
    t.equal(child.clicks, 1, 'Handler should have been triggered');
    comp.rerender();
    child.clicks = 0;
    child.click();
    t.equal(child.clicks, 1, 'Handler should have been triggered');
    t.end();
  });


  /* ##################### Refs: Preserving Components ##########################*/

  test("Children without a ref are not retained", function(t) {
    let comp = TestComponent.create(function($$) {
      return $$('div').append(
        $$(Simple)
      )
    });
    let child = comp.getChildAt(0);
    comp.rerender();
    let newChild = comp.getChildAt(0);
    // as we did not apply a ref, the component should get rerendered from scratch
    t.ok(newChild !== child, 'Child component should have been renewed.');
    t.ok(newChild.el !== child.el, 'Child element should have been renewed.');
    t.end();
  });

  test("A ref must be unique in owner scope (fail on inadvertent reuse)", function(t) {
    class MyComponent extends TestComponent {
      render($$) {
        return $$('div')
          .append($$('div').ref('foo'))
          .append($$('div').ref('foo'))
      }
    }
    t.throws(function() {
      MyComponent.render();
    }, "Should throw an exception when a reference id is used multiple times");
    t.end();
  });

  test("Render a child element with ref", function(t) {
    let comp = TestComponent.create(function($$) {
      return $$('div').addClass('parent')
        .append($$('div').addClass('child').ref('foo'))
    });
    t.notNil(comp.refs.foo, 'Component should have a ref "foo".');
    t.ok(comp.refs.foo.hasClass('child'), 'Referenced component should have class "child".');
    // check that the instance is retained after rerender
    let child = comp.refs.foo;
    let el = child.getNativeElement();
    comp.rerender();
    t.ok(comp.refs.foo === child, 'Child element should have been preserved.');
    t.ok(comp.refs.foo.getNativeElement() === el, 'Native element should be the same.');
    t.end();
  });

  test("Render a child component with ref", function(t) {
    let comp = TestComponent.create(function($$) {
      return $$('div').append(
        $$(Simple).ref('foo')
      )
    });
    let child = comp.refs.foo;
    let el = child.getNativeElement();
    comp.rerender();
    t.ok(comp.refs.foo === child, 'Child component should have been preserved.');
    t.ok(comp.refs.foo.getNativeElement() === el, 'Native element should be the same.');
    t.end();
  });

  test("Rerendering a child component with ref triggers didUpdate()", function(t) {
    let comp = TestComponent.create(function($$) {
      return $$('div').append(
        $$(Simple).ref('foo')
      )
    });
    let child = comp.refs.foo;
    substanceTest.spy(child, 'didUpdate');
    comp.rerender();
    t.ok(child.didUpdate.callCount === 1, "child.didUpdate() should have been called once.");
    t.end();
  });

  test("Trigger didUpdate() on children even when shouldRerender()=false", function(t) {
    class Child extends Component {
      shouldRerender() {
        return false
      }
    }
    let comp = TestComponent.create(function($$) {
      return $$('div').append(
        // change prop randomly
        $$(Child, {foo: Date.now()}).ref('foo')
      )
    });
    let child = comp.refs.foo;
    substanceTest.spy(child, 'didUpdate');
    comp.rerender();
    t.ok(child.didUpdate.callCount === 1, "child.didUpdate() should have been called once.");
    t.end();
  });

  test("didUpdate() provides old props and old state", function(t) {
    let oldProps = null;
    let oldState = null;
    class MyComponent extends Component {
      getInitialState() {
        return {
          val: 1
        }
      }
      didUpdate(_oldProps, _oldState) {
        oldProps = _oldProps;
        oldState = _oldState;
      }
    }
    let comp = MyComponent.mount({
      val: 'a'
    }, getMountPoint(t));
    comp.setState({ val: 2 });
    t.notNil(oldProps, 'old props should have been provided');
    t.equal(oldProps.val, 'a', 'old props should contain original value');
    t.notNil(oldState, 'old state should have been provided');
    t.equal(oldState.val, 1, 'old state should contain original value');
    oldProps = null;
    oldState = null;
    comp.setProps({ val: 'b' });
    t.notNil(oldProps, 'old props should have been provided');
    t.equal(oldProps.val, 'a', 'old props should contain original value');
    t.notNil(oldState, 'old state should have been provided');
    t.equal(oldState.val, 2, 'old state should contain original value');
    t.end();
  });

  test("Refs on grandchild elements.", function(t) {
    let comp = TestComponent.create(function($$) {
      return $$('div').append(
        $$('div').append(
          $$('div').ref(this.props.grandChildRef) // eslint-disable-line no-invalid-this
        )
      )
    }, { grandChildRef: "foo"});

    t.notNil(comp.refs.foo, "Ref 'foo' should be set.");
    let foo = comp.refs.foo;
    comp.rerender();
    t.ok(foo === comp.refs.foo, "Referenced grandchild should have been retained.");
    substanceTest.spy(foo, 'dispose');
    comp.setProps({ grandChildRef: "bar" });
    t.ok(foo.dispose.callCount > 0, "Former grandchild should have been disposed.");
    t.notNil(comp.refs.bar, "Ref 'bar' should be set.");
    t.ok(foo !== comp.refs.bar, "Grandchild should have been recreated.");
    t.end();
  });

  // it happened, that a grandchild component with ref was not preserved
  test("Ref on grandchild component.", function(t) {
    class Grandchild extends TestComponent {
      render($$) {
        return $$('div').append(this.props.foo)
      }
    }

    let comp = TestComponent.create(function($$) {
      let el = $$('div');
      el.append(
        $$('div').append(
          // generating a random property making sure the grandchild gets rerendered
          $$(Grandchild, { foo: String(Date.now()) }).ref('grandchild')
        )
      );
      return el
    });
    t.notNil(comp.refs.grandchild, "Ref 'grandchild' should be set.");
    let grandchild = comp.refs.grandchild;
    comp.rerender();
    t.notNil(comp.refs.grandchild, "Ref 'grandchild' should be set.");
    t.ok(comp.refs.grandchild === grandchild, "'grandchild' should be the same");
    t.end();
  });

  test("Retain refs owned by parent but nested in child component.", function(t) {
    // Note: the child component does not know that there is a ref
    // set by the parent. Still, the component should be retained on rerender
    class Child extends TestComponent {
      render($$) {
        return $$('div').append(
          $$('div').append(this.props.children)
        )
      }
    }

    let comp = TestComponent.create(function($$) {
      let el = $$('div');
      el.append(
        $$('div').append(
          // generating a random property making sure the grandchild gets rerendered
          $$(Child).ref('child').append(
            $$('div').append('foo').ref('grandchild')
          )
        )
      );
      return el
    });
    t.notNil(comp.refs.grandchild, "Ref 'grandchild' should be set.");
    let grandchild = comp.refs.grandchild;
    comp.rerender();
    t.ok(comp.refs.grandchild === grandchild, "'grandchild' should be the same");
    t.end();
  });

  test("Should wipe a referenced component when class changes", function(t) {
    class ComponentA extends TestComponent {
      render($$) {
        return $$('div').addClass('component-a')
      }
    }
    class ComponentB extends TestComponent {
      render($$) {
        return $$('div').addClass('component-b')
      }
    }
    class MainComponent extends TestComponent {
      render($$) {
        let el = $$('div').addClass('context');
        let ComponentClass;
        if (this.props.context ==='A') {
          ComponentClass = ComponentA;
        } else {
          ComponentClass = ComponentB;
        }
        el.append($$(ComponentClass).ref('context'));
        return el
      }
    }

    let comp = MainComponent.render({context: 'A'});
    t.ok(comp.refs.context instanceof ComponentA, 'Context should be of instance ComponentA');
    comp.setProps({context: 'B'});
    t.ok(comp.refs.context instanceof ComponentB, 'Context should be of instance ComponentB');
    t.end();
  });

  test('Should store refs always on owners', function(t) {
    class MyComponent extends TestComponent {
      render($$) {
        return $$('div').append(
          $$(Simple).append(
            $$('div').ref('helloComp')
          ).ref('simpleComp')
        )
      }
    }
    let comp = MyComponent.render(MyComponent);
    t.ok(comp.refs.helloComp, 'There should stil be a ref to the helloComp element/component');
    t.end();
  });

  test("Implicitly retain elements when grandchild elements have refs.", function(t) {
    let comp = TestComponent.create(function($$) {
      return $$('div').append(
        $$('div').append(
          $$('div').ref(this.props.grandChildRef) // eslint-disable-line no-invalid-this
        )
      )
    }, { grandChildRef: "foo"});

    let child = comp.getChildAt(0);
    comp.rerender();
    t.ok(child === comp.getChildAt(0), "Child should be retained.");
    t.ok(child.el === comp.getChildAt(0).el, "Child element should be retained.");
    t.end();
  });

  test("Implicitly retain elements when passing grandchild with ref.", function(t) {
    class Child extends TestComponent {
      render($$) {
        return $$('div').append(
          $$('div').append(this.props.children)
        )
      }
    }

    let comp = TestComponent.create(function($$) {
      let grandchild = $$('div').ref('grandchild');
      return $$('div').append(
        $$(Child).append(grandchild)
      )
    });

    let child = comp.getChildAt(0);
    comp.rerender();
    t.ok(child === comp.getChildAt(0), "Child should be retained.");
    t.ok(child.el === comp.getChildAt(0).el, "Child element should be retained.");
    t.end();
  });

  // In ScrollPane we provied a link into the Srollbar which accesses
  // ScrollPanes ref in didUpdate()
  // This is working fine when didUpdate() is called at the right time,
  // i.e., when ScrollPane has been rendered already
  test("Everthing should be rendered when didUpdate() is triggered.", function(t) {
    let parentIsUpdated = false;
    class Parent extends Component {
      render($$) {
        return $$('div').append(
          $$(Child, {parent: this}).ref('child')
        )
      }
    }
    class Child extends Component {
      render($$) {
        return $$('div')
      }
      didUpdate() {
        if (this.props.parent.el) {
          parentIsUpdated = true;
        }
      }
    }

    let comp = Parent.render();
    // didUpdate() should not have been called (no rerender)
    t.notOk(parentIsUpdated, 'Initially child.didUpdate() should not have been called.');
    comp.rerender();
    t.ok(parentIsUpdated, 'After rerender child.didUpdate() should have access to parent.el');
    t.end();
  });

  /* ##################### Integration tests / Issues ##########################*/

  test('Preserve components when ref matches and rerender when props changed', function(t) {

    class ItemComponent extends TestComponent {
      shouldRerender(nextProps) {
        return !isEqual(nextProps, this.props)
      }
      render($$) {
        return $$('div').append(this.props.name)
      }
    }
    class CompositeComponent extends TestComponent {
      render($$) {
        let el = $$('div').addClass('composite-component');
        this.props.items.forEach(function(item) {
          el.append($$(ItemComponent, {name: item.name}).ref(item.ref));
        });
        return el
      }
    }

    // Initial mount
    let comp = CompositeComponent.render({
      items: [
        {ref: 'a', name: 'A'},
        {ref: 'b', name: 'B'},
        {ref: 'c', name: 'C'}
      ]
    });

    let a = comp.refs.a;
    let b = comp.refs.b;
    let c = comp.refs.c;

    let childNodes = comp.childNodes;
    t.equal(childNodes.length, 3, 'Component should have 3 children.');
    t.equal(childNodes[0].textContent, 'A', '.. first child should have text A');
    t.equal(childNodes[1].textContent, 'B', '.. second child should have text B');
    t.equal(childNodes[2].textContent, 'C', '.. third child should have text C');

    comp.refs.a.render.reset();
    comp.refs.b.render.reset();

    // Props update that preserves some of our components, drops some others
    // and adds some new
    comp.setProps({
      items: [
        {ref: 'a', name: 'X'}, // preserved (props changed)
        {ref: 'd', name: 'Y'}, // new
        {ref: 'b', name: 'B'}, // preserved (same props)
        {ref: 'e', name: 'Z'}  // new
      ]
    });

    childNodes = comp.childNodes;
    t.equal(childNodes.length, 4, 'Component should now have 4 children.');
    // a and b should have been preserved
    t.equal(a, comp.refs.a, '.. a should be the same instance');
    t.equal(b, comp.refs.b, '.. b should be the same component instance');
    // c should be gone
    t.equal(c.dispose.callCount, 1, '.. c should have been unmounted');
    // a should have been rerendered (different props) while b should not (same props)
    t.ok(a.render.callCount > 0, '.. Component a should have been rerendered');
    t.equal(b.render.callCount, 0, '.. Component b should not have been rerendered');
    // check content
    t.equal(childNodes[0].textContent, 'X', '.. first child should have text X');
    t.equal(childNodes[1].textContent, 'Y', '.. second child should have text Y');
    t.equal(childNodes[2].textContent, 'B', '.. third child should have text Y');
    t.equal(childNodes[3].textContent, 'Z', '.. fourth child should have text Z');
    t.end();
  });

  // Note: this is more of an integration test, but I did not manage to isolate the error
  // maybe the solution gets us closer to what actually went wrong.
  // TODO: try to split into useful smaller pieces.
  test("Unspecific integration test:  ref'd component must be retained", function(t) {
    class ComponentWithRefs extends Component {
      getInitialState() {
        return {contextId: 'hello'}
      }
      render($$) {
        let el = $$('div').addClass('lc-lens lc-writer sc-controller');

        let workspace = $$('div').ref('workspace').addClass('le-workspace');

        workspace.append(
          // Main (left column)
          $$('div').ref('main').addClass("le-main").append(
            $$(Simple).ref('toolbar').append($$(Simple)),

            $$(Simple).ref('contentPanel').append(
              $$(Simple).ref('coverEditor'),

              // The full fledged document (ContainerEditor)
              $$("div").ref('content').addClass('document-content').append(
                $$(Simple, {
                }).ref('mainEditor')
              ),
              $$(Simple).ref('bib')
            )
          )
        );

        // Context section (right column)
        workspace.append(
          $$(Simple, {
          }).ref(this.state.contextId)
        );

        el.append(workspace);

        // Status bar
        el.append(
          $$(Simple, {}).ref('statusBar')
        );
        return el
      }
    }

    let comp = ComponentWithRefs.render();
    t.ok(comp.refs.contentPanel, 'There should be a ref to the contentPanel component');
    comp.setState({contextId: 'foo'});
    t.ok(comp.refs.contentPanel, 'There should stil be a ref to the contentPanel component');
    comp.setState({contextId: 'bar'});
    t.ok(comp.refs.contentPanel, 'There should stil be a ref to the contentPanel component');
    comp.setState({contextId: 'baz'});
    t.ok(comp.refs.contentPanel, 'There should stil be a ref to the contentPanel component');
    t.end();
  });

  test("#312: refs should be bound to the owner, not to the parent.", function(t) {
    class Child extends TestComponent {
      render($$) {
        return $$('div').append(this.props.children)
      }
    }
    class Parent extends TestComponent {
      render($$) {
        let el = $$('div');
        el.append(
          $$(Child).append(
            $$('div').ref('foo').append('foo')
          )
        );
        return el
      }
    }
    let comp = Parent.render();
    t.notNil(comp.refs.foo, 'Ref should be bound to owner.');
    t.equal(comp.refs.foo.text(), 'foo', 'Ref should point to the right component.');
    t.end();
  });

  test('#635: Relocating a preserved component', function(t) {
    class Parent extends TestComponent {
      render($$) {
        let el = $$('div');
        el.append('X');
        if (this.props.nested) {
          el.append(
            $$('div').append(
              $$(Simple).ref('foo').append('Y')
            )
          );
        } else {
          el.append(
            $$(Simple).ref('foo').append('Y')
          );
        }
        el.append('Z');
        return el
      }
    }
    let comp = Parent.render();
    t.equal(comp.refs.foo.getParent(), comp, "First 'foo' should be direct child.");
    t.equal(comp.el.textContent, 'XYZ', "... and content should be correct.");
    comp.setProps({ nested: true });
    t.equal(comp.refs.foo.getParent().getParent(), comp, "Then 'foo' should be grand-child.");
    t.equal(comp.el.textContent, 'XYZ', "... and content should be correct.");
    comp.setProps({});
    t.equal(comp.refs.foo.getParent(), comp, "At last 'foo' should be direct child again.");
    t.equal(comp.el.textContent, 'XYZ', "... and content should be correct.");
    t.end();
  });

  test("Combine props and children via append", function(t) {
    class Toolbar extends TestComponent {
      render($$) {
        let el = $$('div').append(
          $$(Simple, this.props.strong).append('Strong')
        );
        return el
      }
    }

    let toolState = {strong: {active: true}};
    Toolbar.render(toolState);
    // the original toolState object should not have been changed
    t.ok(isEqual(toolState, {strong: {active: true}}), "props object should not have been touched");
    t.end();
  });

  test("Pass-through props and add children via append", function(t) {
    class MyComponent extends TestComponent {
      render($$) {
        let el = $$('div').append(
          $$(Simple, this.props).append('Child 1')
        );
        return el
      }
    }
    let props = {foo: 'bar'};
    let comp = MyComponent.render(props);
    let simple = comp.getChildAt(0);
    t.notNil(simple, 'Should have a child component.');
    t.equal(simple.props.foo, props.foo, '.. with props past through');
    t.equal(simple.props.children.length, 1, '.. with props.children having one element');
    t.equal(simple.textContent, 'Child 1', '.. with correct text content');
    t.end();
  });
}

const test$7 = substanceTest.module('ContainerAddress');

test$7("[0,1] is before [1,0]", function(t) {
  var first = new ContainerAddress(0, 1);
  var second = new ContainerAddress(1, 0);
  t.ok(first.isBefore(second, 'strict'));
  t.ok(first.isBefore(second));
  t.end();
});

test$7("[1,0] is not before [0,1]", function(t) {
  var first = new ContainerAddress(1, 0);
  var second = new ContainerAddress(0, 1);
  t.notOk(first.isBefore(second, 'strict'));
  t.notOk(first.isBefore(second));
  t.end();
});

test$7("[0,0] is before [0,1]", function(t) {
  var first = new ContainerAddress(0, 0);
  var second = new ContainerAddress(0, 1);
  t.ok(first.isBefore(second, 'strict'));
  t.ok(first.isBefore(second));
  t.end();
});

test$7("[0,1] is not before [0,0]", function(t) {
  var first = new ContainerAddress(0, 1);
  var second = new ContainerAddress(0, 0);
  t.notOk(first.isBefore(second, 'strict'));
  t.notOk(first.isBefore(second));
  t.end();
});

test$7("[0,1] is not-strictly before [0,1]", function(t) {
  var first = new ContainerAddress(0, 1);
  var second = new ContainerAddress(0, 1);
  t.notOk(first.isBefore(second, 'strict'));
  t.ok(first.isBefore(second));
  t.end();
});

test$7("[0,1] is equal to [0,1]", function(t) {
  var first = new ContainerAddress(0, 1);
  var second = new ContainerAddress(0, 1);
  t.ok(first.isEqual(second));
  t.end();
});

const test$8 = substanceTest.module('ContainerSelection');

test$8("Creating a ContainerSelection", function(t) {
  var sel = new ContainerSelection('body',['p1', 'content'], 1, ['p2', 'content'], 2);
  t.ok(sel.isContainerSelection(), 'Should be a container selection.');
  t.ok(!sel.isNull(), 'Should not be null.');
  t.deepEqual(sel.start.path, ['p1', 'content'], 'startPath should be correct.');
  t.equal(sel.start.offset, 1, 'startOffset should be correct.');
  t.deepEqual(sel.end.path, ['p2', 'content'], 'endPath should be correct.');
  t.equal(sel.end.offset, 2, 'endOffset should be correct.');
  t.ok(!sel.isReverse(), 'Selection should not be reverse');
  t.end();
});

test$8("Collapsed ContainerSelection", function(t) {
  var sel = new ContainerSelection('body', ['p1', 'content'], 1, ['p1', 'content'], 1);
  t.ok(sel.isContainerSelection(), 'Should be a container selection.');
  t.ok(sel.isCollapsed(), 'Selection should be collapsed.');
  t.end();
});

test$8("Reverse ContainerSelection", function(t) {
  var sel = new ContainerSelection('body', ['p1', 'content'], 1, ['p1', 'content'], 3, true);
  t.ok(sel.isReverse(), 'Selection should be reverse.');
  t.end();
});

test$8("isInsideOf: strictly inside other", function(t) {
  var doc = createTestArticle(containerAnnoSample);
  var sel = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 5,
    endPath: ['p3', 'content'],
    endOffset: 4,
  });
  var other = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 4,
    endPath: ['p3', 'content'],
    endOffset: 5,
  });
  t.ok(sel.isInsideOf(other), 'should be inside');
  t.ok(sel.isInsideOf(other, true), 'should be strictly inside');
  t.end();
});

test$8("isInsideOf: not-strictly inside other", function(t) {
  var doc = createTestArticle(containerAnnoSample);
  var sel = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 5,
    endPath: ['p3', 'content'],
    endOffset: 4,
  });
  var other = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 5,
    endPath: ['p3', 'content'],
    endOffset: 5,
  });
  t.ok(sel.isInsideOf(other), 'should be inside');
  t.notOk(sel.isInsideOf(other, true), 'should not be strictly inside');
  t.end();
});

test$8("isInsideOf: inside a PropertySelection", function(t) {
  var doc = createTestArticle(containerAnnoSample);
  var sel = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 4,
    endPath: ['p1', 'content'],
    endOffset: 6,
  });
  var other = doc.createSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 1,
    endOffset: 8,
  });
  t.ok(sel.isInsideOf(other), 'should be inside');
  t.end();
});

test$8("isInsideOf: not inside", function(t) {
  var doc = createTestArticle(containerAnnoSample);
  var other = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 5,
    endPath: ['p3', 'content'],
    endOffset: 5,
  });
  // wrapping
  var sel = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 4,
    endPath: ['p3', 'content'],
    endOffset: 6,
  });
  t.notOk(sel.isInsideOf(other), 'should not be inside');
  // left-boundary not inside
  sel = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 1,
    endPath: ['p2', 'content'],
    endOffset: 2,
  });
  t.notOk(sel.isInsideOf(other), 'should not be inside');
  // right-boundary not inside
  sel = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p2', 'content'],
    startOffset: 1,
    endPath: ['p3', 'content'],
    endOffset: 7,
  });
  t.notOk(sel.isInsideOf(other), 'should not be inside');

  var nullSel = doc.createSelection(null);
  t.notOk(sel.isInsideOf(nullSel), 'should not be inside null selection');
  t.end();
});

test$8("overlaps with other ContainerSelection", function(t) {
  var doc = createTestArticle(containerAnnoSample);
  var sel = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 4,
    endPath: ['p3', 'content'],
    endOffset: 6,
  });
  // equal
  var other = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 4,
    endPath: ['p3', 'content'],
    endOffset: 6,
  });
  t.ok(sel.overlaps(other));
  // inside
  other = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 5,
    endPath: ['p3', 'content'],
    endOffset: 5,
  });
  t.ok(sel.overlaps(other));
  // contained
  other = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 2,
    endPath: ['p3', 'content'],
    endOffset: 8,
  });
  t.ok(sel.overlaps(other));
  // left
  other = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 2,
    endPath: ['p2', 'content'],
    endOffset: 1,
  });
  t.ok(sel.overlaps(other));
  // right
  other = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p2', 'content'],
    startOffset: 2,
    endPath: ['p3', 'content'],
    endOffset: 8,
  });
  t.ok(sel.overlaps(other));
  t.end();
});


test$8("Collapsing to the left", function(t) {
  var sel = new ContainerSelection('body', ['p1', 'content'], 1, ['p3', 'content'], 3);
  sel = sel.collapse('left');
  t.ok(sel.isCollapsed(), 'should be collapsed');
  t.deepEqual(sel.start.path, ['p1', 'content']);
  t.equal(sel.start.offset, 1);
  t.end();
});

test$8("Collapsing to the right", function(t) {
  var sel = new ContainerSelection('body', ['p1', 'content'], 1, ['p3', 'content'], 3);
  sel = sel.collapse('right');
  t.ok(sel.isCollapsed(), 'should be collapsed');
  t.deepEqual(sel.start.path, ['p3', 'content']);
  t.equal(sel.start.offset, 3);
  t.end();
});

test$8("Expanding: other is inside", function(t) {
  var doc = createTestArticle(containerAnnoSample);
  var sel = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 5,
    endPath: ['p3', 'content'],
    endOffset: 4,
  });
  var other = doc.createSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 6,
    endOffset: 8
  });
  var newSel = sel.expand(other);
  t.ok(newSel.equals(sel), "Selection should not have changed.");
  t.end();
});

test$8("Expand: is inside other", function(t) {
  var doc = createTestArticle(containerAnnoSample);
  var sel = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 5,
    endPath: ['p3', 'content'],
    endOffset: 4,
  });
  var other = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 3,
    endPath: ['p3', 'content'],
    endOffset: 5,
  });
  var newSel = sel.expand(other);
  t.ok(newSel.equals(other));
  t.end();
});

test$8("Expand right", function(t) {
  var doc = createTestArticle(containerAnnoSample);
  var sel = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 5,
    endPath: ['p3', 'content'],
    endOffset: 4,
  });
  var other = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p2', 'content'],
    startOffset: 1,
    endPath: ['p3', 'content'],
    endOffset: 6,
  });
  var newSel = sel.expand(other);
  t.ok(newSel.start.equals(sel.start));
  t.ok(newSel.end.equals(other.end));
  t.end();
});

test$8("Expand left", function(t) {
  var doc = createTestArticle(containerAnnoSample);
  var sel = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 5,
    endPath: ['p3', 'content'],
    endOffset: 4,
  });
  var other = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 1,
    endPath: ['p2', 'content'],
    endOffset: 2,
  });
  var newSel = sel.expand(other);
  t.ok(newSel.start.equals(other.start));
  t.ok(newSel.end.equals(sel.end));
  t.end();
});

test$8("Expand left with PropertySelection", function(t) {
  var doc = createTestArticle(containerAnnoSample);
  var sel = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 5,
    endPath: ['p3', 'content'],
    endOffset: 4,
  });
  var other = doc.createSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 1,
    endOffset: 6
  });
  var newSel = sel.expand(other);
  t.ok(newSel.start.equals(other.start));
  t.ok(newSel.end.equals(sel.end));
  t.end();
});

test$8("Expand right with PropertySelection", function(t) {
  var doc = createTestArticle(containerAnnoSample);
  var sel = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 5,
    endPath: ['p3', 'content'],
    endOffset: 4,
  });
  var other = doc.createSelection({
    type: 'property',
    path: ['p3', 'content'],
    startOffset: 1,
    endOffset: 6
  });
  var newSel = sel.expand(other);
  t.ok(newSel.start.equals(sel.start));
  t.ok(newSel.end.equals(other.end));
  t.end();
});

test$8("Truncate with other ContainerSelection", function(t) {
  var doc = createTestArticle(containerAnnoSample);
  var sel = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 5,
    endPath: ['p3', 'content'],
    endOffset: 4,
  });
  // left side overlapping
  var other = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 1,
    endPath: ['p2', 'content'],
    endOffset: 1,
  });
  var newSel = sel.truncateWith(other);
  t.ok(newSel.start.equals(other.end));
  t.ok(newSel.end.equals(sel.end));
  // right side overlapping
  other = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p2', 'content'],
    startOffset: 1,
    endPath: ['p3', 'content'],
    endOffset: 8,
  });
  newSel = sel.truncateWith(other);
  t.ok(newSel.start.equals(sel.start));
  t.ok(newSel.end.equals(other.start));
  // equal
  newSel = sel.truncateWith(sel);
  t.ok(newSel.isNull());
  // wrapping
  other = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 1,
    endPath: ['p3', 'content'],
    endOffset: 8,
  });
  newSel = sel.truncateWith(other);
  t.ok(newSel.isNull());
  // left side aligned
  other = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 5,
    endPath: ['p2', 'content'],
    endOffset: 1,
  });
  newSel = sel.truncateWith(other);
  t.ok(newSel.start.equals(other.end));
  t.ok(newSel.end.equals(sel.end));
  // right side aligned
  other = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p2', 'content'],
    startOffset: 1,
    endPath: ['p3', 'content'],
    endOffset: 4,
  });
  newSel = sel.truncateWith(other);
  t.ok(newSel.start.equals(sel.start));
  t.ok(newSel.end.equals(other.start));
  t.end();
});

test$8("containsNode: inner node", function(t) {
  var doc = createTestArticle(simple);
  var sel = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 1,
    endPath: ['p3', 'content'],
    endOffset: 3
  });
  t.ok(sel.containsNode('p2'), 'Should contain p2.');
  t.end();
});

test$8("containsNode: outer nodes", function(t) {
  var doc = createTestArticle(simple);
  var sel = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p2', 'content'],
    startOffset: 1,
    endPath: ['p3', 'content'],
    endOffset: 3
  });
  t.notOk(sel.containsNode('p1'), 'Should contain p1.');
  t.notOk(sel.containsNode('p4'), 'Should contain p4.');
  t.end();
});

test$8("containsNode: start/end is nodeFragment", function(t) {
  var doc = createTestArticle(simple);
  var sel = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1'],
    startOffset: 0,
    endPath: ['p1'],
    endOffset: 1
  });
  t.ok(sel.containsNode('p1'), 'Should contain node.');
  sel = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1'],
    startOffset: 0,
    endPath: ['p2'],
    endOffset: 1
  });
  t.ok(sel.containsNode('p2'), 'Should contain node.');
  t.end();
});

test$8("containsNode: with partial node fragment", function(t) {
  var doc = createTestArticle(simple);
  var sel = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1'],
    startOffset: 1,
    endPath: ['p2'],
    endOffset: 1
  });
  t.notOk(sel.containsNode('p1'), 'Should not contain node.');
  sel = doc.createSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1'],
    startOffset: 0,
    endPath: ['p2'],
    endOffset: 0
  });
  t.notOk(sel.containsNode('p2'), 'Should not contain node.');
  t.end();
});

if (platform.inBrowser) {
  DOMElementTests('BrowserDOMElement');
}

DOMElementTests('MemoryDOMElement');

function DOMElementTests(impl) {

  const test = substanceTest.module('DefaultDOMElement ('+impl+')', {
    before: () => {
      if (impl === 'MemoryDOMElement') platform.inBrowser = false;
    },
    after: () => {
      platform._reset();
    }
  });

  test("Wrapping a native element", (t) => {
    let doc;
    if (impl === 'BrowserDOMElement') {
      doc = window.document;
    } else {
      doc = MemoryDOMElement.createDocument('html');
    }
    let nativeEl = doc.createElement('div');
    let el = DefaultDOMElement.wrap(nativeEl);
    t.ok(el._isDOMElement, 'native element should be wrapped');
    let el2 = DefaultDOMElement.wrap(nativeEl);
    t.ok(el === el2, 'wrapping an element twice should give the same DOMElement instance');
    t.throws(() => {
      DefaultDOMElement.wrap(undefined);
    }, 'Should throw when trying to wrap nil');
    t.end();
  });

  test("Creating an HTML document", (t) => {
    const doc = DefaultDOMElement.createDocument('html');
    t.notNil(doc.find('head'), 'Document should have a <head>');
    t.notNil(doc.find('body'), '.. and a <body>');
    t.equal(doc.getContentType(), 'text/html', '.. content type should be correct');
    t.end();
  });

  test("Creating an XML document", (t) => {
    const doc = DefaultDOMElement.createDocument('xml');
    t.equal(doc.getContentType(), 'application/xml', 'Document should have correct content type');
    t.equal(doc.children.length, 0, '.. and should be empty');
    t.end();
  });

  test("Parsing a full HTML document", function(t) {
    const html = '<html><head><title>TEST</title></head><body>TEST</body></html>';
    const doc = DefaultDOMElement.parseHTML(html);
    const head = doc.find('head');
    t.notNil(head, '<head> should be there');

    const title = head.find('title');
    t.notNil(title, '<head> should contain <title>');
    t.equal(title.text(), 'TEST', '.. with correct text');

    const body = doc.find('body');
    t.notNil(body, 'document should have a <body> element.');
    t.equal(body.text(), 'TEST', '.. with correct text.');
    t.end();
  });

  test("Parsing one HTML element", function(t) {
    const html = '<p>TEST</p>';
    const p = DefaultDOMElement.parseSnippet(html, 'html');
    t.notNil(p, 'HTML should get parsed.');
    t.equal(p.tagName, 'p', '.. providing one <p> element,');
    t.equal(p.text(), 'TEST', '.. with correct content.');
    t.end();
  });

  test("Parsing multiple HTML elements", function(t) {
    const html = '<p>TEST</p><p>TEST2</p>';
    const els = DefaultDOMElement.parseSnippet(html, 'html');
    t.notNil(els, 'HTML should get parsed.');
    t.equal(els.length, 2, '.. Providing 2 elements');
    t.equal(els[0].tagName, 'p', '.. the first a <p>');
    t.equal(els[0].text(), 'TEST', '.. with correct content');
    t.equal(els[1].tagName, 'p', '.. the second a <p>');
    t.equal(els[1].text(), 'TEST2', '.. with correct content');
    t.end();
  });

  test("Parsing annotated HTML text", function(t) {
    const html = '123<b>456</b>789';
    const els = DefaultDOMElement.parseSnippet(html, 'html');
    t.equal(els.length, 3, 'there are three elements');
    t.equal(els[0].nodeType, 'text', 'first is a text node');
    t.equal(els[0].text(), '123', '... it has correct content');
    t.equal(els[1].nodeType, 'element', 'second is an element');
    t.equal(els[1].tagName, 'b', '... it is a <b>');
    t.equal(els[1].text(), '456', '... it has correct content');
    t.equal(els[2].nodeType, 'text', 'third is a text node again');
    t.equal(els[2].text(), '789', '... it has correct content');
    t.end();
  });

  test("Parsing an XML document", function(t) {
    const xml = "<mydoc><myhead><mytitle>TEST</mytitle></myhead><mybody>TEST</mybody></mydoc>";
    const doc = DefaultDOMElement.parseXML(xml);
    const head = doc.find('myhead');
    t.notNil(head, '<myhead> should be there');
    const title = head.find('mytitle');
    t.notNil(title, '<head> should contain <title>');
    t.equal(title.text(), 'TEST', '.. with correct text');
    const body = doc.find('mybody');
    t.notNil(body, 'document should have a <body> element');
    t.equal(body.text(), 'TEST', '.. with correct text');
    t.end();
  });

  test("Parsing an incomplete HTML document", (t) => {
    const doc = DefaultDOMElement.parseHTML('<title>Foo</title><p>TEST</p>');
    const head = doc.find('head');
    const title = head.find('title');
    const body = doc.find('body');
    t.notNil(head, 'document should have a <head>');
    t.notNil(title, '<head> should contain <title>');
    t.equal(title.textContent, 'Foo', 'title should be correct');
    t.equal(body.innerHTML, '<p>TEST</p>', 'body should have contain the remaining content');
    t.end();
  });

  test("Parsing an HTML document with comments", (t) => {
    let doc = DefaultDOMElement.parseHTML('<body><!--TEST--></body>');
    let body = doc.find('body');
    let comment = body.firstChild;
    t.notNil(comment, 'there should a child element');
    t.ok(comment.isCommentNode(), '.. which should be a comment');
    t.equal(comment.textContent, 'TEST', '.. with correct content');
    // multiple lines
    doc = DefaultDOMElement.parseHTML('<body><!--\n  TEST\n  Foo\n  Bar--></body>');
    body = doc.find('body');
    comment = body.firstChild;
    t.equal(comment.textContent, '\n  TEST\n  Foo\n  Bar', 'multi-line comment should work too');
    /*
      NOTE: from looking at the default implementation in htmlparser2 it seems that they would merge subsequent comments
        but actually this is not the case. So, there is a branch in oncomment() which I do not understand
        The same with TextNodes.
        ```
        if(lastTag && lastTag.type === ElementType.Comment){
          lastTag.data += data
        }
        ```
    */
    // doc = DefaultDOMElement.parseHTML('<body><!--FOO--><!--BAR--></body>')
    // body = doc.find('body')
    // t.deepEqual(body.childNodes.map(c=>c.nodeType), ['comment','comment'], 'subsequent comments should not be merged')
    t.end();
  });

  test("Parsing an XML document with CDATA", (t) => {
    let foo = DefaultDOMElement.parseSnippet(`<foo><![CDATA[TEST]]></foo>`, 'xml');
    let cdata = foo.firstChild;
    t.notNil(cdata, 'there should a child element');
    t.equal(cdata.nodeType, 'cdata', '.. which should be CDATA');
    t.equal(cdata.textContent, 'TEST', '.. with correct content');
    t.end();
  });

  test("Parsing void and self-closing elements", (t) => {
    t.throws(() => {
      DefaultDOMElement.parseSnippet(`<foo>`, 'xml');
    }, 'should not allow void elements in XML');
    t.doesNotThrow(() => {
      DefaultDOMElement.parseSnippet(`<foo />`, 'xml');
    }, 'should allow self-closing elements in XML');
    t.doesNotThrow(() => {
      DefaultDOMElement.parseSnippet(`<input>`, 'html');
      DefaultDOMElement.parseSnippet(`<input />`, 'html');
    }, 'in HTML there are some allowed void elements, such as <input>');
    t.end();
  });

}

const test$9 = substanceTest.module('Document');

test$9("Create null selection.", function(t) {
  let doc = createTestArticle(simple);
  let sel = doc.createSelection(null);
  t.ok(sel.isNull(), 'Selection should be null.');
  t.end();
});

test$9("Create collapsed property selection.", function(t) {
  let doc = createTestArticle(simple);
  let sel = doc.createSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 3
  });
  t.ok(sel.isPropertySelection(), 'Selection should be a property selection.');
  t.ok(sel.isCollapsed(), 'Selection should be collapsed.');
  t.deepEqual(sel.start.path, ['p1', 'content'], 'path should be correct.');
  t.deepEqual(sel.start.offset, 3, 'start offset should be correct.');
  t.end();
});

test$9("Create expanded property selection.", function(t) {
  let doc = createTestArticle(simple);
  let sel = doc.createSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 1,
    endOffset: 4
  });
  t.ok(sel.isPropertySelection(), 'Selection should be a property selection.');
  t.notOk(sel.isCollapsed(), 'Selection should not be collapsed.');
  t.deepEqual(sel.start.path, ['p1', 'content'], 'path should be correct.');
  t.deepEqual(sel.start.offset, 1, 'start offset should be correct.');
  t.deepEqual(sel.end.offset, 4, 'end offset should be correct.');
  t.end();
});

test$9("Node.toJSON() should not export undefined optional properties", function(t) {
  let doc = createTestArticle(simple);
  let p = doc.create({
    type: 'paragraph',
    id: 'p',
    content: ''
  });
  t.deepEqual(p.toJSON(), { type: 'paragraph', id: 'p', content: '' }, 'JSON should be correct.');
  t.end();
});

const test$10 = substanceTest.module('DocumentChange');

test$10("hasUpdated()", (t) => {
  let tx = setup(simple);
  let path = ['p1', 'content'];
  // updating a property
  tx.setSelection({
    type: 'property',
    path,
    startOffset: 1
  });
  tx.insertText('X');
  let change = tx.generateChange();
  t.ok(change.hasUpdated(path), 'should tell that property has been updated');
  t.ok(change.hasUpdated(path[0]), 'should tell that node has been updated');
  // deleting a node
  tx.delete('p2');
  change = tx.generateChange();
  t.ok(change.hasDeleted('p2'), 'should tell that node has been deleted');
  // update but then delete a node
  tx.setSelection({
    type: 'property',
    path,
    startOffset: 1
  });
  tx.insertText('X');
  tx.delete('p1');
  change = tx.generateChange();
  t.notOk(change.hasUpdated(path), 'should not tell that property has been updated');
  t.ok(change.hasDeleted('p1'), 'should tell that node has been deleted');
  t.end();
});

test$10("serialize() and deserialize()", function(t) {
  let tx = setup(simple);
  let path = ['p1', 'content'];
  // updating a property
  let selStart = tx.setSelection({
    type: 'property',
    path,
    startOffset: 1
  });
  tx.insertText('X');
  let change = tx.generateChange();
  let selEnd = tx.selection;
  change.before.selection = selStart;
  change.after.selection = selEnd;

  let serialized = change.serialize();
  let deserialized = DocumentChange.deserialize(serialized);

  t.deepEqual(deserialized.ops, change.ops, 'ops should be equal after deserialization');
  t.ok(selStart.equals(deserialized.before.selection), 'selection (before) should have been deserialized too');
  t.ok(selEnd.equals(deserialized.after.selection), 'selection (after) should have been deserialized too');
  t.end();
});

function setup(...seeds) {
  let config = getTestConfig();
  let doc = new Document(config.getSchema());
  doc.create({ id: 'body', type: 'container' });
  seeds.forEach(s=>s(doc, doc.get('body')));
  return new ChangeRecorder(doc)
}

const H1_TEXT = 'h1:abcdefg';

function _h1(doc, body) {
  doc.create({
    type: 'heading',
    id: 'h1',
    content: H1_TEXT,
    level: 1
  });
  body.show('h1');
}





const P1_TEXT = 'p1:abcdef';

function _p1(doc, body) {
  doc.create({
    type: 'paragraph',
    id: 'p1',
    content: P1_TEXT
  });
  body.show('p1');
}

const P2_TEXT = 'p2:ghijk';

function _p2(doc, body) {
  doc.create({
    type: 'paragraph',
    id: 'p2',
    content: P2_TEXT
  });
  body.show('p2');
}

function _empty(doc, body) {
  doc.create({
    type: 'paragraph',
    id: 'empty',
    content: ''
  });
  body.show('empty');
}

function _s1(doc) {
  doc.create({
    type: 'strong',
    id: 's1',
    start: {
      path: ['p1', 'content'],
      offset: 3,
    },
    end: {
      offset: 5
    }
  });
}

function _il1(doc) {
  let tx = new EditingInterface(doc);
  tx.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 3,
    containerId: 'body'
  });
  tx.insertInlineNode({
    type: 'test-inline-node',
    id: 'il1',
    content: 'X'
  });
}

const LI1_TEXT = 'l1-1:abcdef';
const LI2_TEXT = 'l1-2:0123456';
const LI3_TEXT = 'l1-3:ghij';

// list with two items
function _l1(doc, body) {
  doc.create({
    type: 'list',
    id: 'l1'
  });
  body.show('l1');
}

function _l11(doc) {
  const l1 = doc.get('l1');
  doc.create({
    type: 'list-item',
    id: 'l1-1',
    content: LI1_TEXT
  });
  l1.appendItem('l1-1');
}

function _l12(doc) {
  const l1 = doc.get('l1');
  doc.create({
    type: 'list-item',
    id: 'l1-2',
    content: LI2_TEXT
  });
  l1.appendItem('l1-2');
}

function _l13(doc) {
  const l1 = doc.get('l1');
  doc.create({
    type: 'list-item',
    id: 'l1-3',
    content: LI3_TEXT
  });
  l1.appendItem('l1-3');
}

function _l1_empty(doc) {
  const l1 = doc.get('l1');
  doc.create({
    type: 'list-item',
    id: 'l1-empty',
    content: ''
  });
  l1.appendItem('l1-empty');
}

function _li1plus(doc) {
  doc.set(['l1-1', 'level'], 2);
}

const LI21_TEXT = 'l2-1:abcdef';
const LI22_TEXT = 'l2-2:0123456';

function _l2(doc, body) {
  doc.create({
    type: 'list',
    id: 'l2'
  });
  body.show('l2');
}

function _l21(doc) {
  const l2 = doc.get('l2');
  doc.create({
    type: 'list-item',
    id: 'l2-1',
    content: LI21_TEXT
  });
  l2.appendItem('l2-1');
}

function _l22(doc) {
  const l2 = doc.get('l2');
  doc.create({
    type: 'list-item',
    id: 'l2-2',
    content: LI22_TEXT
  });
  l2.appendItem('l2-2');
}



function _block1(doc, body) {
  doc.create({
    type: 'test-block',
    id: 'block1'
  });
  body.show('block1');
}

function _block2(doc, body) {
  doc.create({
    type: 'test-block',
    id: 'block2'
  });
  body.show('block2');
}

const IN1_TITLE = 'TITLE';
const IN1_BODY = 'BODY';
const IN1_CAPTION = 'CAPTION';
function _in1(doc, body) {
  doc.create({
    type: 'structured-node',
    id: 'in1',
    title: IN1_TITLE,
    body: IN1_BODY,
    caption: IN1_CAPTION
  });
  body.show('in1');
}

const T_CONTENT = [['A1', 'B1'], ['A2', 'B2']];

function _t1(doc, body) {
  doc.create({type: 'table-cell', id: 't1_a1', content:T_CONTENT[0][0]});
  doc.create({type: 'table-cell', id: 't1_b1', content:T_CONTENT[0][1]});
  doc.create({type: 'table-cell', id: 't1_a2', content:T_CONTENT[1][0]});
  doc.create({type: 'table-cell', id: 't1_b2', content:T_CONTENT[1][1]});
  body.show(doc.create({
    type: 'table', id: 't1', cells: [['t1_a1', 't1_b1'], ['t1_a2', 't1_b2']]
  }));
}

function _t1_sparse(doc, body) {
  _t1(doc, body);
  doc.set(['t1', 'cells'], [['t1_a1', null], [null, 't1_b2']]);
}

const test$11 = substanceTest.module('documentHelpers');

test$11("Get annotations for selection", (t) => {
  let doc = createTestArticle(simple);
  const path = ["p1", "content"];
  doc.create({ type: 'strong', path, startOffset: 0, endOffset: 1});
  doc.create({ type: 'emphasis', path, startOffset: 2, endOffset: 3});
  doc.create({ type: 'strong', path, startOffset: 5, endOffset: 6});
  // this lies outside of selection
  doc.create({ type: 'strong', path, startOffset: 6, endOffset: 7});
  let sel = doc.createSelection({
    type: "property",
    path,
    startOffset: 0,
    endOffset: 5
  });
  let annos = documentHelpers.getPropertyAnnotationsForSelection(doc, sel);
  t.equal(annos.length, 3, "should return 3 annos for selection");
  // filtered by type
  annos = documentHelpers.getPropertyAnnotationsForSelection(doc, sel, { type: 'emphasis'});
  t.equal(annos.length, 1, "should return one anno");
  t.equal(annos[0].type, 'emphasis', ".. of type emphasis");
  t.end();
});

test$11("Get text for null selection.", (t) => {
  let doc = createTestArticle(simple);
  t.equal(documentHelpers.getTextForSelection(doc, null), "", "Should be empty for null selection.");
  t.equal(documentHelpers.getTextForSelection(doc, doc.createSelection(null)), "", "Should be empty for null selection.");
  t.end();
});

test$11("Get text for property selection", (t) => {
  let doc = createTestArticle(simple);
  let sel = doc.createSelection({
    type: "property",
    path: ["p1", "content"],
    startOffset: 0,
    endOffset: 5
  });
  t.equal(documentHelpers.getTextForSelection(doc, sel), "01234");
  t.end();
});

test$11("Get text for container selection", (t) => {
  let doc = createTestArticle(simple);
  let sel = doc.createSelection({
    type: "container",
    containerId: "body",
    startPath: ["p1", "content"],
    startOffset: 5,
    endPath: ["p2", "content"],
    endOffset: 5
  });
  t.equal(documentHelpers.getTextForSelection(doc, sel), "56789\n01234");
  t.end();
});

test$11("Get change from document", (t) => {
  let doc = createTestArticle(simple);
  doc.create({ type: 'strong', id: 's1', path: ["p1", "content"], startOffset: 0, endOffset: 1});
  let change = documentHelpers.getChangeFromDocument(doc);
  t.equal(change.ops.length, 6, 'There should be 6 operations');
  t.deepEqual(change.ops.map(op => op.type), new Array(6).fill('create'), 'all should be create ops');
  t.deepEqual(change.ops.map(op => op.path[0]), ['p1', 'p2', 'p3', 'p4', 's1', 'body'], '.. in correct order');
  t.end();
});

test$11("deleteNode()", (t) => {
  let doc = createTestArticle(simple);[_l1, _l11].forEach(f=>f(doc, doc.get('body')));
  doc.create({ type: 'strong', id: 's1', path: ["p1", "content"], startOffset: 0, endOffset: 1});
  documentHelpers.deleteNode(doc, doc.get('p1'));
  t.nil(doc.get('p1'), 'node should have been deleted');
  t.nil(doc.get('s1'), 'annotation should have been deleted too');
  documentHelpers.deleteNode(doc, doc.get('l1'));
  t.nil(doc.get('l1'), 'node should have been deleted');
  t.nil(doc.get('l1-1'), 'list item should have been deleted too');
  t.end();
});

test$11("deleteTextRange()", (t) => {
  const path = ['p1', 'content'];
  // anno is after
  let doc = createTestArticle(simple);
  let s1 = doc.create({ type: 'strong', id: 's1', path: ["p1", "content"], startOffset: 5, endOffset: 6});
  documentHelpers.deleteTextRange(doc, { path, offset: 0 }, { path, offset: 2 });
  t.deepEqual([s1.start.offset, s1.end.offset], [3,4], 'offsets should have been shifted');
  // anno is inside
  doc = createTestArticle(simple);
  s1 = doc.create({ type: 'strong', id: 's1', path: ["p1", "content"], startOffset: 3, endOffset: 4});
  documentHelpers.deleteTextRange(doc, { path, offset: 2 }, { path, offset: 5 });
  t.nil(doc.get('s1'), 'annotation should have been deleted');
  // anno.start between and anno.end after
  doc = createTestArticle(simple);
  s1 = doc.create({ type: 'strong', id: 's1', path: ["p1", "content"], startOffset: 3, endOffset: 6});
  documentHelpers.deleteTextRange(doc, { path, offset: 2 }, { path, offset: 5 });
  t.deepEqual([s1.start.offset, s1.end.offset], [2,3], 'offsets should have been updated correctly');
  // anno.start same and anno.end after
  doc = createTestArticle(simple);
  s1 = doc.create({ type: 'strong', id: 's1', path: ["p1", "content"], startOffset: 3, endOffset: 6});
  documentHelpers.deleteTextRange(doc, { path, offset: 3 }, { path, offset: 5 });
  t.deepEqual([s1.start.offset, s1.end.offset], [3,4], 'offsets should have been updated correctly');
  // anno.start before and anno.end between
  doc = createTestArticle(simple);
  s1 = doc.create({ type: 'strong', id: 's1', path: ["p1", "content"], startOffset: 1, endOffset: 3});
  documentHelpers.deleteTextRange(doc, { path, offset: 2 }, { path, offset: 4 });
  t.deepEqual([s1.start.offset, s1.end.offset], [1,2], 'offsets should have been updated correctly');
  // anno.start before and anno.end after
  doc = createTestArticle(simple);
  s1 = doc.create({ type: 'strong', id: 's1', path: ["p1", "content"], startOffset: 1, endOffset: 6});
  documentHelpers.deleteTextRange(doc, { path, offset: 2 }, { path, offset: 4 });
  t.deepEqual([s1.start.offset, s1.end.offset], [1,4], 'offsets should have been updated correctly');
  t.end();
});

test$11("deleteListRange()", (t) => {
  // first and last not entirely selected, and no item in between
  let doc = createTestArticle(simple);[_l1, _l11, _l12, _l13].forEach(f=>f(doc, doc.get('body')));
  let l1 = doc.get('l1');
  documentHelpers.deleteListRange(doc, l1,
    { path: ['l1-1', 'content'], offset: 2 },
    { path: ['l1-2', 'content'], offset: 3 }
  );
  t.nil(doc.get('l1-2'), 'second list item should have been deleted');
  t.equal(doc.get('l1-1').getText(), LI1_TEXT.slice(0,2)+LI2_TEXT.slice(3), 'list items should have been merged');

  // item in between
  doc = createTestArticle(simple)
  ;[_l1, _l11, _l12, _l13].forEach(f=>f(doc, doc.get('body')));
  l1 = doc.get('l1');
  documentHelpers.deleteListRange(doc, l1,
    { path: ['l1-1', 'content'], offset: 2 },
    { path: ['l1-3', 'content'], offset: 3 }
  );
  t.nil(doc.get('l1-2'), 'second list item should have been deleted');
  t.nil(doc.get('l1-3'), 'third list item should have been deleted');
  t.equal(doc.get('l1-1').getText(), LI1_TEXT.slice(0,2)+LI3_TEXT.slice(3), 'list items should have been merged');

  // entirely selected
  doc = createTestArticle(simple)
  ;[_l1, _l11, _l12, _l13].forEach(f=>f(doc, doc.get('body')));
  l1 = doc.get('l1');
  documentHelpers.deleteListRange(doc, l1,
    { path: ['l1-1', 'content'], offset: 0 },
    { path: ['l1-3', 'content'], offset: LI3_TEXT }
  );
  t.nil(doc.get('l1-1'), 'first list item should have been deleted');
  t.nil(doc.get('l1-2'), 'second list item should have been deleted');
  t.nil(doc.get('l1-3'), 'third list item should have been deleted');

  t.end();
});

test$11("isContainerAnnotation()", (t) => {
  let doc = createTestArticle(simple);
  t.ok(documentHelpers.isContainerAnnotation(doc, 'test-container-anno'));
  t.notOk(documentHelpers.isContainerAnnotation(doc, 'strong'));
  t.end();
});

test$11("Get container annotations for property selection", (t) => {
  let doc = createTestArticle(containerAnnoSample);
  let sel = doc.createSelection({
    type: 'property',
    path: ['p3', 'content'],
    startOffset: 1,
    endOffset: 6
  });
  let annos;
  // without options
  annos = documentHelpers.getContainerAnnotationsForSelection(doc, sel, 'body');
  t.equal(annos.length, 1, 'There should be one container anno');
  // filtered by type
  annos = documentHelpers.getContainerAnnotationsForSelection(doc, sel, 'body', {
    type: 'test-container-anno'
  });
  t.equal(annos.length, 1, 'There should be one container anno');
  t.end();
});

if (platform.inBrowser) {
  DOMElementTests$1('BrowserDOMElement');
}

DOMElementTests$1('MemoryDOMElement');

function DOMElementTests$1(impl) {

  const test = substanceTest.module('DOMElement ('+impl+')', {
    before: () => {
      if (impl === 'MemoryDOMElement') platform.inBrowser = false;
    },
    after: () => {
      platform._reset();
    }
  });

  test("getNativeElement()", (t) => {
    let doc = DefaultDOMElement.createDocument('html');
    let el = doc.createElement('div');
    const nativeEl = el.getNativeElement();
    let isNative;
    if (impl === 'BrowserDOMElement') {
      isNative = nativeEl instanceof window.Element;
    } else {
      isNative = nativeEl._isMemoryDOMElement;
    }
    t.ok(isNative, 'should return a native element.');
    t.end();
  });

  test("getNodeType()", function(t) {
    let doc = DefaultDOMElement.createDocument('html');
    t.equal(doc.getNodeType(), 'document', 'document should have correct node type');
    let el = doc.createElement('div');
    t.equal(el.getNodeType(), 'element', 'element should have correct node type');
    t.equal(el.nodeType, 'element', '.. should also work via property getter');
    t.ok(el.isElementNode(), '.. should tell it is an element node');
    el = doc.createTextNode('foo');
    t.equal(el.getNodeType(), 'text', 'textNode should have correct node type');
    t.ok(el.isTextNode(), '.. and should tell it is a text node');
    el = doc.createComment('Hallo World!');
    t.equal(el.getNodeType(), 'comment', 'comment element should have correct node type');
    t.ok(el.isCommentNode(), '.. and should tell it is a comment node');
    el = doc.createProcessingInstruction('foo', 'bar');
    t.equal(el.getNodeType(), 'directive', 'procession instruction element should have correct node type');
    // cdata only allowed in XML documents
    doc = DefaultDOMElement.createDocument('xml');
    el = doc.createCDATASection('XYZ');
    t.equal(el.getNodeType(), 'cdata', 'cdata element should have correct node type');
    t.end();
  });

  test("getTagName()", function(t) {
    const el = DefaultDOMElement.parseSnippet('<p></p>', 'html');
    t.equal(el.getTagName(), 'p', 'should have correct tagName');
    t.equal(el.tagName, 'p', 'should also work with property accessor');
    t.equal(el.nodeName, 'p', 'nodeName is an alias');
    t.end();
  });

  test("getTagName() in HTML is always lower case", function(t) {
    const el = DefaultDOMElement.parseSnippet('<DIV></DIV>', 'html');
    t.equal(el.tagName, 'div', 'should bet lower case');
    t.end();
  });

  test("getTagName() in XML does not change case", function(t) {
    const el = DefaultDOMElement.parseSnippet('<myNode></myNode>', 'xml');
    t.equal(el.tagName, 'myNode', 'case should be changed');
    t.end();
  });

  test("setTagName()", function(t) {
    const el = DefaultDOMElement.parseSnippet('<p></p>', 'html');
    // NOTE: in the browser this will be done by replacing the native element
    el.setTagName('div');
    t.equal(el.getTagName(), 'div', 'should have changed the tagName');
    el.tagName = 'h1';
    t.equal(el.getTagName(), 'h1', 'should also work via property setter');
    t.end();
  });

  test("getId()", function(t) {
    const el = DefaultDOMElement.parseSnippet('<div id="foo"></div>', 'html');
    t.equal(el.getId(), 'foo', 'should have correct id');
    t.equal(el.id, 'foo', 'should also work with property accessor');
    t.end();
  });

  test("setId()", function(t) {
    const el = DefaultDOMElement.parseSnippet('<div id="foo"></div>', 'html');
    el.setId('bar');
    t.equal(el.getId(), 'bar', 'should have changed id');
    el.id = 'baz';
    t.equal(el.getId(), 'baz', 'should also work using property setter');
    t.end();
  });

  test("setTagName", function(t) {
    var el = DefaultDOMElement.parseSnippet('<p class="foo">ABC<b>DEF</b>GHI</p>', 'html');
    var onClick = substanceTest.spy();
    el.on('click', onClick);
    // this call is brutal as a new element needs to be created
    // and all the content and attributes be copied over
    el.setTagName('h1');
    t.equal(el.tagName, 'h1', 'Now the element should be a heading');
    t.equal(el.textContent, 'ABCDEFGHI', '.. its text content should have been preserved');
    t.equal(el.getChildCount(), 3, '.. and its children should still be there');
    t.notNil(el.find('b'), '.. including the <b> element');
    el.click();
    t.equal(onClick.callCount, 1, '.. and even the click handler should still work');
    t.end();
  });

  test("setTagName on XML should create XML elements", function(t) {
    var el = DefaultDOMElement.parseSnippet('<dummy></dummy>', 'xml');
    // this call is brutal as a new element needs to be created
    // and all the content and attributes be copied over
    el.setTagName('foo');
    t.equal(el.getNativeElement().ownerDocument.contentType, 'application/xml', 'Element should still be an XML element');
    el.setInnerHTML('<link>foo</link>');
    // when using an HTML element <link> will get exported as self-closing
    t.equal(el.outerHTML, '<foo><link>foo</link></foo>');
    t.end();
  });

  test("hasClass()", function(t) {
    const el = DefaultDOMElement.parseSnippet('<div class="foo">TEST</div>', 'html');
    t.ok(el.hasClass('foo'), 'should have class "foo".');
    t.end();
  });

  test("addClass()", function(t) {
    let doc = DefaultDOMElement.createDocument('html');
    const el = doc.createElement('div');
    el.addClass('foo');
    t.ok(el.hasClass('foo'), 'should have class "foo".');
    t.end();
  });

  test("'class' attribute", (t) => {
    let doc = DefaultDOMElement.createDocument('html');
    const el = doc.createElement('div');
    el.setAttribute('class', 'foo bar');
    t.ok(el.hasClass('foo'), 'should have class "foo" set');
    t.ok(el.hasClass('bar'), 'should have class "bar" set');
    t.end();
  });

  test("removeClass()", function(t) {
    const el = DefaultDOMElement.parseSnippet('<div class="foo">TEST</div>', 'html');
    el.removeClass('foo');
    t.notOk(el.hasClass('foo'), 'class should have been removed');
    t.end();
  });

  test("hasAttribute()", (t) => {
    const el = DefaultDOMElement.parseSnippet('<div foo="foo">TEST</div>', 'html');
    t.ok(el.hasAttribute('foo'), 'should have attribute "foo"');
    t.end();
  });

  test("getAttribute()", (t) => {
    const el = DefaultDOMElement.parseSnippet('<div foo="1">TEST</div>', 'html');
    t.equal(el.getAttribute('foo'), "1", 'should have return attribute "foo"');
    t.end();
  });

  test("setAttribute()", (t) => {
    let doc = DefaultDOMElement.createDocument('html');
    const el = doc.createElement('div');
    el.setAttribute('foo', 1);
    t.equal(el.getAttribute('foo'), "1", 'should have set attribute');
    t.end();
  });

  test("removeAttribute()", (t) => {
    const el = DefaultDOMElement.parseSnippet('<div foo="1">TEST</div>', 'html');
    el.removeAttribute('foo');
    t.isNil(el.getAttribute('foo'), 'should removed attribute');
    t.end();
  });

  test("attr() - getter", (t) => {
    const el = DefaultDOMElement.parseSnippet('<div foo="1">TEST</div>', 'html');
    t.equal(el.attr('foo'), "1", 'should return attribute value');
    t.end();
  });

  test("attr() - single attribute", (t) => {
    let doc = DefaultDOMElement.createDocument('html');
    const el = doc.createElement('div');
    el.attr('foo', 1);
    t.equal(el.getAttribute('foo'), "1", 'should have set attribute');
    t.end();
  });

  test("attr() - multiple attributes", (t) => {
    let doc = DefaultDOMElement.createDocument('html');
    const el = doc.createElement('div');
    el.attr({
      foo: 1,
      bar: 2
    });
    t.equal(el.getAttribute('foo'), "1", 'should have set attribute');
    t.equal(el.getAttribute('bar'), "2", 'should have set attribute');
    t.end();
  });

  test("removeAttr() - single attribute", (t) => {
    const el = DefaultDOMElement.parseSnippet('<div foo="1">TEST</div>', 'html');
    el.removeAttr('foo');
    t.isNil(el.getAttribute('foo'), 'should removed attribute');
    t.end();
  });

  test("removeAttr() - multiple attributes", (t) => {
    const el = DefaultDOMElement.parseSnippet('<div foo="1" bar="2">TEST</div>', 'html');
    el.removeAttr('foo bar');
    t.isNil(el.getAttribute('foo'), 'should removed attribute');
    t.isNil(el.getAttribute('bar'), 'should removed attribute');
    t.end();
  });

  test("getAttributes()", (t) => {
    const el = DefaultDOMElement.parseSnippet('<div foo="1" bar="2">TEST</div>', 'html');
    const entries = Array.from(el.getAttributes().entries());
    t.deepEqual(entries, [["foo", '1'], ["bar", '2']], 'should return all attributes');
    t.end();
  });

  test("getProperty()", (t) => {
    const el = DefaultDOMElement.parseSnippet('<input type="text" value="foo">', 'html');
    t.equal(el.getProperty('value'), "foo", 'should have correct value');
    t.end();
  });

  test("getProperty() -- input[type=checkbox]", (t) => {
    const el = DefaultDOMElement.parseSnippet('<input type="checkbox" checked>', 'html');
    t.equal(el.getProperty('value'), 'on', 'should have correct value');
    t.equal(el.getProperty('checked'), true, 'should have correct "checked" property');
    t.end();
  });

  test("setProperty()", (t) => {
    let doc = DefaultDOMElement.createDocument('html');
    const el = doc.createElement('input');
    el.setProperty('value', 1);
    t.equal(el.getProperty('value'), "1", 'should have set property');
    t.end();
  });

  test("setProperty() on XML element is not allowed", (t) => {
    let doc = DefaultDOMElement.createDocument('xml');
    const el = doc.createElement('foo');
    t.throws(()=>{
      el.setProperty('bar', 1);
    }, 'setProperty() on XML element is not allowed');
    t.end();
  });

  test("unsetting a property", (t) => {
    let doc = DefaultDOMElement.createDocument('html');
    const el = doc.createElement('div');
    el.setProperty('foo', 1);
    el.setProperty('foo', undefined);
    t.isNil(el.getProperty('foo'), 'should have unset property');
    t.end();
  });

  test("htmlProp() -- getter", (t) => {
    const el = DefaultDOMElement.parseSnippet('<input type="text" value="foo">', 'html');
    t.equal(el.htmlProp('value'), "foo", 'should return value');
    t.end();
  });

  test("htmlProp() -- single property", (t) => {
    let doc = DefaultDOMElement.createDocument('html');
    const el = doc.createElement('input');
    el.htmlProp('value', 1);
    t.equal(el.getProperty('value'), "1", 'should have set property');
    t.end();
  });

  test("htmlProp() -- multiple properties", (t) => {
    let doc = DefaultDOMElement.createDocument('html');
    const el = doc.createElement('input');
    el.htmlProp({
      'value': 1,
      'foo': 'foo'
    });
    t.equal(el.getProperty('value'), "1", 'should have set property');
    t.equal(el.getProperty('foo'), "foo", 'should have set property');
    t.end();
  });

  test("getValue()", (t) => {
    const el = DefaultDOMElement.parseSnippet('<input type="text" value="foo">', 'html');
    t.equal(el.getValue(), "foo", 'should have correct value');
    t.equal(el.val(), "foo", 'should work via jquery style getter');
    t.equal(el.value, "foo", 'should work via property getter');
    t.end();
  });

  test("setValue()", (t) => {
    let doc = DefaultDOMElement.createDocument('html');
    const el = doc.createElement('input');
    el.setValue(1);
    t.equal(el.getValue(), "1", 'should have set value');
    el.value = 2;
    t.equal(el.value, "2", 'should also work via property accessors');
    el.val(3);
    t.equal(el.value, "3", 'should also work via jquery style setter');
    t.end();
  });

  test("getStyle()", (t) => {
    let el = DefaultDOMElement.parseSnippet('<div style="color:blue"></div>', 'html');
    t.equal(el.getStyle('color'), 'blue', 'element should have style');
    t.equal(el.getAttribute('style'), 'color:blue', 'style can be retrieved via "style" attribute');
    t.end();
  });

  test("setStyle()", (t) => {
    let doc = DefaultDOMElement.createDocument('html');
    let el = doc.createElement('div');
    el.setStyle('color', 'blue');
    t.equal(el.getStyle('color'), 'blue', 'element should have style set');
    el.setAttribute('style', 'color:green');
    t.equal(el.getStyle('color'), 'green', 'style can be set via attribute "style" too');
    t.end();
  });

  test("setStyle() -- px styles", (t) => {
    let doc = DefaultDOMElement.createDocument('html');
    let el = doc.createElement('div');
    el.css({
      top: 10,
      bottom: 11,
      left: 12,
      right: 13,
      height: 100,
      width: 50
    });
    t.equal(el.getStyle('top'), '10px', 'px should have been added automatically');
    t.equal(el.getStyle('bottom'), '11px', 'px should have been added automatically');
    t.equal(el.getStyle('left'), '12px', 'px should have been added automatically');
    t.equal(el.getStyle('right'), '13px', 'px should have been added automatically');
    t.equal(el.getStyle('height'), '100px', 'px should have been added automatically');
    t.equal(el.getStyle('width'), '50px', 'px should have been added automatically');
    t.end();
  });

  test("css()", (t) => {
    let el = DefaultDOMElement.parseSnippet('<div style="color:blue"></div>', 'html');
    t.equal(el.css('color'), 'blue', 'can access style via jquery style getter');
    el.css('color', 'green');
    t.equal(el.getStyle('color'), 'green', 'can change style via jquery style setter');
    el.css({
      color: 'red',
      background: 'orange'
    });
    t.deepEqual([el.getStyle('color'), el.getStyle('background')], ['red', 'orange'], 'can change multiple styles at once via jquery style setter');
    t.end();
  });

  test("setTextContent", (t) => {
    let doc = DefaultDOMElement.createDocument('html');
    let el = doc.createElement('div');
    el.setTextContent('foo');
    t.equal(el.getTextContent(), 'foo', 'text content should have been set');
    el.textContent = 'bar';
    t.equal(el.getTextContent(), 'bar', 'should work also via property setter');
    el.text('baz');
    t.equal(el.getTextContent(), 'baz', 'should work also via jquery style setter');
    t.end();
  });

  test("innerHTML", (t) => {
    let el = DefaultDOMElement.parseSnippet('<p>abc<span class="foo" data-foo="foo">TEST</span>def</p>', 'html');
    t.equal(el.innerHTML, 'abc<span class="foo" data-foo="foo">TEST</span>def', 'el.innerHTML should give the HTML of the childNodes');
    el.innerHTML = 'TEST';
    t.equal(el.innerHTML, 'TEST', 'setting innerHTML should replace the content');
    t.end();
  });

  test("innerHTML on XML", (t) => {
    let el = DefaultDOMElement.parseSnippet('<foo>abc<bar bla="foo" blupp="foo">TEST</bar>def</foo>', 'xml');
    t.equal(el.innerHTML, 'abc<bar bla="foo" blupp="foo">TEST</bar>def', 'el.innerHTML should give the XML of the element content');
    el.innerHTML = 'TEST';
    t.equal(el.innerHTML, 'TEST', 'setting innerHTML should replace the content');
    t.end();
  });

  test("html()", (t) => {
    let el = DefaultDOMElement.parseSnippet('<div>TEST</div>', 'html');
    t.equal(el.html(), 'TEST', 'should return innerHTML');
    el.html('<span>FOO</span>');
    t.equal(el.innerHTML, '<span>FOO</span>', 'should have set innerHTML');
    t.end();
  });

  test("outerHTML", function(t) {
    var p = DefaultDOMElement.parseSnippet('<p class="foo">TEST</p>', 'html');
    t.equal(p.outerHTML, '<p class="foo">TEST</p>', 'outerHTML should be complete');
    t.end();
  });

  test("outerHTML of XML with camelCase tagnames", function(t) {
    let el = DefaultDOMElement.parseSnippet('<dummy><myNode></myNode></dummy>', 'xml');
    t.equal(el.outerHTML, '<dummy><myNode/></dummy>', 'XML tags should be serialized preserving case.');
    t.end();
  });

  test("find via tagname", function(t) {
    let el = DefaultDOMElement.parseSnippet('<dummy><foo></foo></dummy>', 'xml');
    let foo = el.find('foo');
    t.notNil(foo, 'Should find a <foo> element');
    t.end();
  });

  test("find multiple elements", function(t) {
    let el = DefaultDOMElement.parseSnippet('<dummy><foo>Bla</foo><foo>Blupp</foo></dummy>', 'xml');
    let foos = el.findAll('foo');
    t.equal(foos.length, 2, 'Should find two <foo> elements');
    t.equal(foos[0].textContent, 'Bla', '... with correct textContent');
    t.equal(foos[1].textContent, 'Blupp', '... with correct textContent');
    t.end();
  });

  test("find via #id", function(t) {
    let el = DefaultDOMElement.parseSnippet('<dummy><bla id="foo"></bla></dummy>', 'xml');
    let foo = el.find('#foo');
    t.notNil(foo, 'Should find a #foo element');
    t.end();
  });

  test("find via attr", function(t) {
    let el = DefaultDOMElement.parseSnippet('<dummy><bla data-id="foo"></bla></dummy>', 'xml');
    let foo = el.find('[data-id=foo]');
    t.notNil(foo, 'Should find a element with data-id=foo');
    t.end();
  });

  test("find in XML with camelCase tagnames", function(t) {
    let el = DefaultDOMElement.parseSnippet('<dummy><myNode myValue="foo"></myNode></dummy>', 'xml');
    t.notNil(el.find('myNode'), 'Should find a <myNode> element');
    t.notNil(el.find('[myValue="foo"]'), 'Should find a [myValue] element');
    t.end();
  });

  test("find in XML with namespace tagnames", function(t) {
    let xmlDoc = DefaultDOMElement.parseXML('<dummy xmlns:foo="foo"><foo:myNode></foo:myNode></dummy>');
    let myNode = xmlDoc.find('myNode');
    t.notNil(myNode, 'Should find a <foo:myNode> element');
    t.equal(myNode.tagName, 'foo:myNode', 'tagName should be qualified');
    t.end();
  });

  test("insertAt", function(t) {
    let dummy = DefaultDOMElement.parseSnippet('<dummy><bla></bla></dummy>', 'xml');
    let doc = dummy.getOwnerDocument();
    let newChild = doc.createElement('blupp');
    dummy.insertAt(1, newChild);
    t.equal(dummy.getChildCount(), 2, 'There should be 2 children now');
    t.ok(dummy.getChildAt(0).is('bla'), '.. the first should be <bla>');
    t.ok(dummy.getChildAt(1).is('blupp'), '.. the second should be <blupp>');
    t.end();
  });

  test("reattaching an element via insertAt", function(t) {
    let dummy = DefaultDOMElement.parseSnippet('<dummy><bla><blupp></blupp></bla></dummy>', 'xml');
    let blupp = dummy.find('blupp');
    dummy.insertAt(0, blupp);
    t.equal(dummy.getChildCount(), 2, 'There should be 2 children now');
    t.ok(dummy.getChildAt(0).is('blupp'), '.. the first should be <blupp>');
    t.ok(dummy.getChildAt(1).is('bla'), '.. the second should be <bla>');
    t.equal(dummy.getChildAt(1).getChildCount(), 0, '.. and <bla> should have no children');
    t.end();
  });

  test("reattaching an element via append", function(t) {
    let dummy = DefaultDOMElement.parseSnippet('<dummy><bla><blupp></blupp></bla></dummy>', 'xml');
    let blupp = dummy.find('blupp');
    dummy.appendChild(blupp);
    t.equal(dummy.getChildCount(), 2, 'There should be 2 children now');
    t.ok(dummy.getChildAt(0).is('bla'), '.. the first should be <bla>');
    t.ok(dummy.getChildAt(1).is('blupp'), '.. the second should be <blupp>');
    t.equal(dummy.getChildAt(0).getChildCount(), 0, '.. and <bla> should have no children');
    t.end();
  });

  test("children should only include elements", function(t) {
    let dummy = DefaultDOMElement.parseSnippet('<dummy>bli<bla></bla>blupp</dummy>', 'xml');
    dummy.children.forEach((c)=>{
      t.ok(c.isElementNode(), 'child should be an element');
    });
    t.end();
  });

  test("event listeners for custom events", (t) => {
    let doc = DefaultDOMElement.createDocument('html');
    let el = doc.createElement('div');
    let event = null;
    let handleFoo = function(e) {
      event = e;
    };
    el.addEventListener('foo', handleFoo);
    t.deepEqual(el.getEventListeners().map(l=>l.eventName), ['foo'], 'Element should have one listener');
    el.emit('foo', { bar: 'bar' });
    t.notNil(event, 'should have handled event');
    t.notNil(event.detail, 'event should have event.detail');
    t.deepEqual(event.detail, { bar: 'bar' });
    el.removeEventListener('foo', handleFoo);
    t.equal(el.getEventListeners().length, 0, 'element should not have a listener anymore');
    event = null;
    el.emit('foo', { baz: 'baz' });
    t.nil(event, 'event handler should not be called this time');
    t.end();
  });

  test("getChildNodes()", (t) => {
    let el = DefaultDOMElement.parseSnippet('<p>Foo<span>Bar</span>Baz</p>', 'html');
    let childNodes = el.childNodes;
    t.deepEqual(childNodes.map(c => c.nodeType), ['text', 'element', 'text'], "should return all childNodes");
    t.end();
  });

  test("getChildren()", (t) => {
    let el = DefaultDOMElement.parseSnippet('<p>Foo<span>Bar</span>Baz</p>', 'html');
    let children = el.children;
    t.deepEqual(children.map(c => c.tagName), ['span'], "should return all children");
    t.end();
  });

  test("firstChild", (t) => {
    let el = DefaultDOMElement.parseSnippet('<p>Foo<span>Bar</span>Baz</p>', 'html');
    let firstChild = el.firstChild;
    t.equal(firstChild.textContent, 'Foo', "should return all firstChild");
    t.end();
  });

  test("lastChild", (t) => {
    let el = DefaultDOMElement.parseSnippet('<p>Foo<span>Bar</span>Baz</p>', 'html');
    let lastChild = el.lastChild;
    t.equal(lastChild.textContent, 'Baz', "should return all lastChild");
    t.end();
  });

  test("is()", (t) => {
    let doc = DefaultDOMElement.createDocument('html');
    let el = doc.createElement('div').attr('data-id', 'foo');
    t.ok(el.is('div[data-id=foo]'));
    t.notOk(el.is('foo'));
    t.end();
  });

  test("getOwnerDocument()", (t) => {
    let doc = DefaultDOMElement.createDocument('html');
    let el = doc.createElement('div');
    t.equal(el.getOwnerDocument(), doc, 'should return the ownerDocument');
    t.equal(el.ownerDocument, doc, 'should also work via property getter');
    t.end();
  });

}

const test$12 = substanceTest.module('DOMImporter');

const pConverter = {
  type: 'paragraph',
  tagName: 'p',
  import(el, node, converter) {
    node.content = converter.annotatedText(el, [node.id, 'content']);
  }
};

test$12("creating a DOMImporter", (t) => {
  const schema = getTestConfig().getSchema();
  // fail without converters
  t.throws(() => {
    new DOMImporter({ schema });
  }, 'should throw if no converters are given');
  // fail without DocumentClass
  t.throws(() => {
    new DOMImporter({ converters: [] });
  }, 'should throw if schema is not given');
  // fail if a converter has no associated type
  t.throws(() => {
    new DOMImporter({ schema, converters: [{
      matchElement() { return true }
    }] });
  }, 'should throw if a converter does not have associated type');
  // fail if a converter has no matcher
  t.throws(() => {
    new DOMImporter({ schema, converters: [{
      type: 'paragraph'
    }] });
  }, 'should throw if a converter does not have a matcher function');
  t.throws(() => {
    new DOMImporter({ schema, converters: [{
      type: 'heading', tagName: 'h1'
    }] });
  }, 'should throw if there is no converter for the defaultTextType');
  t.throws(() => {
    new DOMImporter({ schema, converters: [{
      type: 'foo', tagName: 'h1'
    }] });
  }, 'should throw if a converter is associated with an unknown node type');
  t.end();
});

test$12("default matchElement()", (t) => {
  const testConverter = {
    type: 'paragraph', tagName: 'p'
  };
  const importer = createImporter([testConverter]);
  t.notNil(importer._allConverters[0].matchElement, 'registered converter should have a matchElement() function');
  t.end();
});

test$12("using a Converter class", (t) => {
  class MyConverter {
    get type() { return 'paragraph' }
    get tagName() { return 'p' }
  }
  const importer = createImporter([MyConverter]);
  t.equal(importer._allConverters[0].tagName, 'p', 'there should be one converter registered for <p>');
  t.end();
});

test$12("convertElement() -- block element", (t) => {
  const importer = createImporter();
  const el = DefaultDOMElement.parseSnippet('<p>TEST</p>', 'html');
  const node = importer.convertElement(el);
  t.equal(node.type, 'paragraph', 'should have converted element to node');
  t.end();
});

// Note: it is not common to convert inline nodes outside of their context -- that's why they are called inline
test$12("convertElement() -- inline node", (t) => {
  const testConverter = {
    type: 'test-inline-node',
    matchElement(el) { return el.is('[data-type=test-inline-node]') },
    import(el, node) {
      node.content = el.textContent;
    }
  };
  const importer = createImporter([pConverter, testConverter]);
  const el = DefaultDOMElement.parseSnippet('<span data-type="test-inline-node">TEST</span>', 'html');
  const node = importer.convertElement(el);
  t.equal(node.type, 'test-inline-node', 'should have converted element to node');
  t.end();
});

// Note: it is not common to convert annotations outside of their context
test$12("convertElement() -- annotation element", (t) => {
  const importer = createImporter();
  const el = DefaultDOMElement.parseSnippet('<b>TEST</b>', 'html');
  const node = importer.convertElement(el);
  t.equal(node.type, 'strong', 'should have converted element to node');
  t.end();
});

test$12("convertElement() should throw if no converter found", (t) => {
  const importer = createImporter([pConverter]);
  const el = DefaultDOMElement.parseSnippet('<h1>TEST</h1>', 'html');
  t.throws(() => {
    importer.convertElement(el);
  }, 'should throw if no converter found');
  t.end();
});

test$12("converting paragraph with inline node", (t) => {
  const testConverter = {
    type: 'test-inline-node',
    matchElement(el) { return el.is('[data-type=test-inline-node]') }
  };
  const importer = createImporter([pConverter, testConverter]);
  const el = DefaultDOMElement.parseSnippet('<p>abc <span data-type="test-inline-node">TEST</span> def</p>', 'html');
  const node = importer.convertElement(el);
  t.equal(node.content, `abc ${DOMImporter.INVISIBLE_CHARACTER} def`, 'should have inserted an invisible character');
  t.end();
});

test$12("converting an annotated paragraph", (t) => {
  const importer = createImporter();
  const el = DefaultDOMElement.parseSnippet('<p>abc <b>TEST</b> def</p>', 'html');
  const node = importer.convertElement(el);
  const doc = node.getDocument();
  const annos = doc.getAnnotations([node.id]);
  t.equal(annos.length, 1, 'there should be one annotation');
  t.end();
});


test$12("plainText()", (t) => {
  const testConverter = Object.assign({}, pConverter, {
    import(el, node, converter) {
      node.content = converter.plainText(el);
    }
  });
  const importer = createImporter([testConverter]);
  const el = DefaultDOMElement.parseSnippet('<p>abc <b>TEST</b> def</p>', 'html');
  const node = importer.convertElement(el);
  t.equal(node.content, `abc TEST def`, 'should have converted plain text');
  t.end();
});

test$12("convertContainer() -- trailing text", (t) => {
  const importer = createImporter();
  const els = DefaultDOMElement.parseSnippet('<p>A paragraph</p> and some trailing text', 'html');
  let container = importer.convertContainer(els, 'body');
  t.equal(container.nodes.length, 2, 'should have two nodes');
  t.deepEqual(container.getNodes().map(n=>n.type), ['paragraph', 'paragraph'], 'both should be paragraphs');
  t.end();
});

test$12("resolving id collisions", (t) => {
  // converter with 'tagName' gets a default matcher
  const importer = createImporter();
  const els = DefaultDOMElement.parseSnippet('<p id="foo">A paragraph</p><p id="foo">and another one with the same id</p>', 'html');
  let container = importer.convertContainer(els, 'body');
  t.equal(container.nodes.length, 2, 'should have two nodes');
  t.equal(container.nodes[0], 'foo', 'first should have the original id');
  t.notEqual(container.nodes[1], 'foo', 'second should have a new one');
  t.end();
});

/*
TODO:
  - import with options
*/

function createImporter(converters) {
  const config = getTestConfig();
  const schema = config.getSchema();
  if (!converters) {
    converters = config.getConverterRegistry().get('html').values();
  }
  return new DOMImporter({
    schema,
    converters: converters
  })
}

function checkValues(t, actual, expected) {
  if (!isObject$1(actual)) {
    t.fail('Provided object is invalid: actual=' + String(actual));
    return
  }
  forEach(expected, (expectedVal, key) => {
    if (isPlainObject$1(expectedVal) || isArray$1(expectedVal)) {
      t.deepEqual(actual[key], expectedVal, key + ' should be correct');
    } else {
      t.equal(actual[key], expectedVal, key + ' should be correct');
    }
  });
}

const test$13 = substanceTest.module('DOMSelection');

test$13.UI("Mapping a cursor inside a TextProperty from DOM to model", function(t) {
  let { editor, surface } = setupEditor(t, _p1);
  let domSelection = surface.context.domSelection;
  let node = editor.el.find('[data-path="p1.content"]').getFirstChild();
  setDOMSelection(node, 3);
  let sel = domSelection.getSelection();
  t.notOk(!sel || sel.isNull(), 'Selection should not be null');
  checkValues(t, sel.toJSON(), {
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 3,
    containerId: 'body',
    surfaceId: 'body'
  });
  t.end();
});

test$13.UI("Mapping a cursor in an empty paragraph from DOM to model", function(t) {
  let { editor, surface } = setupEditor(t, _empty);
  let domSelection = surface.context.domSelection;
  let node = editor.el.find('[data-path="empty.content"]').getFirstChild();
  setDOMSelection(node, 0);
  let sel = domSelection.getSelection();
  t.notOk(!sel || sel.isNull(), 'Selection should not be null');
  checkValues(t, sel.toJSON(), {
    type: 'property',
    path: ['empty', 'content'],
    startOffset: 0,
    containerId: 'body',
    surfaceId: 'body'
  });
  t.end();
});

test$13.UI("Mapping a ContainerSelection from DOM to model", function(t) {
  let { editor, surface } = setupEditor(t, surfaceWithParagraphs);
  let domSelection = surface.context.domSelection;
  let p1Text = editor.el.find('[data-path="p1.content"]').getFirstChild();
  let p2Text = editor.el.find('[data-path="p2.content"]').getFirstChild();
  setDOMSelection(p1Text, 1, p2Text, 2);
  let sel = domSelection.getSelection();
  t.notOk(!sel || sel.isNull(), 'Selection should not be null');
  checkValues(t, sel.toJSON(), {
    type: 'container',
    startPath: ['p1', 'content'],
    startOffset: 1,
    endPath: ['p2', 'content'],
    endOffset: 2,
    containerId: 'body',
    surfaceId: 'body'
  });
  t.end();
});

test$13.FF("Issue #354: Wrong selection in FF when double clicking between lines", function(t) {
  let { editor, surface } = setupEditor(t, surfaceWithParagraphs);
  let domSelection = surface.context.domSelection;
  let surfaceEl = editor.el.find('[data-id="body"]');
  setDOMSelection(surfaceEl, 0, surfaceEl, 1);
  let sel = domSelection.getSelection();
  t.notOk(!sel || sel.isNull(), 'Selection should not be null');
  checkValues(t, sel.toJSON(), {
    type: 'node',
    nodeId: 'p1',
    mode: 'full',
    containerId: 'body',
    surfaceId: 'body'
  });
  t.end();
});

// happens when using the same selection as in #354 in Chrome
test$13.UI("DOM selection that starts in a TextNode and ends in a paragraph on element level", function(t) {
  let { editor, surface } = setupEditor(t, surfaceWithParagraphs);
  let domSelection = surface.context.domSelection;
  let p1Text = editor.el.find('[data-path="p1.content"]').getFirstChild();
  let p2El = editor.el.find('[data-id="p2"]');
  setDOMSelection(p1Text, 0, p2El, 0);
  let sel = domSelection.getSelection();
  t.notOk(!sel || sel.isNull(), 'Selection should not be null');
  checkValues(t, sel.toJSON(), {
    type: 'container',
    startPath: ['p1', 'content'],
    startOffset: 0,
    endPath: ['p2', 'content'],
    endOffset: 0,
    containerId: 'body',
    surfaceId: 'body'
  });
  t.end();
});

test$13.UI("Issue #376: Wrong selection mapping at end of paragraph", function(t) {
  let { editor, surface } = setupEditor(t, surfaceWithParagraphs);
  let domSelection = surface.context.domSelection;
  let p1span = editor.el.find('[data-id="p1"] span');
  let p2El = editor.el.find('[data-id="p2"]');
  setDOMSelection(p1span, 1, p2El, 0);
  let sel = domSelection.getSelection();
  t.notOk(!sel || sel.isNull(), 'Selection should not be null');
  checkValues(t, sel.toJSON(), {
    type: 'container',
    startPath: ['p1', 'content'],
    startOffset: 2,
    endPath: ['p2', 'content'],
    endOffset: 0,
    containerId: 'body',
    surfaceId: 'body'
  });
  t.end();
});

test$13.UI("Rendering a ContainerSelection", function(t) {
  let { editor, doc, surface } = setupEditor(t, surfaceWithParagraphs);
  let domSelection = surface.context.domSelection;
  let sel = doc.createSelection({
    type: 'container',
    startPath: ['p1', 'content'],
    startOffset: 1,
    endPath: ['p2', 'content'],
    endOffset: 1,
    containerId: 'body',
    surfaceId: 'body'
  });
  domSelection.setSelection(sel);

  let p1Text = editor.el.find('[data-path="p1.content"]').getFirstChild();
  let p2Text = editor.el.find('[data-path="p2.content"]').getFirstChild();

  let wSel = window.getSelection();
  t.equal(wSel.anchorNode, p1Text.getNativeElement(), 'anchorNode should be in first paragraph.');
  t.equal(wSel.anchorOffset, 1, 'anchorOffset should be correct.');
  t.equal(wSel.focusNode, p2Text.getNativeElement(), 'focusNode should be in second paragraph.');
  t.equal(wSel.focusOffset, 1, 'focusOffset should be correct.');
  t.end();
});

test$13.UI("Rendering a cursor after inline node", function(t) {
  let { editor, doc, surface } = setupEditor(t, paragraphWithInlineNodes);
  let domSelection = surface.context.domSelection;
  let sel = doc.createSelection({
    type: 'property',
    path: ['p', 'content'],
    startOffset: 3,
    containerId: 'body',
    surfaceId: 'body'
  });
  let {start, end} = domSelection.mapModelToDOMCoordinates(sel);
  let pSpan = editor.el.find('[data-path="p.content"]');
  let third = pSpan.getChildAt(2);
  t.ok(start.container === third.getNativeElement(), 'anchorNode should be correct.');
  t.equal(start.offset, 0, 'anchorOffset should be correct.');
  t.ok(end.container === start.container, 'focusNode should be correct.');
  t.equal(end.offset, 0, 'focusOffset should be correct.');
  t.end();
});

test$13.UI("Rendering a cursor after inline node at the end of a property", function(t) {
  let { editor, doc, surface } = setupEditor(t, paragraphWithInlineNodes);
  let domSelection = surface.context.domSelection;
  let sel = doc.createSelection({
    type: 'property',
    path: ['p', 'content'],
    startOffset: 13,
    containerId: 'body',
    surfaceId: 'body'
  });
  let {start, end} = domSelection.mapModelToDOMCoordinates(sel);
  let pSpan = editor.el.find('[data-path="p.content"]');
  t.ok(start.container === pSpan.getNativeElement(), 'anchorNode should be correct.');
  t.equal(start.offset, 6, 'anchorOffset should be correct.');
  t.ok(end.container === start.container, 'focusNode should be correct.');
  t.equal(end.offset, 6, 'focusOffset should be correct.');
  t.end();
});

// test.UI("Rendering a cursor before isolated node", function(t) {
//   let { doc, surface } = setupEditor(t, _t1)
//   let domSelection = surface.context.domSelection
//   let sel = doc.createSelection({
//     type: 'node',
//     nodeId: 't1',
//     mode: 'before',
//     containerId: 'body',
//     surfaceId: 'body'
//   })
//   domSelection.setSelection(sel)
//   // let {start, end} = domSelection.mapModelToDOMCoordinates(sel)
//   // let pSpan = editor.el.find('[data-path="p.content"]')
//   // t.ok(start.container === pSpan.getNativeElement(), 'anchorNode should be correct.')
//   // t.equal(start.offset, 6, 'anchorOffset should be correct.')
//   // t.ok(end.container === start.container, 'focusNode should be correct.')
//   // t.equal(end.offset, 6, 'focusOffset should be correct.')
//   t.end()
// })

// test.UI("Rendering a cursor after isolated node", function(t) {
//   let { doc, surface } = setupEditor(t, _t1)
//   let domSelection = surface.context.domSelection
//   let sel = doc.createSelection({
//     type: 'node',
//     nodeId: 't1',
//     mode: 'after',
//     containerId: 'body',
//     surfaceId: 'body'
//   })
//   domSelection.setSelection(sel)
//   // let {start, end} = domSelection.mapModelToDOMCoordinates(sel)
//   // let pSpan = editor.el.find('[data-path="p.content"]')
//   // t.ok(start.container === pSpan.getNativeElement(), 'anchorNode should be correct.')
//   // t.equal(start.offset, 6, 'anchorOffset should be correct.')
//   // t.ok(end.container === start.container, 'focusNode should be correct.')
//   // t.equal(end.offset, 6, 'focusOffset should be correct.')
//   t.end()
// })

function surfaceWithParagraphs(doc, body) {
  let tx = new EditingInterface(doc);
  body.show(tx.create({
    type: 'paragraph',
    id: 'p1',
    content: 'AA'
  }));
  body.show(tx.create({
    type: 'paragraph',
    id: 'p2',
    content: 'BBB'
  }));
  body.show(tx.create({
    type: 'paragraph',
    id: 'p3',
    content: 'CCCC'
  }));
}

function paragraphWithInlineNodes(doc, body) {
  let tx = new EditingInterface(doc);
  body.show(tx.create({
    type: 'paragraph',
    id: 'p',
    content: '0123456789'
  }));
  // -> 01X234X56789X
  tx.setSelection({type: 'property', path: ['p', 'content'], startOffset: 2});
  tx.insertInlineNode({ type: 'test-inline-node', id: 'in1', content: '[1]'});
  tx.setSelection({type: 'property', path: ['p', 'content'], startOffset: 6});
  tx.insertInlineNode({ type: 'test-inline-node', id: 'in2', content: '[2]'});
  tx.setSelection({type: 'property', path: ['p', 'content'], startOffset: 12});
  tx.insertInlineNode({ type: 'test-inline-node', id: 'in3', content: '[3]'});
}

function headersAndParagraphs(tx) {
  var body = tx.get('body');
  tx.create({
    type: 'heading',
    id: 'h1',
    content: 'Section 1',
    level: 1
  });
  body.show('h1');
  tx.create({
    type: 'paragraph',
    id: 'p1',
    content: 'Paragraph 1'
  });
  body.show('p1');
  tx.create({
    type: 'heading',
    id: 'h2',
    content: 'Section 2',
    level: 1
  });
  body.show('h2');
  tx.create({
    type: 'paragraph',
    id: 'p2',
    content: 'Paragraph with annotation'
  });
  body.show('p2');
  tx.create({
    type: 'emphasis',
    id: 'em1',
    start: {
      path: ['p2', 'content'],
      offset: 15,
    },
    end: {
      offset: 25
    }
  });
  tx.create({
    type: 'heading',
    id: 'h3',
    content: 'Section 2.2',
    level: 2
  });
  body.show('h3');
  tx.create({
    type: 'paragraph',
    id: 'p3',
    content: 'Paragraph 3'
  });
  body.show('p3');
  // tx.create({
  //   type: "test-node",
  //   id: "test",
  //   boolVal: true,
  //   stringVal: "Test",
  //   arrayVal: [1, 2, 3, 4],
  //   objectVal: { "a": 1, "b": 2 }
  // })
}

/* eslint-disable no-use-before-define */
const test$14 = substanceTest.module('Editing');

// TODO: consolidate specification and 'category labels' of tests

// TODO: we should enable t.sandbox for all tests so that this implementation is
// tested on all platforms

test$14("IT1: Inserting text with cursor in the middle of a TextProperty", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1);
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 3,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.insertText('xxx');
  }, { action: 'type' });
  let sel = editorSession.getSelection();
  let p1 = doc.get('p1');
  t.equal(p1.getText(), P1_TEXT.slice(0,3)+'xxx'+P1_TEXT.slice(3), 'Text should have been inserted correctly.');
  t.equal(sel.start.offset, 6, 'Cursor should be after inserted text');
  t.end();
});

test$14("IT2: Inserting text with cursor within TextProperty inside annotation", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _s1);
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 4,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.insertText('xxx');
  }, { action: 'type' });
  let sel = editorSession.getSelection();
  let p1 = doc.get('p1');
  let s1 = doc.get('s1');
  t.equal(p1.getText(), P1_TEXT.slice(0,4)+'xxx'+P1_TEXT.slice(4), 'Text should have been inserted correctly.');
  t.equal(s1.end.offset, 8, 'Annotation should have been expanded.');
  t.equal(sel.start.offset, 7, 'Cursor should be after inserted text');
  t.end();
});

test$14("IT3: Inserting text with cursor within TextProperty at the start of an annotation", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _s1);
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 3,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.insertText('xxx');
  }, { action: 'type' });
  let sel = editorSession.getSelection();
  let p1 = doc.get('p1');
  let s1 = doc.get('s1');
  t.equal(p1.getText(), P1_TEXT.slice(0,3)+'xxx'+P1_TEXT.slice(3), 'Text should have been inserted correctly.');
  t.equal(s1.start.offset, 6, 'Annotation should have been moved.');
  t.equal(s1.end.offset, 8, 'Annotation should have been moved.');
  t.equal(sel.start.offset, 6, 'Cursor should be after inserted text');
  t.end();
});

test$14("IT4: Inserting text with cursor within TextProperty at the end of an annotation", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _s1);
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 5,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.insertText('xxx');
  }, { action: 'type' });
  let sel = editorSession.getSelection();
  let p1 = doc.get('p1');
  let s1 = doc.get('s1');
  t.equal(p1.getText(), P1_TEXT.slice(0,5)+'xxx'+P1_TEXT.slice(5), 'Text should have been inserted correctly.');
  t.equal(s1.start.offset, 3, 'Annotation should have been moved.');
  t.equal(s1.end.offset, 8, 'Annotation should have been moved.');
  t.equal(sel.start.offset, 8, 'Cursor should be after inserted text');
  t.end();
});

test$14("IT5: Inserting text with range within TextProperty", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1);
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 2,
    endOffset: 5,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.insertText('xxx');
  }, { action: 'type' });
  let sel = editorSession.getSelection();
  let p1 = doc.get('p1');
  t.equal(p1.getText(), P1_TEXT.slice(0, 2)+'xxx'+P1_TEXT.slice(5), 'Text should have been inserted correctly.');
  t.equal(sel.start.offset, 5, 'Cursor should be after inserted text');
  t.ok(sel.isCollapsed(), '... collapsed');
  t.end();
});

test$14("IT6: Inserting text with range within TextProperty overlapping an annotion aligned at the left side", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _s1);
  doc.set(['s1', 'end', 'offset'], 6);
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 3,
    endOffset: 5,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.insertText('xxx');
  }, { action: 'type' });
  let sel = editorSession.getSelection();
  let p1 = doc.get('p1');
  let s1 = doc.get('s1');
  t.equal(p1.getText(), P1_TEXT.slice(0, 3)+'xxx'+P1_TEXT.slice(5), 'Text should have been inserted correctly.');
  t.equal(s1.start.offset, 3, 'Annotation should have been moved.');
  t.equal(s1.end.offset, 7, 'Annotation should have been moved.');
  t.equal(sel.start.offset, 6, 'Cursor should be after inserted text');
  t.ok(sel.isCollapsed(), '... collapsed');
  t.end();
});

test$14("IT7: Inserting text with range within TextProperty inside an annotion", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _s1);
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 3,
    endOffset: 5,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.insertText('xxx');
  }, { action: 'type' });
  let sel = editorSession.getSelection();
  let p1 = doc.get('p1');
  let s1 = doc.get('s1');
  t.equal(p1.getText(), P1_TEXT.slice(0, 3)+'xxx'+P1_TEXT.slice(5), 'Text should have been inserted correctly.');
  t.equal(s1.start.offset, 3, 'Annotation should have been moved.');
  t.equal(s1.end.offset, 6, 'Annotation should have been moved.');
  t.equal(sel.start.offset, 6, 'Cursor should be after inserted text');
  t.ok(sel.isCollapsed(), '... collapsed');
  t.end();
});

test$14("IT8: Inserting text with range within TextProperty starting inside an annotion", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _s1);
  doc.set(['s1', 'end', 'offset'], 6);
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 4,
    endOffset: 6,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.insertText('xxx');
  }, { action: 'type' });
  let sel = editorSession.getSelection();
  let p1 = doc.get('p1');
  let s1 = doc.get('s1');
  t.equal(p1.getText(), P1_TEXT.slice(0, 4)+'xxx'+P1_TEXT.slice(6), 'Text should have been inserted correctly.');
  t.equal(s1.start.offset, 3, 'Annotation should have been moved.');
  t.equal(s1.end.offset, 7, 'Annotation should have been moved.');
  t.equal(sel.start.offset, 7, 'Cursor should be after inserted text');
  t.ok(sel.isCollapsed(), '... collapsed');
  t.end();
});

test$14("IT9: Inserting text after an InlineNode node", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _il1);
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 4,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.insertText('Y');
  }, { action: 'type' });
  let sel = editorSession.getSelection();
  let p1 = doc.get('p1');
  let il1 = doc.get('il1');
  t.equal(p1.getText(), P1_TEXT.slice(0,3)+'\uFEFF'+'Y'+P1_TEXT.slice(3), 'Text should have been inserted correctly.'); // eslint-disable-line no-useless-concat
  t.equal(sel.start.offset, 5, 'Cursor should be after inserted character');
  t.deepEqual([il1.start.offset, il1.end.offset], [3,4], 'InlineNode should have correct dimensions');
  t.end();
});

test$14("IT10: Typing over an InlineNode node", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _il1);
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 3,
    endOffset: 4,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.insertText('Y');
  }, { action: 'type' });
  let p1 = doc.get('p1');
  let il1 = doc.get('il1');
  t.nil(il1, 'InlineNode should have been deleted.');
  t.equal(p1.getText(), P1_TEXT.slice(0,3)+'Y'+P1_TEXT.slice(3), 'Text should have been inserted correctly.'); // eslint-disable-line no-useless-concat
  t.end();
});

test$14("IT11: Typing over a selected IsolatedNode", (t) => {
  let { editorSession, doc } = setupEditor(t, _in1);
  editorSession.setSelection({
    type: 'node',
    nodeId: 'in1',
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.insertText('Y');
  }, { action: 'type' });
  let body = doc.get('body');
  t.isNil(doc.get('in1'), 'IsolatedNode should have been deleted.');
  t.equal(body.getLength(), 1, 'There should be one node.');
  t.equal(body.getChildAt(0).getText(), 'Y', '.. containing the typed text.');
  t.end();
});

test$14("IT12: Typing before an IsolatedNode", (t) => {
  let { editorSession, doc } = setupEditor(t, _in1);
  editorSession.setSelection({
    type: 'node',
    nodeId: 'in1',
    mode: 'before',
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.insertText('Y');
  }, { action: 'type' });

  let body = doc.get('body');
  t.equal(body.getLength(), 2, 'There should be two nodes.');
  t.equal(body.getChildAt(0).getText(), 'Y', '.. first one containing the typed text.');
  t.end();
});

test$14("IT13: Typing after an IsolatedNode", (t) => {
  let { editorSession, doc } = setupEditor(t, _in1);
  editorSession.setSelection({
    type: 'node',
    nodeId: 'in1',
    mode: 'after',
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.insertText('Y');
  }, { action: 'type' });

  let body = doc.get('body');
  t.equal(body.getLength(), 2, 'There should be two nodes.');
  t.equal(body.getChildAt(1).getText(), 'Y', '.. second one containing the typed text.');
  t.end();
});

test$14("IT14: Typing over a ContainerSelection", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _p2);
  let p1 = doc.get('p1');
  let p2 = doc.get('p2');
  editorSession.setSelection({
    type: 'container',
    startPath: p1.getPath(),
    startOffset: 3,
    endPath: p2.getPath(),
    endOffset: 4,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.insertText('Y');
  }, { action: 'type' });

  let body = doc.get('body');
  t.deepEqual(body.nodes, ['p1'], 'Only the first paragraph should be left.');
  t.equal(p1.getText(), P1_TEXT.slice(0,3)+'Y'+P2_TEXT.slice(4), 'The content should have been merged correctly');
  t.end();
});

test$14("IT15: Inserting text into a ListItem", (t) => {
  let { editorSession, doc } = setupEditor(t, _l1, _l11, _l12);
  editorSession.setSelection({
    type: 'property',
    path: ['l1-1', 'content'],
    startOffset: 3,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.insertText('xxx');
  }, { action: 'type' });
  let sel = editorSession.getSelection();
  let li = doc.get('l1-1');
  t.equal(li.getText(), LI1_TEXT.slice(0,3)+'xxx'+LI1_TEXT.slice(3), 'Text should have been inserted correctly.');
  t.equal(sel.start.offset, 6, 'Cursor should be after inserted text');
  t.end();
});

test$14("IT16: Inserting text in an IsolatedNode", (t) => {
  let { doc, editorSession } = setupEditor(t, _p1, _in1, _p2);
  editorSession.setSelection({
    type: 'property',
    path: ['in1', 'title'],
    startOffset: 3
  });

  editorSession.transaction((tx) => {
    tx.insertText('xxx');
  });
  let in1 = doc.get('in1');
  t.equal(in1.title, IN1_TITLE.slice(0,3)+'xxx'+IN1_TITLE.slice(3), 'Text should be inserted into field');
  t.end();
});

test$14("II1: Inserting InlineNode node into a TextProperty", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1);
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 3,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.insertInlineNode({
      type: 'test-inline-node',
      id: 'il1',
      content: 'X'
    });
  });
  let sel = editorSession.getSelection();
  let p1 = doc.get('p1');
  let il1 = doc.get('il1');
  t.equal(p1.getText(), P1_TEXT.slice(0,3)+'\uFEFF'+P1_TEXT.slice(3), 'Text should have been inserted correctly.');
  t.equal(il1.start.offset, 3, 'Annotation should have been moved.');
  t.equal(il1.end.offset, 4, 'Annotation should have been moved.');
  t.equal(sel.start.offset, 4, 'Cursor should be after inserted inline node');
  t.end();
});

test$14("IB2: Inserting BlockNode using cursor at start of a TextNode", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1);
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 0,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.insertBlockNode({
      type: 'test-block',
      id: 'ib1'
    });
  });
  let sel = editorSession.getSelection();
  let body = doc.get('body');
  let p1 = doc.get('p1');
  t.equal(body.nodes[0], 'ib1', 'First node should be inserted block node.');
  t.equal(body.nodes[1], 'p1', 'Second node should be inserted block node.');
  t.ok(sel.isCollapsed(), 'Selection should be collapsed');
  t.deepEqual(sel.start.path, p1.getTextPath(), '... on paragraph');
  t.equal(sel.start.offset, 0, '... first position');
  t.end();
});

test$14("IB3: Inserting an existing BlockNode", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1);
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 0,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.create({
      type: 'test-block',
      id: 'ib1'
    });
  });
  const ib1 = doc.get('ib1');
  editorSession.transaction((tx) => {
    tx.insertBlockNode(ib1);
  });
  let body = doc.get('body');
  t.ok(body.getChildAt(0) === ib1, 'First node should be the very block node instance created before.');
  t.end();
});

test$14("IB4: Inserting a BlockNode when an IsolatedNode is selected", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _in1, _p2);
  editorSession.setSelection({
    type: 'node',
    nodeId: 'in1',
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.insertBlockNode({
      type: 'test-block',
      id: 'ib1'
    });
  });
  let body = doc.get('body');
  t.equals(body.length, 3, 'There should be three nodes.');
  t.equals(body.nodes[1], 'ib1', 'The second one should be the inserted block node.');
  t.isNil(doc.get('in1'), 'The IsolatedNode should have been deleted.');
  t.end();
});

test$14("IB5: Inserting a BlockNode before an IsolatedNode", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _in1, _p2);
  editorSession.setSelection({
    type: 'node',
    nodeId: 'in1',
    mode: 'before',
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.insertBlockNode({
      type: 'test-block',
      id: 'ib1'
    });
  });
  let body = doc.get('body');
  t.deepEqual(body.nodes, ['p1', 'ib1', 'in1', 'p2'], 'The body nodes should be in proper order.');
  t.end();
});

test$14("IB6: Inserting a BlockNode after an IsolatedNode", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _in1, _p2);
  editorSession.setSelection({
    type: 'node',
    nodeId: 'in1',
    mode: 'after',
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.insertBlockNode({
      type: 'test-block',
      id: 'ib1'
    });
  });
  let body = doc.get('body');
  t.deepEqual(body.nodes, ['p1', 'in1', 'ib1', 'p2'], 'The body nodes should be in proper order.');
  t.end();
});

test$14("IB7: Inserting a BlockNode with an expanded PropertySelection", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1);
  let p1 = doc.get('p1');
  editorSession.setSelection({
    type: 'property',
    path: p1.getPath(),
    startOffset: 3,
    endOffset: 5,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.insertBlockNode({
      type: 'test-block',
      id: 'ib1'
    });
  });
  let body = doc.get('body');
  t.equal(body.getLength(), 3, 'There should be 3 nodes.');
  t.equal(body.nodes[1], 'ib1', 'The second should be the inserted block node.');
  t.equal(p1.getText(), P1_TEXT.slice(0, 3), 'The paragraph should have been truncated.');
  t.equal(body.getChildAt(2).getText(), P1_TEXT.slice(5), '.. and the tail stored in a new paragraph.');
  t.end();
});

test$14("IB8: Inserting a BlockNode into an empty paragraph", (t) => {
  let { editorSession, doc } = setupEditor(t, _empty);
  let empty = doc.get('empty');
  editorSession.setSelection({
    type: 'property',
    path: empty.getPath(),
    startOffset: 0,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.insertBlockNode({
      type: 'test-block',
      id: 'ib1'
    });
  });
  let body = doc.get('body');
  t.deepEqual(body.nodes, ['ib1'], 'There should only be the inserted block node.');
  t.nil(doc.get('empty'), 'The empty paragraph should have been deleted.');
  t.end();
});

test$14("IB9: Inserting a BlockNode after a text node", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1);
  let p1 = doc.get('p1');
  editorSession.setSelection({
    type: 'property',
    path: p1.getPath(),
    startOffset: p1.getLength(),
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.insertBlockNode({
      type: 'test-block',
      id: 'ib1'
    });
  });
  let body = doc.get('body');
  t.deepEqual(body.nodes, ['p1', 'ib1'], 'The block node should have been inserted after the paragraph.');
  t.end();
});

test$14("IB10: Inserting a BlockNode with a collapsed ContainerSelection", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1);
  let p1 = doc.get('p1');
  editorSession.setSelection({
    type: 'container',
    startPath: p1.getPath(),
    startOffset: 3,
    endPath: p1.getPath(),
    endOffset: 3,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.insertBlockNode({
      type: 'test-block',
      id: 'ib1'
    });
  });
  let body = doc.get('body');
  t.equal(body.getLength(), 3, 'There should be three nodes.');
  t.equal(body.nodes[1], 'ib1', '.. the second one being the inserted block node.');
  t.end();
});

test$14("IB11: Inserting a BlockNode with a ContainerSelection", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _p2);
  let p1 = doc.get('p1');
  let p2 = doc.get('p2');
  editorSession.setSelection({
    type: 'container',
    startPath: p1.getPath(),
    startOffset: 3,
    endPath: p2.getPath(),
    endOffset: 4,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.insertBlockNode({
      type: 'test-block',
      id: 'ib1'
    });
  });
  let body = doc.get('body');
  t.deepEqual(body.nodes, ['p1', 'ib1', 'p2'], 'There should be three nodes.');
  t.equal(p1.getText(), P1_TEXT.slice(0, 3), 'The first paragraph should have been truncated correctly');
  t.equal(p2.getText(), P2_TEXT.slice(4), 'The second paragraph should have been truncated correctly');
  t.end();
});

test$14("DEL0: Deleting without selection", (t) => {
  let { editorSession } = setupEditor(t, _p1, _p2);
  editorSession.setSelection(null);
  t.doesNotThrow(() => {
    editorSession.transaction((tx) => {
      tx.deleteCharacter('right');
    });
  }, 'Should not throw');
  t.end();
});

test$14("DEL1: Deleting a property selection", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1);
  let p1 = doc.get('p1');
  editorSession.setSelection({
    type: 'property',
    path: p1.getPath(),
    startOffset: 0,
    endOffset: 3
  });
  editorSession.transaction((tx) => {
    tx.deleteSelection();
  });
  t.equal(p1.getText(), P1_TEXT.slice(3), 'Parargaph should be truncated');
  t.end();
});

test$14("DEL2: Deleting using DELETE with cursor at the end of a TextNode at the end of a Container", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _p2);
  editorSession.setSelection({
    type: 'property',
    path: ['p2', 'content'],
    startOffset: P2_TEXT.length,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteCharacter('right');
  });
  // nothing should have happened
  let sel = editorSession.getSelection();
  let body = doc.get('body');
  t.equal(body.nodes.length, 2, 'There should still be 2 nodes in body');
  let p2 = doc.get('p2');
  t.equal(p2.getText(), P2_TEXT, 'p2 should still have same content.');
  t.equal(sel.start.offset, P2_TEXT.length, 'Cursor should still be at the same position');
  t.end();
});

test$14("DEL3: Deleting using DELETE with cursor in the middle of a TextProperty", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1);
  editorSession.transaction((tx) => {
    tx.setSelection({
      type: 'property',
      path: ['p1', 'content'],
      startOffset: 3
    });
    tx.deleteCharacter('right');
  });
  let sel = editorSession.getSelection();
  let p1 = doc.get('p1');
  t.equal(p1.getText(), P1_TEXT.slice(0,3)+P1_TEXT.slice(4), 'One character should have been deleted');
  t.equal(sel.start.offset, 3, 'Cursor should be at the same position');
  t.end();
});

test$14("DEL4: Deleting using DELETE with cursor inside an empty TextNode and TextNode as successor", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _empty, _p2);
  editorSession.setSelection({
    type: 'property',
    path: ['empty', 'content'],
    startOffset: 0,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteCharacter('right');
  });
  let sel = editorSession.getSelection();
  let body = doc.get('body');
  let p2 = doc.get('p2');
  t.equal(body.nodes.length, 2, 'There should be only 2 nodes left.');
  t.equal(p2.getText(), P2_TEXT, 'p2 should not be affected');
  t.ok(sel.isCollapsed(), 'Selection should be collapsed');
  t.deepEqual(sel.start.path, p2.getTextPath(), '... on p2');
  t.equal(sel.start.offset, 0, '... at first position');
  t.end();
});

test$14("DEL5: Deleting using DELETE with cursor inside an empty TextNode and IsolatedNode as successor", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _empty, _block1);
  editorSession.setSelection({
    type: 'property',
    path: ['empty', 'content'],
    startOffset: 0,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteCharacter('right');
  });
  let sel = editorSession.getSelection();
  let body = doc.get('body');
  t.equal(body.nodes.length, 2, 'There should be only 2 nodes left.');
  t.ok(sel.isNodeSelection(), 'Selection should be a NodeSelection');
  t.equal(sel.getNodeId(), 'block1', '... block1');
  t.end();
});

test$14("DEL6: Deleting using DELETE with cursor inside an empty TextNode and List as successor", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _empty, _l1, _l11, _l12);
  editorSession.setSelection({
    type: 'property',
    path: ['empty', 'content'],
    startOffset: 0,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteCharacter('right');
  });
  let sel = editorSession.getSelection();
  let body = doc.get('body');
  t.equal(body.nodes.length, 2, 'There should be only 2 nodes left.');
  t.ok(sel.isCollapsed(), 'Selection should be collapsed');
  t.deepEqual(sel.start.path, ['l1-1', 'content'], '... on first list item');
  t.equal(sel.start.offset, 0,'... at first position');
  t.end();
});

test$14("DEL7: Deleting using DELETE with cursor at the end of a non-empty TextNode and TextNode as successor", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _p2);
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: P1_TEXT.length,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteCharacter('right');
  });
  let sel = editorSession.getSelection();
  let body = doc.get('body');
  let p1 = doc.get('p1');
  t.equal(body.nodes.length, 1, 'There should be only one node left.');
  t.ok(sel.isCollapsed(), 'Selection should be a collapsed');
  t.deepEqual(sel.start.path, p1.getTextPath(), '... on p1');
  t.equal(sel.start.offset, P1_TEXT.length, '... at the same position as before');
  t.end();
});

test$14("DEL8: Deleting using DELETE with cursor at the end of a non-empty TextNode and IsolatedNode as successor", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _block1);
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: P1_TEXT.length,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteCharacter('right');
  });
  // NOTE: if there is no merge possible, nothing should happen
  // only the selection should be updated
  let sel = editorSession.getSelection();
  let body = doc.get('body');
  t.equal(body.nodes.length, 2, 'There should still be 2 nodes left.');
  t.ok(sel.isNodeSelection(), 'Selection should be a NodeSelection');
  t.equal(sel.getNodeId(), 'block1', '... in block1');
  t.ok(sel.isFull(), '... selecting the whole node');
  t.end();
});

test$14("DEL9: Deleting using DELETE with cursor after an IsolatedNode and a TextNode as successor", (t) => {
  let { editorSession, doc } = setupEditor(t, _block1, _p1);
  editorSession.setSelection({
    type: 'node',
    nodeId: 'block1',
    mode: 'after',
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteCharacter('right');
  });
  let p1 = doc.get('p1');
  t.equal(p1.getText(), P1_TEXT.slice(1), 'First character of paragraph should have been deleted.');
  _checkSelection(t, editorSession.getSelection(), {
    type: 'property',
    path: p1.getPath(),
    startOffset: 0
  });
  t.end();
});

test$14("DEL10: Deleting using DELETE with cursor after an IsolatedNode and an IsolatedNode as successor", (t) => {
  let { editorSession } = setupEditor(t, _block1, _block2);
  editorSession.setSelection({
    type: 'node',
    nodeId: 'block1',
    mode: 'after',
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteCharacter('right');
  });
  _checkSelection(t, editorSession.getSelection(), {
    type: 'node',
    nodeId: 'block2',
    mode: 'full'
  });
  t.end();
});

test$14("DEL11: Deleting using BACKSPACE with cursor in the middle of a TextProperty", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1);
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 4,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteCharacter('left');
  });
  let sel = editorSession.getSelection();
  let p1 = doc.get('p1');
  t.equal(p1.getText(), P1_TEXT.slice(0,3)+P1_TEXT.slice(4), 'one character should have been deleted');
  t.equal(sel.start.offset, 3, 'Cursor should have shifted by one character');
  t.end();
});

test$14("DEL12: Deleting using BACKSPACE with cursor inside an empty TextNode and TextNode as predecessor", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _empty, _p2);
  editorSession.setSelection({
    type: 'property',
    path: ['empty', 'content'],
    startOffset: 0,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteCharacter('left');
  });
  let sel = editorSession.getSelection();
  let p1 = doc.get('p1');
  t.isNil(doc.get('empty'), 'empty node should have been deleted');
  t.equal(p1.getText(), P1_TEXT, 'p1 should be untouched');
  t.deepEqual(sel.start.path, p1.getTextPath(), 'Cursor should be in p1');
  t.equal(sel.start.offset, P1_TEXT.length, '... at last position');
  t.end();
});

test$14("DEL13: Deleting using BACKSPACE with cursor inside an empty TextNode and IsolatedNode as predecessor", (t) => {
  let { editorSession, doc } = setupEditor(t, _block1, _empty, _p2);
  editorSession.setSelection({
    type: 'property',
    path: ['empty', 'content'],
    startOffset: 0,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteCharacter('left');
  });
  let sel = editorSession.getSelection();
  t.isNil(doc.get('empty'), 'empty node should have been deleted');
  t.ok(sel.isNodeSelection(), 'Selection should be a node selection');
  t.equal(sel.getNodeId(), 'block1', '... on block1');
  t.end();
});

test$14("DEL14: Deleting using BACKSPACE with cursor at the start of a non-empty TextNode and TextNode as predecessor", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _p2);
  editorSession.setSelection({
    type: 'property',
    path: ['p2', 'content'],
    startOffset: 0,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteCharacter('left');
  });
  let sel = editorSession.getSelection();
  let p1 = doc.get('p1');
  t.isNil(doc.get('p2'), 'p2 should have been deleted');
  t.equal(p1.getText(), P1_TEXT+P2_TEXT, 'Text should have been merged');
  t.ok(sel.isCollapsed(), 'Selection should be collapsed');
  t.deepEqual(sel.start.path, p1.getTextPath(), '... on p1');
  t.ok(sel.start.offset, P1_TEXT.length, '... cursor should after the original text of p1');
  t.end();
});

test$14("DEL15: Deleting using BACKSPACE with cursor at the start of a non-empty TextNode and IsolatedNode as predecessor", (t) => {
  let { editorSession, doc } = setupEditor(t, _block1, _p2);
  editorSession.setSelection({
    type: 'property',
    path: ['p2', 'content'],
    startOffset: 0,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteCharacter('left');
  });
  let sel = editorSession.getSelection();
  let body = doc.get('body');
  t.equal(body.getLength(), 2, 'There should still be two nodes');
  t.ok(sel.isNodeSelection(), 'Selection should be a NodeSelection');
  t.equal(sel.getNodeId(), 'block1', '... on block1');
  t.ok(sel.isFull(), '... selecting the full node');
  t.end();
});

test$14("DEL16: Deleting using BACKSPACE with cursor after IsolatedNode", (t) => {
  let { editorSession, doc } = setupEditor(t, _block1, _p2);
  editorSession.setSelection({
    type: 'node',
    mode: 'after',
    nodeId: 'block1',
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteCharacter('left');
  });
  let sel = editorSession.getSelection();
  let body = doc.get('body');
  t.isNil(doc.get('block1'), 'IsolatedNode should have been deleted');
  t.equal(body.getLength(), 2, 'There should still be two nodes');
  let pnew = body.getChildAt(0);
  t.ok(sel.isCollapsed(), 'Selection should be collapsed');
  t.deepEqual(sel.start.path, pnew.getTextPath(), '... on new paragraph');
  t.equal(sel.start.offset, 0, '... at first position');
  t.end();
});

test$14("DEL17: Deleting using BACKSPACE with cursor before IsolatedNode with TextNode as predecessor", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _block2);
  editorSession.setSelection({
    type: 'node',
    mode: 'before',
    nodeId: 'block2',
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteCharacter('left');
  });
  let sel = editorSession.getSelection();
  let body = doc.get('body');
  let p1 = doc.get('p1');
  t.equal(body.getLength(), 2, 'There should still be two nodes');
  t.equal(p1.getText(), P1_TEXT.slice(0, -1), 'Last character of p1 should have been deleted');
  t.ok(sel.isCollapsed(), 'Selection should be collapsed');
  t.deepEqual(sel.start.path, p1.getTextPath(), '... on p1');
  t.equal(sel.start.offset, P1_TEXT.length-1, '... at last position');
  t.end();
});

test$14("DEL18: Deleting using BACKSPACE with cursor before IsolatedNode and IsolatedNode as predecessor", (t) => {
  let { editorSession, doc } = setupEditor(t, _block1, _block2);
  editorSession.setSelection({
    type: 'node',
    mode: 'before',
    nodeId: 'block2',
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteCharacter('left');
  });
  let sel = editorSession.getSelection();
  let body = doc.get('body');
  t.equal(body.getLength(), 2, 'There should still be two nodes');
  t.ok(sel.isNodeSelection(), 'Selection should be a NodeSelection');
  t.equal(sel.getNodeId(), 'block1', '... on block1');
  t.ok(sel.isFull(), '... selection the entire node');
  t.end();
});

test$14("DEL19: Deleting an entirely selected IsolatedNode", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _block1, _p2);
  editorSession.setSelection({
    type: 'node',
    nodeId: 'block1',
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteSelection();
  });
  let sel = editorSession.getSelection();
  let body = doc.get('body');
  t.isNil(doc.get('block1'), 'IsolatedNode should have been deleted');
  let pnew = body.getChildAt(1);
  t.equal(body.getLength(), 3, 'There should be 3 nodes');
  t.ok(sel.isCollapsed(), 'Selection should be collapsed');
  t.deepEqual(sel.start.path, pnew.getTextPath(), '... on new paragraph');
  t.equal(sel.start.offset, 0, '... at first position');
  t.end();
});

test$14("DEL20: Deleting a range starting before a TextNode and ending after a TextNode", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _block1, _block2, _p2);
  let p1 = doc.get('p1');
  let p2 = doc.get('p2');
  editorSession.setSelection({
    type: 'container',
    startPath: p1.getTextPath(),
    startOffset: 0,
    endPath: p2.getTextPath(),
    endOffset: p2.getLength(),
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteSelection();
  });
  let sel = editorSession.getSelection();
  let body = doc.get('body');
  t.equal(body.nodes.length, 1, 'There should be only one node left');
  let first = body.getChildAt(0);
  t.ok(first.isText(), '... which is a TextNode');
  t.ok(first.isEmpty(), '... which is empty');
  t.ok(sel.isCollapsed(), 'Selection should be collapsed');
  t.deepEqual(sel.start.path, first.getTextPath(), '... on that TextNode');
  t.equal(sel.start.offset, 0, '... at first position');
  t.end();
});

test$14("DEL21: Deleting a range starting in the middle of a TextNode and ending after a TextNode", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _block1, _block2, _p2);
  let p1 = doc.get('p1');
  let p2 = doc.get('p2');
  editorSession.setSelection({
    type: 'container',
    startPath: p1.getTextPath(),
    startOffset: 3,
    endPath: p2.getTextPath(),
    endOffset: p2.getLength(),
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteSelection();
  });
  let sel = editorSession.getSelection();
  let body = doc.get('body');
  t.equal(body.nodes.length, 1, 'There should be only one node left');
  let first = body.getChildAt(0);
  t.ok(first.isText(), '... which is a TextNode');
  t.equal(first.getText(), P1_TEXT.slice(0, 3), '... with truncated content');
  t.ok(sel.isCollapsed(), 'Selection should be collapsed');
  t.deepEqual(sel.start.path, first.getTextPath(), '... on that TextNode');
  t.equal(sel.start.offset, 3, '... at last position');
  t.end();
});

test$14("DEL22: Deleting a range starting before a TextNode and ending in the middle of a TextNode", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _block1, _block2, _p2);
  let p1 = doc.get('p1');
  let p2 = doc.get('p2');
  editorSession.setSelection({
    type: 'container',
    startPath: p1.getTextPath(),
    startOffset: 0,
    endPath: p2.getTextPath(),
    endOffset: 3,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteSelection();
  });
  let sel = editorSession.getSelection();
  let body = doc.get('body');
  t.equal(body.nodes.length, 1, 'There should be only one node left');
  let first = body.getChildAt(0);
  t.ok(first.isText(), '... which is a TextNode');
  t.equal(first.getText(), P2_TEXT.slice(3), '... with sliced content');
  t.ok(sel.isCollapsed(), 'Selection should be collapsed');
  t.deepEqual(sel.start.path, first.getTextPath(), '... on that TextNode');
  t.equal(sel.start.offset, 0, '... at first position');
  t.end();
});

test$14("DEL23: Deleting a range starting in the middle of a TextNode and ending in the middle of a TextNode", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _block1, _block2, _p2);
  let p1 = doc.get('p1');
  let p2 = doc.get('p2');
  editorSession.setSelection({
    type: 'container',
    startPath: p1.getTextPath(),
    startOffset: 3,
    endPath: p2.getTextPath(),
    endOffset: 3,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteSelection();
  });
  let sel = editorSession.getSelection();
  let body = doc.get('body');
  t.equal(body.nodes.length, 1, 'There should be only one node left');
  let first = body.getChildAt(0);
  t.ok(first.isText(), '... which is a TextNode');
  t.equal(first.getText(), P1_TEXT.slice(0,3)+P2_TEXT.slice(3), '... with merged content');
  t.ok(sel.isCollapsed(), 'Selection should be collapsed');
  t.deepEqual(sel.start.path, first.getTextPath(), '... on that TextNode');
  t.equal(sel.start.offset, 3, '... at correct position');
  t.end();
});

test$14("DEL24: Deleting a range starting in the middle of a TextNode and ending in the middle of a ListItem", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _block1, _block2, _l1, _l11, _l12);
  let p1 = doc.get('p1');
  let l1 = doc.get('l1');
  editorSession.setSelection({
    type: 'container',
    startPath: p1.getTextPath(),
    startOffset: 3,
    endPath: ['l1-1', 'content'],
    endOffset: 3,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteSelection();
  });
  let sel = editorSession.getSelection();
  let body = doc.get('body');
  t.equal(body.nodes.length, 2, 'There should be 2 nodes left');
  t.equal(p1.getText(), P1_TEXT.slice(0,3)+LI1_TEXT.slice(3), '... with merged content');
  t.equal(l1.items.length, 1, 'The list should have only 1 item left');
  t.ok(sel.isCollapsed(), 'Selection should be collapsed');
  t.deepEqual(sel.start.path, p1.getTextPath(), '... on p1');
  t.equal(sel.start.offset, 3, '... at correct position');
  t.end();
});

test$14("DEL25: Deleting a range starting in the middle of a ListItem and ending in the middle of a TextNode", (t) => {
  let { editorSession, doc } = setupEditor(t, _l1, _l11, _l12, _block1, _block2, _p1);
  let p1 = doc.get('p1');
  let l1 = doc.get('l1');
  editorSession.setSelection({
    type: 'container',
    startPath: ['l1-2', 'content'],
    startOffset: 3,
    endPath: p1.getTextPath(),
    endOffset: 3,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteSelection();
  });
  let sel = editorSession.getSelection();
  let body = doc.get('body');
  t.equal(body.nodes.length, 1, 'There should be 1 node left');
  t.equal(l1.items.length, 2, 'The list should have 2 items');
  let li2 = l1.getItemAt(1);
  t.equal(li2.getText(), LI2_TEXT.slice(0,3)+P1_TEXT.slice(3), 'The second item should container merged content');
  t.ok(sel.isCollapsed(), 'Selection should be collapsed');
  t.deepEqual(sel.start.path, li2.getTextPath(), '... on second list item');
  t.equal(sel.start.offset, 3, '... at correct position');
  t.end();
});

test$14("DEL26: Deleting a range within a ListItem", (t) => {
  let { editorSession, doc } = setupEditor(t, _l1, _l11, _l12);
  let l1 = doc.get('l1');
  editorSession.setSelection({
    type: 'container',
    startPath: ['l1-2', 'content'],
    startOffset: 3,
    endPath: ['l1-2', 'content'],
    endOffset: 6,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteSelection();
  });
  let sel = editorSession.getSelection();
  t.equal(l1.items.length, 2, 'The list should have 2 items');
  let li2 = l1.getItemAt(1);
  t.equal(li2.getText(), LI2_TEXT.slice(0,3)+LI2_TEXT.slice(6), 'The second item should be changed');
  t.ok(sel.isCollapsed(), 'Selection should be collapsed');
  t.deepEqual(sel.start.path, li2.getTextPath(), '... on second list item');
  t.equal(sel.start.offset, 3, '... at correct position');
  t.end();
});

test$14("DEL27: Deleting a range across two ListItems within the same List", (t) => {
  let { editorSession, doc } = setupEditor(t, _l1, _l11, _l12);
  let l1 = doc.get('l1');
  editorSession.setSelection({
    type: 'container',
    startPath: ['l1-1', 'content'],
    startOffset: 3,
    endPath: ['l1-2', 'content'],
    endOffset: 3,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteSelection();
  });
  let sel = editorSession.getSelection();
  t.equal(l1.items.length, 1, 'The list should have 1 item');
  let li1 = l1.getItemAt(0);
  t.equal(li1.getText(), LI1_TEXT.slice(0,3)+LI2_TEXT.slice(3), 'The items should be merged');
  t.ok(sel.isCollapsed(), 'Selection should be collapsed');
  t.deepEqual(sel.start.path, li1.getTextPath(), '... on the list item');
  t.equal(sel.start.offset, 3, '... at correct position');
  t.end();
});

test$14("DEL28: Deleting a ContainerSelection within a single TextNode", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1);
  let p1 = doc.get('p1');
  editorSession.setSelection({
    type: 'container',
    startPath: p1.getPath(),
    startOffset: 3,
    endPath: p1.getPath(),
    endOffset: 5,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteSelection();
  });
  t.equal(p1.getText(), P1_TEXT.slice(0,3)+P1_TEXT.slice(5), 'The text should have been deleted');
  _checkSelection(t, editorSession.getSelection(), {
    type: 'property',
    path: p1.getPath(),
    startOffset: 3
  });
  t.end();
});

test$14("DEL29: Deleting a character inside an IsolatedNode", (t) => {
  let { doc, editorSession } = setupEditor(t, _p1, _in1, _p2);
  editorSession.setSelection({
    type: 'property',
    path: ['in1', 'title'],
    startOffset: 3
  });
  editorSession.transaction((tx) => {
    tx.deleteCharacter('right');
  });
  let in1 = doc.get('in1');
  t.equal(in1.title, IN1_TITLE.slice(0,3)+IN1_TITLE.slice(4), 'Text should be inserted into field');
  t.end();
});

test$14("DEL30: Merging two ListItems using DELETE", (t) => {
  let { editorSession, doc } = setupEditor(t, _l1, _l11, _l12);
  editorSession.setSelection({
    type: 'property',
    path: ['l1-1', 'content'],
    startOffset: LI1_TEXT.length,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteCharacter('right');
  });
  let sel = editorSession.getSelection();
  let l = doc.get('l1');
  t.equal(l.items.length, 1, 'Only one list item should be left');
  let li = l.getItemAt(0);
  t.equal(li.getText(), LI1_TEXT+LI2_TEXT, 'The list item should have the merged text');
  t.ok(sel.isCollapsed(), 'The selection should be collapsed');
  t.deepEqual(sel.start.path, li.getTextPath(), '... on the list item');
  t.equal(sel.start.offset, LI1_TEXT.length, '... at the end of the original content');
  t.end();
});

test$14("DEL31: Merging a List into previous List using DELETE", (t) => {
  let { editorSession, doc } = setupEditor(t, _l1, _l11, _l12, _l2, _l21, _l22);
  editorSession.setSelection({
    type: 'property',
    path: ['l1-2', 'content'],
    startOffset: LI2_TEXT.length,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteCharacter('right');
  });
  let sel = editorSession.getSelection();
  let l1 = doc.get('l1');
  t.equal(l1.items.length, 4, 'First list should have 4 items');
  t.ok(sel.isCollapsed(), 'The selection should be collapsed');
  t.deepEqual(sel.start.path, ['l1-2', 'content'], '... on the same item');
  t.equal(sel.start.offset, LI2_TEXT.length, '... at the same position');
  t.end();
});

test$14("DEL32: Merging an empty List into previous TextNode using DELETE", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _l1);
  let p1 = doc.get('p1');
  editorSession.setSelection({
    type: 'property',
    path: p1.getPath(),
    startOffset: p1.getLength(),
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteCharacter('right');
  });
  t.isNil(doc.get('l2'), 'Second list should have been removed.');
  t.end();
});

test$14("BR1: Breaking without selection", (t) => {
  let { editorSession } = setupEditor(t, _p1);
  editorSession.setSelection(null);
  t.doesNotThrow(() => {
    editorSession.transaction((tx) => {
      tx.break();
    });
  });
  t.end();
});

test$14("BR2: Breaking a TextNode", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1);
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 3,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.break();
  }, { action: 'break' });
  let body = doc.get('body');
  t.equals(body.length, 2, 'There should be 2 nodes');
  let first = body.getChildAt(0);
  let second = body.getChildAt(1);
  t.equals(second.type, 'paragraph', 'Second should be a paragraph');
  t.equals(first.getText(), P1_TEXT.slice(0,3), 'First should be truncated');
  t.equals(second.getText(), P1_TEXT.slice(3), '.. and second should contain the tail');
  t.end();
});

test$14("BR3: Breaking annotated text with cursor before the annotation", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _s1);
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 1,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.break();
  }, { action: 'break' });
  let body = doc.get('body');
  let second = body.getChildAt(1);
  let anno = doc.get('s1');
  t.deepEqual(anno.start.path, second.getPath(), 'Annotation should now be on the second paragraph');
  t.equal(anno.start.offset, 2, '.. starting at character 2');
  t.equal(anno.end.offset, 4, '.. ending at character 4');
  t.end();
});

test$14("BR4: Breaking annotated text with cursor inside the annotation", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _s1);
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 4,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.break();
  }, { action: 'break' });
  let body = doc.get('body');
  let secondPara = body.getChildAt(1);
  let anno = doc.get('s1');
  t.equal(anno.start.offset, 3, '.. starting at character 3');
  t.equal(anno.end.offset, 4, '.. ending at character 4');
  let secondAnno = doc.getIndex('annotations').get([secondPara.id, 'content'])[0];
  t.equal(secondAnno.start.offset, 0, '.. starting at character 0');
  t.equal(secondAnno.end.offset, 1, '.. ending at character 1');
  t.end();
});

test$14("BR5: Breaking a TextNode at the first position", (t) => {
  let { editorSession, doc } = setupEditor(t, _h1);
  let h1 = doc.get('h1');
  editorSession.setSelection({
    type: 'property',
    path: h1.getPath(),
    startOffset: 0,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.break();
  }, { action: 'break' });
  let body = doc.get('body');
  t.equals(body.length, 2, 'There should be 2 nodes');
  let first = body.getChildAt(0);
  let second = body.getChildAt(1);
  t.equals(second, h1, 'Second should be the original node');
  t.equals(first.type, h1.type, 'First should be of the same type');
  t.equals(first.getText(), '', '.. but empty');
  _checkSelection(t, editorSession.getSelection(), {
    type: 'property',
    path: h1.getPath(),
    startOffset: 0
  });
  t.end();
});

test$14("BR5: Breaking a TextNode at the last position", (t) => {
  let { editorSession, doc } = setupEditor(t, _h1);
  let h1 = doc.get('h1');
  editorSession.setSelection({
    type: 'property',
    path: h1.getPath(),
    startOffset: h1.getLength(),
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.break();
  }, { action: 'break' });
  let body = doc.get('body');
  t.equals(body.length, 2, 'There should be 2 nodes');
  let first = body.getChildAt(0);
  let second = body.getChildAt(1);
  t.equals(first, h1, 'First should be the original node');
  t.equals(second.type, 'paragraph', 'Second should be a paragraph.');
  t.equals(second.getText(), '', '.. without content');
  _checkSelection(t, editorSession.getSelection(), {
    type: 'property',
    path: second.getPath(),
    startOffset: 0
  });
  t.end();
});

test$14("BR6: Breaking a ContainerSelection", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _p2);
  editorSession.setSelection({
    type: 'container',
    startPath: ['p1', 'content'],
    startOffset: 3,
    endPath: ['p2', 'content'],
    endOffset: 4,
    containerId: 'body'
  });
  editorSession.transaction((tx)=>{
    tx.break();
  });
  let p2 = doc.get('p2');
  t.equal(p2.getText(), P2_TEXT.slice(4));
  _checkSelection(t, editorSession.getSelection(), {
    type: 'property',
    path: ['p2', 'content'],
    startOffset: 0,
    endOffset: 0
  });
  t.end();
});

test$14("BR7: Breaking a NodeSelection", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _in1, _p2);
  editorSession.setSelection({
    type: 'node',
    nodeId: 'in1',
    containerId: 'body'
  });
  editorSession.transaction((tx)=>{
    tx.break();
  });
  let body = doc.get('body');
  t.equal(body.getLength(), 4, 'There should be one more node');
  let newP = body.getChildAt(2);
  _checkSelection(t, editorSession.getSelection(), {
    type: 'property',
    path: newP.getPath(),
    startOffset: 0,
    endOffset: 0
  });
  t.end();
});

test$14("BR8: Breaking a NodeSelection (before)", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _in1, _p2);
  editorSession.setSelection({
    type: 'node',
    nodeId: 'in1',
    mode: 'before',
    containerId: 'body'
  });
  editorSession.transaction((tx)=>{
    tx.break();
  });
  let body = doc.get('body');
  t.equal(body.getLength(), 4, 'There should be one more node');
  _checkSelection(t, editorSession.getSelection(), {
    type: 'node',
    nodeId: 'in1',
    mode: 'before',
    containerId: 'body'
  });
  t.end();
});

test$14("BR9: Breaking with an expanded PropertySelection", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _p2);
  editorSession.setSelection({
    type: 'property',
    path: ['p1','content'],
    startOffset: 3,
    endOffset: 5,
    containerId: 'body'
  });
  editorSession.transaction((tx)=>{
    tx.break();
  });
  let body = doc.get('body');
  let first = body.getChildAt(0);
  let second = body.getChildAt(1);
  t.equal(first.getText(), P1_TEXT.slice(0, 3), 'First paragraph should be truncated');
  t.equal(second.getText(), P1_TEXT.slice(5), '.. and tail should be moved to second paragraph');
  _checkSelection(t, editorSession.getSelection(), {
    type: 'property',
    path: second.getPath(),
    startOffset: 0
  });
  t.end();
});

test$14("BR10: Breaking a ListItem with cursor in the middle of text", (t) => {
  let { editorSession, doc } = setupEditor(t, _l1, _l11, _l12);
  editorSession.setSelection({
    type: 'property',
    path: ['l1-1', 'content'],
    startOffset: 3,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.break();
  }, { action: 'break' });
  let sel = editorSession.getSelection();
  let l = doc.get('l1');
  t.equal(l.items.length, 3, 'List should have 3 items');
  let li1 = doc.get(l.items[0]);
  let li2 = doc.get(l.items[1]);
  t.equal(li1.getText(), LI1_TEXT.slice(0,3), 'First item should have been truncated.');
  t.equal(li2.getText(), LI1_TEXT.slice(3), 'remaining line should have been inserted into new list item.');
  t.deepEqual(sel.start.path, li2.getTextPath(), 'Cursor should in second item');
  t.equal(sel.start.offset, 0, 'Cursor should be at begin of item.');
  t.end();
});

test$14("BR11: Breaking a ListItem with cursor at begin of text", (t) => {
  let { editorSession, doc } = setupEditor(t, _l1, _l11, _l12);
  editorSession.setSelection({
    type: 'property',
    path: ['l1-1', 'content'],
    startOffset: 0,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.break();
  }, { action: 'break' });
  let sel = editorSession.getSelection();
  let l = doc.get('l1');
  t.equal(l.items.length, 3, 'List should have 3 items');
  let li1 = doc.get(l.items[0]);
  let li2 = doc.get(l.items[1]);
  t.equal(li1.getText(), '', 'First item should be empty.');
  t.equal(li2.getText(), LI1_TEXT, 'Text should have moved to next item.');
  t.deepEqual(sel.start.path, li2.getTextPath(), 'Cursor should be in second item');
  t.equal(sel.start.offset, 0, '... at begin of item.');
  t.end();
});

test$14("BR12: Splitting a List by breaking an empty ListItem", (t) => {
  let { editorSession, doc } = setupEditor(t, _l1, _l11, _l1_empty, _l12);
  editorSession.setSelection({
    type: 'property',
    path: ['l1-empty', 'content'],
    startOffset: 0,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.break();
  }, { action: 'break' });
  let sel = editorSession.getSelection();
  let body = doc.get('body');
  t.equal(body.nodes.length, 3, 'There should be three nodes');
  let l1 = body.getChildAt(0);
  let p = body.getChildAt(1);
  let l2 = body.getChildAt(2);
  t.ok(l1.isList() && p.isText() && l2.isList(), '... list, paragraph, and list');
  t.equal(l1.items.length, 1, 'The first list should now have only one item');
  t.equal(l1.items[0], 'l1-1', '... with id "li-1"');
  t.equal(p.getText(), '', 'The paragraph should be empty');
  t.equal(l2.items.length, 1, 'The second list should now have only one item');
  t.equal(l2.items[0], 'l1-2', '... with id "li-2"');
  t.ok(sel.isCollapsed(), 'The selection should be collapsed');
  t.deepEqual(sel.start.path, p.getTextPath(), '... on the new paragraph');
  t.equal(sel.start.offset, 0, '... at first position');
  t.end();
});

test$14("BR13: Breaking the last empty list item", (t) => {
  let { editorSession, doc } = setupEditor(t, _l1, _l11, _l12, _l1_empty);
  editorSession.setSelection({
    type: 'property',
    path: ['l1-empty', 'content'],
    startOffset: 0,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.break();
  }, { action: 'break' });
  let sel = editorSession.getSelection();
  let body = doc.get('body');
  t.equal(body.nodes.length, 2, 'There should be two nodes');
  let l1 = body.getChildAt(0);
  let p = body.getChildAt(1);
  t.ok(l1.isList() && p.isText(), '... list, and paragraph');
  t.equal(l1.items.length, 2, 'The list should now have only two items');
  t.equal(p.getText(), '', 'The paragraph should be empty');
  t.ok(sel.isCollapsed(), 'The selection should be collapsed');
  t.deepEqual(sel.start.path, p.getTextPath(), '... on the new paragraph');
  t.equal(sel.start.offset, 0, '... at first position');
  t.end();
});

test$14("BR14: Breaking a list with only one empty item", (t) => {
  let { editorSession, doc } = setupEditor(t, _l1, _l1_empty);
  editorSession.setSelection({
    type: 'property',
    path: ['l1-empty', 'content'],
    startOffset: 0,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.break();
  }, { action: 'break' });
  let body = doc.get('body');
  t.equal(body.nodes.length, 1, 'There should be only one node left');
  let p = body.getChildAt(0);
  t.equal(p.type, 'paragraph', '... of type paragraph');
  t.isNil(doc.get('l1'), 'The list should have been removed');
  _checkSelection(t, editorSession.getSelection(), {
    type: 'property',
    path: p.getPath(),
    startOffset: 0
  });
  t.end();
});

test$14("BR15: Breaking a list with a first empty item", (t) => {
  let { editorSession, doc } = setupEditor(t, _l1, _l1_empty, _l11);
  editorSession.setSelection({
    type: 'property',
    path: ['l1-empty', 'content'],
    startOffset: 0,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.break();
  }, { action: 'break' });
  let body = doc.get('body');
  t.equal(body.nodes.length, 2, 'There should be two nodes now');
  let p = body.getChildAt(0);
  let l1 = body.getChildAt(1);
  t.equal(p.type, 'paragraph', 'First should be a paragraph');
  t.equal(p.getText(), '', '.. without content');
  t.equal(l1.type, 'list', 'Second should be a list');
  t.equal(l1.getLength(), 1, '.. with only one item left.');
  _checkSelection(t, editorSession.getSelection(), {
    type: 'property',
    path: p.getPath(),
    startOffset: 0
  });
  t.end();
});


test$14("L5-1: Toggling a ListItem using BACKSPACE", (t) => {
  let { editorSession, doc } = setupEditor(t, _l1, _l11, _l12);
  editorSession.setSelection({
    type: 'property',
    path: ['l1-2', 'content'],
    startOffset: 0,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.deleteCharacter('left');
  });
  let sel = editorSession.getSelection();
  let body = doc.get('body');
  t.equal(body.getLength(), 2, 'There should be two nodes now');
  let l = doc.get('l1');
  let p = body.getChildAt(1);
  t.equal(l.items.length, 1, 'Only one list item should be left');
  t.equal(p.type, 'paragraph', 'The second one should be a paragraph');
  t.equal(p.getText(), LI2_TEXT, '.. with the text of the second item');
  t.ok(sel.isCollapsed(), 'The selection should be collapsed');
  t.deepEqual(sel.start.path, p.getTextPath(), '... on the new paragraph');
  t.equal(sel.start.offset, 0, '... at first position');
  t.end();
});

test$14("L8-1: Toggling a paragraph into a List", (t) => {
  let { doc, editorSession} = setupEditor(t, _p1);
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 3,
    containerId: 'body'
  });
  editorSession.transaction((tx)=>{
    tx.toggleList({ ordered: false});
  });
  let body = doc.get('body');
  t.equal(body.getLength(), 1, 'There should be 1 node');
  let l = body.getNodeAt(0);
  t.equal(l.type, 'list', '.. a list');
  t.equal(l.ordered, false, '.. unordered');
  t.equal(l.getLength(), 1, '.. with one item');
  let li = l.getItemAt(0);
  t.equal(li.getText(), P1_TEXT, '.. with text of p1');
  t.end();
});

test$14("L8-2: Toggling first list-item into a paragraph", (t) => {
  let { doc, editorSession} = setupEditor(t, _l1, _l11, _l12);
  editorSession.setSelection({
    type: 'property',
    path: ['l1-1', 'content'],
    startOffset: 3,
    containerId: 'body'
  });
  editorSession.transaction((tx)=>{
    tx.toggleList();
  });
  let body = doc.get('body');
  t.equal(body.getLength(), 2, 'There should be 2 nodes');
  let p = body.getNodeAt(0);
  let l = body.getNodeAt(1);
  t.equal(p.type, 'paragraph', '.. first should be paragraph');
  t.equal(p.getText(), LI1_TEXT, '.. with text of first list item');
  t.equal(l.type, 'list', '.. second should be alist');
  t.equal(l.getLength(), 1, '.. with one item');
  t.end();
});

test$14("L8-3: Toggling last list-item into a paragraph", (t) => {
  let { doc, editorSession} = setupEditor(t, _l1, _l11, _l12);
  editorSession.setSelection({
    type: 'property',
    path: ['l1-2', 'content'],
    startOffset: 3,
    containerId: 'body'
  });
  editorSession.transaction((tx)=>{
    tx.toggleList();
  });
  let body = doc.get('body');
  t.equal(body.getLength(), 2, 'There should be 2 nodes');
  let l = body.getNodeAt(0);
  let p = body.getNodeAt(1);
  t.equal(l.type, 'list', '.. first should be alist');
  t.equal(l.getLength(), 1, '.. with one item');
  t.equal(p.type, 'paragraph', '.. second should be paragraph');
  t.equal(p.getText(), LI2_TEXT, '.. with text of second list item');
  t.end();
});

test$14("L8-4: Toggling a middle list-item into a paragraph splitting the list", (t) => {
  let { doc, editorSession} = setupEditor(t, _l1, _l11, _l13, _l12);
  editorSession.setSelection({
    type: 'property',
    path: ['l1-3', 'content'],
    startOffset: 3,
    containerId: 'body'
  });
  editorSession.transaction((tx)=>{
    tx.toggleList();
  });
  let body = doc.get('body');
  t.equal(body.getLength(), 3, 'There should be 3 nodes');
  let l = body.getNodeAt(0);
  let p = body.getNodeAt(1);
  let l2 = body.getNodeAt(2);
  t.equal(l.type, 'list', '.. first should be alist');
  t.equal(l.getLength(), 1, '.. with one item');
  t.equal(p.type, 'paragraph', '.. second should be paragraph');
  t.equal(p.getText(), LI3_TEXT, '.. with text of second list item');
  t.equal(l2.type, 'list', '.. last should be a list again');
  t.equal(l2.getLength(), 1, '.. with one item');
  t.end();
});

test$14("L8-5: Toggling the only list-item of a list into a paragraph", (t) => {
  let { doc, editorSession} = setupEditor(t, _l1, _l11);
  editorSession.setSelection({
    type: 'property',
    path: ['l1-1', 'content'],
    startOffset: 0,
    containerId: 'body'
  });
  editorSession.transaction((tx)=>{
    tx.toggleList();
  });
  let body = doc.get('body');
  t.equal(body.getLength(), 1, 'There should be one node left');
  let p = body.getNodeAt(0);
  t.equal(p.type, 'paragraph', '.. which should be paragraph');
  t.equal(p.getText(), LI1_TEXT, '.. with text of first list item');
  t.end();
});

test$14("IND1: Indenting a ListItem", (t) => {
  let { doc, editorSession } = setupEditor(t, _l1, _l11, _l12);
  editorSession.setSelection({
    type: 'property',
    path: ['l1-1', 'content'],
    startOffset: 3,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.indent();
  });
  let li1 = doc.get('l1-1');
  t.equal(li1.level, 2, 'Indentation level should have increased');
  t.end();
});

test$14("IND2: Dedenting a ListItem", (t) => {
  let { doc, editorSession } = setupEditor(t, _l1, _l11, _l12, _li1plus);
  editorSession.setSelection({
    type: 'property',
    path: ['l1-1', 'content'],
    startOffset: 3,
    containerId: 'body'
  });
  editorSession.transaction((tx) => {
    tx.dedent();
  });
  let li1 = doc.get('l1-1');
  t.equal(li1.level, 1, 'Indentation level should have decreased');
  t.end();
});

test$14("CP1: Copying a property selection", (t) => {
  let { doc } = setupEditor(t, headersAndParagraphs);
  let editor = new EditingInterface(doc);
  editor.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 4,
    endOffset: 9,
    containerId: 'body'
  });
  let copy = editor.copySelection();
  let textNode = copy.get(Document.TEXT_SNIPPET_ID);
  t.notNil(textNode, 'There should be a text node for the property fragment.');
  t.equal(textNode.content, 'graph', 'Selected text should be copied.');
  t.end();
});

test$14("CP2: Copying a property selection with annotated text", (t) => {
  let { doc } = setupEditor(t, headersAndParagraphs);
  let editor = new EditingInterface(doc);
  editor.setSelection({
    type: 'property',
    path: ['p2', 'content'],
    startOffset: 10,
    endOffset: 19,
    containerId: 'body'
  });
  let copy = editor.copySelection();
  t.equal(copy.get([Document.TEXT_SNIPPET_ID, 'content']), 'with anno', 'Selected text should be copied.');
  let annos = copy.getIndex('annotations').get([Document.TEXT_SNIPPET_ID, 'content']);
  t.equal(annos.length, 1, 'There should be one annotation on copied text.');
  let anno = annos[0];
  t.equal(anno.type, "emphasis", "The annotation should be 'emphasis'.");
  t.deepEqual([anno.start.offset, anno.end.offset], [5, 9], 'The annotation should be over the text "anno".');
  t.end();
});

test$14("CP3: Copying a container selection", (t) => {
  let { doc } = setupEditor(t, headersAndParagraphs);
  let editor = new EditingInterface(doc);
  editor.setSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['h1', 'content'],
    startOffset: 4,
    endPath: ['p2', 'content'],
    endOffset: 9,
  });
  let copy = editor.copySelection();
  let content = copy.get(Document.SNIPPET_ID);
  t.notNil(content, 'There should be a container node with id "content".');
  // 4 nodes? 'body', 'snippets', 'p1', 'p2'
  t.equal(content.nodes.length, 4, 'There should be 4 nodes in the copied document.');
  let first = copy.get(content.nodes[0]);
  t.equal(first.type, 'heading', "The first node should be a heading.");
  t.equal(first.content, 'ion 1', "Its content should be truncated to 'ion 1'.");
  let last$$1 = copy.get(content.nodes[3]);
  t.equal(last$$1.type, 'paragraph', "The last node should be a paragraph.");
  t.equal(last$$1.content, 'Paragraph', "Its content should be truncated to 'Paragraph'.");
  t.end();
});

test$14("CP4: Copying a paragraph", (t) => {
  let { doc } = setupEditor(t, _p1, _p2);
  let editor = new EditingInterface(doc);
  editor.setSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p2'],
    startOffset: 0,
    endPath: ['p2'],
    endOffset: 1
  });
  let copy = editor.copySelection();
  let p2 = copy.get('p2');
  t.equal(p2.content, doc.get('p2').content, 'The whole paragraph should be copied.');
  t.end();
});

test$14("CP5: Copying two ListItems", (t) => {
  let { doc } = setupEditor(t, _l1, _l11, _l12);
  let editor = new EditingInterface(doc);
  editor.setSelection({
    type: 'container',
    startPath: ['l1-1', 'content'],
    startOffset: 3,
    endPath: ['l1-2', 'content'],
    endOffset: 3,
    containerId: 'body'
  });
  let copy = editor.copySelection();
  let content = copy.getContainer();
  t.equal(content.getLength(), 1, 'There should be one node');
  let l1 = copy.get('l1');
  t.notNil(l1, 'l1 should exist');
  t.equal(l1.getLength(), 2, '.. having 2 items');
  let li1 = l1.getItemAt(0);
  let li2 = l1.getItemAt(1);
  t.equal(li1.getText(), LI1_TEXT.slice(3), 'The first item should have correct content');
  t.equal(li2.getText(), LI2_TEXT.slice(0, 3), 'The second item should have correct content');
  t.end();
});

test$14("CP6: Copying a paragraph and a ListItem", (t) => {
  let { doc } = setupEditor(t, _p1, _l1, _l11, _l12);
  let editor = new EditingInterface(doc);
  editor.setSelection({
    type: 'container',
    startPath: ['p1', 'content'],
    startOffset: 3,
    endPath: ['l1-1', 'content'],
    endOffset: 3,
    containerId: 'body'
  });
  let copy = editor.copySelection();
  let content = copy.getContainer();
  t.equal(content.getLength(), 2, 'There should be 2 nodes');
  let l1 = copy.get('l1');
  t.notNil(l1, 'l1 should exist');
  t.equal(l1.getLength(), 1, '.. having 1 item');
  let li1 = l1.getItemAt(0);
  t.equal(li1.getText(), LI1_TEXT.slice(0, 3), 'The first item should have correct content');
  t.end();
});

test$14("CP7 Copying a table", (t) => {
  let { doc } = setupEditor(t, _t1);
  let editor = new EditingInterface(doc);
  editor.setSelection({
    type: 'node',
    nodeId: 't1',
    containerId: 'body',
  });
  let copy = editor.copySelection();
  let t1 = copy.get('t1');
  t.notNil(t1, 'Table node should exist');
  t.equals(t1.getRowCount(), 2, '.. with two rows');
  t.equals(t1.getColCount(), 2, '.. with two cols');
  for (let row = 0; row < 2; row++) {
    for (let col = 0; col < 2; col++) {
      let cell = t1.getCellAt(row,col);
      t.notNil(cell, `Cell (${row+1},${col+1}) should exist`);
      t.equals(cell.content, T_CONTENT[row][col], '.. and have correct content');
    }
  }
  t.end();
});

test$14("CP8 Copying a sparse table", (t) => {
  let { doc } = setupEditor(t, _t1_sparse);
  let editor = new EditingInterface(doc);
  editor.setSelection({
    type: 'node',
    nodeId: 't1',
    containerId: 'body',
  });
  let copy = editor.copySelection();
  let t1 = copy.get('t1');
  t.notNil(t1, 'Table node should exist');
  t.equals(t1.getRowCount(), 2, '.. with two rows');
  t.equals(t1.getColCount(), 2, '.. with two cols');
  for (let i = 0; i < 2; i++) {
    let cell = t1.getCellAt(i,i);
    t.notNil(cell, `Cell (${i+1},${i+1}) should exist`);
    t.equals(cell.content, T_CONTENT[i][i], '.. and have correct content');
    t.nil(t1.getCellAt(1-i, i), `Cell (${1-i+1},${i+1}) should not exist`);
  }
  t.end();
});

// NOTE: with Copy'n'Paste within the same document,
// we must make sure to check that things are cloned correctly,
// e.g., not that inadvertantly the copy references one of the original children

test$14("CP9: Copy and Pasting a List", (t) => {
  let { doc, editorSession } = setupEditor(t, _p1, _empty, _l1, _l11, _l12);
  editorSession.transaction((tx) => {
    tx.setSelection({
      type: 'container',
      startPath: ['l1-1', 'content'],
      startOffset: 0,
      endPath: ['l1-2', 'content'],
      endOffset: LI2_TEXT.length,
      containerId: 'body'
    });
    let copy = tx.copySelection();
    tx.setSelection({
      type: 'property',
      path: ['empty', 'content'],
      startOffset: 0,
      containerId: 'body'
    });
    tx.paste(copy);
  });
  let body = doc.get('body');
  let nodes = body.getNodes();
  t.equal(body.length, 3, "There should be 3 nodes");
  t.deepEqual(['paragraph', 'list', 'list'], nodes.map(n => n.type), '.. a paragraph and 2 lists');
  let li1 = body.getChildAt(1);
  let li2 = body.getChildAt(2);
  t.notDeepEqual(li1.items, li2.items, 'List items should not be the same.');
  t.equal(li1.getItemAt(0).getText(), li2.getItemAt(0).getText(), 'First list item should have same content');
  t.equal(li1.getItemAt(1).getText(), li2.getItemAt(1).getText(), 'Second list item should have same content');
  t.end();
});

test$14("CP10: Copy and Pasting a List partially", (t) => {
  let { doc, editorSession } = setupEditor(t, _p1, _empty, _l1, _l11, _l12);
  editorSession.transaction((tx) => {
    tx.setSelection({
      type: 'container',
      startPath: ['l1-1', 'content'],
      startOffset: 3,
      endPath: ['l1-2', 'content'],
      endOffset: 3,
      containerId: 'body'
    });
    let copy = tx.copySelection();
    tx.setSelection({
      type: 'property',
      path: ['empty', 'content'],
      startOffset: 0,
      containerId: 'body'
    });
    tx.paste(copy);
  });
  let body = doc.get('body');
  let nodes = body.getNodes();
  t.equal(body.length, 3, "There should be 3 nodes");
  t.deepEqual(['paragraph', 'list', 'list'], nodes.map(n => n.type), '.. a paragraph and 2 lists');
  let li1 = body.getChildAt(1);
  let li2 = body.getChildAt(2);
  t.notDeepEqual(li1.items, li2.items, 'List items should not be the same.');
  t.equal(li1.getItemAt(0).getText(), LI1_TEXT.slice(3), 'First list item should be truncated');
  t.equal(li1.getItemAt(1).getText(), LI2_TEXT.slice(0,3), 'Second list item should be truncated');
  t.end();
});


test$14("AN1: Annotating without unknown annotation type", (t) => {
  let { editorSession } = setupEditor(t, _p1);
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 0,
    endOffset: 3
  });
  t.throws(() => {
    editorSession.transaction((tx)=>{
      tx.annotate({type: 'unknown-annotation'});
    });
  }, 'Should throw an exception');
  t.end();
});

test$14("AN2: Tying to annotate using a non-annotation type", (t) => {
  let { editorSession } = setupEditor(t, _p1);
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 0,
    endOffset: 3
  });
  t.throws(() => {
    editorSession.transaction((tx)=>{
      tx.annotate({type: 'paragraph'});
    });
  }, 'Should throw an exception');
  t.end();
});

test$14("ST1: Switching text type", (t) => {
  let { editorSession, doc } = setupEditor(t, _p1, _s1);
  let p1 = doc.get('p1');
  editorSession.setSelection({
    type: 'property',
    path: p1.getPath(),
    startOffset: 0,
    endOffset: 0,
    containerId: 'body'
  });
  editorSession.transaction((tx)=>{
    tx.switchTextType({type: 'heading', level: 1});
  });
  let body = doc.get('body');
  t.equal(body.getLength(), 1, 'There should be one node.');
  let first = body.getChildAt(0);
  t.equal(first.type, 'heading', '.. of type heading');
  t.equal(first.getText(), P1_TEXT, '.. with correct content');
  let annos = doc.getAnnotations(first.getPath());
  t.equal(annos.length, 1, '.. and with one annotation');
  t.end();
});

function _checkSelection(t, actual, exp) {
  let _data = actual.toJSON();
  let data = {};
  forEach(exp, (value, key) => {
    data[key] = _data[key];
  });
  t.deepEqual(data, exp, 'Selection should be correct');
}

const test$15 = substanceTest.module('EditorSession');

test$15("Keeping TransactionDocument up-to-date.", function(t) {
  let { editorSession, doc } = setupEditor(t, simple);
  let stageDoc = editorSession._transaction._stageDoc;
  stageDoc._apply = substanceTest.spy(stageDoc, '_apply');
  doc.create({ type: 'paragraph', id: 'foo', content: 'foo'});
  var p = stageDoc.get('foo');
  t.equal(stageDoc._apply.callCount, 1, "Stage should have been updated.");
  t.notNil(p, "Stage should contain new paragraph node.");
  t.equal(p.content, "foo");
  t.end();
});

test$15("Undoing and redoing a change.", function(t) {
  let { editorSession, doc } = setupEditor(t, simple);
  editorSession.transaction(function(tx) {
    tx.update(['p1', 'content'], { type: 'insert', start: 3, text: "XXX" });
  });
  t.equal(doc.get(['p1', 'content']), '012XXX3456789', 'Text should have been inserted.');
  t.equal(editorSession.canUndo(), true, 'Undo should be possible');
  editorSession.undo();
  t.equal(doc.get(['p1', 'content']), '0123456789', 'Original text should have been recovered.');
  t.equal(editorSession.canUndo(), false, 'Undo should be disabled');
  t.equal(editorSession.canRedo(), true, 'Redo should be possible');
  editorSession.redo();
  t.equal(doc.get(['p1', 'content']), '012XXX3456789', 'Text should have been changed again.');
  t.end();
});

test$15("Selections after undo/redo.", function(t) {
  let { editorSession, doc } = setupEditor(t, simple);
  var path = ['p1', 'content'];
  editorSession.setSelection({
    type: 'property',
    path: path,
    startOffset: 3
  });
  editorSession.transaction(function(tx) {
    tx.update(path, { type: 'insert', start: 3, text: "XXX" });
    tx.setSelection({
      type: 'property',
      path: path,
      startOffset: 6
    });
  });
  editorSession.undo();
  var sel = editorSession.getSelection();
  t.ok(sel.equals(doc.createSelection({
    type: 'property',
    path: path,
    startOffset: 3
  })), 'Selection should be set correctly after undo.');
  editorSession.redo();
  sel = editorSession.getSelection();
  t.ok(sel.equals(doc.createSelection({
    type: 'property',
    path: path,
    startOffset: 6
  })), 'Selection should be set correctly after redo.');
  t.end();
});

const test$16 = substanceTest.module('Fragmenter');

var TEXT$1 = 'ABCDEFGHI';

test$16("No annos.", function(t) {
  var annos = [];
  var html = _render$1(TEXT$1, annos);
  t.equal(html, TEXT$1);
  t.end();
});

test$16("With one anno.", function(t) {
  var annos = [new Anno('b', 'b1', 3, 6)];
  var html = _render$1(TEXT$1, annos);
  t.equal(html, 'ABC<b>DEF</b>GHI');
  t.end();
});

test$16("With one anchor.", function(t) {
  var annos = [new Anno('a', 'a1', 3, 3, {
    isAnchor: true
  })];
  var html = _render$1(TEXT$1, annos);
  t.equal(html, 'ABC<a></a>DEFGHI');
  t.end();
});

test$16("With one inline.", function(t) {
  var annos = [new Anno('i', 'i1', 3, 4)];
  var html = _render$1(TEXT$1, annos);
  t.equal(html, 'ABC<i>D</i>EFGHI');
  t.end();
});

test$16("One nested anno.", function(t) {
  var annos = [new Anno('b', 'b1', 3, 6), new Anno('i', 'i1', 4, 5)];
  var html = _render$1(TEXT$1, annos);
  t.equal(html, 'ABC<b>D<i>E</i>F</b>GHI');
  t.end();
});

test$16("Overlapping annos.", function(t) {
  var annos = [new Anno('b', 'b1', 3, 6), new Anno('i', 'i1', 4, 8)];
  var html = _render$1(TEXT$1, annos);
  t.equal(html, 'ABC<b>D<i>EF</i></b><i>GH</i>I');
  t.end();
});

test$16("Equal annos.", function(t) {
  var annos = [new Anno('b', 'b1', 3, 6), new Anno('i', 'i1', 3, 6)];
  var html = _render$1(TEXT$1, annos);
  t.equal(html, 'ABC<b><i>DEF</i></b>GHI');
  t.end();
});

test$16("Overlapping with fragmentation hint.", function(t) {
  var annos = [
    new Anno('b', 'b1', 3, 6),
    new Anno('a', 'link1', 4, 8, {
      fragmentation: Fragmenter.SHOULD_NOT_SPLIT
    })
  ];
  var html = _render$1(TEXT$1, annos);
  t.equal(html, 'ABC<b>D</b><a><b>EF</b>GH</a>I');
  t.end();
});

test$16("Anchors should rendered as early as possible.", function(t) {
  var annos = [
    new Anno('b', 'b1', 3, 6),
    new Anno('a', 'a1', 3, 3, {
      isAnchor: true
    })
  ];
  var html = _render$1(TEXT$1, annos);
  t.equal(html, 'ABC<a></a><b>DEF</b>GHI');
  t.end();
});


test$16("Two subsequent inline nodes.", function(t) {
  var annos = [
    new Anno('a', 'inline1', 3, 4, {
      isInline: true
    }),
    new Anno('b', 'inline2', 4, 5, {
      isInline: true
    })
  ];
  var html = _render$1(TEXT$1, annos);
  t.equal(html, 'ABC<a>D</a><b>E</b>FGHI');
  t.end();
});

test$16("Collapsed annotation.", function(t) {
  var annos = [
    new Anno('a', 'a1', 0, 0, {
    })
  ];
  var html = _render$1(TEXT$1, annos);
  t.equal(html, '<a></a>ABCDEFGHI');
  t.end();
});

test$16("Two collapsed annotations.", function(t) {
  var annos = [
    new Anno('a', 'a1', 0, 0, {
    }),
    new Anno('b', 'b2', 0, 0, {
    })
  ];
  var html = _render$1(TEXT$1, annos);
  t.equal(html, '<a></a><b></b>ABCDEFGHI');
  t.end();
});

test$16("Anchors should not fragment other annotations.", function(t) {
  var annos = [
    new Anno('a', 'a1', 3, 6),
    new Anno('b', 'b1', 4, 4, {
      isAnchor: true
    })
  ];
  var html = _render$1(TEXT$1, annos);
  t.equal(html, 'ABC<a>D<b></b>EF</a>GHI');
  t.end();
});

class Anno extends Annotation {

  constructor(tagName, id, startOffset, endOffset, opts) {
    super(null, {
      id: id,
      start: { path: [id, 'content'], offset: startOffset},
      end: { path: [id, 'content'], offset: endOffset }
    });

    opts = opts || {};
    this.tagName = tagName;

    this._isAnchor = false;
    this._isInline = false;

    if (opts.hasOwnProperty('fragmentation')) {
      this.fragmentationHint = opts.fragmentation;
    }

    if (opts.hasOwnProperty('isAnchor')) {
      this._isAnchor = opts.isAnchor;
      this.zeroWidth = true;
      this.offset = startOffset;
    }

    if (opts.hasOwnProperty('isInline')) {
      this._isInline = opts.isInline;
    }
  }

  // anchors are special annotations that have zero width
  isAnchor() {
    return this._isAnchor
  }

  // inline nodes are implementated as annotations bound to a single character
  // I.e. the always have a length of 1
  isInline() {
    return this._isInline
  }

}

function _render$1(text, annotations, opts) {
  opts = opts || {};
  var output = [];
  var fragmenter = new Fragmenter();
  fragmenter.onText = function(context, text) {
    output.push(text);
  };
  fragmenter.onEnter = function(fragment) {
    var node = fragment.node;
    if (opts.withId) {
      output.push('<' + node.tagName + ' id="' + node.id +'">');
    } else {
      output.push('<' + node.tagName + '>');
    }
  };
  fragmenter.onExit = function(fragment) {
    var node = fragment.node;
    output.push('</' + node.tagName + '>');
  };
  fragmenter.start(output, text, annotations);
  return output.join('')
}

const CONTENT = '0123456789';

htmlExporterTests();

if (platform.inBrowser) {
  htmlExporterTests('memory');
}

function htmlExporterTests(memory) {

  const test = substanceTest.module('HTMLExporter' + ( memory ? ' [memory]' : ''), {
    before: function(t) {
      t.elementFactory = memory ? MemoryDOMElement.createDocument('html') : DefaultDOMElement.createDocument('html');
    }
  });

  test("Exporting paragraph", function(t) {
    let { doc, exporter } = setup(t);
    let p1 = doc.create({ type: 'paragraph', id: 'p1', content: CONTENT });
    let el = exporter.convertNode(p1);
    let actual = el.outerHTML;
    let expected = '<p data-id="p1">' + CONTENT + '</p>';
    t.equal(actual, expected, 'Exported HTML should be correct');
    t.end();
  });

  test("Exporting paragraph with strong", function(t) {
    let { doc, exporter } = setup(t);
    let p1 = doc.create({ type: 'paragraph', id: 'p1', content: CONTENT });
    doc.create({
      type: 'strong',
      id: 's1',
      start: {
        path: ['p1', 'content'],
        offset: 4
      },
      end: {
        offset: 7
      }
    });
    let el = exporter.convertNode(p1);
    let actual = el.outerHTML;
    let expected = '<p data-id="p1">0123<strong data-id="s1">456</strong>789</p>';
    t.equal(actual, expected, 'Exported HTML should be correct');
    t.end();
  });

  test("Exporting h1", function(t) {
    let { doc, exporter } = setup(t);
    let h1 = doc.create({ type: 'heading', id: 'h1', level: 1, content: CONTENT });
    let el = exporter.convertNode(h1);
    let actual = el.outerHTML;
    let expected = '<h1 data-id="h1">' + CONTENT + '</h1>';
    t.equal(actual, expected, 'Exported HTML should be correct');
    t.end();
  });

  test("Exporting h2", function(t) {
    let { doc, exporter } = setup(t);
    let h2 = doc.create({ type: 'heading', id: 'h2', level: 2, content: CONTENT });
    let el = exporter.convertNode(h2);
    let actual = el.outerHTML;
    let expected = '<h2 data-id="h2">' + CONTENT + '</h2>';
    t.equal(actual, expected, 'Exported HTML should be correct');
    t.end();
  });

  test("Exporting simple document", function(t) {
    let { doc, exporter } = setup(t, simple);
    let el = exporter.exportDocument(doc);
    let actual = el.html();
    let expected = [
      '<p data-id="p1">' + CONTENT + '</p>',
      '<p data-id="p2">' + CONTENT + '</p>',
      '<p data-id="p3">' + CONTENT + '</p>',
      '<p data-id="p4">' + CONTENT + '</p>'
    ].join('');
    t.equal(actual, expected, 'Exported HTML should be correct');
    t.end();
  });

  test("Exporting a link", function(t) {
    let { doc, exporter } = setup(t);
    let p1 = doc.create({ type: 'paragraph', id: 'p1', content: CONTENT });
    doc.create({
      type: 'link',
      id: 'l1',
      start: {
        path: ['p1', 'content'],
        offset: 4,
      },
      end: {
        offset: 7
      },
      url: 'foo',
      title: 'bar'
    });
    let el = exporter.convertNode(p1);
    let childNodes = el.getChildNodes();
    t.equal(childNodes.length, 3, 'Exported paragraph should have 3 child nodes');
    t.equal(childNodes[0].textContent, "0123", '.. 1. should have correct text');
    t.equal(childNodes[1].textContent, "456", '.. 2. should have correct text');
    t.equal(childNodes[2].textContent, "789", '.. 3. should have correct text');
    let a = childNodes[1];
    t.equal(a.attr('data-id'), 'l1', '.. <a> should have data-id set');
    t.equal(a.attr('href'), 'foo', '.. and correct href attribute');
    t.equal(a.attr('title'), 'bar', '.. and correct title attribute');
    t.end();
  });

  test("Exporting an unordered list", function(t) {
    let { doc, exporter } = setup(t);
    let l1 = _l1$1(doc);
    let el = exporter.convertNode(l1);
    let childNodes = el.getChildNodes();
    t.equal(el.tagName, 'ul', 'Exported element should be a <ul>');
    t.equal(el.attr('data-id'), 'l1', '.. with correct id');
    t.equal(childNodes.length, 2, '.. and two child nodes');
    t.equal(childNodes[0].tagName, "li", '.. a <li>');
    t.equal(childNodes[0].textContent, 'Foo', ".. with content 'Foo'");
    t.equal(childNodes[1].tagName, "li", '.. and a <li>');
    t.equal(childNodes[1].textContent, 'Bar', ".. with content 'Bar'");
    t.end();
  });

  test("Exporting an ordered list", function(t) {
    let { doc, exporter } = setup(t);
    let ol = doc.create({
      type: 'list',
      id: 'ol1',
      ordered: true
    });
    let el = exporter.convertNode(ol);
    t.equal(el.tagName, 'ol', 'Exported element should be a <ol>');
    t.end();
  });

  test("Exporting a nested list", function(t) {
    let { doc, exporter } = setup(t);
    let l = _l2$1(doc);
    let el = exporter.convertNode(l);
    let items = el.findAll('li');
    let nestedList = el.find('ul');
    t.equal(items.length, 4, 'Exported should contain 4 list items');
    t.notNil(nestedList, '.. and a nested list');
    t.equal(nestedList.childNodes.length, 2, '.. which has 2 child nodes');
    t.end();
  });

  function setup(t, fixture) {
    let config = getTestConfig();
    let exporter = config.createExporter('html', {}, { elementFactory: t.elementFactory });
    let doc = createTestArticle(fixture);
    return { exporter, doc }
  }

}

function _li1(doc) {
  doc.create({
    type: 'list-item',
    id: 'li1',
    content: 'Foo',
    level: 1
  });
}

function _li2(doc) {
  doc.create({
    type: 'list-item',
    id: 'li2',
    content: 'Bar',
    level: 1
  });
}

function _l1$1(doc) {
  _li1(doc);
  _li2(doc);
  return doc.create({
    type: 'list',
    id: 'l1',
    items: ['li1', 'li2'],
    ordered: false
  })
}

function _li3(doc) {
  doc.create({
    type: 'list-item',
    id: 'li3',
    content: 'Bla',
    level: 2
  });
}

function _li4(doc) {
  doc.create({
    type: 'list-item',
    id: 'li4',
    content: 'Blupp',
    level: 2
  });
}

function _l2$1(doc) {
  _li1(doc);
  _li2(doc);
  _li3(doc);
  _li4(doc);
  return doc.create({
    type: 'list',
    id: 'l1',
    items: ['li1', 'li3', 'li4', 'li2'],
    ordered: false
  })
}

const CONTENT$1 = '0123456789';


HTMLImporterTests();

if (platform.inBrowser) {
  HTMLImporterTests('memory');
}

function HTMLImporterTests(memory) {

  const test = substanceTest.module('HTMLImporter' + ( memory ? ' [memory]' : ''), {
    before: () => {
      if (memory) platform.inBrowser = false;
    },
    after: () => {
      platform._reset();
    }
  });

  test("Importing paragraph", function(t) {
    let importer = _setupImporter({ 'stand-alone': true });
    let html = '<p data-id="p1">' + CONTENT$1 + '</p>';
    let el = DefaultDOMElement.parseSnippet(html, 'html');
    let node = importer.convertElement(el);
    checkValues(t, node, {
      id: "p1",
      type: "paragraph",
      content: CONTENT$1
    }, 'paragraph should have been imported correctly');
    t.end();
  });

  test("Importing paragraph with strong", function(t) {
    let importer = _setupImporter({ 'stand-alone': true });
    let html = '<p data-id="p1">0123<strong data-id="s1">456</strong>789</p>';
    let el = DefaultDOMElement.parseSnippet(html, 'html');
    let p1 = importer.convertElement(el);
    let doc = p1.getDocument();
    let s1 = doc.get('s1');
    t.equal(p1.content, CONTENT$1, 'paragraph should have correct content');
    t.equal(s1.getText(), '456', 'annotation should provide correct text');
    t.end();
  });

  test("Importing h1", function(t) {
    let importer = _setupImporter({ 'stand-alone': true });
    let html = '<h1 data-id="h1">' + CONTENT$1 + '</h1>';
    let el = DefaultDOMElement.parseSnippet(html, 'html');
    let node = importer.convertElement(el);
    checkValues(t, node, {
      id: "h1",
      type: "heading",
      level: 1,
      content: CONTENT$1
    }, 'heading should have been imported correctly');
    t.end();
  });

  test("Importing h2", function(t) {
    let importer = _setupImporter({ 'stand-alone': true });
    let html = '<h2 data-id="h2">' + CONTENT$1 + '</h2>';
    let el = DefaultDOMElement.parseSnippet(html, 'html');
    let node = importer.convertElement(el);
    checkValues(t, node, {
      id: "h2",
      type: "heading",
      level: 2,
      content: CONTENT$1
    }, 'heading should have been imported correctly');
    t.end();
  });

  test("Importing an unordered list", function(t) {
    let importer = _setupImporter({ 'stand-alone': true });
    let html = '<ul data-id="l1"><li>Foo</li><li>Bar</li></ul>';
    let el = DefaultDOMElement.parseSnippet(html, 'html');
    let node = importer.convertElement(el);
    t.equal(node.type, 'list', 'Imported node should be a list');
    t.equal(node.id, 'l1', 'id should be correct');
    t.equal(node.ordered, false, 'node should unordered');
    t.equal(node.items.length, 2, 'it should have 2 items');
    let li1 = node.getItemAt(0);
    let li2 = node.getItemAt(1);
    t.equal(li1.getText(), 'Foo', 'First item should have correct text');
    t.equal(li2.getText(), 'Bar', 'Second item should have correct text');
    t.end();
  });

  test("Importing an ordered list", function(t) {
    let importer = _setupImporter({ 'stand-alone': true });
    let html = '<ol data-id="l1"></ol>';
    let el = DefaultDOMElement.parseSnippet(html, 'html');
    let node = importer.convertElement(el);
    t.equal(node.type, 'list', 'Imported node should be a list');
    t.equal(node.id, 'l1', 'id should be correct');
    t.equal(node.ordered, true, 'node should ordered');
    t.end();
  });

  test("Importing a nested list", function(t) {
    let importer = _setupImporter({ 'stand-alone': true });
    let html = '<ul data-id="l1"><li>Foo</li><ul><li>Bla</li><li>Blupp</li></ul><li>Bar</li></ul>';
    let el = DefaultDOMElement.parseSnippet(html, 'html');
    let node = importer.convertElement(el);
    t.equal(node.items.length, 4, 'Imported node should have 4 items');
    let levels = node.getItems().map(item=>item.level);
    let content = node.getItems().map(item=>item.content);
    t.equal(String(levels), String([1,2,2,1]), 'Node levels should be correct');
    t.equal(String(content), String(['Foo','Bla','Blupp','Bar']), 'Items should have correct content');
    t.end();
  });

  test("Importing a nested list (bad style)", function(t) {
    let importer = _setupImporter({ 'stand-alone': true });
    let html = '<ul data-id="l1"><li>Foo<ul><li>Bla</li><li>Blupp</li></ul></li><li>Bar</li></ul>';
    let el = DefaultDOMElement.parseSnippet(html, 'html');
    let node = importer.convertElement(el);
    t.equal(node.items.length, 4, 'Imported node should have 4 items');
    let levels = node.getItems().map(item=>item.level);
    let content = node.getItems().map(item=>item.content);
    t.equal(String(levels), String([1,2,2,1]), 'Node levels should be correct');
    t.equal(String(content), String(['Foo','Bla','Blupp','Bar']), 'Items should have correct content');
    t.end();
  });

}

function _setupImporter(options = {}) {
  let config = getTestConfig();
  let importer = config.createImporter('html', {}, options);
  return importer
}

const test$17 = substanceTest.module('InlineNode');

// NOTE: surface ids are a bit ids of Surfaces and IsolatedNodes are not very intuitive
// body/in1 means parent surface of in1 is body -- while in1 is actually on p1.content, which is not a surface on its own

test$17("InlineNodes should be not selected when selection is null", function(t) {
  let { editorSession, editor } = setupEditor(t, paragraphsWithInlineNodes);
  let nodes = editor.findAll('.sc-inline-node');
  editorSession.setSelection(null);
  nodes.forEach(function(node){
    t.ok(node.isNotSelected(), "node '"+node.getId()+"' should not be selected.");
  });
  t.end();
});

test$17("InlineNodes should be not selected when selection is somewhere else", function(t) {
  let { editorSession, editor } = setupEditor(t, paragraphsWithInlineNodes);
  let nodes = editor.findAll('.sc-inline-node');
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 5,
    surfaceId: 'body'
  });
  nodes.forEach(function(node){
    t.ok(node.isNotSelected(), "node '"+node.getId()+"' should not be selected.");
  });
  t.end();
});

test$17("InlineNode should be 'selected' with when the inline node is selected", function(t) {
  let { editorSession, editor } = setupEditor(t, paragraphsWithInlineNodes);
  let nodes = editor.findAll('.sc-inline-node');
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 2,
    endOffset: 3,
    surfaceId: 'body'
  });
  var expected = {
    'body/in1': 'selected',
    'body/in2': undefined,
  };
  nodes.forEach(function(node){
    var id = node.getId();
    t.equal(node.getMode(), expected[id], "node '" + id + "' should be " + (expected[id] || 'not selected') );
  });
  t.end();
});

test$17("InlineNode should be 'co-selected' when selection is spanning an inline node", function(t) {
  let { editorSession, editor } = setupEditor(t, paragraphsWithInlineNodes);
  let nodes = editor.findAll('.sc-inline-node');
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 1,
    endOffset: 4,
    surfaceId: 'body'
  });
  var expected = {
    'body/in1': 'co-selected',
    'body/in2': undefined,
  };
  nodes.forEach(function(node){
    var id = node.getId();
    t.equal(node.getMode(), expected[id], "node '" + id + "' should be " + (expected[id] || 'not selected') );
  });
  t.end();
});

test$17("InlineNode should be 'focused' when having the selection", function(t) {
  let { editorSession, editor } = setupEditor(t, paragraphsWithInlineNodes);
  let nodes = editor.findAll('.sc-inline-node');
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 1,
    endOffset: 4,
    surfaceId: 'body/in1/sn1.title'
  });
  var expected = {
    'body/in1': 'focused',
    'body/in2': undefined,
  };
  nodes.forEach(function(node){
    var id = node.getId();
    t.equal(node.getMode(), expected[id], "node '" + id + "' should be " + (expected[id] || 'not selected') );
  });
  t.end();
});

// Similar to the previous but with another inline node being focused
test$17("InlineNode should be 'focused' when having the selection (II)", function(t) {
  let { editorSession, editor } = setupEditor(t, paragraphsWithInlineNodes);
  let nodes = editor.findAll('.sc-inline-node');
  editorSession.setSelection({
    type: 'property',
    path: ['c_p', 'content'],
    startOffset: 1,
    endOffset: 4,
    surfaceId: 'body/in2/c'
  });
  var expected = {
    'body/in1': undefined,
    'body/in2': 'focused',
  };
  nodes.forEach(function(node){
    var id = node.getId();
    t.equal(node.getMode(), expected[id], "node '" + id + "' should be " + (expected[id] || 'not selected') );
  });
  t.end();
});

test$17("InlineNode should be 'co-focused' when a nested inline node has the selection", function(t) {
  let { editorSession, editor } = setupEditor(t, nestedInlineNode);
  let nodes = editor.findAll('.sc-inline-node');
  editorSession.setSelection({
    type: 'property',
    path: ['sn2', 'title'],
    startOffset: 2,
    surfaceId: 'body/in1/sn1.title/in2/sn2.title'
  });
  var expected = {
    'body/in1': 'co-focused',
    'body/in1/sn1.title/in2': 'focused',
  };
  nodes.forEach(function(node){
    var id = node.getId();
    t.equal(node.getMode(), expected[id], "node '" + id + "' should be " + (expected[id] || 'not selected') );
  });
  t.end();
});

// fixtures

function paragraphsWithInlineNodes(doc) {
  var tx = new EditingInterface(doc);
  twoParagraphs(tx);
  var sn1 = tx.create({
    type: "structured-node",
    id: "sn1",
    title: "ABCDEFG"
  });
  tx.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 2
  });
  tx.insertInlineNode({
    type: 'inline-wrapper',
    id: 'in1',
    wrappedNode: sn1.id
  });
  var c = tx.create({
    type: "container",
    id: "c"
  });
  var c_p = tx.create({
    type: 'paragraph',
    id: "c_p",
    content: "ABCDEFG"
  });
  c.show(c_p);
  tx.setSelection({
    type: 'property',
    path: ['p2', 'content'],
    startOffset: 2
  });
  tx.insertInlineNode({
    type: 'inline-wrapper',
    id: 'in2',
    wrappedNode: c.id
  });
}

// co-focusing an inline node is only possible, if the inline node itself contains
// content with an inline node (or isolated node)
function nestedInlineNode(doc) {
  let tx = new EditingInterface(doc);
  twoParagraphs(tx);
  let sn1 = tx.create({
    type: "structured-node",
    id: "sn1",
    title: "ABCDEFG"
  });
  tx.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 2
  });
  tx.insertInlineNode({
    type: 'inline-wrapper',
    id: 'in1',
    wrappedNode: sn1.id
  });
  let sn2 = doc.create({
    type: "structured-node",
    id: "sn2",
    title: "ABCDEFG"
  });
  tx.setSelection({
    type: 'property',
    path: ['sn1', 'title'],
    startOffset: 4
  });
  tx.insertInlineNode({
    type: 'inline-wrapper',
    id: 'in2',
    wrappedNode: sn2.id
  });
}

function nestedContainers(tx) {
  var body = tx.get('body');

  var p1 = tx.create({
    type: 'paragraph',
    id: 'p1',
    content: 'ABCDEF'
  });
  body.show(p1);

  var c1 = tx.create({
    type: 'container',
    id: 'c1',
    nodes: []
  });
  var c1p1 = tx.create({
    type: 'paragraph',
    id: 'c1_p1',
    content: 'Lorem ipsum dolor sit amet, consectetur adipiscing elit.'
  });
  c1.show(c1p1);

  var c2 = tx.create({
    type: 'container',
    id: 'c2',
    nodes: []
  });
  var c2p1 = tx.create({
    type: 'paragraph',
    id: 'c2_p1',
    content: 'Nunc turpis erat, sodales id aliquet eget, rutrum vel libero.'
  });
  c2.show(c2p1);
  c1.show(c2);

  var c1p2 = tx.create({
    type: 'paragraph',
    id: 'c1_p2',
    content: 'Donec dapibus vel leo sit amet auctor. Curabitur at diam urna.'
  });
  c1.show(c1p2);
  body.show(c1);

  var p2 = tx.create({
    type: 'paragraph',
    id: 'p2',
    content: 'ABCDEF'
  });
  body.show(p2);
}

const test$18 = substanceTest.module('IsolatedNode');

test$18("IsolatedNodes should be 'not-selected' when selection is null", function(t) {
  let { editorSession, editor } = setupEditor(t, nestedContainers);
  let isolatedNodes = editor.findAll('.sc-isolated-node');
  editorSession.setSelection(null);
  isolatedNodes.forEach(n => _notSelected(t, n));
  t.end();
});

test$18("IsolatedNodes should be 'not-selected' when selection is somewhere else", function(t) {
  let { editorSession, editor } = setupEditor(t, nestedContainers);
  let isolatedNodes = editor.findAll('.sc-isolated-node');
  editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 0,
    surfaceId: 'body'
  });
  isolatedNodes.forEach(function(isolated){
    t.ok(isolated.isNotSelected(), "isolated node '"+isolated.getId()+"' should not be selected.");
  });
  t.end();
});

test$18("IsolatedNode should be 'selected' with node selection", function(t) {
  let { editorSession, editor } = setupEditor(t, nestedContainers);
  let isolatedNodes = editor.findAll('.sc-isolated-node');
  editorSession.setSelection({
    type: 'node',
    nodeId: 'c1',
    containerId: 'body',
    surfaceId: 'body'
  });
  let expected = {
    'body/c1': 'selected',
    'body/c1/c2': undefined,
  };
  isolatedNodes.forEach(n => _modeOk(t, n, expected));
  t.end();
});

test$18("IsolatedNode should be 'co-selected' with spanning container selection", function(t) {
  let { editorSession, editor } = setupEditor(t, nestedContainers);
  let isolatedNodes = editor.findAll('.sc-isolated-node');
  editorSession.setSelection({
    type: 'container',
    containerId: 'body',
    startPath: ['p1', 'content'],
    startOffset: 1,
    endPath: ['p2', 'content'],
    endOffset: 2,
    surfaceId: 'body'
  });
  let expected = {
    'body/c1': 'co-selected',
    // Note: 'co-selection' does not propagate down
    // it is a state related to the parent container
    'body/c1/c2': undefined,
  };
  isolatedNodes.forEach(n => _modeOk(t, n, expected));
  t.end();
});

test$18("IsolatedNode should be 'focused' when having the selection", function(t) {
  let { editorSession, editor } = setupEditor(t, nestedContainers);
  let isolatedNodes = editor.findAll('.sc-isolated-node');
  editorSession.setSelection({
    type: 'property',
    path: ['c1_p1', 'content'],
    startOffset: 0,
    surfaceId: 'body/c1'
  });
  let expected = {
    'body/c1': 'focused',
    'body/c1/c2': undefined,
  };
  isolatedNodes.forEach(n => _modeOk(t, n, expected));
  t.end();
});

test$18("IsolatedNode should be 'co-focused' when child is having the selection", function(t) {
  let { editorSession, editor } = setupEditor(t, nestedContainers);
  let isolatedNodes = editor.findAll('.sc-isolated-node');
  editorSession.setSelection({
    type: 'property',
    path: ['c2_p1', 'content'],
    startOffset: 0,
    surfaceId: 'body/c1/c2'
  });
  let expected = {
    'body/c1': 'co-focused',
    'body/c1/c2': 'focused',
  };
  isolatedNodes.forEach(n => _modeOk(t, n, expected));
  t.end();
});

test$18("Issue #696: IsolatedNode should detect 'co-focused' robustly in presence of surface ids with same prefix", function(t) {
  // as experienced in #696 it happened that co-focused state was infered just
  // by using startsWith on the surface path. This was leading to wrong
  // co-focused states when e.g. two isolated nodes `body/entity` and `body/entity-1`
  // exist. I.e. one surfaceId was a prefix of another one.
  let { editorSession, editor } = setupEditor(t, _twoStructuredNodes);
  let isolatedNodes = editor.findAll('.sc-isolated-node');
  editorSession.setSelection({
    type: 'property',
    path: ['sn2', 'title'],
    startOffset: 0,
    surfaceId: 'body/sn2/sn2.title'
  });
  let expected = {
    'body/sn': null,
    'body/sn2': 'focused',
  };
  isolatedNodes.forEach(n => _modeOk(t, n, expected));
  t.end();
});

function _notSelected(t, isolated) {
  t.ok(isolated.isNotSelected(), "isolated node '"+isolated.getId()+"' should not be selected.");
}

function _modeOk(t, isolated, expected) {
  let id = isolated.getId();
  t.looseEqual(isolated.getMode(), expected[id], "mode of isolated node '" + id + "' should be correct");
}

function _twoStructuredNodes(doc) {
  let body = doc.get('body');
  body.show(doc.create({
    type: 'structured-node',
    id: 'sn',
    title: 'Foo'
  }));
  body.show(doc.create({
    type: 'structured-node',
    id: 'sn2',
    title: 'Bar'
  }));
}

const test$19 = substanceTest.module('ObjectOperation');

function checkObjectOperationTransform(test, a, b, input, expected) {
  let ops = ObjectOperation.transform(a, b);
  let output = ops[1].apply(a.apply(cloneDeep(input)));
  test.deepEqual(output, expected, `(b' o a)('${JSON.stringify(input)}') == '${JSON.stringify(expected)}' with a=${a.toString()}, b'=${ops[1].toString()}`);
  output = ops[0].apply(b.apply(cloneDeep(input)));
  test.deepEqual(output, expected, `(a' o b)('${JSON.stringify(input)}') == '${JSON.stringify(expected)}' with b=${b.toString()}, a'=${ops[0].toString()}`);
}

test$19("Creating values.", (t) => {
  let path = ["a"];
  let val = { bla: "blupp" };
  let expected = {a: { bla: "blupp" } };
  let op = ObjectOperation.Create(path, val);
  let obj = {};
  op.apply(obj);
  t.deepEqual(obj, expected, 'Should create value.');
  t.end();
});

test$19("Creating nested values.", (t) => {
  let path = ["a", "b"];
  let val = { bla: "blupp" };
  let expected = {a: { b: { bla: "blupp" } } };
  let op = ObjectOperation.Create(path, val);
  let obj = {"a": {}};
  op.apply(obj);
  t.deepEqual(obj, expected, 'Should create nested value.');
  t.end();
});

test$19("Deleting values.", (t) => {
  let path = ["a"];
  let val = "bla";
  let op = ObjectOperation.Delete(path, val);
  let expected = {};
  let obj = {"a": "bla"};
  op.apply(obj);
  t.deepEqual(obj, expected, 'Should delete value.');
  t.end();
});

test$19("Deleting nested values.", (t) => {
  let path = ["a", "b"];
  let val = "bla";
  let op = ObjectOperation.Delete(path, val);
  let expected = { a: {} };
  let obj = { a: { b: "bla"} };
  op.apply(obj);
  t.deepEqual(obj, expected, 'Should delete nested value.');
  t.end();
});

test$19("Updating a text property.", (t) => {
  let obj = {a: "bla"};
  let path = ["a"];
  let op1 = ObjectOperation.Update(path, TextOperation.Delete(2, 'a'));
  let op2 = ObjectOperation.Update(path, TextOperation.Insert(2, 'upp'));
  let expected = {a: "blupp"};
  op1.apply(obj);
  op2.apply(obj);
  t.deepEqual(obj, expected);
  t.end();
});

test$19("Updating an array property.", (t) => {
  let obj = {a: [1,2,3,4,5]};
  let path = ["a"];
  let op1 = ObjectOperation.Update(path, ArrayOperation.Delete(2, 3));
  let op2 = ObjectOperation.Update(path, ArrayOperation.Insert(4, 6));
  let expected = {a: [1,2,4,5,6]};
  op1.apply(obj);
  op2.apply(obj);
  t.deepEqual(obj, expected);
  t.end();
});

test$19("Creating an update operation with invalid diff.", (t) => {
  t.throws(function() {
    ObjectOperation.Update(['test'], "foo");
  }, "Should throw.");
  t.end();
});

test$19("Creating a top-level property using id.", (t) => {
  let obj = {};
  let id = "foo";
  let op = ObjectOperation.Create(id, "bar");
  let expected = { "foo": "bar" };
  op.apply(obj);
  t.deepEqual(obj, expected);
  t.end();
});

test$19("Deleting a top-level property using id.", (t) => {
  let obj = { foo: "bar" };
  let id = "foo";
  let op = ObjectOperation.Delete(id, "bar");
  let expected = {};
  op.apply(obj);
  t.deepEqual(obj, expected);
  t.end();
});

test$19("Apply operation on PathObject.", (t) => {
  let myObj = new PathObject();
  let op = ObjectOperation.Set(['foo', 'bar'], null, 'bla');
  op.apply(myObj);
  t.equal(myObj.get(['foo', 'bar']), 'bla');
  t.end();
});

test$19("Creating operation with invalid data.", (t) => {
  t.throws(function() {
    new ObjectOperation();
  }, "Should throw when data is undefined.");
  t.throws(function() {
    new ObjectOperation({});
  }, "Should throw when type is missing.");
  t.throws(function() {
    new ObjectOperation({ type: 'foo', path: ['test'] });
  }, "Should throw when type is invalid.");
  t.throws(function() {
    new ObjectOperation({
      type: ObjectOperation.CREATE
    });
  }, "Should throw when path is missing.");
  t.throws(function() {
    new ObjectOperation({
      type: ObjectOperation.CREATE,
      path: ["test"],
    });
  }, "Should throw when created value is missing.");
  t.throws(function() {
    new ObjectOperation({
      type: ObjectOperation.UPDATE,
      path: ["test"]
    });
  }, "Should throw when update diff is missing.");
  t.throws(function() {
    new ObjectOperation({
      type: ObjectOperation.UPDATE,
      path: ["test"],
      diff: "foo"
    });
  }, "Should throw when update diff is invalid.");
  // we have relaxed that, so that it is possible to 'delete' a property by setting it to undefined
  // t.throws(function() {
  //   new ObjectOperation({
  //     type: ObjectOperation.SET,
  //     path: ["test"],
  //   })
  // }, "Should throw when value is missing.")
  // t.throws(function() {
  //   new ObjectOperation({
  //     type: ObjectOperation.SET,
  //     path: ["test"],
  //     val: 1
  //   })
  // }, "Should throw when old value is missing.")
  t.end();
});

test$19("Inverse of NOP is NOP", (t) => {
  let op = new ObjectOperation({type: ObjectOperation.NOP});
  let inverse = op.invert();
  t.ok(inverse.isNOP());
  t.end();
});

test$19("Inverse of Create is Delete and vice versa", (t) => {
  let op = ObjectOperation.Create('test', 'foo');
  let inverse = op.invert();
  t.ok(inverse.isDelete());
  t.equal(inverse.getValue(), 'foo');
  inverse = inverse.invert();
  t.ok(inverse.isCreate());
  t.equal(inverse.getValue(), 'foo');
  t.end();
});

test$19("Inverse of Update is Update", (t) => {
  let op = ObjectOperation.Update(['test', 'foo'], ArrayOperation.Insert(1, 'a'));
  let inverse = op.invert();
  t.ok(inverse.isUpdate());
  t.deepEqual(inverse.getPath(), ['test', 'foo']);
  t.ok(inverse.diff instanceof ArrayOperation);
  t.deepEqual(inverse.diff.toJSON(), op.diff.invert().toJSON());
  // same with a text operation as diff
  op = ObjectOperation.Update(['test', 'foo'], TextOperation.Insert(1, 'a'));
  inverse = op.invert();
  t.ok(inverse.diff instanceof TextOperation);
  t.deepEqual(inverse.diff.toJSON(), op.diff.invert().toJSON());
  t.end();
});

test$19("Inverse of Set is Set", (t) => {
  let op = ObjectOperation.Set(['test', 'foo'], 'foo', 'bar');
  let inverse = op.invert();
  t.ok(inverse.isSet());
  t.deepEqual(inverse.getPath(), ['test', 'foo']);
  t.equal(inverse.getValue(), 'foo');
  t.equal(inverse.original, 'bar');
  t.end();
});

test$19("Transformation: everything easy when not the same property", (t) => {
  let path1 = ["a"];
  let path2 = ["b"];
  let val1 = "bla";
  let val2 = "blupp";
  let a = ObjectOperation.Create(path1, val1);
  let b = ObjectOperation.Create(path2, val2);
  let ops = ObjectOperation.transform(a, b);
  t.deepEqual(ops[0].toJSON(), a.toJSON());
  t.deepEqual(ops[1].toJSON(), b.toJSON());
  t.end();
});

test$19("Transformation: everything easy when NOP involved", (t) => {
  let path1 = ["a"];
  let val1 = "bla";
  let a = ObjectOperation.Create(path1, val1);
  let b = new ObjectOperation({type: ObjectOperation.NOP});
  let ops = ObjectOperation.transform(a, b);
  t.deepEqual(ops[0].toJSON(), a.toJSON());
  t.deepEqual(ops[1].toJSON(), b.toJSON());
  t.end();
});

test$19("Transformation: creating the same value (unresolvable conflict)", (t) => {
  let path = ["a"];
  let val1 = "bla";
  let val2 = "blupp";
  let a = ObjectOperation.Create(path, val1);
  let b = ObjectOperation.Create(path, val2);
  t.throws(function() {
    ObjectOperation.transform(a, b);
  });
  t.end();
});

test$19("Transformation: creating and updating the same value (unresolvable conflict)", (t) => {
  let path = ["a"];
  let val1 = "bla";
  let a = ObjectOperation.Create(path, val1);
  let b = ObjectOperation.Update(path, TextOperation.Insert(1, 'b'));
  t.throws(function() {
    ObjectOperation.transform(a, b);
  });
  t.end();
});

test$19("Transformation: creating and setting the same value (unresolvable conflict)", (t) => {
  let path = ["a"];
  let val1 = "bla";
  let val2 = "blupp";
  let a = ObjectOperation.Create(path, val1);
  let b = ObjectOperation.Set(path, val1, val2);
  t.throws(function() {
    ObjectOperation.transform(a, b);
  });
  t.end();
});

test$19("Transformation: creating and deleting the same value (unresolvable conflict)", (t) => {
  let path = ["a"];
  let val1 = "bla";
  let a = ObjectOperation.Create(path, val1);
  let b = ObjectOperation.Delete(path, val1);
  t.throws(function() {
    ObjectOperation.transform(a, b);
  });
  t.end();
});

test$19("Transformation: deleting the same value", (t) => {
  let path = ["a"];
  let val = "bla";
  let input = {"a": val};
  let expected = {};
  let a = ObjectOperation.Delete(path, val);
  let b = ObjectOperation.Delete(path, val);
  checkObjectOperationTransform(t, a, b, input, expected);
  checkObjectOperationTransform(t, b, a, input, expected);
  t.end();
});

test$19("Transformation: deleting and updating the same value", (t) => {
  let path = ["a"];
  let a = ObjectOperation.Delete(path, "bla");
  let b = ObjectOperation.Update(path, TextOperation.Insert(3, "pp"));
  let input = {a : "bla"};
  let expected1 = {a: "blapp"};
  let expected2 = {};
  checkObjectOperationTransform(t, a, b, input, expected1);
  checkObjectOperationTransform(t, b, a, input, expected2);
  // same with an array operation
  a = ObjectOperation.Delete(path, [1,2,3]);
  b = ObjectOperation.Update(path, ArrayOperation.Insert(3, 4));
  input = {a : [1,2,3]};
  expected1 = {a: [1,2,3,4]};
  expected2 = {};
  checkObjectOperationTransform(t, a, b, input, expected1);
  checkObjectOperationTransform(t, b, a, input, expected2);
  t.end();
});

test$19("Transformation: deleting and setting the same value", (t) => {
  let path = ["a"];
  let a = ObjectOperation.Delete(path, "bla");
  let b = ObjectOperation.Set(path, "bla", "blupp");
  let input = {a : "bla"};
  let expected1 = {a: "blupp"};
  let expected2 = {};
  checkObjectOperationTransform(t, a, b, input, expected1);
  checkObjectOperationTransform(t, b, a, input, expected2);
  t.end();
});

test$19("Transformation: updating the same value", (t) => {
  let path = ["a"];
  let a = ObjectOperation.Update(path, TextOperation.Insert(3, "pp"));
  let b = ObjectOperation.Update(path, TextOperation.Insert(3, "ff"));
  let input = {a : "bla"};
  let expected1 = {a: "blappff"};
  let expected2 = {a: "blaffpp"};
  checkObjectOperationTransform(t, a, b, input, expected1);
  checkObjectOperationTransform(t, b, a, input, expected2);
  // same with array updates
  a = ObjectOperation.Update(path, ArrayOperation.Insert(2, 3));
  b = ObjectOperation.Update(path, ArrayOperation.Insert(2, 4));
  input = {a : [1,2,5]};
  expected1 = {a: [1,2,3,4,5]};
  expected2 = {a: [1,2,4,3,5]};
  checkObjectOperationTransform(t, a, b, input, expected1);
  checkObjectOperationTransform(t, b, a, input, expected2);
  t.end();
});

test$19("Transformation: updating and setting the same value (unresolvable conflict)", (t) => {
  let path = ["a"];
  let a = ObjectOperation.Update(path, TextOperation.Insert(3, "ff"));
  let b = ObjectOperation.Set(path, "bla", "blupp");
  t.throws(function() {
    ObjectOperation.transform(a, b);
  });
  t.end();
});

test$19("Transformation: setting the same value", (t) => {
  let path = ["a"];
  let a = ObjectOperation.Set(path, "bla", "blapp");
  let b = ObjectOperation.Set(path, "bla", "blupp");
  let input = {a : "bla"};
  let expected1 = {a: "blupp"};
  let expected2 = {a: "blapp"};
  checkObjectOperationTransform(t, a, b, input, expected1);
  checkObjectOperationTransform(t, b, a, input, expected2);
  t.end();
});

test$19("ObjectOperation with the same path are conflicts.", (t) => {
  let a = ObjectOperation.Set(['bla', 'blupp'], null, 'foo');
  let b = ObjectOperation.Set(['bla', 'blupp'], null, 'bar');
  t.ok(a.hasConflict(b) && b.hasConflict(a));
  t.end();
});

test$19("NOPs have never conflicts.", (t) => {
  let a = ObjectOperation.Set(['bla', 'blupp'], null, 'foo');
  let b = new ObjectOperation({type: ObjectOperation.NOP});
  t.ok(!a.hasConflict(b) && !b.hasConflict(a));
  t.end();
});

test$19("With option 'no-conflict' conflicting operations can not be transformed.", (t) => {
  let a = ObjectOperation.Create('bla', 'blupp');
  let b = ObjectOperation.Create('bla', 'blupp');
  t.throws(function() {
    ObjectOperation.transform(a, b, { "no-conflict": true });
  }, 'Transforming conflicting ops should throw when option "no-conflict" is enabled.');
  t.end();
});

test$19("JSON deserialisation", (t) => {
  let path = ["test", "foo"];

  let op = ObjectOperation.fromJSON({
    type: ObjectOperation.SET,
    path: path,
    val: "bla",
    original: "blupp"
  });
  t.equal(op.getType(), ObjectOperation.SET);
  t.equal(op.getValue(), "bla");

  op = ObjectOperation.fromJSON({
    type: ObjectOperation.UPDATE,
    path: path,
    diff: ArrayOperation.Insert(1, 'a'),
    propertyType: 'array'
  });
  t.ok(op.diff instanceof ArrayOperation);
  t.deepEqual(op.getPath(), path);
  op = ObjectOperation.fromJSON({
    type: ObjectOperation.UPDATE,
    path: path,
    diff: TextOperation.Insert(1, 'a'),
    propertyType: 'string'
  });
  t.ok(op.diff instanceof TextOperation);
  t.deepEqual(op.getPath(), path);

  t.throws(function() {
    ObjectOperation.fromJSON({
      type: ObjectOperation.UPDATE,
      path: path,
      diff: "bla",
      propertyType: 'foo'
    });
  }, "Should throw for unknown update diff type.");
  t.end();
});

test$19("JSON serialisation", (t) => {
  let data = ObjectOperation.Create('test', "bla").toJSON();
  t.equal(data.type, ObjectOperation.CREATE);
  t.equal(data.val, "bla");
  data = ObjectOperation.Delete('test', "bla").toJSON();
  t.equal(data.type, ObjectOperation.DELETE);
  t.equal(data.val, "bla");
  data = ObjectOperation.Update(['test', 'foo'], ArrayOperation.Insert(1, 'a')).toJSON();
  t.equal(data.type, ObjectOperation.UPDATE);
  t.equal(data.propertyType, 'array');
  data = ObjectOperation.Update(['test', 'foo'], TextOperation.Insert(1, 'a')).toJSON();
  t.equal(data.propertyType, 'string');
  data = ObjectOperation.Set(['test', 'foo'], 'foo', 'bar').toJSON();
  t.equal(data.type, ObjectOperation.SET);
  t.equal(data.val, 'bar');
  t.equal(data.original, 'foo');
  t.end();
});

const test$20 = substanceTest.module('OperationSerializer');

test$20('Serializing create', function(t) {
  var p = { id: 'p1', type: 'paragraph', content: 'foo' };
  var op = ObjectOperation.Create(['p1'], p);
  var io = new OperationSerializer();
  var data = io.serialize(op);
  t.deepEqual(data, ['c', p.id, p]);
  t.end();
});

test$20('Deserializing create', function(t) {
  var p = { id: 'p1', type: 'paragraph', content: 'foo' };
  var io = new OperationSerializer();
  var data = ['c', 'p1', p];
  var op = io.deserialize(data);
  t.ok(op.type, 'create');
  t.deepEqual(op.getValue(), p);
  t.end();
});

test$20('Serializing delete', function(t) {
  var p = { id: 'p1', type: 'paragraph', content: 'foo' };
  var op = ObjectOperation.Delete(['p1'], p);
  var io = new OperationSerializer();
  var data = io.serialize(op);
  t.deepEqual(data, ['d', p.id, p]);
  t.end();
});

test$20('Deserializing delete', function(t) {
  var p = { id: 'p1', type: 'paragraph', content: 'foo' };
  var io = new OperationSerializer();
  var data = ['d', 'p1', p];
  var op = io.deserialize(data);
  t.ok(op.type, 'delete');
  t.deepEqual(op.getValue(), p);
  t.end();
});

test$20('Serializing set', function(t) {
  var op = ObjectOperation.Set(['p1', 'content'], 'foo', 'bar');
  var io = new OperationSerializer();
  var data = io.serialize(op);
  t.deepEqual(data, ['s', 'p1.content', 'bar', 'foo']);
  t.end();
});

test$20('Serializing set with null', function(t) {
  var op = ObjectOperation.Set(['p1', 'content'], null, null);
  var io = new OperationSerializer();
  var data = io.serialize(op);
  t.deepEqual(data, ['s', 'p1.content', null, null]);
  t.end();
});

test$20('Serializing set with undefined', function(t) {
  // ATTENTION: undefined is not a valid JSON value, so it gets replaced
  // by null
  var op = ObjectOperation.Set(['p1', 'content'], undefined, undefined);
  var io = new OperationSerializer();
  var data = io.serialize(op);
  t.deepEqual(data, ['s', 'p1.content', undefined, undefined]);
  t.end();
});

test$20('Deserializing set', function(t) {
  var io = new OperationSerializer();
  var data = ['s', 'p1.content', 'bar', 'foo'];
  var op = io.deserialize(data);
  t.equal(op.type, 'set');
  t.deepEqual(op.getPath(), ['p1', 'content']);
  t.deepEqual(op.getValue(), 'bar');
  t.deepEqual(op.getOldValue(), 'foo');
  t.end();
});


test$20('Serializing text insert', function(t) {
  var op = ObjectOperation.Update(['p1', 'content'], TextOperation.Insert(3, 'foo'));
  var io = new OperationSerializer();
  var data = io.serialize(op);
  t.deepEqual(data, ['u', 'p1.content', 't+', 3, 'foo']);
  t.end();
});

test$20('Deserializing text insert', function(t) {
  var io = new OperationSerializer();
  var data = ['u', 'p1.content', 't+', 3, 'foo'];
  var op = io.deserialize(data);
  t.ok(op.type, 'update');
  t.deepEqual(op.getPath(), ['p1', 'content']);
  var valueOp = op.getValueOp();
  t.ok(valueOp instanceof TextOperation, 'Value operation should be a TextOperation');
  t.equal(valueOp.type, 'insert');
  t.equal(valueOp.pos, 3);
  t.equal(valueOp.str, 'foo');
  t.end();
});

test$20('Serializing text delete', function(t) {
  var op = ObjectOperation.Update(['p1', 'content'], TextOperation.Delete(3, 'foo'));
  var io = new OperationSerializer();
  var data = io.serialize(op);
  t.deepEqual(data, ['u', 'p1.content', 't-', 3, 'foo']);
  t.end();
});

test$20('Deserializing text delete', function(t) {
  var io = new OperationSerializer();
  var data = ['u', 'p1.content', 't-', 3, 'foo'];
  var op = io.deserialize(data);
  t.ok(op.type, 'update');
  t.deepEqual(op.getPath(), ['p1', 'content']);
  var valueOp = op.getValueOp();
  t.ok(valueOp instanceof TextOperation, 'Value operation should be a TextOperation');
  t.equal(valueOp.type, 'delete');
  t.equal(valueOp.pos, 3);
  t.equal(valueOp.str, 'foo');
  t.end();
});

test$20('Serializing an array insert', function(t) {
  var op = ObjectOperation.Update(['test', 'numbers'], ArrayOperation.Insert(3, 1234));
  var io = new OperationSerializer();
  var data = io.serialize(op);
  t.deepEqual(data, ['u', 'test.numbers', 'a+', 3, 1234]);
  t.end();
});

test$20('Deserializing an array insert', function(t) {
  var io = new OperationSerializer();
  var data = ['u', 'test.numbers', 'a+', 3, 1234];
  var op = io.deserialize(data);
  t.ok(op.type, 'update');
  t.deepEqual(op.getPath(), ['test', 'numbers']);
  var valueOp = op.getValueOp();
  t.ok(valueOp instanceof ArrayOperation, 'Value operation should be a ArrayOperation');
  t.equal(valueOp.type, 'insert');
  t.equal(valueOp.pos, 3);
  t.equal(valueOp.val, 1234);
  t.end();
});

test$20('Serializing an array delete', function(t) {
  var op = ObjectOperation.Update(['test', 'numbers'], ArrayOperation.Delete(3, 1234));
  var io = new OperationSerializer();
  var data = io.serialize(op);
  t.deepEqual(data, ['u', 'test.numbers', 'a-', 3, 1234]);
  t.end();
});

test$20('Deserializing an array delete', function(t) {
  var io = new OperationSerializer();
  var data = ['u', 'test.numbers', 'a-', 3, 1234];
  var op = io.deserialize(data);
  t.ok(op.type, 'update');
  t.deepEqual(op.getPath(), ['test', 'numbers']);
  var valueOp = op.getValueOp();
  t.ok(valueOp instanceof ArrayOperation, 'Value operation should be a ArrayOperation');
  t.equal(valueOp.type, 'delete');
  t.equal(valueOp.pos, 3);
  t.equal(valueOp.val, 1234);
  t.end();
});

const test$21 = substanceTest.module('packages');

test$21('import all packages', (t) => {
  [
    BasePackage,
    BlockquotePackage,
    BodyScrollPanePackage,
    ButtonPackage,
    CodePackage,
    CodeblockPackage,
    ContextMenuPackage,
    DropzonesPackage,
    EmphasisPackage,
    FilePackage,
    GridPackage,
    GutterPackage,
    HeadingPackage,
    ImagePackage,
    InlineWrapperPackage,
    InputPackage,
    LayoutPackage,
    LinkPackage,
    ListPackage,
    ModalPackage,
    OverlayPackage,
    ParagraphPackage,
    PersistencePackage,
    ProseEditorPackage,
    QuoteMarksPackage,
    ScrollPanePackage,
    ScrollbarPackage,
    SpellCheckPackage,
    SplitPanePackage,
    StrongPackage,
    SubscriptPackage,
    SuperscriptPackage,
    SwitchTextTypePackage,
    TabbedPanePackage,
    TablePackage,
    TextInputPackage,
    ToolbarPackage
  ].forEach((p) => {
    t.doesNotThrow(()=>{
      let configurator = new Configurator();
      configurator.import(p);
    });
  });
  t.end();
});

const test$22 = substanceTest.module('paste');

test$22("Pasting plain text", function(t) {
  let { tx } = _fixture$3(simple);
  tx.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 3
  });
  tx.paste('XXX');
  let p1 = tx.get('p1');
  t.equal(p1.content, '012XXX3456789');
  t.end();
});

test$22("Pasting a single paragraph", function(t) {
  let { tx } = _fixture$3(simple);
  let snippet = tx.createSnippet();
  let container = snippet.getContainer();
  let p = snippet.create({
    type: 'paragraph',
    id: Document.TEXT_SNIPPET_ID,
    content: 'AABBCC'
  });
  container.show(p.id);
  tx.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 3,
    containerId: 'body'
  });
  tx.paste(snippet);
  let p1 = tx.get('p1');
  t.equal(p1.content, '012AABBCC3456789', 'Plain text should be inserted.');
  t.end();
});

test$22("Pasting annotated text", function(t) {
  let { tx } = _fixture$3(simple);
  tx.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 3,
    containerId: 'body'
  });
  let snippet = tx.createSnippet();
  let container = snippet.getContainer();
  let p = snippet.create({
    type: 'paragraph',
    id: Document.TEXT_SNIPPET_ID,
    content: 'AABBCC'
  });
  container.show(p.id);
  snippet.create({
    type: 'strong',
    id: 's1',
    start: {
      path: [p.id, 'content'],
      offset: 2,
    },
    end: {
      offset: 4
    }
  });
  tx.paste(snippet);
  let p1 = tx.get('p1');
  t.equal(p1.content, '012AABBCC3456789', 'Plain text should be inserted.');
  let s1 = tx.get('s1');
  t.deepEqual(s1.path, [p1.id, 'content'], 'Annotation is bound to the correct path.');
  t.deepEqual([s1.start.offset, s1.end.offset], [5, 7], 'Annotation has correct range.');
  t.end();
});

test$22("Pasting two paragraphs", function(t) {
  let { tx } = _fixture$3(simple);
  let snippet = tx.createSnippet();
  let container = snippet.getContainer();
  let test1 = snippet.create({
    type: 'paragraph',
    id: 'test1',
    content: 'AA'
  });
  container.show(test1.id);
  let test2 = snippet.create({
    type: 'paragraph',
    id: 'test2',
    content: 'BB'
  });
  container.show(test2.id);
  tx.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 3,
    containerId: 'body'
  });
  tx.paste(snippet);
  let body = tx.get('body');
  let p1 = tx.get('p1');
  t.equal(p1.content, '012AA', 'First part should be inserted into first paragraph.');
  t.equal(body.nodes[1], test2.id, 'Second part should go into a single paragraph.');
  t.equal(tx.get(body.nodes[2]).content, '3456789', 'Remaining part of first paragraph should be in a new paragraph.');
  t.equal(body.nodes[3], 'p2', 'After that should follow p2.');
  t.end();
});

function _fixture$3(seed) {
  let doc = createTestArticle(seed);
  let tx = new EditingInterface(doc);
  return { doc: doc, tx: tx }
}

const test$23 = substanceTest.module('PathEventProxy');

function docWithTestNodes(tx) {
  headersAndParagraphs(tx);
  tx.create({
    type: "test-node",
    id: "test",
    arrayVal: [1,2,3]
  });
}

test$23("Updating a property", function(t) {
  var doc = createTestArticle(docWithTestNodes);
  var callCount = 0;
  doc.getEventProxy('path').on(['test', 'arrayVal'], function() {
    callCount++;
  });
  doc.update(['test', 'arrayVal'], { type: 'insert', pos: 1, value: '1000' } );
  t.equal(callCount, 1, "Event proxy listener should have been called.");
  t.end();
});

test$23("Setting a property", function(t) {
  var doc = createTestArticle(docWithTestNodes);
  var callCount = 0;
  doc.getEventProxy('path').on(['test', 'arrayVal'], function() {
    callCount++;
  });
  doc.set(['test', 'arrayVal'], [1,1,1]);
  t.equal(callCount, 1, "Event proxy listener should have been called.");
  t.end();
});

test$23("Setting a property and deleting the node afterwards", function(t) {
  var doc = createTestArticle(docWithTestNodes);
  var editorSession = new EditorSession(doc, { configurator: new Configurator() });
  var callCount = 0;
  doc.getEventProxy('path').on(['test', 'arrayVal'], function() {
    callCount++;
  });
  editorSession.transaction(function(tx) {
    tx.set(['test', 'arrayVal'], [1,1,1]);
    tx.delete('test');
  });
  t.equal(callCount, 0, "Event proxy listener doesn't get called when node is deleted.");
  t.end();
});

/* eslint-disable no-invalid-this, indent */
const Simple$1 = TestComponent.Simple;
const internal = RenderingEngine._internal;

// runs RenderingEngine in regular mode
RenderingEngineTests();

// runs RenderingEngine in debug mode
RenderingEngineTests('debug');


function RenderingEngineTests(debug) {

  const test = substanceTest.module('RenderingEngine' + (debug ? ' [debug-mode]' : ''), {
    before: function() {
      substanceGlobals.DEBUG_RENDERING = Boolean(debug);
    }
  });

  // NOTE: this is a set of white-box tests for the internal implementation
  // of ui/RenderingEngine.

  // __isMapped__ is an internal state variable that is used
  // by RenderingEnging to know, that a component can be reused. I.e.,
  // the VirtualComponent has been mapped successfully to a Component

  test('Components without refs are not mapped', function(t) {
    var comp = TestComponent.create(function($$) {
      return $$('div').append($$(Simple$1))
    });
    var vc = _capture$1(comp);
    t.notOk(vc._isMapped(vc._content.children[0]), 'child element should not be mapped.');
    t.end();
  });

  test('A component with ref is mapped', function(t) {
    var comp = TestComponent.create(function($$) {
      return $$('div').append($$(Simple$1).ref('foo'))
    });
    var vc = _capture$1(comp);
    t.ok(vc._isMapped(vc._content), 'root element should be mapped');
    t.ok(vc._isMapped(vc._getRef('foo')), "'foo' should be mapped");
    t.end();
  });

  // TODO: there are more case related nesting of components
  // and injecting components via props

  // __isRelocated__ is an internal state variable that is used
  // when a reused component has been rendered with a different parent

  test('Detecting relocation when injecting a new parent element', function(t) {
    function _render($$) {
      var el = $$('div');
      var parent = el;
      if (this.props.extraLayer) {
        var middle = $$('div');
        el.append(middle);
        parent = middle;
      }
      parent.append($$(Simple$1).ref('foo'));
      return el
    }
    var comp = TestComponent.create(_render);
    var vc = _capture$1(comp);
    t.notOk(vc._isRelocated(vc._getRef('foo')), "'foo' is not relocated the first time.");

    comp = TestComponent.create(_render, { extraLayer: false });
    _setProps(comp, { extraLayer: true });
    vc = _capture$1(comp);
    t.ok(vc._isRelocated(vc._getRef('foo')), "'foo' is relocated the second time.");

    comp = TestComponent.create(_render, { extraLayer: true });
    _setProps(comp, { extraLayer: false });
    vc = _capture$1(comp);
    t.ok(vc._isRelocated(vc._getRef('foo')), "'foo' is relocated the third time.");

    t.end();
  });

  test('Detecting relocation when injecting components (TextProperty use-case)', function(t) {
    /*
      This simulates a situation found often when rendering a TextProperty.
      Say a text property contains an inline node.
      Regularly, the content looks roughly like this:
      ```
        AAAA<div class="inline-node">BBBB</div>CCCC
      ```
      When selected there is a wrapper around this
      ```
        AAAA<div class='selection'><div>BBBB</div></div>CCCC
      ```
      Adding and removing this selection element leads to a situation
      where the inline node needs to be attached to varying parent
      elements.
    */
    function _render($$) {
      var el = $$('div');
      var parent = el;
      el.append('AAAA');
      if (this.props.extraLayer) {
        var middle = $$(Simple$1).ref('selection');
        el.append(middle);
        parent = middle;
      }
      parent.append($$(Simple$1).ref('foo').append('BBBB'));
      el.append('CCCC');
      return el
    }
    var comp = TestComponent.create(_render);
    var vc = _capture$1(comp);
    t.notOk(vc._isRelocated(vc._getRef('foo')), "'foo' is not relocated the first time.");

    comp = TestComponent.create(_render, { extraLayer: false });
    _setProps(comp, { extraLayer: true });
    vc = _capture$1(comp);
    t.ok(vc._isRelocated(vc._getRef('foo')), "'foo' is relocated the second time.");

    comp = TestComponent.create(_render, { extraLayer: true });
    _setProps(comp, { extraLayer: false });
    vc = _capture$1(comp);
    t.ok(vc._isRelocated(vc._getRef('foo')), "'foo' is relocated the third time.");

    t.end();
  });

  // NOTE: Dunno if this is a good test. Obviously it depends too much on the performance of the device
  test('(Performance) Rendering a large number of components with ref', (t) => {
    const N = 100;
    const M = 52;
    const data = [];
    for (let i = 0; i < N; i++) {
      let row = [];
      for (let j = 0; j < M; j++) {
        row.push(`${i}-${j}`);
      }
      data.push(row);
    }
    class Matrix extends Component {
      render($$) {
        let el = $$('table').css({
          overflow: 'auto',
          height: '100px'
        });
        const data = this.props.data;
        for (let i = 0; i < data.length; i++) {
          let row = data[i];
          let rowEl = $$('tr');
          for (let j = 0; j < row.length; j++) {
            rowEl.append(
              $$('td')
                .ref(row[j])
                .text(row[j]));
          }
          el.append(rowEl);
        }
        return el
      }
    }
    let t0 = Date.now();
    let comp = Matrix.mount({data}, getMountPoint(t));
    let t1 = Date.now() - t0;
    t0 = Date.now();
    comp.setProps({data});
    let t2 = Date.now() - t0;
    t.ok(t1 < 1000, 'First rendering should be finished in reasonable time');
    t.ok(t2 < 1000, 'Second rendering should be finished in reasonable time');
    t.end();
  });
}

function _capture$1(comp) {
  var vc = internal._wrap(comp);
  var state = new RenderingEngine.State();
  internal._capture(state, vc, 'force');
  vc._state = state;
  vc._isMapped = function(o) { return state.isMapped(o); };
  vc._isRelocated = function(o) { return state.isRelocated(o); };
  vc._getRef = function(ref) { return _getRef(vc, ref); };
  return vc
}

function _getRef(vc, ref) {
  return vc._content._context.refs[ref] || {}
}

function _setProps(comp, props) {
  comp.props = props;
  Object.freeze(props);
}

/* eslint-disable no-invalid-this */
const test$24 = substanceTest.module('Surface');

// This test was added to cover issue #82
test$24.UI("Set the selection after creating annotation.", function(t) {
  window.getSelection().removeAllRanges();
  let { editorSession, surface } = setupEditor(t, _p1$1);
  let sel = editorSession.setSelection({
    type: 'property',
    path: ['p1', 'content'],
    startOffset: 0,
    endOffset: 5,
    containerId: 'body',
    surfaceId: 'body'
  });
  editorSession.transaction(function(tx) {
    tx.annotate({ type: "strong" });
  });
  let wsel = window.getSelection();
  let newSel = surface.domSelection.getSelection();
  t.equal(wsel.rangeCount, 1, "There should be a DOM selection.");
  t.ok(newSel.equals(sel), "New selection should be equal to initial selection.");
  t.end();
});

test$24.UI("Render a reverse selection.", function(t) {
  window.getSelection().removeAllRanges();
  let { editorSession } = setupEditor(t, _p1$1, _p2$1);
  editorSession.setSelection({
    type: 'container',
    startPath:['p1', 'content'],
    startOffset: 3,
    endPath: ['p2', 'content'],
    endOffset: 2,
    reverse: true,
    containerId: 'body',
    surfaceId: 'body'
  });
  var wsel = BrowserDOMElement.getWindowSelection();
  t.ok(BrowserDOMElement.isReverse(wsel.anchorNode, wsel.anchorOffset, wsel.focusNode, wsel.focusOffset));
  t.end();
});

const P1_TEXT$1 = 'abcdef';

function _p1$1(doc, body) {
  doc.create({
    type: 'paragraph',
    id: 'p1',
    content: P1_TEXT$1
  });
  body.show('p1');
}

const P2_TEXT$1 = 'ghijk';

function _p2$1(doc, body) {
  doc.create({
    type: 'paragraph',
    id: 'p2',
    content: P2_TEXT$1
  });
  body.show('p2');
}

const test$25 = substanceTest.module('TextOperation');

function checkTextTransform(test, a, b, input, expected) {
  let ops = TextOperation.transform(a, b);
  let output = ops[1].apply(a.apply(input));
  test.deepEqual(output, expected, `(b' o a)('${input}') == '${expected}' with a=${a.toString()}, b'=${ops[1].toString()}`);
  output = ops[0].apply(b.apply(input));
  test.deepEqual(output, expected, `(a' o b)('${input}') == '${expected}' with b=${b.toString()}, a'=${ops[0].toString()}`);
}

test$25("Insert string", (t) => {
  let input = "Lorem ipsum";
  let expected = "Lorem bla ipsum";
  let a = TextOperation.Insert(6, "bla ");
  t.equal(expected, a.apply(input));
  t.end();
});

test$25("Insert at last position", (t) => {
  let input = "Lorem ipsum";
  let expected = "Lorem ipsum.";
  let a = TextOperation.Insert(11, ".");
  t.equal(expected, a.apply(input));
  t.end();
});

test$25("Invalid arguments", (t) => {
  t.throws(function() {
    new TextOperation({});
  }, "Should throw for incomplete data.");
  t.throws(function() {
    new TextOperation({type: "foo", pos: 0, str: ""});
  }, "Should throw for invalid type.");
  t.throws(function() {
    TextOperation.Insert(-1, "");
  }, "Should throw for invalid position.");
  t.throws(function() {
    TextOperation.Insert(-1, null);
  }, "Should throw for invalid string.");
  t.end();
});

test$25("TextOperation has length", (t) => {
  let op = TextOperation.Delete(1, 'bla');
  t.equal(op.getLength(), 3, 'Length of ' + op.toString() + ' should be 3.');
  op = TextOperation.Insert(1, 'blupp');
  t.equal(op.getLength(), 5, 'Length of ' + op.toString() + ' should be 5.');
  t.end();
});

test$25("JSON serialisation", (t) => {
  let op = TextOperation.Delete(1, 'bla');
  let expected = {
    type: TextOperation.DELETE,
    pos: 1,
    str: 'bla'
  };
  t.deepEqual(op.toJSON(), expected, 'Serialized operation should ok.');
  t.end();
});

test$25("JSON deserialisation", (t) => {
  let data = {
    type: TextOperation.INSERT,
    pos: 1,
    str: "bla"
  };
  let op = TextOperation.fromJSON(data);
  t.ok(op.isInsert(), 'Deserialized operation should be an insert operation.');
  t.ok(op.pos === 1, 'Deserialized operation should have offset==1.');
  t.ok(op.str === "bla", 'Deserialized operation should have string=="bla".');
  t.end();
});

test$25("Empty TextOperations are NOPs", (t) => {
  let op = TextOperation.Insert(0, "");
  t.ok(op.isNOP(), 'Empty operations should be NOPs.');
  t.end();
});

test$25("Can't apply on a too short string", (t) => {
  let op = TextOperation.Insert(6, "bla");
  t.throws(function() {
    op.apply('bla');
  }, "Should throw if string is too short.");
  op = TextOperation.Delete(2, "bla");
  t.throws(function() {
    op.apply('bla');
  }, "Should throw if string is too short.");
  t.end();
});

test$25("Can be applied on custom String implementation", (t) => {
  let CustomString = function(str) {
    this.arr = str.split('');
    this.splice = function(pos, remove, insert) {
      this.arr.splice(pos, remove);
      if (insert) {
        this.arr = this.arr.slice(0, pos).concat(insert.split('')).concat(this.arr.slice(pos));
      }
    };
    this.toString = function() {
      return this.arr.join('')
    };
  };
  Object.defineProperty(CustomString.prototype, 'length', {
    get: function() {
      return this.arr.length
    }
  });
  let str = new CustomString('Lorem ipsum.');
  let op = TextOperation.Insert(6, "bla ");
  op.apply(str);
  t.equal(str.toString(), 'Lorem bla ipsum.', 'Insert operation should work on custom string.');
  str = new CustomString('Lorem bla ipsum.');
  op = TextOperation.Delete(6, "bla ");
  op.apply(str);
  t.equal(str.toString(), 'Lorem ipsum.', 'Delete operation should work on custom string.');
  t.end();
});

test$25("Inversion of Insert = Delete", (t) => {
  let op = TextOperation.Insert(6, "bla");
  let inverse = op.invert();
  t.ok(inverse.isDelete(), 'Inverted operation should be a delete op.');
  t.equal(inverse.pos, op.pos, 'Inverted operation should have the same offset.');
  t.equal(inverse.str, op.str, 'Inverted operation should have the same string data.');
  t.end();
});

test$25("Inversion of Delete = Insert", (t) => {
  let op = TextOperation.Delete(6, "bla");
  let inverse = op.invert();
  t.ok(inverse.isInsert(), 'Inverted operation should be a insert op.');
  t.equal(inverse.pos, op.pos, 'Inverted operation should have the same offset.');
  t.equal(inverse.str, op.str, 'Inverted operation should have the same string data.');
  t.end();
});

test$25("Transformation: a=Insert, b=Insert, a before b", (t) => {
  let input = "Lorem ipsum";
  let expected = "Lorem bla ipsum blupp";
  let a = TextOperation.Insert(6, "bla ");
  let b = TextOperation.Insert(11, " blupp");
  checkTextTransform(t, a, b, input, expected);
  checkTextTransform(t, b, a, input, expected);
  t.end();
});

test$25("Transformation: a=Insert, b=Insert, same position", (t) => {
  let input = "Lorem ipsum";
  let a = TextOperation.Insert(6, "bla ");
  let b = TextOperation.Insert(6, "blupp ");
  let expected = "Lorem bla blupp ipsum";
  // applying b first gives a different result
  let expected_2 = "Lorem blupp bla ipsum";
  checkTextTransform(t, a, b, input, expected);
  checkTextTransform(t, b, a, input, expected_2);
  t.end();
});

test$25("Transformation: a=Delete, b=Delete, a before b", (t) => {
  let input = "Lorem ipsum dolor sit amet";
  let expected = "Lorem dolor amet";
  let a = TextOperation.Delete(6, "ipsum ");
  let b = TextOperation.Delete(18, "sit ");
  checkTextTransform(t, a, b, input, expected);
  checkTextTransform(t, b, a, input, expected);
  t.end();
});

test$25("Transformation: a=Delete, b=Delete, overlapping", (t) => {
  let input = "Lorem ipsum dolor sit amet";
  let expected = "Lorem amet";
  let a = TextOperation.Delete(6, "ipsum dolor sit ");
  let b = TextOperation.Delete(12, "dolor ");
  checkTextTransform(t, a, b, input, expected);
  checkTextTransform(t, b, a, input, expected);
  t.end();
});

test$25("Transformation: a=Delete, b=Delete, same position", (t) => {
  let input = "Lorem ipsum dolor sit amet";
  let expected = "Lorem amet";
  let a = TextOperation.Delete(6, "ipsum dolor ");
  let b = TextOperation.Delete(6, "ipsum dolor sit ");
  checkTextTransform(t, a, b, input, expected);
  checkTextTransform(t, b, a, input, expected);
  t.end();
});

test$25("Transformation: a=Insert, b=Delete", (t) => {
  let input = "Lorem dolor sit amet";
  let expected = "Lorem ipsum dolor amet";
  let a = TextOperation.Insert(6, "ipsum ");
  let b = TextOperation.Delete(12, "sit ");
  checkTextTransform(t, a, b, input, expected);
  checkTextTransform(t, b, a, input, expected);
  t.end();
});

test$25("Transformation: a=Insert, b=Delete, a after b", (t) => {
  let input = "Lorem ipsum dolor amet";
  let expected = "Lorem dolor sit amet";
  let a = TextOperation.Insert(18, "sit ");
  let b = TextOperation.Delete(6, "ipsum ");
  checkTextTransform(t, a, b, input, expected);
  checkTextTransform(t, b, a, input, expected);
  t.end();
});

test$25("Transformation: a=Insert, b=Delete, overlap", (t) => {
  let input = "Lorem dolor sit amet";
  let expected = "Lorem amet";
  let a = TextOperation.Insert(12, "ipsum ");
  let b = TextOperation.Delete(6, "dolor sit ");
  checkTextTransform(t, a, b, input, expected);
  checkTextTransform(t, b, a, input, expected);
  t.end();
});

test$25("Transformations can be done inplace (optimzation for internal use)", (t) => {
  let a = TextOperation.Insert(6, "bla ");
  let b = TextOperation.Insert(6, "blupp ");
  let ops = TextOperation.transform(a, b, {inplace: true});
  t.ok(a.pos === ops[0].pos && b.pos === ops[1].pos, "Transformation should be done inplace.");
  t.end();
});

// Note: In the case of TextOperations conflicts are soft, i.e., there is a defined result
// in such cases. However in certain situations it makes sense to detect such cases, e.g. to notify
// the user to review the result.

test$25("Conflict: Insert at the same position", (t) => {
  let a = TextOperation.Insert(6, "bla");
  let b = TextOperation.Insert(6, "blupp");
  t.ok(a.hasConflict(b), 'Two inserts are considered a conflict if they are at the same position.');
  t.end();
});

test$25("Conflict: Delete with overlapping range", (t) => {
  let a = TextOperation.Delete(4, "bla");
  let b = TextOperation.Delete(6, "blupp");
  t.ok(a.hasConflict(b) && b.hasConflict(a), 'Two deletes are considered a conflict if they overlap.');
  t.end();
});

test$25("Conflict: Delete and Insert with overlapping range", (t) => {
  let a = TextOperation.Insert(4, "bla");
  let b = TextOperation.Delete(2, "blupp");
  t.ok(a.hasConflict(b) && b.hasConflict(a), 'Inserts and Deletes are considered a conflict if they overlap.');
  t.end();
});

test$25("With option 'no-conflict' conflicting operations can not be transformed.", (t) => {
  let a = TextOperation.Insert(4, "bla");
  let b = TextOperation.Delete(2, "blupp");
  t.throws(function() {
    TextOperation.transform(a, b, { "no-conflict": true });
  }, 'Transforming conflicting ops should throw when option "no-conflict" is enabled.');
  t.end();
});

const test$26 = substanceTest.module('TextPropertyComponent');

test$26("Get coordinate of empty property", function(t) {
  var doc = createTestArticle(simple);
  doc.create({
    type: 'paragraph',
    id: 'empty',
    content: ''
  });
  var comp = TextPropertyComponent.mount({
    doc: doc,
    path: ['empty', 'content']
  }, getMountPoint(t));

  var coor = comp.getDOMCoordinate(0);

  t.notNil(coor, 'Coordinate should be not null.');
  t.equal(coor.container, comp.el.getNativeElement(), 'element should be property element');
  t.equal(coor.offset, 0, 'offset should be 0');
  t.end();
});

const test$27 = substanceTest.module('TreeIndex');

test$27("Setting and getting values from a TreeIndex", function(t) {
  var adapter = new TreeIndex();
  adapter.set(['a'], 1);
  adapter.set(['c', 'b'], 2);
  t.equal(adapter.get('a'), 1, 'Top level value should be retrieved correctly.');
  t.equal(adapter.get(['c', 'b']), 2, 'Second level value should be retrieved correctly.');
  t.end();
});

test$27("Getting with invalid arguments", function(t) {
  var adapter = new TreeIndex();
  adapter.set(['a'], 1);
  adapter.set(['c', 'b'], 2);
  t.isNil(adapter.get('e'), 'Should return no value for unknown id');
  t.isNil(adapter.get(['c', 'd']), 'Should return no value for unknown path');
  t.isNil(adapter.get(null), 'Should return no value for null');
  t.isNil(adapter.get(), 'Should return no value for no path');
  t.isNil(adapter.get([]), 'Should return no value for empty path');
  t.isNil(adapter.get({}), 'Should return no value for object');
  t.isNil(adapter.get(1), 'Should return no value for a number');
  t.end();
});

// test("Wrapping an existing object", function(t) {
//   // Note: this way you can create an adapter for plain objects
//   var mine = {}
//   var adapter = TreeIndex.wrap(mine)
//   adapter.set(['a', 'b'], 2)
//   t.equal(adapter.get(['a', 'b']), 2, 'Value should be retrievable.')
//   t.equal(mine.a.b, 2, 'Value should have been written inplace.')
//   t.deepEqual(adapter.get('a'), mine.a, 'Get on first level should return plain content.')
// })

test$27("Getting values recursively", function(t) {
  var adapter = new TreeIndex();
  adapter.set(['a', 'b'], 'foo');
  adapter.set(['a', 'c', 'd'], 'bar');
  t.deepEqual(adapter.getAll('a'), { b: 'foo', d: 'bar'}, 'Values should be collected together into an object');
  t.end();
});

test$27("Arrays: basic usage", function(t) {
  var adapter = new TreeIndex.Arrays();
  t.throws(function() {
    adapter.set('a', []);
  }, 'TreeIndex.set is not allowed for array type');
  adapter.add('a', 1);
  t.deepEqual(adapter.get('a'), [1], 'Get should return an array.');
  adapter.add(['a', 'b'], 2);
  t.deepEqual(adapter.get('a'), [1], 'Content should still be the same after adding a nested value.');
  t.deepEqual(adapter.getAll('a'), [1, 2], 'Collect all values into an array.');
  adapter.add(['a', 'b'], 3);
  adapter.remove(['a', 'b'], 2);
  t.deepEqual(adapter.get(['a', 'b']), [3], 'Only one value should be left after removal.');
  adapter.delete(['a','b']);
  t.isNil(adapter.get(['a', 'b']), 'Value should now be deleted.');
  t.end();
});

const test$28 = substanceTest.module('utils');

test$28('flattenOften should flatten multiple rounds', (t) => {
  let arr = [1, [2, 3, [4, 5]]];
  let result = flattenOften(arr, 2);
  t.deepEqual([1, 2, 3, 4, 5], result, 'array should be flattened.');
  t.end();
});

test$28('flattenOften should stop at max', (t) => {
  let arr = [1, [2, 3, [4, 5]]];
  let result = flattenOften(arr, 1);
  t.deepEqual([1, 2, 3, [4, 5]], result, 'array should be flattened only once.');
  t.end();
});

test$28('isEqual on primitives', (t) => {
  t.ok(isEqual(1,1), '1 === 1');
  t.ok(isEqual("foo", "foo"), '"foo" === "foo"');
  t.notOk(isEqual("foo", "bar"), '"foo" !== "bar"');
  t.notOk(isEqual(null, undefined), 'null !== undefined');
  t.end();
});

test$28('isEqual on arrays', (t) => {
  t.ok(isEqual([1, [2, 3, [4, 5]]], [1, [2, 3, [4, 5]]]), 'should do a deep array comparison');
  t.notOk(isEqual([1, [2, 3, [4, 5]]], [1, [2, 3, [4, 5, 6]]]), 'should detect inequality on deeper level');
  t.end();
});

test$28('isEqual on objects', (t) => {
  t.ok(isEqual({ foo: 'foo', bar: { bla: 'bla', blupp: {} }, baz: [1, 2, 3] }, { foo: 'foo', bar: { bla: 'bla', blupp: {} }, baz: [1, 2, 3] }), 'should do a deep object comparison');
  t.notOk(isEqual({ foo: 'foo', bar: { bla: 'bla', blupp: {} }, baz: [1, 2, 3] }, { foo: 'foo', bar: { bla: 'bla', blupp: { a: 'a'} }, baz: [1, 2, 3] }), 'should detect inequality on deeper level');
  t.end();
});

const CONTENT$2 = '0123456789';

XMLExporterTests();

if (platform.inBrowser) {
  XMLExporterTests('memory');
}

function XMLExporterTests(memory) {

  const test = substanceTest.module('XMLExporter' + ( memory ? ' [memory]' : ''), {
    before: function(t) {
      t.elementFactory = memory ? MemoryDOMElement.createDocument('xml') : DefaultDOMElement.createDocument('xml');
    }
  });

  test("Exporting paragraph", function(t) {
    let { doc, exporter } = setup(t);
    let p1 = doc.create({ type: 'paragraph', id: 'p1', content: CONTENT$2 });
    let el = exporter.convertNode(p1);
    let actual = el.serialize();
    let expected = '<p id="p1">' + CONTENT$2 + '</p>';
    t.equal(actual, expected);
    t.end();
  });

  test("Exporting paragraph with strong", function(t) {
    let { doc, exporter } = setup(t);
    let p1 = doc.create({ type: 'paragraph', id: 'p1', content: CONTENT$2 });
    doc.create({
      type: 'strong',
      id: 's1',
      start: {
        path: ['p1', 'content'],
        offset: 4
      },
      end: {
        offset: 7
      }
    });
    let el = exporter.convertNode(p1);
    let actual = el.serialize();
    let expected = '<p id="p1">0123<strong id="s1">456</strong>789</p>';
    t.equal(actual, expected);
    t.end();
  });

  test("Exporting h1", function(t) {
    let { doc, exporter } = setup(t);
    let h1 = doc.create({ type: 'heading', id: 'h1', level: 1, content: CONTENT$2 });
    let el = exporter.convertNode(h1);
    let actual = el.serialize();
    let expected = '<h1 id="h1">' + CONTENT$2 + '</h1>';
    t.equal(actual, expected);
    t.end();
  });

  test("Exporting h2", function(t) {
    let { doc, exporter } = setup(t);
    let h2 = doc.create({ type: 'heading', id: 'h2', level: 2, content: CONTENT$2 });
    let el = exporter.convertNode(h2);
    let actual = el.serialize();
    let expected = '<h2 id="h2">' + CONTENT$2 + '</h2>';
    t.equal(actual, expected);
    t.end();
  });

  test("Exporting simple document", function(t) {
    let { doc, exporter } = setup(t, simple);
    let rootEl = exporter.exportDocument(doc);
    let actual = rootEl.serialize();
    let expected = [
      '<article>',
      '<p id="p1">' + CONTENT$2 + '</p>',
      '<p id="p2">' + CONTENT$2 + '</p>',
      '<p id="p3">' + CONTENT$2 + '</p>',
      '<p id="p4">' + CONTENT$2 + '</p>',
      '</article>'
    ].join('');
    t.equal(expected, actual);
    t.end();
  });

  test("Exporting meta", function(t) {
    let { doc, exporter } = setup(t);
    let meta = doc.create({
      type: "meta",
      id: "meta",
      title: 'Untitled'
    });
    let el = exporter.convertNode(meta);
    let actual = el.serialize();
    let expected = '<meta id="meta"><title>Untitled</title></meta>';
    t.equal(actual, expected);
    t.ok(true);
    t.end();
  });

  // FIXME: broken since introduction of file nodes
  // test("Exporting image", function(t) {
  //   let data = { type: 'image', id: 'img1', 'src': 'img1.png', 'previewSrc': 'img1preview.png' }
  //   let img = doc.create(data)
  //   let el = exporter.convertNode(img)
  //   t.equal(el.tagName.toLowerCase(), 'image')
  //   t.equal(el.id, 'img1')
  //   t.equal(el.getAttribute('src'), 'img1.png')
  //   t.end()
  // })

  function setup(t, fixture) {
    let config = getTestConfig();
    let exporter = config.createExporter('xml', {}, { elementFactory: t.elementFactory });
    let doc = createTestArticle(fixture);
    return { exporter, doc }
  }
}

const test$29 = substanceTest.module('XMLImporter');

const CONTENT$3 = '0123456789';

test$29("Importing paragraph", function(t) {
  let importer = _setupImporter$1({ 'stand-alone': true });
  let xml = '<p id="p1">' + CONTENT$3 + '</p>';
  let el = DefaultDOMElement.parseSnippet(xml, 'xml');
  let node = importer.convertElement(el);
  checkValues(t, node, {
    id: "p1",
    type: "paragraph",
    content: CONTENT$3
  });
  t.end();
});

test$29("Importing paragraph with strong", function(t) {
  let importer = _setupImporter$1({ 'stand-alone': true });
  let xml = '<p id="p1">0123<strong id="s1">456</strong>789</p>';
  let el = DefaultDOMElement.parseSnippet(xml, 'xml');
  let p1 = importer.convertElement(el);
  let doc = p1.getDocument();
  let s1 = doc.get('s1');
  t.equal(CONTENT$3, p1.content);
  t.equal('456', s1.getText());
  t.end();
});

test$29("Importing h1", function(t) {
  let importer = _setupImporter$1({ 'stand-alone': true });
  let xml = '<h1 id="h1">' + CONTENT$3 + '</h1>';
  let el = DefaultDOMElement.parseSnippet(xml, 'xml');
  let node = importer.convertElement(el);
  checkValues(t, node, {
    id: "h1",
    type: "heading",
    level: 1,
    content: CONTENT$3
  });
  t.end();
});

test$29("Importing h2", function(t) {
  let importer = _setupImporter$1({ 'stand-alone': true });
  let xml = '<h2 id="h2">' + CONTENT$3 + '</h2>';
  let el = DefaultDOMElement.parseSnippet(xml, 'xml');
  let node = importer.convertElement(el);
  checkValues(t, node, {
    id: "h2",
    type: "heading",
    level: 2,
    content: CONTENT$3
  });
  t.end();
});

test$29("Importing meta", function(t) {
  let importer = _setupImporter$1({ 'stand-alone': true });
  let xml = '<meta><title>' + CONTENT$3 + '</title></meta>';
  let el = DefaultDOMElement.parseSnippet(xml, 'xml');
  let node = importer.convertElement(el);
  checkValues(t, node, {
    id: 'meta',
    type: 'meta',
    title: CONTENT$3
  });
  t.end();
});

// FIXME: broken since introduction of file nodes
test$29("Importing image", function(t) {
  let importer = _setupImporter$1({ 'stand-alone': true });
  let xml = '<img id="img1" src="someimage.png"></img>';
  let el = DefaultDOMElement.parseSnippet(xml, 'xml');
  let node = importer.convertElement(el);
  t.equal(node.type, 'image', 'Created node should be of type "image"');
  t.equal(node.id, 'img1', '.. with correct id');
  let file = node.getImageFile();
  t.notNil(file, 'A file node should have been created');
  t.equal(file.url, 'someimage.png', '.. containing the correct URL');
  t.end();
});

function _setupImporter$1(options = {}) {
  let config = getTestConfig();
  let importer = config.createImporter('xml', {}, options);
  return importer
}

//# sourceMappingURL=./tests.cjs.js.map